/**
 * @file benchmark/bench_bcsr2.c
 * @author Serban Georgescu
 * @since 2009
 * 
 * Test for BCSR format
 * Yusa codes in winter, 2009.
 */

//#undef _GLIBCXX_ATOMIC_BUILTINS

#define LOGGER_NAME "benchmark.bench_bcsr"

/* Standard includes */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
//#include <mkl.h>

/* CUKr includes */
//#include "../mat_vec/sp/matrix_sp.h" 
//#include "../blas/sp/blas2_sp.h" 
#include "../cukr_all.h" 
//#include "../mat_vec/sp/bcsr/bcsr_matrix_sp_gen.h" // Poisson generator here

/* Blas includes */
#include "../blas/impl/generic/generic_blas.h"
//#include "../blas/impl/mkl/mkl_blas.h"
//#include "../blas/impl/gpu/gpu_blas.h" 

/* Access the global database*/
extern CukrDB cukr_db;


// ------------------------------------------------------------------- //

int main(int argc, char* argv[]) 
{

	/* MKL Option */
	//mkl_set_num_threads(1);
	//OMP_NUM_THREADS = 1;
	
	/* Timer */
	clock_t t11, t12, t21, t22, t31, t32, t41, t42, t51, t52, t61, t62;

	/*
	 * Test parameters
	 */
	int nn, n;   // Matrix order
	int r, c;   // Block size
	CUKR_DOUBLE alpha = 1, beta = 0; // y=aAx+by
	int loop; // SPMV loop times
	int i; // iterator
	
	if(argc == 4){
		nn = atoi(argv[1]);
		r  = atoi(argv[2]);
		c  = atoi(argv[2]);
		loop = atoi(argv[3]);
	}else{
		printf("Input matrix order: ");
		scanf("%d", &nn);
		printf("Input block size: ");
		scanf("%d", &r);
		c = r;
		printf("Input loop times: ");
		scanf("%d", &loop);
	}
	n = nn*nn;

	/* print env. */
	printf("------------------\n");
	printf("n = %d, r = c = %d\n", n, r);
	printf("alpha = %.lf, beta = %.lf\n", alpha, beta);
	printf("SPMV %d times loop\n", loop);
	printf("in Double Precision!\n");
	printf("------------------\n");

	/*
	 * Init CUKr 
	 */
	CukrInit(0, NULL, CUKR_DEFAULT);					// Init
	CukrDBSetS(CUKR_DB_DOUBLE_PRECISION_BLAS, "GPUBLAS");		//GPU
	//CukrDBSetS(CUKR_DB_MATRIX_FORMAT, "BCSR");			// Set format to BCSR
	CukrDBSetS(CUKR_DB_MATRIX_FORMAT, "CSR");			// Set format to CSR
	CukrDBSetS(CUKR_DB_MATRIX_FORMAT_VAR, "none");		// Set variation to none (since there is no BCSR variation)
	CukrStart(CUKR_DEFAULT);							// Start with default settings
	
	/*
	 * Create generic BLAS objects, all precisions
	 */

	/* For double */
//	CUKR_BLAS blas_generic_d;
//	CukrBlasCreate(&blas_generic_d);
//	blas_create_generic(blas_generic_d);
//	blas_generic_d->precision = CUKR_TYPE_DOUBLE;

	/* For qdouble */
//	CUKR_BLAS blas_generic_q;
//	CukrBlasCreate(&blas_generic_q);
//	blas_create_generic(blas_generic_q);
//	blas_generic_q->precision = CUKR_TYPE_QDOUBLE;

	/* For single */
//	CUKR_BLAS blas_generic_s;
//	CukrBlasCreate(&blas_generic_s);
//	blas_create_generic(blas_generic_s);
//	blas_generic_s->precision = CUKR_TYPE_SINGLE;
	
	/* For double, MKL */
//	CUKR_BLAS blas_mkl_s;
//	CukrBlasCreate(&blas_mkl_s);
//	blas_create_mkl(blas_mkl_s);
//	blas_mkl_s->precision = CUKR_TYPE_DOUBLE;

	/* For double, GPU */
	CUKR_BLAS blas_gpu_d;
	CukrBlasCreate(&blas_gpu_d);
	blas_create_gpu(blas_gpu_d);
	blas_gpu_d->precision = CUKR_TYPE_DOUBLE;

	/*
	 * Generate BCSR matrix
	 */
	CUKR_MATRIX_SP mat, mat2, mat3;
	
	/* This will generate your Poisson matrix.
	 * The matrix will be created from inside here
	 * (e.g. all memory allocated and so on)
	 * Write the code to ../mat_vec/sp/bcsr/bcsr_matrix_sp_gen.c
	 */
	printf("Generating Poisson Matrix...\n");
	CukrCsrMatspGenPoisson2D(&mat, nn, "A");
	CukrBcsrMatspGenPoisson2D(&mat2, nn, r, c, "A2");
	CukrBcsrMatspGenPoisson2D(&mat3, nn, r, c, "A3");
	
	/* format and variation set. (nennotame) */
	mat->format = CUKR_MATRIX_FORMAT_CSR;
	mat->variation = CUKR_MATRIX_FORMAT_VARIATION_NONE;
	mat2->format = CUKR_MATRIX_FORMAT_BCSR;
	mat2->variation = CUKR_MATRIX_FORMAT_VARIATION_NONE;
	mat3->format = CUKR_MATRIX_FORMAT_BCSR;
	mat3->variation = CUKR_MATRIX_FORMAT_VARIATION_NONE;
	
	printf("Finished!\n");

	/* Create test vectors */
	CUKR_VECTOR_SP x = NULL, y = NULL;
	CukrVecspCreate(&x,      n, CUKR_TYPE_DOUBLE, "x");
	CukrVecspCreate(&y,      n, CUKR_TYPE_DOUBLE, "y");

	/* Set the Flo */
	double flo = (double)(mat->nz) * 2.0 * (double)loop;

	/***********************************************************
	 * GPU CSR SPMV Testing
	 ***********************************************************/

	/* Fill vectors with values */
	CukrVecspSetFromVal(x, 1, CUKR_TYPE_DOUBLE);
	CukrVecspSetFromVal(y, 1, CUKR_TYPE_DOUBLE);
	
	/* calculate SPMV */
	t11 = clock();
	for(i=0; i<loop; i++) CukrSPMV_SP(blas_gpu_d, alpha, mat, x, beta, y);
	t12 = clock();
	
	/* print Flops */
	printf("------------------GPU CSR SPMV\n");
	printf("%lf GFlops\n", (1e-9)*flo/((double)(t12-t11)/(double)CLOCKS_PER_SEC));

	/***********************************************************
	 * GPU BCSR SPMV Testing
	 ***********************************************************/
	
	/* Fill vectors with values */
	CukrVecspSetFromVal(x, 1, CUKR_TYPE_DOUBLE);
	CukrVecspSetFromVal(y, 1, CUKR_TYPE_DOUBLE);
	
	/* calculate SPMV */
	t21 = clock();
	for(i=0; i<loop; i++) CukrSPMV_SP(blas_gpu_d, alpha, mat2, x, beta, y);
	t22 = clock();
	
	/* print Flops */
	printf("------------------GPU BCSR SPMV\n");
	printf("%lf GFlops\n", (1e-9)*flo/((double)(t22-t21)/(double)CLOCKS_PER_SEC));

	/***********************************************************
	 * GPU BCSR4 SPMV Testing
	 ***********************************************************/

	/* Convert BCSR to BCSR4 */
	CukrBcsrMatspConvert(mat2, CUKR_MATRIX_FORMAT_BCSR, CUKR_MATRIX_FORMAT_VARIATION_BCSR4);

	/* Fill vectors with values */
	CukrVecspSetFromVal(x, 1, CUKR_TYPE_DOUBLE);
	CukrVecspSetFromVal(y, 1, CUKR_TYPE_DOUBLE);
	
	/* calculate SPMV */
	t31 = clock();
	for(i=0; i<loop; i++) CukrSPMV_SP(blas_gpu_d, alpha, mat2, x, beta, y);
	t32 = clock();
	
	/* print Flops */
	printf("------------------GPU BCSR4 SPMV\n");
	printf("%lf GFlops\n", (1e-9)*flo/((double)(t32-t31)/(double)CLOCKS_PER_SEC));

	/***********************************************************
	 * GPU HYB SPMV Testing
	 ***********************************************************/
	
	/* Convert CSR to HYB */
	CukrCsrMatspConvert(mat, CUKR_MATRIX_FORMAT_HYB, CUKR_MATRIX_FORMAT_VARIATION_NONE);
	
	/* Fill vectors with values */
	CukrVecspSetFromVal(x, 1, CUKR_TYPE_DOUBLE);
	CukrVecspSetFromVal(y, 1, CUKR_TYPE_DOUBLE);
	
	/* calculate SPMV */
	t41 = clock();
	for(i=0; i<loop; i++) CukrSPMV_SP(blas_gpu_d, alpha, mat, x, beta, y);
	t42 = clock();
	
	/* print Flops */
	printf("------------------GPU HYB SPMV\n");
	printf("%lf GFlops\n", (1e-9)*flo/((double)(t42-t41)/(double)CLOCKS_PER_SEC));

	/***********************************************************
	 * GPU BHYB SPMV Testing
	 ***********************************************************/
	
	/* Convert BCSR to BHYB */
	CukrBcsrMatspConvert(mat3, CUKR_MATRIX_FORMAT_BHYB, CUKR_MATRIX_FORMAT_VARIATION_NONE);
	
	/* Fill vectors with values */
	CukrVecspSetFromVal(x, 1, CUKR_TYPE_DOUBLE);
	CukrVecspSetFromVal(y, 1, CUKR_TYPE_DOUBLE);
	
	/* calculate SPMV */
	t51 = clock();
	for(i=0; i<loop; i++) CukrSPMV_SP(blas_gpu_d, alpha, mat3, x, beta, y);
	t52 = clock();
	
	/* print Flops */
	printf("------------------GPU BHYB SPMV\n");
	printf("%lf GFlops\n", (1e-9)*flo/((double)(t52-t51)/(double)CLOCKS_PER_SEC));

	/***********************************************************
	 * GPU BHYB4 SPMV Testing
	 ***********************************************************/
	
	/* Convert BHYB to BHYB4 */
	CukrBhybMatspConvert(mat3, CUKR_MATRIX_FORMAT_BHYB, CUKR_MATRIX_FORMAT_VARIATION_BHYB4);
	
	/* Fill vectors with values */
	CukrVecspSetFromVal(x, 1, CUKR_TYPE_DOUBLE);
	CukrVecspSetFromVal(y, 1, CUKR_TYPE_DOUBLE);
	
	/* calculate SPMV */
	t61 = clock();
	for(i=0; i<loop; i++) CukrSPMV_SP(blas_gpu_d, alpha, mat3, x, beta, y);
	t62 = clock();
	
	/* print Flops */
	printf("------------------GPU BHYB4 SPMV\n");
	printf("%lf GFlops\n", (1e-9)*flo/((double)(t62-t61)/(double)CLOCKS_PER_SEC));


	/***********************************************************
	 * Performance ratio
	 ***********************************************************/

	/* print speedup */
	printf("------------------Performance ratio\n");
	printf("BCSR  : CSR = %lf\n", (double)(t12-t11)/(double)(t22-t21));
	printf("BCSR4 : CSR = %lf\n", (double)(t12-t11)/(double)(t32-t31));
	printf("BHYB  : HYB = %lf\n", (double)(t42-t41)/(double)(t52-t51));
	printf("BHYB4 : HYB = %lf\n", (double)(t42-t41)/(double)(t62-t61));


	/*
	 * Print the matrix and vector to see if all is correct
	 * Do this for various block sizes, until you are sure there are no bugs
	 * Maybe a smaller n might be better.
	 *
	 */
	//CukrMatPrint(mat);
	//CukrVecspPrint(y);

	/* Destroy */
	CukrMatspDestroy(&mat);
	CukrMatspDestroy(&mat2);
	CukrVecspDestroy(&x);
	CukrVecspDestroy(&y);

	/* 
	 * Shutdown 
	 */
	CukrFinalize(CUKR_DEFAULT); 


	return 0;	
	
	
}	

