#! /usr/bin/env python

##################################################
##
##       bench_blas1_blax1_plot plotting utility
##
##  Draws figurs and compiles a pdf report from 
##  the results outputted by the bench_blas1 
##  benchmark program.
##
##  Dependencies:
##      1. Gnuplot python module 
##         ---> http://gnuplot-py.sourceforge.net/
##         Requires numpy module to compile 
##         ---> http://numpy.scipy.org/ (or repository) 
##
##################################################


import sys
import os
from optparse import OptionParser
from Numeric import *
import Gnuplot, Gnuplot.funcutils

##########################
# Globals
##########################

columns = {'n':0, 'blas':1, 'precision':2, 'performance':3,  'bandwidth':4, 'error':5}

REPORT_SUBFIX         = "_report"
PERF_ALL_SUBFIX       = "_all_perf"
BAND_ALL_SUBFIX       = "_all_band"
ERR_ALL_SUBFIX        = "_all_err"
PERF_TAIL_ALL_SUBFIX  = "_all_perf-tail"
OUTPUT_MODE           = "Report"
OUTPUT_COLOR          = 1

##########################
# Functions
##########################

def set_output_mode(g, output):
    
    # General options
    g('set size square')
    g('set pointsize 1')
    
    # Options for presentation mode
    if (OUTPUT_MODE != "Report"):
        g('set terminal gif giant transparent size 480x480')
        g('set output "%s.gif"' % output)
        g('set pointsize 1.25')
    else:
        g('set terminal postscript eps size 3.5,3.5 small')
        g('set output "%s.eps"' % output)

# Get operation specific stuff (e.g. report title, 
# operation name as it will be printed in the report)
def get_operation_specifics(operation):
    formula = ""
    if (operation == "DOT"):
        formula = "d=\sum_{i=1}^{n}{x_iy_i}"
    elif (operation == "AXPY"):
        formula = "y_i=a \cdot x_i + y_i, i=\overline{1,n}"
    elif (operation == "AYPX"):
        formula = "y_i=a \cdot y_i + x_i, i=\overline{1,n}"
    elif (operation == "SCAL"):
        formula = "x_i=a \cdot x_i, i=\overline{1,n}"
    elif (operation == "PWPR"):
        formula = "z_i=x_i \cdot y_i, i=\overline{1,n}"    
    else:
        print "Invalid operation: " + operation
        parser.print_help()
        sys.exit(1)
        
    return formula
        
# Read from the input file all available BLASes
# and thir precisions    
def get_blas_and_precision(filename):
    # Open file
    f = open(input_file, "r")
    
    # Skip header
    f.readline()
    f.readline()
    
    
    # Create workspace list
    res_list = []
    
    # Read and add new blases and precisions to list
    for line in f:
        list = line.split(",")
        if (len(list) < 5):
            continue
        blas = list[columns['blas']]
        prec = list[columns['precision']]
        
        # See if current blas and precision are in results list
        found = 0
        for (b, p) in res_list:
            if (b == blas and p == prec):
                found = 1
        
        # If not found, add to list
        if (not found):
            res_list.append((blas, prec))
            
    f.close()
    return res_list

# Read the data and return it in a big list
def get_operation_and_data(filename,  blas_list):
    
    # Open file
    f = open(input_file, "r")
    
    # Read operation name
    op = f.readline().rstrip("\n\r")
    
    # Skip header
    header = f.readline()
    
    # Create a sublist for every blas combination
    res_list = []
    for (b, p) in blas_list:
        dot_data = {'blas':"",  'precision':"",  'n_list':[], 'perf_list':[], 'band_list':[], 'error_list':[]}
        dot_data['blas'] = b
        dot_data['precision'] = p
        res_list.append(dot_data)
    
    # Read and add new blases and precisions to list
    for line in f:
        list = line.split(",")
        if (len(list) < 5):
            continue
        n = list[columns['n']]
        blas = list[columns['blas']]
        prec = list[columns['precision']] 
        perf = list[columns['performance']] 
        band = list[columns['bandwidth']] 
        error = list[columns['error']] 
        
        # Find the blas it belongs to
        pos = 0
        for (b, p) in blas_list:
            if (b == blas and p == prec):
                break
            pos += 1
            
        # Error if not found
        if pos == len(blas_list):
            print "ERROR: BLAS not found"
        
        # Add data to the right blas
        res_list[pos]['perf_list'].append([int(n), float(perf)])
        res_list[pos]['band_list'].append([int(n), float(band)])
        res_list[pos]['error_list'].append([int(n), float(error)])
        
    f.close()
    return (op, res_list)
   
# Do the overall performance plot
def plot_performance(blas_list,  data_list,  operation):
    
    # Find max
    max = 0
    for i in range(len(blas_list)):
        for perf in data_list[i]['perf_list']:
            if (max < float(perf[1])):
                max = float(perf[1])
    
    # Init graph
    g = Gnuplot.Gnuplot(debug=1)
    g.xlabel('Vector size')
    g.ylabel('Performance (MFlop/s)')
    g('set logscale x')
    g('set key horiz left top')
    g('set yrange [1:%g]' % (max * 1.3))
    
    # Make the dataset list
    for i in range(len(blas_list)):
        # Extract name and precision
        (b, p) = blas_list[i]
        # Extract data
        data = data_list[i]
        # Add it to data sources if is different from QDOUBLE on CPU
        if (not (b == "Generic" and p == "(QDOUBLE)")):
            set_output_mode(g, operation + PERF_ALL_SUBFIX)
            d = Gnuplot.Data(data_list[i]['perf_list'],  title = b + " " + p)
            g.replot(d)
     
# Do the overall bandwidth plot
def plot_bandwidth(blas_list,  data_list,  operation):
    
    # Find max
    max = 0
    for i in range(len(blas_list)):
        for perf in data_list[i]['band_list']:
            if (max < float(perf[1])):
                max = float(perf[1])
    
    # Init graph
    g = Gnuplot.Gnuplot(debug=1)
    g.xlabel('Vector size')
    g.ylabel('Bandwidth (MB/s)')
    g('set logscale x')
    g('set key horiz left top')
    g('set yrange [1:%g]' % (max * 1.3))
    
    # Make the dataset list
    for i in range(len(blas_list)):
        
        # Extract name and precision
        (b, p) = blas_list[i]
        # Extract data
        data = data_list[i]
        # Add it to data sources if is different from QDOUBLE on CPU
        if (not (b == "Generic" and p == "(QDOUBLE)")):
            set_output_mode(g, operation + BAND_ALL_SUBFIX)
            d = Gnuplot.Data(data_list[i]['band_list'],  title = b + " " + p)
            g.replot(d)
    
    
# Do the overall error plot
def plot_error(blas_list,  data_list,  operation):
    
    # Init graph
    g = Gnuplot.Gnuplot(debug=1)
    g.xlabel('Vector size')
    g.ylabel('Relative error')
    g('set key horiz left top')
    g('set yrange [1e-16:1]')
    g('set logscale x')
    g('set logscale y')
    
    
    # Make the dataset list
    for i in range(len(blas_list)):
        
        # Extract name and precision
        (b, p) = blas_list[i]
        # Extract data
        data = data_list[i]
        # Add it to data sources
        if (not(b == "MKL" and p == "(DOUBLE)")):
            set_output_mode(g, operation + ERR_ALL_SUBFIX)
            d = Gnuplot.Data(data_list[i]['error_list'],  title = b + " " + p)
            g.replot(d)
    
# Do the performance plot for the last elements (e.g. very large vectors)
def plot_performance_tail(blas_list,  data_list,  operation):
    
    # Init graph
    g = Gnuplot.Gnuplot(debug=1)
    g.ylabel('Performance (MFlop/s)')
    g('set style data histograms')
    g('set style fill pattern 1 border -1')
    g('set style histogram clustered gap 5 title  offset character 0, 0, 0')
    g('set key horiz left top')
    g('set xtics border in scale 1,0.5 nomirror rotate by 90 offset character 0, 0, 0')
    g('set noxtics ')
    
    # Make the dataset list
    for i in range(len(blas_list)):
        
        # Extract name and precision
        (b, p) = blas_list[i]
        # Extract data
        data = data_list[i]
        # Add it to data sources
        tail_list = data_list[i]['perf_list'][len(data_list[i]['perf_list'])-1]
        tail_val = tail_list[1]
        set_output_mode(g, operation + PERF_TAIL_ALL_SUBFIX)
        d = Gnuplot.Data([tail_val],  title = b + " " + p)
        g.replot(d)

# Make a tex report in latex form    
def make_tex(operation,  formula):
    REPORT_NAME = operation + REPORT_SUBFIX
    
    latexfile = open(REPORT_NAME +".tex", "w")
    latexfile.write(r"""
    %% Automatically generated LaTeX file
    \documentclass[10pt]{article}
    \usepackage{graphicx}
    \usepackage{geometry}
    \geometry{a4paper,tmargin=1cm,bmargin=0cm,lmargin=2cm,rmargin=2cm}
    \pagestyle{empty}

    \begin{document}
    \title{%s performance report \\ $%s$}
    \pagestyle{empty}
    \author{Olav Fagerlund, OKUDA Lab}
    \maketitle

    \begin{figure}[htbp]
    \begin{minipage}[b]{0.45\hsize}	
    \begin{center} 		
    \includegraphics[]{%s}
    \caption{%s performance for all BLAS libraries in all available precisions in CUKr}
    \end{center}
    \end{minipage}	
    \hspace{0.1\hsize}
    \begin{minipage}[b]{0.45\hsize}	
    \begin{center} 		
    \includegraphics[]{%s}
    \caption{%s bandwidth for all BLAS libraries in all available precisions in CUKr}
    \end{center}
    \end{minipage}		
    \end{figure}

    \begin{figure}[htbp]
    \begin{minipage}[b]{0.45\hsize}	
    \begin{center} 		
    \includegraphics[]{%s}
    \caption{%s Relative error for all BLAS libraries in all available precisions in CUKr}
    \end{center}
    \end{minipage}	
    \hspace{0.1\hsize}
    \begin{minipage}[b]{0.45\hsize}	
    \begin{center} 		
    \includegraphics[]{%s}
    \caption{%s performance for very large vector sizes (around one million elements) for all BLAS libraries in all available precisions in CUKr}
    \end{center}
    \end{minipage}		
    \end{figure}	

    \end{document}
    """ % (operation,  formula,  operation + PERF_ALL_SUBFIX, operation,  
           operation + BAND_ALL_SUBFIX, operation,  operation + ERR_ALL_SUBFIX, 
           operation,  operation + PERF_TAIL_ALL_SUBFIX, operation) )
    
    latexfile.close()
    os.system("latex " + REPORT_NAME +".tex")
    os.system("dvipdfm " + REPORT_NAME + ".dvi")
    os.rename(REPORT_NAME + ".pdf", "../" + REPORT_NAME + ".pdf")
    
##########################
# Main
##########################

# Declare the option parser
usage = "usage: %prog [options] input_file"
descr = "Plots graphs out of the results outputed by the CUKr BLAS1 and BLASX1 benchmarks." 
parser = OptionParser(usage)
parser.set_description(descr)                      

parser.add_option("-c", "--color", dest="color", action="store_true", default = False,
                      help="Generate colored graphs [default: %default]")
parser.add_option("-p", "--point-size", dest="point_size", default = 1,
                      help="Choose point size [default: %default]")     
parser.add_option("-g", "--graphs-for-ppt", dest="graphs_for_ppt", action="store_true", default = False,
                      help="Output graphs for inserting in ppt [default: %default]")                        

# Read arguments
(options, args) = parser.parse_args()
if (len(args) < 1): 
    parser.print_help()
    sys.exit(1) 

# Process arguments
input_file = args[0]
print "\nReading data from: %s" % input_file

# Process options
if (options.graphs_for_ppt):
   OUTPUT_MODE = "GraphsForPPT" 
if (options.color):
   OUTPUT_COLOR = 1 

# Read the files and make a list for every blas and precision
blas_list = get_blas_and_precision(input_file)

# Read data
(operation, data_list) = get_operation_and_data(input_file,  blas_list)

# Get operation-specific text
formula = get_operation_specifics(operation)

# Make files folded (if not there) and cd into it
REPORT_FOLDER = operation + REPORT_SUBFIX + "_files"
if (not os.path.exists(REPORT_FOLDER)):
    os.mkdir(REPORT_FOLDER)
os.chdir(REPORT_FOLDER)    

# Do the overall performance plot
plot_performance(blas_list,  data_list, operation)

# Do the overall bandwidth plot
plot_bandwidth(blas_list,  data_list, operation)

# Do the overall error plot
plot_error(blas_list,  data_list, operation)

# Do overall performance for tail
plot_performance_tail(blas_list,  data_list, operation)

# Make tex
make_tex(operation, formula)

print "Done.\n"
