/**
 * @file benchmark/bench_blasx1.c
 * @author Serban Georgescu
 * @since 2008
 * 
 * BLASX1 benchmark routine, serial version.
 * 
 * Benchmarks the performance and 
 * accuracy for:
 * - PWPR
 * 
 * Benchmarks are done in 
 * 	- double precision
 * 	- quasi-double precision
 * 	- single precision
 * on
 * 	- CPU (all precisions)
 * 	- GPU (single, qdouble)
 * 
 */
 
#define LOGGER_NAME "benchmark.bench_blasx1"

/* Standard includes */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

/* Dependency includes */
#include <argtable2.h>

/* CUKr includes */
#include "../blasx/sp/blasx1_sp.h" 
#include "../cukr_all.h"

/* Blas includes 
 * @todo Add all blas includes here */
#include "../blasx/impl/generic/generic_blasx.h"
#include "../blas/impl/generic/generic_blas.h"
#include "../blasx/impl/gpu/gpu_blasx.h"
#include "../blas/impl/gpu/gpu_blas.h" 

/* Access the global database*/
extern CukrDB cukr_db;

// ----------------------- INITIAL CONFIGURATION---------------------- //

/** Lower and upper bounds and step size
 * for tests. The increment will be done
 * by multiplication with the step size,
 * not by addition */ 
#define TEST_LO   10*1024
#define TEST_UP   1*1024*1024
#define TEST_STEP 1.2

/** No of trials for the test 
 * Remember: more than one trials means
 * the smaller tests are done in-cache */
#define TEST_TRIALS 10  


// ----------------------------- GLOBALS ----------------------------- //

/**
 * Struct defining the option database
 */
typedef struct
{
	const char* progname;
	struct arg_str *test_op;
	struct arg_str *test_out_file;
	struct arg_int *test_trials;
	struct arg_int *test_lo;
	struct arg_int *test_up;
	struct arg_dbl *test_step;
	struct arg_lit *no_gpu;	
	struct arg_lit *no_cpu_qdouble;
	struct arg_lit *no_gpu_double;
	struct arg_lit *help;
    struct arg_end *end;
	void *argtable[13];
} Bench_OptionDB;

/*
 * Interface to the BLAS1 test functions
 */
 
typedef void (*BLASX1_TEST)(int n, double alpha, CUKR_BLAS blas, CUKR_VECTOR_SP x,
                            CUKR_VECTOR_SP y_orig, CUKR_VECTOR_SP y, 
                            double *perf, double *band); 

/*
 * Test parameters 
 */
char test_op[255];  
char test_out_file[2550];
int test_trials, test_lo, test_up, test_ref_id, no_gpu, no_cpu_qdouble, no_gpu_double;
double test_step;

/*
 * Blas parameters
 */
CUKR_BLAS blas_vector[100]; 
int blas_count;


// ------------------------- BENCHMARK FUNCTIONS -------------------------- //

void test_pwpr(int n, double alpha, CUKR_BLAS blas, CUKR_VECTOR_SP x, CUKR_VECTOR_SP y_orig, CUKR_VECTOR_SP y, double *perf, double *band)
{
	int t;
	
	/* Start from a copy of the original vector */
	CukrCOPY_SP(blas, y_orig, y); 
	
	/* Warm-up run */
	CukrPWPR_SP(blas, x, x, y);
	
	/* Set counters to zero */
	CukrBlasxResetCounters(blas->blasx);
	
	/* Do a no. of trials */
	for (t = 0; t < test_trials; t++)
		CukrPWPR_SP(blas, x, x, y);
		
	/* Read the counters */
	float elapsed = blas->blasx->counters[0][CUKR_BLAS_IDX_TIME]    [CUKR_BLASX_OPS_PWPR][CukrTypeToBlasPrec(blas->precision)];
	float flops   = blas->blasx->counters[0][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLASX_OPS_PWPR][CukrTypeToBlasPrec(blas->precision)];
	float loads   = blas->blasx->counters[0][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLASX_OPS_PWPR][CukrTypeToBlasPrec(blas->precision)];
	float stores  = blas->blasx->counters[0][CUKR_BLAS_IDX_FLSTORES][CUKR_BLASX_OPS_PWPR][CukrTypeToBlasPrec(blas->precision)];
			
	/* Compute performance */
	*perf = 1e-3 * flops / elapsed;
	*band = CukrBlasWordSize(blas) * 1e-3 * (loads + stores) / elapsed;					
} 

// -------------------------- DRIVER FUNCTIONS ---------------------------- //

/**
 * \brief Creates a vector that contains all
 *        the BLAS types supported by the
 *        CUKr distribution. 
 * 		  If support for other BLASes has been
 *        included, all should be defined, 
 *        created and registered here.
 * \param blas A vector of BLAS objects
 * \param blas_count The total number of
 *        registered objects.
 */ 
void create_blas_vector()
{
	
	blas_count = 0;
	
	/* 
	 * Add the Generic BLAS, all precisions 
	 * Create and associate blasx objects
	 */
	
	/* For double */
	CUKR_BLAS blas_generic_d; 
	CukrBlasCreate(&blas_generic_d);
	blas_create_generic(blas_generic_d);
	blas_generic_d->precision = CUKR_TYPE_DOUBLE;
	  
	CUKR_BLASX blasx_generic_d;
	CukrBlasxCreate(&blasx_generic_d);
	blasx_create_generic(blasx_generic_d); 
	blas_generic_d->blasx = blasx_generic_d;
	
	/* For qdouble */
	CUKR_BLAS blas_generic_q;
	CukrBlasCreate(&blas_generic_q);
	blas_create_generic(blas_generic_q); 
	blas_generic_q->precision = CUKR_TYPE_QDOUBLE; 
	
	CUKR_BLASX blasx_generic_q;
	CukrBlasxCreate(&blasx_generic_q);
	blasx_create_generic(blasx_generic_q); 
	blas_generic_q->blasx = blasx_generic_q;
	         
	/* For single */
	CUKR_BLAS blas_generic_s;         
	CukrBlasCreate(&blas_generic_s);          
	blas_create_generic(blas_generic_s);
	blas_generic_s->precision = CUKR_TYPE_SINGLE; 
	
	CUKR_BLASX blasx_generic_s;
	CukrBlasxCreate(&blasx_generic_s);
	blasx_create_generic(blasx_generic_s); 
	blas_generic_s->blasx = blasx_generic_s;
	
	/* Register */
	blas_vector[0] = blas_generic_d;
	blas_vector[1] = blas_generic_s;
	blas_count += 2;
	
	if (!no_cpu_qdouble) {
		blas_vector[2] = blas_generic_q;
		blas_count++;
	}
	
	/* Set reference blas to generic double */
	test_ref_id = 0;
	
	/* 
	 * Add the GPU BLAS, qdouble and single
	 * (no double support included yet) 
	 */
	 
	#ifdef CUKR_USE_GPU
	if (!no_gpu) {
		
		#ifdef CUKR_GPU_DOUBLE
		if (!no_gpu_double) {	
			
			/* For double */
			CUKR_BLAS blas_GPUBLAS_d;
			CukrBlasCreate(&blas_GPUBLAS_d);
			blas_create_gpu(blas_GPUBLAS_d);
			blas_GPUBLAS_d->precision = CUKR_TYPE_DOUBLE;
			
			CUKR_BLASX blasx_GPUBLAS_d;
			CukrBlasxCreate(&blasx_GPUBLAS_d);
			blasx_create_gpu(blasx_GPUBLAS_d);
			blas_GPUBLAS_d->blasx = blasx_GPUBLAS_d;
			
			blas_vector[blas_count + 0] = blas_GPUBLAS_d;
			blas_count++;
		}
		#endif
		
		/* For qdouble */
		CUKR_BLAS blas_GPUBLAS_q;
		CukrBlasCreate(&blas_GPUBLAS_q);
		blas_create_gpu(blas_GPUBLAS_q);
		blas_GPUBLAS_q->precision = CUKR_TYPE_QDOUBLE;
		
		CUKR_BLASX blasx_GPUBLAS_q;
		CukrBlasxCreate(&blasx_GPUBLAS_q);
		blasx_create_gpu(blasx_GPUBLAS_q);
		blas_GPUBLAS_q->blasx = blasx_GPUBLAS_q;
	
		/* For single */
		CUKR_BLAS blas_GPUBLAS_s;
		CukrBlasCreate(&blas_GPUBLAS_s);
		blas_create_gpu(blas_GPUBLAS_s);
		blas_GPUBLAS_s->precision = CUKR_TYPE_SINGLE;
		
		CUKR_BLASX blasx_GPUBLAS_s;
		CukrBlasxCreate(&blasx_GPUBLAS_s);
		blasx_create_gpu(blasx_GPUBLAS_s);
		blas_GPUBLAS_s->blasx = blasx_GPUBLAS_s;
		
		/* Register */
		blas_vector[blas_count + 0] = blas_GPUBLAS_q;
		blas_vector[blas_count + 1] = blas_GPUBLAS_s;
		blas_count += 2;
		
		/* Set GPU support */
		cukr_db.compute_on_gpu = CUKR_TRUE;
	}
		
	#endif  
}

// ------------------------------------------------------------------- //

/** 
 * \brief Process cmdline options and arguments
 * 		  and return the option database
 * 
 * \param argc the number of arguments
 * \param argv a vector containing the list of
 *        arguments 
 * \param optiondb a pointer to a OptionDB 
 *        structure, where the database is returned
 *  
 **/
void bench_blasx1_process_arguments(int argc, char* argv[], Bench_OptionDB *optiondb)
{
	/*
	 * Declare options
	 */
	 optiondb->progname      = "bench_blasx1";
	 optiondb->test_op		 = arg_str0(NULL, NULL, 
	                               		"{PWPR}", 
	                               		"BLAS1 operation to be benchmarked");
	 optiondb->test_out_file = arg_str0("o", NULL, 
	                               		"<file>", 
	                              		"Result file "\
	                              		"(default: bench_<OP>.result.csv)"); 
	 optiondb->test_trials   = arg_int0("t", "trials", 
										"int", 
										"Number of test trials "\
									    "(default: 10)");	
	 optiondb->test_lo  	 = arg_int0("l", "lo", 
										"int", 
										"Start vector size "\
									    "(default: 10*1024)");		
	 optiondb->test_up  	 = arg_int0("u", "up", 
										"int", 
										"End vector size "\
									    "(default: 1*1024*1024)");	
	 optiondb->test_step  	 = arg_dbl0("s", "step", 
										"double", 
										"Size increment (by multiplication) "\
									    "(default: 1.2)");	
	 optiondb->no_gpu         = arg_lit0(NULL,  "no-gpu", 
									    "Do not use a GPU even if possible");
 	 optiondb->no_cpu_qdouble = arg_lit0(NULL,  "no-cpu-qdouble", 
									    "Do not use qdouble on the CPU (because it's slow)");									    								    	
	 optiondb->no_gpu_double  = arg_lit0(NULL,  "no-gpu-double", 
									    "Do not use double on the GPU (when is not supported)");									    								    	
	 optiondb->help          = arg_lit0("hH",  "help", 
									    "display this help and exit");
	 optiondb->end           = arg_end(11);								    									    
									    
	 optiondb->argtable[0]  = optiondb->test_op; 
	 optiondb->argtable[1]  = optiondb->test_out_file;
	 optiondb->argtable[2]  = optiondb->test_trials;
	 optiondb->argtable[3]  = optiondb->test_lo;
	 optiondb->argtable[4]  = optiondb->test_up;
	 optiondb->argtable[5]  = optiondb->test_step;
	 optiondb->argtable[6]  = optiondb->no_gpu;
	 optiondb->argtable[7]  = optiondb->no_cpu_qdouble;
	 optiondb->argtable[8]  = optiondb->no_gpu_double;
	 optiondb->argtable[9]  = optiondb->help;
	 optiondb->argtable[10] = optiondb->end;
	 
	 /*
	  * Parse options and arguments
	  */
	  
	 /* Verify the argtable[] entries were allocated sucessfully */	
	 if (arg_nullcheck(optiondb->argtable) != 0)
     {
       /* NULL entries were detected, some allocations must have failed */
       LOG(FATAL, "Argtable initialization failed. \n");
       exit(1);
     }
    
     /* Parse the command line as defined by argtable[] */
     int nerrors = arg_parse(argc, argv, optiondb->argtable); 
    
     /* Special case: '--help' takes precedence over error reporting */
     if (optiondb->help->count > 0)
     {
        printf("\nUsage: bench_blasx1 <test> [options]");
        arg_print_syntax(stdout, optiondb->argtable, "\n\n");
        arg_print_glossary(stdout, optiondb->argtable, "  %-25s %s\n");
        exit(1);
     }

     /* If the parser returned any errors then display them and exit */
     if (nerrors > 0)
     {
        /* Display the error details contained in the arg_end struct.*/
        arg_print_errors(stdout, optiondb->end, optiondb->progname);
        printf("Errors. Try '--help' for more information.\n");
        exit(1);
     } 
     
     /*
	  * Do validation
	  */
	 								    							    								    							                                  		                             
	 /* The test name is compulsory */
	 if (!optiondb->test_op->count)
	 {
	 	printf("Please specify the test name.\n");
        exit(1);
	 }
	 	                           
	 /*
	  * Extract values and add them to 
	  * global options
	  */	          
	  
	  sprintf(test_op, "%s", optiondb->test_op->sval[0]);
	  
	  if (optiondb->test_out_file->count)
	  	  sprintf(test_out_file, "%s", optiondb->test_out_file->sval[0]);  
	  else
	  	  sprintf(test_out_file, "bench_%s.result.csv", test_op); 	
	  	  
	  test_trials = optiondb->test_trials->count ? optiondb->test_trials->ival[0] : TEST_TRIALS;
	  test_lo     = optiondb->test_lo->count     ? optiondb->test_lo->ival[0]     : TEST_LO;
	  test_up     = optiondb->test_up->count     ? optiondb->test_up->ival[0]     : TEST_UP;
	  test_step   = optiondb->test_step->count   ? optiondb->test_step->dval[0]   : TEST_STEP;
	  no_gpu   	  	 = optiondb->no_gpu->count;
	  no_cpu_qdouble = optiondb->no_cpu_qdouble->count; 
	  no_gpu_double  = optiondb->no_gpu_double->count;		                 
}

// ------------------------------------------------------------------- //

/**
 * \brief Runs the test indicated by the tid using
 *        all BLAS libraries found in the blas vector
 * \param test_func Function pointer of the operation
 *                  being tested
 * \param f Output file
 */
void run(BLASX1_TEST test_func, FILE *f)
{	
	int i, n;
	
	/* Init random number generator */
	srand(time(NULL));
	
	/* Define vars to hold counters */
	double perf, band, error;
	char buffer[255];
		
	/* Run the selected test */	
	for (n = test_lo; n <= test_up; n *= test_step)
	{	
		/* Get random alpha */
		CUKR_DOUBLE alpha = rand() / (CUKR_DOUBLE)RAND_MAX;
		
		/* Create test vectors, in double */
		CUKR_VECTOR_SP x = NULL, y_orig = NULL, y[blas_count], y_ref;
		CukrVecspCreate(&x,      n, CUKR_TYPE_DOUBLE, "x");
		CukrVecspCreate(&y_orig, n, CUKR_TYPE_DOUBLE, "y");
		
		/* Fill vectors with random values */
		CukrVecspSetFromVal(x,      1.1234567890123456789, blas_vector[test_ref_id]->precision);
		CukrVecspSetFromVal(y_orig, 2.2345678912345678901, blas_vector[test_ref_id]->precision);
						
		/* Run test with the reference BLAS */
		y_ref = NULL;
		CukrVecspCreate(&y_ref, n, CUKR_TYPE_DOUBLE, "y_ref");
		(test_func)(n, alpha, blas_vector[test_ref_id], x, y_orig, y_ref, &perf, &band);
		CukrVecspTransfer(y_ref, CUKR_LOCATION_CPU, CUKR_TYPE_DOUBLE);
		
		/* Output performance for reference */
		CukrTypeIdToStr(blas_vector[test_ref_id]->precision, buffer);
		printf("%10d  %10s   %10s   %11.6g  %11.6g  %d\n", n, blas_vector[test_ref_id]->name, buffer, 
			   perf, band, 0);
		fprintf(f, "%d,%s,%s,%g,%g,%d\n", n, blas_vector[test_ref_id]->name, buffer, perf, band, 0);
		
		/* Run tests for the rest of the BLASes */
		int b;
		for (b = 0; b < blas_count; b++)
			if (b != test_ref_id)
			{
				/* Run test */				
				y[b] = NULL;
				sprintf(buffer, "y[%d]", b);
				CukrVecspCreate(&y[b], n, CUKR_TYPE_DOUBLE, "y[ref]"); 
				(test_func)(n, alpha, blas_vector[b], x, y_orig, y[b], &perf, &band);
				CukrVecspTransfer(y[b], CUKR_LOCATION_CPU, CUKR_TYPE_DOUBLE);
				
				/* Compute error */
				double max_err = 0;				
				for (i = 0; i < n; i++) {
					double err = fabs(y[b]->ph_dval[i] - y_ref->ph_dval[i]);
					double rel_err = err / fabs(y_ref->ph_dval[i]); 
					if (max_err < rel_err)
						max_err = rel_err; 
				}
				
				/* Output performance */
				CukrTypeIdToStr(blas_vector[b]->precision, buffer);
				printf("%10d  %10s   %10s   %11.6g  %11.6g  %11.6g\n", n, blas_vector[b]->name, buffer, 
					   perf, band, max_err);
				fprintf(f, "%d,%s,%s,%g,%g,%g\n", n, blas_vector[b]->name, buffer, perf, band, max_err);
				
				/* Destroy result vector */
				CukrVecspDestroy(&y[b]);
			}
			
		/* Destroy the rest of the vectors */
		CukrVecspDestroy(&x);
		CukrVecspDestroy(&y_orig); 
		CukrVecspDestroy(&y_ref);
	}
}
// ------------------------------------------------------------------- //

int main(int argc, char* argv[]) 
{
	/* 
	 * Parse command-line options 
	 */
	
	Bench_OptionDB optiondb; 
	bench_blasx1_process_arguments(argc, argv, &optiondb);
	
	printf("\n" \
	       "Test configuration: \n"
	       "====================================\n"\
	       "Operation: %s\n"\
	       "Start vector size: %d\n"\
	       "End vector size: %d\n"\
	       "Vector size step: %g\n"\
	       "\n"\
	       "Result file: %s\n"\
	       "====================================\n",
	       test_op, test_lo, test_up, test_step, 
	       test_out_file);	                    
	
	/*
	 * Init CUKr
	 * (selective initialization)
	 */
	
	 /* No options to the library */
	 CukrInit(0, NULL, CUKR_DEFAULT);
	 CukrStart(CUKR_DEFAULT); 
	
	/*
	 * Create BLAS vector
	 */
	 
	
	int i;
	create_blas_vector();
	
	printf("\n" \
	       "BLAS libraries used: \n"
	       "====================================\n");
	       
	for (i = 0; i < blas_count; i++)
	{
		char type_str[255];
		CukrTypeIdToStr(blas_vector[i]->precision, type_str);
		printf("%s in %s precision\n", 
		        blas_vector[i]->name, type_str);
	}  
	printf("====================================\n");
	
	/*
	 * Get test function
	 */
	
	BLASX1_TEST test_func;
	
	if (!strcmp(test_op, "PWPR") || !strcmp(test_op, "pwpr"))
		test_func = test_pwpr;
	else {
		printf("Error: Unknown test: %s\n", test_op);
		exit(1);
	}			
	
	/*
	 * Run test 
	 */
	 
	/* Open output file */
	FILE *f;
	if ( (f = fopen(test_out_file, "w")) == NULL) {
		CUKR_FATALX("Could not open file %s", test_out_file);
	} 
	
	/* Write test name and header */
	fprintf(f, "%s\n", test_op);
	printf("\n     N          BLAS        Precision     MFlops/s      MB/s      Rel.err.    \n");
	fprintf(f, "N,BLAS,Precision,MFlops/s,MB/s,Error\n");
	
	/* Run test */
	run(test_func, f);
	
	/* Close output file */
	fclose(f);	 
	
	/*
	 * Close CUKr
	 */
	  
	CukrFinalize(CUKR_DEFAULT);	 
}

