#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "../cukr_all.h" 
#include "../db/db.h" 
#include "../util/log.h" 
#include "../util/macros.h"

#include "../blas/impl/gpu/cuda/kernels.h"
#include "../blas/impl/generic/generic_blas.h"

/* Access the global database*/
extern CukrDB cukr_db;

/* GPU memory paramters */
#define PI 3.1415

#define TEST_FUNC 1

/*
 * Forward decl
 */
double rhs(double x, double y, double z);
double anal(double x, double y, double z);


// ------- First test rhs and anal solution

#if TEST_FUNC == 0
double rhs(double x, double y, double z) {

	return 2*y*z*(1-y)*(1-z) +
	2*x*y*(1-x)*(1-y) +
	2*x*z*(1-x)*(1-z);
}
double anal(double x, double y, double z) {
	return -x*(1-x)*y*(1-y)*z*(1-z);
}
#endif

// ------- 2nd test rhs and anal solution

#if TEST_FUNC == 1
double rhs(double x, double y, double z) {

	return -exp(x+y+z)*x*y*z*(y*(z-5)-5*z+x*(3*z*y+y+z-5)+9);
}
double anal(double x, double y, double z) {
	return exp(x+y+z)*x*(1-x)*y*(1-y)*z*(1-z);
}
#endif

// ------- 3rd test rhs and anal solution

#if TEST_FUNC == 2
double rhs(double x, double y, double z) {
	return -3*PI*PI*sin(PI*x)*sin(PI*y)*sin(PI*z);
}
double anal(double x, double y, double z) {
	return sin(PI*x)*sin(PI*y)*sin(PI*z);
}
#endif


/*
 * Run CG for the specified matrix
 */
void test_matrix(CUKR_MATRIX A, CUKR_VECTOR b, CUKR_DOUBLE* correct_sol, const char* filename)
{
	int i,j,k,l;
	CUKR_VECTOR x;
	CukrVecCreate(&x, A->cols, CUKR_DEFAULT, CUKR_SHTYPE_DISTRIBUTED, "x");


	/* Set x = 0, b = 1 */
	CukrVecSetFromVal(x, 0.0, CUKR_DEFAULT);
	//CukrVecPrint(b);
	   
	/*
	 * Solve the system
	 */   
	       
	/* Create solver and assign operators */
	CUKR_SOLVER solver;
	CukrSolverCreate(&solver, "CG Solver");
	CukrSolverSetOperators(solver, A);
	/* Solve the system */  
	CukrSolverSolve(solver, b, x);
	printf("%s\n", CukrSolverGetConvergedReasonAsStr(solver, CUKR_OUTER));

	/* Extract performance info */
	double flops = solver->flops / cukr_db.solverTime / 1000.0;
	int    iters = solver->its;
	double time_per_iter = cukr_db.solverTime / iters;
	double spmv_time = solver->blas->counters_mp[CUKR_BLAS_IDX_TIME][CUKR_BLAS_OPS_SPMV_CSR]; // ms
	double spmv_time_per_iter = spmv_time / iters ;
	double spmv_comm_time_iter = solver->blas->comm_inner / iters ;
	double spmv_flops = (2*A->nz) / spmv_time_per_iter / 1e3;
	printf("Number of non-zeroes: %d\n", A->nz);
	printf("Solver %s: %g Mflop/s, %d iters, %g ms\n", A->name, flops, iters, cukr_db.solverTime);
	printf("SPMV: %g ms/iter, SPMV_COMM = %g ms/iter, SPMV flops = %g Mflop/s\n", spmv_time_per_iter, spmv_comm_time_iter, spmv_flops);

	/* Get solution */
	CUKR_DOUBLE* sol = (CUKR_DOUBLE*)malloc(A->rows*sizeof(CUKR_DOUBLE));
	CukrVecGetToVectorD(solver->sol, 0, A->rows, sol, CUKR_LOCATION_CPU);

	/* Compute error */
	double max_err=0;
	for (i = 0; i < A->rows; i++) {
		double err = fabs(correct_sol[i]-sol[i])/fabs(sol[i]);
		//printf("i = %d, sol = %g, correct_sol = %g, err = %g\n", i, sol[i], correct_sol[i], err);
		if (err > max_err) max_err = err;
	}
	printf("Error to analytical solution = %g\n", max_err);
	
	
	/* 
	 * Write performance data to file
	 */ 
	
	/* Open output file */
	FILE *f;
	if ( (f = fopen("solver_perf.csv", "w")) == NULL) {
		CUKR_FATAL("Unable to open solver perf. output file\n");
	}
	
	/* Write solver performance */
	int rows = A->rows;
	int cols = A->cols;
	int nz = A->nz;
	int exit_reason = solver->reason;
	double exit_residual = sqrt(solver->rnorm2 / solver->rnorm02);
	//double flops = solver->flops / cukr_db.solverTime / 1000.0;
	double time = cukr_db.solverTime;
	//int    iters = solver->its;
	//double time_per_iter = time / iters;
	
	double dot_time_iter  = solver->blas->counters[0][CUKR_BLAS_IDX_TIME][CUKR_BLAS_OPS_DOT][CukrTypeToBlasPrec(solver->blas->precision)] / solver->its;
	double axpy_time_iter = solver->blas->counters[0][CUKR_BLAS_IDX_TIME][CUKR_BLAS_OPS_AXPY][CukrTypeToBlasPrec(solver->blas->precision)] / solver->its;
	double aypx_time_iter = solver->blas->counters[0][CUKR_BLAS_IDX_TIME][CUKR_BLAS_OPS_AYPX][CukrTypeToBlasPrec(solver->blas->precision)] / solver->its;
	//double spmv_time_iter = solver->blas->counters[0][CUKR_BLAS_IDX_TIME][CUKR_BLAS_OPS_SPMV_CSR][CukrTypeToBlasPrec(solver->blas->precision)] / solver->its;
	//double pwpr_time_iter = solver->blas->blasx->counters[0][CUKR_BLAS_IDX_TIME][CUKR_BLASX_OPS_PWPR][CukrTypeToBlasPrec(solver->blas->precision)] / solver->its;
	
	
	/* Read the counters */
	float elapsed_axpy = solver->blas->counters[0][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_AXPY][CukrTypeToBlasPrec(solver->blas->precision)];
	float flops_axpy   = solver->blas->counters[0][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_AXPY][CukrTypeToBlasPrec(solver->blas->precision)];
	float loads_axpy   = solver->blas->counters[0][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_AXPY][CukrTypeToBlasPrec(solver->blas->precision)];
	float stores_axpy  = solver->blas->counters[0][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_AXPY][CukrTypeToBlasPrec(solver->blas->precision)];
	
	/* Compute performance */
	double perf_axpy = 1e-3 * flops_axpy / elapsed_axpy;
	double band_axpy = CukrBlasWordSize(solver->blas) * 1e-3 * (loads_axpy + stores_axpy) / elapsed_axpy;
	
	/* Read the counters */
	float elapsed_aypx = solver->blas->counters[0][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_AYPX][CukrTypeToBlasPrec(solver->blas->precision)];
	float flops_aypx   = solver->blas->counters[0][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_AYPX][CukrTypeToBlasPrec(solver->blas->precision)];
	float loads_aypx   = solver->blas->counters[0][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_AYPX][CukrTypeToBlasPrec(solver->blas->precision)];
	float stores_aypx  = solver->blas->counters[0][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_AYPX][CukrTypeToBlasPrec(solver->blas->precision)];
	
	/* Compute performance */
	double perf_aypx = 1e-3 * flops_aypx / elapsed_aypx;
	double band_aypx = CukrBlasWordSize(solver->blas) * 1e-3 * (loads_aypx + stores_aypx) / elapsed_aypx;
	
	/* Read the counters */
	float elapsed_dot = solver->blas->counters[0][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_DOT][CukrTypeToBlasPrec(solver->blas->precision)];
	float flops_dot   = solver->blas->counters[0][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_DOT][CukrTypeToBlasPrec(solver->blas->precision)];
	float loads_dot   = solver->blas->counters[0][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_DOT][CukrTypeToBlasPrec(solver->blas->precision)];
	float stores_dot  = solver->blas->counters[0][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_DOT][CukrTypeToBlasPrec(solver->blas->precision)];
	
	/* Compute performance */
	double perf_dot = 1e-3 * flops_dot / elapsed_dot;
	double band_dot = CukrBlasWordSize(solver->blas) * 1e-3 * (loads_dot + stores_dot) / elapsed_dot;
	
	/* Read the thread counters */
	float elapsed_spmv = solver->blas->counters_mp[CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR];//[CukrTypeToBlasPrec(solver->blas->precision)];
	float flops_spmv   = solver->blas->counters_mp[CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR];//[CukrTypeToBlasPrec(solver->blas->precision)];
	/*
	float floads_spmv  = solver->blas->counters[0][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CukrTypeToBlasPrec(solver->blas->precision)];
	float fstores_spmv = solver->blas->counters[0][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CukrTypeToBlasPrec(solver->blas->precision)];
	float iloads_spmv  = solver->blas->counters[0][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CukrTypeToBlasPrec(solver->blas->precision)];
	float istores_spmv = solver->blas->counters[0][CUKR_BLAS_IDX_INTSTORES][CUKR_BLAS_OPS_SPMV_CSR][CukrTypeToBlasPrec(solver->blas->precision)];
	*/
	//printf("%f %f %f %f\n", floads_spmv, fstores_spmv, iloads_spmv, istores_spmv);
	//printf("%g %g %g %g\n", floads_spmv, fstores_spmv, iloads_spmv, istores_spmv);
	/* Compute performance */
	double perf_spmv = 1e-3 * flops_spmv / elapsed_spmv;
	//double band_spmv = 1e-3 * (CukrBlasWordSize(solver->blas)*(floads_spmv + fstores_spmv) + 4*(iloads_spmv + istores_spmv)) / elapsed_spmv;	

	/*
	// Read the counters
	float elapsed_spmv = solver->blas->counters[0][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_SPMV][CukrTypeToBlasPrec(solver->blas->precision)];
	float flops_spmv   = solver->blas->counters[0][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_SPMV][CukrTypeToBlasPrec(solver->blas->precision)];
	float loads_spmv   = solver->blas->counters[0][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_SPMV][CukrTypeToBlasPrec(solver->blas->precision)];
	float stores_spmv  = solver->blas->counters[0][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_SPMV][CukrTypeToBlasPrec(solver->blas->precision)];
	
	// Compute performance
	double perf_spmv = 1e-3 * flops_spmv / elapsed_spmv;
	double band_spmv = CukrBlasWordSize(solver->blas) * 1e-3 * (loads_spmv + stores_spmv) / elapsed_spmv;
	 */
	/*
	// Read the counters
	float elapsed_scal = solver->blas->counters[0][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_SCAL][CukrTypeToBlasPrec(solver->blas->precision)];
	float flops_scal   = solver->blas->counters[0][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_SCAL][CukrTypeToBlasPrec(solver->blas->precision)];
	float loads_scal   = solver->blas->counters[0][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_SCAL][CukrTypeToBlasPrec(solver->blas->precision)];
	float stores_scal  = solver->blas->counters[0][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_SCAL][CukrTypeToBlasPrec(solver->blas->precision)];
	
	// Compute performance
	double perf_scal = 1e-3 * flops_scal / elapsed_scal;
	double band_scal = CukrBlasWordSize(solver->blas) * 1e-3 * (loads_scal + stores_scal) / elapsed_scal;
	*/
	
	printf("Solver ran at %g Mflop/s and spent %g ms/iter\n", flops, time_per_iter);
	//fprintf(f, "rows;cols;nz;cg_perf;time;iters;dot_time_iter;axpy_time_iter;aypx_time_iter;pwpr_time_iter;spmv_time_iter;exit_reason;exit_residual;dot_perf; dot_band; dot_elapsed; axpy_perf; axpy_band; axpy_elapsed; aypx_perf; aypx_band; aypy_elapsed; spmv_perf; spmv_band; spmv_elapsed\n");
	/*
	fprintf(f, "%s;%d;%d;%d;%g;%g;%d;%g;%g;%g;%g;%d;%g;%g;%g;", filename,rows,cols,nz,flops,time,iters,dot_time_iter,axpy_time_iter,aypx_time_iter,spmv_time__per_iter,exit_reason,exit_residual, 
			perf_dot, band_dot);
	fprintf(f, "%g;%g;%g;%g;%g;%g;%g\n", elapsed_dot, perf_axpy, band_axpy, elapsed_axpy, perf_aypx, band_aypx, elapsed_aypx, perf_spmv, band_spmv, elapsed_spmv);
	*/
	fprintf(f, "%s;%d;%d;%d;%g;%g;%d;%g;%g;%g;%g;%d;%g;%g;%g;%g;%g;%g;%g;%g;%g;%g;%g;%s;%g\n", filename,rows,cols,nz,flops,time,iters,dot_time_iter,axpy_time_iter,aypx_time_iter,spmv_time_per_iter,exit_reason,exit_residual, 
			perf_dot, band_dot, elapsed_dot, perf_axpy, band_axpy, elapsed_axpy, perf_aypx, band_aypx, elapsed_aypx, perf_spmv, "N/A", elapsed_spmv);
	
	/*
	 * Shutdown
	 */
	
	/* Close output file */
	fclose(f);

    /* Cleanup */
	CukrVecDestroy(&x);
	CukrVecDestroy(&b);
	CukrSolverDestroy(&solver);
}

int main(int argc, char* argv[])
{
	int i, j, k, l, p;

	/*
	 * Init CUKr
	 */

	/* Init runtime */
	CukrInit(argc, argv, CUKR_OPTS_STRICT_SYNTAX);

	/* Start runtime */
	CukrStart(CUKR_DEFAULT);

	/* Set common objects */
	CUKR_MATRIX A;
	CUKR_VECTOR b;
	CUKR_DOUBLE* correct_sol;

	/*
	 * Create system
	 */

	const char* matrix_file = CukrDBGetS(CUKR_DB_MATRIX);
	if (!CukrStricmp(matrix_file, ""))										// Generate Poisson
	{
		/* Create entries for diagonals */
		int bands = 7;
		char buffer[255];

		/* Chose problem size */
		int N = 64;//(int)pow((double)n, 1/3.0);
		int n = N*N*N; //chose_problem_size(bands);

		/* Create band vectors */
		CUKR_DOUBLE *main_diag, *diag;
		main_diag = (CUKR_DOUBLE*)malloc(n*sizeof(CUKR_DOUBLE));
			 diag = (CUKR_DOUBLE*)malloc(n*sizeof(CUKR_DOUBLE));

		/* Set values */
		double h = 1.0 / (N+1);
		for (i = 0; i < n; i++)
			main_diag[i] = -6/(h*h);
			//main_diag[i] = -1;

		/* Create Poisson matrix */
		sprintf(buffer, "%dBands(%d,%d)", bands, n, N*N);
		CukrMatCreate(&A, n, n, CUKR_MATRIX_FORMAT_CSR, CUKR_DEFAULT, buffer);

		/* Add diagonals to matrix */
		for (i = 0; i < n; i++) {
			diag[i] = 1/(h*h);
			if ((i+1) % N == 0) diag[i] = 0;
			//diag[i] = 1;
		}
		CukrMatAddDiag(A,  -1,    diag);
		CukrMatAddDiag(A,   0,    main_diag);
		CukrMatAddDiag(A,   1 ,   diag);

		for (i = 0; i < n; i++) {
			diag[i] = 1/(h*h);
			if (i % (N*N) >= N*N - N) {diag[i] = 0;}
		}
		CukrMatAddDiag(A,  -N,    diag);
		CukrMatAddDiag(A,   N,    diag);

		for (i = 0; i < n; i++) {
			diag[i] = 1/(h*h);
		}
		CukrMatAddDiag(A,  -N*N,  diag);
		CukrMatAddDiag(A,   N*N,  diag);

		/* Assemble matrix */
		CukrMatAssemble(A);
		//CukrMatPrint(A);
		sprintf(buffer, "poisson3D_%d.csr", N);
		//CukrCsrMatspSaveToFile(A, buffer);
		//exit(1);

		/* Set RHS */
		printf("Setting rhs\n");
		CukrVecCreate(&b, A->rows, CUKR_TYPE_DOUBLE, CUKR_SHTYPE_DISTRIBUTED, "b");
		l = 0; p = 0;
		for (k = 0; k < N; k++)
			for (j = 0; j < N; j++)
				for (i = 0; i < N; i++) {
					double xx = h + i*h;
					double yy = h + j*h;
					double zz = h + k*h;
					#ifdef CUKR_MP
						b->cells[p]->ph_dval[l] = rhs(xx,yy,zz);
						l++;
						if (l >= b->cells[p]->n) {p++; l = 0;}
					#else
						b->ph_dval[l] = rhs(xx,yy,zz);
						l++;
					#endif
				}

		// Set anal solution vector
		correct_sol = (CUKR_DOUBLE*)malloc(A->rows*sizeof(CUKR_DOUBLE));
		l = 0;
		for (k = 0; k < N; k++)
			for (j = 0; j < N; j++)
				for (i = 0; i < N; i++) {
					double xx = h + i*h;
					double yy = h + j*h;
					double zz = h + k*h;
					correct_sol[l++] = anal(xx,yy,zz);
				}

	}

	else																// Read from file
	{
		/* Get desired matrix format */
		int mat_format_id = CukrMatFormatStrToId(CukrDBGetS(CUKR_DB_MATRIX_FORMAT));

		// If hybrid format, load as csr and let it be automatically converted later
		if (mat_format_id == CUKR_MATRIX_FORMAT_HYB)
			CukrMatCreateFromFile(&A, matrix_file, CUKR_MATRIX_FORMAT_CSR, CUKR_DEFAULT, "A");
		// If not, read using the default format
		else
			CukrMatCreateFromFile(&A, matrix_file, CUKR_DEFAULT, CUKR_DEFAULT, "A");
		CukrMatAssemble(A);
		//CukrMatPrint(A);

		// Set solution vector
		CukrVecCreate(&b, A->rows, CUKR_TYPE_DOUBLE, CUKR_SHTYPE_DISTRIBUTED, "b");
		correct_sol = (CUKR_DOUBLE*)malloc(A->rows*sizeof(CUKR_DOUBLE));

		// If 3d poisson matrix
		if (strstr(matrix_file, "oisson3"))
		{
			int N = round(pow(A->rows, 1/3.));
			double h = 1.0 / (N+1);
			printf("Loading 3D poisson matrix (%d^3). Setting special RHS and SOL\n", N);

			// Set rhs
			l = 0; p = 0;
			for (k = 0; k < N; k++)
				for (j = 0; j < N; j++)
					for (i = 0; i < N; i++) {
						double xx = h + i*h;
						double yy = h + j*h;
						double zz = h + k*h;
						#ifdef CUKR_MP
							b->cells[p]->ph_dval[l] = rhs(xx,yy,zz);
							l++;
							if (l >= b->cells[p]->n) {p++; l = 0;}
						#else
							b->ph_dval[l] = rhs(xx,yy,zz);
							l++;
						#endif
					}

			// Set correct sol
			l = 0;
			for (k = 0; k < N; k++)
				for (j = 0; j < N; j++)
					for (i = 0; i < N; i++) {
						double xx = h + i*h;
						double yy = h + j*h;
						double zz = h + k*h;
						correct_sol[l++] = anal(xx,yy,zz);
					}
		}
		else
		{
			for (i = 0; i < A->rows; i++) correct_sol[i] = 1.12312312;//rand();
			CUKR_VECTOR desired_sol;
			CukrVecCreate(&desired_sol, A->rows, CUKR_TYPE_DOUBLE, CUKR_SHTYPE_SHARED, "desired_sol");
			CukrVecSetFromVectorD(desired_sol, 0, A->rows, correct_sol, CUKR_LOCATION_CPU);

			// Get a double precision blas
			CUKR_BLAS blas_generic_d;
			CukrBlasCreate(&blas_generic_d);
			blas_create_generic(blas_generic_d);
			blas_generic_d->precision = CUKR_TYPE_DOUBLE;

			// Get rhs
			CukrSPMV(blas_generic_d, 1, A, desired_sol, 0, b);
		}

	}

	/* Run test */
	test_matrix(A, b, correct_sol, matrix_file);

	/* Destroy matrix */
	CukrMatDestroy(&A);

	/* Return */  
	return 0;  
}    
