#include <stdio.h>
#include "../cukr_all.h" 
#include "../db/db.h" 
#include "../util/log.h" 

/* Access the global database*/
extern CukrDB cukr_db;

int main(int argc, char* argv[]) 
{        
	/*  
	 * Init CUKr
	 */         
	               
	/* Init runtime */      
	CukrInit(argc, argv, CUKR_OPTS_STRICT_SYNTAX);	
	
	/* Enable gathering of eigenvalue info */
	CukrDBSetI(CUKR_DB_SOLVER_EIGENINFO, 1);
	                       	     	        
	/* Start runtime */
	CukrStart(CUKR_DEFAULT);
	                       
	/*  
	 * Create system
	 */     
	
	/* Exit if no matrix file was specified */
	const char* matrix_file = CukrDBGetS(CUKR_DB_MATRIX);
	if (!CukrStricmp(matrix_file, "")) {
		CUKR_FATAL("Please privide a matrix filename using "\
		           "the --matrix=MATRIX_FILE syntax");
	}   
	    
	/* Otherwise, create it from file */
	CUKR_MATRIX A;  
	CukrMatCreateFromFile(&A, matrix_file, CUKR_DEFAULT, CUKR_DEFAULT, "A");
	CukrMatAssemble(A);    
	                     
	/* Create initial solution and rhs */
	CUKR_VECTOR x,b;
	CukrVecCreate(&x, A->cols, CUKR_DEFAULT, CUKR_SHTYPE_SHARED, "x");
	CukrVecCreate(&b, A->rows, CUKR_DEFAULT, CUKR_SHTYPE_DISTRIBUTED, "b");
	      
	/* Set x = 0, b = 1 */
	CukrVecSetFromVal(x, 0.0, CUKR_DEFAULT);
	CukrVecSetFromVal(b, 1.0, CUKR_DEFAULT);	                 	   
	   
	/*
	 * Solve the system
	 */   
	       
	/* Create solver and assign operators */
	CUKR_SOLVER solver;
	CukrSolverCreate(&solver, "CG Solver");
	CukrSolverSetOperators(solver, A);
	 
	/* Solve the system */  
	CukrSolverSolve(solver, b, x);
	printf("%s\n", CukrSolverGetConvergedReasonAsStr(solver, CUKR_OUTER));      
	 
	/* Get cond. estimate */
	CUKR_DOUBLE condest = CukrSolverCondest(solver);
	printf("condest = %g\n", condest); 
	     
	/* 
	 * Shutdown 
	 */ 
	
	/* Open output file */
	FILE *f;
	if ( (f = fopen("solver.condest.txt", "w")) == NULL) {
		CUKR_FATAL("Unable to open solver perf. output file\n");
	}
	
	/* Write to file */
    int iters = solver->its; 
    double solver_time = solver->counters[SOLVER_IDX_NORMAL   + SOLVER_OFFSET_TIME];
    double condest_time = solver->counters[SOLVER_IDX_CONDEST + SOLVER_OFFSET_TIME];   	
	printf("Solver spent %d iter. Condest = %g\n", iters, condest);	
	fprintf(f, "%g,%g,%g,%d\n", condest, condest_time, solver_time, iters);
	
	/* Close output file */
	fclose(f);
	     
	/* Destroy solver */ 
	CukrSolverDestroy(&solver);
	    
	/* Destroy vectors and matrices */        
	CukrVecDestroy(&x); 
	CukrVecDestroy(&b);
	CukrMatDestroy(&A);
		         
	/* Shutdown CUKr */ 
	CukrFinalize(CUKR_DEFAULT);


	/* Return */  
	return 0;  
}    
