#include <stdio.h>
#include "../cukr_all.h" 
#include "../db/db.h" 
#include "../util/log.h" 

/* Access the global database*/
extern CukrDB cukr_db;

int main(int argc, char* argv[]) 
{        
	/*  
	 * Init CUKr
	 */         
	               
	/* Init runtime */      
	CukrInit(argc, argv, CUKR_OPTS_STRICT_SYNTAX);	
	                       	     	        
	/* Start runtime */
	CukrStart(CUKR_DEFAULT);
	                       
	/*  
	 * Create system
	 */     
	
	/* Exit if no matrix file was specified */
	const char* matrix_file = CukrDBGetS(CUKR_DB_MATRIX);
	if (!CukrStricmp(matrix_file, "")) {
		CUKR_FATAL("Please privide a matrix filename using "\
		           "the --matrix=MATRIX_FILE syntax");
	}   
	    
	/* Otherwise, create it from file */
	CUKR_MATRIX A;  
	CukrMatCreateFromFile(&A, matrix_file, CUKR_DEFAULT, CUKR_DEFAULT, "A");
	CukrMatAssemble(A);    
	                     
	/* Create initial solution and rhs */
	CUKR_VECTOR x,b;
	CukrVecCreate(&x, A->cols, CUKR_DEFAULT, CUKR_SHTYPE_SHARED, "x");
	CukrVecCreate(&b, A->rows, CUKR_DEFAULT, CUKR_SHTYPE_DISTRIBUTED, "b");
	      
	/* Set x = 0, b = 1 */
	CukrVecSetFromVal(x, 0.0, CUKR_DEFAULT);
	CukrVecSetFromVal(b, 1.0, CUKR_DEFAULT);	                 	   
	   
	/*
	 * Solve the system
	 */   
	       
	/* Create solver and assign operators */
	CUKR_SOLVER solver;
	CukrSolverCreate(&solver, "CG Solver");
	CukrSolverSetOperators(solver, A);
	 
	/* Solve the system */  
	CukrSolverSolve(solver, b, x);
	printf("%s\n", CukrSolverGetConvergedReasonAsStr(solver, CUKR_OUTER));      
	     
	/* 
	 * Shutdown 
	 */ 
	
	/* Open output file */
	FILE *f;
	if ( (f = fopen("solver.blas.perf.txt", "w")) == NULL) {
		CUKR_FATAL("Unable to open solver perf. output file\n");
	}
	
	/* Write time data from counters, per iteration */	
	//fprintf(f, "rows,cols,nz,dot,axpy,aypx,scal,copy,spmv,pwpr\n");	
	fprintf(f, "%d,%d,%d,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g\n",
	           A->rows, A->cols, A->nz, 
	           solver->blas->counters_mp[CUKR_BLAS_IDX_TIME][CUKR_BLAS_OPS_DOT]  / solver->its,
	           solver->blas->counters_mp[CUKR_BLAS_IDX_TIME][CUKR_BLAS_OPS_AXPY] / solver->its,
	           solver->blas->counters_mp[CUKR_BLAS_IDX_TIME][CUKR_BLAS_OPS_AYPX] / solver->its,
	           solver->blas->counters_mp[CUKR_BLAS_IDX_TIME][CUKR_BLAS_OPS_SCAL] / solver->its,
	           solver->blas->counters_mp[CUKR_BLAS_IDX_TIME][CUKR_BLAS_OPS_COPY] / solver->its,
	           solver->blas->counters_mp[CUKR_BLAS_IDX_TIME][CUKR_BLAS_OPS_SPMV_CSR] / solver->its,
	           solver->blas->blasx->counters_mp[CUKR_BLAS_IDX_TIME][CUKR_BLASX_OPS_PWPR] / solver->its,
	           solver->blas->counters[0][CUKR_BLAS_IDX_TIME][CUKR_BLAS_OPS_DOT][CukrTypeToBlasPrec(solver->blas->precision)] / solver->its,
	           solver->blas->counters[0][CUKR_BLAS_IDX_TIME][CUKR_BLAS_OPS_AXPY][CukrTypeToBlasPrec(solver->blas->precision)] / solver->its,
	           solver->blas->counters[0][CUKR_BLAS_IDX_TIME][CUKR_BLAS_OPS_AYPX][CukrTypeToBlasPrec(solver->blas->precision)] / solver->its,
	           solver->blas->counters[0][CUKR_BLAS_IDX_TIME][CUKR_BLAS_OPS_SCAL][CukrTypeToBlasPrec(solver->blas->precision)] / solver->its,
	           solver->blas->counters[0][CUKR_BLAS_IDX_TIME][CUKR_BLAS_OPS_COPY][CukrTypeToBlasPrec(solver->blas->precision)] / solver->its,
	           solver->blas->counters[0][CUKR_BLAS_IDX_TIME][CUKR_BLAS_OPS_SPMV_CSR][CukrTypeToBlasPrec(solver->blas->precision)] / solver->its,
	           solver->blas->blasx->counters[0][CUKR_BLAS_IDX_TIME][CUKR_BLASX_OPS_PWPR][CukrTypeToBlasPrec(solver->blas->precision)] / solver->its,
	           cukr_db.solverTime / solver->its);
			
	/* Close output file */
	fclose(f);
     
	/* Destroy solver */ 
	CukrSolverDestroy(&solver);
	    
	/* Destroy vectors and matrices */        
	CukrVecDestroy(&x); 
	CukrVecDestroy(&b);
	CukrMatDestroy(&A);
		         
	/* Shutdown CUKr */ 
	CukrFinalize(CUKR_DEFAULT);


	/* Return */  
	return 0;  
}    
