/**
 * @file benchmark/bench_spmv.c
 * @author Serban Georgescu
 * @since 2008
 * 
 * BLAS2 benchmark routine, serial version.
 * 
 * Benchmarks the performance and 
 * accuracy for:
 * - SpMV
 * 
 * Benchmarks are done in 
 * 	- double precision
 * 	- quasi-double precision
 * 	- single precision
 * on
 * 	- CPU (all precisions)
 * 	- GPU (single, qdouble)
 * 
 */
 
#define LOGGER_NAME "benchmark.bench_spmv"

/* Standard includes */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

/* Dependency includes */
#include <argtable2.h>


/* CUKr includes */
#include "../mat_vec/sp/matrix_sp.h" 
#include "../blas/sp/blas2_sp.h" 
#include "../cukr_all.h" 

/* Blas includes 
 * @todo Add all blas includes here */
#include "../blas/impl/generic/generic_blas.h"
#include "../blas/impl/mkl/mkl_blas.h"
#include "../blas/impl/gpu/gpu_blas.h" 

/* Access the global database*/
extern CukrDB cukr_db;

/* Constant value for x vector */
#define X_VAL 1.12346

/**
 * \brief Perform quasi-double addition
 * \param c0,c1 The result (headvand tail)
 * \param a0,a1 The first number (head
 *              and tail)
 * \param b0,b1 The second number (head 
 *              and tail)
 */
extern void dsadd(volatile float* c0, volatile float* c1, const volatile float a0, const volatile float a1, 
                                                   const volatile float b0, const volatile float b1);
extern void dssub(volatile float* c0, volatile float* c1, const volatile float a0, const volatile float a1, 
                                                   const volatile float b0, const volatile float b1);
extern void dsmul(volatile float *c0, volatile float *c1, const volatile float a0, const volatile float a1, 
                                                   const volatile float b0, const volatile float b1);

// ----------------------- INITIAL CONFIGURATION---------------------- //

/** Lower and upper bounds and step size
 * for tests. The increment will be done
 * by multiplication with the step size,
 * not by addition */ 
#define TEST_LO   		  	  10*1024		// Lower bound for matrix order
#define TEST_UP   		  	  1*1024*1024	// Upper bound for matrix order
#define TEST_STEP 		  	  1.2			// Matrix order increase step
#define TEST_NZ     		  30			// Number of nonzeros/row
#define TEST_HB     		  30			// Matrix halfbandwidth
#define MAT_FORMAT 			  "CSR"			// Matrix format
#define MAT_FORMAT_VARIATION  "CSR4"		// Matrix format variation

/** No of trials for the test 
 * Remember: more than one trials means
 * the smaller tests are done in-cache */
#define TEST_TRIALS 1  						


// ----------------------------- GLOBALS ----------------------------- //

/**
 * Struct defining the option database
 */
typedef struct
{
	const char* progname;
	struct arg_str *test_out_file;
	struct arg_int *test_trials;
	struct arg_str *mat_file;
	struct arg_str *mat_format;
	struct arg_str *mat_format_variation;
	struct arg_int *test_lo;
	struct arg_int *test_up;
	struct arg_dbl *test_step;
	struct arg_int *test_nz;
	struct arg_int *test_hb;
	struct arg_lit *no_gpu;	
	struct arg_lit *no_cpu_qdouble;
	struct arg_lit *help;
    struct arg_end *end;
	void *argtable[20];
} Bench_OptionDB;

/*
 * Test parameters 
 */
char mat_file[2550], mat_format[255], mat_format_variation[255], mat_name[255], test_out_file[2550];
int test_trials, test_lo, test_up, test_nz, test_hb, test_ref_id, no_gpu, banded_mat, no_cpu_qdouble;
double test_step;

/*
 * Blas parameters
 */
CUKR_BLAS blas_vector[100]; 
int blas_count;

// ------------------------- FORWARD DECLARATIONS ------------------------- //

int mat_halfbw(CUKR_MATRIX_SP csr_mat);

// ------------------------- BENCHMARK FUNCTIONS -------------------------- //

void test_spmv(int n, CUKR_DOUBLE alpha, CUKR_DOUBLE beta, CUKR_BLAS blas, 
               CUKR_MATRIX_SP mat, CUKR_VECTOR_SP x, CUKR_VECTOR_SP y_orig, 
               CUKR_VECTOR_SP y, CUKR_DOUBLE *perf, CUKR_DOUBLE *band)
{
	int t;
	
	/* Start from a copy of the original vector */
	CukrCOPY_SP(blas, y_orig, y); 
	
	/* Warm-up run */
	CukrSPMV_SP(blas, alpha, mat, x, beta, y);
	
	/* Set counters to zero */
	CukrBlasResetCounters(blas);
	
	/* Do a no. of trials */
	for (t = 0; t < test_trials; t++) {
		CukrSPMV_SP(blas, alpha, mat, x, beta, y);
	}
			
	/* Read the thread counters */
	float elapsed = blas->counters[0][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CukrTypeToBlasPrec(blas->precision)];
	float flops   = blas->counters[0][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CukrTypeToBlasPrec(blas->precision)];
	float floads  = blas->counters[0][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CukrTypeToBlasPrec(blas->precision)];
	float fstores = blas->counters[0][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CukrTypeToBlasPrec(blas->precision)];
	float iloads  = blas->counters[0][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CukrTypeToBlasPrec(blas->precision)];
	float istores = blas->counters[0][CUKR_BLAS_IDX_INTSTORES][CUKR_BLAS_OPS_SPMV_CSR][CukrTypeToBlasPrec(blas->precision)];

	/* Compute performance */
	*perf = 1e-3 * flops / elapsed;
	*band = 1e-3 * (CukrBlasWordSize(blas)*(floads + fstores) + 4*(iloads + istores)) / elapsed;			
}


// -------------------------- DRIVER FUNCTIONS ---------------------------- //

/**
 * \brief Creates a vector that contains all
 *        the BLAS types supported by the
 *        CUKr distribution. 
 * 		  If support for other BLASes has been
 *        included, all should be defined, 
 *        created and registered here.
 * \param blas A vector of BLAS objects
 * \param blas_count The total number of
 *        registered objects.
 */ 
void create_blas_vector()
{
	
	blas_count = 0;
	
	/* 
	 * Add the Generic BLAS, all precisions 
	 */
	
	// For double
	CUKR_BLAS blas_generic_d; 
	CukrBlasCreate(&blas_generic_d);
	blas_create_generic(blas_generic_d);
	blas_generic_d->precision = CUKR_TYPE_DOUBLE;  
	
	// For qdouble
	/*
	CUKR_BLAS blas_generic_q;
	CukrBlasCreate(&blas_generic_q);
	blas_create_generic(blas_generic_q); 
	blas_generic_q->precision = CUKR_TYPE_QDOUBLE; 
	         
	// For single 
	CUKR_BLAS blas_generic_s;         
	CukrBlasCreate(&blas_generic_s);          
	blas_create_generic(blas_generic_s);
	blas_generic_s->precision = CUKR_TYPE_SINGLE; 
	*/
	// Register
	//blas_vector[0] = blas_generic_s;
	//blas_count++;
	
	blas_vector[0] = blas_generic_d;
	blas_count++;
	/*
	blas_vector[1] = blas_generic_s;
	blas_count += 2;	
	if (!no_cpu_qdouble) {
		blas_vector[2] = blas_generic_q;
		blas_count++;
	}
	*/
	/* Set reference blas to generic double */
	test_ref_id = 0;
	
	/* 
	 * Add the MKL BLAS, double and single
	 * (no qdouble support in MKL) 
	 */
	
	#ifdef CUKR_USE_MKLs
	
		/* For double */
		CUKR_BLAS blas_MKL_d;
		CukrBlasCreate(&blas_MKL_d);
		blas_create_mkl(blas_MKL_d);
		blas_MKL_d->precision = CUKR_TYPE_DOUBLE;
	
		/* For single */
		/*
		CUKR_BLAS blas_MKL_s;
		CukrBlasCreate(&blas_MKL_s);
		blas_create_mkl(blas_MKL_s);
		blas_MKL_s->precision = CUKR_TYPE_SINGLE;
		*/
	
		/* Register */
		blas_vector[blas_count + 0] = blas_MKL_d;
		//blas_vector[blas_count + 1] = blas_MKL_s;
		blas_count += 1;
		
		/* Set reference blas to MKL double */
		test_ref_id = blas_count - 1;
		
	#endif 
	
	/* 
	 * Add the GPU BLAS, qdouble and single
	 * (no double support included yet) 
	 */
	 
	#ifdef CUKR_USE_GPU
	if (!no_gpu) {
		/*
		// For double
		CUKR_BLAS blas_GPUBLAS_d;
		CukrBlasCreate(&blas_GPUBLAS_d);
		blas_create_gpu(blas_GPUBLAS_d);
		blas_GPUBLAS_d->precision = CUKR_TYPE_DOUBLE;
		
		// For qdouble
		CUKR_BLAS blas_GPUBLAS_q;
		CukrBlasCreate(&blas_GPUBLAS_q);
		blas_create_gpu(blas_GPUBLAS_q);
		blas_GPUBLAS_q->precision = CUKR_TYPE_QDOUBLE;
		*/
		/* For single */
		CUKR_BLAS blas_GPUBLAS_s;
		CukrBlasCreate(&blas_GPUBLAS_s);
		blas_create_gpu(blas_GPUBLAS_s);
		blas_GPUBLAS_s->precision = CUKR_TYPE_SINGLE;
		
		/* Register */
		//blas_vector[blas_count + 0] = blas_GPUBLAS_d;
		//blas_vector[blas_count + 1] = blas_GPUBLAS_q;
		blas_vector[blas_count + 0] = blas_GPUBLAS_s;
		//blas_vector[blas_count + 1] = blas_GPUBLAS_q;
		//blas_vector[blas_count + 2] = blas_GPUBLAS_d;
		blas_count += 1;
		
		/* Set GPU support */
		cukr_db.compute_on_gpu = CUKR_TRUE;
	}
	#endif  
}

// ------------------------------------------------------------------- //

/** 
 * \brief Process cmdline options and arguments
 * 		  and return the option database
 * 
 * \param argc the number of arguments
 * \param argv a vector containing the list of
 *        arguments 
 * \param optiondb a pointer to a OptionDB 
 *        structure, where the database is returned
 *  
 **/
void bench_blas1_process_arguments(int argc, char* argv[], Bench_OptionDB *optiondb)
{
	/*
	 * Declare options
	 */
	 optiondb->progname      			= "bench_spmv";
	 optiondb->test_out_file 			= arg_str0("o", NULL, 
				                               		"<file>", 
				                              		"Result file "\
				                              		"(default: bench_<OP>.result.csv)");                          		
	 optiondb->test_trials   			= arg_int0("t", "trials", 
													"int", 
													"Number of test trials "\
												    "(default: 10)");	
	 optiondb->mat_format 	 			= arg_str0("f", "matrix-format", 
				                               		"string", 
				                              		"Matrix format "\
				                              		"(default: CSR)");     
	 optiondb->mat_format_variation	 	= arg_str0("v", "matrix-format-variation", 
				                               		"string", 
				                              		"Matrix format variation "\
				                              		"(default: CSR4)");                                 										    
	 optiondb->test_lo  	 			= arg_int0("l", "lo", 
													"int", 
													"Start vector size "\
												    "(default: 10*1024)");		
	 optiondb->test_up  	 			= arg_int0("u", "up", 
													"int", 
													"End vector size "\
												    "(default: 1*1024*1024)");	
	 optiondb->test_step  	 			= arg_dbl0("s", "step", 
													"double", 
													"Size increment (by multiplication) "\
												    "(default: 1.2)");	
	 optiondb->test_nz  	 			= arg_int0(NULL, "nz", 
													"int", 
													"Nonzeros/row "\
												    "(default: 30)");	
	 optiondb->test_hb  	 			= arg_int0(NULL, "hb", 
													"int", 
													"Matrix halfbandwidth"\
												    "(default: 30)");	
	 optiondb->mat_file 	 			= arg_str0("m", "matrix-file", 
				                               		"<file>", 
				                              		"Matrix file "\
				                              		"(default: none)");   								    								    							    	
	 optiondb->no_gpu        			= arg_lit0(NULL, "no-gpu", 
									    			"Do not use a GPU even if possible");										    
	 optiondb->no_cpu_qdouble 			= arg_lit0(NULL, "no-cpu-qdouble", 
									    			"Do not use qdouble on the CPU (because it's slow)");									    								    									    											    
		 
	 optiondb->help          			= arg_lit0("hH",  "help", 
									    			"display this help and exit");
	 optiondb->end           			= arg_end(14);								    									    
									    
	 optiondb->argtable[0]  = optiondb->test_out_file;
	 optiondb->argtable[1]  = optiondb->test_trials;
	 optiondb->argtable[2]  = optiondb->mat_file;	
	 optiondb->argtable[3]  = optiondb->mat_format; 
	 optiondb->argtable[4]  = optiondb->mat_format_variation; 
	 optiondb->argtable[5]  = optiondb->test_lo;
	 optiondb->argtable[6]  = optiondb->test_up;
	 optiondb->argtable[7]  = optiondb->test_step;
	 optiondb->argtable[8]  = optiondb->test_nz;
	 optiondb->argtable[9]  = optiondb->test_hb;
	 optiondb->argtable[10] = optiondb->no_gpu;
	 optiondb->argtable[11] = optiondb->no_cpu_qdouble;
	 optiondb->argtable[12] = optiondb->help;
	 optiondb->argtable[13] = optiondb->end;
	 
	 /*
	  * Parse options and arguments
	  */
	  
	 /* Verify the argtable[] entries were allocated sucessfully */	
	 if (arg_nullcheck(optiondb->argtable) != 0)
     {
       /* NULL entries were detected, some allocations must have failed */
       LOG(FATAL, "Argtable initialization failed. \n");
       exit(1);
     }
    
     /* Parse the command line as defined by argtable[] */
     int nerrors = arg_parse(argc, argv, optiondb->argtable); 
    
     /* Special case: '--help' takes precedence over error reporting */
     if (optiondb->help->count > 0)
     {
        printf("\nUsage: bench_spmv [matrix_file] [options]");
        arg_print_syntax(stdout, optiondb->argtable, "\n\n");
        arg_print_glossary(stdout, optiondb->argtable, "  %-25s %s\n");
        exit(1);
     }

     /* If the parser returned any errors then display them and exit */
     if (nerrors > 0)
     {
        /* Display the error details contained in the arg_end struct.*/
        arg_print_errors(stdout, optiondb->end, optiondb->progname);
        printf("Errors. Try '--help' for more information.\n");
        exit(1);
     } 
     	 	                           
	 /*
	  * Extract values and add them to 
	  * global options
	  */	          
	  
	  if (optiondb->mat_file->count) {
	  	  char drive[255], dir[2550], ext[255];
	  	  sprintf(mat_file, "%s", optiondb->mat_file->sval[0]);
		  CukrSplitPath(mat_file, drive, dir, mat_name, ext);
		  banded_mat = CUKR_FALSE;
	  } 
	  else {
	  	banded_mat = CUKR_TRUE;
	  	sprintf(mat_name, "%s", "banded");
	  }
	  	  
	  if (optiondb->test_out_file->count)
	  	  sprintf(test_out_file, "%s", optiondb->test_out_file->sval[0]);  
	  else {
	  	  if (banded_mat)
	  	  	sprintf(test_out_file, "bench_spmv_banded.result.csv");
	  	  else
	  	  	sprintf(test_out_file, "bench_spmv_%s.result.csv", mat_name);
	  } 	
	  
	  if (optiondb->mat_format->count)
	  	  sprintf(mat_format, "%s", optiondb->mat_format->sval[0]); 
	  else
	  	  sprintf(mat_format, MAT_FORMAT);	 
	  	  
	  if (optiondb->mat_format_variation->count)
	  	  sprintf(mat_format_variation, "%s", optiondb->mat_format_variation->sval[0]); 
	  else
	  	  sprintf(mat_format_variation, MAT_FORMAT_VARIATION);	
	  	   	  
	  test_trials = optiondb->test_trials->count ? optiondb->test_trials->ival[0] : TEST_TRIALS;
	  test_lo     = optiondb->test_lo->count     ? optiondb->test_lo->ival[0]     : TEST_LO;
	  test_up     = optiondb->test_up->count     ? optiondb->test_up->ival[0]     : TEST_UP;
	  test_step   = optiondb->test_step->count   ? optiondb->test_step->dval[0]   : TEST_STEP;
	  test_nz     = optiondb->test_nz->count     ? optiondb->test_nz->ival[0]     : TEST_NZ;
	  test_hb     = optiondb->test_hb->count     ? optiondb->test_hb->ival[0]     : TEST_HB;
	  no_gpu      = optiondb->no_gpu->count;
	  	                  
}

// ------------------------------------------------------------------- //

/**
 * \brief Runs the test indicated by the tid using
 *        all BLAS libraries found in the blas vector
 * \param test_func Function pointer of the operation
 *                  being tested
 * \param f Output file
 */
void run(FILE *f)
{	
	int i, n;
	
		
	/* Define vars to hold counters */
	double perf, band, error;
	char buffer[255];
	
	/* Read matrix if that is the case */
	CUKR_MATRIX_SP mat;
	int mat_format_id = CukrMatFormatStrToId(mat_format);
	if (!banded_mat) {
		
		CukrMatspCreateFromFile(&mat, mat_file, mat_format_id, CUKR_TYPE_DOUBLE, "A");
		CukrMatspAssemble(mat);
		test_lo = test_up = mat->rows;
		test_nz = mat->nz;
		test_hb = mat_halfbw(mat);	
		test_step = 10;
	}
		
	/* Run the selected test */	
	for (n = test_lo; n <= test_up; n *= test_step)
	{	
		/* Create banded matrix if that is the case */
		if (banded_mat) {
			CukrMatGenBanded(&mat, n, 1, 15, 0, mat_format_id, CUKR_TYPE_DOUBLE, "A");
			CukrMatspAssemble(mat);
		}
		
		/* Set alpha = 1, beta = 0*/
		CUKR_DOUBLE alpha = 1, beta = 0;
		
		/* Create test vectors, in double */
		CUKR_VECTOR_SP x = NULL, y_orig = NULL, y[blas_count], y_ref;
		CukrVecspCreate(&x,      n, CUKR_TYPE_DOUBLE, "x");
		CukrVecspCreate(&y_orig, n, CUKR_TYPE_DOUBLE, "y");
		
		/* Fill vectors with random values */
		CukrVecspSetFromVal(x,      1.123456789012345, blas_vector[test_ref_id]->precision);
		//CukrVecspSetFromVal(y_orig, 2.2345678912345678901, blas_vector[test_ref_id]->precision);
		//CukrVecspSetFromVal(x,      1.1, blas_vector[test_ref_id]->precision);
		CukrVecspSetFromVal(y_orig, 0, blas_vector[test_ref_id]->precision);
						
		/* Run test with the reference BLAS */
		y_ref = NULL;
		CukrVecspCreate(&y_ref, n, CUKR_TYPE_DOUBLE, "y_ref");
		test_spmv(n, alpha, beta, blas_vector[test_ref_id], mat, x, y_orig, y_ref, &perf, &band);
		CukrVecspTransfer(y_ref, CUKR_LOCATION_CPU, CUKR_TYPE_DOUBLE);
		
		/* Output performance for reference */
		CukrTypeIdToStr(blas_vector[test_ref_id]->precision, buffer);
		printf("%15s %10d  %10d  %10d  %10s   %10s   %11.6g  %11.6g  %d\n", 
		       mat_name, n, test_nz, test_hb, blas_vector[test_ref_id]->name, 
		       buffer, perf, band, 0);
		// this line can be commented out if reference is not to show up in file
		//fprintf(f, "%s,%d,%d,%d,%s,%s,%g,%g,%d\n", mat_name, n, test_nz, 
		//           test_hb, blas_vector[test_ref_id]->name, buffer, perf, band, 0);
		
		/* Run tests for the rest of the BLASes */
		int b;
		for (b = 0; b < blas_count; b++)
			if (b != test_ref_id)
			{
				/* Run test */				
				y[b] = NULL;
				sprintf(buffer, "y[%d]", b);
				
				CukrVecspCreate(&y[b], n, CUKR_TYPE_DOUBLE, "buffer");
				
				test_spmv(n, alpha, beta, blas_vector[b], mat, x, y_orig, y[b], &perf, &band);
				CukrVecspTransfer(y[b], CUKR_LOCATION_CPU, CUKR_TYPE_DOUBLE);
				
				/* Compute error */
				double max_err = 0, err, rel_err;	
				double sum_nrm = 0, diff_nrm = 0;			
				for (i = 0; i < n; i++) {	
					err = fabs(y[b]->ph_dval[i] - y_ref->ph_dval[i]);	
					sum_nrm += pow(y[b]->ph_dval[i] + y_ref->ph_dval[i], 2.0);
					diff_nrm += pow(y[b]->ph_dval[i] - y_ref->ph_dval[i], 2.0);
					
					/* If ref is not zero,
					 * compute the regular way */				  			
					if (err != 0)
						rel_err = err / fabs(y_ref->ph_dval[i] + y[b]->ph_dval[i]) / 2.0;
					/* If error is zero, leave it zero */	
					else
						rel_err = 0;
			
					if (max_err < rel_err) {				
						max_err = rel_err; 
						printf("%d: y=%g, ref=%g => err = %g, %g\n", i, y[b]->ph_dval[i], y_ref->ph_dval[i], err, fabs(y_ref->ph_dval[i] + y[b]->ph_dval[i]));						
					}
				}
				printf("Common digits: %g\n", sqrt(sum_nrm) / (2 * sqrt(n) * sqrt(diff_nrm))); 
				
				/* Output performance */
				CukrTypeIdToStr(blas_vector[b]->precision, buffer);
				printf("%15s %10d  %10d  %10d  %10s   %10s   %11.6g  %11.6g  %11.6g\n", 
				       mat_name, n, test_nz, test_hb, blas_vector[b]->name, buffer, 
					   perf, band, max_err);
				fprintf(f, "%s,%d,%d,%d,%s,%s,%g,%g,%g\n", mat_name, n, test_nz, 
				           test_hb, blas_vector[b]->name, buffer, perf, band, max_err);
				
				/* Destroy result vector */
				CukrVecspDestroy(&y[b]);
			}
		
		/* Destroy the matrix */
		CukrMatspDestroy(&mat);
			
		/* Destroy the rest of the vectors */
		CukrVecspDestroy(&x);
		CukrVecspDestroy(&y_orig); 
		CukrVecspDestroy(&y_ref);
	}
}
// ------------------------------------------------------------------- //

int main(int argc, char* argv[]) 
{
	int l;
	
	//double row[5] = {-180, 860.0000000000559, -250, -180, -250};
	double row[5] = {430.1689189189469, -125, -180, -125, -125};
	double vec[5] = {X_VAL, X_VAL, X_VAL, X_VAL, X_VAL};
	float  row_h[5], row_t[5], vec_h[5], vec_t[5];
	
	for (l = 0; l < 5; l++) {
		row_h[l] = (float)row[l];
		row_t[l] = (float)(row[l] - row_h[l]);
		vec_h[l] = (float)vec[l];
		vec_t[l] = (float)(vec[l] - vec_h[l]);
	}
	
	for (l = 0; l < 5; l++) {
		printf("Row[%d]: double = %g, qdouble = (%g,%g), single = %g\n", l + 1, row[l], row_h[l], row_t[l], row_h[l]);
		printf("X[%d]: double = %g, qdouble = (%g,%g), single = %g\n", l + 1, vec[l], vec_h[l], vec_t[l], vec_h[l]);
	}
	
	double d_sum = 0;
	float s_sum = 0, h_sum = 0, t_sum = 0, c0, c1; 
	for (l = 0; l < 5; l++) {
		d_sum += row[l] * vec[l];
		s_sum += row_h[l] * vec_h[l];
		
		dsmul(&c0, &c1, row_h[l], row_t[l], vec_h[l], vec_t[l]);
		dsadd(&h_sum, &t_sum, h_sum, t_sum, c0, c1);  

	}
	
	
	
	printf("Result in double: %g\n", d_sum);
	printf("Result in qdouble: (%g,%g)\n", h_sum, t_sum);
	printf("Result in single: %g\n", s_sum);
	
	//return;
	
	
	
	/* 
	 * Parse command-line options 
	 */
	
	Bench_OptionDB optiondb; 
	bench_blas1_process_arguments(argc, argv, &optiondb);
	
	printf("\n" \
	       "Test configuration: \n"
	       "====================================\n"\
	       "Matrix: %s\n"\
	       "Matrix format: %s\n"\
	       "Matrix format variation: %s\n"\
	       "Start vector size: %d\n"\
	       "End vector size: %d\n"\
	       "Vector size step: %g\n"\
	       "Nonzeros/row: %d\n"\
	       "Halfbandwidth: %d\n"\
	       "\n"\
	       "Result file: %s\n"\
	       "====================================\n",
	       mat_name, mat_format, mat_format_variation,
	       test_lo, test_up, test_step, test_nz, test_hb, test_out_file);	                    
	
	/*
	 * Create BLAS vector
	 */
	 
	
	int i;
	create_blas_vector();
	
	printf("\n" \
	       "BLAS libraries used: \n"
	       "====================================\n");
	       
	for (i = 0; i < blas_count; i++)
	{
		char type_str[255];
		CukrTypeIdToStr(blas_vector[i]->precision, type_str);
		printf("%s in %s precision\n", 
		        blas_vector[i]->name, type_str);
	}  
	printf("====================================\n");
	
	/*
	 * Init CUKr 
	 */
	 
	CukrInit(0, NULL, CUKR_DEFAULT);
	CukrDBSetS(CUKR_DB_MATRIX_FORMAT, mat_format); 
	CukrDBSetS(CUKR_DB_MATRIX_FORMAT_VAR, mat_format_variation);
	
	#ifdef CUKR_USE_OPENCL
		CukrDBSetS(CUKR_DB_SINGLE_PRECISION_BLAS, "CLBLAS");
		//CukrDBSetS(CUKR_DB_SINGLE_PRECISION_BLAS, "MKL");
	#endif
	
	CukrStart(CUKR_DEFAULT);
	
	/*
	 * Run test 
	 */
	 
	/* Open output file */
	FILE *f;
	if ( (f = fopen(test_out_file, "w")) == NULL) {
		CUKR_FATALX("Could not open file %s", test_out_file);
	} 
	
	/* Write test name and header */
	printf("\n         Mat             N       NZ/ROW      HALF-BAND     BLAS        Precision     MFlops/s      MB/s      Rel.err.    \n");
	fprintf(f, "Mat,N,nz,halfb,BLAS,Precision,MFlops/s,MB/s,Error\n");
	
	/* Run test */
	run(f);
	
	/* 
	 * Shutdown 
	 */
	fclose(f);	
	CukrFinalize(CUKR_DEFAULT); 
	
	
	
}


// ------------------------- PRIVATE FUNCTIONS ------------------------- //

/**
 * \brief Computes the halfbandwidth of a matrix
 *        in CSR format
 * \param mat The CSR matrix
 */
int mat_halfbw(CUKR_MATRIX_SP mat)
{
	CUKR_INT i, j, hb = 0;
	
	CUKR_INT* ptr = mat->csr_mat->ptr->ph_ival;
	CUKR_INT* idx = mat->csr_mat->idx->ph_ival;
	int n = mat->rows;
	
	for (i = 0; i < n; i++) 
		for (j = ptr[i] - 1; j < ptr[i+1] - 1; j++)
			{
				CUKR_INT diff = labs(i - (idx[j] - 1));
				if (diff > hb)
					hb = diff;
			}
	return hb;
}	

