/**
 * @file blas/blas.c
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains:
 *  - BLAS initialization function
 *  - Wrappers for the bindings 
 * 	    - counters included
 * 
 * Solvers and preconditioners use, where possible,
 * instructions from the BLAS. There may be more
 * then one implementation of the BLAS, so for
 * generality each BLAS must implement a set of
 * standard operations.
 *  
 * A BLAS implementing the following interface is
 * in fact a wrapper for an external BLAS (e.g.
 * MKL or CUBLAS).  
 * 
 * Each BLAS must implement the following functions:
 * - dot
 * - axpy & aypx
 * - scal
 * - copy
 * - SpMV
 * 
 * The BLAS is also responsible for counting ops,
 * memory loads and stores and providing performance
 * reports.
 * 	
 */
  
#define LOGGER_NAME "CUKr.blas.blas"  
  
#include <stdio.h>
#include <stdlib.h>
#include "blas.h"
#include "../db/db.h"
#include "../util/macros.h"
#include "../util/log.h"
#ifdef CUKR_MPI
	#include <mpi.h>
#endif	

/* Get access to the database */
extern CukrDB cukr_db;

/* Define a max function */
#define MAX(a,b) (a > b ? a : b)

/**
 * \brief Init a BLAS object 
 * - set all counters to zero
 * - set all bindings to null
 * \param blas Pointer to the BLAS 
 *        structure being created  
 */
void CukrBlasCreate(CUKR_BLAS* blas)  
{
	int i;
		
	/* Alloc memory for the BLAS */
	if ( ((*blas) = (CUKR_BLAS)malloc(sizeof(struct _CUKR_BLAS))) == NULL)
		CUKR_FATAL("Error allocating memory for BLAS object");
	 
	/* Set default location to the CPU */
	(*blas)->location = CUKR_LOCATION_CPU;
	
	/* Set counters to 0 */
	CukrBlasResetCounters(*blas);
		
	/* Set bindings to NULL */
	(*blas)->bind.ddot = NULL;	
	(*blas)->bind.qdot = NULL;
	(*blas)->bind.sdot = NULL;
	
	(*blas)->bind.daxpy = NULL;
	(*blas)->bind.qaxpy = NULL;
	(*blas)->bind.saxpy = NULL;
	
	(*blas)->bind.daypx = NULL;
	(*blas)->bind.qaypx = NULL;
	(*blas)->bind.saypx = NULL;
		
	(*blas)->bind.dcopy = NULL;
	(*blas)->bind.qcopy = NULL;
	(*blas)->bind.scopy = NULL;
	
	(*blas)->bind.dspmv_csr = NULL;
	(*blas)->bind.qspmv_csr = NULL;
	(*blas)->bind.sspmv_csr = NULL;
	
	(*blas)->bind.dspmv_csr4 = NULL;
	(*blas)->bind.qspmv_csr4 = NULL;
	(*blas)->bind.sspmv_csr4 = NULL;

	(*blas)->bind.dspmv_bcsr = NULL;
	(*blas)->bind.qspmv_bcsr = NULL;
	(*blas)->bind.sspmv_bcsr = NULL;
	
	(*blas)->bind.dspmv_bcsr4 = NULL;
	(*blas)->bind.qspmv_bcsr4 = NULL;
	(*blas)->bind.sspmv_bcsr4 = NULL;

	(*blas)->bind.dspmv_hyb = NULL;
	(*blas)->bind.qspmv_hyb = NULL;
	(*blas)->bind.sspmv_hyb = NULL;

	(*blas)->bind.dspmv_bhyb = NULL;
	(*blas)->bind.qspmv_bhyb = NULL;
	(*blas)->bind.sspmv_bhyb = NULL;
	
	(*blas)->bind.dspmv_bhyb4 = NULL;
	(*blas)->bind.qspmv_bhyb4 = NULL;
	(*blas)->bind.sspmv_bhyb4 = NULL;
	
#ifdef CUKR_USE_OPENCL
	(*blas)->bind.cl_ddot = NULL;	
	(*blas)->bind.cl_qdot = NULL;
	(*blas)->bind.cl_sdot = NULL;
	
	(*blas)->bind.cl_daypx = NULL;
	(*blas)->bind.cl_qaypx = NULL;
	(*blas)->bind.cl_saypx = NULL;
	
	(*blas)->bind.cl_daxpy = NULL;
	(*blas)->bind.cl_qaxpy = NULL;
	(*blas)->bind.cl_saxpy = NULL;
	
	(*blas)->bind.cl_dcopy = NULL;
	(*blas)->bind.cl_qcopy = NULL;
	(*blas)->bind.cl_scopy = NULL;
	
	(*blas)->bind.cl_dspmv_csr = NULL;
	(*blas)->bind.cl_qspmv_csr = NULL;
	(*blas)->bind.cl_sspmv_csr = NULL;
	
	(*blas)->bind.cl_dspmv_csr4 = NULL;
	(*blas)->bind.cl_qspmv_csr4 = NULL;
	(*blas)->bind.cl_sspmv_csr4 = NULL;
	
	(*blas)->bind.cl_dspmv_hyb = NULL;
	(*blas)->bind.cl_qspmv_hyb = NULL;
	(*blas)->bind.cl_sspmv_hyb = NULL;	
	
	(*blas)->bind.cl_dspmv_bcsr = NULL;
	(*blas)->bind.cl_qspmv_bcsr = NULL;
	(*blas)->bind.cl_sspmv_bcsr = NULL;
	
	(*blas)->bind.cl_dspmv_bcsr4 = NULL;
	(*blas)->bind.cl_qspmv_bcsr4 = NULL;
	(*blas)->bind.cl_sspmv_bcsr4 = NULL;
#endif

}  	

// ------------------------------------------------------------------- //

/**
 * \brief Reset all the counters to zero
 * \param blas The blas structure 
 */
void CukrBlasResetCounters(CUKR_BLAS blas)
{
	int i, j, k, l;
	
	/* Empty thread-level counters */
	for (i = 0; i < CUKR_MAX_PE; i++)
		for (j = 0; j < CUKR_BLAS_IDX_MAXCOUNT; j++)
			for (k = 0; k < CUKR_BLAS_OPS_MAXCOUNT; k++)
				for (l = 0; l < CUKR_BLAS_PREC_MAXCOUNT; l++)
					blas->counters[i][j][k][l] = 0;
					
	/* Empty node-level and cluster-level counters */
	for (j = 0; j < CUKR_BLAS_IDX_MAXCOUNT; j++)
		for (k = 0; k < CUKR_BLAS_OPS_MAXCOUNT; k++) {
			blas->counters_mp[j][k] = 0;	
			blas->counters_mpi[j][k] = 0;		
		}	
		
	blas->comm_inner = blas->comm_outer = 0;	
}

// ------------------------------------------------------------------- //


/**
 * \brief Merges all node-level counters 
 * \param blas The blas structure 
 */
void CukrBlasMergeNodeCounters(CUKR_BLAS blas)
{
	int i, p; 
	
	#ifdef CUKR_MPI
	
	/* 
	 * Ops and load/stores should be added 
	 */
	 
	 /* Add flops */			 
	 MPI_Allreduce(&(blas->counters_mp [CUKR_BLAS_IDX_FLOPS]), 
	               &(blas->counters_mpi[CUKR_BLAS_IDX_FLOPS]), 
                   CUKR_BLAS_OPS_MAXCOUNT, MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD);
	 
	 /* Add int. ops */
	 MPI_Allreduce(&(blas->counters_mp [CUKR_BLAS_IDX_INTOPS]), 
	               &(blas->counters_mpi[CUKR_BLAS_IDX_INTOPS]), 
                   CUKR_BLAS_OPS_MAXCOUNT, MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD);
	 
	 /* Add fl. loads */
	 MPI_Allreduce(&(blas->counters_mp [CUKR_BLAS_IDX_FLLOADS]), 
	               &(blas->counters_mpi[CUKR_BLAS_IDX_FLLOADS]), 
                   CUKR_BLAS_OPS_MAXCOUNT, MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD);		 
	 
	 /* Add fl. stores */
	 MPI_Allreduce(&(blas->counters_mp [CUKR_BLAS_IDX_FLSTORES]), 
	               &(blas->counters_mpi[CUKR_BLAS_IDX_FLSTORES]), 
                   CUKR_BLAS_OPS_MAXCOUNT, MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD);
	 
	 /* Add int. loads */
	 MPI_Allreduce(&(blas->counters_mp [CUKR_BLAS_IDX_INTLOADS]), 
	               &(blas->counters_mpi[CUKR_BLAS_IDX_INTLOADS]), 
                   CUKR_BLAS_OPS_MAXCOUNT, MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD);
	 
	 /* Add int. stores */
	 MPI_Allreduce(&(blas->counters_mp [CUKR_BLAS_IDX_INTSTORES]), 
	               &(blas->counters_mpi[CUKR_BLAS_IDX_INTSTORES]), 
                   CUKR_BLAS_OPS_MAXCOUNT, MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD);
	
	/* 
	 * Timings should be max-ed
	 */
	 
	 MPI_Allreduce(&(blas->counters_mp [CUKR_BLAS_IDX_TIME]), 
	               &(blas->counters_mpi[CUKR_BLAS_IDX_TIME]), 
                   CUKR_BLAS_OPS_MAXCOUNT, MPI_FLOAT, MPI_MAX, MPI_COMM_WORLD);
     
     #endif              		 
}

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DDOT 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */
CUKR_DOUBLE CukrBlasDDOT(CUKR_BLAS blas, CUKR_INT n, CUKR_DOUBLE* x, CUKR_DOUBLE* y)
{ 
	CUKR_DOUBLE res = 0; 
	
	/* Check for valid binding */
	if (blas->bind.ddot == NULL)
		CUKR_FATALX("BLAS %s does not implement the function DDOT\n", blas->name); 
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	res = (blas->bind.ddot)(n, x, y);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int *cukr_activeThread = pthread_getspecific(cukr_db.thread_key);
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]   [CUKR_BLAS_OPS_DOT][CUKR_BLAS_PREC_DOUBLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]  [CUKR_BLAS_OPS_DOT][CUKR_BLAS_PREC_DOUBLE] += 2*n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS][CUKR_BLAS_OPS_DOT][CUKR_BLAS_PREC_DOUBLE] += 2*n;
	#endif
	                                    
	return res;                              
}

/**
 * \brief Wrapper for QDOT 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
CUKR_DOUBLE CukrBlasQDOT(CUKR_BLAS blas, CUKR_INT n, CUKR_SINGLE* xh, CUKR_SINGLE* xl,
       											     CUKR_SINGLE* yh, CUKR_SINGLE* yl)
{ 
	CUKR_DOUBLE res = 0; 
	
	/* Check for valid binding */
	if (blas->bind.qdot == NULL)
		CUKR_FATALX("BLAS %s does not implement the function QDOT\n", blas->name);
	 
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	res = (blas->bind.qdot)(n, xh, xl, yh, yl);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]   [CUKR_BLAS_OPS_DOT][CUKR_BLAS_PREC_QDOUBLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]  [CUKR_BLAS_OPS_DOT][CUKR_BLAS_PREC_QDOUBLE] += 2*n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS][CUKR_BLAS_OPS_DOT][CUKR_BLAS_PREC_QDOUBLE] += 2*n;
	#endif
	                                    
	return res;                              
}

/**
 * \brief Wrapper for SDOT 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */
CUKR_SINGLE CukrBlasSDOT(CUKR_BLAS blas, CUKR_INT n, CUKR_SINGLE* x, CUKR_SINGLE* y)
{ 
	CUKR_SINGLE res = 0; 
	
	/* Check for valid binding */
	if (blas->bind.sdot == NULL)
		CUKR_FATALX("BLAS %s does not implement the function SDOT\n", blas->name);
	 
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	res = (blas->bind.sdot)(n, x, y);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]   [CUKR_BLAS_OPS_DOT][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]  [CUKR_BLAS_OPS_DOT][CUKR_BLAS_PREC_SINGLE] += 2*n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS][CUKR_BLAS_OPS_DOT][CUKR_BLAS_PREC_SINGLE] += 2*n;
	#endif
	                                    
	return res;                              
}

#ifdef CUKR_USE_OPENCL
/**
 * \brief Wrapper for DDOT 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */
CUKR_DOUBLE CukrCLBlasDDOT(CUKR_BLAS blas, CUKR_INT n, CUKR_CL_DOUBLE* x, CUKR_CL_DOUBLE* y)
{ 
	CUKR_DOUBLE res = 0; 
	
	/* Check for valid binding */
	if (blas->bind.cl_ddot == NULL)
		CUKR_FATALX("BLAS %s does not implement the function DDOT\n", blas->name); 
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	res = (blas->bind.cl_ddot)(n, x, y);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int *cukr_activeThread = pthread_getspecific(cukr_db.thread_key);
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]   [CUKR_BLAS_OPS_DOT][CUKR_BLAS_PREC_DOUBLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]  [CUKR_BLAS_OPS_DOT][CUKR_BLAS_PREC_DOUBLE] += 2*n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS][CUKR_BLAS_OPS_DOT][CUKR_BLAS_PREC_DOUBLE] += 2*n;
#endif
	
	return res;                              
}

/**
 * \brief Wrapper for QDOT 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
CUKR_DOUBLE CukrCLBlasQDOT(CUKR_BLAS blas, CUKR_INT n, CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xl,
													   CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yl)
{ 
	CUKR_DOUBLE res = 0; 
	
	/* Check for valid binding */
	if (blas->bind.cl_qdot == NULL)
		CUKR_FATALX("BLAS %s does not implement the function QDOT\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	res = (blas->bind.cl_qdot)(n, xh, xl, yh, yl);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]   [CUKR_BLAS_OPS_DOT][CUKR_BLAS_PREC_QDOUBLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]  [CUKR_BLAS_OPS_DOT][CUKR_BLAS_PREC_QDOUBLE] += 2*n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS][CUKR_BLAS_OPS_DOT][CUKR_BLAS_PREC_QDOUBLE] += 2*n;
#endif
	
	return res;                              
}

/**
 * \brief Wrapper for SDOT 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */
CUKR_SINGLE CukrCLBlasSDOT(CUKR_BLAS blas, CUKR_INT n, CUKR_CL_SINGLE* x, CUKR_CL_SINGLE* y)
{ 
	CUKR_SINGLE res = 0; 
	
	/* Check for valid binding */
	if (blas->bind.cl_sdot == NULL)
		CUKR_FATALX("BLAS %s does not implement the function SDOT\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	res = (blas->bind.cl_sdot)(n, x, y);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]   [CUKR_BLAS_OPS_DOT][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]  [CUKR_BLAS_OPS_DOT][CUKR_BLAS_PREC_SINGLE] += 2*n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS][CUKR_BLAS_OPS_DOT][CUKR_BLAS_PREC_SINGLE] += 2*n;
#endif
	
	return res;                              
}
#endif // CUKR_USE_OPENCL

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DAXPY 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */
void CukrBlasDAXPY(CUKR_BLAS blas, CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x, CUKR_DOUBLE* y)
{ 
	/* Check for valid binding */
	if (blas->bind.daxpy == NULL)
		CUKR_FATALX("BLAS %s does not implement the function DAXPY\n", blas->name);
	  
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blas->bind.daxpy)(n, a, x, y);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_DOUBLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_DOUBLE] += 2*n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_DOUBLE] += 2*n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_DOUBLE] += n;
	#endif                        
}

/**
 * \brief Wrapper for DAXPY 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
void CukrBlasQAXPY(CUKR_BLAS blas, CUKR_INT n, CUKR_DOUBLE a, CUKR_SINGLE* xh, CUKR_SINGLE* xl,
															  CUKR_SINGLE* yh, CUKR_SINGLE* yl)
{ 
	
	/* Check for valid binding */
	if (blas->bind.qaxpy == NULL)
		CUKR_FATALX("BLAS %s does not implement the function QAXPY\n", blas->name);
	  
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blas->bind.qaxpy)(n, a, xh, xl, yh, yl);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_QDOUBLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_QDOUBLE] += 2*n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_QDOUBLE] += 2*n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_QDOUBLE] += n;
    #endif                                 
}

/**
 * \brief Wrapper for SAXPY 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */

void CukrBlasSAXPY(CUKR_BLAS blas, CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x, CUKR_SINGLE* y) 
{ 
	
	/* Check for valid binding */
	if (blas->bind.saxpy == NULL)
		CUKR_FATALX("BLAS %s does not implement the function SAXPY\n", blas->name);
	 
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blas->bind.saxpy)(n, a, x, y);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_SINGLE] += 2*n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_SINGLE] += 2*n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_SINGLE] += n;
    #endif                          
}

#ifdef CUKR_USE_OPENCL
/**
 * \brief Wrapper for DAXPY 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */
void CukrCLBlasDAXPY(CUKR_BLAS blas, CUKR_INT n, CUKR_DOUBLE a, CUKR_CL_DOUBLE* x, CUKR_CL_DOUBLE* y)
{ 
	/* Check for valid binding */
	if (blas->bind.cl_daxpy == NULL)
		CUKR_FATALX("BLAS %s does not implement the function DAXPY\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	(blas->bind.cl_daxpy)(n, a, x, y);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_DOUBLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_DOUBLE] += 2*n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_DOUBLE] += 2*n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_DOUBLE] += n;
#endif                        
}

/**
 * \brief Wrapper for DAXPY 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
void CukrCLBlasQAXPY(CUKR_BLAS blas, CUKR_INT n, CUKR_DOUBLE a, CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xl,
																CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yl)
{ 
	
	/* Check for valid binding */
	if (blas->bind.cl_qaxpy == NULL)
		CUKR_FATALX("BLAS %s does not implement the function QAXPY\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	(blas->bind.cl_qaxpy)(n, a, xh, xl, yh, yl);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_QDOUBLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_QDOUBLE] += 2*n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_QDOUBLE] += 2*n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_QDOUBLE] += n;
#endif                                
}

/**
 * \brief Wrapper for SAXPY 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */

void CukrCLBlasSAXPY(CUKR_BLAS blas, CUKR_INT n, CUKR_SINGLE a, CUKR_CL_SINGLE* x, CUKR_CL_SINGLE* y) 
{ 
	
	/* Check for valid binding */
	if (blas->bind.cl_saxpy == NULL)
		CUKR_FATALX("BLAS %s does not implement the function SAXPY\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	(blas->bind.cl_saxpy)(n, a, x, y);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_SINGLE] += 2*n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_SINGLE] += 2*n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_AXPY][CUKR_BLAS_PREC_SINGLE] += n;
#endif                          
}
#endif //CUKR_USE_OPENCL

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DAYPX 
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param x,y The vectors 
 */
void CukrBlasDAYPX(CUKR_BLAS blas, CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x, CUKR_DOUBLE* y)
{ 

	/* Check for valid binding */
	if (blas->bind.daypx == NULL)
		CUKR_FATALX("BLAS %s does not implement the function DAYPX\n", blas->name);
	  
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blas->bind.daypx)(n, a, x, y);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_DOUBLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_DOUBLE] += 2*n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_DOUBLE] += 2*n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_DOUBLE] += n;
    #endif		                      
}

/**
 * \brief Wrapper for QAYPX 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
void CukrBlasQAYPX(CUKR_BLAS blas, CUKR_INT n, CUKR_DOUBLE a, CUKR_SINGLE* xh, CUKR_SINGLE* xl,
															  CUKR_SINGLE* yh, CUKR_SINGLE* yl)
{ 
	/* Check for valid binding */
	if (blas->bind.qaypx == NULL)
		CUKR_FATALX("BLAS %s does not implement the function QAYPX\n", blas->name);
	  
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blas->bind.qaypx)(n, a, xh, xl, yh, yl);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_QDOUBLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_QDOUBLE] += 2*n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_QDOUBLE] += 2*n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_QDOUBLE] += n;
    #endif                                           
}

/**
 * \brief Wrapper for SAYPX 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */
void CukrBlasSAYPX(CUKR_BLAS blas, CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x, CUKR_SINGLE* y)
{ 

	/* Check for valid binding */
	if (blas->bind.saypx == NULL)
		CUKR_FATALX("BLAS %s does not implement the function SAYPX\n", blas->name);
	  
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blas->bind.saypx)(n, a, x, y);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_SINGLE] += 2*n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_SINGLE] += 2*n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_SINGLE] += n;
    #endif                                      
}

#ifdef CUKR_USE_OPENCL
/**
 * \brief Wrapper for DAYPX 
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param x,y The vectors 
 */
void CukrCLBlasDAYPX(CUKR_BLAS blas, CUKR_INT n, CUKR_DOUBLE a, CUKR_CL_DOUBLE* x, CUKR_CL_DOUBLE* y)
{ 
	
	/* Check for valid binding */
	if (blas->bind.cl_daypx == NULL)
		CUKR_FATALX("BLAS %s does not implement the function DAYPX\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	(blas->bind.cl_daypx)(n, a, x, y);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_DOUBLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_DOUBLE] += 2*n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_DOUBLE] += 2*n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_DOUBLE] += n;
#endif		                      
}

/**
 * \brief Wrapper for QAYPX 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
void CukrCLBlasQAYPX(CUKR_BLAS blas, CUKR_INT n, CUKR_DOUBLE a, CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xl,
				   CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yl)
{ 
	/* Check for valid binding */
	if (blas->bind.cl_qaypx == NULL)
		CUKR_FATALX("BLAS %s does not implement the function QAYPX\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	(blas->bind.cl_qaypx)(n, a, xh, xl, yh, yl);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_QDOUBLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_QDOUBLE] += 2*n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_QDOUBLE] += 2*n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_QDOUBLE] += n;
#endif                                           
}

/**
 * \brief Wrapper for SAYPX 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */
void CukrCLBlasSAYPX(CUKR_BLAS blas, CUKR_INT n, CUKR_SINGLE a, CUKR_CL_SINGLE* x, CUKR_CL_SINGLE* y)
{ 
	
	/* Check for valid binding */
	if (blas->bind.cl_saypx == NULL)
		CUKR_FATALX("BLAS %s does not implement the function SAYPX\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	(blas->bind.cl_saypx)(n, a, x, y);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_SINGLE] += 2*n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_SINGLE] += 2*n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_AYPX][CUKR_BLAS_PREC_SINGLE] += n;
#endif                                      
}
#endif //CUKR_USE_OPENCL


// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSCAL
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param a The scaling factor
 * \param x The vector 
 */
void CukrBlasDSCAL(CUKR_BLAS blas, CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x)
{ 
	/* Check for valid binding */
	if (blas->bind.dscal == NULL)
		CUKR_FATALX("BLAS %s does not implement the function DSCAL\n", blas->name);
	  
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blas->bind.dscal)(n, a, x);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_DOUBLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_DOUBLE] += n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_DOUBLE] += n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_DOUBLE] += n;
    #endif                                   
}

/**
 * \brief Wrapper for QSCAL
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param a The scaling factor
 * \param xh,xt The head and tail of the
 *              vector 
 */
void CukrBlasQSCAL(CUKR_BLAS blas, CUKR_INT n, CUKR_DOUBLE a, CUKR_SINGLE* xh, CUKR_SINGLE* xt)
{ 
	
	/* Check for valid binding */
	if (blas->bind.qscal == NULL)
		CUKR_FATALX("BLAS %s does not implement the function QSCAL\n", blas->name);
	  
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blas->bind.qscal)(n, a, xh, xt);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_QDOUBLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_QDOUBLE] += n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_QDOUBLE] += n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_QDOUBLE] += n;
    #endif	                                        
}

/**
 * \brief Wrapper for SAYPX 
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param a The scaling factor
 * \param x The vector 
 */
void CukrBlasSSCAL(CUKR_BLAS blas, CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x)
{ 
	
	/* Check for valid binding */
	if (blas->bind.sscal == NULL)
		CUKR_FATALX("BLAS %s does not implement the function SSCAL\n", blas->name);
	  
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blas->bind.sscal)(n, a, x);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_SINGLE] += n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_SINGLE] += n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_SINGLE] += n;
    #endif                                      
}

#ifdef CUKR_USE_OPENCL
/**
 * \brief Wrapper for DSCAL
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param a The scaling factor
 * \param x The vector 
 */
void CukrCLBlasDSCAL(CUKR_BLAS blas, CUKR_INT n, CUKR_DOUBLE a, CUKR_CL_DOUBLE* x)
{ 
	/* Check for valid binding */
	if (blas->bind.cl_dscal == NULL)
		CUKR_FATALX("BLAS %s does not implement the function DSCAL\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	(blas->bind.cl_dscal)(n, a, x);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_DOUBLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_DOUBLE] += n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_DOUBLE] += n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_DOUBLE] += n;
#endif                                   
}

/**
 * \brief Wrapper for QSCAL
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param a The scaling factor
 * \param xh,xt The head and tail of the
 *              vector 
 */
void CukrCLBlasQSCAL(CUKR_BLAS blas, CUKR_INT n, CUKR_DOUBLE a, CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt)
{ 
	
	/* Check for valid binding */
	if (blas->bind.cl_qscal == NULL)
		CUKR_FATALX("BLAS %s does not implement the function QSCAL\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	(blas->bind.cl_qscal)(n, a, xh, xt);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_QDOUBLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_QDOUBLE] += n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_QDOUBLE] += n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_QDOUBLE] += n;
#endif	                                        
}

/**
 * \brief Wrapper for SAYPX 
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param a The scaling factor
 * \param x The vector 
 */
void CukrCLBlasSSCAL(CUKR_BLAS blas, CUKR_INT n, CUKR_SINGLE a, CUKR_CL_SINGLE* x)
{ 
	
	/* Check for valid binding */
	if (blas->bind.cl_sscal == NULL)
		CUKR_FATALX("BLAS %s does not implement the function SSCAL\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	(blas->bind.cl_sscal)(n, a, x);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_SINGLE] += n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_SINGLE] += n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_SCAL][CUKR_BLAS_PREC_SINGLE] += n;
#endif                                      
}

#endif // CUKR_USE_OPENCL


// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DCOPY 
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param x The source vector
 * \param y The destination vector 
 */
void CukrBlasDCOPY(CUKR_BLAS blas, CUKR_INT n, CUKR_DOUBLE* x, CUKR_DOUBLE* y)
{ 
	
	/* Check for valid binding */
	if (blas->bind.dcopy == NULL)
		CUKR_FATALX("BLAS %s does not implement the function DCOPY\n", blas->name);
	  
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blas->bind.dcopy)(n, x, y);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_COPY][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_COPY][CUKR_BLAS_PREC_SINGLE] += n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_COPY][CUKR_BLAS_PREC_SINGLE] += n;
    #endif	                                 
}

/**
 * \brief Wrapper for QCOPY 
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
void CukrBlasQCOPY(CUKR_BLAS blas, CUKR_INT n, CUKR_SINGLE* xh, CUKR_SINGLE* xt,
											   CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{ 
	/* Check for valid binding */
	if (blas->bind.qcopy == NULL)
		CUKR_FATALX("BLAS %s does not implement the function QCOPY\n", blas->name);
	  
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blas->bind.qcopy)(n, xh, xt, yh, yt); 
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_COPY][CUKR_BLAS_PREC_QDOUBLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_COPY][CUKR_BLAS_PREC_QDOUBLE] += n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_COPY][CUKR_BLAS_PREC_QDOUBLE] += n;
    #endif	                                     
}

/**
 * \brief Wrapper for SCOPY 
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param x The source vector
 * \param y The destination vector 
 */
void CukrBlasSCOPY(CUKR_BLAS blas, CUKR_INT n, CUKR_SINGLE* x, CUKR_SINGLE* y)
{ 
	/* Check for valid binding */
	if (blas->bind.scopy == NULL)
		CUKR_FATALX("BLAS %s does not implement the function SCOPY\n", blas->name);
	  
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blas->bind.scopy)(n, x, y);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_COPY][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_COPY][CUKR_BLAS_PREC_SINGLE] += n;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_COPY][CUKR_BLAS_PREC_SINGLE] += n;
    #endif                                 
}

#ifdef CUKR_USE_OPENCL
/**
 * \brief Wrapper for DCOPY 
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param x The source vector
 * \param y The destination vector 
 */
void CukrCLBlasDCOPY(CUKR_BLAS blas, CUKR_INT n, CUKR_CL_DOUBLE* x, CUKR_CL_DOUBLE* y)
{ 
	
	/* Check for valid binding */
	if (blas->bind.cl_dcopy == NULL)
		CUKR_FATALX("BLAS %s does not implement the function DCOPY\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	(blas->bind.cl_dcopy)(n, x, y);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_COPY][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_COPY][CUKR_BLAS_PREC_SINGLE] += n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_COPY][CUKR_BLAS_PREC_SINGLE] += n;
#endif	                                 
}

/**
 * \brief Wrapper for QCOPY 
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
void CukrCLBlasQCOPY(CUKR_BLAS blas, CUKR_INT n, CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt, CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt)
{ 
	/* Check for valid binding */
	if (blas->bind.cl_qcopy == NULL)
		CUKR_FATALX("BLAS %s does not implement the function QCOPY\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	(blas->bind.cl_qcopy)(n, xh, xt, yh, yt); 
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_COPY][CUKR_BLAS_PREC_QDOUBLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_COPY][CUKR_BLAS_PREC_QDOUBLE] += n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_COPY][CUKR_BLAS_PREC_QDOUBLE] += n;
#endif	                                     
}

/**
 * \brief Wrapper for SCOPY 
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param x The source vector
 * \param y The destination vector 
 */
void CukrCLBlasSCOPY(CUKR_BLAS blas, CUKR_INT n, CUKR_CL_SINGLE* x, CUKR_CL_SINGLE* y)
{ 
	/* Check for valid binding */
	if (blas->bind.cl_scopy == NULL)
		CUKR_FATALX("BLAS %s does not implement the function SCOPY\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	(blas->bind.cl_scopy)(n, x, y);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLAS_OPS_COPY][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLAS_OPS_COPY][CUKR_BLAS_PREC_SINGLE] += n;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLAS_OPS_COPY][CUKR_BLAS_PREC_SINGLE] += n;
#endif                                 
}
#endif // CUKR_USE_OPENCL

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param a Constant
 * \param y Result vector
 */
void CukrBlasDSPMV_CSR(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, 
                       CUKR_INT* ptr, CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x, 
                       CUKR_DOUBLE beta, CUKR_DOUBLE* y)
{  
	/* Check for valid binding */
	if (blas->bind.dspmv_csr == NULL)
		CUKR_FATALX("BLAS %s does not implement the function DSPMV_CSR\n", blas->name);
	  
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blas->bind.dspmv_csr)(rows, cols, nz, alpha, ptr, idx, val, x, beta, y);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += nz + cols; // from val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += cols;      // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += nz + rows; // from idx and ptr
    #endif	
}

/**
 * \brief Wrapper for QSPMV_CSR 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param beta Scale factor for y
 * \param yh,xt Head and tail for the result
 *              vector
 */
void CukrBlasQSPMV_CSR(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, 
                       CUKR_INT* ptr, CUKR_INT* idx, CUKR_SINGLE* valh, CUKR_SINGLE* valt, 
                       CUKR_SINGLE* xh,  CUKR_SINGLE* xt, CUKR_DOUBLE beta, 
                       CUKR_SINGLE* yh,  CUKR_SINGLE* yt)
{  	
	/* Check for valid binding */
	if (blas->bind.qspmv_csr == NULL)
		CUKR_FATALX("BLAS %s does not implement the function QSPMV_CSR\n", blas->name);
	  
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blas->bind.qspmv_csr)(rows, cols, nz, alpha, ptr, idx, valh, valt, xh, xt, beta, yh, yt);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += nz + cols; // from val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += cols;      // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += nz + rows; // from idx and ptr
    #endif                                                      
}

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param a Constant
 * \param y Result vector
 */
void CukrBlasSSPMV_CSR(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, 
			    	   CUKR_INT* ptr, CUKR_INT* idx, CUKR_SINGLE* val, CUKR_SINGLE* x, 
			    	   CUKR_SINGLE beta, CUKR_SINGLE* y)
{  
	
	/* Check for valid binding */
	if (blas->bind.sspmv_csr == NULL)
		CUKR_FATALX("BLAS %s does not implement the function SSPMV_CSR\n", blas->name);
	  
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blas->bind.sspmv_csr)(rows, cols, nz, alpha, ptr, idx, val, x, beta, y);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + cols; // from val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += cols;      // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + rows; // from idx and ptr
    #endif                                                  
}

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR4 format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param a Constant
 * \param y Result vector
 */
void CukrBlasDSPMV_CSR4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, 
                        CUKR_INT* ptr, CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x, 
                        CUKR_DOUBLE beta, CUKR_DOUBLE* y)
{  	
	/* Check for valid binding */
	if (blas->bind.dspmv_csr4 == NULL)
		CUKR_FATALX("BLAS %s does not implement the function DSPMV_CSR4\n", blas->name);
	  
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blas->bind.dspmv_csr4)(rows, cols, nz, alpha, ptr, idx, val, x, beta, y);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += nz + cols; // from val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += cols;      // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += nz + rows; // from idx and ptr
    #endif	                                                    
}

/**
 * \brief Wrapper for QSPMV_CSR4 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param beta Scale factor for y
 * \param yh,xt Head and tail for the result
 *              vector
 */
void CukrBlasQSPMV_CSR4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, 
                        CUKR_INT* ptr, CUKR_INT* idx, CUKR_SINGLE* valh, CUKR_SINGLE* valt, 
                        CUKR_SINGLE* xh,  CUKR_SINGLE* xt, CUKR_DOUBLE beta, 
                        CUKR_SINGLE* yh,  CUKR_SINGLE* yt)
{  
	
	/* Check for valid binding */
	if (blas->bind.qspmv_csr4 == NULL)
		CUKR_FATALX("BLAS %s does not implement the function QSPMV_CSR4\n", blas->name);
	  
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blas->bind.qspmv_csr4)(rows, cols, nz, alpha, ptr, idx, valh, valt, xh, xt, beta, yh, yt);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += nz + cols; // from val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += cols;      // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += nz + rows; // from idx and ptr
    #endif                                                    
}

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR4 format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param a Constant
 * \param y Result vector
 */
void CukrBlasSSPMV_CSR4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, 
			    	    CUKR_INT* ptr, CUKR_INT* idx, CUKR_SINGLE* val, CUKR_SINGLE* x, 
			    	    CUKR_SINGLE beta, CUKR_SINGLE* y)
{  
	
	/* Check for valid binding */
	if (blas->bind.sspmv_csr4 == NULL)
		CUKR_FATALX("BLAS %s does not implement the function SSPMV_CSR4\n", blas->name);
	  
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blas->bind.sspmv_csr4)(rows, cols, nz, alpha, ptr, idx, val, x, beta, y);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + cols; // from val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += cols;      // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + rows; // from idx and ptr
	#endif		                                                   
}

#ifdef CUKR_USE_OPENCL

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param a Constant
 * \param y Result vector
 */
void CukrCLBlasDSPMV_CSR(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, 
                       CUKR_CL_INT* ptr, CUKR_CL_INT* idx, CUKR_CL_DOUBLE* val, CUKR_CL_DOUBLE* x, 
                       CUKR_DOUBLE beta, CUKR_CL_DOUBLE* y)
{  
	/* Check for valid binding */
	if (blas->bind.cl_dspmv_csr == NULL)
		CUKR_FATALX("BLAS %s does not implement the function DSPMV_CSR\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	(blas->bind.cl_dspmv_csr)(rows, cols, nz, alpha, ptr, idx, val, x, beta, y);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += 2*nz;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += nz + cols; // from val and x
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += cols;      // from y;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += nz + rows; // from idx and ptr
#endif	
}

/**
 * \brief Wrapper for QSPMV_CSR 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param beta Scale factor for y
 * \param yh,xt Head and tail for the result
 *              vector
 */
void CukrCLBlasQSPMV_CSR(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, 
                       CUKR_CL_INT* ptr, CUKR_CL_INT* idx, CUKR_CL_SINGLE* valh, CUKR_CL_SINGLE* valt, 
                       CUKR_CL_SINGLE* xh,  CUKR_CL_SINGLE* xt, CUKR_DOUBLE beta, 
                       CUKR_CL_SINGLE* yh,  CUKR_CL_SINGLE* yt)
{  	
	/* Check for valid binding */
	if (blas->bind.cl_qspmv_csr == NULL)
		CUKR_FATALX("BLAS %s does not implement the function QSPMV_CSR\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	(blas->bind.cl_qspmv_csr)(rows, cols, nz, alpha, ptr, idx, valh, valt, xh, xt, beta, yh, yt);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += 2*nz;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += nz + cols; // from val and x
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += cols;      // from y;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += nz + rows; // from idx and ptr
#endif                                                      
}

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param a Constant
 * \param y Result vector
 */
void CukrCLBlasSSPMV_CSR(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, 
			    	   CUKR_CL_INT* ptr, CUKR_CL_INT* idx, CUKR_CL_SINGLE* val, CUKR_CL_SINGLE* x, 
			    	   CUKR_SINGLE beta, CUKR_CL_SINGLE* y)
{  
	
	/* Check for valid binding */
	if (blas->bind.cl_sspmv_csr == NULL)
		CUKR_FATALX("BLAS %s does not implement the function SSPMV_CSR\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	(blas->bind.cl_sspmv_csr)(rows, cols, nz, alpha, ptr, idx, val, x, beta, y);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += 2*nz;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + cols; // from val and x
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += cols;      // from y;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + rows; // from idx and ptr
#endif                                                  
}

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR4 format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param a Constant
 * \param y Result vector
 */
void CukrCLBlasDSPMV_CSR4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, 
                        CUKR_CL_INT* ptr, CUKR_CL_INT* idx, CUKR_CL_DOUBLE* val, CUKR_CL_DOUBLE* x, 
                        CUKR_DOUBLE beta, CUKR_CL_DOUBLE* y)
{  	
	/* Check for valid binding */
	if (blas->bind.cl_dspmv_csr4 == NULL)
		CUKR_FATALX("BLAS %s does not implement the function DSPMV_CSR4\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	(blas->bind.cl_dspmv_csr4)(rows, cols, nz, alpha, ptr, idx, val, x, beta, y);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += 2*nz;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += nz + cols; // from val and x
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += cols;      // from y;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += nz + rows; // from idx and ptr
#endif	                                                    
}

/**
 * \brief Wrapper for QSPMV_CSR4 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param beta Scale factor for y
 * \param yh,xt Head and tail for the result
 *              vector
 */
void CukrCLBlasQSPMV_CSR4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, 
                        CUKR_CL_INT* ptr, CUKR_CL_INT* idx, CUKR_CL_SINGLE* valh, CUKR_CL_SINGLE* valt, 
                        CUKR_CL_SINGLE* xh,  CUKR_CL_SINGLE* xt, CUKR_DOUBLE beta, 
                        CUKR_CL_SINGLE* yh,  CUKR_CL_SINGLE* yt)
{  
	
	/* Check for valid binding */
	if (blas->bind.cl_qspmv_csr4 == NULL)
		CUKR_FATALX("BLAS %s does not implement the function QSPMV_CSR4\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	(blas->bind.cl_qspmv_csr4)(rows, cols, nz, alpha, ptr, idx, valh, valt, xh, xt, beta, yh, yt);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += 2*nz;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += nz + cols; // from val and x
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += cols;      // from y;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_QDOUBLE] += nz + rows; // from idx and ptr
#endif                                                    
}

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR4 format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param a Constant
 * \param y Result vector
 */
void CukrCLBlasSSPMV_CSR4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, 
			    	    CUKR_CL_INT* ptr, CUKR_CL_INT* idx, CUKR_CL_SINGLE* val, CUKR_CL_SINGLE* x, 
			    	    CUKR_SINGLE beta, CUKR_CL_SINGLE* y)
{  
	
	/* Check for valid binding */
	if (blas->bind.cl_sspmv_csr4 == NULL)
		CUKR_FATALX("BLAS %s does not implement the function SSPMV_CSR4\n", blas->name);
	
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
#endif	
	
	/* Work */
	(blas->bind.cl_sspmv_csr4)(rows, cols, nz, alpha, ptr, idx, val, x, beta, y);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += 2*nz;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + cols; // from val and x
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += cols;      // from y;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + rows; // from idx and ptr
#endif		                                                   
}

/**
 * \brief Wrapper for DSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr_hi/lo CSR ptr vectors
 * \param csr_idx CSR idx vectors
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrCLBlasDSPMV_HYB(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                       CUKR_CL_INT* ell_idx, CUKR_CL_DOUBLE* ell_val, CUKR_CL_INT* csr_ptr, CUKR_CL_INT* csr_idx, CUKR_CL_DOUBLE* csr_val,
                       CUKR_CL_DOUBLE* x, CUKR_DOUBLE beta, CUKR_CL_DOUBLE* y)
{
    /* Check for valid binding */
    if (blas->bind.cl_dspmv_hyb == NULL)
        CUKR_FATALX("BLAS %s does not implement the function DSPMV_HYB\n", blas->name);
	
    /*
     * Do the actual work
     */
	
    /* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone;
	CLOCK_START;
#endif
	
    /* Work */
    (blas->bind.cl_dspmv_hyb)(rows, cols, nz, alpha, ell_nz_row, ell_stride, csr_nz,
    					   ell_idx, ell_val, csr_ptr, csr_idx, csr_val, x, beta, y);
	
    /* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += 2*nz;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += ell_nz_row*ell_stride + csr_nz + rows; // from ell_val and csr_val and x
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += cols;                  // from y;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += ell_nz_row*ell_stride  // from ell_idx
																																	+ rows + csr_nz;       // from csr_ptr, csr_idx
#endif
}

/**
 * \brief Wrapper for QSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
void CukrCLBlasQSPMV_HYB(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
	                   CUKR_CL_INT* ell_idx, CUKR_CL_SINGLE* ell_valh, CUKR_CL_SINGLE* ell_valt, CUKR_CL_INT* csr_ptr, CUKR_CL_INT* csr_idx, CUKR_CL_SINGLE* csr_valh, CUKR_CL_SINGLE* csr_valt,
	                   CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt, CUKR_DOUBLE beta, CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt)
{
	/* Check for valid binding */
    if (blas->bind.cl_qspmv_hyb == NULL)
        CUKR_FATALX("BLAS %s does not implement the function QSPMV_HYB\n", blas->name);
	
    /*
     * Do the actual work
     */
	
    /* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone;
	CLOCK_START;
#endif
	
    /* Work */
    (blas->bind.cl_qspmv_hyb)(rows, cols, nz, alpha, ell_nz_row, ell_stride, csr_nz,
    					   ell_idx, ell_valh, ell_valt, csr_ptr, csr_idx, csr_valh, csr_valt,
    					   xh, xt, beta, yh, yt);
	
    /* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += 2*nz;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += ell_nz_row*ell_stride + csr_nz + rows; // from ell_val and csr_val and x
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += cols;                  // from y;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += ell_nz_row*ell_stride  // from ell_idx
																																	+ rows + csr_nz;       // from csr_ptr, csr_idx
#endif
}

/**
 * \brief Wrapper for SSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vecto
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vectors
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrCLBlasSSPMV_HYB(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                       CUKR_CL_INT* ell_idx, CUKR_CL_SINGLE* ell_val, CUKR_CL_INT* csr_ptr, CUKR_CL_INT* csr_idx, CUKR_CL_SINGLE* csr_val,
                       CUKR_CL_SINGLE* x, CUKR_SINGLE beta, CUKR_CL_SINGLE* y)
{
	/* Check for valid binding */
    if (blas->bind.cl_sspmv_hyb == NULL)
        CUKR_FATALX("BLAS %s does not implement the function SSPMV_HYB\n", blas->name);
	
    /*
     * Do the actual work
     */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone;
	CLOCK_START;
#endif
	
    /* Work */
    (blas->bind.cl_sspmv_hyb)(rows, cols, nz, alpha, ell_nz_row, ell_stride, csr_nz,
    					   ell_idx, ell_val, csr_ptr, csr_idx, csr_val, x, beta, y);
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += 2*nz;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += ell_nz_row*ell_stride + csr_nz + rows; // from ell_val and csr_val and x
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += cols;                  // from y;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += ell_nz_row*ell_stride  // from ell_idx
																																	+ rows + csr_nz;       // from csr_ptr, csr_idx
#endif
}

#endif // CUKR_USE_OPENCL

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param y Result vector
 */
void CukrBlasDSPMV_BCSR(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
					    CUKR_DOUBLE alpha, CUKR_INT* ptr, CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x,
                        CUKR_DOUBLE beta, CUKR_DOUBLE* y)
{
	/* Check for valid binding */
	if (blas->bind.dspmv_bcsr == NULL)
		CUKR_FATALX("BLAS %s does not implement the function DSPMV_BCSR\n", blas->name);

	/*
	 * Do the actual work
	 */

	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone;
		CLOCK_START;
	#endif

	/* Work */
	(blas->bind.dspmv_bcsr)(rows, cols, nz, r, c, alpha, ptr, idx, val, x, beta, y);

	/* TODO
	 * Correctly update counters
	 */

	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += nz + cols; // from val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += cols;      // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += nz + rows; // from idx and ptr
	#endif
}

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector (head,tail)
 * \param x Vector being multiplied (head,tail)
 * \param beta Constant
 * \param y Result vector (head,tail)
 */
void CukrBlasQSPMV_BCSR(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
					    CUKR_DOUBLE alpha, CUKR_INT* ptr, CUKR_INT* idx, CUKR_SINGLE* valh, CUKR_SINGLE* valt,
					    CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{

	/* Check for valid binding */
	if (blas->bind.qspmv_bcsr == NULL)
		CUKR_FATALX("BLAS %s does not implement the function QSPMV_BCSR\n", blas->name);

	/*
	 * Do the actual work
	 */

	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone;
		CLOCK_START;
	#endif

	/* Work */
	(blas->bind.qspmv_bcsr)(rows, cols, nz, r, c, alpha, ptr, idx, valh, valt, xh, xt, beta, yh, yt);

	/* TODO
	 * Correctly update counters
	 */

	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + cols; // from val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += cols;      // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + rows; // from idx and ptr
	#endif
}

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR format, in single precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param y Result vector
 */
void CukrBlasSSPMV_BCSR(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
						CUKR_SINGLE alpha, CUKR_INT* ptr, CUKR_INT* idx, CUKR_SINGLE* val, CUKR_SINGLE* x,
			    	    CUKR_SINGLE beta, CUKR_SINGLE* y)
{

	/* Check for valid binding */
	if (blas->bind.sspmv_bcsr == NULL)
		CUKR_FATALX("BLAS %s does not implement the function SSPMV_BCSR\n", blas->name);

	/*
	 * Do the actual work
	 */

	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone;
		CLOCK_START;
	#endif

	/* Work */
	(blas->bind.sspmv_bcsr)(rows, cols, nz, r, c, alpha, ptr, idx, val, x, beta, y);

	/* TODO
	 * Correctly update counters
	 */

	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + cols; // from val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += cols;      // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + rows; // from idx and ptr
	#endif
}

#ifdef CUKR_USE_OPENCL
/**
 * \brief Sparse matrix vector multiply in
 *        BCSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param y Result vector
 */
void CukrCLBlasDSPMV_BCSR(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
					    CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, CUKR_CL_INT* idx, CUKR_CL_DOUBLE* val, CUKR_CL_DOUBLE* x,
                        CUKR_DOUBLE beta, CUKR_CL_DOUBLE* y)
{
	/* Check for valid binding */
	if (blas->bind.cl_dspmv_bcsr == NULL)
		CUKR_FATALX("BLAS %s does not implement the function DSPMV_BCSR\n", blas->name);
	
	/*
	 * Do the actual work
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone;
	CLOCK_START;
#endif
	
	/* Work */
	(blas->bind.cl_dspmv_bcsr)(rows, cols, nz, r, c, alpha, ptr, idx, val, x, beta, y);
	
	/* TODO
	 * Correctly update counters
	 */
	
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += 2*nz;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += nz + cols; // from val and x
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += cols;      // from y;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += nz + rows; // from idx and ptr
#endif
}

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector (head,tail)
 * \param x Vector being multiplied (head,tail)
 * \param beta Constant
 * \param y Result vector (head,tail)
 */
void CukrCLBlasQSPMV_BCSR(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
					    CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, CUKR_CL_INT* idx, CUKR_CL_SINGLE* valh, CUKR_CL_SINGLE* valt,
					    CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt, CUKR_DOUBLE beta, CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt)
{
	
	/* Check for valid binding */
	if (blas->bind.cl_qspmv_bcsr == NULL)
		CUKR_FATALX("BLAS %s does not implement the function QSPMV_BCSR\n", blas->name);
	
	/*
	 * Do the actual work
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone;
	CLOCK_START;
#endif
	
	/* Work */
	(blas->bind.cl_qspmv_bcsr)(rows, cols, nz, r, c, alpha, ptr, idx, valh, valt, xh, xt, beta, yh, yt);
	
	/* TODO
	 * Correctly update counters
	 */
	
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += 2*nz;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + cols; // from val and x
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += cols;      // from y;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + rows; // from idx and ptr
#endif
}

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR format, in single precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param y Result vector
 */
void CukrCLBlasSSPMV_BCSR(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
						CUKR_SINGLE alpha, CUKR_CL_INT* ptr, CUKR_CL_INT* idx, CUKR_CL_SINGLE* val, CUKR_CL_SINGLE* x,
			    	    CUKR_SINGLE beta, CUKR_CL_SINGLE* y)
{
	
	/* Check for valid binding */
	if (blas->bind.cl_sspmv_bcsr == NULL)
		CUKR_FATALX("BLAS %s does not implement the function SSPMV_BCSR\n", blas->name);
	
	/*
	 * Do the actual work
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone;
	CLOCK_START;
#endif
	
	/* Work */
	(blas->bind.cl_sspmv_bcsr)(rows, cols, nz, r, c, alpha, ptr, idx, val, x, beta, y);
	
	/* TODO
	 * Correctly update counters
	 */
	
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += 2*nz;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + cols; // from val and x
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += cols;      // from y;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + rows; // from idx and ptr
#endif
}
#endif //CUKR_USE_OPENCL
// ------------------------------------------------------------------- //

#ifdef CUKR_USE_OPENCL
/**
 * \brief Sparse matrix vector multiply in
 *        BCSR4 format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param y Result vector
 */
void CukrCLBlasDSPMV_BCSR4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
					    CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, CUKR_CL_INT* idx, CUKR_CL_DOUBLE* val, CUKR_CL_DOUBLE* x,
                        CUKR_DOUBLE beta, CUKR_CL_DOUBLE* y)
{
	/* Check for valid binding */
	if (blas->bind.cl_dspmv_bcsr4 == NULL)
		CUKR_FATALX("BLAS %s does not implement the function DSPMV_BCSR4\n", blas->name);

	/*
	 * Do the actual work
	 */

	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone;
		CLOCK_START;
	#endif

	/* Work */
	(blas->bind.cl_dspmv_bcsr4)(rows, cols, nz, r, c, alpha, ptr, idx, val, x, beta, y);

	/* TODO
	 * Correctly update counters
	 */

	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += nz + cols; // from val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += cols;      // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += nz + rows; // from idx and ptr
	#endif
}

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR4 format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector (head,tail)
 * \param x Vector being multiplied (head,tail)
 * \param beta Constant
 * \param y Result vector (head,tail)
 */
void CukrCLBlasQSPMV_BCSR4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
					    CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, CUKR_CL_INT* idx, CUKR_CL_SINGLE* valh, CUKR_CL_SINGLE* valt,
					    CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt, CUKR_DOUBLE beta, CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt)
{

	/* Check for valid binding */
	if (blas->bind.cl_qspmv_bcsr4 == NULL)
		CUKR_FATALX("BLAS %s does not implement the function QSPMV_BCSR4\n", blas->name);

	/*
	 * Do the actual work
	 */

	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone;
		CLOCK_START;
	#endif

	/* Work */
	(blas->bind.cl_qspmv_bcsr4)(rows, cols, nz, r, c, alpha, ptr, idx, valh, valt, xh, xt, beta, yh, yt);

	/* TODO
	 * Correctly update counters
	 */

	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + cols; // from val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += cols;      // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + rows; // from idx and ptr
	#endif
}

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR4 format, in single precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param y Result vector
 */
void CukrCLBlasSSPMV_BCSR4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
						CUKR_SINGLE alpha, CUKR_CL_INT* ptr, CUKR_CL_INT* idx, CUKR_CL_SINGLE* val, CUKR_CL_SINGLE* x,
			    	    CUKR_SINGLE beta, CUKR_CL_SINGLE* y)
{

	/* Check for valid binding */
	if (blas->bind.cl_sspmv_bcsr4 == NULL)
		CUKR_FATALX("BLAS %s does not implement the function SSPMV_BCSR4\n", blas->name);

	/*
	 * Do the actual work
	 */

	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone;
		CLOCK_START;
	#endif

	/* Work */
	(blas->bind.cl_sspmv_bcsr4)(rows, cols, nz, r, c, alpha, ptr, idx, val, x, beta, y);

	/* TODO
	 * Correctly update counters
	 */

	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + cols; // from val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += cols;      // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + rows; // from idx and ptr
	#endif
}

#endif // CUKR_USE_OPENCL

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR4 format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param y Result vector
 */
void CukrBlasDSPMV_BCSR4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
						 CUKR_DOUBLE alpha, CUKR_INT* ptr, CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x,
						 CUKR_DOUBLE beta, CUKR_DOUBLE* y)
{
	/* Check for valid binding */
	if (blas->bind.dspmv_bcsr4 == NULL)
		CUKR_FATALX("BLAS %s does not implement the function DSPMV_BCSR4\n", blas->name);
	
	/*
	 * Do the actual work
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone;
	CLOCK_START;
#endif
	
	/* Work */
	(blas->bind.dspmv_bcsr4)(rows, cols, nz, r, c, alpha, ptr, idx, val, x, beta, y);
	
	/* TODO
	 * Correctly update counters
	 */
	
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += 2*nz;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += nz + cols; // from val and x
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += cols;      // from y;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_DOUBLE] += nz + rows; // from idx and ptr
#endif
}

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR4 format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector (head,tail)
 * \param x Vector being multiplied (head,tail)
 * \param beta Constant
 * \param y Result vector (head,tail)
 */
void CukrBlasQSPMV_BCSR4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
						 CUKR_DOUBLE alpha, CUKR_INT* ptr, CUKR_INT* idx, CUKR_SINGLE* valh, CUKR_SINGLE* valt,
						 CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	
	/* Check for valid binding */
	if (blas->bind.qspmv_bcsr4 == NULL)
		CUKR_FATALX("BLAS %s does not implement the function QSPMV_BCSR4\n", blas->name);
	
	/*
	 * Do the actual work
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone;
	CLOCK_START;
#endif
	
	/* Work */
	(blas->bind.qspmv_bcsr4)(rows, cols, nz, r, c, alpha, ptr, idx, valh, valt, xh, xt, beta, yh, yt);
	
	/* TODO
	 * Correctly update counters
	 */
	
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += 2*nz;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + cols; // from val and x
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += cols;      // from y;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + rows; // from idx and ptr
#endif
}

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR4 format, in single precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param y Result vector
 */
void CukrBlasSSPMV_BCSR4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
						 CUKR_SINGLE alpha, CUKR_INT* ptr, CUKR_INT* idx, CUKR_SINGLE* val, CUKR_SINGLE* x,
						 CUKR_SINGLE beta, CUKR_SINGLE* y)
{
	
	/* Check for valid binding */
	if (blas->bind.sspmv_bcsr4 == NULL)
		CUKR_FATALX("BLAS %s does not implement the function SSPMV_BCSR4\n", blas->name);
	
	/*
	 * Do the actual work
	 */
	
	/* Counter block */
#ifdef CUKR_THREAD_COUNTERS
	struct timeval t_start, t_end;
	struct timezone t_zone;
	CLOCK_START;
#endif
	
	/* Work */
	(blas->bind.sspmv_bcsr4)(rows, cols, nz, r, c, alpha, ptr, idx, val, x, beta, y);
	
	/* TODO
	 * Correctly update counters
	 */
	
#ifdef CUKR_THREAD_COUNTERS
	CLOCK_END;
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += 2*nz;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + cols; // from val and x
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += cols;      // from y;
	blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_CSR][CUKR_BLAS_PREC_SINGLE] += nz + rows; // from idx and ptr
#endif
}

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr_hi/lo CSR ptr vectors
 * \param csr_idx CSR idx vectors
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrBlasDSPMV_HYB(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                       CUKR_INT* ell_idx, CUKR_DOUBLE* ell_val, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_DOUBLE* csr_val,
                       CUKR_DOUBLE* x, CUKR_DOUBLE beta, CUKR_DOUBLE* y)
{
    /* Check for valid binding */
    if (blas->bind.dspmv_hyb == NULL)
        CUKR_FATALX("BLAS %s does not implement the function DSPMV_HYB\n", blas->name);

    /*
     * Do the actual work
     */

    /* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone;
		CLOCK_START;
	#endif

    /* Work */
    (blas->bind.dspmv_hyb)(rows, cols, nz, alpha, ell_nz_row, ell_stride, csr_nz,
    					   ell_idx, ell_val, csr_ptr, csr_idx, csr_val, x, beta, y);

    /* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += ell_nz_row*ell_stride + csr_nz + rows; // from ell_val and csr_val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += cols;                  // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += ell_nz_row*ell_stride  // from ell_idx
																												      + rows + csr_nz;       // from csr_ptr, csr_idx
	#endif
}

/**
 * \brief Wrapper for QSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
void CukrBlasQSPMV_HYB(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
	                   CUKR_INT* ell_idx, CUKR_SINGLE* ell_valh, CUKR_SINGLE* ell_valt, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_SINGLE* csr_valh, CUKR_SINGLE* csr_valt,
	                   CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
   /* Check for valid binding */
    if (blas->bind.qspmv_hyb == NULL)
        CUKR_FATALX("BLAS %s does not implement the function QSPMV_HYB\n", blas->name);

    /*
     * Do the actual work
     */

    /* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone;
		CLOCK_START;
	#endif

    /* Work */
    (blas->bind.qspmv_hyb)(rows, cols, nz, alpha, ell_nz_row, ell_stride, csr_nz,
    					   ell_idx, ell_valh, ell_valt, csr_ptr, csr_idx, csr_valh, csr_valt,
    					   xh, xt, beta, yh, yt);

    /* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += ell_nz_row*ell_stride + csr_nz + rows; // from ell_val and csr_val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += cols;                  // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += ell_nz_row*ell_stride  // from ell_idx
																												      + rows + csr_nz;       // from csr_ptr, csr_idx
	#endif
}

/**
 * \brief Wrapper for SSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vecto
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vectors
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrBlasSSPMV_HYB(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                       CUKR_INT* ell_idx, CUKR_SINGLE* ell_val, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_SINGLE* csr_val,
                       CUKR_SINGLE* x, CUKR_SINGLE beta, CUKR_SINGLE* y)
{
   /* Check for valid binding */
    if (blas->bind.sspmv_hyb == NULL)
        CUKR_FATALX("BLAS %s does not implement the function SSPMV_HYB\n", blas->name);

    /*
     * Do the actual work
     */

		/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone;
		CLOCK_START;
	#endif

    /* Work */
    (blas->bind.sspmv_hyb)(rows, cols, nz, alpha, ell_nz_row, ell_stride, csr_nz,
    					   ell_idx, ell_val, csr_ptr, csr_idx, csr_val, x, beta, y);

	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += ell_nz_row*ell_stride + csr_nz + rows; // from ell_val and csr_val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += cols;                  // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_HYB][CUKR_BLAS_PREC_SINGLE] += ell_nz_row*ell_stride  // from ell_idx
																												     + rows + csr_nz;        // from csr_ptr, csr_idx
	#endif
}


// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr_hi/lo CSR ptr vectors
 * \param csr_idx CSR idx vectors
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrBlasDSPMV_BHYB(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
                       CUKR_INT* bell_idx, CUKR_DOUBLE* bell_val, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_DOUBLE* bcsr_val,
                       CUKR_DOUBLE* x, CUKR_DOUBLE beta, CUKR_DOUBLE* y)
{
    /* Check for valid binding */
    if (blas->bind.dspmv_bhyb == NULL)
        CUKR_FATALX("BLAS %s does not implement the function DSPMV_BHYB\n", blas->name);

    /*
     * Do the actual work
     */

    /* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone;
		CLOCK_START;
	#endif

    /* Work */
    (blas->bind.dspmv_bhyb)(rows, cols, nz, r, c, alpha, bell_nz_row, bell_stride, bcsr_nz,
    					   bell_idx, bell_val, bcsr_ptr, bcsr_idx, bcsr_val, x, beta, y);

    /* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += bell_nz_row*bell_stride + bcsr_nz + rows; // from bell_val and bcsr_val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += cols;                  // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += bell_nz_row*bell_stride  // from bell_idx
																												      + rows + bcsr_nz;       // from bcsr_ptr, bcsr_idx
	#endif
}


/**
 * \brief Wrapper for QSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
void CukrBlasQSPMV_BHYB(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
	                   CUKR_INT* bell_idx, CUKR_SINGLE* bell_valh, CUKR_SINGLE* bell_valt, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_SINGLE* bcsr_valh, CUKR_SINGLE* bcsr_valt,
	                   CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
    /* Check for valid binding */
    if (blas->bind.qspmv_bhyb == NULL)
        CUKR_FATALX("BLAS %s does not implement the function QSPMV_BHYB\n", blas->name);

    /*
     * Do the actual work
     */

    /* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone;
		CLOCK_START;
	#endif

    /* Work */
    (blas->bind.qspmv_bhyb)(rows, cols, nz, r, c, alpha, bell_nz_row, bell_stride, bcsr_nz,
    					   bell_idx, bell_valh, bell_valt, bcsr_ptr, bcsr_idx, bcsr_valh, bcsr_valt,
    					   xh, xt, beta, yh, yt);

    /* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += bell_nz_row*bell_stride + bcsr_nz + rows; // from bell_val and bcsr_val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += cols;                  // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += bell_nz_row*bell_stride  // from ell_idx
																												      + rows + bcsr_nz;       // from bcsr_ptr, bcsr_idx
	#endif
}


/**
 * \brief Wrapper for SSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vectors
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrBlasSSPMV_BHYB(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
                       CUKR_INT* bell_idx, CUKR_SINGLE* bell_val, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_SINGLE* bcsr_val,
                       CUKR_SINGLE* x, CUKR_SINGLE beta, CUKR_SINGLE* y)
{
   /* Check for valid binding */
    if (blas->bind.sspmv_bhyb == NULL)
        CUKR_FATALX("BLAS %s does not implement the function SSPMV_BHYB\n", blas->name);

    /*
     * Do the actual work
     */

		/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone;
		CLOCK_START;
	#endif

    /* Work */
    (blas->bind.sspmv_bhyb)(rows, cols, nz, r, c, alpha, bell_nz_row, bell_stride, bcsr_nz,
    					   bell_idx, bell_val, bcsr_ptr, bcsr_idx, bcsr_val, x, beta, y);

	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += bell_nz_row*bell_stride + bcsr_nz + rows; // from bell_val and bcsr_val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += cols;                  // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += bell_nz_row*bell_stride  // from bell_idx
																												     + rows + bcsr_nz;        // from bcsr_ptr, bcsr_idx
	#endif
}

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_BHYB4
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr_hi/lo CSR ptr vectors
 * \param csr_idx CSR idx vectors
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrBlasDSPMV_BHYB4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
                       CUKR_INT* bell_idx, CUKR_DOUBLE* bell_val, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_DOUBLE* bcsr_val,
                       CUKR_DOUBLE* x, CUKR_DOUBLE beta, CUKR_DOUBLE* y)
{
    /* Check for valid binding */
    if (blas->bind.dspmv_bhyb4 == NULL)
        CUKR_FATALX("BLAS %s does not implement the function DSPMV_BHYB4\n", blas->name);

    /*
     * Do the actual work
     */

    /* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone;
		CLOCK_START;
	#endif

    /* Work */
    (blas->bind.dspmv_bhyb4)(rows, cols, nz, r, c, alpha, bell_nz_row, bell_stride, bcsr_nz,
    					   bell_idx, bell_val, bcsr_ptr, bcsr_idx, bcsr_val, x, beta, y);

    /* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += bell_nz_row*bell_stride + bcsr_nz + rows; // from bell_val and bcsr_val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += cols;                  // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += bell_nz_row*bell_stride  // from bell_idx
																												      + rows + bcsr_nz;       // from bcsr_ptr, bcsr_idx
	#endif
}


/**
 * \brief Wrapper for QSPMV_BHYB4
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
void CukrBlasQSPMV_BHYB4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
	                   CUKR_INT* bell_idx, CUKR_SINGLE* bell_valh, CUKR_SINGLE* bell_valt, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_SINGLE* bcsr_valh, CUKR_SINGLE* bcsr_valt,
	                   CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
    /* Check for valid binding */
    if (blas->bind.qspmv_bhyb4 == NULL)
        CUKR_FATALX("BLAS %s does not implement the function QSPMV_BHYB\n", blas->name);

    /*
     * Do the actual work
     */

    /* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone;
		CLOCK_START;
	#endif

    /* Work */
    (blas->bind.qspmv_bhyb4)(rows, cols, nz, r, c, alpha, bell_nz_row, bell_stride, bcsr_nz,
    					   bell_idx, bell_valh, bell_valt, bcsr_ptr, bcsr_idx, bcsr_valh, bcsr_valt,
    					   xh, xt, beta, yh, yt);

    /* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += bell_nz_row*bell_stride + bcsr_nz + rows; // from bell_val and bcsr_val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += cols;                  // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += bell_nz_row*bell_stride  // from ell_idx
																												      + rows + bcsr_nz;       // from bcsr_ptr, bcsr_idx
	#endif
}

/**
 * \brief Wrapper for SSPMV_BHYB4
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vectors
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrBlasSSPMV_BHYB4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
                       CUKR_INT* bell_idx, CUKR_SINGLE* bell_val, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_SINGLE* bcsr_val,
                       CUKR_SINGLE* x, CUKR_SINGLE beta, CUKR_SINGLE* y)
{
   /* Check for valid binding */
    if (blas->bind.sspmv_bhyb4 == NULL)
        CUKR_FATALX("BLAS %s does not implement the function SSPMV_BHYB4\n", blas->name);

    /*
     * Do the actual work
     */

		/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone;
		CLOCK_START;
	#endif

    /* Work */
    (blas->bind.sspmv_bhyb4)(rows, cols, nz, r, c, alpha, bell_nz_row, bell_stride, bcsr_nz,
    					   bell_idx, bell_val, bcsr_ptr, bcsr_idx, bcsr_val, x, beta, y);

	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;
		int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]     [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]    [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += 2*nz;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS]  [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += bell_nz_row*bell_stride + bcsr_nz + rows; // from bell_val and bcsr_val and x
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES] [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += cols;                  // from y;
		blas->counters[*cukr_activeThread][CUKR_BLAS_IDX_INTLOADS] [CUKR_BLAS_OPS_SPMV_BHYB][CUKR_BLAS_PREC_SINGLE] += bell_nz_row*bell_stride  // from bell_idx
																												     + rows + bcsr_nz;        // from bcsr_ptr, bcsr_idx
	#endif
}

// ------------------------------------------------------------------- //
