/**
 * @file blas/blas.h
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains the interface to the BLAS routines.
 *  
 * Solvers and preconditioners use, where possible
 * instructions from the BLAS. There may be more
 * then one implementation of the BLAS, so for
 * generality each BLAS must implement a set of
 * standard operations.
 *  
 * A BLAS implementing the following interface is
 * in fact a wrapper for an external BLAS (e.g.
 * MKL or OSKI).    
 * 
 * Each BLAS must implement the following functions:
 * - dot
 * - axpy & aypx
 * - scal
 * - copy
 * - SpMV
 * 
 * The BLAS is also responsible for counting ops,
 * memory loads and stores and providing performance
 * reports.
 * 	
 */

#ifndef BLAS_H_
#define BLAS_H_

#include "../cukr.h"

/* Define the max no. of operations 
 * in the BLAS */
#define CUKR_BLAS_OPS_MAXCOUNT 10

/* Define all operations in the BLAS */
#define CUKR_BLAS_OPS_DOT      0
#define CUKR_BLAS_OPS_AXPY     1
#define CUKR_BLAS_OPS_AYPX     2
#define CUKR_BLAS_OPS_SCAL     3
#define CUKR_BLAS_OPS_COPY     4
#define CUKR_BLAS_OPS_SPMV_CSR 5
#define CUKR_BLAS_OPS_SPMV     5
#define CUKR_BLAS_OPS_SPMV_BCSR 5
#define CUKR_BLAS_OPS_SPMV_HYB 5
#define CUKR_BLAS_OPS_SPMV_BHYB 5

/* Define the maximum number of
 * counter types in the BLAS */

#define CUKR_BLAS_IDX_MAXCOUNT 10

/**
 * Counter index for measuring time (in msec) 
 * They should be in successive order,
 * and with a space between larger than the
 * largest offset.
 */ 
#define CUKR_BLAS_IDX_TIME      0
/**
 * Counter index for measuring the number
 * of executed floating point operations (in flops) 
 */
#define CUKR_BLAS_IDX_FLOPS	    1
/**
 * Counter index for measuring the number
 * of executed integer operations (in iops) 
 */
#define CUKR_BLAS_IDX_INTOPS	2
/**
 * Counter index for measuring the number
 * of floating point loads (in loads) 
 */
#define CUKR_BLAS_IDX_FLLOADS   3
/**
 * Counter index for measuring the number
 * of integer loads (in loads) 
 */
#define CUKR_BLAS_IDX_INTLOADS  4
/**
 * Counter index for measuring the number
 * of floating point stores (in stores) 
 */
#define CUKR_BLAS_IDX_FLSTORES  5
/**
 * Counter index for measuring the number
 * of integer stores (in loads) 
 */
#define CUKR_BLAS_IDX_INTSTORES 6
 
/* We have different operations for different
 * precisions, so we should index after prec.
 * as well */
 
/* Define the maximum number of
 * precisions in the BLAS */ 

#define CUKR_BLAS_PREC_MAXCOUNT 5

/* Define double precision */
#define CUKR_BLAS_PREC_DOUBLE  0

/* Define qdouble precision */
#define CUKR_BLAS_PREC_QDOUBLE 1

/* Define single precision */
#define CUKR_BLAS_PREC_SINGLE  2

/* Include to blasx.h put here, so
 * that the declarations above the
 * include can be accessed */
#include "../blasx/blasx.h"
 
/*
 * The actual BLAS functions found in the BLAS
 * implementations are linked to the BLAS object
 * using function pointers. Here we declare these
 * pointers for each implemented operation.
 */

/* Dot product (d := x'x) -> d = _DOT (n,x,y) */
typedef CUKR_DOUBLE(*CUKR_BLAS_DDOT)(CUKR_INT, CUKR_DOUBLE*, CUKR_DOUBLE*);
typedef CUKR_DOUBLE(*CUKR_BLAS_QDOT)(CUKR_INT, CUKR_SINGLE*, CUKR_SINGLE*,
                                               CUKR_SINGLE*, CUKR_SINGLE*);       
typedef CUKR_SINGLE(*CUKR_BLAS_SDOT)(CUKR_INT, CUKR_SINGLE*, CUKR_SINGLE*);   

/* AXPY (Y := aX + Y) -> _AXPY (n,a,x,y) */
typedef void (*CUKR_BLAS_DAXPY)(CUKR_INT, CUKR_DOUBLE, CUKR_DOUBLE*, CUKR_DOUBLE*); 
typedef void (*CUKR_BLAS_QAXPY)(CUKR_INT, CUKR_DOUBLE, CUKR_SINGLE*, CUKR_SINGLE*,
													   CUKR_SINGLE*, CUKR_SINGLE*); 
typedef void (*CUKR_BLAS_SAXPY)(CUKR_INT, CUKR_SINGLE, CUKR_SINGLE*, CUKR_SINGLE*);

/* AYPX (Y := aY + X) -> _AYPX (n,a,x,y) */
typedef void (*CUKR_BLAS_DAYPX)(CUKR_INT, CUKR_DOUBLE, CUKR_DOUBLE*, CUKR_DOUBLE*); 
typedef void (*CUKR_BLAS_QAYPX)(CUKR_INT, CUKR_DOUBLE, CUKR_SINGLE*, CUKR_SINGLE*,
													   CUKR_SINGLE*, CUKR_SINGLE*);
typedef void (*CUKR_BLAS_SAYPX)(CUKR_INT, CUKR_SINGLE, CUKR_SINGLE*, CUKR_SINGLE*);

/* SCAL (Y := aY) -> d = _SCAL(n,a,x) */
typedef void (*CUKR_BLAS_DSCAL)(CUKR_INT, CUKR_DOUBLE, CUKR_DOUBLE*); 
typedef void (*CUKR_BLAS_QSCAL)(CUKR_INT, CUKR_DOUBLE, CUKR_SINGLE*,
                                                       CUKR_SINGLE*); 
typedef void (*CUKR_BLAS_SSCAL)(CUKR_INT, CUKR_SINGLE, CUKR_SINGLE*);

/* COPY (Y := X) -> _COPY(n,x,y) */
typedef void (*CUKR_BLAS_DCOPY)(CUKR_INT, CUKR_DOUBLE*, CUKR_DOUBLE*); 
typedef void (*CUKR_BLAS_QCOPY)(CUKR_INT, CUKR_SINGLE*, CUKR_SINGLE*,
                                          CUKR_SINGLE*, CUKR_SINGLE*); 
typedef void (*CUKR_BLAS_SCOPY)(CUKR_INT, CUKR_SINGLE*, CUKR_SINGLE*);

/* SpMV (Y := AX + aY) -> y = _SPMV(n,nz,ptr,idx,val,x,a,y) */
typedef void (*CUKR_BLAS_DSPMV_CSR)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_INT*, CUKR_INT*, 
                                    CUKR_DOUBLE*, CUKR_DOUBLE*, CUKR_DOUBLE, CUKR_DOUBLE*);
typedef void (*CUKR_BLAS_QSPMV_CSR)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_INT*, CUKR_INT*, 
                                    CUKR_SINGLE*, CUKR_SINGLE*, CUKR_SINGLE*, CUKR_SINGLE*,
                                    CUKR_DOUBLE, CUKR_SINGLE*, CUKR_SINGLE*);                                     
typedef void (*CUKR_BLAS_SSPMV_CSR)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_SINGLE, CUKR_INT*, CUKR_INT*, 
                                    CUKR_SINGLE*, CUKR_SINGLE*, CUKR_SINGLE, CUKR_SINGLE*); 

/* SpMV (Y := AX + aY) -> y = _SPMV_BCSR(n,nz,r,c,ptr,idx,val,x,a,y) */
typedef void (*CUKR_BLAS_DSPMV_BCSR)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_INT*, CUKR_INT*,
                                    CUKR_DOUBLE*, CUKR_DOUBLE*, CUKR_DOUBLE, CUKR_DOUBLE*);
typedef void (*CUKR_BLAS_QSPMV_BCSR)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_INT*, CUKR_INT*,
                                    CUKR_SINGLE*, CUKR_SINGLE*, CUKR_SINGLE*, CUKR_SINGLE*,
                                    CUKR_DOUBLE, CUKR_SINGLE*, CUKR_SINGLE*);
typedef void (*CUKR_BLAS_SSPMV_BCSR)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_SINGLE, CUKR_INT*, CUKR_INT*,
                                    CUKR_SINGLE*, CUKR_SINGLE*, CUKR_SINGLE, CUKR_SINGLE*);

/* SpMV in HYB (Y := aAX + aY) -> y = _SPMV(rows,cols,nz, a, ell_nz_row,ell_stride,csr_nz, ell_idx,ell_val,csr_ptr,csr_idx,csr_val,x,a,y) */
typedef void (*CUKR_BLAS_DSPMV_HYB)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_INT, CUKR_INT, CUKR_INT,
                                    CUKR_INT*, CUKR_DOUBLE*, CUKR_INT*, CUKR_INT*, CUKR_DOUBLE*,
                                    CUKR_DOUBLE*, CUKR_DOUBLE, CUKR_DOUBLE*);
typedef void (*CUKR_BLAS_QSPMV_HYB)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_INT, CUKR_INT, CUKR_INT,
                                    CUKR_INT*, CUKR_SINGLE*, CUKR_SINGLE*, CUKR_INT*, CUKR_INT*, CUKR_SINGLE*, CUKR_SINGLE*,
                                    CUKR_SINGLE*, CUKR_SINGLE*, CUKR_DOUBLE, CUKR_SINGLE*, CUKR_SINGLE*);
typedef void (*CUKR_BLAS_SSPMV_HYB)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_SINGLE, CUKR_INT, CUKR_INT, CUKR_INT,
                                    CUKR_INT*, CUKR_SINGLE*, CUKR_INT*, CUKR_INT*, CUKR_SINGLE*,
                                    CUKR_SINGLE*, CUKR_SINGLE, CUKR_SINGLE*);

/* SpMV in BHYB (Y := aAX + aY) -> y = _SPMV(rows,cols,nz, r,c, a, ell_nz_row,ell_stride,csr_nz, ell_idx,ell_val,csr_ptr,csr_idx,csr_val,x,a,y) */
typedef void (*CUKR_BLAS_DSPMV_BHYB)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_INT, CUKR_INT, CUKR_INT,
                                    CUKR_INT*, CUKR_DOUBLE*, CUKR_INT*, CUKR_INT*, CUKR_DOUBLE*,
                                    CUKR_DOUBLE*, CUKR_DOUBLE, CUKR_DOUBLE*);
typedef void (*CUKR_BLAS_QSPMV_BHYB)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_INT, CUKR_INT, CUKR_INT,
                                    CUKR_INT*, CUKR_SINGLE*, CUKR_SINGLE*, CUKR_INT*, CUKR_INT*, CUKR_SINGLE*, CUKR_SINGLE*,
                                    CUKR_SINGLE*, CUKR_SINGLE*, CUKR_DOUBLE, CUKR_SINGLE*, CUKR_SINGLE*);
typedef void (*CUKR_BLAS_SSPMV_BHYB)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_SINGLE, CUKR_INT, CUKR_INT, CUKR_INT,
                                    CUKR_INT*, CUKR_SINGLE*, CUKR_INT*, CUKR_INT*, CUKR_SINGLE*,
                                    CUKR_SINGLE*, CUKR_SINGLE, CUKR_SINGLE*);

#ifdef CUKR_USE_OPENCL
typedef CUKR_DOUBLE(*CUKR_CL_BLAS_DDOT)(CUKR_INT, CUKR_CL_DOUBLE*, CUKR_CL_DOUBLE*);
typedef CUKR_DOUBLE(*CUKR_CL_BLAS_QDOT)(CUKR_INT, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*,
												  CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);       
typedef CUKR_SINGLE(*CUKR_CL_BLAS_SDOT)(CUKR_INT, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);

/* AYPX (Y := aY + X) -> _AYPX (n,a,x,y) */
typedef void (*CUKR_CL_BLAS_DAYPX)(CUKR_INT, CUKR_DOUBLE, CUKR_CL_DOUBLE*, CUKR_CL_DOUBLE*); 
typedef void (*CUKR_CL_BLAS_QAYPX)(CUKR_INT, CUKR_DOUBLE, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);
typedef void (*CUKR_CL_BLAS_SAYPX)(CUKR_INT, CUKR_SINGLE, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);

/* AXPY (Y := aX + Y) -> _AXPY (n,a,x,y) */
typedef void (*CUKR_CL_BLAS_DAXPY)(CUKR_INT, CUKR_DOUBLE, CUKR_CL_DOUBLE*, CUKR_CL_DOUBLE*); 
typedef void (*CUKR_CL_BLAS_QAXPY)(CUKR_INT, CUKR_DOUBLE, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);
typedef void (*CUKR_CL_BLAS_SAXPY)(CUKR_INT, CUKR_SINGLE, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);

/* SCAL (Y := aY) -> d = _SCAL(n,a,x) */
typedef void (*CUKR_CL_BLAS_DSCAL)(CUKR_INT, CUKR_DOUBLE, CUKR_CL_DOUBLE*); 
typedef void (*CUKR_CL_BLAS_QSCAL)(CUKR_INT, CUKR_DOUBLE, CUKR_CL_SINGLE*,
														  CUKR_CL_SINGLE*); 
typedef void (*CUKR_CL_BLAS_SSCAL)(CUKR_INT, CUKR_SINGLE, CUKR_CL_SINGLE*);

typedef void (*CUKR_CL_BLAS_DCOPY)(CUKR_INT, CUKR_CL_DOUBLE*, CUKR_CL_DOUBLE*); 
typedef void (*CUKR_CL_BLAS_QCOPY)(CUKR_INT, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*); 
typedef void (*CUKR_CL_BLAS_SCOPY)(CUKR_INT, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);

/* SpMV (Y := AX + aY) -> y = _SPMV(n,nz,ptr,idx,val,x,a,y) */
typedef void (*CUKR_CL_BLAS_DSPMV_CSR)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_CL_INT*, CUKR_CL_INT*, 
                                    CUKR_CL_DOUBLE*, CUKR_CL_DOUBLE*, CUKR_DOUBLE, CUKR_CL_DOUBLE*);
typedef void (*CUKR_CL_BLAS_QSPMV_CSR)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_CL_INT*, CUKR_CL_INT*, 
                                    CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*,
                                    CUKR_DOUBLE, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);                                     
typedef void (*CUKR_CL_BLAS_SSPMV_CSR)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_SINGLE, CUKR_CL_INT*, CUKR_CL_INT*, 
                                    CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_SINGLE, CUKR_CL_SINGLE*);

/* SpMV in HYB (Y := aAX + aY) -> y = _SPMV(rows,cols,nz, a, ell_nz_row,ell_stride,csr_nz, ell_idx,ell_val,csr_ptr,csr_idx,csr_val,x,a,y) */
typedef void (*CUKR_CL_BLAS_DSPMV_HYB)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_INT, CUKR_INT, CUKR_INT,
                                    CUKR_CL_INT*, CUKR_CL_DOUBLE*, CUKR_CL_INT*, CUKR_CL_INT*, CUKR_CL_DOUBLE*,
                                    CUKR_CL_DOUBLE*, CUKR_DOUBLE, CUKR_CL_DOUBLE*);
typedef void (*CUKR_CL_BLAS_QSPMV_HYB)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_INT, CUKR_INT, CUKR_INT,
                                    CUKR_CL_INT*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_CL_INT*, CUKR_CL_INT*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*,
                                    CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_DOUBLE, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);
typedef void (*CUKR_CL_BLAS_SSPMV_HYB)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_SINGLE, CUKR_INT, CUKR_INT, CUKR_INT,
                                    CUKR_CL_INT*, CUKR_CL_SINGLE*, CUKR_CL_INT*, CUKR_CL_INT*, CUKR_CL_SINGLE*,
                                    CUKR_CL_SINGLE*, CUKR_SINGLE, CUKR_CL_SINGLE*);

/* SpMV (Y := AX + aY) -> y = _SPMV_BCSR(n,nz,r,c,ptr,idx,val,x,a,y) */
typedef void (*CUKR_CL_BLAS_DSPMV_BCSR)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_CL_INT*, CUKR_CL_INT*,
									 CUKR_CL_DOUBLE*, CUKR_CL_DOUBLE*, CUKR_DOUBLE, CUKR_CL_DOUBLE*);
typedef void (*CUKR_CL_BLAS_QSPMV_BCSR)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_CL_INT*, CUKR_CL_INT*,
									 CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*,
									 CUKR_DOUBLE, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);
typedef void (*CUKR_CL_BLAS_SSPMV_BCSR)(CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_SINGLE, CUKR_CL_INT*, CUKR_CL_INT*,
									 CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_SINGLE, CUKR_CL_SINGLE*);
#endif

/** 
 * Here we define the BLAS structure. 
 * This structure will include:
 * - a <bind> group where the actual function 
 *   pointers will be stored
 * - a vector where the counters will be stored
 * - name to use for logging 
 */                        

typedef struct _CUKR_BLAS_OPS{
	CUKR_BLAS_DDOT ddot;
	CUKR_BLAS_QDOT qdot;
	CUKR_BLAS_SDOT sdot;
	
	CUKR_BLAS_DAXPY daxpy;
	CUKR_BLAS_QAXPY qaxpy;
	CUKR_BLAS_SAXPY saxpy;
	
	CUKR_BLAS_DAYPX daypx;
	CUKR_BLAS_QAYPX qaypx;
	CUKR_BLAS_SAYPX saypx;
	
	CUKR_BLAS_DSCAL dscal;
	CUKR_BLAS_QSCAL qscal;
	CUKR_BLAS_SSCAL sscal;
	
	CUKR_BLAS_DCOPY dcopy;
	CUKR_BLAS_QCOPY qcopy;
	CUKR_BLAS_SCOPY scopy;
	
	CUKR_BLAS_DSPMV_CSR dspmv_csr, dspmv_csr4;
	CUKR_BLAS_QSPMV_CSR qspmv_csr, qspmv_csr4;
	CUKR_BLAS_SSPMV_CSR sspmv_csr, sspmv_csr4;

	CUKR_BLAS_DSPMV_BCSR dspmv_bcsr, dspmv_bcsr4;
	CUKR_BLAS_QSPMV_BCSR qspmv_bcsr, qspmv_bcsr4;
	CUKR_BLAS_SSPMV_BCSR sspmv_bcsr, sspmv_bcsr4;

	CUKR_BLAS_DSPMV_HYB dspmv_hyb;
	CUKR_BLAS_QSPMV_HYB qspmv_hyb;
	CUKR_BLAS_SSPMV_HYB sspmv_hyb;
	
	CUKR_BLAS_DSPMV_BHYB dspmv_bhyb, dspmv_bhyb4;
	CUKR_BLAS_QSPMV_BHYB qspmv_bhyb, qspmv_bhyb4;
	CUKR_BLAS_SSPMV_BHYB sspmv_bhyb, sspmv_bhyb4;
	
#ifdef CUKR_USE_OPENCL
	CUKR_CL_BLAS_DDOT cl_ddot;
	CUKR_CL_BLAS_QDOT cl_qdot;
	CUKR_CL_BLAS_SDOT cl_sdot;
	
	CUKR_CL_BLAS_DAYPX cl_daypx;
	CUKR_CL_BLAS_QAYPX cl_qaypx;
	CUKR_CL_BLAS_SAYPX cl_saypx;
	
	CUKR_CL_BLAS_DAXPY cl_daxpy;
	CUKR_CL_BLAS_QAXPY cl_qaxpy;
	CUKR_CL_BLAS_SAXPY cl_saxpy;
	
	CUKR_CL_BLAS_DSCAL cl_dscal;
	CUKR_CL_BLAS_QSCAL cl_qscal;
	CUKR_CL_BLAS_SSCAL cl_sscal;
	
	CUKR_CL_BLAS_DCOPY cl_dcopy;
	CUKR_CL_BLAS_QCOPY cl_qcopy;
	CUKR_CL_BLAS_SCOPY cl_scopy;
	
	CUKR_CL_BLAS_DSPMV_CSR cl_dspmv_csr, cl_dspmv_csr4;
	CUKR_CL_BLAS_QSPMV_CSR cl_qspmv_csr, cl_qspmv_csr4;
	CUKR_CL_BLAS_SSPMV_CSR cl_sspmv_csr, cl_sspmv_csr4;
	
	CUKR_CL_BLAS_DSPMV_HYB cl_dspmv_hyb;
	CUKR_CL_BLAS_QSPMV_HYB cl_qspmv_hyb;
	CUKR_CL_BLAS_SSPMV_HYB cl_sspmv_hyb;
	
	CUKR_CL_BLAS_DSPMV_BCSR cl_dspmv_bcsr, cl_dspmv_bcsr4;
	CUKR_CL_BLAS_QSPMV_BCSR cl_qspmv_bcsr, cl_qspmv_bcsr4;
	CUKR_CL_BLAS_SSPMV_BCSR cl_sspmv_bcsr, cl_sspmv_bcsr4;
#endif

}CUKR_BLAS_OPS;

typedef struct _CUKR_BLAS{
	
	/** Bindings to the function pointers */
	CUKR_BLAS_OPS bind;
	
	/** Counter matrix, by thread*/
	float counters[CUKR_MAX_PE][CUKR_BLAS_IDX_MAXCOUNT][CUKR_BLAS_OPS_MAXCOUNT][CUKR_BLAS_PREC_MAXCOUNT];
	
	/** Counter matrix, by node */
	float counters_mp[CUKR_BLAS_IDX_MAXCOUNT][CUKR_BLAS_OPS_MAXCOUNT];
	
	/** Counter matrix, by cluster */
	float counters_mpi[CUKR_BLAS_IDX_MAXCOUNT][CUKR_BLAS_OPS_MAXCOUNT];
	
	/** Communication time for internal and external 
	 * ops (inside one node and outside). Written
	 * only from the _MP and _MPI operations */
	float comm_inner, comm_outer;
	
	/** BLAS name */
	char name[255];
	
	/** BLAS location (default: CPU) */
	int location;
	
	/** BLAS precision (set when the BLAS is first
	 *  created; the same blas implementation can 
	 *  be used with more than one precision but
	 *  a blas object must have only one 
	 */ 
	int precision;
	
	/** Associated BLASX object */
	CUKR_BLASX blasx;

}*CUKR_BLAS;

/* 
 * Additional functions for initializing the BLAS
 * plus wrappers for the founction pointers to
 * automatically take care of updating the counters
 * when the associated functions are being used.
 */

 
/**
 * \brief Init a BLAS object 
 * \param blas Pointer to yhe BLAS 
 *         structure 
 */
void CukrBlasCreate(CUKR_BLAS* blas);

// ------------------------------------------------------------------- //

/**
 * \brief Reset all the counters to zero
 * \param blas The blas structure 
 */
void CukrBlasResetCounters(CUKR_BLAS blas);

// ------------------------------------------------------------------- //

/**
 * \brief Merges the counters for all threads
 *        into one (the first one) 
 * \param blas The blas structure 
 */
void CukrBlasMergeThreadCounters(CUKR_BLAS blas);

// ------------------------------------------------------------------- //

/**
 * \brief Merges the counters for all nodes 
 * \param blas The blas structure 
 */
void CukrBlasMergeNodeCounters(CUKR_BLAS blas);
 
// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DDOT 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */
CUKR_DOUBLE CukrBlasDDOT(CUKR_BLAS, CUKR_INT, CUKR_DOUBLE*, CUKR_DOUBLE*); 

/**
 * \brief Wrapper for QDOT 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
CUKR_DOUBLE CukrBlasQDOT(CUKR_BLAS, CUKR_INT, CUKR_SINGLE*, CUKR_SINGLE*,
                                              CUKR_SINGLE*, CUKR_SINGLE*); 

/**
 * \brief Wrapper for SDOT 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */
CUKR_SINGLE CukrBlasSDOT(CUKR_BLAS , CUKR_INT, CUKR_SINGLE*, CUKR_SINGLE*);

#ifdef CUKR_USE_OPENCL
/**
 * \brief Wrapper for DDOT 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */
CUKR_DOUBLE CukrCLBlasDDOT(CUKR_BLAS, CUKR_INT, CUKR_CL_DOUBLE*, CUKR_CL_DOUBLE*); 

/**
 * \brief Wrapper for QDOT 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
CUKR_DOUBLE CukrCLBlasQDOT(CUKR_BLAS, CUKR_INT, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*,
												CUKR_CL_SINGLE*, CUKR_CL_SINGLE*); 

/**
 * \brief Wrapper for SDOT 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */
CUKR_SINGLE CukrCLBlasSDOT(CUKR_BLAS , CUKR_INT, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);
#endif // CUKR_USE_OPENCL


// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DAXPY 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */
void CukrBlasDAXPY(CUKR_BLAS, CUKR_INT, CUKR_DOUBLE, CUKR_DOUBLE*, CUKR_DOUBLE*);

/**
 * \brief Wrapper for QAXPY 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
void CukrBlasQAXPY(CUKR_BLAS, CUKR_INT, CUKR_DOUBLE, CUKR_SINGLE*, CUKR_SINGLE*,
													 CUKR_SINGLE*, CUKR_SINGLE*);

/**
 * \brief Wrapper for SAXPY 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */ 
void CukrBlasSAXPY(CUKR_BLAS, CUKR_INT, CUKR_SINGLE, CUKR_SINGLE*, CUKR_SINGLE*);

#ifdef CUKR_USE_OPENCL
/**
 * \brief Wrapper for DAXPY 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */
void CukrCLBlasDAXPY(CUKR_BLAS, CUKR_INT, CUKR_DOUBLE, CUKR_CL_DOUBLE*, CUKR_CL_DOUBLE*);

/**
 * \brief Wrapper for QAXPY 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
void CukrCLBlasQAXPY(CUKR_BLAS, CUKR_INT, CUKR_DOUBLE, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*,
				   CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);

/**
 * \brief Wrapper for SAXPY 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */ 
void CukrCLBlasSAXPY(CUKR_BLAS, CUKR_INT, CUKR_SINGLE, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);

#endif // CUKR_USE_OPENCL


// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DAYPX 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */
void CukrBlasDAYPX(CUKR_BLAS, CUKR_INT, CUKR_DOUBLE, CUKR_DOUBLE*, CUKR_DOUBLE*); 

/**
 * \brief Wrapper for QAYPX 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
void CukrBlasQAYPX(CUKR_BLAS, CUKR_INT, CUKR_DOUBLE, CUKR_SINGLE*, CUKR_SINGLE*,
													 CUKR_SINGLE*, CUKR_SINGLE*);

/**
 * \brief Wrapper for SAYPX 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */
void CukrBlasSAYPX(CUKR_BLAS, CUKR_INT, CUKR_SINGLE, CUKR_SINGLE*, CUKR_SINGLE*);

#ifdef CUKR_USE_OPENCL
/**
 * \brief Wrapper for DAYPX 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */
void CukrCLBlasDAYPX(CUKR_BLAS, CUKR_INT, CUKR_DOUBLE, CUKR_CL_DOUBLE*, CUKR_CL_DOUBLE*); 

/**
 * \brief Wrapper for QAYPX 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
void CukrCLBlasQAYPX(CUKR_BLAS, CUKR_INT, CUKR_DOUBLE, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*,
													   CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);

/**
 * \brief Wrapper for SAYPX 
 * \param blas The BLAS structure
 * \param n The vector size
 * \param x,y The vectors 
 */
void CukrCLBlasSAYPX(CUKR_BLAS, CUKR_INT, CUKR_SINGLE, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);

#endif // CUKR_USE_OPENCL

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSCAL
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param a The scaling factor
 * \param x The vector 
 */
void CukrBlasDSCAL(CUKR_BLAS, CUKR_INT, CUKR_DOUBLE, CUKR_DOUBLE*); 

/**
 * \brief Wrapper for QSCAL
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param a The scaling factor
 * \param xh,xt The head and tail of the
 *              vector 
 */
void CukrBlasQSCAL(CUKR_BLAS, CUKR_INT, CUKR_DOUBLE, CUKR_SINGLE*,
                                 					 CUKR_SINGLE*); 

/**
 * \brief Wrapper for SSCAL
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param a The scaling factor
 * \param x The vector 
 */
void CukrBlasSSCAL(CUKR_BLAS, CUKR_INT, CUKR_SINGLE, CUKR_SINGLE*);

#ifdef CUKR_USE_OPENCL
/**
 * \brief Wrapper for DSCAL
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param a The scaling factor
 * \param x The vector 
 */
void CukrCLBlasDSCAL(CUKR_BLAS, CUKR_INT, CUKR_DOUBLE, CUKR_CL_DOUBLE*); 

/**
 * \brief Wrapper for QSCAL
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param a The scaling factor
 * \param xh,xt The head and tail of the
 *              vector 
 */
void CukrCLBlasQSCAL(CUKR_BLAS, CUKR_INT, CUKR_DOUBLE, CUKR_CL_SINGLE*,
													  CUKR_CL_SINGLE*); 

/**
 * \brief Wrapper for SSCAL
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param a The scaling factor
 * \param x The vector 
 */
void CukrCLBlasSSCAL(CUKR_BLAS, CUKR_INT, CUKR_SINGLE, CUKR_CL_SINGLE*);

#endif // CUKR_USE_OPENCL

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DCOPY 
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param x The source vector
 * \param y The destination vector 
 */
void CukrBlasDCOPY(CUKR_BLAS, CUKR_INT, CUKR_DOUBLE*, CUKR_DOUBLE*);

/**
 * \brief Wrapper for QCOPY 
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
void CukrBlasQCOPY(CUKR_BLAS, CUKR_INT, CUKR_SINGLE*, CUKR_SINGLE*,
										CUKR_SINGLE*, CUKR_SINGLE*);

/**
 * \brief Wrapper for SCOPY 
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param x The source vector
 * \param y The destination vector 
 */ 
void CukrBlasSCOPY(CUKR_BLAS, CUKR_INT, CUKR_SINGLE*, CUKR_SINGLE*);

#ifdef CUKR_USE_OPENCL
/**
 * \brief Wrapper for DCOPY 
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param x The source vector
 * \param y The destination vector 
 */
void CukrCLBlasDCOPY(CUKR_BLAS, CUKR_INT, CUKR_CL_DOUBLE*, CUKR_CL_DOUBLE*);

/**
 * \brief Wrapper for QCOPY 
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
void CukrCLBlasQCOPY(CUKR_BLAS, CUKR_INT, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);

/**
 * \brief Wrapper for SCOPY 
 * \param blas The BLAS structure
 * \param n The vectors size
 * \param x The source vector
 * \param y The destination vector 
 */ 
void CukrCLBlasSCOPY(CUKR_BLAS, CUKR_INT, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);

#endif // CUKR_USE_OPENCL

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_CSR 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrBlasDSPMV_CSR(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_INT*, CUKR_INT*, 
                       CUKR_DOUBLE*, CUKR_DOUBLE*, CUKR_DOUBLE, CUKR_DOUBLE*); 
                       
/**
 * \brief Wrapper for QSPMV_CSR 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param beta Scale factor for y
 * \param yh,xt Head and tail for the result
 *              vector
 */
void CukrBlasQSPMV_CSR(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_INT*, CUKR_INT*, 
                       CUKR_SINGLE*, CUKR_SINGLE*, CUKR_SINGLE*, CUKR_SINGLE*,
                       CUKR_DOUBLE, CUKR_SINGLE*, CUKR_SINGLE*);                        
                   
/**
 * \brief Wrapper for SSPMV_CSR 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrBlasSSPMV_CSR(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_SINGLE, CUKR_INT*, CUKR_INT*, 
                       CUKR_SINGLE*, CUKR_SINGLE*, CUKR_SINGLE, CUKR_SINGLE*);
                   
                                              
// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_CSR 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrBlasDSPMV_CSR4(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_INT*, CUKR_INT*, 
                        CUKR_DOUBLE*, CUKR_DOUBLE*, CUKR_DOUBLE, CUKR_DOUBLE*); 
                       
/**
 * \brief Wrapper for QSPMV_CSR 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param beta Scale factor for y
 * \param yh,xt Head and tail for the result
 *              vector
 */
void CukrBlasQSPMV_CSR4(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_INT*, CUKR_INT*, 
                        CUKR_SINGLE*, CUKR_SINGLE*, CUKR_SINGLE*, CUKR_SINGLE*,
                        CUKR_DOUBLE, CUKR_SINGLE*, CUKR_SINGLE*);                        
                   
/**
 * \brief Wrapper for SSPMV_CSR 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrBlasSSPMV_CSR4(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_SINGLE, CUKR_INT*, CUKR_INT*, 
                        CUKR_SINGLE*, CUKR_SINGLE*, CUKR_SINGLE, CUKR_SINGLE*);
                   
#ifdef CUKR_USE_OPENCL
/**
 * \brief Wrapper for DSPMV_CSR 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrCLBlasDSPMV_CSR(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_CL_INT*, CUKR_CL_INT*, 
                       CUKR_CL_DOUBLE*, CUKR_CL_DOUBLE*, CUKR_DOUBLE, CUKR_CL_DOUBLE*); 

/**
 * \brief Wrapper for QSPMV_CSR 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param beta Scale factor for y
 * \param yh,xt Head and tail for the result
 *              vector
 */
void CukrCLBlasQSPMV_CSR(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_CL_INT*, CUKR_CL_INT*, 
                       CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*,
                       CUKR_DOUBLE, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);                        

/**
 * \brief Wrapper for SSPMV_CSR 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrCLBlasSSPMV_CSR(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_SINGLE, CUKR_CL_INT*, CUKR_CL_INT*, 
                       CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_SINGLE, CUKR_CL_SINGLE*);


// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_CSR 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrCLBlasDSPMV_CSR4(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_CL_INT*, CUKR_CL_INT*, 
                        CUKR_CL_DOUBLE*, CUKR_CL_DOUBLE*, CUKR_DOUBLE, CUKR_CL_DOUBLE*); 

/**
 * \brief Wrapper for QSPMV_CSR 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param beta Scale factor for y
 * \param yh,xt Head and tail for the result
 *              vector
 */
void CukrCLBlasQSPMV_CSR4(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_CL_INT*, CUKR_CL_INT*, 
                        CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*,
                        CUKR_DOUBLE, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);                        

/**
 * \brief Wrapper for SSPMV_CSR 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrCLBlasSSPMV_CSR4(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_SINGLE, CUKR_CL_INT*, CUKR_CL_INT*, 
                        CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_SINGLE, CUKR_CL_SINGLE*);


#endif // CUKR_USE_OPENCL

// ------------------------------------------------------------------- //

/* Yusa adds BCSR */

/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Constant
 * \param y Result vector
 */
void CukrBlasDSPMV_BCSR(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
					    CUKR_DOUBLE alpha, CUKR_INT* ptr, CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x,
                        CUKR_DOUBLE beta, CUKR_DOUBLE* y);


/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector (head,tail)
 * \param x Vector being multiplied (head,tail)
 * \param beta Constant
 * \param y Result vector (head,tail)
 */
void CukrBlasQSPMV_BCSR(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
					    CUKR_DOUBLE alpha, CUKR_INT* ptr, CUKR_INT* idx, CUKR_SINGLE* valh, CUKR_SINGLE* valt,
					    CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt);


/**
 * \brief Sparse matrix vector multiply in
 *        BCSR format, in single precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Constant
 * \param y Result vector
 */
void CukrBlasSSPMV_BCSR(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
						CUKR_SINGLE alpha, CUKR_INT* ptr, CUKR_INT* idx, CUKR_SINGLE* val, CUKR_SINGLE* x,
			    	    CUKR_SINGLE beta, CUKR_SINGLE* y);

#ifdef CUKR_USE_OPENCL
/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Constant
 * \param y Result vector
 */
void CukrCLBlasDSPMV_BCSR(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
					    CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, CUKR_CL_INT* idx, CUKR_CL_DOUBLE* val, CUKR_CL_DOUBLE* x,
                        CUKR_DOUBLE beta, CUKR_CL_DOUBLE* y);


/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector (head,tail)
 * \param x Vector being multiplied (head,tail)
 * \param beta Constant
 * \param y Result vector (head,tail)
 */
void CukrCLBlasQSPMV_BCSR(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
					    CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, CUKR_CL_INT* idx, CUKR_CL_SINGLE* valh, CUKR_CL_SINGLE* valt,
					    CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt, CUKR_DOUBLE beta, CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt);


/**
 * \brief Sparse matrix vector multiply in
 *        BCSR format, in single precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Constant
 * \param y Result vector
 */
void CukrCLBlasSSPMV_BCSR(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
						CUKR_SINGLE alpha, CUKR_CL_INT* ptr, CUKR_CL_INT* idx, CUKR_CL_SINGLE* val, CUKR_CL_SINGLE* x,
			    	    CUKR_SINGLE beta, CUKR_CL_SINGLE* y);

#endif // CUKR_USE_OPENCL
// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR4 format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param y Result vector
 */
void CukrBlasDSPMV_BCSR4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
					    CUKR_DOUBLE alpha, CUKR_INT* ptr, CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x,
                        CUKR_DOUBLE beta, CUKR_DOUBLE* y);

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR4 format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector (head,tail)
 * \param x Vector being multiplied (head,tail)
 * \param beta Constant
 * \param y Result vector (head,tail)
 */
void CukrBlasQSPMV_BCSR4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
					    CUKR_DOUBLE alpha, CUKR_INT* ptr, CUKR_INT* idx, CUKR_SINGLE* valh, CUKR_SINGLE* valt,
					    CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt);

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR4 format, in single precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Constant
 * \param y Result vector
 */
void CukrBlasSSPMV_BCSR4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
						CUKR_SINGLE alpha, CUKR_INT* ptr, CUKR_INT* idx, CUKR_SINGLE* val, CUKR_SINGLE* x,
			    	    CUKR_SINGLE beta, CUKR_SINGLE* y);

#ifdef CUKR_USE_OPENCL
/**
 * \brief Sparse matrix vector multiply in
 *        BCSR4 format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param y Result vector
 */
void CukrCLBlasDSPMV_BCSR4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
						 CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, CUKR_CL_INT* idx, CUKR_CL_DOUBLE* val, CUKR_CL_DOUBLE* x,
						 CUKR_DOUBLE beta, CUKR_CL_DOUBLE* y);

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR4 format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector (head,tail)
 * \param x Vector being multiplied (head,tail)
 * \param beta Constant
 * \param y Result vector (head,tail)
 */
void CukrCLBlasQSPMV_BCSR4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
						 CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, CUKR_CL_INT* idx, CUKR_CL_SINGLE* valh, CUKR_CL_SINGLE* valt,
						 CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt, CUKR_DOUBLE beta, CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt);

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR4 format, in single precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Constant
 * \param y Result vector
 */
void CukrCLBlasSSPMV_BCSR4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
						 CUKR_SINGLE alpha, CUKR_CL_INT* ptr, CUKR_CL_INT* idx, CUKR_CL_SINGLE* val, CUKR_CL_SINGLE* x,
						 CUKR_SINGLE beta, CUKR_CL_SINGLE* y);
#endif
// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrBlasDSPMV_HYB(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_INT, CUKR_INT, CUKR_INT,
                       CUKR_INT*, CUKR_DOUBLE*, CUKR_INT*, CUKR_INT*, CUKR_DOUBLE*,
                       CUKR_DOUBLE*, CUKR_DOUBLE, CUKR_DOUBLE*);

/**
 * \brief Wrapper for QSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
void CukrBlasQSPMV_HYB(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_INT, CUKR_INT, CUKR_INT,
                       CUKR_INT*, CUKR_SINGLE*, CUKR_SINGLE*, CUKR_INT*, CUKR_INT*, CUKR_SINGLE*, CUKR_SINGLE*,
                       CUKR_SINGLE*, CUKR_SINGLE*, CUKR_DOUBLE, CUKR_SINGLE*, CUKR_SINGLE*);



/**
 * \brief Wrapper for SSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vectors
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrBlasSSPMV_HYB(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_SINGLE, CUKR_INT, CUKR_INT, CUKR_INT,
                       CUKR_INT*, CUKR_SINGLE*, CUKR_INT*, CUKR_INT*, CUKR_SINGLE*,
                       CUKR_SINGLE*, CUKR_SINGLE, CUKR_SINGLE*);

#ifdef CUKR_USE_OPENCL
/**
 * \brief Wrapper for DSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrCLBlasDSPMV_HYB(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_INT, CUKR_INT, CUKR_INT,
                       CUKR_CL_INT*, CUKR_CL_DOUBLE*, CUKR_CL_INT*, CUKR_CL_INT*, CUKR_CL_DOUBLE*,
                       CUKR_CL_DOUBLE*, CUKR_DOUBLE, CUKR_CL_DOUBLE*);

/**
 * \brief Wrapper for QSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
void CukrCLBlasQSPMV_HYB(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_DOUBLE, CUKR_INT, CUKR_INT, CUKR_INT,
                       CUKR_CL_INT*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_CL_INT*, CUKR_CL_INT*, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*,
                       CUKR_CL_SINGLE*, CUKR_CL_SINGLE*, CUKR_DOUBLE, CUKR_CL_SINGLE*, CUKR_CL_SINGLE*);



/**
 * \brief Wrapper for SSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vectors
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrCLBlasSSPMV_HYB(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_SINGLE, CUKR_INT, CUKR_INT, CUKR_INT,
                       CUKR_CL_INT*, CUKR_CL_SINGLE*, CUKR_CL_INT*, CUKR_CL_INT*, CUKR_CL_SINGLE*,
                       CUKR_CL_SINGLE*, CUKR_SINGLE, CUKR_CL_SINGLE*);

#endif

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr_hi/lo CSR ptr vectors
 * \param csr_idx CSR idx vectors
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrBlasDSPMV_BHYB(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
                       CUKR_INT* bell_idx, CUKR_DOUBLE* bell_val, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_DOUBLE* bcsr_val,
                       CUKR_DOUBLE* x, CUKR_DOUBLE beta, CUKR_DOUBLE* y);

/**
 * \brief Wrapper for QSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
void CukrBlasQSPMV_BHYB(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
	                   CUKR_INT* bell_idx, CUKR_SINGLE* bell_valh, CUKR_SINGLE* bell_valt, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_SINGLE* bcsr_valh, CUKR_SINGLE* bcsr_valt,
	                   CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt);

/**
 * \brief Wrapper for SSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vectors
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrBlasSSPMV_BHYB(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_SINGLE, CUKR_INT, CUKR_INT, CUKR_INT,
                       CUKR_INT*, CUKR_SINGLE*, CUKR_INT*, CUKR_INT*, CUKR_SINGLE*,
                       CUKR_SINGLE*, CUKR_SINGLE, CUKR_SINGLE*);

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_BHYB4
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr_hi/lo CSR ptr vectors
 * \param csr_idx CSR idx vectors
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrBlasDSPMV_BHYB4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
                       CUKR_INT* bell_idx, CUKR_DOUBLE* bell_val, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_DOUBLE* bcsr_val,
                       CUKR_DOUBLE* x, CUKR_DOUBLE beta, CUKR_DOUBLE* y);

/**
 * \brief Wrapper for QSPMV_BHYB4
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
void CukrBlasQSPMV_BHYB4(CUKR_BLAS blas, CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
	                   CUKR_INT* bell_idx, CUKR_SINGLE* bell_valh, CUKR_SINGLE* bell_valt, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_SINGLE* bcsr_valh, CUKR_SINGLE* bcsr_valt,
	                   CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt);

/**
 * \brief Wrapper for SSPMV_BHYB4
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vectors
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void CukrBlasSSPMV_BHYB4(CUKR_BLAS, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_INT, CUKR_SINGLE, CUKR_INT, CUKR_INT, CUKR_INT,
                       CUKR_INT*, CUKR_SINGLE*, CUKR_INT*, CUKR_INT*, CUKR_SINGLE*,
                       CUKR_SINGLE*, CUKR_SINGLE, CUKR_SINGLE*);

// ------------------------------------------------------------------- //

#endif /*BLAS_H_*/
