/**
 * @file blas/impl/generic/generic_blas.c
 * @author Serban Georgescu
 * @since 2008
 *  
 * This defines the "generic blas", with
 * all functions written here. It does not
 * use any libraries. Used for control.
 *
 * Yusa adds the BCSR SPMV.
 */

#define LOGGER_NAME "CUKr.blas.impl.generic_blas"
 
#include <stdio.h>
#include <stdlib.h> 
#include <math.h>
#include "generic_blas.h"
#include "../../../util/log.h"

// ====================== FORWARD DECLARATIONS ======================= //

void dsadd(volatile float *c0, volatile float *c1, const volatile float a0, const volatile float a1, 
                                				   const volatile float b0, const volatile float b1);
void dssub(volatile float *c0, volatile float *c1, const volatile float a0, const volatile float a1, 
                                				   const volatile float b0, const volatile float b1);                     
void dsmul(volatile float *c0, volatile float *c1, const volatile float a0, const volatile float a1, 
                                                   const volatile float b0, const volatile float b1);
                                

// =================================================================== //


// ------------------------------------------------------------------- //

/*
 * \brief Dot product in double precision
 * \param n Vector size
 * \param x,y The two vectors whose product 
 *            is being computed 
 */
CUKR_DOUBLE generic_ddot(CUKR_INT n, CUKR_DOUBLE* x, CUKR_DOUBLE* y)
{
	int n_4 = 4 * (int)(n / 4);
	int i;
	CUKR_DOUBLE s[4];
	
	//return 0;
	
	/* Init sums to zero */
	s[0] = 0; s[1] = 0; s[2] = 0; s[3] = 0; 
	
	/* 4-unrolling */
	for (i = 0; i < n_4 ; i += 4)
	{ 
		s[0] += x[i]	 * y[i];
		s[1] += x[i + 1] * y[i + 1];
		s[2] += x[i + 2] * y[i + 2];
		s[3] += x[i + 3] * y[i + 3];
	}
	 
	/* The rest */
	for (i = n_4; i < n; i++)
	  s[0] += x[i] * y[i];
	
	return s[0] + s[1] + s[2] + s[3];   
}

/**
 * \brief Dot product in quasi-double precision
 * \param n Vector size
 * \param n The vector size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
CUKR_DOUBLE generic_qdot(CUKR_INT n, CUKR_SINGLE* xh, CUKR_SINGLE* xt,
									 CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	CUKR_DOUBLE dval;
	CUKR_SINGLE volatile sum0 = 0, sum1 = 0;
	CUKR_SINGLE volatile c0, c1;
	int i;
	
	
	/* 
	 * Compute the sum 
	 */
	 
	for (i = 0; i < n; i++)
	{
		/* c = x(i)*y(i) */
		dsmul(&c0, &c1, xh[i], xt[i], yh[i], yt[i]);
		 
		/* s += c */
		dsadd(&sum0, &sum1, c0, c1, sum0, sum1);
	}
	
	/*
	 * Recombine and return
	 */
	 
	dval = (double)sum0 + (double)sum1;
	return dval; 
}									 

/*
 * \brief Dot product in single precision
 * \param n Vector size
 * \param x,y The two vectors whose product 
 *            is being computed 
 */
CUKR_SINGLE generic_sdot(CUKR_INT n, CUKR_SINGLE* x, CUKR_SINGLE* y)
{
	int n_4 = 4 * (int)(n / 4);
	int i;
	CUKR_SINGLE s = 0;

	//printf("value 10 on cpu x:%f y:%f\n", x[10], y[10]);
	/* 4-unrolling */
	for (i = 0; i < n_4 ; i += 4)
	{ 
		s += x[i]     * y[i];
		s += x[i + 1] * y[i + 1];
		s += x[i + 2] * y[i + 2];
		s += x[i + 3] * y[i + 3];
	}
	 
	/* The rest */
	for (i = n_4; i < n; i++)
	  s += x[i] * y[i];
	
	//printf("cpu res: %f\n", s);
	return s; 
}

// ------------------------------------------------------------------- //
  
/*
 * \brief AXPY in double precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
void generic_daxpy(CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x, CUKR_DOUBLE* y)
{
	int n_4 = 4 * (int)(n / 4);
	int i;
	
	/* 4-unrolling */
	for (i = 0; i < n_4 ; i += 4)
	{ 
		y[i]     = a*x[i]     + y[i]; 
		y[i + 1] = a*x[i + 1] + y[i + 1];
		y[i + 2] = a*x[i + 2] + y[i + 2];
		y[i + 3] = a*x[i + 3] + y[i + 3];
	}
	 
	/* The rest */
	for (i = n_4; i < n; i++) 
	  y[i] = a*x[i] + y[i];	 
} 

/**
 * \brief AXPY in quasi-double precision
 * \param n Vector size
 * \param a The constant
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector 
 */
void generic_qaxpy(CUKR_INT n, CUKR_DOUBLE a, CUKR_SINGLE* xh, CUKR_SINGLE* xt, 
											  CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	int i;
	CUKR_SINGLE volatile a0, a1, c0, c1;
	
	/* 
	 * Split a 
	 */
	 
	a0 = (CUKR_SINGLE)a;
	a1 = (CUKR_SINGLE)(a - a0);
	
	/* 
	 * Do the update 
	 */
	 
	for (i = 0; i < n; i++)
	{
		/* c = ax(i) */
		dsmul(&c0, &c1, a0, a1, xh[i], xt[i]);
		/* y(i) += c */
		dsadd(&yh[i], &yt[i], c0, c1, yh[i], yt[i]);
	} 
}
											  	

/*
 * \brief AXPY in single precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
void generic_saxpy(CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x, CUKR_SINGLE* y)
{
	int n_4 = 4 * (int)(n / 4);
	int i;
	
	/* 4-unrolling */
	for (i = 0; i < n_4 ; i += 4)
	{ 
		y[i]     = a*x[i]     + y[i]; 
		y[i + 1] = a*x[i + 1] + y[i + 1];
		y[i + 2] = a*x[i + 2] + y[i + 2];
		y[i + 3] = a*x[i + 3] + y[i + 3];
	}
	 
	/* The rest */
	for (i = n_4; i < n; i++)
	  y[i] = a*x[i] + y[i]; 
}


// ------------------------------------------------------------------- //

/*
 * \brief AYPX in double precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
void generic_daypx(CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x, CUKR_DOUBLE* y)
{
	int n_4 = 4 * (int)(n / 4);
	int i;
	
	/* 4-unrolling */
	for (i = 0; i < n_4 ; i += 4)
	{ 
		y[i]     = x[i]     + a*y[i]; 
		y[i + 1] = x[i + 1] + a*y[i + 1];
		y[i + 2] = x[i + 2] + a*y[i + 2];
		y[i + 3] = x[i + 3] + a*y[i + 3];
	}
	 
	/* The rest */
	for (i = n_4; i < n; i++)
	  y[i] = x[i] + a*y[i]; 
} 

/**
 * \brief AYPX in quasi-double precision
 * \param n Vector size
 * \param a The constant
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector 
 */
void generic_qaypx(CUKR_INT n, CUKR_DOUBLE a, CUKR_SINGLE* xh, CUKR_SINGLE* xt, 
											  CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	int i;
	CUKR_SINGLE volatile a0, a1, c0, c1;
	
	/* 
	 * Split a 
	 */
	 
	a0 = (CUKR_SINGLE)a;
	a1 = (CUKR_SINGLE)(a - a0);
	
	/* 
	 * Do the update 
	 */
	 
	for (i = 0; i < n; i++)
	{
		/* c = ay(i) */
		dsmul(&c0, &c1, a0, a1, yh[i], yt[i]);
		/* y(i) = x(i) + c */
		dsadd(&yh[i], &yt[i], c0, c1, xh[i], xt[i]);
	} 
}

/*
 * \brief AYPX in single precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
void generic_saypx(CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x, CUKR_SINGLE* y)
{
	int n_4 = 4 * (int)(n / 4);
	int i;
	
	/* 4-unrolling */
	for (i = 0; i < n_4 ; i += 4)
	{ 
		y[i]     = x[i]     + a*y[i]; 
		y[i + 1] = x[i + 1] + a*y[i + 1];
		y[i + 2] = x[i + 2] + a*y[i + 2];
		y[i + 3] = x[i + 3] + a*y[i + 3];
	}
	 
	/* The rest */
	for (i = n_4; i < n; i++)
	  y[i] = x[i] + a*y[i]; 
}


// ------------------------------------------------------------------- //

/*
 * \brief Vector scaling in double precision
 * \param n Vector size
 * \param a Scaling factor
 * \param x The vector being scaled
 */
void generic_dscal(CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x)
{
	int n_4 = 4 * (int)(n / 4);
	int i;
	
	/* 4-unrolling */
	for (i = 0; i < n_4 ; i += 4)
	{ 
		x[i]     = a*x[i]; 
		x[i + 1] = a*x[i + 1];
		x[i + 2] = a*x[i + 2];
		x[i + 3] = a*x[i + 3];
	}
	 
	/* The rest */
	for (i = n_4; i < n; i++)
	  x[i] = a*x[i]; 
} 

/**
 * \brief Vector scaling in quasi-double 
 *        precision
 * \param n Vector size
 * \param a Scaling factor
 * \param xh,xt The head and tail of the
 *              vector 
 */
void generic_qscal(CUKR_INT n, CUKR_DOUBLE a, CUKR_SINGLE* xh,
											  CUKR_SINGLE* xt)
{
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	int i;
	CUKR_SINGLE volatile a0, a1;
	
	/* 
	 * Split a 
	 */
	 
	a0 = (CUKR_SINGLE)a;
	a1 = (CUKR_SINGLE)(a - a0);
	
	/* 
	 * Do the update 
	 */
	 
	for (i = 0; i < n; i++)
	{
		/* x(i) = ax(i) */
		dsmul(&xh[i], &xt[i], a0, a1, xh[i], xt[i]);
	} 	
}											  

/*
 * \brief Vector scaling in single precision
 * \param n Vector size
 * \param a Scaling factor
 * \param x The vector being scaled
 */
void generic_sscal(CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x)
{
	int n_4 = 4 * (int)(n / 4);
	int i;
	
	/* 4-unrolling */
	for (i = 0; i < n_4 ; i += 4)
	{ 
		x[i]     = a*x[i]; 
		x[i + 1] = a*x[i + 1];
		x[i + 2] = a*x[i + 2];
		x[i + 3] = a*x[i + 3];
	}
	 
	/* The rest */
	for (i = n_4; i < n; i++)
	  x[i] = a*x[i]; 
}  

// ------------------------------------------------------------------- //

/*
 * \brief Vector copy in double precision
 * \param n Vector size
 * \param x The source vector
 * \param y The destination vector
 */
void generic_dcopy(CUKR_INT n, CUKR_DOUBLE* x, CUKR_DOUBLE* y)
{
	int n_4 = 4 * (int)(n / 4);
	int i;
	
	/* 4-unrolling */
	for (i = 0; i < n_4 ; i += 4)
	{ 
		y[i]     = x[i]; 
		y[i + 1] = x[i + 1];
		y[i + 2] = x[i + 2];
		y[i + 3] = x[i + 3];
	}
	 
	/* The rest */
	for (i = n_4; i < n; i++)
	  y[i] = x[i]; 
}  

/**
 * \brief Vector copy in quasi-double 
 *        precision
 * \param n Vector size
 * \param xh,xt The head and tail of the
 *              source vector 
 * \param yh,yt The head and tail of the
 *              destination vector 
 */
void generic_qcopy(CUKR_INT n, CUKR_SINGLE* xh, CUKR_SINGLE* xt, 
							   CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	int n_2 = 2 * (int)(n / 2);
	int i;
	
	/* 2-unrolling */
	for (i = 0; i < n_2 ; i += 2)
	{ 
		yh[i]     = xh[i]; 
		yt[i]     = xt[i];
		yh[i + 1] = xh[i + 1];
		yt[i + 1] = xt[i + 1];
	}
	 
	/* The rest */
	for (i = n_2; i < n; i++) 
	{
	  yh[i] = xh[i]; 
	  yt[i] = xt[i];
	} 	
}							   

/*
 * \brief Vector copy in single precision
 * \param n Vector size
 * \param x The source vector
 * \param y The destination vector
 */
void generic_scopy(CUKR_INT n, CUKR_SINGLE* x, CUKR_SINGLE* y)
{
	int n_4 = 4 * (int)(n / 4);
	int i;
	
	/* 4-unrolling */
	for (i = 0; i < n_4 ; i += 4)
	{ 
		y[i]     = x[i]; 
		y[i + 1] = x[i + 1];
		y[i + 2] = x[i + 2];
		y[i + 3] = x[i + 3];
	}
	 
	/* The rest */
	for (i = n_4; i < n; i++)
	  y[i] = x[i]; 
}   

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
void generic_dspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
                       CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x, CUKR_DOUBLE beta, 
            		   CUKR_DOUBLE* y)
{
   CUKR_DOUBLE volatile sum;
   CUKR_DOUBLE *pval = val;
   CUKR_INT jb, je;
   CUKR_DOUBLE volatile yy, t, c;
   int i, j;
     
   if (beta != 0)
   {
   	   if (beta != 1)
   			generic_dscal(rows, beta, y); 
   			
   	   if (alpha == 1) 
   	   {
   	   	   pval = val;
		   for (i = 0; i != rows; i++) 
		   {
		     jb = ptr[i];
		     je = ptr[i + 1];
		  
		     sum = 0; 
		     jb = ptr[i];
		     je = ptr[i + 1];
		     for (j = jb; j != je; j++)
		     	sum += x[idx[j - 1] - 1] * (*pval++);
		     y[i] += sum;
		   }	
   	   }
   	   else 
   	   {
   	   	   pval = val;
		   for (i = 0; i != rows; i++) 
		   {
		     jb = ptr[i];
		     je = ptr[i + 1];
		  
		     sum = 0; 
		     jb = ptr[i];
		     je = ptr[i + 1];
		     for (j = jb; j != je; j++) {
		     	sum += x[idx[j - 1] - 1] * (*pval++);		     	
		     }
		     y[i] += alpha *sum;
		   }	
   	   }		  	  
   }
   else 
   {
   	   if (alpha == 1) 
   	   {
   	   	   pval = val;
		   for (i = 0; i != rows; i++) 
		   {
		     jb = ptr[i];
		     je = ptr[i + 1];
		  
		     sum = 0; 
		     jb = ptr[i];
		     je = ptr[i + 1];
		     for (j = jb; j != je; j++) {
		     	sum += x[idx[j - 1] - 1] * (*pval++);
		     }
		     y[i] = sum;
		   }	
   	   }
   	   else 
   	   {
   	   	   pval = val;
		   for (i = 0; i != rows; i++) 
		   {
		     jb = ptr[i];
		     je = ptr[i + 1];
		  
		     sum = 0; 
		     jb = ptr[i];
		     je = ptr[i + 1];
		     for (j = jb; j != je; j++)
		     	sum += x[idx[j - 1] - 1] * (*pval++);
		     y[i] = alpha *sum;
		   }	
   	   }		  	
   }
   
}

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param beta Scale factor for y
 * \param yh,xt Head and tail for the result
 *              vector
 */ 
void generic_qspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
               		   CUKR_INT* idx, CUKR_SINGLE* valh, CUKR_SINGLE* valt, 
               		   CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, 
               		   CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{	
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	CUKR_SINGLE volatile alpha0, alpha1, beta0, beta1;
	CUKR_SINGLE volatile sum0, sum1, c0, c1;
    CUKR_INT jb, je;
    int i, j;
    CUKR_SINGLE *pval0 = valh, *pval1 = valt;
    
	
	/* 
	 * Split alpha & beta 
	 */
	 
	alpha0 = (CUKR_SINGLE)alpha;
	alpha1 = (CUKR_SINGLE)(alpha - alpha0);
	
	beta0 = (CUKR_SINGLE)beta;
	beta1 = (CUKR_SINGLE)(beta - beta0);
	
	/* 
	 * Do SpMV
	 */
	
	if (beta && beta != 1)
   		generic_qscal(rows, beta, yh, yt);
	
	for (i = 0; i != rows; i++) 
    {
     	sum0 = sum1 = 0; 
     	jb = ptr[i];
     	je = ptr[i + 1];
     	
     	for (j = jb; j != je; j++) {	
     		/* c = x[idx[j - 1] - 1] * (*pval++) */
     		dsmul(&c0, &c1, xh[idx[j - 1] - 1], xt[idx[j - 1] - 1], (*pval0++), (*pval1++));
     		/* s += c */
       		dsadd(&sum0, &sum1, c0, c1, sum0, sum1);
     	}
     	/* c = alpha * sum */
     	dsmul(&c0, &c1, alpha0, alpha1, sum0, sum1);
     	if (beta != 0) {
     		/* y += c */
     		dsadd(&yh[i], &yt[i], c0, c1, yh[i], yt[i]);
     	}
     	else {
     		/* y = c */ 
     		yh[i] = c0;
     		yt[i] = c1;
     	}
    } 
}
               		   	  

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in single precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
void generic_sspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_INT* ptr, 
           			   CUKR_INT* idx, CUKR_SINGLE* val, CUKR_SINGLE* x, CUKR_SINGLE beta, 
          		       CUKR_SINGLE* y)
{
   CUKR_SINGLE volatile sum;
   CUKR_SINGLE *pval = val;
   CUKR_INT jb, je;
   CUKR_SINGLE volatile yy, t, c;
   int i, j;
   
   if (beta != 0)
   {
   	   if (beta != 1)
   			generic_sscal(rows, beta, y); 
   			
   	   if (alpha == 1) 
   	   {
   	   	   pval = val;
		   for (i = 0; i != rows; i++) 
		   {
		     jb = ptr[i];
		     je = ptr[i + 1];
		  
		     sum = 0; 
		     jb = ptr[i];
		     je = ptr[i + 1];
		     for (j = jb; j != je; j++)
		     	sum += x[idx[j - 1] - 1] * (*pval++);
		     y[i] += sum;		     
		   }	
   	   }
   	   else 
   	   {
   	   	   pval = val;
		   for (i = 0; i != rows; i++) 
		   {
		     jb = ptr[i];
		     je = ptr[i + 1];
		  
		     sum = 0; 
		     jb = ptr[i];
		     je = ptr[i + 1];
		     for (j = jb; j != je; j++)
		     	sum += x[idx[j - 1] - 1] * (*pval++);
		     y[i] += alpha *sum;
		   }	
   	   }		  	  
   }
   else 
   {
   	   if (alpha == 1) 
   	   {
   	   	   volatile float c = 0, yy, t;	
   	   	   pval = val;
		   for (i = 0; i != rows; i++) 
		   {
		     jb = ptr[i];
		     je = ptr[i + 1];
		  
		     sum = 0; 
		     jb = ptr[i];
		     je = ptr[i + 1];
		     for (j = jb; j != je; j++) {
		     	float prod = x[idx[j - 1] - 1] * (*pval++);
		     	yy = prod - c ;
		     	t = sum + yy;
		     	c = (t - sum) - yy;
		     	sum = t;
		     }
		     y[i] = sum;	     
		   }
   	   }
   	   else 
   	   {
   	   	   pval = val;
		   for (i = 0; i != rows; i++) 
		   {
		     jb = ptr[i];
		     je = ptr[i + 1];
		  
		     sum = 0; 
		     jb = ptr[i];
		     je = ptr[i + 1];
		     for (j = jb; j != je; j++)
		     	sum += x[idx[j - 1] - 1] * (*pval++);
		     y[i] = alpha *sum;
		   }	
   	   }		  	
   }
   
}

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzero blocks
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param y Result vector
 */
void generic_dspmv_bcsr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
					    CUKR_DOUBLE alpha, CUKR_INT* ptr, CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x,
                        CUKR_DOUBLE beta, CUKR_DOUBLE* y)
{
   /* 2x2, 3x3, 4x4, 5x5, 6x6 are suported */
   if(r != c) CUKR_FATALX("Block size (%d, %d) is NOT supported at generic_dspmv_bcsr", r, c);

   CUKR_INT jb, je;
   int i, j;
   
   /* Get num of blocks per row */
   CUKR_INT nb = (rows%r ? rows/r+1 : rows/r);
   
     
   if (beta != 0)
   {
   	   if (beta != 1)
   			generic_dscal(rows, beta, y); 
   			
   	   if (alpha == 1) 
   	   {
		   for (i = 0; i != nb; i++, y+=r) 
		   {
		     /* a row ptr */
		     jb = ptr[i] - 1;
		     je = ptr[i + 1] - 1;
		     
		     /* 2x2 to 6x6 */
		     if(r == 2){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]];
		     		/* GEMV, Block by Vector */
		     		y0 += val[0]*x0 + val[1]*x1;
		     	 	y1 += val[2]*x0 + val[3]*x1;
		     	}
		     	/* Copy register to memory */
		     	y[0] += y0; y[1] += y1;
		     	
		     }else if(r == 3){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0, y2 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1];
		     		/* GEMV, Block by Vector */
		     		y0 += val[0]*x0 + val[1]*x1 + val[2]*x2;
		     	 	y1 += val[3]*x0 + val[4]*x1 + val[5]*x2;
		     		y2 += val[6]*x0 + val[7]*x1 + val[8]*x2;
		     	}
		     	/* Copy register to memory */
		     	y[0] += y0; y[1] += y1; y[2] += y2;
		     	
		     }else if(r == 4){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3;
		     	 	y1 += val[ 4]*x0 + val[ 5]*x1 + val[ 6]*x2 + val[ 7]*x3;
		     	 	y2 += val[ 8]*x0 + val[ 9]*x1 + val[10]*x2 + val[11]*x3;
		     	 	y3 += val[12]*x0 + val[13]*x1 + val[14]*x2 + val[15]*x3;
		     	}
		     	/* Copy register to memory */
		     	y[0] += y0; y[1] += y1; y[2] += y2; y[3] += y3;
		     	
		     }else if(r == 5){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2], x4 = x[idx[j]+3];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3 + val[ 4]*x4;
		     	 	y1 += val[ 5]*x0 + val[ 6]*x1 + val[ 7]*x2 + val[ 8]*x3 + val[ 9]*x4;
		     	 	y2 += val[10]*x0 + val[11]*x1 + val[12]*x2 + val[13]*x3 + val[14]*x4;
		     	 	y3 += val[15]*x0 + val[16]*x1 + val[17]*x2 + val[18]*x3 + val[19]*x4;
		     	 	y4 += val[20]*x0 + val[21]*x1 + val[22]*x2 + val[23]*x3 + val[24]*x4;
		     	}
		     	/* Copy register to memory */
		     	y[0] += y0; y[1] += y1; y[2] += y2; y[3] += y3; y[4] += y4;
		     	
		     }else if(r == 6){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 = 0.0, y5 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2], x4 = x[idx[j]+3], x5 = x[idx[j]+4];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3 + val[ 4]*x4 + val[ 5]*x5;
		     		y1 += val[ 6]*x0 + val[ 7]*x1 + val[ 8]*x2 + val[ 9]*x3 + val[10]*x4 + val[11]*x5;
		     		y2 += val[12]*x0 + val[13]*x1 + val[14]*x2 + val[15]*x3 + val[16]*x4 + val[17]*x5;
		     		y3 += val[18]*x0 + val[19]*x1 + val[20]*x2 + val[21]*x3 + val[22]*x4 + val[23]*x5;
		     		y4 += val[24]*x0 + val[25]*x1 + val[26]*x2 + val[27]*x3 + val[28]*x4 + val[29]*x5;
		     		y5 += val[30]*x0 + val[31]*x1 + val[32]*x2 + val[33]*x3 + val[34]*x4 + val[35]*x5;
		     	}
		     	/* Copy register to memory */
		     	y[0] += y0; y[1] += y1; y[2] += y2; y[3] += y3; y[4] += y4; y[5] += y5;
		     	
		     }else{
		     	CUKR_FATALX("Block size (%d, %d) is NOT supported at generic_dspmv_bcsr", r, c);
		     }
		   }
   	   }
   	   else 
   	   {
		   for (i = 0; i != nb; i++, y+=r) 
		   {
		     /* a row ptr */
		     jb = ptr[i] - 1;
		     je = ptr[i + 1] - 1;
		     
		     /* 2x2 to 6x6 */
		     if(r == 2){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]];
		     		/* GEMV, Block by Vector */
		     		y0 += val[0]*x0 + val[1]*x1;
		     	 	y1 += val[2]*x0 + val[3]*x1;
		     	}
		     	/* Copy register to memory */
		     	y[0] += alpha*y0; y[1] += alpha*y1;
		     	
		     }else if(r == 3){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0, y2 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1];
		     		/* GEMV, Block by Vector */
		     		y0 += val[0]*x0 + val[1]*x1 + val[2]*x2;
		     	 	y1 += val[3]*x0 + val[4]*x1 + val[5]*x2;
		     		y2 += val[6]*x0 + val[7]*x1 + val[8]*x2;
		     	}
		     	/* Copy register to memory */
		     	y[0] += alpha*y0; y[1] += alpha*y1; y[2] += alpha*y2;
		     	
		     }else if(r == 4){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3;
		     	 	y1 += val[ 4]*x0 + val[ 5]*x1 + val[ 6]*x2 + val[ 7]*x3;
		     	 	y2 += val[ 8]*x0 + val[ 9]*x1 + val[10]*x2 + val[11]*x3;
		     	 	y3 += val[12]*x0 + val[13]*x1 + val[14]*x2 + val[15]*x3;
		     	}
		     	/* Copy register to memory */
		     	y[0] += alpha*y0; y[1] += alpha*y1; y[2] += alpha*y2; y[3] += alpha*y3;
		     	
		     }else if(r == 5){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2], x4 = x[idx[j]+3];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3 + val[ 4]*x4;
		     	 	y1 += val[ 5]*x0 + val[ 6]*x1 + val[ 7]*x2 + val[ 8]*x3 + val[ 9]*x4;
		     	 	y2 += val[10]*x0 + val[11]*x1 + val[12]*x2 + val[13]*x3 + val[14]*x4;
		     	 	y3 += val[15]*x0 + val[16]*x1 + val[17]*x2 + val[18]*x3 + val[19]*x4;
		     	 	y4 += val[20]*x0 + val[21]*x1 + val[22]*x2 + val[23]*x3 + val[24]*x4;
		     	}
		     	/* Copy register to memory */
		     	y[0] += alpha*y0; y[1] += alpha*y1; y[2] += alpha*y2; y[3] += alpha*y3; y[4] += alpha*y4;
		     	
		     }else if(r == 6){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 = 0.0, y5 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2], x4 = x[idx[j]+3], x5 = x[idx[j]+4];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3 + val[ 4]*x4 + val[ 5]*x5;
		     		y1 += val[ 6]*x0 + val[ 7]*x1 + val[ 8]*x2 + val[ 9]*x3 + val[10]*x4 + val[11]*x5;
		     		y2 += val[12]*x0 + val[13]*x1 + val[14]*x2 + val[15]*x3 + val[16]*x4 + val[17]*x5;
		     		y3 += val[18]*x0 + val[19]*x1 + val[20]*x2 + val[21]*x3 + val[22]*x4 + val[23]*x5;
		     		y4 += val[24]*x0 + val[25]*x1 + val[26]*x2 + val[27]*x3 + val[28]*x4 + val[29]*x5;
		     		y5 += val[30]*x0 + val[31]*x1 + val[32]*x2 + val[33]*x3 + val[34]*x4 + val[35]*x5;
		     	}
		     	/* Copy register to memory */
		     	y[0] += alpha*y0; y[1] += alpha*y1; y[2] += alpha*y2; y[3] += alpha*y3; y[4] += alpha*y4; y[5] += alpha*y5;
		     	
		     }else{
		     	CUKR_FATALX("Block size (%d, %d) is NOT supported at generic_dspmv_bcsr", r, c);
		     }
		   }	
   	   }		  	  
   }
   else 
   {
   	   if (alpha == 1) 
   	   {
		   for (i = 0; i != nb; i++, y+=r) 
		   {
		     /* a row ptr */
		     jb = ptr[i] - 1;
		     je = ptr[i + 1] - 1;
		     
		     /* 2x2 to 6x6 */
		     if(r == 2){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]];
		     		/* GEMV, Block by Vector */
		     		y0 += val[0]*x0 + val[1]*x1;
		     	 	y1 += val[2]*x0 + val[3]*x1;
		     	}
		     	/* Copy register to memory */
		     	y[0] = y0; y[1] = y1;
		     	
		     }else if(r == 3){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0, y2 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1];
		     		/* GEMV, Block by Vector */
		     		y0 += val[0]*x0 + val[1]*x1 + val[2]*x2;
		     	 	y1 += val[3]*x0 + val[4]*x1 + val[5]*x2;
		     		y2 += val[6]*x0 + val[7]*x1 + val[8]*x2;
		     	}
		     	/* Copy register to memory */
		     	y[0] = y0; y[1] = y1; y[2] = y2;
		     	
		     }else if(r == 4){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3;
		     	 	y1 += val[ 4]*x0 + val[ 5]*x1 + val[ 6]*x2 + val[ 7]*x3;
		     	 	y2 += val[ 8]*x0 + val[ 9]*x1 + val[10]*x2 + val[11]*x3;
		     	 	y3 += val[12]*x0 + val[13]*x1 + val[14]*x2 + val[15]*x3;
		     	}
		     	/* Copy register to memory */
		     	y[0] = y0; y[1] = y1; y[2] = y2; y[3] = y3;
		     	
		     }else if(r == 5){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2], x4 = x[idx[j]+3];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3 + val[ 4]*x4;
		     	 	y1 += val[ 5]*x0 + val[ 6]*x1 + val[ 7]*x2 + val[ 8]*x3 + val[ 9]*x4;
		     	 	y2 += val[10]*x0 + val[11]*x1 + val[12]*x2 + val[13]*x3 + val[14]*x4;
		     	 	y3 += val[15]*x0 + val[16]*x1 + val[17]*x2 + val[18]*x3 + val[19]*x4;
		     	 	y4 += val[20]*x0 + val[21]*x1 + val[22]*x2 + val[23]*x3 + val[24]*x4;
		     	}
		     	/* Copy register to memory */
		     	y[0] = y0; y[1] = y1; y[2] = y2; y[3] = y3; y[4] = y4;
		     	
		     }else if(r == 6){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 = 0.0, y5 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2], x4 = x[idx[j]+3], x5 = x[idx[j]+4];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3 + val[ 4]*x4 + val[ 5]*x5;
		     		y1 += val[ 6]*x0 + val[ 7]*x1 + val[ 8]*x2 + val[ 9]*x3 + val[10]*x4 + val[11]*x5;
		     		y2 += val[12]*x0 + val[13]*x1 + val[14]*x2 + val[15]*x3 + val[16]*x4 + val[17]*x5;
		     		y3 += val[18]*x0 + val[19]*x1 + val[20]*x2 + val[21]*x3 + val[22]*x4 + val[23]*x5;
		     		y4 += val[24]*x0 + val[25]*x1 + val[26]*x2 + val[27]*x3 + val[28]*x4 + val[29]*x5;
		     		y5 += val[30]*x0 + val[31]*x1 + val[32]*x2 + val[33]*x3 + val[34]*x4 + val[35]*x5;
		     	}
		     	/* Copy register to memory */
		     	y[0] = y0; y[1] = y1; y[2] = y2; y[3] = y3; y[4] = y4; y[5] = y5;
		     	
		     }else{
		     	CUKR_FATALX("Block size (%d, %d) is NOT supported at generic_dspmv_bcsr", r, c);
		     }
		   }
   	   }
   	   else 
   	   {
		   for (i = 0; i != nb; i++, y+=r) 
		   {
		     /* a row ptr */
		     jb = ptr[i] - 1;
		     je = ptr[i + 1] - 1;
		     
		     /* 2x2 to 6x6 */
		     if(r == 2){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]];
		     		/* GEMV, Block by Vector */
		     		y0 += val[0]*x0 + val[1]*x1;
		     	 	y1 += val[2]*x0 + val[3]*x1;
		     	}
		     	/* Copy register to memory */
		     	y[0] = alpha*y0; y[1] = alpha*y1;
		     	
		     }else if(r == 3){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0, y2 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1];
		     		/* GEMV, Block by Vector */
		     		y0 += val[0]*x0 + val[1]*x1 + val[2]*x2;
		     	 	y1 += val[3]*x0 + val[4]*x1 + val[5]*x2;
		     		y2 += val[6]*x0 + val[7]*x1 + val[8]*x2;
		     	}
		     	/* Copy register to memory */
		     	y[0] = alpha*y0; y[1] = alpha*y1; y[2] = alpha*y2;
		     	
		     }else if(r == 4){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3;
		     	 	y1 += val[ 4]*x0 + val[ 5]*x1 + val[ 6]*x2 + val[ 7]*x3;
		     	 	y2 += val[ 8]*x0 + val[ 9]*x1 + val[10]*x2 + val[11]*x3;
		     	 	y3 += val[12]*x0 + val[13]*x1 + val[14]*x2 + val[15]*x3;
		     	}
		     	/* Copy register to memory */
		     	y[0] = alpha*y0; y[1] = alpha*y1; y[2] = alpha*y2; y[3] = alpha*y3;
		     	
		     }else if(r == 5){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2], x4 = x[idx[j]+3];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3 + val[ 4]*x4;
		     	 	y1 += val[ 5]*x0 + val[ 6]*x1 + val[ 7]*x2 + val[ 8]*x3 + val[ 9]*x4;
		     	 	y2 += val[10]*x0 + val[11]*x1 + val[12]*x2 + val[13]*x3 + val[14]*x4;
		     	 	y3 += val[15]*x0 + val[16]*x1 + val[17]*x2 + val[18]*x3 + val[19]*x4;
		     	 	y4 += val[20]*x0 + val[21]*x1 + val[22]*x2 + val[23]*x3 + val[24]*x4;
		     	}
		     	/* Copy register to memory */
		     	y[0] = alpha*y0; y[1] = alpha*y1; y[2] = alpha*y2; y[3] = alpha*y3; y[4] = alpha*y4;
		     	
		     }else if(r == 6){
		     	
		     	/* register y */
		     	register CUKR_DOUBLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 = 0.0, y5 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_DOUBLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2], x4 = x[idx[j]+3], x5 = x[idx[j]+4];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3 + val[ 4]*x4 + val[ 5]*x5;
		     		y1 += val[ 6]*x0 + val[ 7]*x1 + val[ 8]*x2 + val[ 9]*x3 + val[10]*x4 + val[11]*x5;
		     		y2 += val[12]*x0 + val[13]*x1 + val[14]*x2 + val[15]*x3 + val[16]*x4 + val[17]*x5;
		     		y3 += val[18]*x0 + val[19]*x1 + val[20]*x2 + val[21]*x3 + val[22]*x4 + val[23]*x5;
		     		y4 += val[24]*x0 + val[25]*x1 + val[26]*x2 + val[27]*x3 + val[28]*x4 + val[29]*x5;
		     		y5 += val[30]*x0 + val[31]*x1 + val[32]*x2 + val[33]*x3 + val[34]*x4 + val[35]*x5;
		     	}
		     	/* Copy register to memory */
		     	y[0] = alpha*y0; y[1] = alpha*y1; y[2] = alpha*y2; y[3] = alpha*y3; y[4] = alpha*y4; y[5] = alpha*y5;
		     	
		     }else{
		     	CUKR_FATALX("Block size (%d, %d) is NOT supported at generic_dspmv_bcsr", r, c);
		     }
		   }	
   	   }		  	  
   }
}

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzero blocks
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector (head,tail)
 * \param x Vector being multiplied (head,tail)
 * \param beta Constant
 * \param y Result vector (head,tail)
 */

void generic_qspmv_bcsr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
					    CUKR_DOUBLE alpha, CUKR_INT* ptr, CUKR_INT* idx, CUKR_SINGLE* valh, CUKR_SINGLE* valt,
					    CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	CUKR_FATAL("generic_qspmv_bcsr has not been implemented yet");
}

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR format, in single precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzero blocks
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param y Result vector
 */
void generic_sspmv_bcsr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
						CUKR_SINGLE alpha, CUKR_INT* ptr, CUKR_INT* idx, CUKR_SINGLE* val, CUKR_SINGLE* x,
			    	    CUKR_SINGLE beta, CUKR_SINGLE* y)
{
   /* 2x2, 3x3, 4x4, 5x5, 6x6 are suported */
   if(r != c) CUKR_FATALX("Block size (%d, %d) is NOT supported at generic_sspmv_bcsr", r, c);

   CUKR_INT jb, je;
   int i, j;
   
   /* Get num of blocks per row */
   CUKR_INT nb = (rows%r ? rows/r+1 : rows/r);
   
     
   if (beta != 0)
   {
   	   if (beta != 1)
   			generic_sscal(rows, beta, y); 
   			
   	   if (alpha == 1) 
   	   {
		   for (i = 0; i != nb; i++, y+=r) 
		   {
		     /* a row ptr */
		     jb = ptr[i] - 1;
		     je = ptr[i + 1] - 1;
		     
		     /* 2x2 to 6x6 */
		     if(r == 2){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]];
		     		/* GEMV, Block by Vector */
		     		y0 += val[0]*x0 + val[1]*x1;
		     	 	y1 += val[2]*x0 + val[3]*x1;
		     	}
		     	/* Copy register to memory */
		     	y[0] += y0; y[1] += y1;
		     	
		     }else if(r == 3){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0, y2 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1];
		     		/* GEMV, Block by Vector */
		     		y0 += val[0]*x0 + val[1]*x1 + val[2]*x2;
		     	 	y1 += val[3]*x0 + val[4]*x1 + val[5]*x2;
		     		y2 += val[6]*x0 + val[7]*x1 + val[8]*x2;
		     	}
		     	/* Copy register to memory */
		     	y[0] += y0; y[1] += y1; y[2] += y2;
		     	
		     }else if(r == 4){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3;
		     	 	y1 += val[ 4]*x0 + val[ 5]*x1 + val[ 6]*x2 + val[ 7]*x3;
		     	 	y2 += val[ 8]*x0 + val[ 9]*x1 + val[10]*x2 + val[11]*x3;
		     	 	y3 += val[12]*x0 + val[13]*x1 + val[14]*x2 + val[15]*x3;
		     	}
		     	/* Copy register to memory */
		     	y[0] += y0; y[1] += y1; y[2] += y2; y[3] += y3;
		     	
		     }else if(r == 5){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2], x4 = x[idx[j]+3];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3 + val[ 4]*x4;
		     	 	y1 += val[ 5]*x0 + val[ 6]*x1 + val[ 7]*x2 + val[ 8]*x3 + val[ 9]*x4;
		     	 	y2 += val[10]*x0 + val[11]*x1 + val[12]*x2 + val[13]*x3 + val[14]*x4;
		     	 	y3 += val[15]*x0 + val[16]*x1 + val[17]*x2 + val[18]*x3 + val[19]*x4;
		     	 	y4 += val[20]*x0 + val[21]*x1 + val[22]*x2 + val[23]*x3 + val[24]*x4;
		     	}
		     	/* Copy register to memory */
		     	y[0] += y0; y[1] += y1; y[2] += y2; y[3] += y3; y[4] += y4;
		     	
		     }else if(r == 6){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 = 0.0, y5 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2], x4 = x[idx[j]+3], x5 = x[idx[j]+4];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3 + val[ 4]*x4 + val[ 5]*x5;
		     		y1 += val[ 6]*x0 + val[ 7]*x1 + val[ 8]*x2 + val[ 9]*x3 + val[10]*x4 + val[11]*x5;
		     		y2 += val[12]*x0 + val[13]*x1 + val[14]*x2 + val[15]*x3 + val[16]*x4 + val[17]*x5;
		     		y3 += val[18]*x0 + val[19]*x1 + val[20]*x2 + val[21]*x3 + val[22]*x4 + val[23]*x5;
		     		y4 += val[24]*x0 + val[25]*x1 + val[26]*x2 + val[27]*x3 + val[28]*x4 + val[29]*x5;
		     		y5 += val[30]*x0 + val[31]*x1 + val[32]*x2 + val[33]*x3 + val[34]*x4 + val[35]*x5;
		     	}
		     	/* Copy register to memory */
		     	y[0] += y0; y[1] += y1; y[2] += y2; y[3] += y3; y[4] += y4; y[5] += y5;
		     	
		     }else{
		     	CUKR_FATALX("Block size (%d, %d) is NOT supported at generic_sspmv_bcsr", r, c);
		     }
		   }
   	   }
   	   else 
   	   {
		   for (i = 0; i != nb; i++, y+=r) 
		   {
		     /* a row ptr */
		     jb = ptr[i] - 1;
		     je = ptr[i + 1] - 1;
		     
		     /* 2x2 to 6x6 */
		     if(r == 2){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]];
		     		/* GEMV, Block by Vector */
		     		y0 += val[0]*x0 + val[1]*x1;
		     	 	y1 += val[2]*x0 + val[3]*x1;
		     	}
		     	/* Copy register to memory */
		     	y[0] += alpha*y0; y[1] += alpha*y1;
		     	
		     }else if(r == 3){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0, y2 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1];
		     		/* GEMV, Block by Vector */
		     		y0 += val[0]*x0 + val[1]*x1 + val[2]*x2;
		     	 	y1 += val[3]*x0 + val[4]*x1 + val[5]*x2;
		     		y2 += val[6]*x0 + val[7]*x1 + val[8]*x2;
		     	}
		     	/* Copy register to memory */
		     	y[0] += alpha*y0; y[1] += alpha*y1; y[2] += alpha*y2;
		     	
		     }else if(r == 4){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3;
		     	 	y1 += val[ 4]*x0 + val[ 5]*x1 + val[ 6]*x2 + val[ 7]*x3;
		     	 	y2 += val[ 8]*x0 + val[ 9]*x1 + val[10]*x2 + val[11]*x3;
		     	 	y3 += val[12]*x0 + val[13]*x1 + val[14]*x2 + val[15]*x3;
		     	}
		     	/* Copy register to memory */
		     	y[0] += alpha*y0; y[1] += alpha*y1; y[2] += alpha*y2; y[3] += alpha*y3;
		     	
		     }else if(r == 5){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2], x4 = x[idx[j]+3];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3 + val[ 4]*x4;
		     	 	y1 += val[ 5]*x0 + val[ 6]*x1 + val[ 7]*x2 + val[ 8]*x3 + val[ 9]*x4;
		     	 	y2 += val[10]*x0 + val[11]*x1 + val[12]*x2 + val[13]*x3 + val[14]*x4;
		     	 	y3 += val[15]*x0 + val[16]*x1 + val[17]*x2 + val[18]*x3 + val[19]*x4;
		     	 	y4 += val[20]*x0 + val[21]*x1 + val[22]*x2 + val[23]*x3 + val[24]*x4;
		     	}
		     	/* Copy register to memory */
		     	y[0] += alpha*y0; y[1] += alpha*y1; y[2] += alpha*y2; y[3] += alpha*y3; y[4] += alpha*y4;
		     	
		     }else if(r == 6){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 = 0.0, y5 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2], x4 = x[idx[j]+3], x5 = x[idx[j]+4];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3 + val[ 4]*x4 + val[ 5]*x5;
		     		y1 += val[ 6]*x0 + val[ 7]*x1 + val[ 8]*x2 + val[ 9]*x3 + val[10]*x4 + val[11]*x5;
		     		y2 += val[12]*x0 + val[13]*x1 + val[14]*x2 + val[15]*x3 + val[16]*x4 + val[17]*x5;
		     		y3 += val[18]*x0 + val[19]*x1 + val[20]*x2 + val[21]*x3 + val[22]*x4 + val[23]*x5;
		     		y4 += val[24]*x0 + val[25]*x1 + val[26]*x2 + val[27]*x3 + val[28]*x4 + val[29]*x5;
		     		y5 += val[30]*x0 + val[31]*x1 + val[32]*x2 + val[33]*x3 + val[34]*x4 + val[35]*x5;
		     	}
		     	/* Copy register to memory */
		     	y[0] += alpha*y0; y[1] += alpha*y1; y[2] += alpha*y2; y[3] += alpha*y3; y[4] += alpha*y4; y[5] += alpha*y5;
		     	
		     }else{
		     	CUKR_FATALX("Block size (%d, %d) is NOT supported at generic_sspmv_bcsr", r, c);
		     }
		   }	
   	   }		  	  
   }
   else 
   {
   	   if (alpha == 1) 
   	   {
		   for (i = 0; i != nb; i++, y+=r) 
		   {
		     /* a row ptr */
		     jb = ptr[i] - 1;
		     je = ptr[i + 1] - 1;
		     
		     /* 2x2 to 6x6 */
		     if(r == 2){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]];
		     		/* GEMV, Block by Vector */
		     		y0 += val[0]*x0 + val[1]*x1;
		     	 	y1 += val[2]*x0 + val[3]*x1;
		     	}
		     	/* Copy register to memory */
		     	y[0] = y0; y[1] = y1;
		     	
		     }else if(r == 3){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0, y2 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1];
		     		/* GEMV, Block by Vector */
		     		y0 += val[0]*x0 + val[1]*x1 + val[2]*x2;
		     	 	y1 += val[3]*x0 + val[4]*x1 + val[5]*x2;
		     		y2 += val[6]*x0 + val[7]*x1 + val[8]*x2;
		     	}
		     	/* Copy register to memory */
		     	y[0] = y0; y[1] = y1; y[2] = y2;
		     	
		     }else if(r == 4){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3;
		     	 	y1 += val[ 4]*x0 + val[ 5]*x1 + val[ 6]*x2 + val[ 7]*x3;
		     	 	y2 += val[ 8]*x0 + val[ 9]*x1 + val[10]*x2 + val[11]*x3;
		     	 	y3 += val[12]*x0 + val[13]*x1 + val[14]*x2 + val[15]*x3;
		     	}
		     	/* Copy register to memory */
		     	y[0] = y0; y[1] = y1; y[2] = y2; y[3] = y3;
		     	
		     }else if(r == 5){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2], x4 = x[idx[j]+3];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3 + val[ 4]*x4;
		     	 	y1 += val[ 5]*x0 + val[ 6]*x1 + val[ 7]*x2 + val[ 8]*x3 + val[ 9]*x4;
		     	 	y2 += val[10]*x0 + val[11]*x1 + val[12]*x2 + val[13]*x3 + val[14]*x4;
		     	 	y3 += val[15]*x0 + val[16]*x1 + val[17]*x2 + val[18]*x3 + val[19]*x4;
		     	 	y4 += val[20]*x0 + val[21]*x1 + val[22]*x2 + val[23]*x3 + val[24]*x4;
		     	}
		     	/* Copy register to memory */
		     	y[0] = y0; y[1] = y1; y[2] = y2; y[3] = y3; y[4] = y4;
		     	
		     }else if(r == 6){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 = 0.0, y5 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2], x4 = x[idx[j]+3], x5 = x[idx[j]+4];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3 + val[ 4]*x4 + val[ 5]*x5;
		     		y1 += val[ 6]*x0 + val[ 7]*x1 + val[ 8]*x2 + val[ 9]*x3 + val[10]*x4 + val[11]*x5;
		     		y2 += val[12]*x0 + val[13]*x1 + val[14]*x2 + val[15]*x3 + val[16]*x4 + val[17]*x5;
		     		y3 += val[18]*x0 + val[19]*x1 + val[20]*x2 + val[21]*x3 + val[22]*x4 + val[23]*x5;
		     		y4 += val[24]*x0 + val[25]*x1 + val[26]*x2 + val[27]*x3 + val[28]*x4 + val[29]*x5;
		     		y5 += val[30]*x0 + val[31]*x1 + val[32]*x2 + val[33]*x3 + val[34]*x4 + val[35]*x5;
		     	}
		     	/* Copy register to memory */
		     	y[0] = y0; y[1] = y1; y[2] = y2; y[3] = y3; y[4] = y4; y[5] = y5;
		     	
		     }else{
		     	CUKR_FATALX("Block size (%d, %d) is NOT supported at generic_sspmv_bcsr", r, c);
		     }
		   }
   	   }
   	   else 
   	   {
		   for (i = 0; i != nb; i++, y+=r) 
		   {
		     /* a row ptr */
		     jb = ptr[i] - 1;
		     je = ptr[i + 1] - 1;
		     
		     /* 2x2 to 6x6 */
		     if(r == 2){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]];
		     		/* GEMV, Block by Vector */
		     		y0 += val[0]*x0 + val[1]*x1;
		     	 	y1 += val[2]*x0 + val[3]*x1;
		     	}
		     	/* Copy register to memory */
		     	y[0] = alpha*y0; y[1] = alpha*y1;
		     	
		     }else if(r == 3){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0, y2 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1];
		     		/* GEMV, Block by Vector */
		     		y0 += val[0]*x0 + val[1]*x1 + val[2]*x2;
		     	 	y1 += val[3]*x0 + val[4]*x1 + val[5]*x2;
		     		y2 += val[6]*x0 + val[7]*x1 + val[8]*x2;
		     	}
		     	/* Copy register to memory */
		     	y[0] = alpha*y0; y[1] = alpha*y1; y[2] = alpha*y2;
		     	
		     }else if(r == 4){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3;
		     	 	y1 += val[ 4]*x0 + val[ 5]*x1 + val[ 6]*x2 + val[ 7]*x3;
		     	 	y2 += val[ 8]*x0 + val[ 9]*x1 + val[10]*x2 + val[11]*x3;
		     	 	y3 += val[12]*x0 + val[13]*x1 + val[14]*x2 + val[15]*x3;
		     	}
		     	/* Copy register to memory */
		     	y[0] = alpha*y0; y[1] = alpha*y1; y[2] = alpha*y2; y[3] = alpha*y3;
		     	
		     }else if(r == 5){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2], x4 = x[idx[j]+3];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3 + val[ 4]*x4;
		     	 	y1 += val[ 5]*x0 + val[ 6]*x1 + val[ 7]*x2 + val[ 8]*x3 + val[ 9]*x4;
		     	 	y2 += val[10]*x0 + val[11]*x1 + val[12]*x2 + val[13]*x3 + val[14]*x4;
		     	 	y3 += val[15]*x0 + val[16]*x1 + val[17]*x2 + val[18]*x3 + val[19]*x4;
		     	 	y4 += val[20]*x0 + val[21]*x1 + val[22]*x2 + val[23]*x3 + val[24]*x4;
		     	}
		     	/* Copy register to memory */
		     	y[0] = alpha*y0; y[1] = alpha*y1; y[2] = alpha*y2; y[3] = alpha*y3; y[4] = alpha*y4;
		     	
		     }else if(r == 6){
		     	
		     	/* register y */
		     	register CUKR_SINGLE y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 = 0.0, y5 = 0.0;
		     	for (j = jb; j != je; j++, val+=r*c){
		     		/* register x */
		     		register CUKR_SINGLE x0 = x[idx[j]-1], x1 = x[idx[j]], x2 = x[idx[j]+1], x3 = x[idx[j]+2], x4 = x[idx[j]+3], x5 = x[idx[j]+4];
		     		/* GEMV, Block by Vector */
		     		y0 += val[ 0]*x0 + val[ 1]*x1 + val[ 2]*x2 + val[ 3]*x3 + val[ 4]*x4 + val[ 5]*x5;
		     		y1 += val[ 6]*x0 + val[ 7]*x1 + val[ 8]*x2 + val[ 9]*x3 + val[10]*x4 + val[11]*x5;
		     		y2 += val[12]*x0 + val[13]*x1 + val[14]*x2 + val[15]*x3 + val[16]*x4 + val[17]*x5;
		     		y3 += val[18]*x0 + val[19]*x1 + val[20]*x2 + val[21]*x3 + val[22]*x4 + val[23]*x5;
		     		y4 += val[24]*x0 + val[25]*x1 + val[26]*x2 + val[27]*x3 + val[28]*x4 + val[29]*x5;
		     		y5 += val[30]*x0 + val[31]*x1 + val[32]*x2 + val[33]*x3 + val[34]*x4 + val[35]*x5;
		     	}
		     	/* Copy register to memory */
		     	y[0] = alpha*y0; y[1] = alpha*y1; y[2] = alpha*y2; y[3] = alpha*y3; y[4] = alpha*y4; y[5] = alpha*y5;
		     	
		     }else{
		     	CUKR_FATALX("Block size (%d, %d) is NOT supported at generic_sspmv_bcsr", r, c);
		     }
		   }	
   	   }		  	  
   }
}


// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void generic_dspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                       CUKR_INT* ell_idx, CUKR_DOUBLE* ell_val, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_DOUBLE* csr_val,
                       CUKR_DOUBLE* x, CUKR_DOUBLE beta, CUKR_DOUBLE* y)
{
	int i, n;

	if (beta != 1)
   		generic_dscal(rows, beta, y);

	/* ELL part */
	for(n = 0; n < ell_nz_row; n++){
        CUKR_INT*          Aj_n = ell_idx + n * ell_stride;
        const CUKR_DOUBLE* Ax_n = ell_val + n * ell_stride;
        for(i = 0; i < rows; i++){
            y[i] += alpha * Ax_n[i] * x[Aj_n[i] - 1];
        }
    }

   /* CSR part */
    generic_dspmv_csr(rows, cols, csr_nz, alpha, csr_ptr,
           			  csr_idx, csr_val, x, 1.0, y);
}

/**
 * \brief Wrapper for QSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
void generic_qspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
	                   CUKR_INT* ell_idx, CUKR_SINGLE* ell_valh, CUKR_SINGLE* ell_valt, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_SINGLE* csr_valh, CUKR_SINGLE* csr_valt,
	                   CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	CUKR_SINGLE volatile alpha0, alpha1, beta0, beta1;
	CUKR_SINGLE volatile sum0, sum1, c0, c1;
    CUKR_INT jb, je;
    int i, j, n;


	/*
	 * Split alpha & beta
	 */

	alpha0 = (CUKR_SINGLE)alpha;
	alpha1 = (CUKR_SINGLE)(alpha - alpha0);

	beta0 = (CUKR_SINGLE)beta;
	beta1 = (CUKR_SINGLE)(beta - beta0);

	if (beta != 1)
   		generic_qscal(rows, beta, yh, yt);

	/* ELL part */
	for(n = 0; n < ell_nz_row; n++){
        CUKR_INT*          Aj_n   = ell_idx  + n * ell_stride;
        const CUKR_SINGLE* Ax_n_h = ell_valh + n * ell_stride;
        const CUKR_SINGLE* Ax_n_t = ell_valt + n * ell_stride;
        for(i = 0; i < rows; i++){
        	dsmul(&c0, &c1, Ax_n_h[i], Ax_n_t[i], xh[Aj_n[i] - 1], xt[Aj_n[i] - 1]);
        	dsmul(&c0, &c1, c0, c1, alpha0, alpha1);
        	dsadd(&yh[i], &yt[i], yh[i], yt[i], c0, c1);
        }
    }

    /* CSR part */
    generic_qspmv_csr(rows, cols, csr_nz, alpha, csr_ptr,
           			  csr_idx, csr_valh, csr_valt, xh, xt, 1.0, yh, yt);
}


/**
 * \brief Wrapper for SSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void generic_sspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                       CUKR_INT* ell_idx, CUKR_SINGLE* ell_val, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_SINGLE* csr_val,
                       CUKR_SINGLE* x, CUKR_SINGLE beta, CUKR_SINGLE* y)
{
	int i, n;

	if (beta != 1)
   		generic_sscal(rows, beta, y);

	/* ELL part */
	for(n = 0; n < ell_nz_row; n++){
        CUKR_INT*          Aj_n = ell_idx + n * ell_stride;
        const CUKR_SINGLE* Ax_n = ell_val + n * ell_stride;
        for(i = 0; i < rows; i++){
            y[i] += alpha * Ax_n[i] * x[Aj_n[i] - 1];
        }
    }

    /* CSR part */
    generic_sspmv_csr(rows, cols, csr_nz, alpha, csr_ptr,
           			  csr_idx, csr_val, x, 1.0, y);
}

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void generic_dspmv_bhyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
                       CUKR_INT* bell_idx, CUKR_DOUBLE* bell_val, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_DOUBLE* bcsr_val,
                       CUKR_DOUBLE* x, CUKR_DOUBLE beta, CUKR_DOUBLE* y)
{
	/* (r,c) = (2,2), (3,3), (4,4), (5,5), (6,6) are supported */
	if(r != c) CUKR_FATALX("Block size (%d, %d) is NOT supported at generic_dspmv_bhyb", r, c);

	int i, n;

	/* Get num of blocks per row */
	CUKR_INT nb = (rows%r ? rows/r+1 : rows/r);

	if(beta != 1)
		generic_dscal(rows, beta, y);

	/* BELL part */
	if(r == 2){
		for(n=0; n<bell_nz_row; n++){
			CUKR_INT*          Aj_n = bell_idx + n * bell_stride * 4;
			const CUKR_DOUBLE* Ax_n = bell_val + n * bell_stride * 4;
			for(i=0; i<nb; i++){
				y[i*2  ] += alpha * ( Ax_n[i*4  ] * x[ Aj_n[i*2]-1 ] + Ax_n[i*4+1] * x[ Aj_n[i*2] ] );
				y[i*2+1] += alpha * ( Ax_n[i*4+2] * x[ Aj_n[i*2]-1 ] + Ax_n[i*4+3] * x[ Aj_n[i*2] ] );
			}
		}
	}else if(r == 3){
		for(n=0; n<bell_nz_row; n++){
			CUKR_INT*          Aj_n = bell_idx + n * bell_stride * 9;
			const CUKR_DOUBLE* Ax_n = bell_val + n * bell_stride * 9;
			for(i=0; i<nb; i++){
				y[i*3  ] += alpha * ( Ax_n[i*9  ] * x[ Aj_n[i*3]-1 ] + Ax_n[i*9+1] * x[ Aj_n[i*3] ] + Ax_n[i*9+2] * x[ Aj_n[i*3]+1 ] );
				y[i*3+1] += alpha * ( Ax_n[i*9+3] * x[ Aj_n[i*3]-1 ] + Ax_n[i*9+4] * x[ Aj_n[i*3] ] + Ax_n[i*9+5] * x[ Aj_n[i*3]+1 ] );
				y[i*3+2] += alpha * ( Ax_n[i*9+6] * x[ Aj_n[i*3]-1 ] + Ax_n[i*9+7] * x[ Aj_n[i*3] ] + Ax_n[i*9+8] * x[ Aj_n[i*3]+1 ] );
			}
		}
	}else if(r == 4){
		for(n=0; n<bell_nz_row; n++){
			CUKR_INT*          Aj_n = bell_idx + n * bell_stride * 16;
			const CUKR_DOUBLE* Ax_n = bell_val + n * bell_stride * 16;
			for(i=0; i<nb; i++){
				y[i*4  ] += alpha *
				            ( Ax_n[i*16   ] * x[ Aj_n[i*4]-1 ]
				            + Ax_n[i*16+ 1] * x[ Aj_n[i*4] ]
				            + Ax_n[i*16+ 2] * x[ Aj_n[i*4]+1 ]
				            + Ax_n[i*16+ 3] * x[ Aj_n[i*4]+2 ] );
				y[i*4+1] += alpha *
				            ( Ax_n[i*16+ 4] * x[ Aj_n[i*4]-1 ]
				            + Ax_n[i*16+ 5] * x[ Aj_n[i*4] ]
				            + Ax_n[i*16+ 6] * x[ Aj_n[i*4]+1 ]
				            + Ax_n[i*16+ 7] * x[ Aj_n[i*4]+2 ] );
				y[i*4+2] += alpha *
				            ( Ax_n[i*16+ 8] * x[ Aj_n[i*4]-1 ]
				            + Ax_n[i*16+ 9] * x[ Aj_n[i*4] ]
				            + Ax_n[i*16+10] * x[ Aj_n[i*4]+1 ]
				            + Ax_n[i*16+11] * x[ Aj_n[i*4]+2 ] );
				y[i*4+3] += alpha * 
				            ( Ax_n[i*16+12] * x[ Aj_n[i*4]-1 ]
				            + Ax_n[i*16+13] * x[ Aj_n[i*4] ]
				            + Ax_n[i*16+14] * x[ Aj_n[i*4]+1 ]
				            + Ax_n[i*16+15] * x[ Aj_n[i*4]+2 ] );
			}
		}
	}else if(r == 5){
		for(n=0; n<bell_nz_row; n++){
			CUKR_INT*          Aj_n = bell_idx + n * bell_stride * 25;
			const CUKR_DOUBLE* Ax_n = bell_val + n * bell_stride * 25;
			for(i=0; i<nb; i++){
				y[i*5  ] += alpha *
				            ( Ax_n[i*25   ] * x[ Aj_n[i*5]-1 ]
				            + Ax_n[i*25+ 1] * x[ Aj_n[i*5] ]
				            + Ax_n[i*25+ 2] * x[ Aj_n[i*5]+1 ]
				            + Ax_n[i*25+ 3] * x[ Aj_n[i*5]+2 ]
				            + Ax_n[i*25+ 4] * x[ Aj_n[i*5]+3 ] );
				y[i*5+1] += alpha *
				            ( Ax_n[i*25+ 5] * x[ Aj_n[i*5]-1 ]
				            + Ax_n[i*25+ 6] * x[ Aj_n[i*5] ]
				            + Ax_n[i*25+ 7] * x[ Aj_n[i*5]+1 ]
				            + Ax_n[i*25+ 8] * x[ Aj_n[i*5]+2 ]
				            + Ax_n[i*25+ 9] * x[ Aj_n[i*5]+3 ] );
				y[i*5+2] += alpha *
				            ( Ax_n[i*25+10] * x[ Aj_n[i*5]-1 ]
				            + Ax_n[i*25+11] * x[ Aj_n[i*5] ]
				            + Ax_n[i*25+12] * x[ Aj_n[i*5]+1 ]
				            + Ax_n[i*25+13] * x[ Aj_n[i*5]+2 ]
				            + Ax_n[i*25+14] * x[ Aj_n[i*5]+3 ] );
				y[i*5+3] += alpha *
				            ( Ax_n[i*25+15] * x[ Aj_n[i*5]-1 ]
				            + Ax_n[i*25+16] * x[ Aj_n[i*5] ]
				            + Ax_n[i*25+17] * x[ Aj_n[i*5]+1 ]
				            + Ax_n[i*25+18] * x[ Aj_n[i*5]+2 ]
				            + Ax_n[i*25+19] * x[ Aj_n[i*5]+3 ] );
				y[i*5+4] += alpha *
				            ( Ax_n[i*25+20] * x[ Aj_n[i*5]-1 ]
				            + Ax_n[i*25+21] * x[ Aj_n[i*5] ]
				            + Ax_n[i*25+22] * x[ Aj_n[i*5]+1 ]
				            + Ax_n[i*25+23] * x[ Aj_n[i*5]+2 ]
				            + Ax_n[i*25+24] * x[ Aj_n[i*5]+3 ] );
			}
		}
	}else if(r == 6){
		for(n=0; n<bell_nz_row; n++){
			CUKR_INT*          Aj_n = bell_idx + n * bell_stride * 36;
			const CUKR_DOUBLE* Ax_n = bell_val + n * bell_stride * 36;
			for(i=0; i<nb; i++){
				y[i*6  ] += alpha *
				            ( Ax_n[i*36   ] * x[ Aj_n[i*6]-1 ]
				            + Ax_n[i*36+ 1] * x[ Aj_n[i*6] ]
				            + Ax_n[i*36+ 2] * x[ Aj_n[i*6]+1 ]
				            + Ax_n[i*36+ 3] * x[ Aj_n[i*6]+2 ]
				            + Ax_n[i*36+ 4] * x[ Aj_n[i*6]+3 ]
				            + Ax_n[i*36+ 5] * x[ Aj_n[i*6]+4 ] );
				y[i*6  ] += alpha *
				            ( Ax_n[i*36+ 6] * x[ Aj_n[i*6]-1 ]
				            + Ax_n[i*36+ 7] * x[ Aj_n[i*6] ]
				            + Ax_n[i*36+ 8] * x[ Aj_n[i*6]+1 ]
				            + Ax_n[i*36+ 9] * x[ Aj_n[i*6]+2 ]
				            + Ax_n[i*36+10] * x[ Aj_n[i*6]+3 ]
				            + Ax_n[i*36+11] * x[ Aj_n[i*6]+4 ] );
				y[i*6  ] += alpha *
				            ( Ax_n[i*36+12] * x[ Aj_n[i*6]-1 ]
				            + Ax_n[i*36+13] * x[ Aj_n[i*6] ]
				            + Ax_n[i*36+14] * x[ Aj_n[i*6]+1 ]
				            + Ax_n[i*36+15] * x[ Aj_n[i*6]+2 ]
				            + Ax_n[i*36+16] * x[ Aj_n[i*6]+3 ]
				            + Ax_n[i*36+17] * x[ Aj_n[i*6]+4 ] );
				y[i*6  ] += alpha *
				            ( Ax_n[i*36+18] * x[ Aj_n[i*6]-1 ]
				            + Ax_n[i*36+19] * x[ Aj_n[i*6] ]
				            + Ax_n[i*36+20] * x[ Aj_n[i*6]+1 ]
				            + Ax_n[i*36+21] * x[ Aj_n[i*6]+2 ]
				            + Ax_n[i*36+22] * x[ Aj_n[i*6]+3 ]
				            + Ax_n[i*36+23] * x[ Aj_n[i*6]+4 ] );
				y[i*6  ] += alpha *
				            ( Ax_n[i*36+24] * x[ Aj_n[i*6]-1 ]
				            + Ax_n[i*36+25] * x[ Aj_n[i*6] ]
				            + Ax_n[i*36+26] * x[ Aj_n[i*6]+1 ]
				            + Ax_n[i*36+27] * x[ Aj_n[i*6]+2 ]
				            + Ax_n[i*36+28] * x[ Aj_n[i*6]+3 ]
				            + Ax_n[i*36+29] * x[ Aj_n[i*6]+4 ] );
				y[i*6  ] += alpha *
				            ( Ax_n[i*36+30] * x[ Aj_n[i*6]-1 ]
				            + Ax_n[i*36+31] * x[ Aj_n[i*6] ]
				            + Ax_n[i*36+32] * x[ Aj_n[i*6]+1 ]
				            + Ax_n[i*36+33] * x[ Aj_n[i*6]+2 ]
				            + Ax_n[i*36+34] * x[ Aj_n[i*6]+3 ]
				            + Ax_n[i*36+35] * x[ Aj_n[i*6]+4 ] );
			}
		}
	}else{
		CUKR_FATALX("Block size (%d, %d) is NOT supported at generic_dspmv_bhyb", r, c);
	}
	
	/* BCSR part */
	generic_dspmv_bcsr(rows, cols, bcsr_nz, r, c, alpha, bcsr_ptr, bcsr_idx, bcsr_val, x, 1.0, y);
}

/**
 * \brief Wrapper for QSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
void generic_qspmv_bhyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
	                   CUKR_INT* bell_idx, CUKR_SINGLE* bell_valh, CUKR_SINGLE* bell_valt, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_SINGLE* bcsr_valh, CUKR_SINGLE* bcsr_valt,
	                   CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	CUKR_FATAL("generic_qspmv_bhyb has not been implemented yet");
}

/**
 * \brief Wrapper for SSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void generic_sspmv_bhyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
                       CUKR_INT* bell_idx, CUKR_SINGLE* bell_val, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_SINGLE* bcsr_val,
                       CUKR_SINGLE* x, CUKR_SINGLE beta, CUKR_SINGLE* y)
{
	/* (r,c) = (2,2), (3,3), (4,4), (5,5), (6,6) are supported */
	if(r != c) CUKR_FATALX("Block size (%d, %d) is NOT supported at generic_sspmv_bhyb", r, c);

	int i, n;

	/* Get num of blocks per row */
	CUKR_INT nb = (rows%r ? rows/r+1 : rows/r);

	if(beta != 1)
		generic_sscal(rows, beta, y);

	/* BELL part */
	if(r == 2){
		for(n=0; n<bell_nz_row; n++){
			CUKR_INT*          Aj_n = bell_idx + n * bell_stride * 4;
			const CUKR_SINGLE* Ax_n = bell_val + n * bell_stride * 4;
			for(i=0; i<nb; i++){
				y[i*2  ] += alpha * ( Ax_n[i*4  ] * x[ Aj_n[i*2]-1 ] + Ax_n[i*4+1] * x[ Aj_n[i*2] ] );
				y[i*2+1] += alpha * ( Ax_n[i*4+2] * x[ Aj_n[i*2]-1 ] + Ax_n[i*4+3] * x[ Aj_n[i*2] ] );
			}
		}
	}else if(r == 3){
		for(n=0; n<bell_nz_row; n++){
			CUKR_INT*          Aj_n = bell_idx + n * bell_stride * 9;
			const CUKR_SINGLE* Ax_n = bell_val + n * bell_stride * 9;
			for(i=0; i<nb; i++){
				y[i*3  ] += alpha * ( Ax_n[i*9  ] * x[ Aj_n[i*3]-1 ] + Ax_n[i*9+1] * x[ Aj_n[i*3] ] + Ax_n[i*9+2] * x[ Aj_n[i*3]+1 ] );
				y[i*3+1] += alpha * ( Ax_n[i*9+3] * x[ Aj_n[i*3]-1 ] + Ax_n[i*9+4] * x[ Aj_n[i*3] ] + Ax_n[i*9+5] * x[ Aj_n[i*3]+1 ] );
				y[i*3+2] += alpha * ( Ax_n[i*9+6] * x[ Aj_n[i*3]-1 ] + Ax_n[i*9+7] * x[ Aj_n[i*3] ] + Ax_n[i*9+8] * x[ Aj_n[i*3]+1 ] );
			}
		}
	}else if(r == 4){
		for(n=0; n<bell_nz_row; n++){
			CUKR_INT*          Aj_n = bell_idx + n * bell_stride * 16;
			const CUKR_SINGLE* Ax_n = bell_val + n * bell_stride * 16;
			for(i=0; i<nb; i++){
				y[i*4  ] += alpha *
				            ( Ax_n[i*16   ] * x[ Aj_n[i*4]-1 ]
				            + Ax_n[i*16+ 1] * x[ Aj_n[i*4] ]
				            + Ax_n[i*16+ 2] * x[ Aj_n[i*4]+1 ]
				            + Ax_n[i*16+ 3] * x[ Aj_n[i*4]+2 ] );
				y[i*4+1] += alpha *
				            ( Ax_n[i*16+ 4] * x[ Aj_n[i*4]-1 ]
				            + Ax_n[i*16+ 5] * x[ Aj_n[i*4] ]
				            + Ax_n[i*16+ 6] * x[ Aj_n[i*4]+1 ]
				            + Ax_n[i*16+ 7] * x[ Aj_n[i*4]+2 ] );
				y[i*4+2] += alpha *
				            ( Ax_n[i*16+ 8] * x[ Aj_n[i*4]-1 ]
				            + Ax_n[i*16+ 9] * x[ Aj_n[i*4] ]
				            + Ax_n[i*16+10] * x[ Aj_n[i*4]+1 ]
				            + Ax_n[i*16+11] * x[ Aj_n[i*4]+2 ] );
				y[i*4+3] += alpha * 
				            ( Ax_n[i*16+12] * x[ Aj_n[i*4]-1 ]
				            + Ax_n[i*16+13] * x[ Aj_n[i*4] ]
				            + Ax_n[i*16+14] * x[ Aj_n[i*4]+1 ]
				            + Ax_n[i*16+15] * x[ Aj_n[i*4]+2 ] );
			}
		}
	}else if(r == 5){
		for(n=0; n<bell_nz_row; n++){
			CUKR_INT*          Aj_n = bell_idx + n * bell_stride * 25;
			const CUKR_SINGLE* Ax_n = bell_val + n * bell_stride * 25;
			for(i=0; i<nb; i++){
				y[i*5  ] += alpha *
				            ( Ax_n[i*25   ] * x[ Aj_n[i*5]-1 ]
				            + Ax_n[i*25+ 1] * x[ Aj_n[i*5] ]
				            + Ax_n[i*25+ 2] * x[ Aj_n[i*5]+1 ]
				            + Ax_n[i*25+ 3] * x[ Aj_n[i*5]+2 ]
				            + Ax_n[i*25+ 4] * x[ Aj_n[i*5]+3 ] );
				y[i*5+1] += alpha *
				            ( Ax_n[i*25+ 5] * x[ Aj_n[i*5]-1 ]
				            + Ax_n[i*25+ 6] * x[ Aj_n[i*5] ]
				            + Ax_n[i*25+ 7] * x[ Aj_n[i*5]+1 ]
				            + Ax_n[i*25+ 8] * x[ Aj_n[i*5]+2 ]
				            + Ax_n[i*25+ 9] * x[ Aj_n[i*5]+3 ] );
				y[i*5+2] += alpha *
				            ( Ax_n[i*25+10] * x[ Aj_n[i*5]-1 ]
				            + Ax_n[i*25+11] * x[ Aj_n[i*5] ]
				            + Ax_n[i*25+12] * x[ Aj_n[i*5]+1 ]
				            + Ax_n[i*25+13] * x[ Aj_n[i*5]+2 ]
				            + Ax_n[i*25+14] * x[ Aj_n[i*5]+3 ] );
				y[i*5+3] += alpha *
				            ( Ax_n[i*25+15] * x[ Aj_n[i*5]-1 ]
				            + Ax_n[i*25+16] * x[ Aj_n[i*5] ]
				            + Ax_n[i*25+17] * x[ Aj_n[i*5]+1 ]
				            + Ax_n[i*25+18] * x[ Aj_n[i*5]+2 ]
				            + Ax_n[i*25+19] * x[ Aj_n[i*5]+3 ] );
				y[i*5+4] += alpha *
				            ( Ax_n[i*25+20] * x[ Aj_n[i*5]-1 ]
				            + Ax_n[i*25+21] * x[ Aj_n[i*5] ]
				            + Ax_n[i*25+22] * x[ Aj_n[i*5]+1 ]
				            + Ax_n[i*25+23] * x[ Aj_n[i*5]+2 ]
				            + Ax_n[i*25+24] * x[ Aj_n[i*5]+3 ] );
			}
		}
	}else if(r == 6){
		for(n=0; n<bell_nz_row; n++){
			CUKR_INT*          Aj_n = bell_idx + n * bell_stride * 36;
			const CUKR_SINGLE* Ax_n = bell_val + n * bell_stride * 36;
			for(i=0; i<nb; i++){
				y[i*6  ] += alpha *
				            ( Ax_n[i*36   ] * x[ Aj_n[i*6]-1 ]
				            + Ax_n[i*36+ 1] * x[ Aj_n[i*6] ]
				            + Ax_n[i*36+ 2] * x[ Aj_n[i*6]+1 ]
				            + Ax_n[i*36+ 3] * x[ Aj_n[i*6]+2 ]
				            + Ax_n[i*36+ 4] * x[ Aj_n[i*6]+3 ]
				            + Ax_n[i*36+ 5] * x[ Aj_n[i*6]+4 ] );
				y[i*6  ] += alpha *
				            ( Ax_n[i*36+ 6] * x[ Aj_n[i*6]-1 ]
				            + Ax_n[i*36+ 7] * x[ Aj_n[i*6] ]
				            + Ax_n[i*36+ 8] * x[ Aj_n[i*6]+1 ]
				            + Ax_n[i*36+ 9] * x[ Aj_n[i*6]+2 ]
				            + Ax_n[i*36+10] * x[ Aj_n[i*6]+3 ]
				            + Ax_n[i*36+11] * x[ Aj_n[i*6]+4 ] );
				y[i*6  ] += alpha *
				            ( Ax_n[i*36+12] * x[ Aj_n[i*6]-1 ]
				            + Ax_n[i*36+13] * x[ Aj_n[i*6] ]
				            + Ax_n[i*36+14] * x[ Aj_n[i*6]+1 ]
				            + Ax_n[i*36+15] * x[ Aj_n[i*6]+2 ]
				            + Ax_n[i*36+16] * x[ Aj_n[i*6]+3 ]
				            + Ax_n[i*36+17] * x[ Aj_n[i*6]+4 ] );
				y[i*6  ] += alpha *
				            ( Ax_n[i*36+18] * x[ Aj_n[i*6]-1 ]
				            + Ax_n[i*36+19] * x[ Aj_n[i*6] ]
				            + Ax_n[i*36+20] * x[ Aj_n[i*6]+1 ]
				            + Ax_n[i*36+21] * x[ Aj_n[i*6]+2 ]
				            + Ax_n[i*36+22] * x[ Aj_n[i*6]+3 ]
				            + Ax_n[i*36+23] * x[ Aj_n[i*6]+4 ] );
				y[i*6  ] += alpha *
				            ( Ax_n[i*36+24] * x[ Aj_n[i*6]-1 ]
				            + Ax_n[i*36+25] * x[ Aj_n[i*6] ]
				            + Ax_n[i*36+26] * x[ Aj_n[i*6]+1 ]
				            + Ax_n[i*36+27] * x[ Aj_n[i*6]+2 ]
				            + Ax_n[i*36+28] * x[ Aj_n[i*6]+3 ]
				            + Ax_n[i*36+29] * x[ Aj_n[i*6]+4 ] );
				y[i*6  ] += alpha *
				            ( Ax_n[i*36+30] * x[ Aj_n[i*6]-1 ]
				            + Ax_n[i*36+31] * x[ Aj_n[i*6] ]
				            + Ax_n[i*36+32] * x[ Aj_n[i*6]+1 ]
				            + Ax_n[i*36+33] * x[ Aj_n[i*6]+2 ]
				            + Ax_n[i*36+34] * x[ Aj_n[i*6]+3 ]
				            + Ax_n[i*36+35] * x[ Aj_n[i*6]+4 ] );
			}
		}
	}else{
		CUKR_FATALX("Block size (%d, %d) is NOT supported at generic_sspmv_bhyb", r, c);
	}
	
	/* BCSR part */
	generic_sspmv_bcsr(rows, cols, bcsr_nz, r, c, alpha, bcsr_ptr, bcsr_idx, bcsr_val, x, 1.0, y);
}

// ------------------------------------------------------------------- //

/*
 * \brief Create a generic blas object by
 *        binding the above functions to a
 * 		  CUKR_BLAS structure 
 * \param blas pointer to a blas structure
 */
void blas_create_generic(CUKR_BLAS blas)
{
	/* Set the BLAS name */
	sprintf(blas->name, "GENERIC");
	blas->location = CUKR_LOCATION_CPU;
	
	/* Set the bindings */
	blas->bind.ddot = generic_ddot;
	blas->bind.qdot = generic_qdot;
	blas->bind.sdot = generic_sdot;
	
	blas->bind.daxpy = generic_daxpy;
	blas->bind.qaxpy = generic_qaxpy;
	blas->bind.saxpy = generic_saxpy;
	
	blas->bind.daypx = generic_daypx;
	blas->bind.qaypx = generic_qaypx;
	blas->bind.saypx = generic_saypx;
	
	blas->bind.dscal = generic_dscal;
	blas->bind.qscal = generic_qscal;
	blas->bind.sscal = generic_sscal;
	
	blas->bind.dcopy = generic_dcopy;
	blas->bind.qcopy = generic_qcopy;
	blas->bind.scopy = generic_scopy;
	 
	blas->bind.dspmv_csr = generic_dspmv_csr;
	blas->bind.qspmv_csr = generic_qspmv_csr;
	blas->bind.sspmv_csr = generic_sspmv_csr;
	
	blas->bind.dspmv_bcsr = generic_dspmv_bcsr;
	blas->bind.qspmv_bcsr = generic_qspmv_bcsr;
	blas->bind.sspmv_bcsr = generic_sspmv_bcsr;

	blas->bind.dspmv_hyb = generic_dspmv_hyb;
	blas->bind.qspmv_hyb = generic_qspmv_hyb;
	blas->bind.sspmv_hyb = generic_sspmv_hyb;
	
	blas->bind.dspmv_bhyb = generic_dspmv_bhyb;
	blas->bind.qspmv_bhyb = generic_qspmv_bhyb;
	blas->bind.sspmv_bhyb = generic_sspmv_bhyb;

}               

// ====================== PRIVATE AUX FUNCTIONS ======================= //

/**
 * \brief Perform quasi-double addition
 * \param c0,c1 The result (headvand tail)
 * \param a0,a1 The first number (head
 *              and tail)
 * \param b0,b1 The second number (head 
 *              and tail)
 */
void dsadd(volatile float* c0, volatile float* c1, const volatile float a0, const volatile float a1, 
                                                   const volatile float b0, const volatile float b1)
{
	/*
	 * Use volatile keyword to discourage the
	 * compiler to mess with the functions 
	 */
	
	/* Compute dsa + dsb using Knuth's trick. */
    float t1 = a0 + b0;
    float volatile e = t1 - a0;
    float t2 = ((b0 - e) + (a0 - (t1 - e))) + a1 + b1;

    /* The result is t1 + t2, after normalization. */
    *c0 = e = t1 + t2;
    *c1 = t2 - (e - t1);
   
}

// ------------------------------------------------------------------- //

/**
 * \brief Perform quasi-double subtraction
 * \param c0,c1 The result (headvand tail)
 * \param a0,a1 The first number (head
 *              and tail)
 * \param b0,b1 The second number (head 
 *              and tail)
 */
void dssub(volatile float* c0, volatile float* c1, const volatile float a0, const volatile float a1, 
                                                   const volatile float b0, const volatile float b1)

{
	/*
	 * Use volatile keyword to discourage the
	 * compiler to mess with the functions 
	 */	
	
    // Compute dsa - dsb using Knuth's trick.
    float t1 = a0 - b0;
    float volatile e = t1 - a0;
    float t2 = ((-b0 - e) + (a0 - (t1 - e))) + a1 - b1;

    // The result is t1 + t2, after normalization.
    *c0 = e = t1 + t2;
    *c1 = t2 - (e - t1);
}
   
// ------------------------------------------------------------------- //

/**
 * \brief Perform quasi-double multiplication
 * \param c0,c1 The result (headvand tail)
 * \param a0,a1 The first number (head
 *              and tail)
 * \param b0,b1 The second number (head 
 *              and tail)
 */
void dsmul(volatile float *c0, volatile float *c1, const volatile float a0, const volatile float a1, 
                                                   const volatile float b0, const volatile float b1)
{
	/*
	 * Use volatile keyword to discourage the
	 * compiler to mess with the functions 
	 */	
	
	/* This splits dsa(1) and dsb(1) into 
	 * high-order and low-order words. */
	float cona = a0 * 8193.0f;
	float conb = b0 * 8193.0f;
	
	float sa1 = cona - (cona - a0);
	float sb1 = conb - (conb - b0);
	float sa2 = a0 - sa1;
	float sb2 = b0 - sb1;

	/* Multilply a0 * b0 using Dekker's method */
	float c11 = a0 * b0;
	float c21 = (((sa1 * sb1 - c11) + sa1 * sb2) + sa2 * sb1) + sa2 * sb2;

    /* Compute a0 * b1 + a1 * b0 
    (only high-order word is needed) */
    float c2 = a0 * b1 + a1 * b0;
	
    /* Compute (c11, c21) + c2 using Knuth's trick, 
     * also adding low-order product. */
    float t1 = c11 + c2;
    float volatile e = t1 - c11;
    float t2 = ((c2 - e) + (c11 - (t1 - e))) + c21 + a1 * b1;
	
    /* The result is t1 + t2, after normalization. */
    *c0 = e = t1 + t2;
    *c1 = t2 - (e - t1);	
}
                              	
