/**
 * @file blas/impl/generic/generic_blas.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * This defines the "generic blas", with
 * all functions written here. It does not
 * use any libraries. Used for control.
 */

#ifndef CUKR_GENERIC_BLAS_H_
#define CUKR_GENERIC_BLAS_H_

 
#include "../../blas.h"

// ------------------------------------------------------------------- //

/**
 * \brief Dot product in double precision
 * \param n Vector size
 * \param x,y The two vectors whose product 
 *            is being computed 
 */
CUKR_DOUBLE generic_ddot(CUKR_INT n, CUKR_DOUBLE* x, CUKR_DOUBLE* y);

/**
 * \brief Dot product in quasi-double precision
 * \param n Vector size
 * \param n The vector size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
CUKR_DOUBLE generic_qdot(CUKR_INT n, CUKR_SINGLE* xh, CUKR_SINGLE* xt,
									 CUKR_SINGLE* yh, CUKR_SINGLE* yt);


/**
 * \brief Dot product in single precision
 * \param n Vector size
 * \param x,y The two vectors whose product 
 *            is being computed 
 */
CUKR_SINGLE generic_sdot(CUKR_INT n, CUKR_SINGLE* x, CUKR_SINGLE* y);
    

// ------------------------------------------------------------------- //
  
/**
 * \brief AXPY in double precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
void generic_daxpy(CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x, CUKR_DOUBLE* y); 

/**
 * \brief AXPY in quasi-double precision
 * \param n Vector size
 * \param a The constant
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector 
 */
void generic_qaxpy(CUKR_INT n, CUKR_DOUBLE a, CUKR_SINGLE* xh, CUKR_SINGLE* xt, 
											  CUKR_SINGLE* yh, CUKR_SINGLE* yt); 


/**
 * \brief AXPY in single precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
void generic_saxpy(CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x, CUKR_SINGLE* y); 


// ------------------------------------------------------------------- //

/**
 * \brief AYPX in double precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
void generic_daypx(CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x, CUKR_DOUBLE* y); 

/**
 * \brief AYPX in quasi-double precision
 * \param n Vector size
 * \param a The constant
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector 
 */
void generic_qaypx(CUKR_INT n, CUKR_DOUBLE a, CUKR_SINGLE* xh, CUKR_SINGLE* xt, 
											  CUKR_SINGLE* yh, CUKR_SINGLE* yt); 


/**
 * \brief AYPX in single precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
void generic_saypx(CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x, CUKR_SINGLE* y); 


// ------------------------------------------------------------------- //

/**
 * \brief Vector scaling in double precision
 * \param n Vector size
 * \param a Scaling factor
 * \param x The vector being scaled
 */
void generic_dscal(CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x); 

/**
 * \brief Vector scaling in quasi-double 
 *        precision
 * \param n Vector size
 * \param a Scaling factor
 * \param xh,xt The head and tail of the
 *              vector 
 */
void generic_qscal(CUKR_INT n, CUKR_DOUBLE a, CUKR_SINGLE* xh,
											  CUKR_SINGLE* xt); 

/**
 * \brief Vector scaling in single precision
 * \param n Vector size
 * \param a Scaling factor
 * \param x The vector being scaled
 */
void generic_sscal(CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x); 

// ------------------------------------------------------------------- //

/**
 * \brief Vector copy in double precision
 * \param n Vector size
 * \param x The source vector
 * \param y The destination vector
 */
void generic_dcopy(CUKR_INT n, CUKR_DOUBLE* x, CUKR_DOUBLE* y); 

/**
 * \brief Vector copy in quasi-double 
 *        precision
 * \param n Vector size
 * \param xh,xt The head and tail of the
 *              source vector 
 * \param yh,yt The head and tail of the
 *              destination vector 
 */
void generic_qcopy(CUKR_INT n, CUKR_SINGLE* xh, CUKR_SINGLE* xt, 
							   CUKR_SINGLE* yh, CUKR_SINGLE* yt); 

/**
 * \brief Vector copy in single precision
 * \param n Vector size
 * \param x The source vector
 * \param y The destination vector
 */
void generic_scopy(CUKR_INT n, CUKR_SINGLE* x, CUKR_SINGLE* y); 

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
void generic_dspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
               		   CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x, CUKR_DOUBLE beta, 
               		   CUKR_DOUBLE* y);  

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param beta Scale factor for y
 * \param yh,xt Head and tail for the result
 *              vector
 */
void generic_qspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
               		   CUKR_INT* idx, CUKR_SINGLE* valh, CUKR_SINGLE* valt, 
               		   CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, 
               		   CUKR_SINGLE* yh, CUKR_SINGLE* yt);  


/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in single precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
void generic_sspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_INT* ptr, 
          	           CUKR_INT* idx, CUKR_SINGLE* val, CUKR_SINGLE* x, CUKR_SINGLE beta, 
            		   CUKR_SINGLE* y);                   

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param y Result vector
 */
void generic_dspmv_bcsr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
					    CUKR_DOUBLE alpha, CUKR_INT* ptr, CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x,
                        CUKR_DOUBLE beta, CUKR_DOUBLE* y);

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector (head,tail)
 * \param x Vector being multiplied (head,tail)
 * \param beta Constant
 * \param y Result vector (head,tail)
 */
void generic_qspmv_bcsr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
					    CUKR_DOUBLE alpha, CUKR_INT* ptr, CUKR_INT* idx, CUKR_SINGLE* valh, CUKR_SINGLE* valt,
					    CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt);

/**
 * \brief Sparse matrix vector multiply in
 *        BCSR format, in single precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param r,c Block dims
 * \param alpha Constant
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param y Result vector
 */
void generic_sspmv_bcsr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c,
						CUKR_SINGLE alpha, CUKR_INT* ptr, CUKR_INT* idx, CUKR_SINGLE* val, CUKR_SINGLE* x,
			    	    CUKR_SINGLE beta, CUKR_SINGLE* y);

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void generic_dspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                       CUKR_INT* ell_idx, CUKR_DOUBLE* ell_val, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_DOUBLE* csr_val,
                       CUKR_DOUBLE* x, CUKR_DOUBLE beta, CUKR_DOUBLE* y);

/**
 * \brief Wrapper for QSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
void generic_qspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
	                   CUKR_INT* ell_idx, CUKR_SINGLE* ell_valh, CUKR_SINGLE* ell_valt, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_SINGLE* csr_valh, CUKR_SINGLE* csr_valt,
	                   CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt);

/**
 * \brief Wrapper for SSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void generic_sspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                       CUKR_INT* ell_idx, CUKR_SINGLE* ell_val, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_SINGLE* csr_val,
                       CUKR_SINGLE* x, CUKR_SINGLE beta, CUKR_SINGLE* y);

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void generic_dspmv_bhyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
                       CUKR_INT* bell_idx, CUKR_DOUBLE* bell_val, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_DOUBLE* bcsr_val,
                       CUKR_DOUBLE* x, CUKR_DOUBLE beta, CUKR_DOUBLE* y);

/**
 * \brief Wrapper for QSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
void generic_qspmv_bhyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
	                   CUKR_INT* bell_idx, CUKR_SINGLE* bell_valh, CUKR_SINGLE* bell_valt, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_SINGLE* bcsr_valh, CUKR_SINGLE* bcsr_valt,
	                   CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt);

/**
 * \brief Wrapper for SSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void generic_sspmv_bhyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
                       CUKR_INT* bell_idx, CUKR_SINGLE* bell_val, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_SINGLE* bcsr_val,
                       CUKR_SINGLE* x, CUKR_SINGLE beta, CUKR_SINGLE* y);

// ------------------------------------------------------------------- //

/**
 * \brief Create a generic blas object by
 *        binding the above functions to a
 * 		  CUKR_BLAS structure 
 * \param blas pointer to a blas structure
 */
void blas_create_generic(CUKR_BLAS blas);                


#endif /*CUKR_GENERIC_BLAS_H_*/               
