/**
 * @file cublas_wrappers.cu
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains wrappers for cublas functions.
 * 
 * To keep the library in ANSI C and only
 * at little as possible in C++, wrappers
 * for the CUBLAS functions are defined.
 * 
 * A wrapper is basically a call to the
 * cublas function + calls to cuda functions
 * (e.g. cudaThreadSychronize) that need to
 * be compiled by the CUDA compiler. 
 * 
 */

#include <stdio.h>
#include <cuda.h>
#include <cutil.h>
#include <cublas.h>
#include "cublas_wrappers.h"


/**
 * \brief DDOT wrapper
 */
#ifdef __cplusplus
extern "C"
#endif  
double cuda_cublasDdot(int n, double* x, int incx, double* y, int incy)
{
	/* Execute CUBLAS function */
	double res = cublasDdot(n, x, incx, y, incy);
	
	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize());
    
    return res;
}

/**
 * \brief SDOT wrapper
 */
#ifdef __cplusplus
extern "C"
#endif  
float cuda_cublasSdot(int n, float* x, int incx, float* y, int incy)
{
	/* Execute CUBLAS function */
	float res = cublasSdot(n, x, incx, y, incy);
	
	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize());
    
    return res;
}

/**
 * \brief DAXPY wrapper
 */
#ifdef __cplusplus
extern "C"
#endif  
void cuda_cublasDaxpy(int n, double a, double *x, int incx, double *y, int incy)
{
	/* Execute CUBLAS function */
	cublasDaxpy(n, a, x, incx, y, incy);
	
	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize());
}

/**
 * \brief SAXPY wrapper
 */
#ifdef __cplusplus
extern "C"
#endif  
void cuda_cublasSaxpy(int n, float a, float *x, int incx, float *y, int incy)
{
	/* Execute CUBLAS function */
	cublasSaxpy(n, a, x, incx, y, incy);
	
	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize());
}

/**
 * \brief DSCAL wrapper
 */
#ifdef __cplusplus
extern "C"
#endif  
void cuda_cublasDscal(int n, double a, double *x, int incx)
{
	/* Execute CUBLAS function */
	cublasDscal(n, a, x, incx);
	
	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize());
}

/**
 * \brief SSCAL wrapper
 */
#ifdef __cplusplus
extern "C"
#endif  
void cuda_cublasSscal(int n, float a, float *x, int incx)
{
	/* Execute CUBLAS function */
	cublasSscal(n, a, x, incx);
	
	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize());
}

/**
 * \brief DCOPY wrapper
 */
#ifdef __cplusplus
extern "C"
#endif  
void cuda_cublasDcopy(int n, double *x, int incx, double *y, int incy)
{
	/* Execute CUBLAS function */
	cublasDcopy(n, x, incx, y, incy);
	
	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize());
}

/**
 * \brief SCOPY wrapper
 */
#ifdef __cplusplus
extern "C"
#endif  
void cuda_cublasScopy(int n, float *x, int incx, float *y, int incy)
{
	/* Execute CUBLAS function */
	cublasScopy(n, x, incx, y, incy);
	
	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize());
}

