/**
 * @file blas/gpu/cuda/kernel_dspmv_bell.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Yusa codes.
 *
 * SPMV in BELL format
 */
 
#include "../kernels_textures.h"
#include "../kernels_utils.h"

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x 
	#define bx          blockIdx.x
#endif


/**
 * Calculate SPMV when block size is 2x2
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_dspmv_bell_2x2(const int n, const double alpha, const int nz_row, const int stride, double *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_DSPMV_BELL_2X2_THREADS + tx; i < n; i += CUKR_DSPMV_BELL_2X2_CTAS * CUKR_DSPMV_BELL_2X2_THREADS)
	{
		/* register y */
		double y0 = 0.0, y1 = 0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			int2 x0 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))-1);
			int2 x1 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i)));
			
			/* val */
			int2 val0 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*4);
			int2 val1 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*4+1);
			int2 val2 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*4+2);
			int2 val3 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*4+3);

			/* GEMV */
			y0 += __hiloint2double(val0.y, val0.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val1.y, val1.x) * __hiloint2double(x1.y, x1.x);
			y1 += __hiloint2double(val2.y, val2.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val3.y, val3.x) * __hiloint2double(x1.y, x1.x);
		}

		/* copy register y to global memory d_y */
		d_y[i*2  ] += alpha*y0;
		d_y[i*2+1] += alpha*y1;
	}
}


/**
 * Calculate SPMV when block size is 3x3
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_dspmv_bell_3x3(const int n, const double alpha, const int nz_row, const int stride, double *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_DSPMV_BELL_3X3_THREADS + tx; i < n; i += CUKR_DSPMV_BELL_3X3_CTAS * CUKR_DSPMV_BELL_3X3_THREADS)
	{
		/* register y */
		double y0 = 0.0, y1 = 0.0, y2 = 0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			int2 x0 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))-1);
			int2 x1 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i)));
			int2 x2 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+1);

			/* val */
			int2 val0 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*9);
			int2 val1 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*9+1);
			int2 val2 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*9+2);
			int2 val3 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*9+3);
			int2 val4 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*9+4);
			int2 val5 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*9+5);
			int2 val6 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*9+6);
			int2 val7 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*9+7);
			int2 val8 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*9+8);

			/* GEMV */
			y0 += __hiloint2double(val0.y, val0.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val1.y, val1.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val2.y, val2.x) * __hiloint2double(x2.y, x2.x);
			y1 += __hiloint2double(val3.y, val3.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val4.y, val4.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val5.y, val5.x) * __hiloint2double(x2.y, x2.x);
			y2 += __hiloint2double(val6.y, val6.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val7.y, val7.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val8.y, val8.x) * __hiloint2double(x2.y, x2.x);
		}

		/* copy register y to global memory d_y */
		d_y[i*3  ] += alpha*y0;
		d_y[i*3+1] += alpha*y1;
		d_y[i*3+2] += alpha*y2;
	}
}


/**
 * Calculate SPMV when block size is 4x4
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_dspmv_bell_4x4(const int n, const double alpha, const int nz_row, const int stride, double *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_DSPMV_BELL_4X4_THREADS + tx; i < n; i += CUKR_DSPMV_BELL_4X4_CTAS * CUKR_DSPMV_BELL_4X4_THREADS)
	{
		/* register y */
		double y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			int2 x0 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))-1);
			int2 x1 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i)));
			int2 x2 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+1);
			int2 x3 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+2);

			/* val */
			int2 val0 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*16);
			int2 val1 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*16+1);
			int2 val2 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*16+2);
			int2 val3 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*16+3);
			int2 val4 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*16+4);
			int2 val5 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*16+5);
			int2 val6 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*16+6);
			int2 val7 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*16+7);
			int2 val8 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*16+8);
			int2 val9 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*16+9);
			int2 val10 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*16+10);
			int2 val11 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*16+11);
			int2 val12 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*16+12);
			int2 val13 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*16+13);
			int2 val14 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*16+14);
			int2 val15 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*16+15);

			/* GEMV */
			y0 += __hiloint2double(val0.y, val0.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val1.y, val1.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val2.y, val2.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val3.y, val3.x) * __hiloint2double(x3.y, x3.x);
			y1 += __hiloint2double(val4.y, val4.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val5.y, val5.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val6.y, val6.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val7.y, val7.x) * __hiloint2double(x3.y, x3.x);
			y2 += __hiloint2double(val8.y, val8.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val9.y, val9.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val10.y, val10.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val11.y, val11.x) * __hiloint2double(x3.y, x3.x);
			y3 += __hiloint2double(val12.y, val12.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val13.y, val13.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val14.y, val14.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val15.y, val15.x) * __hiloint2double(x3.y, x3.x);
		}

		/* copy register y to global memory d_y */
		d_y[i*4  ] += alpha*y0;
		d_y[i*4+1] += alpha*y1;
		d_y[i*4+2] += alpha*y2;
		d_y[i*4+3] += alpha*y3;
	}
}


/**
 * Calculate SPMV when block size is 5x5
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_dspmv_bell_5x5(const int n, const double alpha, const int nz_row, const int stride, double *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_DSPMV_BELL_5X5_THREADS + tx; i < n; i += CUKR_DSPMV_BELL_5X5_CTAS * CUKR_DSPMV_BELL_5X5_THREADS)
	{
		/* register y */
		double y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 =0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			int2 x0 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))-1);
			int2 x1 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i)));
			int2 x2 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+1);
			int2 x3 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+2);
			int2 x4 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+3);

			/* val */
			int2 val0 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25);
			int2 val1 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+1);
			int2 val2 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+2);
			int2 val3 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+3);
			int2 val4 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+4);
			int2 val5 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+5);
			int2 val6 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+6);
			int2 val7 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+7);
			int2 val8 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+8);
			int2 val9 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+9);
			int2 val10 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+10);
			int2 val11 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+11);
			int2 val12 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+12);
			int2 val13 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+13);
			int2 val14 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+14);
			int2 val15 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+15);
			int2 val16 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+16);
			int2 val17 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+17);
			int2 val18 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+18);
			int2 val19 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+19);
			int2 val20 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+20);
			int2 val21 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+21);
			int2 val22 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+22);
			int2 val23 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+23);
			int2 val24 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*25+24);

			/* GEMV */
			y0 += __hiloint2double(val0.y, val0.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val1.y, val1.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val2.y, val2.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val3.y, val3.x) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val4.y, val4.x) * __hiloint2double(x4.y, x4.x);
			y1 += __hiloint2double(val5.y, val5.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val6.y, val6.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val7.y, val7.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val8.y, val8.x) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val9.y, val9.x) * __hiloint2double(x4.y, x4.x);
			y2 += __hiloint2double(val10.y, val10.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val11.y, val11.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val12.y, val12.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val13.y, val13.x) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val14.y, val14.x) * __hiloint2double(x4.y, x4.x);
			y3 += __hiloint2double(val15.y, val15.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val16.y, val16.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val17.y, val17.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val18.y, val18.x) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val19.y, val19.x) * __hiloint2double(x4.y, x4.x);
			y4 += __hiloint2double(val20.y, val20.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val21.y, val21.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val22.y, val22.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val23.y, val23.x) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val24.y, val24.x) * __hiloint2double(x4.y, x4.x);
		}

		/* copy register y to global memory d_y */
		d_y[i*5  ] += alpha*y0;
		d_y[i*5+1] += alpha*y1;
		d_y[i*5+2] += alpha*y2;
		d_y[i*5+3] += alpha*y3;
		d_y[i*5+4] += alpha*y4;
	}
}


/**
 * Calculate SPMV when block size is 6x6
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_dspmv_bell_6x6(const int n, const double alpha, const int nz_row, const int stride, double *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_DSPMV_BELL_6X6_THREADS + tx; i < n; i += CUKR_DSPMV_BELL_6X6_CTAS * CUKR_DSPMV_BELL_6X6_THREADS)
	{
		/* register y */
		double y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 =0.0, y5 = 0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			int2 x0 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))-1);
			int2 x1 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i)));
			int2 x2 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+1);
			int2 x3 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+2);
			int2 x4 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+3);
			int2 x5 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+4);

			int2 val0 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36);
			int2 val1 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+1);
			int2 val2 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+2);
			int2 val3 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+3);
			int2 val4 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+4);
			int2 val5 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+5);
			int2 val6 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+6);
			int2 val7 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+7);
			int2 val8 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+8);
			int2 val9 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+9);
			int2 val10 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+10);
			int2 val11 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+11);
			int2 val12 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+12);
			int2 val13 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+13);
			int2 val14 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+14);
			int2 val15 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+15);
			int2 val16 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+16);
			int2 val17 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+17);
			int2 val18 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+18);
			int2 val19 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+19);
			int2 val20 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+20);
			int2 val21 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+21);
			int2 val22 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+22);
			int2 val23 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+23);
			int2 val24 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+24);
			int2 val25 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+25);
			int2 val26 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+26);
			int2 val27 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+27);
			int2 val28 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+28);
			int2 val29 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+29);
			int2 val30 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+30);
			int2 val31 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+31);
			int2 val32 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+32);
			int2 val33 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+33);
			int2 val34 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+34);
			int2 val35 = tex1Dfetch(TEX_VAL2_I, (j*stride+i)*36+35);

			/* GEMV */
			y0 += __hiloint2double(val0.y, val0.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val1.y, val1.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val2.y, val2.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val3.y, val3.x) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val4.y, val4.x) * __hiloint2double(x4.y, x4.x)
			    + __hiloint2double(val5.y, val5.x) * __hiloint2double(x5.y, x5.x);
			y1 += __hiloint2double(val6.y, val6.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val7.y, val7.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val8.y, val8.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val9.y, val9.x) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val10.y, val10.x) * __hiloint2double(x4.y, x4.x)
			    + __hiloint2double(val11.y, val11.x) * __hiloint2double(x5.y, x5.x);
			y2 += __hiloint2double(val12.y, val12.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val13.y, val13.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val14.y, val14.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val15.y, val15.x) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val16.y, val16.x) * __hiloint2double(x4.y, x4.x)
			    + __hiloint2double(val17.y, val17.x) * __hiloint2double(x5.y, x5.x);
			y3 += __hiloint2double(val18.y, val18.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val19.y, val19.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val20.y, val20.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val21.y, val21.x) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val22.y, val22.x) * __hiloint2double(x4.y, x4.x)
			    + __hiloint2double(val23.y, val23.x) * __hiloint2double(x5.y, x5.x);
			y4 += __hiloint2double(val24.y, val24.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val25.y, val25.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val26.y, val26.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val27.y, val27.x) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val28.y, val28.x) * __hiloint2double(x4.y, x4.x)
			    + __hiloint2double(val29.y, val29.x) * __hiloint2double(x5.y, x5.x);
			y5 += __hiloint2double(val30.y, val30.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val31.y, val31.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val32.y, val32.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val33.y, val33.x) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val34.y, val34.x) * __hiloint2double(x4.y, x4.x)
			    + __hiloint2double(val35.y, val35.x) * __hiloint2double(x5.y, x5.x);
		}

		/* copy register y to global memory d_y */
		d_y[i*6  ] += alpha*y0;
		d_y[i*6+1] += alpha*y1;
		d_y[i*6+2] += alpha*y2;
		d_y[i*6+3] += alpha*y3;
		d_y[i*6+4] += alpha*y4;
		d_y[i*6+5] += alpha*y5;
	}
}



