/**
 * @file blas/gpu/cuda/kernel_dspmv_bell4.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Yusa codes.
 *
 * SPMV in BELL4 format
 *
 * Num of registers is {(r+c)*2 + #float4 *4 + 7}
 */
 
#include "../kernels_textures.h"
#include "../kernels_utils.h"

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x 
	#define bx          blockIdx.x
#endif


/**
 * Calculate SPMV when block size is 2x2
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_dspmv_bell4_2x2(const int n, const double alpha, const int nz_row, const int stride, double *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_DSPMV_BELL4_2X2_THREADS + tx; i < n; i += CUKR_DSPMV_BELL4_2X2_CTAS * CUKR_DSPMV_BELL4_2X2_THREADS)
	{
		/* register y */
		double y0 = 0.0, y1 = 0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			int2 x0 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))-1);
			int2 x1 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i)));
			
			/* val */
			int4 val0 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*2);
			int4 val1 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*2+1);

			/* GEMV */
			y0 += __hiloint2double(val0.y, val0.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val0.w, val0.z) * __hiloint2double(x1.y, x1.x);
			y1 += __hiloint2double(val1.y, val1.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val1.w, val1.z) * __hiloint2double(x1.y, x1.x);
		}

		/* copy register y to global memory d_y */
		d_y[i*2  ] += alpha*y0;
		d_y[i*2+1] += alpha*y1;
	}
}


/**
 * Calculate SPMV when block size is 3x3
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_dspmv_bell4_3x3(const int n, const double alpha, const int nz_row, const int stride, double *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_DSPMV_BELL4_3X3_THREADS + tx; i < n; i += CUKR_DSPMV_BELL4_3X3_CTAS * CUKR_DSPMV_BELL4_3X3_THREADS)
	{
		/* register y */
		double y0 = 0.0, y1 = 0.0, y2 = 0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			int2 x0 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))-1);
			int2 x1 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i)));
			int2 x2 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+1);

			/* val */
			int4 val0 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*5);
			int4 val1 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*5+1);
			int4 val2 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*5+2);
			int4 val3 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*5+3);
			int4 val4 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*5+4);

			/* GEMV */
			y0 += __hiloint2double(val0.y, val0.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val0.w, val0.z) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val1.y, val1.x) * __hiloint2double(x2.y, x2.x);
			y1 += __hiloint2double(val1.w, val1.z) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val2.y, val2.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val2.w, val2.z) * __hiloint2double(x2.y, x2.x);
			y2 += __hiloint2double(val3.y, val3.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val3.w, val3.z) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val4.y, val4.x) * __hiloint2double(x2.y, x2.x);
		}

		/* copy register y to global memory d_y */
		d_y[i*3  ] += alpha*y0;
		d_y[i*3+1] += alpha*y1;
		d_y[i*3+2] += alpha*y2;
	}
}


/**
 * Calculate SPMV when block size is 4x4
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_dspmv_bell4_4x4(const int n, const double alpha, const int nz_row, const int stride, double *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_DSPMV_BELL4_4X4_THREADS + tx; i < n; i += CUKR_DSPMV_BELL4_4X4_CTAS * CUKR_DSPMV_BELL4_4X4_THREADS)
	{
		/* register y */
		double y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			int2 x0 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))-1);
			int2 x1 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i)));
			int2 x2 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+1);
			int2 x3 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+2);

			/* val */
			int4 val0 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*8);
			int4 val1 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*8+1);
			int4 val2 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*8+2);
			int4 val3 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*8+3);
			int4 val4 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*8+4);
			int4 val5 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*8+5);
			int4 val6 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*8+6);
			int4 val7 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*8+7);

			/* GEMV */
			y0 += __hiloint2double(val0.y, val0.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val0.w, val0.z) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val1.y, val1.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val1.w, val1.z) * __hiloint2double(x3.y, x3.x);
			y1 += __hiloint2double(val2.y, val2.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val2.w, val2.z) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val3.y, val3.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val3.w, val3.z) * __hiloint2double(x3.y, x3.x);
			y2 += __hiloint2double(val4.y, val4.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val4.w, val4.z) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val5.y, val5.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val5.w, val5.z) * __hiloint2double(x3.y, x3.x);
			y3 += __hiloint2double(val6.y, val6.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val6.w, val6.z) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val7.y, val7.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val7.w, val7.z) * __hiloint2double(x3.y, x3.x);
		}

		/* copy register y to global memory d_y */
		d_y[i*4  ] += alpha*y0;
		d_y[i*4+1] += alpha*y1;
		d_y[i*4+2] += alpha*y2;
		d_y[i*4+3] += alpha*y3;
	}
}


/**
 * Calculate SPMV when block size is 5x5
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_dspmv_bell4_5x5(const int n, const double alpha, const int nz_row, const int stride, double *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_DSPMV_BELL4_5X5_THREADS + tx; i < n; i += CUKR_DSPMV_BELL4_5X5_CTAS * CUKR_DSPMV_BELL4_5X5_THREADS)
	{
		/* register y */
		double y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 =0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			int2 x0 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))-1);
			int2 x1 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i)));
			int2 x2 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+1);
			int2 x3 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+2);
			int2 x4 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+3);

			/* val */
			int4 val0 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*13);
			int4 val1 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*13+1);
			int4 val2 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*13+2);
			int4 val3 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*13+3);
			int4 val4 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*13+4);
			int4 val5 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*13+5);
			int4 val6 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*13+6);
			int4 val7 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*13+7);
			int4 val8 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*13+8);
			int4 val9 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*13+9);
			int4 val10 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*13+10);
			int4 val11 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*13+11);
			int4 val12 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*13+12);

			/* GEMV */
			y0 += __hiloint2double(val0.y, val0.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val0.w, val0.z) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val1.y, val1.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val1.w, val1.z) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val2.y, val2.x) * __hiloint2double(x4.y, x4.x);
			y1 += __hiloint2double(val2.w, val2.z) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val3.y, val3.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val3.w, val3.z) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val4.y, val4.x) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val4.w, val4.z) * __hiloint2double(x4.y, x4.x);
			y2 += __hiloint2double(val5.y, val5.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val5.w, val5.z) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val6.y, val6.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val6.w, val6.z) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val7.y, val7.x) * __hiloint2double(x4.y, x4.x);
			y3 += __hiloint2double(val7.w, val7.z) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val8.y, val8.x) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val8.w, val8.z) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val9.y, val9.x) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val9.w, val9.z) * __hiloint2double(x4.y, x4.x);
			y4 += __hiloint2double(val10.y, val10.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val10.w, val10.z) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val11.y, val11.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val11.w, val11.z) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val12.y, val12.x) * __hiloint2double(x4.y, x4.x);
		}

		/* copy register y to global memory d_y */
		d_y[i*5  ] += alpha*y0;
		d_y[i*5+1] += alpha*y1;
		d_y[i*5+2] += alpha*y2;
		d_y[i*5+3] += alpha*y3;
		d_y[i*5+4] += alpha*y4;
	}
}


/**
 * Calculate SPMV when block size is 6x6
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_dspmv_bell4_6x6(const int n, const double alpha, const int nz_row, const int stride, double *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_DSPMV_BELL4_6X6_THREADS + tx; i < n; i += CUKR_DSPMV_BELL4_6X6_CTAS * CUKR_DSPMV_BELL4_6X6_THREADS)
	{
		/* register y */
		double y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 =0.0, y5 = 0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			int2 x0 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))-1);
			int2 x1 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i)));
			int2 x2 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+1);
			int2 x3 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+2);
			int2 x4 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+3);
			int2 x5 = tex1Dfetch(TEX_X2_I, tex1Dfetch(TEX_IDX, (j*stride+i))+4);

			/* val */
			int4 val0 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*18);
			int4 val1 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*18+1);
			int4 val2 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*18+2);
			int4 val3 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*18+3);
			int4 val4 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*18+4);
			int4 val5 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*18+5);
			int4 val6 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*18+6);
			int4 val7 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*18+7);
			int4 val8 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*18+8);
			int4 val9 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*18+9);
			int4 val10 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*18+10);
			int4 val11 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*18+11);
			int4 val12 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*18+12);
			int4 val13 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*18+13);
			int4 val14 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*18+14);
			int4 val15 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*18+15);
			int4 val16 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*18+16);
			int4 val17 = tex1Dfetch(TEX_VAL4_I, (j*stride+i)*18+17);

			/* GEMV */
			y0 += __hiloint2double(val0.y, val0.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val0.w, val0.z) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val1.y, val1.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val1.w, val1.z) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val2.y, val2.x) * __hiloint2double(x4.y, x4.x)
			    + __hiloint2double(val2.w, val2.z) * __hiloint2double(x5.y, x5.x);
			y1 += __hiloint2double(val3.y, val3.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val3.w, val3.z) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val4.y, val4.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val4.w, val4.z) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val5.y, val5.x) * __hiloint2double(x4.y, x4.x)
			    + __hiloint2double(val5.w, val5.z) * __hiloint2double(x5.y, x5.x);
			y2 += __hiloint2double(val6.y, val6.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val6.w, val6.z) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val7.y, val7.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val7.w, val7.z) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val8.y, val8.x) * __hiloint2double(x4.y, x4.x)
			    + __hiloint2double(val8.w, val8.z) * __hiloint2double(x5.y, x5.x);
			y3 += __hiloint2double(val9.y, val9.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val9.w, val9.z) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val10.y, val10.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val10.w, val10.z) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val11.y, val11.x) * __hiloint2double(x4.y, x4.x)
			    + __hiloint2double(val11.w, val11.z) * __hiloint2double(x5.y, x5.x);
			y4 += __hiloint2double(val12.y, val12.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val12.w, val12.z) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val13.y, val13.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val13.w, val13.z) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val14.y, val14.x) * __hiloint2double(x4.y, x4.x)
			    + __hiloint2double(val14.w, val14.z) * __hiloint2double(x5.y, x5.x);
			y5 += __hiloint2double(val15.y, val15.x) * __hiloint2double(x0.y, x0.x)
			    + __hiloint2double(val15.w, val15.z) * __hiloint2double(x1.y, x1.x)
			    + __hiloint2double(val16.y, val16.x) * __hiloint2double(x2.y, x2.x)
			    + __hiloint2double(val16.w, val16.z) * __hiloint2double(x3.y, x3.x)
			    + __hiloint2double(val17.y, val17.x) * __hiloint2double(x4.y, x4.x)
			    + __hiloint2double(val17.w, val17.z) * __hiloint2double(x5.y, x5.x);
		}

		/* copy register y to global memory d_y */
		d_y[i*6  ] += alpha*y0;
		d_y[i*6+1] += alpha*y1;
		d_y[i*6+2] += alpha*y2;
		d_y[i*6+3] += alpha*y3;
		d_y[i*6+4] += alpha*y4;
		d_y[i*6+5] += alpha*y5;
	}
}



