/**
 * @file blas/gpu/cuda/kernel_dspmv_csr.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Simplest version of SpMV (no data 
 * structure modification, coalescing 
 * the writes)
 * 
 * It can do 2.25 GFlops = 8.86 GB/s 
 * (n = 1000000, 416 threads, 2650 blocks)
 * 
 + Both alpha and beta are considered. 
 */
 
#include "../kernels_textures.h"
#include "../kernels_utils.h"

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x 
	#define bx          blockIdx.x
#endif

 
__global__ void kernel_dspmv_csr( const int     n, 
								  const double *d_val,
								  const double *d_x,     
						  	      const double  alpha,
						                double *d_y)
{
		
	int i;
	
    /* Starting point for this block */
	int ctaStart = bx * CUKR_DSPMV_CSR_THREADS; 
	
	/* Total no. of threads in the kernel */
	int totalThreads = CUKR_DSPMV_CSR_CTAS * CUKR_DSPMV_CSR_THREADS;
	
	/* Read the data*/
	for (i = ctaStart + tx; i < n; i += totalThreads)
	{ 	
		/* Read the beginning and end of the row 
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* Read and sum for the column vectors */
		double sum = 0;		
		for (int j = iRowBeg; j < iRowEnd; j++) {	

			/* Get idx */
			int idx = tex1Dfetch(TEX_IDX, j);
			if (idx) idx--;

			/* Fetch the double value into an int2 struct */
			int2 v_2 = tex1Dfetch(TEX_VAL2_I, j);
			int2 x_2 = tex1Dfetch(TEX_X2_I,   idx);

			/* Add to sum */
            sum += __hiloint2double(v_2.y, v_2.x) * __hiloint2double(x_2.y, x_2.x);
		}

		/* Write the result to shared memory */
		d_y[i] += alpha * sum;		 	
	}  
	
}
