/**
 * @file blas/gpu/cuda/kernel_dspmv_csr4.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * SpMV in single precision for 
 * CSR4 (padded CSR) format
 * 
 * Optimized version
 * Reads 4 elements at a time
 * In order to do that, the vectors
 * have been padded so that the no.
 * of elements in each row is multiple of 4.
 * 
 * It can do up to 3-4GFlop/s 
 * 
 * Both alpha and beta are considered. 
 */

#include "../kernels_textures.h"
#include "../kernels_utils.h"

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x 
	#define bx          blockIdx.x
#endif


 
__global__ void kernel_dspmv_csr4(const int     n, 
								  const double *d_val,
								  const double *d_x,     
						  	      const double  alpha,	
						                double *d_y)
{	
	int i;
	
	/* Starting point for this block */
	int ctaStart = bx * CUKR_DSPMV_CSR4_THREADS; 
	
	/* Total no. of threads in the kernel */
	int totalThreads = CUKR_DSPMV_CSR4_CTAS * CUKR_DSPMV_CSR4_THREADS;
	
	/* Read the data*/
	for (i = ctaStart + tx; i < n; i += totalThreads)
	{ 	
		/* Read the beginning and end of the row 
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* Read and sum for the column vectors */
		double sum = 0;		
		for (int j = iRowBeg / 4; j < iRowEnd / 4; j++) 
		{			
			/* Get idx */
			int4   idx = tex1Dfetch(TEX_IDX4, j);			
			if (idx.x) idx.x--;
			if (idx.y) idx.y--;
			if (idx.z) idx.z--;
			if (idx.w) idx.w--;			

			/* Fetch double values into int4 structs */
			int4 v_4_1 = tex1Dfetch(TEX_VAL4_I, 2*j);
			int4 v_4_2 = tex1Dfetch(TEX_VAL4_I, 2*j + 1);

			/* Fech double values into int4 structs */
			int2 x_2_1 = tex1Dfetch(TEX_X2_I, idx.x);
			int2 x_2_2 = tex1Dfetch(TEX_X2_I, idx.y);
			int2 x_2_3 = tex1Dfetch(TEX_X2_I, idx.z);
			int2 x_2_4 = tex1Dfetch(TEX_X2_I, idx.w);

			/* Make one double out of each two consec. ints and sum */
			sum += __hiloint2double(v_4_1.y, v_4_1.x) * __hiloint2double(x_2_1.y, x_2_1.x) +
				   __hiloint2double(v_4_1.w, v_4_1.z) * __hiloint2double(x_2_2.y, x_2_2.x) +
				   __hiloint2double(v_4_2.y, v_4_2.x) * __hiloint2double(x_2_3.y, x_2_3.x) +
				   __hiloint2double(v_4_2.w, v_4_2.z) * __hiloint2double(x_2_4.y, x_2_4.x);
		}

		/* Write the result to shared memory */		
		d_y[i] += alpha * sum;			
	}  
}
