/**
 * @file blas/gpu/cuda/kernels.cu
 * @author Serban Georgescu
 * @since 2008
 * 
 * Acts as a bridge between the C++ CUDA
 * implementation and the rest of the C 
 * files. All the CUDA kernels must be 
 * included here.
 * 
 * The following kernels are included:
 *  - aypx
 * 
 */

#ifndef CUKR_GPUBLAS_KERNELS_H_
#define CUKR_GPUBLAS_KERNELS_H_


// ****************************  FOR SINGLE  ***************************** //


/**
 * \brief AYPX kernel on the GPU
 * \param n Vector size
 * \param a The constant
 * \param d_x, d_y Device pointers to the 
 *        vectors involved in the operation.
 */ 
#ifdef __cplusplus
extern "C"
#endif 
void cuda_kernel_saypx(int n, float a, float *d_x, float *d_y);


// ------------------------------------------------------------------- // 

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in double precision
 * \param n Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param d_ptr Row index vector (device ptr)
 * \param d_idx Column index vector (device ptr)
 * \param d_val Value vector (device ptr)
 * \param d_x Vector being multiplied (device ptr)
 * \param beta Scaling factor for y
 * \param d_y Result vector (device ptr)
 */
#ifdef __cplusplus
extern "C"
#endif 
void
cuda_sspmv_csr(int rows, int cols, int nz, float alpha, int* d_ptr, int* d_idx, 
			   float* d_val, float* d_x, float beta, float* d_y); 
			   
// ------------------------------------------------------------------- // 

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR4 format, in double precision
 * \param n Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param d_ptr Row index vector (device ptr)
 * \param d_idx Column index vector (device ptr)
 * \param d_val Value vector (device ptr)
 * \param d_x Vector being multiplied (device ptr)
 * \param beta Scaling factor for y
 * \param d_y Result vector (device ptr)
 */
#ifdef __cplusplus
extern "C"
#endif 
void
cuda_sspmv_csr4(int rows, int cols, int nz, float alpha, int* d_ptr, int* d_idx, 
			    float* d_val, float* d_x, float beta, float* d_y); 			   


// ------------------------------------------------------------------- // 

/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR format, in single precision
 * \param rows, cols Matrix size
 * \param r, c Block size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param d_ptr Row index vector (device ptr)
 * \param d_idx Column index vector (device ptr)
 * \param d_val Value vector (device ptr)
 * \param d_x Vector being multiplied (device ptr)
 * \param beta Scaling factor for y
 * \param d_y Result vector (device ptr)
 */
#ifdef __cplusplus
extern "C"
#endif 
void
cuda_sspmv_bcsr(int rows, int cols, int nz, int r, int c, float alpha, int* d_ptr, int* d_idx, 
			   float* d_val, float* d_x, float beta, float* d_y);

// ------------------------------------------------------------------- // 

/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR4 format, in single precision
 * \param rows, cols Matrix size
 * \param r, c Block size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param d_ptr Row index vector (device ptr)
 * \param d_idx Column index vector (device ptr)
 * \param d_val Value vector (device ptr)
 * \param d_x Vector being multiplied (device ptr)
 * \param beta Scaling factor for y
 * \param d_y Result vector (device ptr)
 */
#ifdef __cplusplus
extern "C"
#endif 
void
cuda_sspmv_bcsr4(int rows, int cols, int nz, int r, int c, float alpha, int* d_ptr, int* d_idx, 
			   float* d_val, float* d_x, float beta, float* d_y);

// ------------------------------------------------------------------- // 

/**
 * \brief Wrapper for SSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
#ifdef __cplusplus
extern "C"
#endif
void cuda_sspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                    CUKR_INT* d_ell_idx, CUKR_SINGLE* d_ell_val, CUKR_INT* d_csr_ptr, CUKR_INT* d_csr_idx, CUKR_SINGLE* d_csr_val,
                    CUKR_SINGLE* d_x, CUKR_SINGLE beta, CUKR_SINGLE* d_y);

// ------------------------------------------------------------------- // 

/**
 * \brief Wrapper for SSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
#ifdef __cplusplus
extern "C"
#endif
void cuda_sspmv_bhyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                    CUKR_INT* d_ell_idx, CUKR_SINGLE* d_ell_val, CUKR_INT* d_csr_ptr, CUKR_INT* d_csr_idx, CUKR_SINGLE* d_csr_val,
                    CUKR_SINGLE* d_x, CUKR_SINGLE beta, CUKR_SINGLE* d_y);

// ------------------------------------------------------------------- // 

/**
 * \brief Wrapper for SSPMV_BHYB4
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
#ifdef __cplusplus
extern "C"
#endif
void cuda_sspmv_bhyb4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                    CUKR_INT* d_ell_idx, CUKR_SINGLE* d_ell_val, CUKR_INT* d_csr_ptr, CUKR_INT* d_csr_idx, CUKR_SINGLE* d_csr_val,
                    CUKR_SINGLE* d_x, CUKR_SINGLE beta, CUKR_SINGLE* d_y);


// ****************************  FOR DOUBLE  ***************************** //


/**
 * \brief AYPX kernel on the GPU
 * \param n Vector size
 * \param a The constant
 * \param d_x, d_y Device pointers to the 
 *        vectors involved in the operation.
 */ 
#ifdef __cplusplus
extern "C"
#endif 
void cuda_kernel_daypx(int n, double a, double *d_x, double *d_y);


// ------------------------------------------------------------------- // 

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in double precision
 * \param n Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param d_ptr Row index vector (device ptr)
 * \param d_idx Column index vector (device ptr)
 * \param d_val Value vector (device ptr)
 * \param d_x Vector being multiplied (device ptr)
 * \param beta Scaling factor for y
 * \param d_y Result vector (device ptr)
 */
#ifdef __cplusplus
extern "C"
#endif 
void
cuda_dspmv_csr(int rows, int cols, int nz, double alpha, int* d_ptr, int* d_idx, 
			   double* d_val, double* d_x, double beta, double* d_y); 
			   
// ------------------------------------------------------------------- // 

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR4 format, in double precision
 * \param n Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param d_ptr Row index vector (device ptr)
 * \param d_idx Column index vector (device ptr)
 * \param d_val Value vector (device ptr)
 * \param d_x Vector being multiplied (device ptr)
 * \param beta Scaling factor for y
 * \param d_y Result vector (device ptr)
 */
#ifdef __cplusplus
extern "C"
#endif 
void
cuda_dspmv_csr4(int rows, int cols, int nz, double alpha, int* d_ptr, int* d_idx, 
			    double* d_val, double* d_x, double beta, double* d_y); 			   

// ------------------------------------------------------------------- // 

/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR format, in double precision
 * \param rows, cols Matrix size
 * \param r, c Block size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param d_ptr Row index vector (device ptr)
 * \param d_idx Column index vector (device ptr)
 * \param d_val Value vector (device ptr)
 * \param d_x Vector being multiplied (device ptr)
 * \param beta Scaling factor for y
 * \param d_y Result vector (device ptr)
 */
#ifdef __cplusplus
extern "C"
#endif 
void
cuda_dspmv_bcsr(int rows, int cols, int nz, int r, int c, double alpha, int* d_ptr, int* d_idx, 
			   double* d_val, double* d_x, double beta, double* d_y);

// ------------------------------------------------------------------- // 

/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR4 format, in double precision
 * \param rows, cols Matrix size
 * \param r, c Block size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param d_ptr Row index vector (device ptr)
 * \param d_idx Column index vector (device ptr)
 * \param d_val Value vector (device ptr)
 * \param d_x Vector being multiplied (device ptr)
 * \param beta Scaling factor for y
 * \param d_y Result vector (device ptr)
 */
#ifdef __cplusplus
extern "C"
#endif 
void
cuda_dspmv_bcsr4(int rows, int cols, int nz, int r, int c, double alpha, int* d_ptr, int* d_idx, 
			   double* d_val, double* d_x, double beta, double* d_y);

// ------------------------------------------------------------------- // 

/**
 * \brief Wrapper for DSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
#ifdef __cplusplus
extern "C"
#endif
void cuda_dspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                    CUKR_INT* d_ell_idx, CUKR_DOUBLE* d_ell_val, CUKR_INT* d_csr_ptr, CUKR_INT* d_csr_idx, CUKR_DOUBLE* d_csr_val,
                    CUKR_DOUBLE* d_x, CUKR_DOUBLE beta, CUKR_DOUBLE* d_y);

// ------------------------------------------------------------------- // 

/**
 * \brief Wrapper for DSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
#ifdef __cplusplus
extern "C"
#endif
void cuda_dspmv_bhyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                    CUKR_INT* d_ell_idx, CUKR_DOUBLE* d_ell_val, CUKR_INT* d_csr_ptr, CUKR_INT* d_csr_idx, CUKR_DOUBLE* d_csr_val,
                    CUKR_DOUBLE* d_x, CUKR_DOUBLE beta, CUKR_DOUBLE* d_y);

// ------------------------------------------------------------------- // 

/**
 * \brief Wrapper for DSPMV_BHYB4
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
#ifdef __cplusplus
extern "C"
#endif
void cuda_dspmv_bhyb4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                    CUKR_INT* d_ell_idx, CUKR_DOUBLE* d_ell_val, CUKR_INT* d_csr_ptr, CUKR_INT* d_csr_idx, CUKR_DOUBLE* d_csr_val,
                    CUKR_DOUBLE* d_x, CUKR_DOUBLE beta, CUKR_DOUBLE* d_y);


// *************************  FOR DOUBLE-SINGLE  ************************** //

// ------------------------------------------------------------------- //

/**
 * \brief QDOT kernel in double-single,
 *        on the GPU
 * \param n Vector size
 * \param d_xh, d_xt Device pointers to the 
 *        first vector (head and tail).
 * \param d_yh, d_yt Device pointers to the 
 *        second (result) vector (head and tail).
 */ 
#ifdef __cplusplus
extern "C"
#endif 
double cuda_kernel_qdot(int n, float *d_xh, float *d_xt, float *d_yh, float *d_yt);

// ------------------------------------------------------------------- //

/**
 * \brief QAXPY kernel in double-single,
 *        on the GPU
 * \param n Vector size
 * \param a0,a1 The constant (head and tail)
 * \param d_xh, d_xt Device pointers to the 
 *        first vector (head and tail).
 * \param d_yh, d_yt Device pointers to the 
 *        second (result) vector (head and tail).
 */ 
#ifdef __cplusplus
extern "C"
#endif 
void cuda_kernel_qaxpy(int n, float a0, float a1, float *d_xh, float *d_xt, 
					    	   	                  float *d_yh, float *d_yt);

// ------------------------------------------------------------------- //

/**
 * \brief QAYPX kernel in double-single,
 *        on the GPU
 * \param n Vector size
 * \param a The constant
 * \param d_xh, d_xt Device pointers to the 
 *        first vector (head and tail).
 * \param d_yh, d_yt Device pointers to the 
 *        second (result) vector (head and tail).
 */ 
#ifdef __cplusplus
extern "C"
#endif 
void cuda_kernel_qaypx(int n, float a0, float a1, float *d_xh, float *d_xt, 
					   				              float *d_yh, float *d_yt);
					   				              
// ------------------------------------------------------------------- //

/**
 * \brief QSCAL kernel in double-single,
 *        on the GPU
 * \param n Vector size
 * \param a0,a1 The constant (head and tail)
 * \param d_xh, d_xt Device pointers to the 
 *        vector (head and tail).
 */ 
#ifdef __cplusplus
extern "C"
#endif 
void cuda_kernel_qscal(int n, float a0, float a1, float *d_xh, float *d_xt);					   				              

// ------------------------------------------------------------------- // 
				   				              

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in quasi-double precision
 * \param n Matrix size
 * \param nz Number of nonzeros
 * \param a0,a1 Scaling factor for Ax (head,tail)
 * \param ptr Row index vector 
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param b0, b1 Scale factor for y (head, tail)
 * \param yh,yt Head and tail for the result
 *              vector

 */
#ifdef __cplusplus
extern "C"
#endif 
void
cuda_qspmv_csr(int rows, int cols, int nz, float a0, float a1, int* d_ptr, int* d_idx, 
			   float* d_valh, float* d_valt, float* d_xh, float* d_xt, 
			   float b0, float b1, float* d_yh, float* d_yt);
			   
// ------------------------------------------------------------------- // 

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR4 format, in quasi-double precision
 * \param n Matrix size
 * \param nz Number of nonzeros
 * \param a0,a1 Scaling factor for Ax (head,tail)
 * \param ptr Row index vector 
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param b0, b1 Scale factor for y (head, tail)
 * \param yh,yt Head and tail for the result
 *              vector

 */
#ifdef __cplusplus 
extern "C"
#endif 
void
cuda_qspmv_csr4(int rows, int cols, int nz, float a0, float a1, int* d_ptr, int* d_idx, 
			    float* d_valh, float* d_valt, float* d_xh, float* d_xt, 
			    float b0, float b1, float* d_yh, float* d_yt);

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR format, in quasi-double precision
 * \param n Matrix size
 * \param nz Number of nonzeros
 * \param a0,a1 Scaling factor for Ax (head,tail)
 * \param ptr Row index vector 
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param b0, b1 Scale factor for y (head, tail)
 * \param yh,yt Head and tail for the result
 *              vector

 */
#ifdef __cplusplus
extern "C"
#endif 
void
cuda_qspmv_bcsr(int rows, int cols, int nz, int r, int c, float a0, float a1, int* d_ptr, int* d_idx, 
			   float* d_valh, float* d_valt, float* d_xh, float* d_xt, 
			   float b0, float b1, float* d_yh, float* d_yt);

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR4 format, in quasi-double precision
 * \param n Matrix size
 * \param nz Number of nonzeros
 * \param a0,a1 Scaling factor for Ax (head,tail)
 * \param ptr Row index vector 
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param b0, b1 Scale factor for y (head, tail)
 * \param yh,yt Head and tail for the result
 *              vector

 */
#ifdef __cplusplus
extern "C"
#endif 
void
cuda_qspmv_bcsr4(int rows, int cols, int nz, int r, int c, float a0, float a1, int* d_ptr, int* d_idx, 
			   float* d_valh, float* d_valt, float* d_xh, float* d_xt, 
			   float b0, float b1, float* d_yh, float* d_yt);

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for QSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr_hi/lo CSR ptr vectors
 * \param csr_idx CSR idx vectors
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
#ifdef __cplusplus
extern "C"
#endif
void cuda_qspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha0, CUKR_SINGLE alpha1, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
	                CUKR_INT* d_ell_idx, CUKR_SINGLE* d_ell_valh, CUKR_SINGLE* d_ell_valt, CUKR_INT* d_csr_ptr, CUKR_INT* d_csr_idx, CUKR_SINGLE* d_csr_valh, CUKR_SINGLE* d_csr_valt,
	                CUKR_SINGLE* d_xh, CUKR_SINGLE* d_xt, CUKR_SINGLE beta0, CUKR_SINGLE beta1, CUKR_SINGLE* d_yh, CUKR_SINGLE* d_yt);

// ------------------------------------------------------------------- // 

/**
 * \brief Wrapper for QSPMV_BHYB 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions 
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr_hi/lo CSR ptr vectors
 * \param csr_idx CSR idx vectors 
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y 
 * \param y Result vector (head, tail)
 */
#ifdef __cplusplus
extern "C"
#endif   
void cuda_qspmv_bhyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha0, CUKR_SINGLE alpha1, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz, 
	                CUKR_INT* d_bell_idx, CUKR_SINGLE* d_bell_valh, CUKR_SINGLE* d_bell_valt, CUKR_INT* d_bcsr_ptr, CUKR_INT* d_bcsr_idx, CUKR_SINGLE* d_bcsr_valh, CUKR_SINGLE* d_bcsr_valt,  
	                CUKR_SINGLE* d_xh, CUKR_SINGLE* d_xt, CUKR_SINGLE beta0, CUKR_SINGLE beta1, CUKR_SINGLE* d_yh, CUKR_SINGLE* d_yt);

// ------------------------------------------------------------------- // 

/**
 * \brief Wrapper for QSPMV_BHYB4
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions 
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr_hi/lo CSR ptr vectors
 * \param csr_idx CSR idx vectors 
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y 
 * \param y Result vector (head, tail)
 */
#ifdef __cplusplus
extern "C"
#endif   
void cuda_qspmv_bhyb4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha0, CUKR_SINGLE alpha1, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz, 
	                CUKR_INT* d_bell_idx, CUKR_SINGLE* d_bell_valh, CUKR_SINGLE* d_bell_valt, CUKR_INT* d_bcsr_ptr, CUKR_INT* d_bcsr_idx, CUKR_SINGLE* d_bcsr_valh, CUKR_SINGLE* d_bcsr_valt,  
	                CUKR_SINGLE* d_xh, CUKR_SINGLE* d_xt, CUKR_SINGLE beta0, CUKR_SINGLE beta1, CUKR_SINGLE* d_yh, CUKR_SINGLE* d_yt);


#endif /*CUKR_GPUBLAS_KERNELS_H_*/
