/**
 * @file blas/gpu/cuda/kernels_config.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Contains configuration stuff for the
 * CUDA kernels (e.g. threads, blocks, etc).
 * 
 */

#ifndef KERNELS_CONFIG_G80_H_
#define KERNELS_CONFIG_G80_H_

// ****************************  SETTINGS  ***************************** //

/** Set to 1 to use Kahan summation
 * in the SpMV kernels; 
 * Set to 0 to disable */
#define USE_KAHAN_IN_SPMV 0

// ****************************  FOR SINGLE  ***************************** //


/**
 * Number of thread blocks to be used
 * for the AYPX kernel 
 */
#define CUKR_SAYPX_CTAS  256
/** 
 * Number of threads per block to be
 * used for the AYPX kernel 
 */
#define CUKR_SAYPX_THREADS 128

// -------------------------------------------------------------- //

/** 
 * Number of thread blocks to be used
 * for the SPMV_CSR 
 *
 * use 7 regs
 */
#define CUKR_SSPMV_CSR_CTAS  256
/** 
 * Number of threads per block to be
 * used for the SPMV_CSR kernel 
 */
#define CUKR_SSPMV_CSR_THREADS 512

// -------------------------------------------------------------- //

/** 
 * Number of thread blocks to be used
 * for the SPMV_CSR4 
 *
 * use 14 regs
 */
#define CUKR_SSPMV_CSR4_CTAS   256
/** 
 * Number of threads per block to be
 * used for the SPMV_CSR kernel 
 */
#define CUKR_SSPMV_CSR4_THREADS 512

// ****************************  FOR DOUBLE  ***************************** //


/**
 * Number of thread blocks to be used
 * for the AYPX kernel 
 */
#define CUKR_DAYPX_CTAS  256
/** 
 * Number of threads per block to be
 * used for the AYPX kernel 
 */
#define CUKR_DAYPX_THREADS 128

// -------------------------------------------------------------- //

/** 
 * Number of thread blocks to be used
 * for the SPMV_CSR 
 *
 * use 10 regs
 */
#define CUKR_DSPMV_CSR_CTAS  256
/** 
 * Number of threads per block to be
 * used for the SPMV_CSR kernel 
 */
#define CUKR_DSPMV_CSR_THREADS 512

// -------------------------------------------------------------- //

/** 
 * Number of thread blocks to be used
 * for the SPMV_CSR4 
 *
 * use 17 regs
 */
#define CUKR_DSPMV_CSR4_CTAS  256
/** 
 * Number of threads per block to be
 * used for the SPMV_CSR kernel 
 */
#define CUKR_DSPMV_CSR4_THREADS 480

// *************************  FOR SINGLE-DOUBLE  ************************** //

/**
 * Number of thread blocks to be used
 * for the double-single DOT kernel 
 */
#define CUKR_QDOT_CTAS  256
/** 
 * Number of threads per block to be
 * used for the double-single DOT kernel 
 * IT MUST BE A POWER OF 2 !!! 
 */ 
#define CUKR_QDOT_THREADS 64 

// -------------------------------------------------------------- //

/**
 * Number of thread blocks to be used
 * for the double-single AXPY kernel 
 */
#define CUKR_QAXPY_CTAS  256
/** 
 * Number of threads per block to be
 * used for the double-single AXPY kernel 
 */
#define CUKR_QAXPY_THREADS 192


// -------------------------------------------------------------- //

/**
 * Number of thread blocks to be used
 * for the double-single AYPX kernel 
 */
#define CUKR_QAYPX_CTAS  256
/** 
 * Number of threads per block to be
 * used for the double-single AYPX kernel 
 */
#define CUKR_QAYPX_THREADS 192

// -------------------------------------------------------------- //

/**
 * Number of thread blocks to be used
 * for the double-single SCAL kernel 
 */
#define CUKR_QSCAL_CTAS  256
/** 
 * Number of threads per block to be
 * used for the double-single SCAL kernel 
 */ 
#define CUKR_QSCAL_THREADS 128

// -------------------------------------------------------------- //

/**
 * Number of thread blocks to be used
 * for the double-single CSR kernel,
 * where A = 1 (requires 18 regs) 
 */
#define CUKR_QSPMV_CSR_A1_CTAS  256
/** 
 * Number of threads per block to be used
 * for the double-single CSR kernel,
 * where A = 1 (requires 18 regs) 
 */
#define CUKR_QSPMV_CSR_A1_THREADS 448

/**
 * Number of thread blocks to be used
 * for the double-single CSR kernel,
 * where A != 1 (requires 20 regs) 
 */
#define CUKR_QSPMV_CSR_CTAS  256
/** 
 * Number of threads per block to be used
 * for the double-single CSR kernel,
 * where A != 1 (requires 20 regs) 
 */
#define CUKR_QSPMV_CSR_THREADS 384


// -------------------------------------------------------------- //

/**
 * Number of thread blocks to be used
 * for the double-single SPMV kernel 
 *
 * use 30 regs
 */
#define CUKR_QSPMV_CSR4_CTAS  256

/** 
 * Number of threads per block to be
 * used for the double-single SPMV kernel 
 */
#define CUKR_QSPMV_CSR4_THREADS 256

// -------------------------------------------------------------- //

/**
 * BLOCK FORMATS
 * Define block and threads by the num. of registers
 *
 * Rule is following
 * #blocks  : 256
 * #threads : (int)(8192/32/regs)*32  (if >512 then 512)
 *
 * Changed some #threads because of error (I don't know the reason)
 */

/* single */

/* 15 regs */
#define CUKR_SSPMV_BCSR_2X2_CTAS 256
#define CUKR_SSPMV_BCSR_2X2_THREADS 512
/* 16 regs */
#define CUKR_SSPMV_BCSR_3X3_CTAS 256
#define CUKR_SSPMV_BCSR_3X3_THREADS 512
/* 19 regs. 416 couldn't */
#define CUKR_SSPMV_BCSR_4X4_CTAS 256
#define CUKR_SSPMV_BCSR_4X4_THREADS 384
/* 22 regs. 352 couldn't */
#define CUKR_SSPMV_BCSR_5X5_CTAS 256
#define CUKR_SSPMV_BCSR_5X5_THREADS 320
/* 24 regs */
#define CUKR_SSPMV_BCSR_6X6_CTAS 256
#define CUKR_SSPMV_BCSR_6X6_THREADS 320

/* 14 regs */
#define CUKR_SSPMV_BCSR4_2X2_CTAS 256
#define CUKR_SSPMV_BCSR4_2X2_THREADS 512
/* 22 regs. 352 couldn't. */
#define CUKR_SSPMV_BCSR4_3X3_CTAS 256
#define CUKR_SSPMV_BCSR4_3X3_THREADS 320
/* 23 regs. 352 couldn't */
#define CUKR_SSPMV_BCSR4_4X4_CTAS 256
#define CUKR_SSPMV_BCSR4_4X4_THREADS 320
/* 31 regs */
#define CUKR_SSPMV_BCSR4_5X5_CTAS 256
#define CUKR_SSPMV_BCSR4_5X5_THREADS 256
/* 31 regs */
#define CUKR_SSPMV_BCSR4_6X6_CTAS 256
#define CUKR_SSPMV_BCSR4_6X6_THREADS 256

/* 19 regs. 416 couldn't */
#define CUKR_SSPMV_BELL_2X2_CTAS 256
#define CUKR_SSPMV_BELL_2X2_THREADS 384
/* 27 regs. 288 couldn't */
#define CUKR_SSPMV_BELL_3X3_CTAS 256
#define CUKR_SSPMV_BELL_3X3_THREADS 256
/* 35 regs. 224 couldn't */
#define CUKR_SSPMV_BELL_4X4_CTAS 256
#define CUKR_SSPMV_BELL_4X4_THREADS 192
/* 47 regs. 160 couldn't */
#define CUKR_SSPMV_BELL_5X5_CTAS 256
#define CUKR_SSPMV_BELL_5X5_THREADS 128
/* 60 regs */
#define CUKR_SSPMV_BELL_6X6_CTAS 256
#define CUKR_SSPMV_BELL_6X6_THREADS 128

/* 14 regs */
#define CUKR_SSPMV_BELL4_2X2_CTAS 256
#define CUKR_SSPMV_BELL4_2X2_THREADS 512
/* 25 regs */
#define CUKR_SSPMV_BELL4_3X3_CTAS 256
#define CUKR_SSPMV_BELL4_3X3_THREADS 320
/* 31 regs */
#define CUKR_SSPMV_BELL4_4X4_CTAS 256
#define CUKR_SSPMV_BELL4_4X4_THREADS 256
/* 38 regs */
#define CUKR_SSPMV_BELL4_5X5_CTAS 256
#define CUKR_SSPMV_BELL4_5X5_THREADS 192
/* 42 regs */
#define CUKR_SSPMV_BELL4_6X6_CTAS 256
#define CUKR_SSPMV_BELL4_6X6_THREADS 192

/* quasi-double */

/* 31 regs */
#define CUKR_QSPMV_BCSR_2X2_CTAS 256
#define CUKR_QSPMV_BCSR_2X2_THREADS 256
/* 39 regs */
#define CUKR_QSPMV_BCSR_3X3_CTAS 256
#define CUKR_QSPMV_BCSR_3X3_THREADS 192
/* 42 regs */
#define CUKR_QSPMV_BCSR_4X4_CTAS 256
#define CUKR_QSPMV_BCSR_4X4_THREADS 192
/* 55 regs */
#define CUKR_QSPMV_BCSR_5X5_CTAS 256
#define CUKR_QSPMV_BCSR_5X5_THREADS 128
/* 58 regs */
#define CUKR_QSPMV_BCSR_6X6_CTAS 256
#define CUKR_QSPMV_BCSR_6X6_THREADS 128

/* 32 regs */
#define CUKR_QSPMV_BCSR4_2X2_CTAS 256
#define CUKR_QSPMV_BCSR4_2X2_THREADS 256
/* 41 regs */
#define CUKR_QSPMV_BCSR4_3X3_CTAS 256
#define CUKR_QSPMV_BCSR4_3X3_THREADS 192
/* 59 regs */
#define CUKR_QSPMV_BCSR4_4X4_CTAS 256
#define CUKR_QSPMV_BCSR4_4X4_THREADS 128
/* 60 regs */
#define CUKR_QSPMV_BCSR4_5X5_CTAS 256
#define CUKR_QSPMV_BCSR4_5X5_THREADS 128
/* 60 regs */
#define CUKR_QSPMV_BCSR4_6X6_CTAS 256
#define CUKR_QSPMV_BCSR4_6X6_THREADS 128

/* 31 regs */
#define CUKR_QSPMV_BELL_2X2_CTAS 256
#define CUKR_QSPMV_BELL_2X2_THREADS 256
/* 39 regs */
#define CUKR_QSPMV_BELL_3X3_CTAS 256
#define CUKR_QSPMV_BELL_3X3_THREADS 192
/* 42 regs */
#define CUKR_QSPMV_BELL_4X4_CTAS 256
#define CUKR_QSPMV_BELL_4X4_THREADS 192
/* 55 regs */
#define CUKR_QSPMV_BELL_5X5_CTAS 256
#define CUKR_QSPMV_BELL_5X5_THREADS 128
/* 58 regs */
#define CUKR_QSPMV_BELL_6X6_CTAS 256
#define CUKR_QSPMV_BELL_6X6_THREADS 128

/* 32 regs */
#define CUKR_QSPMV_BELL4_2X2_CTAS 256
#define CUKR_QSPMV_BELL4_2X2_THREADS 256
/* 49 regs */
#define CUKR_QSPMV_BELL4_3X3_CTAS 256
#define CUKR_QSPMV_BELL4_3X3_THREADS 160
/* 60 regs */
#define CUKR_QSPMV_BELL4_4X4_CTAS 256
#define CUKR_QSPMV_BELL4_4X4_THREADS 128
/* 60 regs */
#define CUKR_QSPMV_BELL4_5X5_CTAS 256
#define CUKR_QSPMV_BELL4_5X5_THREADS 128
/* 60 regs */
#define CUKR_QSPMV_BELL4_6X6_CTAS 256
#define CUKR_QSPMV_BELL4_6X6_THREADS 128

/**
 * double
 * Though G80 cannot run in double precision,
 * i set from #regs; 8192 here.
 * Should be set from 16384 (GT200's spec)
 */

/* 19 regs */
#define CUKR_DSPMV_BCSR_2X2_CTAS 256
#define CUKR_DSPMV_BCSR_2X2_THREADS 416
/* 31 regs */
#define CUKR_DSPMV_BCSR_3X3_CTAS 256
#define CUKR_DSPMV_BCSR_3X3_THREADS 256
/* 46 regs */
#define CUKR_DSPMV_BCSR_4X4_CTAS 256
#define CUKR_DSPMV_BCSR_4X4_THREADS 160
/* 60 regs */
#define CUKR_DSPMV_BCSR_5X5_CTAS 256
#define CUKR_DSPMV_BCSR_5X5_THREADS 128
/* 60 regs */
#define CUKR_DSPMV_BCSR_6X6_CTAS 256
#define CUKR_DSPMV_BCSR_6X6_THREADS 128

/* 21 regs */
#define CUKR_DSPMV_BCSR4_2X2_CTAS 256
#define CUKR_DSPMV_BCSR4_2X2_THREADS 384
/* 32 regs */
#define CUKR_DSPMV_BCSR4_3X3_CTAS 256
#define CUKR_DSPMV_BCSR4_3X3_THREADS 256
/* 48 regs */
#define CUKR_DSPMV_BCSR4_4X4_CTAS 256
#define CUKR_DSPMV_BCSR4_4X4_THREADS 160
/* 60 regs */
#define CUKR_DSPMV_BCSR4_5X5_CTAS 256
#define CUKR_DSPMV_BCSR4_5X5_THREADS 128
/* 57 regs */
#define CUKR_DSPMV_BCSR4_6X6_CTAS 256
#define CUKR_DSPMV_BCSR4_6X6_THREADS 128

/* 24 regs */
#define CUKR_DSPMV_BELL_2X2_CTAS 256
#define CUKR_DSPMV_BELL_2X2_THREADS 320
/* 40 regs */
#define CUKR_DSPMV_BELL_3X3_CTAS 256
#define CUKR_DSPMV_BELL_3X3_THREADS 192
/* 60 regs */
#define CUKR_DSPMV_BELL_4X4_CTAS 256
#define CUKR_DSPMV_BELL_4X4_THREADS 128
/* 59 regs */
#define CUKR_DSPMV_BELL_5X5_CTAS 256
#define CUKR_DSPMV_BELL_5X5_THREADS 128
/* 59 regs */
#define CUKR_DSPMV_BELL_6X6_CTAS 256
#define CUKR_DSPMV_BELL_6X6_THREADS 128

/* 23 regs */
#define CUKR_DSPMV_BELL4_2X2_CTAS 256
#define CUKR_DSPMV_BELL4_2X2_THREADS 352
/* 37 regs */
#define CUKR_DSPMV_BELL4_3X3_CTAS 256
#define CUKR_DSPMV_BELL4_3X3_THREADS 192
/* 56 regs */
#define CUKR_DSPMV_BELL4_4X4_CTAS 256
#define CUKR_DSPMV_BELL4_4X4_THREADS 128
/* 57 regs */
#define CUKR_DSPMV_BELL4_5X5_CTAS 256
#define CUKR_DSPMV_BELL4_5X5_THREADS 128
/* 58 regs */
#define CUKR_DSPMV_BELL4_6X6_CTAS 256
#define CUKR_DSPMV_BELL4_6X6_THREADS 128

#endif /*KERNELS_CONFIG_G80_H_*/
