/**
 * @file blas/gpu/cuda/kernels_config.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Contains configuration stuff for the
 * CUDA kernels (e.g. threads, blocks, etc).
 * 
 */

#ifndef KERNELS_CONFIG_G80_H_
#define KERNELS_CONFIG_G80_H_

// ****************************  SETTINGS  ***************************** //

/** Set to 1 to use Kahan summation
 * in the SpMV kernels; 
 * Set to 0 to disable */
#define USE_KAHAN_IN_SPMV 0

// ****************************  FOR SINGLE  ***************************** //


/**
 * Number of thread blocks to be used
 * for the AYPX kernel 
 */
#define CUKR_SAYPX_CTAS  256
/** 
 * Number of threads per block to be
 * used for the AYPX kernel 
 */
#define CUKR_SAYPX_THREADS 128

// -------------------------------------------------------------- //

/** 
 * Number of thread blocks to be used
 * for the SPMV_CSR 
 *
 * use 7 regs
 */
#define CUKR_SSPMV_CSR_CTAS  480
/** 
 * Number of threads per block to be
 * used for the SPMV_CSR kernel 
 */
#define CUKR_SSPMV_CSR_THREADS 512

// -------------------------------------------------------------- //

/** 
 * Number of thread blocks to be used
 * for the SPMV_CSR4 
 *
 * use 14 regs
 */
#define CUKR_SSPMV_CSR4_CTAS   480
/** 
 * Number of threads per block to be
 * used for the SPMV_CSR4 kernel 
 */
#define CUKR_SSPMV_CSR4_THREADS 512

// ****************************  FOR DOUBLE  ***************************** //


/**
 * Number of thread blocks to be used
 * for the AYPX kernel 
 */
#define CUKR_DAYPX_CTAS  256
/** 
 * Number of threads per block to be
 * used for the AYPX kernel 
 */
#define CUKR_DAYPX_THREADS 128

// -------------------------------------------------------------- //

/** 
 * Number of thread blocks to be used
 * for the SPMV_CSR 
 *
 * use 13 regs
 */
#define CUKR_DSPMV_CSR_CTAS  480
/** 
 * Number of threads per block to be
 * used for the SPMV_CSR kernel 
 */
#define CUKR_DSPMV_CSR_THREADS 512

// -------------------------------------------------------------- //

/** 
 * Number of thread blocks to be used
 * for the SPMV_CSR4 
 *
 * use 29 regs
 */
#define CUKR_DSPMV_CSR4_CTAS  480
/** 
 * Number of threads per block to be
 * used for the SPMV_CSR4 kernel 
 */
#define CUKR_DSPMV_CSR4_THREADS 512

// *************************  FOR SINGLE-DOUBLE  ************************** //

/**
 * Number of thread blocks to be used
 * for the double-single DOT kernel 
 */
#define CUKR_QDOT_CTAS  256
/** 
 * Number of threads per block to be
 * used for the double-single DOT kernel 
 * IT MUST BE A POWER OF 2 !!! 
 */ 
#define CUKR_QDOT_THREADS 64 

// -------------------------------------------------------------- //

/**
 * Number of thread blocks to be used
 * for the double-single AXPY kernel 
 */
#define CUKR_QAXPY_CTAS  256
/** 
 * Number of threads per block to be
 * used for the double-single AXPY kernel 
 */
#define CUKR_QAXPY_THREADS 192


// -------------------------------------------------------------- //

/**
 * Number of thread blocks to be used
 * for the double-single AYPX kernel 
 */
#define CUKR_QAYPX_CTAS  256
/** 
 * Number of threads per block to be
 * used for the double-single AYPX kernel 
 */
#define CUKR_QAYPX_THREADS 192

// -------------------------------------------------------------- //

/**
 * Number of thread blocks to be used
 * for the double-single SCAL kernel 
 */
#define CUKR_QSCAL_CTAS  256
/** 
 * Number of threads per block to be
 * used for the double-single SCAL kernel 
 */ 
#define CUKR_QSCAL_THREADS 128

// -------------------------------------------------------------- //

/**
 * Number of thread blocks to be used
 * for the double-single CSR kernel,
 * where A = 1 (requires 18 regs) 
 */
#define CUKR_QSPMV_CSR_A1_CTAS  480
/** 
 * Number of threads per block to be used
 * for the double-single CSR kernel,
 * where A = 1 (requires 18 regs) 
 */
#define CUKR_QSPMV_CSR_A1_THREADS 512

/**
 * Number of thread blocks to be used
 * for the double-single CSR kernel,
 * where A != 1 (requires 20 regs) 
 */
#define CUKR_QSPMV_CSR_CTAS  480
/** 
 * Number of threads per block to be used
 * for the double-single CSR kernel,
 * where A != 1 (requires 20 regs) 
 */
#define CUKR_QSPMV_CSR_THREADS 512


// -------------------------------------------------------------- //

/**
 * Number of thread blocks to be used
 * for the double-single SPMV kernel 
 *
 * use 30 regs
 */
#define CUKR_QSPMV_CSR4_CTAS  480

/** 
 * Number of threads per block to be
 * used for the double-single SPMV kernel 
 */
#define CUKR_QSPMV_CSR4_THREADS 512

// -------------------------------------------------------------- //


/**
 * BLOCK FORMATS
 * Define block and threads by the num. of registers
 *
 * Rule is following
 * #blocks  : 480
 * #threads : (int)(16384/32/regs)*32  (if >512 then 512)
 *
 * Changed some #threads because of error (I don't know the reason)
 */

/* single */

/* 15 regs */
#define CUKR_SSPMV_BCSR_2X2_CTAS 480
#define CUKR_SSPMV_BCSR_2X2_THREADS 512
/* 16 regs */
#define CUKR_SSPMV_BCSR_3X3_CTAS 480
#define CUKR_SSPMV_BCSR_3X3_THREADS 512
/* 18 regs */
#define CUKR_SSPMV_BCSR_4X4_CTAS 480
#define CUKR_SSPMV_BCSR_4X4_THREADS 512
/* 22 regs */
#define CUKR_SSPMV_BCSR_5X5_CTAS 480
#define CUKR_SSPMV_BCSR_5X5_THREADS 512
/* 24 regs */
#define CUKR_SSPMV_BCSR_6X6_CTAS 480
#define CUKR_SSPMV_BCSR_6X6_THREADS 512

/* 14 regs */
#define CUKR_SSPMV_BCSR4_2X2_CTAS 480
#define CUKR_SSPMV_BCSR4_2X2_THREADS 512
/* 21 regs */
#define CUKR_SSPMV_BCSR4_3X3_CTAS 480
#define CUKR_SSPMV_BCSR4_3X3_THREADS 512
/* 19 regs */
#define CUKR_SSPMV_BCSR4_4X4_CTAS 480
#define CUKR_SSPMV_BCSR4_4X4_THREADS 512
/* 25 regs */
#define CUKR_SSPMV_BCSR4_5X5_CTAS 480
#define CUKR_SSPMV_BCSR4_5X5_THREADS 512
/* 31 regs */
#define CUKR_SSPMV_BCSR4_6X6_CTAS 480
#define CUKR_SSPMV_BCSR4_6X6_THREADS 512

/* 17 regs */
#define CUKR_SSPMV_BELL_2X2_CTAS 480
#define CUKR_SSPMV_BELL_2X2_THREADS 512
/* 27 regs */
#define CUKR_SSPMV_BELL_3X3_CTAS 480
#define CUKR_SSPMV_BELL_3X3_THREADS 512
/* 35 regs */
#define CUKR_SSPMV_BELL_4X4_CTAS 480
#define CUKR_SSPMV_BELL_4X4_THREADS 448
/* 47 regs */
#define CUKR_SSPMV_BELL_5X5_CTAS 480
#define CUKR_SSPMV_BELL_5X5_THREADS 320
/* 60 regs */
#define CUKR_SSPMV_BELL_6X6_CTAS 480
#define CUKR_SSPMV_BELL_6X6_THREADS 256

/* 14 regs */
#define CUKR_SSPMV_BELL4_2X2_CTAS 480
#define CUKR_SSPMV_BELL4_2X2_THREADS 512
/* 24 regs */
#define CUKR_SSPMV_BELL4_3X3_CTAS 480
#define CUKR_SSPMV_BELL4_3X3_THREADS 512
/* 27 regs */
#define CUKR_SSPMV_BELL4_4X4_CTAS 480
#define CUKR_SSPMV_BELL4_4X4_THREADS 512
/* 32 regs */
#define CUKR_SSPMV_BELL4_5X5_CTAS 480
#define CUKR_SSPMV_BELL4_5X5_THREADS 448
/* 42 regs */
#define CUKR_SSPMV_BELL4_6X6_CTAS 480
#define CUKR_SSPMV_BELL4_6X6_THREADS 352

/* quasi-double */

/* 31 regs */
#define CUKR_QSPMV_BCSR_2X2_CTAS 480
#define CUKR_QSPMV_BCSR_2X2_THREADS 512
/* 39 regs */
#define CUKR_QSPMV_BCSR_3X3_CTAS 480
#define CUKR_QSPMV_BCSR_3X3_THREADS 416
/* 47 regs */
#define CUKR_QSPMV_BCSR_4X4_CTAS 480
#define CUKR_QSPMV_BCSR_4X4_THREADS 320
/* 50 regs */
#define CUKR_QSPMV_BCSR_5X5_CTAS 480
#define CUKR_QSPMV_BCSR_5X5_THREADS 320
/* 60 regs */
#define CUKR_QSPMV_BCSR_6X6_CTAS 480
#define CUKR_QSPMV_BCSR_6X6_THREADS 256

/* 32 regs */
#define CUKR_QSPMV_BCSR4_2X2_CTAS 480
#define CUKR_QSPMV_BCSR4_2X2_THREADS 512
/* 41 regs */
#define CUKR_QSPMV_BCSR4_3X3_CTAS 480
#define CUKR_QSPMV_BCSR4_3X3_THREADS 384
/* 51 regs */
#define CUKR_QSPMV_BCSR4_4X4_CTAS 480
#define CUKR_QSPMV_BCSR4_4X4_THREADS 320
/* 64 regs */
#define CUKR_QSPMV_BCSR4_5X5_CTAS 480
#define CUKR_QSPMV_BCSR4_5X5_THREADS 256
/* 71 regs */
#define CUKR_QSPMV_BCSR4_6X6_CTAS 480
#define CUKR_QSPMV_BCSR4_6X6_THREADS 224

/* 31 regs */
#define CUKR_QSPMV_BELL_2X2_CTAS 480
#define CUKR_QSPMV_BELL_2X2_THREADS 512
/* 39 regs */
#define CUKR_QSPMV_BELL_3X3_CTAS 480
#define CUKR_QSPMV_BELL_3X3_THREADS 416
/* 47 regs */
#define CUKR_QSPMV_BELL_4X4_CTAS 480
#define CUKR_QSPMV_BELL_4X4_THREADS 320
/* 50 regs */
#define CUKR_QSPMV_BELL_5X5_CTAS 480
#define CUKR_QSPMV_BELL_5X5_THREADS 320
/* 59 regs */
#define CUKR_QSPMV_BELL_6X6_CTAS 480
#define CUKR_QSPMV_BELL_6X6_THREADS 256

/* 32 regs */
#define CUKR_QSPMV_BELL4_2X2_CTAS 480
#define CUKR_QSPMV_BELL4_2X2_THREADS 512
/* 50 regs */
#define CUKR_QSPMV_BELL4_3X3_CTAS 480
#define CUKR_QSPMV_BELL4_3X3_THREADS 320
/* 51 regs */
#define CUKR_QSPMV_BELL4_4X4_CTAS 480
#define CUKR_QSPMV_BELL4_4X4_THREADS 320
/* 63 regs */
#define CUKR_QSPMV_BELL4_5X5_CTAS 480
#define CUKR_QSPMV_BELL4_5X5_THREADS 256
/* 80 regs */
#define CUKR_QSPMV_BELL4_6X6_CTAS 480
#define CUKR_QSPMV_BELL4_6X6_THREADS 192

/* double */

/* 21 regs */
#define CUKR_DSPMV_BCSR_2X2_CTAS 480
#define CUKR_DSPMV_BCSR_2X2_THREADS 512
/* 31 regs */
#define CUKR_DSPMV_BCSR_3X3_CTAS 480
#define CUKR_DSPMV_BCSR_3X3_THREADS 512
/* 33 regs. 480 couldn"t */
#define CUKR_DSPMV_BCSR_4X4_CTAS 480
#define CUKR_DSPMV_BCSR_4X4_THREADS 448
/* 40 regs */
#define CUKR_DSPMV_BCSR_5X5_CTAS 480
#define CUKR_DSPMV_BCSR_5X5_THREADS 384
/* 41 regs */
#define CUKR_DSPMV_BCSR_6X6_CTAS 480
#define CUKR_DSPMV_BCSR_6X6_THREADS 384

/* 21 regs */
#define CUKR_DSPMV_BCSR4_2X2_CTAS 480
#define CUKR_DSPMV_BCSR4_2X2_THREADS 512
/* 33 regs. 480 couldn't */
#define CUKR_DSPMV_BCSR4_3X3_CTAS 480
#define CUKR_DSPMV_BCSR4_3X3_THREADS 448
/* 41 regs */
#define CUKR_DSPMV_BCSR4_4X4_CTAS 480
#define CUKR_DSPMV_BCSR4_4X4_THREADS 384
/* 47 regs */
#define CUKR_DSPMV_BCSR4_5X5_CTAS 480
#define CUKR_DSPMV_BCSR4_5X5_THREADS 320
/* 50 regs */
#define CUKR_DSPMV_BCSR4_6X6_CTAS 480
#define CUKR_DSPMV_BCSR4_6X6_THREADS 320

/* 25 regs */
#define CUKR_DSPMV_BELL_2X2_CTAS 480
#define CUKR_DSPMV_BELL_2X2_THREADS 512
/* 40 regs */
#define CUKR_DSPMV_BELL_3X3_CTAS 480
#define CUKR_DSPMV_BELL_3X3_THREADS 384
/* 49 regs */
#define CUKR_DSPMV_BELL_4X4_CTAS 480
#define CUKR_DSPMV_BELL_4X4_THREADS 320
/* 62 regs */
#define CUKR_DSPMV_BELL_5X5_CTAS 480
#define CUKR_DSPMV_BELL_5X5_THREADS 256
/* 79 regs */
#define CUKR_DSPMV_BELL_6X6_CTAS 480
#define CUKR_DSPMV_BELL_6X6_THREADS 192

/* 27 regs */
#define CUKR_DSPMV_BELL4_2X2_CTAS 480
#define CUKR_DSPMV_BELL4_2X2_THREADS 512
/* 36 regs */
#define CUKR_DSPMV_BELL4_3X3_CTAS 480
#define CUKR_DSPMV_BELL4_3X3_THREADS 448
/* 49 regs */
#define CUKR_DSPMV_BELL4_4X4_CTAS 480
#define CUKR_DSPMV_BELL4_4X4_THREADS 320
/* 60 regs */
#define CUKR_DSPMV_BELL4_5X5_CTAS 480
#define CUKR_DSPMV_BELL4_5X5_THREADS 256
/* 63 regs */
#define CUKR_DSPMV_BELL4_6X6_CTAS 480
#define CUKR_DSPMV_BELL4_6X6_THREADS 256

#endif /*KERNELS_CONFIG_G80_H_*/
