/**
 * @file blas/gpu/cuda/kernels_double.cu
 * @author Serban Georgescu
 * @since 2008
 * 
 * Acts as a bridge between the C++ CUDA
 * implementation and the rest of the C 
 * files. All the CUDA double precision
 * kernels must be included here.
 * 
 * The following kernels are included:
 *  - aypx
 *  + spmv_csr
 * 	- spmv_csr
 * 	- spmv_csr_b0
 *  + spmv_csr4
 * 	- spmv_csr4
 * 	- spmv_csr4_b0
 *  + spmv_bcsr
 * 	- spmv_bcsr
 * 	- spmv_bcsr_b0
 *  + spmv_bcsr4
 * 	- spmv_bcsr4
 * 	- spmv_bcsr4_b0
 */

#include <stdio.h>
#include <cuda.h>
#include <cutil.h>
#include <cublas.h>
#include <unistd.h>
#include "../../../../cukr.h"
#include "../../../../util/macros.h"

#include "kernels_config.h"
#include "double/kernel_daypx.h"
#include "double/kernel_dspmv_csr.h"
#include "double/kernel_dspmv_csr_b0.h"
#include "double/kernel_dspmv_csr4.h"
#include "double/kernel_dspmv_csr4_b0.h"

// BCSR
#include "double/kernel_dspmv_bcsr_b0.h"
#include "double/kernel_dspmv_bcsr.h"

// BCSR4
#include "double/kernel_dspmv_bcsr4_b0.h"
#include "double/kernel_dspmv_bcsr4.h"

// ELL
#include "double/kernel_dspmv_ell.h"

// BELL
#include "double/kernel_dspmv_bell.h"

// BELL4
#include "double/kernel_dspmv_bell4.h"

#define MAX(a,b) (a > b ? a : b)
#define CUKR_TRUE  1
#define CUKR_FALSE 0
#define DIVIDE_INTO(x,y) ((x + y - 1)/y) 


/**
 * \brief AYPX kernl on the GPU
 * \param n Vector size
 * \param a The constant
 * \param d_x, d_y Device pointers to the 
 *        vectors involved in the operation.
 */
#ifdef __cplusplus
extern "C"
#endif 
void cuda_kernel_daypx(int n, double a, double *d_x, double *d_y)
{	
	/* Preparations */
	int dynMem = 2 * CUKR_DAYPX_THREADS * sizeof(double); 
	
	/* Kernel invocation */
	kernel_daypx<<<CUKR_DAYPX_CTAS, CUKR_DAYPX_THREADS, dynMem>>> (d_x, d_y, a, n);  
	CUT_CHECK_ERROR("kernel_daypx() execution failed\n"); 
	
	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize()); 
		
}

// ------------------------------------------------------------------- // 

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in double precision
 * \param n Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param d_ptr Row index vector (device ptr)
 * \param d_idx Column index vector (device ptr)
 * \param d_val Value vector (device ptr)
 * \param d_x Vector being multiplied (device ptr)
 * \param beta Scaling factor for y
 * \param d_y Result vector (device ptr)
 */
#ifdef __cplusplus
extern "C"
#endif 
void
cuda_dspmv_csr(int rows, int cols, int nz, double alpha, int* d_ptr, int* d_idx, 
			   double* d_val, double* d_x, double beta, double* d_y) 
{
			
	/* Bind textures */		
	cudaBindTexture(0, TEX_X2_I  , d_x  , cols       * sizeof(d_x[0]));
  	cudaBindTexture(0, TEX_PTR   , d_ptr, (rows + 1) * sizeof(d_ptr[0]));
   	cudaBindTexture(0, TEX_IDX   , d_idx, nz         * sizeof(d_idx[0]));
   	cudaBindTexture(0, TEX_VAL2_I, d_val, nz         * sizeof(double));
   	
	/*
	 * Chose which kernel to run
	 */
	
	if (beta != 1 && beta)
	 	cublasDscal(rows, beta, d_y, 1);
	
	if (beta) {
		kernel_dspmv_csr <<<CUKR_DSPMV_CSR_CTAS, CUKR_DSPMV_CSR_THREADS>>> (rows, d_val, d_x, alpha, d_y);
    	CUT_CHECK_ERROR("kernel_dspmv_csr() execution failed\n");
	}   
	else {
		kernel_dspmv_csr_b0 <<<CUKR_DSPMV_CSR_CTAS, CUKR_DSPMV_CSR_THREADS>>> (rows, d_val, d_x, alpha, d_y);
    	CUT_CHECK_ERROR("kernel_dspmv_csr_b0() execution failed\n");
	} 		
		
	 
   	/* Un-bind textures */
	cudaUnbindTexture(TEX_X2_I);
   	cudaUnbindTexture(TEX_PTR);
   	cudaUnbindTexture(TEX_IDX);
   	cudaUnbindTexture(TEX_VAL2_I);
   	
   	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize());
 	
}		

// ------------------------------------------------------------------- // 

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR4 format, in double precision
 * \param n Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param d_ptr Row index vector (device ptr)
 * \param d_idx Column index vector (device ptr)
 * \param d_val Value vector (device ptr)
 * \param d_x Vector being multiplied (device ptr)
 * \param beta Scaling factor for y
 * \param d_y Result vector (device ptr)
 */
#ifdef __cplusplus
extern "C"
#endif 
void
cuda_dspmv_csr4(int rows, int cols, int nz, double alpha, int* d_ptr, int* d_idx, 
			   double* d_val, double* d_x, double beta, double* d_y) 
{			
	/* Bind textures */		
	cudaBindTexture(0, TEX_X2_I  , d_x  , cols       * sizeof(d_x[0]));
  	cudaBindTexture(0, TEX_PTR   , d_ptr, (rows + 1) * sizeof(d_ptr[0]));
   	cudaBindTexture(0, TEX_IDX4  , d_idx, nz         * sizeof(d_idx[0]));
   	cudaBindTexture(0, TEX_VAL4_I, d_val, nz         * sizeof(d_val[0]));
   	
	/*
	 * Chose which kernel to run
	 */
	
	if (beta != 1 && beta)
	 	cublasDscal(rows, beta, d_y, 1);
	
	if (beta) {
		kernel_dspmv_csr4 <<<CUKR_DSPMV_CSR4_CTAS, CUKR_DSPMV_CSR4_THREADS>>> (rows, d_val, d_x, alpha, d_y);
	    CUT_CHECK_ERROR("kernel_dspmv_csr4() execution failed\n");
	}
	else {
		kernel_dspmv_csr4_b0 <<<CUKR_DSPMV_CSR4_CTAS, CUKR_DSPMV_CSR4_THREADS>>> (rows, d_val, d_x, alpha, d_y);
	    CUT_CHECK_ERROR("kernel_dspmv_csr4_b0() execution failed\n");
	}
	 
   	/* Un-bind textures */
	cudaUnbindTexture(TEX_X2_I);
   	cudaUnbindTexture(TEX_PTR);
   	cudaUnbindTexture(TEX_IDX4);
   	cudaUnbindTexture(TEX_VAL4_I);
   	
   	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize());
 	
}		 

// ------------------------------------------------------------------- // 
/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR format, in double precision
 * \param rows, cols Matrix size
 * \param nz Number of nonzeros
 * \param r, c Block size
 * \param alpha Scaling factor for Ax
 * \param d_ptr Row index vector (device ptr)
 * \param d_idx Column index vector (device ptr)
 * \param d_val Value vector (device ptr)
 * \param d_x Vector being multiplied (device ptr)
 * \param beta Scaling factor for y
 * \param d_y Result vector (device ptr)
 */
#ifdef __cplusplus
extern "C"
#endif 
void
cuda_dspmv_bcsr(int rows, int cols, int nz, int r, int c, double alpha, int* d_ptr, int* d_idx, 
			   double* d_val, double* d_x, double beta, double* d_y) 
{
	/* num of blocks per row or per col */
	int nb = (rows%r ? rows/r+1 : rows/r);
			
	/* Bind textures */		
  	cudaBindTexture(0, TEX_PTR,    d_ptr, (nb+1) * sizeof(d_ptr[0]));
   	cudaBindTexture(0, TEX_IDX,    d_idx, nz     * sizeof(d_idx[0]));
      	cudaBindTexture(0, TEX_VAL2_I, d_val, nz*r*c * sizeof(d_val[0]));
   	cudaBindTexture(0, TEX_X2_I,   d_x  , cols   * sizeof(d_x[0]));
   	
	/*
	 * Chose which kernel to run
	 */
	
	if (beta != 1 && beta)
	 	cublasDscal(rows, beta, d_y, 1);
	
	if (beta) {
	 	if     (r == 2) kernel_dspmv_bcsr_2x2 <<<CUKR_DSPMV_BCSR_2X2_CTAS, CUKR_DSPMV_BCSR_2X2_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 3) kernel_dspmv_bcsr_3x3 <<<CUKR_DSPMV_BCSR_3X3_CTAS, CUKR_DSPMV_BCSR_3X3_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 4) kernel_dspmv_bcsr_4x4 <<<CUKR_DSPMV_BCSR_4X4_CTAS, CUKR_DSPMV_BCSR_4X4_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 5) kernel_dspmv_bcsr_5x5 <<<CUKR_DSPMV_BCSR_5X5_CTAS, CUKR_DSPMV_BCSR_5X5_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 6) kernel_dspmv_bcsr_6x6 <<<CUKR_DSPMV_BCSR_6X6_CTAS, CUKR_DSPMV_BCSR_6X6_THREADS>>> (nb, alpha, d_y);
	 	else printf("Block size (%d, %d) is NOT supported at cuda_dspmv_bcsr", r, c);
    		CUT_CHECK_ERROR("kernel_dspmv_bcsr() execution failed\n");
	}   
	else {
	 	if     (r == 2) kernel_dspmv_bcsr_2x2_b0 <<<CUKR_DSPMV_BCSR_2X2_CTAS, CUKR_DSPMV_BCSR_2X2_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 3) kernel_dspmv_bcsr_3x3_b0 <<<CUKR_DSPMV_BCSR_3X3_CTAS, CUKR_DSPMV_BCSR_3X3_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 4) kernel_dspmv_bcsr_4x4_b0 <<<CUKR_DSPMV_BCSR_4X4_CTAS, CUKR_DSPMV_BCSR_4X4_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 5) kernel_dspmv_bcsr_5x5_b0 <<<CUKR_DSPMV_BCSR_5X5_CTAS, CUKR_DSPMV_BCSR_5X5_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 6) kernel_dspmv_bcsr_6x6_b0 <<<CUKR_DSPMV_BCSR_6X6_CTAS, CUKR_DSPMV_BCSR_6X6_THREADS>>> (nb, alpha, d_y);
	 	else printf("Block size (%d, %d) is NOT supported at cuda_dspmv_bcsr", r, c);
    		CUT_CHECK_ERROR("kernel_dspmv_bcsr_b0() execution failed\n");
	} 		
		
	 
   	/* Un-bind textures */
   	cudaUnbindTexture(TEX_PTR);
   	cudaUnbindTexture(TEX_IDX);
   	cudaUnbindTexture(TEX_VAL2_I);
	cudaUnbindTexture(TEX_X2_I);
   	
   	/* Do a syncronize, for timing */
    	CUDA_SAFE_CALL(cudaThreadSynchronize());
 	
}

// ------------------------------------------------------------------- // 
/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR4 format, in double precision
 * \param rows, cols Matrix size
 * \param nz Number of nonzeros
 * \param r, c Block size
 * \param alpha Scaling factor for Ax
 * \param d_ptr Row index vector (device ptr)
 * \param d_idx Column index vector (device ptr)
 * \param d_val Value vector (device ptr)
 * \param d_x Vector being multiplied (device ptr)
 * \param beta Scaling factor for y
 * \param d_y Result vector (device ptr)
 */
#ifdef __cplusplus
extern "C"
#endif 
void
cuda_dspmv_bcsr4(int rows, int cols, int nz, int r, int c, double alpha, int* d_ptr, int* d_idx, 
			   double* d_val, double* d_x, double beta, double* d_y) 
{

	/* num of blocks per row or per col */
	int nb = (rows%r ? rows/r+1 : rows/r);
			
	/* Bind textures */		
  	cudaBindTexture(0, TEX_PTR,    d_ptr, (nb+1) * sizeof(d_ptr[0]));
   	cudaBindTexture(0, TEX_IDX,    d_idx, nz     * sizeof(d_idx[0]));
   	if(r%2 == 0) cudaBindTexture(0, TEX_VAL4_I, d_val, nz*r*c     * sizeof(d_val[0]));
   	else         cudaBindTexture(0, TEX_VAL4_I, d_val, nz*(r*c+1) * sizeof(d_val[0]));
   	cudaBindTexture(0, TEX_X2_I,   d_x  , cols   * sizeof(d_x[0]));
   	
	/*   	if(r%2 == 0) cudaBindTexture(0, TEX_VAL4_I, d_val, nz*r*c     * sizeof(d_val[0]));
   	else         cudaBindTexture(0, TEX_VAL4_I, d_val, nz*(r*c+1) * sizeof(d_val[0]));
	 * Chose which kernel to run
	 */
	
	if (beta != 1 && beta)
	 	cublasDscal(rows, beta, d_y, 1);
	
	if (beta) {
	 	if     (r == 2) kernel_dspmv_bcsr4_2x2 <<<CUKR_DSPMV_BCSR4_2X2_CTAS, CUKR_DSPMV_BCSR4_2X2_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 3) kernel_dspmv_bcsr4_3x3 <<<CUKR_DSPMV_BCSR4_3X3_CTAS, CUKR_DSPMV_BCSR4_3X3_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 4) kernel_dspmv_bcsr4_4x4 <<<CUKR_DSPMV_BCSR4_4X4_CTAS, CUKR_DSPMV_BCSR4_4X4_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 5) kernel_dspmv_bcsr4_5x5 <<<CUKR_DSPMV_BCSR4_5X5_CTAS, CUKR_DSPMV_BCSR4_5X5_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 6) kernel_dspmv_bcsr4_6x6 <<<CUKR_DSPMV_BCSR4_6X6_CTAS, CUKR_DSPMV_BCSR4_6X6_THREADS>>> (nb, alpha, d_y);
	 	else printf("Block size (%d, %d) is NOT supported at cuda_dspmv_bcsr4", r, c);
    		CUT_CHECK_ERROR("kernel_dspmv_bcsr4() execution failed\n");
	}
	else {
	 	if     (r == 2) kernel_dspmv_bcsr4_2x2_b0 <<<CUKR_DSPMV_BCSR4_2X2_CTAS, CUKR_DSPMV_BCSR4_2X2_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 3) kernel_dspmv_bcsr4_3x3_b0 <<<CUKR_DSPMV_BCSR4_3X3_CTAS, CUKR_DSPMV_BCSR4_3X3_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 4) kernel_dspmv_bcsr4_4x4_b0 <<<CUKR_DSPMV_BCSR4_4X4_CTAS, CUKR_DSPMV_BCSR4_4X4_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 5) kernel_dspmv_bcsr4_5x5_b0 <<<CUKR_DSPMV_BCSR4_5X5_CTAS, CUKR_DSPMV_BCSR4_5X5_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 6) kernel_dspmv_bcsr4_6x6_b0 <<<CUKR_DSPMV_BCSR4_6X6_CTAS, CUKR_DSPMV_BCSR4_6X6_THREADS>>> (nb, alpha, d_y);
	 	else printf("Block size (%d, %d) is NOT supported at cuda_dspmv_bcsr4", r, c);
    		CUT_CHECK_ERROR("kernel_dspmv_bcsr4_b0() execution failed\n");
	}
		
	 
   	/* Un-bind textures */
   	cudaUnbindTexture(TEX_PTR);
   	cudaUnbindTexture(TEX_IDX);
   	cudaUnbindTexture(TEX_VAL4_I);
	cudaUnbindTexture(TEX_X2_I);
   	
   	/* Do a syncronize, for timing */
    	CUDA_SAFE_CALL(cudaThreadSynchronize());
 	
}

// ------------------------------------------------------------------- // 

/**
 * \brief Wrapper for DSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions 
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector 
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */ 
#ifdef __cplusplus
extern "C"
#endif    
void cuda_dspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz, 
                    CUKR_INT* d_ell_idx, CUKR_DOUBLE* d_ell_val, CUKR_INT* d_csr_ptr, CUKR_INT* d_csr_idx, CUKR_DOUBLE* d_csr_val,  
                    CUKR_DOUBLE* d_x, CUKR_DOUBLE beta, CUKR_DOUBLE* d_y)    
{		
   	
   	//struct timeval t_start, t_end;
	//struct timezone t_zone;
 	//CLOCK_START;
   	
   	/* Bind texture */
   	cudaBindTexture(0, TEX_X2_I, d_x, cols * sizeof(d_x[0]));		

	/* If beta != 1, do a scaling first */
	//if (beta != 1)
	//	cublasSscal(rows, beta, d_y, 1);	
	
	const unsigned int BLOCK_SIZE = 256;
	const unsigned int num_blocks = DIVIDE_INTO(rows, BLOCK_SIZE);
    const dim3 grid = dim3(num_blocks);
	
	/* Do the SPMV in ELL */
	kernel_dspmv_ell <<<grid, BLOCK_SIZE, 0>>>(rows, alpha, ell_nz_row, ell_stride, d_ell_idx, d_ell_val, beta, d_y);
   	CUT_CHECK_ERROR("kernel_dspmv_ell() execution failed\n");
   	
   	/* Do the rest in CSR */
   	if (csr_nz) {
		cudaBindTexture(0, TEX_PTR   , d_csr_ptr, (rows+1) * sizeof(d_csr_ptr[0]));
	   	cudaBindTexture(0, TEX_IDX4  , d_csr_idx, csr_nz   * sizeof(d_csr_idx[0]));
	   	cudaBindTexture(0, TEX_VAL4_I, d_csr_val, csr_nz   * sizeof(d_csr_val[0]));
	                
		kernel_dspmv_csr4 <<<CUKR_DSPMV_CSR4_CTAS, CUKR_DSPMV_CSR4_THREADS>>> (rows, d_csr_val, d_x, alpha, d_y);
		CUT_CHECK_ERROR("kernel_dspmv_csr4() execution failed\n");
		 	  
	    /* Unbind textures */ 	    
	   	cudaUnbindTexture(TEX_PTR);
	   	cudaUnbindTexture(TEX_IDX4);
	   	cudaUnbindTexture(TEX_VAL4_I);
   	}
   	
   	cudaUnbindTexture(TEX_X2_I);
   	CUDA_SAFE_CALL(cudaThreadSynchronize());
   	
   	//CLOCK_END;
   	//double elapsed = CLOCK_SPAN;
   	//printf("[%d] -> %g ms\n", cukr_activeThread, elapsed);   	   		 
}

// ------------------------------------------------------------------- // 

/**
 * \brief Wrapper for DSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions 
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector 
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */ 
#ifdef __cplusplus
extern "C"
#endif    
void cuda_dspmv_bhyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz, 
                    CUKR_INT* d_bell_idx, CUKR_DOUBLE* d_bell_val, CUKR_INT* d_bcsr_ptr, CUKR_INT* d_bcsr_idx, CUKR_DOUBLE* d_bcsr_val,  
                    CUKR_DOUBLE* d_x, CUKR_DOUBLE beta, CUKR_DOUBLE* d_y)   
{		
   	/* num of blocks per row or per col */
	int nb = (rows%r ? rows/r+1 : rows/r);

	if (beta != 1 && beta)
	 	cublasDscal(rows, beta, d_y, 1);

   	/* Bind textures */
   	cudaBindTexture(0, TEX_X2_I, d_x, cols * sizeof(d_x[0]));
   	cudaBindTexture(0, TEX_IDX,    d_bell_idx, bell_nz_row*bell_stride     * sizeof(d_bell_idx[0]));
      	cudaBindTexture(0, TEX_VAL2_I, d_bell_val, bell_nz_row*bell_stride*r*c * sizeof(d_bell_val[0]));

	/* Do the SPMV in BELL */
 	if     (r == 2) kernel_dspmv_bell_2x2 <<<CUKR_DSPMV_BELL_2X2_CTAS, CUKR_DSPMV_BELL_2X2_THREADS>>> (nb, alpha, bell_nz_row, bell_stride, d_y);
 	else if(r == 3) kernel_dspmv_bell_3x3 <<<CUKR_DSPMV_BELL_3X3_CTAS, CUKR_DSPMV_BELL_3X3_THREADS>>> (nb, alpha, bell_nz_row, bell_stride, d_y);
 	else if(r == 4) kernel_dspmv_bell_4x4 <<<CUKR_DSPMV_BELL_4X4_CTAS, CUKR_DSPMV_BELL_4X4_THREADS>>> (nb, alpha, bell_nz_row, bell_stride, d_y);
 	else if(r == 5) kernel_dspmv_bell_5x5 <<<CUKR_DSPMV_BELL_5X5_CTAS, CUKR_DSPMV_BELL_5X5_THREADS>>> (nb, alpha, bell_nz_row, bell_stride, d_y);
 	else if(r == 6) kernel_dspmv_bell_6x6 <<<CUKR_DSPMV_BELL_6X6_CTAS, CUKR_DSPMV_BELL_6X6_THREADS>>> (nb, alpha, bell_nz_row, bell_stride, d_y);
	else
		printf("Block size (%d, %d) is NOT supported at cuda_dspmv_bhyb", r, c);
   	CUT_CHECK_ERROR("kernel_dspmv_bell() execution failed\n");
   
   	/* Unbind textures */ 
   	cudaUnbindTexture(TEX_IDX);
   	cudaUnbindTexture(TEX_VAL2_I);
   
   	/* Do the rest in BCSR */
   	if (bcsr_nz) {
	   	
		/* Bind textures */
	  	cudaBindTexture(0, TEX_PTR,    d_bcsr_ptr, (nb+1)      * sizeof(d_bcsr_ptr[0]));
	   	cudaBindTexture(0, TEX_IDX,    d_bcsr_idx, bcsr_nz     * sizeof(d_bcsr_idx[0]));
	   	cudaBindTexture(0, TEX_VAL2_I, d_bcsr_val, bcsr_nz*r*c * sizeof(d_bcsr_val[0]));
		
	 	if     (r == 2) kernel_dspmv_bcsr_2x2 <<<CUKR_DSPMV_BCSR_2X2_CTAS, CUKR_DSPMV_BCSR_2X2_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 3) kernel_dspmv_bcsr_3x3 <<<CUKR_DSPMV_BCSR_3X3_CTAS, CUKR_DSPMV_BCSR_3X3_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 4) kernel_dspmv_bcsr_4x4 <<<CUKR_DSPMV_BCSR_4X4_CTAS, CUKR_DSPMV_BCSR_4X4_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 5) kernel_dspmv_bcsr_5x5 <<<CUKR_DSPMV_BCSR_5X5_CTAS, CUKR_DSPMV_BCSR_5X5_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 6) kernel_dspmv_bcsr_6x6 <<<CUKR_DSPMV_BCSR_6X6_CTAS, CUKR_DSPMV_BCSR_6X6_THREADS>>> (nb, alpha, d_y);
	        CUT_CHECK_ERROR("kernel_dspmv_bcsr_a1() execution failed\n");

		 	  
	   	/* Unbind textures */ 
	   	cudaUnbindTexture(TEX_PTR);
	   	cudaUnbindTexture(TEX_IDX);
	   	cudaUnbindTexture(TEX_VAL2_I);
   	}
   	
   	/* Unbind texture */ 
   	cudaUnbindTexture(TEX_X2_I);
   	
   	CUDA_SAFE_CALL(cudaThreadSynchronize());  	   		 
}

// ------------------------------------------------------------------- // 

/**
 * \brief Wrapper for DSPMV_BHYB4
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions 
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector 
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */ 
#ifdef __cplusplus
extern "C"
#endif    
void cuda_dspmv_bhyb4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz, 
                    CUKR_INT* d_bell_idx, CUKR_DOUBLE* d_bell_val, CUKR_INT* d_bcsr_ptr, CUKR_INT* d_bcsr_idx, CUKR_DOUBLE* d_bcsr_val,  
                    CUKR_DOUBLE* d_x, CUKR_DOUBLE beta, CUKR_DOUBLE* d_y)   
{		
   	/* num of blocks per row or per col */
	int nb = (rows%r ? rows/r+1 : rows/r);

	if (beta != 1 && beta)
	 	cublasDscal(rows, beta, d_y, 1);

   	/* Bind textures */
   	cudaBindTexture(0, TEX_X2_I, d_x, cols * sizeof(d_x[0]));
   	cudaBindTexture(0, TEX_IDX,  d_bell_idx, bell_nz_row*bell_stride * sizeof(d_bcsr_idx[0]));
   	if((r*c)%2 == 1) cudaBindTexture(0, TEX_VAL4_I, d_bell_val, bell_nz_row*bell_stride*(r*c+1) * sizeof(d_bcsr_val[0]));
   	else             cudaBindTexture(0, TEX_VAL4_I, d_bell_val, bell_nz_row*bell_stride* r*c    * sizeof(d_bcsr_val[0]));	

	/* Do the SPMV in BELL4 */
 	if     (r == 2) kernel_dspmv_bell4_2x2 <<<CUKR_DSPMV_BELL4_2X2_CTAS, CUKR_DSPMV_BELL4_2X2_THREADS>>> (nb, alpha, bell_nz_row, bell_stride, d_y);
 	else if(r == 3) kernel_dspmv_bell4_3x3 <<<CUKR_DSPMV_BELL4_3X3_CTAS, CUKR_DSPMV_BELL4_3X3_THREADS>>> (nb, alpha, bell_nz_row, bell_stride, d_y);
 	else if(r == 4) kernel_dspmv_bell4_4x4 <<<CUKR_DSPMV_BELL4_4X4_CTAS, CUKR_DSPMV_BELL4_4X4_THREADS>>> (nb, alpha, bell_nz_row, bell_stride, d_y);
 	else if(r == 5) kernel_dspmv_bell4_5x5 <<<CUKR_DSPMV_BELL4_5X5_CTAS, CUKR_DSPMV_BELL4_5X5_THREADS>>> (nb, alpha, bell_nz_row, bell_stride, d_y);
 	else if(r == 6) kernel_dspmv_bell4_6x6 <<<CUKR_DSPMV_BELL4_6X6_CTAS, CUKR_DSPMV_BELL4_6X6_THREADS>>> (nb, alpha, bell_nz_row, bell_stride, d_y);
	else
		printf("Block size (%d, %d) is NOT supported at cuda_dspmv_bhyb4", r, c);
   	CUT_CHECK_ERROR("kernel_dspmv_bell4() execution failed\n");
   	
   	/* Unbind textures */ 
   	cudaUnbindTexture(TEX_IDX);
   	cudaUnbindTexture(TEX_VAL4_I);
   	
   	/* Do the rest in BCSR4 */
   	if (bcsr_nz) {
	   	
		/* Bind textures */
	  	cudaBindTexture(0, TEX_PTR,    d_bcsr_ptr, (nb+1)  * sizeof(d_bcsr_ptr[0]));
	   	cudaBindTexture(0, TEX_IDX,    d_bcsr_idx, bcsr_nz * sizeof(d_bcsr_idx[0]));
	   	if((r*c)%2 == 1) cudaBindTexture(0, TEX_VAL4_I, d_bcsr_val, bcsr_nz*(r*c+1) * sizeof(d_bcsr_val[0]));
	   	else             cudaBindTexture(0, TEX_VAL4_I, d_bcsr_val, bcsr_nz* r*c    * sizeof(d_bcsr_val[0]));
		
	 	if     (r == 2) kernel_dspmv_bcsr4_2x2 <<<CUKR_DSPMV_BCSR4_2X2_CTAS, CUKR_DSPMV_BCSR4_2X2_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 3) kernel_dspmv_bcsr4_3x3 <<<CUKR_DSPMV_BCSR4_3X3_CTAS, CUKR_DSPMV_BCSR4_3X3_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 4) kernel_dspmv_bcsr4_4x4 <<<CUKR_DSPMV_BCSR4_4X4_CTAS, CUKR_DSPMV_BCSR4_4X4_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 5) kernel_dspmv_bcsr4_5x5 <<<CUKR_DSPMV_BCSR4_5X5_CTAS, CUKR_DSPMV_BCSR4_5X5_THREADS>>> (nb, alpha, d_y);
	 	else if(r == 6) kernel_dspmv_bcsr4_6x6 <<<CUKR_DSPMV_BCSR4_6X6_CTAS, CUKR_DSPMV_BCSR4_6X6_THREADS>>> (nb, alpha, d_y);
	        CUT_CHECK_ERROR("kernel_dspmv_bcsr4_a1() execution failed\n");

		 	  
	   	/* Unbind textures */ 
	   	cudaUnbindTexture(TEX_PTR);
	   	cudaUnbindTexture(TEX_IDX);
	   	cudaUnbindTexture(TEX_VAL4_I);
   	}
   	
   	/* Unbind texture */
   	cudaUnbindTexture(TEX_X2_I);
   	
   	CUDA_SAFE_CALL(cudaThreadSynchronize());  	   		 
}


