/**
 * @file blas/gpu/cuda/kernels_qdouble.cu
 * @author Serban Georgescu
 * @since 2008
 * 
 * Acts as a bridge between the C++ CUDA
 * implementation and the rest of the C 
 * files. All the CUDA single-double
 * (qdouble, emulated-double) precision
 * kernels must be included here.
 * 
 */
 
#include <stdio.h>
#include <unistd.h>
#include <cuda.h>
#include <cutil.h>
#include <cublas.h>


#include "kernels_textures.h"

/* Get access to the array of texture
 * bindings */
//extern CUKR_TEXTURES cukr_textures[CUKR_MAX_PE];

#include "../../../../cukr.h" 
#include "kernels_config.h"
#include "qdouble/kernel_qdot.h"
#include "qdouble/kernel_qaxpy.h"
#include "qdouble/kernel_qaypx.h"
#include "qdouble/kernel_qscal.h"
#include "qdouble/kernel_qspmv_csr.h"
#include "qdouble/kernel_qspmv_csr_a1.h"
#include "qdouble/kernel_qspmv_csr_b0.h"
#include "qdouble/kernel_qspmv_csr_a1_b0.h"
#include "qdouble/kernel_qspmv_csr4.h"
#include "qdouble/kernel_qspmv_csr4_a1.h"
#include "qdouble/kernel_qspmv_csr4_b0.h"
#include "qdouble/kernel_qspmv_csr4_a1_b0.h"

// BCSR
#include "qdouble/kernel_qspmv_bcsr_a1_b0.h"
#include "qdouble/kernel_qspmv_bcsr_b0.h"
#include "qdouble/kernel_qspmv_bcsr_a1.h"
#include "qdouble/kernel_qspmv_bcsr.h"

// BCSR4
#include "qdouble/kernel_qspmv_bcsr4_a1_b0.h"
#include "qdouble/kernel_qspmv_bcsr4_b0.h"
#include "qdouble/kernel_qspmv_bcsr4_a1.h"
#include "qdouble/kernel_qspmv_bcsr4.h"

//ELL
#include "qdouble/kernel_qspmv_ell.h"

//BELL
#include "qdouble/kernel_qspmv_bell.h"

//BELL4
#include "qdouble/kernel_qspmv_bell4.h"


#define DIVIDE_INTO(x,y) ((x + y - 1)/y) 

//========================== FORWARD DECLARATIONS ========================= //

void HandleCudaMallocErrors(cublasStatus status);


//============================ PUBLIC FUNCTIONS =========================== //

// ------------------------------------------------------------------- //

/**
 * \brief QDOT kernel in double-single,
 *        on the GPU
 * \param n Vector size
 * \param d_xh, d_xt Device pointers to the 
 *        first vector (head and tail).
 * \param d_yh, d_yt Device pointers to the 
 *        second (result) vector (head and tail).
 */   
#ifdef __cplusplus
extern "C"
#endif 
double cuda_kernel_qdot(int n, float *d_xh, float *d_xt, float *d_yh, float *d_yt)
{
	cublasStatus status; 
	int i;
	
	/* Alloc device memory for 
	 * CTA level results */
	float *d_sh, *d_st;
	status = cublasAlloc (CUKR_QDOT_CTAS, sizeof(d_sh[0]), (void**)&d_sh);
	HandleCudaMallocErrors(status); 
	status = cublasAlloc (CUKR_QDOT_CTAS, sizeof(d_st[0]), (void**)&d_st);
	HandleCudaMallocErrors(status); 
	
	/* Alloc host memory to collect CTA 
	 * level results */
	float *h_sh, *h_st;
	if ( (h_sh = (float*)malloc(CUKR_QDOT_CTAS*sizeof(h_sh[0]))) == NULL) {
		printf("FATAL: Error allocating memory for QDOT result vectors");
		exit(1);
	}
	if ( (h_st = (float*)malloc(CUKR_QDOT_CTAS*sizeof(h_st[0]))) == NULL) {
		printf("FATAL: Error allocating memory for QDOT result vectors");
		exit(1);
	}
	
	/* Preparations */
	int dynMem = 2 * CUKR_QDOT_THREADS * sizeof(float);
	
	/* Kernel invocation */
	kernel_qdot<<<CUKR_QDOT_CTAS, CUKR_QDOT_THREADS, dynMem>>> 
	           (d_xh, d_xt, d_yh, d_yt, d_sh, d_st, n);
	CUT_CHECK_ERROR("kernel_qdot() execution failed\n");           
	   
   	
   	/* Get CTA-level sums */
   	status = cublasGetVector (CUKR_QDOT_CTAS, sizeof(d_sh[0]), d_sh, 1, h_sh, 1);
   	HandleCudaMallocErrors(status); 
   	status = cublasGetVector (CUKR_QDOT_CTAS, sizeof(d_st[0]), d_st, 1, h_st, 1);
   	HandleCudaMallocErrors(status); 
   	
   	/* Do the final sum */
   	double sum0 = 0, sum1 = 0;
   	for (i = 0; i < CUKR_QDOT_CTAS; i++) {
   		sum0 += h_sh[i];
   		sum1 += h_st[i];
   	}
   	
   	/* Free the alocated memory */
   	cublasFree(d_sh);
   	cublasFree(d_st);
   	free(h_sh);
   	free(h_st);
   	
   	return sum0 + sum1;
}

// ------------------------------------------------------------------- //

/**
 * \brief QAXPY kernel in double-single,
 *        on the GPU
 * \param n Vector size
 * \param a0,a1 The constant (head and tail)
 * \param d_xh, d_xt Device pointers to the 
 *        first vector (head and tail).
 * \param d_yh, d_yt Device pointers to the 
 *        second (result) vector (head and tail).
 */ 
#ifdef __cplusplus
extern "C"
#endif 
void cuda_kernel_qaxpy(int n, float a0, float a1, float *d_xh, float *d_xt, 
					     			              float *d_yh, float *d_yt)
{
	/* Preparations */
	int dynMem = 4 * CUKR_QAXPY_THREADS * sizeof(float); 
		
	/* Kernel invocation */
	kernel_qaxpy<<<CUKR_QAXPY_CTAS, CUKR_QAXPY_THREADS, dynMem>>> 
	               (d_xh, d_xt, d_yh, d_yt, a0, a1, n);   	
	CUT_CHECK_ERROR("kernel_qaxpy() execution failed\n");   
	
	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize());
}
					   		 		                	

// ------------------------------------------------------------------- //

/**
 * \brief QAYPX kernel in double-single,
 *        on the GPU
 * \param n Vector size
 * \param a0,a1 The constant (head and tail)
 * \param d_xh, d_xt Device pointers to the 
 *        first vector (head and tail).
 * \param d_yh, d_yt Device pointers to the 
 *        second (result) vector (head and tail).
 */ 
#ifdef __cplusplus
extern "C"
#endif 
void cuda_kernel_qaypx(int n, float a0, float a1, float *d_xh, float *d_xt, 
					   				              float *d_yh, float *d_yt)
{	
	
	/* Preparations */
	int dynMem = 4 * CUKR_QAYPX_THREADS * sizeof(float); 
	
	/* Kernel invocation */
	kernel_qaypx<<<CUKR_QAYPX_CTAS, CUKR_QAYPX_THREADS, dynMem>>> 
	               (d_xh, d_xt, d_yh, d_yt, a0, a1, n); 
	CUT_CHECK_ERROR("kernel_qaypx() execution failed\n");   	               
	               
	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize());
} 

// ------------------------------------------------------------------- //

/**
 * \brief QSCAL kernel in double-single,
 *        on the GPU
 * \param n Vector size
 * \param a0,a1 The constant (head and tail)
 * \param d_xh, d_xt Device pointers to the 
 *        vector (head and tail).
 */ 
#ifdef __cplusplus
extern "C"
#endif 
void cuda_kernel_qscal(int n, float a0, float a1, float *d_xh, float *d_xt)
{	
	
	/* Preparations */
	int dynMem = 2 * CUKR_QSCAL_THREADS * sizeof(float); 

	/* Kernel invocation */
	kernel_qscal<<<CUKR_QSCAL_CTAS, CUKR_QSCAL_THREADS, dynMem>>> (d_xh, d_xt, a0, a1, n);
	CUT_CHECK_ERROR("kernel_qscal() execution failed\n");  	               
	               
	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize());
} 				   	      
					   	      
// ------------------------------------------------------------------- // 					   	      

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in quasi-double precision
 * \param n Matrix size
 * \param nz Number of nonzeros
 * \param a0,a1 Scaling factor for Ax (head,tail)
 * \param ptr Row index vector 
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param b0, b1 Scale factor for y (head, tail)
 * \param yh,yt Head and tail for the result
 *              vector

 */

/**
 * TODO Fix bug in cuda_qspmv_csr. Results are not correct on my macbooc pro.
 *      They seem however to be correct on nana ...
 */

#ifdef __cplusplus
extern "C"
#endif 
void
cuda_qspmv_csr(int rows, int cols, int nz, float a0, float a1, int* d_ptr, int* d_idx, 
			   float* d_valh, float* d_valt, float* d_xh, float* d_xt, 
			   float b0, float b1, float* d_yh, float* d_yt)
{	 
	/* Bind to texture */
	cudaBindTexture(0, TEX_PTR   , d_ptr, (rows + 1) * sizeof(d_ptr[0]));
   	cudaBindTexture(0, TEX_IDX   , d_idx,  nz        * sizeof(d_idx[0]));
   	cudaBindTexture(0, TEX_VAL   , d_valh, nz        * sizeof(d_valh[0]));
   	cudaBindTexture(0, TEX_VAL_T , d_valt, nz        * sizeof(d_valt[0]));
   	cudaBindTexture(0, TEX_X     , d_xh  , cols      * sizeof(d_xh[0]));	
   	cudaBindTexture(0, TEX_X_T   , d_xt  , cols      * sizeof(d_xt[0]));
	 
	/*
	 * Chose which kernel to run 
	 */
	
	double a = a0 + a1;
	double b = b0 + b1;
	
//	kernel_qspmv_csr <<<CUKR_QSPMV_CSR_CTAS, CUKR_QSPMV_CSR_THREADS>>>
//	   	             	            (rows, a0, a1, d_yh, d_yt);
//	CUT_CHECK_ERROR("kernel_qspmv_csr() execution failed\n");

	 /* If beta != 0 */
	if (b != 0)
	{
	 	/* If beta != 1, do a scaling first */
	 	if (b != 1) 
			cuda_kernel_qscal(rows, b0, b1, d_yh, d_yt);
	 	
	 	/* If alpha = 1, no need to consider */                         
   		if (a == 1) {
   	   		kernel_qspmv_csr_a1 <<<CUKR_QSPMV_CSR_A1_CTAS, CUKR_QSPMV_CSR_A1_THREADS>>> 
   	             	               (rows, d_yh, d_yt);
   	        CUT_CHECK_ERROR("kernel_qspmv_csr_a1() execution failed\n");
   		} 
	 	/* Else, take the most general case */
	 	else {
	 		kernel_qspmv_csr <<<CUKR_QSPMV_CSR_CTAS, CUKR_QSPMV_CSR_THREADS>>> 
   	             	            (rows, a0, a1, d_yh, d_yt);
            CUT_CHECK_ERROR("kernel_qspmv_csr() execution failed\n");    	             	           
	 	}
	 	  
	}
	/* If beta = 0 */
	else {
	 	/* If alpha = 1 as well, no need to
	 	 * consider them both */
	 	 if (a == 1) {
	 	 	kernel_qspmv_csr_a1_b0 <<<CUKR_QSPMV_CSR_A1_CTAS, CUKR_QSPMV_CSR_A1_THREADS>>>
   	             	                  (rows, d_yh, d_yt);
   	        CUT_CHECK_ERROR("kernel_qspmv_csr_a1_b0() execution failed\n");
	 	 }
   	    /* If alpha != 1, have to consider it */
   	    else {
   	    	kernel_qspmv_csr_b0 <<<CUKR_QSPMV_CSR_CTAS, CUKR_QSPMV_CSR_THREADS>>> 
   	              	               (rows, a0, a1, d_yh, d_yt);
   	        CUT_CHECK_ERROR("kernel_qspmv_csr_b0() execution failed\n");
   	    }

	}   	


  	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize());
   
   	/* Un-bind textures */
   	cudaUnbindTexture(TEX_PTR);
   	cudaUnbindTexture(TEX_IDX);
   	cudaUnbindTexture(TEX_VAL);
   	cudaUnbindTexture(TEX_VAL_T);
   	cudaUnbindTexture(TEX_X);	
   	cudaUnbindTexture(TEX_X_T);
}		 

// ------------------------------------------------------------------- // 

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR4 format, in quasi-double precision
 * \param n Matrix size
 * \param nz Number of nonzeros
 * \param a0,a1 Scaling factor for Ax (head,tail)
 * \param ptr Row index vector 
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param b0, b1 Scale factor for y (head, tail)
 * \param yh,yt Head and tail for the result
 *              vector

 */
#ifdef __cplusplus
extern "C"
#endif 
void
cuda_qspmv_csr4(int rows, int cols, int nz, float a0, float a1, int* d_ptr, int* d_idx, 
			    float* d_valh, float* d_valt, float* d_xh, float* d_xt, 
			    float b0, float b1, float* d_yh, float* d_yt)
{
	
	/* Bind to texture */
   	cudaBindTexture(0, TEX_PTR   , d_ptr, (rows + 1) * sizeof(d_ptr[0]));
   	cudaBindTexture(0, TEX_IDX4  , d_idx,  nz        * sizeof(d_idx[0]));
   	cudaBindTexture(0, TEX_VAL4  , d_valh, nz        * sizeof(d_valh[0]));
   	cudaBindTexture(0, TEX_VAL4_T, d_valt, nz        * sizeof(d_valt[0]));
   	cudaBindTexture(0, TEX_X     , d_xh  , cols      * sizeof(d_xh[0]));	
   	cudaBindTexture(0, TEX_X_T   , d_xt  , cols      * sizeof(d_xt[0]));
	
   	
   	/*
	 * Chose which kernel to run 
	 */
	
	double a = a0 + a1;
	double b = b0 + b1;
	
	/* If beta != 0 */
	if (b != 0)
	{
	 	/* If beta != 1, do a scaling first */
	 	if (b != 1) 
			cuda_kernel_qscal(rows, b0, b1, d_yh, d_yt);
	 	
	 	/* If alpha = 1, no need to consider */                         
   		if (a == 1) {
   	   		kernel_qspmv_csr4_a1 <<<CUKR_QSPMV_CSR4_CTAS, CUKR_QSPMV_CSR4_THREADS>>> 
   	        	                    (rows, d_yh, d_yt);
   	        CUT_CHECK_ERROR("kernel_qspmv_csr4_a1() execution failed\n");
   		} 
	 	/* Else, take the most general case */
	 	else {
	 		kernel_qspmv_csr4 <<<CUKR_QSPMV_CSR4_CTAS, CUKR_QSPMV_CSR4_THREADS>>> 
   	             	             (rows, a0, a1, d_yh, d_yt);
            CUT_CHECK_ERROR("kernel_qspmv_csr4() execution failed\n");    	             	           
	 	}
	 	  
	}
	/* If beta = 0 */
	else {
	 	/* If alpha = 1 as well, no need to
	 	 * consider them both */
	 	 if (a == 1) {
	 	 	kernel_qspmv_csr4_a1_b0 <<<CUKR_QSPMV_CSR4_CTAS, CUKR_QSPMV_CSR4_THREADS>>> (rows, d_yh, d_yt);
   	        CUT_CHECK_ERROR("kernel_qspmv_csr4_a1_b0() execution failed\n");
	 	 }
   	    /* If alpha != 1, have to consider it */
   	    else {
   	    	kernel_qspmv_csr4_b0 <<<CUKR_QSPMV_CSR4_CTAS, CUKR_QSPMV_CSR4_THREADS>>> 
   	              	                (rows, a0, a1, d_yh, d_yt);
   	        CUT_CHECK_ERROR("kernel_qspmv_csr4_b0() execution failed\n");
   	    }
   	    	
	}   	
   	
   	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize());
   
   	/* Un-bind textures */
   	cudaUnbindTexture(TEX_PTR);
   	cudaUnbindTexture(TEX_IDX4);
   	cudaUnbindTexture(TEX_VAL4);
   	cudaUnbindTexture(TEX_VAL4_T);
   	cudaUnbindTexture(TEX_X);	
   	cudaUnbindTexture(TEX_X_T);
}		

// ------------------------------------------------------------------- // 

/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR format, in quasi-double precision
 * \param n Matrix size
 * \param nz Number of nonzeros
 * \param a0,a1 Scaling factor for Ax (head,tail)
 * \param ptr Row index vector 
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param b0, b1 Scale factor for y (head, tail)
 * \param yh,yt Head and tail for the result
 *              vector

 */
#ifdef __cplusplus
extern "C"
#endif 
void
cuda_qspmv_bcsr(int rows, int cols, int nz, int r, int c, float a0, float a1, int* d_ptr, int* d_idx, 
			   float* d_valh, float* d_valt, float* d_xh, float* d_xt, 
			   float b0, float b1, float* d_yh, float* d_yt)
{	
	/* num of blocks per row or per col */
	int nb = (rows%r ? rows/r+1 : rows/r);

	/* Bind to texture */
  	cudaBindTexture(0, TEX_PTR  , d_ptr , (nb+1) * sizeof(d_ptr[0]));
   	cudaBindTexture(0, TEX_IDX  , d_idx , nz     * sizeof(d_idx[0]));
   	cudaBindTexture(0, TEX_VAL  , d_valh, nz*r*c * sizeof(d_valh[0]));
   	cudaBindTexture(0, TEX_VAL_T, d_valt, nz*r*c * sizeof(d_valt[0]));
   	cudaBindTexture(0, TEX_X    , d_xh  , cols   * sizeof(d_xh[0]));
   	cudaBindTexture(0, TEX_X_T  , d_xt  , cols   * sizeof(d_xt[0]));
	 
	/*
	 * Chose which kernel to run 
	 */
	
	double a = a0 + a1;
	double b = b0 + b1;
	
	 /* If beta != 0 */
	if (b != 0)
	{
	 	/* If beta != 1, do a scaling first */
	 	if (b != 1) 
			cuda_kernel_qscal(rows, b0, b1, d_yh, d_yt);
	 	
	 	/* If alpha = 1, no need to consider */                         
   		if (a == 1) {
   	   		if     (r == 2) kernel_qspmv_bcsr_2x2_a1 <<<CUKR_QSPMV_BCSR_2X2_CTAS, CUKR_QSPMV_BCSR_2X2_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 3) kernel_qspmv_bcsr_3x3_a1 <<<CUKR_QSPMV_BCSR_3X3_CTAS, CUKR_QSPMV_BCSR_3X3_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 4) kernel_qspmv_bcsr_4x4_a1 <<<CUKR_QSPMV_BCSR_4X4_CTAS, CUKR_QSPMV_BCSR_4X4_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 5) kernel_qspmv_bcsr_5x5_a1 <<<CUKR_QSPMV_BCSR_5X5_CTAS, CUKR_QSPMV_BCSR_5X5_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 6) kernel_qspmv_bcsr_6x6_a1 <<<CUKR_QSPMV_BCSR_6X6_CTAS, CUKR_QSPMV_BCSR_6X6_THREADS>>> (nb, d_yh, d_yt);
   	   		else printf("Block size (%d, %d) is NOT supported at cuda_qspmv_bcsr", r, c);
   	        	CUT_CHECK_ERROR("kernel_qspmv_bcsr_a1() execution failed\n");
   		} 
	 	/* Else, take the most general case */
	 	else {
	 		if     (r == 2) kernel_qspmv_bcsr_2x2 <<<CUKR_QSPMV_BCSR_2X2_CTAS, CUKR_QSPMV_BCSR_2X2_THREADS>>> (nb, a0, a1, d_yh, d_yt);
	 		else if(r == 3) kernel_qspmv_bcsr_3x3 <<<CUKR_QSPMV_BCSR_3X3_CTAS, CUKR_QSPMV_BCSR_3X3_THREADS>>> (nb, a0, a1, d_yh, d_yt);
	 		else if(r == 4) kernel_qspmv_bcsr_4x4 <<<CUKR_QSPMV_BCSR_4X4_CTAS, CUKR_QSPMV_BCSR_4X4_THREADS>>> (nb, a0, a1, d_yh, d_yt);
	 		else if(r == 5) kernel_qspmv_bcsr_5x5 <<<CUKR_QSPMV_BCSR_5X5_CTAS, CUKR_QSPMV_BCSR_5X5_THREADS>>> (nb, a0, a1, d_yh, d_yt);
	 		else if(r == 6) kernel_qspmv_bcsr_6x6 <<<CUKR_QSPMV_BCSR_6X6_CTAS, CUKR_QSPMV_BCSR_6X6_THREADS>>> (nb, a0, a1, d_yh, d_yt);
	 		else printf("Block size (%d, %d) is NOT supported at cuda_qspmv_bcsr", r, c);
	 		CUT_CHECK_ERROR("kernel_qspmv_bcsr() execution failed\n");    	             	           
	 	}
	 	  
	}
	/* If beta = 0 */
	else {
	 	/* If alpha = 1 as well, no need to
	 	 * consider them both */
	 	 if (a == 1) {
   	   		if     (r == 2) kernel_qspmv_bcsr_2x2_a1_b0 <<<CUKR_QSPMV_BCSR_2X2_CTAS, CUKR_QSPMV_BCSR_2X2_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 3) kernel_qspmv_bcsr_3x3_a1_b0 <<<CUKR_QSPMV_BCSR_3X3_CTAS, CUKR_QSPMV_BCSR_3X3_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 4) kernel_qspmv_bcsr_4x4_a1_b0 <<<CUKR_QSPMV_BCSR_4X4_CTAS, CUKR_QSPMV_BCSR_4X4_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 5) kernel_qspmv_bcsr_5x5_a1_b0 <<<CUKR_QSPMV_BCSR_5X5_CTAS, CUKR_QSPMV_BCSR_5X5_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 6) kernel_qspmv_bcsr_6x6_a1_b0 <<<CUKR_QSPMV_BCSR_6X6_CTAS, CUKR_QSPMV_BCSR_6X6_THREADS>>> (nb, d_yh, d_yt);
	 	 	else printf("Block size (%d, %d) is NOT supported at cuda_qspmv_bcsr", r, c);
   	        	CUT_CHECK_ERROR("kernel_qspmv_bcsr_a1_b0() execution failed\n");
	 	 }
	 	/* If alpha != 1, have to consider it */
	 	else {
	 		if     (r == 2) kernel_qspmv_bcsr_2x2_b0 <<<CUKR_QSPMV_BCSR_2X2_CTAS, CUKR_QSPMV_BCSR_2X2_THREADS>>> (nb, a0, a1, d_yh, d_yt);
	 		else if(r == 3) kernel_qspmv_bcsr_3x3_b0 <<<CUKR_QSPMV_BCSR_3X3_CTAS, CUKR_QSPMV_BCSR_3X3_THREADS>>> (nb, a0, a1, d_yh, d_yt);
	 		else if(r == 4) kernel_qspmv_bcsr_4x4_b0 <<<CUKR_QSPMV_BCSR_4X4_CTAS, CUKR_QSPMV_BCSR_4X4_THREADS>>> (nb, a0, a1, d_yh, d_yt);
	 		else if(r == 5) kernel_qspmv_bcsr_5x5_b0 <<<CUKR_QSPMV_BCSR_5X5_CTAS, CUKR_QSPMV_BCSR_5X5_THREADS>>> (nb, a0, a1, d_yh, d_yt);
	 		else if(r == 6) kernel_qspmv_bcsr_6x6_b0 <<<CUKR_QSPMV_BCSR_6X6_CTAS, CUKR_QSPMV_BCSR_6X6_THREADS>>> (nb, a0, a1, d_yh, d_yt);
   	    		else printf("Block size (%d, %d) is NOT supported at cuda_qspmv_bcsr", r, c);
   	        	CUT_CHECK_ERROR("kernel_qspmv_bcsr_b0() execution failed\n");
	 	}

	}   	


  	/* Do a syncronize, for timing */
  	CUDA_SAFE_CALL(cudaThreadSynchronize());
   
   	/* Un-bind textures */
   	cudaUnbindTexture(TEX_PTR);
   	cudaUnbindTexture(TEX_IDX);
   	cudaUnbindTexture(TEX_VAL);
   	cudaUnbindTexture(TEX_VAL_T);
   	cudaUnbindTexture(TEX_X);	
   	cudaUnbindTexture(TEX_X_T);
}	

// ------------------------------------------------------------------- // 

/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR4 format, in quasi-double precision
 * \param n Matrix size
 * \param nz Number of nonzeros
 * \param a0,a1 Scaling factor for Ax (head,tail)
 * \param ptr Row index vector 
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param b0, b1 Scale factor for y (head, tail)
 * \param yh,yt Head and tail for the result
 *              vector

 */
#ifdef __cplusplus
extern "C"
#endif 
void
cuda_qspmv_bcsr4(int rows, int cols, int nz, int r, int c, float a0, float a1, int* d_ptr, int* d_idx, 
			   float* d_valh, float* d_valt, float* d_xh, float* d_xt, 
			   float b0, float b1, float* d_yh, float* d_yt)
{	
	/* num of blocks per row or per col */
	int nb = (rows%r ? rows/r+1 : rows/r);

	/* Bind to texture */
  	cudaBindTexture(0, TEX_PTR  , d_ptr , (nb+1) * sizeof(d_ptr[0]));
   	cudaBindTexture(0, TEX_IDX  , d_idx , nz     * sizeof(d_idx[0]));
   	if((r*c)%4 == 3)     { cudaBindTexture(0, TEX_VAL4, d_valh, nz*(r*c+1) * sizeof(d_valh[0]));
   	                       cudaBindTexture(0, TEX_VAL4_T, d_valt, nz*(r*c+1) * sizeof(d_valt[0])); }
   	else if((r*c)%4 == 2){ cudaBindTexture(0, TEX_VAL4, d_valh, nz*(r*c+2) * sizeof(d_valh[0]));
   	                       cudaBindTexture(0, TEX_VAL4_T, d_valt, nz*(r*c+2) * sizeof(d_valt[0])); }
   	else if((r*c)%4 == 1){ cudaBindTexture(0, TEX_VAL4, d_valh, nz*(r*c+3) * sizeof(d_valh[0]));
   	                       cudaBindTexture(0, TEX_VAL4_T, d_valt, nz*(r*c+3) * sizeof(d_valt[0])); }
   	else                 { cudaBindTexture(0, TEX_VAL4, d_valh, nz* r*c    * sizeof(d_valh[0]));
   	                       cudaBindTexture(0, TEX_VAL4_T, d_valt, nz* r*c    * sizeof(d_valt[0])); }
   	cudaBindTexture(0, TEX_X    , d_xh  , cols   * sizeof(d_xh[0]));
   	cudaBindTexture(0, TEX_X_T  , d_xt  , cols   * sizeof(d_xt[0]));
	 
	/*
	 * Chose which kernel to run 
	 */
	
	double a = a0 + a1;
	double b = b0 + b1;
	
	 /* If beta != 0 */
	if (b != 0)
	{
	 	/* If beta != 1, do a scaling first */
	 	if (b != 1) 
			cuda_kernel_qscal(rows, b0, b1, d_yh, d_yt);
	 	
	 	/* If alpha = 1, no need to consider */                         
   		if (a == 1) {
   	   		if     (r == 2) kernel_qspmv_bcsr4_2x2_a1 <<<CUKR_QSPMV_BCSR4_2X2_CTAS, CUKR_QSPMV_BCSR4_2X2_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 3) kernel_qspmv_bcsr4_3x3_a1 <<<CUKR_QSPMV_BCSR4_3X3_CTAS, CUKR_QSPMV_BCSR4_3X3_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 4) kernel_qspmv_bcsr4_4x4_a1 <<<CUKR_QSPMV_BCSR4_4X4_CTAS, CUKR_QSPMV_BCSR4_4X4_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 5) kernel_qspmv_bcsr4_5x5_a1 <<<CUKR_QSPMV_BCSR4_5X5_CTAS, CUKR_QSPMV_BCSR4_5X5_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 6) kernel_qspmv_bcsr4_6x6_a1 <<<CUKR_QSPMV_BCSR4_6X6_CTAS, CUKR_QSPMV_BCSR4_6X6_THREADS>>> (nb, d_yh, d_yt);
   	   		else printf("Block size (%d, %d) is NOT supported at cuda_qspmv_bcsr4", r, c);
   	        	CUT_CHECK_ERROR("kernel_qspmv_bcsr4_a1() execution failed\n");
   		} 
	 	/* Else, take the most general case */
	 	else {
	 		if     (r == 2) kernel_qspmv_bcsr4_2x2 <<<CUKR_QSPMV_BCSR4_2X2_CTAS, CUKR_QSPMV_BCSR4_2X2_THREADS>>> (nb, a0, a1, d_yh, d_yt);
	 		else if(r == 3) kernel_qspmv_bcsr4_3x3 <<<CUKR_QSPMV_BCSR4_3X3_CTAS, CUKR_QSPMV_BCSR4_3X3_THREADS>>> (nb, a0, a1, d_yh, d_yt);
	 		else if(r == 4) kernel_qspmv_bcsr4_4x4 <<<CUKR_QSPMV_BCSR4_4X4_CTAS, CUKR_QSPMV_BCSR4_4X4_THREADS>>> (nb, a0, a1, d_yh, d_yt);
	 		else if(r == 5) kernel_qspmv_bcsr4_5x5 <<<CUKR_QSPMV_BCSR4_5X5_CTAS, CUKR_QSPMV_BCSR4_5X5_THREADS>>> (nb, a0, a1, d_yh, d_yt);
	 		else if(r == 6) kernel_qspmv_bcsr4_6x6 <<<CUKR_QSPMV_BCSR4_6X6_CTAS, CUKR_QSPMV_BCSR4_6X6_THREADS>>> (nb, a0, a1, d_yh, d_yt);
	 		else printf("Block size (%d, %d) is NOT supported at cuda_qspmv_bcsr4", r, c);
	 		CUT_CHECK_ERROR("kernel_qspmv_bcsr4() execution failed\n");    	             	           
	 	}
	 	  
	}
	/* If beta = 0 */
	else {
	 	/* If alpha = 1 as well, no need to
	 	 * consider them both */
	 	 if (a == 1) {
   	   		if     (r == 2) kernel_qspmv_bcsr4_2x2_a1_b0 <<<CUKR_QSPMV_BCSR4_2X2_CTAS, CUKR_QSPMV_BCSR4_2X2_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 3) kernel_qspmv_bcsr4_3x3_a1_b0 <<<CUKR_QSPMV_BCSR4_3X3_CTAS, CUKR_QSPMV_BCSR4_3X3_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 4) kernel_qspmv_bcsr4_4x4_a1_b0 <<<CUKR_QSPMV_BCSR4_4X4_CTAS, CUKR_QSPMV_BCSR4_4X4_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 5) kernel_qspmv_bcsr4_5x5_a1_b0 <<<CUKR_QSPMV_BCSR4_5X5_CTAS, CUKR_QSPMV_BCSR4_5X5_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 6) kernel_qspmv_bcsr4_6x6_a1_b0 <<<CUKR_QSPMV_BCSR4_6X6_CTAS, CUKR_QSPMV_BCSR4_6X6_THREADS>>> (nb, d_yh, d_yt);
	 	 	else printf("Block size (%d, %d) is NOT supported at cuda_qspmv_bcsr4", r, c);
   	        	CUT_CHECK_ERROR("kernel_qspmv_bcsr4_a1_b0() execution failed\n");
	 	 }
	 	/* If alpha != 1, have to consider it */
	 	else {
	 		if     (r == 2) kernel_qspmv_bcsr4_2x2_b0 <<<CUKR_QSPMV_BCSR4_2X2_CTAS, CUKR_QSPMV_BCSR4_2X2_THREADS>>> (nb, a0, a1, d_yh, d_yt);
	 		else if(r == 3) kernel_qspmv_bcsr4_3x3_b0 <<<CUKR_QSPMV_BCSR4_3X3_CTAS, CUKR_QSPMV_BCSR4_3X3_THREADS>>> (nb, a0, a1, d_yh, d_yt);
	 		else if(r == 4) kernel_qspmv_bcsr4_4x4_b0 <<<CUKR_QSPMV_BCSR4_4X4_CTAS, CUKR_QSPMV_BCSR4_4X4_THREADS>>> (nb, a0, a1, d_yh, d_yt);
	 		else if(r == 5) kernel_qspmv_bcsr4_5x5_b0 <<<CUKR_QSPMV_BCSR4_5X5_CTAS, CUKR_QSPMV_BCSR4_5X5_THREADS>>> (nb, a0, a1, d_yh, d_yt);
	 		else if(r == 6) kernel_qspmv_bcsr4_6x6_b0 <<<CUKR_QSPMV_BCSR4_6X6_CTAS, CUKR_QSPMV_BCSR4_6X6_THREADS>>> (nb, a0, a1, d_yh, d_yt);
   	    		else printf("Block size (%d, %d) is NOT supported at cuda_qspmv_bcsr4", r, c);
   	        	CUT_CHECK_ERROR("kernel_qspmv_bcsr4_b0() execution failed\n");
	 	}

	}   	


  	/* Do a syncronize, for timing */
  	CUDA_SAFE_CALL(cudaThreadSynchronize());
   
   	/* Un-bind textures */
   	cudaUnbindTexture(TEX_PTR);
   	cudaUnbindTexture(TEX_IDX);
   	cudaUnbindTexture(TEX_VAL);
   	cudaUnbindTexture(TEX_VAL_T);
   	cudaUnbindTexture(TEX_X);	
   	cudaUnbindTexture(TEX_X_T);
}	


// ------------------------------------------------------------------- // 

/**
 * \brief Wrapper for QSPMV_HYB 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions 
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr_hi/lo CSR ptr vectors
 * \param csr_idx CSR idx vectors 
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y 
 * \param y Result vector (head, tail)
 */
#ifdef __cplusplus
extern "C"
#endif   
void cuda_qspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha0, CUKR_SINGLE alpha1, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz, 
	                CUKR_INT* d_ell_idx, CUKR_SINGLE* d_ell_valh, CUKR_SINGLE* d_ell_valt, CUKR_INT* d_csr_ptr, CUKR_INT* d_csr_idx, CUKR_SINGLE* d_csr_valh, CUKR_SINGLE* d_csr_valt,  
	                CUKR_SINGLE* d_xh, CUKR_SINGLE* d_xt, CUKR_SINGLE beta0, CUKR_SINGLE beta1, CUKR_SINGLE* d_yh, CUKR_SINGLE* d_yt)
{		
   	
   	/* Bind texture */
   	cudaBindTexture(0, TEX_X  , d_xh, cols * sizeof(d_xh[0]));		
   	cudaBindTexture(0, TEX_X_T, d_xt, cols * sizeof(d_xt[0]));
	
	const unsigned int BLOCK_SIZE = 256;
	const unsigned int num_blocks = DIVIDE_INTO(rows, BLOCK_SIZE);
    const dim3 grid = dim3(num_blocks);
	
	/* Do the SPMV in ELL */
	kernel_qspmv_ell <<<grid, BLOCK_SIZE, 0>>>(rows, alpha0, alpha1, ell_nz_row, ell_stride, d_ell_idx, d_ell_valh, d_ell_valt, beta0, beta1, d_yh, d_yt);
   	CUT_CHECK_ERROR("kernel_qspmv_ell() execution failed\n");
   	
   	/* Do the rest in CSR4 */
   	if (csr_nz) {
		cudaBindTexture(0, TEX_PTR   , d_csr_ptr , (rows+1) * sizeof(d_csr_ptr[0]));
	   	cudaBindTexture(0, TEX_IDX4  , d_csr_idx , csr_nz   * sizeof(d_csr_idx[0]));
	   	cudaBindTexture(0, TEX_VAL4  , d_csr_valh, csr_nz   * sizeof(d_csr_valh[0]));
	   	cudaBindTexture(0, TEX_VAL4_T, d_csr_valt, csr_nz   * sizeof(d_csr_valt[0]));
	   	double alpha = alpha0 + alpha1;
		if (alpha == 1) {
	 	 	kernel_qspmv_csr4_a1 <<<CUKR_QSPMV_CSR4_CTAS, CUKR_QSPMV_CSR4_THREADS>>> (rows, d_yh, d_yt);
	        CUT_CHECK_ERROR("kernel_qspmv_csr4_a1() execution failed\n");
	 	 }
	    else {
	    	kernel_qspmv_csr4<<<CUKR_QSPMV_CSR4_CTAS, CUKR_QSPMV_CSR4_THREADS>>>
	              	                (rows, alpha0, alpha1, d_yh, d_yt);
	        CUT_CHECK_ERROR("kernel_qspmv_csr4() execution failed\n");
	    }
		 	  
	    /* Unbind textures */ 
	   	cudaUnbindTexture(TEX_PTR);
	   	cudaUnbindTexture(TEX_IDX4);
	   	cudaUnbindTexture(TEX_VAL4);
	   	cudaUnbindTexture(TEX_VAL4_T);	   	
   	}
   	
   	cudaUnbindTexture(TEX_X);
	cudaUnbindTexture(TEX_X_T);
   	
   	CUDA_SAFE_CALL(cudaThreadSynchronize());	  	
}	

// ------------------------------------------------------------------- // 

/**
 * \brief Wrapper for QSPMV_BHYB 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions 
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr_hi/lo CSR ptr vectors
 * \param csr_idx CSR idx vectors 
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y 
 * \param y Result vector (head, tail)
 */
#ifdef __cplusplus
extern "C"
#endif   
void cuda_qspmv_bhyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha0, CUKR_SINGLE alpha1, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz, 
	                CUKR_INT* d_bell_idx, CUKR_SINGLE* d_bell_valh, CUKR_SINGLE* d_bell_valt, CUKR_INT* d_bcsr_ptr, CUKR_INT* d_bcsr_idx, CUKR_SINGLE* d_bcsr_valh, CUKR_SINGLE* d_bcsr_valt,  
	                CUKR_SINGLE* d_xh, CUKR_SINGLE* d_xt, CUKR_SINGLE beta0, CUKR_SINGLE beta1, CUKR_SINGLE* d_yh, CUKR_SINGLE* d_yt)
{		
   
   	/* num of blocks per row or per col */
	int nb = (rows%r ? rows/r+1 : rows/r);
   
   	/* Bind texture */
   	cudaBindTexture(0, TEX_X  , d_xh, cols * sizeof(d_xh[0]));		
   	cudaBindTexture(0, TEX_X_T, d_xt, cols * sizeof(d_xt[0]));
   	cudaBindTexture(0, TEX_IDX,   d_bell_idx , bell_nz_row*bell_stride         * sizeof(d_bell_idx[0]));
   	cudaBindTexture(0, TEX_VAL,   d_bell_valh, bell_nz_row*bell_stride* r*c    * sizeof(d_bell_valh[0]));
   	cudaBindTexture(0, TEX_VAL_T, d_bell_valt, bell_nz_row*bell_stride* r*c    * sizeof(d_bell_valt[0]));
	
	/* Do the SPMV in BELL */
	if     (r==2) kernel_qspmv_bell_2x2 <<<CUKR_QSPMV_BELL_2X2_CTAS, CUKR_QSPMV_BELL_2X2_THREADS>>> (nb, alpha0, alpha1, bell_nz_row, bell_stride, beta0, beta1, d_yh, d_yt);
	else if(r == 3) kernel_qspmv_bell_3x3 <<<CUKR_QSPMV_BELL_3X3_CTAS, CUKR_QSPMV_BELL_3X3_THREADS>>> (nb, alpha0, alpha1, bell_nz_row, bell_stride, beta0, beta1, d_yh, d_yt);
	else if(r == 4) kernel_qspmv_bell_4x4 <<<CUKR_QSPMV_BELL_4X4_CTAS, CUKR_QSPMV_BELL_4X4_THREADS>>> (nb, alpha0, alpha1, bell_nz_row, bell_stride, beta0, beta1, d_yh, d_yt);
	else if(r == 5) kernel_qspmv_bell_5x5 <<<CUKR_QSPMV_BELL_5X5_CTAS, CUKR_QSPMV_BELL_5X5_THREADS>>> (nb, alpha0, alpha1, bell_nz_row, bell_stride, beta0, beta1, d_yh, d_yt);
	else if(r == 6) kernel_qspmv_bell_6x6 <<<CUKR_QSPMV_BELL_6X6_CTAS, CUKR_QSPMV_BELL_6X6_THREADS>>> (nb, alpha0, alpha1, bell_nz_row, bell_stride, beta0, beta1, d_yh, d_yt);
	else printf("Block size (%d, %d) is NOT supported at cuda_qspmv_bhyb", r, c);
   	CUT_CHECK_ERROR("kernel_qspmv_bell() execution failed\n");
   	
   	/* Unbind textures */
 	 cudaUnbindTexture(TEX_IDX);
 	 cudaUnbindTexture(TEX_VAL);
 	 cudaUnbindTexture(TEX_VAL_T);
   	
   	/* Do the rest in BCSR */
   	if (bcsr_nz) {
		cudaBindTexture(0, TEX_PTR   , d_bcsr_ptr , (nb+1)    * sizeof(d_bcsr_ptr[0]));
	   	cudaBindTexture(0, TEX_IDX  , d_bcsr_idx , bcsr_nz   * sizeof(d_bcsr_idx[0]));
	   	cudaBindTexture(0, TEX_VAL,   d_bcsr_valh, bcsr_nz* r*c    * sizeof(d_bcsr_valh[0]));
	   	cudaBindTexture(0, TEX_VAL_T, d_bcsr_valt, bcsr_nz* r*c    * sizeof(d_bcsr_valt[0]));
	   	
	   	double alpha = alpha0 + alpha1;
   		if (alpha == 1) {
   	   		if     (r == 2) kernel_qspmv_bcsr_2x2_a1 <<<CUKR_QSPMV_BCSR_2X2_CTAS, CUKR_QSPMV_BCSR_2X2_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 3) kernel_qspmv_bcsr_3x3_a1 <<<CUKR_QSPMV_BCSR_3X3_CTAS, CUKR_QSPMV_BCSR_3X3_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 4) kernel_qspmv_bcsr_4x4_a1 <<<CUKR_QSPMV_BCSR_4X4_CTAS, CUKR_QSPMV_BCSR_4X4_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 5) kernel_qspmv_bcsr_5x5_a1 <<<CUKR_QSPMV_BCSR_5X5_CTAS, CUKR_QSPMV_BCSR_5X5_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 6) kernel_qspmv_bcsr_6x6_a1 <<<CUKR_QSPMV_BCSR_6X6_CTAS, CUKR_QSPMV_BCSR_6X6_THREADS>>> (nb, d_yh, d_yt);
   	   		else printf("Block size (%d, %d) is NOT supported at cuda_qspmv_bhyb", r, c);
   	        	CUT_CHECK_ERROR("kernel_qspmv_bcsr_a1() execution failed\n");
   		} else {
	 		if     (r == 2) kernel_qspmv_bcsr_2x2 <<<CUKR_QSPMV_BCSR_2X2_CTAS, CUKR_QSPMV_BCSR_2X2_THREADS>>> (nb, alpha0, alpha1, d_yh, d_yt);
	 		else if(r == 3) kernel_qspmv_bcsr_3x3 <<<CUKR_QSPMV_BCSR_3X3_CTAS, CUKR_QSPMV_BCSR_3X3_THREADS>>> (nb, alpha0, alpha1, d_yh, d_yt);
	 		else if(r == 4) kernel_qspmv_bcsr_4x4 <<<CUKR_QSPMV_BCSR_4X4_CTAS, CUKR_QSPMV_BCSR_4X4_THREADS>>> (nb, alpha0, alpha1, d_yh, d_yt);
	 		else if(r == 5) kernel_qspmv_bcsr_5x5 <<<CUKR_QSPMV_BCSR_5X5_CTAS, CUKR_QSPMV_BCSR_5X5_THREADS>>> (nb, alpha0, alpha1, d_yh, d_yt);
	 		else if(r == 6) kernel_qspmv_bcsr_6x6 <<<CUKR_QSPMV_BCSR_6X6_CTAS, CUKR_QSPMV_BCSR_6X6_THREADS>>> (nb, alpha0, alpha1, d_yh, d_yt);
	 		else printf("Block size (%d, %d) is NOT supported at cuda_qspmv_bhyb", r, c);
	 		CUT_CHECK_ERROR("kernel_qspmv_bcsr() execution failed\n");    	             	           
	 	}
		 	  
	 	 /* Unbind textures */
	 	 cudaUnbindTexture(TEX_PTR);
	 	 cudaUnbindTexture(TEX_IDX);
	 	 cudaUnbindTexture(TEX_VAL);
	 	 cudaUnbindTexture(TEX_VAL_T);
   	}
   	
   	/* Unbind textures */
   	cudaUnbindTexture(TEX_X);
	cudaUnbindTexture(TEX_X_T);
   	
   	CUDA_SAFE_CALL(cudaThreadSynchronize());	  	
}	

// ------------------------------------------------------------------- // 

/**
 * \brief Wrapper for QSPMV_BHYB4
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions 
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr_hi/lo CSR ptr vectors
 * \param csr_idx CSR idx vectors 
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y 
 * \param y Result vector (head, tail)
 */
#ifdef __cplusplus
extern "C"
#endif   
void cuda_qspmv_bhyb4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha0, CUKR_SINGLE alpha1, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz, 
	                CUKR_INT* d_bell_idx, CUKR_SINGLE* d_bell_valh, CUKR_SINGLE* d_bell_valt, CUKR_INT* d_bcsr_ptr, CUKR_INT* d_bcsr_idx, CUKR_SINGLE* d_bcsr_valh, CUKR_SINGLE* d_bcsr_valt,  
	                CUKR_SINGLE* d_xh, CUKR_SINGLE* d_xt, CUKR_SINGLE beta0, CUKR_SINGLE beta1, CUKR_SINGLE* d_yh, CUKR_SINGLE* d_yt)
{		
   
   	/* num of blocks per row or per col */
	int nb = (rows%r ? rows/r+1 : rows/r);
   
   	/* Bind texture */
   	cudaBindTexture(0, TEX_X  , d_xh, cols * sizeof(d_xh[0]));		
   	cudaBindTexture(0, TEX_X_T, d_xt, cols * sizeof(d_xt[0]));
   	cudaBindTexture(0, TEX_IDX, d_bell_idx, bell_nz_row*bell_stride * sizeof(d_bcsr_idx[0]));
   	if((r*c)%4 == 3){
   		cudaBindTexture(0, TEX_VAL4,   d_bell_valh, bell_nz_row*bell_stride*(r*c+1) * sizeof(d_bell_valh[0]));
   		cudaBindTexture(0, TEX_VAL4_T, d_bell_valt, bell_nz_row*bell_stride*(r*c+1) * sizeof(d_bell_valt[0]));
   	}if((r*c)%4 == 2){
   		cudaBindTexture(0, TEX_VAL4,   d_bell_valh, bell_nz_row*bell_stride*(r*c+2) * sizeof(d_bell_valh[0]));
   		cudaBindTexture(0, TEX_VAL4_T, d_bell_valt, bell_nz_row*bell_stride*(r*c+2) * sizeof(d_bell_valt[0]));
   	}if((r*c)%4 == 1){
   		cudaBindTexture(0, TEX_VAL4,   d_bell_valh, bell_nz_row*bell_stride*(r*c+3) * sizeof(d_bell_valh[0]));
   		cudaBindTexture(0, TEX_VAL4_T, d_bell_valt, bell_nz_row*bell_stride*(r*c+3) * sizeof(d_bell_valt[0]));
   	}else{
   		cudaBindTexture(0, TEX_VAL4,   d_bell_valh, bell_nz_row*bell_stride* r*c    * sizeof(d_bell_valh[0]));
   		cudaBindTexture(0, TEX_VAL4_T, d_bell_valt, bell_nz_row*bell_stride* r*c    * sizeof(d_bell_valt[0]));
   	}
	
	/* Do the SPMV in BELL */
	if     (r==2) kernel_qspmv_bell4_2x2 <<<CUKR_QSPMV_BELL4_2X2_CTAS, CUKR_QSPMV_BELL4_2X2_THREADS>>> (nb, alpha0, alpha1, bell_nz_row, bell_stride, beta0, beta1, d_yh, d_yt);
	else if(r == 3) kernel_qspmv_bell4_3x3 <<<CUKR_QSPMV_BELL4_3X3_CTAS, CUKR_QSPMV_BELL4_3X3_THREADS>>> (nb, alpha0, alpha1, bell_nz_row, bell_stride, beta0, beta1, d_yh, d_yt);
	else if(r == 4) kernel_qspmv_bell4_4x4 <<<CUKR_QSPMV_BELL4_4X4_CTAS, CUKR_QSPMV_BELL4_4X4_THREADS>>> (nb, alpha0, alpha1, bell_nz_row, bell_stride, beta0, beta1, d_yh, d_yt);
	else if(r == 5) kernel_qspmv_bell4_5x5 <<<CUKR_QSPMV_BELL4_5X5_CTAS, CUKR_QSPMV_BELL4_5X5_THREADS>>> (nb, alpha0, alpha1, bell_nz_row, bell_stride, beta0, beta1, d_yh, d_yt);
	else if(r == 6) kernel_qspmv_bell4_6x6 <<<CUKR_QSPMV_BELL4_6X6_CTAS, CUKR_QSPMV_BELL4_6X6_THREADS>>> (nb, alpha0, alpha1, bell_nz_row, bell_stride, beta0, beta1, d_yh, d_yt);
	else printf("Block size (%d, %d) is NOT supported at cuda_qspmv_bhyb4", r, c);
   	CUT_CHECK_ERROR("kernel_qspmv_bell() execution failed\n");

 	 /* Unbind textures */ 
 	 cudaUnbindTexture(TEX_IDX4);
 	 cudaUnbindTexture(TEX_VAL4);
 	 cudaUnbindTexture(TEX_VAL4_T);
   	
   	/* Do the rest in BCSR */
   	if (bcsr_nz) {
   	   	/* Bind texture */
		cudaBindTexture(0, TEX_PTR   , d_bcsr_ptr , rows * sizeof(d_bcsr_ptr[0]));
	   	cudaBindTexture(0, TEX_IDX4  , d_bcsr_idx , bcsr_nz   * sizeof(d_bcsr_idx[0]));
	   	if((r*c)%4 == 3){
	   		cudaBindTexture(0, TEX_VAL4,   d_bcsr_valh, bcsr_nz*(r*c+1) * sizeof(d_bcsr_valh[0]));
	   		cudaBindTexture(0, TEX_VAL4_T, d_bcsr_valt, bcsr_nz*(r*c+1) * sizeof(d_bcsr_valt[0]));
	   	}else if((r*c)%4 == 2){
	   		cudaBindTexture(0, TEX_VAL4,   d_bcsr_valh, bcsr_nz*(r*c+2) * sizeof(d_bcsr_valh[0]));
	   		cudaBindTexture(0, TEX_VAL4_T, d_bcsr_valt, bcsr_nz*(r*c+2) * sizeof(d_bcsr_valt[0]));
	   	}else if((r*c)%4 == 1){
	   		cudaBindTexture(0, TEX_VAL4,   d_bcsr_valh, bcsr_nz*(r*c+3) * sizeof(d_bcsr_valh[0]));
	   		cudaBindTexture(0, TEX_VAL4_T, d_bcsr_valt, bcsr_nz*(r*c+3) * sizeof(d_bcsr_valt[0]));
	   	}else{
	   		cudaBindTexture(0, TEX_VAL4,   d_bcsr_valh, bcsr_nz* r*c    * sizeof(d_bcsr_valh[0]));
	   		cudaBindTexture(0, TEX_VAL4_T, d_bcsr_valt, bcsr_nz* r*c    * sizeof(d_bcsr_valt[0]));
	   	}
	   	
	   	double alpha = alpha0 + alpha1;
   		if (alpha == 1) {
   	   		if     (r == 2) kernel_qspmv_bcsr4_2x2_a1 <<<CUKR_QSPMV_BCSR4_2X2_CTAS, CUKR_QSPMV_BCSR4_2X2_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 3) kernel_qspmv_bcsr4_3x3_a1 <<<CUKR_QSPMV_BCSR4_3X3_CTAS, CUKR_QSPMV_BCSR4_3X3_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 4) kernel_qspmv_bcsr4_4x4_a1 <<<CUKR_QSPMV_BCSR4_4X4_CTAS, CUKR_QSPMV_BCSR4_4X4_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 5) kernel_qspmv_bcsr4_5x5_a1 <<<CUKR_QSPMV_BCSR4_5X5_CTAS, CUKR_QSPMV_BCSR4_5X5_THREADS>>> (nb, d_yh, d_yt);
   	   		else if(r == 6) kernel_qspmv_bcsr4_6x6_a1 <<<CUKR_QSPMV_BCSR4_6X6_CTAS, CUKR_QSPMV_BCSR4_6X6_THREADS>>> (nb, d_yh, d_yt);
   	   		else printf("Block size (%d, %d) is NOT supported at cuda_qspmv_bcsr4", r, c);
   	        	CUT_CHECK_ERROR("kernel_qspmv_bcsr4_a1() execution failed\n");
   		} else {
	 		if     (r == 2) kernel_qspmv_bcsr4_2x2 <<<CUKR_QSPMV_BCSR4_2X2_CTAS, CUKR_QSPMV_BCSR4_2X2_THREADS>>> (nb, alpha0, alpha1, d_yh, d_yt);
	 		else if(r == 3) kernel_qspmv_bcsr4_3x3 <<<CUKR_QSPMV_BCSR4_3X3_CTAS, CUKR_QSPMV_BCSR4_3X3_THREADS>>> (nb, alpha0, alpha1, d_yh, d_yt);
	 		else if(r == 4) kernel_qspmv_bcsr4_4x4 <<<CUKR_QSPMV_BCSR4_4X4_CTAS, CUKR_QSPMV_BCSR4_4X4_THREADS>>> (nb, alpha0, alpha1, d_yh, d_yt);
	 		else if(r == 5) kernel_qspmv_bcsr4_5x5 <<<CUKR_QSPMV_BCSR4_5X5_CTAS, CUKR_QSPMV_BCSR4_5X5_THREADS>>> (nb, alpha0, alpha1, d_yh, d_yt);
	 		else if(r == 6) kernel_qspmv_bcsr4_6x6 <<<CUKR_QSPMV_BCSR4_6X6_CTAS, CUKR_QSPMV_BCSR4_6X6_THREADS>>> (nb, alpha0, alpha1, d_yh, d_yt);
	 		else printf("Block size (%d, %d) is NOT supported at cuda_qspmv_bcsr4", r, c);
	 		CUT_CHECK_ERROR("kernel_qspmv_bcsr4() execution failed\n");    	             	           
	 	}
		 	  
	 	 /* Unbind textures */
	 	 cudaUnbindTexture(TEX_PTR);
	 	 cudaUnbindTexture(TEX_IDX4);
	 	 cudaUnbindTexture(TEX_VAL4);
	 	 cudaUnbindTexture(TEX_VAL4_T);	   	
   	}
   	
   	/* Unbind textures */
   	cudaUnbindTexture(TEX_X);
	cudaUnbindTexture(TEX_X_T);
   	
   	CUDA_SAFE_CALL(cudaThreadSynchronize());	  	
}	

//============================ PRIVATE FUNCTIONS =========================== //



/**
 * \brief Aux function to check for malloc
 *        errors on the GPU 
 * \param status The status returned by the
 *        malloc function
 */
void HandleCudaMallocErrors(cublasStatus status)
{
	if (status == CUBLAS_STATUS_NOT_INITIALIZED) {
		printf("FATAL: Error allocating GPU memory for QDOT. "\
		             "Reason: CUBLAS has not been initialized");
		exit(1);
	}
	else if (status == CUBLAS_STATUS_INVALID_VALUE) {
		printf("FATAL: Error allocating GPU memory for QDOT. "\
		            "Reason: Invalid value");
		exit(1);
	}
	else if (status == CUBLAS_STATUS_ALLOC_FAILED) {
		printf("FATAL: Error allocating GPU memory for QDOT. "\
		            "Reason: Allocation failure");
		exit(1);
	}
	else if (status == CUBLAS_STATUS_MAPPING_ERROR) {
		printf("FATAL: Error allocating GPU memory for QDOT. "\
		            "Reason: Mapping error");
		exit(1);
	}
}
