/**
 * @file blas/gpu/cuda/kernels_single.cu
 * @author Serban Georgescu
 * @since 2008
 *
 * Yusa adds BCSR in winter, 2009.
 *
 * Acts as a bridge between the C++ CUDA
 * implementation and the rest of the C
 * files. All the CUDA single precision
 * kernels must be included here.
 *
 * The following kernels are included:
 *  - aypx
 *  + spmv_csr
 * 	- spmv_csr
 * 	- spmv_csr_a1
 * 	- spmv_csr_b0
 * 	- spmv_csr_a1_b0
 *  + spmv_csr4
 * 	- spmv_csr4
 * 	- spmv_csr4_a1
 * 	- spmv_csr4_b0
 * 	- spmv_csr4_a1_b0
 *  + spmv_bcsr
 * 	- spmv_bcsr
 * 	- spmv_bcsr_a1
 * 	- spmv_bcsr_b0
 * 	- spmv_bcsr_a1_b0
 *  + spmv_bcsr4
 * 	- spmv_bcsr4
 * 	- spmv_bcsr4_a1
 * 	- spmv_bcsr4_b0
 * 	- spmv_bcsr4_a1_b0
 */

#define LOGGER_NAME "CUKr.blas.impl.gpu.cuda.kernels_single"

#include <stdio.h>
#include <cuda.h>
#include <cutil.h>
#include <cublas.h>
#include <unistd.h>

#include "../../../../cukr.h"
#include "../../../../util/macros.h"

#include "kernels_config.h"
#include "single/kernel_saypx.h"
#include "single/kernel_sspmv_csr_a1_b0.h"
#include "single/kernel_sspmv_csr_b0.h"
#include "single/kernel_sspmv_csr_a1.h"
#include "single/kernel_sspmv_csr.h"
#include "single/kernel_sspmv_csr4_a1_b0.h"
#include "single/kernel_sspmv_csr4_b0.h"
#include "single/kernel_sspmv_csr4_a1.h"
#include "single/kernel_sspmv_csr4.h"

// BCSR
#include "single/kernel_sspmv_bcsr_a1_b0.h"
#include "single/kernel_sspmv_bcsr_b0.h"
#include "single/kernel_sspmv_bcsr_a1.h"
#include "single/kernel_sspmv_bcsr.h"

// BCSR4
#include "single/kernel_sspmv_bcsr4_a1_b0.h"
#include "single/kernel_sspmv_bcsr4_b0.h"
#include "single/kernel_sspmv_bcsr4_a1.h"
#include "single/kernel_sspmv_bcsr4.h"
 
// ELL
#include "single/kernel_sspmv_ell.h"

// BELL
#include "single/kernel_sspmv_bell.h"

// BELL4
#include "single/kernel_sspmv_bell4.h"

#define MAX(a,b) (a > b ? a : b)
#define CUKR_TRUE  1
#define CUKR_FALSE 0
#define DIVIDE_INTO(x,y) ((x + y - 1)/y) 

/**
 * \brief AYPX kernl on the GPU
 * \param n Vector size
 * \param a The constant
 * \param d_x, d_y Device pointers to the
 *        vectors involved in the operation.
 */
#ifdef __cplusplus
extern "C"
#endif
void cuda_kernel_saypx(int n, float a, float *d_x, float *d_y)
{
	/* Preparations */
	int dynMem = 2 * CUKR_SAYPX_THREADS * sizeof(float);

	/* Kernel invocation */
	kernel_saypx<<<CUKR_SAYPX_CTAS, CUKR_SAYPX_THREADS, dynMem>>> (d_x, d_y, a, n);
	CUT_CHECK_ERROR("kernel_saypx() execution failed\n");

	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize());

}

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in
 *        CSR format, in double precision
 * \param n Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param d_ptr Row index vector (device ptr)
 * \param d_idx Column index vector (device ptr)
 * \param d_val Value vector (device ptr)
 * \param d_x Vector being multiplied (device ptr)
 * \param beta Scaling factor for y
 * \param d_y Result vector (device ptr)
 */
#ifdef __cplusplus
extern "C"
#endif
void
cuda_sspmv_csr(int rows, int cols, int nz, float alpha, int* d_ptr, int* d_idx,
			   float* d_val, float* d_x, float beta, float* d_y)
{


  	cudaBindTexture(0, TEX_PTR , d_ptr, (rows + 1) * sizeof(d_ptr[0]));
   	cudaBindTexture(0, TEX_IDX , d_idx, nz         * sizeof(d_idx[0]));
   	cudaBindTexture(0, TEX_VAL , d_val, nz         * sizeof(d_val[0]));
   	cudaBindTexture(0, TEX_X   , d_x  , cols       * sizeof(d_x[0]));

	/*
	 * Chose which kernel to run
	 */

	 /* If beta != 0 */
	 if (beta != 0)
	 {
	 	/* If beta != 1, do a scaling first */
	 	if (beta != 1)
	 		cublasSscal(rows, beta, d_y, 1);

	 	/* If alpha = 1, no need to consider */
   		if (alpha == 1) {
   	   		kernel_sspmv_csr_a1 <<<CUKR_SSPMV_CSR_CTAS, CUKR_SSPMV_CSR_THREADS,
   	        	                   CUKR_SSPMV_CSR_THREADS * sizeof(float)>>>
   	             	               (rows, d_y);
   	        CUT_CHECK_ERROR("kernel_spmv_csr_a1() execution failed\n");
   		}
	 	/* Else, take the most general case */
	 	else {
	 		kernel_sspmv_csr <<<CUKR_SSPMV_CSR_CTAS, CUKR_SSPMV_CSR_THREADS>>> (rows, alpha, d_y);
            CUT_CHECK_ERROR("kernel_spmv_csr() execution failed\n");
	 	}

	 }
	 /* If beta = 0 */
	 else {
	 	/* If alpha = 1 as well, no need to
	 	 * consider them both */
	 	 if (alpha == 1) {
	 	 	kernel_sspmv_csr_a1_b0 <<<CUKR_SSPMV_CSR_CTAS, CUKR_SSPMV_CSR_THREADS>>> (rows, d_y);
	 	 	//kernel_sspmv_csr_a1_b0_c <<<CUKR_SSPMV_CSR_CTAS, CUKR_SSPMV_CSR_THREADS>>> (rows, d_x, d_y);
   	        CUT_CHECK_ERROR("kernel_spmv_csr_a1_b0() execution failed\n");
	 	 }
   	    /* If alpha != 1, have to consider it */
   	    else {
   	    	kernel_sspmv_csr_b0 <<<CUKR_SSPMV_CSR_CTAS, CUKR_SSPMV_CSR_THREADS,
   	         	                   CUKR_SSPMV_CSR_THREADS * sizeof(float)>>>
   	              	               (rows, alpha, d_y);
   	        CUT_CHECK_ERROR("kernel_spmv_csr_b0() execution failed\n");
   	    }

	 }


   	/* Un-bind textures */
   	cudaUnbindTexture(TEX_PTR);
   	cudaUnbindTexture(TEX_IDX);
   	cudaUnbindTexture(TEX_VAL);
   	cudaUnbindTexture(TEX_X);


   	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize());

}

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in
 *        CSR4 format, in double precision
 * \param n Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param d_ptr Row index vector (device ptr)
 * \param d_idx Column index vector (device ptr)
 * \param d_val Value vector (device ptr)
 * \param d_x Vector being multiplied (device ptr)
 * \param beta Scaling factor for y
 * \param d_y Result vector (device ptr)
 */
#ifdef __cplusplus
extern "C"
#endif
void
cuda_sspmv_csr4(int rows, int cols, int nz, float alpha, int* d_ptr, int* d_idx,
			    float* d_val, float* d_x, float beta, float* d_y)
{

	/* Bind to texture */
   	cudaBindTexture(0, TEX_PTR , d_ptr, (rows + 1) * sizeof(d_ptr[0]));
   	cudaBindTexture(0, TEX_IDX4, d_idx, nz         * sizeof(d_idx[0]));
   	cudaBindTexture(0, TEX_VAL4, d_val, nz         * sizeof(d_val[0]));
   	cudaBindTexture(0, TEX_X   , d_x  , cols       * sizeof(d_x[0]));


	/*
	 * Chose which kernel to run
	 */

	 /* If beta != 0 */
	 if (beta != 0)
	 {
	 	/* If beta != 1, do a scaling first */
	 	if (beta != 1)
	 		cublasSscal(rows, beta, d_y, 1);

	 	/* If alpha = 1, no need to consider */
   		if (alpha == 1) {
   	   		kernel_sspmv_csr4_a1 <<<CUKR_SSPMV_CSR4_CTAS, CUKR_SSPMV_CSR4_THREADS>>>(rows, d_y);
   	        CUT_CHECK_ERROR("kernel_spmv_csr4_a1() execution failed\n");
   		}
	 	/* Else, take the most general case */
	 	else {
	 		kernel_sspmv_csr4 <<<CUKR_SSPMV_CSR4_CTAS, CUKR_SSPMV_CSR4_THREADS>>>(rows, alpha, d_y);
            CUT_CHECK_ERROR("kernel_spmv_csr4() execution failed\n");
	 	}

	 }
	 /* If beta = 0 */
	 else {
	 	/* If alpha = 1 as well, no need to
	 	 * consider them both */
	 	 if (alpha == 1) {
	 	 	kernel_sspmv_csr4_a1_b0 <<<CUKR_SSPMV_CSR4_CTAS, CUKR_SSPMV_CSR4_THREADS>>> (rows, d_y);
   	        CUT_CHECK_ERROR("kernel_spmv_csr4_a1_b0() execution failed\n");
	 	 }
   	    /* If alpha != 1, have to consider it */
   	    else {
   	    	kernel_sspmv_csr4_b0 <<<CUKR_SSPMV_CSR4_CTAS, CUKR_SSPMV_CSR4_THREADS>>> (rows, alpha, d_y);
   	        CUT_CHECK_ERROR("kernel_spmv_csr4_b0() execution failed\n");
   	    }

	 }

   	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize());

    /* Unbind textures */
   	cudaUnbindTexture(TEX_PTR);
   	cudaUnbindTexture(TEX_IDX4);
   	cudaUnbindTexture(TEX_VAL4);
   	cudaUnbindTexture(TEX_X);

}


// ------------------------------------------------------------------- //

/**
 * Yusa writes in winter, 2009.
 * \brief Sparse matrix vector multiply in
 *        BCSR format, in single precision
 * \param rows, cols Matrix size
 * \param r, c Block size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param d_ptr Row index vector (device ptr)
 * \param d_idx Column index vector (device ptr)
 * \param d_val Value vector (device ptr)
 * \param d_x Vector being multiplied (device ptr)
 * \param beta Scaling factor for y
 * \param d_y Result vector (device ptr)
 */
#ifdef __cplusplus
extern "C"
#endif
void
cuda_sspmv_bcsr(int rows, int cols, int nz, int r, int c, float alpha, int* d_ptr, int* d_idx,
			   float* d_val, float* d_x, float beta, float* d_y)
{
	/* num of blocks per row or per col */
	int nb = (rows%r ? rows/r+1 : rows/r);

  	cudaBindTexture(0, TEX_PTR , d_ptr, (nb+1) * sizeof(d_ptr[0]));
   	cudaBindTexture(0, TEX_IDX , d_idx, nz     * sizeof(d_idx[0]));
   	cudaBindTexture(0, TEX_VAL , d_val, nz*r*c * sizeof(d_val[0]));
   	cudaBindTexture(0, TEX_X   , d_x  , cols   * sizeof(d_x[0]));

	/*
	 * Chose which kernel to run
	 */

	 /* If beta != 0 */
	 if (beta != 0)
	 {
	 	/* If beta != 1, do a scaling first */
	 	if (beta != 1)
	 		cublasSscal(rows, beta, d_y, 1);

	 	/* If alpha = 1, no need to consider */
   		if (alpha == 1) {
   	   		if     (r == 2) kernel_sspmv_bcsr_2x2_a1 <<<CUKR_SSPMV_BCSR_2X2_CTAS, CUKR_SSPMV_BCSR_2X2_THREADS>>> (nb, d_y);
   	   		else if(r == 3) kernel_sspmv_bcsr_3x3_a1 <<<CUKR_SSPMV_BCSR_3X3_CTAS, CUKR_SSPMV_BCSR_3X3_THREADS>>> (nb, d_y);
   	   		else if(r == 4) kernel_sspmv_bcsr_4x4_a1 <<<CUKR_SSPMV_BCSR_4X4_CTAS, CUKR_SSPMV_BCSR_4X4_THREADS>>> (nb, d_y);
   	   		else if(r == 5) kernel_sspmv_bcsr_5x5_a1 <<<CUKR_SSPMV_BCSR_5X5_CTAS, CUKR_SSPMV_BCSR_5X5_THREADS>>> (nb, d_y);
   	   		else if(r == 6) kernel_sspmv_bcsr_6x6_a1 <<<CUKR_SSPMV_BCSR_6X6_CTAS, CUKR_SSPMV_BCSR_6X6_THREADS>>> (nb, d_y);
   	   		else printf("Block size (%d, %d) is NOT supported at cuda_sspmv_bcsr", r, c);
   	        	CUT_CHECK_ERROR("kernel_spmv_bcsr_a1() execution failed\n");
   		}
	 	/* Else, take the most general case */
	 	else {
   	   		if     (r == 2) kernel_sspmv_bcsr_2x2 <<<CUKR_SSPMV_BCSR_2X2_CTAS, CUKR_SSPMV_BCSR_2X2_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 3) kernel_sspmv_bcsr_3x3 <<<CUKR_SSPMV_BCSR_3X3_CTAS, CUKR_SSPMV_BCSR_3X3_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 4) kernel_sspmv_bcsr_4x4 <<<CUKR_SSPMV_BCSR_4X4_CTAS, CUKR_SSPMV_BCSR_4X4_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 5) kernel_sspmv_bcsr_5x5 <<<CUKR_SSPMV_BCSR_5X5_CTAS, CUKR_SSPMV_BCSR_5X5_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 6) kernel_sspmv_bcsr_6x6 <<<CUKR_SSPMV_BCSR_6X6_CTAS, CUKR_SSPMV_BCSR_6X6_THREADS>>> (nb, alpha, d_y);
	 		else printf("Block size (%d, %d) is NOT supported at cuda_sspmv_bcsr", r, c);
            		CUT_CHECK_ERROR("kernel_spmv_bcsr() execution failed\n");
	 	}

	 }
	 /* If beta = 0 */
	 else {
	 	/* If alpha = 1 as well, no need to
	 	 * consider them both */
	 	 if (alpha == 1) {
   	   		if     (r == 2) kernel_sspmv_bcsr_2x2_a1_b0 <<<CUKR_SSPMV_BCSR_2X2_CTAS, CUKR_SSPMV_BCSR_2X2_THREADS>>> (nb, d_y);
   	   		else if(r == 3) kernel_sspmv_bcsr_3x3_a1_b0 <<<CUKR_SSPMV_BCSR_3X3_CTAS, CUKR_SSPMV_BCSR_3X3_THREADS>>> (nb, d_y);
   	   		else if(r == 4) kernel_sspmv_bcsr_4x4_a1_b0 <<<CUKR_SSPMV_BCSR_4X4_CTAS, CUKR_SSPMV_BCSR_4X4_THREADS>>> (nb, d_y);
   	   		else if(r == 5) kernel_sspmv_bcsr_5x5_a1_b0 <<<CUKR_SSPMV_BCSR_5X5_CTAS, CUKR_SSPMV_BCSR_5X5_THREADS>>> (nb, d_y);
   	   		else if(r == 6) kernel_sspmv_bcsr_6x6_a1_b0 <<<CUKR_SSPMV_BCSR_6X6_CTAS, CUKR_SSPMV_BCSR_6X6_THREADS>>> (nb, d_y);
	 	 	else printf("Block size (%d, %d) is NOT supported at cuda_sspmv_bcsr", r, c);
   	         	CUT_CHECK_ERROR("kernel_spmv_bcsr_a1_b0() execution failed\n");
	 	 }
   	    	/* If alpha != 1, have to consider it */
   	    	else {
   	   		if     (r == 2) kernel_sspmv_bcsr_2x2_b0 <<<CUKR_SSPMV_BCSR_2X2_CTAS, CUKR_SSPMV_BCSR_2X2_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 3) kernel_sspmv_bcsr_3x3_b0 <<<CUKR_SSPMV_BCSR_3X3_CTAS, CUKR_SSPMV_BCSR_3X3_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 4) kernel_sspmv_bcsr_4x4_b0 <<<CUKR_SSPMV_BCSR_4X4_CTAS, CUKR_SSPMV_BCSR_4X4_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 5) kernel_sspmv_bcsr_5x5_b0 <<<CUKR_SSPMV_BCSR_5X5_CTAS, CUKR_SSPMV_BCSR_5X5_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 6) kernel_sspmv_bcsr_6x6_b0 <<<CUKR_SSPMV_BCSR_6X6_CTAS, CUKR_SSPMV_BCSR_6X6_THREADS>>> (nb, alpha, d_y);
   	    		else printf("Block size (%d, %d) is NOT supported at cuda_sspmv_bcsr", r, c);
   	        	CUT_CHECK_ERROR("kernel_spmv_bcsr_b0() execution failed\n");
   	    }

	 }


   	/* Un-bind textures */
   	cudaUnbindTexture(TEX_PTR);
   	cudaUnbindTexture(TEX_IDX);
   	cudaUnbindTexture(TEX_VAL);
   	cudaUnbindTexture(TEX_X);


   	/* Do a syncronize, for timing */
    	CUDA_SAFE_CALL(cudaThreadSynchronize());

}

// ------------------------------------------------------------------- //

/**
 * Yusa writes in winter, 2009.
 * \brief Sparse matrix vector multiply in
 *        BCSR4 format, in single precision
 * \param rows, cols Matrix size
 * \param r, c Block size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param d_ptr Row index vector (device ptr)
 * \param d_idx Column index vector (device ptr)
 * \param d_val Value vector (device ptr)
 * \param d_x Vector being multiplied (device ptr)
 * \param beta Scaling factor for y
 * \param d_y Result vector (device ptr)
 */
#ifdef __cplusplus
extern "C"
#endif
void
cuda_sspmv_bcsr4(int rows, int cols, int nz, int r, int c, float alpha, int* d_ptr, int* d_idx,
			   float* d_val, float* d_x, float beta, float* d_y)
{


	/* num of blocks per row or per col */
	int nb = (rows%r ? rows/r+1 : rows/r);

  	cudaBindTexture(0, TEX_PTR , d_ptr, (nb+1) * sizeof(d_ptr[0]));
   	cudaBindTexture(0, TEX_IDX , d_idx, nz     * sizeof(d_idx[0]));
   	if((r*c)%4 == 3) cudaBindTexture(0, TEX_VAL4, d_val, nz*(r*c+1) * sizeof(d_val[0]));
   	if((r*c)%4 == 2) cudaBindTexture(0, TEX_VAL4, d_val, nz*(r*c+2) * sizeof(d_val[0]));
   	if((r*c)%4 == 1) cudaBindTexture(0, TEX_VAL4, d_val, nz*(r*c+3) * sizeof(d_val[0]));
   	else             cudaBindTexture(0, TEX_VAL4, d_val, nz* r*c    * sizeof(d_val[0]));
   	cudaBindTexture(0, TEX_X   , d_x  , cols   * sizeof(d_x[0]));

	/*
	 * Chose which kernel to run
	 */

	 /* If beta != 0 */
	 if (beta != 0)
	 {
	 	/* If beta != 1, do a scaling first */
	 	if (beta != 1)
	 		cublasSscal(rows, beta, d_y, 1);

	 	/* If alpha = 1, no need to consider */
   		if (alpha == 1) {
   	   		if     (r == 2) kernel_sspmv_bcsr4_2x2_a1 <<<CUKR_SSPMV_BCSR4_2X2_CTAS, CUKR_SSPMV_BCSR4_2X2_THREADS>>> (nb, d_y);
   	   		else if(r == 3) kernel_sspmv_bcsr4_3x3_a1 <<<CUKR_SSPMV_BCSR4_3X3_CTAS, CUKR_SSPMV_BCSR4_3X3_THREADS>>> (nb, d_y);
   	   		else if(r == 4) kernel_sspmv_bcsr4_4x4_a1 <<<CUKR_SSPMV_BCSR4_4X4_CTAS, CUKR_SSPMV_BCSR4_4X4_THREADS>>> (nb, d_y);
   	   		else if(r == 5) kernel_sspmv_bcsr4_5x5_a1 <<<CUKR_SSPMV_BCSR4_5X5_CTAS, CUKR_SSPMV_BCSR4_5X5_THREADS>>> (nb, d_y);
   	   		else if(r == 6) kernel_sspmv_bcsr4_6x6_a1 <<<CUKR_SSPMV_BCSR4_6X6_CTAS, CUKR_SSPMV_BCSR4_6X6_THREADS>>> (nb, d_y);
	 		else printf("Block size (%d, %d) is NOT supported at cuda_sspmv_bcsr4", r, c);
   	        	CUT_CHECK_ERROR("kernel_spmv_bcsr4_a1() execution failed\n");
   		}
	 	/* Else, take the most general case */
	 	else {
   	   		if     (r == 2) kernel_sspmv_bcsr4_2x2 <<<CUKR_SSPMV_BCSR4_2X2_CTAS, CUKR_SSPMV_BCSR4_2X2_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 3) kernel_sspmv_bcsr4_3x3 <<<CUKR_SSPMV_BCSR4_3X3_CTAS, CUKR_SSPMV_BCSR4_3X3_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 4) kernel_sspmv_bcsr4_4x4 <<<CUKR_SSPMV_BCSR4_4X4_CTAS, CUKR_SSPMV_BCSR4_4X4_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 5) kernel_sspmv_bcsr4_5x5 <<<CUKR_SSPMV_BCSR4_5X5_CTAS, CUKR_SSPMV_BCSR4_5X5_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 6) kernel_sspmv_bcsr4_6x6 <<<CUKR_SSPMV_BCSR4_6X6_CTAS, CUKR_SSPMV_BCSR4_6X6_THREADS>>> (nb, alpha, d_y);
	 		else printf("Block size (%d, %d) is NOT supported at cuda_sspmv_bcsr4", r, c);
            		CUT_CHECK_ERROR("kernel_spmv_bcsr4() execution failed\n");
	 	}

	 }
	 /* If beta = 0 */
	 else {
	 	/* If alpha = 1 as well, no need to
	 	 * consider them both */
	 	 if (alpha == 1) {
   	   		if     (r == 2) kernel_sspmv_bcsr4_2x2_a1_b0 <<<CUKR_SSPMV_BCSR4_2X2_CTAS, CUKR_SSPMV_BCSR4_2X2_THREADS>>> (nb, d_y);
   	   		else if(r == 3) kernel_sspmv_bcsr4_3x3_a1_b0 <<<CUKR_SSPMV_BCSR4_3X3_CTAS, CUKR_SSPMV_BCSR4_3X3_THREADS>>> (nb, d_y);
   	   		else if(r == 4) kernel_sspmv_bcsr4_4x4_a1_b0 <<<CUKR_SSPMV_BCSR4_4X4_CTAS, CUKR_SSPMV_BCSR4_4X4_THREADS>>> (nb, d_y);
   	   		else if(r == 5) kernel_sspmv_bcsr4_5x5_a1_b0 <<<CUKR_SSPMV_BCSR4_5X5_CTAS, CUKR_SSPMV_BCSR4_5X5_THREADS>>> (nb, d_y);
   	   		else if(r == 6) kernel_sspmv_bcsr4_6x6_a1_b0 <<<CUKR_SSPMV_BCSR4_6X6_CTAS, CUKR_SSPMV_BCSR4_6X6_THREADS>>> (nb, d_y);
	 	 	else printf("Block size (%d, %d) is NOT supported at cuda_sspmv_bcsr4", r, c);
   	         	CUT_CHECK_ERROR("kernel_spmv_bcsr4_a1_b0() execution failed\n");
	 	 }
   	    	/* If alpha != 1, have to consider it */
   	    	else {
   	   		if     (r == 2) kernel_sspmv_bcsr4_2x2_b0 <<<CUKR_SSPMV_BCSR4_2X2_CTAS, CUKR_SSPMV_BCSR4_2X2_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 3) kernel_sspmv_bcsr4_3x3_b0 <<<CUKR_SSPMV_BCSR4_3X3_CTAS, CUKR_SSPMV_BCSR4_3X3_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 4) kernel_sspmv_bcsr4_4x4_b0 <<<CUKR_SSPMV_BCSR4_4X4_CTAS, CUKR_SSPMV_BCSR4_4X4_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 5) kernel_sspmv_bcsr4_5x5_b0 <<<CUKR_SSPMV_BCSR4_5X5_CTAS, CUKR_SSPMV_BCSR4_5X5_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 6) kernel_sspmv_bcsr4_6x6_b0 <<<CUKR_SSPMV_BCSR4_6X6_CTAS, CUKR_SSPMV_BCSR4_6X6_THREADS>>> (nb, alpha, d_y);
   	    		else printf("Block size (%d, %d) is NOT supported at cuda_sspmv_bcsr4", r, c);
   	        	CUT_CHECK_ERROR("kernel_spmv_bcsr4_b0() execution failed\n");
   	    }

	 }


   	/* Un-bind textures */
   	cudaUnbindTexture(TEX_PTR);
   	cudaUnbindTexture(TEX_IDX);
   	cudaUnbindTexture(TEX_VAL4);
   	cudaUnbindTexture(TEX_X);


   	/* Do a syncronize, for timing */
    	CUDA_SAFE_CALL(cudaThreadSynchronize());

}

// ------------------------------------------------------------------- // 

/**
 * \brief Wrapper for SSPMV_HYB 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions 
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector 
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
#ifdef __cplusplus
extern "C"
#endif   
void cuda_sspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz, 
                    CUKR_INT* d_ell_idx, CUKR_SINGLE* d_ell_val, CUKR_INT* d_csr_ptr, CUKR_INT* d_csr_idx, CUKR_SINGLE* d_csr_val,  
                    CUKR_SINGLE* d_x, CUKR_SINGLE beta, CUKR_SINGLE* d_y)
{
   	/* Bind texture */
   	cudaBindTexture(0, TEX_X, d_x, cols * sizeof(d_x[0]));		
	
	const unsigned int BLOCK_SIZE_ELL = 256;
	unsigned int num_blocks = DIVIDE_INTO(rows, BLOCK_SIZE_ELL);
    const dim3 grid = dim3(num_blocks);
	
	/* Do the SPMV in ELL */
	kernel_sspmv_ell <<<grid, BLOCK_SIZE_ELL, 0>>>(rows, alpha, ell_nz_row, ell_stride, d_ell_idx, d_ell_val, beta, d_y);
   	CUT_CHECK_ERROR("kernel_sspmv_ell() execution failed\n");
   	
   	if (csr_nz) {
	   	
		/* Do the rest in CSR4 */
		cudaBindTexture(0, TEX_PTR  ,d_csr_ptr, (rows+1) * sizeof(d_csr_ptr[0]));
	   	cudaBindTexture(0, TEX_IDX4 ,d_csr_idx, csr_nz   * sizeof(d_csr_idx[0]));
	   	cudaBindTexture(0, TEX_VAL4 ,d_csr_val, csr_nz   * sizeof(d_csr_val[0])); 
	                
		if (alpha == 1) {
	   		kernel_sspmv_csr4_a1 <<<CUKR_SSPMV_CSR4_CTAS, CUKR_SSPMV_CSR4_THREADS, 0>>>(rows, d_y);
	        CUT_CHECK_ERROR("kernel_spmv_csr4_a1() execution failed\n");
		} 
	 	else {
	 		kernel_sspmv_csr4 <<<CUKR_SSPMV_CSR4_CTAS, CUKR_SSPMV_CSR4_THREADS, 0>>>(rows, alpha, d_y);
	        CUT_CHECK_ERROR("kernel_spmv_csr4() execution failed\n");    	             	           
	 	}
		 	  
	    /* Unbind textures */ 
	   	cudaUnbindTexture(TEX_PTR);
	   	cudaUnbindTexture(TEX_IDX4);
	   	cudaUnbindTexture(TEX_VAL4);
   	}
   	
   	cudaUnbindTexture(TEX_X);
   	
   	CUDA_SAFE_CALL(cudaThreadSynchronize());	  	
}	


// ------------------------------------------------------------------- // 

/**
 * \brief Wrapper for SSPMV_BHYB 
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions 
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector 
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
#ifdef __cplusplus
extern "C"
#endif   
void cuda_sspmv_bhyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz, 
                    CUKR_INT* d_bell_idx, CUKR_SINGLE* d_bell_val, CUKR_INT* d_bcsr_ptr, CUKR_INT* d_bcsr_idx, CUKR_SINGLE* d_bcsr_val,  
                    CUKR_SINGLE* d_x, CUKR_SINGLE beta, CUKR_SINGLE* d_y)
{
   	/* num of blocks per row or per col */
	int nb = (rows%r ? rows/r+1 : rows/r);

	 /* If beta != 1, do a scaling first */
	 if (beta != 1)
	 	cublasSscal(rows, beta, d_y, 1);
   	
   	/* Bind texture */
   	cudaBindTexture(0, TEX_X, d_x, cols * sizeof(d_x[0]));
	cudaBindTexture(0, TEX_IDX , d_bell_idx, bell_nz_row*bell_stride     * sizeof(d_bell_idx[0]));
	cudaBindTexture(0, TEX_VAL , d_bell_val, bell_nz_row*bell_stride*r*c * sizeof(d_bell_val[0]));

	/* Do the SPMV in BELL */
	if(r==2)
		kernel_sspmv_bell_2x2 <<<CUKR_SSPMV_BELL_2X2_CTAS, CUKR_SSPMV_BELL_2X2_THREADS>>>(nb, alpha, bell_nz_row, bell_stride, d_y);
	else if(r==3)
		kernel_sspmv_bell_3x3 <<<CUKR_SSPMV_BELL_3X3_CTAS, CUKR_SSPMV_BELL_3X3_THREADS>>>(nb, alpha, bell_nz_row, bell_stride, d_y);
	else if(r==4)
		kernel_sspmv_bell_4x4 <<<CUKR_SSPMV_BELL_4X4_CTAS, CUKR_SSPMV_BELL_4X4_THREADS>>>(nb, alpha, bell_nz_row, bell_stride, d_y);
	else if(r==5)
		kernel_sspmv_bell_5x5 <<<CUKR_SSPMV_BELL_5X5_CTAS, CUKR_SSPMV_BELL_5X5_THREADS>>>(nb, alpha, bell_nz_row, bell_stride, d_y);
	else if(r==6)
		kernel_sspmv_bell_6x6 <<<CUKR_SSPMV_BELL_6X6_CTAS, CUKR_SSPMV_BELL_6X6_THREADS>>>(nb, alpha, bell_nz_row, bell_stride, d_y);
	else
		printf("Block size (%d, %d) is NOT supported at cuda_sspmv_bhyb", r, c);
   	CUT_CHECK_ERROR("kernel_sspmv_bell() execution failed\n");
   	
   	/* Unbind textures */
	cudaUnbindTexture(TEX_IDX);
	cudaUnbindTexture(TEX_VAL);
   	
   	if (bcsr_nz) {
	   	
		/* Do the rest in BCSR */
	  	cudaBindTexture(0, TEX_PTR , d_bcsr_ptr, (nb+1)      * sizeof(d_bcsr_ptr[0]));
	   	cudaBindTexture(0, TEX_IDX , d_bcsr_idx, bcsr_nz     * sizeof(d_bcsr_idx[0]));
	   	cudaBindTexture(0, TEX_VAL , d_bcsr_val, bcsr_nz*r*c * sizeof(d_bcsr_val[0]));
		
		if (alpha == 1) {
   	   		if     (r == 2) kernel_sspmv_bcsr_2x2_a1 <<<CUKR_SSPMV_BCSR_2X2_CTAS, CUKR_SSPMV_BCSR_2X2_THREADS>>> (nb, d_y);
   	   		else if(r == 3) kernel_sspmv_bcsr_3x3_a1 <<<CUKR_SSPMV_BCSR_3X3_CTAS, CUKR_SSPMV_BCSR_3X3_THREADS>>> (nb, d_y);
   	   		else if(r == 4) kernel_sspmv_bcsr_4x4_a1 <<<CUKR_SSPMV_BCSR_4X4_CTAS, CUKR_SSPMV_BCSR_4X4_THREADS>>> (nb, d_y);
   	   		else if(r == 5) kernel_sspmv_bcsr_5x5_a1 <<<CUKR_SSPMV_BCSR_5X5_CTAS, CUKR_SSPMV_BCSR_5X5_THREADS>>> (nb, d_y);
   	   		else if(r == 6) kernel_sspmv_bcsr_6x6_a1 <<<CUKR_SSPMV_BCSR_6X6_CTAS, CUKR_SSPMV_BCSR_6X6_THREADS>>> (nb, d_y);
	        	CUT_CHECK_ERROR("kernel_sspmv_bcsr_a1() execution failed\n");
		} 
	 	else {
   	   		if     (r == 2) kernel_sspmv_bcsr_2x2 <<<CUKR_SSPMV_BCSR_2X2_CTAS, CUKR_SSPMV_BCSR_2X2_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 3) kernel_sspmv_bcsr_3x3 <<<CUKR_SSPMV_BCSR_3X3_CTAS, CUKR_SSPMV_BCSR_3X3_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 4) kernel_sspmv_bcsr_4x4 <<<CUKR_SSPMV_BCSR_4X4_CTAS, CUKR_SSPMV_BCSR_4X4_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 5) kernel_sspmv_bcsr_5x5 <<<CUKR_SSPMV_BCSR_5X5_CTAS, CUKR_SSPMV_BCSR_5X5_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 6) kernel_sspmv_bcsr_6x6 <<<CUKR_SSPMV_BCSR_6X6_CTAS, CUKR_SSPMV_BCSR_6X6_THREADS>>> (nb, alpha, d_y);
	        	CUT_CHECK_ERROR("kernel_sspmv_bcsr() execution failed\n");    	             	           
	 	}
		 	  
	   	/* Unbind textures */ 
	   	cudaUnbindTexture(TEX_PTR);
	   	cudaUnbindTexture(TEX_IDX);
	   	cudaUnbindTexture(TEX_VAL);
   	}
   	
   	/* Unbind textures */
   	cudaUnbindTexture(TEX_X);
   	
   	CUDA_SAFE_CALL(cudaThreadSynchronize());	  	
}

// ------------------------------------------------------------------- // 

/**
 * \brief Wrapper for SSPMV_BHYB4
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions 
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector 
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
#ifdef __cplusplus
extern "C"
#endif   
void cuda_sspmv_bhyb4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz, 
                    CUKR_INT* d_bell_idx, CUKR_SINGLE* d_bell_val, CUKR_INT* d_bcsr_ptr, CUKR_INT* d_bcsr_idx, CUKR_SINGLE* d_bcsr_val,  
                    CUKR_SINGLE* d_x, CUKR_SINGLE beta, CUKR_SINGLE* d_y)
{
   	/* num of blocks per row or per col */
	int nb = (rows%r ? rows/r+1 : rows/r);

	 /* If beta != 1, do a scaling first */
	 if (beta != 1)
	 	cublasSscal(rows, beta, d_y, 1);

   	/* Bind texture */
   	cudaBindTexture(0, TEX_X, d_x, cols * sizeof(d_x[0]));
   	cudaBindTexture(0, TEX_IDX , d_bell_idx, bell_nz_row*bell_stride * sizeof(d_bcsr_idx[0]));
   	if((r*c)%4 == 3) cudaBindTexture(0, TEX_VAL4, d_bell_val, bell_nz_row*bell_stride*(r*c+1) * sizeof(d_bell_val[0]));
   	if((r*c)%4 == 2) cudaBindTexture(0, TEX_VAL4, d_bell_val, bell_nz_row*bell_stride*(r*c+2) * sizeof(d_bell_val[0]));
   	if((r*c)%4 == 1) cudaBindTexture(0, TEX_VAL4, d_bell_val, bell_nz_row*bell_stride*(r*c+3) * sizeof(d_bell_val[0]));
   	else             cudaBindTexture(0, TEX_VAL4, d_bell_val, bell_nz_row*bell_stride* r*c    * sizeof(d_bell_val[0]));
	
	/* Do the SPMV in BELL4 */
	if(r==2)
		kernel_sspmv_bell4_2x2 <<<CUKR_SSPMV_BELL4_2X2_CTAS, CUKR_SSPMV_BELL4_2X2_THREADS>>>(nb, alpha, bell_nz_row, bell_stride, d_y);
	else if(r==3)
		kernel_sspmv_bell4_3x3 <<<CUKR_SSPMV_BELL4_3X3_CTAS, CUKR_SSPMV_BELL4_3X3_THREADS>>>(nb, alpha, bell_nz_row, bell_stride, d_y);
	else if(r==4)
		kernel_sspmv_bell4_4x4 <<<CUKR_SSPMV_BELL4_4X4_CTAS, CUKR_SSPMV_BELL4_4X4_THREADS>>>(nb, alpha, bell_nz_row, bell_stride, d_y);
	else if(r==5)
		kernel_sspmv_bell4_5x5 <<<CUKR_SSPMV_BELL4_5X5_CTAS, CUKR_SSPMV_BELL4_5X5_THREADS>>>(nb, alpha, bell_nz_row, bell_stride, d_y);
	else if(r==6)
		kernel_sspmv_bell4_6x6 <<<CUKR_SSPMV_BELL4_6X6_CTAS, CUKR_SSPMV_BELL4_6X6_THREADS>>>(nb, alpha, bell_nz_row, bell_stride, d_y);
	else
		printf("Block size (%d, %d) is NOT supported at cuda_sspmv_bhyb", r, c);
   	CUT_CHECK_ERROR("kernel_sspmv_bell4() execution failed\n");
   	
    	/* Unbind textures */ 
	cudaUnbindTexture(TEX_IDX);
	cudaUnbindTexture(TEX_VAL4);
   	
   	if (bcsr_nz) {
	   	
		/* Do the rest in BCSR4 */
	  	cudaBindTexture(0, TEX_PTR , d_bcsr_ptr, (nb+1) * sizeof(d_bcsr_ptr[0]));
	   	cudaBindTexture(0, TEX_IDX , d_bcsr_idx, bcsr_nz * sizeof(d_bcsr_idx[0]));
	   	if((r*c)%4 == 3) cudaBindTexture(0, TEX_VAL4, d_bcsr_val, bcsr_nz*(r*c+1) * sizeof(d_bcsr_val[0]));
	   	if((r*c)%4 == 2) cudaBindTexture(0, TEX_VAL4, d_bcsr_val, bcsr_nz*(r*c+2) * sizeof(d_bcsr_val[0]));
	   	if((r*c)%4 == 1) cudaBindTexture(0, TEX_VAL4, d_bcsr_val, bcsr_nz*(r*c+3) * sizeof(d_bcsr_val[0]));
	   	else             cudaBindTexture(0, TEX_VAL4, d_bcsr_val, bcsr_nz* r*c    * sizeof(d_bcsr_val[0]));
		
		if (alpha == 1) {
   	   		if     (r == 2) kernel_sspmv_bcsr4_2x2_a1 <<<CUKR_SSPMV_BCSR4_2X2_CTAS, CUKR_SSPMV_BCSR4_2X2_THREADS>>> (nb, d_y);
   	   		else if(r == 3) kernel_sspmv_bcsr4_3x3_a1 <<<CUKR_SSPMV_BCSR4_3X3_CTAS, CUKR_SSPMV_BCSR4_3X3_THREADS>>> (nb, d_y);
   	   		else if(r == 4) kernel_sspmv_bcsr4_4x4_a1 <<<CUKR_SSPMV_BCSR4_4X4_CTAS, CUKR_SSPMV_BCSR4_4X4_THREADS>>> (nb, d_y);
   	   		else if(r == 5) kernel_sspmv_bcsr4_5x5_a1 <<<CUKR_SSPMV_BCSR4_5X5_CTAS, CUKR_SSPMV_BCSR4_5X5_THREADS>>> (nb, d_y);
   	   		else if(r == 6) kernel_sspmv_bcsr4_6x6_a1 <<<CUKR_SSPMV_BCSR4_6X6_CTAS, CUKR_SSPMV_BCSR4_6X6_THREADS>>> (nb, d_y);
	        	CUT_CHECK_ERROR("kernel_sspmv_bcsr4_a1() execution failed\n");
		} 
	 	else {
   	   		if     (r == 2) kernel_sspmv_bcsr4_2x2 <<<CUKR_SSPMV_BCSR4_2X2_CTAS, CUKR_SSPMV_BCSR4_2X2_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 3) kernel_sspmv_bcsr4_3x3 <<<CUKR_SSPMV_BCSR4_3X3_CTAS, CUKR_SSPMV_BCSR4_3X3_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 4) kernel_sspmv_bcsr4_4x4 <<<CUKR_SSPMV_BCSR4_4X4_CTAS, CUKR_SSPMV_BCSR4_4X4_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 5) kernel_sspmv_bcsr4_5x5 <<<CUKR_SSPMV_BCSR4_5X5_CTAS, CUKR_SSPMV_BCSR4_5X5_THREADS>>> (nb, alpha, d_y);
   	   		else if(r == 6) kernel_sspmv_bcsr4_6x6 <<<CUKR_SSPMV_BCSR4_6X6_CTAS, CUKR_SSPMV_BCSR4_6X6_THREADS>>> (nb, alpha, d_y);
	        	CUT_CHECK_ERROR("kernel_sspmv_bcsr4() execution failed\n");    	             	           
	 	}
	    	/* Unbind textures */ 
		cudaUnbindTexture(TEX_PTR);
		cudaUnbindTexture(TEX_IDX);
		cudaUnbindTexture(TEX_VAL4);
   	}
   	
   	cudaUnbindTexture(TEX_X);
   	
   	CUDA_SAFE_CALL(cudaThreadSynchronize());	  	
}


