/**
 * @file kernel_qaypx.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines a double-single (emulated double)
 * AYPX kernel.
 * 
 * AYPX: Y := X + aY 
 * 
 * With 192 threads and 256 blocks does up to 5GB/s = 5.5GFlops 
 * This is about half of the single precision one, so it works at maximum 
 * performance, so there is no room for improvement here.
 * 
 * For around than 16,000 elements, it has the same speed as the single
 * precision variant.
 * 
 */
 
#include "../kernels_config.h"
#include "ds_ops.h"

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x 
	#define bx          blockIdx.x
#endif
#define XH_OFFSET_QAYPX    0
#define XT_OFFSET_QAYPX        CUKR_QAYPX_THREADS
#define YH_OFFSET_QAYPX    2 * CUKR_QAYPX_THREADS
#define YT_OFFSET_QAYPX    3 * CUKR_QAYPX_THREADS

extern __shared__ float shared[];

__global__ void kernel_qaypx(const float *d_xh, const float *d_xt, 
						    	   float *d_yh,       float *d_yt, 
							 const float  alpha0,     float alpha1, 
                             const int    n)
{	
	int i;
	float c0, c1;
				
	/* Alloc shared memory */
	float  *buffer_xh = &shared[XH_OFFSET_QAYPX]; 
	float  *buffer_xt = &shared[XT_OFFSET_QAYPX]; 
	float  *buffer_yh = &shared[YH_OFFSET_QAYPX]; 
	float  *buffer_yt = &shared[YT_OFFSET_QAYPX]; 
	
	/* Starting point for this block */
	int ctaStart = bx * CUKR_QAYPX_THREADS; 
	/* Total no. of threads in the kernel */
	int totalThreads = CUKR_QAYPX_CTAS * CUKR_QAYPX_THREADS;
	
	/* Read the data (full lines) */
	for (i = ctaStart + tx; i < n; i += totalThreads)
	{  
		/* Prefetch x and y */
		buffer_xh[tx] = d_xh[i];
		buffer_xt[tx] = d_xt[i];
		buffer_yh[tx] = d_yh[i];
		buffer_yt[tx] = d_yt[i]; 
		
		/* 
		 * Multiply and write
		 */
		
		/* c = alpha*y */
		dsmul(c0, c1, alpha0, alpha1, buffer_yh[tx], buffer_yt[tx]);
		/* y = x + c */
		dsadd(d_yh[i], d_yt[i], buffer_xh[tx], buffer_xt[tx], c0, c1);   	 
	}  
	 
}

