/**
 * @file kernel_qaxpy.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines a double-single (emulated double)
 * DOT kernel.
 * 
 * DOT: d = sum(XiYi), i=1,n 
 * 
 * With 64 threads and 256 blocks does up to 45GB/s = 6GFlops 
 * This is about 1/3 of the single precision one, so there would be room for 
 * improvement, if there there were more registers.
 * 
 * However, for less than 100,000 elements, the speed is better than
 * half the one of the single precision dot product.
 * 
 */
 
#include "../kernels_config.h"
#include "ds_ops.h"

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x 
	#define bx          blockIdx.x
#endif

#define PSH_OFFSET_QDOT 0
#define PST_OFFSET_QDOT CUKR_QDOT_THREADS 

extern __shared__ float shared[];

__global__ void kernel_qdot(const float *d_xh, const float *d_xt, 
						    const float *d_yh, const float *d_yt, 
							      float *d_sh,       float *d_st, 
                            int n)
{	
/*	
	if (tx == 0 && bx == 0)
	{
		volatile double a = 1.12345678876543, b = 1.231242354324324;
		volatile float a0 = (float)a; volatile float a1 = (float)(a - a0);
		volatile float b0 = (float)b; volatile float b1 = (float)(b - b0);
		
		volatile double cc = a + b;
		volatile float cc0, cc1;
		dsadd(cc0, cc1, a0, a1, b0, b1);
		
		double err = cc - (double)cc0 - (double)cc1;
		
		printf("a + b = %g, a = %g, b = %g\n", cc, a, b);
		printf("cc = %g, c0 = %g, c1 = %g\n", cc, cc0, cc1);
		
		printf("a0 = %g,  a1 = %g,  b0 = %g,  b1 = %g\n", a0, a1, b0, b1);
		printf("err = %g\n", err);
	}
*/	
	
	float sum0 = 0, sum1 = 0, c0, c1;
				
	/* Alloc shared memory */
	float *partial_sum_h = &shared[PSH_OFFSET_QDOT]; 
	float *partial_sum_t = &shared[PST_OFFSET_QDOT];  
	
	/* Starting point for this block */
	int ctaStart = bx * CUKR_QDOT_THREADS; 
	/* Total no. of threads in the kernel */
	int totalThreads = CUKR_QDOT_CTAS * CUKR_QDOT_THREADS;
	
	/* Read the sum data */
	for (int i = ctaStart + tx; i < n; i += totalThreads)
	{ 
		/* Multiply */
		dsmul(c0, c1, d_xh[i], d_xt[i], d_yh[i], d_yt[i]);
		/* Accumulate */
		dsadd(sum0, sum1, c0, c1, sum0, sum1);
	}  
	partial_sum_h[tx] = sum0;
	partial_sum_t[tx] = sum1;
	
	/* Reduce data for the block */
	for (int i = CUKR_QDOT_THREADS >> 1; i > 0; i >>= 1)
	{
		__syncthreads(); 
		if (tx < i) {
            dsadd(partial_sum_h[tx],     partial_sum_t[tx],
                  partial_sum_h[tx],     partial_sum_t[tx], 
                  partial_sum_h[tx + i], partial_sum_t[tx + i]); 
        }
	} 
	if (tx == 0) {
       	 	d_sh[bx] = partial_sum_h[tx];
       	 	d_st[bx] = partial_sum_t[tx];
    }
	
}
