/**
 * @file kernel_qscal.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines a double-single (emulated double)
 * SCAL kernel.
 * 
 * SCAL: X := aX
 * 
 * For 256 blocks and 128 threads per block does around 62.5BG/s, which
 * means around 4 GFlops. This is exact half of the single precision
 * version, so there is no room for improvement.
 * 
 */
 
#include "../kernels_config.h"
#include "ds_ops.h"

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x 
	#define bx          blockIdx.x
#endif
#define XH_OFFSET_QSCAL    0
#define XT_OFFSET_QSCAL    CUKR_QSCAL_THREADS

extern __shared__ float shared[];

__global__ void kernel_qscal(float *d_xh, float *d_xt, 
								const float alpha0, const float alpha1, 
								const int n)
{				
	/* Alloc shared memory */
	float  *buffer_xh = &shared[XH_OFFSET_QSCAL]; 
	float  *buffer_xt = &shared[XT_OFFSET_QSCAL]; 
	
	/* Starting point for this block */
	int ctaStart = bx * CUKR_QSCAL_THREADS; 
	/* Total no. of threads in the kernel */
	int totalThreads = CUKR_QSCAL_CTAS * CUKR_QSCAL_THREADS;
	
	/* Read the data (full lines) */
	for (int i = ctaStart + tx; i < n; i += totalThreads)
	{  
		/* Prefetch x and y */
		buffer_xh[tx] = d_xh[i];
		buffer_xt[tx] = d_xt[i]; 
		
		/* Scale and write */
		dsmul(d_xh[i], d_xt[i], buffer_xh[tx], buffer_xt[tx], alpha0, alpha1); 	 
	}  	 
}
