/**
 * src/blas/impl/gpu/cuda/qdouble/kernel_qspmv_bcsr4_a1_b0.h
 * Yusa programs following "kernel_qspmv_bcsr4_a1_b0.h" in winter, 2009.
 * SPMV in CUDA kernel.
 * SPMV is y=aAx+by, a=1, b=0.
 */

#include "../kernels_textures.h"

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x
	#define bx          blockIdx.x
#endif


/**
 * Calculate SPMV when block size is 2x2
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_qspmv_bcsr4_2x2_a1_b0(const int n, float *d_yh, float* d_yt)
{

	int i, j;
	float ch, ct;

	/* Read the data*/
	for (i = bx * CUKR_QSPMV_BCSR4_2X2_THREADS + tx; i < n; i+=CUKR_QSPMV_BCSR4_2X2_CTAS * CUKR_QSPMV_BCSR4_2X2_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0h=0.0, y0t=0.0,
		      y1h=0.0, y1t=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) - 1);
			float x0t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) - 1);
			float x1h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j));
			float x1t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j));
			
			/* float4 val */
			float4 val0h = tex1Dfetch(TEX_VAL4,   j);
			float4 val0t = tex1Dfetch(TEX_VAL4_T, j);
			
			/* GEMV, Block by Vector */
			dsmul(ch, ct, val0h.x, val0t.x, x0h, x0t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.y, val0t.y, x1h, x1t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.z, val0t.z, x0h, x0t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val0h.w, val0t.w, x1h, x1t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
		}
		
		/* copy register to memory */
		d_yh[i*2  ] = y0h; d_yt[i*2  ] = y0t;
		d_yh[i*2+1] = y1h; d_yt[i*2+1] = y1t;
	}
}

/**
 * Calculate SPMV when block size is 3x3
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_qspmv_bcsr4_3x3_a1_b0(const int n, float *d_yh, float* d_yt)
{

	int i, j;
	float ch, ct;

	/* Read the data*/
	for (i = bx * CUKR_QSPMV_BCSR4_3X3_THREADS + tx; i < n; i+=CUKR_QSPMV_BCSR4_3X3_CTAS * CUKR_QSPMV_BCSR4_3X3_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0h=0.0, y0t=0.0,
		      y1h=0.0, y1t=0.0,
		      y2h=0.0, y2t=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) - 1);
			float x0t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) - 1);
			float x1h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j));
			float x1t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j));
			float x2h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 1);
			float x2t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 1);
			
			/* float4 val */
			float4 val0h = tex1Dfetch(TEX_VAL4,   j*3);
			float4 val0t = tex1Dfetch(TEX_VAL4_T, j*3);
			float4 val1h = tex1Dfetch(TEX_VAL4,   j*3+1);
			float4 val1t = tex1Dfetch(TEX_VAL4_T, j*3+1);
			float4 val2h = tex1Dfetch(TEX_VAL4,   j*3+2);
			float4 val2t = tex1Dfetch(TEX_VAL4_T, j*3+2);
			
			/* GEMV, Block by Vector */
			dsmul(ch, ct, val0h.x, val0t.x, x0h, x0t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.y, val0t.y, x1h, x1t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.z, val0t.z, x2h, x2t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.w, val0t.w, x0h, x0t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val1h.x, val1t.x, x1h, x1t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val1h.y, val1t.y, x2h, x2t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val1h.z, val1t.z, x0h, x0t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val1h.w, val1t.w, x1h, x1t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val2h.x, val2t.x, x2h, x2t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
		}
		
		/* copy register to memory */
		d_yh[i*3  ] = y0h; d_yt[i*3  ] = y0t;
		d_yh[i*3+1] = y1h; d_yt[i*3+1] = y1t;
		d_yh[i*3+2] = y2h; d_yt[i*3+2] = y2t;
	}
}

/**
 * Calculate SPMV when block size is 4x4
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_qspmv_bcsr4_4x4_a1_b0(const int n, float *d_yh, float* d_yt)
{

	int i, j;
	float ch, ct;

	/* Read the data*/
	for (i = bx * CUKR_QSPMV_BCSR4_4X4_THREADS + tx; i < n; i+=CUKR_QSPMV_BCSR4_4X4_CTAS * CUKR_QSPMV_BCSR4_4X4_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0h=0.0, y0t=0.0,
		      y1h=0.0, y1t=0.0,
		      y2h=0.0, y2t=0.0,
		      y3h=0.0, y3t=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) - 1);
			float x0t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) - 1);
			float x1h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j));
			float x1t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j));
			float x2h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 1);
			float x2t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 1);
			float x3h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 2);
			float x3t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 2);
			
			/* float4 val */
			float4 val0h = tex1Dfetch(TEX_VAL4,   j*4);
			float4 val0t = tex1Dfetch(TEX_VAL4_T, j*4);
			float4 val1h = tex1Dfetch(TEX_VAL4,   j*4+1);
			float4 val1t = tex1Dfetch(TEX_VAL4_T, j*4+1);
			float4 val2h = tex1Dfetch(TEX_VAL4,   j*4+2);
			float4 val2t = tex1Dfetch(TEX_VAL4_T, j*4+2);
			float4 val3h = tex1Dfetch(TEX_VAL4,   j*4+3);
			float4 val3t = tex1Dfetch(TEX_VAL4_T, j*4+3);
			
			/* GEMV, Block by Vector */
			dsmul(ch, ct, val0h.x, val0t.x, x0h, x0t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.y, val0t.y, x1h, x1t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.z, val0t.z, x2h, x2t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.w, val0t.w, x3h, x3t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val1h.x, val1t.x, x0h, x0t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val1h.y, val1t.y, x1h, x1t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val1h.z, val1t.z, x2h, x2t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val1h.w, val1t.w, x3h, x3t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val2h.x, val2t.x, x0h, x0t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val2h.y, val2t.y, x1h, x1t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val2h.z, val2t.z, x2h, x2t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val2h.w, val2t.w, x3h, x3t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val3h.x, val3t.x, x0h, x0t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val3h.y, val3t.y, x1h, x1t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val3h.z, val3t.z, x2h, x2t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val3h.w, val3t.w, x3h, x3t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
		}
		
		/* copy register to memory */
		d_yh[i*4  ] = y0h; d_yt[i*4  ] = y0t;
		d_yh[i*4+1] = y1h; d_yt[i*4+1] = y1t;
		d_yh[i*4+2] = y2h; d_yt[i*4+2] = y2t;
		d_yh[i*4+3] = y3h; d_yt[i*4+3] = y3t;
	}
}

/**
 * Calculate SPMV when block size is 5x5
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_qspmv_bcsr4_5x5_a1_b0(const int n, float *d_yh, float* d_yt)
{

	int i, j;
	float ch, ct;

	/* Read the data*/
	for (i = bx * CUKR_QSPMV_BCSR4_5X5_THREADS + tx; i < n; i+=CUKR_QSPMV_BCSR4_5X5_CTAS * CUKR_QSPMV_BCSR4_5X5_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0h=0.0, y0t=0.0,
		      y1h=0.0, y1t=0.0,
		      y2h=0.0, y2t=0.0,
		      y3h=0.0, y3t=0.0,
		      y4h=0.0, y4t=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) - 1);
			float x0t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) - 1);
			float x1h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j));
			float x1t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j));
			float x2h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 1);
			float x2t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 1);
			float x3h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 2);
			float x3t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 2);
			float x4h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 3);
			float x4t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 3);
			
			/* float4 val */
			float4 val0h = tex1Dfetch(TEX_VAL4,   j*7);
			float4 val0t = tex1Dfetch(TEX_VAL4_T, j*7);
			float4 val1h = tex1Dfetch(TEX_VAL4,   j*7+1);
			float4 val1t = tex1Dfetch(TEX_VAL4_T, j*7+1);
			float4 val2h = tex1Dfetch(TEX_VAL4,   j*7+2);
			float4 val2t = tex1Dfetch(TEX_VAL4_T, j*7+2);
			float4 val3h = tex1Dfetch(TEX_VAL4,   j*7+3);
			float4 val3t = tex1Dfetch(TEX_VAL4_T, j*7+3);
			float4 val4h = tex1Dfetch(TEX_VAL4,   j*7+4);
			float4 val4t = tex1Dfetch(TEX_VAL4_T, j*7+4);
			float4 val5h = tex1Dfetch(TEX_VAL4,   j*7+5);
			float4 val5t = tex1Dfetch(TEX_VAL4_T, j*7+5);
			float4 val6h = tex1Dfetch(TEX_VAL4,   j*7+6);
			float4 val6t = tex1Dfetch(TEX_VAL4_T, j*7+6);
			
			/* GEMV, Block by Vector */
			dsmul(ch, ct, val0h.x, val0t.x, x0h, x0t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.y, val0t.y, x1h, x1t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.z, val0t.z, x2h, x2t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.w, val0t.w, x3h, x3t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val1h.x, val1t.x, x4h, x4t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val1h.y, val1t.y, x0h, x0t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val1h.z, val1t.z, x1h, x1t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val1h.w, val1t.w, x2h, x2t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val2h.x, val2t.x, x3h, x3t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val2h.y, val2t.y, x4h, x4t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val2h.z, val2t.z, x0h, x0t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val2h.w, val2t.w, x1h, x1t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val3h.x, val3t.x, x2h, x2t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val3h.y, val3t.y, x3h, x3t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val3h.z, val3t.z, x4h, x4t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val3h.w, val3t.w, x0h, x0t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val4h.x, val4t.x, x1h, x1t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val4h.y, val4t.y, x2h, x2t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val4h.z, val4t.z, x3h, x3t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val4h.w, val4t.w, x4h, x4t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val5h.x, val5t.x, x0h, x0t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val5h.y, val5t.y, x1h, x1t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val5h.z, val5t.z, x2h, x2t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val5h.w, val5t.w, x3h, x3t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val6h.x, val6t.x, x4h, x4t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
		}
		
		/* copy register to memory */
		d_yh[i*5  ] = y0h; d_yt[i*5  ] = y0t;
		d_yh[i*5+1] = y1h; d_yt[i*5+1] = y1t;
		d_yh[i*5+2] = y2h; d_yt[i*5+2] = y2t;
		d_yh[i*5+3] = y3h; d_yt[i*5+3] = y3t;
		d_yh[i*5+4] = y4h; d_yt[i*5+4] = y4t;
	}
}

/**
 * Calculate SPMV when block size is 6x6
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_qspmv_bcsr4_6x6_a1_b0(const int n, float *d_yh, float* d_yt)
{

	int i, j;
	float ch, ct;

	/* Read the data*/
	for (i = bx * CUKR_QSPMV_BCSR4_6X6_THREADS + tx; i < n; i+=CUKR_QSPMV_BCSR4_6X6_CTAS * CUKR_QSPMV_BCSR4_6X6_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0h=0.0, y0t=0.0,
		      y1h=0.0, y1t=0.0,
		      y2h=0.0, y2t=0.0,
		      y3h=0.0, y3t=0.0,
		      y4h=0.0, y4t=0.0,
		      y5h=0.0, y5t=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) - 1);
			float x0t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) - 1);
			float x1h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j));
			float x1t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j));
			float x2h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 1);
			float x2t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 1);
			float x3h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 2);
			float x3t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 2);
			float x4h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 3);
			float x4t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 3);
			float x5h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 4);
			float x5t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 4);
			
			/* float4 val */
			float4 val0h = tex1Dfetch(TEX_VAL4,   j*9);
			float4 val0t = tex1Dfetch(TEX_VAL4_T, j*9);
			float4 val1h = tex1Dfetch(TEX_VAL4,   j*9+1);
			float4 val1t = tex1Dfetch(TEX_VAL4_T, j*9+1);
			float4 val2h = tex1Dfetch(TEX_VAL4,   j*9+2);
			float4 val2t = tex1Dfetch(TEX_VAL4_T, j*9+2);
			float4 val3h = tex1Dfetch(TEX_VAL4,   j*9+3);
			float4 val3t = tex1Dfetch(TEX_VAL4_T, j*9+3);
			float4 val4h = tex1Dfetch(TEX_VAL4,   j*9+4);
			float4 val4t = tex1Dfetch(TEX_VAL4_T, j*9+4);
			float4 val5h = tex1Dfetch(TEX_VAL4,   j*9+5);
			float4 val5t = tex1Dfetch(TEX_VAL4_T, j*9+5);
			float4 val6h = tex1Dfetch(TEX_VAL4,   j*9+6);
			float4 val6t = tex1Dfetch(TEX_VAL4_T, j*9+6);
			float4 val7h = tex1Dfetch(TEX_VAL4,   j*9+7);
			float4 val7t = tex1Dfetch(TEX_VAL4_T, j*9+7);
			float4 val8h = tex1Dfetch(TEX_VAL4,   j*9+8);
			float4 val8t = tex1Dfetch(TEX_VAL4_T, j*9+8);
			
			/* GEMV, Block by Vector */
			dsmul(ch, ct, val0h.x, val0t.x, x0h, x0t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.y, val0t.y, x1h, x1t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.z, val0t.z, x2h, x2t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.w, val0t.w, x3h, x3t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val1h.x, val1t.x, x4h, x4t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val1h.y, val1t.y, x5h, x5t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val1h.z, val1t.z, x0h, x0t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val1h.w, val1t.w, x1h, x1t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val2h.x, val2t.x, x2h, x2t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val2h.y, val2t.y, x3h, x3t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val2h.z, val2t.z, x4h, x4t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val2h.w, val2t.w, x5h, x5t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val3h.x, val3t.x, x0h, x0t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val3h.y, val3t.y, x1h, x1t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val3h.z, val3t.z, x2h, x2t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val3h.w, val3t.w, x3h, x3t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val4h.x, val4t.x, x4h, x4t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val4h.y, val4t.y, x5h, x5t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val4h.z, val4t.z, x0h, x0t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val4h.w, val4t.w, x1h, x1t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val5h.x, val5t.x, x2h, x2t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val5h.y, val5t.y, x3h, x3t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val5h.z, val5t.z, x4h, x4t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val5h.w, val5t.w, x5h, x5t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val6h.x, val6t.x, x0h, x0t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val6h.y, val6t.y, x1h, x1t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val6h.z, val6t.z, x2h, x2t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val6h.w, val6t.w, x3h, x3t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val7h.x, val7t.x, x4h, x4t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val7h.y, val7t.y, x5h, x5t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val7h.z, val7t.z, x0h, x0t); dsadd(y5h, y5t, ch, ct, y5h, y5t);
			dsmul(ch, ct, val7h.w, val7t.w, x1h, x1t); dsadd(y5h, y5t, ch, ct, y5h, y5t);
			dsmul(ch, ct, val8h.x, val8t.x, x2h, x2t); dsadd(y5h, y5t, ch, ct, y5h, y5t);
			dsmul(ch, ct, val8h.y, val8t.y, x3h, x3t); dsadd(y5h, y5t, ch, ct, y5h, y5t);
			dsmul(ch, ct, val8h.z, val8t.z, x4h, x4t); dsadd(y5h, y5t, ch, ct, y5h, y5t);
			dsmul(ch, ct, val8h.w, val8t.w, x5h, x5t); dsadd(y5h, y5t, ch, ct, y5h, y5t);
		}
		
		/* copy register to memory */
		d_yh[i*6  ] = y0h; d_yt[i*6  ] = y0t;
		d_yh[i*6+1] = y1h; d_yt[i*6+1] = y1t;
		d_yh[i*6+2] = y2h; d_yt[i*6+2] = y2t;
		d_yh[i*6+3] = y3h; d_yt[i*6+3] = y3t;
		d_yh[i*6+4] = y4h; d_yt[i*6+4] = y4t;
		d_yh[i*6+5] = y5h; d_yt[i*6+5] = y5t;
	}
}
