/**
 * src/blas/impl/gpu/cuda/qdouble/kernel_qspmv_bcsr_b0.h
 * Yusa programs following "kernel_qspmv_bcsr_b0.h" in winter, 2009.
 * SPMV in CUDA kernel.
 * SPMV is y=aAx+by, b=0.
 */

#include "../kernels_textures.h"

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x
	#define bx          blockIdx.x
#endif


/**
 * Calculate SPMV when block size is 2x2
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_qspmv_bcsr_2x2_b0(const int n, const float alphah, const float alphat, float *d_yh, float* d_yt)
{

	int i, j;
	float ch, ct;

	/* Read the data*/
	for (i = bx * CUKR_QSPMV_BCSR_2X2_THREADS + tx; i < n; i+=CUKR_QSPMV_BCSR_2X2_CTAS * CUKR_QSPMV_BCSR_2X2_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0h=0.0, y0t=0.0,
		      y1h=0.0, y1t=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) - 1);
			float x0t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) - 1);
			float x1h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j));
			float x1t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j));
			
			/* GEMV, Block by Vector */
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*4  ), tex1Dfetch(TEX_VAL_T, j*4  ), x0h, x0t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*4+1), tex1Dfetch(TEX_VAL_T, j*4+1), x1h, x1t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*4+2), tex1Dfetch(TEX_VAL_T, j*4+2), x0h, x0t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*4+3), tex1Dfetch(TEX_VAL_T, j*4+3), x1h, x1t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
		}
		
		/* mul alpha */
		dsmul(y0h, y0t, alphah, alphat, y0h, y0t);
		dsmul(y1h, y1t, alphah, alphat, y1h, y1t);
		
		/* copy register to memory */
		d_yh[i*2  ] = y0h; d_yt[i*2  ] = y0t;
		d_yh[i*2+1] = y1h; d_yt[i*2+1] = y1t;
	}
}

/**
 * Calculate SPMV when block size is 3x3
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_qspmv_bcsr_3x3_b0(const int n, const float alphah, const float alphat, float *d_yh, float* d_yt)
{

	int i, j;
	float ch, ct;

	/* Read the data*/
	for (i = bx * CUKR_QSPMV_BCSR_3X3_THREADS + tx; i < n; i+=CUKR_QSPMV_BCSR_3X3_CTAS * CUKR_QSPMV_BCSR_3X3_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0h=0.0, y0t=0.0,
		      y1h=0.0, y1t=0.0,
		      y2h=0.0, y2t=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) - 1);
			float x0t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) - 1);
			float x1h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j));
			float x1t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j));
			float x2h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 1);
			float x2t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 1);
			
			/* GEMV, Block by Vector */
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*9  ), tex1Dfetch(TEX_VAL_T, j*9  ), x0h, x0t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*9+1), tex1Dfetch(TEX_VAL_T, j*9+1), x1h, x1t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*9+2), tex1Dfetch(TEX_VAL_T, j*9+2), x2h, x2t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*9+3), tex1Dfetch(TEX_VAL_T, j*9+3), x0h, x0t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*9+4), tex1Dfetch(TEX_VAL_T, j*9+4), x1h, x1t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*9+5), tex1Dfetch(TEX_VAL_T, j*9+5), x2h, x2t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*9+6), tex1Dfetch(TEX_VAL_T, j*9+6), x0h, x0t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*9+7), tex1Dfetch(TEX_VAL_T, j*9+7), x1h, x1t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*9+8), tex1Dfetch(TEX_VAL_T, j*9+8), x2h, x2t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
		}
		
		/* mul alpha */
		dsmul(y0h, y0t, alphah, alphat, y0h, y0t);
		dsmul(y1h, y1t, alphah, alphat, y1h, y1t);
		dsmul(y2h, y2t, alphah, alphat, y2h, y2t);
		
		/* copy register to memory */
		d_yh[i*3  ] = y0h; d_yt[i*3  ] = y0t;
		d_yh[i*3+1] = y1h; d_yt[i*3+1] = y1t;
		d_yh[i*3+2] = y2h; d_yt[i*3+2] = y2t;
	}
}

/**
 * Calculate SPMV when block size is 4x4
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_qspmv_bcsr_4x4_b0(const int n, const float alphah, const float alphat, float *d_yh, float* d_yt)
{

	int i, j;
	float ch, ct;

	/* Read the data*/
	for (i = bx * CUKR_QSPMV_BCSR_4X4_THREADS + tx; i < n; i+=CUKR_QSPMV_BCSR_4X4_CTAS * CUKR_QSPMV_BCSR_4X4_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0h=0.0, y0t=0.0,
		      y1h=0.0, y1t=0.0,
		      y2h=0.0, y2t=0.0,
		      y3h=0.0, y3t=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) - 1);
			float x0t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) - 1);
			float x1h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j));
			float x1t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j));
			float x2h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 1);
			float x2t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 1);
			float x3h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 2);
			float x3t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 2);
			
			/* GEMV, Block by Vector */
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*16  ), tex1Dfetch(TEX_VAL_T, j*16  ), x0h, x0t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*16+1), tex1Dfetch(TEX_VAL_T, j*16+1), x1h, x1t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*16+2), tex1Dfetch(TEX_VAL_T, j*16+2), x2h, x2t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*16+3), tex1Dfetch(TEX_VAL_T, j*16+3), x3h, x3t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*16+4), tex1Dfetch(TEX_VAL_T, j*16+4), x0h, x0t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*16+5), tex1Dfetch(TEX_VAL_T, j*16+5), x1h, x1t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*16+6), tex1Dfetch(TEX_VAL_T, j*16+6), x2h, x2t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*16+7), tex1Dfetch(TEX_VAL_T, j*16+7), x3h, x3t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*16+8), tex1Dfetch(TEX_VAL_T, j*16+8), x0h, x0t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*16+9), tex1Dfetch(TEX_VAL_T, j*16+9), x1h, x1t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*16+10), tex1Dfetch(TEX_VAL_T, j*16+10), x2h, x2t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*16+11), tex1Dfetch(TEX_VAL_T, j*16+11), x3h, x3t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*16+12), tex1Dfetch(TEX_VAL_T, j*16+12), x0h, x0t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*16+13), tex1Dfetch(TEX_VAL_T, j*16+13), x1h, x1t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*16+14), tex1Dfetch(TEX_VAL_T, j*16+14), x2h, x2t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*16+15), tex1Dfetch(TEX_VAL_T, j*16+15), x3h, x3t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
		}
		
		/* mul alpha */
		dsmul(y0h, y0t, alphah, alphat, y0h, y0t);
		dsmul(y1h, y1t, alphah, alphat, y1h, y1t);
		dsmul(y2h, y2t, alphah, alphat, y2h, y2t);
		dsmul(y3h, y3t, alphah, alphat, y3h, y3t);
		
		/* copy register to memory */
		d_yh[i*4  ] = y0h; d_yt[i*4  ] = y0t;
		d_yh[i*4+1] = y1h; d_yt[i*4+1] = y1t;
		d_yh[i*4+2] = y2h; d_yt[i*4+2] = y2t;
		d_yh[i*4+3] = y3h; d_yt[i*4+3] = y3t;
	}
}

/**
 * Calculate SPMV when block size is 5x5
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_qspmv_bcsr_5x5_b0(const int n, const float alphah, const float alphat, float *d_yh, float* d_yt)
{

	int i, j;
	float ch, ct;

	/* Read the data*/
	for (i = bx * CUKR_QSPMV_BCSR_5X5_THREADS + tx; i < n; i+=CUKR_QSPMV_BCSR_5X5_CTAS * CUKR_QSPMV_BCSR_5X5_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0h=0.0, y0t=0.0,
		      y1h=0.0, y1t=0.0,
		      y2h=0.0, y2t=0.0,
		      y3h=0.0, y3t=0.0,
		      y4h=0.0, y4t=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) - 1);
			float x0t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) - 1);
			float x1h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j));
			float x1t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j));
			float x2h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 1);
			float x2t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 1);
			float x3h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 2);
			float x3t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 2);
			float x4h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 3);
			float x4t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 3);
			
			/* GEMV, Block by Vector */
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25  ), tex1Dfetch(TEX_VAL_T, j*25  ), x0h, x0t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+1), tex1Dfetch(TEX_VAL_T, j*25+1), x1h, x1t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+2), tex1Dfetch(TEX_VAL_T, j*25+2), x2h, x2t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+3), tex1Dfetch(TEX_VAL_T, j*25+3), x3h, x3t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+4), tex1Dfetch(TEX_VAL_T, j*25+4), x4h, x4t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+5), tex1Dfetch(TEX_VAL_T, j*25+5), x0h, x0t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+6), tex1Dfetch(TEX_VAL_T, j*25+6), x1h, x1t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+7), tex1Dfetch(TEX_VAL_T, j*25+7), x2h, x2t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+8), tex1Dfetch(TEX_VAL_T, j*25+8), x3h, x3t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+9), tex1Dfetch(TEX_VAL_T, j*25+9), x4h, x4t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+10), tex1Dfetch(TEX_VAL_T, j*25+10), x0h, x0t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+11), tex1Dfetch(TEX_VAL_T, j*25+11), x1h, x1t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+12), tex1Dfetch(TEX_VAL_T, j*25+12), x2h, x2t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+13), tex1Dfetch(TEX_VAL_T, j*25+13), x3h, x3t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+14), tex1Dfetch(TEX_VAL_T, j*25+14), x4h, x4t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+15), tex1Dfetch(TEX_VAL_T, j*25+15), x0h, x0t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+16), tex1Dfetch(TEX_VAL_T, j*25+16), x1h, x1t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+17), tex1Dfetch(TEX_VAL_T, j*25+17), x2h, x2t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+18), tex1Dfetch(TEX_VAL_T, j*25+18), x3h, x3t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+19), tex1Dfetch(TEX_VAL_T, j*25+19), x4h, x4t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+20), tex1Dfetch(TEX_VAL_T, j*25+20), x0h, x0t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+21), tex1Dfetch(TEX_VAL_T, j*25+21), x1h, x1t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+22), tex1Dfetch(TEX_VAL_T, j*25+22), x2h, x2t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+23), tex1Dfetch(TEX_VAL_T, j*25+23), x3h, x3t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*25+24), tex1Dfetch(TEX_VAL_T, j*25+24), x4h, x4t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
		}
		
		/* mul alpha */
		dsmul(y0h, y0t, alphah, alphat, y0h, y0t);
		dsmul(y1h, y1t, alphah, alphat, y1h, y1t);
		dsmul(y2h, y2t, alphah, alphat, y2h, y2t);
		dsmul(y3h, y3t, alphah, alphat, y3h, y3t);
		dsmul(y4h, y4t, alphah, alphat, y4h, y4t);
		
		/* copy register to memory */
		d_yh[i*5  ] = y0h; d_yt[i*5  ] = y0t;
		d_yh[i*5+1] = y1h; d_yt[i*5+1] = y1t;
		d_yh[i*5+2] = y2h; d_yt[i*5+2] = y2t;
		d_yh[i*5+3] = y3h; d_yt[i*5+3] = y3t;
		d_yh[i*5+4] = y4h; d_yt[i*5+4] = y4t;
	}
}

/**
 * Calculate SPMV when block size is 6x6
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_qspmv_bcsr_6x6_b0(const int n, const float alphah, const float alphat, float *d_yh, float* d_yt)
{

	int i, j;
	float ch, ct;

	/* Read the data*/
	for (i = bx * CUKR_QSPMV_BCSR_6X6_THREADS + tx; i < n; i+=CUKR_QSPMV_BCSR_6X6_CTAS * CUKR_QSPMV_BCSR_6X6_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0h=0.0, y0t=0.0,
		      y1h=0.0, y1t=0.0,
		      y2h=0.0, y2t=0.0,
		      y3h=0.0, y3t=0.0,
		      y4h=0.0, y4t=0.0,
		      y5h=0.0, y5t=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) - 1);
			float x0t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) - 1);
			float x1h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j));
			float x1t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j));
			float x2h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 1);
			float x2t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 1);
			float x3h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 2);
			float x3t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 2);
			float x4h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 3);
			float x4t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 3);
			float x5h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j) + 4);
			float x5t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j) + 4);
			
			/* GEMV, Block by Vector */
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36  ), tex1Dfetch(TEX_VAL_T, j*36  ), x0h, x0t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+1), tex1Dfetch(TEX_VAL_T, j*36+1), x1h, x1t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+2), tex1Dfetch(TEX_VAL_T, j*36+2), x2h, x2t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+3), tex1Dfetch(TEX_VAL_T, j*36+3), x3h, x3t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+4), tex1Dfetch(TEX_VAL_T, j*36+4), x4h, x4t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+5), tex1Dfetch(TEX_VAL_T, j*36+5), x5h, x5t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+6), tex1Dfetch(TEX_VAL_T, j*36+6), x0h, x0t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+7), tex1Dfetch(TEX_VAL_T, j*36+7), x1h, x1t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+8), tex1Dfetch(TEX_VAL_T, j*36+8), x2h, x2t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+9), tex1Dfetch(TEX_VAL_T, j*36+9), x3h, x3t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+10), tex1Dfetch(TEX_VAL_T, j*36+10), x4h, x4t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+11), tex1Dfetch(TEX_VAL_T, j*36+11), x5h, x5t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+12), tex1Dfetch(TEX_VAL_T, j*36+12), x0h, x0t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+13), tex1Dfetch(TEX_VAL_T, j*36+13), x1h, x1t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+14), tex1Dfetch(TEX_VAL_T, j*36+14), x2h, x2t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+15), tex1Dfetch(TEX_VAL_T, j*36+15), x3h, x3t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+16), tex1Dfetch(TEX_VAL_T, j*36+16), x4h, x4t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+17), tex1Dfetch(TEX_VAL_T, j*36+17), x5h, x5t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+18), tex1Dfetch(TEX_VAL_T, j*36+18), x0h, x0t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+19), tex1Dfetch(TEX_VAL_T, j*36+19), x1h, x1t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+20), tex1Dfetch(TEX_VAL_T, j*36+20), x2h, x2t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+21), tex1Dfetch(TEX_VAL_T, j*36+21), x3h, x3t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+22), tex1Dfetch(TEX_VAL_T, j*36+22), x4h, x4t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+23), tex1Dfetch(TEX_VAL_T, j*36+23), x5h, x5t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+24), tex1Dfetch(TEX_VAL_T, j*36+24), x0h, x0t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+25), tex1Dfetch(TEX_VAL_T, j*36+25), x1h, x1t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+26), tex1Dfetch(TEX_VAL_T, j*36+26), x2h, x2t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+27), tex1Dfetch(TEX_VAL_T, j*36+27), x3h, x3t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+28), tex1Dfetch(TEX_VAL_T, j*36+28), x4h, x4t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+29), tex1Dfetch(TEX_VAL_T, j*36+29), x5h, x5t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+30), tex1Dfetch(TEX_VAL_T, j*36+30), x0h, x0t); dsadd(y5h, y5t, ch, ct, y5h, y5t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+31), tex1Dfetch(TEX_VAL_T, j*36+31), x1h, x1t); dsadd(y5h, y5t, ch, ct, y5h, y5t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+32), tex1Dfetch(TEX_VAL_T, j*36+23), x2h, x2t); dsadd(y5h, y5t, ch, ct, y5h, y5t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+33), tex1Dfetch(TEX_VAL_T, j*36+33), x3h, x3t); dsadd(y5h, y5t, ch, ct, y5h, y5t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+34), tex1Dfetch(TEX_VAL_T, j*36+34), x4h, x4t); dsadd(y5h, y5t, ch, ct, y5h, y5t);
			dsmul(ch, ct, tex1Dfetch(TEX_VAL, j*36+35), tex1Dfetch(TEX_VAL_T, j*36+35), x5h, x5t); dsadd(y5h, y5t, ch, ct, y5h, y5t);
		}
		
		/* mul alpha */
		dsmul(y0h, y0t, alphah, alphat, y0h, y0t);
		dsmul(y1h, y1t, alphah, alphat, y1h, y1t);
		dsmul(y2h, y2t, alphah, alphat, y2h, y2t);
		dsmul(y3h, y3t, alphah, alphat, y3h, y3t);
		dsmul(y4h, y4t, alphah, alphat, y4h, y4t);
		dsmul(y5h, y5t, alphah, alphat, y5h, y5t);
		
		/* copy register to memory */
		d_yh[i*6  ] = y0h; d_yt[i*6  ] = y0t;
		d_yh[i*6+1] = y1h; d_yt[i*6+1] = y1t;
		d_yh[i*6+2] = y2h; d_yt[i*6+2] = y2t;
		d_yh[i*6+3] = y3h; d_yt[i*6+3] = y3t;
		d_yh[i*6+4] = y4h; d_yt[i*6+4] = y4t;
		d_yh[i*6+5] = y5h; d_yt[i*6+5] = y5t;
	}
}
