/**
 * @file blas/gpu/cuda/kernel_spmv_bell4.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * SPMV in BELL4 format
 * Follows NVIDIA's implementation.
 * 
 + Both alpha and beta are considered. 
 */
 
#include "../kernels_textures.h"
#include "ds_ops.h"

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x 
	#define bx          blockIdx.x
#endif


/**
 * Calculate SPMV when block size is 2x2
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * beta   constant number b
 * d_y    output vector y
 */ 
__global__ void kernel_qspmv_bell4_2x2(const int n, float alpha0, float alpha1, const int nz_row, const int stride, float beta0, float beta1, float *d_yh, float* d_yt)
{
	int i, j;
	float ch, ct;

	for (i = bx * CUKR_QSPMV_BELL_2X2_THREADS + tx; i < n; i += CUKR_QSPMV_BELL_2X2_CTAS * CUKR_QSPMV_BELL_2X2_THREADS){

		/* register y */
		float y0h = 0, y0t = 0,
		      y1h = 0, y1t = 0;

		for(j = 0; j < nz_row; j++){
			/* register x */
			const float x0h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i)-1);
			const float x0t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i)-1);
			const float x1h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i));
			const float x1t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i));
		
			/* flot4 val */
			float4 val0h = tex1Dfetch(TEX_VAL4,   j*stride+i);
			float4 val0t = tex1Dfetch(TEX_VAL4_T, j*stride+i);

			/* GEMV */
			dsmul(ch, ct, val0h.x, val0t.x, x0h, x0t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.y, val0t.y, x1h, x1t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.z, val0t.z, x0h, x0t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val0h.w, val0t.w, x1h, x1t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
		}

		/* mul alpha */
		dsmul(y0h, y0t, alpha0, alpha1, y0h, y0t);
		dsmul(y1h, y1t, alpha0, alpha1, y1h, y1t);
	
		/* copy register to memory */
		dsadd(d_yh[i*2  ], d_yt[i*2  ], y0h, y0t, d_yh[i*2  ], d_yt[i*2  ]);
		dsadd(d_yh[i*2+1], d_yt[i*2+1], y1h, y1t, d_yh[i*2+1], d_yt[i*2+1]);
	}
	
}

/**
 * Calculate SPMV when block size is 3x3
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * beta   constant number b
 * d_y    output vector y
 */ 
__global__ void kernel_qspmv_bell4_3x3(const int n, float alpha0, float alpha1, const int nz_row, const int stride, float beta0, float beta1, float *d_yh, float* d_yt)
{
	int i, j;
	float ch, ct;

	for (i = bx * CUKR_QSPMV_BELL_3X3_THREADS + tx; i < n; i += CUKR_QSPMV_BELL_3X3_CTAS * CUKR_QSPMV_BELL_3X3_THREADS){

		/* register y */
		float y0h = 0, y0t = 0,
		      y1h = 0, y1t = 0,
		      y2h = 0, y2t = 0;

		for(j = 0; j < nz_row; j++){
			/* register x */
			const float x0h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i)-1);
			const float x0t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i)-1);
			const float x1h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i));
			const float x1t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i));
			const float x2h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i)+1);
			const float x2t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i)+1);
		
			/* flot4 val */
			float4 val0h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*3);
			float4 val0t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*3);
			float4 val1h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*3+1);
			float4 val1t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*3+1);
			float4 val2h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*3+2);
			float4 val2t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*3+2);

			/* GEMV */
			dsmul(ch, ct, val0h.x, val0t.x, x0h, x0t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.y, val0t.y, x1h, x1t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.z, val0t.z, x2h, x2t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.w, val0t.w, x0h, x0t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val1h.x, val1t.x, x1h, x1t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val1h.y, val1t.y, x2h, x2t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val1h.z, val1t.z, x0h, x0t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val1h.w, val1t.w, x1h, x1t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val2h.x, val2t.x, x2h, x2t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
		}

		/* mul alpha */
		dsmul(y0h, y0t, alpha0, alpha1, y0h, y0t);
		dsmul(y1h, y1t, alpha0, alpha1, y1h, y1t);
		dsmul(y2h, y2t, alpha0, alpha1, y2h, y2t);
	
		/* copy register to memory */
		dsadd(d_yh[i*3  ], d_yt[i*3  ], y0h, y0t, d_yh[i*3  ], d_yt[i*3  ]);
		dsadd(d_yh[i*3+1], d_yt[i*3+1], y1h, y1t, d_yh[i*3+1], d_yt[i*3+1]);
		dsadd(d_yh[i*3+2], d_yt[i*3+2], y2h, y2t, d_yh[i*3+2], d_yt[i*3+2]);
	}
	
}

/**
 * Calculate SPMV when block size is 4x4
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * beta   constant number b
 * d_y    output vector y
 */ 
__global__ void kernel_qspmv_bell4_4x4(const int n, float alpha0, float alpha1, const int nz_row, const int stride, float beta0, float beta1, float *d_yh, float* d_yt)
{
	int i, j;
	float ch, ct;

	for (i = bx * CUKR_QSPMV_BELL_4X4_THREADS + tx; i < n; i += CUKR_QSPMV_BELL_4X4_CTAS * CUKR_QSPMV_BELL_4X4_THREADS){

		/* register y */
		float y0h = 0, y0t = 0,
		      y1h = 0, y1t = 0,
		      y2h = 0, y2t = 0,
		      y3h = 0, y3t = 0;

		for(j = 0; j < nz_row; j++){
			/* register x */
			const float x0h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i)-1);
			const float x0t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i)-1);
			const float x1h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i));
			const float x1t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i));
			const float x2h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i)+1);
			const float x2t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i)+1);
			const float x3h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i)+2);
			const float x3t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i)+2);
		
			/* flot4 val */
			float4 val0h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*4);
			float4 val0t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*4);
			float4 val1h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*4+1);
			float4 val1t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*4+1);
			float4 val2h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*4+2);
			float4 val2t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*4+2);
			float4 val3h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*4+3);
			float4 val3t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*4+3);

			/* GEMV */
			dsmul(ch, ct, val0h.x, val0t.x, x0h, x0t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.y, val0t.y, x1h, x1t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.z, val0t.z, x2h, x2t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.w, val0t.w, x3h, x3t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val1h.x, val1t.x, x0h, x0t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val1h.y, val1t.y, x1h, x1t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val1h.z, val1t.z, x2h, x2t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val1h.w, val1t.w, x3h, x3t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val2h.x, val2t.x, x0h, x0t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val2h.y, val2t.y, x1h, x1t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val2h.z, val2t.z, x2h, x2t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val2h.w, val2t.w, x3h, x3t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val3h.x, val3t.x, x0h, x0t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val3h.y, val3t.y, x1h, x1t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val3h.z, val3t.z, x2h, x2t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val3h.w, val3t.w, x3h, x3t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
		}

		/* mul alpha */
		dsmul(y0h, y0t, alpha0, alpha1, y0h, y0t);
		dsmul(y1h, y1t, alpha0, alpha1, y1h, y1t);
		dsmul(y2h, y2t, alpha0, alpha1, y2h, y2t);
		dsmul(y3h, y3t, alpha0, alpha1, y3h, y3t);
	
		/* copy register to memory */
		dsadd(d_yh[i*4  ], d_yt[i*4  ], y0h, y0t, d_yh[i*4  ], d_yt[i*4  ]);
		dsadd(d_yh[i*4+1], d_yt[i*4+1], y1h, y1t, d_yh[i*4+1], d_yt[i*4+1]);
		dsadd(d_yh[i*4+2], d_yt[i*4+2], y2h, y2t, d_yh[i*4+2], d_yt[i*4+2]);
		dsadd(d_yh[i*4+3], d_yt[i*4+3], y3h, y3t, d_yh[i*4+3], d_yt[i*4+3]);
	}
	
}

/**
 * Calculate SPMV when block size is 5x5
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * beta   constant number b
 * d_y    output vector y
 */ 
__global__ void kernel_qspmv_bell4_5x5(const int n, float alpha0, float alpha1, const int nz_row, const int stride, float beta0, float beta1, float *d_yh, float* d_yt)
{
	int i, j;
	float ch, ct;

	for (i = bx * CUKR_QSPMV_BELL_5X5_THREADS + tx; i < n; i += CUKR_QSPMV_BELL_5X5_CTAS * CUKR_QSPMV_BELL_5X5_THREADS){

		/* register y */
		float y0h = 0, y0t = 0,
		      y1h = 0, y1t = 0,
		      y2h = 0, y2t = 0,
		      y3h = 0, y3t = 0,
		      y4h = 0, y4t = 0;

		for(j = 0; j < nz_row; j++){
			/* register x */
			const float x0h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i)-1);
			const float x0t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i)-1);
			const float x1h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i));
			const float x1t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i));
			const float x2h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i)+1);
			const float x2t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i)+1);
			const float x3h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i)+2);
			const float x3t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i)+2);
			const float x4h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i)+3);
			const float x4t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i)+3);

			/* flot4 val */
			float4 val0h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*7);
			float4 val0t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*7);
			float4 val1h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*7+1);
			float4 val1t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*7+1);
			float4 val2h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*7+2);
			float4 val2t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*7+2);
			float4 val3h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*7+3);
			float4 val3t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*7+3);
			float4 val4h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*7+4);
			float4 val4t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*7+4);
			float4 val5h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*7+5);
			float4 val5t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*7+5);
			float4 val6h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*7+6);
			float4 val6t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*7+6);

			/* GEMV */
			dsmul(ch, ct, val0h.x, val0t.x, x0h, x0t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.y, val0t.y, x1h, x1t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.z, val0t.z, x2h, x2t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.w, val0t.w, x3h, x3t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val1h.x, val1t.x, x4h, x4t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val1h.y, val1t.y, x0h, x0t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val1h.z, val1t.z, x1h, x1t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val1h.w, val1t.w, x2h, x2t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val2h.x, val2t.x, x3h, x3t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val2h.y, val2t.y, x4h, x4t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val2h.z, val2t.z, x0h, x0t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val2h.w, val2t.w, x1h, x1t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val3h.x, val3t.x, x2h, x2t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val3h.y, val3t.y, x3h, x3t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val3h.z, val3t.z, x4h, x4t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val3h.w, val3t.w, x0h, x0t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val4h.x, val4t.x, x1h, x1t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val4h.y, val4t.y, x2h, x2t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val4h.z, val4t.z, x3h, x3t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val4h.w, val4t.w, x4h, x4t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val5h.x, val5t.x, x0h, x0t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val5h.y, val5t.y, x1h, x1t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val5h.z, val5t.z, x2h, x2t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val5h.w, val5t.w, x3h, x3t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val6h.x, val6t.x, x4h, x4t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
		}

		/* mul alpha */
		dsmul(y0h, y0t, alpha0, alpha1, y0h, y0t);
		dsmul(y1h, y1t, alpha0, alpha1, y1h, y1t);
		dsmul(y2h, y2t, alpha0, alpha1, y2h, y2t);
		dsmul(y3h, y3t, alpha0, alpha1, y3h, y3t);
		dsmul(y4h, y4t, alpha0, alpha1, y4h, y4t);
	
		/* copy register to memory */
		dsadd(d_yh[i*5  ], d_yt[i*5  ], y0h, y0t, d_yh[i*5  ], d_yt[i*5  ]);
		dsadd(d_yh[i*5+1], d_yt[i*5+1], y1h, y1t, d_yh[i*5+1], d_yt[i*5+1]);
		dsadd(d_yh[i*5+2], d_yt[i*5+2], y2h, y2t, d_yh[i*5+2], d_yt[i*5+2]);
		dsadd(d_yh[i*5+3], d_yt[i*5+3], y3h, y3t, d_yh[i*5+3], d_yt[i*5+3]);
		dsadd(d_yh[i*5+4], d_yt[i*5+4], y4h, y4t, d_yh[i*5+4], d_yt[i*5+4]);
	}
	
}

/**
 * Calculate SPMV when block size is 6x6
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * beta   constant number b
 * d_y    output vector y
 */ 
__global__ void kernel_qspmv_bell4_6x6(const int n, float alpha0, float alpha1, const int nz_row, const int stride, float beta0, float beta1, float *d_yh, float* d_yt)
{
	int i, j;
	float ch, ct;

	for (i = bx * CUKR_QSPMV_BELL_6X6_THREADS + tx; i < n; i += CUKR_QSPMV_BELL_6X6_CTAS * CUKR_QSPMV_BELL_6X6_THREADS){

		/* register y */
		float y0h = 0, y0t = 0,
		      y1h = 0, y1t = 0,
		      y2h = 0, y2t = 0,
		      y3h = 0, y3t = 0,
		      y4h = 0, y4t = 0,
		      y5h = 0, y5t = 0;

		for(j = 0; j < nz_row; j++){
			/* register x */
			const float x0h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i)-1);
			const float x0t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i)-1);
			const float x1h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i));
			const float x1t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i));
			const float x2h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i)+1);
			const float x2t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i)+1);
			const float x3h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i)+2);
			const float x3t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i)+2);
			const float x4h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i)+3);
			const float x4t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i)+3);
			const float x5h = tex1Dfetch(TEX_X,   tex1Dfetch(TEX_IDX, j*stride+i)+4);
			const float x5t = tex1Dfetch(TEX_X_T, tex1Dfetch(TEX_IDX, j*stride+i)+4);

			/* flot4 val */
			float4 val0h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*9);
			float4 val0t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*9);
			float4 val1h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*9+1);
			float4 val1t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*9+1);
			float4 val2h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*9+2);
			float4 val2t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*9+2);
			float4 val3h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*9+3);
			float4 val3t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*9+3);
			float4 val4h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*9+4);
			float4 val4t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*9+4);
			float4 val5h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*9+5);
			float4 val5t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*9+5);
			float4 val6h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*9+6);
			float4 val6t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*9+6);
			float4 val7h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*9+7);
			float4 val7t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*9+7);
			float4 val8h = tex1Dfetch(TEX_VAL4,   (j*stride+i)*9+8);
			float4 val8t = tex1Dfetch(TEX_VAL4_T, (j*stride+i)*9+8);

			/* GEMV */
			dsmul(ch, ct, val0h.x, val0t.x, x0h, x0t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.y, val0t.y, x1h, x1t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.z, val0t.z, x2h, x2t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val0h.w, val0t.w, x3h, x3t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val1h.x, val1t.x, x4h, x4t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val1h.y, val1t.y, x5h, x5t); dsadd(y0h, y0t, ch, ct, y0h, y0t);
			dsmul(ch, ct, val1h.z, val1t.z, x0h, x0t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val1h.w, val1t.w, x1h, x1t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val2h.x, val2t.x, x2h, x2t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val2h.y, val2t.y, x3h, x3t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val2h.z, val2t.z, x4h, x4t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val2h.w, val2t.w, x5h, x5t); dsadd(y1h, y1t, ch, ct, y1h, y1t);
			dsmul(ch, ct, val3h.x, val3t.x, x0h, x0t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val3h.y, val3t.y, x1h, x1t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val3h.z, val3t.z, x2h, x2t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val3h.w, val3t.w, x3h, x3t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val4h.x, val4t.x, x4h, x4t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val4h.y, val4t.y, x5h, x5t); dsadd(y2h, y2t, ch, ct, y2h, y2t);
			dsmul(ch, ct, val4h.z, val4t.z, x0h, x0t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val4h.w, val4t.w, x1h, x1t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val5h.x, val5t.x, x2h, x2t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val5h.y, val5t.y, x3h, x3t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val5h.z, val5t.z, x4h, x4t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val5h.w, val5t.w, x5h, x5t); dsadd(y3h, y3t, ch, ct, y3h, y3t);
			dsmul(ch, ct, val6h.x, val6t.x, x0h, x0t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val6h.y, val6t.y, x1h, x1t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val6h.z, val6t.z, x2h, x2t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val6h.w, val6t.w, x3h, x3t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val7h.x, val7t.x, x4h, x4t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val7h.y, val7t.y, x5h, x5t); dsadd(y4h, y4t, ch, ct, y4h, y4t);
			dsmul(ch, ct, val7h.z, val7t.z, x0h, x0t); dsadd(y5h, y5t, ch, ct, y5h, y5t);
			dsmul(ch, ct, val7h.w, val7t.w, x1h, x1t); dsadd(y5h, y5t, ch, ct, y5h, y5t);
			dsmul(ch, ct, val8h.x, val8t.x, x2h, x2t); dsadd(y5h, y5t, ch, ct, y5h, y5t);
			dsmul(ch, ct, val8h.y, val8t.y, x3h, x3t); dsadd(y5h, y5t, ch, ct, y5h, y5t);
			dsmul(ch, ct, val8h.z, val8t.z, x4h, x4t); dsadd(y5h, y5t, ch, ct, y5h, y5t);
			dsmul(ch, ct, val8h.w, val8t.w, x5h, x5t); dsadd(y5h, y5t, ch, ct, y5h, y5t);
		}

		/* mul alpha */
		dsmul(y0h, y0t, alpha0, alpha1, y0h, y0t);
		dsmul(y1h, y1t, alpha0, alpha1, y1h, y1t);
		dsmul(y2h, y2t, alpha0, alpha1, y2h, y2t);
		dsmul(y3h, y3t, alpha0, alpha1, y3h, y3t);
		dsmul(y4h, y4t, alpha0, alpha1, y4h, y4t);
		dsmul(y5h, y5t, alpha0, alpha1, y5h, y5t);
	
		/* copy register to memory */
		dsadd(d_yh[i*6  ], d_yt[i*6  ], y0h, y0t, d_yh[i*6  ], d_yt[i*6  ]);
		dsadd(d_yh[i*6+1], d_yt[i*6+1], y1h, y1t, d_yh[i*6+1], d_yt[i*6+1]);
		dsadd(d_yh[i*6+2], d_yt[i*6+2], y2h, y2t, d_yh[i*6+2], d_yt[i*6+2]);
		dsadd(d_yh[i*6+3], d_yt[i*6+3], y3h, y3t, d_yh[i*6+3], d_yt[i*6+3]);
		dsadd(d_yh[i*6+4], d_yt[i*6+4], y4h, y4t, d_yh[i*6+4], d_yt[i*6+4]);
		dsadd(d_yh[i*6+5], d_yt[i*6+5], y5h, y5t, d_yh[i*6+5], d_yt[i*6+5]);
	}
	
}
