/**
 * @file blas/gpu/cuda/kernel_spmv_csr4_a1.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * SpMV in quasi-double precision for 
 * CSR4 (padded CSR) format
 * 
 * Optimized version
 * Reads 4 elements at a time from
 * val and idx vectors, using float4
 * and int4 structures.
 * In order to do that, the vectors
 * have been padded so that the no.
 * of elements in each row is multiple
 * of 4.
 * 
 * It can do up to 6-7GFlop/s 
 * 
 * Simplified version: alpha = 1, so
 * the multiplication is no longer
 * necessary.
 */


#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x 
	#define bx          blockIdx.x
#endif
 
__global__ void kernel_qspmv_csr4_a1( const int    n,      	
						                    float *d_yh, float *d_yt)
{
	float c0, c1;

	/* Starting point for this block */
	int ctaStart = bx * CUKR_QSPMV_CSR4_THREADS; 
	
	/* Total no. of threads in the kernel */
	int totalThreads = CUKR_QSPMV_CSR4_CTAS * CUKR_QSPMV_CSR4_THREADS;
	
	/* Read the data*/
	for (int i = ctaStart + tx; i < n; i += totalThreads)
	{ 	
		/* Read the beginning and end of the row 
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* Read and sum for the column vectors */
		float sum0 = 0, sum1 = 0;		
		for (int j = iRowBeg / 4; j < iRowEnd / 4; j++) {
			float4 valh, valt, xh, xt;
			int4 idx;
			
			/* Read idx and val */ 
			idx  = tex1Dfetch(TEX_IDX4,   j);	
			valh = tex1Dfetch(TEX_VAL4,   j);
			valt = tex1Dfetch(TEX_VAL4_T, j);
			
			/* Idx is base 1, change to 0 */
			idx.x -= 1;
			idx.y -= 1;
			idx.z -= 1;
			idx.w -= 1;
			
			/* Read head part of x */
			xh.x = tex1Dfetch(TEX_X, idx.x);
			xh.y = tex1Dfetch(TEX_X, idx.y);
			xh.z = tex1Dfetch(TEX_X, idx.z);
			xh.w = tex1Dfetch(TEX_X, idx.w); 
			
			/* Read tail part of x */
			xt.x = tex1Dfetch(TEX_X_T, idx.x);
			xt.y = tex1Dfetch(TEX_X_T, idx.y);
			xt.z = tex1Dfetch(TEX_X_T, idx.z);
			xt.w = tex1Dfetch(TEX_X_T, idx.w); 
			
			/* Mupliply and add */
			dsmul(c0, c1, xh.x, xt.x, valh.x, valt.x);
			dsadd(sum0, sum1, c0, c1, sum0, sum1);
			dsmul(c0, c1, xh.y, xt.y, valh.y, valt.y);
			dsadd(sum0, sum1, c0, c1, sum0, sum1);
			dsmul(c0, c1, xh.z, xt.z, valh.z, valt.z);
			dsadd(sum0, sum1, c0, c1, sum0, sum1);
			dsmul(c0, c1, xh.w, xt.w, valh.w, valt.w);
			dsadd(sum0, sum1, c0, c1, sum0, sum1);
		}

		/* Write the result to memory 
		 * (update (add), not overwrite)*/
		dsadd(d_yh[i], d_yt[i], d_yh[i], d_yt[i], sum0, sum1);	 	
	}  
}
