/**
 * @file blas/gpu/cuda/kernel_qspmv_csr_b0.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Simplest version of SpMV (no data 
 * structure modification, coalescing 
 * the writes) in qdouble precision
 * 
 * Simplificated variant: beta = 0, so need
 * to consider only alpha
 */

#include "../kernels_textures.h"

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x 
	#define bx          blockIdx.x
#endif

__global__ void kernel_qspmv_csr_b0( const int  n,      
						  	         float      alpha0, float alpha1,
						             float     *d_yh,   float* d_yt)
{
	float c0, c1;
	
	/* Starting point for this block */
	int ctaStart = bx * CUKR_QSPMV_CSR_THREADS; 
	
	/* Total no. of threads in the kernel */
	int totalThreads = CUKR_QSPMV_CSR_CTAS * CUKR_QSPMV_CSR_THREADS;
	
	/* Read the data*/
	for (int i = ctaStart + tx; i < n; i += totalThreads)
	{ 	
		/* Read the beginning and end of the row 
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* Read and sum for the column vectors */
		float sum0 = 0, sum1 = 0;
		#if USE_KAHAN_IN_SPMV
		
			int col = tex1Dfetch(TEX_IDX, iRowBeg) - 1;
			dsmul(sum0, sum1, tex1Dfetch(TEX_VAL, iRowBeg), tex1Dfetch(TEX_VAL_T, iRowBeg),
				              tex1Dfetch(TEX_X, col), tex1Dfetch(TEX_X_T, col));
			float cc0 = 0.0, cc1 = 0.0;
			for (int j = iRowBeg + 1; j < iRowEnd; j++) {
				float y0, y1, t0, t1;
				col = tex1Dfetch(TEX_IDX, j) - 1;
				dsmul(c0, c1, tex1Dfetch(TEX_VAL, j), tex1Dfetch(TEX_VAL_T, j),
				              tex1Dfetch(TEX_X, col), tex1Dfetch(TEX_X_T, col));
				dssub(y0, y1, c0, c1, cc0, cc1);
				dsadd(t0, t1, sum0, sum1, y0, y1);
				dssub(c0, c1, t0, t1, sum0, sum1);
				dssub(cc0, cc1, c0, c1, y0, y1);
				sum0 = t0; sum1 = t1; 
			}
		#else				
			for (int j = iRowBeg; j < iRowEnd; j++) {
				int col = tex1Dfetch(TEX_IDX, j) - 1;
				dsmul(c0, c1, tex1Dfetch(TEX_VAL, j), tex1Dfetch(TEX_VAL_T, j),
				              tex1Dfetch(TEX_X, col), tex1Dfetch(TEX_X_T, col));
				dsadd(sum0, sum1, c0, c1, sum0, sum1);
			}
		#endif
		
		/* Multiply with alpha and write 
		 * the result to memory (overwrite) */
		dsmul(d_yh[i], d_yt[i], alpha0, alpha1, sum0, sum1);
		
	}  
}

