/**
 * src/blas/impl/gpu/cuda/single/kernel_spmv_bcsr.h
 * Yusa programs following "kernel_sspmv_bcsr.h" in winter, 2009.
 * SPMV in CUDA kernel.
 * SPMV is y=aAx+by.
 */

#include "../kernels_textures.h"

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x
	#define bx          blockIdx.x
#endif


/**
 * Calculate SPMV when block size is 2x2
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bcsr_2x2(const int n, const float alpha, float *d_y)
{

	int i, j;

	/* Read the data*/
	for (i = bx * CUKR_SSPMV_BCSR_2X2_THREADS + tx; i < n; i += CUKR_SSPMV_BCSR_2X2_CTAS * CUKR_SSPMV_BCSR_2X2_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0=0.0, y1=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) - 1);
			float x1=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j));
			
			/* GEMV, Block by Vector */
			y0 += tex1Dfetch(TEX_VAL, j*4  )*x0 + tex1Dfetch(TEX_VAL, j*4+1)*x1;
			y1 += tex1Dfetch(TEX_VAL, j*4+2)*x0 + tex1Dfetch(TEX_VAL, j*4+3)*x1;
		}
		
		/* copy register to memory */
		d_y[i*2  ] += alpha*y0;
		d_y[i*2+1] += alpha*y1;
	}
}


/**
 * Calculate SPMV when block size is 3x3
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bcsr_3x3(const int n, const float alpha, float *d_y)
{

	int i, j;

	/* Read the data*/
	for (i = bx * CUKR_SSPMV_BCSR_3X3_THREADS + tx; i < n; i += CUKR_SSPMV_BCSR_3X3_CTAS * CUKR_SSPMV_BCSR_3X3_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0=0.0, y1=0.0, y2=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) - 1);
			float x1=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j));
			float x2=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 1);
			
			/* GEMV, Block by Vector */
			y0 += tex1Dfetch(TEX_VAL, j*9  )*x0 + tex1Dfetch(TEX_VAL, j*9+1)*x1 + tex1Dfetch(TEX_VAL, j*9+2)*x2;
			y1 += tex1Dfetch(TEX_VAL, j*9+3)*x0 + tex1Dfetch(TEX_VAL, j*9+4)*x1 + tex1Dfetch(TEX_VAL, j*9+5)*x2;
			y2 += tex1Dfetch(TEX_VAL, j*9+6)*x0 + tex1Dfetch(TEX_VAL, j*9+7)*x1 + tex1Dfetch(TEX_VAL, j*9+8)*x2;
		}
		
		/* copy register to memory */
		d_y[i*3  ] += alpha*y0;
		d_y[i*3+1] += alpha*y1;
		d_y[i*3+2] += alpha*y2;
	}
}


/**
 * Calculate SPMV when block size is 4x4
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bcsr_4x4(const int n, const float alpha, float *d_y)
{

	int i, j;

	/* Read the data*/
	for (i = bx * CUKR_SSPMV_BCSR_4X4_THREADS + tx; i < n; i += CUKR_SSPMV_BCSR_4X4_CTAS * CUKR_SSPMV_BCSR_4X4_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0=0.0, y1=0.0, y2=0.0, y3=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) - 1);
			float x1=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j));
			float x2=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 1);
			float x3=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 2);
			
			/* GEMV, Block by Vector */
			y0 += tex1Dfetch(TEX_VAL, j*16   )*x0 + tex1Dfetch(TEX_VAL, j*16+ 1)*x1 + tex1Dfetch(TEX_VAL, j*16+ 2)*x2 + tex1Dfetch(TEX_VAL, j*16+ 3)*x3;
			y1 += tex1Dfetch(TEX_VAL, j*16+ 4)*x0 + tex1Dfetch(TEX_VAL, j*16+ 5)*x1 + tex1Dfetch(TEX_VAL, j*16+ 6)*x2 + tex1Dfetch(TEX_VAL, j*16+ 7)*x3;
			y2 += tex1Dfetch(TEX_VAL, j*16+ 8)*x0 + tex1Dfetch(TEX_VAL, j*16+ 9)*x1 + tex1Dfetch(TEX_VAL, j*16+10)*x2 + tex1Dfetch(TEX_VAL, j*16+11)*x3;
			y3 += tex1Dfetch(TEX_VAL, j*16+12)*x0 + tex1Dfetch(TEX_VAL, j*16+13)*x1 + tex1Dfetch(TEX_VAL, j*16+14)*x2 + tex1Dfetch(TEX_VAL, j*16+15)*x3;
		}
		
		/* copy register to memory */
		d_y[i*4  ] += alpha*y0;
		d_y[i*4+1] += alpha*y1;
		d_y[i*4+2] += alpha*y2;
		d_y[i*4+3] += alpha*y3;
	}
}

/**
 * Calculate SPMV when block size is 5x5
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bcsr_5x5(const int n, const float alpha, float *d_y)
{

	int i, j;

	/* Read the data*/
	for (i = bx * CUKR_SSPMV_BCSR_5X5_THREADS + tx; i < n; i += CUKR_SSPMV_BCSR_5X5_CTAS * CUKR_SSPMV_BCSR_5X5_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0=0.0, y1=0.0, y2=0.0, y3=0.0, y4=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) - 2);
			float x1=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j));
			float x2=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 1);
			float x3=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 2);
			float x4=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 3);
			
			/* GEMV, Block by Vector */
			y0 += tex1Dfetch(TEX_VAL, j*25   )*x0 + tex1Dfetch(TEX_VAL, j*25+ 1)*x1 + tex1Dfetch(TEX_VAL, j*25+ 2)*x2 + tex1Dfetch(TEX_VAL, j*25+ 3)*x3 + tex1Dfetch(TEX_VAL, j*25+ 4)*x4;
			y1 += tex1Dfetch(TEX_VAL, j*25+ 5)*x0 + tex1Dfetch(TEX_VAL, j*25+ 6)*x1 + tex1Dfetch(TEX_VAL, j*25+ 7)*x2 + tex1Dfetch(TEX_VAL, j*25+ 8)*x3 + tex1Dfetch(TEX_VAL, j*25+ 9)*x4;
			y2 += tex1Dfetch(TEX_VAL, j*25+10)*x0 + tex1Dfetch(TEX_VAL, j*25+11)*x1 + tex1Dfetch(TEX_VAL, j*25+12)*x2 + tex1Dfetch(TEX_VAL, j*25+13)*x3 + tex1Dfetch(TEX_VAL, j*25+14)*x4;
			y3 += tex1Dfetch(TEX_VAL, j*25+15)*x0 + tex1Dfetch(TEX_VAL, j*25+16)*x1 + tex1Dfetch(TEX_VAL, j*25+17)*x2 + tex1Dfetch(TEX_VAL, j*25+18)*x3 + tex1Dfetch(TEX_VAL, j*25+19)*x4;
			y4 += tex1Dfetch(TEX_VAL, j*25+20)*x0 + tex1Dfetch(TEX_VAL, j*25+21)*x1 + tex1Dfetch(TEX_VAL, j*25+22)*x2 + tex1Dfetch(TEX_VAL, j*25+23)*x3 + tex1Dfetch(TEX_VAL, j*25+24)*x4;
		}
		
		/* copy register to memory */
		d_y[i*5  ] += alpha*y0;
		d_y[i*5+1] += alpha*y1;
		d_y[i*5+2] += alpha*y2;
		d_y[i*5+3] += alpha*y3;
		d_y[i*5+4] += alpha*y4;
	}
}

/**
 * Calculate SPMV when block size is 6x6
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bcsr_6x6(const int n, const float alpha, float *d_y)
{

	int i, j;

	/* Read the data*/
	for (i = bx * CUKR_SSPMV_BCSR_6X6_THREADS + tx; i < n; i += CUKR_SSPMV_BCSR_6X6_CTAS * CUKR_SSPMV_BCSR_6X6_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0=0.0, y1=0.0, y2=0.0, y3=0.0, y4=0.0, y5=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) - 1);
			float x1=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j));
			float x2=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 1);
			float x3=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 2);
			float x4=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 3);
			float x5=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 4);
			
			/* GEMV, Block by Vector */
			y0 += tex1Dfetch(TEX_VAL, j*36   )*x0 + tex1Dfetch(TEX_VAL, j*36+ 1)*x1 + tex1Dfetch(TEX_VAL, j*36+ 2)*x2 + tex1Dfetch(TEX_VAL, j*36+ 3)*x3 + tex1Dfetch(TEX_VAL, j*36+ 4)*x4 + tex1Dfetch(TEX_VAL, j*36+ 5)*x5;
			y1 += tex1Dfetch(TEX_VAL, j*36+ 6)*x0 + tex1Dfetch(TEX_VAL, j*36+ 7)*x1 + tex1Dfetch(TEX_VAL, j*36+ 8)*x2 + tex1Dfetch(TEX_VAL, j*36+ 9)*x3 + tex1Dfetch(TEX_VAL, j*36+10)*x4 + tex1Dfetch(TEX_VAL, j*36+11)*x5;
			y2 += tex1Dfetch(TEX_VAL, j*36+12)*x0 + tex1Dfetch(TEX_VAL, j*36+13)*x1 + tex1Dfetch(TEX_VAL, j*36+14)*x2 + tex1Dfetch(TEX_VAL, j*36+15)*x3 + tex1Dfetch(TEX_VAL, j*36+16)*x4 + tex1Dfetch(TEX_VAL, j*36+17)*x5;
			y3 += tex1Dfetch(TEX_VAL, j*36+18)*x0 + tex1Dfetch(TEX_VAL, j*36+19)*x1 + tex1Dfetch(TEX_VAL, j*36+20)*x2 + tex1Dfetch(TEX_VAL, j*36+21)*x3 + tex1Dfetch(TEX_VAL, j*36+22)*x4 + tex1Dfetch(TEX_VAL, j*36+23)*x5;
			y4 += tex1Dfetch(TEX_VAL, j*36+24)*x0 + tex1Dfetch(TEX_VAL, j*36+25)*x1 + tex1Dfetch(TEX_VAL, j*36+26)*x2 + tex1Dfetch(TEX_VAL, j*36+27)*x3 + tex1Dfetch(TEX_VAL, j*36+28)*x4 + tex1Dfetch(TEX_VAL, j*36+29)*x5;
			y5 += tex1Dfetch(TEX_VAL, j*36+30)*x0 + tex1Dfetch(TEX_VAL, j*36+31)*x1 + tex1Dfetch(TEX_VAL, j*36+32)*x2 + tex1Dfetch(TEX_VAL, j*36+33)*x3 + tex1Dfetch(TEX_VAL, j*36+34)*x4 + tex1Dfetch(TEX_VAL, j*36+35)*x5;
		}
		
		/* copy register to memory */
		d_y[i*6  ] += alpha*y0;
		d_y[i*6+1] += alpha*y1;
		d_y[i*6+2] += alpha*y2;
		d_y[i*6+3] += alpha*y3;
		d_y[i*6+4] += alpha*y4;
		d_y[i*6+5] += alpha*y5;
	}
}
