/**
 * src/blas/impl/gpu/cuda/single/kernel_sspmv_bcsr4.h
 * SPMV in CUDA kernel.
 * SPMV is y=aAx+by.
 */

#include "../kernels_textures.h"

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x
	#define bx          blockIdx.x
#endif



/**
 * Calculate SPMV 2x2
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bcsr4_2x2(const int n, const float alpha, float *d_y)
{

	int i, j;

	/* Read the data*/
	for (i = bx * CUKR_SSPMV_BCSR4_2X2_THREADS + tx; i < n; i += CUKR_SSPMV_BCSR4_2X2_CTAS * CUKR_SSPMV_BCSR4_2X2_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0=0.0, y1=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) - 1);
			float x1=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j));
			
			/* flot4 */
			float4 val0 = tex1Dfetch(TEX_VAL4, j);
			
			/* GEMV, Block by Vector */
			y0 += val0.x * x0 + val0.y * x1;
			y1 += val0.z * x0 + val0.w * x1;
		}
		
		/* copy register to memory */
		d_y[i*2  ] += alpha*y0;
		d_y[i*2+1] += alpha*y1;
	}
}

/**
 * Calculate SPMV 3x3
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bcsr4_3x3(const int n, const float alpha, float *d_y)
{

	int i, j;

	/* Read the data*/
	for (i = bx * CUKR_SSPMV_BCSR4_3X3_THREADS + tx; i < n; i += CUKR_SSPMV_BCSR4_3X3_CTAS * CUKR_SSPMV_BCSR4_3X3_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0=0.0, y1=0.0 ,y2=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) - 1);
			float x1=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j));
			float x2=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 1);
			
			/* flot4 */
			float4 val0 = tex1Dfetch(TEX_VAL4, j*3);
			float4 val1 = tex1Dfetch(TEX_VAL4, j*3+1);
			float4 val2 = tex1Dfetch(TEX_VAL4, j*3+2);
			
			/* GEMV, Block by Vector */
			y0 += val0.x * x0 + val0.y * x1 + val0.z * x2;
			y1 += val0.w * x0 + val1.x * x1 + val1.y * x2;
			y2 += val1.z * x0 + val1.w * x1 + val2.x * x2;
		}
		
		/* copy register to memory */
		d_y[i*3  ] += alpha*y0;
		d_y[i*3+1] += alpha*y1;
		d_y[i*3+2] += alpha*y2;
	}
}

/**
 * Calculate SPMV 4x4
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bcsr4_4x4(const int n, const float alpha, float *d_y)
{

	int i, j;

	/* Read the data*/
	for (i = bx * CUKR_SSPMV_BCSR4_4X4_THREADS + tx; i < n; i += CUKR_SSPMV_BCSR4_4X4_CTAS * CUKR_SSPMV_BCSR4_4X4_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0=0.0, y1=0.0 ,y2=0.0, y3=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) - 1);
			float x1=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j));
			float x2=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 1);
			float x3=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 2);
			
			/* flot4 */
			float4 val0 = tex1Dfetch(TEX_VAL4, j*4);
			float4 val1 = tex1Dfetch(TEX_VAL4, j*4+1);
			float4 val2 = tex1Dfetch(TEX_VAL4, j*4+2);
			float4 val3 = tex1Dfetch(TEX_VAL4, j*4+3);
			
			/* GEMV, Block by Vector */
			y0 += val0.x * x0 + val0.y * x1 + val0.z * x2 + val0.w * x3;
			y1 += val1.x * x0 + val1.y * x1 + val1.z * x2 + val1.w * x3;
			y2 += val2.x * x0 + val2.y * x1 + val2.z * x2 + val2.w * x3;
			y3 += val3.x * x0 + val3.y * x1 + val3.z * x2 + val3.w * x3;
		}
		
		/* copy register to memory */
		d_y[i*4  ] += alpha*y0;
		d_y[i*4+1] += alpha*y1;
		d_y[i*4+2] += alpha*y2;
		d_y[i*4+3] += alpha*y3;
	}
}

/**
 * Calculate SPMV 5x5
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bcsr4_5x5(const int n, const float alpha, float *d_y)
{

	int i, j;

	/* Read the data*/
	for (i = bx * CUKR_SSPMV_BCSR4_5X5_THREADS + tx; i < n; i += CUKR_SSPMV_BCSR4_5X5_CTAS * CUKR_SSPMV_BCSR4_5X5_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0=0.0, y1=0.0 ,y2=0.0, y3=0.0, y4=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) - 1);
			float x1=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j));
			float x2=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 1);
			float x3=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 2);
			float x4=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 3);
			
			/* flot4 */
			float4 val0 = tex1Dfetch(TEX_VAL4, j*7);
			float4 val1 = tex1Dfetch(TEX_VAL4, j*7+1);
			float4 val2 = tex1Dfetch(TEX_VAL4, j*7+2);
			float4 val3 = tex1Dfetch(TEX_VAL4, j*7+3);
			float4 val4 = tex1Dfetch(TEX_VAL4, j*7+4);
			float4 val5 = tex1Dfetch(TEX_VAL4, j*7+5);
			float4 val6 = tex1Dfetch(TEX_VAL4, j*7+6);
			
			/* GEMV, Block by Vector */
			y0 += val0.x * x0 + val0.y * x1 + val0.z * x2 + val0.w * x3 + val1.x * x4;
			y1 += val1.y * x0 + val1.z * x1 + val1.w * x2 + val2.x * x3 + val2.y * x4;
			y2 += val2.z * x0 + val2.w * x1 + val3.x * x2 + val3.y * x3 + val3.z * x4;
			y3 += val3.w * x0 + val4.x * x1 + val4.y * x2 + val4.z * x3 + val4.w * x4;
			y4 += val5.x * x0 + val5.y * x1 + val5.z * x2 + val5.w * x3 + val6.x * x4;
		}
		
		/* copy register to memory */
		d_y[i*5  ] += alpha*y0;
		d_y[i*5+1] += alpha*y1;
		d_y[i*5+2] += alpha*y2;
		d_y[i*5+3] += alpha*y3;
		d_y[i*5+4] += alpha*y4;
	}
}

/**
 * Calculate SPMV 6x6
 * n      num of blocks per row or per col
 * alpha  constant number a
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bcsr4_6x6(const int n, const float alpha, float *d_y)
{

	int i, j;

	/* Read the data*/
	for (i = bx * CUKR_SSPMV_BCSR4_6X6_THREADS + tx; i < n; i += CUKR_SSPMV_BCSR4_6X6_CTAS * CUKR_SSPMV_BCSR4_6X6_THREADS)
	{
		/* Read the beginning and end of the row
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* register y */
		float y0=0.0, y1=0.0 ,y2=0.0, y3=0.0, y4=0.0, y5=0.0;

		/* Read and sum for the column vectors */
		for (j = iRowBeg; j < iRowEnd; j++) {
			/* register x */
			float x0=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) - 1);
			float x1=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j));
			float x2=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 1);
			float x3=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 2);
			float x4=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 3);
			float x5=tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) + 4);
			
			/* flot4 */
			float4 val0 = tex1Dfetch(TEX_VAL4, j*9);
			float4 val1 = tex1Dfetch(TEX_VAL4, j*9+1);
			float4 val2 = tex1Dfetch(TEX_VAL4, j*9+2);
			float4 val3 = tex1Dfetch(TEX_VAL4, j*9+3);
			float4 val4 = tex1Dfetch(TEX_VAL4, j*9+4);
			float4 val5 = tex1Dfetch(TEX_VAL4, j*9+5);
			float4 val6 = tex1Dfetch(TEX_VAL4, j*9+6);
			float4 val7 = tex1Dfetch(TEX_VAL4, j*9+7);
			float4 val8 = tex1Dfetch(TEX_VAL4, j*9+8);
			
			/* GEMV, Block by Vector */
			y0 += val0.x * x0 + val0.y * x1 + val0.z * x2 + val0.w * x3 + val1.x * x4 + val1.y * x5;
			y1 += val1.z * x0 + val1.w * x1 + val2.x * x2 + val2.y * x3 + val2.z * x4 + val2.w * x5;
			y2 += val3.x * x0 + val3.y * x1 + val3.z * x2 + val3.w * x3 + val4.x * x4 + val4.y * x5;
			y3 += val4.z * x0 + val4.w * x1 + val5.x * x2 + val5.y * x3 + val5.z * x4 + val5.w * x5;
			y4 += val6.x * x0 + val6.y * x1 + val6.z * x2 + val6.w * x3 + val7.x * x4 + val7.y * x5;
			y5 += val7.z * x0 + val7.w * x1 + val8.x * x2 + val8.y * x3 + val8.z * x4 + val8.w * x5;
		}
		
		/* copy register to memory */
		d_y[i*6  ] += alpha*y0;
		d_y[i*6+1] += alpha*y1;
		d_y[i*6+2] += alpha*y2;
		d_y[i*6+3] += alpha*y3;
		d_y[i*6+4] += alpha*y4;
		d_y[i*6+5] += alpha*y5;
	}
}
