/**
 * @file blas/gpu/cuda/kernel_sspmv_bell.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Yusa codes.
 *
 * SPMV in BELL format
 *
 * X  Num of registers is (r + c + 6)
 * O  Num of registers is noted in kernels_single.sm_10.cubin
 */
 
#include "../kernels_textures.h"
#include "../kernels_utils.h"

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x 
	#define bx          blockIdx.x
#endif


/**
 * Calculate SPMV when block size is 2x2
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bell_2x2(const int n, const float alpha, const int nz_row, const int stride, float *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_SSPMV_BELL_2X2_THREADS + tx; i < n; i += CUKR_SSPMV_BELL_2X2_CTAS * CUKR_SSPMV_BELL_2X2_THREADS)
	{
		/* register y */
		float y0 = 0.0, y1 = 0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			float x0 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)-1);
			float x1 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i));

			/* GEMV */
			y0 += tex1Dfetch(TEX_VAL, (j*stride+i)*4  )*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*4+1)*x1;
			y1 += tex1Dfetch(TEX_VAL, (j*stride+i)*4+2)*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*4+3)*x1;
		}

		/* copy register y to global memory d_y */
		d_y[i*2  ] += alpha*y0;
		d_y[i*2+1] += alpha*y1;
	}
}


/**
 * Calculate SPMV when block size is 3x3
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bell_3x3(const int n, const float alpha, const int nz_row, const int stride, float *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_SSPMV_BELL_3X3_THREADS + tx; i < n; i += CUKR_SSPMV_BELL_3X3_CTAS * CUKR_SSPMV_BELL_3X3_THREADS)
	{
		/* register y */
		float y0 = 0.0, y1 = 0.0, y2 = 0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			float x0 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)-1);
			float x1 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i));
			float x2 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+1);

			/* GEMV */
			y0 += tex1Dfetch(TEX_VAL, (j*stride+i)*9  )*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*9+1)*x1
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*9+2)*x2;
			y1 += tex1Dfetch(TEX_VAL, (j*stride+i)*9+3)*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*9+4)*x1
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*9+5)*x2;
			y2 += tex1Dfetch(TEX_VAL, (j*stride+i)*9+6)*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*9+7)*x1
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*9+8)*x2;
		}

		/* copy register y to global memory d_y */
		d_y[i*3  ] += alpha*y0;
		d_y[i*3+1] += alpha*y1;
		d_y[i*3+2] += alpha*y2;
	}
}


/**
 * Calculate SPMV when block size is 4x4
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bell_4x4(const int n, const float alpha, const int nz_row, const int stride, float *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_SSPMV_BELL_4X4_THREADS + tx; i < n; i += CUKR_SSPMV_BELL_4X4_CTAS * CUKR_SSPMV_BELL_4X4_THREADS)
	{
		/* register y */
		float y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			float x0 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)-1);
			float x1 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i));
			float x2 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+1);
			float x3 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+2);

			/* GEMV */
			y0 += tex1Dfetch(TEX_VAL, (j*stride+i)*16  )*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*16+1)*x1
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*16+2)*x2
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*16+3)*x3;
			y1 += tex1Dfetch(TEX_VAL, (j*stride+i)*16+4)*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*16+5)*x1
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*16+6)*x2
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*16+7)*x3;
			y2 += tex1Dfetch(TEX_VAL, (j*stride+i)*16+8)*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*16+9)*x1
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*16+10)*x2
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*16+11)*x3;
			y3 += tex1Dfetch(TEX_VAL, (j*stride+i)*16+12)*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*16+13)*x1
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*16+14)*x2
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*16+15)*x3;
		}

		/* copy register y to global memory d_y */
		d_y[i*4  ] += alpha*y0;
		d_y[i*4+1] += alpha*y1;
		d_y[i*4+2] += alpha*y2;
		d_y[i*4+3] += alpha*y3;
	}
}


/**
 * Calculate SPMV when block size is 5x5
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bell_5x5(const int n, const float alpha, const int nz_row, const int stride, float *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_SSPMV_BELL_5X5_THREADS + tx; i < n; i += CUKR_SSPMV_BELL_5X5_CTAS * CUKR_SSPMV_BELL_5X5_THREADS)
	{
		/* register y */
		float y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 =0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			float x0 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)-1);
			float x1 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i));
			float x2 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+1);
			float x3 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+2);
			float x4 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+3);

			/* GEMV */
			y0 += tex1Dfetch(TEX_VAL, (j*stride+i)*25  )*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+1)*x1
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+2)*x2
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+3)*x3
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+4)*x4;
			y1 += tex1Dfetch(TEX_VAL, (j*stride+i)*25+5)*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+6)*x1
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+7)*x2
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+8)*x3
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+9)*x4;
			y2 += tex1Dfetch(TEX_VAL, (j*stride+i)*25+10)*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+11)*x1
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+12)*x2
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+13)*x3
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+14)*x4;
			y3 += tex1Dfetch(TEX_VAL, (j*stride+i)*25+15)*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+16)*x1
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+17)*x2
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+18)*x3
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+19)*x4;
			y4 += tex1Dfetch(TEX_VAL, (j*stride+i)*25+20)*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+21)*x1
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+22)*x2
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+23)*x3
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*25+24)*x4;
		}

		/* copy register y to global memory d_y */
		d_y[i*5  ] += alpha*y0;
		d_y[i*5+1] += alpha*y1;
		d_y[i*5+2] += alpha*y2;
		d_y[i*5+3] += alpha*y3;
		d_y[i*5+4] += alpha*y4;
	}
}


/**
 * Calculate SPMV when block size is 6x6
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bell_6x6(const int n, const float alpha, const int nz_row, const int stride, float *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_SSPMV_BELL_6X6_THREADS + tx; i < n; i += CUKR_SSPMV_BELL_6X6_CTAS * CUKR_SSPMV_BELL_6X6_THREADS)
	{
		/* register y */
		float y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 =0.0, y5 = 0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			float x0 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)-1);
			float x1 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i));
			float x2 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+1);
			float x3 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+2);
			float x4 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+3);
			float x5 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+4);

			/* GEMV */
			y0 += tex1Dfetch(TEX_VAL, (j*stride+i)*36  )*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+1)*x1
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+2)*x2
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+3)*x3
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+4)*x4
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+5)*x5;
			y1 += tex1Dfetch(TEX_VAL, (j*stride+i)*36+6)*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+7)*x1
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+8)*x2
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+9)*x3
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+10)*x4
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+11)*x5;
			y2 += tex1Dfetch(TEX_VAL, (j*stride+i)*36+12)*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+13)*x1
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+14)*x2
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+15)*x3
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+16)*x4
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+17)*x5;
			y3 += tex1Dfetch(TEX_VAL, (j*stride+i)*36+18)*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+19)*x1
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+20)*x2
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+21)*x3
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+22)*x4
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+23)*x5;
			y4 += tex1Dfetch(TEX_VAL, (j*stride+i)*36+24)*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+25)*x1
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+26)*x2
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+27)*x3
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+28)*x4
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+29)*x5;
			y5 += tex1Dfetch(TEX_VAL, (j*stride+i)*36+30)*x0
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+31)*x1
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+32)*x2
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+33)*x3
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+34)*x4
			    + tex1Dfetch(TEX_VAL, (j*stride+i)*36+35)*x5;
		}

		/* copy register y to global memory d_y */
		d_y[i*6  ] += alpha*y0;
		d_y[i*6+1] += alpha*y1;
		d_y[i*6+2] += alpha*y2;
		d_y[i*6+3] += alpha*y3;
		d_y[i*6+4] += alpha*y4;
		d_y[i*6+5] += alpha*y5;
	}
}



