/**
 * @file blas/gpu/cuda/kernel_spmv_bell4.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Yusa codes.
 *
 * SPMV in BELL4 format
 */
 
#include "../kernels_textures.h"
#include "../kernels_utils.h"

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x 
	#define bx          blockIdx.x
#endif


/**
 * Calculate SPMV when block size is 2x2
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bell4_2x2(const int n, const float alpha, const int nz_row, const int stride, float *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_SSPMV_BELL4_2X2_THREADS + tx; i < n; i += CUKR_SSPMV_BELL4_2X2_CTAS * CUKR_SSPMV_BELL4_2X2_THREADS)
	{
		/* register y */
		float y0 = 0.0, y1 = 0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			float x0 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)-1);
			float x1 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i));
			
			/* flot4 */
			float4 val0 = tex1Dfetch(TEX_VAL4, j*stride+i);
			
			/* GEMV */
			y0 += val0.x * x0 + val0.y * x1;
			y1 += val0.z * x0 + val0.w * x1;
		}

		/* copy register y to global memory d_y */
		d_y[i*2  ] += alpha*y0;
		d_y[i*2+1] += alpha*y1;
	}
}


/**
 * Calculate SPMV when block size is 3x3
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bell4_3x3(const int n, const float alpha, const int nz_row, const int stride, float *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_SSPMV_BELL4_3X3_THREADS + tx; i < n; i += CUKR_SSPMV_BELL4_3X3_CTAS * CUKR_SSPMV_BELL4_3X3_THREADS)
	{
		/* register y */
		float y0 = 0.0, y1 = 0.0, y2 = 0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			float x0 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)-1);
			float x1 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i));
			float x2 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+1);
			
			/* flot4 */
			float4 val0 = tex1Dfetch(TEX_VAL4, (j*stride+i)*3);
			float4 val1 = tex1Dfetch(TEX_VAL4, (j*stride+i)*3+1);
			float4 val2 = tex1Dfetch(TEX_VAL4, (j*stride+i)*3+2);

			/* GEMV */
			y0 += val0.x * x0 + val0.y * x1 + val0.z * x2;
			y1 += val0.w * x0 + val1.x * x1 + val1.y * x2;
			y2 += val1.z * x0 + val1.w * x1 + val2.x * x2;
		}

		/* copy register y to global memory d_y */
		d_y[i*3  ] += alpha*y0;
		d_y[i*3+1] += alpha*y1;
		d_y[i*3+2] += alpha*y2;
	}
}


/**
 * Calculate SPMV when block size is 4x4
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bell4_4x4(const int n, const float alpha, const int nz_row, const int stride, float *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_SSPMV_BELL4_4X4_THREADS + tx; i < n; i += CUKR_SSPMV_BELL4_4X4_CTAS * CUKR_SSPMV_BELL4_4X4_THREADS)
	{
		/* register y */
		float y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			float x0 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)-1);
			float x1 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i));
			float x2 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+1);
			float x3 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+2);
			
			/* flot4 */
			float4 val0 = tex1Dfetch(TEX_VAL4, (j*stride+i)*4);
			float4 val1 = tex1Dfetch(TEX_VAL4, (j*stride+i)*4+1);
			float4 val2 = tex1Dfetch(TEX_VAL4, (j*stride+i)*4+2);
			float4 val3 = tex1Dfetch(TEX_VAL4, (j*stride+i)*4+3);

			/* GEMV */
			y0 += val0.x * x0 + val0.y * x1 + val0.z * x2 + val0.w * x3;
			y1 += val1.x * x0 + val1.y * x1 + val1.z * x2 + val1.w * x3;
			y2 += val2.x * x0 + val2.y * x1 + val2.z * x2 + val2.w * x3;
			y3 += val3.x * x0 + val3.y * x1 + val3.z * x2 + val3.w * x3;
		}

		/* copy register y to global memory d_y */
		d_y[i*4  ] += alpha*y0;
		d_y[i*4+1] += alpha*y1;
		d_y[i*4+2] += alpha*y2;
		d_y[i*4+3] += alpha*y3;
	}
}


/**
 * Calculate SPMV when block size is 5x5
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bell4_5x5(const int n, const float alpha, const int nz_row, const int stride, float *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_SSPMV_BELL4_5X5_THREADS + tx; i < n; i += CUKR_SSPMV_BELL4_5X5_CTAS * CUKR_SSPMV_BELL4_5X5_THREADS)
	{
		/* register y */
		float y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 =0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			float x0 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)-1);
			float x1 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i));
			float x2 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+1);
			float x3 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+2);
			float x4 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+3);

			/* flot4 */
			float4 val0 = tex1Dfetch(TEX_VAL4, (j*stride+i)*7);
			float4 val1 = tex1Dfetch(TEX_VAL4, (j*stride+i)*7+1);
			float4 val2 = tex1Dfetch(TEX_VAL4, (j*stride+i)*7+2);
			float4 val3 = tex1Dfetch(TEX_VAL4, (j*stride+i)*7+3);
			float4 val4 = tex1Dfetch(TEX_VAL4, (j*stride+i)*7+4);
			float4 val5 = tex1Dfetch(TEX_VAL4, (j*stride+i)*7+5);
			float4 val6 = tex1Dfetch(TEX_VAL4, (j*stride+i)*7+6);

			/* GEMV */
			y0 += val0.x * x0 + val0.y * x1 + val0.z * x2 + val0.w * x3 + val1.x * x4;
			y1 += val1.y * x0 + val1.z * x1 + val1.w * x2 + val2.x * x3 + val2.y * x4;
			y2 += val2.z * x0 + val2.w * x1 + val3.x * x2 + val3.y * x3 + val3.z * x4;
			y3 += val3.w * x0 + val4.x * x1 + val4.y * x2 + val4.z * x3 + val4.w * x4;
			y4 += val5.x * x0 + val5.y * x1 + val5.z * x2 + val5.w * x3 + val6.x * x4;
		}

		/* copy register y to global memory d_y */
		d_y[i*5  ] += alpha*y0;
		d_y[i*5+1] += alpha*y1;
		d_y[i*5+2] += alpha*y2;
		d_y[i*5+3] += alpha*y3;
		d_y[i*5+4] += alpha*y4;
	}
}


/**
 * Calculate SPMV when block size is 6x6
 * n      num of blocks per row or per col
 * alpha  constant number a
 * nz_row nonzero blocks per row
 * stride stride
 * d_y    output vector y
 */
__global__ void kernel_sspmv_bell4_6x6(const int n, const float alpha, const int nz_row, const int stride, float *d_y)
{
	int i, j;
	
	for (i = bx * CUKR_SSPMV_BELL4_6X6_THREADS + tx; i < n; i += CUKR_SSPMV_BELL4_6X6_CTAS * CUKR_SSPMV_BELL4_6X6_THREADS)
	{
		/* register y */
		float y0 = 0.0, y1 = 0.0, y2 = 0.0, y3 = 0.0, y4 =0.0, y5 = 0.0;
	
		for(j = 0; j < nz_row; j++){
			/* register x */
			float x0 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)-1);
			float x1 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i));
			float x2 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+1);
			float x3 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+2);
			float x4 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+3);
			float x5 = tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j*stride+i)+4);

			/* flot4 */
			float4 val0 = tex1Dfetch(TEX_VAL4, (j*stride+i)*9);
			float4 val1 = tex1Dfetch(TEX_VAL4, (j*stride+i)*9+1);
			float4 val2 = tex1Dfetch(TEX_VAL4, (j*stride+i)*9+2);
			float4 val3 = tex1Dfetch(TEX_VAL4, (j*stride+i)*9+3);
			float4 val4 = tex1Dfetch(TEX_VAL4, (j*stride+i)*9+4);
			float4 val5 = tex1Dfetch(TEX_VAL4, (j*stride+i)*9+5);
			float4 val6 = tex1Dfetch(TEX_VAL4, (j*stride+i)*9+6);
			float4 val7 = tex1Dfetch(TEX_VAL4, (j*stride+i)*9+7);
			float4 val8 = tex1Dfetch(TEX_VAL4, (j*stride+i)*9+8);

			/* GEMV */
			y0 += val0.x * x0 + val0.y * x1 + val0.z * x2 + val0.w * x3 + val1.x * x4 + val1.y * x5;
			y1 += val1.z * x0 + val1.w * x1 + val2.x * x2 + val2.y * x3 + val2.z * x4 + val2.w * x5;
			y2 += val3.x * x0 + val3.y * x1 + val3.z * x2 + val3.w * x3 + val4.x * x4 + val4.y * x5;
			y3 += val4.z * x0 + val4.w * x1 + val5.x * x2 + val5.y * x3 + val5.z * x4 + val5.w * x5;
			y4 += val6.x * x0 + val6.y * x1 + val6.z * x2 + val6.w * x3 + val7.x * x4 + val7.y * x5;
			y5 += val7.z * x0 + val7.w * x1 + val8.x * x2 + val8.y * x3 + val8.z * x4 + val8.w * x5;
		}

		/* copy register y to global memory d_y */
		d_y[i*6  ] += alpha*y0;
		d_y[i*6+1] += alpha*y1;
		d_y[i*6+2] += alpha*y2;
		d_y[i*6+3] += alpha*y3;
		d_y[i*6+4] += alpha*y4;
		d_y[i*6+5] += alpha*y5;
	}
}



