/**
 * @file blas/gpu/cuda/kernel_spmv_csr4_a1_b0.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * SpMV in single precision for 
 * CSR4 (padded CSR) format
 * 
 * Optimized version
 * Reads 4 elements at a time from
 * val and idx vectors, using float4
 * and int4 structures.
 * In order to do that, the vectors
 * have been padded so that the no.
 * of elements in each row is multiple
 * of 4.
 * 
 * It can do up to 6-7GFlop/s 
 * 
 * Most simplified version: alpha=1, beta=0.
 * This means the multiplication by alpha is
 * no longer needed and the result vector
 * is simply overwriten (not updated)
 */


#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x 
	#define bx          blockIdx.x
#endif

 
__global__ void kernel_sspmv_csr4_a1_b0_2( const int     n,  
										   const float  *d_x,
										   const int    *d_idx,
										   const float  *d_val,   	
						          	             float  *d_y)
{
	int i;

	/* Starting point for this block */
	int ctaStart = bx * CUKR_SSPMV_CSR4_THREADS; 
	
	/* Total no. of threads in the kernel */
	int totalThreads = CUKR_SSPMV_CSR4_CTAS * CUKR_SSPMV_CSR4_THREADS;
	
	/* Read the data*/
	for (i = ctaStart + tx; i < n; i += totalThreads)
	{ 	
		/* Read the beginning and end of the row 
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* Read and sum for the column vectors */
		float sum = 0;		
		for (int j = iRowBeg / 1; j < iRowEnd / 1; j++) {
			float val, x;
			int idx;
			
			val = d_val[j];
			x = d_x[idx];
			sum += x*val;  //x.x*val.x + x.y*val.y + x.z*val.z + x.w*val.w ;  
		}

		/* Write the result to global memory 
		 * (overwrite) */
		d_y[i] = sum;
					 	
	}  
}
