/**
 * @file blas/gpu/cuda/kernel_spmv_csr4_b0.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * SpMV in single precision for 
 * CSR4 (padded CSR) format
 * 
 * Optimized version
 * Reads 4 elements at a time from
 * val and idx vectors, using float4
 * and int4 structures.
 * In order to do that, the vectors
 * have been padded so that the no.
 * of elements in each row is multiple
 * of 4.
 * 
 * It can do up to 6-7GFlop/s 
 * 
 * Simplified version: only alpha is 
 * considered. Taking beta zero means
 * the final vector does not have to
 * be updated, but simply overriten
 * (so a read is saved)  
 */


#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x 
	#define bx          blockIdx.x
#endif

 
__global__ void kernel_sspmv_csr4_b0( const int    n,      	
						  	          const float  alpha,
						          	        float *d_y)
{
	int i;

	/* Starting point for this block */
	int ctaStart = bx * CUKR_SSPMV_CSR4_THREADS; 
	
	/* Total no. of threads in the kernel */
	int totalThreads = CUKR_SSPMV_CSR4_CTAS * CUKR_SSPMV_CSR4_THREADS;
	
	/* Read the data*/
	for (i = ctaStart + tx; i < n; i += totalThreads)
	{ 	
		/* Read the beginning and end of the row 
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* Read and sum for the column vectors */
		float sum = 0;		
		for (int j = iRowBeg / 4; j < iRowEnd / 4; j++) {
			float4 val, x;
			int4 idx;
			 
			idx = tex1Dfetch(TEX_IDX4, j);	
			val = tex1Dfetch(TEX_VAL4, j);
			
			idx.x -= 1;
			idx.y -= 1;
			idx.z -= 1;
			idx.w -= 1;
			
			x.x = tex1Dfetch(TEX_X, idx.x);
			x.y = tex1Dfetch(TEX_X, idx.y);
			x.z = tex1Dfetch(TEX_X, idx.z);
			x.w = tex1Dfetch(TEX_X, idx.w); 
			 		
			sum += x.x*val.x + x.y*val.y + x.z*val.z + x.w*val.w;
		}

		/* Write the result to shared memory 
		 * (overwrite)*/
		d_y[i] = alpha * sum;		 	
	}  
}

