/**
 * @file blas/gpu/cuda/kernel_spmv_csr_a1_b9.cu
 * @author Serban Georgescu
 * @since 2008
 * 
 * Simplest version of SpMV (no data 
 * structure modification, coalescing 
 * the writes)
 * 
 * It can do 2.25 GFlops = 8.86 GB/s 
 * (n = 1000000, 416 threads, 2650 blocks)
 * 
 * Simplest variant: alpha = 1, beta = 0
 * meaning that they are not considered.
 */

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x 
	#define bx          blockIdx.x
#endif

/* Textures */
#include "../kernels_textures.h"

/* Shared memory */
extern float __shared__ s_y[]; 
 
__global__ void kernel_sspmv_csr_a1_b0( const int    n,      
						        	          float *d_y)
{	
	/* Get texture bindings */
	
	int i;
	
	/* Starting point for this block */
	int ctaStart = bx * CUKR_SSPMV_CSR_THREADS; 
	
	/* Total no. of threads in the kernel */
	int totalThreads = CUKR_SSPMV_CSR_CTAS * CUKR_SSPMV_CSR_THREADS;
	
	/* Read the data*/
	for (i = ctaStart + tx; i < n; i += totalThreads)
	{ 	
		/* Read the beginning and end of the row 
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		
		/* Read and sum for the column vectors */
		#if USE_KAHAN_IN_SPMV
			float sum = tex1Dfetch(TEX_VAL, iRowBeg) * tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, iRowBeg) - 1);
			float c = 0.0;
			for (int j = iRowBeg + 1; j < iRowEnd; j++) {
				float y = tex1Dfetch(TEX_VAL, j) * tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) - 1) - c;
				float t = sum + y;
	        	c = (t - sum) - y;
	        	sum = t;
			}
		#else		
			float sum = 0;		
			for (int j = iRowBeg; j < iRowEnd; j++) {
				sum += tex1Dfetch(TEX_VAL, j) * tex1Dfetch(TEX_X, tex1Dfetch(TEX_IDX, j) - 1);					
			} 
			
		#endif

		/* Write the result to shared memory */
		d_y[i] = sum;	
		 	
	}  
}

