/**
 * @file blas/gpu/cuda/kernel_spmv_csr_a1_b9.cu
 * @author Serban Georgescu
 * @since 2008
 * 
 * Simplest version of SpMV (no data 
 * structure modification, coalescing 
 * the writes)
 * 
 * It can do 2.25 GFlops = 8.86 GB/s 
 * (n = 1000000, 416 threads, 2650 blocks)
 * 
 * Simplest variant: alpha = 1, beta = 0
 * meaning that they are not considered.
 */

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x 
	#define bx          blockIdx.x
#endif

/* Textures */
#include "../kernels_textures.h"

 
__global__ void kernel_sspmv_csr_a1_b0_c( const int    n,  
											    float *d_x,    
						        	            float *d_y)
{	
	int i;
	
	/* Declare cache */
	__shared__ float x_cache[CUKR_SSPMV_CSR_THREADS];
	
	/* Starting point for this block */
	int ctaStart = bx * CUKR_SSPMV_CSR_THREADS; 
	
	/* Total no. of threads in the kernel */
	int totalThreads = CUKR_SSPMV_CSR_CTAS * CUKR_SSPMV_CSR_THREADS;
	
	/* Read the data*/
	for (i = ctaStart + tx; i < n; i += totalThreads)
	{ 	
		printf("i = %d\n", i);
		
		/* Prefetch */
		if (tx < n) {
			x_cache[tx] = d_x[i];
			printf("  x_cache[%d] = d_x[%d] = %g\n", tx, i, x_cache[tx]);  
		}
		//__syncthreads();	
		
		/* Read the beginning and end of the row 
		 * which will be processed by this thread */
		int iRowBeg = tex1Dfetch(TEX_PTR, i) - 1;
		int iRowEnd = tex1Dfetch(TEX_PTR, i + 1) - 1;
		printf("  iRowBeg = %d, iRowEnd = %d\n", iRowBeg, iRowEnd);
		
		/* Read and sum for the column vectors */
		float sum = 0;		
		for (int j = iRowBeg; j < iRowEnd; j++) 
		{
			printf("  j = %d\n", j);
			
			float x;
			int col, col_shifted;
			
			col = tex1Dfetch(TEX_IDX, j) - 1;
			printf(" col = %d\n", col);
			col_shifted = col - i;
			printf(" col_shifted = %d\n", col_shifted);
			
			if ((0 <= col_shifted) && (col_shifted < totalThreads)) {						
				x = x_cache[col_shifted];
				printf("    x = x_cache[%d] = %g\n", col_shifted, x);	
			}
			else {					
				x = tex1Dfetch(TEX_X, col);
				printf("    x = tex1Dfetch[%d] = %g\n", col, x);	
			}
		
			sum += tex1Dfetch(TEX_VAL, j) * x;					
		} 
			
		/* Write the result to shared memory */
		d_y[i] = sum;	
		 	
	}  
}

