/**
 * @file blas/impl/gpu/gpu_blas.c
 * @author Serban Georgescu
 * @since 2008
 * 
 *@ OpenCL additions Olav Aanes Fagerlund
 *
 * Defines wrappers for BLAS functions 
 * on the GPU. 
 * 
 * The functions take as parameter device
 * pointers, so they only act on data that
 * has been PREVIOUSLY COPIED to the device.
 * 
 * The functions are implemented in the 
 * following way:
 * - single
 * 	- BLAS 1: CUBLAS + own
 *  - SpMV: gpublas
 *  - double
 */

#define LOGGER_NAME "CUKr.blas.impl.gpu_blas"

#include <stdio.h>
#include <stdlib.h> 
#include <pthread.h>
#include "../../../util/log.h"
#include "gpu_blas.h"
#ifdef CUKR_USE_CUDA
	#include "cuda/cublas_wrappers.h"
	#include "cuda/kernels.h"
#endif
#ifdef CUKR_USE_OPENCL
	#include "opencl/kernels_single.h"
	#include "opencl/kernels_qdouble.h"
	#include "opencl/kernels_double.h"
#endif	


// ------------------------------------------------------------------- //

#ifdef CUKR_USE_GPU

/**
 * \brief Dot product in double precision
 * \param n Vector size
 * \param x,y The two vectors whose product 
 *            is being computed 
 */
#ifdef CUKR_USE_OPENCL
CUKR_DOUBLE gpu_ddot(CUKR_INT n, CUKR_CL_DOUBLE* x, CUKR_CL_DOUBLE* y)
{
	return opencl_kernel_ddot(n, x, y);
}
#else
CUKR_DOUBLE gpu_ddot(CUKR_INT n, CUKR_DOUBLE* x, CUKR_DOUBLE* y)
{
	return cuda_cublasDdot(n, x, 1, y, 1);
}
#endif

/**
 * \brief Dot product in quasi-double precision
 * \param n Vector size
 * \param n The vector size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
#ifdef CUKR_USE_OPENCL
CUKR_DOUBLE gpu_qdot(CUKR_INT n, CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt,
								 CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt)
{
	return opencl_kernel_qdot(n, xh, xt, yh, yt);
}
#else
CUKR_DOUBLE gpu_qdot(CUKR_INT n, CUKR_SINGLE* xh, CUKR_SINGLE* xt,
					 CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	return cuda_kernel_qdot(n, xh, xt, yh, yt);
}
#endif

/**
 * \brief Dot product in single precision
 * \param n Vector size
 * \param x,y The two vectors whose product 
 *            is being computed 
 */
#ifdef CUKR_USE_OPENCL
CUKR_SINGLE gpu_sdot(CUKR_INT n, CUKR_CL_SINGLE* x, CUKR_CL_SINGLE* y)
{
	return opencl_kernel_sdot(n, x, y);
}
#else
CUKR_SINGLE gpu_sdot(CUKR_INT n, CUKR_SINGLE* x, CUKR_SINGLE* y)
{
	return cuda_cublasSdot(n, x, 1, y, 1);
}
#endif
    
// ------------------------------------------------------------------- //
  
/**
 * \brief AXPY in double precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
#ifdef CUKR_USE_OPENCL
void gpu_daxpy(CUKR_INT n, CUKR_DOUBLE a, CUKR_CL_DOUBLE* x, CUKR_CL_DOUBLE* y)
{
	opencl_kernel_daxpy(n, a, x, y);
}
#else
void gpu_daxpy(CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x, CUKR_DOUBLE* y)
{
	cuda_cublasDaxpy(n, a, x, 1, y, 1);
}
#endif

/**
 * \brief AXPY in quasi-double precision
 * \param n Vector size
 * \param a The constant
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector 
 */
#ifdef CUKR_USE_OPENCL
void gpu_qaxpy(CUKR_INT n, CUKR_DOUBLE a, CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt, 
										  CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt)
{
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	/* Split a */
	float volatile a0 = (float)a;
	float volatile a1 = (float)(a - a0);
	
	/* Run the kernel */
	opencl_kernel_qaxpy(n, a0, a1, xh, xt, yh, yt);
}	
#else
void gpu_qaxpy(CUKR_INT n, CUKR_DOUBLE a, CUKR_SINGLE* xh, CUKR_SINGLE* xt, 
										  CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	/* Split a */
	float volatile a0 = (float)a;
	float volatile a1 = (float)(a - a0);
	
	/* Run the kernel */
	cuda_kernel_qaxpy(n, a0, a1, xh, xt, yh, yt);
}	
#endif

/**
 * \brief AXPY in single precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
#ifdef CUKR_USE_OPENCL
void gpu_saxpy(CUKR_INT n, CUKR_SINGLE a, CUKR_CL_SINGLE* x, CUKR_CL_SINGLE* y)
{
	opencl_kernel_saxpy(n, a, x, y);
}
#else
void gpu_saxpy(CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x, CUKR_SINGLE* y)
{
	cuda_cublasSaxpy(n, a, x, 1, y, 1);
}
#endif

// ------------------------------------------------------------------- //
  
/**
 * \brief AYPX in double precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
#ifdef CUKR_USE_OPENCL
void gpu_daypx(CUKR_INT n, CUKR_DOUBLE a, CUKR_CL_DOUBLE* cl_x, CUKR_CL_DOUBLE* cl_y)
{
	opencl_kernel_daypx(n, a, cl_x, cl_y);
	
}
#else
void gpu_daypx(CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x, CUKR_DOUBLE* y)
{
	cuda_kernel_daypx(n, a, x, y);
}
#endif

/**
 * \brief AYPX in quasi-double precision
 * \param n Vector size
 * \param a The constant
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector 
 */

#ifdef CUKR_USE_OPENCL
void gpu_qaypx(CUKR_INT n, CUKR_DOUBLE a, CUKR_CL_SINGLE* cl_xh, CUKR_CL_SINGLE* cl_xt,
										  CUKR_CL_SINGLE* cl_yh, CUKR_CL_SINGLE* cl_yt)
{
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	/* Split a */
	float volatile a0 = (float)a;
	float volatile a1 = (float)(a - a0);
	
	/* Run the kernel */
	opencl_kernel_qaypx(n, a0, a1, cl_xh, cl_xt, cl_yh, cl_yt);
}
#else
void gpu_qaypx(CUKR_INT n, CUKR_DOUBLE a, CUKR_SINGLE* xh, CUKR_SINGLE* xt, 
										  CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	/* Split a */
	float volatile a0 = (float)a;
	float volatile a1 = (float)(a - a0);
	
	/* Run the kernel */
	cuda_kernel_qaypx(n, a0, a1, xh, xt, yh, yt);
}
#endif

/**
 * \brief AYPX in single precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */

#ifdef CUKR_USE_OPENCL
void gpu_saypx(CUKR_INT n, CUKR_SINGLE a, CUKR_CL_SINGLE* x, CUKR_CL_SINGLE* y)
{	
	opencl_kernel_saypx(n, a, x, y);
}
#else
void gpu_saypx(CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x, CUKR_SINGLE* y)
{
	cuda_kernel_saypx(n, a, x, y);
}
#endif

// ------------------------------------------------------------------- //

/**
 * \brief Vector scaling in double precision
 * \param n Vector size
 * \param a Scaling factor
 * \param x The vector being scaled
 */
#ifdef CUKR_USE_OPENCL
void gpu_dscal(CUKR_INT n, CUKR_DOUBLE a, CUKR_CL_DOUBLE* x)
{
	opencl_kernel_dscal(n, a, x);
}
#else
void gpu_dscal(CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x)
{
	cuda_cublasDscal(n, a, x, 1);
}
#endif

/**
 * \brief Vector scaling in quasi-double 
 *        precision
 * \param n Vector size
 * \param a Scaling factor
 * \param xh,xt The head and tail of the
 *              vector 
 */
#ifdef CUKR_USE_OPENCL
void gpu_qscal(CUKR_INT n, CUKR_DOUBLE a, CUKR_CL_SINGLE* xh,
										  CUKR_CL_SINGLE* xt)
{
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	/* Split a */
	float volatile a0 = (float)a;
	float volatile a1 = (float)(a - a0);

	opencl_kernel_qscal(n, a0, a1, xh, xt);
}
#else
void gpu_qscal(CUKR_INT n, CUKR_DOUBLE a, CUKR_SINGLE* xh,
			   CUKR_SINGLE* xt)
{
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	/* Split a */
	float volatile a0 = (float)a;
	float volatile a1 = (float)(a - a0);
	
	cuda_kernel_qscal(n, a0, a1, xh, xt);
}
#endif

/**
 * \brief Vector scaling in single precision
 * \param n Vector size
 * \param a Scaling factor
 * \param x The vector being scaled
 */

#ifdef CUKR_USE_OPENCL
void gpu_sscal(CUKR_INT n, CUKR_SINGLE a, CUKR_CL_SINGLE* x)
{
	opencl_kernel_sscal(n, a, x);
}
#else
void gpu_sscal(CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x)
{
	cuda_cublasSscal(n, a, x, 1);
}
#endif

// ------------------------------------------------------------------- //

/**
 * \brief Vector copy in double precision
 * \param n Vector size
 * \param x The source vector
 * \param y The destination vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_dcopy(CUKR_INT n, cl_mem* x, cl_mem* y)
{
	opencl_dcopy(n, x, 1, y, 1);
} 
#else
void gpu_dcopy(CUKR_INT n, CUKR_DOUBLE* x, CUKR_DOUBLE* y)
{
	cuda_cublasDcopy(n, x, 1, y, 1);
} 
#endif

/**
 * \brief Vector copy in quasi-double 
 *        precision
 * \param n Vector size
 * \param xh,xt The head and tail of the
 *              source vector 
 * \param yh,yt The head and tail of the
 *              destination vector 
 */
#ifdef CUKR_USE_OPENCL
void gpu_qcopy(CUKR_INT n, cl_mem* xh, cl_mem* xt, 
						   cl_mem* yh, cl_mem* yt)
{
	opencl_scopy(n, xh, 1, yh, 1); 
	opencl_scopy(n, xt, 1, yt, 1);
}						   	
#else
void gpu_qcopy(CUKR_INT n, CUKR_SINGLE* xh, CUKR_SINGLE* xt, 
						   CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	cuda_cublasScopy(n, xh, 1, yh, 1); 
	cuda_cublasScopy(n, xt, 1, yt, 1);
}
#endif

/**
 * \brief Vector copy in single precision
 * \param n Vector size
 * \param x The source vector
 * \param y The destination vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_scopy(CUKR_INT n, cl_mem* x, cl_mem* y)
{
	opencl_scopy(n, x, 1, y, 1);
} 
#else
void gpu_scopy(CUKR_INT n, CUKR_SINGLE* x, CUKR_SINGLE* y)
{
	cuda_cublasScopy(n, x, 1, y, 1);
} 
#endif

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_dspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, 
                   CUKR_CL_INT* idx, CUKR_CL_DOUBLE* val, CUKR_CL_DOUBLE* x, CUKR_DOUBLE beta, 
                   CUKR_CL_DOUBLE* y)
{
	/* Call the kernel */
	opencl_dspmv_csr(rows, cols, nz, alpha, ptr, idx, val, x, beta, y);
}
#else
void gpu_dspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
                   CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x, CUKR_DOUBLE beta, 
                   CUKR_DOUBLE* y)
{
	/* Call the kernel */
	cuda_dspmv_csr(rows, cols, nz, alpha, ptr, idx, val, x, beta, y);
} 
#endif
                   
/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector 
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param beta Scale factor for y
 * \param yh,yt Head and tail for the result
 *              vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_qspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, 
             	   CUKR_CL_INT* idx, CUKR_CL_SINGLE* valh, CUKR_CL_SINGLE* valt, 
               	   CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt, CUKR_DOUBLE beta, 
               	   CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt)
{
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	/* Split alpha and beta */
	float volatile a0 = (CUKR_SINGLE)alpha;
	float volatile a1 = (CUKR_SINGLE)(alpha - a0);
	float volatile b0 = (CUKR_SINGLE)beta;
	float volatile b1 = (CUKR_SINGLE)(beta - b0); 
	
	/* Call the kernel */
	opencl_qspmv_csr(rows, cols , nz, a0, a1, ptr, idx, valh, valt, xh, xt, b0, b1, yh, yt);
}
#else
void gpu_qspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
             	   CUKR_INT* idx, CUKR_SINGLE* valh, CUKR_SINGLE* valt, 
               	   CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, 
               	   CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	/* Split alpha and beta */
	float volatile a0 = (CUKR_SINGLE)alpha;
	float volatile a1 = (CUKR_SINGLE)(alpha - a0);
	float volatile b0 = (CUKR_SINGLE)beta;
	float volatile b1 = (CUKR_SINGLE)(beta - b0); 
	
	/* Call the kernel */
	cuda_qspmv_csr(rows, cols , nz, a0, a1, ptr, idx, valh, valt, xh, xt, b0, b1, yh, yt);
}
#endif
                   
/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_sspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_CL_INT* ptr, 
                   CUKR_CL_INT* idx, CUKR_CL_SINGLE* val, CUKR_CL_SINGLE* x, CUKR_SINGLE beta, 
                   CUKR_CL_SINGLE* y)
{
	/* Call the kernel */
	opencl_sspmv_csr(rows, cols, nz, alpha, ptr, idx, val, x, beta, y);
}
#else
void gpu_sspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_INT* ptr, 
                   CUKR_INT* idx, CUKR_SINGLE* val, CUKR_SINGLE* x, CUKR_SINGLE beta, 
                   CUKR_SINGLE* y)
{
	/* Call the kernel */
	cuda_sspmv_csr(rows, cols, nz, alpha, ptr, idx, val, x, beta, y);
}
#endif
                   	                            
// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR4 format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_dspmv_csr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, 
                    CUKR_CL_INT* idx, CUKR_CL_DOUBLE* val, CUKR_CL_DOUBLE* x, CUKR_DOUBLE beta, 
                    CUKR_CL_DOUBLE* y)
{
	/* Run the kernel */
	opencl_dspmv_csr4(rows, cols, nz, alpha, ptr, idx, val, x, beta, y);
}
#else
void gpu_dspmv_csr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
                    CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x, CUKR_DOUBLE beta, 
                    CUKR_DOUBLE* y)
{
	/* Run the kernel */
	cuda_dspmv_csr4(rows, cols, nz, alpha, ptr, idx, val, x, beta, y);
}
#endif
                   
/**
 * \brief Sparse matrix vector multiply in 
 *        CSR4 format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector 
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param beta Scale factor for y
 * \param yh,yt Head and tail for the result
 *              vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_qspmv_csr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, 
             	    CUKR_CL_INT* idx, CUKR_CL_SINGLE* valh, CUKR_CL_SINGLE* valt, 
               	    CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt, CUKR_DOUBLE beta, 
               	    CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt)
{
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	/* Split alpha and beta */
	float volatile a0 = (CUKR_SINGLE)alpha;
	float volatile a1 = (CUKR_SINGLE)(alpha - a0);
	float volatile b0 = (CUKR_SINGLE)beta;
	float volatile b1 = (CUKR_SINGLE)(beta - b0); 
	
	/* Call the kernel */
	opencl_qspmv_csr4(rows, cols, nz, a0, a1, ptr, idx, valh, valt, xh, xt, b0, b1, yh, yt);
}
#else
void gpu_qspmv_csr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
             	    CUKR_INT* idx, CUKR_SINGLE* valh, CUKR_SINGLE* valt, 
               	    CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, 
               	    CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	/* Split alpha and beta */
	float volatile a0 = (CUKR_SINGLE)alpha;
	float volatile a1 = (CUKR_SINGLE)(alpha - a0);
	float volatile b0 = (CUKR_SINGLE)beta;
	float volatile b1 = (CUKR_SINGLE)(beta - b0); 
	
	/* Call the kernel */
	cuda_qspmv_csr4(rows, cols, nz, a0, a1, ptr, idx, valh, valt, xh, xt, b0, b1, yh, yt);
} 
#endif
                   
/**
 * \brief Sparse matrix vector multiply in 
 *        CSR4 format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_sspmv_csr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_CL_INT* ptr, 
                    CUKR_CL_INT* idx, CUKR_CL_SINGLE* val, CUKR_CL_SINGLE* x, CUKR_SINGLE beta, 
                    CUKR_CL_SINGLE* y)
{
	/* Run the kernel */
	opencl_sspmv_csr4(rows, cols, nz, alpha, ptr, idx, val, x, beta, y);
}
#else
void gpu_sspmv_csr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_INT* ptr, 
                    CUKR_INT* idx, CUKR_SINGLE* val, CUKR_SINGLE* x, CUKR_SINGLE beta, 
                    CUKR_SINGLE* y)
{
	/* Run the kernel */
	cuda_sspmv_csr4(rows, cols, nz, alpha, ptr, idx, val, x, beta, y);
}
#endif


// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR format, in double precision
 * \param rows,cols Matrix size
 * \param r,c Block size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_dspmv_bcsr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, 
                   CUKR_CL_INT* idx, CUKR_CL_DOUBLE* val, CUKR_CL_DOUBLE* x, CUKR_DOUBLE beta, 
                   CUKR_CL_DOUBLE* y)
{
	/* Call the kernel */
	opencl_dspmv_bcsr(rows, cols, nz, r, c, alpha, ptr, idx, val, x, beta, y);
}
#else
void gpu_dspmv_bcsr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
					CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x, CUKR_DOUBLE beta, 
					CUKR_DOUBLE* y)
{
	/* Call the kernel */
	cuda_dspmv_bcsr(rows, cols, nz, r, c, alpha, ptr, idx, val, x, beta, y);
}
#endif
                   
/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param r,c Block size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector 
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param beta Scale factor for y
 * \param yh,yt Head and tail for the result
 *              vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_qspmv_bcsr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, 
             	   CUKR_CL_INT* idx, CUKR_CL_SINGLE* valh, CUKR_CL_SINGLE* valt, 
               	   CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt, CUKR_DOUBLE beta, 
               	   CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt)
{
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	/* Split alpha and beta */
	float volatile a0 = (CUKR_SINGLE)alpha;
	float volatile a1 = (CUKR_SINGLE)(alpha - a0);
	float volatile b0 = (CUKR_SINGLE)beta;
	float volatile b1 = (CUKR_SINGLE)(beta - b0); 
	
	/* Call the kernel */
	
		opencl_qspmv_bcsr(rows, cols , nz, r, c, a0, a1, ptr, idx, valh, valt, xh, xt, b0, b1, yh, yt);
	/* Not coded */
}
#else
void gpu_qspmv_bcsr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
					CUKR_INT* idx, CUKR_SINGLE* valh, CUKR_SINGLE* valt, 
					CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, 
					CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	/* Split alpha and beta */
	float volatile a0 = (CUKR_SINGLE)alpha;
	float volatile a1 = (CUKR_SINGLE)(alpha - a0);
	float volatile b0 = (CUKR_SINGLE)beta;
	float volatile b1 = (CUKR_SINGLE)(beta - b0); 
	
	/* Call the kernel */
	
	cuda_qspmv_bcsr(rows, cols , nz, r, c, a0, a1, ptr, idx, valh, valt, xh, xt, b0, b1, yh, yt);
	/* Not coded */
}  
#endif
                   
/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR format, in single precision
 * \param rows,cols Matrix size
 * \param r,c Block size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_sspmv_bcsr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_CL_INT* ptr, 
                   CUKR_CL_INT* idx, CUKR_CL_SINGLE* val, CUKR_CL_SINGLE* x, CUKR_SINGLE beta, 
                   CUKR_CL_SINGLE* y)
{
	/* Call the kernel */
	opencl_sspmv_bcsr(rows, cols, nz, r, c, alpha, ptr, idx, val, x, beta, y);
}
#else
void gpu_sspmv_bcsr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_INT* ptr, 
					CUKR_INT* idx, CUKR_SINGLE* val, CUKR_SINGLE* x, CUKR_SINGLE beta, 
					CUKR_SINGLE* y)
{
	/* Call the kernel */
	cuda_sspmv_bcsr(rows, cols, nz, r, c, alpha, ptr, idx, val, x, beta, y);
}
#endif


// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR4 format, in double precision
 * \param rows,cols Matrix size
 * \param r,c Block size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_dspmv_bcsr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, 
					 CUKR_CL_INT* idx, CUKR_CL_DOUBLE* val, CUKR_CL_DOUBLE* x, CUKR_DOUBLE beta, 
					 CUKR_CL_DOUBLE* y)
{
	/* Call the kernel */
	opencl_dspmv_bcsr4(rows, cols, nz, r, c, alpha, ptr, idx, val, x, beta, y);
}
#else
void gpu_dspmv_bcsr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
					 CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x, CUKR_DOUBLE beta, 
					 CUKR_DOUBLE* y)
{
	/* Call the kernel */
	cuda_dspmv_bcsr4(rows, cols, nz, r, c, alpha, ptr, idx, val, x, beta, y);
}
#endif

/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR4 format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param r,c Block size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector 
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param beta Scale factor for y
 * \param yh,yt Head and tail for the result
 *              vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_qspmv_bcsr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, 
             	   CUKR_CL_INT* idx, CUKR_CL_SINGLE* valh, CUKR_CL_SINGLE* valt, 
               	   CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt, CUKR_DOUBLE beta, 
               	   CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt)
{
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	/* Split alpha and beta */
	float volatile a0 = (CUKR_SINGLE)alpha;
	float volatile a1 = (CUKR_SINGLE)(alpha - a0);
	float volatile b0 = (CUKR_SINGLE)beta;
	float volatile b1 = (CUKR_SINGLE)(beta - b0); 
	
	/* Call the kernel */
	opencl_qspmv_bcsr4(rows, cols , nz, r, c, a0, a1, ptr, idx, valh, valt, xh, xt, b0, b1, yh, yt);
	/* Not coded */
}
#else
void gpu_qspmv_bcsr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
					 CUKR_INT* idx, CUKR_SINGLE* valh, CUKR_SINGLE* valt, 
					 CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, 
					 CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	/*
	 * Volatile keywork must be used to ensure
	 * the compiler does not mess with the 
	 * double-single method 
	 */
	
	/* Split alpha and beta */
	float volatile a0 = (CUKR_SINGLE)alpha;
	float volatile a1 = (CUKR_SINGLE)(alpha - a0);
	float volatile b0 = (CUKR_SINGLE)beta;
	float volatile b1 = (CUKR_SINGLE)(beta - b0); 
	
	/* Call the kernel */
	cuda_qspmv_bcsr4(rows, cols , nz, r, c, a0, a1, ptr, idx, valh, valt, xh, xt, b0, b1, yh, yt);
	/* Not coded */
}
#endif

/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR4 format, in double precision
 * \param rows,cols Matrix size
 * \param r,c Block size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
#ifdef	CUKR_USE_OPENCL
void gpu_sspmv_bcsr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_CL_INT* ptr, 
                   CUKR_CL_INT* idx, CUKR_CL_SINGLE* val, CUKR_CL_SINGLE* x, CUKR_SINGLE beta, 
                   CUKR_CL_SINGLE* y)
{
	/* Call the kernel */
		opencl_sspmv_bcsr4(rows, cols, nz, r, c, alpha, ptr, idx, val, x, beta, y);
}
#else
void gpu_sspmv_bcsr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_INT* ptr, 
					 CUKR_INT* idx, CUKR_SINGLE* val, CUKR_SINGLE* x, CUKR_SINGLE beta, 
					 CUKR_SINGLE* y)
{
	/* Call the kernel */
	cuda_sspmv_bcsr4(rows, cols, nz, r, c, alpha, ptr, idx, val, x, beta, y);
}
#endif

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_dspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                   CUKR_CL_INT* ell_idx, CUKR_CL_DOUBLE* ell_val, CUKR_CL_INT* csr_ptr, CUKR_CL_INT* csr_idx, CUKR_CL_DOUBLE* csr_val,
                   CUKR_CL_DOUBLE* x, CUKR_DOUBLE beta, CUKR_CL_DOUBLE* y)
{
	opencl_dspmv_hyb(rows, cols, nz, alpha, ell_nz_row, ell_stride, csr_nz, ell_idx, ell_val, csr_ptr, csr_idx, csr_val, x, beta, y);
}
#else
void gpu_dspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                   CUKR_INT* ell_idx, CUKR_DOUBLE* ell_val, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_DOUBLE* csr_val,
                   CUKR_DOUBLE* x, CUKR_DOUBLE beta, CUKR_DOUBLE* y)
{
	cuda_dspmv_hyb(rows, cols, nz, alpha, ell_nz_row, ell_stride, csr_nz, ell_idx, ell_val, csr_ptr, csr_idx, csr_val, x, beta, y);
}
#endif

/**
 * \brief Wrapper for QSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
#ifdef CUKR_USE_OPENCL
void gpu_qspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
	               CUKR_CL_INT* ell_idx, CUKR_CL_SINGLE* ell_valh, CUKR_CL_SINGLE* ell_valt, CUKR_CL_INT* csr_ptr, CUKR_CL_INT* csr_idx, CUKR_CL_SINGLE* csr_valh, CUKR_CL_SINGLE* csr_valt,
	               CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt, CUKR_DOUBLE beta, CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt)
{
	/* Split alpha and beta */
	float volatile a0 = (CUKR_SINGLE)alpha;
	float volatile a1 = (CUKR_SINGLE)(alpha - a0);
	float volatile b0 = (CUKR_SINGLE)beta;
	float volatile b1 = (CUKR_SINGLE)(beta - b0);

	opencl_qspmv_hyb(rows, cols, nz, a0, a1, ell_nz_row, ell_stride, csr_nz, ell_idx, ell_valh, ell_valt, csr_ptr, csr_idx, csr_valh, csr_valt, xh, xt, b0, b1, yh, yt);
}
#else
void gpu_qspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
	               CUKR_INT* ell_idx, CUKR_SINGLE* ell_valh, CUKR_SINGLE* ell_valt, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_SINGLE* csr_valh, CUKR_SINGLE* csr_valt,
	               CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	/* Split alpha and beta */
	float volatile a0 = (CUKR_SINGLE)alpha;
	float volatile a1 = (CUKR_SINGLE)(alpha - a0);
	float volatile b0 = (CUKR_SINGLE)beta;
	float volatile b1 = (CUKR_SINGLE)(beta - b0);
	
	cuda_qspmv_hyb(rows, cols, nz, a0, a1, ell_nz_row, ell_stride, csr_nz, ell_idx, ell_valh, ell_valt, csr_ptr, csr_idx, csr_valh, csr_valt, xh, xt, b0, b1, yh, yt);
}
#endif

/**
 * \brief Wrapper for SSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_sspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                   CUKR_CL_INT* ell_idx, CUKR_CL_SINGLE* ell_val, CUKR_CL_INT* csr_ptr, CUKR_CL_INT* csr_idx, CUKR_CL_SINGLE* csr_val,
                   CUKR_CL_SINGLE* x, CUKR_SINGLE beta, CUKR_CL_SINGLE* y)
{
	opencl_sspmv_hyb(rows, cols, nz, alpha, ell_nz_row, ell_stride, csr_nz, ell_idx, ell_val, csr_ptr, csr_idx, csr_val, x, beta, y);
}
#else
void gpu_sspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                   CUKR_INT* ell_idx, CUKR_SINGLE* ell_val, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_SINGLE* csr_val,
                   CUKR_SINGLE* x, CUKR_SINGLE beta, CUKR_SINGLE* y)
{
	cuda_sspmv_hyb(rows, cols, nz, alpha, ell_nz_row, ell_stride, csr_nz, ell_idx, ell_val, csr_ptr, csr_idx, csr_val, x, beta, y);
}
#endif

#endif /* CUKR_USE_GPU */


// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void gpu_dspmv_bhyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
                   CUKR_INT* bell_idx, CUKR_DOUBLE* bell_val, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_DOUBLE* bcsr_val,
                   CUKR_DOUBLE* x, CUKR_DOUBLE beta, CUKR_DOUBLE* y)
{
	#ifdef CUKR_USE_CUDA
		cuda_dspmv_bhyb(rows, cols, nz, r, c, alpha, bell_nz_row, bell_stride, bcsr_nz, bell_idx, bell_val, bcsr_ptr, bcsr_idx, bcsr_val, x, beta, y);
	#endif
}

/**
 * \brief Wrapper for QSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
void gpu_qspmv_bhyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
	               CUKR_INT* ell_idx, CUKR_SINGLE* ell_valh, CUKR_SINGLE* ell_valt, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_SINGLE* csr_valh, CUKR_SINGLE* csr_valt,
	               CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	/* Split alpha and beta */
	float volatile a0 = (CUKR_SINGLE)alpha;
	float volatile a1 = (CUKR_SINGLE)(alpha - a0);
	float volatile b0 = (CUKR_SINGLE)beta;
	float volatile b1 = (CUKR_SINGLE)(beta - b0);

	#ifdef CUKR_USE_CUDA
		cuda_qspmv_bhyb(rows, cols, nz, r, c, a0, a1, ell_nz_row, ell_stride, csr_nz, ell_idx, ell_valh, ell_valt, csr_ptr, csr_idx, csr_valh, csr_valt, xh, xt, b0, b1, yh, yt);
	#endif
}

/**
 * \brief Wrapper for SSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void gpu_sspmv_bhyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
                   CUKR_INT* bell_idx, CUKR_SINGLE* bell_val, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_SINGLE* bcsr_val,
                   CUKR_SINGLE* x, CUKR_SINGLE beta, CUKR_SINGLE* y)
{
	#ifdef CUKR_USE_CUDA
		cuda_sspmv_bhyb(rows, cols, nz, r, c, alpha, bell_nz_row, bell_stride, bcsr_nz, bell_idx, bell_val, bcsr_ptr, bcsr_idx, bcsr_val, x, beta, y);
	#endif
}


// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_BHYB4
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void gpu_dspmv_bhyb4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
                   CUKR_INT* bell_idx, CUKR_DOUBLE* bell_val, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_DOUBLE* bcsr_val,
                   CUKR_DOUBLE* x, CUKR_DOUBLE beta, CUKR_DOUBLE* y)
{
	#ifdef CUKR_USE_CUDA
		cuda_dspmv_bhyb4(rows, cols, nz, r, c, alpha, bell_nz_row, bell_stride, bcsr_nz, bell_idx, bell_val, bcsr_ptr, bcsr_idx, bcsr_val, x, beta, y);
	#endif
}

/**
 * \brief Wrapper for QSPMV_BHYB4
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
void gpu_qspmv_bhyb4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
	               CUKR_INT* ell_idx, CUKR_SINGLE* ell_valh, CUKR_SINGLE* ell_valt, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_SINGLE* csr_valh, CUKR_SINGLE* csr_valt,
	               CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt)
{
	/* Split alpha and beta */
	float volatile a0 = (CUKR_SINGLE)alpha;
	float volatile a1 = (CUKR_SINGLE)(alpha - a0);
	float volatile b0 = (CUKR_SINGLE)beta;
	float volatile b1 = (CUKR_SINGLE)(beta - b0);

	#ifdef CUKR_USE_CUDA
		cuda_qspmv_bhyb4(rows, cols, nz, r, c, a0, a1, ell_nz_row, ell_stride, csr_nz, ell_idx, ell_valh, ell_valt, csr_ptr, csr_idx, csr_valh, csr_valt, xh, xt, b0, b1, yh, yt);
	#endif
}

/**
 * \brief Wrapper for SSPMV_BHYB4
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void gpu_sspmv_bhyb4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
                   CUKR_INT* bell_idx, CUKR_SINGLE* bell_val, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_SINGLE* bcsr_val,
                   CUKR_SINGLE* x, CUKR_SINGLE beta, CUKR_SINGLE* y)
{
	#ifdef CUKR_USE_CUDA
		cuda_sspmv_bhyb4(rows, cols, nz, r, c, alpha, bell_nz_row, bell_stride, bcsr_nz, bell_idx, bell_val, bcsr_ptr, bcsr_idx, bcsr_val, x, beta, y);
	#endif
}


// ------------------------------------------------------------------- //


/**
 * \brief Create a generic blas object by
 *        binding the above functions to a
 * 		  CUKR_BLAS structure 
 * \param blas pointer to a blas structure
 */
void blas_create_gpu(CUKR_BLAS blas)
#ifndef CUKR_USE_GPU
{
	CUKR_FATAL("GPU support was not compiled in the library");
}
#else 
{
	/* Set the BLAS name */
#ifdef CUKR_USE_OPENCL
	sprintf(blas->name, "CLBLAS");
#else
	sprintf(blas->name, "GPUBLAS");
#endif
	
	/* Set the BLAS location */
	blas->location = CUKR_LOCATION_GPU;
	
	/* Set the bindings */
#ifdef CUKR_USE_OPENCL
	blas->bind.cl_ddot = gpu_ddot;
	blas->bind.cl_qdot = gpu_qdot;
	blas->bind.cl_sdot = gpu_sdot;
	
	blas->bind.cl_daxpy = gpu_daxpy;
	blas->bind.cl_qaxpy = gpu_qaxpy;
	blas->bind.cl_saxpy = gpu_saxpy;
	
	blas->bind.cl_daypx = gpu_daypx;
	blas->bind.cl_qaypx = gpu_qaypx;
	blas->bind.cl_saypx = gpu_saypx;
	
	blas->bind.cl_dscal = gpu_dscal;
	blas->bind.cl_qscal = gpu_qscal;
	blas->bind.cl_sscal = gpu_sscal;
	
	blas->bind.cl_dcopy = gpu_dcopy;
	blas->bind.cl_qcopy = gpu_qcopy;
	blas->bind.cl_scopy = gpu_scopy;
	
	blas->bind.cl_dspmv_csr = gpu_dspmv_csr;
	blas->bind.cl_qspmv_csr = gpu_qspmv_csr;
	blas->bind.cl_sspmv_csr = gpu_sspmv_csr;	
	
	blas->bind.cl_dspmv_csr4 = gpu_dspmv_csr4;
	blas->bind.cl_qspmv_csr4 = gpu_qspmv_csr4;
	blas->bind.cl_sspmv_csr4 = gpu_sspmv_csr4;
 	
	blas->bind.cl_dspmv_hyb = gpu_dspmv_hyb;
	blas->bind.cl_qspmv_hyb = gpu_qspmv_hyb;
	blas->bind.cl_sspmv_hyb = gpu_sspmv_hyb;
	
	blas->bind.cl_dspmv_bcsr = gpu_dspmv_bcsr;
	blas->bind.cl_qspmv_bcsr = gpu_qspmv_bcsr;
	blas->bind.cl_sspmv_bcsr = gpu_sspmv_bcsr;
	
	blas->bind.cl_dspmv_bcsr4 = gpu_dspmv_bcsr4;
	blas->bind.cl_qspmv_bcsr4 = gpu_qspmv_bcsr4;
	blas->bind.cl_sspmv_bcsr4 = gpu_sspmv_bcsr4;
#else
	blas->bind.ddot = gpu_ddot;
	blas->bind.qdot = gpu_qdot;
	blas->bind.sdot = gpu_sdot;
	
	blas->bind.daxpy = gpu_daxpy;
	blas->bind.qaxpy = gpu_qaxpy;
	blas->bind.saxpy = gpu_saxpy;
	
	blas->bind.daypx = gpu_daypx;
	blas->bind.qaypx = gpu_qaypx;
	blas->bind.saypx = gpu_saypx;
	
	blas->bind.dscal = gpu_dscal;
	blas->bind.qscal = gpu_qscal;
	blas->bind.sscal = gpu_sscal;
	
	blas->bind.dcopy = gpu_dcopy;
	blas->bind.qcopy = gpu_qcopy;
	blas->bind.scopy = gpu_scopy;
	
	blas->bind.dspmv_csr = gpu_dspmv_csr;
	blas->bind.qspmv_csr = gpu_qspmv_csr;
	blas->bind.sspmv_csr = gpu_sspmv_csr;	
	
	blas->bind.dspmv_csr4 = gpu_dspmv_csr4;
	blas->bind.qspmv_csr4 = gpu_qspmv_csr4;
	blas->bind.sspmv_csr4 = gpu_sspmv_csr4;
	
	blas->bind.dspmv_bcsr = gpu_dspmv_bcsr;
	blas->bind.qspmv_bcsr = gpu_qspmv_bcsr;
	blas->bind.sspmv_bcsr = gpu_sspmv_bcsr;
	
	blas->bind.dspmv_bcsr4 = gpu_dspmv_bcsr4;
	blas->bind.qspmv_bcsr4 = gpu_qspmv_bcsr4;
	blas->bind.sspmv_bcsr4 = gpu_sspmv_bcsr4;
	
	blas->bind.dspmv_hyb = gpu_dspmv_hyb;
	blas->bind.qspmv_hyb = gpu_qspmv_hyb;
	blas->bind.sspmv_hyb = gpu_sspmv_hyb;
#endif
	
	blas->bind.dspmv_bhyb = gpu_dspmv_bhyb;
	blas->bind.qspmv_bhyb = gpu_qspmv_bhyb;
	blas->bind.sspmv_bhyb = gpu_sspmv_bhyb;
	
	blas->bind.dspmv_bhyb4 = gpu_dspmv_bhyb4;
	blas->bind.qspmv_bhyb4 = gpu_qspmv_bhyb4;
	blas->bind.sspmv_bhyb4 = gpu_sspmv_bhyb4;

}  
#endif /* CUKR_USE_GPU */   
