/**
 * @file blas/impl/gpu/gpu_blas.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines wrappers for BLAS functions 
 * on the GPU. 
 * 
 * The functions take as parameter device
 * pointers, so they only act on data that
 * has been PREVIOUSLY COPIED to the device.
 * 
 * The functions are implemented in the 
 * following way:
 * - single
 * 	- BLAS 1: CUBLAS + own
 *  - SpMV: gpublas
 *  - double
 */

#ifndef GPU_BLAS_H_
#define GPU_BLAS_H_

#include "../../blas.h"
#include "../../../init/init_opencl.h"

// ------------------------------------------------------------------- //

/**
 * \brief Dot product in double precision
 * \param n Vector size
 * \param x,y The two vectors whose product 
 *            is being computed 
 */
#ifdef CUKR_USE_OPENCL
CUKR_DOUBLE gpu_ddot(CUKR_INT n, CUKR_CL_DOUBLE* x, CUKR_CL_DOUBLE* y);
#else
CUKR_DOUBLE gpu_ddot(CUKR_INT n, CUKR_DOUBLE* x, CUKR_DOUBLE* y);
#endif

/**
 * \brief Dot product in quasi-double precision
 * \param n Vector size
 * \param n The vector size
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector
 */
#ifdef CUKR_USE_OPENCL
CUKR_DOUBLE gpu_qdot(CUKR_INT n, CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt,
								 CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt);
#else
CUKR_DOUBLE gpu_qdot(CUKR_INT n, CUKR_SINGLE* xh, CUKR_SINGLE* xt,
								 CUKR_SINGLE* yh, CUKR_SINGLE* yt);
#endif

/**
 * \brief Dot product in single precision
 * \param n Vector size
 * \param x,y The two vectors whose product 
 *            is being computed 
 */
#ifdef CUKR_USE_OPENCL
CUKR_SINGLE gpu_sdot(CUKR_INT n, CUKR_CL_SINGLE* x, CUKR_CL_SINGLE* y);
#else
CUKR_SINGLE gpu_sdot(CUKR_INT n, CUKR_SINGLE* x, CUKR_SINGLE* y);
#endif
    
// ------------------------------------------------------------------- //
  
/**
 * \brief AXPY in double precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
#ifdef CUKR_USE_OPENCL
void gpu_daxpy(CUKR_INT n, CUKR_DOUBLE a, CUKR_CL_DOUBLE* x, CUKR_CL_DOUBLE* y);
#else
void gpu_daxpy(CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x, CUKR_DOUBLE* y);
#endif

/**
 * \brief AXPY in quasi-double precision
 * \param n Vector size
 * \param a The constant
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector 
 */
#ifdef CUKR_USE_OPENCL
void gpu_qaxpy(CUKR_INT n, CUKR_DOUBLE a, CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt, 
										  CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt);
#else
void gpu_qaxpy(CUKR_INT n, CUKR_DOUBLE a, CUKR_SINGLE* xh, CUKR_SINGLE* xt, 
										  CUKR_SINGLE* yh, CUKR_SINGLE* yt);
#endif

/**
 * \brief AXPY in single precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
#ifdef CUKR_USE_OPENCL
void gpu_saxpy(CUKR_INT n, CUKR_SINGLE a, CUKR_CL_SINGLE* x, CUKR_CL_SINGLE* y);
#else
void gpu_saxpy(CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x, CUKR_SINGLE* y);
#endif

// ------------------------------------------------------------------- //
  
/**
 * \brief AYPX in double precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
#ifdef CUKR_USE_OPENCL
void gpu_daypx(CUKR_INT n, CUKR_DOUBLE a, CUKR_CL_DOUBLE* cl_x, CUKR_CL_DOUBLE* cl_y);
#else
void gpu_daypx(CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x, CUKR_DOUBLE* y);
#endif

/**
 * \brief AYPX in quasi-double precision
 * \param n Vector size
 * \param a The constant
 * \param xh,xt The head and tail of the
 *              first vector 
 * \param yh,yt The head and tail of the
 *              second vector 
 */
#ifdef CUKR_USE_OPENCL
void gpu_qaypx(CUKR_INT n, CUKR_DOUBLE a, CUKR_CL_SINGLE* cl_xh, CUKR_CL_SINGLE* cl_xt, 
										  CUKR_CL_SINGLE* cl_yh, CUKR_CL_SINGLE* cl_yt);
#else
void gpu_qaypx(CUKR_INT n, CUKR_DOUBLE a, CUKR_SINGLE* xh, CUKR_SINGLE* xt, 
										  CUKR_SINGLE* yh, CUKR_SINGLE* yt);
#endif

/**
 * \brief AYPX in single precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
#ifdef CUKR_USE_OPENCL
void gpu_saypx(CUKR_INT n, CUKR_SINGLE a, CUKR_CL_SINGLE* cl_x, CUKR_CL_SINGLE* cl_y);
#else
void gpu_saypx(CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x, CUKR_SINGLE* y);
#endif
 

// ------------------------------------------------------------------- //

/**
 * \brief Vector scaling in double precision
 * \param n Vector size
 * \param a Scaling factor
 * \param x The vector being scaled
 */
#ifdef CUKR_USE_OPENCL
void gpu_dscal(CUKR_INT n, CUKR_DOUBLE a, CUKR_CL_DOUBLE* x); 
#else
void gpu_dscal(CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x);
#endif

/**
 * \brief Vector scaling in quasi-double 
 *        precision
 * \param n Vector size
 * \param a Scaling factor
 * \param xh,xt The head and tail of the
 *              vector 
 */
#ifdef CUKR_USE_OPENCL
void gpu_qscal(CUKR_INT n, CUKR_DOUBLE a, CUKR_CL_SINGLE* xh,
										  CUKR_CL_SINGLE* xt);
#else
void gpu_qscal(CUKR_INT n, CUKR_DOUBLE a, CUKR_SINGLE* xh,
										  CUKR_SINGLE* xt);
#endif
/**
 * \brief Vector scaling in single precision
 * \param n Vector size
 * \param a Scaling factor
 * \param x The vector being scaled
 */
#ifdef CUKR_USE_OPENCL
void gpu_sscal(CUKR_INT n, CUKR_SINGLE a, CUKR_CL_SINGLE* x); 
#else
void gpu_sscal(CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x);
#endif
// ------------------------------------------------------------------- //

/**
 * \brief Vector copy in double precision
 * \param n Vector size
 * \param x The source vector
 * \param y The destination vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_dcopy(CUKR_INT n, cl_mem* x, cl_mem* y);
#else
void gpu_dcopy(CUKR_INT n, CUKR_DOUBLE* x, CUKR_DOUBLE* y);
#endif

/**
 * \brief Vector copy in quasi-double 
 *        precision
 * \param n Vector size
 * \param xh,xt The head and tail of the
 *              source vector 
 * \param yh,yt The head and tail of the
 *              destination vector 
 */
#ifdef CUKR_USE_OPENCL
void gpu_qcopy(CUKR_INT n, cl_mem* xh, cl_mem* xt, 
						   cl_mem* yh, cl_mem* yt);
#else
void gpu_qcopy(CUKR_INT n, CUKR_SINGLE* xh, CUKR_SINGLE* xt, 
			   CUKR_SINGLE* yh, CUKR_SINGLE* yt);
#endif

/**
 * \brief Vector copy in single precision
 * \param n Vector size
 * \param x The source vector
 * \param y The destination vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_scopy(CUKR_INT n, cl_mem* x, cl_mem* y);
#else
void gpu_scopy(CUKR_INT n, CUKR_SINGLE* x, CUKR_SINGLE* y);
#endif

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_dspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, 
                   CUKR_CL_INT* idx, CUKR_CL_DOUBLE* val, CUKR_CL_DOUBLE* x, CUKR_DOUBLE beta, 
                   CUKR_CL_DOUBLE* y);            
#else
void gpu_dspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
                   CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x, CUKR_DOUBLE beta, 
                   CUKR_DOUBLE* y);
#endif
/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector 
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param beta Scale factor for y
 * \param yh,yt Head and tail for the result
 *              vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_qspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, 
             	   CUKR_CL_INT* idx, CUKR_CL_SINGLE* valh, CUKR_CL_SINGLE* valt, 
               	   CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt, CUKR_DOUBLE beta, 
               	   CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt);  
#else
void gpu_qspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
             	   CUKR_INT* idx, CUKR_SINGLE* valh, CUKR_SINGLE* valt, 
               	   CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, 
               	   CUKR_SINGLE* yh, CUKR_SINGLE* yt); 
#endif                   
/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_sspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_CL_INT* ptr, 
                   CUKR_CL_INT* idx, CUKR_CL_SINGLE* val, CUKR_CL_SINGLE* x, CUKR_SINGLE beta, 
                   CUKR_CL_SINGLE* y);                               
#else
void gpu_sspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_INT* ptr, 
                   CUKR_INT* idx, CUKR_SINGLE* val, CUKR_SINGLE* x, CUKR_SINGLE beta, 
                   CUKR_SINGLE* y); 
#endif              

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR4 format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_dspmv_csr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, 
                    CUKR_CL_INT* idx, CUKR_CL_DOUBLE* val, CUKR_CL_DOUBLE* x, CUKR_DOUBLE beta, 
                    CUKR_CL_DOUBLE* y);
#else
void gpu_dspmv_csr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
                    CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x, CUKR_DOUBLE beta, 
                    CUKR_DOUBLE* y);
#endif
/**
 * \brief Sparse matrix vector multiply in 
 *        CSR4 format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector 
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param beta Scale factor for y
 * \param yh,yt Head and tail for the result
 *              vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_qspmv_csr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, 
             	    CUKR_CL_INT* idx, CUKR_CL_SINGLE* valh, CUKR_CL_SINGLE* valt, 
               	    CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt, CUKR_DOUBLE beta, 
               	    CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt);
#else
void gpu_qspmv_csr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
             	    CUKR_INT* idx, CUKR_SINGLE* valh, CUKR_SINGLE* valt, 
               	    CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, 
               	    CUKR_SINGLE* yh, CUKR_SINGLE* yt);
#endif
/**
 * \brief Sparse matrix vector multiply in 
 *        CSR4 format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_sspmv_csr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_CL_INT* ptr, 
                    CUKR_CL_INT* idx, CUKR_CL_SINGLE* val, CUKR_CL_SINGLE* x, CUKR_SINGLE beta, 
                    CUKR_CL_SINGLE* y);
#else
void gpu_sspmv_csr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_INT* ptr, 
                    CUKR_INT* idx, CUKR_SINGLE* val, CUKR_SINGLE* x, CUKR_SINGLE beta, 
                    CUKR_SINGLE* y);
#endif
// ------------------------------------------------------------------- //

/* Yusa adds BCSR in winter, 2009. */

/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR format, in double precision
 * \param rows,cols Matrix size
 * \param r,c Block size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_dspmv_bcsr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, 
                   CUKR_CL_INT* idx, CUKR_CL_DOUBLE* val, CUKR_CL_DOUBLE* x, CUKR_DOUBLE beta, 
                   CUKR_CL_DOUBLE* y);
#else
void gpu_dspmv_bcsr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
					CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x, CUKR_DOUBLE beta, 
					CUKR_DOUBLE* y);
#endif
                   
/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR format, in quasi-double precision
 * \param rows,cols Matrix size
 * \param r,c Block size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector 
 * \param idx Column index vector
 * \param valh,valt Head and tail for the
 *                   value vector 
 * \param xh,xt Head and tail for the vector
 *              being multiplied
 * \param beta Scale factor for y
 * \param yh,yt Head and tail for the result
 *              vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_qspmv_bcsr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, 
             	   CUKR_CL_INT* idx, CUKR_CL_SINGLE* valh, CUKR_CL_SINGLE* valt, 
               	   CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt, CUKR_DOUBLE beta, 
               	   CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt);
#else
void gpu_qspmv_bcsr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
					CUKR_INT* idx, CUKR_SINGLE* valh, CUKR_SINGLE* valt, 
					CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, 
					CUKR_SINGLE* yh, CUKR_SINGLE* yt);
#endif
/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR format, in double precision
 * \param rows,cols Matrix size
 * \param r,c Block size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_sspmv_bcsr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_CL_INT* ptr, 
                   CUKR_CL_INT* idx, CUKR_CL_SINGLE* val, CUKR_CL_SINGLE* x, CUKR_SINGLE beta, 
                   CUKR_CL_SINGLE* y);
#else
void gpu_sspmv_bcsr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_INT* ptr, 
					CUKR_INT* idx, CUKR_SINGLE* val, CUKR_SINGLE* x, CUKR_SINGLE beta, 
					CUKR_SINGLE* y);
#endif

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR4 format, in double precision
 * \param rows,cols Matrix size
 * \param r,c Block size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_dspmv_bcsr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, 
                   CUKR_CL_INT* idx, CUKR_CL_DOUBLE* val, CUKR_CL_DOUBLE* x, CUKR_DOUBLE beta, 
                   CUKR_CL_DOUBLE* y);
#else
void gpu_dspmv_bcsr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
					 CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x, CUKR_DOUBLE beta, 
					 CUKR_DOUBLE* y);
#endif

/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR4 format, in quasi double precision
 * \param rows,cols Matrix size
 * \param r,c Block size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_qspmv_bcsr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_CL_INT* ptr, 
					CUKR_CL_INT* idx, CUKR_CL_SINGLE* valh, CUKR_CL_SINGLE* valt, 
					CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt, CUKR_DOUBLE beta, 
					CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt);
#else
void gpu_qspmv_bcsr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
					 CUKR_INT* idx, CUKR_SINGLE* valh, CUKR_SINGLE* valt, 
					 CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, 
					 CUKR_SINGLE* yh, CUKR_SINGLE* yt);
#endif

/**
 * \brief Sparse matrix vector multiply in 
 *        BCSR4 format, in single precision
 * \param rows,cols Matrix size
 * \param r,c Block size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_sspmv_bcsr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_CL_INT* ptr, 
                   CUKR_CL_INT* idx, CUKR_CL_SINGLE* val, CUKR_CL_SINGLE* x, CUKR_SINGLE beta, 
                   CUKR_CL_SINGLE* y);
#else
void gpu_sspmv_bcsr4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_INT* ptr, 
					 CUKR_INT* idx, CUKR_SINGLE* val, CUKR_SINGLE* x, CUKR_SINGLE beta, 
					 CUKR_SINGLE* y);
#endif

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_dspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                   CUKR_CL_INT* ell_idx, CUKR_CL_DOUBLE* ell_val, CUKR_CL_INT* csr_ptr, CUKR_CL_INT* csr_idx, CUKR_CL_DOUBLE* csr_val,
                   CUKR_CL_DOUBLE* x, CUKR_DOUBLE beta, CUKR_CL_DOUBLE* y);
#else
void gpu_dspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                   CUKR_INT* ell_idx, CUKR_DOUBLE* ell_val, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_DOUBLE* csr_val,
                   CUKR_DOUBLE* x, CUKR_DOUBLE beta, CUKR_DOUBLE* y);
#endif

/**
 * \brief Wrapper for QSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
#ifdef CUKR_USE_OPENCL
void gpu_qspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
	               CUKR_CL_INT* ell_idx, CUKR_CL_SINGLE* ell_valh, CUKR_CL_SINGLE* ell_valt, CUKR_CL_INT* csr_ptr, CUKR_CL_INT* csr_idx, CUKR_CL_SINGLE* csr_valh, CUKR_CL_SINGLE* csr_valt,
	               CUKR_CL_SINGLE* xh, CUKR_CL_SINGLE* xt, CUKR_DOUBLE beta, CUKR_CL_SINGLE* yh, CUKR_CL_SINGLE* yt);
#else
void gpu_qspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
	               CUKR_INT* ell_idx, CUKR_SINGLE* ell_valh, CUKR_SINGLE* ell_valt, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_SINGLE* csr_valh, CUKR_SINGLE* csr_valt,
	               CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt);
#endif



/**
 * \brief Wrapper for SSPMV_HYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
#ifdef CUKR_USE_OPENCL
void gpu_sspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                   CUKR_CL_INT* ell_idx, CUKR_CL_SINGLE* ell_val, CUKR_CL_INT* csr_ptr, CUKR_CL_INT* csr_idx, CUKR_CL_SINGLE* csr_val,
                   CUKR_CL_SINGLE* x, CUKR_SINGLE beta, CUKR_CL_SINGLE* y);
#else
void gpu_sspmv_hyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
                   CUKR_INT* ell_idx, CUKR_SINGLE* ell_val, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_SINGLE* csr_val,
                   CUKR_SINGLE* x, CUKR_SINGLE beta, CUKR_SINGLE* y);
#endif

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void gpu_dspmv_bhyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
                   CUKR_INT* bell_idx, CUKR_DOUBLE* bell_val, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_DOUBLE* bcsr_val,
                   CUKR_DOUBLE* x, CUKR_DOUBLE beta, CUKR_DOUBLE* y);


/**
 * \brief Wrapper for QSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
void gpu_qspmv_bhyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
	               CUKR_INT* ell_idx, CUKR_SINGLE* ell_valh, CUKR_SINGLE* ell_valt, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_SINGLE* csr_valh, CUKR_SINGLE* csr_valt,
	               CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt);

/**
 * \brief Wrapper for SSPMV_BHYB
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void gpu_sspmv_bhyb(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
                   CUKR_INT* bell_idx, CUKR_SINGLE* bell_val, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_SINGLE* bcsr_val,
                   CUKR_SINGLE* x, CUKR_SINGLE beta, CUKR_SINGLE* y);

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DSPMV_BHYB4
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void gpu_dspmv_bhyb4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
                   CUKR_INT* bell_idx, CUKR_DOUBLE* bell_val, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_DOUBLE* bcsr_val,
                   CUKR_DOUBLE* x, CUKR_DOUBLE beta, CUKR_DOUBLE* y);

/**
 * \brief Wrapper for QSPMV_BHYB4
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector (head, tail)
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector (head, tail)
 * \param x Vector being multiplied (head, tail)
 * \param beta Scale factor for y
 * \param y Result vector (head, tail)
 */
void gpu_qspmv_bhyb4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_DOUBLE alpha, CUKR_INT ell_nz_row, CUKR_INT ell_stride, CUKR_INT csr_nz,
	               CUKR_INT* ell_idx, CUKR_SINGLE* ell_valh, CUKR_SINGLE* ell_valt, CUKR_INT* csr_ptr, CUKR_INT* csr_idx, CUKR_SINGLE* csr_valh, CUKR_SINGLE* csr_valt,
	               CUKR_SINGLE* xh, CUKR_SINGLE* xt, CUKR_DOUBLE beta, CUKR_SINGLE* yh, CUKR_SINGLE* yt);

/**
 * \brief Wrapper for SSPMV_BHYB4
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scale factor for Ax
 * \param ell_nz_row,ell_stride ELL dimensions
 * \param csr_nz CSR dimension
 * \param ell_idx ELL column index vector
 * \param ell_val ELL value vector
 * \param csr_ptr CSR ptr vector
 * \param csr_idx CSR idx vector
 * \param csr_val CSR value vector
 * \param x Vector being multiplied
 * \param beta Scale factor for y
 * \param y Result vector
 */
void gpu_sspmv_bhyb4(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_INT r, CUKR_INT c, CUKR_SINGLE alpha, CUKR_INT bell_nz_row, CUKR_INT bell_stride, CUKR_INT bcsr_nz,
                   CUKR_INT* bell_idx, CUKR_SINGLE* bell_val, CUKR_INT* bcsr_ptr, CUKR_INT* bcsr_idx, CUKR_SINGLE* bcsr_val,
                   CUKR_SINGLE* x, CUKR_SINGLE beta, CUKR_SINGLE* y);

// ------------------------------------------------------------------- //

/**
 * \brief Create a generic blas object by
 *        binding the above functions to a
 * 		  CUKR_BLAS structure 
 * \param blas pointer to a blas structure
 */
void blas_create_gpu(CUKR_BLAS blas);                


#endif /*GENERIC_BLAS_H_*/               
