/*
 *  kernels_single.c
 *  
 *
 *  Created by Olav Aanes Fagerlund.
 *
 */

#define LOGGER_NAME "CUKr.blas.impl.gpu.opencl.kernels_qdouble"
#include "kernels_qdouble.h"

#define AUTO_LOCAL 0
#define ORIGINAL_VALUES 0

#if ORIGINAL_VALUES
// Work-group sizes and the number of those equal to the CUDA version
#include "kernels_config_orig.h"
#else
// Work-group sizes and the number of those giving better OpenCL performance
#include "kernels_config.h"
#endif

double opencl_kernel_qdot(int n, cl_mem* xh, cl_mem* xt, cl_mem* yh, cl_mem* yt)
{	
	size_t global[1];
	size_t local[1];
	
	local[0] = CUKR_QDOT_THREADS;
	global[0] = CUKR_QDOT_CTAS * local[0];
	
	size_t sizes[9];
	void *values[9];
	
	double res = 0;
	
	// Setup work-group level result
	cl_mem res_bufh = clCreateBuffer(ComputeContext, CL_MEM_WRITE_ONLY, CUKR_QDOT_CTAS * sizeof(cl_float), NULL, &err);
	if (err != CL_SUCCESS) {
		printf("Setting up res_bufh for qdot failed!\n");
	}
	cl_mem res_buft = clCreateBuffer(ComputeContext, CL_MEM_WRITE_ONLY, CUKR_QDOT_CTAS * sizeof(cl_float), NULL, &err);
	if (err != CL_SUCCESS) {
		printf("Setting up res_buft for qdot failed!\n");
	}
	
	// Host memory to collect results from work-groups
	float *h_sh, *h_st;
	if ( (h_sh = (float*)malloc(CUKR_QDOT_CTAS*sizeof(h_sh[0]))) == NULL) {
		printf("FATAL: Error allocating memory for QDOT result vectors");
		exit(1);
	}
	if ( (h_st = (float*)malloc(CUKR_QDOT_CTAS*sizeof(h_st[0]))) == NULL) {
		printf("FATAL: Error allocating memory for QDOT result vectors");
		exit(1);
	}
	
	sizes[0] = sizeof(int);
	values[0] = (void *)&n;
	sizes[1] = sizeof(cl_mem);
	values[1] = (void *)xh;
	sizes[2] = sizeof(cl_mem);
	values[2] = (void *)xt;
	sizes[3] = sizeof(cl_mem);
	values[3] = (void *)yh;
	sizes[4] = sizeof(cl_mem);
	values[4] = (void *)yt;
	sizes[5] = sizeof(cl_mem);
	values[5] = (void *)&res_bufh;
	sizes[6] = sizeof(cl_mem);
	values[6] = (void *)&res_buft;
	sizes[7] =  local[0] * sizeof(cl_float);
	values[7] = NULL;
	sizes[8] =  local[0] * sizeof(cl_float);
	values[8] = NULL;
	
	int x_inc;
	for (x_inc = 0; x_inc<9; x_inc++) {
		err = clSetKernelArg(ComputeKernel_qdot, x_inc, sizes[x_inc], values[x_inc]);
		if (err == CL_INVALID_KERNEL)
		{
			printf("Failed %d arg qdot\n", x_inc);
			//return -1;
		}
		if (err != CL_SUCCESS)
		{
			printf("clSetKernelArg %d failed\n", x_inc);
			//return -1;
		}
	}
	
	// Enqueue kernel for execution
	err = clEnqueueNDRangeKernel(ComputeCommands, ComputeKernel_qdot, 1, NULL, global, local, 0, NULL, &qdot_event);
	if (err != CL_SUCCESS)
	{
		printf("clExecuteKernel failed %d\n",n);
	}
	
	// Synchronize for timing
	err = clWaitForEvents(1, &qdot_event);
	if (err != CL_SUCCESS) {
		printf("clWaitForEvents failed!\n");
	}
	
	// Get result from work-group level
	err = clEnqueueReadBuffer(ComputeCommands, res_bufh, CL_TRUE, 0, CUKR_QDOT_CTAS * sizeof(cl_float), (void*)h_sh, 0, NULL, NULL);
	if (err != CL_SUCCESS) {
		printf("readback of res_buf in qdot failed!\n");
	}
	err = clEnqueueReadBuffer(ComputeCommands, res_buft, CL_TRUE, 0, CUKR_QDOT_CTAS * sizeof(cl_float), (void*)h_st, 0, NULL, NULL);
	if (err != CL_SUCCESS) {
		printf("readback of res_buf in qdot failed!\n");
	}
	
	// Releas memory objects
	err = clReleaseMemObject(res_bufh);
	if (err != CL_SUCCESS) {
		printf("release of cl memobject res_bufh in qdot failed!\n");
	}
	err = clReleaseMemObject(res_buft);
	if (err != CL_SUCCESS) {
		printf("release of cl memobject res_buft in qdot failed!\n");
	}
	
	// Do the final sum
   	double sum0 = 0, sum1 = 0;
	int i;
   	for (i = 0; i < CUKR_QDOT_CTAS; i++) {
		sum0 += h_sh[i];
   		sum1 += h_st[i];
   	}
	
	// Free allocated memory
	free(h_sh);
   	free(h_st);
   	
   	return sum0 + sum1;
}

void opencl_kernel_qaxpy(int n, float a0, float a1, cl_mem* cl_xh, cl_mem* cl_xt, cl_mem* cl_yh, cl_mem* cl_yt)
{
	size_t sizes[11];
	void *values[11];
	
	sizes[0] = sizeof(int);
	values[0] = (void *)&n;
	sizes[1] = sizeof(float);
	values[1] = (void *)&a0;
	sizes[2] = sizeof(float);
	values[2] = (void *)&a1;
	sizes[3] = sizeof(cl_mem);
	values[3] = (void *)cl_xh;
	sizes[4] = sizeof(cl_mem);
	values[4] = (void *)cl_xt;
	sizes[5] = sizeof(cl_mem);
	values[5] = (void *)cl_yh;
	sizes[6] = sizeof(cl_mem);
	values[6] = (void *)cl_yt;
	
	sizes[7] = CUKR_QAXPY_THREADS * sizeof(cl_float);
	values[7] = NULL;
	sizes[8] = CUKR_QAXPY_THREADS * sizeof(cl_float);
	values[8] = NULL;
	sizes[9] = CUKR_QAXPY_THREADS * sizeof(cl_float);
	values[9] = NULL;
	sizes[10] = CUKR_QAXPY_THREADS * sizeof(cl_float);
	values[10] = NULL;
	
	/* Kernel invocation */
	int x;
	for (x = 0; x<11; x++) {
		err = clSetKernelArg(ComputeKernel_qaxpy, x, sizes[x], values[x]);
		if (err == CL_INVALID_KERNEL)
		{
			printf("Failed %d failed\n", x);
			//return -1;
		}
		if (err != CL_SUCCESS)
		{
			printf("clSetKernelArg %d failed\n", x);
			//return -1;
		}
	}
	
	size_t global[1];
	size_t local[1];
	
	local[0] = CUKR_QAXPY_THREADS;
	global[0] = CUKR_QAXPY_CTAS * local[0];
	
	/* Enqueue kernel for execution */
	err = clEnqueueNDRangeKernel(ComputeCommands, ComputeKernel_qaxpy, 1, NULL, global, local, 0, NULL, &qaxpy_event);
	if (err != CL_SUCCESS)
	{
		printf("clExecuteKernel failed %d\n",n);
	}
	
	// Synchronize for timing
	err = clWaitForEvents(1, &qaxpy_event);
	if (err != CL_SUCCESS) {
		printf("clWaitForEvents failed!\n");
	}
	
}

void opencl_kernel_qaypx(int n, float a0, float a1, cl_mem* cl_xh, cl_mem* cl_xt, cl_mem* cl_yh, cl_mem* cl_yt)
{
	size_t sizes[11];
	void *values[11];
	
	sizes[0] = sizeof(int);
	values[0] = (void *)&n;
	sizes[1] = sizeof(float);
	values[1] = (void *)&a0;
	sizes[2] = sizeof(float);
	values[2] = (void *)&a1;
	sizes[3] = sizeof(cl_mem);
	values[3] = (void *)cl_xh;
	sizes[4] = sizeof(cl_mem);
	values[4] = (void *)cl_xt;
	sizes[5] = sizeof(cl_mem);
	values[5] = (void *)cl_yh;
	sizes[6] = sizeof(cl_mem);
	values[6] = (void *)cl_yt;
	
	sizes[7] = CUKR_QAYPX_THREADS * sizeof(cl_float);
	values[7] = NULL;
	sizes[8] = CUKR_QAYPX_THREADS * sizeof(cl_float);
	values[8] = NULL;
	sizes[9] = CUKR_QAYPX_THREADS * sizeof(cl_float);
	values[9] = NULL;
	sizes[10] = CUKR_QAYPX_THREADS * sizeof(cl_float);
	values[10] = NULL;
	
	/* Kernel invocation */
	int x;
	for (x = 0; x<11; x++) {
		err = clSetKernelArg(ComputeKernel_qaypx, x, sizes[x], values[x]);
		if (err == CL_INVALID_KERNEL)
		{
			printf("Failed %d failed\n", x);
			//return -1;
		}
		if (err != CL_SUCCESS)
		{
			printf("clSetKernelArg %d failed\n", x);
			//return -1;
		}
	}
	
	size_t global[1];
	size_t local[1];
	
	local[0] = CUKR_QAYPX_THREADS;
	global[0] = CUKR_QAYPX_CTAS * local[0];
	
	/* Enqueue kernel for execution */
	err = clEnqueueNDRangeKernel(ComputeCommands, ComputeKernel_qaypx, 1, NULL, global, local, 0, NULL, &qaypx_event);
	if (err != CL_SUCCESS)
	{
		printf("clExecuteKernel failed %d\n",n);
	}
	
	// Synchronize for timing
	err = clWaitForEvents(1, &qaypx_event);
	if (err != CL_SUCCESS) {
		printf("clWaitForEvents failed!\n");
	}
	
}

void opencl_kernel_qscal(int n, float a0, float a1, cl_mem* xh, cl_mem* xt){
	
	size_t sizes[7];
	void *values[7];
	
	sizes[0] = sizeof(int);
	values[0] = (void *)&n;
	sizes[1] = sizeof(float);
	values[1] = (void *)&a0;
	sizes[2] = sizeof(float);
	values[2] = (void *)&a1;
	sizes[3] = sizeof(cl_mem);
	values[3] = (void *)xh;
	sizes[4] = sizeof(cl_mem);
	values[4] = (void *)xt;
	sizes[5] = CUKR_QSCAL_THREADS * sizeof(cl_float);
	values[5] = NULL;
	sizes[6] = CUKR_QSCAL_THREADS * sizeof(cl_float);
	values[6] = NULL;
	
	/* Kernel invocation */
	int i;
	for (i = 0; i<7; i++) {
		err = clSetKernelArg(ComputeKernel_qscal, i, sizes[i], values[i]);
		if (err == CL_INVALID_KERNEL)
		{
			printf("Failed %d failed\n", i);
			//return -1;
		}
		if (err != CL_SUCCESS)
		{
			printf("clSetKernelArg %d failed\n", i);
			//return -1;
		}
	}
	
	size_t global[1];
	size_t local[1];
	
	local[0] = CUKR_QSCAL_THREADS;
	global[0] = CUKR_QSCAL_CTAS * local[0];
	
	/* Enqueue kernel for execution */
	err = clEnqueueNDRangeKernel(ComputeCommands, ComputeKernel_qscal, 1, NULL, global, local, 0, NULL, &qscal_event);
	if (err != CL_SUCCESS)
	{
		printf("clExecuteKernel qscal failed %d\n",n);
	}
	
	// Synchronize for timing
	err = clWaitForEvents(1, &qscal_event);
	if (err != CL_SUCCESS) {
		printf("clWaitForEvents failed!\n");
	}
}


void opencl_qspmv_csr(int rows, int cols, int nz, double a0, double a1, cl_mem* ptr, 
					cl_mem* idx, cl_mem* valh, cl_mem* valt, 
					cl_mem* xh, cl_mem* xt, double b0, double b1, 
					cl_mem* yh, cl_mem* yt)
{
	
	size_t sizes[11];
	void *values[11];
	
	sizes[0] = sizeof(int);
	values[0] = (void *)&rows;
	
	sizes[1] = sizeof(cl_mem);
	values[1] = (void *)ptr;
	
	sizes[2] = sizeof(cl_mem);
	values[2] = (void *)idx;
	
	sizes[3] = sizeof(cl_mem);
	values[3] = (void *)valh;
	
	sizes[4] = sizeof(cl_mem);
	values[4] = (void *)valt;
	
	sizes[5] = sizeof(cl_mem);
	values[5] = (void *)xh;
	
	sizes[6] = sizeof(cl_mem);
	values[6] = (void *)xt;
	
	sizes[7] = sizeof(cl_mem);
	values[7] = (void *)yh;
	
	sizes[8] = sizeof(cl_mem);
	values[8] = (void *)yt;
	
	sizes[9] = sizeof(float);
	values[9] = (void *)&a0;
	
	sizes[10] = sizeof(float);
	values[10] = (void *)&a1;
	
	size_t global[1];
	size_t local[1];
	
	int inc;
	
	/*
	 * Chose which kernel to run 
	 */
	
	double a = a0 + a1;
	double b = b0 + b1;
	
	/* If beta != 0 */
	if (b != 0)
	{
	 	/* If beta != 1, do a scaling first */
	 	if (b != 1)
	 		opencl_kernel_qscal(rows, b0, b1, yh, yt);
		
	 	/* If alpha = 1, no need to consider */
   		if (a == 1) {
			// Set workgroup sizes
			local[0] = CUKR_QSPMV_CSR_A1_THREADS; //512;
			global[0] = CUKR_QSPMV_CSR_A1_CTAS * local[0];
			
			// Set kernel args
			for (inc = 0; inc < 9; inc++) {
				err = clSetKernelArg(ComputeKernel_qspmv_csr_a1, inc, sizes[inc], values[inc]);
				if (err == CL_INVALID_KERNEL)
				{
					printf("Failed %d failed\n", inc);
					//return -1;
				}
				if (err != CL_SUCCESS)
				{
					printf("clSetKernelArg %d failed\n", inc);
					//return -1;
				}
			}
			err = clEnqueueNDRangeKernel(ComputeCommands, ComputeKernel_qspmv_csr_a1, 1, NULL, global, local, 0, NULL, &qspmv_csr_a1_event);
			if (err != CL_SUCCESS)
			{
				printf("clExecuteKernel qspmv_csr_a1 failed %d\n", nz);
			}
			// Synchronize for timing
			err = clWaitForEvents(1, &qspmv_csr_a1_event);
			if (err != CL_SUCCESS) {
				printf("clWaitForEvents failed!\n");
			}
   		}
	 	/* Else, take the most general case */
	 	else {
			// Set workgroup sizes
			local[0] = CUKR_QSPMV_CSR_THREADS; //512;
			global[0] = CUKR_QSPMV_CSR_CTAS * local[0];
			
			// Set kernel args
			for (inc = 0; inc < 11; inc++) {
				err = clSetKernelArg(ComputeKernel_qspmv_csr, inc, sizes[inc], values[inc]);
				if (err == CL_INVALID_KERNEL)
				{
					printf("Failed %d failed\n", inc);
					//return -1;
				}
				if (err != CL_SUCCESS)
				{
					printf("clSetKernelArg %d failed\n", inc);
					//return -1;
				}
			}
			err = clEnqueueNDRangeKernel(ComputeCommands, ComputeKernel_qspmv_csr, 1, NULL, global, local, 0, NULL, &qspmv_csr_event);
			if (err != CL_SUCCESS)
			{
				printf("clExecuteKernel qspmv_csr failed %d\n", nz);
			}
			// Synchronize for timing
			err = clWaitForEvents(1, &qspmv_csr_event);
			if (err != CL_SUCCESS) {
				printf("clWaitForEvents failed!\n");
			}
	 	}
		
	}
	/* If beta = 0 */
	else {
	 	/* If alpha = 1 as well, no need to
	 	 * consider them both */
		if (a == 1) {
	 	 	// Set workgroup sizes
			local[0] = CUKR_QSPMV_CSR_THREADS; //512;
			global[0] = CUKR_QSPMV_CSR_CTAS * local[0];
			
			// Set kernel args
			for (inc = 0; inc < 9; inc++) {
				err = clSetKernelArg(ComputeKernel_qspmv_csr_a1_b0, inc, sizes[inc], values[inc]);
				if (err == CL_INVALID_KERNEL)
				{
					printf("Failed %d failed\n", inc);
					//return -1;
				}
				if (err != CL_SUCCESS)
				{
					printf("clSetKernelArg %d failed\n", inc);
					//return -1;
				}
			}
			err = clEnqueueNDRangeKernel(ComputeCommands, ComputeKernel_qspmv_csr_a1_b0, 1, NULL, global, local, 0, NULL, &qspmv_csr_a1_b0_event);
			if (err != CL_SUCCESS)
			{
				printf("clExecuteKernel qspmv_csr_a1_b0 failed %d\n", nz);
			}
			// Synchronize for timing
			err = clWaitForEvents(1, &qspmv_csr_a1_b0_event);
			if (err != CL_SUCCESS) {
				printf("clWaitForEvents failed!\n");
			}
		}
   	    /* If alpha != 1, have to consider it */
   	    else {
   	    	// Set workgroup sizes
			local[0] = CUKR_QSPMV_CSR_THREADS; //512;
			global[0] = CUKR_QSPMV_CSR_CTAS * local[0];
			
			// Set kernel args
			for (inc = 0; inc < 11; inc++) {
				err = clSetKernelArg(ComputeKernel_qspmv_csr_b0, inc, sizes[inc], values[inc]);
				if (err == CL_INVALID_KERNEL)
				{
					printf("Failed %d failed\n", inc);
					//return -1;
				}
				if (err != CL_SUCCESS)
				{
					printf("clSetKernelArg %d failed\n", inc);
					//return -1;
				}
			}
			err = clEnqueueNDRangeKernel(ComputeCommands, ComputeKernel_qspmv_csr_b0, 1, NULL, global, local, 0, NULL, &qspmv_csr_b0_event);
			if (err != CL_SUCCESS)
			{
				printf("clExecuteKernel qspmv_csr_b0 failed %d\n", nz);
			}
			// Synchronize for timing
			err = clWaitForEvents(1, &qspmv_csr_b0_event);
			if (err != CL_SUCCESS) {
				printf("clWaitForEvents failed!\n");
			}
   	    }
		
	}	
}

void opencl_qspmv_csr4(int rows, int cols, int nz, double a0, double a1, cl_mem* ptr, 
					   cl_mem* idx, cl_mem* valh, cl_mem* valt, 
					   cl_mem* xh, cl_mem* xt, double b0, double b1, 
					   cl_mem* yh, cl_mem* yt)
{	
	size_t sizes[11];
	void *values[11];
	
	sizes[0] = sizeof(int);
	values[0] = (void *)&rows;
	
	sizes[1] = sizeof(cl_mem);
	values[1] = (void *)ptr;
	
	sizes[2] = sizeof(cl_mem);
	values[2] = (void *)idx;
	
	sizes[3] = sizeof(cl_mem);
	values[3] = (void *)valh;
	
	sizes[4] = sizeof(cl_mem);
	values[4] = (void *)valt;
	
	sizes[5] = sizeof(cl_mem);
	values[5] = (void *)xh;
	
	sizes[6] = sizeof(cl_mem);
	values[6] = (void *)xt;
	
	sizes[7] = sizeof(cl_mem);
	values[7] = (void *)yh;
	
	sizes[8] = sizeof(cl_mem);
	values[8] = (void *)yt;
	
	sizes[9] = sizeof(float);
	values[9] = (void *)&a0;
	
	sizes[10] = sizeof(float);
	values[10] = (void *)&a1;
	
	size_t global[1];
	size_t local[1];
	
	int inc;
	
	/*
	 * Chose which kernel to run 
	 */
	
	double a = a0 + a1;
	double b = b0 + b1;
	
	/* If beta != 0 */
	if (b != 0)
	{
	 	/* If beta != 1, do a scaling first */
	 	if (b != 1)
	 		opencl_kernel_qscal(rows, b0, b1, yh, yt);
		
	 	/* If alpha = 1, no need to consider */
   		if (a == 1) {
			// Set workgroup sizes
			local[0] = CUKR_QSPMV_CSR4_THREADS; //512;
			global[0] = CUKR_QSPMV_CSR4_CTAS * local[0];
			
			// Set kernel args
			for (inc = 0; inc < 9; inc++) {
				err = clSetKernelArg(ComputeKernel_qspmv_csr4_a1, inc, sizes[inc], values[inc]);
				if (err == CL_INVALID_KERNEL)
				{
					printf("Failed %d failed\n", inc);
					//return -1;
				}
				if (err != CL_SUCCESS)
				{
					printf("clSetKernelArg %d failed\n", inc);
					//return -1;
				}
			}
			err = clEnqueueNDRangeKernel(ComputeCommands, ComputeKernel_qspmv_csr4_a1, 1, NULL, global, local, 0, NULL, &qspmv_csr4_a1_event);
			if (err != CL_SUCCESS)
			{
				printf("clExecuteKernel qspmv_csr_a1 failed %d\n", nz);
			}
			// Synchronize for timing
			err = clWaitForEvents(1, &qspmv_csr4_a1_event);
			if (err != CL_SUCCESS) {
				printf("clWaitForEvents failed!\n");
			}
   		}
	 	/* Else, take the most general case */
	 	else {
			// Set workgroup sizes
			local[0] = CUKR_QSPMV_CSR4_THREADS; //512;
			global[0] = CUKR_QSPMV_CSR4_CTAS * local[0];
			
			// Set kernel args
			for (inc = 0; inc < 11; inc++) {
				err = clSetKernelArg(ComputeKernel_qspmv_csr4, inc, sizes[inc], values[inc]);
				if (err == CL_INVALID_KERNEL)
				{
					printf("Failed %d failed\n", inc);
					//return -1;
				}
				if (err != CL_SUCCESS)
				{
					printf("clSetKernelArg %d failed\n", inc);
					//return -1;
				}
			}
			err = clEnqueueNDRangeKernel(ComputeCommands, ComputeKernel_qspmv_csr4, 1, NULL, global, local, 0, NULL, &qspmv_csr4_event);
			if (err != CL_SUCCESS)
			{
				printf("clExecuteKernel qspmv_csr failed %d\n", nz);
			}
			// Synchronize for timing
			err = clWaitForEvents(1, &qspmv_csr4_event);
			if (err != CL_SUCCESS) {
				printf("clWaitForEvents failed!\n");
			}
	 	}
		
	}
	/* If beta = 0 */
	else {
	 	/* If alpha = 1 as well, no need to
	 	 * consider them both */
		if (a == 1) {
	 	 	// Set workgroup sizes
			local[0] = CUKR_QSPMV_CSR4_THREADS; //512;
			global[0] = CUKR_QSPMV_CSR4_CTAS * local[0];
			
			// Set kernel args
			for (inc = 0; inc < 9; inc++) {
				err = clSetKernelArg(ComputeKernel_qspmv_csr4_a1_b0, inc, sizes[inc], values[inc]);
				if (err == CL_INVALID_KERNEL)
				{
					printf("Failed %d failed\n", inc);
					//return -1;
				}
				if (err != CL_SUCCESS)
				{
					printf("clSetKernelArg %d failed\n", inc);
					//return -1;
				}
			}
			err = clEnqueueNDRangeKernel(ComputeCommands, ComputeKernel_qspmv_csr4_a1_b0, 1, NULL, global, local, 0, NULL, &qspmv_csr4_a1_b0_event);
			if (err != CL_SUCCESS)
			{
				printf("clExecuteKernel qspmv_csr_a1_b0 failed %d\n", nz);
			}
			// Synchronize for timing
			err = clWaitForEvents(1, &qspmv_csr4_a1_b0_event);
			if (err != CL_SUCCESS) {
				printf("clWaitForEvents failed!\n");
			}
		}
   	    /* If alpha != 1, have to consider it */
   	    else {
   	    	// Set workgroup sizes
			local[0] = CUKR_QSPMV_CSR4_THREADS; //512;
			global[0] = CUKR_QSPMV_CSR4_CTAS * local[0];
			
			// Set kernel args
			for (inc = 0; inc < 11; inc++) {
				err = clSetKernelArg(ComputeKernel_qspmv_csr4_b0, inc, sizes[inc], values[inc]);
				if (err == CL_INVALID_KERNEL)
				{
					printf("Failed %d failed\n", inc);
					//return -1;
				}
				if (err != CL_SUCCESS)
				{
					printf("clSetKernelArg %d failed\n", inc);
					//return -1;
				}
			}
			err = clEnqueueNDRangeKernel(ComputeCommands, ComputeKernel_qspmv_csr4_b0, 1, NULL, global, local, 0, NULL, &qspmv_csr4_b0_event);
			if (err != CL_SUCCESS)
			{
				printf("clExecuteKernel qspmv_csr_b0 failed %d\n", nz);
			}
			// Synchronize for timing
			err = clWaitForEvents(1, &qspmv_csr4_b0_event);
			if (err != CL_SUCCESS) {
				printf("clWaitForEvents failed!\n");
			}
   	    }
		
	}		
}

void opencl_qspmv_hyb(int rows, int cols, int nz, float alpha0, float alpha1, int ell_nz_row, int ell_stride, int csr_nz, 
	                cl_mem* d_ell_idx, cl_mem* d_ell_valh, cl_mem* d_ell_valt, cl_mem* d_csr_ptr, cl_mem* d_csr_idx, cl_mem* d_csr_valh, cl_mem* d_csr_valt,  
	                cl_mem* d_xh, cl_mem* d_xt, float beta0, float beta1, cl_mem* d_yh, cl_mem* d_yt)
{	
	#define DIVIDE_INTO(x, y) (((x) + (y) - 1)/(y))
	
	const unsigned int BLOCK_SIZE_ELL = 256;
	unsigned int num_blocks = DIVIDE_INTO(rows, BLOCK_SIZE_ELL);
	
	// Prepare data for the ell kernel
	size_t sizes[14];
	void *values[14];
	
	sizes[0] = sizeof(int);
	values[0] = (void *)&rows;
	sizes[1] = sizeof(float);
	values[1] = (void *)&alpha0;
	sizes[2] = sizeof(float);
	values[2] = (void *)&alpha1;
	sizes[3] = sizeof(int);
	values[3] = (void *)&ell_nz_row;
	sizes[4] = sizeof(int);
	values[4] = (void *)&ell_stride;
	sizes[5] = sizeof(cl_mem);
	values[5] = (void *)d_ell_idx;
	sizes[6] = sizeof(cl_mem);
	values[6] = (void *)d_ell_valh;
	sizes[7] = sizeof(cl_mem);
	values[7] = (void *)d_ell_valt;
	sizes[8] = sizeof(float);
	values[8] = (void *)&beta0;
	sizes[9] = sizeof(float);
	values[9] = (void *)&beta1;
	sizes[10] = sizeof(cl_mem);
	values[10] = (void *)d_yh;
	sizes[11] = sizeof(cl_mem);
	values[11] = (void *)d_yt;
	sizes[12] = sizeof(cl_mem);
	values[12] = (void *)d_xh;
	sizes[13] = sizeof(cl_mem);
	values[13] = (void *)d_xt;
	
	size_t global[1];
	size_t local[1];
	
	local[0] = BLOCK_SIZE_ELL;
	global[0] = num_blocks * local[0];
	
	int inc;
	// Set kernel args
	for (inc = 0; inc < 14; inc++) {
		err = clSetKernelArg(ComputeKernel_qspmv_ell, inc, sizes[inc], values[inc]);
		if (err == CL_INVALID_KERNEL)
		{
			printf("Failed %d failed\n", inc);
		}
		if (err != CL_SUCCESS)
		{
			printf("clSetKernelArg %d failed\n", inc);
		}
	}
	
	// Launch the kernel
	err = clEnqueueNDRangeKernel(ComputeCommands, ComputeKernel_qspmv_ell, 1, NULL, global, local, 0, NULL, &qspmv_hyb_event);
	if (err != CL_SUCCESS)
	{
		printf("clExecuteKernel qspmv_ell failed %d\n", nz);
	}
	
	local[0] = CUKR_SSPMV_CSR4_THREADS;
	global[0] = CUKR_SSPMV_CSR4_CTAS * local[0];
	
	if(csr_nz){
		// The rest in CSR4
		double alpha = alpha0 + alpha1;
		if (alpha == 1){
			sizes[0] = sizeof(int);
			values[0] = (void *)&rows;
			sizes[1] = sizeof(cl_mem);
			values[1] = (void *)d_csr_ptr;
			sizes[2] = sizeof(cl_mem);
			values[2] = (void *)d_csr_idx;
			sizes[3] = sizeof(cl_mem);
			values[3] = (void *)d_csr_valh;
			sizes[4] = sizeof(cl_mem);
			values[4] = (void *)d_csr_valt;
			sizes[5] = sizeof(cl_mem);
			values[5] = (void *)d_xh;
			sizes[6] = sizeof(cl_mem);
			values[6] = (void *)d_xt;
			sizes[7] = sizeof(cl_mem);
			values[7] = (void *)d_yh;
			sizes[8] = sizeof(cl_mem);
			values[8] = (void *)d_yt;
			
			// Set kernel args
			for (inc = 0; inc < 9; inc++) {
				err = clSetKernelArg(ComputeKernel_qspmv_csr4_a1, inc, sizes[inc], values[inc]);
				if (err == CL_INVALID_KERNEL)
				{
					printf("Failed %d failed\n", inc);
				}
				if (err != CL_SUCCESS)
				{
					printf("clSetKernelArg %d failed\n", inc);
				}
			}
			err = clEnqueueNDRangeKernel(ComputeCommands, ComputeKernel_qspmv_csr4_a1, 1, NULL, global, local, 0, NULL, &qspmv_hyb_event);
			if (err != CL_SUCCESS)
			{
				printf("clExecuteKernel qspmv_csr4_a1 failed %d\n", nz);
			}
		}
		else {
			sizes[0] = sizeof(int);
			values[0] = (void *)&rows;
			sizes[1] = sizeof(cl_mem);
			values[1] = (void *)d_csr_ptr;
			sizes[2] = sizeof(cl_mem);
			values[2] = (void *)d_csr_idx;
			sizes[3] = sizeof(cl_mem);
			values[3] = (void *)d_csr_valh;
			sizes[4] = sizeof(cl_mem);
			values[4] = (void *)d_csr_valt;
			sizes[5] = sizeof(cl_mem);
			values[5] = (void *)d_xh;
			sizes[6] = sizeof(cl_mem);
			values[6] = (void *)d_xt;
			sizes[7] = sizeof(cl_mem);
			values[7] = (void *)d_yh;
			sizes[8] = sizeof(cl_mem);
			values[8] = (void *)d_yt;
			sizes[9] = sizeof(float);
			values[9] = (void *)&alpha0;
			sizes[10] = sizeof(float);
			values[10] = (void *)&alpha1;
			
			// Set kernel args
			for (inc = 0; inc < 11; inc++) {
				err = clSetKernelArg(ComputeKernel_qspmv_csr4, inc, sizes[inc], values[inc]);
				if (err == CL_INVALID_KERNEL)
				{
					printf("Failed %d failed\n", inc);
				}
				if (err != CL_SUCCESS)
				{
					printf("clSetKernelArg %d failed\n", inc);
				}
			}
			err = clEnqueueNDRangeKernel(ComputeCommands, ComputeKernel_qspmv_csr4, 1, NULL, global, local, 0, NULL, &qspmv_hyb_event);
			if (err != CL_SUCCESS)
			{
				printf("clExecuteKernel qspmv_csr4 failed %d\n", nz);
			}
		}
	}
	
	// Synchronize for timing
	err = clWaitForEvents(1, &qspmv_hyb_event);
	if (err != CL_SUCCESS) {
		printf("clWaitForEvents failed!\n");
	}	
}	


void opencl_qspmv_bcsr(int rows, int cols, int nz, int r, int c, double a0, double a1, cl_mem* ptr, 
					   cl_mem* idx, cl_mem* valh, cl_mem* valt, cl_mem* xh, cl_mem* xt, double b0, double b1, 
					   cl_mem* yh, cl_mem* yt)
{
	printf("test opencl_qspmv_bcsr\n");
}

void opencl_qspmv_bcsr4(int rows, int cols, int nz, int r, int c, double a0, double a1, cl_mem* ptr, 
						cl_mem* idx, cl_mem* valh, cl_mem* valt, 
						cl_mem* xh, cl_mem* xt, double b0, double b1, 
						cl_mem* yh, cl_mem* yt)
{
	printf("test opencl_qspmv_bcsr4\n");
}