#define USE_KAHAN_IN_SPMV 0

/**
 * @ds_ops
 * @author NVIDIA
 * @since 2008
 *  
 * Defines a double-single (qdouble)
 * operations:
 * 	- double-single add (addition)
 *	- double-single sub (subtraction)
 * 	- double-single mul (multiplication)
 * 
 *	OpenCL port Olav Aanes Fagerlund 2010
 */

/** 
 * \brief Double-single (qdouble) addition
 * \param c0,c1 Head and tail for the result
 * \param a0,a1 Head and tail for the first
 *              operand
 * \param b0,b1 Head and tail for the second
 *              operand
 */
float2 dsadd(const float a0, const float a1, const float b0, const float b1)
{
	//printf("gets here\n");
	float c0, c1;
	float t1, t2, e;
	
    // Compute dsa + dsb using Knuth's trick.
    t1 = a0 + b0;
    e = t1 - a0;
    t2 = ((b0 - e) + (a0 - (t1 - e))) + a1 + b1;
	
    // The result is t1 + t2, after normalization.
    c0 = e = t1 + t2;
    c1 = t2 - (e - t1);
	
	return (float2)(c0, c1);
}

/** 
 * \brief Double-single (qdouble) subtraction
 * \param c0,c1 Head and tail for the result
 * \param a0,a1 Head and tail for the first
 *              operand
 * \param b0,b1 Head and tail for the second
 *              operand
 */
float2 dssub(const float a0, const float  a1, const float b0, const float b1){
	
	float t1, t2, e, c0, c1;;

	// Compute dsa - dsb using Knuth's trick.
    t1 = a0 - b0;
    e = t1 - a0;
    t2 = ((-b0 - e) + (a0 - (t1 - e))) + a1 - b1;

    // The result is t1 + t2, after normalization.
    c0 = e = t1 + t2;
    c1 = t2 - (e - t1);
	
	return (float2)(c0, c1);
}


/** 
 * \brief Double-single (qdouble) multiplication
 * \param c0,c1 Head and tail for the result
 * \param a0,a1 Head and tail for the first
 *              operand
 * \param b0,b1 Head and tail for the second
 *              operand
 */
 
float2 dsmul(const float a0, const float a1, const float b0, const float b1)
{
	float c0, c1;
	float cona, conb, sa1, sa2, sb1, sb2, c11, c21, c2, t1, e, t2;
	
	cona = a0 * 8193.0f;
	conb = b0 * 8193.0f;
	sa1 = cona - (cona - a0);
	sb1 = conb - (conb - b0);
	sa2 = a0 - sa1;
	sb2 = b0 - sb1;
	
	// Multilply a0 * b0 using Dekker's method.
	c11 = a0 * b0;
	c21 = (((sa1 * sb1 - c11) + sa1 * sb2) + sa2 * sb1) + sa2 * sb2;
	
	// Compute a0 * b1 + a1 * b0 (only high-order word is needed).
	c2 = a0 * b1 + a1 * b0;
	
	// Compute (c11, c21) + c2 using Knuth's trick, also adding low-order product.
	t1 = c11 + c2;
    e = t1 - c11;
	t2 = ((c2 - e) + (c11 - (t1 - e))) + c21 + a1 * b1;
	
	// The result is t1 + t2, after normalization.
	c0 = e = t1 + t2;
	c1 = t2 - (e - t1);
	
	return (float2)(c0, c1);
}

__kernel void kernel_qaxpy(
		const int n,  
		const float a0,
		const float a1,  
		__global const float* xh,
		__global const float* xt,
		__global float* yh,
		__global float* yt,
		__local float* buffer_xh,
		__local float* buffer_xt,
		__local float* buffer_yh,
		__local float* buffer_yt
)
{
	float2 ret, ret2;

	/* Starting point for this block */
	int ctaStart = get_group_id(0) * get_local_size(0);
	 
	/* Total no. of threads in the kernel */
	int totalThreads = get_global_size(0);
	
	/* Get current thread */
	int tx = get_local_id(0);
	
	/* Read the data (full lines) */
	for (int i = ctaStart + tx; i < n; i += totalThreads)
	{  
		/* Prefetch x and y */
		buffer_xh[tx] = xh[i];
		buffer_xt[tx] = xt[i];
		buffer_yh[tx] = yh[i];
		buffer_yt[tx] = yt[i];	
		
		/* Multiply and write */ 	 
		ret = dsmul(a0, a1, buffer_xh[tx], buffer_xt[tx]);
		ret2 = dsadd(buffer_yh[tx], buffer_yt[tx], ret.x, ret.y);
		yh[i] = ret2.x;
		yt[i] = ret2.y;  	
	}
}

__kernel void kernel_qaypx(
		const int n,  
		const float a0,
		const float a1,  
		__global const float* xh,
		__global const float* xt,
		__global float* yh,
		__global float* yt,
		__local float* buffer_xh,
		__local float* buffer_xt,
		__local float* buffer_yh,
		__local float* buffer_yt
)
{
	float2 ret, ret2;
	
	// Starting point for this block
	int ctaStart = get_group_id(0) * get_local_size(0);
	 
	// Total no. of threads in the kernel
	int totalThreads = get_global_size(0);
	
	// Get current thread
	int tx = get_local_id(0);
	
	// Read the data (full lines)
	for (int i = ctaStart + tx; i < n; i += totalThreads)
	{  
		// Prefetch x and y
		
		buffer_xh[tx] = xh[i];
		buffer_xt[tx] = xt[i];
		buffer_yh[tx] = yh[i];
		buffer_yt[tx] = yt[i];
		
		// Multiply and write 	 
		ret = dsmul(a0, a1, buffer_yh[tx], buffer_yt[tx]);
		ret2 = dsadd(buffer_xh[tx], buffer_xt[tx], ret.x, ret.y);
		yh[i] = ret2.x;
		yt[i] = ret2.y; 
	}

}

__kernel void kernel_qdot(
		const int n, 
		__global const float* xh,
		__global const float* xt,
		__global const float* yh,
		__global const float* yt,
		__global float* d_sh,
		__global float* d_st,
		__local float* partial_sum_h,
		__local float* partial_sum_t
)
{
	float2 c, sum = (float2)(0, 0), p_sum;
	
	// Starting point for this block
	int ctaStart = get_group_id(0) * get_local_size(0); 
	// Total no. of threads in the kernel
	int totalThreads = get_global_size(0);
	
	// Get current thread
	int tx = get_local_id(0); 
	int bx = get_group_id(0); 
	
	// Read the sum data */
	for (int i = ctaStart + tx; i < n; i += totalThreads)
	{ 
		// Multiply
		c = dsmul(xh[i], xt[i], yh[i], yt[i]);
		
		// Accumulate
		sum = dsadd(c.x, c.y, sum.x, sum.y);
	}  
	partial_sum_h[tx] = sum.x;
	partial_sum_t[tx] = sum.y;
	
	// Reduce data for the block
	for (int i = get_local_size(0) >> 1; i > 0; i >>= 1)
	{
		barrier(CLK_LOCAL_MEM_FENCE);
		if (tx < i) {
			p_sum = dsadd(partial_sum_h[tx], partial_sum_t[tx], partial_sum_h[tx + i], partial_sum_t[tx + i]);
			partial_sum_h[tx] = p_sum.x;
			partial_sum_t[tx] = p_sum.y;
		}
	} 
	if (tx == 0) {
		d_sh[bx] = partial_sum_h[tx];
		d_st[bx] = partial_sum_t[tx];
    }
}

__kernel void kernel_qscal(
		const int n,    
		const float a0,
		const float a1,
		__global float* xh,
		__global float* xt,
		__local float* buffer_xh,
		__local float* buffer_xt
)
{
	float2 res;
	
    // Starting point for this block
	int ctaStart = get_group_id(0) * get_local_size(0);
	 
	// Total no. of threads in the kernel
	int totalThreads = get_global_size(0);
	
	// Get current thread
	int tx = get_local_id(0);
	
	// Read the data (full lines)
	for (int i = ctaStart + tx; i < n; i += totalThreads)
	{  
		// Prefetch x
		buffer_xh[tx] = xh[i];
		buffer_xt[tx] = xt[i];
		
		// Scale and write
		res = dsmul (buffer_xh[tx], buffer_xt[tx], a0, a1);
		xh[i] = res.x; //xh[i] * a0;
		xt[i] = res.y; //xt[i] * a1;  
	}
}

__kernel void kernel_qspmv_csr_a1_b0(
		const int rows,
		__global const int* d_ptr,
		__global const int* d_idx,
		__global const float* d_valh,
		__global const float* d_valt,
		__global const float* d_xh,
		__global const float* d_xt,
		__global float* d_yh,
		__global float* d_yt
)
{
	/* Starting point for this block */
	int ctaStart = get_group_id(0) * get_local_size(0);
	 
	/* Total no. of threads in the kernel */
	int totalThreads = get_global_size(0);
	
	/* Get current thread */
	int tx = get_local_id(0);

	int i;
	float2 c;

	/* Read the data*/
	for (i = ctaStart + tx; i < rows; i += totalThreads)
	{ 	
		/* Read the beginning and end of the row 
		 * which will be processed by this thread */
		int iRowBeg = d_ptr[i] - 1;
		int iRowEnd = d_ptr[i+1] - 1;
		
		float2 sum = (float2)(0, 0);
		
		/* Read and sum for the column vectors */
		#if USE_KAHAN_IN_SPMV
			int col = d_idx[iRowBeg] - 1;
			sum = dsmul(d_valh[iRowBeg], d_valt[iRowBeg], d_xh[col], d_xt[col]);
			float2 cc = (float2)(0.0f, 0.0f);
			for (int j= iRowBeg + 1; j < iRowEnd; j++) {
				float2 y, t;
				col = d_idx[j] - 1;
				c = dsmul(d_valh[j], d_valt[j], d_xh[col], d_xt[col]);
				y = dssub(c.x, c.y, cc.x, cc.y);
				t = dsadd(sum.x, sum.y, y.x, y.y);
				c = dssub(t.x, t.y, sum.x, sum.y); 
				cc = dssub(c.x, c.y, y.x, y.y);
				sum.x = t.x;
				sum.y = t.y;
			}
		#else
			for ( int j = iRowBeg; j < iRowEnd; j++) {
				int col = d_idx[j] - 1;
				c = dsmul(d_valh[j], d_valt[j], d_xh[col], d_xt[col]);
				sum = dsadd(c.x, c.y, sum.x, sum.y);
			}
		#endif

		/* Write the result to global memory */
		d_yh[i] = sum.x;
		d_yt[i] = sum.y;		 	
	}
}

__kernel void kernel_qspmv_csr_a1(
		const int rows,
		__global const int* d_ptr,
		__global const int* d_idx,
		__global const float* d_valh,
		__global const float* d_valt,
		__global const float* d_xh,
		__global const float* d_xt,
		__global float* d_yh,
		__global float* d_yt
)
{
	/* Starting point for this block */
	int ctaStart = get_group_id(0) * get_local_size(0);
	 
	/* Total no. of threads in the kernel */
	int totalThreads = get_global_size(0);
	
	/* Get current thread */
	int tx = get_local_id(0);

	int i;
	float2 c;

	/* Read the data*/
	for (i = ctaStart + tx; i < rows; i += totalThreads)
	{ 	
		/* Read the beginning and end of the row 
		 * which will be processed by this thread */
		int iRowBeg = d_ptr[i] - 1;
		int iRowEnd = d_ptr[i+1] - 1;
		
		float2 sum = (float2)(0, 0);
		
		/* Read and sum for the column vectors */
		#if USE_KAHAN_IN_SPMV
			int col = d_idx[iRowBeg] - 1;
			sum = dsmul(d_valh[iRowBeg], d_valt[iRowBeg], d_xh[col], d_xt[col]);
			float2 cc = (float2)(0.0f, 0.0f);
			for (int j= iRowBeg + 1; j < iRowEnd; j++) {
				float2 y, t;
				col = d_idx[j] - 1;
				c = dsmul(d_valh[j], d_valt[j], d_xh[col], d_xt[col]);
				y = dssub(c.x, c.y, cc.x, cc.y);
				t = dsadd(sum.x, sum.y, y.x, y.y);
				c = dssub(t.x, t.y, sum.x, sum.y); 
				cc = dssub(c.x, c.y, y.x, y.y);
				sum.x = t.x;
				sum.y = t.y;
			}
		#else
			for ( int j = iRowBeg; j < iRowEnd; j++) {
				int col = d_idx[j] - 1;
				c = dsmul(d_valh[j], d_valt[j], d_xh[col], d_xt[col]);
				sum = dsadd(c.x, c.y, sum.x, sum.y);			
			}
		#endif

		/* Write the result to global memory */
		d_yh[i] += sum.x;
		d_yt[i] += sum.y;		 	
	}
}

__kernel void kernel_qspmv_csr_b0(
		const int rows,
		__global const int* d_ptr,
		__global const int* d_idx,
		__global const float* d_valh,
		__global const float* d_valt,
		__global const float* d_xh,
		__global const float* d_xt,
		__global float* d_yh,
		__global float* d_yt,
		const float alpha0,
		const float alpha1
)
{
	/* Starting point for this block */
	int ctaStart = get_group_id(0) * get_local_size(0);
	 
	/* Total no. of threads in the kernel */
	int totalThreads = get_global_size(0);
	
	/* Get current thread */
	int tx = get_local_id(0);

	int i;
	float2 c, res;

	/* Read the data*/
	for (i = ctaStart + tx; i < rows; i += totalThreads)
	{ 	
		/* Read the beginning and end of the row 
		 * which will be processed by this thread */
		int iRowBeg = d_ptr[i] - 1;
		int iRowEnd = d_ptr[i+1] - 1;
		
		float2 sum = (float2)(0, 0);
		
		/* Read and sum for the column vectors */
		#if USE_KAHAN_IN_SPMV
			int col = d_idx[iRowBeg] - 1;
			sum = dsmul(d_valh[iRowBeg], d_valt[iRowBeg], d_xh[col], d_xt[col]);
			float2 cc = (float2)(0.0f, 0.0f);
			for (int j= iRowBeg + 1; j < iRowEnd; j++) {
				float2 y, t;
				col = d_idx[j] - 1;
				c = dsmul(d_valh[j], d_valt[j], d_xh[col], d_xt[col]);
				y = dssub(c.x, c.y, cc.x, cc.y);
				t = dsadd(sum.x, sum.y, y.x, y.y);
				c = dssub(t.x, t.y, sum.x, sum.y); 
				cc = dssub(c.x, c.y, y.x, y.y);
				sum.x = t.x;
				sum.y = t.y;
			}
		#else
			for ( int j = iRowBeg; j < iRowEnd; j++) {
				int col = d_idx[j] - 1;
				c = dsmul(d_valh[j], d_valt[j], d_xh[col], d_xt[col]);
				sum = dsadd(c.x, c.y, sum.x, sum.y);			
			}
		#endif
		
		/* Write the result to global memory */
		res = dsmul(alpha0, alpha1, sum.x, sum.y);
		d_yh[i] = res.x;
		d_yt[i] = res.y;		 	
	}
}

__kernel void kernel_qspmv_csr(
		const int rows,
		__global const int* d_ptr,
		__global const int* d_idx,
		__global const float* d_valh,
		__global const float* d_valt,
		__global const float* d_xh,
		__global const float* d_xt,
		__global float* d_yh,
		__global float* d_yt,
		const float alpha0,
		const float alpha1
)
{
	/* Starting point for this block */
	int ctaStart = get_group_id(0) * get_local_size(0);
	 
	/* Total no. of threads in the kernel */
	int totalThreads = get_global_size(0);
	
	/* Get current thread */
	int tx = get_local_id(0);

	int i;
	float2 c;
	
	/* Read the data*/
	for (i = ctaStart + tx; i < rows; i += totalThreads)
	{ 	
		/* Read the beginning and end of the row 
		 * which will be processed by this thread */
		int iRowBeg = d_ptr[i] - 1;
		int iRowEnd = d_ptr[i+1] - 1;
		
		float2 sum = (float2)(0, 0);
		
		/* Read and sum for the column vectors */
		#if USE_KAHAN_IN_SPMV
			int col = d_idx[iRowBeg] - 1;
			sum = dsmul(d_valh[iRowBeg], d_valt[iRowBeg], d_xh[col], d_xt[col]);
			float2 cc = (float2)(0.0f, 0.0f);
			for (int j= iRowBeg + 1; j < iRowEnd; j++) {
				float2 y, t;
				col = d_idx[j] - 1;
				c = dsmul(d_valh[j], d_valt[j], d_xh[col], d_xt[col]);
				y = dssub(c.x, c.y, cc.x, cc.y);
				t = dsadd(sum.x, sum.y, y.x, y.y);
				c = dssub(t.x, t.y, sum.x, sum.y); 
				cc = dssub(c.x, c.y, y.x, y.y);
				sum.x = t.x;
				sum.y = t.y;
			}
		#else
			for ( int j = iRowBeg; j < iRowEnd; j++) {
				int col = d_idx[j] - 1;
				c = dsmul(d_valh[j], d_valt[j], d_xh[col], d_xt[col]);
				sum = dsadd(c.x, c.y, sum.x, sum.y);			
			}
		#endif

		// Multiply with alpha
		sum = dsmul(alpha0, alpha1, sum.x, sum.y);

		/* Write the result to global memory */
		d_yh[i] += sum.x;
		d_yt[i] += sum.y;		 	
	}
}

__kernel void kernel_qspmv_csr4_a1_b0(
		const int rows,
		__global const int* d_ptr,
		__global const int4* d_idx,
		__global const float4* d_valh,
		__global const float4* d_valt,
		__global const float* d_xh,
		__global const float* d_xt,
		__global float* d_yh,
		__global float* d_yt
)
{
	/* Starting point for this block */
	int ctaStart = get_group_id(0) * get_local_size(0);
	 
	/* Total no. of threads in the kernel */
	int totalThreads = get_global_size(0);
	
	/* Get current thread */
	int tx = get_local_id(0);

	float2 c;
	
	/* Read the data*/
	for (int i = ctaStart + tx; i < rows; i += totalThreads)
	{ 	
		/* Read the beginning and end of the row 
		 * which will be processed by this thread */
		int iRowBeg = d_ptr[i] - 1;
		int iRowEnd = d_ptr[i+1] - 1;
		
		float2 sum = (float2)(0, 0); 

		//#pragma unroll
		for (int j = iRowBeg / 4; j < iRowEnd / 4; j++) {
			float4 valh, valt, xh, xt;
			int4 idx;
			
			// Read idx and val
			idx = d_idx[j];
			valh = d_valh[j];
			valt = d_valt[j];
			
			// Idx is base 1, change to 0
			idx.x -= 1;
			idx.y -= 1;
			idx.z -= 1;
			idx.w -= 1;
			
			// Read head part of x
			xh.x = d_xh[idx.x];
			xh.y = d_xh[idx.y];
			xh.z = d_xh[idx.z];
			xh.w = d_xh[idx.w];
			
			// Read tail part of x
			xt.x = d_xt[idx.x];
			xt.y = d_xt[idx.y];
			xt.z = d_xt[idx.z];
			xt.w = d_xt[idx.w];
			
			// Multiply and add
			c = dsmul(xh.x, xt.x, valh.x, valt.x);
			sum = dsadd(c.x, c.y, sum.x, sum.y);
			
			c = dsmul(xh.y, xt.y, valh.y, valt.y);
			sum = dsadd(c.x, c.y, sum.x, sum.y);
			
			c = dsmul(xh.z, xt.z, valh.z, valt.z);
			sum =  dsadd(c.x, c.y, sum.x, sum.y);

			c = dsmul(xh.w, xt.w, valh.w, valt.w);
			sum = dsadd(c.x, c.y, sum.x, sum.y);
		}

		// Write the result to global memory
		d_yh[i] = sum.x;
		d_yt[i] = sum.y;		 	
	}
}

__kernel void kernel_qspmv_csr4_a1(
		const int rows,
		__global const int* d_ptr,
		__global const int4* d_idx,
		__global const float4* d_valh,
		__global const float4* d_valt,
		__global const float* d_xh,
		__global const float* d_xt,
		__global float* d_yh,
		__global float* d_yt
)
{
	/* Starting point for this block */
	int ctaStart = get_group_id(0) * get_local_size(0);
	 
	/* Total no. of threads in the kernel */
	int totalThreads = get_global_size(0);
	
	/* Get current thread */
	int tx = get_local_id(0);

	float2 c, res;
	
	/* Read the data*/
	for (int i = ctaStart + tx; i < rows; i += totalThreads)
	{ 	
		/* Read the beginning and end of the row 
		 * which will be processed by this thread */
		int iRowBeg = d_ptr[i] - 1;
		int iRowEnd = d_ptr[i+1] - 1;
		
		float2 sum = (float2)(0, 0); 

		for (int j = iRowBeg / 4; j < iRowEnd / 4; j++) {
			float4 valh, valt, xh, xt;
			int4 idx;
			
			// Read idx and val
			idx = d_idx[j];
			valh = d_valh[j];
			valt = d_valt[j];
			
			// Idx is base 1, change to 0
			idx.x -= 1;
			idx.y -= 1;
			idx.z -= 1;
			idx.w -= 1;
			
			// Read head part of x
			xh.x = d_xh[idx.x];
			xh.y = d_xh[idx.y];
			xh.z = d_xh[idx.z];
			xh.w = d_xh[idx.w];
			
			// Read tail part of x
			xt.x = d_xt[idx.x];
			xt.y = d_xt[idx.y];
			xt.z = d_xt[idx.z];
			xt.w = d_xt[idx.w];
			
			// Multiply and add
			c = dsmul(xh.x, xt.x, valh.x, valt.x);
			sum = dsadd(c.x, c.y, sum.x, sum.y);
			
			c = dsmul(xh.y, xt.y, valh.y, valt.y);
			sum = dsadd(c.x, c.y, sum.x, sum.y);
			
			c = dsmul(xh.z, xt.z, valh.z, valt.z);
			sum =  dsadd(c.x, c.y, sum.x, sum.y);

			c = dsmul(xh.w, xt.w, valh.w, valt.w);
			sum = dsadd(c.x, c.y, sum.x, sum.y);
		}

		// Add/write the result to global memory
		res = dsadd(d_yh[i], d_yt[i], sum.x, sum.y);
		d_yh[i] = res.x;
		d_yt[i] = res.y; 		 	
	}
}

__kernel void kernel_qspmv_csr4_b0(
		const int rows,
		__global const int* d_ptr,
		__global const int4* d_idx,
		__global const float4* d_valh,
		__global const float4* d_valt,
		__global const float* d_xh,
		__global const float* d_xt,
		__global float* d_yh,
		__global float* d_yt,
		const float alpha0,
		const float alpha1
)
{
	/* Starting point for this block */
	int ctaStart = get_group_id(0) * get_local_size(0);
	 
	/* Total no. of threads in the kernel */
	int totalThreads = get_global_size(0);
	
	/* Get current thread */
	int tx = get_local_id(0);

	float2 c, res;

	/* Read the data*/
	for (int i = ctaStart + tx; i < rows; i += totalThreads)
	{ 	
		/* Read the beginning and end of the row 
		 * which will be processed by this thread */
		int iRowBeg = d_ptr[i] - 1;
		int iRowEnd = d_ptr[i+1] - 1;
		
		float2 sum = (float2)(0, 0); 

		for (int j = iRowBeg / 4; j < iRowEnd / 4; j++) {
			float4 valh, valt, xh, xt;
			int4 idx;
			
			// Read idx and val
			idx = d_idx[j];
			valh = d_valh[j];
			valt = d_valt[j];
			
			// Idx is base 1, change to 0
			idx.x -= 1;
			idx.y -= 1;
			idx.z -= 1;
			idx.w -= 1;
			
			// Read head part of x
			xh.x = d_xh[idx.x];
			xh.y = d_xh[idx.y];
			xh.z = d_xh[idx.z];
			xh.w = d_xh[idx.w];
			
			// Read tail part of x
			xt.x = d_xt[idx.x];
			xt.y = d_xt[idx.y];
			xt.z = d_xt[idx.z];
			xt.w = d_xt[idx.w];
			
			// Multiply and add
			c = dsmul(xh.x, xt.x, valh.x, valt.x);
			sum = dsadd(c.x, c.y, sum.x, sum.y);
			
			c = dsmul(xh.y, xt.y, valh.y, valt.y);
			sum = dsadd(c.x, c.y, sum.x, sum.y);
			
			c = dsmul(xh.z, xt.z, valh.z, valt.z);
			sum =  dsadd(c.x, c.y, sum.x, sum.y);

			c = dsmul(xh.w, xt.w, valh.w, valt.w);
			sum = dsadd(c.x, c.y, sum.x, sum.y);
		}

		// Multiply with alpha and write to global memory
		res = dsmul(alpha0, alpha1, sum.x, sum.y);
		
		d_yh[i] = res.x;
		d_yt[i] = res.y;	 	
	}
}

__kernel void kernel_qspmv_csr4(
		const int rows,
		__global const int* d_ptr,
		__global const int4* d_idx,
		__global const float4* d_valh,
		__global const float4* d_valt,
		__global const float* d_xh,
		__global const float* d_xt,
		__global float* d_yh,
		__global float* d_yt,
		const float alpha0,
		const float alpha1
)
{
	/* Starting point for this block */
	int ctaStart = get_group_id(0) * get_local_size(0);
	 
	/* Total no. of threads in the kernel */
	int totalThreads = get_global_size(0);
	
	/* Get current thread */
	int tx = get_local_id(0);

	float2 c, res;

	/* Read the data*/
	for (int i = ctaStart + tx; i < rows; i += totalThreads)
	{ 	
		/* Read the beginning and end of the row 
		 * which will be processed by this thread */
		int iRowBeg = d_ptr[i] - 1;
		int iRowEnd = d_ptr[i+1] - 1;
		
		float2 sum = (float2)(0, 0); 

		//#pragma unroll
		for (int j = iRowBeg / 4; j < iRowEnd / 4; j++) {
			float4 valh, valt, xh, xt;
			int4 idx;
			
			// Read idx and val
			idx = d_idx[j];
			valh = d_valh[j];
			valt = d_valt[j];
			
			// Idx is base 1, change to 0
			idx.x -= 1;
			idx.y -= 1;
			idx.z -= 1;
			idx.w -= 1;
			
			// Read head part of x
			xh.x = d_xh[idx.x];
			xh.y = d_xh[idx.y];
			xh.z = d_xh[idx.z];
			xh.w = d_xh[idx.w];
			
			// Read tail part of x
			xt.x = d_xt[idx.x];
			xt.y = d_xt[idx.y];
			xt.z = d_xt[idx.z];
			xt.w = d_xt[idx.w];
			
			// Multiply and add
			c = dsmul(xh.x, xt.x, valh.x, valt.x);
			sum = dsadd(c.x, c.y, sum.x, sum.y);
			
			c = dsmul(xh.y, xt.y, valh.y, valt.y);
			sum = dsadd(c.x, c.y, sum.x, sum.y);
			
			c = dsmul(xh.z, xt.z, valh.z, valt.z);
			sum =  dsadd(c.x, c.y, sum.x, sum.y);

			c = dsmul(xh.w, xt.w, valh.w, valt.w);
			sum = dsadd(c.x, c.y, sum.x, sum.y);
		}

		// Multiply with alpha
		sum = dsmul(alpha0, alpha1, sum.x, sum.y);

		// Write the result to global memory
		res = dsadd(d_yh[i], d_yt[i], sum.x, sum.y);
		
		d_yh[i] = res.x;
		d_yt[i] = res.y;
	}
}

#define large_grid_thread_id(void) (((uint)mul24((uint)get_local_size(0),(uint)get_group_id(0) + (uint)mul24((uint)get_group_id(1),(uint)get_num_groups(0))) + (uint)get_local_id(0)))

__kernel void kernel_qspmv_ell(
		const int		rows,      
		const float	alpha0,
		const float	alpha1,
		const int    ell_nz_row,
		const int    ell_stride,
		__global const int   *ell_idx, 
		__global const float *ell_valh,
		__global const float *ell_valt,
		const float  beta0,
		const float  beta1,
		__global	float *d_yh,
		__global	float *d_yt,
		__global	float *d_xh,
		__global	float *d_xt
)
{
	const int row = large_grid_thread_id();
	
	if(row >= rows){
		return;
	}
	
	float2 sum = (float2)(0, 0);
	if (beta0 || beta1)
		sum = dsmul(beta0, beta1, d_yh[row], d_yt[row]);
	
	ell_idx += row;
	ell_valh += row;
	ell_valt += row;
	
	for(int n = 0; n < ell_nz_row; n++){
		const float A_ij_h = *ell_valh;
		const float A_ij_t = *ell_valt;
		
		if(A_ij_h != 0){
			float2 c;
			const int col = *ell_idx - 1;
			c = dsmul(A_ij_h, A_ij_t, d_xh[col], d_xt[col]); // this last d_x.. -> can be replaced by image access..
			sum = dsadd(sum.x, sum.y, c.x, c.y);
		}
		ell_idx += ell_stride;
		ell_valh += ell_stride;
		ell_valt += ell_stride;
	}
	
	d_yh[row] = sum.x;
	d_yt[row] = sum.y;
}