/**
 * @file blas/impl/mkl/mkl_blas.c
 * @author Serban Georgescu
 * @since 2008
 * 
 * It defines wrapper for the Intel
 * MKL BLAS.
 * 
 * The AYPX functions, not present in
 * MKL, are taken from the generic BLAS.
 */

#define LOGGER_NAME "CUKr.blas.impl.mkl_blas" 
 
#include <stdio.h>
#include <stdlib.h>
#include "../../../util/log.h"
#include "../generic/generic_blas.h"
#include "mkl_blas.h"
 
#ifdef CUKR_USE_MKL
	#include <mkl_cblas.h>
	#include <mkl_types.h>
	#include <mkl_spblas.h>
#endif	




#ifdef CUKR_USE_MKL

// ------------------------------------------------------------------- //

/**
 * \brief Dot product in double precision
 * \param n Vector size
 * \param x,y The two vectors whose product 
 *            is being computed 
 */
CUKR_DOUBLE mkl_ddot(CUKR_INT n, CUKR_DOUBLE* x, CUKR_DOUBLE* y)
{
	return cblas_ddot(n, x, 1, y, 1);
}

/**
 * \brief Dot product in single precision
 * \param n Vector size
 * \param x,y The two vectors whose product 
 *            is being computed 
 */
CUKR_SINGLE mkl_sdot(CUKR_INT n, CUKR_SINGLE* x, CUKR_SINGLE* y)
{
	return cblas_sdot(n, x, 1, y, 1);
}
    
// ------------------------------------------------------------------- //
  
/**
 * \brief AXPY in double precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
void mkl_daxpy(CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x, CUKR_DOUBLE* y)
{
	cblas_daxpy(n, a, x, 1, y, 1);
}

/**
 * \brief AXPY in single precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
void mkl_saxpy(CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x, CUKR_SINGLE* y)
{
	cblas_saxpy(n, a, x, 1, y, 1);
}

// ------------------------------------------------------------------- //

/**
 * \brief Vector scaling in double precision
 * \param n Vector size
 * \param a Scaling factor
 * \param x The vector being scaled
 */
void mkl_dscal(CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x)
{
	cblas_dscal(n, a, x, 1);
}

/**
 * \brief Vector scaling in single precision
 * \param n Vector size
 * \param a Scaling factor
 * \param x The vector being scaled
 */
void mkl_sscal(CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x)
{
	cblas_sscal(n, a, x, 1);
}

// ------------------------------------------------------------------- //

/**
 * \brief Vector copy in double precision
 * \param n Vector size
 * \param x The source vector
 * \param y The destination vector
 */
void mkl_dcopy(CUKR_INT n, CUKR_DOUBLE* x, CUKR_DOUBLE* y)
{
	cblas_dcopy(n, x, 1, y, 1);
} 

/**
 * \brief Vector copy in single precision
 * \param n Vector size
 * \param x The source vector
 * \param y The destination vector
 */
void mkl_scopy(CUKR_INT n, CUKR_SINGLE* x, CUKR_SINGLE* y)
{
	cblas_scopy(n, x, 1, y, 1);
} 

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param a Constant
 * \param y Result vector
 */
void mkl_dspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT* ptr, CUKR_INT* idx, 
             	   CUKR_DOUBLE* val, CUKR_DOUBLE* x, CUKR_DOUBLE beta, CUKR_DOUBLE* y)
{
	mkl_dcsrmv("N", &rows, &cols, &alpha, "G", val , idx, ptr, &(ptr[1]), x, &beta, y);
}

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param a Constant
 * \param y Result vector
 */
void mkl_sspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_SINGLE alpha, CUKR_INT* ptr, CUKR_INT* idx, 
             	   CUKR_SINGLE* val, CUKR_SINGLE* x, CUKR_SINGLE beta, CUKR_SINGLE* y)
{
	mkl_scsrmv("N", &rows, &cols, &alpha, "G", val , idx, ptr, &(ptr[1]), x, &beta, y);
}
                
#endif /* CUKR_USE_MKL */


       
// ------------------------------------------------------------------- //

 
/**
 * \brief Create a generic blas object by
 *        binding the above functions to a
 * 		  CUKR_BLAS structure 
 * \param blas pointer to a blas structure
 */
void blas_create_mkl(CUKR_BLAS blas)
#ifndef CUKR_USE_MKL
{
	CUKR_FATAL("MKL support was not compiled in the library");
}
#else	
{
	/* Set the BLAS name */
	sprintf(blas->name, "MKL");
	blas->location = CUKR_LOCATION_CPU;
	
	/* Set the bindings */
	blas->bind.ddot = mkl_ddot;
	blas->bind.sdot = mkl_sdot;
	
	blas->bind.daxpy = mkl_daxpy;
	blas->bind.saxpy = mkl_saxpy;
	
	blas->bind.daypx = generic_daypx;
	blas->bind.saypx = generic_saypx;
	
	blas->bind.dscal = mkl_dscal;
	blas->bind.sscal = mkl_sscal;
	
	blas->bind.dcopy = mkl_dcopy;
	blas->bind.scopy = mkl_scopy;
	
	blas->bind.dspmv_csr = mkl_dspmv_csr;
	blas->bind.sspmv_csr = mkl_sspmv_csr;
}    
#endif /* CUKR_USE_MKL */           
