/**
 * @file blas/impl/mkl/mkl_blas.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * It defines wrapper for the Intel
 * MKL BLAS.
 */

#ifndef MKL_BLAS_H_
#define MKL_BLAS_H_

#include "../../blas.h"

// ------------------------------------------------------------------- //

/**
 * \brief Dot product in double precision
 * \param n Vector size
 * \param x,y The two vectors whose product 
 *            is being computed 
 */
CUKR_DOUBLE mkl_ddot(CUKR_INT n, CUKR_DOUBLE* x, CUKR_DOUBLE* y);

/**
 * \brief Dot product in single precision
 * \param n Vector size
 * \param x,y The two vectors whose product 
 *            is being computed 
 */
CUKR_SINGLE mkl_sdot(CUKR_INT n, CUKR_SINGLE* x, CUKR_SINGLE* y);
    
// ------------------------------------------------------------------- //
  
/**
 * \brief AXPY in double precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
void mkl_daxpy(CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x, CUKR_DOUBLE* y); 

/**
 * \brief AXPY in single precision
 * \param n Vector size
 * \param a The constant
 * \param x,y The two vectors used in the
 *            operation.  
 */
void mkl_saxpy(CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x, CUKR_SINGLE* y); 

// ------------------------------------------------------------------- //

/**
 * \brief Vector scaling in double precision
 * \param n Vector size
 * \param a Scaling factor
 * \param x The vector being scaled
 */
void mkl_dscal(CUKR_INT n, CUKR_DOUBLE a, CUKR_DOUBLE* x); 

/**
 * \brief Vector scaling in single precision
 * \param n Vector size
 * \param a Scaling factor
 * \param x The vector being scaled
 */
void mkl_sscal(CUKR_INT n, CUKR_SINGLE a, CUKR_SINGLE* x); 

// ------------------------------------------------------------------- //

/**
 * \brief Vector copy in double precision
 * \param n Vector size
 * \param x The source vector
 * \param y The destination vector
 */
void mkl_dcopy(CUKR_INT n, CUKR_DOUBLE* x, CUKR_DOUBLE* y); 

/**
 * \brief Vector copy in single precision
 * \param n Vector size
 * \param x The source vector
 * \param y The destination vector
 */
void mkl_scopy(CUKR_INT n, CUKR_SINGLE* x, CUKR_SINGLE* y); 

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiply in 
 *        CSR format, in double precision
 * \param rows,cols Matrix size
 * \param nz Number of nonzeros
 * \param alpha Scaling factor for Ax
 * \param ptr Row index vector
 * \param idx Column index vector
 * \param val Value vector
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */
void mkl_dspmv_csr(CUKR_INT rows, CUKR_INT cols, CUKR_INT nz, CUKR_DOUBLE alpha, CUKR_INT* ptr, 
                   CUKR_INT* idx, CUKR_DOUBLE* val, CUKR_DOUBLE* x, CUKR_DOUBLE beta, 
                   CUKR_DOUBLE* y);            
              
// ------------------------------------------------------------------- //

/**
 * \brief Create a generic blas object by
 *        binding the above functions to a
 * 		  CUKR_BLAS structure 
 * \param blas pointer to a blas structure
 */
void blas_create_mkl(CUKR_BLAS blas);                


#endif /*GENERIC_BLAS_H_*/               
