/**
 * @file blas/blas1_mp.h
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains the extension of the
 * BLAS1 routines to MP vectors
 * 	 
 */ 

#define LOGGER_NAME "CUKr.blas.mp.blas1_mp"

#include <stdio.h>
#include "blas1_mp.h"
#include "../sp/blas1_sp.h"
#include "../../util/log.h"
#include "../../util/utils.h"
#include "../../util/macros.h"
#include "../../threading/threading.h"
#include "../../db/db.h"

/* Get access to the database */
extern CukrDB cukr_db;

// ------------------------------------------------------------------- //

/**
 * \brief Dot product between two MP vectors
 * \param blas The blas used for the operation
 * \param x,y The MP vectors
 */
CUKR_DOUBLE CukrDOT_MP(CUKR_BLAS blas, CUKR_VECTOR_MP x, CUKR_VECTOR_MP y)
{
	CUKR_DOUBLE dval = 0;
	int i;
	
	/* Start timing */
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START; 
	
	/* Log */
	LOGX(DEBUG, "[N_%d] Running dot product between MP vectors "\
				"%s and %s", cukr_db.mpi_rank, x->name, y->name);			
		
	/* Error if the vectors do not have
	 * the same number of elements */
	if (x->n != y->n)
	 	 CUKR_FATALX("[N_%d] Cannot compute the dot product "\
		             "for MP vectors %s and %s since they "\
		             "do not have the same number of "\
		             "elements", cukr_db.mpi_rank, x->name, y->name); 
	
	/* If the vectors do not have the same
	 * location/precision as the BLAS, the 
	 * vectors should be moved to that location
	 * and/or their precision changed */
	 
	CukrVecmpTransfer(x, blas->location, blas->precision);
	CukrVecmpTransfer(y, blas->location, blas->precision);  
	
	/* The operation happens in the same way
	 * regardless of the sharing type. For
	 * both cases, only the diagonal elements
	 * are used */
	
	/* Schedule DOT between all pairs
	 * of parts */
	
	for (i = 0; i < x->nparts; i++)
		CukrThreadsRun(i, (void*)CukrDOT_SP, CUKR_FUNCID_DOT, blas, x->cells[i], y->cells[i]);	 
	
	/* Wait for all threads to finish */
	CukrThreadsWaitCompletion(x->nparts);
	
	/* Add the returned values */
	for (i = 0; i < x->nparts; i++) {
		dval += CukrThreadsGetRetval_double(i);
	}  
	
	/* Log*/
	LOGX(DEBUG, "[N_%d] Dot product between MP vectors %s "\
				"and %s returned %g", cukr_db.mpi_rank, 
				x->name, y->name, dval);				
				
	/* Update counters */
	CLOCK_END;
	blas->counters_mp[CUKR_BLAS_IDX_TIME] [CUKR_BLAS_OPS_DOT] += CLOCK_SPAN;
	blas->counters_mp[CUKR_BLAS_IDX_FLOPS][CUKR_BLAS_OPS_DOT] += 2 * x->n;
				
	return dval;
	
}

// ------------------------------------------------------------------- //

/**
 * \brief AXPY (ax + y) between two MP
 *        vectors
 * \param blas The blas used for the operation
 * \param a The constant
 * \param x,y The MP vectors
 */
void CukrAXPY_MP(CUKR_BLAS blas, CUKR_DOUBLE a, CUKR_VECTOR_MP x, CUKR_VECTOR_MP y)
{
	int i;
	
	/* Start timing */
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START; 
	
	/* Log */
	LOGX(DEBUG, "[N_%d] Running AXPY between MP vectors "\
				"%s and %s, with a=%g", cukr_db.mpi_rank, 
				x->name, y->name, a);				
		
	/* Error if the vectors do not have
	 * the same number of elements */
	if (x->n != y->n)
		CUKR_FATALX("[N_%d] Cannot compute AXPY "\
		            "for MP vectors %s and %s since they "\
		            "do not have the same number of "\
		            "elements", cukr_db.mpi_rank, x->name, y->name); 
	
	/* If the vectors do not have the same
	 * location/precision as the BLAS, the 
	 * vectors should be moved to that location
	 * and/or their precision changed */
	 
	CukrVecmpTransfer(x, blas->location, blas->precision);
	CukrVecmpTransfer(y, blas->location, blas->precision);  
	
	/* The operation happens in the same way
	 * regardless of the sharing type. For
	 * both cases, only the diagonal elements
	 * are used */
	
	/* Schedule AXPY between all pairs
	 * of parts */
	for (i = 0; i < x->nparts; i++)
		CukrThreadsRun(i, (void*)CukrAXPY_SP, CUKR_FUNCID_AXPY, blas, a, x->cells[i], y->cells[i]);	 	
	
	/* Wait for all threads to finish */
	CukrThreadsWaitCompletion(x->nparts);
	
	/* Invalidate the insync flag for
	 * the y vector, since it is modified*/
	y->insync = CUKR_FALSE;
	
	/* Log */
	LOGX(DEBUG, "[N_%d] AXPY between MP vectors %s "\
				"and %s was computed", cukr_db.mpi_rank, 
				x->name, y->name);				
	
	/* Update counters */
	CLOCK_END;	
	blas->counters_mp[CUKR_BLAS_IDX_TIME] [CUKR_BLAS_OPS_AXPY] += CLOCK_SPAN;
	blas->counters_mp[CUKR_BLAS_IDX_FLOPS][CUKR_BLAS_OPS_AXPY] += 2 * x->n;
					
}



// ------------------------------------------------------------------- //

/**
 * \brief AYPX (ay + x) bewteen two MP
 *        vectors
 * \param blas The blas used for the operation
 * \param a The constant
 * \param x,y The MP vectors
 */
void CukrAYPX_MP(CUKR_BLAS blas, CUKR_DOUBLE a, CUKR_VECTOR_MP x, CUKR_VECTOR_MP y)
{
	int i;
	
	/* Start timing */
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START; 	
	
	/* Log */
	LOGX(DEBUG, "[N_%d] Running AXPY between MP vectors "\
				"%s and %s, with a=%g",cukr_db.mpi_rank, 
			 	x->name, y->name, a);		 	
		
	/* Error if the vectors do not have
	 * the same number of elements */
	if (x->n != y->n) {
		CUKR_FATALX("[N_%d] Cannot compute AXPY "\
		            "for MP vectors %s and %s since they "\
		            "do not have the same number of "\
		            "elements", cukr_db.mpi_rank, x->name, y->name); 
	} 
	
	/* If the vectors do not have the same
	 * location/precision as the BLAS, the 
	 * vectors should be moved to that location
	 * and/or their precision changed */
	 
	CukrVecmpTransfer(x, blas->location, blas->precision);
	CukrVecmpTransfer(y, blas->location, blas->precision);  
	
	/* The operation happens in the same way
	 * regardless of the sharing type. For
	 * both cases, only the diagonal elements
	 * are used */
	
	/* Schedule AYPX between all pairs
	 * of parts */
	for (i = 0; i < x->nparts; i++)
		CukrThreadsRun(i, (void*)CukrAYPX_SP, CUKR_FUNCID_AYPX, blas, a, x->cells[i], y->cells[i]);	 
	
	/* Wait for all threads to finish */
	CukrThreadsWaitCompletion(x->nparts);
	
	/* Invalidate the insync flag for
	 * the y vector, since it is modified*/
	y->insync = CUKR_FALSE;
	
	/* Log message */
	LOGX(DEBUG, "[N_%d] AYPX between MP vectors %s "\
				"and %s was computed", cukr_db.mpi_rank, 
				x->name, y->name);						
	
	/* Update counters */	
	CLOCK_END;	
	blas->counters_mp[CUKR_BLAS_IDX_TIME] [CUKR_BLAS_OPS_AYPX] += CLOCK_SPAN;
	blas->counters_mp[CUKR_BLAS_IDX_FLOPS][CUKR_BLAS_OPS_AYPX] += 2 * x->n;	
}

// ------------------------------------------------------------------- //

/**
 * \brief MP vector scaling 
 * \param blas The blas used for the operation
 * \param a The scaling constant
 * \param x The MP vector being scaled
 */
void CukrSCAL_MP(CUKR_BLAS blas, CUKR_DOUBLE a, CUKR_VECTOR_MP x)
{
	int i;
	
	/* Start timing */
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START; 
	
	/* Log */
	LOGX(DEBUG, "[N_%d] Scaling MP vector %s by a = %g",
	        	cukr_db.mpi_rank, x->name, a);    	
	
	/* If the vectors do not have the same
	 * location/precision as the BLAS, the 
	 * vectors should be moved to that location
	 * and/or their precision changed */
	 
	CukrVecmpTransfer(x, blas->location, blas->precision);
	
	/* The operation happens in the same way
	 * regardless of the sharing type. For
	 * both cases, only the diagonal elements
	 * are used */
	
	/* Schedule SCAL between all pairs
	 * of parts */
	for (i = 0; i < x->nparts; i++)
		CukrThreadsRun(i, (void*)CukrSCAL_SP, CUKR_FUNCID_SCAL, blas, a, x->cells[i]);	 
	
	/* Wait for all threads to finish */
	CukrThreadsWaitCompletion(x->nparts);
	
	/* Invalidate the insync flag for
	 * the x vector, since it is modified*/
	x->insync = CUKR_FALSE;
	
	/* Log message */
	LOGX(DEBUG, "[N_%d] MP vector %s has been scaled", 
			    cukr_db.mpi_rank, x->name);	    
			    
	/* Update counters */		    
	CLOCK_END;		
	blas->counters_mp[CUKR_BLAS_IDX_TIME] [CUKR_BLAS_OPS_SCAL] += CLOCK_SPAN;
	blas->counters_mp[CUKR_BLAS_IDX_FLOPS][CUKR_BLAS_OPS_SCAL] += x->n;    
}

// ------------------------------------------------------------------- //

/**
 * \brief MP vector copy
 * \param blas The blas used for the operation
 * \param x The source MP vector
 * \param y The destination MP vector
 */
void CukrCOPY_MP(CUKR_BLAS blas, CUKR_VECTOR_MP x, CUKR_VECTOR_MP y)
{
	int i;
	
	/* Start timing */
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;  
	
	/* Log */
	LOGX(DEBUG, "[N_%d] Copying MP vector %s to vector %s ",
				cukr_db.mpi_rank, x->name, y->name);		
		
	/* Error if the vectors do not have
	 * the same number of elements */
	if (x->n != y->n) {
		CUKR_FATALX("[N_%d] Cannot run COPY "\
		            "for MP vectors %s and %s since they "\
		            "do not have the same number of "\
		            "elements", cukr_db.mpi_rank, x->name, y->name); 
	} 
	
	/* Malloc at destination. Have to do it now, in 
	 * order to make use of the speciall malloc for
	 * shared vectors. */
	CukrVecmpMalloc(y, blas->location, blas->precision); 
	
	/* If the vectors do not have the same
	 * location/precision as the BLAS, the 
	 * vectors should be moved to that location
	 * and/or their precision changed */
	 
	CukrVecmpTransfer(x, blas->location, blas->precision); 
	
	/* The operation happens in the same way
	 * regardless of the sharing type. For
	 * both cases, only the diagonal elements
	 * are used */
	
	/* Schedule COPY between all pairs
	 * of parts */

	for (i = 0; i < x->nparts; i++)
		CukrThreadsRun(i, (void*)CukrCOPY_SP, CUKR_FUNCID_COPY, blas, x->cells[i], y->cells[i]);	 
	
	/* Wait for all threads to finish */
	CukrThreadsWaitCompletion(x->nparts);	
	
	/* If shared, sync. the status of the rows
	 * to the status of the cells, for the dest
	 * vector */
	if (y->shtype == CUKR_SHTYPE_SHARED) 
		for (i = 0; i < x->nparts; i++)
			 y->rows[i]->status = y->cells[i]->status;  	 
	
	/* Invalidate the insync flag for
	 * the y vector, since it is modified*/
	y->insync = CUKR_FALSE;
	
	/* Log */
	LOGX(DEBUG, "[N_%d] MP vector %s has been copied to vector %s", 
				cukr_db.mpi_rank, x->name, y->name);			
				
	/* Update counters */			
	CLOCK_END;	
	blas->counters_mp[CUKR_BLAS_IDX_TIME] [CUKR_BLAS_OPS_COPY] += CLOCK_SPAN;	
}
