/**
 * @file blas/blas2_mp.c
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains the interface to BLAS2 routines.
 * MP version.
 * 
 * The difference between these routines 
 * and the ones in blas.h is that these
 * are written as function of abstract
 * vectors and matrices, therefor work
 * independent of format, type and
 * location.
 * 	 
 */

#define LOGGER_NAME "CUKr.blas.mp.blas2_mp"

#include <stdio.h>
#include "../../util/log.h"
#include "../../util/macros.h"
#include "../../util/utils.h"
#include "../../db/db.h"
#include "../../threading/threading.h"
#include "../sp/blas2_sp.h"
#include "blas2_mp.h"

/* Get access to the database */ 
extern CukrDB cukr_db;  

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiplication
 * \param alpha Scaling factor for Ax
 * \param mat The MP matrix
 * \param x MP vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result MP vector
 */ 
void CukrSPMV_MP(CUKR_BLAS blas, CUKR_DOUBLE alpha, CUKR_MATRIX_MP mat, 
                 CUKR_VECTOR_MP x, CUKR_DOUBLE beta, CUKR_VECTOR_MP y)
{	
	/* Start timing */
 	struct timeval t_start, t_end,  t_start2, t_end2;
	struct timezone t_zone;
 	CLOCK_START;
	
	
	int i;
	
	/* Log */
	LOGX(DEBUG, "Running product between MP matrix %s " \
	            "and MP vector %s (result in MP vector %s)",
	            mat->name, x->name, y->name);
	            
	/* 
	 * Check for agreement between dimmensions 
	 */
	 
	if (mat->cols != x->n || mat->rows != y->n)
		CUKR_FATALX("Cannot compute SpMV for MP matrix %s "\
	    	        "and MP vectors %s and %s since the "\
	        	    "dimmensions do not agree ", 
	        	    mat->name, x->name, y->name); 
	
	/*
	 * The x vector must be shared
	 */
	 
	if (x->shtype != CUKR_SHTYPE_SHARED)
		CUKR_FATALX("Cannot compute SpMV between MP matrix %s "\
	    	        "and MP vector %s because the vector is not shared ",
	    	        mat->name, x->name);  
 	
 	/*
 	 * Convert if necessary
 	 */
 	
 	int mat_format           = CukrMatFormatStrToId(CukrDBGetS(CUKR_DB_MATRIX_FORMAT)); 
	int mat_format_variation = CukrMatFormatVariationStrToId(CukrDBGetS(CUKR_DB_MATRIX_FORMAT_VAR)); 	 	
	CukrMatmpConvert(mat, mat_format, mat_format_variation);
 	
 	
 	/*
	 * Transfer if necessary
	 */
	
	/* If the vectors or the matrix do not have 
	 * the same location/precision as the BLAS, the 
	 * vectors should be moved to that location
	 * and/or their precision changed */
	 
	CukrMatmpTransfer(mat, blas->location, blas->precision); 
	CukrVecmpTransfer(x,   blas->location, blas->precision);
	
	/* Transfer only if beta not zero. If so, y 
	 * would be overwritten anyway, no need to 
	 * transfer, just malloc */ 
	if (beta)
		CukrVecmpTransfer(y, blas->location, blas->precision);
	else
		CukrVecmpMalloc(y, blas->location, blas->precision);  
 	
 	/*
 	 * Synchronize only if using more then one
 	 * thread 
 	 */
 	 	
 	int n_threads = CukrDBGetI(CUKR_DB_PE);	 
 	if (n_threads > 1)
 	{	
		/*
		 * Sychronize shared vector
		 */
		 
		CLOCK_START2;
		CukrVecmpSynchronize(x, mat->comm_table, blas->location, blas->precision);
		CLOCK_END2;
		blas->comm_inner += CLOCK_SPAN2;				
 	}
 	
	/*
	 * Compute 
	 */
	 
	/* Schedule SPMV between all pairs
	 * of parts */
	for (i = 0; i < mat->nparts; i++)
		CukrThreadsRun(i, (void*)CukrSPMV_SP, CUKR_FUNCID_SPMV, blas, alpha, mat->part[i], x->rows[i], beta, y->cells[i]);	 
		
	/* Wait for all threads to finish */
	CukrThreadsWaitCompletion(x->nparts);
	
	/* Invalidate the insync flag for
	 * the y vector, since it is modified*/
	y->insync = CUKR_FALSE;
	
	
	/*
	 * End 
	 */
	 
	 /* Log */
	 LOGX(DEBUG, "Succefully computed product between MP matrix %s " \
	             "and MP vector %s (result in MP vector %s)",
	             mat->name, x->name, y->name);	             
	 
	 /* Update counters */ 
	 CLOCK_END;
	 blas->counters_mp[CUKR_BLAS_IDX_TIME] [CUKR_BLAS_OPS_SPMV] += CLOCK_SPAN;
	 blas->counters_mp[CUKR_BLAS_IDX_FLOPS][CUKR_BLAS_OPS_SPMV] += 2 * mat->nz;	                       
}

