/**
 * @file blas/sp/blas1_sp.c
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains the interface to BLAS1 routines.
 * 
 * The difference between these routines 
 * and the ones in blas.h is that these
 * are written as function of abstract
 * vectors and matrices, therefor work
 * independent of format, type and
 * location.
 * 	
 */ 

#define LOGGER_NAME "CUKr.blas.sp.blas1_sp"  

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
 
#include "blas1_sp.h"
#include "../../util/log.h"
#include "../../util/utils.h"
#include "../../util/macros.h"
#include "../../db/db.h"

/* Get access to the database */
extern CukrDB cukr_db;

// ------------------------------------------------------------------- //

/**
 * \brief Dot product
 * \param blas Pointer to blas used for 
 *             the operation
 * \param x,y The vectors
 */

CUKR_DOUBLE CukrDOT_SP(CUKR_BLAS blas, CUKR_VECTOR_SP x, CUKR_VECTOR_SP y)
{
	#ifndef CUKR_MP
		/* Start timing */
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START; 
	#endif
	
	CUKR_DOUBLE dval = 0;
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));
	
	/* Log */
	LOGX(DEBUG, "[N_%d][T_%d] Running dot product between "\
	            "vectors %s and %s", cukr_db.mpi_rank,
	            *cukr_activeThread, x->name, y->name);
	
	/* Error if the vectors do not have
	 * the same number of elements */
	if (x->n != y->n)
		CUKR_FATALX("[N_%d][T_%d] Cannot compute the dot product "\
		            "for vectors %s and %s since they do not "\
		            "have the same number of elements", cukr_db.mpi_rank,
	            	*cukr_activeThread, x->name, y->name);
	
	/* If the vectors do not have the same
	 * location/precision as the BLAS, the 
	 * vectors should be moved to that location
	 * and/or their precision changed */
	 
	CukrVecspTransfer(x, blas->location, blas->precision);
	CukrVecspTransfer(y, blas->location, blas->precision);  
	 
	 
	/* Now run the BLAS operation at the BLAS
	 * location and in the BLAS precision.
	 * DOT does not change the vectors, so the
	 * status will stay the same. */
	
	
	/* If the blas is on the CPU */
	if (blas->location == CUKR_LOCATION_CPU)
	{
		if (blas->precision == CUKR_TYPE_DOUBLE)
			dval = CukrBlasDDOT(blas, x->n, x->ph_dval, y->ph_dval);

		else if (blas->precision == CUKR_TYPE_QDOUBLE)
			dval = CukrBlasQDOT(blas, x->n, x->ph_sval, x->ph_qvalt,
											y->ph_sval, y->ph_qvalt);
		else
			dval = CukrBlasSDOT(blas, x->n, x->ph_sval, y->ph_sval);
	}	
	/* If the blas is on the GPU */
	else
	{		
		if (blas->precision == CUKR_TYPE_DOUBLE)
			#ifdef	CUKR_USE_OPENCL
				dval = CukrCLBlasDDOT(blas, x->n, x->pcl_dval, y->pcl_dval);
			#else
				dval = CukrBlasDDOT(blas, x->n, x->pd_dval, y->pd_dval);
			#endif
		else if (blas->precision == CUKR_TYPE_QDOUBLE)
		{
			/* Do the padded version, if needed */
			#ifdef CUKR_QDOT_PADDING
				CukrVecPadToMultGPU(x, CUKR_TYPE_QDOUBLE, PADDING_QDOT); 
				CukrVecPadToMultGPU(y, CUKR_TYPE_QDOUBLE, PADDING_QDOT);
				#ifdef CUKR_USE_OPENCL
					dval = CukrCLBlasQDOT(blas, x->n_padded, x->pcl_sval, x->pcl_qvalt,
															 y->pcl_sval, y->pcl_qvalt);
				#else
					dval = CukrBlasQDOT(blas, x->n_padded, x->pd_sval, x->pd_qvalt,
														   y->pd_sval, y->pd_qvalt);
				#endif
			#else
				/* If not, do the regular one */
				#ifdef CUKR_USE_OPENCL
					dval = CukrCLBlasQDOT(blas, x->n, x->pcl_sval, x->pcl_qvalt,
													  y->pcl_sval, y->pcl_qvalt);
				#else
					dval = CukrBlasQDOT(blas, x->n, x->pd_sval, x->pd_qvalt,
													y->pd_sval, y->pd_qvalt);
				#endif
			#endif
		}
		else
		{
			/* Do the padded version, if needed */
			#ifdef CUKR_SDOT_PADDING
				CukrVecPadToMultGPU(x, CUKR_TYPE_SINGLE, PADDING_SDOT); 
				CukrVecPadToMultGPU(y, CUKR_TYPE_SINGLE, PADDING_SDOT);
				#ifdef CUKR_USE_OPENCL
					dval = CukrCLBlasSDOT(blas, x->n_padded, x->pcl_sval, y->pcl_sval);
				#else
					dval = CukrBlasSDOT(blas, x->n_padded, x->pd_sval, y->pd_sval);
				#endif
			#else
				/* If not, do the regular one */
				#ifdef CUKR_USE_OPENCL
					dval = CukrCLBlasSDOT(blas, x->n, x->pcl_sval, y->pcl_sval);
				#else
					dval = CukrBlasSDOT(blas, x->n, x->pd_sval, y->pd_sval);
				#endif
			#endif
		}
	}
	
	/* Log */
	LOGX(DEBUG, "[N_%d][T_%d] Dot product between vectors "\
	            "%s and %s returned %g", cukr_db.mpi_rank, 
	            *cukr_activeThread, x->name, y->name, dval);
	
	#ifndef CUKR_MP
		/* Update counters */
		CLOCK_END;
		blas->counters_mp[CUKR_BLAS_IDX_TIME] [CUKR_BLAS_OPS_DOT] += CLOCK_SPAN;
		blas->counters_mp[CUKR_BLAS_IDX_FLOPS][CUKR_BLAS_OPS_DOT] += 2 * x->n;  
	#endif

	return dval;
}

// ------------------------------------------------------------------- //

/**
 * \brief AXPY (ax + y)
 * \param blas Pointer to blas used for 
 *             the operation
 * \param a The constant
 * \param x,y The vectors
 */ 
void CukrAXPY_SP(CUKR_BLAS blas, CUKR_DOUBLE a, CUKR_VECTOR_SP x, CUKR_VECTOR_SP y)
{
	#ifndef CUKR_MP
		/* Start timing */
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START; 
	#endif
	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));

	/* Log */
	LOGX(DEBUG, "[N_%d][T_%d] Running AXPY between vectors " \
	            "%s and %s with a = %g", cukr_db.mpi_rank, 
	            *cukr_activeThread, x->name, y->name, a);
	
	/* Error if the vectors do not have
	 * the same number of elements */
	if (x->n != y->n)
		CUKR_FATALX("[N_%d][T_%d] Cannot compute AXPY for "\
		            "vectors %s and %s since they do not "\
		            "have the same number of elements", 
		            cukr_db.mpi_rank, *cukr_activeThread,
		            x->name, y->name); 
	
	/* If the vectors do not have the same
	 * location/precision as the BLAS, the 
	 * vectors should be moved to that location
	 * and/or their precision changed */
	 
	CukrVecspTransfer(x, blas->location, blas->precision);
	CukrVecspTransfer(y, blas->location, blas->precision); 
	 
	/* Now run the BLAS operation at the BLAS
	 * location and in the BLAS precision.
	 * AXPY changes the second vector, so its
	 * status will be resetted */
	
	/* If the blas is on the CPU */
	if (blas->location == CUKR_LOCATION_CPU)
	{
		if (blas->precision == CUKR_TYPE_DOUBLE)
			CukrBlasDAXPY(blas, x->n, a , x->ph_dval, y->ph_dval);
		else if (blas->precision == CUKR_TYPE_QDOUBLE)
			CukrBlasQAXPY(blas, x->n, a ,x->ph_sval, x->ph_qvalt, 
										 y->ph_sval, y->ph_qvalt);			
		else 
			CukrBlasSAXPY(blas, x->n, (CUKR_SINGLE)a, x->ph_sval, y->ph_sval);
	}	
	/* If the blas is on the GPU */
	else
	{
		if (blas->precision == CUKR_TYPE_DOUBLE)
			#ifdef CUKR_USE_OPENCL
				CukrCLBlasDAXPY(blas, x->n, a, x->pcl_dval, y->pcl_dval);
			#else
				CukrBlasDAXPY(blas, x->n, a, x->pd_dval, y->pd_dval);
			#endif
		else if (blas->precision == CUKR_TYPE_QDOUBLE)
		{
			/* Do the padded version, if needed */
			#ifdef CUKR_QAXPY_PADDING
				CukrVecPadToMultGPU(x, CUKR_TYPE_QDOUBLE, CUKR_QAXPY_PADDING); 
				CukrVecPadToMultGPU(y, CUKR_TYPE_QDOUBLE, CUKR_QAXPY_PADDING);
					#ifdef CUKR_USE_OPENCL
						CukrCLBlasQAXPY(blas, x->n_padded, a, x->pcl_sval, x->pcl_qvalt, 
															  y->pcl_sval, y->pcl_qvalt);
					#else
						CukrBlasQAXPY(blas, x->n_padded, a ,x->pd_sval, x->pd_qvalt, 
															y->pd_sval, y->pd_qvalt);
					#endif
			#else
				/* If not, do the regular one */
				#ifdef CUKR_USE_OPENCL
					CukrCLBlasQAXPY(blas, x->n, a, x->pcl_sval, x->pcl_qvalt, 
												   y->pcl_sval, y->pcl_qvalt);
				#else
					CukrBlasQAXPY(blas, x->n, a ,x->pd_sval, x->pd_qvalt, 
												 y->pd_sval, y->pd_qvalt);
				#endif
			#endif

		}	
		else
		{	
			/* Do the padded version, if needed */
			#ifdef CUKR_SAXPY_PADDING
				CukrVecPadToMultGPU(x, CUKR_TYPE_SINGLE, CUKR_SAXPY_PADDING); 
				CukrVecPadToMultGPU(y, CUKR_TYPE_SINGLE, CUKR_SAXPY_PADDING);
					#ifdef CUKR_USE_OPENCL
						CukrCLBlasSAXPY(blas, x->n_padded, (CUKR_SINGLE)a, x->pcl_sval, y->pcl_sval);
					#else
						CukrBlasSAXPY(blas, x->n, (CUKR_SINGLE)a, x->pd_sval, y->pd_sval);
					#endif
			#else 
					/* If not, do the regular one */
					#ifdef CUKR_USE_OPENCL
						CukrCLBlasSAXPY(blas, x->n, (CUKR_SINGLE)a, x->pcl_sval, y->pcl_sval);
					#else
						CukrBlasSAXPY(blas, x->n, (CUKR_SINGLE)a, x->pd_sval, y->pd_sval);
					#endif
			#endif
		}
	}
	
	/* Make vector y, which was modified, 
	 * valid only at the location of the BLAS */
	y->status = CukrLocIdAndTypeIdToStatusId(blas->location, blas->precision); 
	/* Special case: precision qdouble
	 * In this case, set single as well since
	 * it is automatically valid */
	if (blas->precision == CUKR_TYPE_QDOUBLE)
		y->status |= CukrLocIdAndTypeIdToStatusId(blas->location, CUKR_TYPE_SINGLE); 

	/* Log */
	LOGX(DEBUG, "[N_%d][T_%d] AXPY between vectors " \
				"%s and %s was computed", cukr_db.mpi_rank, 
				*cukr_activeThread, x->name, y->name);
				
	#ifndef CUKR_MP
		/* Update counters */
		CLOCK_END;
		blas->counters_mp[CUKR_BLAS_IDX_TIME] [CUKR_BLAS_OPS_AXPY] += CLOCK_SPAN;
		blas->counters_mp[CUKR_BLAS_IDX_FLOPS][CUKR_BLAS_OPS_AXPY] += 2 * x->n; 
	#endif				
}

// ------------------------------------------------------------------- //

/**
 * \brief AXPY (ay + x)
 * \param blas Pointer to blas used for 
 *             the operation
 * \param a The constant
 * \param x,y The vectors
 */
void CukrAYPX_SP(CUKR_BLAS blas, CUKR_DOUBLE a, CUKR_VECTOR_SP x, CUKR_VECTOR_SP y)
{
	#ifndef CUKR_MP
		/* Start timing */
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START; 
	#endif
	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));

	/* Log */
	LOGX(DEBUG, "[N_%d][T_%d] Running AYPX between vectors "\
				"%s and %s with a = %g", cukr_db.mpi_rank, 
				*cukr_activeThread, x->name, y->name, a);
	
	/* Error if the vectors do not have
	 * the same number of elements */
	if (x->n != y->n)
		CUKR_FATALX("[N_%d][T_%d] Cannot compute AYPX "\
		            "for vectors %s and %s since they "\
		            "do not have the same number of "\
		            "elements", cukr_db.mpi_rank, *cukr_activeThread,
		            x->name, y->name); 
	
	/* If the vectors do not have the same
	 * location/precision as the BLAS, the 
	 * vectors should be moved to that location
	 * and/or their precision changed */
	 
	CukrVecspTransfer(x, blas->location, blas->precision);
	CukrVecspTransfer(y, blas->location, blas->precision);  
	 
	/* Now run the BLAS operation at the BLAS
	 * location and in the BLAS precision 
	 * AXPY changes the second vector, so its
	 * status will be resetted */
	
	/* If the blas is on the CPU */
	if (blas->location == CUKR_LOCATION_CPU)
	{
		//printf("Just to call CukrBlasDAYPX\n");
		if (blas->precision == CUKR_TYPE_DOUBLE)
			CukrBlasDAYPX(blas, x->n, a , x->ph_dval, y->ph_dval);
		else if (blas->precision == CUKR_TYPE_QDOUBLE)
			CukrBlasQAYPX(blas, x->n, a ,x->ph_sval, x->ph_qvalt, 
										 y->ph_sval, y->ph_qvalt);
		else
			CukrBlasSAYPX(blas, x->n, (CUKR_SINGLE)a, x->ph_sval, y->ph_sval);
	}	
	/* If the blas is on the GPU */
	else
	{ 			   
		if (blas->precision == CUKR_TYPE_DOUBLE)
			#ifdef CUKR_USE_OPENCL
				CukrCLBlasDAYPX(blas, x->n, a, x->pcl_dval, y->pcl_dval);
			#else
				CukrBlasDAYPX(blas, x->n, a, x->pd_dval, y->pd_dval);
			#endif
		else if (blas->precision == CUKR_TYPE_QDOUBLE)
		{ 
			//printf("Just to call CukrBlasQAYPX\n");
			/* Do the padded version, if needed */
			#ifdef CUKR_QAYPX_PADDING
				CukrVecPadToMultGPU(x, CUKR_TYPE_QDOUBLE, CUKR_QAYPX_PADDING); 
				CukrVecPadToMultGPU(y, CUKR_TYPE_QDOUBLE, CUKR_QAYPX_PADDING);
					#ifdef CUKR_USE_OPENCL
						CukrCLBlasQAYPX(blas, x->n_padded, a, x->pcl_sval, x->pcl_qvalt, 
															  y->pcl_sval, y->pcl_qvalt);
					#else
						CukrBlasQAYPX(blas, x->n_padded, a,	x->pd_sval, x->pd_qvalt, 
															y->pd_sval, y->pd_qvalt);
					#endif
			#else
				/* If not, do the regular one */
					#ifdef CUKR_USE_OPENCL
						CukrCLBlasQAYPX(blas, x->n, a, x->pcl_sval, x->pcl_qvalt, 
													   y->pcl_sval, y->pcl_qvalt);
					#else
						CukrBlasQAYPX(blas, x->n, a, x->pd_sval, x->pd_qvalt, 
													 y->pd_sval, y->pd_qvalt);
					#endif
			#endif
					
		}
		else
		{	
			//printf("Just to call CukrBlasSAYPX\n");
			/* Do the padded version, if ncukr_activeThreadeeded */
			#ifdef CUKR_SAYPX_PADDING
				CukrVecPadToMultGPU(x, CUKR_TYPE_SINGLE, CUKR_SAYPX_PADDING); 
				CukrVecPadToMultGPU(y, CUKR_TYPE_SINGLE, CUKR_SAYPX_PADDING);
					#ifdef CUKR_USE_OPENCL
						CukrCLBlasSAYPX(blas, x->n_padded, (CUKR_SINGLE)a, x->pcl_sval, y->pcl_sval);
					#else
						CukrBlasSAYPX(blas, x->n_padded, (CUKR_SINGLE)a, x->pd_sval, y->pd_sval);
					#endif
			#else
					/* If not, do the regular one */
					#ifdef CUKR_USE_OPENCL
						CukrCLBlasSAYPX(blas, x->n, (CUKR_SINGLE)a, x->pcl_sval, y->pcl_sval);
					#else
						CukrBlasSAYPX(blas, x->n, (CUKR_SINGLE)a, x->pd_sval, y->pd_sval);
					#endif
			#endif
		}
	}
	
	/* Make vector y, which was modified, 
	 * valid only at the location of the BLAS */
	y->status = CukrLocIdAndTypeIdToStatusId(blas->location, blas->precision);
	/* Special case: precision qdouble
	 * In this case, set single as well since
	 * it is automatically valid */
	if (blas->precision == CUKR_TYPE_QDOUBLE)
		y->status |= CukrLocIdAndTypeIdToStatusId(blas->location, CUKR_TYPE_SINGLE); 

	/* Log */
	LOGX(DEBUG, "[N_%d][T_%d] AYPX between vectors " \
				"%s and %s was computed", cukr_db.mpi_rank, 
				*cukr_activeThread, x->name, y->name);
				
	#ifndef CUKR_MP
		/* Update counters */
		CLOCK_END;
		blas->counters_mp[CUKR_BLAS_IDX_TIME] [CUKR_BLAS_OPS_AYPX] += CLOCK_SPAN;
		blas->counters_mp[CUKR_BLAS_IDX_FLOPS][CUKR_BLAS_OPS_AYPX] += 2 * x->n;	
	#endif			
}

// ------------------------------------------------------------------- //

/**
 * \brief Vector scaling 
 * \param blas Pointer to blas used for 
 *             the operation
 * \param a The scaling constant
 * \param x The vectors being scaled
 */
void CukrSCAL_SP(CUKR_BLAS blas, CUKR_DOUBLE a, CUKR_VECTOR_SP x)
{
	#ifndef CUKR_MP
		/* Start timing */
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START; 
	#endif
	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));

	/* Log */
	LOGX(DEBUG, "[N_%d][T_%d] Scaling vector %s by a = %g", 
	            cukr_db.mpi_rank, *cukr_activeThread, x->name, a);
	
	/* If the vector does not have the same
	 * location/precision as the BLAS, the 
	 * vector should be moved to that location
	 * and/or its precision changed */
	 
	CukrVecspTransfer(x, blas->location, blas->precision);
	 
	/* Now run the BLAS operation at the BLAS
	 * location and in the BLAS precision.
	 * Scalling modifies the vector so the 
	 * status will be resetted */
	
	/* If the blas is on the CPU */
	if (blas->location == CUKR_LOCATION_CPU)
	{
		if (blas->precision == CUKR_TYPE_DOUBLE)
			CukrBlasDSCAL(blas, x->n, a , x->ph_dval);
		else if (blas->precision == CUKR_TYPE_QDOUBLE)
			CukrBlasQSCAL(blas, x->n, a , x->ph_sval, x->ph_qvalt);			
		else
			CukrBlasSSCAL(blas, x->n, (CUKR_SINGLE)a, x->ph_sval);
	}	
	/* If the blas is on the GPU */
	else
	{
		
		if (blas->precision == CUKR_TYPE_DOUBLE)
			#ifdef CUKR_USE_OPENCL
				CukrCLBlasDSCAL(blas, x->n, a, x->pcl_dval);
			#else
				CukrBlasDSCAL(blas, x->n, a, x->pd_dval);
			#endif
		else if (blas->precision == CUKR_TYPE_QDOUBLE) 
		{	
			/* Do the padded version, if needed */
			#ifdef CUKR_QSCAL_PADDING
				CukrVecPadToMultGPU(x, CUKR_TYPE_QDOUBLE, CUKR_QSCAL_PADDING);
				#ifdef CUKR_USE_OPENCL
					CukrCLBlasQSCAL(blas, x->n_padded, a , x->pcl_sval, x->pcl_qvalt);
				#else
					CukrBlasQSCAL(blas, x->n_padded, a , x->pd_sval, x->pd_qvalt);
				#endif
			#else
				/* If not, do the regular one */
				#ifdef CUKR_USE_OPENCL
					CukrCLBlasQSCAL(blas, x->n, a , x->pcl_sval, x->pcl_qvalt);
				#else
					CukrBlasQSCAL(blas, x->n, a , x->pd_sval, x->pd_qvalt);
				#endif
			#endif
		}				
		else{
			/* Do the padded version, if needed */
			#ifdef CUKR_SSCAL_PADDING
				CukrVecPadToMultGPU(x, CUKR_TYPE_SINGLE, CUKR_SAXPY_PADDING);
				#ifdef CUKR_USE_OPENCL
					CukrCLBlasSSCAL(blas, x->n_padded, (CUKR_SINGLE)a, x->pcl_sval);
				#else
					CukrBlasSSCAL(blas, x->n_padded, (CUKR_SINGLE)a, x->pd_sval);
				#endif
			#else
				/* If not, do the regular one */
				#ifdef CUKR_USE_OPENCL
					CukrCLBlasSSCAL(blas, x->n, (CUKR_SINGLE)a, x->pcl_sval);
				#else
					CukrBlasSSCAL(blas, x->n, (CUKR_SINGLE)a, x->pd_sval);
				#endif
			#endif		
		}
	}
	
	/* Make vector x, which was modified, 
	 * valid only at the location of the BLAS */
	x->status = CukrLocIdAndTypeIdToStatusId(blas->location, blas->precision);
	/* Special case: precision qdouble
	 * In this case, set single as well since
	 * it is automatically valid */
	if (blas->precision == CUKR_TYPE_QDOUBLE)
		x->status |= CukrLocIdAndTypeIdToStatusId(blas->location, CUKR_TYPE_SINGLE); 
	
	/* Log */
	LOGX(DEBUG, "[N_%d][T_%d] Vector %s has been scaled", 
	             cukr_db.mpi_rank, *cukr_activeThread, x->name);
	             
	#ifndef CUKR_MP
		/* Update counters */
		CLOCK_END;
		blas->counters_mp[CUKR_BLAS_IDX_TIME] [CUKR_BLAS_OPS_SCAL] += CLOCK_SPAN;
		blas->counters_mp[CUKR_BLAS_IDX_FLOPS][CUKR_BLAS_OPS_SCAL] += x->n;  	
	#endif	             
}

// ------------------------------------------------------------------- //

/**
 * \brief Vector copy
 * \param blas Pointer to blas used for 
 *             the operation
 * \param x The source vector
 * \param y The destination vector
 */
void CukrCOPY_SP(CUKR_BLAS blas, CUKR_VECTOR_SP x, CUKR_VECTOR_SP y)
{
	#ifndef CUKR_MP
		/* Start timing */
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START; 
	#endif
	
	int* cukr_activeThread = pthread_getspecific(cukr_db.thread_key);

	/* Log */
	LOGX(DEBUG, "[N_%d][T_%d] Copying vector %s to vector %s",
	            cukr_db.mpi_rank, *cukr_activeThread, x->name, y->name);
	
	/* Error if the vectors do not have
	 * the same number of elements */
	if (x->n != y->n)
		CUKR_FATALX("[N_%d][T_%d] Cannot run COPY "\
		            "for vectors %s and %s since they "\
		            "do not have the same number of "\
		            "elements", cukr_db.mpi_rank, *cukr_activeThread,
		            x->name, y->name); 
	
	/* If the vectors do not have the same
	 * location/precision as the BLAS, the 
	 * vectors should be moved to that location
	 * and/or their precision changed */
	CukrVecspTransfer(x, blas->location, blas->precision);
	
	/* For copy, the second vector does not 
	 * actually has to be moved, only memory
	 * has to be allocated at the destination */	
	 
	CukrVecspMalloc(y, blas->location, blas->precision);  
		
	/* Now run the BLAS operation at the BLAS
	 * location and in the BLAS precision.
	 * The source vector is not modified, so
	 * its status will be left as it is. */
	
	/* If the blas is on the CPU */
	if (blas->location == CUKR_LOCATION_CPU)
	{
		if (blas->precision == CUKR_TYPE_DOUBLE)
			CukrBlasDCOPY(blas, x->n, x->ph_dval, y->ph_dval);
		else if (blas->precision == CUKR_TYPE_QDOUBLE)
			CukrBlasQCOPY(blas, x->n, x->ph_sval, x->ph_qvalt, 
									  y->ph_sval, y->ph_qvalt);
		else
			CukrBlasSCOPY(blas, x->n, x->ph_sval, y->ph_sval);
	}	
	/* If the blas is on the GPU */
	else 
	{		
		if (blas->precision == CUKR_TYPE_DOUBLE)
			#ifdef CUKR_USE_OPENCL
				CukrCLBlasDCOPY(blas, x->n, x->pcl_dval, y->pcl_dval);
			#else
				CukrBlasDCOPY(blas, x->n, x->pd_dval, y->pd_dval);
			#endif
		else if (blas->precision == CUKR_TYPE_QDOUBLE) 
		{
			/* Do the padded version, if needed */
			#ifdef CUKR_QCOPY_PADDING
				CukrVecPadToMultGPU(x, CUKR_TYPE_QDOUBLE, PADDING_QCOPY); 
				CukrVecPadToMultGPU(y, CUKR_TYPE_QDOUBLE, PADDING_QCOPY);
					#ifdef CUKR_USE_OPENCL
						CukrCLBlasQCOPY(blas, x->n_padded, x->pcl_sval, x->pcl_qvalt,
														   y->pcl_sval, y->pcl_qvalt); 
					#else	
						CukrBlasQCOPY(blas, x->n_padded, x->pd_sval, x->pd_qvalt,
														 y->pd_sval, y->pd_qvalt);
					#endif
			#else
				/* If not, do the regular one */
					#ifdef CUKR_USE_OPENCL
						CukrCLBlasQCOPY(blas, x->n, x->pcl_sval, x->pcl_qvalt, 
													y->pcl_sval, y->pcl_qvalt);
					#else
						CukrBlasQCOPY(blas, x->n, x->pd_sval, x->pd_qvalt, 
												  y->pd_sval, y->pd_qvalt);
					#endif
			#endif
		
		}
		else
		{
			/* Do the padded version, if needed */
			#ifdef CUKR_SCOPY_PADDING
				CukrVecPadToMultGPU(x, CUKR_TYPE_SINGLE, PADDING_SCOPY); 
				CukrVecPadToMultGPU(y, CUKR_TYPE_SINGLE, PADDING_SCOPY);
					#ifdef CUKR_USE_OPENCL
						CukrCLBlasSCOPY(blas, x->n_padded, x->pcl_sval, y->pcl_sval);
					#else
						CukrBlasSCOPY(blas, x->n_padded, x->pd_sval, y->pd_sval);
					#endif
			#else
				/* If not, do the regular one */
					#ifdef CUKR_USE_OPENCL
						CukrCLBlasSCOPY(blas, x->n, x->pcl_sval, y->pcl_sval);
					#else
						CukrBlasSCOPY(blas, x->n, x->pd_sval, y->pd_sval);
					#endif
			#endif
		}
	}
	
	/* Make vector y, which was modified, 
	 * valid only at the location of the BLAS */
	y->status = CukrLocIdAndTypeIdToStatusId(blas->location, blas->precision);
	/* Special case: precision qdouble
	 * In this case, set single as well since
	 * it is automatically valid */
	if (blas->precision == CUKR_TYPE_QDOUBLE)
		y->status |= CukrLocIdAndTypeIdToStatusId(blas->location, CUKR_TYPE_SINGLE);
	
	/* Log */
	LOGX(DEBUG, "[N_%d][T_%d] Vector %s has been copied " \
			    "to vector %s ", cukr_db.mpi_rank, 
			    *cukr_activeThread, x->name, y->name);
			    
	#ifndef CUKR_MP
		/* Update counters */
		CLOCK_END;
		blas->counters_mp[CUKR_BLAS_IDX_TIME] [CUKR_BLAS_OPS_COPY] += CLOCK_SPAN;  	
	#endif			    	
}
