/**
 * @file blas/sp/blas2_sp.c
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains the interface to BLAS2 routines.
 * 
 * The difference between these routines 
 * and the ones in blas->h is that these
 * are written as function of abstract
 * vectors and matrices, therefor work
 * independent of format, type and
 * location.
 * 	
 */

#define LOGGER_NAME "CUKr.blas.sp.blas2_sp"  

#include <stdio.h>
#include <stdlib.h>
#include "../../mat_vec/matrix.h"
#include "../sp/blas1_sp.h"
#include "../sp/blas2_sp.h"
#include "../../util/log.h"
#include "../../util/utils.h"
#include "../../util/macros.h"
#include "../../db/db.h"


/* Get access to the database */
extern CukrDB cukr_db;

// ------------------------------------------------------------------- //

/**
 * \brief Sparse matrix vector multiplication
 * \param alpha Scaling factor for Ax
 * \param mat The matrix
 * \param x Vector being multiplied
 * \param beta Scaling factor for y
 * \param y Result vector
 */ 
void CukrSPMV_SP(CUKR_BLAS blas, CUKR_DOUBLE alpha, CUKR_MATRIX_SP mat, 
              CUKR_VECTOR_SP x, CUKR_DOUBLE beta, CUKR_VECTOR_SP y)
{
	#ifndef CUKR_MP
		/* Start timing */
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START; 
	#endif
	
	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));

	/* Log */
	LOGX(DEBUG, "[N_%d][T_%d] Running product between matrix %s " \
	            "and vector %s (result in vector %s)", cukr_db.mpi_rank,
	            *cukr_activeThread, mat->name, x->name, y->name);
		
	/*
	 * Validate parameters
	 */
	            
	/* Error if the vectors and matrix do not 
	 * have the same number of elements */
	  
	if (mat->cols != x->n || mat->rows != y->n)
		CUKR_FATALX("[N_%d][T_%d] Cannot compute SpMV %s = %s*%s since the "\
	        	    "dimmensions do not agree (%s.cols=%d, "\
	        	    "%s.n=%d; %s.rows=%d, %s.n=%d)", 
	        	    cukr_db.mpi_rank, *cukr_activeThread,
	        	    y->name, mat->name, x->name, mat->name,
	        	    mat->cols, x->name, x->n,
	        	    mat->name, mat->rows, y->name, y->n); 

	
	/*
	 * Convert if the options have been set 
	 */
	
	//int mat_format           = CukrMatFormatStrToId(CukrDBGetS(CUKR_DB_MATRIX_FORMAT)); 
	//int mat_format_variation = CukrMatFormatVariationStrToId(CukrDBGetS("CUKR_DB_MATRIX_FORMAT_VARIATION"));	
	//CukrMatspConvert(mat, mat_format, mat_format_variation);	
	
	/*
	 * Transfer if necessary
	 */
	
	/* If the vectors or the matrix do not have 
	 * the same location/precision as the BLAS, the 
	 * vectors should be moved to that location
	 * and/or their precision changed */
	 
	CukrMatspTransfer(mat, blas->location, blas->precision); 
	CukrVecspTransfer(x, blas->location, blas->precision);
	
	/* Transfer only if beta not zero. If so, y 
	 * would be overwritten anyway, no need to 
	 * transfer, just malloc */ 
	if (beta)
		CukrVecspTransfer(y, blas->location, blas->precision);
	else
		CukrVecspMalloc(y, blas->location, blas->precision);    
		
	/*
	 * Run the operation
	 * The SpMV operation changes only the
	 * destiantion vector, so only the status
	 * of the destination vector will be reset.
	 */
	
	 /* If in CSR */ 
	if (mat->format == CUKR_MATRIX_FORMAT_CSR)
	{ 
	 	
		/* If on the CPU, do not care about the
	 	 * variation */
	 	if (blas->location == CUKR_LOCATION_CPU)
		{
			if (blas->precision == CUKR_TYPE_DOUBLE) {
				CukrBlasDSPMV_CSR(blas, mat->rows, mat->cols, mat->nz, alpha, 
				                  mat->csr_mat->ptr->ph_ival, mat->csr_mat->idx->ph_ival, 
				                  mat->csr_mat->val->ph_dval, x->ph_dval, beta, y->ph_dval);
			} 
			if (blas->precision == CUKR_TYPE_QDOUBLE) {
				CukrBlasQSPMV_CSR(blas, mat->rows, mat->cols, mat->nz, alpha, 
				                  mat->csr_mat->ptr->ph_ival, mat->csr_mat->idx->ph_ival, 
				                  mat->csr_mat->val->ph_sval, mat->csr_mat->val->ph_qvalt,
				                  x->ph_sval, x->ph_qvalt, beta, y->ph_sval, y->ph_qvalt);
			} 
			else if (blas->precision == CUKR_TYPE_SINGLE) {
				CukrBlasSSPMV_CSR(blas, mat->rows, mat->cols, mat->nz, alpha, 
				                  mat->csr_mat->ptr->ph_ival, mat->csr_mat->idx->ph_ival, 
				                  mat->csr_mat->val->ph_sval, x->ph_sval, beta, y->ph_sval);
			} 
		}	
		/* If the blas is on the GPU,
		 * we care about the variation
		 * (of course, because the variations
		 * are made to optimize SpMV on GPU)*/
		else
		{
			#ifdef CUKR_USE_OPENCL
			if (blas->precision == CUKR_TYPE_DOUBLE) {
				/* If CSR4 variation, use the specific function */
				if (mat->variation == CUKR_MATRIX_FORMAT_VARIATION_CSR4)
					CukrCLBlasDSPMV_CSR4(blas, mat->rows, mat->cols, mat->nz, alpha, 
				                  	   mat->csr_mat->ptr->pcl_ival, mat->csr_mat->idx->pcl_ival, 
				                       mat->csr_mat->val->pcl_dval, x->pcl_dval, beta, y->pcl_dval);
				/* Else, use generic function */
				else
					CukrCLBlasDSPMV_CSR(blas, mat->rows, mat->cols, mat->nz, alpha, 
				                  	  mat->csr_mat->ptr->pcl_ival, mat->csr_mat->idx->pcl_ival, 
				                      mat->csr_mat->val->pcl_dval, x->pcl_dval, beta, y->pcl_dval);
					
			} 
			if (blas->precision == CUKR_TYPE_QDOUBLE) {
				/* If CSR4 variation, use the specific function */
				if (mat->variation == CUKR_MATRIX_FORMAT_VARIATION_CSR4)
					CukrCLBlasQSPMV_CSR4(blas, mat->rows, mat->cols, mat->nz, alpha, 
				                  	   mat->csr_mat->ptr->pcl_ival, mat->csr_mat->idx->pcl_ival, 
				                  	   mat->csr_mat->val->pcl_sval, mat->csr_mat->val->pcl_qvalt,
				                  	   x->pcl_sval, x->pcl_qvalt, beta, y->pcl_sval, y->pcl_qvalt);
				/* Else, use generic function */
				else
					CukrCLBlasQSPMV_CSR(blas, mat->rows, mat->cols, mat->nz, alpha, 
				                  	  mat->csr_mat->ptr->pcl_ival, mat->csr_mat->idx->pcl_ival, 
				                      mat->csr_mat->val->pcl_sval, mat->csr_mat->val->pcl_qvalt,
				                      x->pcl_sval, x->pcl_qvalt, beta, y->pcl_sval, y->pcl_qvalt);
			} 
			else if (blas->precision == CUKR_TYPE_SINGLE) {
				/* If CSR4 variation, use the specific function */
				if (mat->variation == CUKR_MATRIX_FORMAT_VARIATION_CSR4)
					CukrCLBlasSSPMV_CSR4(blas, mat->rows, mat->cols, mat->nz, alpha, 
				                       mat->csr_mat->ptr->pcl_ival, mat->csr_mat->idx->pcl_ival, 
				                  	   mat->csr_mat->val->pcl_sval, x->pcl_sval, beta, y->pcl_sval);
				/* Else, use generic function */
				else
					CukrCLBlasSSPMV_CSR(blas, mat->rows, mat->cols, mat->nz, alpha, 
				                      mat->csr_mat->ptr->pcl_ival, mat->csr_mat->idx->pcl_ival, 
				                  	  mat->csr_mat->val->pcl_sval, x->pcl_sval, beta, y->pcl_sval);				                  	   
			}
			
			#else
			
			if (blas->precision == CUKR_TYPE_DOUBLE) {
				/* If CSR4 variation, use the specific function */
				if (mat->variation == CUKR_MATRIX_FORMAT_VARIATION_CSR4)
					CukrBlasDSPMV_CSR4(blas, mat->rows, mat->cols, mat->nz, alpha, 
				                  	   mat->csr_mat->ptr->pd_ival, mat->csr_mat->idx->pd_ival, 
				                       mat->csr_mat->val->pd_dval, x->pd_dval, beta, y->pd_dval);
				/* Else, use generic function */
				else
					CukrBlasDSPMV_CSR(blas, mat->rows, mat->cols, mat->nz, alpha, 
				                  	  mat->csr_mat->ptr->pd_ival, mat->csr_mat->idx->pd_ival, 
				                      mat->csr_mat->val->pd_dval, x->pd_dval, beta, y->pd_dval);
				
			} 
			if (blas->precision == CUKR_TYPE_QDOUBLE) {
				/* If CSR4 variation, use the specific function */
				if (mat->variation == CUKR_MATRIX_FORMAT_VARIATION_CSR4)
					CukrBlasQSPMV_CSR4(blas, mat->rows, mat->cols, mat->nz, alpha, 
				                  	   mat->csr_mat->ptr->pd_ival, mat->csr_mat->idx->pd_ival, 
				                  	   mat->csr_mat->val->pd_sval, mat->csr_mat->val->pd_qvalt,
				                  	   x->pd_sval, x->pd_qvalt, beta, y->pd_sval, y->pd_qvalt);
				/* Else, use generic function */
				else
					CukrBlasQSPMV_CSR(blas, mat->rows, mat->cols, mat->nz, alpha, 
				                  	  mat->csr_mat->ptr->pd_ival, mat->csr_mat->idx->pd_ival, 
				                      mat->csr_mat->val->pd_sval, mat->csr_mat->val->pd_qvalt,
				                      x->pd_sval, x->pd_qvalt, beta, y->pd_sval, y->pd_qvalt);
			} 
			else if (blas->precision == CUKR_TYPE_SINGLE) {
				/* If CSR4 variation, use the specific function */
				if (mat->variation == CUKR_MATRIX_FORMAT_VARIATION_CSR4)
					CukrBlasSSPMV_CSR4(blas, mat->rows, mat->cols, mat->nz, alpha, 
				                       mat->csr_mat->ptr->pd_ival, mat->csr_mat->idx->pd_ival, 
				                  	   mat->csr_mat->val->pd_sval, x->pd_sval, beta, y->pd_sval);
				/* Else, use generic function */
				else
					CukrBlasSSPMV_CSR(blas, mat->rows, mat->cols, mat->nz, alpha, 
				                      mat->csr_mat->ptr->pd_ival, mat->csr_mat->idx->pd_ival, 
				                  	  mat->csr_mat->val->pd_sval, x->pd_sval, beta, y->pd_sval);				                  	   
			} 
			#endif // CUKR_USE_OPENCL
		}
	}	
	
    /* If in BCSR */
	else if (mat->format == CUKR_MATRIX_FORMAT_BCSR)
	{

		/* If on the CPU */
		if (blas->location == CUKR_LOCATION_CPU)
		{
			if (blas->precision == CUKR_TYPE_DOUBLE) {
				CukrBlasDSPMV_BCSR(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
								  mat->bcsr_mat->ptr->ph_ival, mat->bcsr_mat->idx->ph_ival,
								  mat->bcsr_mat->val->ph_dval, x->ph_dval, beta, y->ph_dval);
			}
			if (blas->precision == CUKR_TYPE_QDOUBLE) {
				CukrBlasQSPMV_BCSR(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
								  mat->bcsr_mat->ptr->ph_ival, mat->bcsr_mat->idx->ph_ival,
								  mat->bcsr_mat->val->ph_sval, mat->bcsr_mat->val->ph_qvalt,
								  x->ph_sval, x->ph_qvalt, beta, y->ph_sval, y->ph_qvalt);
			}
			else if (blas->precision == CUKR_TYPE_SINGLE) {
				CukrBlasSSPMV_BCSR(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
								  mat->bcsr_mat->ptr->ph_ival, mat->bcsr_mat->idx->ph_ival,
								  mat->bcsr_mat->val->ph_sval, x->ph_sval, beta, y->ph_sval);
			}
		}
		/* If the blas is on the GPU */
		else
		{
			#ifdef CUKR_USE_OPENCL
			if (blas->precision == CUKR_TYPE_DOUBLE) {
				if (mat->variation == CUKR_MATRIX_FORMAT_VARIATION_BCSR4)
					CukrCLBlasDSPMV_BCSR4(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
									  mat->bcsr_mat->ptr->pcl_ival, mat->bcsr_mat->idx->pcl_ival,
									  mat->bcsr_mat->val->pcl_dval, x->pcl_dval, beta, y->pcl_dval);
				else
					CukrCLBlasDSPMV_BCSR(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
									  mat->bcsr_mat->ptr->pcl_ival, mat->bcsr_mat->idx->pcl_ival,
									  mat->bcsr_mat->val->pcl_dval, x->pcl_dval, beta, y->pcl_dval);

			}
			if (blas->precision == CUKR_TYPE_QDOUBLE) {
				if (mat->variation == CUKR_MATRIX_FORMAT_VARIATION_BCSR4)
					CukrCLBlasQSPMV_BCSR4(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
									  mat->bcsr_mat->ptr->pcl_ival, mat->bcsr_mat->idx->pcl_ival,
									  mat->bcsr_mat->val->pcl_sval, mat->bcsr_mat->val->pcl_qvalt,
									  x->pcl_sval, x->pcl_qvalt, beta, y->pcl_sval, y->pcl_qvalt);
				else
					CukrCLBlasQSPMV_BCSR(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
									  mat->bcsr_mat->ptr->pcl_ival, mat->bcsr_mat->idx->pcl_ival,
									  mat->bcsr_mat->val->pcl_sval, mat->bcsr_mat->val->pcl_qvalt,
									  x->pcl_sval, x->pcl_qvalt, beta, y->pcl_sval, y->pcl_qvalt);
			}
			else if (blas->precision == CUKR_TYPE_SINGLE) {
				if (mat->variation == CUKR_MATRIX_FORMAT_VARIATION_BCSR4)
					CukrCLBlasSSPMV_BCSR4(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
									  mat->bcsr_mat->ptr->pcl_ival, mat->bcsr_mat->idx->pcl_ival,
									  mat->bcsr_mat->val->pcl_sval, x->pcl_sval, beta, y->pcl_sval);
				else
					CukrCLBlasSSPMV_BCSR(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
									  mat->bcsr_mat->ptr->pcl_ival, mat->bcsr_mat->idx->pcl_ival,
									  mat->bcsr_mat->val->pcl_sval, x->pcl_sval, beta, y->pcl_sval);
			}
			
			#else
			
			if (blas->precision == CUKR_TYPE_DOUBLE) {
				if (mat->variation == CUKR_MATRIX_FORMAT_VARIATION_BCSR4)
					CukrBlasDSPMV_BCSR4(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
										mat->bcsr_mat->ptr->pd_ival, mat->bcsr_mat->idx->pd_ival,
										mat->bcsr_mat->val->pd_dval, x->pd_dval, beta, y->pd_dval);
				else
					CukrBlasDSPMV_BCSR(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
									   mat->bcsr_mat->ptr->pd_ival, mat->bcsr_mat->idx->pd_ival,
									   mat->bcsr_mat->val->pd_dval, x->pd_dval, beta, y->pd_dval);
				
			}
			if (blas->precision == CUKR_TYPE_QDOUBLE) {
				if (mat->variation == CUKR_MATRIX_FORMAT_VARIATION_BCSR4)
					CukrBlasQSPMV_BCSR4(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
										mat->bcsr_mat->ptr->pd_ival, mat->bcsr_mat->idx->pd_ival,
										mat->bcsr_mat->val->pd_sval, mat->bcsr_mat->val->pd_qvalt,
										x->pd_sval, x->pd_qvalt, beta, y->pd_sval, y->pd_qvalt);
				else
					CukrBlasQSPMV_BCSR(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
									   mat->bcsr_mat->ptr->pd_ival, mat->bcsr_mat->idx->pd_ival,
									   mat->bcsr_mat->val->pd_sval, mat->bcsr_mat->val->pd_qvalt,
									   x->pd_sval, x->pd_qvalt, beta, y->pd_sval, y->pd_qvalt);
			}
			else if (blas->precision == CUKR_TYPE_SINGLE) {
				if (mat->variation == CUKR_MATRIX_FORMAT_VARIATION_BCSR4)
					CukrBlasSSPMV_BCSR4(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
										mat->bcsr_mat->ptr->pd_ival, mat->bcsr_mat->idx->pd_ival,
										mat->bcsr_mat->val->pd_sval, x->pd_sval, beta, y->pd_sval);
				else
					CukrBlasSSPMV_BCSR(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
									   mat->bcsr_mat->ptr->pd_ival, mat->bcsr_mat->idx->pd_ival,
									   mat->bcsr_mat->val->pd_sval, x->pd_sval, beta, y->pd_sval);
			}
			#endif // CUKR_USE_OPENCL
		}
	}

	/* If in HYB */
	else if (mat->format == CUKR_MATRIX_FORMAT_HYB)
	{

		/* If on the CPU, do not care about the
		 * variation */
		if (blas->location == CUKR_LOCATION_CPU)
		{
			if (blas->precision == CUKR_TYPE_DOUBLE) {
				CukrBlasDSPMV_HYB(blas, mat->rows, mat->cols, mat->nz, alpha,
								  mat->hyb_mat->ell_nz_per_row, mat->hyb_mat->ell_stride, mat->hyb_mat->csr_nz,
								  mat->hyb_mat->ell_idx->ph_ival, mat->hyb_mat->ell_val->ph_dval, mat->hyb_mat->csr_ptr->ph_ival, mat->hyb_mat->csr_idx->ph_ival, mat->hyb_mat->csr_val->ph_dval,
								  x->ph_dval, beta, y->ph_dval);
			}
			if (blas->precision == CUKR_TYPE_QDOUBLE) {
				CukrBlasQSPMV_HYB(blas, mat->rows, mat->cols, mat->nz, alpha,
								  mat->hyb_mat->ell_nz_per_row, mat->hyb_mat->ell_stride, mat->hyb_mat->csr_nz,
								  mat->hyb_mat->ell_idx->ph_ival, mat->hyb_mat->ell_val->ph_sval, mat->hyb_mat->ell_val->ph_qvalt, mat->hyb_mat->csr_ptr->ph_ival, mat->hyb_mat->csr_idx->ph_ival, mat->hyb_mat->csr_val->ph_sval, mat->hyb_mat->csr_val->ph_qvalt,
								  x->ph_sval, x->ph_qvalt, beta, y->ph_sval, y->ph_qvalt);
			}
			else if (blas->precision == CUKR_TYPE_SINGLE) {
				CukrBlasSSPMV_HYB(blas, mat->rows, mat->cols, mat->nz, alpha,
								  mat->hyb_mat->ell_nz_per_row, mat->hyb_mat->ell_stride, mat->hyb_mat->csr_nz,
								  mat->hyb_mat->ell_idx->ph_ival, mat->hyb_mat->ell_val->ph_sval, mat->hyb_mat->csr_ptr->ph_ival, mat->hyb_mat->csr_idx->ph_ival, mat->hyb_mat->csr_val->ph_sval,
								  x->ph_sval, beta, y->ph_sval);
			}
		}
		/* If the blas is on the GPU,
		 * we care about the variation
		 * (of course, because the variations
		 * are made to optimize SpMV on GPU)*/
		else
		{
			#ifdef CUKR_USE_OPENCL
			if (blas->precision == CUKR_TYPE_DOUBLE) {
					CukrCLBlasDSPMV_HYB(blas, mat->rows, mat->cols, mat->nz, alpha,
									  mat->hyb_mat->ell_nz_per_row, mat->hyb_mat->ell_stride, mat->hyb_mat->csr_nz,
									  mat->hyb_mat->ell_idx->pcl_ival, mat->hyb_mat->ell_val->pcl_dval, mat->hyb_mat->csr_ptr->pcl_ival, mat->hyb_mat->csr_idx->pcl_ival, mat->hyb_mat->csr_val->pcl_dval,
									  x->pcl_dval, beta, y->pcl_dval);

			}
			if (blas->precision == CUKR_TYPE_QDOUBLE) {
					CukrCLBlasQSPMV_HYB(blas, mat->rows, mat->cols, mat->nz, alpha,
								  mat->hyb_mat->ell_nz_per_row, mat->hyb_mat->ell_stride, mat->hyb_mat->csr_nz,
								  mat->hyb_mat->ell_idx->pcl_ival, mat->hyb_mat->ell_val->pcl_sval, mat->hyb_mat->ell_val->pcl_qvalt, mat->hyb_mat->csr_ptr->pcl_ival, mat->hyb_mat->csr_idx->pcl_ival, mat->hyb_mat->csr_val->pcl_sval, mat->hyb_mat->csr_val->pcl_qvalt,
								  x->pcl_sval, x->pcl_qvalt, beta, y->pcl_sval, y->pcl_qvalt);

			}
			else if (blas->precision == CUKR_TYPE_SINGLE) {
					CukrCLBlasSSPMV_HYB(blas, mat->rows, mat->cols, mat->nz, alpha,
								  mat->hyb_mat->ell_nz_per_row, mat->hyb_mat->ell_stride, mat->hyb_mat->csr_nz,
								  mat->hyb_mat->ell_idx->pcl_ival, mat->hyb_mat->ell_val->pcl_sval, mat->hyb_mat->csr_ptr->pcl_ival, mat->hyb_mat->csr_idx->pcl_ival, mat->hyb_mat->csr_val->pcl_sval,
								  x->pcl_sval, beta, y->pcl_sval);
			}
			#else
			if (blas->precision == CUKR_TYPE_DOUBLE) {
				CukrBlasDSPMV_HYB(blas, mat->rows, mat->cols, mat->nz, alpha,
								  mat->hyb_mat->ell_nz_per_row, mat->hyb_mat->ell_stride, mat->hyb_mat->csr_nz,
								  mat->hyb_mat->ell_idx->pd_ival, mat->hyb_mat->ell_val->pd_dval, mat->hyb_mat->csr_ptr->pd_ival, mat->hyb_mat->csr_idx->pd_ival, mat->hyb_mat->csr_val->pd_dval,
								  x->pd_dval, beta, y->pd_dval);
				
			}
			if (blas->precision == CUKR_TYPE_QDOUBLE) {
				CukrBlasQSPMV_HYB(blas, mat->rows, mat->cols, mat->nz, alpha,
								  mat->hyb_mat->ell_nz_per_row, mat->hyb_mat->ell_stride, mat->hyb_mat->csr_nz,
								  mat->hyb_mat->ell_idx->pd_ival, mat->hyb_mat->ell_val->pd_sval, mat->hyb_mat->ell_val->pd_qvalt, mat->hyb_mat->csr_ptr->pd_ival, mat->hyb_mat->csr_idx->pd_ival, mat->hyb_mat->csr_val->pd_sval, mat->hyb_mat->csr_val->pd_qvalt,
								  x->pd_sval, x->pd_qvalt, beta, y->pd_sval, y->pd_qvalt);
				
			}
			else if (blas->precision == CUKR_TYPE_SINGLE) {
				CukrBlasSSPMV_HYB(blas, mat->rows, mat->cols, mat->nz, alpha,
								  mat->hyb_mat->ell_nz_per_row, mat->hyb_mat->ell_stride, mat->hyb_mat->csr_nz,
								  mat->hyb_mat->ell_idx->pd_ival, mat->hyb_mat->ell_val->pd_sval, mat->hyb_mat->csr_ptr->pd_ival, mat->hyb_mat->csr_idx->pd_ival, mat->hyb_mat->csr_val->pd_sval,
								  x->pd_sval, beta, y->pd_sval);
			}
			#endif
		}
	}


	/* If in BHYB */
	else if (mat->format == CUKR_MATRIX_FORMAT_BHYB)
	{

		/* If on the CPU, do not care about the
		 * variation */
		if (blas->location == CUKR_LOCATION_CPU)
		{
			if (blas->precision == CUKR_TYPE_DOUBLE) {
				CukrBlasDSPMV_BHYB(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
								  mat->bhyb_mat->bell_nz_per_row, mat->bhyb_mat->bell_stride, mat->bhyb_mat->bcsr_nz,
								  mat->bhyb_mat->bell_idx->ph_ival, mat->bhyb_mat->bell_val->ph_dval, mat->bhyb_mat->bcsr_ptr->ph_ival, mat->bhyb_mat->bcsr_idx->ph_ival, mat->bhyb_mat->bcsr_val->ph_dval,
								  x->ph_dval, beta, y->ph_dval);

			}
			if (blas->precision == CUKR_TYPE_QDOUBLE) {
				CukrBlasQSPMV_BHYB(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
								  mat->bhyb_mat->bell_nz_per_row, mat->bhyb_mat->bell_stride, mat->bhyb_mat->bcsr_nz,
								  mat->bhyb_mat->bell_idx->ph_ival, mat->bhyb_mat->bell_val->ph_sval, mat->bhyb_mat->bell_val->ph_qvalt, mat->bhyb_mat->bcsr_ptr->ph_ival, mat->bhyb_mat->bcsr_idx->ph_ival, mat->bhyb_mat->bcsr_val->ph_sval, mat->bhyb_mat->bcsr_val->ph_qvalt,
								  x->ph_sval, x->ph_qvalt, beta, y->ph_sval, y->ph_qvalt);

			}
			else if (blas->precision == CUKR_TYPE_SINGLE) {
				CukrBlasSSPMV_BHYB(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
								  mat->bhyb_mat->bell_nz_per_row, mat->bhyb_mat->bell_stride, mat->bhyb_mat->bcsr_nz,
								  mat->bhyb_mat->bell_idx->ph_ival, mat->bhyb_mat->bell_val->ph_sval, mat->bhyb_mat->bcsr_ptr->ph_ival, mat->bhyb_mat->bcsr_idx->ph_ival, mat->bhyb_mat->bcsr_val->ph_sval,
								  x->ph_sval, beta, y->ph_sval);
			}
		}
		/* If the blas is on the GPU,
		 * we care about the variation
		 * (of course, because the variations
		 * are made to optimize SpMV on GPU)*/
		else
		{
			if (blas->precision == CUKR_TYPE_DOUBLE) {
				if (mat->variation == CUKR_MATRIX_FORMAT_VARIATION_BHYB4)
					CukrBlasDSPMV_BHYB4(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
									  mat->bhyb_mat->bell_nz_per_row, mat->bhyb_mat->bell_stride, mat->bhyb_mat->bcsr_nz,
									  mat->bhyb_mat->bell_idx->pd_ival, mat->bhyb_mat->bell_val->pd_dval, mat->bhyb_mat->bcsr_ptr->pd_ival, mat->bhyb_mat->bcsr_idx->pd_ival, mat->bhyb_mat->bcsr_val->pd_dval,
									  x->pd_dval, beta, y->pd_dval);
				else
					CukrBlasDSPMV_BHYB(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
									  mat->bhyb_mat->bell_nz_per_row, mat->bhyb_mat->bell_stride, mat->bhyb_mat->bcsr_nz,
									  mat->bhyb_mat->bell_idx->pd_ival, mat->bhyb_mat->bell_val->pd_dval, mat->bhyb_mat->bcsr_ptr->pd_ival, mat->bhyb_mat->bcsr_idx->pd_ival, mat->bhyb_mat->bcsr_val->pd_dval,
									  x->pd_dval, beta, y->pd_dval);
			}
			if (blas->precision == CUKR_TYPE_QDOUBLE) {
				if (mat->variation == CUKR_MATRIX_FORMAT_VARIATION_BHYB4)
					CukrBlasQSPMV_BHYB4(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
								  mat->bhyb_mat->bell_nz_per_row, mat->bhyb_mat->bell_stride, mat->bhyb_mat->bcsr_nz,
								  mat->bhyb_mat->bell_idx->pd_ival, mat->bhyb_mat->bell_val->pd_sval, mat->bhyb_mat->bell_val->pd_qvalt, mat->bhyb_mat->bcsr_ptr->pd_ival, mat->bhyb_mat->bcsr_idx->pd_ival, mat->bhyb_mat->bcsr_val->pd_sval, mat->bhyb_mat->bcsr_val->pd_qvalt,
								  x->pd_sval, x->pd_qvalt, beta, y->pd_sval, y->pd_qvalt);
				else
					CukrBlasQSPMV_BHYB(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
								  mat->bhyb_mat->bell_nz_per_row, mat->bhyb_mat->bell_stride, mat->bhyb_mat->bcsr_nz,
								  mat->bhyb_mat->bell_idx->pd_ival, mat->bhyb_mat->bell_val->pd_sval, mat->bhyb_mat->bell_val->pd_qvalt, mat->bhyb_mat->bcsr_ptr->pd_ival, mat->bhyb_mat->bcsr_idx->pd_ival, mat->bhyb_mat->bcsr_val->pd_sval, mat->bhyb_mat->bcsr_val->pd_qvalt,
								  x->pd_sval, x->pd_qvalt, beta, y->pd_sval, y->pd_qvalt);

			}
			else if (blas->precision == CUKR_TYPE_SINGLE) {
				if (mat->variation == CUKR_MATRIX_FORMAT_VARIATION_BHYB4)
					CukrBlasSSPMV_BHYB4(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
								  mat->bhyb_mat->bell_nz_per_row, mat->bhyb_mat->bell_stride, mat->bhyb_mat->bcsr_nz,
								  mat->bhyb_mat->bell_idx->pd_ival, mat->bhyb_mat->bell_val->pd_sval, mat->bhyb_mat->bcsr_ptr->pd_ival, mat->bhyb_mat->bcsr_idx->pd_ival, mat->bhyb_mat->bcsr_val->pd_sval,
								  x->pd_sval, beta, y->pd_sval);
				else
					CukrBlasSSPMV_BHYB(blas, mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c, alpha,
								  mat->bhyb_mat->bell_nz_per_row, mat->bhyb_mat->bell_stride, mat->bhyb_mat->bcsr_nz,
								  mat->bhyb_mat->bell_idx->pd_ival, mat->bhyb_mat->bell_val->pd_sval, mat->bhyb_mat->bcsr_ptr->pd_ival, mat->bhyb_mat->bcsr_idx->pd_ival, mat->bhyb_mat->bcsr_val->pd_sval,
								  x->pd_sval, beta, y->pd_sval);
			}
		}
	}


	/* Else, error */
	else {
		CUKR_FATALX("Unknown format for matrix %s: %d",
				    mat->name, mat->format);
	}

	/* Make vector y, which was modified, 
	 * valid only at the location of the BLAS */
	y->status = CukrLocIdAndTypeIdToStatusId(blas->location, blas->precision);
	
	/* Special case: precision qdouble
	 * In this case, set single as well since
	 * it is automatically valid */
	if (blas->precision == CUKR_TYPE_QDOUBLE)
		y->status |= CukrLocIdAndTypeIdToStatusId(blas->location, CUKR_TYPE_SINGLE);
	
	/* Log */
	LOGX(DEBUG, "[N_%d][T_%d] Finished computing product between " \
	            "matrix %s and vector %s ", cukr_db.mpi_rank, 
	            *cukr_activeThread, mat->name, x->name );

	#ifndef CUKR_MP
		/* Update counters */
		CLOCK_END;
		blas->counters_mp[CUKR_BLAS_IDX_TIME] [CUKR_BLAS_OPS_SPMV] += CLOCK_SPAN;
		blas->counters_mp[CUKR_BLAS_IDX_FLOPS][CUKR_BLAS_OPS_SPMV] += 2 * mat->nz;  	
	#endif            	
}

