/**
 * @file blasx/blasx.c
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains the interface to the BLASX routines.
 * 
 * BLASX routines are extended BLAS routines, 
 * that are kind of blas-like in functionality
 * (have the potential of becoming building
 * blocks) but are not implemented in the
 * standard blas libraries.
 * 
 * The BLASX object is attached to the main
 * BLAS object. In this way, each operation
 * can choose the correct BLASX, which has to
 * match the BLAS object.
 * For example, a BLASX running on the CPU shoud
 * be associated to a BLAS running on the CPU and
 * never to one running on the GPU 
 */
  
#define LOGGER_NAME "CUKr.blasx.blasx"  
  
#include <stdio.h>
#include <stdlib.h>
#include "../blas/blas.h"
#include "../util/macros.h"
#include "../util/log.h"
#include "../db/db.h"
#include "blasx.h"
#ifdef CUKR_MPI
	#include <mpi.h>
#endif

/* Get access to the option database */
extern CukrDB cukr_db;

/* Define a max function */
#define MAX(a,b) (a > b ? a : b)

/* 
 * Init a BLASX object 
 * - set all counters to zero
 * - set all bindings to null 
 */
void CukrBlasxCreate(CUKR_BLASX* blasx)  
{
	int i;
	
	/* Alloc memory for the BLAS */
	if ( ((*blasx) = (CUKR_BLASX)malloc(sizeof(struct _CUKR_BLASX))) == NULL)
		CUKR_FATAL("Error allocating memory for BLASX object");
			
	/* Set bindings to NULL */
	(*blasx)->bind.dpwpr = NULL;	
	(*blasx)->bind.qpwpr = NULL;
	(*blasx)->bind.spwpr = NULL;

}	

// ------------------------------------------------------------------- //

/**
 * \brief Reset all the counters to zero
 * \param blasx The blasx structure 
 */
void CukrBlasxResetCounters(CUKR_BLASX blasx)
{
	int i, j, k, l;
	
	/* Empty thread-level counters */
	for (i = 0; i < CUKR_MAX_PE; i++)
		for (j = 0; j < CUKR_BLAS_IDX_MAXCOUNT; j++)
			for (k = 0; k < CUKR_BLASX_OPS_MAXCOUNT; k++)
				for (l = 0; l < CUKR_BLAS_PREC_MAXCOUNT; l++)
					blasx->counters[i][j][k][l] = 0;
					
	/* Empty node-level and cluster-level counters */
	for (j = 0; j < CUKR_BLAS_IDX_MAXCOUNT; j++)
		for (k = 0; k < CUKR_BLASX_OPS_MAXCOUNT; k++) {
			blasx->counters_mp[j][k] = 0;	
			blasx->counters_mpi[j][k] = 0;		
		}	
		
	blasx->comm_inner = blasx->comm_outer = 0;	
}		

// ------------------------------------------------------------------- //

/**
 * \brief Merges the counters for all threads
 *        into one (the first one) 
 * \param blasx The blasx structure 
 */
void CukrBlasxMergeNodeCounters(CUKR_BLASX blasx)
{
	#ifdef CUKR_MPI
	
	int i, p; 
	
	/* 
	 * Ops and load/stores should be added 
	 */
	 
	 /* Add flops */			 
	 MPI_Allreduce(&(blasx->counters_mp [CUKR_BLAS_IDX_FLOPS]), 
	               &(blasx->counters_mpi[CUKR_BLAS_IDX_FLOPS]), 
                   CUKR_BLASX_OPS_MAXCOUNT, MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD);
	 
	 /* Add int. ops */
	 MPI_Allreduce(&(blasx->counters_mp [CUKR_BLAS_IDX_INTOPS]), 
	               &(blasx->counters_mpi[CUKR_BLAS_IDX_INTOPS]), 
                   CUKR_BLASX_OPS_MAXCOUNT, MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD);
	 
	 /* Add fl. loads */
	 MPI_Allreduce(&(blasx->counters_mp [CUKR_BLAS_IDX_FLLOADS]), 
	               &(blasx->counters_mpi[CUKR_BLAS_IDX_FLLOADS]), 
                   CUKR_BLASX_OPS_MAXCOUNT, MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD);		 
	 
	 /* Add fl. stores */
	 MPI_Allreduce(&(blasx->counters_mp [CUKR_BLAS_IDX_FLSTORES]), 
	               &(blasx->counters_mpi[CUKR_BLAS_IDX_FLSTORES]), 
                   CUKR_BLASX_OPS_MAXCOUNT, MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD);
	 
	 /* Add int. loads */
	 MPI_Allreduce(&(blasx->counters_mp [CUKR_BLAS_IDX_INTLOADS]), 
	               &(blasx->counters_mpi[CUKR_BLAS_IDX_INTLOADS]), 
                   CUKR_BLASX_OPS_MAXCOUNT, MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD);
	 
	 /* Add int. stores */
	 MPI_Allreduce(&(blasx->counters_mp [CUKR_BLAS_IDX_INTSTORES]), 
	               &(blasx->counters_mpi[CUKR_BLAS_IDX_INTSTORES]), 
                   CUKR_BLASX_OPS_MAXCOUNT, MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD);
	
	/* 
	 * Timings should be max-ed
	 */
	 
	 MPI_Allreduce(&(blasx->counters_mp [CUKR_BLAS_IDX_TIME]), 
	               &(blasx->counters_mpi[CUKR_BLAS_IDX_TIME]), 
                   CUKR_BLASX_OPS_MAXCOUNT, MPI_FLOAT, MPI_MAX, MPI_COMM_WORLD);	
                   
     #endif              	
}		   	


// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DPWPR (pointwise
 *        product in double precision) 
 * \param blas The BLASX structure
 * \param n The vector size
 * \param x,y The intput vectors
 * \param z The result vector 
 */
void CukrBlasxDPWPR(CUKR_BLASX blasx, CUKR_INT n, CUKR_DOUBLE* x, CUKR_DOUBLE* y, CUKR_DOUBLE* z) 
{	
	/* Check for valid binding */
	if (blasx->bind.dpwpr == NULL)
		CUKR_FATALX("BLASX %s does not implement the function DPWPR\n", blasx->name);
	 
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif

	/* Work */
	(blasx->bind.dpwpr)(n, x, y, z);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = pthread_getspecific(cukr_db.thread_key);
		blasx->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLASX_OPS_PWPR][CUKR_BLAS_PREC_DOUBLE] += CLOCK_SPAN;
		blasx->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLASX_OPS_PWPR][CUKR_BLAS_PREC_DOUBLE] += n;
		blasx->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLASX_OPS_PWPR][CUKR_BLAS_PREC_DOUBLE] += 2*n;
		blasx->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLASX_OPS_PWPR][CUKR_BLAS_PREC_DOUBLE] += n;
	#endif	                                         
}

/**
 * \brief Wrapper for QPWPR (pointwise
 *        product in quasi-double precision) 
 * \param blas The BLASX structure
 * \param n The vector size
 * \param xh, xt, yh, yt The intput vectors 
 *        (heads & tails)
 * \param zh, zt The result vector (head & tail)
 */
void CukrBlasxQPWPR(CUKR_BLASX blasx, CUKR_INT n, CUKR_SINGLE* xh, CUKR_SINGLE* xt,
				 								  CUKR_SINGLE* yh, CUKR_SINGLE* yt,
												  CUKR_SINGLE* zh, CUKR_SINGLE* zt) 
{ 
	/* Check for valid binding */
	if (blasx->bind.qpwpr == NULL)
		CUKR_FATALX("BLASX %s does not implement the function QPWPR\n", blasx->name)
	 
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blasx->bind.qpwpr)(n, xh, xt, yh, yt, zh, zt);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = pthread_getspecific(cukr_db.thread_key);
		blasx->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLASX_OPS_PWPR][CUKR_BLAS_PREC_QDOUBLE] += CLOCK_SPAN;
		blasx->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLASX_OPS_PWPR][CUKR_BLAS_PREC_QDOUBLE] += n;
		blasx->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLASX_OPS_PWPR][CUKR_BLAS_PREC_QDOUBLE] += 2*n;
		blasx->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLASX_OPS_PWPR][CUKR_BLAS_PREC_QDOUBLE] += n;
	#endif		
}

/**
 * \brief Wrapper for SPWPR (pointwise
 *        product in single precision) 
 * \param blas The BLASX structure
 * \param n The vector size
 * \param x,y The intput vectors
 * \param z The result vector 
 */
void CukrBlasxSPWPR(CUKR_BLASX blasx, CUKR_INT n, CUKR_SINGLE* x, CUKR_SINGLE* y, CUKR_SINGLE* z) 
{  
	/* Check for valid binding */
	if (blasx->bind.spwpr == NULL)
		CUKR_FATALX("BLASX %s does not implement the function SPWPR\n", blasx->name);
	
	 
	/* 
	 * Do the actual work 
	 */
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START;
	#endif	
	
	/* Work */
	(blasx->bind.spwpr)(n, x, y, z);
	
	/* Counter block */
	#ifdef CUKR_THREAD_COUNTERS
		CLOCK_END;	
		int* cukr_activeThread = pthread_getspecific(cukr_db.thread_key);
		blasx->counters[*cukr_activeThread][CUKR_BLAS_IDX_TIME]    [CUKR_BLASX_OPS_PWPR][CUKR_BLAS_PREC_SINGLE] += CLOCK_SPAN;
		blasx->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLOPS]   [CUKR_BLASX_OPS_PWPR][CUKR_BLAS_PREC_SINGLE] += n;
		blasx->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLLOADS] [CUKR_BLASX_OPS_PWPR][CUKR_BLAS_PREC_SINGLE] += 2*n;
		blasx->counters[*cukr_activeThread][CUKR_BLAS_IDX_FLSTORES][CUKR_BLASX_OPS_PWPR][CUKR_BLAS_PREC_SINGLE] += n;
	#endif                                          
}
