/**
 * @file blasx/blasx.h
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains the interface to the BLASX routines.
 * 
 * BLASX routines are extended BLAS routines, 
 * that are kind of blas-like in functionality
 * (have the potential of becoming building
 * blocks) but are not implemented in the
 * standard blas libraries.
 * 
 * The BLASX object is attached to the main
 * BLAS object. In this way, each operation
 * can choose the correct BLASX, which has to
 * match the BLAS object.
 * For example, a BLASX running on the CPU shoud
 * be associated to a BLAS running on the CPU and
 * never to one running on the GPU 
 */

#ifndef BLASX_H_
#define BLASX_H_

/* =============== FORWARD DECLARATIONS ============== */

typedef struct _CUKR_BLASX *CUKR_BLASX;

/* =================================================== */

#include "../cukr.h"
#include "../mat_vec/vector.h"
#include "../blas/blas.h"

/* Define the max no. of operations 
 * in the BLAS */
#define CUKR_BLASX_OPS_MAXCOUNT 5

/* Define all operations in the BLAS */
#define CUKR_BLASX_OPS_PWPR 0

/*
 * The actual BLASX functions found in the BLAS
 * implementations are linked to the BLAS object
 * using function pointers. Here we declare these
 * pointers for each implemented operation.
 */

/* PWPROD (Zi := Xi * Yi) -> _PWPR (n,x,y,z) */
typedef void (*CUKR_BLASX_DPWPR)(CUKR_INT, CUKR_DOUBLE*, CUKR_DOUBLE*, CUKR_DOUBLE*); 
typedef void (*CUKR_BLASX_QPWPR)(CUKR_INT, CUKR_SINGLE*, CUKR_SINGLE*,
										   CUKR_SINGLE*, CUKR_SINGLE*,
										   CUKR_SINGLE*, CUKR_SINGLE*); 
typedef void (*CUKR_BLASX_SPWPR)(CUKR_INT, CUKR_SINGLE*, CUKR_SINGLE*, CUKR_SINGLE*);


                           
/** 
 * Here we define the BLASX structure. 
 * This structure will include:
 * - a <bind> group where the actual function 
 *   pointers will be stored
 * - a vector where the counters will be stored
 * - name to use for logging 
 */                        

typedef struct _CUKR_BLASX_OPS{
	CUKR_BLASX_DPWPR dpwpr;
	CUKR_BLASX_QPWPR qpwpr;
	CUKR_BLASX_SPWPR spwpr;
}CUKR_BLASX_OPS;

struct _CUKR_BLASX{
	
	/** Bindings to the function pointers */
	CUKR_BLASX_OPS bind;
		
	/** BLASX name */
	char name[255];
	
	/** Counter matrix, by thread*/
	float counters[CUKR_MAX_PE][CUKR_BLAS_IDX_MAXCOUNT][CUKR_BLASX_OPS_MAXCOUNT][CUKR_BLAS_PREC_MAXCOUNT];
	
	/** Counter matrix, by node */
	float counters_mp[CUKR_BLAS_IDX_MAXCOUNT][CUKR_BLASX_OPS_MAXCOUNT];
	
	/** Counter matrix, by cluster */
	float counters_mpi[CUKR_BLAS_IDX_MAXCOUNT][CUKR_BLASX_OPS_MAXCOUNT];
	
	/** Communication time for internal and external 
	 * ops (inside one node and outside). Written
	 * only from the _MP and _MPI operations */
	float comm_inner, comm_outer;
};

/* 
 * Additional functions for initializing the BLASX
 * plus wrappers for the founction pointers to
 * automatically take care of updating the counters
 * when the associated functions are being used.
 */
 
/* Init a BLAS object */
void CukrBlasxCreate(CUKR_BLASX* blasx);

// ------------------------------------------------------------------- //

/**
 * \brief Reset all the counters to zero
 * \param blasx The blasx structure 
 */
void CukrBlasxResetCounters(CUKR_BLASX blasx);

// ------------------------------------------------------------------- //

/**
 * \brief Merges the counters for all nodes 
 * \param blasx The blasx structure 
 */
void CukrBlasxMergeNodeCounters(CUKR_BLASX blasx);

// ------------------------------------------------------------------- //

/**
 * \brief Wrapper for DPWPR (pointwise
 *        product in double precision) 
 * \param blas The BLASX structure
 * \param n The vector size
 * \param x,y The intput vectors
 * \param z The result vector 
 */
void CukrBlasxDPWPR(CUKR_BLASX, CUKR_INT, CUKR_DOUBLE*, CUKR_DOUBLE*, CUKR_DOUBLE*); 


/**
 * \brief Wrapper for QPWPR (pointwise
 *        product in quasi-double precision) 
 * \param blas The BLASX structure
 * \param n The vector size
 * \param xh, xt, yh, yt The intput vectors 
 *        (heads & tails)
 * \param zh, zt The result vector (head & tail)
 */
void CukrBlasxQPWPR(CUKR_BLASX, CUKR_INT, CUKR_SINGLE*, CUKR_SINGLE*,
					  			   		 CUKR_SINGLE*, CUKR_SINGLE*,
										 CUKR_SINGLE*, CUKR_SINGLE*); 

/**
 * \brief Wrapper for SPWPR (pointwise
 *        product in single precision) 
 * \param blas The BLASX structure
 * \param n The vector size
 * \param x,y The intput vectors
 * \param z The result vector 
 */
void CukrBlasxSPWPR(CUKR_BLASX, CUKR_INT, CUKR_SINGLE*, CUKR_SINGLE*, CUKR_SINGLE*); 


#endif /*BLASX_H_*/
