/**
 * @author Serban Georgescu
 * @since 2008
 *  
 * This defines the "generic blasx", with
 * all functions written here. It does not
 * use any libraries. 
 */

#define LOGGER_NAME "CUKr.blasx.generic_blasx"
 
#include <stdio.h>
#include <stdlib.h> 
#include <math.h>
#include "generic_blasx.h"

// ====================== FORWARD DECLARATIONS ======================= //

extern void dsadd(volatile float *c0, volatile float *c1, const volatile float a0, const volatile float a1, 
                                				   		  const volatile float b0, const volatile float b1);
extern void dssub(volatile float *c0, volatile float *c1, const volatile float a0, const volatile float a1, 
                                				   		  const volatile float b0, const volatile float b1);                     
extern void dsmul(volatile float *c0, volatile float *c1, const volatile float a0, const volatile float a1, 
                                                   		  const volatile float b0, const volatile float b1);
                        

// =================================================================== //

  
/**
 * \brief Wrapper for DPWPR (pointwise
 *        product in double precision) 
 * \param blas The BLASX structure
 * \param n The vector size
 * \param x,y The intput vectors
 * \param z The result vector 
 */
void generic_dpwpr(CUKR_INT n, CUKR_DOUBLE* x, CUKR_DOUBLE* y, CUKR_DOUBLE* z)
{
	int n_4 = 4 * (int)(n / 4);
	int i;
	
	/* 4-unrolling */
	for (i = 0; i < n_4 ; i += 4)
	{ 
		z[i]     = x[i]     * y[i]; 
		z[i + 1] = x[i + 1] * y[i + 1];
		z[i + 2] = x[i + 2] * y[i + 2];
		z[i + 3] = x[i + 3] * y[i + 3];
	}
	 
	/* The rest */
	for (i = n_4; i < n; i++)
	  z[i] = x[i] * y[i]; 
} 

/**
 * \brief Wrapper for QPWPR (pointwise
 *        product in quasi-double precision) 
 * \param blas The BLASX structure
 * \param n The vector size
 * \param xh, xt, yh, yt The intput vectors 
 *        (heads & tails)
 * \param zh, zt The result vector (head & tail)
 */
void generic_qpwpr(CUKR_INT n, CUKR_SINGLE* xh, CUKR_SINGLE* xt,
				 			   CUKR_SINGLE* yh, CUKR_SINGLE* yt,
							   CUKR_SINGLE* zh, CUKR_SINGLE* zt)
{
	int i;
	CUKR_SINGLE c0, c1;
	
	/* 
	 * Do the update 
	 */
	 
	for (i = 0; i < n; i++)
		dsmul(&zh[i], &zt[i], xh[i], xt[i], yh[i], yt[i]); 
}
											  	

/**
 * \brief Wrapper for SPWPR (pointwise
 *        product in single precision) 
 * \param blas The BLASX structure 
 * \param n The vector size
 * \param x,y The intput vectors
 * \param z The result vector 
 */
void generic_spwpr(CUKR_INT n, CUKR_SINGLE* x, CUKR_SINGLE* y, CUKR_SINGLE* z)
{
	int n_4 = 4 * (int)(n / 4);
	int i;
	
	/* 4-unrolling */
	for (i = 0; i < n_4 ; i += 4)
	{ 
		z[i]     = x[i]     * y[i]; 
		z[i + 1] = x[i + 1] * y[i + 1];
		z[i + 2] = x[i + 2] * y[i + 2];
		z[i + 3] = x[i + 3] * y[i + 3];
	}
	 
	/* The rest */
	for (i = n_4; i < n; i++)
	  z[i] = x[i] * y[i]; 
}

               
// ------------------------------------------------------------------- //

/*
 * \brief Create a generic blasx object by
 *        binding the above functions to a
 * 		  CUKR_BLASX structure 
 * \param blas pointer to a blas structure
 */
void blasx_create_generic(CUKR_BLASX blasx)
{
	/* Set the BLAS name */
	sprintf(blasx->name, "GENERIC");
	
	/* Set the bindings */
	blasx->bind.dpwpr = generic_dpwpr;
	blasx->bind.qpwpr = generic_qpwpr;
	blasx->bind.spwpr = generic_spwpr;
}                                           	
