/**
 * @file kernel_dpwpr.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines a double precision
 * PWPR (point-wise product) kernel.
 * 
 * PWPR: Zi := Xi * Yi, i = 1,n 
 * 
 * For 256 blocks of 128 threads, does around 67GB/s, which 
 * translates to around 5.8GFlops. 
 * 
 */
 
#include "../kernels_config.h"

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x 
	#define bx          blockIdx.x
#endif
#define X_OFFSET_DPWPR    0
#define Y_OFFSET_DPWPR    CUKR_DPWPR_THREADS

extern __shared__ double shared[];

__global__ void kernel_dpwpr(const double *d_x,
						     const double *d_y,
						    	   double *d_z,
                             const int     n)
{	
				
	/* Alloc shared memory */
	double  *buffer_x = &shared[X_OFFSET_DPWPR]; 
	double  *buffer_y = &shared[Y_OFFSET_DPWPR]; 
	
	/* Starting point for this block */
	int ctaStart = bx * CUKR_DPWPR_THREADS; 
	/* Total no. of threads in the kernel */
	int totalThreads = CUKR_DPWPR_CTAS * CUKR_DPWPR_THREADS;
	
	/* Read the data (full lines) */
	for (int i = ctaStart + tx; i < n; i += totalThreads)
	{  
		/* Prefetch x and y */
		buffer_x[tx] = d_x[i];
		buffer_y[tx] = d_y[i];
		
		/* Multiply and write */
		d_z[i] = buffer_x[tx] * buffer_y[tx];  	 
	}  	 
}
