#ifndef CUKR_DS_OPS_H_
#define CUKR_DS_OPS_H_

/**
 * @file kernel_ds_ops.h
 * @author NVIDIA
 * @since 2008
 *  
 * Defines a double-single (qdouble)
 * operations:
 * 	- double-single add (addition)
 *  - double-single sub (subtraction)
 * 	- double-single mul (multiplication)
 * 
 * The functions declared heren are device
 * functions, so they can only be used
 * inside CUDA kernels
 * 
 */


/** 
 * \brief Double-single (qdouble) addition
 * \param c0,c1 Head and tail for the result
 * \param a0,a1 Head and tail for the first
 *              operand
 * \param b0,b1 Head and tail for the second
 *              operand
 */
__device__ inline void dsadd(float &c0, float &c1, const float a0, const float a1, const float b0, const float b1)
{
	
	#if defined(__DEVICE_EMULATION__)
   		volatile float t1, t2, e;
	#else
   		float t1, t2, e;
	#endif
		 	
    /* Compute dsa + dsb using Knuth's trick. */
    t1 = a0 + b0;
    e = t1 - a0;
    t2 = ((b0 - e) + (a0 - (t1 - e))) + a1 + b1;
	
    /* The result is t1 + t2, after normalization. */
    c0 = e = t1 + t2;
    c1 = t2 - (e - t1);	
}

// ------------------------------------------------------------------- //

/** 
 * \brief Double-single (qdouble) subtraction
 * \param c0,c1 Head and tail for the result
 * \param a0,a1 Head and tail for the first
 *              operand
 * \param b0,b1 Head and tail for the second
 *              operand
 */
__device__ void dssub(float  &c0, float  &c1, const float a0, const float  a1, const float b0, const float b1);

// ------------------------------------------------------------------- //

/** 
 * \brief Double-single (qdouble) multiplication
 * \param c0,c1 Head and tail for the result
 * \param a0,a1 Head and tail for the first
 *              operand
 * \param b0,b1 Head and tail for the second
 *              operand
 */
__device__ inline void dsmul(float &c0, float &c1, const float a0, const float a1, const float b0, const float b1)
{
  #if defined(__DEVICE_EMULATION__)
   		volatile float cona, conb, sa1, sa2, sb1, sb2, c11, c21, c2, t1, e, t2;
   #else
   		float cona, conb, sa1, sa2, sb1, sb2, c11, c21, c2, t1, e, t2;
   #endif

   cona = a0 * 8193.0f;
   conb = b0 * 8193.0f;
   sa1 = cona - (cona - a0);
   sb1 = conb - (conb - b0);
   sa2 = a0 - sa1;
   sb2 = b0 - sb1;

   /* Multilply a0 * b0 using Dekker's method. */
   c11 = a0 * b0;
   c21 = (((sa1 * sb1 - c11) + sa1 * sb2) + sa2 * sb1) + sa2 * sb2;

   /* Compute a0 * b1 + a1 * b0 (only high-order word is needed). */
   c2 = a0 * b1 + a1 * b0;

   /* Compute (c11, c21) + c2 using Knuth's trick, also adding low-order product. */
   t1 = c11 + c2;
    e = t1 - c11;
   t2 = ((c2 - e) + (c11 - (t1 - e))) + c21 + a1 * b1;

   /* The result is t1 + t2, after normalization. */
   c0 = e = t1 + t2;
   c1 = t2 - (e - t1);

}

// ------------------------------------------------------------------- //

#endif /*CUKR_DS_OPS_H_*/
