/**
 * @file kernel_qpwpr.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines a double-single (emulated double)
 * PWPR (point-wise product) kernel.
 * 
 * PWPR: Zi := Xi * Yi, i = 1,n 
 * 
 * For 256 blocks of 128 threads, does around 67GB/s, which 
 * translates to around 2.8GFlops. This is exactly half the single,
 * so there is no more to be improved.
 * 
 * 
 */
 
#include "../kernels_config.h"
#include "ds_ops.h"

#ifndef _BX_TX
	#define _BX_TX
	#define tx          threadIdx.x 
	#define bx          blockIdx.x
#endif
#define XH_OFFSET_QPWPR    0
#define XT_OFFSET_QPWPR        CUKR_QPWPR_THREADS
#define YH_OFFSET_QPWPR    2 * CUKR_QPWPR_THREADS
#define YT_OFFSET_QPWPR    3 * CUKR_QPWPR_THREADS

extern __shared__ float shared[];

__global__ void kernel_qpwpr(const float *d_xh, const float *d_xt, 
						     const float *d_yh, const float *d_yt, 
						    	   float *d_zh,       float *d_zt, 
                             const int    n)
{	
				
	/* Alloc shared memory */
	float  *buffer_xh = &shared[XH_OFFSET_QPWPR]; 
	float  *buffer_xt = &shared[XT_OFFSET_QPWPR]; 
	float  *buffer_yh = &shared[YH_OFFSET_QPWPR]; 
	float  *buffer_yt = &shared[YT_OFFSET_QPWPR]; 
	
	/* Starting point for this block */
	int ctaStart = bx * CUKR_QPWPR_THREADS; 
	/* Total no. of threads in the kernel */
	int totalThreads = CUKR_QPWPR_CTAS * CUKR_QPWPR_THREADS;
	
	/* Read the data (full lines) */
	for (int i = ctaStart + tx; i < n; i += totalThreads)
	{  
		/* Prefetch x and y */
		buffer_xh[tx] = d_xh[i];
		buffer_xt[tx] = d_xt[i];
		buffer_yh[tx] = d_yh[i];
		buffer_yt[tx] = d_yt[i]; 
		
		/* Multiply and write */
		dsmul(d_zh[i], d_zt[i], buffer_xh[tx], buffer_xt[tx], buffer_yh[tx], buffer_yt[tx]);  	 
	}  	 
}
