/**
 * @file blas/gpu/gpu_blas.c
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines wrappers for BLAS functions 
 * on the GPU. 
 * 
 * The functions take as parameter device
 * pointers, so they only act on data that
 * has been PREVIOUSLY COPIED to the device.
 * 
 */

#include <stdio.h>
#include <stdlib.h> 
#include "gpu_blasx.h"
#include "cuda/kernels.h"
#include "../../../util/log.h"
#ifdef CUKR_USE_CUDA
	#include <cuda.h> 
	#include <cublas.h>
#endif 	


// ------------------------------------------------------------------- //

#ifdef CUKR_USE_GPU

/**
 * \brief Wrapper for DPWPR (pointwise
 *        product in double precision) 
 * \param blas The BLASX structure
 * \param n The vector size
 * \param x, y The intput vectors 
 * \param z The result vector 
 */
void gpu_dpwpr(CUKR_INT n, CUKR_DOUBLE* x, CUKR_DOUBLE* y, CUKR_DOUBLE* z)
{
	#ifdef CUKR_USE_CUDA
		cuda_kernel_dpwpr(n, x, y, z);
	#endif
	#ifdef CUKR_USE_OPENCL
//		opencl_kernel_dpwpr(n, x, y, z);
	#endif
}

/**
 * \brief Wrapper for QPWPR (pointwise
 *        product in quasi-double precision) 
 * \param blas The BLASX structure
 * \param n The vector size
 * \param xh, xt, yh, yt The intput vectors 
 *        (heads & tails)
 * \param zh, zt The result vector (head & tail)
 */
void gpu_qpwpr(CUKR_INT n, CUKR_SINGLE* xh, CUKR_SINGLE* xt,
				 		   CUKR_SINGLE* yh, CUKR_SINGLE* yt,
						   CUKR_SINGLE* zh, CUKR_SINGLE* zt)
{
	#ifdef CUKR_USE_CUDA
		cuda_kernel_qpwpr(n, xh, xt, yh, yt, zh, zt);
	#endif
	#ifdef CUKR_USE_OPENCL
//		opencl_kernel_qpwpr(n, xh, xt, yh, yt, zh, zt);
	#endif
	
}
											  	

/**
 * \brief Wrapper for SPWPR (pointwise
 *        product in single precision) 
 * \param blas The BLASX structure 
 * \param n The vector size
 * \param x,y The intput vectors
 * \param z The result vector 
 */
void gpu_spwpr(CUKR_INT n, CUKR_SINGLE* x, CUKR_SINGLE* y, CUKR_SINGLE* z)
{
	#ifdef CUKR_USE_CUDA
		cuda_kernel_spwpr(n, x, y, z);
	#endif
	#ifdef CUKR_USE_OPENCL
//		opencl_kernel_spwpr(n, x, y, z);
	#endif
}

#endif /* CUKR_USE_CUDA */



// ------------------------------------------------------------------- //



/*
 * \brief Create a generic blasx object by
 *        binding the above functions to a
 * 		  CUKR_BLASX structure 
 * \param blas pointer to a blas structure
 */
void blasx_create_gpu(CUKR_BLASX blasx)
#ifndef CUKR_USE_GPU
{
	CUKR_FATAL("GPU support was not compiled in the library");
}
#else 
{
	/* Set the BLAS name */
	sprintf(blasx->name, "GPUBLASX");
	
	/* Set the bindings */
	blasx->bind.dpwpr = gpu_dpwpr;
	blasx->bind.qpwpr = gpu_qpwpr;
	blasx->bind.spwpr = gpu_spwpr;
} 
#endif /* CUKR_USE_CUDA */   
