/**
 * @file blasx/blasx1_mp.h
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains the extension of the
 * BLASX1 routines to multigpu vectors
 * 	 
 */ 

#define LOGGER_NAME "CUKr.blasx.blasx1_mp"  

#include <stdio.h>
#include <stdlib.h>
 
#include "blasx1_mp.h" 
#include "../sp/blasx1_sp.h"
#include "../../util/log.h"
#include "../../util/macros.h"
#include "../../threading/threading.h"
#include "../../db/db.h"

/* Get access to the database */
extern CukrDB cukr_db;
// ------------------------------------------------------------------- //

/**
/**
 * \brief Pointwise-product between
 *        two MP vectors
 * \param blas The BLAS structure
 * \param x,y The input MP vectors
 * \param z The result MP vector 
 */
void CukrPWPR_MP(CUKR_BLAS blas, CUKR_VECTOR_MP x, CUKR_VECTOR_MP y, CUKR_VECTOR_MP z)
{	
	int i;
	char str_shtype_x[255], str_shtype_y[255];
	
	/* Start timing */
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START; 
	
	/* Log */
	LOGX(DEBUG, "[N_%d] Running PWPR between MP vectors "\
				"%s and %s, with result in %s", cukr_db.mpi_rank, 
				x->name, y->name, z->name);
		
	/* Error if the vectors do not have
	 * the same number of elements */
	if (x->n != y->n || x->n != z->n)
		CUKR_FATALX("[N_%d] Cannot compute PWPR for vectors which" \
		            "do not have the same number of elements",
		            cukr_db.mpi_rank); 
	
	/* The operation happens in the same way
	 * regardless of the sharing type. For
	 * both cases, only the diagonal elements
	 * are used */
	
	/* Malloc at destination. Have to do it now, in 
	 * order to make use of the speciall malloc for
	 * shared vectors. */
	
	CukrVecmpMalloc(z, blas->location, blas->precision); 
	
	/* Schedule PWPR between all pairs
	 * of parts */
	for (i = 0; i < x->nparts; i++)
		CukrThreadsRun(i, (void*)CukrPWPR_SP, CUKR_FUNCID_PWPR, blas, x->cells[i], y->cells[i], z->cells[i]);	 
	
	/* Wait for all threads to finish */
	CukrThreadsWaitCompletion(x->nparts);
	
	/* If shared, sync. the status of the rows
	 * to the status of the cells, for the dest
	 * vector */
	if (z->shtype == CUKR_SHTYPE_SHARED) 
		for (i = 0; i < z->nparts; i++)
			 z->rows[i]->status = z->cells[i]->status;  	
	
	/* Log */
	LOGX(DEBUG, "[N_%d] PWPR between MP vectors %s and %s "\
				"(res: %s) was computed", cukr_db.mpi_rank,
				x->name, y->name, z->name);	
				
	/* Update counters */
	CLOCK_END;
	blas->blasx->counters_mp[CUKR_BLAS_IDX_TIME] [CUKR_BLASX_OPS_PWPR] += CLOCK_SPAN;
	blas->blasx->counters_mp[CUKR_BLAS_IDX_FLOPS][CUKR_BLASX_OPS_PWPR] += x->n;
							
}
