/**
 * @file blasx/blasx1.c
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains the interface to BLASX1 routines.
 * 
 * The difference between these routines 
 * and the ones in blasx.h is that these
 * are written as function of abstract
 * vectors and matrices, therefor work
 * independent of format, type and
 * location.
 * 	
 */

#define LOGGER_NAME "CUKr.blasx->blasx1"  

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
 
#include "blasx1_sp.h"
#include "../../util/log.h"
#include "../../util/macros.h"
#include "../../util/utils.h"
#include "../../db/db.h"

/* Get access to the database */
extern CukrDB cukr_db;

// ------------------------------------------------------------------- //

/**
/**
 * \brief Pointwise-product between
 *        two vectors
 * \param blas The BLAS structure
 * \param x,y The intput vectors
 * \param z The result vector 
 */
void CukrPWPR_SP(CUKR_BLAS blas, CUKR_VECTOR_SP x, CUKR_VECTOR_SP y, CUKR_VECTOR_SP z)
{
	#ifndef CUKR_MP
		/* Start timing */
		struct timeval t_start, t_end;
		struct timezone t_zone; 
		CLOCK_START; 
	#endif	

	int* cukr_activeThread = (int*)(pthread_getspecific(cukr_db.thread_key));

	/* Log */
	LOGX(DEBUG, "[N_%d][T_%d] Running pointwise product between " \
	            "vectors %s and %s (writing result in vector %s)", 
	            cukr_db.mpi_rank, cukr_activeThread, 
	            x->name, y->name, z->name );
	
	/* Error if the vectors do not have
	 * the same number of elements */
	if (x->n != y->n || x->n != z->n)
		CUKR_FATALX("[N_%d][T_%d] Cannot compute pointwise product "\
		            "for vectors %s, %s and %s since they "\
		            "do not have the same number of "\
		            "elements", cukr_db.mpi_rank, *cukr_activeThread,
		            x->name, y->name, z->name); 
	
	/* If the vectors do not have the same
	 * location/precision as the BLAS, the 
	 * vectors should be moved to that location
	 * and/or their precision changed */
	 
	CukrVecspTransfer(x, blas->location, blas->precision);
	CukrVecspTransfer(y, blas->location, blas->precision); 
	
	/* For pwpr, the result vector does not 
	 * actually has to be moved, only memory
	 * has to be allocated at the destination */	
	 
	CukrVecspMalloc(z, blas->location, blas->precision); 
	 
	/* Now run the BLASX operation at the BLAS
	 * location and in the BLAS precision. */
	/* If the blas is on the CPU */
	if (blas->location == CUKR_LOCATION_CPU)
	{
		if (blas->precision == CUKR_TYPE_DOUBLE)
			CukrBlasxDPWPR(blas->blasx, x->n, x->ph_dval, y->ph_dval, z->ph_dval);
		else if (blas->precision == CUKR_TYPE_QDOUBLE)
			CukrBlasxQPWPR(blas->blasx, x->n, x->ph_sval, x->ph_qvalt,
											  y->ph_sval, y->ph_qvalt,
											  z->ph_sval, z->ph_qvalt);			
		else
			CukrBlasxSPWPR(blas->blasx, x->n, x->ph_sval, y->ph_sval, z->ph_sval);
	}	
	/* If the blas is on the GPU */ 
	else 
	{		
		if (blas->precision == CUKR_TYPE_DOUBLE)
			CukrBlasxDPWPR(blas->blasx, x->n, x->pd_dval, y->pd_dval, z->pd_dval);
		else if (blas->precision == CUKR_TYPE_QDOUBLE)
		{
			/* Do the padded version, if needed */ 
			#ifdef CUKR_QPWPR_PADDING
				CukrVecPadToMultGPU(x, CUKR_TYPE_QDOUBLE, CUKR_QPWPR_PADDING); 
				CukrVecPadToMultGPU(y, CUKR_TYPE_QDOUBLE, CUKR_QPWPR_PADDING); 
				CukrVecPadToMultGPU(z, CUKR_TYPE_QDOUBLE, CUKR_QPWPR_PADDING); 
				CukrBlasxQPWPR(blas->blasx, x->n_padded, x->pd_sval, x->pd_qvalt,
											  	  		 y->pd_sval, y->pd_qvalt,
											  	  		 z->pd_sval, z->pd_qvalt);	
			#else
				/* If not, do the regular one */ 
				CukrBlasxQPWPR(blas->blasx, x->n, x->pd_sval, x->pd_qvalt,
											  	  y->pd_sval, y->pd_qvalt,
											  	  z->pd_sval, z->pd_qvalt);	
			#endif
			
		}
		else
		{
			/* Do the padded version, if needed */
			#ifdef CUKR_SPWPR_PADDING
				CukrVecPadToMultGPU(x, CUKR_TYPE_SINGLE, CUKR_SPWPR_PADDING); 
				CukrVecPadToMultGPU(y, CUKR_TYPE_SINGLE, CUKR_SPWPR_PADDING); 
				CukrVecPadToMultGPU(z, CUKR_TYPE_SINGLE, CUKR_SPWPR_PADDING);
				CukrBlasxSPWPR(blas->blasx, x->n_padded, x->pd_sval, y->pd_sval, z->pd_sval);
			#else
				/* If not, do the regular one */	
				CukrBlasxSPWPR(blas->blasx, x->n, x->pd_sval, y->pd_sval, z->pd_sval);
			#endif
		}
	}
	
	/* Make vector z, which was modified, 
	 * valid only at the location of the BLAS */
	z->status = CukrLocIdAndTypeIdToStatusId(blas->location, blas->precision); 
	/* Special case: precision qdouble
	 * In this case, set single as well since
	 * it is automatically valid */
	if (blas->precision == CUKR_TYPE_QDOUBLE)
		z->status |= CukrLocIdAndTypeIdToStatusId(blas->location, CUKR_TYPE_SINGLE); 
	
	/* Log */
	LOGX(DEBUG, "[N_%d][T_%d] Pointwise product between vectors " \
			    "%s and %s finished", cukr_db.mpi_rank, *cukr_activeThread,
			    x->name, y->name);
			    
	#ifndef CUKR_MP
		/* Update counters */
		CLOCK_END;
		blas->blasx->counters_mp[CUKR_BLAS_IDX_TIME] [CUKR_BLASX_OPS_PWPR] += CLOCK_SPAN;
		blas->blasx->counters_mp[CUKR_BLAS_IDX_FLOPS][CUKR_BLASX_OPS_PWPR] += 2 * x->n;  
	#endif
}
