/**
 * @file cukr.h
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains global defines, types
 * error cores and such that will 
 * be used throughout all the 
 * application
 */
 
#ifndef CUKR_H_
#define CUKR_H_

#include "cukr_config.h"

#ifdef CUKR_USE_OPENCL
	#include "init/init_opencl.h"
#endif

// ================================== TYPE DEFS ================================== //


/*
 * The following float data types are supported:
 * - double (64bit)
 * - qdouble (quasi-double 2x32bit)
 * - single (32bit)
 * 
 * The following integer data types are supported:
 * - int (32bit)
 */

// #ifdef CUKR_USE_OPENCL
	#define CUKR_SINGLE  float 
	#define CUKR_DOUBLE  double
	#define CUKR_INT     int

	#define CUKR_CL_SINGLE cl_mem
	#define CUKR_CL_DOUBLE cl_mem
	#define CUKR_CL_INT	   cl_mem
/*
	#define CUKR_GPU_SINGLE  cl_mem 
	#define CUKR_GPU_DOUBLE  cl_mem
	#define CUKR_GPU_INT     cl_mem
#else
	#define CUKR_SINGLE  float 
	#define CUKR_DOUBLE  double
	#define CUKR_INT     int
	#define CUKR_GPU_SINGLE  float 
	#define CUKR_GPU_DOUBLE  double
	#define CUKR_GPU_INT     int
#endif
 */

/* 
 * Booleans 
 */
/** True */
#define CUKR_TRUE  1
/** False */
#define CUKR_FALSE 0

/* 
 * Default
 */
#define CUKR_DEFAULT 1024
 

// ================================= OPTIONS ================================== //

#define CUKR_OPTS_STRICT_SYNTAX   2048
#define CUKR_OPTS_COMPUTES_ON_GPU 2049
#define CUKR_OPTS_RUNTIME_REPORT  4096
#define CUKR_OPTS_BLAS_REPORT     8192
#define CUKR_OPTS_FROMGPU        16384
#define CUKR_OPTS_TOGPU          32768


// ================================== STATES ================================== //


/* Status indicators
 * The status shows in one value both
 * the type and the location of the 
 * variable. The actual status is 
 * usually a combination of states, since
 * the stored data can be good (actualized 
 * data) in more places than one.
 * 
 * For example, consider a vector created in
 * single, in the CPU. Its initial status would
 * be something like CPU_SINGLE. If we want to 
 * compute it's norm on the GPU, the vector will
 * be copied on the GPU. Since the norm does not
 * modify the vector, the state of the vector 
 * will thus become a superposition of CPU_SINGLE 
 * and GPU_SINGLE. 
 * 
 * This behaviour can be emulated using powers of
 * two. Superposition are written as STATE_1 | STATEE_2
 * | ... | STATE_n. A vector is in a particular state i
 * iff state & STATE_i != 0. State i can be added with
 * state |= STATE_i and deleted with state -= STATE_i.
 * A single state can be assigned with state = STATE_i
 */

/** Indicates the vector currently
 * not in any state */
#define CUKR_STATE_NONE 0 

/**
 * Indicates a single precision 
 * (float) variable on CPU
 */
#define CUKR_STATE_CPU_SINGLE  1

/**
 * Indicates a single precision 
 * (float) variable on GPU
 */
#define CUKR_STATE_GPU_SINGLE  2

/**
 * Indicates a quasi-double 
 * variable on CPU
 */
#define CUKR_STATE_CPU_QDOUBLE 4
/**
 * Indicates a quasi-double 
 * variable on GPU
 */
#define CUKR_STATE_GPU_QDOUBLE 8

/**
 * Indicates a double precision 
 * (double) variable on CPU
 */
#define CUKR_STATE_CPU_DOUBLE  16

/** Indicates a double precision 
 * (double) variable on GPU
 */
#define CUKR_STATE_GPU_DOUBLE  32

/**
 * Indicates an integer (int) 
 * variable on CPU
 */
#define CUKR_STATE_CPU_INT     64
/**
 * Indicates an integer (int) 
 * variable on GPU
 */
#define CUKR_STATE_GPU_INT     128


/**
 * Matrix build phases
 */
 
/* Matrix created but not assembled */ 
#define CUKR_MAT_BUILD_CREATED      1  
/* Matrix created from file, 
 * but not assembled */ 
#define CUKR_MAT_BUILD_CREATED_FILE 2 
/* Matrix assembled */
#define CUKR_MAT_BUILD_ASSEMBLED    3 
 

// ============================ LOCATIONS & DATA TYPES =============================== //


/* Vectors and matrices can be on CPU or GPU. In
 * the same way, BLAS objects can act either on 
 * GPU or on CPU. In order to know if data transfer
 * is necessary or that the BLAS operation is valid
 * the following locations are declared.
 * 
 * The locations are created using the powers of 2
 * rule used before
 */

/**
 * Specifies that the object or the BLAS 
 * rezides/acts on the CPU 
 */ 
#define CUKR_LOCATION_CPU (CUKR_STATE_CPU_SINGLE  |  \
 						   CUKR_STATE_CPU_DOUBLE  |  \
 						   CUKR_STATE_CPU_QDOUBLE |  \
 						   CUKR_STATE_CPU_INT ) 
/**
 * Specifies that the object or the BLAS 
 * rezides/acts on the GPU 
 */ 
#define CUKR_LOCATION_GPU (CUKR_STATE_GPU_SINGLE  |  \
 						   CUKR_STATE_GPU_DOUBLE  |  \
 						   CUKR_STATE_GPU_QDOUBLE |  \
 						   CUKR_STATE_GPU_INT )


/* Vectors and matrices also have types. 
 * They are fully expressed by the status,
 * but for convinience, we create defines 
 * for these as well.
 * 
 * The power of 2 rule is used again 
 */

#define CUKR_TYPE_NONE 0 
  
/**
 * Indicates a single precision 
 * (float) variable
 */
#define CUKR_TYPE_SINGLE  (CUKR_STATE_CPU_SINGLE | CUKR_STATE_GPU_SINGLE)

/**
 * Indicates a double precision 
 * (double) variable
 */
#define CUKR_TYPE_DOUBLE  (CUKR_STATE_CPU_DOUBLE | CUKR_STATE_GPU_DOUBLE)

/**
 * Indicates a quasi-double precision 
 * (double) variable
 */
#define CUKR_TYPE_QDOUBLE  (CUKR_STATE_CPU_QDOUBLE | CUKR_STATE_GPU_QDOUBLE)

/**
 * Indicates an integer (int) 
 * variable
 */
#define CUKR_TYPE_INT     (CUKR_STATE_CPU_INT | CUKR_STATE_GPU_INT)
 
 
#endif /*CUKR_H_*/
