/**
 * @file db/db.c
 * @author Serban Georgescu
 * @since 2008
 *
 * This header file declares the options used 
 * throughout the package the solver. The 
 * options are based on the <argtable> library,
 * which should be present in the system. 
 * 
 * Other global data needed throughout the
 * package will also be packed into the database.
 * 
 * All the options are added to a structure
 * we call the CukrDB, the program's
 * database. Access will be global, so
 * all functions have access to it. In this 
 * way, program options will be able to 
 * propagate and be processed at the place 
 * where they are needed.
 * 
 */

#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include "../util/log.h"
#include "../util/utils.h"
#include "../cukr.h"
#include "db.h"


/*
 * GLOBALS
 */

/**
 * This file exports the cukr_db
 * object as global so that even the
 * most distant corners of the application
 * can have full acccess to the option
 * database and extract what they need.
 * 
 * However, for options, it is wise not 
 * deal directly with the database; 
 * and use the provided get/set functions instead.
 */
 
CukrDB cukr_db; 


// ------------------------------------------------------------------- //

/**
 * \brief Create the optiondb 
 */
void CukrDBCreate()
{	
	int i;
	
	/*
	 * Add options to the database
	 */
	
	/* Declare options */
	cukr_db.progname = "CUKr";
	cukr_db.int_args[CUKR_DB_PE]			 					= arg_int0("p",   "pe", 
													  					   "int", 
												      					   "Number of processors/GPUs per node to use"\
												      					   "(default: 1)");
	cukr_db.int_args[CUKR_DB_DEVICE]	 						= arg_int0("d",   "device",
														  				   "int",
													      				   "Start device to use"\
													      				   "(default: 0)");
	cukr_db.str_args[CUKR_DB_DOUBLE_PRECISION_BLAS]   			= arg_str0(NULL,  "double-precision-blas", 
						                                    	           "{Generic, MKL, GPUBLAS}", 
						                                    	           "BLAS library for double precision " \
						                                    	           "(default: Generic)");
	cukr_db.str_args[CUKR_DB_QDOUBLE_PRECISION_BLAS]    		= arg_str0(NULL,  "qdouble-precision-blas", 
					                                        	          "{Generic, GPUBLAS}", 
					                                        	          "BLAS library for quasi-double precision " \
					                                       	           "(default: Generic)");	                                                  
	cukr_db.str_args[CUKR_DB_SINGLE_PRECISION_BLAS]    			= arg_str0(NULL,  "single-precision-blas", 
					                                        	          "{Generic, MKL, GPUBLAS}", 
					                                        	          "BLAS library for single precision "\
					                                                  "(default: Generic)");
    cukr_db.str_args[CUKR_DB_INNER_BLAS]  	  					= arg_str0(NULL,  "inner-blas", 
					                                        	          "{Generic, MKL, GPUBLAS}", 
					                                        	          "Low-precision BLAS for iterative refinement "\
					                                        	          "(default: Generic)");	                                                  
	cukr_db.str_args[CUKR_DB_MATRIX]			  				= arg_str0(NULL, "matrix", 
																		   "<matrix path>",
																		   "sytem matrix file or folder" \
																		   "(default: Empty string)");                                                 
	cukr_db.str_args[CUKR_DB_MATRIX_FORMAT]  					= arg_str0(NULL,  "matrix-format", 
																		  "{CSR, BCSR, HYB, BELL}",
																		  "Sparse matrix format "\
																		  "(default: CSR)");
	cukr_db.str_args[CUKR_DB_MATRIX_FORMAT_VAR]  				= arg_str0(NULL,  "matrix-format-var", 
																	      "{NONE, CSR4, BCSR4}",
																	      "Sparse matrix format variation "\
																	      "(default: CSR4)");
	cukr_db.str_args[CUKR_DB_SOLVER] 		 					= arg_str0("sS",  "solver", 
																	      "{CG}", 
																	      "Solver type "\
																	      "(default: CG)");	
	cukr_db.str_args[CUKR_DB_SOLVER_PRECISION]					= arg_str0(NULL,  "solver-precision", 
																	      "{double, qdouble, single, half}", 
																	      "Solver precision "\
																	      "(default: double)");	
	cukr_db.str_args[CUKR_DB_SOLVER_INNER_PRECISION]   			= arg_str0(NULL,  "solver-inner-precision", 
																	      "{double, qdouble, single, half}", 
																	      "Inner solver precision "\
																    	  "(default: single)");												      	
	cukr_db.int_args[CUKR_DB_SOLVER_MAXIT]	 					= arg_int0(NULL,  "solver-maxit", 
																		  "int", 
																	      "Maximum no. of iterations "\
																	      "(default: 10000)");	
	cukr_db.int_args[CUKR_DB_SOLVER_INNER_MAXIT] 				= arg_int0(NULL,  "solver-inner-maxit", 
																		  "int", 
																	      "Maximum no. of inner iterations "\
																	      "(default: 10000)");												      										      
	cukr_db.int_args[CUKR_DB_SOLVER_INNER_MINIT] 				= arg_int0(NULL,  "solver-inner-minit", 
																		  "int", 
																	      "Minimum no. of inner iterations (overrides inner-rtol and inner-maxit), applies only for it.ref. and only for first outer iteration."\
																	      "(default: 0)");																      
	cukr_db.dbl_args[CUKR_DB_SOLVER_RTOL] 						= arg_dbl0(NULL,  "solver-rtol", 
																		  "double", 
																	      "Relative tolerance (||r|| < rtol||r0||) "\
																	      "(default: 1e-12)");
    cukr_db.dbl_args[CUKR_DB_SOLVER_INNER_RTOL] 				= arg_dbl0(NULL,  "solver-inner-rtol", 
																		  "double", 
																	      "Inner relative tolerance (||r|| < rtol||r0||) "\
																	      "(default: 1e-4)");				
	cukr_db.int_args[CUKR_DB_SOLVER_INNER_RTOL_AUTO]			= arg_int0(NULL,  "solver-inner-rtol-auto", 
																		  "integer", 
																	      "If nonzero, automatically set inner tolerance based on condition number estimation. First guess will be <solver-inner-rtol>"\
																	      "(default: 0");
	cukr_db.int_args[CUKR_DB_SOLVER_OUTER_RTOL_AUTO]			= arg_int0(NULL,  "solver-outer-rtol-auto", 
																		  "integer", 
																	      "If nonzero, automatically set outer tolerance based on condition number estimation and threshold"\
																	      "(default: 0");																			      												      															      								      
	cukr_db.dbl_args[CUKR_DB_SOLVER_OUTER_RTOL_AUTO_THRESHOLD]	= arg_dbl0(NULL,  "solver-outer-rtol-auto-threshold", 
																		  "double", 
																	      "Required level of outer residual reduction in order not to abort"\
																	      "(default: 5");
	cukr_db.dbl_args[CUKR_DB_SOLVER_OUTER_RTOL_AUTO_RELAXATION]	= arg_dbl0(NULL,  "solver-outer-rtol-auto-relaxation", 
																		  "double", 
																	      "Apply the treshold only after 1e16 - relaxation*condition level has been reached"\
																	      "(default: 1e+4");																      																		      																      
    cukr_db.dbl_args[CUKR_DB_SOLVER_ABSTOL]	 					= arg_dbl0(NULL,  "solver-abstol", 
																		  "double", 
																	      "Absolute tolerance (||r|| < abstol) "\
																	      "(default: 1e-50)");
    cukr_db.dbl_args[CUKR_DB_SOLVER_INNER_ABSTOL]				= arg_dbl0(NULL,  "solver-inner-abstol", 
																		  "double", 
																	      "Inner absolute tolerance (||r|| < abstol) "\
																	      "(default: 1e-50)");												      
	cukr_db.dbl_args[CUKR_DB_SOLVER_DIVTOL]	 					= arg_dbl0(NULL,  "solver-divtol", 
																		  "double", 
																	      "Divergence tolerance (||r|| > divtol||r0||) "\
																	      "(default: 1e+5)");	
    cukr_db.dbl_args[CUKR_DB_SOLVER_INNER_DIVTOL]		 		= arg_dbl0(NULL,  "solver-inner-divtol", 
																		  "double", 
																	      "Inner divergence tolerance (||r|| > divtol||r0||) "\
																	      "(default: 1e+5)");
    cukr_db.int_args[CUKR_DB_SOLVER_ITREF]		 				= arg_int0("iI",  "solver-itref", 
    																	  "integer",												   
																	      "Use iterative refinement (0 = false, !0 = true) "\
																	      "(default: 0)");
	cukr_db.int_args[CUKR_DB_SOLVER_EIGENINFO]		 			= arg_int0(NULL,  "solver-eigeninfo", 
    																	  "integer",												   
																	      "Save information for necessary for eigenvalue computations (0 = false, !0 = true) "\
																	      "(default: 0)");																      
	cukr_db.str_args[CUKR_DB_PC]	 		 					= arg_str0(NULL,  "pc", 
																	      "{none, jacobi}", 
																	      "Preconditioner type "\
																	      "(default: none)");													      
	cukr_db.str_args[CUKR_DB_MONITOR] 		 					= arg_strn("mM",  "monitor", 
												      					  "{rres, absres}",
																	      0, 100, 
																	      "Add monitor "\
																	      "(default: none)");												      													      										      													      											      								      
	cukr_db.help           										= arg_lit0("hH",  "help", 
																  		   "display this help and exit");
    cukr_db.version        										= arg_lit0("vV",  "version", 
                                                      					   "display version information and exit");
    cukr_db.end            										= arg_end(30);
    
    /* Add all to argtable */
    cukr_db.argtable[0]   = cukr_db.int_args[CUKR_DB_PE];
    cukr_db.argtable[1]   = cukr_db.int_args[CUKR_DB_DEVICE];
    cukr_db.argtable[2]   = cukr_db.str_args[CUKR_DB_DOUBLE_PRECISION_BLAS];
    cukr_db.argtable[3]   = cukr_db.str_args[CUKR_DB_QDOUBLE_PRECISION_BLAS];
    cukr_db.argtable[4]   = cukr_db.str_args[CUKR_DB_SINGLE_PRECISION_BLAS];
    cukr_db.argtable[5]   = cukr_db.str_args[CUKR_DB_INNER_BLAS];
    cukr_db.argtable[6]   = cukr_db.str_args[CUKR_DB_MATRIX];
    cukr_db.argtable[7]   = cukr_db.str_args[CUKR_DB_MATRIX_FORMAT];
    cukr_db.argtable[8]   = cukr_db.str_args[CUKR_DB_MATRIX_FORMAT_VAR];
    cukr_db.argtable[9]   = cukr_db.str_args[CUKR_DB_SOLVER];
    cukr_db.argtable[10]   = cukr_db.str_args[CUKR_DB_SOLVER_PRECISION];
    cukr_db.argtable[11]  = cukr_db.str_args[CUKR_DB_SOLVER_INNER_PRECISION];
    cukr_db.argtable[12]  = cukr_db.int_args[CUKR_DB_SOLVER_MAXIT];
    cukr_db.argtable[13]  = cukr_db.int_args[CUKR_DB_SOLVER_INNER_MAXIT];
    cukr_db.argtable[14]  = cukr_db.int_args[CUKR_DB_SOLVER_INNER_MINIT];
    cukr_db.argtable[15]  = cukr_db.dbl_args[CUKR_DB_SOLVER_RTOL];
    cukr_db.argtable[16]  = cukr_db.dbl_args[CUKR_DB_SOLVER_INNER_RTOL];
    cukr_db.argtable[17]  = cukr_db.int_args[CUKR_DB_SOLVER_INNER_RTOL_AUTO];
    cukr_db.argtable[18]  = cukr_db.int_args[CUKR_DB_SOLVER_OUTER_RTOL_AUTO];
    cukr_db.argtable[19]  = cukr_db.dbl_args[CUKR_DB_SOLVER_OUTER_RTOL_AUTO_THRESHOLD];
    cukr_db.argtable[20]  = cukr_db.dbl_args[CUKR_DB_SOLVER_OUTER_RTOL_AUTO_RELAXATION];
    cukr_db.argtable[21]  = cukr_db.dbl_args[CUKR_DB_SOLVER_ABSTOL];
    cukr_db.argtable[22]  = cukr_db.dbl_args[CUKR_DB_SOLVER_INNER_ABSTOL];
    cukr_db.argtable[23]  = cukr_db.dbl_args[CUKR_DB_SOLVER_DIVTOL];
    cukr_db.argtable[24]  = cukr_db.dbl_args[CUKR_DB_SOLVER_INNER_DIVTOL];
    cukr_db.argtable[25]  = cukr_db.int_args[CUKR_DB_SOLVER_ITREF];
    cukr_db.argtable[26]  = cukr_db.int_args[CUKR_DB_SOLVER_EIGENINFO];
    cukr_db.argtable[27]  = cukr_db.str_args[CUKR_DB_PC];
    cukr_db.argtable[28]  = cukr_db.str_args[CUKR_DB_MONITOR];
    cukr_db.argtable[29]  = cukr_db.help;
    cukr_db.argtable[30]  = cukr_db.version;
    cukr_db.argtable[31]  = cukr_db.end;
  
    /* Verify the argtable[] entries were allocated sucessfully */	
	if (arg_nullcheck(cukr_db.argtable) != 0)
       CUKR_FATAL("Argtable initialization failed. \n");
    
    /*
     * Init globals which need it 
     */
     
     cukr_db.threadsRunCalls = 0;
     cukr_db.solverTime = 0;

   	 cukr_db.allocated_cpu = 0;
   	 cukr_db.allocated_gpu = 0;   	    	 
}

// ------------------------------------------------------------------- //

/**
 * \brief Init the optiondb with the default
 *        options
 */
void CukrDBSetDefaults()
{	
	/* Integer defaults */
	CukrDBSetI(CUKR_DB_PE,                 1);
	CukrDBSetI(CUKR_DB_DEVICE,             0);
	CukrDBSetI(CUKR_DB_SOLVER_MAXIT,       10000);
	CukrDBSetI(CUKR_DB_SOLVER_INNER_MAXIT, 10000);
	CukrDBSetI(CUKR_DB_SOLVER_INNER_MINIT, 0);
	CukrDBSetI(CUKR_DB_SOLVER_ITREF,       0);
	CukrDBSetI(CUKR_DB_SOLVER_INNER_RTOL_AUTO, 0);
	CukrDBSetI(CUKR_DB_SOLVER_OUTER_RTOL_AUTO, 0);
			
	/* Double defaults */
	CukrDBSetD(CUKR_DB_SOLVER_RTOL,         1e-12);
	CukrDBSetD(CUKR_DB_SOLVER_INNER_RTOL,   1e-4);
	CukrDBSetD(CUKR_DB_SOLVER_ABSTOL,       1e-50);
	CukrDBSetD(CUKR_DB_SOLVER_INNER_ABSTOL, 1e-50);
	CukrDBSetD(CUKR_DB_SOLVER_DIVTOL,       1e+5);
	CukrDBSetD(CUKR_DB_SOLVER_INNER_DIVTOL, 1e+5);
	CukrDBSetD(CUKR_DB_SOLVER_OUTER_RTOL_AUTO_THRESHOLD,  5);
	CukrDBSetD(CUKR_DB_SOLVER_OUTER_RTOL_AUTO_RELAXATION, 1e+4);
		
	/* String defaluts */
	CukrDBSetS(CUKR_DB_MATRIX,                 "");
	CukrDBSetS(CUKR_DB_DOUBLE_PRECISION_BLAS,  "GENERIC");
	CukrDBSetS(CUKR_DB_QDOUBLE_PRECISION_BLAS, "GENERIC");
	CukrDBSetS(CUKR_DB_SINGLE_PRECISION_BLAS,  "GENERIC");
	CukrDBSetS(CUKR_DB_INNER_BLAS,             "GENERIC");
	CukrDBSetS(CUKR_DB_MATRIX_FORMAT,          "CSR");
	CukrDBSetS(CUKR_DB_MATRIX_FORMAT_VAR,      "CSR4");
	CukrDBSetS(CUKR_DB_SOLVER,                 "CG");
	CukrDBSetS(CUKR_DB_SOLVER_PRECISION,       "DOUBLE");
	CukrDBSetS(CUKR_DB_SOLVER_INNER_PRECISION, "SINGLE");
	CukrDBSetS(CUKR_DB_PC,                     "NONE");
}

// ------------------------------------------------------------------- //

/**
 * \brief Parse cmdline and add data from
 *        there.
 * \param argc,argv The arguments as received
 *        by the the main function.
 * \param strict Exit on syntax errors. Should be
 *        disabled if cukr is used as library.
 */
void CukrDBReadFromCmdline(int argc, char* argv[], int strict)
{
    /* Parse the command line as defined by argtable[] */
    int nerrors = arg_parse(argc, argv, cukr_db.argtable); 
 	   
    /* Special case: '--help' takes precedence over error reporting */
    if (cukr_db.help->count > 0)
    {
        printf("\nCUKr options: \n");
        printf("----------------------------------------------------\n");
        arg_print_syntax(stdout, cukr_db.argtable, "\n\n");
        arg_print_glossary(stdout, cukr_db.argtable, "  %-25s %s\n");
        printf("----------------------------------------------------\n");
        exit(1);
    }

    /* Special case: '--version' takes precedence error reporting */
    if (cukr_db.version->count > 0)
    {
        printf("'%s' CUKr (CUDA Krylov Space Solver.\n", cukr_db.progname);
        printf("From Feb. 2008, Serban Georgescu\n");
        exit(1);
     }

    /* If the parser returned any errors then display them and exit */
    if (nerrors > 0)
    {
        /* Display the error details contained in the arg_end struct.*/
        if (strict) {
        	arg_print_errors(stdout, cukr_db.end, cukr_db.progname);
        	printf("try '--help' for more information.\n");
        	exit(1);
        } 
    }
    
}

// ------------------------------------------------------------------- //

/**
 * \brief Return the value of the option
 *        having the specified option_id
 *        Used when the result is integer.
 * \param option_id Option id 
 */
int CukrDBGetI(int option_id)
{
	return cukr_db.int_args[option_id]->ival[0];
}

// ------------------------------------------------------------------- //

/**
 * \brief Return the value of the option
 *        having the specified option_id.
 *        Used for arguments where multiple
 *        instances are allowed. 
 *        Used when the result is integer.
 * \param option_id Option id 
 * \param index The index of the desired 
 *        instance
 */
int CukrDBGetExI(int option_id, unsigned int index)
{	
	if (index >= cukr_db.int_args[option_id]->hdr.maxcount)  
		CUKR_FATALX("Invalid option index for "\
		            "option_id %d: %d", option_id, index);	
		
	return cukr_db.int_args[option_id]->ival[index];
}

// ------------------------------------------------------------------- //

/**
 * \brief Return the value of the option
 *        having the specified option_id
 *        Used when the result is a double. 
 * \param option_id Option id 
 */
double CukrDBGetD(int option_id)
{
	return cukr_db.dbl_args[option_id]->dval[0];	
}

// ------------------------------------------------------------------- //

/**
 * \brief Return the value of the option
 *        having the specified option_id.
 *        Used for arguments where multiple
 *        instances are allowed. 
 * 		  Used when the result is a double.
 * \param option_id Option id 
 * \param index The index of the desired 
 *        instance
 */
double CukrDBGetExD(int option_id, unsigned int index)
{
	if (index >= cukr_db.dbl_args[option_id]->hdr.maxcount)  
		CUKR_FATALX("Invalid option index for "\
		            "option_id %d: %d", option_id, index);		
	
	return cukr_db.dbl_args[option_id]->dval[index];	
}

// ------------------------------------------------------------------- //

/**
 * \brief Return the value of the option
 *        having the specified option_id
 *        Used when the result is a string.
 * \param option_id Option id 
 */
const char* CukrDBGetS(int option_id)
{	
	return cukr_db.str_args[option_id]->sval[0];	
}

// ------------------------------------------------------------------- //

/**
 * \brief Return the value of the option
 *        having the specified option_id.
 *        Used for arguments where multiple
 *        instances are allowed. 
 * 		  Used when the result is a string.
 * \param option_id Option id 
 * \param index The index of the desired 
 *        instance
 */
const char* CukrDBGetExS(int option_id, unsigned int index)
{
	if (index >= cukr_db.str_args[option_id]->hdr.maxcount)  
		CUKR_FATALX("Invalid option index for "\
		            "option_id %d: %d", option_id, index);	
		
	return cukr_db.str_args[option_id]->sval[index];	
}


// ------------------------------------------------------------------- //

/**
 * \brief Set the value of the option
 *        having the specified option id.
 *        Used for integer values.
 * \param option_id The option id 
 * \param ival Value for the option
 */
void CukrDBSetI(int option_id, int ival)
{
	cukr_db.int_args[option_id]->ival[0] = ival;
}

// ------------------------------------------------------------------- //

/**
 * \brief Set the value of the option
 *        having the specified option id.
 *        Used for arguments where multiple
 *        instances are allowed. 
 *        Used for integer values.
 * \param option_id The option id 
 * \param index The index of the desired 
 *        instance
 * \param ival Value for the option
 */
void CukrDBSetExI(int option_id, unsigned int index, int ival)
{
	if (index >= cukr_db.int_args[option_id]->hdr.maxcount)  
		CUKR_FATALX("Invalid option index for "\
		            "option_id %d: %d", option_id, index);	
	
	cukr_db.int_args[option_id]->ival[index] = ival;
}

// ------------------------------------------------------------------- //

/**
 * \brief Append a new value to the option
 *        having the specified option id.
 *        Used for arguments where multiple
 *        instances are allowed. 
 *        Used for integer values.
 * \param option_id The option id 
 * \param ival Value for the option
 */
void CukrDBAppendI(int option_id, int ival)
{
	if (cukr_db.int_args[option_id]->count >= cukr_db.int_args[option_id]->hdr.maxcount)
		CUKR_FATALX("Option with option_id=%d does not "\
		            "allow more than %d instances", option_id, 
		            cukr_db.int_args[option_id]->hdr.maxcount);	
	
	cukr_db.int_args[option_id]->ival[cukr_db.int_args[option_id]->count] = ival;
	cukr_db.int_args[option_id]->count++;
}

// ------------------------------------------------------------------- //

/**
 * \brief Set the value of the option
 *        having the specified option id.
 *        Used for double values.
 * \param option_id The option id 
 * \param dval Value for the option
 */
void CukrDBSetD(int option_id, double dval)
{
	cukr_db.dbl_args[option_id]->dval[0] = dval;
}

// ------------------------------------------------------------------- //

/**
 * \brief Set the value of the option
 *        having the specified option id.
 *        Used for arguments where multiple
 *        instances are allowed. 
 *        Used for double values.
 * \param option_id The option id 
 * \param index The index of the desired 
 *        instance
 * \param dval Value for the option
 */
void CukrDBSetExD(int option_id, unsigned int index, double dval)
{
	if (cukr_db.dbl_args[option_id]->count >= cukr_db.dbl_args[option_id]->hdr.maxcount)  
		CUKR_FATALX("Invalid option index for "\
		            "option_id %d: %d", option_id, index);	
	
	cukr_db.dbl_args[option_id]->dval[index] = dval;
}

// ------------------------------------------------------------------- //

/**
 * \brief Append a new value to the option
 *        having the specified option id.
 *        Used for arguments where multiple
 *        instances are allowed. 
 *        Used for double values.
 * \param option_id The option id 
 * \param dval Value for the option
 */
void CukrDBAppendD(int option_id, double dval)
{
	if (cukr_db.dbl_args[option_id]->count >= cukr_db.dbl_args[option_id]->hdr.maxcount)  
		CUKR_FATALX("Option with option_id=%d does not "\
		            "allow more than %d instances", option_id, 
		            cukr_db.dbl_args[option_id]->hdr.maxcount);	
	
	cukr_db.dbl_args[option_id]->dval[cukr_db.dbl_args[option_id]->count] = dval;
	cukr_db.dbl_args[option_id]->count++;
}

// ------------------------------------------------------------------- //

/**
 * \brief Set the value of the option
 *        having the specified option id.
 *        Used for string values.
 * \param option_id The option id 
 * \param sval Value for the option
 */
void CukrDBSetS(int option_id, char *sval)
{
	cukr_db.str_args[option_id]->sval[0] = sval;
}

// ------------------------------------------------------------------- //

/**
 * \brief Set the value of the option
 *        having the specified option id.
 *        Used for arguments where multiple
 *        instances are allowed. 
 *        Used for string values.
 * \param option_id The option id 
 * \param index The index of the desired 
 *        instance
 * \param sval Value for the option
 */
void CukrDBSetExS(int option_id, unsigned int index, char* sval)
{
	if (cukr_db.str_args[option_id]->count >= cukr_db.str_args[option_id]->hdr.maxcount)  
		CUKR_FATALX("Invalid option index for "\
		            "option_id %d: %d", option_id, index);	
	
	cukr_db.str_args[option_id]->sval[index] = sval;
}

// ------------------------------------------------------------------- //

/**
 * \brief Append a new value to the option
 *        having the specified option id.
 *        Used for arguments where multiple
 *        instances are allowed. 
 *        Used for string values.
 * \param option_id The option id 
 * \param sval Value for the option
 */
void CukrDBAppendS(int option_id, char* sval)
{
	if (cukr_db.str_args[option_id]->count >= cukr_db.str_args[option_id]->hdr.maxcount)  
		CUKR_FATALX("Option with option_id=%d does not "\
		            "allow more than %d instances", option_id, 
		            cukr_db.str_args[option_id]->hdr.maxcount);	
	
	cukr_db.str_args[option_id]->sval[cukr_db.str_args[option_id]->count] = sval;
	cukr_db.str_args[option_id]->count++;
}

// ------------------------------------------------------------------- //

/**
 * \brief Return the number of instances
 *        for the specified option_id
 *        Used when the result is integer.
 * \param option_id Option id 
 */
int CukrDBGetCountI(int option_id)
{
	return cukr_db.int_args[option_id]->count;
}

// ------------------------------------------------------------------- //

/**
 * \brief Return the number of instances
 *        for the specified option_id
 *        Used when the result is double.
 * \param option_id Option id 
 */
int CukrDBGetCountD(int option_id)
{
	return cukr_db.dbl_args[option_id]->count;
}

// ------------------------------------------------------------------- //

/**
 * \brief Return the number of instances
 *        for the specified option_id
 *        Used when the result is string.
 * \param option_id Option id 
 */
int CukrDBGetCountS(int option_id)
{
	return cukr_db.str_args[option_id]->count;
}
