#ifndef CUKR_DB_H_
#define CUKR_DB_H_

/**
 * @file db/db.h
 * @author Serban Georgescu
 * @since 2008
 *
 * This header file declares the options used 
 * throughout the package the solver. The 
 * options are based on the <argtable> library,
 * which should be present in the system. 
 * 
 * Other global data needed throughout the
 * package will also be packed into the database.
 * 
 * All the options are added to a structure
 * we call the CukrDB, the program's
 * database. Access will be global, so
 * all functions have access to it. In this 
 * way, program options will be able to 
 * propagate and be processed at the place 
 * where they are needed.
 * 
 */

#include <argtable2.h>
#include "../blas/blas.h"
#include <pthread.h>

/**
 * Set defines for all options
 * This becomes the option_id, by which
 * the option can be referrenced 
 */

/* Maximum no. of arguments per section */
#define CUKR_DB_MAXARGS 15

/* Integers */
#define CUKR_DB_PE                                  0
#define CUKR_DB_DEVICE                              1
#define CUKR_DB_SOLVER_MAXIT                        2
#define CUKR_DB_SOLVER_INNER_MAXIT                  3
#define CUKR_DB_SOLVER_INNER_MINIT                  4
#define CUKR_DB_SOLVER_ITREF                        5
#define CUKR_DB_SOLVER_EIGENINFO                    6
#define CUKR_DB_SOLVER_INNER_RTOL_AUTO              7
#define CUKR_DB_SOLVER_OUTER_RTOL_AUTO              8

/* Reals */
#define CUKR_DB_SOLVER_RTOL                         0
#define CUKR_DB_SOLVER_INNER_RTOL                   1
#define CUKR_DB_SOLVER_ABSTOL                       2
#define CUKR_DB_SOLVER_INNER_ABSTOL                 3
#define CUKR_DB_SOLVER_DIVTOL                       4
#define CUKR_DB_SOLVER_INNER_DIVTOL                 5
#define CUKR_DB_SOLVER_OUTER_RTOL_AUTO_THRESHOLD    6
#define CUKR_DB_SOLVER_OUTER_RTOL_AUTO_RELAXATION   7

/* Strings */
#define CUKR_DB_MATRIX                              0
#define CUKR_DB_DOUBLE_PRECISION_BLAS               1 
#define CUKR_DB_QDOUBLE_PRECISION_BLAS              2
#define CUKR_DB_SINGLE_PRECISION_BLAS               3
#define CUKR_DB_INNER_BLAS                          4
#define CUKR_DB_MATRIX_FORMAT                       5
#define CUKR_DB_MATRIX_FORMAT_VAR                   6
#define CUKR_DB_SOLVER                              7 
#define CUKR_DB_SOLVER_PRECISION                    8
#define CUKR_DB_SOLVER_INNER_PRECISION              9
#define CUKR_DB_PC                                  10
#define CUKR_DB_MONITOR                             11

/**
 * Struct defining the option database
 */
typedef struct
{
	/* 
	 * Various globals 
	 */
	
	/* Key for accessing thread-local
	 * storage */
	pthread_key_t thread_key;

	/**
	 * The current node ID and the
	 * number of nodes, as obtained from 
	 * the MPI call 
	 */
	int mpi_rank, mpi_size;  
		
	/**
	 * The presence or absence of the
	 * computation on the GPU (actual usage,
	 * not compiler support)
	 */
	int compute_on_gpu;

	/**
	 * Number of GPUs detected
	 */
	int gpu_count;

	/**
	 * Export these BLAS objects so that
	 * all operations could be written using
	 * them. This way changing the BLAS can
	 * be done from command-line options alone.
	 */
	CUKR_BLAS blasD, blasQ, blasS, blasI;
	
	/**
	 * Number of calls to the CukrThreadRun.
	 * Good to have it, has to be as small as
	 * possible */
	int threadsRunCalls; 
	
	/**
	 * Total time spent in solver.
	 * Updated by the solver.
	 * Used for computing runtime efficiency */
	double solverTime;
	
	/**
	 * Total memory allocated (in bytes), 
	 * on CPU and GPU */
	long allocated_cpu, allocated_gpu; 
			
	/* 
	 * Program options 
	 */
	 
	const char* progname;
	struct arg_int*  int_args[CUKR_DB_MAXARGS];
	struct arg_dbl*  dbl_args[CUKR_DB_MAXARGS];
	struct arg_str*  str_args[CUKR_DB_MAXARGS];
	struct arg_lit*  version;
	struct arg_lit*  help;
	struct arg_end*  end;
    void *argtable[CUKR_DB_MAXARGS * 3 + 2];
	  
} CukrDB;

// ------------------------------------------------------------------- //

/**
 * \brief Create the optiondb 
 */
void CukrDBCreate();

// ------------------------------------------------------------------- //

/**
 * \brief Init the optiondb with the default
 *        options
 */
void CukrDBSetDefaults();

// ------------------------------------------------------------------- //

/**
 * \brief Parse cmdline and add data from
 *        there.
 * \param argc,argv The arguments as received
 *        by the the main function.
 * \param strict Exit on syntax errors. Should be
 *        disabled if cukr is used as library.
 */
void CukrDBReadFromCmdline(int argc, char* argv[], int strict);

// ------------------------------------------------------------------- //

/**
 * \brief Return the value of the option
 *        having the specified option_id.
 *        Used for arguments where multiple
 *        instances are allowed. 
 *        Used when the result is integer.
 * \param option_id Option id 
 * \param index The index of the desired 
 *        instance
 */
int CukrDBGetExI(int option_id, unsigned int index);

// ------------------------------------------------------------------- //

/**
 * \brief Return the value of the option
 *        having the specified option_id
 *        Used when the result is a double. 
 * \param option_id Option id 
 */
double CukrDBGetD(int option_id);

// ------------------------------------------------------------------- //

/**
 * \brief Return the value of the option
 *        having the specified option_id.
 *        Used for arguments where multiple
 *        instances are allowed. 
 * 		  Used when the result is a double.
 * \param option_id Option id 
 * \param index The index of the desired 
 *        instance
 */
double CukrDBGetExD(int option_id, unsigned int index);

// ------------------------------------------------------------------- //

/**
 * \brief Return the value of the option
 *        having the specified option_id
 *        Used when the result is a string.
 * \param option_id Option id 
 */
const char* CukrDBGetS(int option_id);

// ------------------------------------------------------------------- //

/**
 * \brief Return the value of the option
 *        having the specified option_id.
 *        Used for arguments where multiple
 *        instances are allowed. 
 * 		  Used when the result is a string.
 * \param option_id Option id 
 * \param index The index of the desired 
 *        instance
 */
const char* CukrDBGetExS(int option_id, unsigned int index);


// ------------------------------------------------------------------- //

/**
 * \brief Set the value of the option
 *        having the specified option id.
 *        Used for integer values.
 * \param option_id The option id 
 * \param ival Value for the option
 */
void CukrDBSetI(int option_id, int ival);

// ------------------------------------------------------------------- //

/**
 * \brief Set the value of the option
 *        having the specified option id.
 *        Used for arguments where multiple
 *        instances are allowed. 
 *        Used for integer values.
 * \param option_id The option id 
 * \param index The index of the desired 
 *        instance
 * \param ival Value for the option
 */
void CukrDBSetExI(int option_id, unsigned int index, int ival);

// ------------------------------------------------------------------- //

/**
 * \brief Append a new value to the option
 *        having the specified option id.
 *        Used for arguments where multiple
 *        instances are allowed. 
 *        Used for integer values.
 * \param option_id The option id 
 * \param ival Value for the option
 */
void CukrDBAppendI(int option_id, int ival);

// ------------------------------------------------------------------- //

/**
 * \brief Set the value of the option
 *        having the specified option id.
 *        Used for double values.
 * \param option_id The option id 
 * \param dval Value for the option
 */
void CukrDBSetD(int option_id, double dval);

// ------------------------------------------------------------------- //

/**
 * \brief Set the value of the option
 *        having the specified option id.
 *        Used for arguments where multiple
 *        instances are allowed. 
 *        Used for double values.
 * \param option_id The option id 
 * \param index The index of the desired 
 *        instance
 * \param dval Value for the option
 */
void CukrDBSetExD(int option_id, unsigned int index, double dval);

// ------------------------------------------------------------------- //

/**
 * \brief Append a new value to the option
 *        having the specified option id.
 *        Used for arguments where multiple
 *        instances are allowed. 
 *        Used for double values.
 * \param option_id The option id 
 * \param dval Value for the option
 */
void CukrDBAppendD(int option_id, double dval);

// ------------------------------------------------------------------- //

/**
 * \brief Set the value of the option
 *        having the specified option id.
 *        Used for string values.
 * \param option_id The option id 
 * \param sval Value for the option
 */
void CukrDBSetS(int option_id, char *sval);

// ------------------------------------------------------------------- //

/**
 * \brief Set the value of the option
 *        having the specified option id.
 *        Used for arguments where multiple
 *        instances are allowed. 
 *        Used for string values.
 * \param option_id The option id 
 * \param index The index of the desired 
 *        instance
 * \param sval Value for the option
 */
void CukrDBSetExS(int option_id, unsigned int index, char* sval);

// ------------------------------------------------------------------- //

/**
 * \brief Append a new value to the option
 *        having the specified option id.
 *        Used for arguments where multiple
 *        instances are allowed. 
 *        Used for string values.
 * \param option_id The option id 
 * \param sval Value for the option
 */
void CukrDBAppendS(int option_id, char* sval);

// ------------------------------------------------------------------- //

/**
 * \brief Return the number of instances
 *        for the specified option_id
 *        Used when the result is integer.
 * \param option_id Option id 
 */
int CukrDBGetCountI(int option_id);

// ------------------------------------------------------------------- //

/**
 * \brief Return the number of instances
 *        for the specified option_id
 *        Used when the result is double.
 * \param option_id Option id 
 */
int CukrDBGetCountD(int option_id);

// ------------------------------------------------------------------- //

/**
 * \brief Return the number of instances
 *        for the specified option_id
 *        Used when the result is string.
 * \param option_id Option id 
 */
int CukrDBGetCountS(int option_id);

#endif /*CUKR_DB_H_*/
