/**
 * @file final/final.c
 * @author Serban Georgescu
 * @since 2008
 * 
 * Here we include all the post tasks
 * that must be performed, including:
 * 	- reporting the BLAS performance
 *  - ...
 */

#define LOGGER_NAME "CUKr.final.final"
  
#include <string.h>
#include "../monitors/reports/reports.h"
#include "../init/init.h"
#include "../threading/threading.h"
#include "../util/utils.h"
#include "final.h"
#ifdef CUKR_USE_CUDA
	#include <cuda.h>
	#include <cublas.h>
#endif
#ifdef CUKR_USE_OPENCL
	#include "../init/init_opencl.h"
#endif
#ifdef CUKR_MPI
	#include <mpi.h>
#endif

/* Access the global option database and blas objects */
extern CukrDB cukr_db;

// ------------------------------------------------------------------- //

/**
 * \brief Final cleanup, shutdown and reporting
 * \param opts Options  
 */
void CukrFinalize(int opts)
{	
	/* Report */
	if (opts & CUKR_OPTS_BLAS_REPORT)	
		CukrFin_ReportBLAS();
	
	if (opts & CUKR_OPTS_RUNTIME_REPORT)	
		CukrFin_ReportRuntime();
	 
	/* Shudtown */
	#ifdef CUKR_CUDA
		CukrFin_ShutdownCublas();
	#endif
	#ifdef CUKR_USE_OPENCL
	CukrFin_ShutdownOpenCL();
	#endif
	CukrFin_ShutdownThreading();
	#ifdef CUKR_MPI
		CukrFin_ShutdownMPI();
	#endif			
}

/**
 * @todo Add to the blas vector below
 *       all global BLAS objects which
 * 		 are used in the application
 */

// ------------------------------------------------------------------- //

/*
 * \brief Generate performance reports for
 *        the BLAS 
 */
void CukrFin_ReportBLAS()
{
	char buffer[255];
	int i, j, k, l, m, n, p;
	
	/*
	 * Since we have more than one blas 
	 * objects, some of which may point 
	 * to the same blas, we need to merge 
	 * the counters if the blases are the 
	 * same.
	 */
	
	/* Make a vector of BLAS pointers
	 * of size 5 (the current no. of
	 * global blas objects) */
	const int BLAS_COUNT = 4;
	CUKR_BLAS blas[BLAS_COUNT];
	
	/* Populate the vectors with the
	 * global blases */
	blas[0] = cukr_db.blasD;
	blas[1] = cukr_db.blasQ;
	blas[2] = cukr_db.blasS;
	blas[3] = cukr_db.blasI; 
	
	/* Merge the counters and NULL the
	 * duplicates */
	for (i = 0; i < BLAS_COUNT - 1; i++)
		for (j = i + 1; j < BLAS_COUNT; j++)
		 	if (blas[j] != NULL && blas[i] != NULL) 
		 		if (!CukrStricmp(blas[i]->name, blas[j]->name))
		 		{
		 			/* Merge thread counters */
		 			for (k = 0; k < CUKR_MAX_PE; k++)
						for (l = 0; l < CUKR_BLAS_IDX_MAXCOUNT; l++) 
						{							
							/* For BLAS */
							for (m = 0; m < CUKR_BLAS_OPS_MAXCOUNT; m++)  
								for (n = 0; n < CUKR_BLAS_PREC_MAXCOUNT; n++)
									blas[i]->counters[k][l][m][n] += blas[j]->counters[k][l][m][n];
									
							/* For BLASX */
							for (m = 0; m < CUKR_BLASX_OPS_MAXCOUNT; m++)  
								for (n = 0; n < CUKR_BLAS_PREC_MAXCOUNT; n++)
									if (blas[i]->blasx != blas[j]->blasx)
										blas[i]->blasx->counters[k][l][m][n] += blas[j]->blasx->counters[k][l][m][n];			
						}
		 			
		 			/* Merge MP and MPI BLAS counters */
					for (l = 0; l < CUKR_BLAS_IDX_MAXCOUNT; l++)
						for (m = 0; m < CUKR_BLAS_OPS_MAXCOUNT; m++) {
								blas[i]->counters_mp [l][m] += blas[j]->counters_mp [l][m];
								blas[i]->counters_mpi[l][m] += blas[j]->counters_mpi[l][m];
						}
	 					 				
		 			/* Merge MP and MPI BLASX counters */
					for (l = 0; l < CUKR_BLAS_IDX_MAXCOUNT; l++)
						for (m = 0; m < CUKR_BLASX_OPS_MAXCOUNT; m++)
						if (blas[i]->blasx != blas[j]->blasx) {
								blas[i]->blasx->counters_mp [l][m] += blas[j]->blasx->counters_mp [l][m];
								blas[i]->blasx->counters_mpi[l][m] += blas[j]->blasx->counters_mpi[l][m];
						}
								 			
		 			/* Merge BLAS comm times */
		 			blas[i]->comm_inner += blas[j]->comm_inner;
		 			blas[i]->comm_outer += blas[j]->comm_outer;
		 			
		 			/* Merge BLASX comm times */
		 			blas[i]->blasx->comm_inner += blas[j]->blasx->comm_inner;
		 			blas[i]->blasx->comm_outer += blas[j]->blasx->comm_outer;
		 			
		 			/* Null the second one */
		 			blas[j] = NULL;	
		 		} 
	
	/* Merge node-level counters for the
	 * ones that remained */
	for (i = 0; i < BLAS_COUNT; i++)
		if (blas[i] != NULL) {
			CukrBlasMergeNodeCounters(blas[i]);
			CukrBlasxMergeNodeCounters(blas[i]->blasx);
		}
	
	/*
	 * Now report the performance of the
	 * ones that remained 
	 */
	
	/* Get no. of threads */
	int pe = CukrDBGetI(CUKR_DB_PE);
	
	for (i = 0; i < BLAS_COUNT; i++)
		if (blas[i] != NULL)
		{
			/*
			 * Report BLAS
			 */
			
			/* Thread level, for all threads */
			for (p = 0; p < pe; p++) {
				sprintf(buffer, "%s.blas.thread%d.txt", blas[i]->name, p);
				CukrReportBlasPerformance(blas[i], p, buffer);
				sprintf(buffer, "%s.blasx.thread%d.txt", blas[i]->name, p);
				CukrReportBlasxPerformance(blas[i]->blasx, p, buffer);
			}	
			
			/* Node level */
			sprintf(buffer, "%s.blas.node.txt", blas[i]->name);
			CukrReportBlasPerformanceMP(blas[i], buffer);
			sprintf(buffer, "%s.blasx.node.txt", blas[i]->name);
			CukrReportBlasxPerformanceMP(blas[i]->blasx, buffer);
			
			/* Cluster level */
			sprintf(buffer, "%s.blas.cluster.txt", blas[i]->name);
			CukrReportBlasPerformanceMPI(blas[i], buffer);
			sprintf(buffer, "%s.blasx.cluster.txt", blas[i]->name);
			CukrReportBlasxPerformanceMPI(blas[i]->blasx, buffer);		
		}		
	 
}

// ------------------------------------------------------------------- //

/**
 * \brief Generate solver report
 * \param solver The solver
 */
void CukrFin_ReportSolver(CUKR_SOLVER solver)
{
	CukrReportSolver(solver, "solver.txt");
}

// ------------------------------------------------------------------- //

/**
 * \brief Generate runtime report
 */
void CukrFin_ReportRuntime()
{
	CukrReportRuntime("runtime.txt");
}

// ------------------------------------------------------------------- //

/**
 * \brief Shutdown CUBLAS
 */
void CukrFin_ShutdownCublas()
{
	#ifdef CUKR_USE_CUDA
		cublasShutdown();
	#endif	
}

// ------------------------------------------------------------------- //

/**
 * \brief Shutdown OpenCL
 */
void CukrFin_ShutdownOpenCL()
{
#ifdef CUKR_USE_OPENCL
	err = clFlush(ComputeCommands);
	if (err != CL_SUCCESS) {
		printf("Flush: OpenCL problems.\n");
	}
	
	clReleaseEvent(qdot_event);
	clReleaseEvent(sdot_event);
	clReleaseEvent(ddot_event);
	
	clReleaseEvent(qaxpy_event);
	clReleaseEvent(saxpy_event);
	clReleaseEvent(daxpy_event);
	
	clReleaseEvent(qaypx_event);
	clReleaseEvent(saypx_event);
	clReleaseEvent(daypx_event);
	
	clReleaseEvent(qscal_event);
	clReleaseEvent(sscal_event);
	clReleaseEvent(dscal_event);
	
	clReleaseEvent(scopy_event);
	clReleaseEvent(dcopy_event);
	//clReleaseEvent();
	
	clReleaseEvent(sspmv_csr_event);
	clReleaseEvent(sspmv_csr_a1_b0_event);
	clReleaseEvent(sspmv_csr_a1_event);
	clReleaseEvent(sspmv_csr_b0_event);
	
	clReleaseEvent(sspmv_csr4_event);
	clReleaseEvent(sspmv_csr4_a1_b0_event);
	clReleaseEvent(sspmv_csr4_a1_event);
	clReleaseEvent(sspmv_csr4_b0_event);
	
	err = clFinish(ComputeCommands);
	if (err != CL_SUCCESS) {
		printf("Finnish: OpenCL problems.\n");
	}
	
	clReleaseCommandQueue(ComputeCommands);
	clReleaseContext(ComputeContext);
	
	printf("OpenCL flushed and shut down!\n");
#endif	
}

// ------------------------------------------------------------------- //

/**
 * \brief Stop all worker threads
 */
void CukrFin_ShutdownThreading()
{	
	int pe = CukrDBGetI(CUKR_DB_PE);
	CukrThreadsStop(pe);
} 

// ------------------------------------------------------------------- //

/**
 * \brief Shutdown MPI 
 */
void CukrFin_ShutdownMPI()
{	
	#ifdef CUKR_MPI
		MPI_Finalize();
	#endif	
} 
 
