/**
 * @file init/init.c
 * @author Serban Georgescu
 * @since 2008
 * 
 * Here we include all initialization
 * tasks that must be performed, including:
 * 	- creating the option library
 *  - creating and initialzing the BLAS
 *  - ...
 */

#define LOGGER_NAME "CUKr.init"

#include <stdio.h> 
#include <stdlib.h> 
#include <string.h> 
#ifdef CUKR_MPI
	#include <mpi.h>
#endif	
#include "init.h"
#include "init_cuda.h"
#ifdef CUKR_USE_OPENCL
	#include "init_opencl.h"
#endif
#include "../util/log.h"
#include "../util/utils.h"
#include "../db/db.h"
#include "../threading/threading.h"
#ifdef CUKR_USE_CUDA
	#include <cublas.h>
#endif
/*
#ifdef CUKR_USE_OPENCL
	#include <opencl.h>
#endif
*/
#include "../blas/impl/generic/generic_blas.h"
#include "../blas/impl/mkl/mkl_blas.h"
#include "../blas/impl/gpu/gpu_blas.h"
#include "../blasx/impl/generic/generic_blasx.h"
#include "../blasx/impl/gpu/gpu_blasx.h"


/*
 * GLOBALS
 */

extern CukrDB cukr_db;

// ------------------------------------------------------------------- //

/**
 * \brief Does all the initialization stuff
 *        needed when using CUKr, but does
 *        not start threading or reading 
 *        /creating the system matrix. 
 * \param argc, argv The program arguments
 * \param opts Additional options
 */
void CukrInit(int argc, char* argv[], int opts)
{		
	/* Initialize logging */
	CukrInit_Logging();
	
	/* Init MPI */
	CukrInit_MPI(argc, argv);
	
	/* Read parameters and arguments and 
	 * populate the option database */  
	CukrInit_OptionDB(argc, argv, opts & CUKR_OPTS_STRICT_SYNTAX);	
}

// ------------------------------------------------------------------- //

/**
 * \brief Start CUKr based on the options set
 *        at initialization and after
 * \param opts Additional options
 */
void CukrStart(int opts)
{
	/* See if GPU is actually being used for
	 * computation */
	if (opts & CUKR_OPTS_COMPUTES_ON_GPU)
		cukr_db.compute_on_gpu = 1;
	else
		CukrInit_ComputesOnGpu();
	
	#ifdef CUKR_USE_GPU
		if (cukr_db.compute_on_gpu) {
			#ifdef CUKR_USE_CUDA
				/* Initialize CUDA */
				CukrInit_Cuda();
		
				/* Initialize CUBLAS */
				CukrInit_Cublas();
			#endif
			#ifdef CUKR_USE_OPENCL
				/* Initialize OpenCL */
				CukrInit_OpenCL();
			
				/* Initialize clBlas : to come */
				// CukrInit_clBlas();
			#endif
		}
	#endif
	
	/* Initialize the BLAS */
	CukrInit_BLAS();	
	/* Initialize threading */
	CukrInit_Threading();
}

// ------------------------------------------------------------------- //

/*
 * \brief Run the option parser and create
 *        the option database.
 * \param argc argument count received by the
 *        <main> function
 * \param argv pointer to the parameter list
 *        received by the <main> function
 * \param strict Strict syntax check (exit on
 *        unknown options)
 */
 
void CukrInit_OptionDB(int argc, char* argv[], int strict)
{
	CukrDBCreate();
	CukrDBSetDefaults();
 	CukrDBReadFromCmdline(argc, argv, strict);
} 

// ------------------------------------------------------------------- //

/*
 * \brief Initialize the logging system
 */
void CukrInit_Logging()
{
	;	
}

// ------------------------------------------------------------------- //

/**
 * \brief Initializes the MPI subsystem
 * \param argc argument count received by the
 *        <main> function
 * \param argv pointer to the parameter list
 *        received by the <main> function
 */
void CukrInit_MPI(int argc, char* argv[])
{	
	#ifdef CUKR_MPI
		/* Init MPI and get rank and size */
		MPI_Init(&argc, &argv);
		MPI_Comm_size(MPI_COMM_WORLD, &cukr_db.mpi_size);
	  	MPI_Comm_rank(MPI_COMM_WORLD, &cukr_db.mpi_rank);
	  	
	  	LOGX(INFO, "MPI initialized: rank %d (from %d nodes)", 
	  	           cukr_db.mpi_rank, cukr_db.mpi_size);
	#endif  	           
}

// ------------------------------------------------------------------- //

/**
 * \brief Do CUDA-related initializations
 */
void CukrInit_Cuda()
#ifndef CUKR_USE_CUDA
{
	CUKR_FATAL("GPU support was not compiled in the library");
}
#else
{
	/* Get the no. of devices in the system */
	cukr_db.gpu_count = CukrInit_Cuda_cu();

	/* Error if no GPUs found */
    if (cukr_db.gpu_count == 0) {
    	CUKR_FATAL("No GPUs found in the system");
    }
    else {
		LOGX(INFO, "%d GPUs present in the system", cukr_db.gpu_count);
    } 
	
	/* See if the number of requested gpus 
	 * (by option) does not exceed this number 
	 * Only if using GPU */
	
 	if (!CukrStricmp(CukrDBGetS(CUKR_DB_DOUBLE_PRECISION_BLAS),  "GPUBLAS") ||
 		!CukrStricmp(CukrDBGetS(CUKR_DB_QDOUBLE_PRECISION_BLAS), "GPUBLAS") ||
 		!CukrStricmp(CukrDBGetS(CUKR_DB_SINGLE_PRECISION_BLAS),  "GPUBLAS") || 
 		!CukrStricmp(CukrDBGetS(CUKR_DB_INNER_BLAS),             "GPUBLAS")   )
	    if (CukrDBGetI(CUKR_DB_PE) > cukr_db.gpu_count) {
	    	CUKR_FATALX("%d GPUs requested when only %d were "\
	    	            "found in the system", 
	    	            CukrDBGetI(CUKR_DB_PE), cukr_db.gpu_count);
	    }
	    else {
	    	LOGX(INFO, "Using %d PEs", CukrDBGetI(CUKR_DB_PE));
	    } 

 	int start_device  = CukrDBGetI(CUKR_DB_DEVICE);
	if (cukr_db.compute_on_gpu) {
		CukrSetDevice(start_device);
	}
}
#endif /* CUKR_USE_CUDA */

// ------------------------------------------------------------------- //

/**
 * \brief Initialize CUBLAS
 */
void CukrInit_Cublas()
#ifndef CUKR_USE_CUDA
{
	CUKR_FATAL("GPU support was not compiled in the library");
}
#else
{
	cublasStatus status;
	
	/* Init CUBLAS */
	status = cublasInit();
	
	/* Check for errors */ 
	if (status != CUBLAS_STATUS_SUCCESS)
		CUKR_FATAL("Error initializing CUBLAS. Possible causes are: \
		             - no CUDA device while compiling in Device mode \
		             - linked with the wrong cublas / cublasemu \
		             - not -lcuda found");
}
#endif /* CUKR_USE_CUDA */

// ------------------------------------------------------------------- //

/**
 * \brief Do OpenCL-related initializations
 */
void CukrInit_OpenCL()
#ifndef CUKR_USE_OPENCL
{
	CUKR_FATAL("OpenCL support was not compiled in the library");
}
#else
{
	/* Get the no. of devices in the system */
	cukr_db.gpu_count = CukrInit_OpenCL_cl();
	
	/* Error if no GPUs found */
    if (cukr_db.gpu_count == 0) {
    	CUKR_FATAL("No OpenCL devices found in the system");
    }
    else {
		LOGX(INFO, "%d OpenCL devices present in the system", cukr_db.gpu_count);
    } 
	
	/* See if the number of requested OpenCL GPUs 
	 * (by option) does not exceed this number 
	 * Only if using GPU */
	
 	if (!CukrStricmp(CukrDBGetS(CUKR_DB_DOUBLE_PRECISION_BLAS),  "CLBLAS") ||
 		!CukrStricmp(CukrDBGetS(CUKR_DB_QDOUBLE_PRECISION_BLAS), "CLBLAS") ||
 		!CukrStricmp(CukrDBGetS(CUKR_DB_SINGLE_PRECISION_BLAS),  "CLBLAS") || 
 		!CukrStricmp(CukrDBGetS(CUKR_DB_INNER_BLAS),             "CLBLAS")   )
	    if (CukrDBGetI(CUKR_DB_PE) > cukr_db.gpu_count) {
	    	CUKR_FATALX("%d OpenCL devices requested when only %d were "\
	    	            "found in the system", 
	    	            CukrDBGetI(CUKR_DB_PE), cukr_db.gpu_count);
	    }
	    else {
	    	LOGX(INFO, "Using %d PEs", CukrDBGetI(CUKR_DB_PE));
	    } 
}
#endif /* CUKR_USE_OPENCL */

// ------------------------------------------------------------------- //

/**
 * \brief Initialize clBlas
 */
void CukrInit_clBlas()
#ifndef CUKR_USE_OPENCL
{
	CUKR_FATAL("OpenCL support was not compiled in the library");
}
#else
{
	//clblasStatus status;
	
	/* Init CLBLAS */
	//status = clblasInit();
	
	/* Check for errors */ 
	//if (status != CLBLAS_STATUS_SUCCESS)
	//	CUKR_FATAL("Error initializing CLBLAS. Possible causes are: \
	//			   - no OpenCL device \
	//			   - not -lOpenCL found");
}
#endif /* CUKR_USE_OPENCL */

// ------------------------------------------------------------------- //

/*
 * \brief Initialize the BLAS acording to options 
 */
void CukrInit_BLAS()
{	
	char buffer[255];
	
	/* 
	 * Declare and create BLAS
	 * and BLASX objects.
	 * 
	 * Declare and create objects
	 * for each different BLAS
	 * implementation. 
	 */
	
	/* Generics */
	CUKR_BLAS blas_generic;
	CukrBlasCreate(&blas_generic);
	blas_create_generic(blas_generic);
	
	CUKR_BLASX blasx_generic;
	CukrBlasxCreate(&blasx_generic);
	blasx_create_generic(blasx_generic);
	
	/* MKL */
	#ifdef CUKR_USE_MKL
		CUKR_BLAS blas_MKL;
		CukrBlasCreate(&blas_MKL);
		blas_create_mkl(blas_MKL);
	#endif	
	
	/* GPUBLAS */
	#ifdef CUKR_USE_GPU
		#ifdef CUKR_USE_OPENCL
		CUKR_BLAS blas_CLBLAS;
		CukrBlasCreate(&blas_CLBLAS);
		blas_create_gpu(blas_CLBLAS);
	
		CUKR_BLASX blasx_CLBLAS;
		CukrBlasxCreate(&blasx_CLBLAS);
		blasx_create_gpu(blasx_CLBLAS);
		#else
		CUKR_BLAS blas_GPUBLAS;
		CukrBlasCreate(&blas_GPUBLAS);
		blas_create_gpu(blas_GPUBLAS);
	
		CUKR_BLASX blasx_GPUBLAS;
		CukrBlasxCreate(&blasx_GPUBLAS);
		blasx_create_gpu(blasx_GPUBLAS);
		#endif
	#endif	
	
	
	/* 
	 * Create BLAS objects for each 
	 * precision.
	 * 
	 * Create here one different BLAS
	 * object for each different precisio
	 * or special instance used in the code
	 */
	 
	CukrBlasCreate(&cukr_db.blasD);
	CukrBlasCreate(&cukr_db.blasQ);
	CukrBlasCreate(&cukr_db.blasS);
	CukrBlasCreate(&cukr_db.blasI);
	
	/*
	 * Choose which implementation corresponds
	 * to each instance
	 */
	
	/* 
	 * Generics 
	 */
	
	/* For double */
	if (!CukrStricmp(blas_generic->name, CukrDBGetS(CUKR_DB_DOUBLE_PRECISION_BLAS))) {
		*cukr_db.blasD = *blas_generic;
		cukr_db.blasD->blasx = blasx_generic;
	}
	
	/* For qdouble */
	if (!CukrStricmp(blas_generic->name, CukrDBGetS(CUKR_DB_QDOUBLE_PRECISION_BLAS))) {
		*cukr_db.blasQ = *blas_generic;
		cukr_db.blasQ->blasx = blasx_generic;
	}
	
	/* For single */
	if (!CukrStricmp(blas_generic->name, CukrDBGetS(CUKR_DB_SINGLE_PRECISION_BLAS))) {
		*cukr_db.blasS = *blas_generic;
		cukr_db.blasS->blasx = blasx_generic;
	}
	
	/* For inner */
	if (!CukrStricmp(blas_generic->name, CukrDBGetS(CUKR_DB_INNER_BLAS))) {
		*cukr_db.blasI = *blas_generic;
		cukr_db.blasI->blasx = blasx_generic;	
	}
	
	/* 
	 * MKL 
	 */
	
	#ifdef CUKR_USE_MKL
	
		/* For double */
		if (!CukrStricmp(blas_MKL->name, CukrDBGetS(CUKR_DB_DOUBLE_PRECISION_BLAS))) {
			*cukr_db.blasD = *blas_MKL;
			cukr_db.blasD->blasx = blasx_generic;		
		}
		
		/* For single */
		if (!CukrStricmp(blas_MKL->name, CukrDBGetS(CUKR_DB_SINGLE_PRECISION_BLAS))) {
			*cukr_db.blasS = *blas_MKL;
			cukr_db.blasS->blasx = blasx_generic;	
		}
		
		/* For inner */
		if (!CukrStricmp(blas_MKL->name, CukrDBGetS(CUKR_DB_INNER_BLAS))) {
			*cukr_db.blasI = *blas_MKL;
			cukr_db.blasI->blasx = blasx_generic;	
		}
	#endif
	
	/*
	 * GPUBLAS
	 */
	
	#ifdef CUKR_USE_GPU
		#ifdef CUKR_USE_OPENCL
		/* For double */
		if (!CukrStricmp(blas_CLBLAS->name, CukrDBGetS(CUKR_DB_DOUBLE_PRECISION_BLAS))) {
			*cukr_db.blasD = *blas_CLBLAS;
			cukr_db.blasD->blasx = blasx_CLBLAS;		
		}
		
		/* For qdouble */
		if (!CukrStricmp(blas_CLBLAS->name, CukrDBGetS(CUKR_DB_QDOUBLE_PRECISION_BLAS))) {
			*cukr_db.blasQ = *blas_CLBLAS;
			cukr_db.blasQ->blasx = blasx_CLBLAS;		
		}
		
		/* For single */
		if (!CukrStricmp(blas_CLBLAS->name, CukrDBGetS(CUKR_DB_SINGLE_PRECISION_BLAS))) {
			*cukr_db.blasS = *blas_CLBLAS;
			cukr_db.blasS->blasx = blasx_CLBLAS;			
		}	 
		
		/* For inner */
		if (!CukrStricmp(blas_CLBLAS->name, CukrDBGetS(CUKR_DB_INNER_BLAS))) {
			*cukr_db.blasI = *blas_CLBLAS;
			cukr_db.blasI->blasx = blasx_CLBLAS;			
		}
		#else
		if (!CukrStricmp(blas_GPUBLAS->name, CukrDBGetS(CUKR_DB_DOUBLE_PRECISION_BLAS))) {
			*cukr_db.blasD = *blas_GPUBLAS;
			cukr_db.blasD->blasx = blasx_GPUBLAS;		
		}
	
		/* For qdouble */
		if (!CukrStricmp(blas_GPUBLAS->name, CukrDBGetS(CUKR_DB_QDOUBLE_PRECISION_BLAS))) {
			*cukr_db.blasQ = *blas_GPUBLAS;
			cukr_db.blasQ->blasx = blasx_GPUBLAS;		
		}
	
		/* For single */
		if (!CukrStricmp(blas_GPUBLAS->name, CukrDBGetS(CUKR_DB_SINGLE_PRECISION_BLAS))) {
			*cukr_db.blasS = *blas_GPUBLAS;
			cukr_db.blasS->blasx = blasx_GPUBLAS;			
		}	 
	
		/* For inner */
		if (!CukrStricmp(blas_GPUBLAS->name, CukrDBGetS(CUKR_DB_INNER_BLAS))) {
			*cukr_db.blasI = *blas_GPUBLAS;
			cukr_db.blasI->blasx = blasx_GPUBLAS;
		}
		#endif
	#endif
	
	/*
	 * Check for un-assigned BLASes (errors)
	 */
	 
	if (!CukrStricmp(cukr_db.blasD->name, ""))
		CUKR_FATALX("Unknown double-precision BLAS: %s", 
		            CukrDBGetS(CUKR_DB_DOUBLE_PRECISION_BLAS));
	
	if (!CukrStricmp(cukr_db.blasQ->name, ""))
		CUKR_FATALX("Unknown quasi-double-precision BLAS: %s", 
		            CukrDBGetS(CUKR_DB_QDOUBLE_PRECISION_BLAS));
	
	if (!CukrStricmp(cukr_db.blasS->name, ""))
		CUKR_FATALX("Unknown single-precision BLAS: %s", 
		            CukrDBGetS(CUKR_DB_SINGLE_PRECISION_BLAS));
	
	if (!CukrStricmp(cukr_db.blasI->name, ""))
		CUKR_FATALX("Unknown inner BLAS: %s", 
					CukrDBGetS(CUKR_DB_INNER_BLAS));
	 
 	/*
	 * Set the the corresponding precision
	 * for the objects above 
	 */
	 
	 /* Store the precision in the .precision member 
	 * for later validation */
	cukr_db.blasD->precision = CUKR_TYPE_DOUBLE; 
	cukr_db.blasQ->precision = CUKR_TYPE_QDOUBLE;
	cukr_db.blasS->precision = CUKR_TYPE_SINGLE;
	
	/* Set the precsion of the inner BLAS acording
	 * to solver options */  
	if (CukrDBGetI(CUKR_DB_SOLVER_ITREF)) 
		cukr_db.blasI->precision = CukrTypeStrToId(CukrDBGetS(CUKR_DB_SOLVER_INNER_PRECISION));
	else
		cukr_db.blasI->precision = CukrTypeStrToId(CukrDBGetS(CUKR_DB_SOLVER_PRECISION));
 
 
	/* 
	 * Report the BLASes that were chosen 
	 */
	 
	LOGX(INFO, "Double precision BLAS: %s", cukr_db.blasD->name); 	
	LOGX(INFO, "Quasi-double precision BLAS: %s", cukr_db.blasQ->name); 
	LOGX(INFO, "Single precision BLAS: %s", cukr_db.blasS->name); 
}

// ------------------------------------------------------------------- //

void keyDestructor(void *value) {;}

/**
 * \brief Create and start as many worker threads
 *        as requested PEs
 */
void CukrInit_Threading()
{
	// Create thread key and malloc one element for it
	pthread_key_create(&(cukr_db.thread_key), keyDestructor);
	int* val = (int*)malloc(sizeof(int));
	*val = 0;
	// Main thread is treaded as thread 0
	pthread_setspecific(cukr_db.thread_key, (void*)val);
	// Create threads
	CukrThreadsCreate(CukrDBGetI(CUKR_DB_PE));
}

// ------------------------------------------------------------------- //

/**
 * \brief Analyze which BLAS libraries have been
 *        chosen and export to cukr_compute_on_gpu
 *        if the gpu is actually used for computation
 *        or not. 
 */
void CukrInit_ComputesOnGpu()
{
#ifdef CUKR_USE_OPENCL
	if (!CukrStricmp("CLBLAS", CukrDBGetS(CUKR_DB_DOUBLE_PRECISION_BLAS))  ||
		!CukrStricmp("CLBLAS", CukrDBGetS(CUKR_DB_QDOUBLE_PRECISION_BLAS)) ||
		!CukrStricmp("CLBLAS", CukrDBGetS(CUKR_DB_SINGLE_PRECISION_BLAS))  ||   
		!CukrStricmp("CLBLAS", CukrDBGetS(CUKR_DB_INNER_BLAS))               )
		cukr_db.compute_on_gpu = CUKR_TRUE;
	else
		cukr_db.compute_on_gpu = CUKR_FALSE;
#else
	if (!CukrStricmp("GPUBLAS", CukrDBGetS(CUKR_DB_DOUBLE_PRECISION_BLAS))  ||
		!CukrStricmp("GPUBLAS", CukrDBGetS(CUKR_DB_QDOUBLE_PRECISION_BLAS)) ||
		!CukrStricmp("GPUBLAS", CukrDBGetS(CUKR_DB_SINGLE_PRECISION_BLAS))  ||   
		!CukrStricmp("GPUBLAS", CukrDBGetS(CUKR_DB_INNER_BLAS))               )
		cukr_db.compute_on_gpu = CUKR_TRUE;
	else
		cukr_db.compute_on_gpu = CUKR_FALSE;	
#endif
}
