/**
 * @file init/init.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Here we include all initialization
 * tasks that must be performed, including:
 * 	- creating the option library
 *  - creating and initialzing the BLAS
 *  - ...
 */

#ifndef INIT_H_
#define INIT_H_

#include "../db/db.h"

/**
 * \brief Does all the initialization stuff
 *        needed when using CUKr, but does
 *        not start the threading engine. 
 * \param argc, argv The program arguments
 * \param opts Additional options
 */
void CukrInit(int argc, char* argv[], int opts);

// ------------------------------------------------------------------- //

/**
 * \brief Start CUKr based on the options set
 *        at initialization and after
 * \param opts Additional options  
 */
void CukrStart(int opts);

// ------------------------------------------------------------------- //

/**
 * \brief Run the option parser and create
 *        the option database.
 * \param argc argument count received by the
 *        <main> function
 * \param argv pointer to the parameter list
 *        received by the <main> function
 * \param strict Strict syntax check (exit on
 *        unknown options)
 */
void CukrInit_OptionDB(int argc, char* argv[], int strict); 

// ------------------------------------------------------------------- //

/**
 * \brief Initialize the logging system
 */
void CukrInit_Logging(); 

// ------------------------------------------------------------------- //

/**
 * \brief Initializes the MPI subsystem
 * \param argc argument count received by the
 *        <main> function
 * \param argv pointer to the parameter list
 *        received by the <main> function
 */
void CukrInit_MPI(int argc, char* argv[]);

// ------------------------------------------------------------------- //

/**
 * \brief Do CUDA-related initializations
 */
#ifdef __cplusplus
extern "C"
#endif 
void CukrInit_Cuda();

// ------------------------------------------------------------------- //

/**
 * \brief Initialize CUBLAS
 */
void CukrInit_Cublas(); 

// ------------------------------------------------------------------- //

/**
 * \brief Initialize OpenCL
 */
void CukrInit_OpenCL(); 

// ------------------------------------------------------------------- //

/**
 * \brief Initialize clBlas
 */
void CukrInit_clBlas(); 

// ------------------------------------------------------------------- //

/**
 * \brief Initialize the BLAS acording to options 
 */
void CukrInit_BLAS();

// ------------------------------------------------------------------- //

/*
 * \brief Get partitioner type acording to options
 */
void CukrInit_Partitioner(); 

// ------------------------------------------------------------------- //

/**
 * \brief Create and start as many worker threads
 *        as requested GPUs
 */
void CukrInit_Threading();
 
// ------------------------------------------------------------------- // 
 
/**
 * \brief Analyze which BLAS libraries have been
 *        chosen and export to cukr_compute_on_gpu
 *        if the gpu is actually used for computation
 *        or not. 
 */
void CukrInit_ComputesOnGpu(); 

// ------------------------------------------------------------------- //

#endif /*INIT_H_*/
