/**
 * @file mat_vec/matrix.c
 * @author Serban Georgescu
 * @since 2008
 * 
 * Interface to the SP/MP/MPI matrix
 * functions
 */

#include "matrix.h"






				/*****************************************/
				//       CONSTRUCTOR / DESTRUCTOR        //
				/*****************************************/         






/**
 * \brief Creates a matrix of the specified
 *        format, on the CPU, with data of 
 *        the specified type
 * \param mat Pointer to the matrix structure 
 *        being created
 * \param rows,cols Matrix size
 * \param format Matrix format (e.g. CSR)
 * \param type Matrix data type 
 *        (float, qdouble, double)
 * \param name Matrix name 
 */

void CukrMatCreate(CUKR_MATRIX* mat, CUKR_INT rows, CUKR_INT cols,
				   int format, int type, char* name)
{
	#ifdef CUKR_MPI
		CukrMatmpiCreate(mat, rows, cols, format, type, name);
	#else
		#ifdef CUKR_MP
			CukrMatmpCreate(mat, rows, cols, format, type, name);	
		#else
			CukrMatspCreate(mat, rows, cols, format, type, name);
		#endif
	#endif			
}				   

// ------------------------------------------------------------------- //

/**
 * \brief Loads a matrix from file.
 * \param mat Pointer to the matrix in which
 *        data is being read
 * \param filename The file where the data is
 *        being read from
 * \param format The matrix storage format (CSR)
 * \param type The data type the values should
 *        be considered to be (double, qdouble,
 *        single).
 * \param name The name given to the matrix
 */
void CukrMatCreateFromFile(CUKR_MATRIX* mat, const char *filename, int format, int type, char* name)
{
	#ifdef CUKR_MPI
		CukrMatmpiCreateFromFile(mat, filename, format, type, name);
	#else
		#ifdef CUKR_MP
			CukrMatmpCreateFromFile(mat, filename, format, type, name);	
		#else
			CukrMatspCreateFromFile(mat, filename, format, type, name);
		#endif
	#endif	
}

// ------------------------------------------------------------------- //

/**
 * \brief Finalizes the creation of the matrix,
 *        after all elements have been added. This
 *        inlcudes the creation of data structures
 *        and building the comm. tables.
 * \param mat The matrix structure to assemble.
 */
void CukrMatAssemble(CUKR_MATRIX mat)
{
	#ifdef CUKR_MPI
		CukrMatmpiAssemble(mat);
	#else
		#ifdef CUKR_MP
			CukrMatmpAssemble(mat);
		#else
			CukrMatspAssemble(mat);
		#endif
	#endif	
}

// ------------------------------------------------------------------- //
                   
/**
 * \brief Frees all the memory used 
 *        by the matrix
 * \param mat The matrix to be destroyed 
 */
void CukrMatDestroy(CUKR_MATRIX* mat)
{
	#ifdef CUKR_MPI
		CukrMatmpiDestroy(mat);
	#else
		#ifdef CUKR_MP
			CukrMatmpDestroy(mat);
		#else
			CukrMatspDestroy(mat);
		#endif
	#endif
}
                 




				/*****************************************/
				//               GET/SET                 // 
				/*****************************************/ 





/**
 * \brief Adds an element to a matrix. This can only
 *        be used after the matrix has been created and
 *        before it is assembled.
 * \param mat The matrix where the element is to be added
 * \param row The row of the element
 * \param col The col of the element
 * \param val The value of the element
 */
void CukrMatAddElement(CUKR_MATRIX mat, int row, int col, CUKR_DOUBLE val)
{
	#ifdef CUKR_MPI
		CukrMatmpiAddElement(mat, row, col, val);
	#else
		#ifdef CUKR_MP
			CukrMatmpAddElement(mat, row, col, val);
		#else
			CukrMatspAddElement(mat, row, col, val);
		#endif
	#endif
}

// ------------------------------------------------------------------- //

/**
 * \brief Adds a row of elements to a matrix. 
 *        This can only be used after the
 *        matrix has been created and before
 *        it is assembled.
 * \param mat The matrix where the row 
 *        is to be added
 * \param row The row to be added
 * \param val Vector of values, having at least as  
 *        many elements as the length of the row.
 */
void CukrMatAddRow(CUKR_MATRIX mat, int row, CUKR_DOUBLE* val)
{
	#ifdef CUKR_MPI
		CukrMatmpiAddRow(mat, row, val);
	#else
		#ifdef CUKR_MP
			CukrMatmpAddRow(mat, row, val);
		#else
			CukrMatspAddRow(mat, row, val);
		#endif
	#endif
}

// ------------------------------------------------------------------- //

/**
 * \brief Adds a column of elements to a matrix. 
 *        This can only be used after the
 *        matrix has been created and before
 *        it is assembled.
 * \param mat The matrix where the column 
 *        is to be added
 * \param col The column to be added
 * \param val Vector of values, having at least as  
 *        many elements as the length of the column.
 */
void CukrMatAddCol(CUKR_MATRIX mat, int col, CUKR_DOUBLE* val)
{
	#ifdef CUKR_MPI
		CukrMatmpiAddCol(mat, col, val);
	#else
		#ifdef CUKR_MP
			CukrMatmpAddCol(mat, col, val);
		#else
			CukrMatspAddCol(mat, col, val);
		#endif
	#endif
}

// ------------------------------------------------------------------- //

/**
 * \brief Adds a diagonal of elements to a matrix. 
 *        This can only be used after the
 *        matrix has been created and before
 *        it is assembled.
 * \param mat The matrix where the diagonal 
 *        is to be added
 * \param diag The diagonal to be added. 0 is the
 *        main diagonal, -1 ... min(rows,cols) are
 *        the lower diagonals and 1 .. min(rows,cols)
 *        are the upper diagonals.
 * \param val Vector of values, having at least as many 
 *        elements as the length of the diagonal.
 */
void CukrMatAddDiag(CUKR_MATRIX mat, int diag, CUKR_DOUBLE* val)
{
	#ifdef CUKR_MPI
		CukrMatmpiAddDiag(mat, diag, val);
	#else
		#ifdef CUKR_MP
			CukrMatmpAddDiag(mat, diag, val);
		#else
			CukrMatspAddDiag(mat, diag, val);
		#endif
	#endif
}

// ------------------------------------------------------------------- //

/**
 * \brief Extracts the diagonal from a matrix
 * \param mat The matrix 
 * \param vec The vector where to store the diagonal
 * \param diag The diagonal to read (zero for the
 *        main diagonal)
 * \param invert Whether to invert it or not 
 */
void CukrMatGetDiag(CUKR_MATRIX mat, CUKR_VECTOR vec, int diag, int invert)
{
	#ifdef CUKR_MPI
		CukrMatmpiGetDiag(mat, vec, diag, invert);
	#else
		#ifdef CUKR_MP
			CukrMatmpGetDiag(mat, vec, diag, invert);
		#else
			CukrMatspGetDiag(mat, vec, diag, invert);
		#endif
	#endif
}







				/*****************************************/
				//          TRANSFER / CONVERT           // 
				/*****************************************/ 

  
  
  
  
  
/**
 * \brief Changes the location of a matrix
 *        by copying the data from one location
 *        to another (if not already there)
 * \param mat Pointer to the matrix 
 * \param dest_loc Destination matrix location (CPU,GPU)
 * \param dest_type Destination matrix value type 
 *        (float, double)
 */
void CukrMatTransfer(CUKR_MATRIX mat, int dest_loc, int dest_type)
{
	#ifdef CUKR_MPI
		CukrMatmpiTransfer(mat, dest_loc, dest_type);
	#else
		#ifdef CUKR_MP
			CukrMatmpTransfer(mat, dest_loc, dest_type);
		#else
			CukrMatspTransfer(mat, dest_loc, dest_type);
		#endif
	#endif
}                   


// ------------------------------------------------------------------- //


/**
 * \brief Changes the storage format of 
 *        the matrix to a diferent variation
 *        or to a completely different format
 *        (if not already in that format)
 * \param mat Pointer to the matrix 
 * \param dest_format Destination format
 * \param dest_variation Destination variation 
 */
void CukrMatConvert(CUKR_MATRIX mat, int dest_format, int dest_variation)
{
	#ifdef CUKR_MPI
		CukrMatmpiConvert(mat, dest_format, dest_variation);
	#else
		#ifdef CUKR_MP
			CukrMatmpConvert(mat, dest_format, dest_variation);
		#else
			CukrMatspConvert(mat, dest_format, dest_variation);
		#endif
	#endif	
}






				/*****************************************/
				//                 I/O                   // 
				/*****************************************/ 







/**
 * \brief Prints a matrix
 * \param mat The matrix being printed
 */
void CukrMatPrint(CUKR_MATRIX mat)
{
	#ifdef CUKR_MPI
		CukrMatmpiPrint(mat);
	#else
		#ifdef CUKR_MP
			CukrMatmpPrint(mat);
		#else
			CukrMatspPrint(mat);
		#endif
	#endif		
}
	
// ------------------------------------------------------------------- //

/**
 * \brief Prints a matrix to file
 * \param vec The matrix being printed
 * \param filename The full path of the file
 *        to write the matrix to
 */
void CukrMatPrintToFilename(CUKR_MATRIX mat, char* filename)
{
	#ifdef CUKR_MPI
		CukrMatmpiPrintToFilename(mat, filename);
	#else
		#ifdef CUKR_MP
			CukrMatmpPrintToFilename(mat, filename);
		#else
			CukrMatspPrintToFilename(mat, filename);
		#endif
	#endif	
}

// ------------------------------------------------------------------- //

/**
 * \brief Prints a matrix to a file stream
 * \param mat The matrix being printed
 * \param f An open file stream to write to
 */
void CukrMatPrintToFile(CUKR_MATRIX mat, FILE *f)
{
	#ifdef CUKR_MPI
		CukrMatmpiPrintToFile(mat, f);
	#else
		#ifdef CUKR_MP
			CukrMatmpPrintToFile(mat, f);
		#else
			CukrMatspPrintToFile(mat, f);
		#endif
	#endif	
}

// ------------------------------------------------------------------- //

/**
 * \brief Prints a matrix to a file stream,
 *        adding an indent. Used for nesting.
 * \param mat The matrix being printed
 * \param f An open file stream to write to
 * \param indent An indentation string, to be inserted
 *        before every print.
 */
void CukrMatPrintToFileWithIndent(CUKR_MATRIX mat, FILE *f, char* indent)
{
	#ifdef CUKR_MPI
		CukrMatmpiPrintToFileWithIndent(mat, f, indent);
	#else
		#ifdef CUKR_MP
			CukrMatmpPrintToFileWithIndent(mat, f, indent);
		#else
			CukrMatspPrintToFileWithIndent(mat, f, indent);
		#endif
	#endif
}

