/**
 * @file mat_vec/matrix.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Interface to the SP/MP/MPI matrix
 * functions
 */

#ifndef CUKR_MATRIX_H_
#define CUKR_MATRIX_H_

#include "../cukr.h"
#include "matrix_defs.h"

/* Decide which include file to use,
 * depending on defines set at built time.*/

#ifdef CUKR_MPI
	#include "mpi/matrix_mpi.h"
#else
	#ifdef CUKR_MP
		#include "mp/matrix_mp.h"
	#else
		#include "sp/matrix_sp.h"
	#endif
#endif
		
#include "matrix_gen.h"



				/*****************************************/
				//       CONSTRUCTOR / DESTRUCTOR        //
				/*****************************************/         






/**
 * \brief Creates a matrix of the specified
 *        format, on the CPU, with data of 
 *        the specified type
 * \param mat Pointer to the matrix structure 
 *        being created
 * \param rows,cols Matrix size
 * \param format Matrix format (e.g. CSR)
 * \param type Matrix data type 
 *        (float, qdouble, double)
 * \param name Matrix name 
 */

void CukrMatCreate(CUKR_MATRIX* mat, CUKR_INT rows, CUKR_INT cols,
				   int format, int type, char* name);

// ------------------------------------------------------------------- //

/**
 * \brief Loads a matrix from file.
 * \param mat Pointer to the matrix in which
 *        data is being read
 * \param filename The file where the data is
 *        being read from
 * \param format The matrix storage format (CSR)
 * \param type The data type the values should
 *        be considered to be (double, qdouble,
 *        single).
 * \param name The name given to the matrix
 */
void CukrMatCreateFromFile(CUKR_MATRIX* mat, const char *filename, int format, int type, char* name);

// ------------------------------------------------------------------- //

/**
 * \brief Finalizes the creation of the matrix,
 *        after all elements have been added. This
 *        inlcudes the creation of data structures
 *        and building the comm. tables.
 * \param mat The matrix structure to assemble.
 */
void CukrMatAssemble(CUKR_MATRIX mat); 

// ------------------------------------------------------------------- //
                   
/**
 * \brief Frees all the memory used 
 *        by the matrix
 * \param mat The matrix to be destroyed 
 */
void CukrMatDestroy(CUKR_MATRIX* mat);
                 




				/*****************************************/
				//               GET/SET                 // 
				/*****************************************/ 





/**
 * \brief Adds an element to a matrix. This can only
 *        be used after the matrix has been created and
 *        before it is assembled.
 * \param mat The matrix where the element is to be added
 * \param row The row of the element
 * \param col The col of the element
 * \param val The value of the element
 */
void CukrMatAddElement(CUKR_MATRIX mat, int row, int col, CUKR_DOUBLE val);

// ------------------------------------------------------------------- //

/**
 * \brief Adds a row of elements to a matrix. 
 *        This can only be used after the
 *        matrix has been created and before
 *        it is assembled.
 * \param mat The matrix where the row 
 *        is to be added
 * \param row The row to be added
 * \param val Vector of values, having at least as  
 *        many elements as the length of the row.
 */
void CukrMatAddRow(CUKR_MATRIX mat, int row, CUKR_DOUBLE* val);

// ------------------------------------------------------------------- //

/**
 * \brief Adds a column of elements to a matrix. 
 *        This can only be used after the
 *        matrix has been created and before
 *        it is assembled.
 * \param mat The matrix where the column 
 *        is to be added
 * \param col The column to be added
 * \param val Vector of values, having at least as  
 *        many elements as the length of the column.
 */
void CukrMatAddCol(CUKR_MATRIX mat, int col, CUKR_DOUBLE* val);

// ------------------------------------------------------------------- //

/**
 * \brief Adds a diagonal of elements to a matrix. 
 *        This can only be used after the
 *        matrix has been created and before
 *        it is assembled.
 * \param mat The matrix where the diagonal 
 *        is to be added
 * \param diag The diagonal to be added. 0 is the
 *        main diagonal, -1 ... min(rows,cols) are
 *        the lower diagonals and 1 .. min(rows,cols)
 *        are the upper diagonals.
 * \param val Vector of values, having at least as many 
 *        elements as the length of the diagonal.
 */
void CukrMatAddDiag(CUKR_MATRIX mat, int diag, CUKR_DOUBLE* val);

// ------------------------------------------------------------------- //

/**
 * \brief Extracts the diagonal from a matrix
 * \param mat The matrix 
 * \param vec The vector where to store the diagonal
 * \param diag The diagonal to read (zero for the
 *        main diagonal)
 * \param invert Whether to invert it or not 
 */
void CukrMatGetDiag(CUKR_MATRIX mat, CUKR_VECTOR vec, int diag, int invert);







				/*****************************************/
				//          TRANSFER / CONVERT           // 
				/*****************************************/ 

  
  
  
  
  
/**
 * \brief Changes the location of a matrix
 *        by copying the data from one location
 *        to another (if not already there)
 * \param mat Pointer to the matrix 
 * \param dest_loc Destination matrix location (CPU,GPU)
 * \param dest_type Destination matrix value type 
 *        (float, double)
 */
void CukrMatTransfer(CUKR_MATRIX mat, int dest_loc, int dest_type);                   


// ------------------------------------------------------------------- //


/**
 * \brief Changes the storage format of 
 *        the matrix to a diferent variation
 *        or to a completely different format
 *        (if not already in that format)
 * \param mat Pointer to the matrix 
 * \param dest_format Destination format
 * \param dest_variation Destination variation 
 */
void CukrMatConvert(CUKR_MATRIX mat, int dest_format, int dest_variation);






				/*****************************************/
				//                 I/O                   // 
				/*****************************************/ 







/**
 * \brief Prints a matrix
 * \param mat The matrix being printed
 */
void CukrMatPrint(CUKR_MATRIX mat);
	
// ------------------------------------------------------------------- //

/**
 * \brief Prints a matrix to file
 * \param vec The matrix being printed
 * \param filename The full path of the file
 *        to write the matrix to
 */
void CukrMatPrintToFilename(CUKR_MATRIX mat, char* filename);

// ------------------------------------------------------------------- //

/**
 * \brief Prints a matrix to a file stream
 * \param mat The matrix being printed
 * \param f An open file stream to write to
 */
void CukrMatPrintToFile(CUKR_MATRIX mat, FILE *f);

// ------------------------------------------------------------------- //

/**
 * \brief Prints a matrix to a file stream,
 *        adding an indent. Used for nesting.
 * \param mat The matrix being printed
 * \param f An open file stream to write to
 * \param indent An indentation string, to be inserted
 *        before every print.
 */
void CukrMatPrintToFileWithIndent(CUKR_MATRIX mat, FILE *f, char* indent); 


#endif /*CUKR_MATRIX_H_*/
