/**
 * @file mat_vec/matrix_gen.c
 * @author Serban Georgescu
 * @since 2008
 * 
 * Interface to the SP/MP/MPI matrix
 * generating functions
 */

#include "matrix_gen.h"
#include <stdlib.h>
#include "../util/log.h"

/**
 * \brief Generates a diagonally-dominated sparse
 *        random matrix. The matrix is made from
 *        <band_group_size> groups of bands, separated
 *        by <space_group_size> empty bands. There are
 *        a total of group_count groups.  
 * \param mat Pointer to the matrix 
 * \param n Matrix size
 * \param group_count Number of band groups (symmetric)
 * \param band_group_size How many bands in a group
 * \param space_group_size How many spaces are used
 *        as separators.
 * \param format Matrix format (CSR) 
 * \param type Matrix type (float or default) 
 * \param name Matrix name
 */
void CukrMatGenBanded(CUKR_MATRIX* mat, int n, int group_count, int band_group_size, 
                      int space_group_size, int format, int type, char* name)                       
{
	int d, g, b, i, j;
	CUKR_DOUBLE val, *vals;
	
	/* Create the matrix */
	CukrMatCreate(mat, n, n, format, type, name);
	
	/* Create a vector of vals */
	if ( (vals = (CUKR_DOUBLE*)malloc(n * sizeof(CUKR_DOUBLE))) == NULL) 
		CUKR_FATALX("Could not allocate memory for value "\
		            "vector while generating banded matrix %s",
		            name);
	
	/* Set the first diagonal position 
	 * to zero */
	d = 0; 
	
	/* To make it diagonally domitaned,
	 * start with max on the main diagonal
	 * and decrease with each new one */
	val = band_group_size * group_count * 2 + 1;
	
	/* For each group */
	for (g = 0; g < group_count; g++) {
		
		/* Add diagonals */
		for (i = 0; i < band_group_size; i++) {
			int band = d + i;
			/* If zero, add only one */
			if (!band) {
				for (j = 0; j < n; j++) vals[j] = val;				
				CukrMatAddDiag(*mat, 0, vals);
				val--;
			}
			/* If not, add two */
			else {
				for (j = 0; j < n; j++) vals[j] = -val;				
				CukrMatAddDiag(*mat,  band, vals);
				for (j = 0; j < n; j++) vals[j] = -val;				
				CukrMatAddDiag(*mat, -band, vals);
				val -= 2;
			}
		}
		
		/* Add spaces */
		d += group_count + space_group_size;		
	}	
	
	/* Cleanup */
	free(vals);	
}                        

