/**
 * @file mat_vec/mp/csr/csr_matrix_mp.c
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains all the functions needed
 * to support a MP CSR matrix
 */

#define LOGGER_NAME "CUKr.mat_vec.mp.csr.csr_matrix_mp"

#include <stdio.h>
#include <stdlib.h>
#include "../../../util/log.h"
#include "../../../util/containers.h"
#include "../../../db/db.h"
#include "csr_matrix_mp.h"

#define MIN(a,b) (a < b ? a : b)

/* Get access to the option database */
extern CukrDB cukr_db;


//============================ PUBLIC FUNCTIONS =========================== //

/**
 * \brief Builds toe communication table for
 *        for the matrix. Has to be called
 *        when data IS PRESENT.
 * \param mat The multi-gpu matrix
 */
void CukrCsrMatmpBuildCommtable(CUKR_MATRIX_MP mat)
{
	int p, i, j, src;
	
	LOGX(INFO, "Building communication table for "\
	           "MP matrix %s", mat->name);
	
	/* 
	 * Analyze idx and find for each part
	 * the elements which would be sent/received
	 * to/by each part
	 */
	 
	/* Create a container for each PE, in which
	 * to add what is needed and what is given */
	 
	 CUKR_CONTAINER_VECTOR_INT  need[CUKR_MAX_PE];
	 CUKR_CONTAINER_VECTOR_INT* give[CUKR_MAX_PE];
	 
	 /* The give containers are only shortucts
	  * to the raw_give_container */
	 
	 for (i = 0; i < mat->nparts; i++) {
	 	CukrContainersVectorOfIntCreate(&need[i]);
	 	give[i] = &(mat->comm_table.raw_give_container[i]);
	 	CukrContainersVectorOfIntCreate(give[i]);
	 }
	
	 LOG(DEBUG, "Analyizing idx and filling containers ... ");
		 
	
	 /* Analyze idx and fill the containers */ 
	 int part_size = mat->part[0]->rows;
	 for (p = 0; p < mat->nparts; p++)
	 {	 	
	 	LOGX(DEBUG, "Processing part %d ... ", p);
	 	for (i = 0; i < mat->part[p]->nz; i++) 
	 	{		 		
	 		/* Get the column of the nonzero element and
	 		 * skip of is -1, since that is padding */
	 			 		
	 		int col = mat->part[p]->csr_mat->idx->ph_ival[i] - 1;
	 		if (col == -1) continue;
	 		
	 		/* 
	 		 * Find the part the data belongs to 
	 		 */
	 		
	 		/* As a first step, translate the column
	 		 * so that it becames local to the part, 
	 		 * if this MP matrix is part of a larger
	 		 * MPI matrix */
	 		src = col -  mat->part_index;
	 			 		
	 		//if (cukr_db.mpi_rank == 3)	 		
	 		//	printf("col = %d, src = %d, idx = %d\n",col, src, mat->part_index); 	 		
	 			 		
	 		/* If the translated value is negative, 
	 		 * or larger than n, no need to compute
	 		 * furter. We set it only as a need,
	 		 * since we don't know whom to give it to.
	 		 * The give will be set outside, when merging
	 		 * the MPI commtable with the MP commtable */
	 		if (src < 0 || src >= mat->rows) {  
	 			CukrContainersVectorOfIntAppend(need[p], col);
	 		}
	 		
	 		/* If not, than it belongs to this node, 
	 		 * so we must compute further to find 
	 		 * to which PE it belongs */	 		
	 		else 
	 		{ 
		 		src = MIN(src / part_size, mat->nparts - 1);		
		 				 	
		 		/* If not in the current part, add it to lists */
		 		if (src != p) {
		 			/* Need is always OK, because is global (one
		 			 * always knows that it needs something, 
		 			 * regardles of the place that things comes
		 			 * from) */
		 			CukrContainersVectorOfIntAppend(need[p], col);
		 			/* Give only applies to the current node, since
		 			 * there is no way of knowing what other nodes
		 			 * might need */
		 			CukrContainersVectorOfIntAppend(*give[src], col);
		 		}
	 		}
	  	}
	 }
	 LOG(DEBUG, "Sorting containers and removing duplicates ... ");
	 
	 /* Sort vectors */
	 for (p = 0; p < mat->nparts; p++) {
	 	CukrContainersVectorOfIntSortUnique(need[p]);
	 	CukrContainersVectorOfIntSortUnique(*give[p]);
	 }
	 
	 /*
	 for (p = 0; p < mat->nparts; p++) {
	 	printf("\n ------ NEED [%d] ----------- \n", p);
	 	CukrContainersVectorOfIntPrint(need[p]);
	 	printf("\n ------ GIVE [%d] ----------- \n", p);
	 	CukrContainersVectorOfIntPrint(give[p]);
	 } 
	 */
	 
	 /*
	  * To optimize transfer, we have to minimze the 
	  * number of transfers. This is done by grouping
	  * consecutive elements into vectors, and
	  * transfering a vector at a time.
	  * A small number of non needed elements can be
	  * added in order to merge smaller vectors into
	  * larger ones
	  */
	  
	  LOG(DEBUG, "Optimize need containers ... ");
	  
	  /* Process <need> containers and see if there
	   * is anything to add */ 
	  for (p = 0; p < mat->nparts; p++) 
	  	for (i = 0; i < need[p]->size - 1; i++)
	  	{
	  		/* Make a bridge if the distance is
	  		 * smaller than CUKR_MP_BRIDGE_SIZE */
	  		if (need[p]->elements[i + 1] - need[p]->elements[i] <= CUKR_MP_BRIDGE_SIZE) 
	  			for (j = need[p]->elements[i] + 1; j < need[p]->elements[i + 1]; j++)
	  				CukrContainersVectorOfIntAppendU(need[p], j);
	  	}
	  
	  LOG(DEBUG, "Optimize give containers ... ");
	  	
	  /* Process <give> containers and see if there
	   * is anything to add */ 
	  for (p = 0; p < mat->nparts; p++) 
	  	for (i = 0; i < (*give[p])->size - 1; i++)
	  	{ 
	  		/* Make a bridge if the distance is
	  		 * smaller than CUKR_MP_BRIDGE_SIZE */
	  		if ((*give[p])->elements[i + 1] - (*give[p])->elements[i] <= CUKR_MP_BRIDGE_SIZE) 
	  			for (j = (*give[p])->elements[i] + 1; j < (*give[p])->elements[i + 1]; j++)
	  				CukrContainersVectorOfIntAppendU(*give[p], j);
	  	}	
	 
	 LOG(DEBUG, "Re-sorting containers and removing duplicates ... ");
	 
	 /* Re-sort vectors */
	 for (p = 0; p < mat->nparts; p++) {
	 	CukrContainersVectorOfIntSortI(need[p]);
	 	CukrContainersVectorOfIntSortI(*give[p]);
	 }
	 
	 /*
	 for (p = 0; p < mat->nparts; p++) {
	 	printf("\n ------ NEED [%d] ----------- \n", p);
	 	CukrContainersVectorOfIntPrint(need[p]);
	 	printf("\n ------ GIVE [%d] ----------- \n", p);
	 	CukrContainersVectorOfIntPrint(give[p]);
	 }
	 */
	 
	 
	 /*
	  * Now that all the elements are present
	  * we can can group consecutive elements
	  * into comm-entries
	  */
	  
	  /* Init the comm-table containers for 
	   * holding comm-entries */
	  for (p = 0; p < mat->nparts; p++) {
	 	CukrContainersVectorOfCommentryCreate(&mat->comm_table.need_container[p]);
	 	CukrContainersVectorOfCommentryCreate(&mat->comm_table.give_container[p]);
	  } 
	  
	  	
	  LOG(DEBUG, "Build commtable ... ");
	  	
	  /* Populate the need containers */
	  CUKR_COMMENTRY comm;
	  for (p = 0; p < mat->nparts; p++) {
	  		
	  	 /* If nothing to add, break */
	  	 if (!need[p]->size) continue;	
	  		
	  	 /* Start the comm-entry from this position 
	  	  * and with size = 1*/
	  	 comm.pos = need[p]->elements[0]; 
	  	 comm.size = 1; 
	  	 
	  	 for (i = 1; i < need[p]->size; i++) 
	  	 {
	  	 	/* If the following element is consecutive,
	  	 	 * increase the size */
	  	 	if (need[p]->elements[i] - need[p]->elements[i-1] == 1)
	  	 		comm.size++;
	  	 	
	  	 	/* If not, add the comm-entry to the container
	  	 	 * and re-initialize the entry */
	  	 	else {
	  	 		CukrContainersVectorOfCommentryAppend(mat->comm_table.need_container[p], comm);
	  	 		comm.pos = need[p]->elements[i];
	  	 		comm.size = 1;
	  	 	} 	
	  	 }
	  	 
	  	 /* Add the last comm-entry to the container */
	  	 CukrContainersVectorOfCommentryAppend(mat->comm_table.need_container[p], comm);
	  }
	  
	  /* Populate the give containers */
	  for (p = 0; p < mat->nparts; p++) {
	  	
	  	/* If nothing to add, break */
	  	 if (!(*give[p])->size) continue;	
	  	
	  	 /* Start the comm-entry from this position 
	  	  * and with size = 1*/
	  	 comm.pos = (*give[p])->elements[0]; 
	  	 comm.size = 1; 
	  	 
	  	 for (i = 1; i < (*give[p])->size; i++) 
	  	 {
	  	 	/* If the following element is consecutive,
	  	 	 * increase the size */
	  	 	if ((*give[p])->elements[i] - (*give[p])->elements[i-1] == 1)
	  	 		comm.size++;
	  	 	
	  	 	/* If not, add the comm-entry to the container
	  	 	 * and re-initialize the entry */
	  	 	else {
	  	 		CukrContainersVectorOfCommentryAppend(mat->comm_table.give_container[p], comm);
	  	 		comm.pos = (*give[p])->elements[i];
	  	 		comm.size = 1;
	  	 	} 	
	  	 }
	  	 
	  	 /* Add the last comm-entry to the container */
	  	 CukrContainersVectorOfCommentryAppend(mat->comm_table.give_container[p], comm);
	  } 
	 
	/* 
	 for (p = 0; p < mat->nparts; p++) {
	 	printf("\n -------- NEED CONTAINER [%d] ----------- \n", p);
	 	CukrContainersVectorOfCommentryPrint(mat->comm_table.need_container[p]);
	 	printf("\n -------- GIVE CONTAINER [%d] ----------- \n", p);
	 	CukrContainersVectorOfCommentryPrint(mat->comm_table.give_container[p]);
	 }
	 */
	 
	 LOG(DEBUG, "Cleanup ... ");
	 
	 /* 
	  * Cleanup
	  */
	   
	 //for (i = 0; i < mat->nparts; i++) {
	 //	printf("PE%d: Needs %d tranfsers, gives %d transfers \n", i, mat->comm_table.need_container[i]->size, mat->comm_table.give_container[i]->size);	 		 		
	 //}
	 
	 /* Desroy only need containers. */
	 for (i = 0; i < mat->nparts; i++) {
	 	CukrContainersVectorOfIntDestroy(&need[i]); 
	 }
	  
	 
	 LOGX(INFO, "Done building communication table for "\
	            "MP matrix %s", mat->name);
}

// ------------------------------------------------------------------- //

/**
 * \brief Permutes the columns of a CSR matrix
 *        according to the permutation defined
 *        in the permutation vector.
 * \param mat The matrix whose colums are to be
 *        permuted
 * \param permv Permutation vector (e.g. i -> permv[i])
 */
void CukrMatmpCsrPermuteCol(CUKR_MATRIX_MP mat, int* permv)
{
	int i, j, p;
	
	LOGX(DEBUG, "Permuting colums for CSR matrix %s",
	            mat->name);
	
	/*
	 * Make sure matrix is in double, on CPU 
	 */
	 
	CukrMatmpTransfer(mat, CUKR_LOCATION_CPU, CUKR_TYPE_DOUBLE);
	
	/*
	 * PTR does not have to be modified 
	 * since the number of nz./row does
	 * not change 
	 */
	 
	 /*
	  * Allocate temp. vector of size col
	  */
	  
	 CUKR_DOUBLE *temp;
	 if ( (temp = (CUKR_DOUBLE*)malloc(mat->cols * sizeof(CUKR_DOUBLE))) == NULL) {
	 	 CUKR_FATALX("Error allocating %d ints for temp vector",
	 	              mat->cols);
	 }
	  
	 /*
	  * Process IDX and VAL
	  */
	   
	  for (p = 0; p < mat->nparts; p++)
	  {
	  	  /* Get current part offset in 
	  	   * permutation vector */	   	   
	  	  int part_offset = (p == 0) ? 0 : mat->part[p - 1]->rows;
	  	
	  	  /* Make shortcuts */
	  	  CUKR_INT*    ptr = mat->part[p]->csr_mat->ptr->ph_ival;
	  	  CUKR_INT*    idx = mat->part[p]->csr_mat->idx->ph_ival;
	  	  CUKR_DOUBLE* val = mat->part[p]->csr_mat->val->ph_dval;
	  	  
	  	  /* Process */
	  	  for (i = 0; i < mat->part[p]->rows; i++) 
	  	  {
	  	  	
	  	  	/* 
	  	  	 * Update IDX 
	  	  	 */
	  	  	 	
	  	  	for (j = ptr[i]; j < ptr[i+1]; j++)
	  	  	{	  	  			  	  		
	  	  		/* Do the permutation (apply the map) */
	  	  		idx[j - 1] = permv[idx[j - 1] - 1] + 1;
	  	  		
	  	  		/* Add corresponding VAL element to 
	  	  		 * temp vector */
	  	  		temp[idx[j-1] - 1] = val[j - 1]; 
	  	  	}	
	  	  	
	  	  	/*
	  	  	 * Update VAL, from the temp vector
	  	  	 */		
	  	  	
	  	  	for (j = ptr[i]; j < ptr[i+1]; j++)
	  	  	{	   	  		
	  	  		/* Get the values from temp */
	  	  		val[j - 1] = temp[idx[j - 1] - 1]; 
	  	  	}	 
	  	  	 
	  	  }
	  } 
	  
	  LOGX(DEBUG, "Done permuting colums for CSR matrix %s",
	              mat->name);
	 
}
// ------------------------------------------------------------------- //

/**
 * \brief Generates a sparse multi-gpu random matrix, 
 *        filled with random elements, in CSR format 
 * \param mat Pointer to the multi-gpu matrix 
 * \param halfb The half-bandwidth of the matrix
 * \param nz_per_row No. of nonzeros per row
 * \param min,max The lower and upper bound for
 *                matrix values
 * \param seed The seed to use for the random no. generator
 */
void CukrCsrMatmpCreateRandom(CUKR_MATRIX_MP* mat, int halfb, int nz_per_row, 
						    CUKR_DOUBLE min, CUKR_DOUBLE max, CUKR_DOUBLE seed);

// ------------------------------------------------------------------- //
					 
/**
 * \brief Generates a 2D Poisson multi-gpu
 *        matrix 
 * \param mat Pointer to the matrix 
 * \param n Order
 * \param name Matrix name
 */
void CukrCsrMatmpCreatePoisson2D(CUKR_MATRIX_MP* mat, int n, char* name);



