/**
 * @file mat_vec/mp/csr/csr_matrix_mp.h
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains all the functions needed
 * to support a MP CSR matrix
 */

#ifndef CUKR_CSR_MATRIX_MP_H_
#define CUKR_CSR_MATRIX_MP_H_

#include "../../sp/csr/csr_matrix_sp.h"
#include "../matrix_mp.h"

/**
 * \brief Loads a CSR matrix from a CSR file.
 *        The CSR file format is the following:
 *        1st_row: m n nz
 *        2nd->(n+1)th_row: ptr
 *        (n+1)th -> (n+nz+1)th_row: idx
 *        (n+nz+1)th -> (n+2*nz+1)th_row: val
 * \param mat Pointer to the matrix in which
 *        data is being read
 * \param filename The file where the data is
 *        being read from
 * \param type The data type the values should
 *        be considered to be (double, single,
 *        short)
 */
void CukrCsrMatmpCreateFromCsrFile(CUKR_MATRIX_MP mat, const char *filename, int type);

// ------------------------------------------------------------------- //

/**
 * \brief Builds toe communication table for
 *        for the matrix. Has to be called
 *        when data IS PRESENT.
 * \param mat The multi-gpu matrix
 */
void CukrCsrMatmpBuildCommtable(CUKR_MATRIX_MP mat);

// ------------------------------------------------------------------- //

/**
 * \brief Extracts the diagonal from a 
 *        multi-gpu matrix
 * \param mat The multi-gpu matrix 
 * \param vec The multi-gpu vector where to 
 *        store the diagonal
 * \param invert Whether to invert it or not 
 */
void CukrCsrMatmpGetDiag(CUKR_MATRIX_MP mat, CUKR_VECTOR_MP vec, int invert);

// ------------------------------------------------------------------- //

/**
 * \brief Permutes the columns of a CSR matrix
 *        according to the permutation defined
 *        in the permutation vector.
 * \param mat The matrix whose colums are to be
 *        permuted
 * \param permv Permutation vector (e.g. i -> permv[i])
 */
void CukrMatmpCsrPermuteCol(CUKR_MATRIX_MP mat, int* permv);

// ------------------------------------------------------------------- //

/**
 * \brief Generates a sparse multi-gpu random matrix, 
 *        filled with random elements, in CSR format 
 * \param mat Pointer to the multi-gpu matrix 
 * \param halfb The half-bandwidth of the matrix
 * \param nz_per_row No. of nonzeros per row
 * \param min,max The lower and upper bound for
 *                matrix values
 * \param seed The seed for the random no. generator
 */
void CukrCsrMatmpCreateRandom(CUKR_MATRIX_MP* mat, int halfb, int nz_per_row, 
						      CUKR_DOUBLE min, CUKR_DOUBLE max, CUKR_DOUBLE seed);

// ------------------------------------------------------------------- //
					 
/**
 * \brief Generates a 2D Poisson multi-gpu
 *        matrix 
 * \param mat Pointer to the matrix 
 * \param n Order
 * \param name Matrix name
 */
void CukrCsrMatmpCreatePoisson2D(CUKR_MATRIX_MP* mat, int n, char* name);

// ------------------------------------------------------------------- //


/**
 * \brief Prints the contents of the CSR
 *        part of a multi-gpu matrix
 * \param mat Pointer to the matrix being
 *        printed
 */
void CukrCsrMatmpPrint(CUKR_MATRIX_MP mat);

// ------------------------------------------------------------------- //


#endif /*CUKR_CSR_MATRIX_MP_H_*/
