/**
 * @file mat_vec/mp/csr/csr_matrix_gen_mp.c
 * @author Serban Georgescu
 * @since 2008
 *
 * CSR matrix generation routines
 */

#define LOGGER_NAME "CUKr.mat_vec.mp.csr.csr_matrix_mp_gen"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <malloc.h>
#include "csr_matrix_mp.h"
#include "../../matrix_defs.h"
#include "../../../util/log.h" 
#include "../../../db/db.h" 

#define MAX(a,b) (a > b ? a : b)
#define MIN(a,b) (a < b ? a : b)
#define ROUND(a) ((int)(a + 0.5) == (int)a ? (int)a : (int)(a+1)) 

/************ Forward declarations *******************/

void gen_unique_int_rnd_mp(int *x, int *w, int n, int lo, int up);

/*****************************************************/


// ------------------------------------------------------------------- //

/**
 * \brief Generates a sparse multi-gpu random matrix, 
 *        filled with random elements, in CSR format 
 * \param mat Pointer to the multi-gpu matrix 
 * \param halfb The half-bandwidth of the matrix
 * \param nz_per_row No. of nonzeros per row
 * \param min,max The lower and upper bound for
 *                matrix values
 * \param seed The seed for the random no. generator
 */
void CukrCsrMatmpCreateRandom(CUKR_MATRIX_MP* mat, int halfb, int nz_per_row, 
						      CUKR_DOUBLE min, CUKR_DOUBLE max, CUKR_DOUBLE seed)
{
	int i, p, irow, icol;
	
	/* 
	 * Validate params 
	 */
	 
	/* Positive halfb */
	if (halfb < 0) {
		CUKR_FATALX("The half-bandwidh specified for matrix %s "\
					" must be non-negative", (*mat)->name);			
	}
	
	/* Strictly positive nz_per_row */
	if (nz_per_row <= 0) {
		CUKR_FATALX("The nonzeros/row specified for matrix %s "\
					" must be strictly positive", (*mat)->name);			
	}
	 
	/* nz_per_row <=  2*halfb + 1, more elements
	 * won't fit the space */
	if (nz_per_row > 2*halfb + 1) {
		CUKR_FATALX("The of %d half-bandwidh specified for matrix "\
					"%s allows for a maximum of %d nonzeros/row "\
					"(but %d were requested)", halfb, (*mat)->name,  
	 				2*halfb + 1, nz_per_row);			
	}
	 
	LOGX(DEBUG, "Filling multi-gpu CSR matrix %s with random vals. "\
	            "between %g and %g, with a half-bandwidth "\
	            "of %d and %d nonzeros/row", (*mat)->name, 
	            min, max, halfb, nz_per_row);
	
	/*
	 * Create the parts
	 */
 	 
	/* Compute the row count for the parts */
	int part_size, last_part_size;
	if ((*mat)->nparts == 1) 
		last_part_size = (*mat)->rows;
	else {
	 	part_size = (*mat)->rows / (*mat)->nparts; 
		last_part_size = (*mat)->rows - ((*mat)->nparts - 1) * part_size;
	}
	
	/* Put the row counts in a vector */
	int part_rows[CUKR_MAX_PE], part_nz[CUKR_MAX_PE]; 
	for (i = 0; i < (*mat)->nparts; i++) {
		if (i < (*mat)->nparts - 1) 
			part_rows[i] = part_size;
		else
			part_rows[i] = last_part_size;
		
		part_nz[i] = part_rows[i] * nz_per_row;	
	} 
	
	 /* Create the parts */
/*	 char buffer[255];
	 for (i = 0; i < (*mat)->nparts; i++) {
	 	sprintf(buffer, "%s.part%d", (*mat)->name, i);
	 	CukrMatspCreate(&((*mat)->part[i]), part_rows[i], (*mat)->cols, part_nz[i], CUKR_MATRIX_FORMAT_CSR,
	 	              CUKR_MATRIX_FORMAT_VARIATION_NONE, CUKR_TYPE_DOUBLE, buffer);
	 }     
	*/
	
	
	/* 
	 * Fill with data  
	 */
	
	/* Init random no. generator */
	srand(seed);
	
	/* Malloc for workspace */
	int *w;
	if ( (w = (int*)malloc((*mat)->rows * sizeof(int))) == NULL) {
		CUKR_FATAL("Cannot alloc memory for worspace vector");
	} 
	
	/* Fill ptr and val 
	 * Filling all values for ALL parts must be
	 * finished BEFORE generating the idx, since
	 * for generating idx we use the random no. 
	 * generator as well. 
	 * If all values for all parts are generated
	 * before, then for the same seed we will 
	 * always have the same values, regardless
	 * of the splitting.
	 * */
	
	for (p = 0; p < (*mat)->nparts; p++)
	{
		/* Make shortcuts for easy reading */
		CUKR_INT* ptr = (*mat)->part[p]->csr_mat->ptr->ph_ival;
		CUKR_INT* idx = (*mat)->part[p]->csr_mat->idx->ph_ival;
		CUKR_DOUBLE* dval = (*mat)->part[p]->csr_mat->val->ph_dval;
		
		/* Generate ptr 
	 	 * Simple, since we have fixed nonzeros per
	 	 * row, so there is a fixed increment */
		ptr[0] = 1; 
		for (i = 1; i <= part_rows[p]; i++)
		 	ptr[i] = ptr[i - 1] + nz_per_row;
		 	
		/* Generate val
	 	 * Simple, since we have fixed nonzeros per 
	 	 * row, just generate n*nz random numbers */
		for (i = 0; i < part_nz[p]; i++) {
			CUKR_DOUBLE rnd = min + ((CUKR_DOUBLE)rand() / (CUKR_DOUBLE)RAND_MAX) * (max - min);
			dval[i] = rnd;
		} 	
	} 
	
	/* Fill idx */
	int rows_until_now = 0;
	for (p = 0; p < (*mat)->nparts; p++)
	{
		/* Make shortcuts for easy reading */
		CUKR_INT* ptr = (*mat)->part[p]->csr_mat->ptr->ph_ival;
		CUKR_INT* idx = (*mat)->part[p]->csr_mat->idx->ph_ival;
		CUKR_DOUBLE* dval = (*mat)->part[p]->csr_mat->val->ph_dval;
		
		/* Generate idx
		 * A little more complicated. Have to generate
		 * unique random numbers in [row-halfb,row+halfb] 
		 * The idx has to be shifted for each row */
		 
		for (irow = 0; irow < part_rows[p]; irow++) {
			
			/* See the bounds for the generator */
			int l = MAX(1           , rows_until_now + irow - halfb + 1);
			int u = MIN((*mat)->rows, rows_until_now + irow + halfb + 1);
			
			
			/* Generate (unorderd) unique numbers */
			if (l != u)
				gen_unique_int_rnd_mp(&idx[ptr[irow] - 1], w, nz_per_row, l, u);
			else {
				idx[ptr[irow] - 1] = l;
				continue;
			} 
			
			/* Sort row elements */
			for (icol = ptr[irow]; icol < ptr[irow + 1] - 1; icol++)
				 for (i = icol + 1; i < ptr[irow + 1]; i++)
				 	if (idx[i - 1] < idx[icol - 1]) {
				 		int aux = idx[i - 1];
				 		idx[i - 1] = idx[icol - 1];
				 		idx[icol -1] = aux;
				 	} 
		} 
		
		rows_until_now += part_rows[p];
	}
	
	/* 
	 * Create communication table 
	 */
	 
	CukrCsrMatmpBuildCommtable(*mat);
	  	 
	free(w);	  	 
	LOG(DEBUG, "Done generating random multi-gpu CSR matrix"); 	  	 
}

// ------------------------------------------------------------------- //
					 
/**
 * \brief Generates a 2D Poisson multi-gpu
 *        matrix 
 * \param mat Pointer to the matrix 
 * \param n Order
 * \param name Matrix name
 */
void CukrCsrMatmpCreatePoisson2D(CUKR_MATRIX_MP* mat, int n, char* name)
{
	;
}

//============================ PRIVATE FUNCTIONS =========================== //

/**
 * \brief Generates n unique random integers,
 *        between lo and lo + n
 * \param x The vector where to put the elements
 * \param w Workspace vector (prev. allocated)
 * \param n The no. of requested numbers
 * \param lo The smallest 
 * \param up The largest 
 */
void gen_unique_int_rnd_mp(int *x, int *w, int n, int lo, int up)
{
	int i;
	
	/* Fill the workspace */
	int nn = up - lo + 1;
	for (i = 0; i <= nn; i++)
		w[i] = i;
		
	/* Extract one by one */
	for (i = 0; i < n; i++) {
		int irnd = rand() % nn + 1;
		x[i] = w[irnd] + lo - 1;
		w[irnd] = w[nn--];
		
		/* Has to be smaller than <up> */
		while (x[i] > up) {
			irnd = rand() % nn + 1;
			x[i] = w[irnd] + lo - 1;
			w[irnd] = w[nn--];
		}
	}	
}
