/**
 * @file mat_vec/mp/csr/csr_matrix_io.c
 * @author Serban Georgescu
 * @since 2008
 *
 * File IO functionality for matrices
 * in CSR format
 */

#define LOGGER_NAME "CUKr.mat_vec.mp.csr.csr_matrix_mp_io"

#include <stdio.h>
#include <stdlib.h>
#include "csr_matrix_mp.h"
#include "../../matrix_defs.h"
#include "../../../util/log.h"
#include "../../../db/db.h"

/** Get access to database */
extern CukrDB cukr_db;

//========================== FORWARD DECLARATIONS ========================= //

void CukrCsrMatmpReadFromCsrFile(CUKR_MATRIX_MP mat, FILE *f, int type);

//============================ PUBLIC FUNCTIONS =========================== // 

/**
 * \brief Loads a CSR matrix from a CSR file.
 *        The CSR file format is the following:
 *        1st_row: m n nz
 *        2nd->(n+1)th_row: ptr
 *        (n+1)th -> (n+nz+1)th_row: idx
 *        (n+nz+1)th -> (n+2*nz+1)th_row: val
 * \param mat Pointer to the matrix in which
 *        data is being read
 * \param filename The file where the data is
 *        being read from
 * \param type The data type the values should
 *        be considered to be (double, single,
 *        short).
 */
void CukrCsrMatmpCreateFromCsrFile(CUKR_MATRIX_MP mat, const char *filename, int type)
{
	FILE *f;
	int i, j, res;
	char buffer[255];
	
	LOGX(DEBUG, "Reading MP matrix from file: %s", filename);
	 		            
	/*
	 * Initialize the matrix
	 */
		 
	/* Open the file for reading */
	if ( (f = fopen(filename, "r")) == NULL) {
		CUKR_FATALX("Unable to open file %s for read",
		            filename);
	}    
	
	/* Read the dimensions */
	res = fscanf(f, "%d %d %d", &mat->rows, &mat->cols, &mat->nz);

	/*
	 * Read the matrix from file. 
	 */
	
	CukrCsrMatmpReadFromCsrFile(mat, f, type);		
		
	/*
	 * Close 
	 */
		
	fclose(f);
	LOGX(INFO, "Finished reading matrix %s from file %s",
	           mat->name, filename);  	 	  
}

//============================ PRIVATE FUNCTIONS =========================== //

/**
 * \brief Reads data from file and stores it
 *        in a matrix which is partitioned
 *        sequentially
 * \param mat The matrix whose name and dimensions
 *        have been already set
 * \param f The already opend file stream to read
 *        the data from
 * \param type The data type to read from the file
 */
void CukrCsrMatmpReadFromCsrFile(CUKR_MATRIX_MP mat, FILE *f, int type)
{
	int i, j, res;
	char buffer[255];
	
	LOGX(DEBUG, "Reading CSR data for MP matrix %s", 
	             mat->name);
		
	/* 
	 * Read the ptr vector
	 * We use it to get the no. of
	 * nonzeros, so it must be read all
	 * in advance.
	 */
	
	 /* Malloc rows+1 elements */
	 CUKR_INT *ptr;
	 if ( (ptr = (CUKR_INT*)malloc( (mat->rows + 1)*sizeof(CUKR_INT))) == NULL) {
	 	CUKR_FATALX("Could not allocate memory %ld bytes of "
	 				"memory for temp. ptr vector", 
	 				(unsigned long)((mat->rows + 1)*sizeof(CUKR_INT)));
	 }
	 
	 /* Read */
	 for (i = 0; i < mat->rows + 1; i++)
	 	res = fscanf(f, "%d", &ptr[i]);
	  
	 /*
	  * Compute row count and nonzero count
	  * for each part.
	  * We use that to know how much memory
	  * to allocate.
	  */	
	 
	 /* Compute the row count for the parts */
	int part_size, last_part_size;
	if (mat->nparts == 1) 
		last_part_size = mat->rows;
	else {
	 	part_size = mat->rows / mat->nparts; 
		last_part_size = mat->rows - (mat->nparts - 1) * part_size;
	}
	
	/* Put the row counts in a vector */
	int part_rows[CUKR_MAX_PE]; 
	for (i = 0; i < mat->nparts; i++)
		if (i < mat->nparts - 1)
			part_rows[i] = part_size;
		else
			part_rows[i] = last_part_size;
	
	 /* Compute the nonzeros for each part */
	 int part_nz[CUKR_MAX_PE];
	 int row_count = 0, nz_count = 0, part_count = 0;
	 for (i = 0; i < mat->rows; i++) {
	 	
	 	/* Update row and nz counters */
	 	row_count++;
	 	nz_count += ptr[i+1] - ptr[i];
	 	
	 	/* If part full, write nz and 
	 	 * pass to the next */
	 	if (row_count == part_rows[part_count]) {
	 		part_nz[part_count] = nz_count;
	 		nz_count = 0;
	 		row_count = 0;
	 		part_count++;
	 	}
	 }
	 
	 #ifdef CUKR_LOGGING
		 /* Log splitting */
		 sprintf(buffer, "Splitting for matrix %s: [", mat->name);
		 for (i = 0; i < mat->nparts; i++)
			 sprintf(buffer, "%s %d (%d el.)", buffer, part_rows[i], part_nz[i]);
		 sprintf(buffer, "%s ]", buffer);
		 LOGX(DEBUG, "%s", buffer);	
	 #endif	 
	 
	  
	 /*
	  * Now that we have the info, create
	  * the parts 
	  */
	 
	 /* Create the parts */	 
	 for (i = 0; i < mat->nparts; i++) {
	 	
	 	/* Create parts */
	 	sprintf(buffer, "%s.part%d", mat->name, i);
	 	CukrMatspCreate(&mat->part[i], part_rows[i], mat->cols, CUKR_MATRIX_FORMAT_CSR,
	 	                type, buffer);
		
		/* Set the the nz per part */
		mat->part[i]->nz = part_nz[i];
		
		/* Set parts as loaded from file */
		mat->part[i]->build_phase = CUKR_MAT_BUILD_CREATED_FILE;	 	                
	 }
	 
	 /* Malloc data members */
	 for (i = 0; i < mat->nparts; i++) {
		/* ptr */ 
		sprintf(buffer, "%s.ptr", mat->part[i]->name);
		CukrVecspCreate(&mat->part[i]->csr_mat->ptr, part_rows[i] + 1, CUKR_TYPE_INT, 		         buffer);
		/* idx */
		sprintf(buffer, "%s.idx",  mat->part[i]->name);
		CukrVecspCreate(&mat->part[i]->csr_mat->idx, part_nz[i],       CUKR_TYPE_INT, 		         buffer);
		/* val */
		sprintf(buffer, "%s.val",  mat->part[i]->name);
		CukrVecspCreate(&mat->part[i]->csr_mat->val, part_nz[i],       mat->part[i]->type_at_creation, buffer);		 		 	
	 } 
	   
	 
	 /*
	  * Add the ptr data 
	  * The ptr vector has to be translated
	  * so it begins from 1 for each matrix
	  */
	  
	 int pos = 0;
	 nz_count = 0;
	 for (i = 0; i < mat->nparts; i++)
	 {	
		for (j = 0; j < part_rows[i] + 1; j++)
	 		mat->part[i]->csr_mat->ptr->ph_ival[j] = ptr[pos + j] - nz_count;
		
		nz_count += part_nz[i];
	 	pos += part_rows[i];	
	 }	 
	 	 
	 /* 
	  * Read the idx and val data
	  * This is to be left as it is 
	  */
	 
	 /* Read the idx data */ 
	 for (i = 0; i < mat->nparts; i++)
		CukrVecspReadDataFromFile(mat->part[i]->csr_mat->idx, f, CUKR_TYPE_INT);       
		
	 /* Read the val data */ 
	 for (i = 0; i < mat->nparts; i++)
		CukrVecspReadDataFromFile(mat->part[i]->csr_mat->val, f, type);			
		
	/*
	 * Cleanup 
	 */
	 
	free(ptr); 	
	
	LOGX(DEBUG, "Done reading CSR data for MP matrix %s", mat->name);
}		
