/**
 * @file mat_vec/mp/matrix_mp.c
 * @author Serban Georgescu
 * @since 2008
 *  
 * Defines the CUKR_MATRIX_MP abstract
 * interace. It's purpose is to extend
 * a matrix across multiple PEs.
 * 
 * The CUKR_MATRIX_MP structure defines
 * a matrix which is split in a number 
 * of parts, so that operations can be
 * performed in parallel on multiple PEs.
 * 
 * In the case of a distributed matrix, we
 * cannot have the two type of data 
 * distribution (shared + distributed) since
 * the matrix cannot be shared. Therefore
 * the sharing type will not be defined. 
 * 
 * The shared matrix is simply split by
 * assigining for each PE a number of
 * n/p rows. 
 */

#define LOGGER_NAME "CUKr.mat_vec.mp.matrix_mp"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "matrix_mp.h"
#include "csr/csr_matrix_mp.h"
#include "../matrix_defs.h"
#include "../../threading/threading.h"
#include "../../util/log.h"
#include "../../util/utils.h"
#include "../../db/db.h" 

#define MAX(a,b) (a > b ? a : b)
#define MIN(a,b) (a < b ? a : b)

/** Get access to database */
extern CukrDB cukr_db;

//============================ PUBLIC FUNCTIONS =========================== //





				/*****************************************/
				//       CONSTRUCTOR / DESTRUCTOR        //
				/*****************************************/         




/**
 * \brief Creates a matrix of the specified
 *        format, on the CPU, with data of 
 *        the specified type
 * \param mat Pointer to the matrix structure 
 *        being created
 * \param rows,cols Matrix size
 * \param format Matrix format (e.g. CSR)
 * \param type Matrix data type 
 *        (float, qdouble, double)
 * \param name Matrix name 
 */
void CukrMatmpCreate(CUKR_MATRIX_MP* mat, CUKR_INT rows, CUKR_INT cols,
				     int format, int type, char* name)
{
	char buffer[255];
	int i;
	
	/** 
	 * Check for valid parameters 
	 */
	
	/* First check for valid name */
	if (name == NULL)
		CUKR_FATAL("Matrix must have a non-null name");
	
	LOGX(DEBUG, "Creating MP matrix %s", name);
	
	
	/*
	* Malloc for the matrix structure
	*/
	 
	/* Alloc memory */
	if ( ((*mat) = (CUKR_MATRIX_MP)malloc(sizeof(struct _CUKR_MATRIX_MP))) == NULL) {
		CUKR_FATAL("Error allocating memory for MP matrix object ");
	}
	else {
		LOG(DEBUG, "Allocated MP matrix structure");
	}
	   
	 /* 
	  * Set all matrix parameters 
	  */
	 
	 /* Set the name */
	 sprintf((*mat)->name, "%s", name); 
	 
	 /* Set sizes */ 
	 (*mat)->rows = rows;
	 (*mat)->cols = cols;
	 (*mat)->nz   = 0;
	 
	 /*
	  * Set partitioning-related parameters
	  */
	  
	 /* Set the no. of parts to be equal to
	  * the no. of processors used */
	 (*mat)->nparts = CukrDBGetI(CUKR_DB_PE);
	
	 /* Set part index is zero, since it is
	  * created directly */
	 (*mat)->part_index = 0;
	 
	 /* Compute the row count for the parts */
	 int part_size, last_part_size;
	 if ((*mat)->nparts == 1) 
		 last_part_size = (*mat)->rows;
	 else {
		 part_size = (*mat)->rows / (*mat)->nparts; 
		 last_part_size = (*mat)->rows - ((*mat)->nparts - 1) * part_size;
	 }
	
	 /* Put the row counts in a vector */
	 int part_rows[CUKR_MAX_PE]; 
	 for (i = 0; i < (*mat)->nparts; i++) {
		 if (i < (*mat)->nparts - 1)
			  part_rows[i] = part_size;
	     else
			  part_rows[i] = last_part_size;
	 }
	 
	 /*
	  * Create parts
	  */
	 
	  for (i = 0; i < (*mat)->nparts; i++) {
	  	  sprintf(buffer, "%s.part%d", name, i);	
	  	  CukrMatspCreate(&(*mat)->part[i], part_rows[i], cols, format, type, buffer);
	  }  	   
	 	 
	 /*
	  * Success message 
	  */
	
	 LOGX(DEBUG, "Successfully created MP matrix %s(%d,%d) on host.", 
	             (*mat)->name, rows, cols);
	             
	
}   

// ------------------------------------------------------------------- //

/**
 * \brief Loads a MP matrix from file
 * \param mat Pointer to the MP matrix 
 *        in which data is being read
 * \param filename The file where the data is
 *        being read from
 * \param format The matrix storage format (CSR)
 * \param type The data type the values should
 *        be considered to be (double, qdouble,
 *        single). 
 * \param name The name given to the matrix
 */
void CukrMatmpCreateFromFile(CUKR_MATRIX_MP* mat, const char *filename, int format, int type, char* name)
{
	/*
	 * Validate parameters
	 */
	
	/* Check for valid name */
	if (name == NULL)
		CUKR_FATAL("Matrix must have a non-null name");
	
	LOGX(DEBUG, "Creating MP matrix %s from file %s", 
	            name, filename);
	            
	/*
	 * Create matrix structure
	 */
	 
	/* Alloc memory */
	if ( ((*mat) = (CUKR_MATRIX_MP)malloc(sizeof(struct _CUKR_MATRIX_MP))) == NULL)
		CUKR_FATAL("Error allocating memory for MP matrix object ");
	 
	/* 
	 * Set all matrix parameters 
	 */		
	 
	/* Set the name */
	sprintf((*mat)->name, "%s", name); 
	
	/*
	 * Set partitioning-related parameters
	 */
	
	/* Set the no. of parts to be equal to
	 * the no. of processors used */
	(*mat)->nparts = CukrDBGetI(CUKR_DB_PE);
	
	/* Set part index is zero, since it is
	 * created directly */
	(*mat)->part_index = 0; 	
	
	/*
	 * Create from file
	 */
	 
	 /* Get format */
	 int mat_format; 
	 if (format == CUKR_DEFAULT)
	 	mat_format = CukrMatFormatStrToId(CukrDBGetS(CUKR_DB_MATRIX_FORMAT));
	 else
	 	mat_format = format;  
	 	
	 /* Get type */
	 int mat_type; 
	 if (type == CUKR_DEFAULT)
	 	 mat_type = CukrTypeStrToId(CukrDBGetS(CUKR_DB_SOLVER_PRECISION));
	 else
	 	 mat_type = type;	
	 	
	 /* If CSR format */
	 if (mat_format == CUKR_MATRIX_FORMAT_CSR) 
		CukrCsrMatmpCreateFromCsrFile(*mat, filename, mat_type);
		
	 /* Else, error */
	 else 
	 	CUKR_FATALX("Invalid format for matrix %s: %d", 
	 	            name, format);
	 	            
	  /*
	  * Success message 
	  */
	
	 LOGX(DEBUG, "Successfully created MP matrix %s(%d,%d) on host (from file).", 
	             (*mat)->name, (*mat)->rows, (*mat)->cols);	            		          
}

// ------------------------------------------------------------------- //

/**
 * \brief Finalizes the creation of the matrix,
 *        after all elements have been added. 
 * \param mat The matrix structure to assemble.
 */
void CukrMatmpAssemble(CUKR_MATRIX_MP mat)
{	
	int i;
	
	LOGX(DEBUG, "Starting the assembly for MP matrix %s",
	             mat->name);
	             
	/* 
	 * Assemble parts
	 */
	
	int nz = 0; 
	for (i = 0; i < mat->nparts; i++) {
		CukrMatspAssemble(mat->part[i]);
		nz += mat->part[i]->nz;	
	} 
	mat->nz = nz;
		
	/*
	 * Format conversion for the parts
	 * is not needed. Done automatically when 
	 * assembling the parts. 
	 */
	 
	/*
	 * Build communication table
	 */
	 
	/* If CSR format */
	if (mat->part[0]->format == CUKR_MATRIX_FORMAT_CSR) {
		CukrCsrMatmpBuildCommtable(mat);
	} 
	/* Else, error */
	else 
	 	CUKR_FATALX("Invalid format for MP matrix %s: %d", 
	 	            mat->name, mat->part[0]->format);
	  
	/* 
	 * Finalize
	 */
	
	LOGX(DEBUG, "MP Matrix %s was succesfully assembled",
	             mat->name);
}


// ------------------------------------------------------------------- //
                   
/**
 * \brief Frees all the memory used 
 *        by the MP matrix
 * \param mat Pointer to the MP matrix 
 *        to be destroyed 
 */
void CukrMatmpDestroy(CUKR_MATRIX_MP* mat)
{
	int i;
	
	/* Check for non-null vector */
	if (*mat == NULL)
		return;
	
	/* Copy the name, so it doesn't get
	 * destroyed */
	char name[255];
	memcpy(name, (*mat)->name, 255);
	LOGX(DEBUG, "Destroying MP matrix %s", name);
	
	/* Schedule the distruction of the parts */ 
	for (i = 0; i < (*mat)->nparts; i++)
		CukrThreadsRun(i, (void*)CukrMatspDestroy, CUKR_FUNCID_MAT_DESTROY, &((*mat)->part[i]));	 
	
	/* Wait for the jobs to finish */
	CukrThreadsWaitCompletion((*mat)->nparts);	
	
	/* Destroy commtable */
	for (i = 0; i < (*mat)->nparts; i++) {
		CukrContainersVectorOfCommentryDestroy(&((*mat)->comm_table.give_container[i]));
		CukrContainersVectorOfCommentryDestroy(&((*mat)->comm_table.need_container[i]));
		CukrContainersVectorOfIntDestroy(&((*mat)->comm_table.raw_give_container[i]));
	}
		
	/* Finally, free the matrix structure */
	free(*mat); 
	LOGX(DEBUG, "MP matrix %s destroyed", name);		
}






				/*****************************************/
				//          TRANSFER / CONVERT           // 
				/*****************************************/ 






/**
 * \brief Changes the location of a MP matrix
 *        by copying the data from one location
 *        to another
 * \param mat The MP matrix 
 * \param dest_loc Destination matrix location (CPU,GPU)
 * \param dest_type Destination matrix value type 
 *        (double, q-double, single)
 */
void CukrMatmpTransfer(CUKR_MATRIX_MP mat, int dest_loc, int dest_type)
{
	int i;
	
	/* Check to see if transfer is really 
	 * necessary. It will be done of the lower
	 * levels anyway, but we don't want to waste
	 * time. Since all parts are synchronized,
	 * it is enough to chech the first */
	int dest_status = CukrLocIdAndTypeIdToStatusId(dest_loc, dest_type); 
	if (mat->part[0]->status & dest_status) 
	 	return;
           
	/* Schedule the transfer of the parts */ 
	for (i = 0; i < mat->nparts; i++)
		CukrThreadsRun(i, (void*)CukrMatspTransfer, CUKR_FUNCID_MAT_TRANSFER, mat->part[i], dest_loc, dest_type);	
	
	/* Wait for the jobs to finish */
	CukrThreadsWaitCompletion(mat->nparts);
}                   

// ------------------------------------------------------------------- //

/**
 * \brief Changes the storage format of 
 *        the MP matrix to a diferent variation
 *        or to a completely different format
 * \param mat The MP matrix 
 * \param dest_format Destination format
 * \param dest_variation Destination variation 
 */
void CukrMatmpConvert(CUKR_MATRIX_MP mat, int dest_format, int dest_variation)
{
	int i;
    
    /* Check to see if conversion is really 
	 * necessary. It will be done of the lower
	 * levels anyway, but we don't want to waste
	 * time. Since all parts are synchronized,
	 * it is enough to chech the first */
	if (mat->part[0]->format    == dest_format && 
	    mat->part[0]->variation == dest_variation)
	    return;
          
	/* Schedule the conversion of the parts
	 * Schedule not because is needed, since
	 * no GPU is involved, but because doing
	 * it multicore may be faster */ 
	for (i = 0; i < mat->nparts; i++)
		CukrThreadsRun(i, (void*)CukrMatspConvert, CUKR_FUNCID_MAT_CONVERT, mat->part[i], dest_format, dest_variation);	
	
	/* Wait for the jobs to finish */
	CukrThreadsWaitCompletion(mat->nparts);	
}






				/*****************************************/
				//               GET/SET                 // 
				/*****************************************/ 






/**
 * \brief Adds an element to a matrix. This can only
 *        be used after the matrix has been created and
 *        before it is assembled.
 * \param mat The matrix where the element is to be added
 * \param row The row of the element 
 *            (global numbering)
 * \param col The col of the element
 *            (global numbering)
 * \param val The value of the element
 */
void CukrMatmpAddElement(CUKR_MATRIX_MP mat, int row, int col, CUKR_DOUBLE val)
{
	/* Add the element to the right part */
	int i, lo = 0;
	for (i = 0; i < mat->nparts; i++) {
		if (row > lo && row <= lo + mat->part[i]->rows) {
			CukrMatspAddElement(mat->part[i], row - lo, col, val);
			return;
		}
		lo += mat->part[i]->rows;
	}
	
	/* Error if still here, means the row
	 * was invalid */
	CUKR_FATALX("Invalid entry (row=%d,col=%d,val=%g) "\
	            "was added to matrix %s(%d,%d). Remember "\
	            "that numbering starts from 1.", row, col, 
	            val, mat->name, mat->rows, mat->cols); 
		
}

// ------------------------------------------------------------------- //

/**
 * \brief Adds a row of elements to a matrix. 
 *        This can only be used after the
 *        matrix has been created and before
 *        it is assembled.
 * \param mat The matrix where the row 
 *        is to be added
 * \param row The row to be added
 *        (global numbering)
 * \param val Vector of values, having at least as  
 *        many elements as the length of the row.
 */
void CukrMatmpAddRow(CUKR_MATRIX_MP mat, int row, CUKR_DOUBLE* val)
{
	/* Add the row to the right part */
	int i, lo = 0;
	for (i = 0; i < mat->nparts; i++) {
		if (row >= lo && row < lo + mat->part[i]->rows) {
			CukrMatspAddRow(mat->part[i], row - lo, val);
			return;
		}
		lo += mat->part[i]->rows;
	}
	
	/* Error if still here, means the row
	 * was invalid */
	CUKR_FATALX("Invalid entry row (%d) was added to the matrix"\
	            "%s(%d,%d). Remember that numbering starts from 1.",
	            row, mat->name, mat->rows, mat->cols); 
}

// ------------------------------------------------------------------- //

/**
 * \brief Adds a column of elements to a matrix. 
 *        This can only be used after the
 *        matrix has been created and before
 *        it is assembled.
 * \param mat The matrix where the column 
 *        is to be added
 * \param col The column to be added
 * 		  (global numbering)
 * \param val Vector of values, having at least as  
 *        many elements as the length of the column.
 */
void CukrMatmpAddCol(CUKR_MATRIX_MP mat, int col, CUKR_DOUBLE* val)
{
	/* Add sections of the column to 
	 * each part */
	int i, lo = 0;
	for (i = 0; i < mat->nparts; i++) {
		CukrMatspAddCol(mat->part[i], col, &val[lo]);
		lo += mat->part[i]->rows;
	}
}

// ------------------------------------------------------------------- //

/**
 * \brief Adds a diagonal of elements to a matrix. 
 *        This can only be used after the
 *        matrix has been created and before
 *        it is assembled.
 * \param mat The matrix where the diagonal 
 *        is to be added
 * \param diag The diagonal to be added. 0 is the
 *        main diagonal, -1 ... min(rows,cols) are
 *        the lower diagonals and 1 .. min(rows,cols)
 *        are the upper diagonals.
 *        (global numbering)
 * \param val Vector of values, having at least as many 
 *        elements as the length of the diagonal.
 */
void CukrMatmpAddDiag(CUKR_MATRIX_MP mat, int diag, CUKR_DOUBLE* val)
{
	int i, lo;
	
	/* Add diagonal for first part */
	CukrMatspAddDiag(mat->part[0], diag, val);
	lo = mat->part[0]->rows;
	/* Add the rest of diagonals */
	for (i = 1; i < mat->nparts; i++) {
		/* For positive diagonal, the lower part
		 * is full, so use the normal way */
		if (diag >= 0) 
			CukrMatspAddDiag(mat->part[i], diag + lo, &val[lo]);
		/* For negative diagonal, we have to shift
		 * the val vector */
		else
			CukrMatspAddDiag(mat->part[i], diag + lo, &val[MAX(0,diag + lo)]); 
			
		lo += mat->part[i]->rows;
	}
}

// ------------------------------------------------------------------- //
 
/**
 * \brief Extracts the diagonal from a matrix
 * \param mat The matrix 
 * \param vec The vector where to store the diagonal
 * \param diag The diagonal to read (zero for the
 *        main diagonal) (global numbering)
 * \param invert Whether to invert it or not 
 */
void CukrMatmpGetDiag(CUKR_MATRIX_MP mat, CUKR_VECTOR_MP vec, int diag, int invert)
{
	
	/* Bring the matrix to CPU in DOUBLE */
	CukrMatmpTransfer(mat, CUKR_LOCATION_CPU, CUKR_TYPE_DOUBLE);
	
	/* Alloc in double on CPU, to make sure we 
	 * have where to put the diagonal */
	CukrVecmpMalloc(vec, CUKR_LOCATION_CPU, CUKR_TYPE_DOUBLE); 
		
	/* Get diagonals for all parts */
	int i, c = 0, lo = 0;
	for (i = 0; i < mat->nparts; i++)
	{		
		/* Get the diagonal */
		CukrMatspGetDiag(mat->part[i], vec->cells[i], lo + diag, invert);
		
		/* Update the pos, so that it will point
		 * to the diagonal which would correspond
		 * to the main diagonal in the big matrix */
		lo += mat->part[i]->rows; 			
	}
}




				/*****************************************/
				//                 I/O                   // 
				/*****************************************/ 





/**
 * \brief Prints a MP matrix
 * \param mat The MP matrix being printed
 */
void CukrMatmpPrint(CUKR_MATRIX_MP mat)
{
	CukrMatmpPrintToFile(mat, stdout);
}
	
// ------------------------------------------------------------------- //

/**
 * \brief Prints a MP matrix to file
 * \param vec The MP matrix being printed
 * \param filename The full path of the file
 *        to write the matrix to
 */
void CukrMatmpPrintToFilename(CUKR_MATRIX_MP mat, char* filename)
{
	FILE *f;
	
	if ( (f = fopen(filename, "w")) == NULL)
		CUKR_FATALX("Could not write MP matrix %s to file "\
		            "%s because the file could not be opened",
		            mat->name, filename);
	
	CukrMatmpPrintToFile(mat, f);
	
	fclose(f);
}

// ------------------------------------------------------------------- //

/**
 * \brief Prints a MP matrix to a file stream
 * \param mat The MP matrix being printed
 * \param f An open file stream to write to
 */
void CukrMatmpPrintToFile(CUKR_MATRIX_MP mat, FILE *f)
{
	CukrMatmpPrintToFileWithIndent(mat, f, "");
}


// ------------------------------------------------------------------- //

/**
 * \brief Prints a MP matrix to a file stream
 * \param mat The MP matrix being printed
 * \param f An open file stream to write to
 * \param indent An indentation string, to be inserted
 *        before every print.
 */
void CukrMatmpPrintToFileWithIndent(CUKR_MATRIX_MP mat, FILE *f, char* indent) 
{	
	int i;
	
	LOGX(DEBUG, "Printing MP matrix %s", mat->name);
	
	/* Merge indents */
	char mindent[255], mmindent[255], mmmindent[255];
	sprintf(mindent,  "%s    ",indent);
	sprintf(mmindent, "%s            ",indent);
	
	fprintf(f, "%sMP Matrix %s (%d,%d,%d) with %d parts \n", indent,
	       mat->name, mat->rows, mat->cols, mat->nz, mat->nparts);
	
	/* Print all the parts */
	fprintf(f, "%sPARTS\n", indent);	
	for (i = 0; i < mat->nparts; i++)
		CukrMatspPrintToFileWithIndent(mat->part[i], f, mindent);
		
	/* Print the commtable */
	fprintf(f, "%sCOMMTABLE\n", indent);
	for (i = 0; i < mat->nparts; i++) {
		fprintf(f, "    %sPE %d\n", indent, i);
		fprintf(f, "        %sNeed\n", indent);
		CukrContainersVectorOfCommentryPrintToFileWithIndent(mat->comm_table.need_container[i], f, mmindent);	
		fprintf(f, "        %sGive\n", indent);
		CukrContainersVectorOfCommentryPrintToFileWithIndent(mat->comm_table.give_container[i], f, mmindent);
	}
		
	LOGX(DEBUG, "Done printing MP matrix %s", mat->name);	
}
