/**
 * @file mat_vec/mp/matrix_mp.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines the CUKR_MATRIX_MP abstract
 * interace. It's purpose is to extend
 * a matrix across multiple PEs.
 * 
 * The CUKR_MATRIX_MP structure defines
 * a matrix which is split in a number 
 * of parts, so that operations can be
 * performed in parallel on multiple PEs.
 * 
 * In the case of a distributed matrix, we
 * cannot have the two type of data 
 * distribution (shared + distributed) since
 * the matrix cannot be shared. Therefore
 * the sharing type will not be defined. 
 * 
 * The shared matrix is simply split by
 * assigining for each PE a number of
 * n/p rows. 
 */
 
#ifndef CUKR_MATRIX_MP_H_
#define CUKR_MATRIX_MP_H_

#ifndef CUKR_MATRIX
	#define CUKR_MATRIX CUKR_MATRIX_MP
#endif

/* =============== FORWARD DECLARATIONS ============== */

typedef struct _CUKR_MATRIX_MP_COMMENTRY CUKR_MATRIX_MP_COMMENTRY;
typedef struct _CUKR_MATRIX_MP_COMMTABLE CUKR_MATRIX_MP_COMMTABLE;
typedef struct _CUKR_MATRIX_MP *CUKR_MATRIX_MP;

/* =================================================== */


#include "vector_mp.h"
#include "matrix_mp_gen.h"
#include "../sp/matrix_sp.h"
#include "../../util/containers.h"


/** All information needed to exchange the
 * needed elements between parts */
struct _CUKR_MATRIX_MP_COMMTABLE {
	
	/** Containers for <needed> comm-entries */
	CUKR_CONTAINER_VECTOR_COMMENTRY need_container[CUKR_MAX_PE];
	
	/** Containers for <given> comm-entries */
	CUKR_CONTAINER_VECTOR_COMMENTRY give_container[CUKR_MAX_PE];
	
	/** Raw element container for <given> elements.
	 * It is kept only if the matrix is a part of
	 * an MPI matrix, case in which the give 
	 * container will be created after merging with
	 * the global give container */
	/* Create a container for each PE, in which
	 * to add what is needed and what is given */
	 CUKR_CONTAINER_VECTOR_INT raw_give_container[CUKR_MAX_PE]; 
	  
};

/** The multi-gpu matrix structure
 * containig a matrix eac parts and
 * a communication table */
struct _CUKR_MATRIX_MP {
	
	// ------------ Public ------------- //
	
	/** The name of the vector */
	char name[255];
	
	/** Number of cols, rows and nonzeros*/
	int cols, rows, nz;
	
	/** Number of parts, equal to the no.
	 * of GPUs used */
	int nparts;
	
	/** The vectors holding the parts */
	CUKR_MATRIX_SP part[CUKR_MAX_PE];
	
	
	// ----------- Protected ------------ //
	
	/** If this matrix is part of an MPI
	 * matrix, which is the starting row */
	int part_index; 
	
	/** The communication table */
	CUKR_MATRIX_MP_COMMTABLE comm_table;
	
	
	// ------------ Private ------------- //
		
	/** Matrix data type at creation */
	int type_at_creation;
};






				/*****************************************/
				//       CONSTRUCTOR / DESTRUCTOR        //
				/*****************************************/         




/**
 * \brief Creates a matrix of the specified
 *        format, on the CPU, with data of 
 *        the specified type
 * \param mat Pointer to the matrix structure 
 *        being created
 * \param rows,cols Matrix size
 * \param format Matrix format (e.g. CSR)
 * \param type Matrix data type 
 *        (float, qdouble, double)
 * \param name Matrix name 
 */
void CukrMatmpCreate(CUKR_MATRIX_MP* mat, CUKR_INT rows, CUKR_INT cols,
				     int format, int type, char* name);

// ------------------------------------------------------------------- //

/**
 * \brief Loads a matrix from file.
 * \param mat Pointer to the matrix in which
 *        data is being read
 * \param filename The file where the data is
 *        being read from
 * \param format The matrix storage format (CSR)
 * \param type The data type the values should
 *        be considered to be (double, qdouble,
 *        single).
 * \param name The name given to the matrix
 */
void CukrMatmpCreateFromFile(CUKR_MATRIX_MP* mat, const char *filename, int format, int type, char* name);

// ------------------------------------------------------------------- //

/**
 * PROTECTED FUNCTION
 * 
 * \brief Loads a matrix from file.
 * \param mat Pointer to the matrix in which
 *        data is being read
 * \param filename The file where the data is
 *        being read from
 * \param format The matrix storage format (CSR)
 * \param type The data type the values should
 *        be considered to be (double, qdouble,
 *        single).
 * \param name The name given to the matrix
 * \param part_index If the matrix is part of an
 *        MPI matrix, this paramter shows from which
 *        row in the MPI matrix it starts.
 */
void CukrMatmpCreateFromFileWithIndex(CUKR_MATRIX_MP* mat, const char *filename, int format, int type, char* name, int part_index);

// ------------------------------------------------------------------- //

/**
 * \brief Finalizes the creation of the matrix,
 *        after all elements have been added. This
 *        inlcudes the creation of data structures
 *        and building the comm. tables.
 * \param mat The matrix structure to assemble.
 */
void CukrMatmpAssemble(CUKR_MATRIX_MP mat); 

// ------------------------------------------------------------------- //
                   
/**
 * \brief Frees all the memory used 
 *        by the MP matrix
 * \param mat Pointer to the MP matrix 
 *        to be destroyed 
 */
void CukrMatmpDestroy(CUKR_MATRIX_MP* mat);






				/*****************************************/
				//               GET/SET                 // 
				/*****************************************/ 





/**
 * \brief Adds an element to a matrix. This can only
 *        be used after the matrix has been created and
 *        before it is assembled.
 * \param mat The matrix where the element is to be added
 * \param row The row of the element 
 *            (global numbering)
 * \param col The col of the element
 *            (global numbering)
 * \param val The value of the element
 */
void CukrMatmpAddElement(CUKR_MATRIX_MP mat, int row, int col, CUKR_DOUBLE val);

// ------------------------------------------------------------------- //

/**
 * \brief Adds a row of elements to a matrix. 
 *        This can only be used after the
 *        matrix has been created and before
 *        it is assembled.
 * \param mat The matrix where the row 
 *        is to be added
 * \param row The row to be added
 *        (global numbering)
 * \param val Vector of values, having at least as  
 *        many elements as the length of the row.
 */
void CukrMatmpAddRow(CUKR_MATRIX_MP mat, int row, CUKR_DOUBLE* val);

// ------------------------------------------------------------------- //

/**
 * \brief Adds a column of elements to a matrix. 
 *        This can only be used after the
 *        matrix has been created and before
 *        it is assembled.
 * \param mat The matrix where the column 
 *        is to be added
 * \param col The column to be added
 * 		  (global numbering)
 * \param val Vector of values, having at least as  
 *        many elements as the length of the column.
 */
void CukrMatmpAddCol(CUKR_MATRIX_MP mat, int col, CUKR_DOUBLE* val);

// ------------------------------------------------------------------- //

/**
 * \brief Adds a diagonal of elements to a matrix. 
 *        This can only be used after the
 *        matrix has been created and before
 *        it is assembled.
 * \param mat The matrix where the diagonal 
 *        is to be added
 * \param diag The diagonal to be added. 0 is the
 *        main diagonal, -1 ... min(rows,cols) are
 *        the lower diagonals and 1 .. min(rows,cols)
 *        are the upper diagonals.
 *        (global numbering)
 * \param val Vector of values, having at least as many 
 *        elements as the length of the diagonal.
 */
void CukrMatmpAddDiag(CUKR_MATRIX_MP mat, int diag, CUKR_DOUBLE* val);

// ------------------------------------------------------------------- //

/**
 * \brief Extracts the diagonal from a matrix
 * \param mat The matrix 
 * \param vec The vector where to store the diagonal
 * \param diag The diagonal to read (zero for the
 *        main diagonal) (global numbering)
 * \param invert Whether to invert it or not 
 */
void CukrMatmpGetDiag(CUKR_MATRIX_MP mat, CUKR_VECTOR_MP vec, int diag, int invert);







				/*****************************************/
				//          TRANSFER / CONVERT           // 
				/*****************************************/ 

  




/**
 * \brief Changes the location of a matrix
 *        by copying the data from one location
 *        to another (if not already there)
 * \param mat Pointer to the matrix 
 * \param dest_loc Destination matrix location (CPU,GPU)
 * \param dest_type Destination matrix value type 
 *        (float, double)
 */
void CukrMatmpTransfer(CUKR_MATRIX_MP mat, int dest_loc, int dest_type);                   

// ------------------------------------------------------------------- //

/**
 * \brief Changes the storage format of 
 *        the matrix to a diferent variation
 *        or to a completely different format
 *        (if not already in that format)
 * \param mat The matrix 
 * \param dest_format Destination format
 * \param dest_variation Destination variation 
 */
void CukrMatmpConvert(CUKR_MATRIX_MP mat, int dest_format, int dest_variation);







				/*****************************************/
				//                 I/O                   // 
				/*****************************************/ 





/**
 * \brief Prints a MP matrix
 * \param mat The MP matrix being printed
 */
void CukrMatmpPrint(CUKR_MATRIX_MP mat);
	
// ------------------------------------------------------------------- //

/**
 * \brief Prints a MP matrix to file
 * \param vec The MP matrix being printed
 * \param filename The full path of the file
 *        to write the matrix to
 */
void CukrMatmpPrintToFilename(CUKR_MATRIX_MP mat, char* filename);

// ------------------------------------------------------------------- //

/**
 * \brief Prints a MP matrix to a file stream
 * \param mat The MP matrix being printed
 * \param f An open file stream to write to
 */
void CukrMatmpPrintToFile(CUKR_MATRIX_MP mat, FILE *f);

// ------------------------------------------------------------------- //

/**
 * \brief Prints a MP matrix to a file stream
 * \param mat The MP matrix being printed
 * \param f An open file stream to write to
 * \param indent An indentation string, to be inserted
 *        before every print.
 */
void CukrMatmpPrintToFileWithIndent(CUKR_MATRIX_MP mat, FILE *f, char* indent); 



#endif /* CUKR_MATRIX_MP_H_ */
