/**
 * @file mat_vec/mp/matrix_gen_mp.c
 * @author Serban Georgescu
 * @since 2008
 * 
 * Holds a list of matrix generators.
 * The generators here are indepenent of
 * the matrix format, so the format-specific
 * implementation of the generator will be
 * found elsewere (most probably in the
 * subfolder assigned to the specific format)
 * 
 * The generator will create the matrix,
 * so the matrix should not be created beforehand
 * 
 * Current generators are:
 * 	  - random 
 * 	  - poisson 2D 
 * 
 */

#define LOGGER_NAME "CUKr.mat_vec.mp.matrix_gen"

#include <stdio.h>
#include <stdlib.h>
#include <time.h> 
#include "matrix_mp_gen.h"
#include "csr/csr_matrix_mp.h"
#include "../matrix_defs.h"
#include "../../util/log.h"
#include "../../util/utils.h"
#include "../../db/db.h"

#define MAX(a,b) (a > b ? a : b)
#define MIN(a,b) (a < b ? a : b)

/**
 * \brief Generates a sparse random multi-gpu matrix, 
 *        filled with random elements 
 * \param mat Pointer to the multi-gpu matrix 
 * \param format Matrix format (CSR)
 * \param n Matrix size
 * \param halfb The half-bandwidth of the matrix
 * \param nz_per_row No. of nonzeros per row
 * \param min,max The lower and upper bound for
 *                matrix values
 * \param seed The seed for the random no. generator
 * \param name Matrix name
 */
void CukrMatmpCreateRandom(CUKR_MATRIX_MP* mat, int format, int n, int halfb, int nz_per_row, 
						   CUKR_DOUBLE min, CUKR_DOUBLE max, CUKR_DOUBLE seed, char* name)
{
	 /*
	  * Alloc memory for structure
	  */
	
	 /* Alloc memory for matrix structure */
	 if ( ((*mat) = (CUKR_MATRIX_MP)malloc(sizeof(struct _CUKR_MATRIX_MP))) == NULL)
	 	CUKR_FATAL("Error allocating memory for PE matrix object ");
	 
	 /*
	  * Set matrix properties 
	  */
	 
	/* Set name */
	sprintf((*mat)->name, "%s", name);
     
    /* Set the no. of parts to be equal to
	* the no. of processors used */
	(*mat)->nparts = CukrDBGetI(CUKR_DB_PE);
	
	/* Set sizes */
	(*mat)->rows = n;
	(*mat)->cols = n;
	(*mat)->nz = n * nz_per_row;
	
	/* Set part index to zero, since it is 
	 * not a part of an MPI matrix yet. If 
	 * it were so, the MPI matrix creation
	 * routine should set this to the right
	 * value. */
	(*mat)->part_index = 0;
		
	/* Do it for CSR structure 
	 * (only CSR is currently supported) */
	if (format == CUKR_MATRIX_FORMAT_CSR) 
		CukrCsrMatmpCreateRandom(mat, halfb, nz_per_row, min, max, seed);
		
	/* Error if unknown format */	
	else
		CUKR_FATALX("Invalid format for MP matrix %s", name);
}                                

// ------------------------------------------------------------------- //
					 
/**
 * \brief Generates a sparse random multi-gpu 
 *        matrix, filled with random elements 
 * \param mat Pointer to the multi-gpu matrix 
 * \param format Matrix format (CSR)
 * \param n Order
 * \param name Matrix name
 */
void CukrMatmpCreatePoisson2D(CUKR_MATRIX_MP* mat, int format, int n, char* name)
{
	/* Do it for CSR structure 
	 * (only CSR is currently supported) */
	if (format == CUKR_MATRIX_FORMAT_CSR) 
		CukrCsrMatmpCreatePoisson2D(mat, n, name);
		
	/* Error if unknown format */	
	else
		CUKR_FATALX("Invalid format for multigpu matrix %s", name); 	
}

/**
 * \brief Generates a diagonally-dominated banded
 *        MP matrix, filled with ones but the main 
 * 		  diagonal computed as such as to be dominant 
 * \param mat Pointer to the matrix 
 * \param format Matrix format (CSR)
 * \param rows*cols Matrix order
 * \param hb Halfbandwidth
 * \param shift The shift 
 * \param name Matrix name
 */
void CukrMatmpCreateBanded(CUKR_MATRIX_MP* mat, int format, int rows, int cols, int hb, char* name)
{
	CukrMatmpCreateBandedWShift(mat, format, rows, cols, hb, 0, name);
}

/**
 * \brief Generates a diagonally-dominated banded
 *        MP matrix, filled with ones but the main 
 * 		  diagonal computed as such as to be dominant 
 * \param mat Pointer to the matrix 
 * \param format Matrix format (CSR)
 * \param rows*cols Matrix order
 * \param hb Halfbandwidth
 * \param shift The shift 
 * \param name Matrix name
 */
void CukrMatmpCreateBandedWShift(CUKR_MATRIX_MP* mat, int format, int rows, int cols, int hb, int shift, char* name)
{
	int p, b;
	
	LOGX(DEBUG, "Creating MP banded matrix %s of " \
				"order (%d,%d) and halfbandwidth %d", 
				name, rows, cols, hb);
	
	/*
	 * Alloc memory for structure
	 */
	
	/* Alloc memory for matrix structure */
	if ( ((*mat) = (CUKR_MATRIX_MP)malloc(sizeof(struct _CUKR_MATRIX_MP))) == NULL)
		CUKR_FATAL("Error allocating memory for MP matrix object ");
	 
	/*
	 * Set matrix properties 
	 */
	 
	/* Set name */
	sprintf((*mat)->name, "%s", name);
     
   /* Set the no. of parts to be equal to
	* the no. of processors used */
	(*mat)->nparts = CukrDBGetI(CUKR_DB_PE);
	
	/* Set sizes */
	(*mat)->rows = rows;
	(*mat)->cols = cols;
	int nz = 0;
	for (b = -hb; b <= hb; b++) 
		nz += cols - abs(b); 
	(*mat)->nz = nz;
	
	/* Set part index to zero, since it is 
	 * not a part of an MPI matrix yet. If 
	 * it were so, the MPI matrix creation
	 * routine should set this to the right
	 * value. */
	(*mat)->part_index = 0;
		
	/*
	 * Compute part sizes
	 */
	
	 int part_rows, last_part_rows;
	 int parts_rows[CUKR_MAX_PE];
	
    /* Compute the size for the parts */
	 if ((*mat)->nparts == 1) 
	 	 last_part_rows = rows;
	 else {
		 part_rows = rows / (*mat)->nparts; 
		 last_part_rows = rows - ((*mat)->nparts - 1) * part_rows;
	 }
	 
	 /* Store the size for part i in part_sizes[i] 
	  * Make all but the last part have part_size 
	  * elements and the last one have last_part_size
	  * elements */
	 for (p = 0; p < (*mat)->nparts; p++) {
	 	if (p < (*mat)->nparts - 1)	
	 		parts_rows[p] = part_rows;
	 	else
	 		parts_rows[p] = last_part_rows;
	 }
	
	/*
	 * Create part matrices as banded
	 * with shift
	 */
	 	
	 for (p = 0; p < (*mat)->nparts; p++)
	 	CukrMatspCreateBandedWShift(&((*mat)->part[p]), format, parts_rows[p], cols, hb, shift + p * part_rows , "B");
	 
	 /*
	  * Build commtables
	  */
	  
	  /* Init the comm-table containers for 
	   * holding comm-entries */
	  for (p = 0; p < (*mat)->nparts; p++) {
	 	CukrContainersVectorOfCommentryCreate(&(*mat)->comm_table.need_container[p]);
	 	CukrContainersVectorOfCommentryCreate(&(*mat)->comm_table.give_container[p]);
	  } 
	  
	  /* For each part, see what it needs 
	   * and gives */
	  int own_rows_start = 0; 
	  for (p = 0; p < (*mat)->nparts; p++)
	  {
	  	 CUKR_COMMENTRY comm;
	  	
	  	 /* See when the rows owned by the part end */
	  	 int own_rows_end = own_rows_start + parts_rows[p];
	  		  	 
	  	 /* Needs from below */
	  	 if (p > 0) {
	  	 	comm.size = hb;
	  	 	comm.pos  = shift + own_rows_start - hb;	
	  	 	CukrContainersVectorOfCommentryAppend((*mat)->comm_table.need_container[p], comm);  	 	
	  	 }
	  	 
	  	 /* Needs from above */
	  	 if (p < (*mat)->nparts - 1) {
	  	 	comm.size = hb;
	  	 	comm.pos  = shift + own_rows_end;	 
	  	 	CukrContainersVectorOfCommentryAppend((*mat)->comm_table.need_container[p], comm); 	 	
	  	 }
	  	 
	  	 /* Gives below */
	  	 if (p > 0) {
	  	 	comm.size = hb;
	  	 	comm.pos  = shift + own_rows_start;	
	  	 	CukrContainersVectorOfCommentryAppend((*mat)->comm_table.give_container[p], comm);  	 	
	  	 }
	  	 
	  	 /* Gives above */
	  	 if (p < (*mat)->nparts - 1) {
	  	 	comm.size = hb;
	  	 	comm.pos  = shift + own_rows_end - hb;	 
	  	 	CukrContainersVectorOfCommentryAppend((*mat)->comm_table.give_container[p], comm); 	 	
	  	 }
	  	 
	  	 /* Set the own_rows_start for the next part */
	  	 own_rows_start = own_rows_end;
	  } 
	 	
	 LOGX(DEBUG, "Done creating MP banded matrix %s", name);			
}
