/**
 * @file mat_vec/mp/matrix_gen_mp.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Holds a list of matrix generators.
 * The generators here are indepenent of
 * the matrix format, so the format-specific
 * implementation of the generator will be
 * found elsewere (most probably in the
 * subfolder assigned to the specific format)
 * 
 * Multi-GPU version
 * 
 * The generator will create the matrix,
 * so the matrix should not be created beforehand
 * 
 * Current generators are:
 * 	  - random 
 * 	  - poisson 2D 
 * 
 */

#ifndef CUKR_MATRIX_GEN_MP_H_
#define CUKR_MATRIX_GEN_MP_H_

#include "matrix_mp.h"

// ------------------------------------------------------------------- //

/**
 * \brief Generates a sparse random multi-gpu matrix, 
 *        filled with random elements 
 * \param mat Pointer to the multi-gpu matrix 
 * \param format Matrix format (CSR)
 * \param n Matrix size
 * \param halfb The half-bandwidth of the matrix
 * \param nz_per_row No. of nonzeros per row
 * \param min,max The lower and upper bound for
 *                matrix values
 * \param seed The seed for the random no. generator
 * \param name Matrix name
 */
void CukrMatmpCreateRandom(CUKR_MATRIX_MP* mat, int format, int n, int halfb, int nz_per_row, 
						   CUKR_DOUBLE min, CUKR_DOUBLE max, CUKR_DOUBLE seed, char* name);

// ------------------------------------------------------------------- //
						 
/**
 * \brief Generates a sparse random multi-gpu 
 *        matrix, filled with random elements 
 * \param mat Pointer to the multi-gpu matrix 
 * \param format Matrix format (CSR)
 * \param n Order
 * \param name Matrix name
 */
void CukrMatmpCreatePoisson2D(CUKR_MATRIX_MP* mat, int format, int n, char* name);						 
						 
// ------------------------------------------------------------------- //						 
						 
/**
 * \brief Generates a diagonally-dominated banded
 *        MP matrix, filled with ones but the main 
 * 		  diagonal computed as such as to be dominant 
 * \param mat Pointer to the matrix 
 * \param format Matrix format (CSR)
 * \param rows*cols Matrix order
 * \param hb Halfbandwidth 
 * \param name Matrix name
 */
void CukrMatmpCreateBanded(CUKR_MATRIX_MP* mat, int format, int rows, int cols, int hb, char* name);

/**
 * \brief Generates a diagonally-dominated banded
 *        MP matrix, filled with ones but the main 
 * 		  diagonal computed as such as to be dominant 
 * \param mat Pointer to the matrix 
 * \param format Matrix format (CSR)
 * \param rows*cols Matrix order
 * \param hb Halfbandwidth 
 * \param shift The shift
 * \param name Matrix name
 */
void CukrMatmpCreateBandedWShift(CUKR_MATRIX_MP* mat, int format, int rows, int cols, int hb, int shift, char* name);						
						

#endif /* CUKR_MATRIX_GEN_MP_H_ */
