/**
 * @file mat_vec/mp/vector_mp.c
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines the CUKR_VECTOR_MP abstract
 * interace. It's purpose is to extend
 * a vector across multiple nodes.
 * 
 * The CUKR_VECTOR_MP structure defines
 * a vector which is split in a number 
 * of parts, so that operations can be
 * performed in parallel on multiple PEs.
 * 
 * There are two types of distributed 
 * vectors:
 *  1. distributed
 *  2. shared
 * 
 * A distributed vector spanning n PEs 
 * allocates len(vec) / n elements on
 * each vector. The sections are completely
 * independent.
 * 
 * A shared vector spanning n PEs allocates
 * len(vec) on each PE. This is becase in
 * cases as SpMV, since the matrix is 
 * divident into rows and each row has to be
 * multiply to a full length vector, all the
 * memory has to be there. Most operation will
 * involve the len(vec) / n elements which 
 * actually correspond to that PE, but there 
 * will be few (hopfully) operations that need
 * other elements. Therefore, before such operation
 * the shared vector must SYNCHRONIZE. This 
 * procedure uses a communication table to 
 * copy to each copy of the shared vector assigned
 * to each PE exactly those extra elements that 
 * are needed, in the right places.
 *   
 * To implement the distributed/shared type of
 * vectors, a MP vector has two types of views
 * for the data: cells and rows. A distributed
 * vector has one cell for each PE, while a
 * shared one has one cell and one row for each
 * PE.
 * 
 * A cell is a CUKR_VECTOR_SP that corresponds to
 * the part of the distributed vector that is
 * owned by a particular PE. If there are n
 * PEs, a distributed vector will have n cells,
 * one for each PE. Operations on distributed
 * vectors (e.g. DOT) will work with on the cells.
 * 
 * A row is a CUKR_VECTOR_SP that corresponds to all
 * the memory allocated on that particular PE, 
 * containg both the elements owned by that PE
 * and the ones owned by the rest. The memory
 * allocated to the cell is part of the memory
 * allocated for the row. The position of the
 * cell is in the row is chosen such as the 
 * cell vector correspond to the elements 
 * owned by the particular PE. In a SpMV operation,
 * the vector multiplied by the matrix must be a 
 * shared vector, the multiplication will be 
 * between the matrix and the row, since the full
 * length is needed. The result will can be either
 * a shared or a distributed vector, since it 
 * will be stored in the cell.
 * 
 */


#define LOGGER_NAME "CUKr.mat_vec.mp.vector_mp"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef __MACH__
	#include <malloc.h>
#endif
#include "vector_mp.h"
#include "../../util/log.h"
#include "../../util/macros.h"
#include "../../util/utils.h"
#include "../../threading/threading.h"
#include "../../db/db.h"
#include "../../blas/mp/blas1_mp.h"
#ifdef CUKR_USE_GPU
	#include <cublas.h>
#endif

/** Get access to database */
extern CukrDB cukr_db;

//========================== FORWARD DECLARATIONS ========================= //

void CukrVecmpMallocSetCellPtrs(CUKR_VECTOR_MP vec, int loc, int type, int row);
void CukrVecmpSynchronizePart_Send_CPU(CUKR_VECTOR_SP rvec, CUKR_VECTOR_SP sync, CUKR_CONTAINER_VECTOR_COMMENTRY give, int loc, int type);
void CukrVecmpSynchronizePart_Send_GPU(CUKR_VECTOR_SP rvec, CUKR_VECTOR_SP sync, CUKR_CONTAINER_VECTOR_COMMENTRY give, int loc, int type);
void CukrVecmpSynchronizePart_Recv_CPU(CUKR_VECTOR_SP rvec, CUKR_VECTOR_SP sync, CUKR_CONTAINER_VECTOR_COMMENTRY need, int loc, int type);
void CukrVecmpSynchronizePart_Recv_GPU(CUKR_VECTOR_SP rvec, CUKR_VECTOR_SP sync, CUKR_CONTAINER_VECTOR_COMMENTRY need, int loc, int type);
#ifdef CUKR_USE_GPU
	extern void CukrHandleCudaMemErrors(CUKR_VECTOR_SP vec, cublasStatus status);
#endif



//============================ PUBLIC FUNCTIONS =========================== // 




				/*****************************************/
				//       CONSTRUCTOR / DESTRUCTOR        //
				/*****************************************/         



/**
 * \brief Creates a MP vector, of the 
 *        specified size, on the CPU, with the 
 *        specified type
 * \param vec Pointer to the MP vector 
 *        being created
 * \param n Vector size
 * \param type Vector type (double,
 *        qdouble, single, int)
 * \param shtype The type of sharing 
 *        {distributed,shared}
 * \param name The name of the vector 
 */
void CukrVecmpCreate(CUKR_VECTOR_MP* vec, CUKR_INT n, int type, int shtype, char* name)
{
	char buffer[255];
	int part_sizes[CUKR_MAX_PE];
	int part_size, last_part_size;
	int i, j;
	
	/*
	 * Validate parameters
	 */
	
	/* First validate name */
	if (name == NULL)
		CUKR_FATAL("Vector must have a non-null name");
	
	/* Log */
	#ifdef CUKR_LOGGING
		char shtype_str[255];
		CukrShtypeIdToStr(shtype, shtype_str);
		LOGX(DEBUG, "Creating MP vector %s(%d) with " \
		            "sharing type %s", 
		            name, n, shtype_str);
	#endif	            
	
	/*
	 * Malloc for the MP vector structure
	 */ 
	 
	 /* Alloc memory */
	 (*vec) = NULL;
	 if ( ((*vec) = (CUKR_VECTOR_MP)malloc(sizeof(struct _CUKR_VECTOR_MP))) == NULL) {
		CUKR_FATAL("Error allocating memory for MP vector object ");
	 }
	 else {
		LOG(DEBUG, "Allocated MP vector structure");
	 }
	 
	 /*
	  * Init and set
	  */

	 /* NULL all the rows */
	 for (i = 0; i < CUKR_MAX_PE; i++)
	 	(*vec)->rows[i] = NULL;

	 /* NULL all the cells */
	 for (i = 0; i < CUKR_MAX_PE; i++) 
	 	(*vec)->cells[i] = NULL;	

	 /* NULL the sync vector */
	 (*vec)->sync = NULL;		 	
	 
	 /* Get vector type */
	 int vec_type;
	 if (type == CUKR_DEFAULT)
		vec_type = CukrTypeStrToId(CukrDBGetS(CUKR_DB_SOLVER_PRECISION));
	 else
	 	vec_type = type;
	 
	 /* Set vector size and name */
	 (*vec)->n = n;
	 sprintf((*vec)->name, "%s", name);
	 
	 /*
	  * Partitioning related
	  */
	 
	 /* Set the no. of parts to be equal to
	  * the no. of PEs used */
	 (*vec)->nparts = CukrDBGetI(CUKR_DB_PE); 
	 
	 /* Invalidate the insync flag */
	 (*vec)->insync = CUKR_FALSE;
	 
	 /* Set sharing type */
	 (*vec)->shtype = shtype;
	 
	 /*
	  * Partition into parts
	  */ 
	 
 	  /* Compute the size for the cells */
	  if ((*vec)->nparts == 1) 
	 	 last_part_size = n;
	  else {
		 part_size = n / (*vec)->nparts; 
		 last_part_size = n - ((*vec)->nparts - 1) * part_size;
	  }
		 
	  /* Store the size for part i in part_sizes[i] 
	   * Make all but the last part have part_size 
	   * elements and the last one have last_part_size
	   * elements */
	  for (i = 0; i < (*vec)->nparts; i++) {
	 	 if (i < (*vec)->nparts - 1)	
	 	 	 part_sizes[i] = part_size;
	 	 else
	 		 part_sizes[i] = last_part_size;
	  }		           
	 
	 /* Creating cells. Since the vectors are created by
	  * default on the CPU, without any connection to the 
	  * GPU, this part does not need to be multithreaded. */
	  
	 /* For distributed vectors, only create the
	  * cells as CUKR_VECTORS, which will
	  * automatically allocate memory as well */ 
	 if (shtype == CUKR_SHTYPE_DISTRIBUTED)  
		 for (i = 0; i < (*vec)->nparts; i++)
		 {	
		 	/* Set part name */
		 	sprintf(buffer, "%s.cell%d", name, i);
		 	
		 	/* Create part */ 
		 	CukrVecspCreate(&((*vec)->cells[i]), part_sizes[i], vec_type, buffer);
		 }
		 
	  /* For shared vectors, the malloc is more 
	   * complicated, so we have to alloc not 
	   * vector by vector but a row at a time.
	   * First the row vectors will be 
	   * created and memory automatically allocated. 
	   * Then the cells will be created, but
	   * with no data inside, since the members
	   * will be pointers to the the corresp.
	   * places in the row data members. 
	   * The cells act therefore as viewes to 
	   * the memory owned by the row vector.*/
	 else 
	 {
	 	/* Creating the row vectors */
	 	for (i = 0; i < (*vec)->nparts; i++)
		{	
		 	/* Set row name */
		 	sprintf(buffer, "%s.row%d", name, i);
		 	 
		 	/* Create row vector */
		 	CukrVecspCreate(&((*vec)->rows[i]), n, vec_type, buffer);
		 }
		 
		/* Creating cell vectors */
		for (i = 0; i < (*vec)->nparts; i++)
		{	
		 	/* Set part name */
		 	sprintf(buffer, "%s.cell%d", name, i);
		 	 
		 	/* Create part vectors */
		 	CukrVecspCreate(&((*vec)->cells[i]), part_sizes[i], CUKR_TYPE_NONE, buffer);
		 		
		 	/* Set status for the part vectors, 
		 	 * since they were created with TYPE_NONE
		 	 * so the status could not be automatically 
		 	 * set by CukrVecspCreate*/
		 	((*vec)->cells[i])->status = CukrLocIdAndTypeIdToStatusId(CUKR_LOCATION_CPU, vec_type);	
		} 
		
		/* Set the cell pointers to the right places 
		 * only if the type is not NONE. If the type
		 * is NONE, then the cell pointers will
		 * be set from somewhere else. */
		
		if (type != CUKR_TYPE_NONE) 
			for (i = 0; i < (*vec)->nparts; i++)
				CukrVecmpMallocSetCellPtrs((*vec), CUKR_LOCATION_CPU, vec_type, i);	
			
		/* Create the sync vector */
		sprintf(buffer, "%s.sync", name);
		CukrVecspCreate(&((*vec)->sync), n, vec_type, buffer);	
	 }  	 
	 
	 /* Success message */
	 LOGX(DEBUG, "Successfully created MP vector %s", name);         
}

// ------------------------------------------------------------------- //

/**
 * \brief Creates a clone of a vector. 
 *        The function currently copies
 *        the data in the highes precision
 *        available, and leaves the rest
 *        of the precision invalidated, so
 *        they will be transformed/convered
 *        to only if needed.
 * \param src The source MP vector
 * \param dst Pointer to the destination vector
 * \param name The new name for the destination
 */
void CukrVecmpCreateClone(CUKR_VECTOR_MP src, CUKR_VECTOR_MP* dst, char* name)
{
	int i;
	LOGX(DEBUG, "Cloning MP vector %s to %s ... ",
	            src->name, name);             
	
	/* 
	 * Create the destinantion vector 
	 * Use the first cell, since
	 * all statuses are synchronized
	 */
	 
	int type = GetHighestPrecFromStatusId(src->cells[0]->status);
	CukrVecmpCreate(dst, src->n, type, src->shtype, name);	               
	
	/*
	 * Now copy the data in the highest precision 
	 */
	
	if (type == CUKR_TYPE_DOUBLE)
		CukrCOPY_MP(cukr_db.blasD, src, *dst);
	else if (type == CUKR_TYPE_QDOUBLE)
		CukrCOPY_MP(cukr_db.blasQ, src, *dst);
	else if (type == CUKR_TYPE_SINGLE)
		CukrCOPY_MP(cukr_db.blasS, src, *dst);	
	else
		CUKR_FATALX("Invalid type when cloning vector %s " \
		            "to vector %s: %d", src->name, 
		            (*dst)->name, type);
	
		
	/* Invalidate the insync flag 
	 * for the clone, since the info.
	 * aquired by synchronizing is not
	 * copied. */
	(*dst)->insync = CUKR_FALSE;
		
	LOGX(DEBUG, "Successfully created clone MP "\
				"vector %s ", (*dst)->name); 
}

// ------------------------------------------------------------------- //

/**
 * \brief Frees all the memory used 
 *        by the MP vector
 * \param vec Pointer to the PE vector 
 *        to be destroyed 
 */
void CukrVecmpDestroy(CUKR_VECTOR_MP* vec)
{
	int i, j;
	
	/* Check for non-null vector */
	if (*vec == NULL)
		return;
	
	/* Copy the name, so it doesn't get
	 * destroyed */
	char name[255];
	memcpy(name, (*vec)->name, 255);
	LOGX(DEBUG, "Destroying MP vector %s", name);
	
	/* 
	 * Schedule the distruction of the parts 
	 */
	  
	/* For distributed, destroy the 
	   cells */
	if ((*vec)->shtype == CUKR_SHTYPE_DISTRIBUTED) 
		for (i = 0; i < (*vec)->nparts; i++)
			CukrThreadsRun(i, (void*)CukrVecspDestroy, CUKR_FUNCID_VEC_DESTROY, &((*vec)->cells[i]));	 
	
	/* For shared, destroy the rows, as the cells
	 * only point to here. Then free the structures for 
	 * the cells */
	else
		for (i = 0; i < (*vec)->nparts; i++) {
			/* Destroy the rows with the data */
			CukrThreadsRun(i, (void*)CukrVecspDestroy, CUKR_FUNCID_VEC_DESTROY, &((*vec)->rows[i]));
			
			/* Destroy now the vector structures 
			 * for the cells, as well */
			free((*vec)->cells[i]);		 
		}
			
	/* Wait for the jobs to finish */
	CukrThreadsWaitCompletion((*vec)->nparts);	
	
	/* For shared, destroy the sync
	 * vector as well */
	if ((*vec)->shtype == CUKR_SHTYPE_SHARED)
		CukrVecspDestroy(&((*vec)->sync));  
		
	/* Finally, free the vector structure */
	free(*vec);
	LOGX(DEBUG, "MP vector %s destroyed", name);	
}


			
				/*****************************************/
				//            MALLOC / FREE              // 
				/*****************************************/ 
				
								      


/**
 * \brief Allocate memory for a MP vector. The 
 *        function checks if the pointer 
 *        corresponding to that particular location
 *        and precision is NULL of not. If not NULL,
 *        on malloc will be performed. Therefore, this
 *        function can be called "just in case" to 
 *        make sure memory allocation where it 
 *        should be.
 * \param vec The vector structure for which
 *        memory is being allocated
 * \param loc Vector location {CPU,GPU}
 * \param type Vector type 
 *        {double, qdouble, single, int}
 */
void CukrVecmpMalloc(CUKR_VECTOR_MP vec, int loc, int type)
{	
	int i;
    
    /*
     * Do a check at this level, to see if the
     * malloc is realy needed or not. The 
     * check will be done anyway at the lowest
     * level, but it will mean a lot of wasted 
     * time. Should be enough to check for the
     * first cell, all should be synhronized 
     * anyway. 
     * This is only a preliminary test: we check
     * the status. If OK, then of course there 
     * is no need to malloc. If not OK, the 
     * memory might still be there, but without
     * the status set. This will the seen at the
     * lowest (CUKR_VECTOR_SP) level. */
     
    int dest_status = CukrLocIdAndTypeIdToStatusId(loc, type);
    if (vec->cells[0]->status & dest_status)
    	return;
           	
	/* For distributed, malloc the cells.
	 * This has to be done done multi-threaded,
	 * since cells might be located on different 
	 * GPUs. */
	if (vec->shtype == CUKR_SHTYPE_DISTRIBUTED) {
		
		/* Malloc the cells */
		for (i = 0; i < vec->nparts; i++)
			CukrThreadsRun(i, (void*)CukrVecspMalloc, CUKR_FUNCID_VEC_MALLOC, vec->cells[i], loc, type);
		
		/* Wait for the jobs to finish */
		CukrThreadsWaitCompletion(vec->nparts);	  	
	}
	/* For shared, malloc the rows then set
	 * the cell pointers to the right places.
	 * This has to be done done multi-threaded,
	 * since rows might be located on different 
	 * GPUs.  */
	else 
	{
		/* Malloc the rows */ 
		for (i = 0; i < vec->nparts; i++)
			CukrThreadsRun(i, (void*)CukrVecspMalloc, CUKR_FUNCID_VEC_MALLOC, vec->rows[i], loc, type);
		
		/* Wait for the jobs to finish */
		CukrThreadsWaitCompletion(vec->nparts);	  
		
		/* Set the cell poiters to the right places */
		for (i = 0; i < vec->nparts; i++)
			CukrVecmpMallocSetCellPtrs(vec, loc, type, i);	
	}		 
}   




				/*****************************************/
				//              TRANSFER                 // 
				/*****************************************/ 




/** 
 * \brief Changes the location of a MP vector
 *        by copying the data from one location
 *        to another. An MP vector does not hold
 *        a status variable by itself; it's status
 *        (location + precision) is the status of
 *        the rows and cells. Therefore, a transfer
 *        operation on an MP vector is basically
 *        a transfer operation of all of its parts.
 *        All parts (both cells and rows) are and 
 *        must be moved in the same way, at the same
 *        time. This is guaranteed by the transfered
 *        procedure, so one can query the status of 
 *        an MP vector by querying, for example, 
 *        the status of the first cell, which is 
 *        always sure to exist. 
 * 		  If the vector has valid data at the destination
 *        location and the destination time, nothing will
 *        happen. As such, one can use this function
 *        as a "precation" that data is indeed in the 
 *        required precision and location. 
 * \param vec The MP vector being transferred 
 * \param dest_loc Vector destination location {CPU,GPU}
 * \param dest_type Vector destination type 
 *        (double, qdouble, single, int)
 */

void CukrVecmpTransfer(CUKR_VECTOR_MP vec, int dest_loc, int dest_type)
{
	int i;
    
    /*
     * Do a check at this level, to see if the
     * transfer is realy needed or not. The 
     * check will be done anyway at the lowest
     * level, but it will mean a lot of wasted 
     * time. Should be enough to check for the
     * first cell, all should be synhronized 
     * anyway. */     
    int dest_status = CukrLocIdAndTypeIdToStatusId(dest_loc, dest_type);
    if (vec->cells[0]->status & dest_status)
    	return;
           
    /* Do the malloc first.
     * For the distributed type, it doesn't 
     * matter if we do it here or do it inside
     * the CukrVecmpTransfer function.
     * For the shared type, we have to do it 
     * before the individual transfer in order
     * to do the special malloc (malloc for row
     * followed by setting cell pointers). */                  
    CukrVecmpMalloc(vec, dest_loc, dest_type);
    
    /* When converting to a different type,
     * memory has to be allocated on host for the data 
     * conversion which has to take place before
     * transfer. For shared type, this has to be
     * a vecmg-malloc. If the destination location
     * is on the CPU, then this is not needed. */
    if (dest_loc != CUKR_LOCATION_CPU)    
    	CukrVecmpMalloc(vec, CUKR_LOCATION_CPU, dest_type); 
    	     
           
	/* Schedule transfers
	 * Valid data (data owned by each PE) is 
	 * only contained in the cells (a row might
	 * participate in an operation, but the result
	 * of the operation will go the cell corresp.
	 * to the PE performing the operation).
	 * Therefore, it is enough to transfer cells. */
	for (i = 0; i < vec->nparts; i++)
		CukrThreadsRun(i, (void*)CukrVecspTransfer, CUKR_FUNCID_VEC_TRANSFER, 
		    	           vec->cells[i], dest_loc, dest_type);		
	
	/* Wait for the jobs to finish */
	CukrThreadsWaitCompletion(vec->nparts);
		
	/* For the shared case, we need to set the
	 * status of the row vectors to the status
	 * of the cells that were just moved, in order
	 * to keep all in sync. */
	if (vec->shtype == CUKR_SHTYPE_SHARED) 
		for (i = 0; i < vec->nparts; i++)
			vec->rows[i]->status = vec->cells[i]->status; 	
			
	/* Invalidate the insync flag, since only
	 * cells are transferred, so the data 
	 * aquired by synchronizing is lost. */
	vec->insync = CUKR_FALSE;		
}

// ------------------------------------------------------------------- //

/**
 * \brief Syncronize a shared vector, using
 *        the specified comm-table. A shared 
 *        vector must be synchronized BEFORE 
 *        operations like SpMV which require 
 *        data from different PEs. The function
 *		  uses the communication table to know
 *        which data has to be sent where. 
 *        Since GPUs cannot currently do P2P
 *        communication, temporary storage in
 *        the require precision, on the CPU, 
 *        is required. 
 * 		  Synchronization is done in two steps:
 *        1. Data needed by destination PEs is 
 *           copied (sent) from the source PEs to
 *           the member with appropriate precision
 *           in the sync vector.
 *        2. All needed data, which is now found 
 *           in the sync vector, is copied back
 *           (received) by the PEs that need it.	
 * \param vec The shared vector to be synchronized
 * \param commtable The commtable describing the
 *        synchronization
 * \param The location to be synchonized  
 * \param The data-type to be synchronize 
 */
void CukrVecmpSynchronize(CUKR_VECTOR_MP vec, CUKR_MATRIX_MP_COMMTABLE commtable, int loc, int type)
{
	int i;
	
	/* Error if not shared */
	if (vec->shtype != CUKR_SHTYPE_SHARED)
		CUKR_FATALX("Cannot synchronize MP vector %s since it's" \
		            "not a shared vector", vec->name);
	
	/* If already in sync, do nothing */
	if (vec->insync) {
		LOGX(DEBUG, "MP vector %s is already in sync", vec->name);
		return;
	}	
	
	/* Make sure the memory for the sync
	 * vector is allocated on the CPU. Check the
	 * status first, so not waste any time. */
	int dest_status = CukrLocIdAndTypeIdToStatusId(CUKR_LOCATION_CPU, type); 
	if (!(vec->sync->status & dest_status)) 
		CukrVecspMalloc(vec->sync, CUKR_LOCATION_CPU, type);
	
	/* Schedule the execution of sending part */
	for (i = 0; i < vec->nparts; i++)
		CukrThreadsRun(i, (void*)CukrVecmpSynchronizePart_Send, CUKR_FUNCID_VECMP_SYNC_SEND, 
		                         vec->rows[i], vec->sync, commtable.give_container[i], loc, type);
		                         
	/* Wait for the jobs to finish */
	CukrThreadsWaitCompletion(vec->nparts);	  
	
	/* Schedule the execution of receiving part */
	for (i = 0; i < vec->nparts; i++)
		CukrThreadsRun(i, (void*)CukrVecmpSynchronizePart_Recv, CUKR_FUNCID_VECMP_SYNC_RECV, 
		                         vec->rows[i], vec->sync, commtable.need_container[i], loc, type);
		                         
	/* Wait for the jobs to finish */
	CukrThreadsWaitCompletion(vec->nparts);	                        
	
	/* Validate the insync flag */
	vec->insync = CUKR_TRUE; 
}

// ------------------------------------------------------------------- //

/**
 * \brief Data needed by destination PEs is 
 *        copied (sent) from the source PEs to
 *        the member with appropriate precision
 *        in the sync vector.
 * \param rvec The part (row) of the shared vector
 *        to be synchronized
 * \param sync The temporary vector used for 
 *        synchronization
 * \param give The comm-entry container describing 
 *        what data has to be sent (given)
 * \param The location to be synchonized  
 * \param The data-type to be synchronized
 */

void CukrVecmpSynchronizePart_Send(CUKR_VECTOR_SP rvec, CUKR_VECTOR_SP sync, CUKR_CONTAINER_VECTOR_COMMENTRY give, int loc, int type)
{	 
	/* If on the CPU */
	if (loc == CUKR_LOCATION_CPU)
		CukrVecmpSynchronizePart_Send_CPU(rvec, sync, give, loc, type);
	
	/* If on the GPU */	
	else if (loc == CUKR_LOCATION_GPU) 
		CukrVecmpSynchronizePart_Send_GPU(rvec, sync, give, loc, type);
		
	/* If none of the above, error */
	else {
		int status = CukrLocIdAndTypeIdToStatusId(loc, type);
		char status_str[255], status_str2[255];
		CukrStatusIdToStr(status, status_str);
		CukrStatusIdToStr(rvec->status, status_str2);
		CUKR_FATALX("Trying to synchronize part (send) but no "\
		            "valid data found with status %s in vector %s. " \
		            "The current status is %s",
		            status_str, rvec->name, status_str2);
	}
} 


// ------------------------------------------------------------------- //

/**
 * \brief All needed data, which is now found 
 *        in the sync vector, is copied back
 *        (received) by the PEs that need it
 * \param rvec The part (row) of the shared vector
 *        to be synchronized
 * \param sync The temporary vector used for 
 *        synchronization
 * \param give The comm-entry container describing 
 *        what data has to be sent (given)
 * \param The location to be synchonized  
 * \param The data-type to be synchronize
 */
void CukrVecmpSynchronizePart_Recv(CUKR_VECTOR_SP rvec, CUKR_VECTOR_SP sync, CUKR_CONTAINER_VECTOR_COMMENTRY need, int loc, int type)
{	
	/* If on the CPU */
	if (loc == CUKR_LOCATION_CPU)
		CukrVecmpSynchronizePart_Recv_CPU(rvec, sync, need, loc, type);
	
	/* If on the GPU */
	else if (loc == CUKR_LOCATION_GPU)
		CukrVecmpSynchronizePart_Recv_GPU(rvec, sync, need, loc, type);
		
	/* If none of the above, error */
	else {
		int status = CukrLocIdAndTypeIdToStatusId(loc, type);
		char status_str[255], status_str2[255];
		CukrStatusIdToStr(status, status_str);
		CukrStatusIdToStr(rvec->status, status_str2);
		CUKR_FATALX("Trying to synchronize part (recv) but no "\
		            "valid data found with status %s in vector %s. " \
		            "The current status is %s",
		            status_str, rvec->name, status_str2);
	}	
} 




				/*****************************************/
				//               GET/SET                 // 
				/*****************************************/ 





/**
 * \brief Returns the element found at the
 *        position <index> in the vector vec. 
 * \param vec The vector
 * \param index The index where to put the value
 * \param type An existing data type in which to 
 * 		  store the value or default to read from
 *        the highest available one
 * \return val The value of the element
 */
CUKR_DOUBLE CukrVecmpGetElement(CUKR_VECTOR_MP vec, int index, int type)
{
	int i, lo = 0;
	for (i = 0; i < vec->nparts; i++) {
		if (index >= lo && index < lo + vec->cells[i]->n) {
			CukrThreadsRun(i, (void*)CukrVecspGetElement, CUKR_FUNCID_VEC_GET_ELEMENT, vec->cells[i], index - lo, type);
			CukrThreadsWaitCompletionForThread(i);
			return CukrThreadsGetRetval_double(i);
		}
		lo += vec->cells[i]->n;
	}
	
	/* If still here, error */
	CUKR_FATALX("Invalid element index (%d) accessed from "\
		        "vector %s", index, vec->name);	
}

// ------------------------------------------------------------------- //

/**
 * \brief Sets the element with index <index>
 *        to the value specified by <val> 
 *        Currently, setting can only
 *        be done on the CPU, so data will be 
 *        set on the CPU and the status set to
 *        there. 
 * \param vec The vector
 * \param index The index where to put the value
 * \param val The value the vector 
 *        elements are set to
 * \param type An existing data type in which to 
 * 		  store the value or default for all 
 *        available precisions
 */
void CukrVecmpSetElement(CUKR_VECTOR_MP vec, int index, CUKR_DOUBLE val, int type)
{
	int i, lo = 0;
	for (i = 0; i < vec->nparts; i++) {
		if (index >= lo && index < lo + vec->cells[i]->n) {
			CukrThreadsRun(i, (void*)CukrVecspSetElement, CUKR_FUNCID_VEC_SET_ELEMENT, vec->cells[i], index - lo, val, type);
			CukrThreadsWaitCompletionForThread(i);
			return;
		}
		lo += vec->cells[i]->n;
	}
	
	/* If still here, error */
	CUKR_FATALX("Invalid element index (%d) accessed from "\
		        "vector %s", index, vec->name);		
}


/**
 * \todo  Should make a simple kernel that
 *        sets all the values of a vector to
 *        a specicific value, so that we can
 *        set directly on the GPU. 
 * \brief Sets all the elements
 *        of the vector to <val>. 
 *        Currently, setting can only
 *        be done on the CPU, so data will be 
 *        set on the CPU and the status set to
 *        there. 
 * \param vec The vector
 * \param val The value the vector 
 *        elements are set to
 * \param type An existing data type in which to 
 * 		  store the values or default for the 
 *        highest one available
 */
void CukrVecmpSetFromVal(CUKR_VECTOR_MP vec, CUKR_DOUBLE val, int type)
{
	int i;
	
	/* Set the value for all cells
	 * to the specified value, regardless of the
	 * sharing type */
	for (i = 0; i < vec->nparts; i++)
		CukrVecspSetFromVal(vec->cells[i], val, type); 	
		
	/* Invalidate the insync flag */
	vec->insync = CUKR_FALSE;	
}	

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets all elements of a 
 *        vector from a double precision vector
 *        <val>. The vector must reside on the host.
 *        Since data
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecmpSetFromVectorD(CUKR_VECTOR_MP vec, int index, int count, CUKR_DOUBLE* val, int loc)
{
	int i, lo;
	
	/* 
	 * Make sure the vector has data in 
	 * the required precision and at the
	 * required location
	 */
	 
	if (loc == CUKR_LOCATION_CPU || loc == CUKR_DEFAULT)
		CukrVecmpMalloc(vec, CUKR_LOCATION_CPU, CUKR_TYPE_DOUBLE);
	else if (loc == CUKR_LOCATION_GPU)		
		CukrVecmpMalloc(vec, CUKR_LOCATION_GPU, CUKR_TYPE_DOUBLE);
	else
		CUKR_FATALX("Invalid location (%d) specified for "\
		            "vector %s while setting from vector "\
		            "of doubles", loc, vec->name);	
	/*
	 * Special case where index = 0 and count = vec->n
	 */
	 
	 if (index == 0 && count == vec->n)
	 {
	 	lo = 0;
	 	for (i = 0; i < vec->nparts; i++) {
	 		CukrThreadsRun(i, (void*)CukrVecspSetFromVectorD, CUKR_FUNCID_VEC_SET_FROMVEC_D, 
	 		                   vec->cells[i], 0,  vec->cells[i]->n, &val[lo], loc);
	 		lo += vec->cells[i]->n;                   
	 	}
	 }
	
	
	/* 
	 * Find first and last parts 
	 */
	 
	int start_part = -1, end_part = -1, start_part_index;
	lo = 0;
	for (i = 0; i < vec->nparts; i++) 
	{			
		/* If start conditions are met,
		 * set start (once) */
		if (index - lo < vec->cells[i]->n && start_part == -1) {
				start_part = i;
				start_part_index = index - lo;
		}				
				
		/* If last part detected, set and exit */	
		if (index + count < lo + vec->cells[i]->n) {
			end_part = i;	
			break;	
		}
						
		lo += vec->cells[i]->n;
	} 
	
	/* If start_part still -1, error */
	if (start_part == -1) 
		CUKR_FATALX("Invalid index (%d) accessed from "\
			         "vector %s when setting from vector", 
			         index, vec->name);	 
	
	/* If end part still -1, then it is the 
	 * last part */
	if (end_part == -1)
		end_part = vec->nparts - 1; 
		
	/* 
	 * If all in the same part 
	 */
	
	/* If start_part and end_part are the same,
	 * set count elements and be done */	
	if (start_part == end_part) {
		CukrThreadsRun(start_part, (void*)CukrVecspSetFromVectorD, CUKR_FUNCID_VEC_SET_FROMVEC_D, 
		                           vec->cells[start_part], start_part_index, count, val, loc);
		CukrThreadsWaitCompletionForThread(start_part);	
		return;
	} 
	 
	/* 
	 * If not, do the parts 
	 */
	 
	/* For start part, set how many elements
	 * are left from index to end */	
	int val_index   = 0;  
	int vals_to_set = vec->cells[start_part]->n - start_part_index;
	CukrThreadsRun(start_part, (void*)CukrVecspSetFromVectorD, CUKR_FUNCID_VEC_SET_FROMVEC_D, 
	                            vec->cells[start_part], start_part_index, vals_to_set, val, loc); 	                            
	val_index += vals_to_set;
	
	/* The ones between start and end are full */	
	for (i = start_part + 1; i < end_part; i++) {
		CukrThreadsRun(i, (void*)CukrVecspSetFromVectorD, CUKR_FUNCID_VEC_SET_FROMVEC_D, 
	                      vec->cells[i], 0, vec->cells[i]->n, &val[val_index], loc); 
	    val_index += vec->cells[i]->n;                  	   
	} 
	
	/* The last one has the leftovers */
	vals_to_set = count - val_index;
	CukrThreadsRun(end_part, (void*)CukrVecspSetFromVectorD, CUKR_FUNCID_VEC_SET_FROMVEC_D, 
	                         vec->cells[end_part], 0, vals_to_set, &val[val_index], loc); 
	                         
	/* Wait for the threads to finish */
	for (i = start_part; i < end_part; i++)
		 CukrThreadsWaitCompletionForThread(i);
		 
	/* Invalidate the insync flag */
	vec->insync = CUKR_FALSE;	                        		
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets all elements of a 
 *        vector from a quasi-double precision 
 *        vector <val>.
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecmpSetFromVectorQ(CUKR_VECTOR_MP vec, int index, int count, CUKR_SINGLE* hval, CUKR_SINGLE* tval, int loc)
{
	int i, lo;
	
	/* 
	 * Make sure the vector has data in 
	 * the required precision and at the
	 * required location
	 */
	 
	if (loc == CUKR_LOCATION_CPU || loc == CUKR_DEFAULT)
		CukrVecmpMalloc(vec, CUKR_LOCATION_CPU, CUKR_TYPE_QDOUBLE);
	else if (loc == CUKR_LOCATION_GPU)		
		CukrVecmpMalloc(vec, CUKR_LOCATION_GPU, CUKR_TYPE_QDOUBLE);
	else
		CUKR_FATALX("Invalid location (%d) specified for "\
		            "vector %s while setting from vector "\
		            "of quasi-doubles", loc, vec->name);			
	
	/*
	 * Special case where index = 0 and count = vec->n
	 */
	 
	 if (index == 0 && count == vec->n)
	 {
	 	lo = 0;
	 	for (i = 0; i < vec->nparts; i++) {
	 		CukrThreadsRun(i, (void*)CukrVecspSetFromVectorQ, CUKR_FUNCID_VEC_SET_FROMVEC_Q, 
	 		                   vec->cells[i], 0,  vec->cells[i]->n, &hval[lo], &tval[lo], loc);
	 		lo += vec->cells[i]->n;                   
	 	}
	 }
	
	
	/* 
	 * Find first and last parts 
	 */
	 
	int start_part = -1, end_part = -1, start_part_index;
	lo = 0;
	for (i = 0; i < vec->nparts; i++) 
	{			
		/* If start conditions are met,
		 * set start (once) */
		if (index - lo < vec->cells[i]->n && start_part == -1) {
				start_part = i;
				start_part_index = index - lo;
		}				
				
		/* If last part detected, set and exit */	
		if (index + count < lo + vec->cells[i]->n) {
			end_part = i;	
			break;	
		}
						
		lo += vec->cells[i]->n;
	} 
	
	/* If start_part still -1, error */
	if (start_part == -1) 
		CUKR_FATALX("Invalid index (%d) accessed from "\
			         "vector %s when setting from vector", 
			         index, vec->name);	 
	
	/* If end part still -1, then it is the 
	 * last part */
	if (end_part == -1)
		end_part = vec->nparts - 1; 
		
	/* 
	 * If all in the same part 
	 */
	
	/* If start_part and end_part are the same,
	 * set count elements and be done */	
	if (start_part == end_part) {
		CukrThreadsRun(start_part, (void*)CukrVecspSetFromVectorQ, CUKR_FUNCID_VEC_SET_FROMVEC_Q, 
		                           vec->cells[start_part], start_part_index, count, hval, tval, loc);
		CukrThreadsWaitCompletionForThread(start_part);	
		return;
	} 
	 
	/* 
	 * If not, do the parts 
	 */
	 
	/* For start part, set how many elements
	 * are left from index to end */	
	int val_index   = 0;  
	int vals_to_set = vec->cells[start_part]->n - start_part_index;
	CukrThreadsRun(start_part, (void*)CukrVecspSetFromVectorQ, CUKR_FUNCID_VEC_SET_FROMVEC_Q, 
	                            vec->cells[start_part], start_part_index, vals_to_set, hval, tval, loc); 	                            
	val_index += vals_to_set;
	
	/* The ones between start and end are full */	
	for (i = start_part + 1; i < end_part; i++) {
		CukrThreadsRun(i, (void*)CukrVecspSetFromVectorQ, CUKR_FUNCID_VEC_SET_FROMVEC_Q, 
	                      vec->cells[i], 0, vec->cells[i]->n, &hval[val_index], &tval[val_index], loc); 
	    val_index += vec->cells[i]->n;                  	   
	} 
	
	/* The last one has the leftovers */
	vals_to_set = count - val_index;
	CukrThreadsRun(end_part, (void*)CukrVecspSetFromVectorQ, CUKR_FUNCID_VEC_SET_FROMVEC_Q, 
	                         vec->cells[end_part], 0, vals_to_set, &hval[val_index], &tval[val_index], loc); 
	                         
	/* Wait for the threads to finish */
	for (i = start_part; i < end_part; i++)
		 CukrThreadsWaitCompletionForThread(i);
		 
	/* Invalidate the insync flag */
	vec->insync = CUKR_FALSE;	 
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets all elements of a 
 *        vector from a single precision vector
 *        <val>.
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecmpSetFromVectorS(CUKR_VECTOR_MP vec, int index, int count, CUKR_SINGLE* val, int loc)
{
	int i, lo;
	
	/* 
	 * Make sure the vector has data in 
	 * the required precision and at the
	 * required location
	 */
	 
	if (loc == CUKR_LOCATION_CPU || loc == CUKR_DEFAULT)
		CukrVecmpMalloc(vec, CUKR_LOCATION_CPU, CUKR_TYPE_SINGLE);
	else if (loc == CUKR_LOCATION_GPU)		
		CukrVecmpMalloc(vec, CUKR_LOCATION_GPU, CUKR_TYPE_SINGLE);
	else
		CUKR_FATALX("Invalid location (%d) specified for "\
		            "vector %s while setting from vector "\
		            "of singles", loc, vec->name);	
	
	/*
	 * Special case where index = 0 and count = vec->n
	 */
	 
	 if (index == 0 && count == vec->n)
	 {
	 	lo = 0;
	 	for (i = 0; i < vec->nparts; i++) {
	 		CukrThreadsRun(i, (void*)CukrVecspSetFromVectorS, CUKR_FUNCID_VEC_SET_FROMVEC_S, 
	 		                   vec->cells[i], 0,  vec->cells[i]->n, &val[lo], loc);
	 		lo += vec->cells[i]->n;                   
	 	}
	 }	
	
	/* 
	 * Find first and last parts 
	 */
	 
	int start_part = -1, end_part = -1, start_part_index;
	lo = 0;
	for (i = 0; i < vec->nparts; i++) 
	{			
		/* If start conditions are met,
		 * set start (once) */
		if (index - lo < vec->cells[i]->n && start_part == -1) {
				start_part = i;
				start_part_index = index - lo;
		}				
				
		/* If last part detected, set and exit */	
		if (index + count < lo + vec->cells[i]->n) {
			end_part = i;	
			break;	
		}
						
		lo += vec->cells[i]->n;
	} 
	
	/* If start_part still -1, error */
	if (start_part == -1) 
		CUKR_FATALX("Invalid index (%d) accessed from "\
			         "vector %s when setting from vector", 
			         index, vec->name);	 
	
	/* If end part still -1, then it is the 
	 * last part */
	if (end_part == -1)
		end_part = vec->nparts - 1; 
		
	/* 
	 * If all in the same part 
	 */
	
	/* If start_part and end_part are the same,
	 * set count elements and be done */	
	if (start_part == end_part) {
		CukrThreadsRun(start_part, (void*)CukrVecspSetFromVectorS, CUKR_FUNCID_VEC_SET_FROMVEC_S, 
		                           vec->cells[start_part], start_part_index, count, val, loc);
		CukrThreadsWaitCompletionForThread(start_part);	
		return;
	} 
	 
	/* 
	 * If not, do the parts 
	 */
	 
	/* For start part, set how many elements
	 * are left from index to end */	
	int val_index   = 0;  
	int vals_to_set = vec->cells[start_part]->n - start_part_index;
	CukrThreadsRun(start_part, (void*)CukrVecspSetFromVectorS, CUKR_FUNCID_VEC_SET_FROMVEC_S, 
	                            vec->cells[start_part], start_part_index, vals_to_set, val, loc); 	                            
	val_index += vals_to_set;
	
	/* The ones between start and end are full */	
	for (i = start_part + 1; i < end_part; i++) {
		CukrThreadsRun(i, (void*)CukrVecspSetFromVectorS, CUKR_FUNCID_VEC_SET_FROMVEC_S, 
	                      vec->cells[i], 0, vec->cells[i]->n, &val[val_index], loc); 
	    val_index += vec->cells[i]->n;                  	   
	} 
	
	/* The last one has the leftovers */
	vals_to_set = count - val_index;
	CukrThreadsRun(end_part, (void*)CukrVecspSetFromVectorS, CUKR_FUNCID_VEC_SET_FROMVEC_S, 
	                         vec->cells[end_part], 0, vals_to_set, &val[val_index], loc); 
	                         
	/* Wait for the threads to finish */
	for (i = start_part; i < end_part; i++)
		 CukrThreadsWaitCompletionForThread(i);
		 
	/* Invalidate the insync flag */
	vec->insync = CUKR_FALSE;	 
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets all elements of a 
 *        vector from a integer vector <val>. 
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecmpSetFromVectorI(CUKR_VECTOR_MP vec, int index, int count, CUKR_INT* val, int loc)
{
	int i, lo;
	
	/* 
	 * Make sure the vector has data in 
	 * the required precision and at the
	 * required location
	 */
	 
	if (loc == CUKR_LOCATION_CPU || loc == CUKR_DEFAULT)
		CukrVecmpMalloc(vec, CUKR_LOCATION_CPU, CUKR_TYPE_INT);
	else if (loc == CUKR_LOCATION_GPU)		
		CukrVecmpMalloc(vec, CUKR_LOCATION_GPU, CUKR_TYPE_INT);
	else
		CUKR_FATALX("Invalid location (%d) specified for "\
		            "vector %s while setting from vector "\
		            "of integers", loc, vec->name);		
	
	/*
	 * Special case where index = 0 and count = vec->n
	 */
	 
	 if (index == 0 && count == vec->n)
	 {
	 	lo = 0;
	 	for (i = 0; i < vec->nparts; i++) {
	 		CukrThreadsRun(i, (void*)CukrVecspSetFromVectorI, CUKR_FUNCID_VEC_SET_FROMVEC_I, 
	 		                   vec->cells[i], 0,  vec->cells[i]->n, &val[lo], loc);
	 		lo += vec->cells[i]->n;                   
	 	}
	 }
	
	
	/* 
	 * Find first and last parts 
	 */
	 
	int start_part = -1, end_part = -1, start_part_index;
	lo = 0;
	for (i = 0; i < vec->nparts; i++) 
	{			
		/* If start conditions are met,
		 * set start (once) */
		if (index - lo < vec->cells[i]->n && start_part == -1) {
				start_part = i;
				start_part_index = index - lo;
		}				
				
		/* If last part detected, set and exit */	
		if (index + count < lo + vec->cells[i]->n) {
			end_part = i;	
			break;	
		}
						
		lo += vec->cells[i]->n;
	} 
	
	/* If start_part still -1, error */
	if (start_part == -1) 
		CUKR_FATALX("Invalid index (%d) accessed from "\
			         "vector %s when setting from vector", 
			         index, vec->name);	 
	
	/* If end part still -1, then it is the 
	 * last part */
	if (end_part == -1)
		end_part = vec->nparts - 1; 
		
	/* 
	 * If all in the same part 
	 */
	
	/* If start_part and end_part are the same,
	 * set count elements and be done */	
	if (start_part == end_part) {
		CukrThreadsRun(start_part, (void*)CukrVecspSetFromVectorI, CUKR_FUNCID_VEC_SET_FROMVEC_I, 
		                           vec->cells[start_part], start_part_index, count, val, loc);
		CukrThreadsWaitCompletionForThread(start_part);	
		return;
	} 
	 
	/* 
	 * If not, do the parts 
	 */
	 
	/* For start part, set how many elements
	 * are left from index to end */	
	int val_index   = 0;  
	int vals_to_set = vec->cells[start_part]->n - start_part_index;
	CukrThreadsRun(start_part, (void*)CukrVecspSetFromVectorI, CUKR_FUNCID_VEC_SET_FROMVEC_I, 
	                            vec->cells[start_part], start_part_index, vals_to_set, val, loc); 	                            
	val_index += vals_to_set;
	
	/* The ones between start and end are full */	
	for (i = start_part + 1; i < end_part; i++) {
		CukrThreadsRun(i, (void*)CukrVecspSetFromVectorI, CUKR_FUNCID_VEC_SET_FROMVEC_I, 
	                      vec->cells[i], 0, vec->cells[i]->n, &val[val_index], loc); 
	    val_index += vec->cells[i]->n;                  	   
	} 
	
	/* The last one has the leftovers */
	vals_to_set = count - val_index;
	CukrThreadsRun(end_part, (void*)CukrVecspSetFromVectorI, CUKR_FUNCID_VEC_SET_FROMVEC_I, 
	                         vec->cells[end_part], 0, vals_to_set, &val[val_index], loc); 
	                         
	/* Wait for the threads to finish */
	for (i = start_part; i < end_part; i++)
		 CukrThreadsWaitCompletionForThread(i);
		 
	/* Invalidate the insync flag */
	vec->insync = CUKR_FALSE;	 
}
// ------------------------------------------------------------------- //
 
/**
 * \brief Reads count elements from a cukr vector
 *        to a double precision vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecmpGetToVectorD(CUKR_VECTOR_MP vec, int index, int count, CUKR_DOUBLE* val, int loc)
{
	int i, lo;
	
	/* 
	 * Make sure the vector has data in 
	 * the required precision and at the
	 * required location
	 */
	 
	if (loc == CUKR_LOCATION_CPU || loc == CUKR_DEFAULT)
		CukrVecmpTransfer(vec, CUKR_LOCATION_CPU, CUKR_TYPE_DOUBLE);
	else if (loc == CUKR_LOCATION_GPU)		
		CukrVecmpTransfer(vec, CUKR_LOCATION_GPU, CUKR_TYPE_DOUBLE);
	else
		CUKR_FATALX("Invalid location (%d) specified for "\
		            "vector %s while setting from vector "\
		            "of doubles", loc, vec->name);	
	
	/*
	 * Special case where index = 0 and count = vec->n
	 */
	 
	 if (index == 0 && count == vec->n)
	 {  
	 	lo = 0;
	 	for (i = 0; i < vec->nparts; i++) {
	 		CukrThreadsRun(i, (void*)CukrVecspGetToVectorD, CUKR_FUNCID_VEC_GET_TOVEC_D, 
	 		                   vec->cells[i], 0,  vec->cells[i]->n, &val[lo], loc);
	 		lo += vec->cells[i]->n;                   
	 	}
	 	CukrThreadsWaitCompletion(vec->nparts);
	 	return;
	 }
	
	/* 
	 * Find first and last parts 
	 */
	 
	int start_part = -1, end_part = -1, start_part_index;
	lo = 0;
	for (i = 0; i < vec->nparts; i++) 
	{			
		/* If start conditions are met,
		 * set start (once) */
		if (index - lo < vec->cells[i]->n && start_part == -1) {
				start_part = i;
				start_part_index = index - lo;
		}				
				
		/* If last part detected, set and exit */	
		if (index + count < lo + vec->cells[i]->n) {
			end_part = i;	
			break;	
		}
						
		lo += vec->cells[i]->n;
	} 
	
	/* If start_part still -1, error */
	if (start_part == -1) 
		CUKR_FATALX("Invalid index (%d) accessed from "\
			         "vector %s when setting from vector", 
			         index, vec->name);	 
	
	/* If end part still -1, then it is the 
	 * last part */
	if (end_part == -1)
		end_part = vec->nparts - 1; 
		
	/* 
	 * If all in the same part 
	 */
	
	/* If start_part and end_part are the same,
	 * set count elements and be done */	
	if (start_part == end_part) {
		CukrThreadsRun(start_part, (void*)CukrVecspGetToVectorD, CUKR_FUNCID_VEC_GET_TOVEC_D, 
		                           vec->cells[start_part], start_part_index, count, val, loc);
		CukrThreadsWaitCompletionForThread(start_part);	
		return;
	} 
	  
	/* 
	 * If not, do the parts 
	 */
	 
	/* For start part, set how many elements
	 * are left from index to end */	
	int val_index   = 0;  
	int vals_to_set = vec->cells[start_part]->n - start_part_index;
	CukrThreadsRun(start_part, (void*)CukrVecspGetToVectorD, CUKR_FUNCID_VEC_GET_TOVEC_D, 
	                            vec->cells[start_part], start_part_index, vals_to_set, val, loc); 	                            
	val_index += vals_to_set;
	
	/* The ones between start and end are full */	
	for (i = start_part + 1; i < end_part; i++) {
		CukrThreadsRun(i, (void*)CukrVecspGetToVectorD, CUKR_FUNCID_VEC_GET_TOVEC_D, 
	                      vec->cells[i], 0, vec->cells[i]->n, &val[val_index], loc); 
	    val_index += vec->cells[i]->n;                  	   
	} 
	
	/* The last one has the leftovers */
	vals_to_set = count - val_index;
	CukrThreadsRun(end_part, (void*)CukrVecspGetToVectorD, CUKR_FUNCID_VEC_GET_TOVEC_D, 
	                         vec->cells[end_part], 0, vals_to_set, &val[val_index], loc); 
	                         
	/* Wait for the threads to finish */
	for (i = start_part; i < end_part; i++)
		 CukrThreadsWaitCompletionForThread(i);                 
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to a quasi-double precision vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecmpGetToVectorQ(CUKR_VECTOR_MP vec, int index, int count, CUKR_SINGLE* hval, CUKR_SINGLE* tval, int loc)
{
	int i, lo;
	
	/* 
	 * Make sure the vector has data in 
	 * the required precision and at the
	 * required location
	 */
	 
	if (loc == CUKR_LOCATION_CPU || loc == CUKR_DEFAULT)
		CukrVecmpTransfer(vec, CUKR_LOCATION_CPU, CUKR_TYPE_QDOUBLE);
	else if (loc == CUKR_LOCATION_GPU)		
		CukrVecmpTransfer(vec, CUKR_LOCATION_GPU, CUKR_TYPE_QDOUBLE);
	else
		CUKR_FATALX("Invalid location (%d) specified for "\
		            "vector %s while setting from vector "\
		            "of quasi-doubles", loc, vec->name);	
	
	/*
	 * Special case where index = 0 and count = vec->n
	 */
	 
	 if (index == 0 && count == vec->n)
	 {  
	 	lo = 0;
	 	for (i = 0; i < vec->nparts; i++) {
	 		CukrThreadsRun(i, (void*)CukrVecspGetToVectorQ, CUKR_FUNCID_VEC_GET_TOVEC_Q, 
	 		                   vec->cells[i], 0,  vec->cells[i]->n, &hval[lo], &tval[lo], loc);
	 		lo += vec->cells[i]->n;                   
	 	}
	 	CukrThreadsWaitCompletion(vec->nparts);
	 	return;
	 }
	
	/* 
	 * Find first and last parts 
	 */
	 
	int start_part = -1, end_part = -1, start_part_index;
	lo = 0;
	for (i = 0; i < vec->nparts; i++) 
	{			
		/* If start conditions are met,
		 * set start (once) */
		if (index - lo < vec->cells[i]->n && start_part == -1) {
				start_part = i;
				start_part_index = index - lo;
		}				
				
		/* If last part detected, set and exit */	
		if (index + count < lo + vec->cells[i]->n) {
			end_part = i;	
			break;	
		}
						
		lo += vec->cells[i]->n;
	} 
	
	/* If start_part still -1, error */
	if (start_part == -1) 
		CUKR_FATALX("Invalid index (%d) accessed from "\
			         "vector %s when setting from vector", 
			         index, vec->name);	 
	
	/* If end part still -1, then it is the 
	 * last part */
	if (end_part == -1)
		end_part = vec->nparts - 1; 
		
	/* 
	 * If all in the same part 
	 */
	
	/* If start_part and end_part are the same,
	 * set count elements and be done */	
	if (start_part == end_part) {
		CukrThreadsRun(start_part, (void*)CukrVecspGetToVectorQ, CUKR_FUNCID_VEC_GET_TOVEC_Q, 
		                           vec->cells[start_part], start_part_index, count, hval, tval, loc);
		CukrThreadsWaitCompletionForThread(start_part);	
		return;
	} 
	  
	/* 
	 * If not, do the parts 
	 */
	 
	/* For start part, set how many elements
	 * are left from index to end */	
	int val_index   = 0;  
	int vals_to_set = vec->cells[start_part]->n - start_part_index;
	CukrThreadsRun(start_part, (void*)CukrVecspGetToVectorQ, CUKR_FUNCID_VEC_GET_TOVEC_Q, 
	                            vec->cells[start_part], start_part_index, vals_to_set, hval, tval, loc); 	                            
	val_index += vals_to_set;
	
	/* The ones between start and end are full */	
	for (i = start_part + 1; i < end_part; i++) {
		CukrThreadsRun(i, (void*)CukrVecspGetToVectorQ, CUKR_FUNCID_VEC_GET_TOVEC_Q, 
	                      vec->cells[i], 0, vec->cells[i]->n, &hval[val_index], &tval[val_index], loc); 
	    val_index += vec->cells[i]->n;                  	   
	} 
	
	/* The last one has the leftovers */
	vals_to_set = count - val_index;
	CukrThreadsRun(end_part, (void*)CukrVecspGetToVectorQ, CUKR_FUNCID_VEC_GET_TOVEC_Q, 
	                         vec->cells[end_part], 0, vals_to_set, &hval[val_index], &tval[val_index], loc); 
	                         
	/* Wait for the threads to finish */
	for (i = start_part; i < end_part; i++)
		 CukrThreadsWaitCompletionForThread(i); 
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to a single precision vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecmpGetToVectorS(CUKR_VECTOR_MP vec, int index, int count, CUKR_SINGLE* val, int loc)
{
	int i, lo;
	
	/* 
	 * Make sure the vector has data in 
	 * the required precision and at the
	 * required location
	 */
	 
	if (loc == CUKR_LOCATION_CPU || loc == CUKR_DEFAULT)
		CukrVecmpTransfer(vec, CUKR_LOCATION_CPU, CUKR_TYPE_SINGLE);
	else if (loc == CUKR_LOCATION_GPU)		
		CukrVecmpTransfer(vec, CUKR_LOCATION_GPU, CUKR_TYPE_SINGLE);
	else
		CUKR_FATALX("Invalid location (%d) specified for "\
		            "vector %s while setting from vector "\
		            "of singles", loc, vec->name);	
	
	/*
	 * Special case where index = 0 and count = vec->n
	 */
	 
	 if (index == 0 && count == vec->n)
	 {  
	 	lo = 0;
	 	for (i = 0; i < vec->nparts; i++) {
	 		CukrThreadsRun(i, (void*)CukrVecspGetToVectorS, CUKR_FUNCID_VEC_GET_TOVEC_S, 
	 		                   vec->cells[i], 0,  vec->cells[i]->n, &val[lo], loc);
	 		lo += vec->cells[i]->n;                   
	 	}
	 	CukrThreadsWaitCompletion(vec->nparts);
	 	return;
	 }
	
	/* 
	 * Find first and last parts 
	 */
	 
	int start_part = -1, end_part = -1, start_part_index;
	lo = 0;
	for (i = 0; i < vec->nparts; i++) 
	{			
		/* If start conditions are met,
		 * set start (once) */
		if (index - lo < vec->cells[i]->n && start_part == -1) {
				start_part = i;
				start_part_index = index - lo;
		}				
				
		/* If last part detected, set and exit */	
		if (index + count < lo + vec->cells[i]->n) {
			end_part = i;	
			break;	
		}
						
		lo += vec->cells[i]->n;
	} 
	
	/* If start_part still -1, error */
	if (start_part == -1) 
		CUKR_FATALX("Invalid index (%d) accessed from "\
			         "vector %s when setting from vector", 
			         index, vec->name);	 
	
	/* If end part still -1, then it is the 
	 * last part */
	if (end_part == -1)
		end_part = vec->nparts - 1; 
		
	/* 
	 * If all in the same part 
	 */
	
	/* If start_part and end_part are the same,
	 * set count elements and be done */	
	if (start_part == end_part) {
		CukrThreadsRun(start_part, (void*)CukrVecspGetToVectorS, CUKR_FUNCID_VEC_GET_TOVEC_S, 
		                           vec->cells[start_part], start_part_index, count, val, loc);
		CukrThreadsWaitCompletionForThread(start_part);	
		return;
	} 
	  
	/* 
	 * If not, do the parts 
	 */
	 
	/* For start part, set how many elements
	 * are left from index to end */	
	int val_index   = 0;  
	int vals_to_set = vec->cells[start_part]->n - start_part_index;
	CukrThreadsRun(start_part, (void*)CukrVecspGetToVectorS, CUKR_FUNCID_VEC_GET_TOVEC_S, 
	                            vec->cells[start_part], start_part_index, vals_to_set, val, loc); 	                            
	val_index += vals_to_set;
	
	/* The ones between start and end are full */	
	for (i = start_part + 1; i < end_part; i++) {
		CukrThreadsRun(i, (void*)CukrVecspGetToVectorS, CUKR_FUNCID_VEC_GET_TOVEC_S, 
	                      vec->cells[i], 0, vec->cells[i]->n, &val[val_index], loc); 
	    val_index += vec->cells[i]->n;                  	   
	} 
	
	/* The last one has the leftovers */
	vals_to_set = count - val_index;
	CukrThreadsRun(end_part, (void*)CukrVecspGetToVectorS, CUKR_FUNCID_VEC_GET_TOVEC_S, 
	                         vec->cells[end_part], 0, vals_to_set, &val[val_index], loc); 
	                         
	/* Wait for the threads to finish */
	for (i = start_part; i < end_part; i++)
		 CukrThreadsWaitCompletionForThread(i);     
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to an integer vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecmpGetToVectorI(CUKR_VECTOR_MP vec, int index, int count, CUKR_INT* val, int loc)
{
	int i, lo;
	
	/* 
	 * Make sure the vector has data in 
	 * the required precision and at the
	 * required location
	 */
	 
	if (loc == CUKR_LOCATION_CPU || loc == CUKR_DEFAULT)
		CukrVecmpTransfer(vec, CUKR_LOCATION_CPU, CUKR_TYPE_INT);
	else if (loc == CUKR_LOCATION_GPU)		
		CukrVecmpTransfer(vec, CUKR_LOCATION_GPU, CUKR_TYPE_INT);
	else
		CUKR_FATALX("Invalid location (%d) specified for "\
		            "vector %s while setting from vector "\
		            "of integers", loc, vec->name);	
	
	/*
	 * Special case where index = 0 and count = vec->n
	 */
	 
	 if (index == 0 && count == vec->n)
	 {  
	 	lo = 0;
	 	for (i = 0; i < vec->nparts; i++) {
	 		CukrThreadsRun(i, (void*)CukrVecspGetToVectorI, CUKR_FUNCID_VEC_GET_TOVEC_I, 
	 		                   vec->cells[i], 0,  vec->cells[i]->n, &val[lo], loc);
	 		lo += vec->cells[i]->n;                   
	 	}
	 	CukrThreadsWaitCompletion(vec->nparts);
	 	return;
	 }
	
	/* 
	 * Find first and last parts 
	 */
	 
	int start_part = -1, end_part = -1, start_part_index;
	lo = 0;
	for (i = 0; i < vec->nparts; i++) 
	{			
		/* If start conditions are met,
		 * set start (once) */
		if (index - lo < vec->cells[i]->n && start_part == -1) {
				start_part = i;
				start_part_index = index - lo;
		}				
				
		/* If last part detected, set and exit */	
		if (index + count < lo + vec->cells[i]->n) {
			end_part = i;	
			break;	
		}
						
		lo += vec->cells[i]->n;
	} 
	
	/* If start_part still -1, error */
	if (start_part == -1) 
		CUKR_FATALX("Invalid index (%d) accessed from "\
			         "vector %s when setting from vector", 
			         index, vec->name);	 
	
	/* If end part still -1, then it is the 
	 * last part */
	if (end_part == -1)
		end_part = vec->nparts - 1; 
		
	/* 
	 * If all in the same part 
	 */
	
	/* If start_part and end_part are the same,
	 * set count elements and be done */	
	if (start_part == end_part) {
		CukrThreadsRun(start_part, (void*)CukrVecspGetToVectorI, CUKR_FUNCID_VEC_GET_TOVEC_I, 
		                           vec->cells[start_part], start_part_index, count, val, loc);
		CukrThreadsWaitCompletionForThread(start_part);	
		return;
	} 
	  
	/* 
	 * If not, do the parts 
	 */
	 
	/* For start part, set how many elements
	 * are left from index to end */	
	int val_index   = 0;  
	int vals_to_set = vec->cells[start_part]->n - start_part_index;
	CukrThreadsRun(start_part, (void*)CukrVecspGetToVectorI, CUKR_FUNCID_VEC_GET_TOVEC_I, 
	                            vec->cells[start_part], start_part_index, vals_to_set, val, loc); 	                            
	val_index += vals_to_set;
	
	/* The ones between start and end are full */	
	for (i = start_part + 1; i < end_part; i++) {
		CukrThreadsRun(i, (void*)CukrVecspGetToVectorI, CUKR_FUNCID_VEC_GET_TOVEC_I, 
	                      vec->cells[i], 0, vec->cells[i]->n, &val[val_index], loc); 
	    val_index += vec->cells[i]->n;                  	   
	} 
	
	/* The last one has the leftovers */
	vals_to_set = count - val_index;
	CukrThreadsRun(end_part, (void*)CukrVecspGetToVectorI, CUKR_FUNCID_VEC_GET_TOVEC_I, 
	                         vec->cells[end_part], 0, vals_to_set, &val[val_index], loc); 
	                         
	/* Wait for the threads to finish */
	for (i = start_part; i < end_part; i++)
		 CukrThreadsWaitCompletionForThread(i);     
}

// ------------------------------------------------------------------- //

/**
 * \brief Sets all elements of the
 *        vector to random values, on the CPUs. 
 *        The rest of the precisions and 
 *        locations are invalidated.
 * \param vec The vector
 * \param min,max The bounds for the random number
 * \param seed The seed of the random number gen.
 * \param type Data type where to write the data
 *             or default to match the solver type
 */
void CukrVecmpSetRandom(CUKR_VECTOR_MP vec, CUKR_DOUBLE min, CUKR_DOUBLE max, int seed, int type)
{
	int i;
	
	/* Set the value for all cells
	 * to a random value, regardless of the
	 * sharing type */
	for (i = 0; i < vec->nparts; i++)
		CukrVecspSetRandom(vec->cells[i], min, max, seed, type); 		
		
	/* Invalidate the insync flag */
	vec->insync = CUKR_FALSE;	
}





				/*****************************************/
				//                 I/O                   // 
				/*****************************************/ 




/**
 * \brief Reads data into a previously
 *        created vector structure from 
 *        an opened file stream. Can be used
 *        for example when reading data for
 *        a CSR matrix, which consists of
 *        three vectors one after another.
 *        Calling this function three times,
 *        once for each vector, would do the 
 *        trick.
 * \param f The open file stream
 * \param type The type of data that is
 *        supposed to be read (as double,
 *        qdouble, single, int)
 */
void CukrVecmpReadDataFromFile(CUKR_VECTOR_MP vec, FILE* f, int type)
{
	int i;
	
	/* Successively read data into the
	 * cells, regardless of the sharing type */
	for (i = 0; i < vec->nparts; i++)
		CukrVecspReadDataFromFile(vec->cells[i], f, type); 
		
	/* Invalidate the insync flag */
	vec->insync = CUKR_FALSE;		
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads data into a previously
 *        created vector structure from 
 *        a file.
 * \param filename The file name
 * \param type The type of data that is
 *        supposed to be read (as double,
 *        qdouble, single, int) or default
 *        to match the solver
 */
void CukrVecmpReadDataFromFilename(CUKR_VECTOR_MP vec, char* filename, int type)
{
	FILE *f;
	
	if ( (f = fopen(filename, "r")) == NULL)
		CUKR_FATALX("Could not read MP vector %s from file "\
		            "%s because the file could not be opened",
		            vec->name, filename);
	
	CukrVecmpReadDataFromFile(vec, f, type);
	
	fclose(f);
} 

// ------------------------------------------------------------------- //

/**
 * \brief Write the values of a vector
 *        to an opened file stream. 
 * \param vec The vector
 * \param f The open file stream
 */
void CukrVecmpWriteDataToFile(CUKR_VECTOR_MP vec, FILE* f)
{
	int i;
	
	LOGX(DEBUG, "Writing MP vector %s to file stream", 
			    vec->name);
		
	/* 
	 * Bring back to host and write
	 * all cells, regardless of
	 * sharing type, since this is
	 * where valid data is for both 
	 */ 
	
	/* Check the type of the first part,
	 * since all are the same. If int,
	 * bring it in int. If float, bring
	 * it in double */
	if (vec->cells[0]->status & CUKR_TYPE_INT)
		CukrVecmpTransfer(vec, CUKR_LOCATION_CPU, CUKR_TYPE_INT);
	else
		CukrVecmpTransfer(vec, CUKR_LOCATION_CPU, CUKR_TYPE_DOUBLE);
	
	/* 
	 * Write the parts 
	 */
	for (i = 0; i < vec->nparts; i++)
		CukrVecspWriteDataToFile(vec->cells[i], f);
	
	LOGX(DEBUG, "Done writing MP vector %s to file stream", 
			    vec->name);		
}

// ------------------------------------------------------------------- //

/**
 * \brief Write the values of a vector
 *        to an opened file stream. 
 * \param vec The vector
 * \param f The file name
 */
void CukrVecmpWriteDataToFilename(CUKR_VECTOR_MP vec, char* filename)
{
	FILE *f;
	
	LOGX(DEBUG, "Writing MP vector %s to file %s", 
			    vec->name, filename);
	
	if ( (f = fopen(filename, "w")) == NULL)
		CUKR_FATALX("Could not write MP vector %s to file "\
		            "%s because the file could not be opened",
		            vec->name, filename);
	
	/* Write the no. of elements */
	fprintf(f, "%d\n", vec->n);
	
	/* Write data to stream */
	CukrVecmpWriteDataToFile(vec, f);
	
	/* Close */
	fclose(f);
	
	LOGX(DEBUG, "Done writing vector %s to file %s", 
			    vec->name, filename);	
}

// ------------------------------------------------------------------- // 

/**
 * \brief Prints the values of a MP vector
 * \param vec The vector being printed
 * \param f An open file stream to write to instead
 */
void CukrVecmpPrint(CUKR_VECTOR_MP vec)
{
	CukrVecmpPrintToFile(vec, stdout);
}

// ------------------------------------------------------------------- //

/**
 * \brief Prints the values of a MP vector
 * \param vec The vector being printed
 * \param filename The full path of the file
 *        to write the vector to
 */
void CukrVecmpPrintToFilename(CUKR_VECTOR_MP vec, char* filename)
{
	FILE *f;
	
	if ( (f = fopen(filename, "w")) == NULL)
		CUKR_FATALX("Could not write vector %s to file "\
		            "%s because the file could not be opened",
		            vec->name, filename);
	
	CukrVecmpPrintToFile(vec, f);
	
	fclose(f);
}

// ------------------------------------------------------------------- //

/**
 * \brief Prints the values of a MP vector
 * \param vec The vector being printed
 * \param f An open file stream to write to instead
 */
void CukrVecmpPrintToFile(CUKR_VECTOR_MP vec, FILE* f)
{
	CukrVecmpPrintToFileWithIndent(vec, f, "");
}
  
// ------------------------------------------------------------------- //  

/**
 * \brief Prints the values of a MP vector
 * \param vec The vector being printed
 * \param f An open file stream to write to
 * \param indent An indentation string, to be inserted
 *        before every print.
 */
void CukrVecmpPrintToFileWithIndent(CUKR_VECTOR_MP vec, FILE* f, char* indent)
{
	int i, j;
	
	/* If on the GPU, do the transfer first, 
	 * since it's a multi-gpu transfer
	 * Enough to check the status of the first
	 * cell, since all have to have the same
	 * status and location */
	if (!(vec->cells[0]->status & CUKR_LOCATION_CPU)) {
		if (vec->cells[0]->status & CUKR_TYPE_QDOUBLE)
			CukrVecmpTransfer(vec, CUKR_LOCATION_CPU, CUKR_TYPE_QDOUBLE);
		else if (vec->cells[0]->status & CUKR_TYPE_SINGLE)
			CukrVecmpTransfer(vec, CUKR_LOCATION_CPU, CUKR_TYPE_SINGLE);		
	} 
	
	/* Merging indents */
	char mindent[255];
	sprintf(mindent, "%s    ", indent);
	
	/* Print all the parts */
	fprintf(f, "%sMP Vector %s\n", indent, vec->name);
	fprintf(f, "%sAddr = %p\n", indent, vec);
	
	/* For distributed vector, print  the cells */	
	if (vec->shtype == CUKR_SHTYPE_DISTRIBUTED) {
		fprintf(f, "%sDISTRIBUTED type -> printing cells\n", indent);
		for (i = 0; i < vec->nparts; i++)
			CukrVecspPrintToFileWithIndent(vec->cells[i], f, mindent);
	}
	/* For shared vectors, print all the rows
	 * and cells */
	else {
		fprintf(f, "%sSHARED type -> printing rows \n", indent);
		for (i = 0; i < vec->nparts; i++)
			CukrVecspPrintToFileWithIndent(vec->rows[i], f, mindent);
		
		fprintf(f, "%sSHARED type -> printing cells \n", indent);
		for (i = 0; i < vec->nparts; i++)
			CukrVecspPrintToFileWithIndent(vec->cells[i], f, mindent);
	} 		
}





//============================ PRIVATE FUNCTIONS =========================== //



/**
 * \brief Malloc used for shared vectors.
 *        First, memory for the entire row is 
 *        allocated. Then, the data vectors in 
 *        the cells are made to point to the
 *        right places in the row data
 * \param vec The vector structure being created
 * \param loc Vector location {CPU,GPU}
 * \param type Vector type 
 *        {double, qdouble, single, int}
 * \param part The part for which cell pointers are
 *        being set
 */
void CukrVecmpMallocSetCellPtrs(CUKR_VECTOR_MP vec, int loc, int type, int part)
{
	int i, pos;
		
	/* 
	 * For the CPU 
	 */
	 
	if (loc == CUKR_LOCATION_CPU)
	{
		
		/* For double */
		if (type == CUKR_TYPE_DOUBLE)
		{
			
			/* Check to see if setting is needed.
			 * Enough to check first one, since 
			 * all would either be allocated or
			 * null */
			 
			if (vec->cells[part]->ph_dval != NULL)
				return;
			
			LOGX(DEBUG, "Setting cell pointer for shared MP " \
			            "vector %s part %d (CPU,double)", vec->name, part);
			
			/* Set pointers for cell */
			pos = 0;
			for (i = 0; i < part; i++)
				pos += vec->cells[i]->n;
			vec->cells[part]->ph_dval = &(vec->rows[part]->ph_dval[pos]);	
			
			LOGX(DEBUG, "Done setting cell pointer for  "
			            "MP vector %s part %d", vec->name, part);
			            
			/* Nothing more to do */
			return;             
		}
		
		/* For single or qdouble head part*/
		if (type == CUKR_TYPE_SINGLE || type == CUKR_TYPE_QDOUBLE)
		{	
			/* Check to see if setting is needed.
			 * Enough to check first one, since 
			 * all would either be allocated or
			 * null */
			if (vec->cells[part]->ph_sval != NULL)
				return;
			
			if (type == CUKR_TYPE_QDOUBLE) {
				LOGX(DEBUG, "Setting cell pointer for shared MP " \
				            "vector %s part %d (CPU,qdouble,head)", vec->name, part); 
			} 
			else {
				LOGX(DEBUG, "Setting cell pointer for shared MP " \
				            "vector %s part %d (CPU,single)", vec->name, part); 
			}
			
			/* Set pointers for cell */
			pos = 0;
			for (i = 0; i < part; i++)
				pos += vec->cells[i]->n;
			vec->cells[part]->ph_sval = &(vec->rows[part]->ph_sval[pos]);	
			
			LOGX(DEBUG, "Done setting cell pointer for  "
			            "MP vector %s part %d", vec->name, part);
			            
			/* Cannot exit yet for qdouble still
			 * need to set tails */
			if (type == CUKR_TYPE_SINGLE)
				return;              	 
		}
		
		/* For qdouble tail part*/
		if (type == CUKR_TYPE_QDOUBLE)
		{
			/* Check to see if setting is needed.
			 * Enough to check first one, since 
			 * all would either be allocated or
			 * null */
			if (vec->cells[part]->ph_qvalt != NULL)
				return;
					
			LOGX(DEBUG, "Setting cell pointer for shared MP " \
				        "vector %s part %d (CPU,qdouble,tail)", vec->name, part); 
				        
			/* Set pointers for cell */
			pos = 0;
			for (i = 0; i < part; i++)
				pos += vec->cells[i]->n;
			vec->cells[part]->ph_qvalt = &(vec->rows[part]->ph_qvalt[pos]);	
			
			LOGX(DEBUG, "Done setting cell pointer for  "
			            "MP vector %s part %d", vec->name, part);
			            
			/* Nothing more to do */
			return;             
		}
	}
	
	/* 
	 * For the GPU 
	 */
	if (loc == CUKR_LOCATION_GPU)
	{
		/* For double */
		if (type == CUKR_TYPE_DOUBLE)
		{
			/* Check to see if setting is needed.
			 * Enough to check first one, since 
			 * all would either be allocated or
			 * null */
			if (vec->cells[part]->pd_dval != NULL)
				return;
			
			LOGX(DEBUG, "Setting cell pointer for shared MP " \
			            "vector %s row %d (GPU,double)", vec->name, part);
			
			/* Set pointers for cell */
			pos = 0;
			for (i = 0; i < part; i++)
				pos += vec->cells[i]->n;
			vec->cells[part]->pd_dval = &(vec->rows[part]->pd_dval[pos]);	
			
			LOGX(DEBUG, "Done setting cell pointersfor  "
			            "MP vector %s part %d", vec->name, part);
			            
			/* Nothing more to do */
			return;              
		}
		
		/* For single or qdouble head part*/
		if (type == CUKR_TYPE_SINGLE || type == CUKR_TYPE_QDOUBLE)
		{
			/* Check to see if setting is needed.
			 * Enough to check first one, since 
			 * all would either be allocated or
			 * null */
			if (vec->cells[part]->pd_sval != NULL)
				return;
			
			if (type == CUKR_TYPE_QDOUBLE) {
				LOGX(DEBUG, "Setting cell pointer for shared MP " \
				            "vector %s part %d (GPU,qdouble,head)", vec->name, part); 
			} 
			else {
				LOGX(DEBUG, "Setting cell pointer for shared MP " \
				            "vector %s part %d (GPU,single)", vec->name, part); 
			}
			
			/* Set pointers for cell */
			pos = 0;
			for (i = 0; i < part; i++)
				pos += vec->cells[i]->n;
			vec->cells[part]->pd_sval = &(vec->rows[part]->pd_sval[pos]);	
			
			LOGX(DEBUG, "Done setting cell pointer for  "
			            "MP vector %s part %d", vec->name, part);
			            
			/* Cannot exit yet for qdouble still
			 * need to set tails */
			if (type == CUKR_TYPE_SINGLE)
				return;               
		}
		
		/* For qdouble tail part*/
		if (type == CUKR_TYPE_QDOUBLE)
		{
			/* Check to see if setting is needed.
			 * Enough to check first one, since 
			 * all would either be allocated or
			 * null */
			if (vec->cells[part]->pd_qvalt != NULL)
				return;
			
			LOGX(DEBUG, "Setting cell pointer for shared MP " \
				        "vector %s part %d (GPU,qdouble,tail)", vec->name, part); 
			
			pos = 0;
			for (i = 0; i < part; i++)
				pos += vec->cells[i]->n;
			vec->cells[part]->pd_qvalt = &(vec->rows[part]->pd_qvalt[pos]);
			
			LOGX(DEBUG, "Done setting cell pointers for  "
			            "MP vector %s part %d", vec->name, part);
			            
			/* Nothing more to do */
			return;              
		}
	}
	
	/* If still here, error */
	CUKR_FATALX("Error setting cell pointers for MP " \
	            "vector %s. Invalid type (%d) or "
	            "location (%d) found", vec->name,
	            type, loc);              
}	

// ------------------------------------------------------------------- //

/**
 * \brief Send to the sync vector the data owned
 *        by the specified part, which is needed
 *        by other parts (for the CPU)
 * \param rvec The part (row) of the shared vector
 *        to be synchronized
 * \param sync The temporary vector used for 
 *        synchronization
 * \param give The comm-entry container describing 
 *        what data has to be sent (given)
 * \param The location to be synchonized  
 * \param The data-type to be synchronize
 */
void CukrVecmpSynchronizePart_Send_CPU(CUKR_VECTOR_SP rvec, CUKR_VECTOR_SP sync, CUKR_CONTAINER_VECTOR_COMMENTRY give, int loc, int type)
{
	int i, pos, size;	

	/* For double */
	if (type == CUKR_TYPE_DOUBLE)
	{
		LOGX(DEBUG, "Sync. (send) vector %s in host,double", rvec->name);
		for (i = 0; i < give->size; i++) {
			pos = give->elements[i].pos;
			size = give->elements[i].size;			
			memcpy((void*)(&(sync->ph_dval[pos])), (void*)(&(rvec->ph_dval[pos])), size * sizeof(CUKR_DOUBLE));
		}
		
		/* Nothing more to do */
		return;
	}
	
	/* For single - qdouble (head) */
	if (type == CUKR_TYPE_SINGLE || type == CUKR_TYPE_QDOUBLE)
	{	
		if (type == CUKR_TYPE_QDOUBLE) {
			LOGX(DEBUG, "Sync. (send)  vector %s in host,qdouble,head", rvec->name);
		}
		else {
			LOGX(DEBUG, "Sync. (send)  vector %s in host,single", rvec->name);
		}
		for (i = 0; i < give->size; i++) {
			pos = give->elements[i].pos;
			size = give->elements[i].size;
			memcpy((void*)(&(sync->ph_sval[pos])), (void*)(&(rvec->ph_sval[pos])), size * sizeof(CUKR_SINGLE));
		}
		
		/* Cannot exit yet for qdouble still
		 * need to set tails */
		 if (type == CUKR_TYPE_SINGLE)
		 	return;
	}
	
	/* For qdouble (tail) */
	if (type == CUKR_TYPE_QDOUBLE)
	{	
		LOGX(DEBUG, "Sync. (send)  vector %s in host,qdouble,tail", rvec->name);
		
		for (i = 0; i < give->size; i++) {
			pos = give->elements[i].pos;
			size = give->elements[i].size;
			memcpy((void*)(&(sync->ph_qvalt[pos])), (void*)(&(rvec->ph_qvalt[pos])), size * sizeof(CUKR_SINGLE));
		}
		
		/* Nothing more to do */
		return;
	}
	
	/* If here, error */
	CUKR_FATALX("No floating-point data to syncronize " \
	            "vector %s, on host", rvec->name);            
	
} 

// ------------------------------------------------------------------- //

/**
 * \brief Send to the sync vector the data owned
 *        by the specified part, which is needed
 *        by other parts (for the GPU)
 * \param rvec The part (row) of the shared vector
 *        to be synchronized
 * \param sync The temporary vector used for 
 *        synchronization
 * \param give The comm-entry container describing 
 *        what data has to be sent (given)
 * \param The location to be synchonized  
 * \param The data-type to be synchronize
 */
void CukrVecmpSynchronizePart_Send_GPU(CUKR_VECTOR_SP rvec, CUKR_VECTOR_SP sync, CUKR_CONTAINER_VECTOR_COMMENTRY give, int loc, int type)
#ifndef CUKR_USE_GPU
{
	CUKR_FATAL("GPU support was not compiled in the library " \
			   "(CukrVecmpSynchronizePart_Send_GPU cannot be executed) ");
}
#else
{
	int i, pos, size;
	cublasStatus cublas_status;
		
	/* For double */
	if ( type == CUKR_TYPE_DOUBLE)
	{	
		LOGX(DEBUG, "Sync. (recv) vector %s in gpu,double", rvec->name);

		for (i = 0; i < give->size; i++) {
			pos = give->elements[i].pos;
			size = give->elements[i].size;
			cublas_status = cublasGetVector(size, sizeof(CUKR_DOUBLE), &(rvec->pd_dval[pos]),
                                     1, &(sync->ph_dval[pos]), 1);
    		CukrHandleCudaMemErrors(rvec, cublas_status);
		}
		
		/* Nothing more to do */
		return;
	}
	
	/* For single - qdouble (head) */
	if (type == CUKR_TYPE_SINGLE || type == CUKR_TYPE_QDOUBLE)
	{	
		if (type == CUKR_TYPE_QDOUBLE) {
			LOGX(DEBUG, "Sync. (recv)  vector %s in gpu,qdouble,head", rvec->name);
		}
		else {
			LOGX(DEBUG, "Sync. (recv) vector %s in gpu,single", rvec->name);
		}
					
		for (i = 0; i < give->size; i++) {
			pos = give->elements[i].pos;
			size = give->elements[i].size;
			cublas_status = cublasGetVector(size, sizeof(CUKR_SINGLE), &(rvec->pd_sval[pos]),
                                            1, &(sync->ph_sval[pos]), 1);
    		CukrHandleCudaMemErrors(rvec, cublas_status);
		}
		
		/* Cannot exit yet for qdouble still
		 * need to set tails */
		 if (type == CUKR_TYPE_SINGLE)
		 	return;
	}
	
	/* For qdouble (tail) */
	if ( type == CUKR_TYPE_QDOUBLE)
	{	
		LOGX(DEBUG, "Sync. (recv) vector %s in gpu,qdouble,tail", rvec->name);

		for (i = 0; i < give->size; i++) {
			pos = give->elements[i].pos;
			size = give->elements[i].size;
			cublas_status = cublasGetVector(size, sizeof(CUKR_SINGLE), &(rvec->pd_qvalt[pos]),
                                     1, &(sync->ph_qvalt[pos]), 1);
    		CukrHandleCudaMemErrors(rvec, cublas_status);
		}
		
		/* Nothing more to do */
		return;
	}
	
	/* If here, error */
	CUKR_FATALX("No floating-point data to syncronize " \
	            "vector %s, on gpu", rvec->name);            
} 
#endif /* CUKR_USE_GPU */

// ------------------------------------------------------------------- //

/**
 * \brief Receive from the sync vector the data needed
 *        by the specified part, which is owned
 *        by other parts (for the CPU)
 * \param rvec The part (row) of the shared vector
 *        to be synchronized
 * \param sync The temporary vector used for 
 *        synchronization
 * \param give The comm-entry container describing 
 *        what data has to be sent (given)
 * \param The location to be synchonized  
 * \param The data-type to be synchronize
 */
void CukrVecmpSynchronizePart_Recv_CPU(CUKR_VECTOR_SP rvec, CUKR_VECTOR_SP sync, CUKR_CONTAINER_VECTOR_COMMENTRY need, int loc, int type)
{
	int i, pos, size;
			 
	/* For double */
	if (type == CUKR_TYPE_DOUBLE)
	{
		LOGX(DEBUG, "Sync. (recv) vector %s in host,double", rvec->name);
		for (i = 0; i < need->size; i++) {
			pos = need->elements[i].pos;
			size = need->elements[i].size;
			memcpy((void*)(&(rvec->ph_dval[pos])), (void*)(&(sync->ph_dval[pos])), size * sizeof(CUKR_DOUBLE));
		}
		
		/* Nothing more to do */
		return;
	}
	
	/* For single - qdouble (head) */
	if (type == CUKR_TYPE_SINGLE || type == CUKR_TYPE_QDOUBLE)
	{	
		if (type == CUKR_TYPE_QDOUBLE) {
			LOGX(DEBUG, "Sync. (recv) vector %s in host,qdouble,head", rvec->name);
		}
		else {
			LOGX(DEBUG, "Sync. (recv) vector %s in host,single", rvec->name);
		}
		for (i = 0; i < need->size; i++) {
			pos = need->elements[i].pos;
			size = need->elements[i].size;
			memcpy((void*)(&(rvec->ph_sval[pos])), (void*)(&(sync->ph_sval[pos])), size * sizeof(CUKR_SINGLE));
		}
		
		/* Cannot exit yet for qdouble still
		 * need to set tails */
		 if (type == CUKR_TYPE_SINGLE)
		 	return;
	}
	
	/* For qdouble (tail) */
	if (type == CUKR_TYPE_QDOUBLE)
	{	
		LOGX(DEBUG, "Sync. (recv) %s in host,qdouble,tail", rvec->name);
		
		for (i = 0; i < need->size; i++) {
			pos = need->elements[i].pos;
			size = need->elements[i].size;
			memcpy((void*)(&(rvec->ph_qvalt[pos])), (void*)(&(sync->ph_qvalt[pos])), size * sizeof(CUKR_SINGLE));
		}
		
		/* Nothing more to do */
		return;
	}
	
	/* If here, error */	
	CUKR_FATALX("No floating-point data to syncronize for" \
                "vector %s, on host", rvec->name);           
	
} 

// ------------------------------------------------------------------- //

/**
 * \brief Receive from the sync vector the data needed
 *        by the specified part, which is owned
 *        by other parts (for the GPU)
 * \param rvec The part (row) of the shared vector
 *        to be synchronized
 * \param sync The temporary vector used for 
 *        synchronization
 * \param give The comm-entry container describing 
 *        what data has to be sent (given)
 * \param The location to be synchonized  
 * \param The data-type to be synchronize
 */
void CukrVecmpSynchronizePart_Recv_GPU(CUKR_VECTOR_SP rvec, CUKR_VECTOR_SP sync, CUKR_CONTAINER_VECTOR_COMMENTRY need, int loc, int type)
#ifndef CUKR_USE_GPU
{
	CUKR_FATAL("GPU support was not compiled in the library " \
			   "(CukrVecmpSynchronizePart_Recv_GPU cannot be executed) ");
}
#else
{
	int i, pos, size;
	cublasStatus cublas_status;

	
	/* For double */
	if (type == CUKR_TYPE_DOUBLE)
	{	
		LOGX(DEBUG, "Sync. (recv) %s in gpu,double", rvec->name);

		for (i = 0; i < need->size; i++) {
			pos = need->elements[i].pos;
			size = need->elements[i].size;
			cublas_status = cublasSetVector(size, sizeof(CUKR_DOUBLE), &(sync->ph_dval[pos]),
                                            1, &(rvec->pd_dval[pos]), 1);
    		CukrHandleCudaMemErrors(rvec, cublas_status);
		}
		
		/* Nothing more to do */
		return;
	}
	
	/* For single - qdouble (head) */
	if (type == CUKR_TYPE_SINGLE || type == CUKR_TYPE_QDOUBLE)
	{	
		if (type == CUKR_TYPE_QDOUBLE) {
			LOGX(DEBUG, "Sync. (recv) %s in gpu,qdouble,head", rvec->name);
		}
		else {
			LOGX(DEBUG, "Sync. (recv) %s in gpu,single", rvec->name);
		}
		
		
		for (i = 0; i < need->size; i++) {
			pos = need->elements[i].pos;
			size = need->elements[i].size;			
			cublas_status = cublasSetVector(size, sizeof(CUKR_SINGLE), &(sync->ph_sval[pos]),
                                            1, &(rvec->pd_sval[pos]), 1);
    		CukrHandleCudaMemErrors(rvec, cublas_status);
		}
		
		/* Cannot exit yet for qdouble still
		 * need to set tails */
		if (type == CUKR_TYPE_SINGLE)
		 	return;
	}
	
	/* For qdouble (tail) */
	if (type == CUKR_TYPE_QDOUBLE)
	{	
		LOGX(DEBUG, "Sync. (recv) %s in gpu,qdouble,tail", rvec->name);

		for (i = 0; i < need->size; i++) {
			pos = need->elements[i].pos;
			size = need->elements[i].size;
			cublas_status = cublasSetVector(size, sizeof(CUKR_SINGLE), &(sync->ph_qvalt[pos]),
                                            1, &(rvec->pd_qvalt[pos]), 1);
    		CukrHandleCudaMemErrors(rvec, cublas_status);
		}
		
		/* Nothing more to do */
		return;
	}
	
	/* If here, error */
	CUKR_FATALX("No floating-point data to syncronize for" \
	            "vector %s, on gpu", rvec->name);            
} 
#endif /* CUKR_USE_GPU */
