/**
 * @file mat_vec/mp/vector_mp.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines the CUKR_VECTOR_MP abstract
 * interace. It's purpose is to extend
 * a vector across multiple nodes.
 * 
 * The CUKR_VECTOR_MP structure defines
 * a vector which is split in a number 
 * of parts, so that operations can be
 * performed in parallel on multiple PEs.
 * 
 * There are two types of distributed 
 * vectors:
 *  1. distributed
 *  2. shared
 * 
 * A distributed vector spanning n PEs 
 * allocates len(vec) / n elements on
 * each vector. The sections are completely
 * independent.
 * 
 * A shared vector spanning n PEs allocates
 * len(vec) on each PE. This is becase in
 * cases as SpMV, since the matrix is 
 * divident into rows and each row has to be
 * multiply to a full length vector, all the
 * memory has to be there. Most operation will
 * involve the len(vec) / n elements which 
 * actually correspond to that PE, but there 
 * will be few (hopfully) operations that need
 * other elements. Therefore, before such operation
 * the shared vector must SYNCHRONIZE. This 
 * procedure uses a communication table to 
 * copy to each copy of the shared vector assigned
 * to each PE exactly those extra elements that 
 * are needed, in the right places.
 *   
 * To implement the distributed/shared type of
 * vectors, a MP vector has two types of views
 * for the data: cells and rows. A distributed
 * vector has one cell for each PE, while a
 * shared one has one cell and one row for each
 * PE.
 * 
 * A cell is a CUKR_VECTOR_SP that corresponds to
 * the part of the distributed vector that is
 * owned by a particular PE. If there are n
 * PEs, a distributed vector will have n cells,
 * one for each PE. Operations on distributed
 * vectors (e.g. DOT) will work with on the cells.
 * 
 * A row is a CUKR_VECTOR_SP that corresponds to all
 * the memory allocated on that particular PE, 
 * containg both the elements owned by that PE
 * and the ones owned by the rest. The memory
 * allocated to the cell is part of the memory
 * allocated for the row. The position of the
 * cell is in the row is chosen such as the 
 * cell vector correspond to the elements 
 * owned by the particular PE. In a SpMV operation,
 * the vector multiplied by the matrix must be a 
 * shared vector, the multiplication will be 
 * between the matrix and the row, since the full
 * length is needed. The result will can be either
 * a shared or a distributed vector, since it 
 * will be stored in the cell.
 * 
 */

#ifndef CUKR_VECTOR_MP_H_
#define CUKR_VECTOR_MP_H_

#ifndef CUKR_VECTOR
	#define CUKR_VECTOR CUKR_VECTOR_MP
#endif

/* =============== FORWARD DECLARATIONS ============== */

typedef struct _CUKR_VECTOR_MP *CUKR_VECTOR_MP;

/* =================================================== */

#include "../sp/vector_sp.h"
#include "matrix_mp.h"
#include "../../util/containers.h"

/* 
 * Function pointers for the functions which
 * have to be scheduled to threads
 */

typedef void (*CUKR_VECMP_SYNC_SEND)(CUKR_VECTOR_SP, CUKR_VECTOR_SP, CUKR_CONTAINER_VECTOR_COMMENTRY, int, int);
typedef void (*CUKR_VECMP_SYNC_RECV)(CUKR_VECTOR_SP, CUKR_VECTOR_SP, CUKR_CONTAINER_VECTOR_COMMENTRY, int, int);

 
struct _CUKR_VECTOR_MP {
	
	// ----------- Protected ------------ //
	
	/** The name of the vector */
	char name[255];
	
	/** Length of the vector */
	int n;
	
	/** Number of parts, equal to the no.
	 * of PEs used */
	int nparts;
	
	
	// ----------- Protected ------------ //
	
	/** Vectors holding the cells. 
	 * Each multigpu vector holds nparts 
	 * vector cells, where cell[i] represents 
	 * the part[i] stored on device[i] that is
	 * actually owned and updated by the device.
	 * For distributed vectors, cells are all
	 * that is needed.
	 */
	CUKR_VECTOR_SP cells[CUKR_MAX_PE];
	
	/** Vectors for holding rows of cells.
	 * For shared vectors, we need to get access to
	 * both the parts (via the cells) and the entire
	 * row (all data from a MP vector taken as one
	 * large vector). Also, when a cell is updated, 
	 * the large vector should be updated as well, so
	 * basically we need the data in the cells to point
	 * to the data in the rows.
	 */
	CUKR_VECTOR_SP rows[CUKR_MAX_PE]; 
	
	/** Vector for holding tranferred data during
	 * synchronization */
	CUKR_VECTOR_SP sync; 
	
	
	// ------------ Private ------------- //
	
	/** Vector sharing type */
	int shtype;
	
	/** Synchronized or not. A shared vector
	 * needs to sync. it's part only if 
	 * modified. This flag, if true, shows
	 * the data is still valid. It should be
	 * invalidated after every operation that
	 * affects the data in the vector */
	int insync; 
	
};





				/*****************************************/
				//       CONSTRUCTOR / DESTRUCTOR        //
				/*****************************************/         





/**
 * \brief Creates a MP vector, of the 
 *        specified size, on the CPU, with the 
 *        specified type
 * \param vec Pointer to the MP vector 
 *        being created
 * \param n Vector size
 * \param type Vector type (double,
 *        qdouble, single, int)
 * \param shtype The type of sharing 
 *        {distributed,shared}
 * \param name The name of the vector 
 */
void CukrVecmpCreate(CUKR_VECTOR_MP* vec, CUKR_INT n, int type, int shtype, char* name);

// ------------------------------------------------------------------- //

/**
 * \brief Creates a clone of a vector. 
 *        The function currently copies
 *        the data in the highest precision
 *        available, and leaves the rest
 *        of the precisions invalidated, so
 *        they will be transformed/converted
 *        to only if needed.
 * \param src The source vector
 * \param dst Pointer to the destination vector
 * \param name The new name for the destination
 */
void CukrVecmpCreateClone(CUKR_VECTOR_MP src, CUKR_VECTOR_MP* dst, char* name);

// ------------------------------------------------------------------- //

/**
 * \brief Frees all the memory used 
 *        by the MP vector
 * \param vec Pointer to the PE vector 
 *        to be destroyed 
 */
void CukrVecmpDestroy(CUKR_VECTOR_MP* vec);





			
				/*****************************************/
				//            MALLOC / FREE              // 
				/*****************************************/ 
				
						      




/**
 * \brief Allocate memory for a MP vector. The 
 *        function checks if the pointer 
 *        corresponding to that particular location
 *        and precision is NULL of not. If not NULL,
 *        on malloc will be performed. Therefore, this
 *        function can be called "just in case" to 
 *        make sure memory allocation where it 
 *        should be. 
 * \param vec The vector structure for which
 *        memory is being allocated
 * \param loc Vector location {CPU,GPU}
 * \param type Vector type 
 *        {double, qdouble, single, int}
 */
void CukrVecmpMalloc(CUKR_VECTOR_MP vec, int loc, int type);   





				/*****************************************/
				//              TRANSFER                 // 
				/*****************************************/ 

  



/** 
 * \brief Changes the location of a MP vector
 *        by copying the data from one location
 *        to another. An MP vector does not hold
 *        a status variable by itself; it's status
 *        (location + precision) is the status of
 *        the rows and cells. Therefore, a transfer
 *        operation on an MP vector is basically
 *        a transfer operation of all of its parts.
 *        All parts (both cells and rows) are and 
 *        must be moved in the same way, at the same
 *        time. This is guaranteed by the transfered
 *        procedure, so one can query the status of 
 *        an MP vector by querying, for example, 
 *        the status of the first cell, which is 
 *        always sure to exist. 
 * 		  If the vector has valid data at the destination
 *        location and the destination time, nothing will
 *        happen. As such, one can use this function
 *        as a "precation" that data is indeed in the 
 *        required precision and location. 
 * \param vec The MP vector being transferred 
 * \param dest_loc Vector destination location {CPU,GPU}
 * \param dest_type Vector destination type 
 *        (double, qdouble, single, int)
 */
void CukrVecmpTransfer(CUKR_VECTOR_MP vec, int dest_loc, int dest_type);

// ------------------------------------------------------------------- //

/**
 * \brief Syncronize a shared vector, using
 *        the specified comm-table. A shared 
 *        vector must be synchronized BEFORE 
 *        operations like SpMV which require 
 *        data from different PEs. The function
 *		  uses the communication table to know
 *        which data has to be sent where. 
 *        Since GPUs cannot currently do P2P
 *        communication, temporary storage in
 *        the require precision, on the CPU, 
 *        is required. 
 * 		  Synchronization is done in two steps:
 *        1. Data needed by destination PEs is 
 *           copied (sent) from the source PEs to
 *           the member with appropriate precision
 *           in the sync vector.
 *        2. All needed data, which is now found 
 *           in the sync vector, is copied back
 *           (received) by the PEs that need it.	
 * \param vec The shared vector to be synchronized
 * \param commtable The commtable describing the
 *        synchronization
 * \param The location to be synchonized  
 * \param The data-type to be synchronize 
 */
void CukrVecmpSynchronize(CUKR_VECTOR_MP vec, CUKR_MATRIX_MP_COMMTABLE commtable, int loc, int type);

// ------------------------------------------------------------------- //

/**
 * \brief Data needed by destination PEs is 
 *        copied (sent) from the source PEs to
 *        the member with appropriate precision
 *        in the sync vector.
 * \param rvec The part (row) of the shared vector
 *        to be synchronized
 * \param sync The temporary vector used for 
 *        synchronization
 * \param give The comm-entry container describing 
 *        what data has to be sent (given)
 * \param The location to be synchonized  
 * \param The data-type to be synchronized
 */
void CukrVecmpSynchronizePart_Send(CUKR_VECTOR_SP rvec, CUKR_VECTOR_SP sync, CUKR_CONTAINER_VECTOR_COMMENTRY give, int loc, int type);

// ------------------------------------------------------------------- //

/**
 * \brief All needed data, which is now found 
 *        in the sync vector, is copied back
 *        (received) by the PEs that need it
 * \param rvec The part (row) of the shared vector
 *        to be synchronized
 * \param sync The temporary vector used for 
 *        synchronization
 * \param give The comm-entry container describing 
 *        what data has to be sent (given)
 * \param The location to be synchonized  
 * \param The data-type to be synchronize
 */
void CukrVecmpSynchronizePart_Recv(CUKR_VECTOR_SP rvec, CUKR_VECTOR_SP sync, CUKR_CONTAINER_VECTOR_COMMENTRY need, int loc, int type);





				/*****************************************/
				//               GET/SET                 // 
				/*****************************************/ 





/**
 * \brief Returns the element found at the
 *        position <index> in the vector vec. 
 * \param vec The vector
 * \param index The index where to put the value
 * \param type An existing data type in which to 
 * 		  store the value or default to read from
 *        the highest available one
 * \return val The value of the element
 */
CUKR_DOUBLE CukrVecmpGetElement(CUKR_VECTOR_MP vec, int index, int type);

// ------------------------------------------------------------------- //

/**
 * \brief Sets the element with index <index>
 *        to the value specified by <val> 
 *        Currently, setting can only
 *        be done on the CPU, so data will be 
 *        set on the CPU and the status set to
 *        there. 
 * \param vec The vector
 * \param index The index where to put the value
 * \param val The value the vector 
 *        elements are set to
 * \param type An existing data type in which to 
 * 		  store the value or default for all 
 *        available precisions
 */
void CukrVecmpSetElement(CUKR_VECTOR_MP vec, int index, CUKR_DOUBLE val, int type);



/**
 * \todo  Should make a simple kernel that
 *        sets all the values of a vector to
 *        a specicific value, so that we can
 *        set directly on the GPU. 
 * \brief Sets all the elements
 *        of the vector to <val>. 
 *        Currently, setting can only
 *        be done on the CPU, so data will be 
 *        set on the CPU and the status set to
 *        there. 
 * \param vec The vector
 * \param val The value the vector 
 *        elements are set to
 * \param type An existing data type in which to 
 * 		  store the values or default for the 
 *        highest one available
 */
void CukrVecmpSetFromVal(CUKR_VECTOR_MP vec, CUKR_DOUBLE val, int type);

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets all elements of a 
 *        vector from a double precision vector
 *        <val>. The vector must reside on the host.
 *        Since data
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecmpSetFromVectorD(CUKR_VECTOR_MP vec, int index, int count, CUKR_DOUBLE* val, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets all elements of a 
 *        vector from a quasi-double precision 
 *        vector <val>.
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecmpSetFromVectorQ(CUKR_VECTOR_MP vec, int index, int count, CUKR_SINGLE* hval, CUKR_SINGLE* tval, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets all elements of a 
 *        vector from a single precision vector
 *        <val>.
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecmpSetFromVectorS(CUKR_VECTOR_MP vec, int index, int count, CUKR_SINGLE* val, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets all elements of a 
 *        vector from a integer vector <val>. 
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecmpSetFromVectorI(CUKR_VECTOR_MP vec, int index, int count, CUKR_INT* val, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to a double precision vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecmpGetToVectorD(CUKR_VECTOR_MP vec, int index, int count, CUKR_DOUBLE* val, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to a quasi-double precision vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecmpGetToVectorQ(CUKR_VECTOR_MP vec, int index, int count, CUKR_SINGLE* hval, CUKR_SINGLE* tval, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to a single precision vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecmpGetToVectorS(CUKR_VECTOR_MP vec, int index, int count, CUKR_SINGLE* val, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to an integer vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecmpGetToVectorI(CUKR_VECTOR_MP vec, int index, int count, CUKR_INT* val, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Sets all elements of the
 *        vector to random values, on the CPUs. 
 *        The rest of the precisions and 
 *        locations are invalidated.
 * \param vec The vector
 * \param min,max The bounds for the random number
 * \param seed The seed of the random number gen.
 * \param type Data type where to write the data
 *             or default to match the solver type
 */
void CukrVecmpSetRandom(CUKR_VECTOR_MP vec, CUKR_DOUBLE min, CUKR_DOUBLE max, int seed, int type);





				/*****************************************/
				//                 I/O                   // 
				/*****************************************/ 






/**
 * \brief Prints the values of a MP vector
 * \param vec The vector being printed
 */
void CukrVecmpPrint(CUKR_VECTOR_MP vec);

// ------------------------------------------------------------------- //

/**
 * \brief Prints the values of a MP vector
 * \param vec The vector being printed
 * \param filename The full path of the file
 *        to write the vector to
 */
void CukrVecmpPrintToFilename(CUKR_VECTOR_MP vec, char* filename);

// ------------------------------------------------------------------- //

/**
 * \brief Prints the values of a MP vector
 * \param vec The vector being printed
 * \param f An open file stream to write to instead
 */
void CukrVecmpPrintToFile(CUKR_VECTOR_MP vec, FILE* f);

// ------------------------------------------------------------------- //

/**
 * \brief Prints the values of a MP vector
 * \param vec The vector being printed
 * \param f An open file stream to write to instead
 * \param indent An indentation string, to be inserted
 *        before every print.
 */
void CukrVecmpPrintToFileWithIndent(CUKR_VECTOR_MP vec, FILE* f, char* indent);


#endif /*CUKR_VECTOR_MP_H_*/
