/**
 * @file mat_vec/sp/csr/csr_matrix_sp.c
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains all the functions needed
 * to support the BCSR matrix format
 */
 

#define LOGGER_NAME "CUKr.mat_vec.sp.bcsr.bcsr_matrix"

#include <stdio.h>
#include <stdlib.h>
#include "../matrix_sp.h"
#include "../../../util/log.h"
#include "../../../util/utils.h"
#include "../../../blas/blas.h"

#define MAX(a,b) (a > b ? a : b)
#define MIN(a,b) (a < b ? a : b)

// ------------------------------------------------------------------- // 

/**
 * \brief Creates a BCSR matrix
 *        with data of the specified type
 * \param mat The matrix structure 
 *        whose BCSR part is being created
 */
void CukrBcsrMatspCreate(CUKR_MATRIX_SP mat)
{
	char buffer[255];
	
		
	 /*
	 * Malloc for the matrix structure
	 */
	 
	 /* Alloc memory */
	 if ( ((mat->bcsr_mat) = (CUKR_BCSR_MATRIX_SP)malloc(sizeof(struct _CUKR_BCSR_MATRIX_SP))) == NULL) {
	 	CUKR_FATAL("Error allocating memory for bcsr matrix object ");
	 } 
	
	LOGX(DEBUG, "Creating matrix %s(%d,%d,%d) as BCSR",
	             mat->name, mat->rows, mat->cols, mat->nz);
	
	
	/* 
	 * Null vectors 
	 */
	
	mat->bcsr_mat->ptr = NULL;
	mat->bcsr_mat->idx = NULL;
	mat->bcsr_mat->val = NULL;
	
}				      

// ------------------------------------------------------------------- // 		      
		      
/**
 * \brief Finalizes the creation of the matrix,
 *        after all elements have been added.
 * \param mat The matrix structure to assemble.
 */
void CukrBcsrMatspAssemble(CUKR_MATRIX_SP mat)
{
	char buffer[255];
	
	CUKR_FATAL("CukrBcsrMatspAssemble has not been implemented yet");
}

// ------------------------------------------------------------------- // 

/**
 * \brief Free memory all memory used
 *        by all data and indexing members
 *        of the matrix, in all precisions,
 *        for both CPU and GPU
 * \param bcsr_mat Pointer to the BCSR matrix
 */
void CukrBcsrMatspDestroy(CUKR_MATRIX_SP mat)
{	
	/* Free all members */
	CukrVecspDestroy(&mat->bcsr_mat->ptr);
	CukrVecspDestroy(&mat->bcsr_mat->idx);
	CukrVecspDestroy(&mat->bcsr_mat->val);
	
	/* Finally, destroy the BCSR matrix structure */
	free(mat->bcsr_mat);
}

// ------------------------------------------------------------------- //   

/**
 * \brief Changes the location of a BCSR matrix
 *        by copying the BCSR data from one location
 *        to another
 * \param mat Pointer to the matrix 
 * \param dest_loc Destination matrix location (CPU,GPU)
 * \param dest_type Destination matrix value type 
 *        (float, double)
 */
void CukrBcsrMatmpTransfer(CUKR_MATRIX_SP mat, int dest_loc, int dest_type)
{
	/*
	 * Validate the parameters 
	 */
	 
	 /* The only allowed destination
	  * types are float ones */
	if (dest_type != CUKR_TYPE_DOUBLE  &&
		dest_type != CUKR_TYPE_QDOUBLE &&
	    dest_type != CUKR_TYPE_SINGLE  ) 
	{
	    CUKR_FATALX(" Trying to convert to integer type " \
	                " floating point data from matrix %s", 
	                mat->name);
	}
	
	/*
	 * Do the transfers (if necessary)
	 */
	
	/* ptr */
	CukrVecspTransfer(mat->bcsr_mat->ptr, dest_loc, CUKR_TYPE_INT);
	/* idx */
	CukrVecspTransfer(mat->bcsr_mat->idx, dest_loc, CUKR_TYPE_INT);
	/* val */
	CukrVecspTransfer(mat->bcsr_mat->val, dest_loc, dest_type);
		 
}

// ------------------------------------------------------------------- //

/**
 * \brief Extracts the diagonal from a matrix
 * \param mat The matrix 
 * \param vec The vector where to store the diagonal
 * \param diag The diagonal to read (zero for the
 *        main diagonal)
 * \param invert Whether to invert it or not 
 */
void CukrBcsrMatspGetDiag(CUKR_MATRIX_SP mat, CUKR_VECTOR_SP vec, int diag, int invert)
{
	int i_row, i_col, k;
	char buffer[255];
	
	LOGX(DEBUG, "Extracting diagonal from BCSR matrix %s", 
	            mat->name);

	/* Make shortcut pointers for more 
	 * readable code */
	CUKR_INT r = mat->bcsr_mat->r;
	CUKR_INT c = mat->bcsr_mat->c;
	CUKR_INT rc = r*c;
	if(mat->variation == CUKR_MATRIX_FORMAT_VARIATION_BCSR4 && (r*c)%2 == 1) rc = r*c + 1;
	CUKR_INT n = mat->rows;
	CUKR_INT nb = ( n%c ? n/c+1 : n/c );
	CUKR_INT    *ptr = mat->bcsr_mat->ptr->ph_ival;
	CUKR_INT    *idx = mat->bcsr_mat->idx->ph_ival;
	CUKR_DOUBLE *val = mat->bcsr_mat->val->ph_dval;
	CUKR_DOUBLE *ddiag = vec->ph_dval;
	
	if(r!=c) CUKR_FATAL("r!=c is NOT supported at CukrBcsrMatspGetDiag");
	
	/*
	 * Create the diag vector 
	 */	
	
	/* Make sure the matrix has double 
	 * precision on host values  */
	CukrBcsrMatmpTransfer(mat, CUKR_LOCATION_CPU, CUKR_TYPE_DOUBLE);

	/*
	 * Put the diagonal elements in the
	 * diagonal vector
	 */

	/* If no inversion */
	if (!invert) {
		for (i_row = -diag; i_row < nb; i_row++)
			if (i_row >= 0)	
				for (i_col = ptr[i_row] - 1; i_col < ptr[i_row + 1] - 1; i_col++)
					if (diag + i_row*r == idx[i_col] - 1) { 
						for(k=0; k<r; k++) ddiag[i_row*r+k] = val[i_col*rc+k*c+k];
						break;
					}
	}

	/* If inversion */
	else {
		for (i_row = -diag; i_row < nb; i_row++)
		{
			if (i_row >= 0)	
				for (i_col = ptr[i_row] - 1; i_col < ptr[i_row + 1] - 1; i_col++)
					if (diag + i_row*r == idx[i_col] - 1) { 
						for(k=0; k<r; k++) ddiag[i_row*r+k] = 1.0 / val[i_col*rc+k*c+k];
						break;
					} 
		}
	}

	LOG(DEBUG, "Finished extracting diagonal ");
}

// ------------------------------------------------------------------- //

/**
 * \brief Prints a matrix to a file stream
 * \param mat The matrix being printed
 * \param f An open file stream to write to
 * \param indent An indentation string, to be inserted
 *        before every print.
 */
void CukrBcsrMatspPrintToFileWithIndent(CUKR_MATRIX_SP mat, FILE *f, char* indent)
{
	char fmt_str[255], var_str[255];
	int i; 
	
	/* 
	 * Print matrix info
	 */
	
	/* Get strings for location and precision*/
	CukrMatFormatIdToStr(mat->format, fmt_str);
	CukrMatFormatVariationIdToStr(mat->variation, var_str);
	
	/* Print the info */
	fprintf(f, "%sMatrix: %s (%dx%d with %d nonzeros) " \
	       "with format BCSR(%d,%d)\n", indent, mat->name,
	       mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c);
	
	/* Merge indents */
	char mindent[255];
	sprintf(mindent, "%s    ", indent);
	
	/*
	 * Print data members 
	 */
	        
	/* ptr */
	fprintf(f, "%sPTR\n", indent);
	CukrVecspPrintToFileWithIndent(mat->bcsr_mat->ptr, f, mindent);
	/* idx */
	fprintf(f, "%sIDX\n", indent);
	CukrVecspPrintToFileWithIndent(mat->bcsr_mat->idx, f, mindent);
	/* val */
	fprintf(f, "%sVAL\n", indent);
	CukrVecspPrintToFileWithIndent(mat->bcsr_mat->val, f, mindent);
}

