/**
 * @file mat_vec/sp/bcsr/bcsr_matrix_sp.h
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains all the functions needed
 * to support the BCSR matrix format
 */

#ifndef CUKR_BCSR_MATRIX_SP_H_
#define CUKR_BCSR_MATRIX_SP_H_

/* =============== FORWARD DECLARATIONS ============== */

typedef struct _CUKR_BCSR_MATRIX_SP *CUKR_BCSR_MATRIX_SP;

/* =================================================== */

#include "../matrix_sp.h"
#include "../vector_sp.h"

/**
 * Holds data for a matrix of type BCSR.
 * Whether the matrix is padded or not, 
 * the matrix is still in BCSR format so
 * data conversion will be done on place.
 * 
 * After the conversion, since only a few
 * zeros are added, multiplication with
 * a standard SpMV routine working in BCSR
 * will give the same results as before.
 * Because of this, the BLASes no not have
 * to be made aware the the format variation.
 */
struct _CUKR_BCSR_MATRIX_SP{
	/** The block size */
	CUKR_INT r,c;
	/** Vector holding the row indices */
	CUKR_VECTOR_SP ptr;
	/** Vector holding the column indices */
	CUKR_VECTOR_SP idx;
	/** Vector holding the values **/
	CUKR_VECTOR_SP val;
};


		




				/*****************************************/
				//       CONSTRUCTOR / DESTRUCTOR        //
				/*****************************************/         







/**
 * \brief Creates a BCSR matrix 
 *        with data of the specified type
 * \param mat The matrix structure 
 *        whose BCSR part is being created
 */
void CukrBcsrMatspCreate(CUKR_MATRIX_SP mat);

// ------------------------------------------------------------------- //

/**
 * \brief Loads a BCSR matrix from a BCSR file.
 *        The BCSR file format is the following:
 *        1st_row: m n nz
 *        2nd->(n+1)th_row: ptr
 *        (n+1)th -> (n+nz+1)th_row: idx
 *        (n+nz+1)th -> (n+2*nz+1)th_row: val
 * \param mat The matrix in which data is being read
 * \param filename The file where the data is
 *        being read from
 */
void CukrBcsrMatspCreateFromFile(CUKR_MATRIX_SP mat, const char *filename);
		       
// ------------------------------------------------------------------- // 		      
		      
/**
 * \brief Finalizes the creation of the matrix,
 *        after all elements have been added.
 * \param mat The matrix structure to assemble.
 */
void CukrBcsrMatspAssemble(CUKR_MATRIX_SP mat);

// ------------------------------------------------------------------- // 

/**
 * \brief Free memory all memory used
 *        by the BCSR part of the matrix
 * \param mat The matrix 
 */
void CukrBcsrMatspDestroy(CUKR_MATRIX_SP mat);








				/*****************************************/
				//          TRANSFER / CONVERT           // 
				/*****************************************/ 

  
  
  
  
  

/**
 * \brief Changes the location of a CSR matrix
 *        by changing the location of the 
 *        data vectors
 * \param mat The matrix
 * \param dest_loc Destination matrix location (CPU,GPU)
 * \param dest_type Destination matrix value type 
 *        (float, double)
 */
void CukrBcsrMatmpTransfer(CUKR_MATRIX_SP mat, int dest_loc, int dest_type);

// ------------------------------------------------------------------- //

/**
 * \brief Changes the storage format of 
 *        the matrix to a diferent variation
 *        or to a completely different format
 * \param mat Pointer to the matrix 
 * \param dest_format Destination format
 * \param dest_variation Destination variation 
 */
void CukrBcsrMatspConvert(CUKR_MATRIX_SP mat, int dest_format, int dest_variation);

                




				/*****************************************/
				//               GET/SET                 // 
				/*****************************************/ 






/**
 * \brief Extracts the diagonal from a matrix
 * \param mat The matrix 
 * \param vec The vector where to store the diagonal
 * \param diag The diagonal to read (zero for the
 *        main diagonal)
 * \param invert Whether to invert it or not 
 */
void CukrBcsrMatspGetDiag(CUKR_MATRIX_SP mat, CUKR_VECTOR_SP vec, int diag, int invert);






				/*****************************************/
				//                 I/O                   // 
				/*****************************************/ 






/**
 * \brief Prints a matrix to a file stream
 * \param mat The matrix being printed
 * \param f An open file stream to write to
 * \param indent An indentation string, to be inserted
 *        before every print.
 */
void CukrBsrMatspPrintToFileWithIndent(CUKR_MATRIX_SP mat, FILE *f, char* indent);


/**
 * \brief Dump a BCSR matrix to a BCSR file.
 * \param mat The matrix
 * \param filename The file
 */
void CukrBcsrMatspDumpToFile(CUKR_MATRIX_SP mat, const char *filename);

#endif /*CUKR_CSR_MATRIX_SP_H_*/
