/**
 * @file mat_vec/sp/bcsr/bcsr_matrix_sp_conv.c
 * @author Serban Georgescu
 * @since 2008
 *
 * Conversion between BCSR format
 * and other formats or variations
 */

#define LOGGER_NAME "CUKr.mat_vec.sp.bcsr.bcsr_matrix_sp_conv"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#ifndef __MACH__
	#include <malloc.h>
#endif
#include "bcsr_matrix_sp.h"
#include "../../../util/log.h"
#include "../../matrix_defs.h"

#define MAX(a,b) (a > b ? a : b)
#define MIN(a,b) (a < b ? a : b)
#define CLOSEST_MUTLIPLE(a,n) (a==0 ? 0 : (a + (n - a%n)))

//========================== FORWARD DECLARATIONS ========================= //

void CukrBcsrMatspConvertToBCSR4(CUKR_MATRIX_SP mat);
void CukrBcsrMatspConvertToBhyb(CUKR_MATRIX_SP mat, int bell_vs_bcsr);

//============================ PUBLIC FUNCTIONS =========================== //

/**
 * \brief Changes the storage format of
 *        the matrix to a diferent variation
 *        or to a completely different format
 * \param mat Pointer to the matrix
 * \param dest_format Destination format
 * \param dest_variation Destination variation
 */
void CukrBcsrMatspConvert(CUKR_MATRIX_SP mat, int dest_format, int dest_variation)
{
	//CUKR_FATAL("CukrBcsrMatspConvert has not been implemented yet");
	
	if(dest_format == CUKR_MATRIX_FORMAT_BCSR){
		if(dest_variation == CUKR_MATRIX_FORMAT_VARIATION_BCSR4){
			CukrBcsrMatspConvertToBCSR4(mat);
		}
	}else if(dest_format == CUKR_MATRIX_FORMAT_BHYB){
		if(mat->variation == CUKR_MATRIX_FORMAT_VARIATION_NONE && dest_variation == CUKR_MATRIX_FORMAT_VARIATION_NONE){
			CukrBcsrMatspConvertToBhyb(mat, CUKR_ELL_VS_COO);
		}else if(mat->variation == CUKR_MATRIX_FORMAT_VARIATION_BCSR4 && dest_variation == CUKR_MATRIX_FORMAT_VARIATION_BHYB4){
			CukrBcsrMatspConvertToBhyb(mat, CUKR_ELL_VS_COO);
		}else if(mat->variation == CUKR_MATRIX_FORMAT_VARIATION_NONE && dest_variation == CUKR_MATRIX_FORMAT_VARIATION_BHYB4){
			CukrBcsrMatspConvertToBhyb(mat, CUKR_ELL_VS_COO);
			CukrBhybMatspConvertToBHYB4(mat);
		}
	}
}

//============================ PRIVATE FUNCTIONS =========================== //

/**
 * \brief Converts a BCSR matrix in generic
 *        format to BCSR4 variation, meaning
 *        that val is padded so that
 *        the no. of elements in each block becomes
 *        multiple of 4.
 * \param mat Pointer to the matrix
 */
void CukrBcsrMatspConvertToBCSR4(CUKR_MATRIX_SP mat)
{
	int i;
	
	/* If the matrix is already in BCSR4
	 * variation, do nothing */
	if (mat->variation ==  CUKR_MATRIX_FORMAT_VARIATION_BCSR4) {
		LOGX(DEBUG, "Matrix %s already in BCSR4 variation. " \
					"Nothing to be done", mat->name);
		return;
	}

	/* block size */
	int r = mat->bcsr_mat->r;
	int c = mat->bcsr_mat->c;
	int rc4; /* r*c + padded-zero */
	
	/* parameters */
	int bcsr_nz = mat->nz; /* # of nonzero blocks */
	
	/* Declare flags to see which data members have
	 * to be updated */
	int pad_double  = mat->bcsr_mat->val->status & CUKR_STATE_CPU_DOUBLE  ? 1 : 0;
	int pad_qdouble = mat->bcsr_mat->val->status & CUKR_STATE_CPU_QDOUBLE ? 1 : 0;
	int pad_single  = mat->bcsr_mat->val->status & CUKR_STATE_CPU_SINGLE  ? 1 : 0;
	
	/* Start Converting */
	if(pad_double){
	
		/* if r*c is multiple of 2, need not to convert */
		if((r*c)%2 == 0){
			mat->variation = CUKR_MATRIX_FORMAT_VARIATION_BCSR4;
			return;
		}
	
		/* padded val size */
		rc4 = r*c + 1; //r*c + (2 - (r*c)%2);

		/*
		 * BCSR val
		 */

		/* tmp_bcsr vector */
		CUKR_DOUBLE *tmp_bcsr;
		if((tmp_bcsr = (CUKR_DOUBLE*)malloc(bcsr_nz * rc4 * sizeof(CUKR_DOUBLE))) == NULL)
			CUKR_FATALX("Error allocating %ld bytes of memory for %s(host,double,padded)",
				(unsigned long)(bcsr_nz * rc4 * sizeof(CUKR_DOUBLE)), mat->bcsr_mat->val->name);
		
		/* val (not padded) */
		CUKR_DOUBLE *bcsr_val = mat->bcsr_mat->val->ph_dval;
		
		/* fill tmp */
		for(i=0; i<bcsr_nz*rc4; i++){
			if(i%rc4 < r*c) tmp_bcsr[i] = bcsr_val[(i/rc4)*r*c + i%rc4];
			else tmp_bcsr[i] = 0.0;
		}
		
		/* Destroy val */
		CukrVecspDestroyData(mat->bcsr_mat->val);
		
		/* New val */
		mat->bcsr_mat->val->n = bcsr_nz * rc4;
		mat->bcsr_mat->val->status = (0|CUKR_STATE_CPU_DOUBLE);
		mat->bcsr_mat->val->ph_dval = tmp_bcsr;
		
	}else if(pad_single){
	
		/* if r*c is multiple of 4, need not to convert */
		if((r*c)%4 == 0){
			mat->variation = CUKR_MATRIX_FORMAT_VARIATION_BCSR4;
			return;
		}
	
		/* padded val size */
		rc4 = r*c + (4 - (r*c)%4);
		
		/*
		 * BCSR val
		 */

		/* tmp_bcsr vector */
		CUKR_SINGLE *tmp_bcsr;
		if((tmp_bcsr = (CUKR_SINGLE*)malloc(bcsr_nz * rc4 * sizeof(CUKR_SINGLE))) == NULL)
			CUKR_FATALX("Error allocating %ld bytes of memory for %s(host,single,padded)",
				(unsigned long)(bcsr_nz * rc4 * sizeof(CUKR_SINGLE)), mat->bcsr_mat->val->name);
		
		/* val (not padded) */
		CUKR_SINGLE *bcsr_val = mat->bcsr_mat->val->ph_sval;
		
		/* fill tmp */
		for(i=0; i<bcsr_nz*rc4; i++){
			if(i%rc4 < r*c) tmp_bcsr[i] = bcsr_val[(i/rc4)*r*c + i%rc4];
			else tmp_bcsr[i] = 0.0;
		}
		
		/* Destroy val */
		CukrVecspDestroyData(mat->bcsr_mat->val);
		
		/* New val */
		mat->bcsr_mat->val->n = bcsr_nz * rc4;
		mat->bcsr_mat->val->status = (0|CUKR_STATE_CPU_SINGLE);
		mat->bcsr_mat->val->ph_sval = tmp_bcsr;
		
	}else if(pad_qdouble){

		/* if r*c is multiple of 4, need not to convert */
		if((r*c)%4 == 0){
			mat->variation = CUKR_MATRIX_FORMAT_VARIATION_BCSR4;
			return;
		}
	
		/* padded val size */
		rc4 = r*c + (4 - (r*c)%4);
		
		/*
		 * BCSR val
		 */

		/* tmp_bcsr vector */
		CUKR_SINGLE *tmp_bcsr, *tmp_bcsr_t;
		if((tmp_bcsr = (CUKR_SINGLE*)malloc(bcsr_nz * rc4 * sizeof(CUKR_SINGLE))) == NULL)
			CUKR_FATALX("Error allocating %ld bytes of memory for %s(host,qdouble,padded)",
				(unsigned long)(bcsr_nz * rc4 * sizeof(CUKR_SINGLE)), mat->bcsr_mat->val->name);
		if((tmp_bcsr_t = (CUKR_SINGLE*)malloc(bcsr_nz * rc4 * sizeof(CUKR_SINGLE))) == NULL)
			CUKR_FATALX("Error allocating %ld bytes of memory for %s(host,qdouble,padded)",
				(unsigned long)(bcsr_nz * rc4 * sizeof(CUKR_SINGLE)), mat->bcsr_mat->val->name);
		
		/* val (not padded) */
		CUKR_SINGLE *bcsr_val  = mat->bcsr_mat->val->ph_sval;
		CUKR_SINGLE *bcsr_valt = mat->bcsr_mat->val->ph_qvalt;
		
		/* fill tmp */
		for(i=0; i<bcsr_nz*rc4; i++){
			if(i%rc4 < r*c){
				tmp_bcsr  [i] = bcsr_val [(i/rc4)*r*c + i%rc4];
				tmp_bcsr_t[i] = bcsr_valt[(i/rc4)*r*c + i%rc4];
			}else{
				tmp_bcsr  [i] = 0.0;
				tmp_bcsr_t[i] = 0.0;
			}
		}
		
		/* Destroy val */
		CukrVecspDestroyData(mat->bcsr_mat->val);
		
		/* New val */
		mat->bcsr_mat->val->n = bcsr_nz * rc4;
		mat->bcsr_mat->val->status = (0|CUKR_STATE_CPU_QDOUBLE);
		mat->bcsr_mat->val->ph_sval  = tmp_bcsr;
		mat->bcsr_mat->val->ph_qvalt = tmp_bcsr_t;

	}else{
		CUKR_FATAL("Invalid format at CukrBcsrMatspConvertToBCSR4");
	}
	
	
	/* change the variation */
	mat->variation = CUKR_MATRIX_FORMAT_VARIATION_BCSR4;

}


/**
 * \brief Converts a BCSR matrix in generic
 *        format to BHYB format
 * \param mat          Pointer to the matrix
 * \param bell_vs_bcsr Number of times BELL format
 *                     is faster than BCSR
 */
void CukrBcsrMatspConvertToBhyb(CUKR_MATRIX_SP mat, int bell_vs_bcsr){
	char buffer[255];
	int i;

	/* If the matrix is already in BHYB
	 * format, do nothing */
	if (mat->format ==  CUKR_MATRIX_FORMAT_BHYB) {
		LOGX(DEBUG, "Matrix %s already in BHYB variation. " \
					"Nothing to be done", mat->name);
		return;
	}
	
	/* Log */
	LOGX(DEBUG, "Converting matrix %s to BHYB format",
	            mat->name);


	/* Get the block size */
	int r = mat->bcsr_mat->r;
	int c = mat->bcsr_mat->c;
	int rc; /* block size (consider zero padding) */
	if(mat->variation == CUKR_MATRIX_FORMAT_VARIATION_NONE) rc = r*c;
	else if(mat->variation == CUKR_MATRIX_FORMAT_VARIATION_BCSR4) rc = r*c + (4 - (r*c)%4);

	/* get the num of row blocks */
	int nb = (mat->rows%r ? mat->rows/r+1 : mat->rows/r);
	int nb_cols = (mat->cols%c ? mat->cols/c+1 : mat->cols/c);
	
	/*
	 * Malloc the BHYB matrix structure
	 */
	 
	/* Alloc memory */
	if ( ((mat->bhyb_mat) = (CUKR_BHYB_MATRIX_SP)malloc(sizeof(struct _CUKR_BHYB_MATRIX_SP))) == NULL) {
		CUKR_FATAL("Error allocating memory for Bhyb matrix object ");
	}

	/* Declare pointers for a more readable code */
	CUKR_INT*    bcsr_ptr_lo = mat->bcsr_mat->ptr->ph_ival;
	CUKR_INT*    bcsr_ptr_hi = &(mat->bcsr_mat->ptr->ph_ival[1]);
	CUKR_INT*    bcsr_idx    = mat->bcsr_mat->idx->ph_ival;
	CUKR_DOUBLE* bcsr_dval   = mat->bcsr_mat->val->ph_dval;
	CUKR_SINGLE* bcsr_sval   = mat->bcsr_mat->val->ph_sval;
	CUKR_SINGLE* bcsr_qvalt  = mat->bcsr_mat->val->ph_qvalt;

	/* set block size */
	mat->bhyb_mat->r = r;
	mat->bhyb_mat->c = c;
	
	/* set # of nonzero blocks per row. Tuning system is not programmed now */
	//mat->bhyb_mat->bell_nz_per_row = nz_per_row;

	/*
	 * Find the optimal no. of nz_per_row for
	 * the BELL format (NVIDIA way)
	 */

	/* Compute max. no of nz/row */
	int max_nz_row = 0;
	for (i = 0; i < nb; i++) {
		int elems = bcsr_ptr_hi[i] - bcsr_ptr_lo[i];
		if (max_nz_row < elems) max_nz_row = elems;
	}

	/* Make nz_per_row histogram */
	int *hist;
	if ( (hist = (int*)(malloc((nb_cols + 1) * sizeof(int)))) == NULL) {
		CUKR_FATALX("Error allocating memory for histogram "\
		            "while converting matrix %s to BHYB format",
		            mat->name);
	}
	memset(hist, 0, (nb_cols + 1) * sizeof(int));
	for (i = 0; i < nb; i++) {
		int elems = bcsr_ptr_hi[i] - bcsr_ptr_lo[i];
		hist[elems]++;
	}

	/* Choose the best nz_per_col
	 * The idea is that the rows which do not
	 * fit in ELL format will be completly removed
	 * and written in COO format (removed, not trimmed).
	 * Therfore the idea is to match the ratio between the
	 * rows in ELL and the ones in COO with the ratio of
	 * the speed of the formats */
	mat->bhyb_mat->bell_nz_per_row = max_nz_row;
	int rows = nb;
	for (i = 0; i < nb_cols + 1; i++) {
		rows -= hist[i];
		if (bell_vs_bcsr*rows < nb) {
			mat->bhyb_mat->bell_nz_per_row = i;
			break;
		}
	}

	free(hist);


	/*
	 * Find the number of nz for both formats
	 */

	/* First get the no. of actual nonzeros (not filling)
	 * for the BELL ans BCSR formats */
	mat->bhyb_mat->bell_nz = 0;
	mat->bhyb_mat->bcsr_nz = 0;
	for (i = 0; i < nb; i++) {
		int elems = bcsr_ptr_hi[i] - bcsr_ptr_lo[i];
		int elems_bell = MIN(mat->bhyb_mat->bell_nz_per_row, elems);
		mat->bhyb_mat->bell_nz += elems_bell;
		mat->bhyb_mat->bcsr_nz += elems - elems_bell;
	}
	
	
	/*
	 * Set BELL stride
	 */

	mat->bhyb_mat->bell_stride = nb;
	
	
	/*
	 * Create BHYB matrix (now that we have the sizes)
	 */

	/* Null vectors */
	mat->bhyb_mat->bell_idx    = NULL;
	mat->bhyb_mat->bell_val    = NULL;
	mat->bhyb_mat->bcsr_ptr    = NULL;
	mat->bhyb_mat->bcsr_idx    = NULL;
	mat->bhyb_mat->bcsr_val    = NULL;
	
	/* Create BELL vectors */
	sprintf(buffer, "%s.bhyb.bell_idx", mat->name);
	CukrVecspCreate(&mat->bhyb_mat->bell_idx, mat->bhyb_mat->bell_nz_per_row * mat->bhyb_mat->bell_stride,      CUKR_TYPE_INT,         buffer);
	sprintf(buffer, "%s.bhyb.bell_val", mat->name);
	CukrVecspCreate(&mat->bhyb_mat->bell_val, mat->bhyb_mat->bell_nz_per_row * mat->bhyb_mat->bell_stride * rc, mat->type_at_creation, buffer);

	/* Declare pointers for more readable code */
	CUKR_INT*    bhyb_bell_idx   = mat->bhyb_mat->bell_idx->ph_ival;
	CUKR_DOUBLE* bhyb_bell_dval  = mat->bhyb_mat->bell_val->ph_dval;
	CUKR_SINGLE* bhyb_bell_sval  = mat->bhyb_mat->bell_val->ph_sval;
	CUKR_SINGLE* bhyb_bell_qvalt = mat->bhyb_mat->bell_val->ph_qvalt;
	
	
	/* Create BCSR vectors*/
	sprintf(buffer, "%s.bhyb.bcsr_ptr", mat->name);
	CukrVecspCreate(&mat->bhyb_mat->bcsr_ptr, nb + 1,                      CUKR_TYPE_INT,         buffer);
	sprintf(buffer, "%s.bhyb.bcsr_idx", mat->name);
	CukrVecspCreate(&mat->bhyb_mat->bcsr_idx, mat->bhyb_mat->bcsr_nz,      CUKR_TYPE_INT,         buffer);
	sprintf(buffer, "%s.bhyb.bcsr_val", mat->name);
	CukrVecspCreate(&mat->bhyb_mat->bcsr_val, mat->bhyb_mat->bcsr_nz * rc, mat->type_at_creation, buffer);

	/* Declare pointers for more readable code */
	CUKR_INT*    bhyb_bcsr_ptr_lo = mat->bhyb_mat->bcsr_ptr->ph_ival;
	CUKR_INT*    bhyb_bcsr_ptr_hi = &(mat->bhyb_mat->bcsr_ptr->ph_ival[1]);
	CUKR_INT*    bhyb_bcsr_idx    = mat->bhyb_mat->bcsr_idx->ph_ival;
	CUKR_DOUBLE* bhyb_bcsr_dval   = mat->bhyb_mat->bcsr_val->ph_dval;
	CUKR_SINGLE* bhyb_bcsr_sval   = mat->bhyb_mat->bcsr_val->ph_sval;
	CUKR_SINGLE* bhyb_bcsr_qvalt  = mat->bhyb_mat->bcsr_val->ph_qvalt;

	
	/*
	 * Fill BHYB matrix (now that we have the sizes)
	 */

	/* Set vectors to zero */
	CukrVecspSetFromVal(mat->bhyb_mat->bell_idx, 0, CUKR_TYPE_INT);
	CukrVecspSetFromVal(mat->bhyb_mat->bell_val, 0, CUKR_DEFAULT);
	CukrVecspSetFromVal(mat->bhyb_mat->bcsr_ptr, 0, CUKR_TYPE_INT);
	CukrVecspSetFromVal(mat->bhyb_mat->bcsr_idx, 0, CUKR_TYPE_INT);
	CukrVecspSetFromVal(mat->bhyb_mat->bcsr_val, 0, CUKR_DEFAULT);
	
	/* Fill BELL and BCSR */
	int bhyb_bcsr_count = 0, bhyb_bcsr_ptr_val = 1;
	for (i = 0; i < nb; i++) {

		int n      = 0;
		int j      = bcsr_ptr_lo[i] - 1;
		int row_up = bcsr_ptr_hi[i] - 1;
		int k; /* iterator in block */

		/* Fill BELL with up to bell_nz_per_row elements */
		while (j < row_up && n < mat->bhyb_mat->bell_nz_per_row) {
			bhyb_bell_idx[mat->bhyb_mat->bell_stride*n + i] = bcsr_idx[j];
			if (mat->type_at_creation == CUKR_TYPE_SINGLE || mat->type_at_creation == CUKR_TYPE_QDOUBLE)
				for(k=0; k<rc; k++) bhyb_bell_sval[(mat->bhyb_mat->bell_stride*n + i)*rc + k] = bcsr_sval[j*rc + k];
			if (mat->type_at_creation == CUKR_TYPE_QDOUBLE)
				for(k=0; k<rc; k++) bhyb_bell_qvalt[(mat->bhyb_mat->bell_stride*n + i)*rc + k] = bcsr_qvalt[j*rc + k];
			if (mat->type_at_creation == CUKR_TYPE_DOUBLE)
				for(k=0; k<rc; k++) bhyb_bell_dval[(mat->bhyb_mat->bell_stride*n + i)*rc + k] = bcsr_dval[j*rc + k];
			j++; n++;
		}

		/* Add the rest to BCSR */
		bhyb_bcsr_ptr_lo[i] = bhyb_bcsr_ptr_val;
		int bhyb_bcsr_count_for_cur_row = 0;
		while (j < row_up) {
			bhyb_bcsr_idx[bhyb_bcsr_count]   = bcsr_idx[j];
			if (mat->type_at_creation == CUKR_TYPE_SINGLE || mat->type_at_creation == CUKR_TYPE_QDOUBLE)
				for(k=0; k<rc; k++) bhyb_bcsr_sval[bhyb_bcsr_count*rc + k]  = bcsr_sval[j*rc + k];
			if (mat->type_at_creation == CUKR_TYPE_QDOUBLE)
				for(k=0; k<rc; k++) bhyb_bcsr_qvalt[bhyb_bcsr_count*rc + k] = bcsr_qvalt[j*rc + k];
			if ( mat->type_at_creation == CUKR_TYPE_DOUBLE)
				for(k=0; k<rc; k++) bhyb_bcsr_dval[bhyb_bcsr_count*rc + k]  = bcsr_dval[j*rc + k];
			j++; bhyb_bcsr_count_for_cur_row++; bhyb_bcsr_count++;
		}

		bhyb_bcsr_ptr_val += bhyb_bcsr_count_for_cur_row;
		bhyb_bcsr_ptr_hi[i] = bhyb_bcsr_ptr_val;

	}
	
	
	/*
	 * Free the BCSR structures
	 */

	CukrVecspDestroyData(mat->bcsr_mat->ptr);
	CukrVecspDestroyData(mat->bcsr_mat->idx);
	CukrVecspDestroyData(mat->bcsr_mat->val);

	/*
	 * Set new matrix format
	 */

	mat->format = CUKR_MATRIX_FORMAT_BHYB;
	if(mat->variation == CUKR_MATRIX_FORMAT_VARIATION_BCSR4) mat->variation = CUKR_MATRIX_FORMAT_VARIATION_BHYB4;


}
