/**
 * @file mat_vec/sp/bcsr/bcsr_matrix_sp_gen.c
 * @author Serban Georgescu
 * @since 2008
 *
 * BCSR matrix generation routines
 * Yusa codes in winter, 2009.
 */

#define LOGGER_NAME "CUKr.mat_vec.sp.bcsr.bcsr_matrix_sp_gen"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#ifndef __MACH__
	#include <malloc.h>
#endif
#include "bcsr_matrix_sp_gen.h"
#include "../../matrix_defs.h"
#include "../../../util/log.h" 

#define MAX(a,b) (a > b ? a : b)
#define MIN(a,b) (a < b ? a : b)
#define ROUND(a) ((int)(a + 0.5) == (int)a ? (int)a : (int)(a+1)) 



/************ Forward declarations *******************/

void gen_unique_int_rnd(int *x, int *w, int n, int lo, int up);


// ------------------------------------------------------------------- //
					 
/**
 * \brief Generates a sparse random matrix, filled
 *        with random elements 
 * \param mat Pointer to the matrix 
 * \param nn Matrix order (Matrix size is nn x nn)
 * \param r,c Block sizes
 * \param name Matrix name
 */
void CukrBcsrMatspGenPoisson2D(CUKR_MATRIX_SP* mat, int nn, int r, int c, char* name)
{

	/* Now, r=c is only suported. */
	if(r!=c) return;
	int b=r; /* b is Size of Block */
	
	/* Matrix size */
	int n=nn*nn;
	/* Half-Bandwidth */
	int m=nn;

	/* Iterator */
	int i, j;
	int ii, jj;
	int iii, jjj;
	int k;
	
	char buffer[255];

	LOGX(DEBUG, "Creating 2D BCSR(%d,%d) Poisson matrix %s of order %d",
			     r, c, name, n);

	/* First create the matrix
	 * Creating in double precision is enough.  */
	CukrMatspCreate(mat, n, n, CUKR_MATRIX_FORMAT_BCSR, CUKR_TYPE_DOUBLE, name);

	/* Set the elements */
	CUKR_DOUBLE diagonal = 4.0;
	CUKR_DOUBLE third = -1.0;
	CUKR_DOUBLE fifth = -1.0;

	/* get # of blocks a row or a col */
	int nb = (n%b ? n/b+1 : n/b);

	/*
	 * Set some matrix properties
	 */
	/* set n to rows and cols */
	(*mat)->rows = n;
	(*mat)->cols = n;

	/* get nz by counting. nz is nun of nonzero blocks */
	k=0;
	for(j=1; j<=nb; j++){
		for(i=1; i<=nb; i++){
			if(i==j) k++; /* Diagonal */
			if((i==j+1) && (j*b)%m!=0) k++; /* i+1 diagonal */
			if((i==j-1) && (i*b)%m!=0) k++; /* i-1 diagonal */
			if((i*b-((j-1)*b+1)>=m) && (((i-1)*b+1)-j*b<=m)) k++; /* i+m diagonal */
			if((j*b-((i-1)*b+1)>=m) && (((j-1)*b+1)-i*b<=m)) k++; /* i-m diagonal */
		}
	}
	(*mat)->nz = k;

	/*
	 * Create data members
	 */
	/* r, c */
	(*mat)->bcsr_mat->r = r;
	(*mat)->bcsr_mat->c = c;
	/* ptr */
	sprintf(buffer, "%s.ptr", (*mat)->name);
	CukrVecspCreate(&(*mat)->bcsr_mat->ptr, nb+1,             CUKR_TYPE_INT,    buffer);
	/* idx */
	sprintf(buffer, "%s.idx", (*mat)->name);
	CukrVecspCreate(&(*mat)->bcsr_mat->idx, (*mat)->nz,       CUKR_TYPE_INT,    buffer);
	/* val */
	sprintf(buffer, "%s.val", (*mat)->name);
	CukrVecspCreate(&(*mat)->bcsr_mat->val, (*mat)->nz * b*b, CUKR_TYPE_DOUBLE, buffer);


	/* Make shortcuts for easy reading */
	CUKR_INT* ptr = (*mat)->bcsr_mat->ptr->ph_ival;
	CUKR_INT* idx = (*mat)->bcsr_mat->idx->ph_ival;
	CUKR_DOUBLE* val = (*mat)->bcsr_mat->val->ph_dval;

	/* Create Poisson Matrix in BCSR */
	k=0;
	for(j=1; j<=nb; j++){
		/* Set ptr once a loop */
		ptr[j-1]=k+1;

		for(i=1; i<=nb; i++){
			/* Diagonal */
			if(i==j){
				for(jj=1; jj<=b; jj++){
					for(ii=1; ii<=b; ii++){
						/* absolute coordinate */
						iii=(i-1)*b+ii;
						jjj=(j-1)*b+jj;

						if(iii==jjj) val[k*b*b + (jj-1)*b+(ii-1)]=diagonal;
						else if((iii==jjj+1) && (jjj%m!=0)) val[k*b*b + (jj-1)*b+(ii-1)]=third;
						else if((iii==jjj-1) && (jjj%m!=1)) val[k*b*b + (jj-1)*b+(ii-1)]=third;
						else val[k*b*b + (jj-1)*b+(ii-1)]=0.0;

						/* Right and Down edges of the Matrix should be Zero */
						if(iii>n && jjj>n) val[k*b*b + (jj-1)*b+(ii-1)]=0.0;
					}
				}
				idx[k]=(i-1)*b+1;
				k++;
			}

			/* i+1 diagonal */
			if((i==j+1) && (j*b)%m!=0){
				for(jj=1; jj<=b; jj++){
					for(ii=1; ii<=b; ii++){
						/* absolute coordinate */
						iii=(i-1)*b+ii;
						jjj=(j-1)*b+jj;

						if(jj-ii==b-1) val[k*b*b + (jj-1)*b+(ii-1)]=third;
						else val[k*b*b + (jj-1)*b+(ii-1)]=0.0;

						/* Right and Down edges of the Matrix should be Zero */
						if(iii>n && jjj>n) val[k*b*b + (jj-1)*b+(ii-1)]=0.0;
					}
				}
				idx[k]=(i-1)*b+1;
				k++;
			}

			/* i-1 diagonal */
			if((i==j-1) && (i*b)%m!=0){
				for(jj=1; jj<=b; jj++){
					for(ii=1; ii<=b; ii++){
						/* absolute coordinate */
						iii=(i-1)*b+ii;
						jjj=(j-1)*b+jj;

						if(ii-jj==b-1) val[k*b*b + (jj-1)*b+(ii-1)]=third;
						else val[k*b*b + (jj-1)*b+(ii-1)]=0.0;

						/* Right and Down edges of the Matrix should be Zero */
						if(iii>n && jjj>n) val[k*b*b + (jj-1)*b+(ii-1)]=0.0;
					}
				}
				idx[k]=(i-1)*b+1;
				k++;
			}

			/* i+m diagonal */
			if((i*b-((j-1)*b+1)>=m) && (((i-1)*b+1)-j*b<=m)){
				for(jj=1; jj<=b; jj++){
					for(ii=1; ii<=b; ii++){
						/* absolute coordinate */
						iii=(i-1)*b+ii;
						jjj=(j-1)*b+jj;

						if(iii-jjj==m) val[k*b*b + (jj-1)*b+(ii-1)]=fifth;
						else if(jjj-iii==m) val[k*b*b + (jj-1)*b+(ii-1)]=fifth;
						else val[k*b*b + (jj-1)*b+(ii-1)]=0.0;

						/* Right and Down edges of the Matrix should be Zero */
						if(iii>n && jjj>n) val[k*b*b + (jj-1)*b+(ii-1)]=0.0;
					}
				}
				idx[k]=(i-1)*b+1;
				k++;
			}

			/* i-m diagonal */
			if((j*b-((i-1)*b+1)>=m) && (((j-1)*b+1)-i*b<=m)){
				for(jj=1; jj<=b; jj++){
					for(ii=1; ii<=b; ii++){
						/* absolute coordinate */
						iii=(i-1)*b+ii;
						jjj=(j-1)*b+jj;

						if(iii-jjj==m) val[k*b*b + (jj-1)*b+(ii-1)]=fifth;
						else if(jjj-iii==m) val[k*b*b + (jj-1)*b+(ii-1)]=fifth;
						else val[k*b*b + (jj-1)*b+(ii-1)]=0.0;

						/* Right and Down edges of the Matrix should be Zero */
						if(iii>n && jjj>n) val[k*b*b + (jj-1)*b+(ii-1)]=0.0;
					}
				}
				idx[k]=(i-1)*b+1;
				k++;
			}

		}
	}

	/* Set the last element of ptr */
	ptr[nb]=k+1;


	LOG(DEBUG, "Done generating 2D Poisson matrix");
}

/**
 * \brief Generates a sparse random matrix, filled
 *        with random elements 
 * \param mat Pointer to the matrix 
 * \param nn Matrix order (Matrix size is nn x nn)
 * \param r,c Block sizes
 * \param name Matrix name
 */
void CukrBcsrMatspGenPoisson2Dsingle(CUKR_MATRIX_SP* mat, int nn, int r, int c, char* name)
{

	/* Now, r=c is only suported. */
	if(r!=c) return;
	int b=r; /* b is Size of Block */
	
	/* Matrix size */
	int n=nn*nn;
	/* Half-Bandwidth */
	int m=nn;

	/* Iterator */
	int i, j;
	int ii, jj;
	int iii, jjj;
	int k;
	
	char buffer[255];

	LOGX(DEBUG, "Creating 2D BCSR(%d,%d) Poisson matrix %s of order %d",
			     r, c, name, n);

	/* First create the matrix
	 * Creating in double precision is enough.  */
	CukrMatspCreate(mat, n, n, CUKR_MATRIX_FORMAT_BCSR, CUKR_TYPE_SINGLE, name);

	/* Set the elements */
	CUKR_SINGLE diagonal = 4.0;
	CUKR_SINGLE third = -1.0;
	CUKR_SINGLE fifth = -1.0;

	/* get # of blocks a row or a col */
	int nb = (n%b ? n/b+1 : n/b);

	/*
	 * Set some matrix properties
	 */
	/* set n to rows and cols */
	(*mat)->rows = n;
	(*mat)->cols = n;

	/* get nz by counting. nz is nun of nonzero blocks */
	k=0;
	for(j=1; j<=nb; j++){
		for(i=1; i<=nb; i++){
			if(i==j) k++; /* Diagonal */
			if((i==j+1) && (j*b)%m!=0) k++; /* i+1 diagonal */
			if((i==j-1) && (i*b)%m!=0) k++; /* i-1 diagonal */
			if((i*b-((j-1)*b+1)>=m) && (((i-1)*b+1)-j*b<=m)) k++; /* i+m diagonal */
			if((j*b-((i-1)*b+1)>=m) && (((j-1)*b+1)-i*b<=m)) k++; /* i-m diagonal */
		}
	}
	(*mat)->nz = k;

	/*
	 * Create data members
	 */
	/* r, c */
	(*mat)->bcsr_mat->r = r;
	(*mat)->bcsr_mat->c = c;
	/* ptr */
	sprintf(buffer, "%s.ptr", (*mat)->name);
	CukrVecspCreate(&(*mat)->bcsr_mat->ptr, nb+1,             CUKR_TYPE_INT,    buffer);
	/* idx */
	sprintf(buffer, "%s.idx", (*mat)->name);
	CukrVecspCreate(&(*mat)->bcsr_mat->idx, (*mat)->nz,       CUKR_TYPE_INT,    buffer);
	/* val */
	sprintf(buffer, "%s.val", (*mat)->name);
	CukrVecspCreate(&(*mat)->bcsr_mat->val, (*mat)->nz * b*b, CUKR_TYPE_SINGLE, buffer);


	/* Make shortcuts for easy reading */
	CUKR_INT* ptr = (*mat)->bcsr_mat->ptr->ph_ival;
	CUKR_INT* idx = (*mat)->bcsr_mat->idx->ph_ival;
	CUKR_SINGLE* val = (*mat)->bcsr_mat->val->ph_sval;

	/* Create Poisson Matrix in BCSR */
	k=0;
	for(j=1; j<=nb; j++){
		/* Set ptr once a loop */
		ptr[j-1]=k+1;

		for(i=1; i<=nb; i++){
			/* Diagonal */
			if(i==j){
				for(jj=1; jj<=b; jj++){
					for(ii=1; ii<=b; ii++){
						/* absolute coordinate */
						iii=(i-1)*b+ii;
						jjj=(j-1)*b+jj;

						if(iii==jjj) val[k*b*b + (jj-1)*b+(ii-1)]=diagonal;
						else if((iii==jjj+1) && (jjj%m!=0)) val[k*b*b + (jj-1)*b+(ii-1)]=third;
						else if((iii==jjj-1) && (jjj%m!=1)) val[k*b*b + (jj-1)*b+(ii-1)]=third;
						else val[k*b*b + (jj-1)*b+(ii-1)]=0.0;

						/* Right and Down edges of the Matrix should be Zero */
						if(iii>n && jjj>n) val[k*b*b + (jj-1)*b+(ii-1)]=0.0;
					}
				}
				idx[k]=(i-1)*r+1;
				k++;
			}

			/* i+1 diagonal */
			if((i==j+1) && (j*b)%m!=0){
				for(jj=1; jj<=b; jj++){
					for(ii=1; ii<=b; ii++){
						/* absolute coordinate */
						iii=(i-1)*b+ii;
						jjj=(j-1)*b+jj;

						if(jj-ii==b-1) val[k*b*b + (jj-1)*b+(ii-1)]=third;
						else val[k*b*b + (jj-1)*b+(ii-1)]=0.0;

						/* Right and Down edges of the Matrix should be Zero */
						if(iii>n && jjj>n) val[k*b*b + (jj-1)*b+(ii-1)]=0.0;
					}
				}
				idx[k]=(i-1)*r+1;
				k++;
			}

			/* i-1 diagonal */
			if((i==j-1) && (i*b)%m!=0){
				for(jj=1; jj<=b; jj++){
					for(ii=1; ii<=b; ii++){
						/* absolute coordinate */
						iii=(i-1)*b+ii;
						jjj=(j-1)*b+jj;

						if(ii-jj==b-1) val[k*b*b + (jj-1)*b+(ii-1)]=third;
						else val[k*b*b + (jj-1)*b+(ii-1)]=0.0;

						/* Right and Down edges of the Matrix should be Zero */
						if(iii>n && jjj>n) val[k*b*b + (jj-1)*b+(ii-1)]=0.0;
					}
				}
				idx[k]=(i-1)*r+1;
				k++;
			}

			/* i+m diagonal */
			if((i*b-((j-1)*b+1)>=m) && (((i-1)*b+1)-j*b<=m)){
				for(jj=1; jj<=b; jj++){
					for(ii=1; ii<=b; ii++){
						/* absolute coordinate */
						iii=(i-1)*b+ii;
						jjj=(j-1)*b+jj;

						if(iii-jjj==m) val[k*b*b + (jj-1)*b+(ii-1)]=fifth;
						else if(jjj-iii==m) val[k*b*b + (jj-1)*b+(ii-1)]=fifth;
						else val[k*b*b + (jj-1)*b+(ii-1)]=0.0;

						/* Right and Down edges of the Matrix should be Zero */
						if(iii>n && jjj>n) val[k*b*b + (jj-1)*b+(ii-1)]=0.0;
					}
				}
				idx[k]=(i-1)*r+1;
				k++;
			}

			/* i-m diagonal */
			if((j*b-((i-1)*b+1)>=m) && (((j-1)*b+1)-i*b<=m)){
				for(jj=1; jj<=b; jj++){
					for(ii=1; ii<=b; ii++){
						/* absolute coordinate */
						iii=(i-1)*b+ii;
					(i-1)*r+1;	jjj=(j-1)*b+jj;

						if(iii-jjj==m) val[k*b*b + (jj-1)*b+(ii-1)]=fifth;
						else if(jjj-iii==m) val[k*b*b + (jj-1)*b+(ii-1)]=fifth;
						else val[k*b*b + (jj-1)*b+(ii-1)]=0.0;

						/* Right and Down edges of the Matrix should be Zero */
						if(iii>n && jjj>n) val[k*b*b + (jj-1)*b+(ii-1)]=0.0;
					}
				}
				idx[k]=(i-1)*r+1;
				k++;
			}

		}
	}

	/* Set the last element of ptr */
	ptr[nb]=k+1;


	LOG(DEBUG, "Done generating 2D Poisson matrix");
}
