/**
 * @file mat_vec/sp/bcsr/bcsr_matrix_sp_io.c
 * @author Serban Georgescu
 * @since 2008
 *
 * File IO functionality for matrices
 * in BCSR format
 */

#define LOGGER_NAME "CUKr.mat_vec.sp.bcsr.bcsr_matrix_sp_io"

#include <stdio.h>
#include <stdlib.h>
#include "../../../util/log.h"

#include "bcsr_matrix_sp.h"

/**
 * \brief Loads a BCSR matrix from a BCSR file.
 * \param mat The matrix in which data is being read
 * \param filename The file where the data is
 *        being read from
 */
void CukrBcsrMatspCreateFromFile(CUKR_MATRIX_SP mat, const char *filename)
{
	FILE *f;
	int i, rows, cols, nz, r, c, res;
	char buffer[255];
	
	LOGX(DEBUG, "Reading BCSR matrix from file: %s", filename);

	
	/*
	 * Malloc for the bcsr member structure
	 */
	 
	 /* Alloc memory */
	 if ( ((mat->bcsr_mat) = (CUKR_BCSR_MATRIX_SP)malloc(sizeof(struct _CUKR_BCSR_MATRIX_SP))) == NULL) {
	 	CUKR_FATAL("Error allocating memory for bcsr matrix object ");
	 } 
			
	 LOGX(DEBUG, "Read matrix %s as BCSR", mat->name);

	 
	/* Open the file and read the
	 * matrix sizes */
	
	/* Open the file for reading */
	if ( (f = fopen(filename, "r")) == NULL) {
		CUKR_FATALX("Unable to open file %s for read",
		            filename);;
	}    
	
	/* Read the dimensions */
	res = fscanf(f, "%d %d %d %d %d", &rows, &cols, &nz, &r, &c);
	mat->rows = rows;
	mat->cols = cols;
	mat->nz = nz;
	mat->bcsr_mat->r = r;
	mat->bcsr_mat->c = c;

	/* get # of blocks a row or a col */
	int nb = (rows%r ? rows/r+1 : rows/r);

	/*
	 * Create data mambers
	 */
	 
	/* ptr */ 
	sprintf(buffer, "%s.ptr", mat->name);
	CukrVecspCreate(&mat->bcsr_mat->ptr, nb+1,   CUKR_TYPE_INT, 	buffer);
	/* idx */
	sprintf(buffer, "%s.idx",  mat->name);
	CukrVecspCreate(&mat->bcsr_mat->idx, nz,     CUKR_TYPE_INT, 	buffer);
	/* val */
	sprintf(buffer, "%s.val",  mat->name);
	CukrVecspCreate(&mat->bcsr_mat->val, nz*r*c, mat->type_at_creation, buffer); 
	
	              
	/*
	 * Read data in the members
	 */              
	   
	/* Read ptr */
	CukrVecspReadDataFromFile(mat->bcsr_mat->ptr, f, CUKR_TYPE_INT); 
	/* Read idx */
	CukrVecspReadDataFromFile(mat->bcsr_mat->idx, f, CUKR_TYPE_INT); 
	/* Read val */
	CukrVecspReadDataFromFile(mat->bcsr_mat->val, f, mat->type_at_creation); 
	
	/*
	 * Close 
	 */
	
	fclose(f);
	LOGX(INFO, "Finished reading matrix %s(%d,%d,%d) "\
	           "from file %s", mat->name, mat->rows,  
	           mat->cols, mat->nz, filename);
}

// ------------------------------------------------------------------- //

/**
 * \brief Dump a BCSR matrix to a BCSR file.
 * \param mat The matrix
 * \param filename The file
 */
void CukrBcsrMatspDumpToFile(CUKR_MATRIX_SP mat, const char *filename)
{
	FILE *f;
	int i;
	
	LOGX(DEBUG, "Dumping BCSR matrix to file: %s", filename);

	/* Reffer ptr, idx, val */
	CUKR_INT *ptr = mat->bcsr_mat->ptr->ph_ival;
	CUKR_INT *idx = mat->bcsr_mat->idx->ph_ival;
	CUKR_DOUBLE *val = mat->bcsr_mat->val->ph_dval;
	
	/* Get size of ptr, idx, val */
	int size_ptr = (mat->rows % mat->bcsr_mat->r ? mat->rows / mat->bcsr_mat->r +1 : mat->rows / mat->bcsr_mat->r) + 1;
	int size_idx = mat->nz;
	int size_val = mat->nz * mat->bcsr_mat->r * mat->bcsr_mat->c;

	/* Open the file for dumping */
	if ( (f = fopen(filename, "w")) == NULL) {
		CUKR_FATALX("Unable to open file %s for dumping",
		            filename);;
	} 
	
	/* Print the header */
	fprintf(f, "%d %d %d %d %d\n", mat->rows, mat->cols, mat->nz, mat->bcsr_mat->r, mat->bcsr_mat->c);
	
	/* Print ptr, idx, val */
	for(i=0; i<size_ptr ;i++) fprintf(f, "%d\n", ptr[i]);
	for(i=0; i<size_idx ;i++) fprintf(f, "%d\n", idx[i]);
	for(i=0; i<size_val ;i++) fprintf(f, "%lf\n", val[i]);

	/* Close the file */	
	fclose(f);
	LOGX(INFO, "Finished dumping matrix %s(%d,%d,%d) "\
	           "to file %s", mat->name, mat->rows,  
	           mat->cols, mat->nz, filename);
}

// ------------------------------------------------------------------- //
