/**
 * @file mat_vec/sp/bhyb/bhyb_matrix_sp.c
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains all the functions needed
 * to support the block hybrid matrix format
 */


#define LOGGER_NAME "CUKr.mat_vec.sp.bhyb.bhyb_matrix"

#include <stdio.h>
#include <stdlib.h>
#include "bhyb_matrix_sp.h"
#include "../matrix_sp.h"
#include "../../../util/log.h"
#include "../../../util/utils.h"
#include "../../../blas/blas.h"

#define MAX(a,b) (a > b ? a : b)
#define MIN(a,b) (a < b ? a : b)



				/*****************************************/
				//       CONSTRUCTOR / DESTRUCTOR        //
				/*****************************************/         



/**
 * \brief Creates a BHYB matrix
 *        with data of the specified type
 * \param mat The matrix structure
 *        whose BHYB part is being created
 */
void CukrBhybMatspCreate(CUKR_MATRIX_SP mat)
{
	char buffer[255];


	 /*
	 * Malloc for the matrix structure
	 */

	 /* Alloc memory */
	 if ( ((mat->bhyb_mat) = (CUKR_BHYB_MATRIX_SP)malloc(sizeof(struct _CUKR_BHYB_MATRIX_SP))) == NULL) {
	 	CUKR_FATAL("Error allocating memory for bhyb matrix object ");
	 }

	LOGX(DEBUG, "Creating matrix %s(%d,%d,%d) as BHYB",
	             mat->name, mat->rows, mat->cols, mat->nz);


	/*
	 * Null vectors
	 */

	mat->bhyb_mat->bell_idx = NULL;
	mat->bhyb_mat->bell_val = NULL;
	mat->bhyb_mat->bcsr_ptr = NULL;
	mat->bhyb_mat->bcsr_idx = NULL;
	mat->bhyb_mat->bcsr_val = NULL;
}

// ------------------------------------------------------------------- //

/**
 * \brief Finalizes the creation of the matrix,
 *        after all elements have been added.
 * \param mat The matrix structure to assemble.
 */
void CukrBhybMatspAssemble(CUKR_MATRIX_SP mat)
{
	char buffer[255];

	CUKR_FATAL("CukrBhybMatspAssemble has not been implemented yet");
}

// ------------------------------------------------------------------- //

/**
 * \brief Free memory all memory used
 *        by the BHYB part of the matrix
 * \param mat The matrix 
 */
void CukrBhybMatspDestroy(CUKR_MATRIX_SP mat)
{
	/* Free all members */
	CukrVecspDestroy(&mat->bhyb_mat->bell_idx);
	CukrVecspDestroy(&mat->bhyb_mat->bell_val);
	CukrVecspDestroy(&mat->bhyb_mat->bcsr_ptr);
	CukrVecspDestroy(&mat->bhyb_mat->bcsr_idx);
	CukrVecspDestroy(&mat->bhyb_mat->bcsr_val);
	
	/* Finally, destroy the BCSR matrix structure */
	free(mat->bhyb_mat);	
} 






				/*****************************************/
				//          TRANSFER / CONVERT           // 
				/*****************************************/ 

  
  
  
  
  

/**
 * \brief Changes the location of a Bhyb matrix
 *        by changing the location of the 
 *        data vectors
 * \param mat The matrix
 * \param dest_loc Destination matrix location (CPU,GPU)
 * \param dest_type Destination matrix value type 
 *        (float, qdouble, double)
 */
void CukrBhybMatmpTransfer(CUKR_MATRIX_SP mat, int dest_loc, int dest_type)
{
	/*
	 * Validate the parameters 
	 */
	 
	 /* The only allowed destination
	  * types are float ones */
	if (dest_type != CUKR_TYPE_DOUBLE  &&
		dest_type != CUKR_TYPE_QDOUBLE &&
	    dest_type != CUKR_TYPE_SINGLE  ) 
	{
	    CUKR_FATALX(" Trying to convert to integer type " \
	                " floating point data from matrix %s", 
	                mat->name);
	}
	
	/*
	 * Do the transfers (if necessary)
	 */
	
	/* bell_idx */
	CukrVecspTransfer(mat->bhyb_mat->bell_idx, dest_loc, CUKR_TYPE_INT);
	/* bell_val */
	CukrVecspTransfer(mat->bhyb_mat->bell_val, dest_loc, dest_type);  
	
	if (mat->bhyb_mat->bcsr_nz) {	
		/* bcsr_ptr */
		CukrVecspTransfer(mat->bhyb_mat->bcsr_ptr, dest_loc, CUKR_TYPE_INT);
		/* bcsr_idx */
		CukrVecspTransfer(mat->bhyb_mat->bcsr_idx, dest_loc, CUKR_TYPE_INT);
		/* bcsr_val */
		CukrVecspTransfer(mat->bhyb_mat->bcsr_val, dest_loc, dest_type);
	}  
}

// ------------------------------------------------------------------- //

/**
 * \brief Changes the storage format of 
 *        the matrix to a diferent variation
 *        or to a completely different format
 * \param mat Pointer to the matrix 
 * \param dest_format Destination format
 * \param dest_variation Destination variation 
 */
/* This function is defined in bhyb_matrix_sp_conv.c
void CukrBhybMatspConvert(CUKR_MATRIX_SP mat, int dest_format, int dest_variation)
{
	CUKR_FATAL("No conversion currently supported from "\
	           "Bhyb format"); 	
}
*/             


				/*****************************************/
				//               GET/SET                 // 
				/*****************************************/ 






/**
 * \brief Extracts the diagonal from a matrix
 * \param mat The matrix 
 * \param vec The vector where to store the diagonal
 * \param diag The diagonal to read (zero for the
 *        main diagonal)
 * \param invert Whether to invert it or not 
 */
void CukrBhybMatspGetDiag(CUKR_MATRIX_SP mat, CUKR_VECTOR_SP vec, int diag, int invert)
{
	CUKR_FATAL("CukrBhybMatspGetDiag not yet implemented");
}







				/*****************************************/
				//                 I/O                   // 
				/*****************************************/ 






/**
 * \brief Prints a matrix to a file stream
 * \param mat The matrix being printed
 * \param f An open file stream to write to
 * \param indent An indentation string, to be inserted
 *        before every print.
 */
void CukrBhybMatspPrintToFileWithIndent(CUKR_MATRIX_SP mat, FILE *f, char* indent)
{
	char fmt_str[255], var_str[255];
	int i; 
	
	/* 
	 * Print matrix info
	 */
	
	/* Get strings for location and precision*/
	CukrMatFormatIdToStr(mat->format, fmt_str);
	CukrMatFormatVariationIdToStr(mat->variation, var_str);
	
	/* Print the info */
	fprintf(f, "%sMatrix: %s (%dx%d with %d nonzeros) " \
	       "with format %s (variation %s)\n", indent, mat->name, 
	       mat->rows, mat->cols, mat->nz,
	       fmt_str, var_str);
	
	/* Merge indents */
	char mindent[255];
	sprintf(mindent, "%s    ", indent);
	
	/*
	 * Print data members 
	 */
	        
	/* bell_idx */
	fprintf(f, "%sBELL_IDX\n", indent);
	CukrVecspPrintToFileWithIndent(mat->bhyb_mat->bell_idx, f, mindent);
	/* bell_val */
	fprintf(f, "%sBELL_VAL\n", indent);
	CukrVecspPrintToFileWithIndent(mat->bhyb_mat->bell_val, f, mindent);
	
	if (mat->bhyb_mat->bcsr_nz) {
		/* bcsr_ptr */
		fprintf(f, "%sBCSR_PTR\n", indent);
		CukrVecspPrintToFileWithIndent(mat->bhyb_mat->bcsr_ptr, f, mindent);
		/* bcsr_idx */
		fprintf(f, "%sBCSR_IDX\n", indent);
		CukrVecspPrintToFileWithIndent(mat->bhyb_mat->bcsr_idx, f, mindent);
		/* bcsr_val */ 
		fprintf(f, "%sBCSR_VAL\n", indent);
		CukrVecspPrintToFileWithIndent(mat->bhyb_mat->bcsr_val, f, mindent);
	}
}

