/**
 * @file mat_vec/sp/bhyb/bhyb_matrix_sp_conv.c
 * @author Serban Georgescu
 * @since 2008
 *
 * Conversion between BHYB format
 * and other formats or variations
 */

#define LOGGER_NAME "CUKr.mat_vec.sp.bhyb.bhyb_matrix_sp_conv"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#ifndef __MACH__
	#include <malloc.h>
#endif
#include "bhyb_matrix_sp.h"
#include "../../../util/log.h"
#include "../../matrix_defs.h"

#define MAX(a,b) (a > b ? a : b)
#define MIN(a,b) (a < b ? a : b)
#define CLOSEST_MUTLIPLE(a,n) (a==0 ? 0 : (a + (n - a%n)))

//========================== FORWARD DECLARATIONS ========================= //

void CukrBhybMatspConvertToBHYB4(CUKR_MATRIX_SP mat);

//============================ PUBLIC FUNCTIONS =========================== //

/**
 * \brief Changes the storage format of
 *        the matrix to a diferent variation
 *        or to a completely different format
 * \param mat Pointer to the matrix
 * \param dest_format Destination format
 * \param dest_variation Destination variation
 */
void CukrBhybMatspConvert(CUKR_MATRIX_SP mat, int dest_format, int dest_variation)
{
	//CUKR_FATAL("CukrBhybMatspConvert has not been implemented yet");
	
	if(dest_format == CUKR_MATRIX_FORMAT_BHYB){
		if(dest_variation == CUKR_MATRIX_FORMAT_VARIATION_BHYB4){
			CukrBhybMatspConvertToBHYB4(mat);
		}
	}
}

//============================ PRIVATE FUNCTIONS =========================== //

/**
 * \brief Converts a BHYB matrix in generic
 *        format to BHYB4 variation, meaning
 *        that val is padded so that
 *        the no. of elements in each block becomes
 *        multiple of 4.
 * \param mat Pointer to the matrix
 */
void CukrBhybMatspConvertToBHYB4(CUKR_MATRIX_SP mat)
{
	int i;
	
	/* If the matrix is already in BHYB4
	 * variation, do nothing */
	if (mat->variation ==  CUKR_MATRIX_FORMAT_VARIATION_BHYB4) {
		LOGX(DEBUG, "Matrix %s already in BHYB4 variation. " \
					"Nothing to be done", mat->name);
		return;
	}

	/* block size */
	int r = mat->bhyb_mat->r;
	int c = mat->bhyb_mat->c;
	int rc4; /* r*c + padded-zero */
	
	/* parameters */
	int bell_nz = mat->bhyb_mat->bell_nz_per_row * mat->bhyb_mat->bell_stride; /* # of nonzero blocks */
	int bcsr_nz = mat->bhyb_mat->bcsr_nz;
	
	/* Declare flags to see which data members have
	 * to be updated */
	int pad_double  = mat->bhyb_mat->bell_val->status & CUKR_STATE_CPU_DOUBLE  ? 1 : 0;
	int pad_qdouble = mat->bhyb_mat->bell_val->status & CUKR_STATE_CPU_QDOUBLE ? 1 : 0;
	int pad_single  = mat->bhyb_mat->bell_val->status & CUKR_STATE_CPU_SINGLE  ? 1 : 0;
	
	/* Start Converting */
	if(pad_double){
	
		/* if r*c is multiple of 2, need not to convert */
		if((r*c)%2 == 0){
			mat->variation = CUKR_MATRIX_FORMAT_VARIATION_BHYB4;
			return;
		}
	
		/* padded val size */
		rc4 = r*c + 1; //r*c + (2 - (r*c)%2);

		/*
		 * BELL val
		 */

		/* tmp_bell vector */
		CUKR_DOUBLE *tmp_bell;
		if((tmp_bell = (CUKR_DOUBLE*)malloc(bell_nz * rc4 * sizeof(CUKR_DOUBLE))) == NULL)
			CUKR_FATALX("Error allocating %ld bytes of memory for %s(host,double,padded)",
				(unsigned long)(bell_nz * rc4 * sizeof(CUKR_DOUBLE)), mat->bhyb_mat->bell_val->name);
		
		/* val (not padded) */
		CUKR_DOUBLE *bell_val = mat->bhyb_mat->bell_val->ph_dval;
		
		/* fill tmp */
		for(i=0; i<bell_nz*rc4; i++){
			if(i%rc4 < r*c) tmp_bell[i] = bell_val[(i/rc4)*r*c + i%rc4];
			else tmp_bell[i] = 0.0;
		}
		
		/* Destroy val */
		CukrVecspDestroyData(mat->bhyb_mat->bell_val);
		
		/* New val */
		mat->bhyb_mat->bell_val->n = bell_nz * rc4;
		mat->bhyb_mat->bell_val->status = (0|CUKR_STATE_CPU_DOUBLE);
		mat->bhyb_mat->bell_val->ph_dval = tmp_bell;
		
		/*
		 * BCSR val
		 */

		/* tmp_bcsr vector */
		CUKR_DOUBLE *tmp_bcsr;
		if((tmp_bcsr = (CUKR_DOUBLE*)malloc(bcsr_nz * rc4 * sizeof(CUKR_DOUBLE))) == NULL)
			CUKR_FATALX("Error allocating %ld bytes of memory for %s(host,double,padded)",
				(unsigned long)(bcsr_nz * rc4 * sizeof(CUKR_DOUBLE)), mat->bhyb_mat->bcsr_val->name);
		
		/* val (not padded) */
		CUKR_DOUBLE *bcsr_val = mat->bhyb_mat->bcsr_val->ph_dval;
		
		/* fill tmp */
		for(i=0; i<bcsr_nz*rc4; i++){
			if(i%rc4 < r*c) tmp_bcsr[i] = bcsr_val[(i/rc4)*r*c + i%rc4];
			else tmp_bcsr[i] = 0.0;
		}
		
		/* Destroy val */
		CukrVecspDestroyData(mat->bhyb_mat->bcsr_val);
		
		/* New val */
		mat->bhyb_mat->bcsr_val->n = bcsr_nz * rc4;
		mat->bhyb_mat->bcsr_val->status = (0|CUKR_STATE_CPU_DOUBLE);
		mat->bhyb_mat->bcsr_val->ph_dval = tmp_bcsr;
		
	}else if(pad_single){
	
		/* if r*c is multiple of 4, need not to convert */
		if((r*c)%4 == 0){
			mat->variation = CUKR_MATRIX_FORMAT_VARIATION_BHYB4;
			return;
		}
	
		/* padded val size */
		rc4 = r*c + (4 - (r*c)%4);

		/*
		 * BELL val
		 */

		/* tmp_bell vector */
		CUKR_SINGLE *tmp_bell;
		if((tmp_bell = (CUKR_SINGLE*)malloc(bell_nz * rc4 * sizeof(CUKR_SINGLE))) == NULL)
			CUKR_FATALX("Error allocating %ld bytes of memory for %s(host,single,padded)",
				(unsigned long)(bell_nz * rc4 * sizeof(CUKR_SINGLE)), mat->bhyb_mat->bell_val->name);
		
		/* val (not padded) */
		CUKR_SINGLE *bell_val = mat->bhyb_mat->bell_val->ph_sval;
		
		/* fill tmp */
		for(i=0; i<bell_nz*rc4; i++){
			if(i%rc4 < r*c) tmp_bell[i] = bell_val[(i/rc4)*r*c + i%rc4];
			else tmp_bell[i] = 0.0;
		}
		
		/* Destroy val */
		CukrVecspDestroyData(mat->bhyb_mat->bell_val);
		
		/* New val */
		mat->bhyb_mat->bell_val->n = bell_nz * rc4;
		mat->bhyb_mat->bell_val->status = (0|CUKR_STATE_CPU_SINGLE);
		mat->bhyb_mat->bell_val->ph_sval = tmp_bell;
		
		/*
		 * BCSR val
		 */

		/* tmp_bcsr vector */
		CUKR_SINGLE *tmp_bcsr;
		if((tmp_bcsr = (CUKR_SINGLE*)malloc(bcsr_nz * rc4 * sizeof(CUKR_SINGLE))) == NULL)
			CUKR_FATALX("Error allocating %ld bytes of memory for %s(host,single,padded)",
				(unsigned long)(bcsr_nz * rc4 * sizeof(CUKR_SINGLE)), mat->bhyb_mat->bcsr_val->name);
		
		/* val (not padded) */
		CUKR_SINGLE *bcsr_val = mat->bhyb_mat->bcsr_val->ph_sval;
		
		/* fill tmp */
		for(i=0; i<bcsr_nz*rc4; i++){
			if(i%rc4 < r*c) tmp_bcsr[i] = bcsr_val[(i/rc4)*r*c + i%rc4];
			else tmp_bcsr[i] = 0.0;
		}
		
		/* Destroy val */
		CukrVecspDestroyData(mat->bhyb_mat->bcsr_val);
		
		/* New val */
		mat->bhyb_mat->bcsr_val->n = bcsr_nz * rc4;
		mat->bhyb_mat->bcsr_val->status = (0|CUKR_STATE_CPU_SINGLE);
		mat->bhyb_mat->bcsr_val->ph_sval = tmp_bcsr;
		
	}else if(pad_qdouble){
	
		/* if r*c is multiple of 4, need not to convert */
		if((r*c)%4 == 0){
			mat->variation = CUKR_MATRIX_FORMAT_VARIATION_BHYB4;
			return;
		}
	
		/* padded val size */
		rc4 = r*c + (4 - (r*c)%4);

		/*
		 * BELL val
		 */

		/* tmp_bell vector */
		CUKR_SINGLE *tmp_bell, *tmp_bell_t;
		if((tmp_bell = (CUKR_SINGLE*)malloc(bell_nz * rc4 * sizeof(CUKR_SINGLE))) == NULL)
			CUKR_FATALX("Error allocating %ld bytes of memory for %s(host,qdouble,padded)",
				(unsigned long)(bell_nz * rc4 * sizeof(CUKR_SINGLE)), mat->bhyb_mat->bell_val->name);
		if((tmp_bell_t = (CUKR_SINGLE*)malloc(bell_nz * rc4 * sizeof(CUKR_SINGLE))) == NULL)
			CUKR_FATALX("Error allocating %ld bytes of memory for %s(host,qdouble,padded)",
				(unsigned long)(bell_nz * rc4 * sizeof(CUKR_SINGLE)), mat->bhyb_mat->bell_val->name);
		
		/* val (not padded) */
		CUKR_SINGLE *bell_val  = mat->bhyb_mat->bell_val->ph_sval;
		CUKR_SINGLE *bell_valt = mat->bhyb_mat->bell_val->ph_qvalt;
		
		/* fill tmp */
		for(i=0; i<bell_nz*rc4; i++){
			if(i%rc4 < r*c){
				tmp_bell  [i] = bell_val [(i/rc4)*r*c + i%rc4];
				tmp_bell_t[i] = bell_valt[(i/rc4)*r*c + i%rc4];
			}else{
				tmp_bell  [i] = 0.0;
				tmp_bell_t[i] = 0.0;
			}
		}
		
		/* Destroy val */
		CukrVecspDestroyData(mat->bhyb_mat->bell_val);
		
		/* New val */
		mat->bhyb_mat->bell_val->n = bell_nz * rc4;
		mat->bhyb_mat->bell_val->status = (0|CUKR_STATE_CPU_QDOUBLE);
		mat->bhyb_mat->bell_val->ph_sval  = tmp_bell;
		mat->bhyb_mat->bell_val->ph_qvalt = tmp_bell_t;
		
		/*
		 * BCSR val
		 */

		/* tmp_bcsr vector */
		CUKR_SINGLE *tmp_bcsr, *tmp_bcsr_t;
		if((tmp_bcsr = (CUKR_SINGLE*)malloc(bcsr_nz * rc4 * sizeof(CUKR_SINGLE))) == NULL)
			CUKR_FATALX("Error allocating %ld bytes of memory for %s(host,qdouble,padded)",
				(unsigned long)(bcsr_nz * rc4 * sizeof(CUKR_SINGLE)), mat->bhyb_mat->bcsr_val->name);
		if((tmp_bcsr_t = (CUKR_SINGLE*)malloc(bcsr_nz * rc4 * sizeof(CUKR_SINGLE))) == NULL)
			CUKR_FATALX("Error allocating %ld bytes of memory for %s(host,qdouble,padded)",
				(unsigned long)(bcsr_nz * rc4 * sizeof(CUKR_SINGLE)), mat->bhyb_mat->bcsr_val->name);
		
		/* val (not padded) */
		CUKR_SINGLE *bcsr_val  = mat->bhyb_mat->bcsr_val->ph_sval;
		CUKR_SINGLE *bcsr_valt = mat->bhyb_mat->bcsr_val->ph_qvalt;
		
		/* fill tmp */
		for(i=0; i<bcsr_nz*rc4; i++){
			if(i%rc4 < r*c){
				tmp_bcsr  [i] = bcsr_val [(i/rc4)*r*c + i%rc4];
				tmp_bcsr_t[i] = bcsr_valt[(i/rc4)*r*c + i%rc4];
			}else{
				tmp_bcsr  [i] = 0.0;
				tmp_bcsr_t[i] = 0.0;
			}
		}
		
		/* Destroy val */
		CukrVecspDestroyData(mat->bhyb_mat->bcsr_val);
		
		/* New val */
		mat->bhyb_mat->bcsr_val->n = bcsr_nz * rc4;
		mat->bhyb_mat->bcsr_val->status = (0|CUKR_STATE_CPU_QDOUBLE);
		mat->bhyb_mat->bcsr_val->ph_sval  = tmp_bcsr;
		mat->bhyb_mat->bcsr_val->ph_qvalt = tmp_bcsr_t;
	
	}else{
		CUKR_FATAL("Invalid format at CukrBcsrMatspConvertToBCSR4");
	}
	
	
	/* change the variation */
	mat->variation = CUKR_MATRIX_FORMAT_VARIATION_BHYB4;

}

