/**
 * @file mat_vec/sp/csr/csr_matrix_sp.c
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains all the functions needed
 * to support the CSR matrix format
 */
 

#define LOGGER_NAME "CUKr.mat_vec.sp.csr.csr_matrix"

#include <stdio.h>
#include <stdlib.h>
#include "../matrix_sp.h"
#include "../../../util/log.h"
#include "../../../util/utils.h"
#include "../../../blas/blas.h"

#define MAX(a,b) (a > b ? a : b)
#define MIN(a,b) (a < b ? a : b)

// ------------------------------------------------------------------- // 

/**
 * \brief Creates a CSR matrix 
 *        with data of the specified type
 * \param mat The matrix structure 
 *        whose CSR part is being created
 */
void CukrCsrMatspCreate(CUKR_MATRIX_SP mat)
{
	char buffer[255];
	
		
	 /*
	 * Malloc for the matrix structure
	 */
	 
	 /* Alloc memory */
	 if ( ((mat->csr_mat) = (CUKR_CSR_MATRIX_SP)malloc(sizeof(struct _CUKR_CSR_MATRIX_SP))) == NULL) {
	 	CUKR_FATAL("Error allocating memory for csr matrix object ");
	 } 
	
	LOGX(DEBUG, "Creating matrix %s(%d,%d,%d) as CSR", 
	             mat->name, mat->rows, mat->cols, mat->nz);
	
	
	/* 
	 * Null vectors 
	 */
	
	mat->csr_mat->ptr = NULL;
	mat->csr_mat->idx = NULL;
	mat->csr_mat->val = NULL;
	
}				      

// ------------------------------------------------------------------- // 		      
		      
/**
 * \brief Finalizes the creation of the matrix,
 *        after all elements have been added.
 * \param mat The matrix structure to assemble.
 */
void CukrCsrMatspAssemble(CUKR_MATRIX_SP mat)
{
	char buffer[255];
	
	/* If the matrix was read from file,
	 * nothing to do */
	
	if (mat->build_phase == CUKR_MAT_BUILD_CREATED_FILE)
		return; 	
	
	LOGX(DEBUG, "Starting the assembly for matrix %s "\
	            "in CSR format", mat->name);
	
	 
	/*
	 * Create data structures 
	 */
	
	/* ptr */ 
	sprintf(buffer, "%s.ptr", mat->name);
	CukrVecspCreate(&mat->csr_mat->ptr, mat->rows + 1, CUKR_TYPE_INT, 		buffer);
	/* idx */
	sprintf(buffer, "%s.idx", mat->name);
	CukrVecspCreate(&mat->csr_mat->idx, mat->nz,       CUKR_TYPE_INT, 		buffer);
	/* val */
	sprintf(buffer, "%s.val", mat->name);
	CukrVecspCreate(&mat->csr_mat->val, mat->nz,       mat->type_at_creation, buffer);
	
	/*
	 * Sort entry container by rows and 
	 * cols and remove any duplicates
	 */
	 
	 CukrContainersVectorOfMatentrySortUnique(mat->entries);
	 
	 /*
	  * Build the CSR data structure
	  */
	 
	 int i, j;
	 int current_ptr = 1;           // Start from 1
	 int current_row_nz = 0;        // No nonzeros
	 int last_row = 1;              // Last row set to 0 (first one)
	 
	 /* First pointer set to 1 */
	 mat->csr_mat->ptr->ph_ival[0] = current_ptr;
	  
	 /* Process all nozeros one by one */
	 for (i = 0; i < mat->nz; i++) 
	 { 	
	 	/* Write idx */
	 	mat->csr_mat->idx->ph_ival[i] = mat->entries->elements[i].col;
	 	
	 	/* Read nz row */
	 	int row = mat->entries->elements[i].row;	 	
	 	
	 	/* If the row is changed */
	 	if (row != last_row) {	 			 	
	 		
	 		/* Set all empty entries between
	 		 * the last_row and row */
	 		current_ptr += current_row_nz;
	 		for (j = last_row; j < row; j++)
	 			 mat->csr_mat->ptr->ph_ival[j] = current_ptr;		
	 				 		 		
	 		/* Update last row and reset current_row_nz*/
	 		last_row = row;	 	 		
	 		current_row_nz = 1;
	 	} 	 
	 	/* If not changed, increase the no. of nz
	 	 * for the current row */	
	 	else
	 		current_row_nz++;
	 }
	 
	 /* Write from last row to the
	  * end of the matrix*/
	 for (j = last_row; j <= mat->rows; j++) 
        mat->csr_mat->ptr->ph_ival[j] = current_ptr + current_row_nz;
	 
	 	 
	 /* Add the nonzeros, depending on type */
	 if (mat->type_at_creation == CUKR_TYPE_DOUBLE)
	 	for (i = 0; i < mat->nz; i++)
	 		mat->csr_mat->val->ph_dval[i] = mat->entries->elements[i].val;
	 else if (mat->type_at_creation == CUKR_TYPE_QDOUBLE)
	 	for (i = 0; i < mat->nz; i++) {
	 		mat->csr_mat->val->ph_sval[i]  = (CUKR_SINGLE)mat->entries->elements[i].val;
	 		mat->csr_mat->val->ph_qvalt[i] = (CUKR_SINGLE)(mat->entries->elements[i].val - mat->csr_mat->val->ph_sval[i]);
	 	}
	 else if (mat->type_at_creation == CUKR_TYPE_SINGLE)
	 	 for (i = 0; i < mat->nz; i++)
	 		mat->csr_mat->val->ph_sval[i]  = (CUKR_SINGLE)mat->entries->elements[i].val;		
	 else
	 	 CUKR_FATALX("Invalid data type for matrix %s: %d",
	 	             mat->name, mat->type_at_creation );
	 
	 LOGX(DEBUG, "Done assembling matrix %s "\
	             "in CSR format", mat->name);
}

// ------------------------------------------------------------------- // 

/**
 * \brief Free memory all memory used
 *        by all data and indexing members
 *        of the matrix, in all precisions,
 *        for both CPU and GPU
 * \param csr_mat Pointer to the CSR matrix
 */
void CukrCsrMatspDestroy(CUKR_MATRIX_SP mat) 
{	
	/* Free all members */
	CukrVecspDestroy(&mat->csr_mat->ptr);
	CukrVecspDestroy(&mat->csr_mat->idx);
	CukrVecspDestroy(&mat->csr_mat->val);
	
	/* Finally, destroy the CSR matrix structure */
	free(mat->csr_mat);
}

// ------------------------------------------------------------------- //   

/**
 * \brief Changes the location of a CSR matrix
 *        by copying the CSR data from one location
 *        to another
 * \param mat Pointer to the matrix 
 * \param dest_loc Destination matrix location (CPU,GPU)
 * \param dest_type Destination matrix value type 
 *        (float, double)
 */
void CukrCsrMatmpTransfer(CUKR_MATRIX_SP mat, int dest_loc, int dest_type)
{
	/*
	 * Validate the parameters 
	 */
	 
	 /* The only allowed destination
	  * types are float ones */
	if (dest_type != CUKR_TYPE_DOUBLE  &&
		dest_type != CUKR_TYPE_QDOUBLE &&
	    dest_type != CUKR_TYPE_SINGLE  ) 
	{
	    CUKR_FATALX(" Trying to convert to integer type " \
	                " floating point data from matrix %s", 
	                mat->name);
	}
	
	/*
	 * Do the transfers (if necessary)
	 */
	
	/* ptr */
	CukrVecspTransfer(mat->csr_mat->ptr, dest_loc, CUKR_TYPE_INT);
	/* idx */
	CukrVecspTransfer(mat->csr_mat->idx, dest_loc, CUKR_TYPE_INT);
	/* val */
	CukrVecspTransfer(mat->csr_mat->val, dest_loc, dest_type);  
		 
}

// ------------------------------------------------------------------- //

/**
 * \brief Extracts the diagonal from a matrix
 * \param mat The matrix 
 * \param vec The vector where to store the diagonal
 * \param diag The diagonal to read (zero for the
 *        main diagonal)
 * \param invert Whether to invert it or not 
 */
void CukrCsrMatspGetDiag(CUKR_MATRIX_SP mat, CUKR_VECTOR_SP vec, int diag, int invert)
{
	int i_row, i_col;
	char buffer[255];
	
	LOGX(DEBUG, "Extracting diagonal from CSR matrix %s", 
	            mat->name);
	
	
	/* Make shortcut pointers for more 
	 * readable code */
	 
	 CUKR_INT    n      = mat->rows;
	 CUKR_INT    *ptr   = mat->csr_mat->ptr->ph_ival;
	 CUKR_INT    *idx   = mat->csr_mat->idx->ph_ival;	
	 CUKR_DOUBLE *ddiag = vec->ph_dval;
	 
	/*
	 * Create the diag vector 
	 */	
	
	/* Make sure the matrix has double 
	 * precision on host values  */
	CukrCsrMatmpTransfer(mat, CUKR_LOCATION_CPU, CUKR_TYPE_DOUBLE);
	
	/*
	 * Put the diagonal elements in the
	 * diagonal vector
	 */
	
	
	/* If no inversion */
	if (!invert) {
		for (i_row = -diag; i_row < n; i_row++)
			if (i_row >= 0)	
				for (i_col = ptr[i_row] - 1; i_col < ptr[i_row + 1] - 1; i_col++)
					if (diag + i_row == idx[i_col] - 1) { 
						ddiag[i_row] = mat->csr_mat->val->ph_dval[i_col];						
						break;
					}
	}
	
	/* If inversion */
	else {
		for (i_row = -diag; i_row < n; i_row++)
		{
			if (i_row >= 0)	
				for (i_col = ptr[i_row] - 1; i_col < ptr[i_row + 1] - 1; i_col++)
					if (diag + i_row == idx[i_col] - 1) { 
						ddiag[i_row] = 1.0 / mat->csr_mat->val->ph_dval[i_col];
						break;
					} 
		}
	}	
	 
	LOG(DEBUG, "Finished extracting diagonal ");		
}

// ------------------------------------------------------------------- //

/**
 * \brief Prints a matrix to a file stream
 * \param mat The matrix being printed
 * \param f An open file stream to write to
 * \param indent An indentation string, to be inserted
 *        before every print.
 */
void CukrCsrMatspPrintToFileWithIndent(CUKR_MATRIX_SP mat, FILE *f, char* indent)
{
	char fmt_str[255], var_str[255];
	int i; 
	
	/* 
	 * Print matrix info
	 */
	
	/* Get strings for location and precision*/
	CukrMatFormatIdToStr(mat->format, fmt_str);
	CukrMatFormatVariationIdToStr(mat->variation, var_str);
	
	/* Print the info */
	fprintf(f, "%sMatrix: %s (%dx%d with %d nonzeros) " \
	       "with format %s (variation %s)\n", indent, mat->name, 
	       mat->rows, mat->cols, mat->nz,
	       fmt_str, var_str);
	
	/* Merge indents */
	char mindent[255];
	sprintf(mindent, "%s    ", indent);
	
	/*
	 * Print data members 
	 */
	        
	/* ptr */
	fprintf(f, "%sPTR\n", indent);
	CukrVecspPrintToFileWithIndent(mat->csr_mat->ptr, f, mindent);
	/* idx */
	fprintf(f, "%sIDX\n", indent);
	CukrVecspPrintToFileWithIndent(mat->csr_mat->idx, f, mindent);
	/* val */
	fprintf(f, "%sVAL\n", indent);
	CukrVecspPrintToFileWithIndent(mat->csr_mat->val, f, mindent);
}

