/**
 * @file mat_vec/sp/csr/csr_matrix_sp.h
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains all the functions needed
 * to support the CSR matrix format
 */

#ifndef CUKR_CSR_MATRIX_SP_H_
#define CUKR_CSR_MATRIX_SP_H_

/* =============== FORWARD DECLARATIONS ============== */

typedef struct _CUKR_CSR_MATRIX_SP *CUKR_CSR_MATRIX_SP;

/* =================================================== */

#include "../matrix_sp.h"
#include "../vector_sp.h"

/**
 * Holds data for a matrix of type CSR.
 * Whether the matrix is padded or not, 
 * the matrix is still in CSR format so
 * data conversion will be done on place.
 * 
 * After the conversion, since only a few
 * zeros are added, multiplication with
 * a standard SpMV routine working in CSR
 * will give the same results as before.
 * Because of this, the BLASes no not have
 * to be made aware the the format variation.
 */
struct _CUKR_CSR_MATRIX_SP{   
	/** Vector holding the row indices */
	CUKR_VECTOR_SP ptr;
	/** Vector holding the column indices */
	CUKR_VECTOR_SP idx;
	/** Vector holding the values **/
	CUKR_VECTOR_SP val;
};


		




				/*****************************************/
				//       CONSTRUCTOR / DESTRUCTOR        //
				/*****************************************/         







/**
 * \brief Creates a CSR matrix 
 *        with data of the specified type
 * \param mat The matrix structure 
 *        whose CSR part is being created
 */
void CukrCsrMatspCreate(CUKR_MATRIX_SP mat);	

// ------------------------------------------------------------------- //

/**
 * \brief Loads a CSR matrix from a CSR file.
 *        The CSR file format is the following:
 *        1st_row: m n nz
 *        2nd->(n+1)th_row: ptr
 *        (n+1)th -> (n+nz+1)th_row: idx
 *        (n+nz+1)th -> (n+2*nz+1)th_row: val
 * \param mat The matrix in which data is being read
 * \param filename The file where the data is
 *        being read from
 */
void CukrCsrMatspCreateFromFile(CUKR_MATRIX_SP mat, const char *filename);
		       
// ------------------------------------------------------------------- // 		      
		      
/**
 * \brief Finalizes the creation of the matrix,
 *        after all elements have been added.
 * \param mat The matrix structure to assemble.
 */
void CukrCsrMatspAssemble(CUKR_MATRIX_SP mat);      	      

// ------------------------------------------------------------------- // 

/**
 * \brief Free memory all memory used
 *        by the CSR part of the matrix
 * \param mat The matrix 
 */
void CukrCsrMatspDestroy(CUKR_MATRIX_SP mat); 








				/*****************************************/
				//          TRANSFER / CONVERT           // 
				/*****************************************/ 

  
  
  
  
  

/**
 * \brief Changes the location of a CSR matrix
 *        by changing the location of the 
 *        data vectors
 * \param mat The matrix
 * \param dest_loc Destination matrix location (CPU,GPU)
 * \param dest_type Destination matrix value type 
 *        (float, double)
 */
void CukrCsrMatmpTransfer(CUKR_MATRIX_SP mat, int dest_loc, int dest_type);

// ------------------------------------------------------------------- //

/**
 * \brief Changes the storage format of 
 *        the matrix to a diferent variation
 *        or to a completely different format
 * \param mat Pointer to the matrix 
 * \param dest_format Destination format
 * \param dest_variation Destination variation 
 */
void CukrCsrMatspConvert(CUKR_MATRIX_SP mat, int dest_format, int dest_variation);

                




				/*****************************************/
				//               GET/SET                 // 
				/*****************************************/ 






/**
 * \brief Extracts the diagonal from a matrix
 * \param mat The matrix 
 * \param vec The vector where to store the diagonal
 * \param diag The diagonal to read (zero for the
 *        main diagonal)
 * \param invert Whether to invert it or not 
 */
void CukrCsrMatspGetDiag(CUKR_MATRIX_SP mat, CUKR_VECTOR_SP vec, int diag, int invert);

// ------------------------------------------------------------------- //

/**
 * \brief Generates a sparse random matrix, filled
 *        with random elements, in CSR format 
 * \param mat Pointer to the matrix 
 * \param n Matrix size
 * \param halfb The half-bandwidth of the matrix
 * \param nz_per_row No. of nonzeros per row
 * \param min,max The lower and upper bound for
 *                matrix values
 * \param name Matrix name
 */
void CukrCsrMatspGenRandom(CUKR_MATRIX_SP* mat, int n, int halfb, int nz_per_row, 
						    CUKR_DOUBLE min, CUKR_DOUBLE max, char* name);

// ------------------------------------------------------------------- //
					 
/**
 * \brief Generates a sparse random matrix, filled
 *        with random elements, in CSR format 
 * \param mat Pointer to the matrix 
 * \param n Order
 * \param name Matrix name
 */
void CukrCsrMatspGenPoisson2D(CUKR_MATRIX_SP* mat, int n, char* name);

// ------------------------------------------------------------------- //

/**
 * \brief Generates a diagonally-dominated banded
 *        matrix, filled with ones but the main 
 * 		  diagonal computed as such as to be dominant.
 * 		  The diagonals can be shifted to the left or 
 *        right by specifying a negative or positive shift 
 * \param mat Pointer to the matrix 
 * \param format Matrix format (CSR)
 * \param rows*cols Matrix order
 * \param hb Halfbandwidth 
 * \param shift The shift  
 * \param name Matrix name
 */
void CukrCsrMatspGenBandedWShift(CUKR_MATRIX_SP* mat, int rows, int cols, int hb, int shift, char* name);







				/*****************************************/
				//                 I/O                   // 
				/*****************************************/ 








/**
 * \brief Prints a matrix to a file stream
 * \param mat The matrix being printed
 * \param f An open file stream to write to
 * \param indent An indentation string, to be inserted
 *        before every print.
 */
void CukrCsrMatspPrintToFileWithIndent(CUKR_MATRIX_SP mat, FILE *f, char* indent);

// ------------------------------------------------------------------- //

/**
 * \brief Open the file storing a CSR matrix
 *        and return the matrix dimensions 
 * 		 (rows,cols) and the no. nonzeros
 * \param format The matrix format ID
 * \param rows Pointer to a variable where to 
 *        store the no. of rows
 * \param cols Pointer to a variable where to 
 *        store the no. of cols
 * \param nz Pointer to a variable where to 
 *        store the no. of nz
 */
void CukrCsrMatspReadSizesFromFile(char* filename, int* rows, int* cols, int* nz);

/**
 * \brief Save a SP matrix to file, in readable .csr format
 */
void CukrCsrMatspSaveToFile(CUKR_MATRIX_SP mat, const char *filename);

#endif /*CUKR_CSR_MATRIX_SP_H_*/
