/**
 * @file mat_vec/sp/csr/csr_matrix_sp_conv.c
 * @author Serban Georgescu
 * @since 2008
 *
 * Conversion between CSR format
 * and other formats or variations
 */

#define LOGGER_NAME "CUKr.mat_vec.sp.csr.csr_matrix_sp_conv"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#ifndef __MACH__
	#include <malloc.h>
#endif
#include "csr_matrix_sp.h"
#include "../../../cukr.h"
#include "../../../util/log.h"
#include "../../matrix_defs.h"

#define MAX(a,b) (a > b ? a : b)
#define MIN(a,b) (a < b ? a : b)
#define CLOSEST_MUTLIPLE(a,n) (a==0 ? 0 : (a + (n - a%n)))

//========================== FORWARD DECLARATIONS ========================= //

void CukrCsrMatspConvertToCSRn(CUKR_MATRIX_SP mat, int n);
void CukrCsrMatspConvertToHyb(CUKR_MATRIX_SP mat, int alignment, int ell_vs_coo);


//============================ PUBLIC FUNCTIONS =========================== //


/**
 * \brief   Changes the storage format of a CSR
 *        matrix to a diferent variation
 *        or to a completely different format.
 * 		    Curently only CSR format is supported,
 *        so the dest_format must be CSR as well.
 *        Only the variation can be different.
 *          Also, it is not useful to revert back to
 *        generic format from a variation, since
 *        the generic format works on all variations
 *        as well, so this is not supported as well
 * \param mat Pointer to the matrix
 * \param dest_format Destination format
 * \param dest_variation Destination variation
 */
void CukrCsrMatspConvert(CUKR_MATRIX_SP mat, int dest_format, int dest_variation)
{
	/*
	 * Validate paramteres
	 */

	/* Current format has been validated
	 * before calling this function */

	/* If the destination is CSR */
	if (dest_format == CUKR_MATRIX_FORMAT_CSR)
	{
		/* Destination variation must not be
		 * NONE. If so, do nothing */
		if (dest_variation == CUKR_MATRIX_FORMAT_VARIATION_NONE) {
			LOGX(WARN, "No need to convert matrix %s "\
			           "to generic format. Skipped. ", mat->name);
			return;
		}

		/* If destination is CSR4, call the
		 * conversion procedure */
		else if (dest_variation == CUKR_MATRIX_FORMAT_VARIATION_CSR4) {
			CukrCsrMatspConvertToCSRn(mat, 4);
			mat->variation = CUKR_MATRIX_FORMAT_VARIATION_CSR4;
		}
	}

	/* If destination is HYB */
	else if (dest_format == CUKR_MATRIX_FORMAT_HYB) {
		CukrCsrMatspConvertToHyb(mat, CUKR_ELL_ALIGNMENT, CUKR_ELL_VS_COO);
		mat->variation = CUKR_MATRIX_FORMAT_VARIATION_NONE;
	}

	/* Else, unsupported format*/
	else {
		CUKR_FATALX("Unsuported destination variation" \
		            "for matrix %s", mat->name);;
	}
}

//============================ PRIVATE FUNCTIONS =========================== //

/**
 * \brief Converts a CSR matrix in generic
 *        format to CSR4 variation, meaning
 *        that idx and val are padded so that
 *        the no. of elements in each row becomes
 *        multiple of n.
 * \param mat Pointer to the matrix
 * \param n   The padding size
 */
void CukrCsrMatspConvertToCSRn(CUKR_MATRIX_SP mat, int n)
{

	int i, i_row;
	int rows = mat->rows;

	/* If the matrix is already in CSR4
	 * variation, do nothing */
	if (mat->variation ==  CUKR_MATRIX_FORMAT_VARIATION_CSR4) {
		LOGX(DEBUG, "Matrix %s already in CSR4 variation. " \
					"Nothing to be done", mat->name);
		return;
	}

	/* Log */
	LOGX(DEBUG, "Converting matrix %s to CSR4 variation",
	            mat->name);

	/* Structure to hold element shifts */
	struct SHIFT {
		int pos;     // The position of the shift
		int val;     // The value of the shift (nr. of zeros to be inserted)
	};

	/* Declare pointers for a more readable code */
	CUKR_INT*    ptr   = mat->csr_mat->ptr->ph_ival;
	CUKR_INT*    idx   = mat->csr_mat->idx->ph_ival;
	CUKR_DOUBLE* dval  = mat->csr_mat->val->ph_dval;
	CUKR_SINGLE* sval  = mat->csr_mat->val->ph_sval;
	CUKR_SINGLE* qvalt = mat->csr_mat->val->ph_qvalt;

	/* Declare flags to see which data members have
	 * to be updated */
	int pad_double  = mat->csr_mat->val->status & CUKR_STATE_CPU_DOUBLE  ? 1 : 0;
	int pad_qdouble = mat->csr_mat->val->status & CUKR_STATE_CPU_QDOUBLE ? 1 : 0;
	int pad_single  = mat->csr_mat->val->status & CUKR_STATE_CPU_SINGLE  ? 1 : 0;


	/* Compute the new nz and shifts  ---------------------------------------*/

	int nz = 0;

	/* Declare the shift vector, which stores the padding
	 * we have to make for each row set, so that the val
	 * and idx vectors will be alligned.
	 * The maximum number of shifts can be <rows> as there is
	 * maximum of one shift per row */

	struct SHIFT *shift = NULL;
	if ( (shift = (struct SHIFT *)malloc((rows + 2)* sizeof(struct SHIFT))) == NULL) {
		CUKR_FATALX("Error allocating %ld bytes of memory for shift vector",
		             (unsigned long)((rows + 2) * sizeof(struct SHIFT)));
	}

	/* For every row, compute the no of elements and what
	 * must be added for it to become divisible by n */

	shift[0].val = 0;
	shift[0].pos = 0;
	for (i_row = 1; i_row < rows + 1; i_row++)
	{
		int elems = ptr[i_row] - ptr[i_row - 1];
		shift[i_row].val = n * ceil(elems / (double)n) - elems;
		shift[i_row].pos = ptr[i_row] - 1;
		nz += elems + shift[i_row].val;
	}

	/* Write the padded idx & val --------------------------------------- */

	/*
	 * Alloc memory for the new idx
	 */
	int *idx_padded = NULL;
	if ( (idx_padded = (int*)malloc( nz * sizeof(CUKR_INT))) == NULL) {
		CUKR_FATALX("Error allocating %ld bytes of memory for %s(host,int,padded)",
			        (unsigned long)(nz * sizeof(CUKR_INT)), mat->csr_mat->idx->name);
	}
	/* Log */
	LOGX(DEBUG, "Allocated %d doubles for %s (host,double,padded)",
		        nz, mat->csr_mat->idx->name);

	/* Set all to 1. If we set them to zero, then
	 * when computing SpMV, the idx-1 will have
	 * the value -1, so out of the bounds of the
	 * matrix.
	 * Still, better with 0, since there will be no
	 * crash if JUST reading from there. On the other hand,
	 * if I set it to 1, it is impossible to separate the
	 * padding from real values of 1. This will introduce
	 * artificial communication when using the MP or MPI
	 * versions.*/
	for (i = 0; i < nz; i++)
		idx_padded[i] = 0;

	/*
	 * Alloc memory for double val,
	 * if neccesary
	 */
	CUKR_DOUBLE *dval_padded = NULL;
	if (pad_double) {
		if ( (dval_padded = (CUKR_DOUBLE*)malloc( nz * sizeof(CUKR_DOUBLE))) == NULL) {
			CUKR_FATALX("Error allocating %ld bytes of memory for %s(host,double,padded)",
			             (unsigned long)(nz * sizeof(CUKR_DOUBLE)), mat->csr_mat->val->name);
		}
		/* Log */
		LOGX(DEBUG, "Allocated %d doubles for %s (host,double,padded)",
			        nz, mat->csr_mat->val->name);
		/* Set all to 0 */
		memset(dval_padded, 0, nz * sizeof(CUKR_DOUBLE));
	}

	/*
	 * Alloc memory for single val
	 * or qdouble head, if neccesary
	 */
	CUKR_SINGLE *sval_padded = NULL;
	if (pad_single | pad_qdouble) {
		if ( (sval_padded = (CUKR_SINGLE*)malloc(nz * sizeof(CUKR_SINGLE))) == NULL) {

			if (pad_single && pad_qdouble) {
				CUKR_FATALX("Error allocating %ld bytes of memory for %s(host,single+qdouble head,padded)",
			            	 (unsigned long)(nz * sizeof(CUKR_SINGLE)),
			            	 mat->csr_mat->val->name);
			}
			else if (pad_single) {
				CUKR_FATALX("Error allocating %ld bytes of memory for %s(host,single,padded)",
			            	 (unsigned long)(nz * sizeof(CUKR_SINGLE)),
			            	 mat->csr_mat->val->name);
			}
			else {
				CUKR_FATALX("Error allocating %ld bytes of memory for %s(host,qdouble,head,padded)",
			            	 (unsigned long)(nz * sizeof(CUKR_SINGLE)),
			            	 mat->csr_mat->val->name);
			}
		}

		/* Log */
		if (pad_single && pad_qdouble) {
			LOGX(DEBUG, "Allocated %d floats for %s(host,single+qdouble head,padded)",
			        	nz, mat->csr_mat->val->name);
		}
		else if (pad_single) {
			LOGX(DEBUG, "Allocated %d floats for %s(host,single,padded)",
			        	nz, mat->csr_mat->val->name);
		}
		else {
			LOGX(DEBUG, "Allocated %d floats for %s(host,qdouble,head,padded)",
			        	nz, mat->csr_mat->val->name);
		}
		/* Set all to 0 */
		memset(sval_padded, 0, nz * sizeof(CUKR_SINGLE));
	}

	/*
	 * Alloc memory for qdouble tail,
	 * if necessary
	 */
	CUKR_SINGLE *qvalt_padded = NULL;
	if (pad_qdouble)
	{
		if ( (qvalt_padded = (CUKR_SINGLE*)malloc( nz * sizeof(CUKR_SINGLE))) == NULL) {
				CUKR_FATALX("Error allocating %ld bytes of memory for %s(host,qdouble,tail,padded)",
				             (unsigned long)(nz * sizeof(CUKR_SINGLE)),
				             mat->csr_mat->val->name);
			}
		/* Log */
		LOGX(DEBUG, "Allocated %d doubles for %s (host,qdouble,tail,padded)",
			        nz, mat->csr_mat->val->name);
		/* Set all to 0 */
		memset(qvalt_padded, 0, nz * sizeof(CUKR_SINGLE));
	}


	/*
	 * Move the data, considering the shifts
	 */
	int sh_val = 0;
	for (i_row = 1; i_row < rows + 1; i_row++) {
		/* Copy all until the shift */
		for (i = shift[i_row - 1].pos ; i < shift[i_row].pos; i++) {
				/* Idx always */
				idx_padded[sh_val + i] = idx[i];
				/* If double */
				if (pad_double)
					dval_padded[sh_val + i] = dval[i];
				/* If single or qdouble */
				if (pad_single || pad_qdouble)
					sval_padded[sh_val + i] = sval[i];
				/* If qdouble */
				if (pad_qdouble)
					qvalt_padded[sh_val + i] = qvalt[i];
		}
		// Increment acording to the shift
		sh_val  += shift[i_row].val;
	}

	/* Reindex the ptr to account for the padding ---------------------------- */

	int sum_shift = 0;
	for (i_row = 0; i_row < rows + 1; i_row++) {
		ptr[i_row] += sum_shift;
		sum_shift += shift[i_row + 1].val;
	}



	/* Free the old structures and add the new -------------------------------- */


	/*
	 * Free the old and repalce
	 */

	/* Set the new nz for the matrix */
	mat->nz = nz;

	/* First destroy all the data in
	 * idx and val vectors */
	CukrVecspDestroyData(mat->csr_mat->idx);
	CukrVecspDestroyData(mat->csr_mat->val);

	/* Now create new vectors and set
	 * their status and length*/
	mat->csr_mat->idx->n = nz;
	mat->csr_mat->val->n = nz;
	mat->csr_mat->idx->status = 0;
	mat->csr_mat->val->status = 0;

	/* Replace the old idx */
	mat->csr_mat->idx->ph_ival = idx_padded;
	mat->csr_mat->idx->status |= CUKR_STATE_CPU_INT;
	LOGX(DEBUG, "Freed and replaced %s - (host,int) member ",
	            mat->csr_mat->idx->name);



	/*
	 * Cleanup
	 */

	/* Free shift */
	free(shift);

	/* Free val in double, if padded */
	if (pad_double) {
		mat->csr_mat->val->ph_dval = dval_padded;
		mat->csr_mat->val->status |= CUKR_STATE_CPU_DOUBLE;
		LOGX(DEBUG, "Freed and replaced  %s - (host,double) member ",
	                mat->csr_mat->val->name);
	}

	/* Free val in single, if padded */
	if (pad_single | pad_qdouble) {
		mat->csr_mat->val->ph_sval = sval_padded;
		if (pad_single && pad_qdouble) {
			mat->csr_mat->val->status |= CUKR_STATE_CPU_SINGLE;
			mat->csr_mat->val->status |= CUKR_STATE_CPU_QDOUBLE;
			LOGX(DEBUG, "Freed and replaced  %s - (host,single + qdouble-head) member ",
	                	mat->csr_mat->val->name);
		}
		else if (pad_single) {
			mat->csr_mat->val->status |= CUKR_STATE_CPU_SINGLE;
			LOGX(DEBUG, "Freed and replaced  %s - (host,single) member ",
	                	mat->csr_mat->val->name);
		}
		else {
			mat->csr_mat->val->status |= CUKR_STATE_CPU_SINGLE;
			mat->csr_mat->val->status |= CUKR_STATE_CPU_QDOUBLE;
			LOGX(DEBUG, "Freed and replaced  %s - (host,qdouble head) member ",
	                	mat->csr_mat->val->name);
		}
	}

	/* Free val in qdouble (tail), if padded */
	if (pad_qdouble) {
		mat->csr_mat->val->ph_qvalt = qvalt_padded;
		mat->csr_mat->val->status |= CUKR_STATE_CPU_QDOUBLE;
		LOGX(DEBUG, "Freed and replaced  %s - (host,qdouble tail) member ",
	                mat->csr_mat->val->name);
	}

	LOG(DEBUG, "Finished converting matrix");
}


/**
 * \brief Converts a CSR matrix in generic
 *        format to HYB format
 * \param mat        Pointer to the matrix
 * \param allignment The padding size
 * \param ell_vs_coo Number of times ELL format
 *                   is faster than COO
 * 					 (e.g. 2.0 for twice faster)
 */
void CukrCsrMatspConvertToHyb(CUKR_MATRIX_SP mat, int alignment, int ell_vs_coo)
{
	char buffer[255];
	int i;

	/* If the matrix is already in HYB
	 * format, do nothing */
	if (mat->format ==  CUKR_MATRIX_FORMAT_HYB) {
		LOGX(DEBUG, "Matrix %s already in HYB variation. " \
					"Nothing to be done", mat->name);
		return;
	}

	/* Log */
	LOGX(DEBUG, "Converting matrix %s to HYB format",
	            mat->name);

	/*
	 * Malloc the HYB matrix structure
	 */

	/* Alloc memory */
	if ( ((mat->hyb_mat) = (CUKR_HYB_MATRIX_SP)malloc(sizeof(struct _CUKR_HYB_MATRIX_SP))) == NULL) {
		CUKR_FATAL("Error allocating memory for Hyb matrix object ");
	}

	/* Declare pointers for a more readable code */
	CUKR_INT*    csr_ptr_lo = mat->csr_mat->ptr->ph_ival;
	CUKR_INT*    csr_ptr_hi = &(mat->csr_mat->ptr->ph_ival[1]);
	CUKR_INT*    csr_idx    = mat->csr_mat->idx->ph_ival;
	CUKR_DOUBLE* csr_dval   = mat->csr_mat->val->ph_dval;
	CUKR_SINGLE* csr_sval   = mat->csr_mat->val->ph_sval;
	CUKR_SINGLE* csr_qvalt  = mat->csr_mat->val->ph_qvalt;

	/*
	 * Find the optimal no. of nz_per_row for
	 * the ELL format (NVIDIA way)
	 */

	/* Compute max. no of nz/row */
	int max_nz_row = 0;
	for (i = 0; i < mat->rows; i++) {
		int elems = csr_ptr_hi[i] - csr_ptr_lo[i];
		if (max_nz_row < elems) max_nz_row = elems;
	}

	/* Make nz_per_row histogram */
	int *hist;
	if ( (hist = (int*)(malloc((mat->cols + 1) * sizeof(int)))) == NULL) {
		CUKR_FATALX("Error allocating memory for histogram "\
		            "while converting matrix %s to HYB format",
		            mat->name);
	}
	memset(hist, 0, (mat->cols + 1) * sizeof(int));
	for (i = 0; i < mat->rows; i++) {
		int elems = csr_ptr_hi[i] - csr_ptr_lo[i];
		hist[elems]++;
	}

	/* Choose the best nz_per_col
	 * The idea is that the rows which do not
	 * fit in ELL format will be completly removed
	 * and written in COO format (removed, not trimmed).
	 * Therfore the idea is to match the ratio between the
	 * rows in ELL and the ones in COO with the ratio of
	 * the speed of the formats */
	mat->hyb_mat->ell_nz_per_row = max_nz_row;
	int rows = mat->rows;
	for (i = 0; i < mat->cols + 1; i++) {
		rows -= hist[i];
		if (ell_vs_coo*rows < mat->rows) {
			mat->hyb_mat->ell_nz_per_row = i;
			break;
		}
	}

	free(hist);

	/*
	 * Find the number of nz for both formats
	 */

	/* First get the no. of actual nonzeros (not filling)
	 * for the ELL ans CSR formats */
	mat->hyb_mat->ell_nz = 0;
	mat->hyb_mat->csr_nz = 0;
	for (i = 0; i < mat->rows; i++) {
		int elems = csr_ptr_hi[i] - csr_ptr_lo[i];
		int elems_ell = MIN(mat->hyb_mat->ell_nz_per_row, elems);
		mat->hyb_mat->ell_nz += elems_ell;
		mat->hyb_mat->csr_nz += CLOSEST_MUTLIPLE( (elems - elems_ell), 4);
	}

	/*
	 * Set ELL stride
	 */

	mat->hyb_mat->ell_stride = alignment * ((mat->rows + alignment - 1)/ alignment);


	/*
	 * Create HYB matrix (now that we have the sizes)
	 */

	/* Null vectors */
	mat->hyb_mat->ell_idx    = NULL;
	mat->hyb_mat->ell_val    = NULL;
	mat->hyb_mat->csr_ptr    = NULL;
	mat->hyb_mat->csr_idx    = NULL;
	mat->hyb_mat->csr_val    = NULL;

	/* Create ELL vectors */
	sprintf(buffer, "%s.hyb.ell_idx", mat->name);
	CukrVecspCreate(&mat->hyb_mat->ell_idx, mat->hyb_mat->ell_nz_per_row * mat->hyb_mat->ell_stride, CUKR_TYPE_INT, 	    buffer);
	sprintf(buffer, "%s.hyb.ell_val", mat->name);
	CukrVecspCreate(&mat->hyb_mat->ell_val, mat->hyb_mat->ell_nz_per_row * mat->hyb_mat->ell_stride, mat->type_at_creation, buffer);

	/* Declare pointers for more readable code */
	CUKR_INT*    hyb_ell_idx   = mat->hyb_mat->ell_idx->ph_ival;
	CUKR_DOUBLE* hyb_ell_dval  = mat->hyb_mat->ell_val->ph_dval;
	CUKR_SINGLE* hyb_ell_sval  = mat->hyb_mat->ell_val->ph_sval;
	CUKR_SINGLE* hyb_ell_qvalt = mat->hyb_mat->ell_val->ph_qvalt;


	/* Create CSR vectors*/
	//if (mat->hyb_mat->coo_nz) {
	sprintf(buffer, "%s.hyb.csr_ptr", mat->name);
	CukrVecspCreate(&mat->hyb_mat->csr_ptr, mat->rows + 1,        CUKR_TYPE_INT,         buffer);
	sprintf(buffer, "%s.hyb.csr_idx", mat->name);
	CukrVecspCreate(&mat->hyb_mat->csr_idx, mat->hyb_mat->csr_nz, CUKR_TYPE_INT,         buffer);
	sprintf(buffer, "%s.hyb.csr_val", mat->name);
	CukrVecspCreate(&mat->hyb_mat->csr_val, mat->hyb_mat->csr_nz, mat->type_at_creation, buffer);

	/* Declare pointers for more readable code */
	CUKR_INT*    hyb_csr_ptr_lo = mat->hyb_mat->csr_ptr->ph_ival;
	CUKR_INT*    hyb_csr_ptr_hi = &(mat->hyb_mat->csr_ptr->ph_ival[1]);
	CUKR_INT*    hyb_csr_idx    = mat->hyb_mat->csr_idx->ph_ival;
	CUKR_DOUBLE* hyb_csr_dval   = mat->hyb_mat->csr_val->ph_dval;
	CUKR_SINGLE* hyb_csr_sval   = mat->hyb_mat->csr_val->ph_sval;
	CUKR_SINGLE* hyb_csr_qvalt  = mat->hyb_mat->csr_val->ph_qvalt;

	/*
	 * Fill HYB matrix (now that we have the sizes)
	 */

	/* Set ELL vectors to zero */

	CukrVecspSetFromVal(mat->hyb_mat->ell_idx, 0, CUKR_TYPE_INT);
	CukrVecspSetFromVal(mat->hyb_mat->ell_val, 0, CUKR_DEFAULT);

	/* Fill ELL and CSR */
	int hyb_csr_count = 0, hyb_csr_ptr_val = 1;
	for (i = 0; i < mat->rows; i++) {

		int n      = 0;
		int j      = csr_ptr_lo[i] - 1;
		int row_up = csr_ptr_hi[i] - 1;

		/* Fill ELL with up to ell_nz_per_row elements */
		while (j < row_up && n < mat->hyb_mat->ell_nz_per_row) {
			hyb_ell_idx[mat->hyb_mat->ell_stride*n + i] = csr_idx[j];
			if (mat->type_at_creation == CUKR_TYPE_SINGLE || mat->type_at_creation == CUKR_TYPE_QDOUBLE)
				hyb_ell_sval[mat->hyb_mat->ell_stride*n + i] = csr_sval[j];
			if (mat->type_at_creation == CUKR_TYPE_QDOUBLE)
				hyb_ell_qvalt[mat->hyb_mat->ell_stride*n + i] = csr_qvalt[j];
			if (mat->type_at_creation == CUKR_TYPE_DOUBLE)
				hyb_ell_dval[mat->hyb_mat->ell_stride*n + i] = csr_dval[j];
			j++; n++;
		}

		/* Add the rest to CSR */
		hyb_csr_ptr_lo[i] = hyb_csr_ptr_val;
		int hyb_csr_count_for_cur_row = 0;
		while (j < row_up) {
			    hyb_csr_idx[hyb_csr_count]   = csr_idx[j];
			if (mat->type_at_creation == CUKR_TYPE_SINGLE || mat->type_at_creation == CUKR_TYPE_QDOUBLE)
				hyb_csr_sval[hyb_csr_count]  = csr_sval[j];
			if (mat->type_at_creation == CUKR_TYPE_QDOUBLE)
				hyb_csr_qvalt[hyb_csr_count] = csr_qvalt[j];
			if ( mat->type_at_creation == CUKR_TYPE_DOUBLE)
				hyb_csr_dval[hyb_csr_count]  = csr_dval[j];
			j++; hyb_csr_count_for_cur_row++; hyb_csr_count++;
		}

		/* Fill with zeros to make multiple of four */
		int row_up_multiple_of_4 = CLOSEST_MUTLIPLE(hyb_csr_count_for_cur_row, 4);
		for (j = hyb_csr_count_for_cur_row; j < row_up_multiple_of_4; j++) {
			    hyb_csr_idx[hyb_csr_count]   = 0;
			if (mat->type_at_creation == CUKR_TYPE_SINGLE || mat->type_at_creation == CUKR_TYPE_QDOUBLE)
				hyb_csr_sval[hyb_csr_count]  = 0;
			if (mat->type_at_creation == CUKR_TYPE_QDOUBLE)
				hyb_csr_qvalt[hyb_csr_count] = 0;
			if ( mat->type_at_creation == CUKR_TYPE_DOUBLE)
				hyb_csr_dval[hyb_csr_count]  = 0;
			hyb_csr_count++;
		}

		hyb_csr_ptr_val += row_up_multiple_of_4;
		hyb_csr_ptr_hi[i] = hyb_csr_ptr_val;
	}

	/*
	 * Free the CSR structures
	 */

	CukrVecspDestroyData(mat->csr_mat->ptr);
	CukrVecspDestroyData(mat->csr_mat->idx);
	CukrVecspDestroyData(mat->csr_mat->val);

	/*
	 * Set new matrix format
	 */

	mat->format = CUKR_MATRIX_FORMAT_HYB;
}
