/**
 * @file mat_vec/sp/csr/csr_matrix_sp_gen.c
 * @author Serban Georgescu
 * @since 2008
 *
 * CSR matrix generation routines
 */

#define LOGGER_NAME "CUKr.mat_vec.sp.csr.csr_matrix_sp_gen"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#ifndef __MACH__
	#include <malloc.h>
#endif
#include "csr_matrix_sp_gen.h"
#include "../../matrix_defs.h"
#include "../../../util/log.h" 

#define MAX(a,b) (a > b ? a : b)
#define MIN(a,b) (a < b ? a : b)
#define ROUND(a) ((int)(a + 0.5) == (int)a ? (int)a : (int)(a+1)) 


/************ Forward declarations *******************/

void gen_unique_int_rnd(int *x, int *w, int n, int lo, int up);

/*****************************************************/


// ------------------------------------------------------------------- //

/**
 * \brief Generates a sparse random matrix, filled
 *        with random elements, in CSR format 
 * \param mat Pointer to the matrix 
 * \param n Matrix size
 * \param halfb The half-bandwidth of the matrix
 * \param nz_per_row No. of nonzeros per row
 * \param min,max The lower and upper bound for
 *                matrix values
 * \param name Matrix name
 */
void CukrCsrMatspGenRandom(CUKR_MATRIX_SP* mat, int n, int halfb, int nz_per_row, 
						    CUKR_DOUBLE min, CUKR_DOUBLE max, char* name)
{
	int i, irow, icol;
	
	/* 
	 * Validate params 
	 */
	 
	/* Positive halfb */
	if (halfb < 0) {
		CUKR_FATALX("The half-bandwidh specified for matrix %s "\
					" must be non-negative", name);			
	}
	
	/* Strictly positive nz_per_row */
	if (nz_per_row <= 0) {
		CUKR_FATALX("The nonzeros/row specified for matrix %s "\
					" must be strictly positive", name);			
	}
	 
	/* nz_per_row <=  2*halfb + 1, more elements
	 * won't fit the space */
	if (nz_per_row > 2*halfb + 1) {
		CUKR_FATALX("The of %d half-bandwidh specified for matrix "\
					"%s allows for a maximum of %d nonzeros/row "\
					"(but %d were requested)", halfb, name,  
	 				2*halfb + 1, nz_per_row);		
	}
	 
	LOGX(DEBUG, "Filling CSR matrix %s with random vals. "\
	            "between %g and %g, with a half-bandwidth "\
	            "of %d and %d nonzeros/row", name, 
	            min, max, halfb, nz_per_row);
	
	
	/* Create the matrix */
	//int nz = n + 2*(n*halfb - halfb*(halfb+1)/2) ;
	int nz = n * nz_per_row;
	//CukrMatspCreate(mat, n, n, nz, CUKR_MATRIX_FORMAT_CSR,
	//              CUKR_MATRIX_FORMAT_VARIATION_NONE, CUKR_TYPE_DOUBLE, name);
	
	/* Init random no. generator */
	//srand ( time(NULL) ); 
	srand(123);
	
	/* Make shortcuts for easy reading */
	CUKR_INT* ptr = (*mat)->csr_mat->ptr->ph_ival;
	CUKR_INT* idx = (*mat)->csr_mat->idx->ph_ival;
	CUKR_DOUBLE* dval = (*mat)->csr_mat->val->ph_dval;
	
	/* Generate ptr 
	 * Simple, since we have fixed nonzeros per
	 * row, so there is a fixed increment */
	ptr[0] = 1; 
	for (i = 1; i <= n; i++)
		 ptr[i] = ptr[i - 1] + nz_per_row;
		 
	/* Generate val
	 * Simple, since we have fixed nonzeros per 
	 * row, just generate n*nz random numbers */
	for (i = 0; i < nz; i++) {
		CUKR_DOUBLE rnd = min + ((CUKR_DOUBLE)rand() / (CUKR_DOUBLE)RAND_MAX) * (max - min);
		dval[i] = rnd;
	}
	 
	/* Generate idx
	 * A little more complicated. Have to generate
	 * unique random numbers in [row-halfb,row+halfb] */
	
	/* Malloc for workspace */
	int *w;
	if ( (w = (int*)malloc( n * sizeof(int))) == NULL) {
		CUKR_FATAL("Cannot alloc memory for worspace vector");
	}
	 
	for (irow = 0; irow < n; irow++) {
		
		/* See the bounds for the generator */
		int l = MAX(1, irow - halfb + 1);
		int u = MIN(n, irow + halfb + 1);
		
		/* Generate (unorderd) unique numbers */
		if (l != u)
			gen_unique_int_rnd(&idx[ptr[irow] - 1], w, nz_per_row, l, u);
		else {
			printf("ptr[irow] - 1 = %d\n", ptr[irow] - 1);
			idx[ptr[irow] - 1] = l;
			continue;
		} 
		
		/* Sort row elements */
		for (icol = ptr[irow]; icol < ptr[irow + 1] - 1; icol++)
			 for (i = icol + 1; i < ptr[irow + 1]; i++)
			 	if (idx[i - 1] < idx[icol - 1]) {
			 		int aux = idx[i - 1];
			 		idx[i - 1] = idx[icol - 1];
			 		idx[icol -1] = aux;
			 	} 
	} 
		  	 
	free(w);	  	 
	LOG(DEBUG, "Done generating random CSR matrix"); 	  	 
}

// ------------------------------------------------------------------- //
					 
/**
 * \brief Generates a sparse random matrix, filled
 *        with random elements 
 * \param mat Pointer to the matrix 
 * \param ord Order
 * \param name Matrix name
 */
void CukrCsrMatspGenPoisson2D(CUKR_MATRIX_SP* mat, int ord, char* name)
{
	int row;
	
	char buffer[255];
	
	LOGX(DEBUG, "Creating 2D Poisson matrix %s of " \
				"order %d", name, ord);
	
	/* Compute size, the no. of nonzeros 
	 * and create matrix */
	int n = ord * ord;
	int nz = ord*ord + 4*(ord-1)*ord;
	
	
	/* ----------------------------------------------------------------------------*/
	/* Yusa changed after CukrBcsrMatspGenPoisson2D() */
	/* Create Matrix */
	CukrMatspCreate(mat, n, n, CUKR_MATRIX_FORMAT_CSR, CUKR_TYPE_DOUBLE, name);
	
	/* set rows, cols and nz */
	(*mat)->rows = n;
	(*mat)->cols = n;
	(*mat)->nz = nz;
	
	/* ptr */
	sprintf(buffer, "%s.ptr", (*mat)->name);
	CukrVecspCreate(&(*mat)->csr_mat->ptr, n+1, CUKR_TYPE_INT,    buffer);
	/* idx */
	sprintf(buffer, "%s.idx", (*mat)->name);
	CukrVecspCreate(&(*mat)->csr_mat->idx, nz,  CUKR_TYPE_INT,    buffer);
	/* val */
	sprintf(buffer, "%s.val", (*mat)->name);
	CukrVecspCreate(&(*mat)->csr_mat->val, nz,  CUKR_TYPE_DOUBLE, buffer);
	
	/* ----------------------------------------------------------------------------*/
	
	
	/* Make shortcuts for easy reading */
	CUKR_INT* ptr = (*mat)->csr_mat->ptr->ph_ival;
	CUKR_INT* idx = (*mat)->csr_mat->idx->ph_ival;
	CUKR_DOUBLE* dval = (*mat)->csr_mat->val->ph_dval;
	
	/*
	 * Build the CSR matrix  
	 */
	 
	int pos = 0;
	ptr[0] = 1;
	for (row = 1; row <= n; row++){
		
		int count = 0;                   
		/* The left-most band */
		if (row > ord) {
			idx[pos] = row - ord;			
			dval[pos] = -1;  
			pos++;
			count++; 
		}
		 
		/* The band to the left of the diagonal */
		if (row > 1 && (row - 1) % ord) {					
			idx[pos] = row - 1;
			dval[pos] = -1;  
			pos++;
			count++; 
		}
		
		/* The diagonal */
		idx[pos] = row;		
		dval[pos] = 4;  
		pos++;
		count++; 
		
		/* The band to the right of the diagonal */
		if (row < n && row % ord) {
			
			idx[pos] = row + 1;
			dval[pos] = -1;  
			pos++;
			count++; 
		}
		
		/* The right-most band */
		if (row <= n - ord) {
			idx[pos] = row + ord;
			dval[pos] = -1;  
			pos++;
			count++; 
		}
		
		ptr[row] = ptr[row - 1] + count;
	}   
	
	LOG(DEBUG, "Done generating 2D Poisson matrix"); 	             
	              			
}


/**
 * \brief Generates a sparse random matrix, filled
 *        with random elements 
 * \param mat Pointer to the matrix 
 * \param ord Order
 * \param name Matrix name
 */
void CukrCsrMatspGenPoisson2Dsingle(CUKR_MATRIX_SP* mat, int ord, char* name)
{
	int row;
	
	char buffer[255];
	
	LOGX(DEBUG, "Creating 2D Poisson matrix %s of " \
				"order %d", name, ord);
	
	/* Compute size, the no. of nonzeros 
	 * and create matrix */
	int n = ord * ord;
	int nz = ord*ord + 4*(ord-1)*ord;
	
	
	/* ----------------------------------------------------------------------------*/
	/* Yusa changed after CukrBcsrMatspGenPoisson2D() */
	/* Create Matrix */
	CukrMatspCreate(mat, n, n, CUKR_MATRIX_FORMAT_CSR, CUKR_TYPE_SINGLE, name);
	
	/* set rows, cols and nz */
	(*mat)->rows = n;
	(*mat)->cols = n;
	(*mat)->nz = nz;
	
	/* ptr */
	sprintf(buffer, "%s.ptr", (*mat)->name);
	CukrVecspCreate(&(*mat)->csr_mat->ptr, n+1, CUKR_TYPE_INT,    buffer);
	/* idx */
	sprintf(buffer, "%s.idx", (*mat)->name);
	CukrVecspCreate(&(*mat)->csr_mat->idx, nz,  CUKR_TYPE_INT,    buffer);
	/* val */
	sprintf(buffer, "%s.val", (*mat)->name);
	CukrVecspCreate(&(*mat)->csr_mat->val, nz,  CUKR_TYPE_SINGLE, buffer);
	
	/* ----------------------------------------------------------------------------*/
	
	
	/* Make shortcuts for easy reading */
	CUKR_INT* ptr = (*mat)->csr_mat->ptr->ph_ival;
	CUKR_INT* idx = (*mat)->csr_mat->idx->ph_ival;
	CUKR_SINGLE* dval = (*mat)->csr_mat->val->ph_sval;
	
	/*
	 * Build the CSR matrix  
	 */
	 
	int pos = 0;
	ptr[0] = 1;
	for (row = 1; row <= n; row++){
		
		int count = 0;                   
		/* The left-most band */
		if (row > ord) {
			idx[pos] = row - ord;			
			dval[pos] = -1;  
			pos++;
			count++; 
		}
		 
		/* The band to the left of the diagonal */
		if (row > 1 && (row - 1) % ord) {					
			idx[pos] = row - 1;
			dval[pos] = -1;  
			pos++;
			count++; 
		}
		
		/* The diagonal */
		idx[pos] = row;		
		dval[pos] = 4;  
		pos++;
		count++; 
		
		/* The band to the right of the diagonal */
		if (row < n && row % ord) {
			
			idx[pos] = row + 1;
			dval[pos] = -1;  
			pos++;
			count++; 
		}
		
		/* The right-most band */
		if (row <= n - ord) {
			idx[pos] = row + ord;
			dval[pos] = -1;  
			pos++;
			count++; 
		}
		
		ptr[row] = ptr[row - 1] + count;
	}   
	
	LOG(DEBUG, "Done generating 2D Poisson matrix"); 	             
	              			
}


// ------------------------------------------------------------------- //

/**
 * \brief Generates a diagonally-dominated banded
 *        matrix, filled with ones but the main 
 * 		  diagonal computed as such as to be dominant.
 * 		  The diagonals can be shifted to the left or 
 *        right by specifying a negative or positive shift 
 * \param mat Pointer to the matrix 
 * \param format Matrix format (CSR)
 * \param rows*cols Matrix order
 * \param hb Halfbandwidth 
 * \param shift The shift  
 * \param name Matrix name
 */
void CukrCsrMatspGenBandedWShift(CUKR_MATRIX_SP* mat, int rows, int cols, int hb, int shift, char* name)
{
	int r, b;
	
	LOGX(DEBUG, "Creating banded matrix %s of " \
				"order (%d,%d) and halfb %d and shift %d", 
				name, rows, cols, hb, shift);
	
	/*  
	 * Compute the no. of nonzeros 
	 * and create matrix 
	 */
	 
	/* Compute the no. of nonzeros */ 
	int nz = 0;
	for (r = 0; r < rows; r++) 
		nz += MIN(hb, r + shift) + MIN(hb, cols - (r + 1 + shift)) + 1;

		
	/* Create matrix */
	//CukrMatspCreate(mat, rows, cols, nz, CUKR_MATRIX_FORMAT_CSR,
	//              CUKR_MATRIX_FORMAT_VARIATION_NONE, CUKR_TYPE_DOUBLE, name);
	
	/* Make shortcuts for easy reading */
	CUKR_INT* ptr = (*mat)->csr_mat->ptr->ph_ival;
	CUKR_INT* idx = (*mat)->csr_mat->idx->ph_ival;
	CUKR_DOUBLE* dval = (*mat)->csr_mat->val->ph_dval;
	
	/*
	 * Set the values
	 */
	 
	 /* Set the val. for the bands to -1 and
	  * the one for the diagonal to dominate */
	 CUKR_DOUBLE band_val = -1;
	 CUKR_DOUBLE diag_val = 2 * hb + 1;
	 
	 /*
	  * Build the CSR matrix
	  */
	  
	  int idx_pos = 0, val_pos = 0;
	  ptr[0] = 1;
	  for (r = 0; r < rows; r++)  
	  {
	  	 int cnt = 0;
	  	 for (b = -hb + shift; b <= hb + shift; b++) 	  	 
	  	 	if (r + b >= 0 && r + b < cols) {	  	 		
	  	 		/* Add to idx */
	  	 		idx[idx_pos++] = r + b + 1;
	  	 		
	  	 		/* Add to val */ 
	  	 		dval[val_pos++] = (b == shift) ? diag_val : band_val;
	  	 		
	  	 		/* Increase counter */
	  	 		cnt++;  
	  	 	} 
	  	 
	  	 /* Set ptr */
	  	 ptr[r + 1] = ptr[r] + cnt;		  	 	
	  }
	  
	  LOGX(DEBUG, "Done creating banded matrix %s of", name);
}




//============================ PRIVATE FUNCTIONS =========================== //




/**
 * \brief Generates n unique random integers,
 *        between lo and lo + n
 * \param x The vector where to put the elements
 * \param w Workspace vector (prev. allocated)
 * \param n The no. of requested numbers
 * \param lo The smallest 
 * \param up The largest 
 */
void gen_unique_int_rnd(int *x, int *w, int n, int lo, int up)
{
	int i;
	
	/* Fill the workspace */
	int nn = up - lo + 1;
	for (i = 0; i <= nn; i++)
		w[i] = i;
		
	/* Extract one by one */
	for (i = 0; i < n; i++) {
		int irnd = rand() % nn + 1;
		x[i] = w[irnd] + lo - 1;
		w[irnd] = w[nn--];
		
		/* Has to be smaller than <up> */
		while (x[i] > up) {
			irnd = rand() % nn + 1;
			x[i] = w[irnd] + lo - 1;
			w[irnd] = w[nn--];
		}
	}	
}
