/**
 * @file mat_vec/sp/csr/csr_matrix_sp_io.c
 * @author Serban Georgescu
 * @since 2008
 *
 * File IO functionality for matrices
 * in CSR format
 */

#define LOGGER_NAME "CUKr.mat_vec.sp.csr.csr_matrix_sp_io"

#include <stdio.h>
#include <stdlib.h>
#include "../../../util/log.h"
#include "../../matrix_defs.h"

#include "csr_matrix_sp.h"

/**
 * \brief Loads a CSR matrix from a CSR file.
 *        The CSR file format is the following:
 *        1st_row: m n nz
 *        2nd->(n+1)th_row: ptr
 *        (n+1)th -> (n+nz+1)th_row: idx
 *        (n+nz+1)th -> (n+2*nz+1)th_row: val
 * \param mat The matrix in which data is being read
 * \param filename The file where the data is
 *        being read from
 */
void CukrCsrMatspCreateFromFile(CUKR_MATRIX_SP mat, const char *filename)
{
	FILE *f;
	int i, rows, cols, nz, res;
	char buffer[255];
	
	LOGX(DEBUG, "Reading CSR matrix from file: %s", filename);

	
	/*
	 * Malloc for the csr member structure
	 */
	 
	 /* Alloc memory */
	 if ( ((mat->csr_mat) = (CUKR_CSR_MATRIX_SP)malloc(sizeof(struct _CUKR_CSR_MATRIX_SP))) == NULL) {
	 	CUKR_FATAL("Error allocating memory for csr matrix object ");
	 } 
			
	 LOGX(DEBUG, "Read matrix %s as CSR", mat->name);

	 
	/* Open the file and read the
	 * matrix sizes */
	
	/* Open the file for reading */
	if ( (f = fopen(filename, "r")) == NULL) {
		CUKR_FATALX("Unable to open file %s for read",
		            filename);;
	}    
	
	/* Read the dimensions */
	res = fscanf(f, "%d %d %d", &mat->rows, &mat->cols, &mat->nz);

	/*
	 * Create data mambers
	 */
	 
	/* ptr */ 
	sprintf(buffer, "%s.ptr", mat->name);
	CukrVecspCreate(&mat->csr_mat->ptr, mat->rows + 1, CUKR_TYPE_INT, 		buffer);
	/* idx */
	sprintf(buffer, "%s.idx",  mat->name);
	CukrVecspCreate(&mat->csr_mat->idx, mat->nz,       CUKR_TYPE_INT, 		buffer);
	/* val */
	sprintf(buffer, "%s.val",  mat->name);
	CukrVecspCreate(&mat->csr_mat->val, mat->nz,       mat->type_at_creation, buffer); 
	
	              
	/*
	 * Read data in the members
	 */              
	   
	/* Read ptr */
	CukrVecspReadDataFromFile(mat->csr_mat->ptr, f, CUKR_TYPE_INT); 
	/* Read idx */
	CukrVecspReadDataFromFile(mat->csr_mat->idx, f, CUKR_TYPE_INT); 
	/* Read val */
	CukrVecspReadDataFromFile(mat->csr_mat->val, f, mat->type_at_creation); 
	
	/*
	 * Close 
	 */
	
	fclose(f);
	LOGX(INFO, "Finished reading matrix %s(%d,%d,%d) "\
	           "from file %s", mat->name, mat->rows,  
	           mat->cols, mat->nz, filename);  	
}

// ------------------------------------------------------------------- //

/**
 * \brief Open the file storing a CSR matrix
 *        and return the matrix dimensions 
 * 		 (rows,cols) and the no. nonzeros
 * \param filename The full path to the file 
 *        storing the matrix
 * \param format The matrix format ID
 * \param rows Pointer to a variable where to 
 *        store the no. of rows
 * \param cols Pointer to a variable where to 
 *        store the no. of cols
 * \param nz Pointer to a variable where to 
 *        store the no. of nz
 */
void CukrCsrMatspReadSizesFromFile(char* filename, int* rows, int* cols, int* nz)
{
	FILE* f;
	int res;
	
	LOGX(DEBUG, "Reading matrix dimensions from file %s",
	             filename);
	             
	/* Open the file for reading */
	if ( (f = fopen(filename, "r")) == NULL) {
		CUKR_FATALX("Unable to open file %s for read",
		            filename);
	}    
	
	/* Read the dimensions */
	res = fscanf(f, "%d %d %d", rows, cols, nz);
	
	/* Done */
	fclose(f);
	LOGX(DEBUG, "Successfully read rows = %d, cols = %d, "\
	            "nz = %d", *rows, *cols, *nz); 
}

/**
 * \brief Save a SP matrix to file, in readable .csr format
 */
void CukrCsrMatspSaveToFile(CUKR_MATRIX_SP mat, const char *filename)
{
	int i,j;
	FILE *f;

	/* Error if variation other then none */
	if (mat->variation != CUKR_MATRIX_FORMAT_VARIATION_NONE)
		CUKR_FATAL("Can only write matrices with no format variation");

	/* Open the file for writing */
	if ( (f = fopen(filename, "w")) == NULL) {
		CUKR_FATALX("Unable to open file %s for read",
					filename);
	}

	/* Write sizes */
	fprintf(f, "%d %d %d\n", mat->rows, mat->cols, mat->nz);

	/* Write ptr and idx*/
	for (i = 0; i < mat->rows + 1; i++)
		fprintf(f, "%d\n", mat->csr_mat->ptr->ph_ival[i]);
	for (i = 0; i < mat->nz; i++)
		fprintf(f, "%d\n", mat->csr_mat->idx->ph_ival[i]);

	/* Write val */
	CukrVecspTransfer(mat->csr_mat->val, CUKR_LOCATION_CPU, CUKR_TYPE_DOUBLE);
	for (i = 0; i < mat->nz; i++)
		fprintf(f, "%g\n", mat->csr_mat->val->ph_dval[i]);

	fclose(f);

}
