/**
 * @file mat_vec/sp/cuda/vector_kernels.cu
 * @author Serban Georgescu
 * @since 2008
 * 
 * Acts as a bridge between the C++ CUDA
 * implementation and the rest of the C 
 * files. All the CUDA copy kernels are
 * included here.
 * 
 */

#include <stdio.h>
#include <cuda.h>
#include <cutil.h>
#include <cublas.h>
#ifndef _WIN32
	#include <unistd.h>
#endif	

#include "kernels_config.h"
#include "kernel_dbl_to_qdbl.h"
#include "kernel_dbl_to_sng.h"
#include "kernel_qdbl_to_dbl.h"
#include "kernel_sng_to_dbl.h"
#include "kernel_set_dbl_to_val.h"
#include "kernel_set_sng_to_val.h"

#define CUKR_TRUE  1
#define CUKR_FALSE 0



/**
 * \brief Dbl->Qdbl kernel on the GPU
 * \param n Vector size
 * \param d_x Device pointer to the source
 *            vector
 * \param d_yh, d_yt Device pointers to the
 *            dest vectors (head and tail)
 */
#ifdef __cplusplus
extern "C"
#endif 
void cuda_kernel_copy_dbl_to_qdbl(int n, double *d_x, float *d_yh, float *d_yt)
{	
	/* Preparations */
	int dynMem = 1 * CUKR_COPY_DBL_QDBL_THREADS * sizeof(double); 
	
	/* Kernel invocation */
	kernel_dbl_to_qdbl<<<CUKR_COPY_DBL_QDBL_CTAS, CUKR_COPY_DBL_QDBL_THREADS, dynMem>>> (d_x, d_yh, d_yt, n);  
	CUT_CHECK_ERROR("kernel_dbl_to_qdbl() execution failed\n"); 
	
	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize()); 	
}

// ------------------------------------------------------------------- //

/**
 * \brief Dbl->Sng kernel on the GPU
 * \param n Vector size
 * \param d_x Device pointer to the source
 *            vector
 * \param d_y Device pointer to the dest.
 *            vector
 */
#ifdef __cplusplus
extern "C"
#endif 
void cuda_kernel_copy_dbl_to_sng(int n, double *d_x, float *d_y)
{	
	/* Preparations */
	int dynMem = 1 * CUKR_COPY_DBL_SNG_THREADS * sizeof(double); 
	
	/* Kernel invocation */
	kernel_dbl_to_sng<<<CUKR_COPY_DBL_SNG_CTAS, CUKR_COPY_DBL_SNG_THREADS, dynMem>>> (d_x, d_y, n);  
	CUT_CHECK_ERROR("kernel_dbl_to_sng() execution failed\n"); 
	
	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize()); 	
}

// ------------------------------------------------------------------- //

/**
 * \brief QDbl->Dbl kernel on the GPU
 * \param n Vector size
 * \param d_xh, d_xt Device pointers to the source
 *            vectors (head, tail)
 * \param d_y Device pointer to the dest.
 *            vector
 */
#ifdef __cplusplus
extern "C"
#endif 
void cuda_kernel_copy_qdbl_to_dbl(int n, float *d_xh, float *d_xt, double *d_y)
{	
	/* Preparations */
	int dynMem = 2 * CUKR_COPY_QDBL_DBL_THREADS * sizeof(float); 
	
	/* Kernel invocation */
	kernel_qdbl_to_dbl<<<CUKR_COPY_QDBL_DBL_CTAS, CUKR_COPY_QDBL_DBL_THREADS, dynMem>>> (d_xh, d_xt, d_y, n);  
	CUT_CHECK_ERROR("kernel_qdbl_to_dbl() execution failed\n"); 
	
	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize()); 	
}

// ------------------------------------------------------------------- //

/**
 * \brief Sng->Dbl kernel on the GPU
 * \param n Vector size
 * \param d_x Device pointer to the source
 *            vector
 * \param d_y Device pointer to the dest.
 *            vector
 */
#ifdef __cplusplus
extern "C"
#endif 
void cuda_kernel_copy_sng_to_dbl(int n, float *d_x, double *d_y)
{	
	/* Preparations */
	int dynMem = 1 * CUKR_COPY_SNG_DBL_THREADS * sizeof(float); 
	
	/* Kernel invocation */
	kernel_sng_to_dbl<<<CUKR_COPY_SNG_DBL_CTAS, CUKR_COPY_SNG_DBL_THREADS, dynMem>>> (d_x, d_y, n);  
	CUT_CHECK_ERROR("kernel_sng_to_dbl() execution failed\n"); 
	
	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize()); 	
}

// ------------------------------------------------------------------- //

/**
 * \brief Dbl.Vec<-dbl kernel on the GPU
 * \param n Vector size
 * \param d_x Device pointer to the vector 
 * \param val Value to set the vector to
 */
#ifdef __cplusplus
extern "C"
#endif 
void cuda_kernel_set_dbl_to_val(int n, double *d_x, double val)
{	
	/* Kernel invocation */
	kernel_set_dbl_to_val<<<CUKR_SET_DBL_VAL_CTAS, CUKR_SET_DBL_VAL_THREADS>>> (d_x, val, n);  
	CUT_CHECK_ERROR("cuda_kernel_set_dbl_to_val() execution failed\n"); 
	
	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize()); 	
}

// ------------------------------------------------------------------- //

/**
 * \brief Sng.Vec <-val kernel on the GPU
 * \param n Vector size
 * \param d_x Device pointer to the vector 
 * \param val Value to set the vector to
 */
#ifdef __cplusplus
extern "C"
#endif 
void cuda_kernel_set_sng_to_val(int n, float *d_x, float val)
{	
	/* Kernel invocation */
	kernel_set_sng_to_val<<<CUKR_SET_SNG_VAL_CTAS, CUKR_SET_SNG_VAL_THREADS, 0>>> (d_x, val, n);  
	CUT_CHECK_ERROR("cuda_kernel_set_sng_to_val() execution failed\n"); 
	
	/* Do a syncronize, for timing */
    CUDA_SAFE_CALL(cudaThreadSynchronize()); 	
}
