/**
 * @file mat_vec/sp/hyb/hyb_matrix_sp.c
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains all the functions needed
 * to support the ELL-COO hybrid matrix format
 */


#define LOGGER_NAME "CUKr.mat_vec.sp.hyb.hyb_matrix"

#include <stdio.h>
#include <stdlib.h>
#include "hyb_matrix_sp.h"
#include "../matrix_sp.h"
#include "../../../util/log.h"
#include "../../../util/utils.h"
#include "../../../blas/blas.h"

#define MAX(a,b) (a > b ? a : b)
#define MIN(a,b) (a < b ? a : b)



				/*****************************************/
				//       CONSTRUCTOR / DESTRUCTOR        //
				/*****************************************/         



/**
 * \brief Creates a HYB matrix
 *        with data of the specified type
 * \param mat The matrix structure
 *        whose HYB part is being created
 */
void CukrHybMatspCreate(CUKR_MATRIX_SP mat)
{
	char buffer[255];


	 /*
	 * Malloc for the matrix structure
	 */

	 /* Alloc memory */
	 if ( ((mat->hyb_mat) = (CUKR_HYB_MATRIX_SP)malloc(sizeof(struct _CUKR_HYB_MATRIX_SP))) == NULL) {
	 	CUKR_FATAL("Error allocating memory for hyb matrix object ");
	 }

	LOGX(DEBUG, "Creating matrix %s(%d,%d,%d) as HYB",
	             mat->name, mat->rows, mat->cols, mat->nz);


	/*
	 * Null vectors
	 */

	mat->hyb_mat->ell_idx = NULL;
	mat->hyb_mat->ell_val = NULL;
	mat->hyb_mat->csr_ptr = NULL;
	mat->hyb_mat->csr_idx = NULL;
	mat->hyb_mat->csr_val = NULL;
}

// ------------------------------------------------------------------- //

/**
 * \brief Finalizes the creation of the matrix,
 *        after all elements have been added.
 * \param mat The matrix structure to assemble.
 */
void CukrHybMatspAssemble(CUKR_MATRIX_SP mat)
{
	char buffer[255];

	CUKR_FATAL("CukrHybMatspAssemble has not been implemented yet");
}

// ------------------------------------------------------------------- //

/**
 * \brief Free memory all memory used
 *        by the HYB part of the matrix
 * \param mat The matrix 
 */
void CukrHybMatspDestroy(CUKR_MATRIX_SP mat)
{
	/* Free all members */
	CukrVecspDestroy(&mat->hyb_mat->ell_idx);
	CukrVecspDestroy(&mat->hyb_mat->ell_val);
	CukrVecspDestroy(&mat->hyb_mat->csr_ptr);
	CukrVecspDestroy(&mat->hyb_mat->csr_idx);
	CukrVecspDestroy(&mat->hyb_mat->csr_val);
	
	/* Finally, destroy the CSR matrix structure */
	free(mat->hyb_mat);	
} 






				/*****************************************/
				//          TRANSFER / CONVERT           // 
				/*****************************************/ 

  
  
  
  
  

/**
 * \brief Changes the location of a Hyb matrix
 *        by changing the location of the 
 *        data vectors
 * \param mat The matrix
 * \param dest_loc Destination matrix location (CPU,GPU)
 * \param dest_type Destination matrix value type 
 *        (float, qdouble, double)
 */
void CukrHybMatmpTransfer(CUKR_MATRIX_SP mat, int dest_loc, int dest_type)
{
	/*
	 * Validate the parameters 
	 */
	 
	 /* The only allowed destination
	  * types are float ones */
	if (dest_type != CUKR_TYPE_DOUBLE  &&
		dest_type != CUKR_TYPE_QDOUBLE &&
	    dest_type != CUKR_TYPE_SINGLE  ) 
	{
	    CUKR_FATALX(" Trying to convert to integer type " \
	                " floating point data from matrix %s", 
	                mat->name);
	}
	
	/*
	 * Do the transfers (if necessary)
	 */
	
	/* ell_idx */
	CukrVecspTransfer(mat->hyb_mat->ell_idx, dest_loc, CUKR_TYPE_INT);
	/* ell_val */
	CukrVecspTransfer(mat->hyb_mat->ell_val, dest_loc, dest_type);  
	
	if (mat->hyb_mat->csr_nz) {	
		/* csr_ptr */
		CukrVecspTransfer(mat->hyb_mat->csr_ptr, dest_loc, CUKR_TYPE_INT);
		/* csr_idx */
		CukrVecspTransfer(mat->hyb_mat->csr_idx, dest_loc, CUKR_TYPE_INT);
		/* csr_val */
		CukrVecspTransfer(mat->hyb_mat->csr_val, dest_loc, dest_type);
	}  
}

// ------------------------------------------------------------------- //

/**
 * \brief Changes the storage format of 
 *        the matrix to a diferent variation
 *        or to a completely different format
 * \param mat Pointer to the matrix 
 * \param dest_format Destination format
 * \param dest_variation Destination variation 
 */
void CukrHybMatspConvert(CUKR_MATRIX_SP mat, int dest_format, int dest_variation)
{
	CUKR_FATAL("No conversion currently supported from "\
	           "Hyb format"); 	
}
               


				/*****************************************/
				//               GET/SET                 // 
				/*****************************************/ 






/**
 * \brief Extracts the diagonal from a matrix
 * \param mat The matrix 
 * \param vec The vector where to store the diagonal
 * \param diag The diagonal to read (zero for the
 *        main diagonal)
 * \param invert Whether to invert it or not 
 */
void CukrHybMatspGetDiag(CUKR_MATRIX_SP mat, CUKR_VECTOR_SP vec, int diag, int invert)
{
	//CUKR_FATAL("CukrHybMatspGetDiag not yet implemented");

	 LOGX(DEBUG, "Extracting diagonal from HYB matrix %s",
		          mat->name);

	/* Make sure the matrix has double
	 * precision on host values  */
	 CukrHybMatmpTransfer(mat, CUKR_LOCATION_CPU, CUKR_TYPE_DOUBLE);

	/* Make shortcut pointers for more
	 * readable code */
	 CUKR_INT    n        		= mat->rows;
	 CUKR_INT    ell_nz_per_row = mat->hyb_mat->ell_nz_per_row;
	 CUKR_INT    ell_stride     = mat->hyb_mat->ell_stride;
	 CUKR_INT    *ell_idx 		= mat->hyb_mat->ell_idx->ph_ival;
	 CUKR_DOUBLE *ell_val 		= mat->hyb_mat->ell_val->ph_dval;
	 CUKR_INT    *csr_ptr 		= mat->hyb_mat->csr_ptr->ph_ival;
	 CUKR_INT    *csr_idx 		= mat->hyb_mat->csr_idx->ph_ival;
	 CUKR_DOUBLE *csr_val 		= mat->hyb_mat->csr_val->ph_dval;
	 CUKR_DOUBLE *ddiag   		= vec->ph_dval;

	/*
	 * Put the diagonal elements in the
	 * diagonal vector
	 */
	int i_col, i_row, found;

	for (i_row = -diag; i_row < n; i_row++)
		if (i_row >= 0){
			found = 0;
			// Search in the ell part
			for (i_col = 0; i_col < ell_nz_per_row; i_col++) {
				int i_strided_col = i_row + i_col*ell_stride;
				if (diag + i_row == ell_idx[i_strided_col] - 1) {
					if (!invert)
						ddiag[i_row] = ell_val[i_strided_col];
					else
						ddiag[i_row] = 1.0/ell_val[i_strided_col];
					found = 1;
					break;
				}
			}
			if (found) continue;
			// Search in the csr part
			for (i_col = csr_ptr[i_row] - 1; i_col < csr_ptr[i_row + 1] - 1; i_col++)
				if (diag + i_row == csr_idx[i_col] - 1) {
					if (!invert)
						ddiag[i_row] = csr_val[i_col];
					else
						ddiag[i_row] = 1.0/csr_val[i_col];
					found = 1;
					break;
			}
			if (found) continue;

			// Error if nothing found
			CUKR_FATALX("No diagonal element stored for matrix %s row %d",
						mat->name, i_row);
		}
}







				/*****************************************/
				//                 I/O                   // 
				/*****************************************/ 






/**
 * \brief Prints a matrix to a file stream
 * \param mat The matrix being printed
 * \param f An open file stream to write to
 * \param indent An indentation string, to be inserted
 *        before every print.
 */
void CukrHybMatspPrintToFileWithIndent(CUKR_MATRIX_SP mat, FILE *f, char* indent)
{
	char fmt_str[255], var_str[255];
	int i; 
	
	/* 
	 * Print matrix info
	 */
	
	/* Get strings for location and precision*/
	CukrMatFormatIdToStr(mat->format, fmt_str);
	CukrMatFormatVariationIdToStr(mat->variation, var_str);
	
	/* Print the info */
	fprintf(f, "%sMatrix: %s (%dx%d with %d nonzeros) " \
	       "with format %s (variation %s)\n", indent, mat->name, 
	       mat->rows, mat->cols, mat->nz,
	       fmt_str, var_str);
	
	/* Merge indents */
	char mindent[255];
	sprintf(mindent, "%s    ", indent);
	
	/*
	 * Print data members 
	 */
	        
	/* ell_idx */
	fprintf(f, "%sELL_IDX\n", indent);
	CukrVecspPrintToFileWithIndent(mat->hyb_mat->ell_idx, f, mindent);
	/* ell_val */
	fprintf(f, "%sELL_VAL\n", indent);
	CukrVecspPrintToFileWithIndent(mat->hyb_mat->ell_val, f, mindent);
	
	if (mat->hyb_mat->csr_nz) {
		/* csr_ptr */
		fprintf(f, "%sCSR_PTR\n", indent);
		CukrVecspPrintToFileWithIndent(mat->hyb_mat->csr_ptr, f, mindent);
		/* csr_idx */
		fprintf(f, "%sCSR_IDX\n", indent);
		CukrVecspPrintToFileWithIndent(mat->hyb_mat->csr_idx, f, mindent);
		/* coo_val */ 
		fprintf(f, "%sCSR_VAL\n", indent);
		CukrVecspPrintToFileWithIndent(mat->hyb_mat->csr_val, f, mindent);
	}
}

