/**
 * @file mat_vec/sp/hyb/hyb_matrix_sp.h
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains all the functions needed
 * to support the ELL-COO hybrid matrix format
 */

#ifndef CUKR_HYB_MATRIX_SP_H_
#define CUKR_HYB_MATRIX_SP_H_

/* =============== FORWARD DECLARATIONS ============== */

typedef struct _CUKR_HYB_MATRIX_SP *CUKR_HYB_MATRIX_SP;

/* =================================================== */

#include "../matrix_sp.h"
#include "../vector_sp.h"

/**
 * Holds data for a matrix of type Hybrid
 */
struct _CUKR_HYB_MATRIX_SP{   
	
	/**
	 * For the ELL part 
	 */
	
	/* Matrix dimensions */
	CUKR_INT ell_nz; 	
	
	/* Number of nonzeros per row */
	CUKR_INT ell_nz_per_row;
	
	/* Read stride */
	CUKR_INT ell_stride;
	
	/* Column vector (ell_nz_per_row * ell_rows length) */
	CUKR_VECTOR_SP ell_idx;
	
	/* Value vector (ell_nz_per_row * ell_rows length) */
	CUKR_VECTOR_SP ell_val;
	
	/*
	 * For the CSR part
	 */
	
	/* Matrix dimensions */
	CUKR_INT csr_nz, csr_start_row, csr_rows;
	
	/* Row and column vectors */
	CUKR_VECTOR_SP csr_ptr, csr_idx;
	
	/* Value vector */
	CUKR_VECTOR_SP csr_val; 
	 
};


		




				/*****************************************/
				//       CONSTRUCTOR / DESTRUCTOR        //
				/*****************************************/         




/**
 * \brief Free memory all memory used
 *        by the CSR part of the matrix
 * \param mat The matrix 
 */
void CukrHybMatspDestroy(CUKR_MATRIX_SP mat); 






				/*****************************************/
				//          TRANSFER / CONVERT           // 
				/*****************************************/ 

  
  
  
  
  

/**
 * \brief Changes the location of a Hyb matrix
 *        by changing the location of the 
 *        data vectors
 * \param mat The matrix
 * \param dest_loc Destination matrix location (CPU,GPU)
 * \param dest_type Destination matrix value type 
 *        (float, double)
 */
void CukrHybMatmpTransfer(CUKR_MATRIX_SP mat, int dest_loc, int dest_type);

// ------------------------------------------------------------------- //

/**
 * \brief Changes the storage format of 
 *        the matrix to a diferent variation
 *        or to a completely different format
 * \param mat Pointer to the matrix 
 * \param dest_format Destination format
 * \param dest_variation Destination variation 
 */
void CukrHybMatspConvert(CUKR_MATRIX_SP mat, int dest_format, int dest_variation);

                




				/*****************************************/
				//               GET/SET                 // 
				/*****************************************/ 






/**
 * \brief Extracts the diagonal from a matrix
 * \param mat The matrix 
 * \param vec The vector where to store the diagonal
 * \param diag The diagonal to read (zero for the
 *        main diagonal)
 * \param invert Whether to invert it or not 
 */
void CukrHybMatspGetDiag(CUKR_MATRIX_SP mat, CUKR_VECTOR_SP vec, int diag, int invert);







				/*****************************************/
				//                 I/O                   // 
				/*****************************************/ 






/**
 * \brief Prints a matrix to a file stream
 * \param mat The matrix being printed
 * \param f An open file stream to write to
 * \param indent An indentation string, to be inserted
 *        before every print.
 */
void CukrHybMatspPrintToFileWithIndent(CUKR_MATRIX_SP mat, FILE *f, char* indent);



#endif /*CUKR_HYB_MATRIX_SP_H_*/
