/**
 * @file mat_vec/sp/matrix_sp.c
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines the CUKR_MATRIX_SP abstract
 * interace.
 * 
 * The CUKR_MATRIX_SP structure stores a 
 * matrix that can be in any (useful) type
 * and reside either on the CPU or GPU.
 * The structure stores, besides the data,
 * both type and location.
 *  
 * The BLAS2 operations are brought one
 * level up. They take paramters only the
 * matrices and vectors; the type, size and 
 * location are automatically read and the 
 * appropriate BLAS operation is chosen.
 * 
 * If matrices resides in a different location
 * than the BLAS, data is automatically moved. 
 */

#define LOGGER_NAME "CUKr.mat_vec.sp.matrix_sp"
  
#include <stdio.h>
#include <stdlib.h> 
#include <string.h> 
#include "matrix_sp.h"
#include "../matrix.h"
#include "../../db/db.h"
#include "../../util/log.h"
#include "../../util/utils.h"

#define MAX(a,b) (a > b ? a : b)
#define MIN(a,b) (a < b ? a : b)





				/*****************************************/
				//       CONSTRUCTOR / DESTRUCTOR        //
				/*****************************************/         






/**
 * \brief Creates a matrix of the specified
 *        format, on the CPU, with data of 
 *        the specified type
 * \param mat Pointer to the matrix structure 
 *        being created
 * \param rows,cols Matrix size
 * \param format Matrix format (e.g. CSR)
 * \param type Matrix data type 
 *        (float, qdouble, double)
 * \param name Matrix name 
 */
void CukrMatspCreate(CUKR_MATRIX_SP* mat, CUKR_INT rows, CUKR_INT cols,
				     int format, int type, char* name)
{
	char buffer[255];
	
	/** 
	 * Check for valid parameters 
	 */
	
	/* First check for valid name */
	if (name == NULL)
		CUKR_FATAL("Matrix must have a non-null name");
	
	LOGX(DEBUG, "Creating SP matrix %s", name);
	
	/* Check for valid type is done
	 * in the VecCreate routine */
	
	/*
	* Malloc for the matrix structure
	*/
	 
	/* Alloc memory */
	if ( ((*mat) = (CUKR_MATRIX_SP)malloc(sizeof(struct _CUKR_MATRIX_SP))) == NULL) {
		CUKR_FATAL("Error allocating memory for SP matrix object ");
	}
	else {
		LOG(DEBUG, "Allocated SP matrix structure");
	}
	  
	 
	 /* 
	  * Set all matrix parameters 
	  */
	 
	 /* Set build phase to created */
	 (*mat)->build_phase = CUKR_MAT_BUILD_CREATED;
	 
	 /* Set format */
	 if (format == CUKR_DEFAULT)
	 	(*mat)->format = CukrMatFormatStrToId(CukrDBGetS(CUKR_DB_MATRIX_FORMAT));
	 else
	 	(*mat)->format = format;
	 	
	 /* Set type */
	 if (type == CUKR_DEFAULT)
	 	 (*mat)->type_at_creation = CukrTypeStrToId(CukrDBGetS(CUKR_DB_SOLVER_PRECISION));
	 else
	 	 (*mat)->type_at_creation = type;
	 
	 /* Set variation to none, from the start */
	 (*mat)->variation = CUKR_MATRIX_FORMAT_VARIATION_NONE;	 
	  
	 /* Set sizes */ 
	 (*mat)->rows = rows;
	 (*mat)->cols = cols;
	 (*mat)->nz   = 0;
	 
	 /* Set matrix status */
	 (*mat)->status = CukrLocIdAndTypeIdToStatusId(CUKR_LOCATION_CPU, (*mat)->type_at_creation);	 
	 sprintf((*mat)->name, "%s", name); 
	 
	 /*
	  * Create data member
	  */
	 
	 /* In case of CSR format */
	 if ((*mat)->format == CUKR_MATRIX_FORMAT_CSR)
	 	CukrCsrMatspCreate(*mat);
	 /* In case of BCSR format */
	 else if ((*mat)->format == CUKR_MATRIX_FORMAT_BCSR)
		 CukrBcsrMatspCreate(*mat);
	 /* In case of HYB format */
	 else if ((*mat)->format == CUKR_MATRIX_FORMAT_HYB)
		 CukrHybMatspCreate(*mat);
	 /* In case of BHYB format */
	 else if ((*mat)->format == CUKR_MATRIX_FORMAT_BHYB)
		 CukrBhybMatspCreate(*mat);
	 /* Else, error */
	 else
		 CUKR_FATALX("Invalid format for matrix %s", name);
	 
	 /*
	  * Initialize temporary storage for entries
	  */
	  
	  CukrContainersVectorOfMatentryCreate(&(*mat)->entries);
	  
	 
	/*
	 * Success message 
	 */
	
	#ifdef CUKR_LOGGING
		char loc_str[255], type_str[255], fmt_str[255];
		CukrTypeIdToStr((*mat)->type_at_creation, type_str);
		CukrMatFormatIdToStr((*mat)->format, fmt_str);
		LOGX(DEBUG, "Successfully created matrix %s(%d,%d) on host with " \
		            "format %s and value type %s", 
		            (*mat)->name, rows, cols, fmt_str, type_str);
	#endif	             
	
}                   

// ------------------------------------------------------------------- //

/**
 * \brief Loads a matrix from file
 * \param mat Pointer to the matrix in which
 *        data is being read
 * \param filename The file where the data is
 *        being read from
 * \param format The matrix storage format (CSR)
 * \param type The data type the values should
 *        be considered to be (double, single,
 *        short).
 * \param name The name given to the matrix
 */
void CukrMatspCreateFromFile(CUKR_MATRIX_SP* mat, const char *filename, int format, int type, char* name)
{
	/*
	 * Validate parameters
	 */
	
	/* First check for valid name */
	if (name == NULL)
		CUKR_FATAL("Matrix must have a non-null name");
	 
	/* Check for valid type */
	if (type != CUKR_TYPE_DOUBLE  &&
		type != CUKR_TYPE_QDOUBLE &&
	    type != CUKR_TYPE_SINGLE  &&
	    type != CUKR_DEFAULT        )
	    CUKR_FATALX("Invalid data type for matrix %s", name);
	 
	
	 /*
	  * Create matrix structure
	  */
	 
	 /* Alloc memory */
	 if ( ((*mat) = (CUKR_MATRIX_SP)malloc(sizeof(struct _CUKR_MATRIX_SP))) == NULL)
		 CUKR_FATAL("Error allocating memory for matrix object ");
	 
	 /* 
	  * Set all matrix parameters 
	  */
	 
	 /* Set the name */
	 sprintf((*mat)->name, "%s", name);   
	 
	 /* Set build phase to created from file */
	 (*mat)->build_phase = CUKR_MAT_BUILD_CREATED_FILE;
	 
	 /* Set format */
	 if (format == CUKR_DEFAULT)
	 	(*mat)->format = CukrMatFormatStrToId(CukrDBGetS(CUKR_DB_MATRIX_FORMAT));
	 else
	 	(*mat)->format = format;
	 	
	 /* Set type */
	 if (type == CUKR_DEFAULT)
	 	 (*mat)->type_at_creation = CukrTypeStrToId(CukrDBGetS(CUKR_DB_SOLVER_PRECISION));
	 else
	 	 (*mat)->type_at_creation = type;
	 
	 /* Set variation to none, from the start */
	 (*mat)->variation = CUKR_MATRIX_FORMAT_VARIATION_NONE;	 
	 /* Set status */
	 (*mat)->status = CukrLocIdAndTypeIdToStatusId(CUKR_LOCATION_CPU, (*mat)->type_at_creation); 
	  
	 /*
	  * Load from file
	  */
	
	 /* If CSR format */
	 if ((*mat)->format == CUKR_MATRIX_FORMAT_CSR) 
		CukrCsrMatspCreateFromFile(*mat, filename);
	 /* if BCSR */
	 else if ((*mat)->format == CUKR_MATRIX_FORMAT_BCSR)
	 		CukrBcsrMatspCreateFromFile(*mat, filename);
	 /* if HYB */
	 else if ((*mat)->format == CUKR_MATRIX_FORMAT_HYB)
	 		CukrHybMatspCreateFromFile(*mat, filename);
	 /* if BHYB */
	 else if ((*mat)->format == CUKR_MATRIX_FORMAT_BHYB)
	 		CukrBhybMatspCreateFromFile(*mat, filename);
	 /* Else, error */
	 else 
	 	CUKR_FATALX("Invalid format for matrix %s: %d", 
	 	            name, (*mat)->format);	
	 	
	 /*
	  * Success message 
	  */
	  
	 #ifdef CUKR_LOGGING
		char loc_str[255], type_str[255], fmt_str[255];
		CukrTypeIdToStr((*mat)->type_at_creation, type_str);
		CukrMatFormatIdToStr((*mat)->format, fmt_str);
		LOGX(DEBUG, "Successfully created matrix %s(%d,%d) on host with " \
		            "format %s and value type %s", 
		            (*mat)->name, (*mat)->rows, (*mat)->cols, fmt_str, type_str);
	#endif	        	

}

// ------------------------------------------------------------------- //

/**
 * \brief Finalizes the creation of the matrix,
 *        after all elements have been added. 
 * \param mat The matrix structure to assemble.
 */
void CukrMatspAssemble(CUKR_MATRIX_SP mat)
{
	/*
	 * Check if the matrix is ready for
	 * assembly */
	
	if (mat->build_phase != CUKR_MAT_BUILD_CREATED &&
	    mat->build_phase != CUKR_MAT_BUILD_CREATED_FILE)
		CUKR_FATALX("Matrix %s is not ready for assembly "\
		            "or it has already been assembled", mat->name);	 
	
	LOGX(DEBUG, "Starting the assembly for SP matrix %s",
	             mat->name);
	             
	 /*
	  * Do this only if the matrix is not loaded 
	  * from file
	  */
	  
	 if (mat->build_phase != CUKR_MAT_BUILD_CREATED_FILE) {
        /*
		 * Set additional attributes
		 */
		 
		 /* Set the no. of nonzeros */
		 mat->nz = mat->entries->size;
		
		/* Call the assembly routine for the
		 * appropriate format */
		 
		/* If CSR format */
		if (mat->format == CUKR_MATRIX_FORMAT_CSR) 
		   CukrCsrMatspAssemble(mat);
		/* Else, error */
		else 
		   CUKR_FATALX("Invalid format for matrix %s", 
		               mat->name);	
		               
		/* Free temporary storage */
		CukrContainersVectorOfMatentryDestroy(&mat->entries);               
	 }             
	
	/*
	 * Format conversion, if necessary 
	 */
	 
	 CukrMatspConvert(mat, CukrMatFormatStrToId(CukrDBGetS(CUKR_DB_MATRIX_FORMAT)),
	                       CukrMatFormatVariationStrToId(CukrDBGetS(CUKR_DB_MATRIX_FORMAT_VAR)));
	
	/* 
	 * Finalize
	 */
	
	/* Set matrix as assembled */
	mat->build_phase = CUKR_MAT_BUILD_ASSEMBLED;
	
	LOGX(DEBUG, "Matrix %s was succesfully assembled",
	             mat->name);
}

// ------------------------------------------------------------------- //

/**
 * \brief Frees all the memory used 
 *        by the matrix
 * \param mat The matrix to be destroyed 
 */
void CukrMatspDestroy(CUKR_MATRIX_SP* mat)
{
	
	/* Copy the name, so it doesn't get
	 * destroyed */
	char name[255];
	memcpy(name, (*mat)->name, 255);
	LOGX(DEBUG, "Destroying matrix %s", name);
	
	/* Destroy the CSR part, currently
	 * the only one available */
	if ((*mat)->format == CUKR_MATRIX_FORMAT_CSR) 
		CukrCsrMatspDestroy(*mat); 
	else if ((*mat)->format == CUKR_MATRIX_FORMAT_BCSR)
		CukrBcsrMatspDestroy(*mat);
	else if ((*mat)->format == CUKR_MATRIX_FORMAT_HYB)
		CukrHybMatspDestroy(*mat);
	else if ((*mat)->format == CUKR_MATRIX_FORMAT_BHYB)
		CukrBhybMatspDestroy(*mat);
	else 
	 	CUKR_FATALX("Invalid format for matrix %s", name);	
		
	/* Finally, destroy the matrix stucture */
	free(*mat); 
	LOGX(DEBUG, "Matrix %s destroyed", name);
}






				/*****************************************/
				//          TRANSFER / CONVERT           // 
				/*****************************************/ 

  




/**
 * \brief Changes the location of a matrix
 *        by copying the data from one location
 *        to another
 * \param mat Pointer to the matrix 
 * \param dest_loc Destination matrix location (CPU,GPU)
 * \param dest_type Destination matrix value type 
 *        (float, double)
 */
void CukrMatspTransfer(CUKR_MATRIX_SP mat, int dest_loc, int dest_type)
{
	if (mat->format == CUKR_MATRIX_FORMAT_CSR) 
		CukrCsrMatmpTransfer(mat, dest_loc, dest_type);
	else if (mat->format == CUKR_MATRIX_FORMAT_BCSR)
		CukrBcsrMatmpTransfer(mat, dest_loc, dest_type);
	else if (mat->format == CUKR_MATRIX_FORMAT_HYB)
		CukrHybMatmpTransfer(mat, dest_loc, dest_type);
	else if (mat->format == CUKR_MATRIX_FORMAT_BHYB)
		CukrBhybMatmpTransfer(mat, dest_loc, dest_type);
	/* Else, error */
	else
		 CUKR_FATALX("Invalid format for matrix %s", mat->name);
		
	/* Update status */
	mat->status = CukrLocIdAndTypeIdToStatusId(dest_loc, dest_type);	
}

// ------------------------------------------------------------------- //

/**
 * \brief Changes the storage format of 
 *        the matrix to a diferent variation
 *        or to a completely different format
 * \param mat Pointer to the matrix 
 * \param dest_format Destination format
 * \param dest_variation Destination variation 
 */
void CukrMatspConvert(CUKR_MATRIX_SP mat, int dest_format, int dest_variation)
{
	/*
	 * Does not work if assembled
	 */
	 
	if (mat->build_phase == CUKR_MAT_BUILD_ASSEMBLED)
		CUKR_FATALX("Matrix %s cannot be converted since "\
		            "it is already in assembled form. Perform "\
		            "this conversion before calling the "\
		            "assembly routine ", mat->name); 
	 

	/* 
	 * Convert the CSR structure  
	 */

	if (mat->format == CUKR_MATRIX_FORMAT_CSR) 
		CukrCsrMatspConvert(mat, dest_format, dest_variation);
	else if (mat->format == CUKR_MATRIX_FORMAT_BCSR)
			CukrBcsrMatspConvert(mat, dest_format, dest_variation);
	else if (mat->format == CUKR_MATRIX_FORMAT_HYB)
			CukrHybMatspConvert(mat, dest_format, dest_variation);
	else if (mat->format == CUKR_MATRIX_FORMAT_BHYB)
			CukrBhybMatspConvert(mat, dest_format, dest_variation);
	else 
	 	CUKR_FATALX("Invalid format for matrix %s: %d", mat->name, mat->format);
		
}






				/*****************************************/
				//               GET/SET                 // 
				/*****************************************/ 





/**
 * \brief Adds an element to a matrix. This can only
 *        be used after the matrix has been created and
 *        before it is assembled.
 * \param mat The matrix where the element is to be added
 * \param row The row of the element
 * \param col The col of the element
 * \param val The value of the element
 */
void CukrMatspAddElement(CUKR_MATRIX_SP mat, int row, int col, CUKR_DOUBLE val)
{
	/*
	 * Error if matrix already assembled
	 * or created from file 
	 */
	 
	 if (mat->build_phase != CUKR_MAT_BUILD_CREATED)
	 	CUKR_FATALX("Cannot add elements to matrix %s" \
	 	       		"since it has already been assembled "\
	 	       		"or has been created from a file", mat->name);
	
	/* 
	 * Perform bounds check. Slow, but
	 * such function will not be used
	 * in production anyway. 
	 */
	
	if (row < 1 || row > mat->rows ||
	    col < 1 || col > mat->cols   )
	    CUKR_FATALX("Invalid entry (row=%d,col=%d,val=%g) "\
	                "was added to matrix %s(%d,%d). Remember "\
	                "that numbering starts from 1.", row, col, 
	           val, mat->name, mat->rows, mat->cols);
	
	/* 
	 * Add to container 
	 */
	 
	CUKR_MATENTRY entry;
	entry.col = col; entry.row = row; entry.val = val; 
	CukrContainersVectorOfMatentryAppend(mat->entries, entry);
	                  
}

// ------------------------------------------------------------------- //

/**
 * \brief Adds a row of elements to a matrix. 
 *        This can only be used after the
 *        matrix has been created and before
 *        it is assembled.
 * \param mat The matrix where the row 
 *        is to be added
 * \param row The row to be added
 * \param val Vector of values, having at least as  
 *        many elements as the length of the row.
 */
void CukrMatspAddRow(CUKR_MATRIX_SP mat, int row, CUKR_DOUBLE* val)
{
	int i;
	
	for (i = 0; i < mat->cols; i++)
		CukrMatspAddElement(mat, row, i, val[i]);
}

// ------------------------------------------------------------------- //

/**
 * \brief Adds a column of elements to a matrix. 
 *        This can only be used after the
 *        matrix has been created and before
 *        it is assembled.
 * \param mat The matrix where the column 
 *        is to be added
 * \param col The column to be added
 * \param val Vector of values, having at least as  
 *        many elements as the length of the column.
 */
void CukrMatspAddCol(CUKR_MATRIX_SP mat, int col, CUKR_DOUBLE* val)
{
	int i;
	
	for (i = 0; i < mat->rows; i++)
		CukrMatspAddElement(mat, i, col, val[i]);
}

// ------------------------------------------------------------------- //

/**
 * \brief Adds a diagonal of elements to a matrix. 
 *        This can only be used after the
 *        matrix has been created and before
 *        it is assembled.
 * \param mat The matrix where the diagonal 
 *        is to be added
 * \param diag The diagonal to be added. 0 is the
 *        main diagonal, -1 ... min(rows,cols) are
 *        the lower diagonals and 1 .. min(rows,cols)
 *        are the upper diagonals.
 * \param val Vector of values, having at least as many 
 *        elements as the length of the diagonal.
 */
void CukrMatspAddDiag(CUKR_MATRIX_SP mat, int diag, CUKR_DOUBLE* val)
{
	int i;
	
	/* Compute start and end */
	int rs = MAX(0, -diag);
	int re = MIN(mat->rows, mat->cols - diag);
	   
	/* Add elements (adjust to start from 1) */
	for (i = 0; i < re - rs; i++) {
		int aux = rs + i + 1;
		CukrMatspAddElement(mat, aux , aux + diag, val[i]);
	}	 
}

/**i
 * \brief Extracts the diagonal from a matrix
 * \param mat The matrix 
 * \param vec The vector where to store the diagonal
 * \param diag The diagonal to read (zero for the
 *        main diagonal)
 * \param invert Whether to invert it or not 
 */
void CukrMatspGetDiag(CUKR_MATRIX_SP mat, CUKR_VECTOR_SP vec, int diag, int invert)
{	
	if (mat->format == CUKR_MATRIX_FORMAT_CSR)
		CukrCsrMatspGetDiag(mat, vec, diag, invert);
	else if (mat->format == CUKR_MATRIX_FORMAT_HYB)
		CukrHybMatspGetDiag(mat, vec, diag, invert);
	else if (mat->format == CUKR_MATRIX_FORMAT_BCSR)
		CukrBcsrMatspGetDiag(mat, vec, diag, invert);
	else
		CUKR_FATALX("Cannot get the diagonal of matrix %s. " \
					"Unsupported format (%d)", mat->name, mat->format);
}



				/*****************************************/
				//                 I/O                   // 
				/*****************************************/ 





/**
 * PROTECTED 
 * 
 * \brief Open the file storing a matrix
 *        in the specified format and return the
 *        matrix dimensions (rows,cols) and the
 *        no. nonzeros
 * \param filename The full path to the file 
 *        storing the matrix
 * \param format The matrix format ID
 * \param rows Pointer to a variable where to 
 *        store the no. of rows
 * \param cols Pointer to a variable where to 
 *        store the no. of cols
 * \param nz Pointer to a variable where to 
 *        store the no. of nz
 */
void CukrMatspReadSizesFromFile(char* filename, int format, int* rows, int* cols, int* nz)
{
	if (format == CUKR_MATRIX_FORMAT_CSR) 
		return CukrCsrMatspReadSizesFromFile(filename, rows, cols, nz);
	else
		CUKR_FATALX("Invalid format was specified for " \
		            "matrix file %s: %d", filename, format);	
}



/**
 * \brief Prints a matrix
 * \param mat The matrix being printed
 */
void CukrMatspPrint(CUKR_MATRIX_SP mat)
{
	CukrMatspPrintToFile(mat, stdout);
}

// ------------------------------------------------------------------- //

/**
 * \brief Prints a matrix to file
 * \param vec The matrix being printed
 * \param filename The full path of the file
 *        to write the matrix to
 */
void CukrMatspPrintToFilename(CUKR_MATRIX_SP mat, char* filename)
{
	FILE *f;
	
	if ( (f = fopen(filename, "w")) == NULL)
		CUKR_FATALX("Could not write matrix %s to file "\
		            "%s because the file could not be opened",
		            mat->name, filename);
	
	CukrMatspPrintToFile(mat, f);
	
	fclose(f);
}

// ------------------------------------------------------------------- //

/**
 * \brief Prints a matrix to a file stream
 * \param mat The matrix being printed
 * \param f An open file stream to write to
 */
void CukrMatspPrintToFile(CUKR_MATRIX_SP mat, FILE *f)
{
	CukrMatspPrintToFileWithIndent(mat, f, "");
}

// ------------------------------------------------------------------- //

/**
 * \brief Prints a matrix to a file stream
 * \param mat The matrix being printed
 * \param f An open file stream to write to
 * \param indent An indentation string, to be inserted
 *        before every print.
 */
void CukrMatspPrintToFileWithIndent(CUKR_MATRIX_SP mat, FILE *f, char* indent) 
{
	LOGX(DEBUG, "Printing matrix %s", mat->name);
	
	/* Print in CSR */
	if (mat->format == CUKR_MATRIX_FORMAT_CSR)
		CukrCsrMatspPrintToFileWithIndent(mat, f, indent);
	/* Print in BCSR */
	else if (mat->format == CUKR_MATRIX_FORMAT_BCSR)
		CukrBcsrMatspPrintToFileWithIndent(mat, f, indent);
	/* Print in HYB */
	else if (mat->format == CUKR_MATRIX_FORMAT_HYB)
		CukrHybMatspPrintToFileWithIndent(mat, f, indent);
	/* Print in BHYB */
	else if (mat->format == CUKR_MATRIX_FORMAT_BHYB)
		CukrBhybMatspPrintToFileWithIndent(mat, f, indent);
	else
		CUKR_FATALX("Invalid format was specified for " \
					"matrix %s: %d", mat->name, mat->format);
		
	LOGX(DEBUG, "Done printing matrix %s", mat->name);		
}
