/**
 * @file mat_vec/sp/matrix_sp.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines the CUKR_MATRIX_SP abstract
 * interace.
 * 
 * The CUKR_MATRIX_SP structure stores a 
 * matrix that can be in any (useful) type
 * and reside either on the CPU or GPU.
 * The structure stores, besides the data,
 * both type and location.
 * 
 * The BLAS2 operations are brought one
 * level up. They take paramters only the
 * matrices and vectors; the type, size and 
 * location are automatically read and the 
 * appropriate BLAS operation is chosen.
 * 
 * If matrices resides in a different location
 * than the BLAS, data is automatically moved. 
 */

#ifndef CUKR_MATRIX_SP_H_
#define CUKR_MATRIX_SP_H_

#ifndef CUKR_MATRIX
	#define CUKR_MATRIX CUKR_MATRIX_SP
#endif

/* =============== FORWARD DECLARATIONS ============== */

typedef struct _CUKR_MATRIX_SP *CUKR_MATRIX_SP;

/* =================================================== */

#include "csr/csr_matrix_sp.h"
#include "bcsr/bcsr_matrix_sp.h"
#include "hyb/hyb_matrix_sp.h"
#include "bhyb/bhyb_matrix_sp.h"
//#include "bell/bell_matrix_sp.h"
#include "matrix_sp_gen.h"
#include "vector_sp.h"
#include "../../util/containers.h"


/**
 * Holds a general matrix, independed
 * of the format, location or value type.
 */
struct _CUKR_MATRIX_SP{
	
	// ------------ Public ------------- //
	
	/** Matrix dimensions */
	int rows, cols, nz;
	/** Matrix name */
	char name[255];
	
	
	// ----------- Protected ------------ //
	
	/** Matrix format: currently only CSR */
	int format;
	/** Matrix format variation: none, CSR4, BCSR */
	int variation;
	/** Member storing the CSR version */
	CUKR_CSR_MATRIX_SP csr_mat;
	/** Member storing the BCSR version */
	CUKR_BCSR_MATRIX_SP bcsr_mat;
	/** Member storing the HYB version */
	CUKR_HYB_MATRIX_SP hyb_mat;
	/** Member storing the BHYB version */
	CUKR_BHYB_MATRIX_SP bhyb_mat;
	/** Member storing the BELL version */
//	CUKR_BELL_MATRIX_SP bell_mat;
	
	// ------------ Private ------------- //
	
	/** Matrix status (location + precision) */
	int status;
	/** Matrix build phase (created -> assembled) */
	int build_phase;
	/** Matrix data type at creation */
	int type_at_creation;
	/** Container for elements that will be added
	 * to the matrix (temporary storage)*/
	CUKR_CONTAINER_VECTOR_MATENTRY entries; 
};

/* 
 * Function pointers for the functions which
 * have to be scheduled to threads
 */

typedef void (*CUKR_MAT_DESTROY)(CUKR_MATRIX_SP*);
typedef void (*CUKR_MAT_TRANSFER)(CUKR_MATRIX_SP, int, int);
typedef void (*CUKR_MAT_CONVERT)(CUKR_MATRIX_SP, int, int);


		




				/*****************************************/
				//       CONSTRUCTOR / DESTRUCTOR        //
				/*****************************************/         






/**
 * \brief Creates a matrix of the specified
 *        format, on the CPU, with data of 
 *        the specified type
 * \param mat Pointer to the matrix structure 
 *        being created
 * \param rows,cols Matrix size
 * \param format Matrix format (e.g. CSR)
 * \param type Matrix data type 
 *        (float, qdouble, double)
 * \param name Matrix name 
 */
void CukrMatspCreate(CUKR_MATRIX_SP* mat, CUKR_INT rows, CUKR_INT cols,
				     int format, int type, char* name);

// ------------------------------------------------------------------- //

/**
 * \brief Loads a matrix from file.
 * \param mat Pointer to the matrix in which
 *        data is being read
 * \param filename The file where the data is
 *        being read from
 * \param format The matrix storage format (CSR)
 * \param type The data type the values should
 *        be considered to be (double, qdouble,
 *        single).
 * \param name The name given to the matrix
 */
void CukrMatspCreateFromFile(CUKR_MATRIX_SP* mat, const char *filename, int format, int type, char* name);

// ------------------------------------------------------------------- //

/**
 * \brief Finalizes the creation of the matrix,
 *        after all elements have been added. This
 *        inlcudes the creation of data structures
 *        and building the comm. tables.
 * \param mat The matrix structure to assemble.
 */
void CukrMatspAssemble(CUKR_MATRIX_SP mat); 

// ------------------------------------------------------------------- //
                   
/**
 * \brief Frees all the memory used 
 *        by the matrix
 * \param mat Pointer to the matrix 
 *        to be destroyed 
 */
void CukrMatspDestroy(CUKR_MATRIX_SP* mat);
                 




				/*****************************************/
				//               GET/SET                 // 
				/*****************************************/ 





/**
 * \brief Adds an element to a matrix. This can only
 *        be used after the matrix has been created and
 *        before it is assembled.
 * \param mat The matrix where the element is to be added
 * \param row The row of the element
 * \param col The col of the element
 * \param val The value of the element
 */
void CukrMatspAddElement(CUKR_MATRIX_SP mat, int row, int col, CUKR_DOUBLE val);

// ------------------------------------------------------------------- //

/**
 * \brief Adds a row of elements to a matrix. 
 *        This can only be used after the
 *        matrix has been created and before
 *        it is assembled.
 * \param mat The matrix where the row 
 *        is to be added
 * \param row The row to be added
 * \param val Vector of values, having at least as  
 *        many elements as the length of the row.
 */
void CukrMatspAddRow(CUKR_MATRIX_SP mat, int row, CUKR_DOUBLE* val);

// ------------------------------------------------------------------- //

/**
 * \brief Adds a column of elements to a matrix. 
 *        This can only be used after the
 *        matrix has been created and before
 *        it is assembled.
 * \param mat The matrix where the column 
 *        is to be added
 * \param col The column to be added
 * \param val Vector of values, having at least as  
 *        many elements as the length of the column.
 */
void CukrMatspAddCol(CUKR_MATRIX_SP mat, int col, CUKR_DOUBLE* val);

// ------------------------------------------------------------------- //

/**
 * \brief Adds a diagonal of elements to a matrix. 
 *        This can only be used after the
 *        matrix has been created and before
 *        it is assembled.
 * \param mat The matrix where the diagonal 
 *        is to be added
 * \param diag The diagonal to be added. 0 is the
 *        main diagonal, -1 ... min(rows,cols) are
 *        the lower diagonals and 1 .. min(rows,cols)
 *        are the upper diagonals.
 * \param val Vector of values, having at least as many 
 *        elements as the length of the diagonal.
 */
void CukrMatspAddDiag(CUKR_MATRIX_SP mat, int diag, CUKR_DOUBLE* val);

// ------------------------------------------------------------------- //

/**
 * \brief Extracts the diagonal from a matrix
 * \param mat The matrix 
 * \param vec The vector where to store the diagonal
 * \param diag The diagonal to read (zero for the
 *        main diagonal)
 * \param invert Whether to invert it or not 
 */
void CukrMatspGetDiag(CUKR_MATRIX_SP mat, CUKR_VECTOR_SP vec, int diag, int invert);







				/*****************************************/
				//          TRANSFER / CONVERT           // 
				/*****************************************/ 

  
  
  
  
  
/**
 * \brief Changes the location of a matrix
 *        by copying the data from one location
 *        to another (if not already there)
 * \param mat Pointer to the matrix 
 * \param dest_loc Destination matrix location (CPU,GPU)
 * \param dest_type Destination matrix value type 
 *        (float, double)
 */
void CukrMatspTransfer(CUKR_MATRIX_SP mat, int dest_loc, int dest_type);                   

// ------------------------------------------------------------------- //

/**
 * \brief Changes the storage format of 
 *        the matrix to a diferent variation
 *        or to a completely different format
 *        (if not already in that format)
 * \param mat The matrix 
 * \param dest_format Destination format
 * \param dest_variation Destination variation 
 */
void CukrMatspConvert(CUKR_MATRIX_SP mat, int dest_format, int dest_variation);






				/*****************************************/
				//                 I/O                   // 
				/*****************************************/ 







/**
 * \brief Prints a matrix
 * \param mat The matrix being printed
 */
void CukrMatspPrint(CUKR_MATRIX_SP mat);
	
// ------------------------------------------------------------------- //

/**
 * \brief Prints a matrix to file
 * \param vec The matrix being printed
 * \param filename The full path of the file
 *        to write the matrix to
 */
void CukrMatspPrintToFilename(CUKR_MATRIX_SP mat, char* filename);

// ------------------------------------------------------------------- //

/**
 * \brief Prints a matrix to a file stream
 * \param mat The matrix being printed
 * \param f An open file stream to write to
 */
void CukrMatspPrintToFile(CUKR_MATRIX_SP mat, FILE *f);

// ------------------------------------------------------------------- //

/**
 * \brief Prints a matrix to a file stream,
 *        adding an indent. Used for nesting.
 * \param mat The matrix being printed
 * \param f An open file stream to write to
 * \param indent An indentation string, to be inserted
 *        before every print.
 */
void CukrMatspPrintToFileWithIndent(CUKR_MATRIX_SP mat, FILE *f, char* indent); 

#endif /*CUKR_MATRIX_SP_H_*/
