/**
 * @file mat_vec/sp/matrix_gen.c
 * @author Serban Georgescu
 * @since 2008
 * 
 * Holds a list of matrix generators.
 * The generators here are indepenent of
 * the matrix format, so the format-specific
 * implementation of the generator will be
 * found elsewere (most probably in the
 * subfolder assigned to the specific format)
 * 
 * The generator will create the matrix,
 * so the matrix should not be created beforehand
 * 
 * Current generators are:
 * 	  - random 
 * 	  - poisson 2D 
 * 
 */

#define LOGGER_NAME "CUKr.mat_vec.sp.matrix_sp_gen"

#include <stdio.h>
#include <stdlib.h>
#include <time.h> 
#include "matrix_sp_gen.h"
#include "../matrix_defs.h"
#include "../../util/log.h"
#include "../../util/utils.h"
#include "csr/csr_matrix_sp.h"






				/*****************************************/
				//                WORKING                //
				/*****************************************/         






                 

/**
 * \brief Generates a diagonally-dominated banded
 *        matrix, filled with ones but the main 
 * 		  diagonal computed as such as to be dominant.
 * 		  The diagonals can be shifted to the left or 
 *        right by specifying a negative or positive shift 
 * \param mat Pointer to the matrix 
 * \param format Matrix format (CSR)
 * \param rows*cols Matrix order
 * \param hb Halfbandwidth 
 * \param shift The shift  
 * \param name Matrix name
 */
void CukrMatspGenBandedWShift(CUKR_MATRIX_SP* mat, int format, int rows, int cols, int hb, int shift, char* name)
{
	/* Do it for CSR structure 
	 * (only CSR is currently supported) */
	if (format == CUKR_MATRIX_FORMAT_CSR) 
		CukrCsrMatspGenBandedWShift(mat, rows, cols, hb, shift, name);
		
	/* Error if unknown format */	
	else
		CUKR_FATALX("Invalid format for matrix %s", name);	
}








				/*****************************************/
				//              NOT  WORKING             //
				/*****************************************/         











/**
 * \brief Generates a sparse random matrix, filled
 *        with random elements 
 * \param mat Pointer to the matrix 
 * \param format Matrix format (CSR)
 * \param n Matrix size
 * \param halfb The half-bandwidth of the matrix
 * \param nz_per_row No. of nonzeros per row
 * \param min,max The lower and upper bound for
 *                matrix values
 * \param name Matrix name
 */
void CukrMatspCreateRandom(CUKR_MATRIX_SP* mat, int format, int n, int halfb, int nz_per_row, 
 						 CUKR_DOUBLE min, CUKR_DOUBLE max, char* name)
{
	
	/* Do it for CSR structure 
	 * (only CSR is currently supported) */
	if (format == CUKR_MATRIX_FORMAT_CSR) 
		CukrCsrMatspGenRandom(mat, n, halfb, nz_per_row, min, max, name);
		
	/* Error if unknown format */	
	else
		CUKR_FATALX("Invalid format for matrix %s", name);
}                                

// ------------------------------------------------------------------- //
					 
/**
 * \brief Generates a sparse random matrix, filled
 *        with random elements 
 * \param mat Pointer to the matrix 
 * \param format Matrix format (CSR)
 * \param ord Order
 * \param name Matrix name
 */
void CukrMatspCreatePoisson2D(CUKR_MATRIX_SP* mat, int format, int ord, char* name)
{
	/* Do it for CSR structure 
	 * (only CSR is currently supported) */
	if (format == CUKR_MATRIX_FORMAT_CSR) 
		CukrCsrMatspGenPoisson2D(mat, ord, name);
		
	/* Error if unknown format */	
	else
		CUKR_FATALX("Invalid format for matrix %s", name);	
}
