/**
 * @file mat_vec/vector_sp.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines the CUKR_VECTOR_SP abstract
 * interace.
 * 
 * The CUKR_VECTOR_SP structure stores a 
 * vector that can be in any (useful) type
 * and reside either on the CPU or GPU.
 * The structure stores, besides the data,
 * both precision and location.
 * 
 * The precision shows can be of type 
 * double, qdouble (quasi-double) single or int.
 * 
 * The location can be CPU or GPU.
 * 
 * Precision and location are combined into a 
 * status, which shows where actual (updated)
 * data exists. For example, if a vector has
 * been created on the CPU in double precision,
 * it will start with the CUKR_STATUS_CPU_DOUBLE
 * status. If afterwards it will be converted to
 * single precision, the new status will of
 * CUKR_STATUS_CPU_DOUBLE | CUKR_STATUS_CPU_SINGLE
 * will reflect that. In this way, we can know
 * at any time if actual data is present where we
 * need it, and only copy / convert where needed.
 *  
 */

#define LOGGER_NAME "CUKr.mat_vec.sp.vector_sp"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h> 
#include "vector_sp.h"
#include "../../util/log.h"
#include "../../util/utils.h"
#include "../../db/db.h"
#include "../../blas/impl/generic/generic_blas.h"
#include "../../blas/impl/gpu/gpu_blas.h"
#ifdef CUKR_USE_CUDA
	#include <cublas.h>
	#include "cuda/vector_kernels.h"
#endif
#ifdef CUKR_USE_OPENCL
	#include "../../init/init_opencl.h"
	#include "opencl/vector_kernels.h"
#endif

/* Access the global option database and blas objects */
extern CukrDB cukr_db;
 
//========================== FORWARD DECLARATIONS ========================= //

void CukrVecspCopyDataCPU2CPU(CUKR_VECTOR_SP vec, int dest_type);
void CukrVecspCopyDataCPU2GPU(CUKR_VECTOR_SP vec, int dest_type); 
void CukrVecspCopyDataGPU2CPU(CUKR_VECTOR_SP vec, int dest_type); 
void CukrVecspCopyDataGPU2GPU(CUKR_VECTOR_SP vec, int dest_type); 
#ifdef CUKR_USE_CUDA
	void CukrHandleCudaMemErrors(CUKR_VECTOR_SP vec, cublasStatus status);
#endif

//============================ PUBLIC FUNCTIONS =========================== //



				/*****************************************/
				//       CONSTRUCTOR / DESTRUCTOR        //
				/*****************************************/         


/**
 * \brief Creates a vector of the specified 
 *        size, on the CPU, with the specified 
 *        type. Since by default the vectors are
 *        created on the CPU, the locatio need
 *        not be specified.
 * \param vec Pointer to the vector structure 
 *        being created
 * \param n Vector size
 * \param type Vector type (double, qddouble, 
 *        single, int)
 * \param name Vector name
 */
void CukrVecspCreate(CUKR_VECTOR_SP* vec, CUKR_INT n, int type, char* name)
{
	 
	/*
	 * Validate parameters
	 */
	
	/* First validate name */
	if (name == NULL)
		CUKR_FATAL("Vector must have a non-null name");
	
	#ifdef CUKR_LOGGING
		LOGX(DEBUG, "Creating SP vector %s(%d)s", 
		            name, n);
	#endif
		
	/*
	 * Malloc for the vector structure
	 */ 
	 
	/* Alloc memory */
	if ( ((*vec) = (CUKR_VECTOR_SP)malloc(sizeof(struct _CUKR_VECTOR_SP))) == NULL) {
		CUKR_FATAL("Error allocating memory for vector object ");
	} 
	else {
		LOGX(DEBUG, "Allocated vector structure [addr = %p]", *vec);
	}
	
	/*
	 * Init and set
	 */
	
	/* NULL all the data members */
	(*vec)->pd_dval  = NULL; (*vec)->ph_dval  = NULL;
	(*vec)->pd_qvalt = NULL; (*vec)->ph_qvalt = NULL; 
	(*vec)->pd_sval  = NULL; (*vec)->ph_sval  = NULL;  
	(*vec)->pd_ival  = NULL; (*vec)->ph_ival  = NULL;
	#ifdef CUKR_USE_OPENCL
	(*vec)->pcl_dval  = NULL;
	(*vec)->pcl_qvalt = NULL;
	(*vec)->pcl_sval  = NULL;
	(*vec)->pcl_ival  = NULL;
	#endif
	
	/* Get vector type */
	int vec_type;
	if (type == CUKR_DEFAULT)
		vec_type = CukrTypeStrToId(CukrDBGetS(CUKR_DB_SOLVER_PRECISION));
	else
	 	vec_type = type;
	
	/* Set vector size, precision, status 
	 * and name */
	(*vec)->n = n;
	(*vec)->status = CukrLocIdAndTypeIdToStatusId(CUKR_LOCATION_CPU, vec_type);
	sprintf((*vec)->name, "%s", name);
	
	/* If type is qdouble, since it includes 
	 * single, set it as well */
	if (vec_type == CUKR_TYPE_QDOUBLE)
		(*vec)->status |= CUKR_STATE_CPU_SINGLE; 
	
	
	/*
	 * Alloc memory
	 */
	 
	/* Alloc memory */
	if (type != CUKR_TYPE_NONE)		
		CukrVecspMalloc((*vec), CUKR_LOCATION_CPU, vec_type);

	/* Success message */
	#ifdef CUKR_LOGGING
		char status_str[255];
		CukrStatusIdToStr((*vec)->status, status_str);
		LOGX(DEBUG, "Successfully created vector %s with status %s", 
		            (*vec)->name, status_str);
	#endif	            

}

// ------------------------------------------------------------------- //

/**
 * \brief Creates a clone of a vector. 
 *        The function currently copies
 *        the data in the highest precision
 *        available, and leaves the rest
 *        of the precisions invalidated, so
 *        they will be transformed/converted
 *        to only if needed.
 * \param src The source vector
 * \param dst Pointer to the destination vector
 * \param name The new name for the destination
 */
void CukrVecspCreateClone(CUKR_VECTOR_SP src, CUKR_VECTOR_SP* dst, char* name)
{
	int i;
		
	/* 
	 * Create the clone in the highest
	 * precision available, and copy it.
	 * The rest will be added automatically
	 * if needed.
	 */
	
	/* If double on CPU */ 
	if (src->status & CUKR_STATE_CPU_DOUBLE) {
		LOGX(DEBUG, "Copy %s.cpu.double -> %s.cpu.double",
		            src->name, name);
		CukrVecspCreate(dst, src->n, CUKR_TYPE_DOUBLE, name);
		generic_dcopy(src->n, src->ph_dval, (*dst)->ph_dval);
		(*dst)->status = CUKR_STATE_CPU_DOUBLE;
		return; 					        
	}
	
	/* If double on GPU */ 
	#ifdef CUKR_USE_GPU 
	if (src->status & CUKR_STATE_GPU_DOUBLE) {
		LOGX(DEBUG, "Copy %s.gpu.double -> %s.gpu.double",
		            src->name, name);
		CukrVecspCreate(dst, src->n, CUKR_TYPE_NONE, name);
		CukrVecspMallocGPU(*dst, CUKR_TYPE_DOUBLE);
		#ifdef CUKR_USE_OPENCL
		gpu_dcopy(src->n, src->pcl_dval, (*dst)->pcl_dval);
		#else
		gpu_dcopy(src->n, src->pd_dval, (*dst)->pd_dval);
		#endif
		(*dst)->status = CUKR_STATE_GPU_DOUBLE;
		return; 					        
	}
	#endif
	
	/* If qdouble on CPU */ 
	if (src->status & CUKR_STATE_CPU_QDOUBLE) {
		LOGX(DEBUG, "Copy %s.cpu.qdouble -> %s.cpu.qdouble",
		            src->name, name);
		CukrVecspCreate(dst, src->n, CUKR_TYPE_QDOUBLE, name);
		generic_qcopy(src->n, src->ph_sval, src->ph_qvalt, (*dst)->ph_sval, (*dst)->ph_qvalt);
		(*dst)->status = CUKR_STATE_CPU_QDOUBLE;
		return; 					        
	}
	
	/* If qdouble on GPU */
	#ifdef CUKR_USE_GPU 
	if (src->status & CUKR_STATE_GPU_QDOUBLE) {
		LOGX(DEBUG, "Copy %s.gpu.qdouble -> %s.pu.qdouble",
		            src->name, name);
		CukrVecspCreate(dst, src->n, CUKR_TYPE_NONE, name);
		CukrVecspMallocGPU(*dst, CUKR_TYPE_DOUBLE);
		#ifdef CUKR_USE_OPENCL
		gpu_qcopy(src->n, src->pcl_sval, src->pcl_qvalt, (*dst)->pcl_sval, (*dst)->pcl_qvalt);
		#else
		gpu_qcopy(src->n, src->pd_sval, src->pd_qvalt, (*dst)->pd_sval, (*dst)->pd_qvalt);
		#endif
		(*dst)->status = CUKR_STATE_GPU_QDOUBLE;
		return; 					        
	}
	#endif
	
	/* If single on CPU */
	if (src->status & CUKR_STATE_CPU_SINGLE) {
		LOGX(DEBUG, "Copy %s.cpu.single -> %s.cpu.single",
		            src->name, name);
		CukrVecspCreate(dst, src->n, CUKR_TYPE_SINGLE, name);
		generic_scopy(src->n, src->ph_sval, (*dst)->ph_sval);
		(*dst)->status = CUKR_STATE_CPU_SINGLE;
		return; 					        
	}
	
	/* If single on GPU */ 
	#ifdef CUKR_USE_GPU 
	if (src->status & CUKR_STATE_GPU_SINGLE) {
		LOGX(DEBUG, "Copy %s.gpu.single -> %s.gpu.single",
		            src->name, name);
		CukrVecspCreate(dst, src->n, CUKR_TYPE_NONE, name);
		CukrVecspMallocGPU(*dst, CUKR_TYPE_SINGLE);
		#ifdef CUKR_USE_OPENCL
		gpu_scopy(src->n, src->pcl_sval, (*dst)->pcl_sval);
		#else
		gpu_scopy(src->n, src->pd_sval, (*dst)->pd_sval);
		#endif
		(*dst)->status = CUKR_STATE_GPU_SINGLE;
		return; 					        
	}
	#endif
	
	/* If int on CPU */ 
	if (src->status & CUKR_STATE_CPU_INT) {
		LOGX(DEBUG, "Copy %s.cpu.int -> %s.cpu.int",
		            src->name, name);
		CukrVecspCreate(dst, src->n, CUKR_TYPE_INT, name);
		CukrVecspTransfer(src, CUKR_LOCATION_CPU, CUKR_TYPE_INT);
		for (i = 0 ; i < src->n ; i++)
			(*dst)->ph_ival[i] = src->ph_ival[i];
		(*dst)->status = CUKR_STATE_CPU_INT;
		return; 					        
	}
	
	/* If int on GPU */ 
	#ifdef CUKR_USE_GPU 
	if (src->status & CUKR_STATE_GPU_INT) {
		/* Currently not supported since it cannot
		 * be done with the blas. Have to make a
		 * kernel */
		 CUKR_FATALX("Cloning for vector %s failed. "\
		             "Cloning an integer vector, on the GPU "\
		             "is not currently supported", src->name);					        
	}
	#endif
	
	/* If went here, error, since no
	 * valid type was found */ 
	CUKR_FATALX("Error cloning vector %s from vector %s. "\
	            "No data was found at source.", 
	            name, src->name); 	             
	
}


// ------------------------------------------------------------------- //

/**
 * \brief Frees all the memory used 
 *        by the data members, but leaves
 *        the vector structure intact
 * \param vec The vector
 */
void CukrVecspDestroyData(CUKR_VECTOR_SP vec)
{
	/* Free all the non-null value vectors,
	 * on CPU */
	CukrVecspFreeCPU(vec, CUKR_TYPE_DOUBLE);
	CukrVecspFreeCPU(vec, CUKR_TYPE_QDOUBLE);
	CukrVecspFreeCPU(vec, CUKR_TYPE_SINGLE);
	CukrVecspFreeCPU(vec, CUKR_TYPE_INT); 
	
	/* Free all the non-null value vectors,
	 * on GPU */
	#ifdef CUKR_USE_GPU
		CukrVecspFreeGPU(vec, CUKR_TYPE_DOUBLE);
		CukrVecspFreeGPU(vec, CUKR_TYPE_QDOUBLE); 
		CukrVecspFreeGPU(vec, CUKR_TYPE_SINGLE);
		CukrVecspFreeGPU(vec, CUKR_TYPE_INT);	
	#endif	
}

// ------------------------------------------------------------------- //

/**
 * \brief Frees all the memory used 
 *        by the vector
 * \param vec Pointer to the vector to 
 *        be destroyed 
 */
void CukrVecspDestroy(CUKR_VECTOR_SP* vec)
{	
	/* Check for non-null vector */
	if (*vec == NULL)
		return;
	
	/* Copy the name, so it doesn't get
	 * destroyed */
	char name[255];
	memcpy(name, (*vec)->name, 255);
	LOGX(DEBUG, "Destroying vector %s", name);
	
	/* Free all the data members */
	CukrVecspDestroyData(*vec);
		
	/* Finally, free the vector structure */
	free(*vec);
	LOGX(DEBUG, "Vector %s destroyed", name);
}

	
			
				/*****************************************/
				//            MALLOC / FREE              // 
				/*****************************************/ 
				
								      

/**
 * \brief Allocate memory for a vector. The 
 *        function checks if the pointer 
 *        corresponding to that particular location
 *        and precision is NULL of not. If not NULL,
 *        on malloc will be performed. Therefore, this
 *        function can be called "just in case" to 
 *        make sure memory allocation where it 
 *        should be.
 * \param vec The vector structure being created
 * \param loc Vector location {CPU,GPU}
 * \param type Vector type {double, qdouble, 
 *        single, int}
 */
void CukrVecspMalloc(CUKR_VECTOR_SP vec, int loc, int type)
{
	
	/* If the location is on the CPU, alloc
	 * memory in the appropriate member */
  	if (loc == CUKR_LOCATION_CPU)
 		CukrVecspMallocCPU(vec, type);
 					
  	/* If the location is on the GPU, alloc
	 * memory in the appropriate member */
	else if (loc == CUKR_LOCATION_GPU) 
		CukrVecspMallocGPU(vec, type);
		
	/* Else, error */
	else
		CUKR_FATALX("Cannot alloc memory for vector %s " \
		            "because of invalid location: %d",
		             vec->name, loc);	
}

// ------------------------------------------------------------------- //

/**
 * \brief Alloc memory for the data member
 *        of a vector corresponding to the
 *        desired precision, on the CPU. The 
 *        function checks if the pointer 
 *        corresponding to that particular location
 *        and precision is NULL of not. If not NULL,
 *        on malloc will be performed. Therefore, this
 *        function can be called "just in case" to 
 *        make sure memory allocation where it 
 *        should be.
 * \param vec Pointer to the vector
 * \param type Vector type {float, double, int}
 */
void CukrVecspMallocCPU(CUKR_VECTOR_SP vec, int type) 
{	
	/* Alloc memory for double */
	if (type == CUKR_TYPE_DOUBLE) { 
		if (vec->ph_dval == NULL){
			if ((vec->ph_dval = (CUKR_DOUBLE*)malloc(vec->n * sizeof(CUKR_DOUBLE))) == NULL)
	 		     CUKR_FATALX("Error allocating %ld bytes of memory for vector %s", 
	 			     	     (unsigned long)(vec->n * sizeof(CUKR_DOUBLE)), vec->name);
			
			/* Set all to zero */
			memset(vec->ph_dval, 0, vec->n * sizeof(CUKR_DOUBLE));
			
			/* Update the counter */
			cukr_db.allocated_cpu += vec->n * sizeof(CUKR_DOUBLE);
			
			LOGX(DEBUG, "Allocated %ld bytes of memory (%d elements) "\
					    "for vector %s (in double) [addr = %p]",  (unsigned long)(vec->n * sizeof(CUKR_DOUBLE)), 
					    vec->n, vec->name, vec->ph_dval);
					    
		    /* Return, since there is nothing
		     * more to do */
		    return;			     		
		}
		/* Already allocated */
		else return;
	}
			 
		
	/* Alloc memory for single and 
	 * head part for qdouble */ 	 
	if (type == CUKR_TYPE_SINGLE || type == CUKR_TYPE_QDOUBLE) { 
		if (vec->ph_sval == NULL)  {									
			if ((vec->ph_sval = (CUKR_SINGLE*)malloc(vec->n * sizeof(CUKR_SINGLE))) == NULL) {
	 			 CUKR_FATALX("Error allocating %ld bytes of memory for vector %s", 
	 				          (unsigned long)(vec->n * sizeof(CUKR_SINGLE)), vec->name);
	 		}
	 		
	 		/* Set all to zero */
			memset(vec->ph_sval, 0, vec->n * sizeof(CUKR_SINGLE));
			
			/* Update the counter */
			cukr_db.allocated_cpu += vec->n * sizeof(CUKR_SINGLE);
			
	 		
	 		/* Log and return */
	 		if (type == CUKR_TYPE_SINGLE)
	 		{
	 			LOGX(DEBUG, "Allocated %ld bytes of memory (%d elements) "\
						    "for vector %s (in single) [addr = %p]",  
						    (unsigned long)(vec->n * sizeof(CUKR_SINGLE)), 
						    vec->n, vec->name, vec->ph_sval);
				return;		    
	 		}
			else
			{
				LOGX(DEBUG, "Allocated %ld bytes of memory (%d elements) "\
						    "for vector %s (in qdouble, head) [addr = %p]",  
						    (unsigned long)(vec->n * sizeof(CUKR_SINGLE)), 
						    vec->n, vec->name, vec->ph_sval);
						    
				/* For qdouble 
				 * do not return yet, might need to
				 * allocate qdouble tail */						    
			}				
		}
		/* Already allocated */	
		else 
		    if (type == CUKR_TYPE_SINGLE)
			  return;		
	}
		
						    
	/* Alloc memory for tail part 
	 * for qdouble */
	if (type == CUKR_TYPE_QDOUBLE) { 
		if (vec->ph_qvalt == NULL) {
			if ((vec->ph_qvalt = (CUKR_SINGLE*)malloc(vec->n * sizeof(CUKR_SINGLE))) == NULL)
	 		     CUKR_FATALX("Error allocating %ld bytes of memory for vector %s", 
	 			     	     (unsigned long)(vec->n * sizeof(CUKR_SINGLE)), vec->name);
			
			/* Set all to zero */
			memset(vec->ph_qvalt, 0, vec->n * sizeof(CUKR_SINGLE));
			
			/* Update the counter */
			cukr_db.allocated_cpu += vec->n * sizeof(CUKR_SINGLE);
			
			LOGX(DEBUG, "Allocated %ld bytes of memory (%d elements) "\
					    "for vector %s (in qdouble, tail) [addr = %p]",  
					    (unsigned long)(vec->n * sizeof(CUKR_SINGLE)), 
					    vec->n, vec->name, vec->ph_qvalt);
					    
			/* Return, nothing more to do */
			return;		    				    
		}
		/* Already allocated */
		else return; 
	}
	
	/* Alloc memory for int */
	if (type == CUKR_TYPE_INT) { 
		if (vec->ph_ival == NULL) {
			if ((vec->ph_ival = (CUKR_INT*)malloc(vec->n * sizeof(CUKR_INT))) == NULL)
	 			 CUKR_FATALX("Error allocating %ld bytes of memory for vector %s",
	 			              vec->n * sizeof(CUKR_INT), vec->name); 
	 		
	 		/* Set all to zero */
			memset(vec->ph_ival, 0, vec->n * sizeof(CUKR_INT));
			
			/* Update the counter */
			cukr_db.allocated_cpu += vec->n * sizeof(CUKR_INT);
	 		
	 		LOGX(DEBUG, "Allocated %ld bytes of memory (%d elements) "\
					    "for vector %s (in int) [addr = %p]",  
					    (unsigned long)(vec->n * sizeof(CUKR_INT)), 
					    vec->n, vec->name, vec->ph_ival);
			
			/* Return, nothing more to do */
			return;		    
		} 
		/* Already allocated */
		else return;
	}
	
	/* If still here, error */
	CUKR_FATALX("Cannot alloc memory for vector %s "\
	            "on host because of invalid type (%d)",
	            vec->name, type);
} 

// ------------------------------------------------------------------- //

/**
 * \brief Free memory allocated to the data 
 		  member of a vector corresponding to the
 *        desired precision, on the CPU. The 
 *        function checks if the pointer 
 *        corresponding to that particular location
 *        and precision is NULL of not. If not NULL,
 *        on malloc will be performed. Therefore, this
 *        function can be called "just in case" to 
 *        make sure memory allocation where it 
 *        should be.
 * \param vec Pointer to the vector
 * \param type Vector type {double,qdouble,single,int}
 */
void CukrVecspMallocGPU(CUKR_VECTOR_SP vec, int type) 
#ifndef CUKR_USE_GPU
{
	CUKR_FATAL("GPU support was not compiled in the library "
			   "(CukrVecMallocGPU cannot be executed) ");;
}
#else
{
	#ifdef CUKR_USE_CUDA
	// printf("tries to malloc on gpu... CUDA!\n");
	cublasStatus status;  
	/* Alloc memory for double */
	if (type == CUKR_TYPE_DOUBLE) { 
		if (vec->pd_dval == NULL)
		{
			status = cublasAlloc(vec->n, sizeof(CUKR_DOUBLE), (void**)(&vec->pd_dval));
			CukrHandleCudaMemErrors(vec, status);
			LOGX(DEBUG, "Allocated %ld bytes of memory for vector %s (GPU, double) [addr = %p]", 
	 			     	 (unsigned long)(vec->n * sizeof(CUKR_DOUBLE)), 
	 			     	 vec->name, vec->pd_dval);
	 		
	 		/* Nothing more to do */
	 		return;	     	  			     	 	
		}
		/* Already allocated */
		else return;
	}
		
	/* Alloc memory for single or head
	 * part for qdouble*/
	if (type == CUKR_TYPE_SINGLE || type == CUKR_TYPE_QDOUBLE){ 
		if (vec->pd_sval == NULL) 
		{
			status = cublasAlloc(vec->n, sizeof(CUKR_SINGLE), (void**)(&vec->pd_sval));
			CukrHandleCudaMemErrors(vec, status);
			
			/* Update the counter */
			cukr_db.allocated_gpu += vec->n * sizeof(CUKR_SINGLE);
			
			/* Log */
			if (type == CUKR_TYPE_SINGLE)
			{
				LOGX(DEBUG, "Allocated %ld bytes of memory for vector %s (GPU, single) [addr = %p]", 
	 			     		  (unsigned long)(vec->n * sizeof(CUKR_SINGLE)), 
	 			     		  vec->name, vec->pd_sval);
	 			     		  
	 			/* Nothing more to do */
	 			return;     		  
			}	
	 		else
	 		{
	 			LOGX(DEBUG, "Allocated %ld bytes of memory for vector %s (GPU, qdouble, head) [addr = %p]", 
	 			     		  (unsigned long)(vec->n * sizeof(CUKR_SINGLE)), 
	 			     		  vec->name, vec->pd_sval);
	 			     		  
	 			/* Cannot return, since it might be 
	 		 	 * qdouble which needs the tail as well */     		  
	 		}	 		
		}
 		/* Already allocated */
		else 
		    if (type == CUKR_TYPE_SINGLE)
			  return; 			
	}
		
	/* Alloc memory for qdouble,
	 * tail part */
	if (type == CUKR_TYPE_QDOUBLE) { 
		if (vec->pd_qvalt == NULL) 
		{
				status = cublasAlloc(vec->n, sizeof(CUKR_SINGLE), (void**)(&vec->pd_qvalt));
				CukrHandleCudaMemErrors(vec, status);
				
				/* Update the counter */
				cukr_db.allocated_gpu += vec->n * sizeof(CUKR_SINGLE);
				
				
				LOGX(DEBUG, "Allocated %ld bytes of memory for vector %s (GPU, qdouble, tail) [addr = %p]", 
	 				     	 (unsigned long)(vec->n * sizeof(CUKR_SINGLE)), 
	 				     	  vec->name, vec->pd_qvalt);
				
				/* Nothing more to to */
				return;		 				     	  
	 				     	  	
		}
		/* Already allocated */			
		else return;
	}
		
		
		/* Alloc memory for int */
		if (type == CUKR_TYPE_INT) {
			if (vec->pd_ival == NULL) {
				status = cublasAlloc(vec->n, sizeof(CUKR_INT), (void**)(&vec->pd_ival));
				CukrHandleCudaMemErrors(vec, status);
				
				/* Update the counter */
				cukr_db.allocated_gpu += vec->n * sizeof(CUKR_INT);
								
				LOGX(DEBUG, "Allocated %ld bytes of memory for vector %s (GPU, int) [addr = %p]", 
		 			     	  (unsigned long)(vec->n * sizeof(CUKR_INT)), 
		 			     	  vec->name, vec->pd_ival);
		 		
		 		/* Nothing more to do */
				return;	     	  	
			}
			/* Already allocated */			
			else return;
		}
	
	/* If still here, error */	
	CUKR_FATALX("Cannot alloc memory for vector %s "\
		        "on GPU because of invalid type (%d)",
		        vec->name, type);

	#endif /* CUKR_USE_CUDA */

	#ifdef CUKR_USE_OPENCL
	
	//printf("tries to malloc on gpu...OPENCL!\n");
	
	//cublasStatus status;
	
	/* Alloc memory for double */
	if (type == CUKR_TYPE_DOUBLE) { 
		if (vec->pcl_dval == NULL)
		{			
			/* Set up arrays in device memory for our calculation */
			vec->pcl_dval = (cl_mem*)malloc(sizeof(cl_mem));
			*(vec->pcl_dval) = clCreateBuffer(ComputeContext, CL_MEM_READ_ONLY|CL_MEM_ALLOC_HOST_PTR,  (vec->n) * sizeof(cl_double), NULL, &err);
			if (err != CL_SUCCESS)
			{
				printf("Error: Failed to allocate device memory pcl_dval!\n");
				exit(1);
			}
			//CukrHandleCudaMemErrors(vec, status);
			LOGX(DEBUG, "Allocated %ld bytes of memory for vector %s (GPU, double) [cl_mem addr = %p]", 
				 (unsigned long)(vec->n * sizeof(CUKR_DOUBLE)), 
				 vec->name, vec->pcl_dval);
	 		
	 		/* Nothing more to do */
	 		return;	     	  			     	 	
		}
		/* Already allocated */
		else return;
	}
	
	/* Alloc memory for single or head
	 * part for qdouble*/
	if (type == CUKR_TYPE_SINGLE || type == CUKR_TYPE_QDOUBLE){ 
		if (vec->pcl_sval == NULL) 
		{
			/* Create buffer on device */
			vec->pcl_sval = (cl_mem*)malloc(sizeof(cl_mem));
			*(vec->pcl_sval) = clCreateBuffer(ComputeContext, CL_MEM_READ_ONLY|CL_MEM_ALLOC_HOST_PTR,  vec->n * sizeof(cl_float), NULL, &err);
			if (err != CL_SUCCESS)
			{
				printf("Error: Failed to allocate device memory pcl_sval!\n");
				exit(1);
			}
			//CukrHandleCudaMemErrors(vec, status);
			
			/* Update the counter */
			cukr_db.allocated_gpu += vec->n * sizeof(CUKR_SINGLE);
			
			// Log
			
			if (type == CUKR_TYPE_SINGLE)
			{
				LOGX(DEBUG, "Allocated %ld bytes of memory for vector %s (GPU, single) [cl_mem addr = %p]", 
					 (unsigned long)(vec->n * sizeof(CUKR_SINGLE)), 
					 vec->name, vec->pcl_sval);
				
	 			// Nothing more to do
	 			return;     		  
			}	
	 		else
	 		{
	 			LOGX(DEBUG, "Allocated %ld bytes of memory for vector %s (GPU, qdouble, head) [cl_mem addr = %p]", 
					 (unsigned long)(vec->n * sizeof(CUKR_SINGLE)), 
					 vec->name, vec->pcl_sval);
				
				// Cannot return, since it might be 
				// qdouble which needs the tail as well     		  
	 		}	 	
		}
 		/* Already allocated */
		else 
		    if (type == CUKR_TYPE_SINGLE)
				return; 			
	}
	
	/* Alloc memory for qdouble,
	 * tail part */
	if (type == CUKR_TYPE_QDOUBLE) { 
		if (vec->pcl_qvalt == NULL) 
		{
			//CukrHandleCudaMemErrors(vec, status);
			
			/* Set up arrays in device memory for our calculation */
			vec->pcl_qvalt = (cl_mem*)malloc(sizeof(cl_mem));
			*(vec->pcl_qvalt) = clCreateBuffer(ComputeContext, CL_MEM_READ_ONLY|CL_MEM_ALLOC_HOST_PTR, vec->n * sizeof(cl_float), NULL, &err);
			if (err != CL_SUCCESS)
			{
				printf("Error: Failed to allocate device memory pcl_qvalt!\n");
				exit(1);
			}
			
			/* Update the counter */
			cukr_db.allocated_gpu += vec->n * sizeof(CUKR_SINGLE);
			
			
			LOGX(DEBUG, "Allocated %ld bytes of memory for vector %s (GPU, qdouble, tail) [cl_mem addr = %p]", 
				 (unsigned long)(vec->n * sizeof(CUKR_SINGLE)), 
				 vec->name, vec->pcl_qvalt);
			
			/* Nothing more to to */
			return;		 				     	  
			
		}
		/* Already allocated */			
		else return;
	}
	
	
	/* Alloc memory for int */
	if (type == CUKR_TYPE_INT) {
		if (vec->pcl_ival == NULL) {
			
			//CukrHandleCudaMemErrors(vec, status);
			
			/* Set up arrays in device memory for our calculation */
			vec->pcl_ival = (cl_mem*)malloc(sizeof(cl_mem));
			*(vec->pcl_ival) = clCreateBuffer(ComputeContext, CL_MEM_READ_ONLY|CL_MEM_ALLOC_HOST_PTR, vec->n * sizeof(cl_int), NULL, &err);
			if (err != CL_SUCCESS)
			{
				printf("Error: Failed to allocate device memory pcl_ival!\n");
				exit(1);
			}
			
			/* Update the counter */
			cukr_db.allocated_gpu += vec->n * sizeof(CUKR_INT);
			
			LOGX(DEBUG, "Allocated %ld bytes of memory for vector %s (GPU, int) [cl_mem addr = %p]", 
				 (unsigned long)(vec->n * sizeof(CUKR_INT)), 
				 vec->name, vec->pcl_ival);
			
			/* Nothing more to do */
			return;	     	  	
		}
		/* Already allocated */			
		else return;
	}
	
	/* If still here, error */	
	CUKR_FATALX("Cannot alloc memory for vector %s "\
		        "on GPU because of invalid type (%d)",
		        vec->name, type);
	#endif /* CUKR_USE_OPENCL */
}
#endif /* CUKR_USE_GPU */



// ------------------------------------------------------------------- // 

/**
 * \brief Free memory allocated to the data 
 		  member of a vector corresponding to the
 *        desired precision, on the GPU
 * \param vec Pointer to the vector
 * \param type Vector type {double,qdouble,single,int}
 */ 
void CukrVecspFree(CUKR_VECTOR_SP vec, int loc, int type)
{	
	/* If the location is on the CPU, 
	 * free for GPU */
	if (loc == CUKR_LOCATION_CPU) 
		CukrVecspFreeCPU(vec, type);
		
	/* If not, free for the GPU */
	else if (loc == CUKR_LOCATION_GPU) 
		CukrVecspFreeGPU(vec, type);
	
	/* If neither, error */
	else {
		/* Check for valid location */
		if (!CukrIsValidLocationId(loc))
			CUKR_FATALX("Cannot free memory for vector %s " \
			            "because of invalid location: %d",
			            vec->name, loc);
	}	
}

// ------------------------------------------------------------------- //

/**
 * \brief Free memory allocated to the data 
 		  member of a vector corresponding to the
 *        desired precision, on the CPU
 * \param vec Pointer to the vector
 * \param type Vector type {float, double, int}
 */
void CukrVecspFreeCPU(CUKR_VECTOR_SP vec, int type) 
{

	/* Free memory for double */
	if (type == CUKR_TYPE_DOUBLE) {
		if (vec->ph_dval != NULL) {
			free (vec->ph_dval);
			vec->ph_dval = NULL;
			LOGX(DEBUG, "Freed memory for vector %s (host,double)", vec->name);
			
			/* Nothing more to do */
			return;
		}
		/* Not allocated */
		else return;
	}
	
	/* Free memory for single or head
	 * part from qdouble */
	if ( (type == CUKR_TYPE_SINGLE || type == CUKR_TYPE_QDOUBLE)) {
		if (vec->ph_sval != NULL) {
			free (vec->ph_sval);
			vec->ph_sval = NULL;
			
			/* Log */
			if (type == CUKR_TYPE_SINGLE) {
				LOGX(DEBUG, "Freed memory for vector %s (host,single)", vec->name);
				
				/* Nothing else to do */
				return;
			}
			else {
				LOGX(DEBUG, "Freed memory for vector %s (host,qdouble,head)", vec->name);
				
				/* Cannot exit yet, if qdouble the tail
			 	 * must be freed as well */
			}
		}				
		/* Not allocated */
		else
			if (type == CUKR_TYPE_SINGLE)
				return;				 	
	}
	 
	/* Free memory for the tail part
	 * of qdouble */
	if (type == CUKR_TYPE_QDOUBLE) {
		if (vec->ph_qvalt != NULL) {
			free (vec->ph_qvalt);
			vec->ph_qvalt = NULL;
			LOGX(DEBUG, "Freed memory for vector %s (host,qdouble,tail)", vec->name);
			
			/* All done, can exit */
			return;
		}
		/* Not allocated */
		else return; 
	}
	
	/* Free memory for int */
	if (type == CUKR_TYPE_INT) { 
		if (vec->ph_ival != NULL){
			free (vec->ph_ival);
			vec->ph_ival = NULL;
			LOGX(DEBUG, "Freed memory for vector %s (host,int)", vec->name);
			
			/* Nothing more to do */
			return;
		}
		/* Not allocated */
		else return;
	}
	
	/* If still here, error */
	CUKR_FATALX("Cannot free memory for vector %s "\
		        "on GPU because of invalid type (%d)",
		        vec->name, type);
		
} 

// ------------------------------------------------------------------- //

/**
 * \brief Free memory allocated to the data 
 		  member of a vector corresponding to the
 *        desired precision, on the GPU
 * \param vec Pointer to the vector
 * \param type Vector type {float, double, int}
 */
void CukrVecspFreeGPU(CUKR_VECTOR_SP vec, int type) 
#ifndef CUKR_USE_GPU
{
	CUKR_FATAL("GPU support was not compiled in the library " \
			   "(CukrVecFreeGPU cannot be executed) ");;
}
#else
{
	//printf("trying to free memory\n");
	/* Free memory for double */
	if (type == CUKR_TYPE_DOUBLE) {
		#ifdef CUKR_USE_CUDA
		if (vec->pd_dval != NULL) {
			cublasFree(vec->pd_dval);
			vec->pd_dval = NULL;
		#endif
		#ifdef CUKR_USE_OPENCL
		if (vec->pcl_dval != NULL) {
			//free(vec->pd_dval);
			err = clReleaseMemObject(*(vec->pcl_dval));
			if (err != CL_SUCCESS) {
				printf("Problems freeing cl_mem double.\n");
			}
			free(vec->pcl_dval);
			vec->pcl_dval = NULL;
		#endif
			//vec->pd_dval = NULL;
			LOGX(DEBUG, "Freed memory for vector %s (GPU,double)", vec->name);
		
			/* Nothing more to do */
			return;
		}
		/* Not allocated */
		else return;
	}
	
	/* Free memory for single for 
	 * for qdouble head */
	if ((type == CUKR_TYPE_SINGLE || type == CUKR_TYPE_QDOUBLE)) {
		#ifdef CUKR_USE_CUDA
		if (vec->pd_sval != NULL) {
			cublasFree(vec->pd_sval);
			vec->pd_sval = NULL;
		#endif
		#ifdef CUKR_USE_OPENCL
		if (vec->pcl_sval != NULL) {
			//free(vec->pd_sval);
			err = clReleaseMemObject(*(vec->pcl_sval));
			if (err != CL_SUCCESS) {
				printf("Problems freeing cl_mem single.\n");
			}
			free(vec->pcl_sval);
			vec->pcl_sval = NULL;
		#endif
			//vec->pd_sval = NULL;
			
			/* Check if pd_qvalt is null to know if 
			 * the free request is for a single or for
			 * a qdouble. If null, is for single. */
			#ifdef CUKR_USE_OPENCL
			if (type == CUKR_TYPE_SINGLE || vec->pcl_qvalt == NULL) {
				LOGX(DEBUG, "Freed memory for vector %s (GPU,single)", vec->name);
				
				/* Nothing else to do */
				return;
			}
			else {
				LOGX(DEBUG, "Freed memory for vector %s (GPU,qdouble,head)", vec->name);
				
				/* Not finished yet, if in qdouble we
			 	 * might have to free the tail */					
			}
			#else
			if (type == CUKR_TYPE_SINGLE || vec->pd_qvalt == NULL) {
				LOGX(DEBUG, "Freed memory for vector %s (GPU,single)", vec->name);
				
				/* Nothing else to do */
				return;
			}
			else {
				LOGX(DEBUG, "Freed memory for vector %s (GPU,qdouble,head)", vec->name);
				
				/* Not finished yet, if in qdouble we
			 	 * might have to free the tail */					
			}
			#endif // CUKR_USE_OPENCL
		}
		/* Not allocated */
		else
			if (type == CUKR_TYPE_SINGLE)
				return;					
	}
	
	/* Free memory for the tail part 
	 * of qdouble */
	if (type == CUKR_TYPE_QDOUBLE) {
		#ifdef CUKR_USE_CUDA
		if (vec->pd_qvalt != NULL) {
			cublasFree(vec->pd_qvalt);
			vec->pd_qvalt = NULL;
		#endif
		#ifdef CUKR_USE_OPENCL
		if (vec->pcl_qvalt != NULL) {
			//free(vec->pd_qvalt);
			err = clReleaseMemObject(*(vec->pcl_qvalt));
			if (err != CL_SUCCESS) {
				printf("Problems freeing cl_mem qsingle tail.\n");
			}
			free(vec->pcl_qvalt);
			vec->pcl_qvalt = NULL;
		#endif
			//vec->pd_qvalt = NULL;
			LOGX(DEBUG, "Freed memory for vector %s (GPU,qdouble,tail)", vec->name);
			
			/* Nothing more to do */
			return;
		}
		/* Not allocated */
		else return;
		
	}
	
	/* Free memory for int */
	else if (type == CUKR_TYPE_INT) {
		#ifdef CUKR_USE_CUDA
		if (vec->pd_ival != NULL) {
			cublasFree(vec->pd_ival);
			vec->pd_ival = NULL;
		#endif
		#ifdef CUKR_USE_OPENCL
		if (vec->pcl_ival != NULL) {
			//free(vec->pd_ival);
			err = clReleaseMemObject(*(vec->pcl_ival));
			if (err != CL_SUCCESS) {
				printf("Problems freeing cl_mem int.\n");
			}
			free(vec->pcl_ival);
			vec->pcl_ival = NULL;
		#endif
			//vec->pd_ival = NULL;
			LOGX(DEBUG, "Freed memory for vector %s (GPU,int)", vec->name);
			
			/* Nothing more to do */
			return;
		}
		/* Not allocated */
		else return;
	}
	
	/* If got up to here, error */
	CUKR_FATALX("Cannot free memory for vector %s "\
	            "on GPU because of invalid type (%d)",
	            vec->name, type);
	
	
} 
#endif /* CUKR_USE_GPU */





				/*****************************************/
				//              TRANSFER                 // 
				/*****************************************/ 





/** 
 * \brief Changes the location of a vector
 *        by copying the data from one location
 *        to another. The status of the vector will
 *        be updated to reflect the new location. 
 * \param vec Pointer to the vector 
 * \param dest_loc Destination vector location (CPU,GPU)
 * \param dest_type Destination vector type 
 *        {double,qdouble,single,int}
 * \param update_attrs If nonzero, the location and 
 *        the precision of the vector will be updated 
 *        after the transfer. A false value should only
 *        be used for operations like print, where one
 *        would want to print the elements from a 
 *        different locatio (which requires transfering 
 *        the data to the CPU) but wishes to see the 
 *        original status of the vector, not the updated
 *        one. 
 */
void CukrVecspTransfer(CUKR_VECTOR_SP vec, int dest_loc, int dest_type)
{
	/* Build the destination status from
	 * the type and location and check if 
	 * data is not already there */
	int dest_status = CukrLocIdAndTypeIdToStatusId(dest_loc, dest_type);
	if (vec->status & dest_status) {
		LOGX(DEBUG, "No need to move vector %s since " \
		            "valid data is already there", vec->name);
		return;                
	}
	
	/* Get the highest precisions available
	 * on CPU and GPU */
	int cpu_highest = GetHighestPrecOnCpuFromStatusId(vec->status);
	int gpu_highest = GetHighestPrecOnGpuFromStatusId(vec->status); 
	//printf("%d, %d\n", cpu_highest, gpu_highest);
	/* If the destination is on the CPU */
	if (dest_loc == CUKR_LOCATION_CPU)
	{
		/* If the CPU has enough precision
		 * use from there */		  
		 if (cpu_highest >= dest_type) {
		 	CukrVecspCopyDataCPU2CPU(vec, dest_type);
		 	return;
		 }
		 
		 /* If the GPU has enough precision
		  * use from there */
 	 	 if (gpu_highest >= dest_type) { 	 	 	
		 	CukrVecspCopyDataGPU2CPU(vec, dest_type); 
		 	return;
 	 	 }
 	 	 
 	 	 /* If none have, then if the cpu has more */
 	 	 if (cpu_highest >= gpu_highest) {
 	 	 	CukrVecspCopyDataCPU2CPU(vec, dest_type);
 	 	 	return;
 	 	 }
		
		 /* If not, get from the GPU */		
		 CukrVecspCopyDataGPU2CPU(vec, dest_type);	
		 return;
	}
	
	/* If the destination is on the GPU */
	else if (dest_loc == CUKR_LOCATION_GPU)
	{
		/* If the GPU has enough precision
		 * use from there */		  
		 if (gpu_highest >= dest_type) {
		 	CukrVecspCopyDataGPU2GPU(vec, dest_type);
		 	return;
		 }
		 
		 /* If not, but the CPU has, use from there 
		  * instead */
		 if (cpu_highest >= dest_type) {
		 	CukrVecspCopyDataCPU2GPU(vec, dest_type);
		 	return;
		 }
		  
		 /* If none have, then if the gpu has more */
		 if (gpu_highest >= cpu_highest) {
		 	CukrVecspCopyDataGPU2GPU(vec, dest_type);
		 	return;
		 }
		 
		 /* If not, from the cpu */
		 CukrVecspCopyDataCPU2GPU(vec, dest_type);	
		 return;	 
	}
	
	/* If none, error */
	else 
	{
		CUKR_FATALX("Could not transfer vector %s: " \
		            "invalid destination location: %d", 
		            vec->name, dest_loc);
	}
}



				/*****************************************/
				//               GET/SET                 // 
				/*****************************************/ 



/**
 * \brief Returns the element found at the
 *        position <index> in the vector vec. 
 * \param vec The vector
 * \param index The index where to put the value
 * \param type An existing data type in which to 
 * 		  store the value or default to read from
 *        the highest available one
 * \return val The value of the element
 */
CUKR_DOUBLE CukrVecspGetElement(CUKR_VECTOR_SP vec, int index, int type)
{	
	printf("gets here: CukrVecspGetElement()\n");
	
	#ifdef CUKR_GPU
		CUKR_DOUBLE dval[1];
		CUKR_SINGLE sval[2];
		CUKR_INT    ival[1];
	#endif

	if (index < 0 && index >= vec->n)
		CUKR_FATALX("Invalid element index (%d) accessed from "\
		            "vector %s", index, vec->name);	

	/* Try integer on CPU */
	if (vec->status & CUKR_STATE_CPU_INT) {
		if (type == CUKR_TYPE_QDOUBLE || type == CUKR_DEFAULT) {
			LOGX(DEBUG, "Read %s[%d] -> %d in int, on CPU", 
		            vec->name, index, vec->ph_ival[index]);
		    return vec->ph_ival[index];        			
		}			
	}
	
	/* Try integer on GPU */
	#ifdef CUKR_CUDA
		if (vec->status & CUKR_STATE_GPU_INT) {
			
			printf("CukrVecspGetElement: get integer on gpu\n");
			
			status = cublasGetVector(1, sizeof(CUKR_INT), &(vec->pd_ival[index]),
	                         	     1, ival, 1);
	    	CukrHandleCudaMemErrors(vec, status);
	    	LOGX(DEBUG, "Read %s[%d] -> %g in int, on GPU", 
		            vec->name, index, ival[0]);
	    	return ival[0];
		}
	#endif		
		
	/* Try double on CPU */
	if (vec->status & CUKR_STATE_CPU_DOUBLE) {
		LOGX(DEBUG, "Read %s[%d] -> %g in double, on CPU", 
		            vec->name, index, vec->ph_dval[index]);
		return vec->ph_dval[index];	
	}
		
	/* Try double on GPU */
 	#ifdef CUKR_CUDA
		if (vec->status & CUKR_STATE_GPU_DOUBLE) {
			
			printf("CukrVecspGetElement: get double on gpu\n");
			
			status = cublasGetVector(1, sizeof(CUKR_DOUBLE), &(vec->pd_dval[index]),
	                         	     1, dval, 1);
	    	CukrHandleCudaMemErrors(vec, status);
	    	return dval[0];
		}
	#endif
			
	
	/* Try qdouble on CPU */
	if (vec->status & CUKR_STATE_CPU_QDOUBLE) {
		LOGX(DEBUG, "Read %s[%d] -> (%g,%g)=%g in qdouble, on CPU", 
		            vec->name, index, vec->ph_sval[index], vec->ph_qvalt[index],
		            vec->ph_sval[index] + vec->ph_qvalt[index]);
		return vec->ph_sval[index] + vec->ph_qvalt[index];
	}	
		
	/* Try qdouble on GPU */
	#ifdef CUKR_CUDA
		if (vec->status & CUKR_STATE_GPU_QDOUBLE) {
			
			printf("CukrVecspGetElement: get qdouble on gpu\n");
			
			status = cublasGetVector(1, sizeof(CUKR_SINGLE), &(vec->pd_sval[index]),
	                         	     1, sval, 1);
	    	CukrHandleCudaMemErrors(vec, status);
	    	status = cublasGetVector(1, sizeof(CUKR_SINGLE), &(vec->pd_sval[index]),
	                         	     1, &(sval[1]), 1);
	    	CukrHandleCudaMemErrors(vec, status);
	    	LOGX(DEBUG, "Read %s[%d] -> (%g,%g)=%g in qdouble, on CPU", 
		            vec->name, index, sval[0], sval[1], sval[0] + sval[1]);
	    	return sval[0] + sval[1];
		}
	#endif	
	
	/* Try single on CPU */
	if (vec->status & CUKR_STATE_CPU_SINGLE) {
		LOGX(DEBUG, "Read %s[%d] -> %g in single, on CPU", 
		            vec->name, index, vec->ph_sval[index]);
		return vec->ph_sval[index];		
	}
		
	/* Try single on GPU */
	#ifdef CUKR_CUDA
		if (vec->status & CUKR_STATE_GPU_SINGLE) {
			
			printf("CukrVecspGetElement: get qdouble on gpu\n");
			
			status = cublasGetVector(1, sizeof(CUKR_SINGLE), &(vec->pd_sval[index]),
	                         	     1, sval, 1);
	    	CukrHandleCudaMemErrors(vec, status);
	    	LOGX(DEBUG, "Read %s[%d] -> %g in single, on GPU", 
		            vec->name, index, sval[0]);
	    	return sval[0];
		}
	#endif	
	
	/* If none, error */
	CUKR_FATALX("Invalid state for vector %s: %d",
	            vec->name, vec->status);	

	return 0;
}

// ------------------------------------------------------------------- //

/**
 * \brief Sets the element with index <index>
 *        to the value specified by <val> 
 *        Currently, setting can only
 *        be done on the CPU, so data will be 
 *        set on the CPU and the status set to
 *        there. 
 * \param vec The vector
 * \param index The index where to put the value
 * \param val The value the vector 
 *        elements are set to
 * \param type An existing data type in which to 
 * 		  store the value or default for all 
 *        available precisions
 */
void CukrVecspSetElement(CUKR_VECTOR_SP vec, int index, CUKR_DOUBLE val, int type)
{
	int success = CUKR_FALSE;
	
	if (index < 0 && index >= vec->n)
		CUKR_FATALX("Invalid element index (%d) accessed from "\
		            "vector %s", index, vec->name);	
	
	/* Try integer */
	if ((vec->status & CUKR_STATE_CPU_INT)) {
		vec->ph_ival[index] = (CUKR_INT)val;
		vec->status = CUKR_STATE_CPU_INT;	
		LOGX(DEBUG, "Set %s[%d] <- %g in integer, on CPU", 
		            vec->name, index, val);
		return;
	}
	
	/* Try double */
	if (vec->status & CUKR_STATE_CPU_DOUBLE)  
		if (type == CUKR_TYPE_DOUBLE || type == CUKR_DEFAULT) {
			vec->ph_dval[index] = val;
			success = CUKR_TRUE;
			LOGX(DEBUG, "Set %s[%d] <- %g in double, on CPU", 
		            vec->name, index, val);
		}
			
	/* Try qdouble */
	if (vec->status & CUKR_STATE_CPU_QDOUBLE || vec->status & CUKR_STATE_CPU_SINGLE){ 
		if (type == CUKR_TYPE_SINGLE || type == CUKR_DEFAULT) {
			vec->ph_sval [index] = (CUKR_SINGLE)val;		
			success = CUKR_TRUE;		
			LOGX(DEBUG, "Set %s[%d] <- %g in single, on CPU", 
		            vec->name, index, val);
		}
		
		if (type == CUKR_TYPE_QDOUBLE || type == CUKR_DEFAULT) {	
			vec->ph_qvalt[index] = (CUKR_SINGLE)(val - vec->ph_sval[index]);
			success = CUKR_TRUE;
			LOGX(DEBUG, "Set %s[%d] <- %g in qdouble, on CPU", 
		            vec->name, index, val);
		}
	}
		
	/* Invalidate all GPU data */
	vec->status != CUKR_STATE_GPU_DOUBLE;
	vec->status != CUKR_STATE_GPU_QDOUBLE;
	vec->status != CUKR_STATE_GPU_SINGLE;
		
	/* Error if no success */
	if (success == CUKR_FALSE)
		CUKR_FATALX("Could not add value %g at position %d "\
	    	        "in vector %s since no valid data was "\
	        	    "found on the CPU", val, index, vec->name); 	
}

// ------------------------------------------------------------------- //

/**
 * \todo  Should make a simple kernel that
 *        sets all the values of a vector to
 *        a specicific value, so that we can
 *        set directly on the GPU. 
 * \brief Sets all the elements
 *        of the vector to <val>. 
 *        Currently, setting can only
 *        be done on the CPU, so data will be 
 *        set on the CPU and the status set to
 *        there. 
 * \param vec The vector
 * \param val The value the vector 
 *        elements are set to
 * \param type An existing data type in which to 
 * 		  store the values or default to match
 *        the solver precision
 */
void CukrVecspSetFromVal(CUKR_VECTOR_SP vec, CUKR_DOUBLE val, int type)
{
	int i, vec_type;
	 	
	/*
	 * Find the type to set the data to
	 */
	
	if (type == CUKR_DEFAULT)
		vec_type = CukrTypeStrToId(CukrDBGetS(CUKR_DB_SOLVER_PRECISION));
	else
	 	vec_type = type; 
	 
	/*
	 * Set data in that type
	 */
	 
	/* Check for allocated memory 
	 * and allocate if needed */ 
	CukrVecspMalloc(vec, CUKR_LOCATION_CPU, vec_type);
		
	/* For int */ 
	if (vec_type == CUKR_TYPE_INT) {          
				
		/* Set value */
		for (i = 0; i < vec->n ; i++) 
			vec->ph_ival[i] = (CUKR_INT)val;
		vec->status = CUKR_STATE_CPU_INT;
		LOGX(DEBUG, "Vector %s set to %d (in int)",
		             vec->name, (CUKR_INT)val);  
		
		/* Nothing more to do */
		return;                       
	} 
	           
	/* For double */	 
	if (vec_type == CUKR_TYPE_DOUBLE) {          
		
		/* Set the values */
		for (i = 0; i < vec->n ; i++) 
			vec->ph_dval[i] = val;
		vec->status = CUKR_STATE_CPU_DOUBLE;
		LOGX(DEBUG, "Vector %s set to %g (in double)",
		             vec->name, val);     
				
		/* Nothing more to do */
		return;                 
	} 
	
	/* For qdouble */	 
	if (vec_type == CUKR_TYPE_QDOUBLE) {
				          
		/* Set value */          
		for (i = 0; i < vec->n ; i++) {
			vec->ph_sval[i]  = (float)val;
			vec->ph_qvalt[i] = (float)(val - vec->ph_sval[i]);
		}
		vec->status = CUKR_STATE_CPU_QDOUBLE;
		LOGX(DEBUG, "Vector %s set to %g (in double)",
		             vec->name, val);  
		
		/* Nothing more to do */
		return;                       
	} 
	
	/* For single */
	if (vec_type == CUKR_TYPE_SINGLE) {          
		
		/* Malloc on CPU */
		CukrVecspMalloc(vec, CUKR_LOCATION_CPU, CUKR_TYPE_DOUBLE);
		
		/* Set value */
		for (i = 0; i < vec->n ; i++) 
			vec->ph_sval[i] = val;
		vec->status = CUKR_STATE_CPU_SINGLE;
		LOGX(DEBUG, "Vector %s set to %g (in single)",
		             vec->name, val);   
		
		/* Noting more to do */
		return;                      
	}  
	
	
	/* If here, error*/
	CUKR_FATALX("Error setting vector %s from value. "\
	            "invalid status (%d)", vec->name, vec->status);  
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets all elements of a 
 *        vector from a double precision vector
 *        <val>.
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecspSetFromVectorD(CUKR_VECTOR_SP vec, int index, int count, CUKR_DOUBLE* val, int loc)
{
	/* 
	 * Validation
	 */
	
	 if (index < 0 || index >= vec->n || index + count > vec->n)
		CUKR_FATALX("Invalid index/count combination when setting "\
		            "vector %s from vector of doubles", vec->name); 
	/*
	 * Set values
	 */ 
	
	/* If double on CPU */ 
	if (loc == CUKR_LOCATION_CPU || loc == CUKR_DEFAULT) {
		LOGX(DEBUG, "Copy vector -> %s.cpu.double",
		            vec->name);
		CukrVecspMallocCPU(vec, CUKR_TYPE_DOUBLE);
		generic_dcopy(count, val, &vec->ph_dval[index]);
		vec->status = CUKR_STATE_CPU_DOUBLE;
		return; 					        
	}
	
	/* If double on GPU */ 
	#ifdef CUKR_USE_GPU 
	if (loc == CUKR_LOCATION_GPU) {
		LOGX(DEBUG, "Copy vec -> %s.gpu.double",
		            vec->name);
		CukrVecspMallocGPU(vec, CUKR_TYPE_DOUBLE);
		#ifdef CUKR_USE_OPENCL
		printf("CukrVecspSetFromVectorD called\n");
		#else
		gpu_dcopy(count, val, &vec->pd_dval[index]);
		#endif
		vec->status = CUKR_STATE_GPU_DOUBLE;
		return; 					        
	}
	#endif       	
	
	/* If still here, error */
	CUKR_FATALX("Invalid location received when setting "\
	            "vector %s from vector of doubles: %d", 
	            vec->name, loc);  
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets all elements of a 
 *        vector from a quasi-double precision 
 *        vector <val>.
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecspSetFromVectorQ(CUKR_VECTOR_SP vec, int index, int count, CUKR_SINGLE* hval, CUKR_SINGLE* tval, int loc)
{

	/* 
	 * Validation
	 */
	
	 if (index < 0 || index >= vec->n || index + count > vec->n)
		CUKR_FATALX("Invalid index/count combination when setting "\
		            "vector %s from vector of quasi-doubles", vec->name); 
	/*
	 * Set values
	 */ 
	
	/* If qdouble on CPU */ 
	if (loc == CUKR_LOCATION_CPU || loc == CUKR_DEFAULT) {
		LOGX(DEBUG, "Copy vector -> %s.cpu.qdouble",
		            vec->name);
		CukrVecspMallocCPU(vec, CUKR_TYPE_QDOUBLE);
		generic_qcopy(count, hval, tval, &vec->ph_sval[index], &vec->ph_qvalt[index]);
		vec->status = CUKR_STATE_CPU_QDOUBLE;
		return; 					        
	}
	
	/* If qdouble on GPU */ 
	#ifdef CUKR_USE_GPU 
	if (loc == CUKR_LOCATION_GPU) {
		LOGX(DEBUG, "Copy vec -> %s.gpu.qdouble",
		            vec->name);
		CukrVecspMallocGPU(vec, CUKR_TYPE_QDOUBLE);
		#ifdef CUKR_USE_OPENCL
		printf("CukrVecspSetFromVectorQ called\n");
		#else
			gpu_qcopy(count, hval, tval, &vec->ph_sval[index], &vec->ph_qvalt[index]);
		#endif
		vec->status = CUKR_STATE_GPU_QDOUBLE;
		return; 					        
	}
	#endif 
	
	/* If still here, error */
	CUKR_FATALX("Invalid location received when setting "\
	            "vector %s from vector of quasi-doubles: %d", 
	            vec->name, loc);        	
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets all elements of a 
 *        vector from a single precision vector
 *        <val>.
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecspSetFromVectorS(CUKR_VECTOR_SP vec, int index, int count, CUKR_SINGLE* val, int loc)
{
	/* 
	 * Validation
	 */
	
	 if (index < 0 || index >= vec->n || index + count > vec->n)
		CUKR_FATALX("Invalid index/count combination when setting "\
		            "vector %s from vector of singles", vec->name); 
	/*
	 * Set values
	 */ 
	
	/* If single on CPU */ 
	if (loc == CUKR_LOCATION_CPU || loc == CUKR_DEFAULT) {
		LOGX(DEBUG, "Copy vector -> %s.cpu.single",
		            vec->name);
		CukrVecspMallocCPU(vec, CUKR_TYPE_SINGLE);
		generic_scopy(count, val, &vec->ph_sval[index]);
		vec->status = CUKR_STATE_CPU_SINGLE;
		return; 					        
	}
	
	/* If single on GPU */ 
	#ifdef CUKR_USE_GPU 
	if (loc == CUKR_LOCATION_GPU) {
		LOGX(DEBUG, "Copy vec -> %s.gpu.single",
		            vec->name);
		CukrVecspMallocGPU(vec, CUKR_TYPE_SINGLE);
		#ifdef CUKR_USE_OPENCL
		printf("CukrVecspSetFromVectorS called\n");
		#else
		gpu_scopy(count, val, &vec->pd_sval[index]);
		#endif
		vec->status = CUKR_STATE_GPU_SINGLE;
		return; 					        
	}
	#endif  
	
	/* If still here, error */
	CUKR_FATALX("Invalid location received when setting "\
	            "vector %s from vector of single: %d", 
	            vec->name, loc);       	
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets all elements of a 
 *        vector from a integer vector <val>. 
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecspSetFromVectorI(CUKR_VECTOR_SP vec, int index, int count, CUKR_INT* val, int loc)
{
	/* 
	 * Validation
	 */
	
	 if (index < 0 || index >= vec->n || index + count > vec->n)
		CUKR_FATALX("Invalid index/count combination when setting "\
		            "vector %s from vector of integers", vec->name); 
	/*
	 * Set values
	 */ 
	
	/* If int on CPU */ 
	if (loc == CUKR_LOCATION_CPU || loc == CUKR_DEFAULT) {
		LOGX(DEBUG, "Copy vector -> %s.cpu.int",
		            vec->name);
		CukrVecspMallocCPU(vec, CUKR_TYPE_INT);
		generic_scopy(count, (CUKR_SINGLE*)val, (CUKR_SINGLE*)(&vec->ph_ival[index]));
		vec->status = CUKR_STATE_CPU_INT;
		return; 					        
	}
	
	/* If int on GPU */ 
	#ifdef CUKR_USE_GPU 
	if (loc == CUKR_LOCATION_GPU) {
		LOGX(DEBUG, "Copy vec -> %s.gpu.int",
		            vec->name);
		CukrVecspMallocGPU(vec, CUKR_TYPE_INT);
		#ifdef CUKR_USE_OPENCL
		printf("CukrVecspSetFromVectorI called\n");
		#else
		gpu_scopy(count, (CUKR_SINGLE*)val, (CUKR_SINGLE*)(&vec->pd_ival[index]));
		#endif
		vec->status = CUKR_STATE_GPU_INT;
		return; 					        
	}
	#endif    
	
	/* If still here, error */
	CUKR_FATALX("Invalid location received when setting "\
	            "vector %s from vector of integers: %d", 
	            vec->name, loc);     	
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to a double precision vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecspGetToVectorD(CUKR_VECTOR_SP vec, int index, int count, CUKR_DOUBLE* val, int loc)
{
	/* Make sure the vector has data 
	 * in the right place */
	CukrVecspTransfer(vec, loc, CUKR_TYPE_DOUBLE); 
	
	/*
	 * Write to the destination vector
	 */
	
	/* If double on CPU */ 
	if (loc == CUKR_LOCATION_CPU) {
		LOGX(DEBUG, "Copy %s.cpu.double -> vector",
		            vec->name);
		generic_dcopy(count, vec->ph_dval, val);
		return; 					        
	}
	
	/* If double on GPU */ 
	#ifdef CUKR_USE_GPU 
	if (loc == CUKR_LOCATION_GPU) {
		LOGX(DEBUG, "Copy %s.gpu.double -> vector",
		            vec->name);
		#ifdef CUKR_USE_OPENCL
		printf("CukrVecspGetToVectorD called\n");
		#else
		gpu_dcopy(count, vec->pd_dval, val);
		#endif
		return; 					        
	}
	#endif       	
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to a quasi-double precision vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecspGetToVectorQ(CUKR_VECTOR_SP vec, int index, int count, CUKR_SINGLE* hval, CUKR_SINGLE* tval, int loc)
{
	/* Make sure the vector has data 
	 * in the right place */
	CukrVecspTransfer(vec, loc, CUKR_TYPE_QDOUBLE); 
	
	/*
	 * Write to the destination vector
	 */
	
	/* If double on CPU */ 
	if (loc == CUKR_LOCATION_CPU) {
		LOGX(DEBUG, "Copy %s.cpu.qdouble -> vector",
		            vec->name);
		generic_qcopy(count, vec->ph_sval, vec->ph_qvalt, hval, tval);
		return; 					        
	}
	
	/* If double on GPU */ 
	#ifdef CUKR_USE_GPU 
	if (loc == CUKR_LOCATION_GPU) {
		LOGX(DEBUG, "Copy %s.gpu.qdouble -> vector",
		            vec->name);
		#ifdef CUKR_USE_OPENCL
		printf("CukrVecspGetToVectorQ called\n");
		#else
		gpu_qcopy(count, vec->pd_sval, vec->pd_qvalt, hval, tval);
		#endif
		return; 					        
	}
	#endif       	
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to a single precision vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecspGetToVectorS(CUKR_VECTOR_SP vec, int index, int count, CUKR_SINGLE* val, int loc)
{
	/* Make sure the vector has data 
	 * in the right place */
	CukrVecspTransfer(vec, loc, CUKR_TYPE_SINGLE); 
	
	/*
	 * Write to the destination vector
	 */
	
	/* If double on CPU */ 
	if (loc == CUKR_LOCATION_CPU) {
		LOGX(DEBUG, "Copy %s.cpu.single -> vector",
		            vec->name);
		generic_scopy(count, vec->ph_sval, val);
		return; 					        
	}
	
	/* If double on GPU */ 
	#ifdef CUKR_USE_GPU 
	if (loc == CUKR_LOCATION_GPU) {
		LOGX(DEBUG, "Copy %s.gpu.single -> vector",
		            vec->name);
		#ifdef CUKR_USE_OPENCL
		printf("CukrVecspGetToVectorS called\n");
		#else
		gpu_scopy(count, vec->pd_sval, val);
		#endif
		return; 					        
	}
	#endif       	
}


// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to an integer vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecspGetToVectorI(CUKR_VECTOR_SP vec, int index, int count, CUKR_INT* val, int loc)
{
	/* Make sure the vector has data 
	 * in the right place */
	CukrVecspTransfer(vec, loc, CUKR_TYPE_INT); 
	
	/*
	 * Write to the destination vector
	 */
	
	/* If double on CPU */ 
	if (loc == CUKR_LOCATION_CPU) {
		LOGX(DEBUG, "Copy %s.cpu.int -> vector",
		            vec->name);
		generic_scopy(count, (CUKR_SINGLE*)vec->ph_sval, (CUKR_SINGLE*)val);
		return; 					        
	}
	
	/* If double on GPU */ 
	#ifdef CUKR_USE_GPU 
	if (loc == CUKR_LOCATION_GPU) {
		LOGX(DEBUG, "Copy %s.gpu.int -> vector",
		            vec->name);
		#ifdef CUKR_USE_OPENCL
		printf("CukrVecspGetToVectorI called\n");
		#else
		gpu_scopy(count, (CUKR_SINGLE*)vec->pd_sval, (CUKR_SINGLE*)val);
		#endif
		return; 					        
	}
	#endif       	
}

// ------------------------------------------------------------------- //

/**
 * \brief Sets all elements of the
 *        vector to random values, on the CPUs. 
 *        The rest of the precisions and 
 *        locations are invalidated.
 * \param vec The vector
 * \param min,max The bounds for the random number
 * \param seed The seed of the random number gen.
 * \param type Data type where to write the data
 *             or default to match the solver type
 */
void CukrVecspSetRandom(CUKR_VECTOR_SP vec, CUKR_DOUBLE min, CUKR_DOUBLE max, int seed, int type)
{
	int i;
	
	LOGX(DEBUG, "Filling vector %s with random values " \
	            "between %g and %g", vec->name, min, max);
	
	/* Find the destination data type */
	int dest_type;
	if (type == CUKR_DEFAULT)
		dest_type = CukrTypeStrToId(CukrDBGetS(CUKR_DB_SOLVER_PRECISION));
	else
	 	dest_type = type;
	
	/* Make sure the vector has data
	 * on the CPU in dest_type*/
	CukrVecspMallocCPU(vec, dest_type);
	
	/* Init random no. generator */
	srand(seed);   
	
	/*
	 *  Set the values 
	 */

	/* In double */
	if (dest_type == CUKR_TYPE_DOUBLE) {	
		for (i = 0 ; i < vec->n ; i ++) {
			CUKR_DOUBLE rnd = min + ((CUKR_DOUBLE)rand() / (CUKR_DOUBLE)RAND_MAX) * (max - min);
			vec->ph_dval[i] = rnd; 
		}
		vec->status = CUKR_STATE_CPU_DOUBLE;
		return;
	}
	
	/* In qdouble */
	if (dest_type == CUKR_TYPE_QDOUBLE) {	
		for (i = 0 ; i < vec->n ; i ++) {
			CUKR_DOUBLE rnd = min + ((CUKR_DOUBLE)rand() / (CUKR_DOUBLE)RAND_MAX) * (max - min);
			vec->ph_sval[i]  = (CUKR_SINGLE)rnd; 
			vec->ph_qvalt[i] = (CUKR_SINGLE)(rnd - vec->ph_sval[i]); 
		}
		vec->status = CUKR_STATE_CPU_QDOUBLE;
		return;
	}
	
	/* In single */
	if (dest_type == CUKR_TYPE_SINGLE) {	
		for (i = 0 ; i < vec->n ; i ++) {
			CUKR_DOUBLE rnd = min + ((CUKR_DOUBLE)rand() / (CUKR_DOUBLE)RAND_MAX) * (max - min);
			vec->ph_sval[i] = (CUKR_SINGLE)rnd; 
		}
		vec->status = CUKR_STATE_CPU_SINGLE;
		return;
	}
	
	/* In int */
	if (dest_type == CUKR_TYPE_SINGLE) {	
		for (i = 0 ; i < vec->n ; i ++) {
			CUKR_DOUBLE rnd = min + ((CUKR_DOUBLE)rand() / (CUKR_DOUBLE)RAND_MAX) * (max - min);
			vec->ph_ival[i] = (CUKR_INT)rnd; 
		}
		vec->status = CUKR_STATE_CPU_INT;
		return;
	}
	
	/* Else, error */
	CUKR_FATALX("Error setting vector %s to random: "\
	            "invalid type: %d", vec->name, dest_type);
}





				/*****************************************/
				//                 I/O                   // 
				/*****************************************/ 



/**
 * \brief Reads data into a previously
 *        created vector structure from 
 *        an opened file stream
 * \param f The open file stream
 * \param type The type the read data is
 *        supposed to be read as or default
 *        to match the solver precision
 */
void CukrVecspReadDataFromFile(CUKR_VECTOR_SP vec, FILE* f, int type)
{
	int i, ival;
	double dval;
	
	/*
	 * Check paramters 
	 */
	 
	/* Check that the file is 
	 * indeed open */	
	if (f == NULL)
		CUKR_FATALX("Trying to read data in "\
		            "vector %s from NULL stream",
		            vec->name);

	
	/* 
	 * Find the destination data type 
	 */
	 
	int dest_type;
	if (type == CUKR_DEFAULT)
		dest_type = CukrTypeStrToId(CukrDBGetS(CUKR_DB_SOLVER_PRECISION));
	else
	 	dest_type = type;
	
	
	/*
	 * Read the data 
	 */
	
	LOGX(DEBUG, "Reading data in vector %s " \
	            "from file stream", vec->name);
	
	/* Make sure memory is allocated */
	CukrVecspMallocCPU(vec, dest_type);
	
	/* Read in double */
	if (dest_type == CUKR_TYPE_DOUBLE) { 
		for (i = 0; i < vec->n; i++) {
			int read_ok = fscanf(f, "%lg", &dval);
			vec->ph_dval[i] = dval;
		}
		vec->status = CUKR_STATE_CPU_DOUBLE;
		LOGX(DEBUG, "Done reading %d elements in "\
		            "vector %s (double)", vec->n, vec->name);
	} 
	/* Read in qdouble */
	else if (dest_type == CUKR_TYPE_QDOUBLE) { 
		for (i = 0; i < vec->n; i++) {
			int read_ok = fscanf(f, "%lg", &dval);
			vec->ph_sval[i]  = (float)dval;
			vec->ph_qvalt[i] = (float)(dval - vec->ph_sval[i]);
		}
		vec->status  = CUKR_STATE_CPU_QDOUBLE;
		vec->status |= CUKR_STATE_CPU_SINGLE;
		LOGX(DEBUG, "Done reading %d elements in "\
		            "vector %s (qdouble)", vec->n, vec->name);
	} 
	/* Read in single */
	else if (dest_type == CUKR_TYPE_SINGLE) { 
		for (i = 0; i < vec->n; i++){
			int read_ok = fscanf(f, "%lg", &dval);
			vec->ph_sval[i] = (CUKR_SINGLE)dval;
		}
		vec->status = CUKR_STATE_CPU_SINGLE;
		LOGX(DEBUG, "Done reading %d elements in "\
		            "vector %s (single)", vec->n, vec->name);
	} 
	/* Read in int */
	else if (dest_type == CUKR_TYPE_INT) { 
		for (i = 0; i < vec->n; i++){
			int read_ok = fscanf(f, "%d", &ival);
			vec->ph_ival[i] = ival;
		}
		vec->status = CUKR_STATE_CPU_INT;
		LOGX(DEBUG, "Done reading %d elements in "\
		            "vector %s (int)", vec->n, vec->name);
	} 
	else
		CUKR_FATALX("Cannot find a valid precision to read "\
		            "from file, in vector %s", vec->name);
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads data into a previously
 *        created vector structure from 
 *        a file.
 * \param filename The file name
 * \param type The type of data that is
 *        supposed to be read (as double,
 *        qdouble, single, int) or default
 *        to match the solver
 */
void CukrVecspReadDataFromFilename(CUKR_VECTOR_SP vec, char* filename, int type)
{
	FILE *f;
	
	if ( (f = fopen(filename, "r")) == NULL)
		CUKR_FATALX("Could not read vector %s from file "\
		            "%s because the file could not be opened",
		            vec->name, filename);
	
	CukrVecspReadDataFromFile(vec, f, type);
	
	fclose(f);
}

// ------------------------------------------------------------------- //

/**
 * \brief Write the values of a vector
 *        to an opened file stream. 
 * \param vec The vector
 * \param f The open file stream
 */
void CukrVecspWriteDataToFile(CUKR_VECTOR_SP vec, FILE* f)
{	
	int i;
	
	/* If INT, bring it in on CPU, in INT */
	if (vec->status & CUKR_TYPE_INT)
		CukrVecspTransfer(vec, CUKR_LOCATION_CPU, CUKR_TYPE_INT);
	/* Else Bring back vector to CPU, in double */
	else
		CukrVecspTransfer(vec, CUKR_LOCATION_CPU, CUKR_TYPE_DOUBLE);
	
	/* Save in int*/
	if (vec->status & CUKR_TYPE_INT) 
	{	
		LOG(DEBUG, "Writing in int");
		for (i = 0; i < vec->n; i++)
			fprintf(f, "%d\n", vec->ph_ival[i]); 	
	}
	/* Else, in double */
	else {
		LOG(DEBUG, "Writing in double");
		for (i = 0; i < vec->n; i++)
			fprintf(f, "%30.50lg\n", vec->ph_dval[i]); 
	}
	
	LOGX(DEBUG, "Done writing vector %s to file stream", 
			    vec->name);
}

// ------------------------------------------------------------------- //

/**
 * \brief Write the values of a vector
 *        to an opened file stream. 
 * \param vec The vector
 * \param f The file name
 */
void CukrVecspWriteDataToFilename(CUKR_VECTOR_SP vec, char* filename)
{	
	FILE *f;
	
	LOGX(DEBUG, "Writing vector %s to file %s", 
			    vec->name, filename);
	
	if ( (f = fopen(filename, "w")) == NULL)
		CUKR_FATALX("Could not write vector %s to file "\
		            "%s because the file could not be opened",
		            vec->name, filename);
	
	/* Write the no. of elements */
	fprintf(f, "%d\n", vec->n);
	
	/* Write data to stream */
	CukrVecspWriteDataToFile(vec, f);
	
	/* Close */
	fclose(f);
	
	LOGX(DEBUG, "Done writing vector %s to file %s", 
			    vec->name, filename);
	
}



// ------------------------------------------------------------------- //

   
/**
 * \brief Prints the values of a vector
 * \param vec Pointer to the vector being
 *        printed
 */
void CukrVecspPrint(CUKR_VECTOR_SP vec)
{
	CukrVecspPrintToFile(vec, stdout);
}

// ------------------------------------------------------------------- //

/**
 * \brief Prints the values of a vector
 * \param vec The vector being printed
 * \param filename The full path of the file
 *        to write the vector to
 */
void CukrVecspPrintToFilename(CUKR_VECTOR_SP vec, char* filename)
{
	FILE *f;
	
	if ( (f = fopen(filename, "w")) == NULL)
		CUKR_FATALX("Could not write vector %s to file "\
		            "%s because the file could not be opened",
		            vec->name, filename);
	
	CukrVecspPrintToFile(vec, f);
	
	fclose(f);
}

// ------------------------------------------------------------------- //

/**
 * \brief Prints the values of a vector
 * \param vec Pointer to the vector being
 *        printed
 * \param f An open file stream to write to instead
 */
void CukrVecspPrintToFile(CUKR_VECTOR_SP vec, FILE* f)
{
	CukrVecspPrintToFileWithIndent(vec, f, "");
}

// ------------------------------------------------------------------- //

/**
 * \brief Prints the values of a vector
 * \param vec Pointer to the vector being
 *        printed
 * \param f An open file stream to write to instead
 * \param indent An indentation string, to be inserted
 *        before every print.
 */
void CukrVecspPrintToFileWithIndent(CUKR_VECTOR_SP vec, FILE* f, char* indent)
{	
	char status_str[255];
	int i; 
	
	/* Get string for status */
	CukrStatusIdToStr(vec->status, status_str);
	 
	fprintf(f, "%sVector: %s with status %s and %d elem\n", 
	            indent, vec->name, status_str, vec->n);
	
	/* If the location of the vector is
	 * on the CPU, print it directly */	 
	if (vec->status & CUKR_LOCATION_CPU)
	{	
		/* Print in double */
		if (vec->status & CUKR_TYPE_DOUBLE) {
			fprintf(f, "%sAddr = %p\n", indent, vec->ph_dval);
			fprintf(f, "%s", indent); 
			for (i = 0; i < vec->n; i++)
				fprintf(f, "%1.16g ", vec->ph_dval[i]);
			fprintf(f, "\n");
		}
		
		/* Print in qdouble */
		else if (vec->status & CUKR_TYPE_QDOUBLE) {
			fprintf(f, "%sAddr = (%p, %p)\n", indent, 
			        vec->ph_sval, vec->ph_qvalt);
			fprintf(f, "%s", indent);         
			for (i = 0; i < vec->n; i++)
				fprintf(f, "%1.16g(%g,%g) ", (double)vec->ph_sval[i] + (double)vec->ph_qvalt[i],
		                vec->ph_sval[i], vec->ph_qvalt[i]);
			fprintf(f, "\n");
		}
		/* Print in single */
		else if (vec->status & CUKR_TYPE_SINGLE) {
			fprintf(f, "%sAddr = %p\n", indent, vec->ph_sval);
			fprintf(f, "%s", indent); 
			for (i = 0; i < vec->n; i++)
				fprintf(f, "%g ", vec->ph_sval[i]);
			fprintf(f, "\n");
		}
		/* Print in int */
		else if (vec->status & CUKR_TYPE_INT) {
			fprintf(f, "%sAddr = %p\n", indent, vec->ph_ival);
			fprintf(f, "%s", indent); 
			for (i = 0; i < vec->n; i++)
				fprintf(f, "%d ", vec->ph_ival[i]);
			fprintf(f, "\n");
		}
	}
	
	/* If the location of the vector is
	 * on the GPU, print it after transfer
	 * on the CPU, but without updating the
	 * location and precision */	 
	 
	else 
	{
		/* Print in double */
		if (vec->status & CUKR_TYPE_DOUBLE) 
		{	
			/* Transfer in double */
			CukrVecspTransfer(vec, CUKR_LOCATION_CPU, CUKR_TYPE_DOUBLE);
			
			/* Print in double */
			fprintf(f, "%sAddr = %p\n", indent, vec->pd_dval);
			fprintf(f, "%s", indent); 
			for (i = 0; i < vec->n; i++)
				fprintf(f, "%g ", vec->ph_dval[i]);
			fprintf(f, "\n"); 	
		}
		
		/* Print in qdouble */
		else if (vec->status & CUKR_TYPE_QDOUBLE) {
					
			/* Transfer in qdouble */
			CukrVecspTransfer(vec, CUKR_LOCATION_CPU, CUKR_TYPE_QDOUBLE);
			
			fprintf(f,"%sAddr = (%p, %p)\n", indent, vec->pd_sval, vec->pd_qvalt);
			fprintf(f, "%s", indent);
			for (i = 0; i < vec->n; i++) 
				fprintf(f, "%g(%g,%g) ", (double)vec->ph_sval[i] + (double)vec->ph_qvalt[i],
								     vec->ph_sval[i], vec->ph_qvalt[i]);
			fprintf(f, "\n");
		}
		
		/* Print in single */
		else if (vec->status & CUKR_TYPE_SINGLE) 
		{	
			/* Transfer in single */
			CukrVecspTransfer(vec, CUKR_LOCATION_CPU, CUKR_TYPE_SINGLE);
			
			/* Print in single */
			fprintf(f, "%sAddr = %p\n", indent, vec->pd_sval);
			fprintf(f, "%s", indent); 
			for (i = 0; i < vec->n; i++)
				fprintf(f, "%g ", vec->ph_sval[i]);
			fprintf(f, "\n"); 	
		}
		
		/* Print in int */
		else if (vec->status & CUKR_TYPE_INT) 
		{	
			/* Transfer in single */
			CukrVecspTransfer(vec, CUKR_LOCATION_CPU, CUKR_TYPE_INT);
			
			/* Print in int */
			fprintf(f, "%sAddr = %p\n", indent, vec->pd_ival);
			fprintf(f, "%s", indent); 
			for (i = 0; i < vec->n; i++)
				fprintf(f, "%d ", vec->ph_ival[i]);
			fprintf(f, "\n"); 	
		}	
	} 
} 


//============================ PRIVATE FUNCTIONS =========================== //




/**
 * \brief Aux function to check for malloc
 *        errors on the GPU 
 * \param status The status returned by the
 *        malloc function
 */
#ifdef CUKR_USE_CUDA 
void CukrHandleCudaMemErrors(CUKR_VECTOR_SP vec, cublasStatus status)
{
	if (status == CUBLAS_STATUS_NOT_INITIALIZED) {
		CUKR_FATALX("Error allocating GPU memory for vector %s. "\
		            "Reason: CUBLAS has not been initialized", vec->name);
	}
	else if (status == CUBLAS_STATUS_INVALID_VALUE) {
		CUKR_FATALX("Error allocating GPU memory for vector %s. " \
		            "Reason: Invalid value", vec->name);
	}
	else if (status == CUBLAS_STATUS_ALLOC_FAILED) {
		CUKR_FATALX("Error allocating GPU memory for vector %s. "\
		            "Reason: Allocation failure", vec->name);
	}
	else if (status == CUBLAS_STATUS_MAPPING_ERROR) {
		CUKR_FATALX("Error transfering vector %s from/to GPU. "\
		            "Reason: Mapping error", vec->name);
	}
}
#endif /* CUKR_USE_CUDA */

// ------------------------------------------------------------------- //

/**
 * \brief Copy data from the current precision
 *        to another, on the CPU
 * \param vec Pointer to the vector
 * \param dest_type The destination type 
 *        (double, qdouble, single)
 */
void CukrVecspCopyDataCPU2CPU(CUKR_VECTOR_SP vec, int dest_type) 
{
	int i, success = CUKR_FALSE;
	
	/* 
	 * Copy only if data in the current
	 * precision does not contain valid
	 * (actual) data
	 */ 
	 
	if (!(vec->status & dest_type))
	{
		/* First alloc memory for the destination,
		 * if necessary */
		CukrVecspMallocCPU(vec, dest_type);
		
		/* If destination is double */
		if (dest_type == CUKR_TYPE_DOUBLE) 
		{
			/* First look for data in qdouble.
			 * If available, promote and update status*/
			if (vec->status & CUKR_STATE_CPU_QDOUBLE) {
				for (i = 0; i < vec->n; i++)
					vec->ph_dval[i] = (CUKR_DOUBLE)vec->ph_sval[i] + (CUKR_DOUBLE)vec->ph_qvalt[i];
				LOGX(DEBUG, "Promoted vector %s (host, qdouble) to double precision", vec->name);
				vec->status |= CUKR_STATE_CPU_DOUBLE;
				return;
			} 		
			/* If not, look for data in single.
			 * If available, promote and update status*/
			else if (vec->status & CUKR_STATE_CPU_SINGLE) {
				for (i = 0; i < vec->n; i++)
					vec->ph_dval[i] = (CUKR_DOUBLE)vec->ph_sval[i];
				LOGX(DEBUG, "Promoted vector %s (host, single) to double precision", vec->name);
				vec->status |= CUKR_STATE_CPU_DOUBLE;
				return;
			} 
			/* If nothing, error */
			else {
				CUKR_FATALX("Trying to convert vector %s to " \
				            "double on host, but no floating " \
				            "point data source found", vec->name);
			}
			
			success = CUKR_TRUE;
		}
		
		/* If destination is in qdouble */
		if (dest_type == CUKR_TYPE_QDOUBLE) 
		{				
			/* First look for double precision
			 * to convert data from */
			if (vec->status & CUKR_STATE_CPU_DOUBLE) {
				for (i = 0; i < vec->n; i++) {
					vec->ph_sval[i]  = (CUKR_SINGLE)vec->ph_dval[i];
					vec->ph_qvalt[i] = (CUKR_SINGLE)(vec->ph_dval[i] - vec->ph_sval[i]);
				}
				LOGX(DEBUG, "Converted vector %s (host, double) to quasi-double precision", vec->name);
				
				/* Set the status to both single
				 * and qdouble, since the single
				 * is included in qdouble (equal
				 * to the head part) */
				vec->status |= CUKR_STATE_CPU_SINGLE;
				vec->status |= CUKR_STATE_CPU_QDOUBLE;
				return;
			}  
			/* If there is no double, try to get
			 * from single (e.g. just set the tail
			 * to zero ) */
			if (vec->status & CUKR_STATE_CPU_SINGLE) {
				for (i = 0; i < vec->n; i++) {
					vec->ph_qvalt[i] = 0;
				}
				LOGX(DEBUG, "Promoted vector %s (host, single) to quasi-double precision", vec->name);
				vec->status |= CUKR_STATE_CPU_SINGLE;
				vec->status |= CUKR_STATE_CPU_QDOUBLE;
				return;
			}
			/* If nothing, error */
			else {
				CUKR_FATALX("Trying to convert vector %s to " \
				            "qdouble on host, but no floating " \
				            "point data source found", vec->name);
			}
			
			success = CUKR_TRUE;  
		}
				
		/* If destination is single */
		if (dest_type == CUKR_TYPE_SINGLE)
		{	
			/* @note No need to search for data in 
			 *       qdouble, since if data in qdouble
			 *       is present, then so is data in
			 *       single (since it is the head part) */
			if (vec->status & CUKR_STATE_CPU_QDOUBLE) {
				LOGX(DEBUG, "No need to convert vector %s to single since "\
				            "qdouble format includes it", vec->name);
			}
			
			/* First look for double precision
			 * to truncate data from */
			else if (vec->status & CUKR_STATE_CPU_DOUBLE) {
				for (i = 0; i < vec->n; i++)
					vec->ph_sval[i] = (CUKR_SINGLE)vec->ph_dval[i];
				LOGX(DEBUG, "Truncated vector %s (host, double) to single precision", vec->name);
				vec->status |= CUKR_STATE_CPU_SINGLE;
				return;
			} 
			/* If nothing, error */
			else {
				CUKR_FATALX("Trying to convert vector %s to " \
				            "single on host, but no floating " \
				            "point data source found", vec->name);
			}
			
			success = CUKR_TRUE;
		}
		
		/* Nothing for integer type, since
		 * it does not have to be converted,
		 * only moved */
		 
		/* If none of the above, error */
		if (success == CUKR_FALSE) {
			CUKR_FATALX("Error copying data CPU->CPU for vector %s. "\
			            "Invalid dest. type (%d)", vec->name, dest_type);              
		}	 
		
	}
	/* If actual data is found, give 
	 * some message */
	else {
		char str_type[255];
		CukrTypeIdToStr(dest_type, str_type);
		LOGX(DEBUG, "No need to change type for " \
			        "vector %s in %s (on host) since data " \
			        "in this type is still valid", 
			         vec->name, str_type);
		}
} 

// ------------------------------------------------------------------- //

/**
 * \brief Copy data (possibly from the current 
 *        precision to another one) from CPU to GPU
 * \param vec Pointer to the vector
 * \param dest_type The destination type 
 *        (double, qdouble, single)
 */
void CukrVecspCopyDataCPU2GPU(CUKR_VECTOR_SP vec, int dest_type) 
#ifndef CUKR_USE_GPU
{
	CUKR_FATAL("GPU support was not compiled in the library " \
			   "(CukrVecCopyDataCPU2GPU cannot be executed) ");
}
#else
{
	#ifdef CUKR_USE_CUDA
	cublasStatus status;
	
	int success = CUKR_FALSE;
	
	/* Alloc memory on the GPU, if needed */
	CukrVecspMallocGPU(vec, dest_type);
		 
	/* Change the precision on the CPU, if 
	 * data in that precision is not
	 * available */
	if (!(vec->status & dest_type)) 
		CukrVecspCopyDataCPU2CPU(vec, dest_type);
	
	
	/* If the destination is double, copy
	 * the data and update vector status */   
	if (dest_type == CUKR_TYPE_DOUBLE)
	{
		status = cublasSetVector(vec->n, sizeof(CUKR_DOUBLE), vec->ph_dval,
	                             1, vec->pd_dval, 1);
	    CukrHandleCudaMemErrors(vec, status);
	    LOGX(DEBUG, "Copied vector %s (host, double) to (GPU, double)", vec->name);
	    vec->status |= CUKR_STATE_GPU_DOUBLE; 
	    
	    success = CUKR_TRUE; 
	}	   
		
	/* If destination is single, copy the
	 * data and update vector status to
	 * single.
	 * If destination is qdouble, copy the
	 * head data and set the status to
	 * single (it will be set to qdouble
	 * in the next step, after copying the
	 * tail part as well)*/
	if (dest_type == CUKR_TYPE_SINGLE || dest_type == CUKR_TYPE_QDOUBLE) {
		status = cublasSetVector(vec->n, sizeof(CUKR_SINGLE), vec->ph_sval,
	                             1, vec->pd_sval, 1);
	    CukrHandleCudaMemErrors(vec, status);
	    if (dest_type == CUKR_TYPE_SINGLE) {
	    	LOGX(DEBUG, "Copied vector %s (host, single) to (GPU, single)", vec->name);
	    }
	    else {
	    	LOGX(DEBUG, "Copied vector %s (host, qdouble, head) to (GPU, qdouble, head)", vec->name);
	    }
	    vec->status |= CUKR_STATE_GPU_SINGLE;
	    
	    success = CUKR_TRUE;
	}
	/* If the destination is qdouble, set
	 * copy the tail part as well and add
	 * qdouble to the status */
	if (dest_type == CUKR_TYPE_QDOUBLE) {
		//printf("write cpu to gpu qdouble tail: %f\n", *(vec->ph_qvalt));
		
		status = cublasSetVector(vec->n, sizeof(CUKR_SINGLE), vec->ph_qvalt,
	                             1, vec->pd_qvalt, 1);
	    CukrHandleCudaMemErrors(vec, status);
    	LOGX(DEBUG, "Copied vector %s (host, qdouble, tail) to (GPU, qdouble, tail)", vec->name);
	    vec->status |= CUKR_STATE_GPU_QDOUBLE;
	    
	    success = CUKR_TRUE;
	} 
	/* If the destination is int, copy
	 * the data and update vector status */   
	if (dest_type == CUKR_TYPE_INT)
	{
		status = cublasSetVector(vec->n, sizeof(CUKR_INT), vec->ph_ival,
	                             1, vec->pd_ival, 1);
	    CukrHandleCudaMemErrors(vec, status);
	    LOGX(DEBUG, "Copied vector %s (host, int) to (GPU, int)", vec->name);
	    vec->status |= CUKR_STATE_GPU_INT;  
	    
	    success = CUKR_TRUE;                                                 
	}  
	
	/* If none of the above, error */
	if (success == CUKR_FALSE) {
		CUKR_FATALX("Error copying data CPU->GPU for vector %s. "\
		            "Invalid dest. type (%d)", vec->name, dest_type);  
	}
	#endif /* CUKR_USE_CUDA */
	
	#ifdef CUKR_USE_OPENCL
	//cublasStatus status;
	
	int success = CUKR_FALSE;
	
	/* Alloc memory on the GPU, if needed */
	CukrVecspMallocGPU(vec, dest_type);
	
	/* Change the precision on the CPU, if 
	 * data in that precision is not
	 * available */
	if (!(vec->status & dest_type)) 
		CukrVecspCopyDataCPU2CPU(vec, dest_type);
		
		
	/* If the destination is double, copy
	 * the data and update vector status */   
		if (dest_type == CUKR_TYPE_DOUBLE)
		{
			err = clEnqueueWriteBuffer(ComputeCommands, *(vec->pcl_dval), CL_TRUE, 0, vec->n * sizeof(CUKR_DOUBLE), (void *) (vec->ph_dval), 0, NULL, NULL);
			if (err != CL_SUCCESS) {
				printf("writing cl_mem double failed!\n");
			}
			//CukrHandleCudaMemErrors(vec, status);
			LOGX(DEBUG, "Copied vector %s (host, double) to (GPU, double)", vec->name);
			vec->status |= CUKR_STATE_GPU_DOUBLE; 
			
			success = CUKR_TRUE; 
		}	   
	
	/* If destination is single, copy the
	 * data and update vector status to
	 * single.
	 * If destination is qdouble, copy the
	 * head data and set the status to
	 * single (it will be set to qdouble
	 * in the next step, after copying the
	 * tail part as well)*/
	if (dest_type == CUKR_TYPE_SINGLE || dest_type == CUKR_TYPE_QDOUBLE) {
		//printf("value 10 to gpu: %f\n", vec->ph_sval[10]);
		/*
		if (vec->pcl_sval == NULL) {
			vec->pcl_sval = (cl_mem*)malloc(sizeof(cl_mem));
			*(vec->pcl_sval) = clCreateBuffer(ComputeContext, CL_MEM_READ_WRITE|CL_MEM_COPY_HOST_PTR,  vec->n * sizeof(cl_float), (void *) (vec->ph_sval), &err);
			if (err != CL_SUCCESS)
			{
				printf("Error: Failed to allocate device memory pcl_sval!\n");
				exit(1);
			}
		}
		else{
		*/
		err = clEnqueueWriteBuffer(ComputeCommands, *(vec->pcl_sval), CL_TRUE, 0, vec->n * sizeof(CUKR_SINGLE), (void *) (vec->ph_sval), 0, NULL, NULL);
		if (err != CL_SUCCESS) {
			printf("writing cl_mem single failed!\n");
		}
		//}
		
	    //CukrHandleCudaMemErrors(vec, status);
	    if (dest_type == CUKR_TYPE_SINGLE) {
	    	LOGX(DEBUG, "Copied vector %s (host, single) to (GPU, single)", vec->name);
	    }
	    else {
	    	LOGX(DEBUG, "Copied vector %s (host, qdouble, head) to (GPU, qdouble, head)", vec->name);
	    }
	    vec->status |= CUKR_STATE_GPU_SINGLE;
	    
	    success = CUKR_TRUE;
	}
	/* If the destination is qdouble, set
	 * copy the tail part as well and add
	 * qdouble to the status */
	if (dest_type == CUKR_TYPE_QDOUBLE) {
	    //CukrHandleCudaMemErrors(vec, status);
		//printf("write cpu to gpu qdouble tail: %f\n", *(vec->ph_qvalt));
		
		err = clEnqueueWriteBuffer(ComputeCommands, *(vec->pcl_qvalt), CL_TRUE, 0, vec->n * sizeof(CUKR_SINGLE), (void *) (vec->ph_qvalt), 0, NULL, NULL);
		if (err != CL_SUCCESS) {
			printf("writing cl_mem qsingle tail failed!\n");
		}
		
    	LOGX(DEBUG, "Copied vector %s (host, qdouble, tail) to (GPU, qdouble, tail)", vec->name);
	    vec->status |= CUKR_STATE_GPU_QDOUBLE;
	    
	    success = CUKR_TRUE;
	} 
	/* If the destination is int, copy
	 * the data and update vector status */   
	if (dest_type == CUKR_TYPE_INT)
	{
	    //CukrHandleCudaMemErrors(vec, status);
		err = clEnqueueWriteBuffer(ComputeCommands, *(vec->pcl_ival), CL_TRUE, 0, vec->n * sizeof(CUKR_INT), (void *) (vec->ph_ival), 0, NULL, NULL);
		if (err != CL_SUCCESS) {
			printf("writing cl_mem double failed!\n");
		}
		
	    LOGX(DEBUG, "Copied vector %s (host, int) to (GPU, int)", vec->name);
	    vec->status |= CUKR_STATE_GPU_INT;  
	    
	    success = CUKR_TRUE;                                                 
	}  
	
	/* If none of the above, error */
	
	if (success == CUKR_FALSE) {
		CUKR_FATALX("Error copying data CPU->GPU for vector %s. "\
		            "Invalid dest. type (%d)", vec->name, dest_type);  
	}
	#endif /* CUKR_USE_OPENCL */
}
#endif /* CUKR_USE_GPU */

// ------------------------------------------------------------------- //

/**
 * \brief Copy data (possibly from the current 
 *        precision to another one) from GPU to CPU
 * \param vec Pointer to the vector
 * \param dest_type The destination type 
 *        (double, qdouble, single)
 */
void CukrVecspCopyDataGPU2CPU(CUKR_VECTOR_SP vec, int dest_type) 
#ifndef CUKR_USE_GPU
{
	CUKR_FATAL("GPU support was not compiled in the library " \
			   "(CukrVecCopyDataGPU2CPU cannot be executed) ");
}
#else
{
	#ifdef CUKR_USE_CUDA
	cublasStatus status;
	int got_data = CUKR_FALSE;
	
	/* 
	 * Alloc memory on the CPU to store 
	 * the transferred data into 
	 */
	switch (dest_type)
	{
		case CUKR_TYPE_DOUBLE:
			if (vec->status & CUKR_STATE_GPU_DOUBLE)
				CukrVecspMallocCPU(vec, CUKR_TYPE_DOUBLE);
			else if	(vec->status & CUKR_STATE_GPU_QDOUBLE)
				CukrVecspMallocCPU(vec, CUKR_TYPE_QDOUBLE);
			else if	(vec->status & CUKR_STATE_GPU_SINGLE) {					
				CukrVecspMallocCPU(vec, CUKR_TYPE_SINGLE);
			}
			else {
				CUKR_FATALX("Error copying vector %s GPU->CPU: No " \
				            "GPU floating point data to copy from", vec->name);         
			}	
			break;
			
		case CUKR_TYPE_QDOUBLE:	
			if (vec->status & CUKR_STATE_GPU_QDOUBLE) {
				CukrVecspMallocCPU(vec, CUKR_TYPE_QDOUBLE);
			}
			else if (vec->status & CUKR_STATE_GPU_DOUBLE)
				CukrVecspMallocCPU(vec, CUKR_TYPE_DOUBLE);
			else if (vec->status & CUKR_STATE_GPU_SINGLE)
				CukrVecspMallocCPU(vec, CUKR_TYPE_SINGLE);
			else {
				CUKR_FATALX("Error copying vector %s GPU->CPU: No " \
				            "GPU floating point data to copy from", vec->name);           
			}	
			break;
		
		case CUKR_TYPE_SINGLE:
			if (vec->status & CUKR_STATE_GPU_SINGLE) {				
				CukrVecspMallocCPU(vec, CUKR_TYPE_SINGLE);
			}
			else if (vec->status & CUKR_STATE_GPU_DOUBLE)
				CukrVecspMallocCPU(vec, CUKR_TYPE_DOUBLE);
			else {
				CUKR_FATALX("Error copying vector %s GPU->CPU: No " \
				            "GPU floating point data to copy from", vec->name);           
			}	
			break;
		
		case CUKR_TYPE_INT:
			CukrVecspMallocCPU(vec, CUKR_TYPE_INT);
			break;
		
		default:
			CUKR_FATALX("Invalid destination type while copying " \
			            "vector %s GPU->CPU (%d)", vec->name, dest_type); 		
	}

	 			 
 	/* If the vector has double data on the GPU */
	if (vec->status & CUKR_STATE_GPU_DOUBLE)
	{
		status = cublasGetVector(vec->n, sizeof(double), vec->pd_dval, 1, vec->ph_dval, 1);
	    CukrHandleCudaMemErrors(vec, status);
	    vec->status |= CUKR_STATE_CPU_DOUBLE;
	    got_data = CUKR_TRUE;    
	    	
	    LOGX(DEBUG, "Copied vector %s (GPU, double) to (host, double)", vec->name);
	}
 	
	
	/* If the vector has data in single
	 * on the GPU, bring it and set the
	 * cpu-single flag.
	 * If the vector has the data in qdouble
	 * on the GPU, do the same thing and set
	 * the cpu-single flag only, since the
	 * cpu-qdouble one will be set after 
	 * copying the tail part as well */
	if (vec->status & CUKR_STATE_GPU_SINGLE || vec->status & CUKR_STATE_GPU_QDOUBLE)
	{
		status = cublasGetVector(vec->n, sizeof(CUKR_SINGLE), vec->pd_sval, 1, vec->ph_sval, 1);
	                         	                	 
	    CukrHandleCudaMemErrors(vec, status);
	    vec->status |= CUKR_STATE_CPU_SINGLE; 
	    got_data = CUKR_TRUE;
	    
	    if (dest_type & CUKR_STATE_GPU_QDOUBLE) {
	    	LOGX(DEBUG, "Copied vector %s (GPU, qdouble, head) to (host, qdouble, head)", vec->name);
	    } 
	    else {
	    	LOGX(DEBUG, "Copied vector %s (GPU, single ) to (host, single)", vec->name);
	    }
		
		//printf("write gpu to cpu single: %10.30lf\n", *(vec->ph_sval));
	}
	/* If the vector has data in qdouble
	 * on the GPU, copy the tail part as
	 * well and add to the status the
	 * cpu-qdouble flag */
	if (vec->status & CUKR_STATE_GPU_QDOUBLE)
	{
		status = cublasGetVector(vec->n, sizeof(CUKR_SINGLE), vec->pd_qvalt,
	                         	1, vec->ph_qvalt, 1);
		CukrHandleCudaMemErrors(vec, status);
		vec->status |= CUKR_STATE_CPU_QDOUBLE; 
		got_data = CUKR_TRUE;
    	LOGX(DEBUG, "Copied vector %s (GPU, qdouble, tail) to (host, qdouble, tail)", vec->name);

		//printf("write gpu to cpu qdouble tail: %10.60lf\n", *(vec->ph_qvalt));
	}
	 
	/* If the vector has data in int
	 * on the GPU */
	if (vec->status & CUKR_STATE_GPU_INT)
	{
		status = cublasGetVector(vec->n, sizeof(CUKR_INT), vec->pd_ival,
	                         	1, vec->ph_ival, 1);
	    CukrHandleCudaMemErrors(vec, status);
	    vec->status |= CUKR_STATE_CPU_INT;
	    got_data = CUKR_TRUE;    
	    
	    LOGX(DEBUG, "Copied vector %s (GPU, int) to (host, int)", vec->name);
	}
	
	/* If neither, error because there
	 * is nowhere to get the data from */
	if (!got_data) {
		CUKR_FATALX("Cound not move vector %s (host) to GPU " \
			        "since data on the GPU is not available", 
			         vec->name);
	} 	
 
	/* 
	 * If necessary, convert on the CPU to the
	 * desired destination type 
	 */
	 
	CukrVecspCopyDataCPU2CPU(vec, dest_type);
	#endif /* CUKR_USE_CUDA */
	
	#ifdef CUKR_USE_OPENCL
	//cublasStatus status;
	int got_data = CUKR_FALSE;
	
	/* 
	 * Alloc memory on the CPU to store 
	 * the transferred data into 
	 */
	switch (dest_type)
	{
		case CUKR_TYPE_DOUBLE:
			if (vec->status & CUKR_STATE_GPU_DOUBLE)
				CukrVecspMallocCPU(vec, CUKR_TYPE_DOUBLE);
			else if	(vec->status & CUKR_STATE_GPU_QDOUBLE)
				CukrVecspMallocCPU(vec, CUKR_TYPE_QDOUBLE);
			else if	(vec->status & CUKR_STATE_GPU_SINGLE) {					
				CukrVecspMallocCPU(vec, CUKR_TYPE_SINGLE);
			}
			else {
				CUKR_FATALX("Error copying vector %s GPU->CPU: No " \
				            "GPU floating point data to copy from", vec->name);         
			}	
			break;
			
		case CUKR_TYPE_QDOUBLE:	
			if (vec->status & CUKR_STATE_GPU_QDOUBLE) {
				CukrVecspMallocCPU(vec, CUKR_TYPE_QDOUBLE);
			}
			else if (vec->status & CUKR_STATE_GPU_DOUBLE)
				CukrVecspMallocCPU(vec, CUKR_TYPE_DOUBLE);
			else if (vec->status & CUKR_STATE_GPU_SINGLE)
				CukrVecspMallocCPU(vec, CUKR_TYPE_SINGLE);
			else {
				CUKR_FATALX("Error copying vector %s GPU->CPU: No " \
				            "GPU floating point data to copy from", vec->name);           
			}	
			break;
			
		case CUKR_TYPE_SINGLE:
			if (vec->status & CUKR_STATE_GPU_SINGLE) {				
				CukrVecspMallocCPU(vec, CUKR_TYPE_SINGLE);
			}
			else if (vec->status & CUKR_STATE_GPU_DOUBLE)
				CukrVecspMallocCPU(vec, CUKR_TYPE_DOUBLE);
			else {
				CUKR_FATALX("Error copying vector %s GPU->CPU: No " \
				            "GPU floating point data to copy from", vec->name);           
			}	
			break;
			
		case CUKR_TYPE_INT:
			CukrVecspMallocCPU(vec, CUKR_TYPE_INT);
			break;
			
		default:
			CUKR_FATALX("Invalid destination type while copying " \
			            "vector %s GPU->CPU (%d)", vec->name, dest_type); 		
	}
	
	
 	/* If the vector has double data on the GPU */
	if (vec->status & CUKR_STATE_GPU_DOUBLE)
	{
		err = clEnqueueReadBuffer(ComputeCommands, *(vec->pcl_dval), CL_TRUE/*blocking or not*/, 0, 
								  vec->n * sizeof(cl_double), (void*)(vec->ph_dval), 0, NULL, NULL);
		if (err != CL_SUCCESS) {
			printf("Reading gpu-double array failed!\n");
		}
	    // CukrHandleCudaMemErrors(vec, status);
	    vec->status |= CUKR_STATE_CPU_DOUBLE;
	    got_data = CUKR_TRUE;    
		
	    LOGX(DEBUG, "Copied vector %s (GPU, double) to (host, double)", vec->name);
	}
 	
	
	/* If the vector has data in single
	 * on the GPU, bring it and set the
	 * cpu-single flag.
	 * If the vector has the data in qdouble
	 * on the GPU, do the same thing and set
	 * the cpu-single flag only, since the
	 * cpu-qdouble one will be set after 
	 * copying the tail part as well */
	if (vec->status & CUKR_STATE_GPU_SINGLE || vec->status & CUKR_STATE_GPU_QDOUBLE)
	{
		//cl_event transfere;
		//status = cublasGetVector(vec->n, sizeof(CUKR_SINGLE), vec->pd_sval, 1, vec->ph_sval, 1);
		err = clEnqueueReadBuffer(ComputeCommands, *(vec->pcl_sval), CL_TRUE/*blocking or not*/, 0, 
								  vec->n * sizeof(cl_float), (void*)(vec->ph_sval), 0, NULL, NULL);//&transfere);
		if (err != CL_SUCCESS) {
			printf("Reading gpu-single/qsingle array failed!\n");
		}/*
		err = clWaitForEvents(1, &transfere);
		if (err != CL_SUCCESS) {
			printf("Transfere event failed!\n");
		}*/
		// printf("value 10 to cpu: %f\n", vec->ph_sval[10]);
	
	    // CukrHandleCudaMemErrors(vec, status);
	    vec->status |= CUKR_STATE_CPU_SINGLE; 
	    got_data = CUKR_TRUE;
	    
	    if (dest_type & CUKR_STATE_GPU_QDOUBLE) {
	    	LOGX(DEBUG, "Copied vector %s (GPU, qdouble, head) to (host, qdouble, head)", vec->name);
	    } 
	    else {
	    	LOGX(DEBUG, "Copied vector %s (GPU, single ) to (host, single)", vec->name);
	    }
		
		//printf("write gpu to cpu single: %10.30lf\n", *(vec->ph_sval));
	}
	/* If the vector has data in qdouble
	 * on the GPU, copy the tail part as
	 * well and add to the status the
	 * cpu-qdouble flag */
	if (vec->status & CUKR_STATE_GPU_QDOUBLE)
	{
		//cl_event transfere;
		//printf("qdouble true\n");
		// status = cublasGetVector(vec->n, sizeof(CUKR_SINGLE), vec->pd_qvalt, 1, vec->ph_qvalt, 1);
		err = clEnqueueReadBuffer(ComputeCommands, *(vec->pcl_qvalt), CL_TRUE/*blocking or not*/, 0, 
								  vec->n * sizeof(cl_float), (void*)(vec->ph_qvalt), 0, NULL, /*&transfere*/ NULL);
		if (err != CL_SUCCESS) {
			printf("Reading gpu qdouble-tail-array failed!\n");
		}/*
		err = clWaitForEvents(1, &transfere);
		if (err != CL_SUCCESS) {
			printf("Transfere event failed!\n");
		}*/
	    // CukrHandleCudaMemErrors(vec, status);
	    vec->status |= CUKR_STATE_CPU_QDOUBLE; 
		got_data = CUKR_TRUE;
    	LOGX(DEBUG, "Copied vector %s (GPU, qdouble, tail) to (host, qdouble, tail)", vec->name);
		//printf("read qdata\n");
		//printf("write gpu to cpu qdouble tail: %10.60lf\n", *(vec->ph_qvalt));
		
	}
	
	/* If the vector has data in int
	 * on the GPU */
	if (vec->status & CUKR_STATE_GPU_INT)
	{
		// status = cublasGetVector(vec->n, sizeof(CUKR_INT), vec->pd_ival, 1, vec->ph_ival, 1);
		err = clEnqueueReadBuffer(ComputeCommands, *(vec->pcl_ival), CL_TRUE/*blocking or not*/, 0, 
								  vec->n * sizeof(cl_int), (void*)(vec->ph_ival), 0, NULL, NULL);
		if (err != CL_SUCCESS) {
			printf("Reading gpu int-array failed!\n");
		}
	    // CukrHandleCudaMemErrors(vec, status);
	    vec->status |= CUKR_STATE_CPU_INT;
	    got_data = CUKR_TRUE;    
	    
	    LOGX(DEBUG, "Copied vector %s (GPU, int) to (host, int)", vec->name);
	}
	
	/* If neither, error because there
	 * is nowhere to get the data from */
	if (!got_data) {
		CUKR_FATALX("Cound not move vector %s (host) to GPU " \
			        "since data on the GPU is not available", 
					vec->name);
	} 	
	
	/* 
	 * If necessary, convert on the CPU to the
	 * desired destination type 
	 */
	
	CukrVecspCopyDataCPU2CPU(vec, dest_type);
	#endif /* CUKR_USE_OPENCL */	
}
#endif /* CUKR_USE_GPU */

// ------------------------------------------------------------------- //

/**
 * \brief Copy data (possibly from the current 
 *        precision to another one) from GPU to GPU
 * \param vec Pointer to the vector
 * \param dest_type The destination type 
 *        (double, qdouble, single)
 */
void CukrVecspCopyDataGPU2GPU(CUKR_VECTOR_SP vec, int dest_type) 
#ifndef CUKR_USE_GPU
{
	CUKR_FATAL("GPU support was not compiled in the library " \
			   "(CukrVecCopyDataGPU2GPU cannot be executed) ");
}
#else
{
	#ifdef CUKR_USE_CUDA
		cublasStatus status;
	#endif
	int got_data = CUKR_FALSE;
	
	/* 
	 * Alloc memory on the GPU to store 
	 * the transferred data into 
	 */
	
	CukrVecspMallocGPU(vec, dest_type);
	//printf("copies gpu to gpu\n");
	 
	/*
	 * Copy the data
	 */ 
	 			 
 	/* If the vector has double data on the GPU */
	if (vec->status & CUKR_STATE_GPU_DOUBLE)
	{
		/* If dest is double, do nothing */
		if (dest_type == CUKR_TYPE_DOUBLE) {
			LOGX(DEBUG, "No need to trasfer data in double precision " \
			 			"on the GPU for vector %s since it is already " \
			 			"there", vec->name);
			return; 			
		}
		
		/* If the destination is in qdouble */
		if (dest_type == CUKR_TYPE_QDOUBLE) {
			#ifdef CUKR_USE_OPENCL
				opencl_kernel_copy_dbl_to_qdbl(vec->n, vec->pcl_dval, vec->pcl_sval, vec->pcl_qvalt);
			#else
				cuda_kernel_copy_dbl_to_qdbl(vec->n, vec->pd_dval, vec->pd_sval, vec->pd_qvalt);
			#endif
			vec->status |= CUKR_STATE_GPU_QDOUBLE;
			vec->status |= CUKR_STATE_GPU_SINGLE;
			return;
		}
		
		/* If the destination is in single */
		if (dest_type == CUKR_TYPE_SINGLE) {
			#ifdef CUKR_USE_OPENCL
				opencl_kernel_copy_dbl_to_sng(vec->n, vec->pcl_dval, vec->pcl_sval);
			#else
				cuda_kernel_copy_dbl_to_sng(vec->n, vec->pd_dval, vec->pd_sval);
			#endif
			vec->status |= CUKR_STATE_GPU_SINGLE;
			return;
		}
		
		/* Else error */
		CUKR_FATALX("Trying to convert double precision GPU "\
		            "data for vector %s to non-floating point "\
		            "type", vec->name);
		
	}
	
	/* If the vector has qdouble data on the GPU */
	else if (vec->status & CUKR_STATE_GPU_QDOUBLE)
	{
		/* If dest is double */
		if (dest_type == CUKR_TYPE_DOUBLE) {
			#ifdef CUKR_USE_OPENCL
				opencl_kernel_copy_qdbl_to_dbl(vec->n, vec->pcl_sval, vec->pcl_qvalt, vec->pcl_dval);
			#else
				cuda_kernel_copy_qdbl_to_dbl(vec->n, vec->pd_sval, vec->pd_qvalt, vec->pd_dval);
			#endif
			vec->status |= CUKR_STATE_GPU_DOUBLE;
			return; 			
		}
		
		/* If the destination is in qdouble or single,
		 * do nothing */
		if (dest_type == CUKR_TYPE_QDOUBLE && dest_type == CUKR_TYPE_SINGLE) {
			LOGX(DEBUG, "No need to trasfer data in qdouble precision " \
			 			"on the GPU for vector %s since it is already " \
			 			"there", vec->name);
			return;
		}
		
		/* Else error */
		CUKR_FATALX("Trying to convert qdouble precision GPU "\
		            "data for vector %s to non-floating point "\
		            "type", vec->name);		
	}
	
	/* If the vector has qdouble data on the GPU */
	else if (vec->status & CUKR_STATE_GPU_SINGLE)
	{
		/* If dest is double */
		if (dest_type == CUKR_TYPE_DOUBLE) {
			#ifdef CUKR_USE_OPENCL
				opencl_kernel_copy_sng_to_dbl(vec->n, vec->pcl_sval, vec->pcl_dval);
			#else
				cuda_kernel_copy_sng_to_dbl(vec->n, vec->pd_sval, vec->pd_dval);
			#endif
			vec->status |= CUKR_STATE_GPU_DOUBLE;
			return; 			
		}
		
		/* If the destination is in qdouble or single */
		if (dest_type == CUKR_TYPE_QDOUBLE) {
			/* The head is there, need to set the
			 * tail to zero */
			#ifdef CUKR_USE_OPENCL
				opencl_kernel_set_sng_to_val(vec->n, vec->pcl_qvalt, 0);
			#else
				cuda_kernel_set_sng_to_val(vec->n, vec->pd_qvalt, 0);
			#endif
			return;
		}
		
		/* If the destination is in single, do nothing */
		if (dest_type == CUKR_TYPE_SINGLE) {
			LOGX(DEBUG, "No need to trasfer data in single precision " \
			 			"on the GPU for vector %s since it is already " \
			 			"there", vec->name);
			return;
		}
		
		/* Else error */
		CUKR_FATALX("Trying to convert single precision GPU "\
		            "data for vector %s to non-floating point "\
		            "type", vec->name);		
	}
	 
	/* If neither, error because there
	 * is nowhere to get the data from */
	CUKR_FATALX("Cound not move vector %s (GPU) to GPU " \
			     "since data on the GPU is not available", 
			     vec->name);
}
#endif /* CUKR_USE_CUDA */
