/**
 * @file mat_vec/sp/vector_sp.h
 * @author Serban Georgescu
 * @since 2008
 *
 * Defines the CUKR_VECTOR_SP abstract
 * interace.
 *
 * The CUKR_VECTOR_SP structure stores a
 * vector that can be in any (useful) type
 * and reside either on the CPU or GPU.
 * The structure stores, besides the data,
 * both precision and location.
 *
 * The precision shows can be of type
 * double, qdouble (quasi-double) single or int.
 *
 * The location can be CPU or GPU.
 *
 * Precision and location are combined into a
 * status, which shows where actual (updated)
 * data exists. For example, if a vector has
 * been created on the CPU in double precision,
 * it will start with the CUKR_STATUS_CPU_DOUBLE
 * status. If afterwards it will be converted to
 * single precision, the new status will of
 * CUKR_STATUS_CPU_DOUBLE | CUKR_STATUS_CPU_SINGLE
 * will reflect that. In this way, we can know
 * at any time if actual data is present where we
 * need it, and only copy / convert where needed.
 *
 */

#ifndef CUKR_VECTOR_SP_H_
#define CUKR_VECTOR_SP_H_

#ifndef CUKR_VECTOR
	#define CUKR_VECTOR CUKR_VECTOR_SP
#endif

#include <stdio.h>
#include "../../cukr.h"


#ifdef CUKR_USE_OPENCL
	#include "../../init/init_opencl.h"
#endif


/**
 * Contains an abstract vector in
 * arbitrary precision at an arbitrary
 * location.
 *
 * Data of various types is stored in the
 * member variables. Which one is used by
 * the BLAS is automatically decided based
 * on the location and precision of the blas.
 */
typedef struct _CUKR_VECTOR_SP {

	// ------------ Public ------------- //

	/** The name of the vector */
	char name[255];

	/** The status of the vector from a data
	 * point of view. The status stores a
	 * collection of location-datatype
	 * combinations which show which of the
	 * data members below contain valid data */
	int status;

	/** The length of the vector */
	int n;


	// ----------- Protected ------------ //
	
#ifdef CUKR_USE_OPENCL
	/** Host and device data for double
	 * precision */
	CUKR_DOUBLE *ph_dval, *pd_dval;
	cl_mem *pcl_dval;
	
	/** Host and device data for single
	 * precision */
	CUKR_SINGLE *ph_sval, *pd_sval;
	cl_mem *pcl_sval;
	
	/** Host and device data for quasi-double
	 * precision (tail part, since the head
	 * part will be stored in the members for
	 * single precision) */
	CUKR_SINGLE *ph_qvalt, *pd_qvalt;
	cl_mem *pcl_qvalt;
	
	/** Host and device data for integer */
	CUKR_INT    *ph_ival, *pd_ival;
	cl_mem *pcl_ival;
#else
	/** Host and device data for double
	 * precision */
	CUKR_DOUBLE *ph_dval, *pd_dval;
	
	/** Host and device data for single
	 * precision */
	CUKR_SINGLE *ph_sval, *pd_sval;
	
	/** Host and device data for quasi-double
	 * precision (tail part, since the head
	 * part will be stored in the members for
	 * single precision) */
	CUKR_SINGLE *ph_qvalt, *pd_qvalt;
	
	/** Host and device data for integer */
	CUKR_INT    *ph_ival, *pd_ival;
#endif
 
} *CUKR_VECTOR_SP;

/*
 * Function pointers for the functions which
 * have to be scheduled to threads
 */
typedef void        (*CUKR_VECSP_MALLOC)       (CUKR_VECTOR_SP, int, int);
typedef void        (*CUKR_VECSP_DESTROY)      (CUKR_VECTOR_SP*);
typedef void        (*CUKR_VECSP_TRANSFER)     (CUKR_VECTOR_SP, int, int);
typedef void        (*CUKR_VECSP_CLONE)        (CUKR_VECTOR_SP, CUKR_VECTOR_SP*, char*);
typedef CUKR_DOUBLE (*CUKR_VECSP_GET_ELEMENT)  (CUKR_VECTOR_SP, int, int);
typedef void        (*CUKR_VECSP_SET_ELEMENT)  (CUKR_VECTOR_SP, int, CUKR_DOUBLE, int);
typedef void        (*CUKR_VECSP_SET_FROMVEC_D)(CUKR_VECTOR_SP, int, int, CUKR_DOUBLE*, int);
typedef void        (*CUKR_VECSP_SET_FROMVEC_Q)(CUKR_VECTOR_SP, int, int, CUKR_SINGLE*, CUKR_SINGLE*, int);
typedef void        (*CUKR_VECSP_SET_FROMVEC_S)(CUKR_VECTOR_SP, int, int, CUKR_SINGLE*, int);
typedef void        (*CUKR_VECSP_SET_FROMVEC_I)(CUKR_VECTOR_SP, int, int, CUKR_INT*, int);
typedef void        (*CUKR_VECSP_GET_TOVEC_D)  (CUKR_VECTOR_SP, int, int, CUKR_DOUBLE*, int);
typedef void        (*CUKR_VECSP_GET_TOVEC_Q)  (CUKR_VECTOR_SP, int, int, CUKR_SINGLE*, CUKR_SINGLE*, int);
typedef void        (*CUKR_VECSP_GET_TOVEC_S)  (CUKR_VECTOR_SP, int, int, CUKR_SINGLE*, int);
typedef void        (*CUKR_VECSP_GET_TOVEC_I)  (CUKR_VECTOR_SP, int, int, CUKR_INT*, int);



				/*****************************************/
				//       CONSTRUCTOR / DESTRUCTOR        //
				/*****************************************/





/**
 * \brief Creates a vector of the specified
 *        size, on the CPU, with the specified
 *        type. Since by default the vectors are
 *        created on the CPU, the location need
 *        not be specified.
 * \param vec Pointer to the vector structure
 *        being created
 * \param n Vector size
 * \param type Vector type (double, qdouble,
 *        single, int) or default to match the
 *        solver precision
 * \param name The vector name
 */
void CukrVecspCreate(CUKR_VECTOR_SP* vec, CUKR_INT n, int type, char* name);

// ------------------------------------------------------------------- //

/**
 * \brief Creates a clone of a vector.
 *        The function currently copies
 *        the data in the highest precision
 *        available, and leaves the rest
 *        of the precisions invalidated, so
 *        they will be transformed/converted
 *        to only if needed.
 * \param src The source vector
 * \param dst Pointer to the destination vector
 * \param name The new name for the destination
 */
void CukrVecspCreateClone(CUKR_VECTOR_SP src, CUKR_VECTOR_SP* dst, char* name);

// ------------------------------------------------------------------- //

/**
 * \brief Frees all the memory used
 *        by the vector
 * \param vec Pointer to the vector to
 *        be destroyed
 */
void CukrVecspDestroy(CUKR_VECTOR_SP* vec);

// ------------------------------------------------------------------- //

/**
 * \brief Frees all the memory used
 *        by the data members, but leaves
 *        the vector structure intact
 * \param vec The vector
 */
void CukrVecspDestroyData(CUKR_VECTOR_SP vec);





				/*****************************************/
				//            MALLOC / FREE              //
				/*****************************************/




/**
 * \brief Allocate memory for a vector. The
 *        function checks if the pointer
 *        corresponding to that particular location
 *        and precision is NULL of not. If not NULL,
 *        on malloc will be performed. Therefore, this
 *        function can be called "just in case" to
 *        make sure memory allocation where it
 *        should be.
 * \param vec The vector structure being created
 * \param loc Vector location {CPU,GPU}
 * \param type Vector type {double, qdouble,
 *        single, int}
 */
void CukrVecspMalloc(CUKR_VECTOR_SP vec, int loc, int type);

// ------------------------------------------------------------------- //

/**
 * \brief Alloc memory for the data member
 *        of a vector corresponding to the
 *        desired precision, on the CPU. The
 *        function checks if the pointer
 *        corresponding to that particular location
 *        and precision is NULL of not. If npt NULL,
 *        on malloc will be performed. Therefore, this
 *        function can be called "just in case" to
 *        make sure memory allocation where it
 *        should be.
 * \param vec Pointer to the vector
 * \param type Vector type {float, double, int}
 */
void CukrVecspMallocCPU(CUKR_VECTOR_SP vec, int type);

// ------------------------------------------------------------------- //
/**
 * \brief Alloc memory for the data member
 *        of a vector corresponding to the
 *        desired precision, on the GPU. The
 *        function checks if the pointer
 *        corresponding to that particular location
 *        and precision is NULL of not. If not NULL,
 *        on malloc will be performed. Therefore, this
 *        function can be called "just in case" to
 *        make sure memory allocation where it
 *        should be.
 * \param vec Pointer to the vector
 * \param type Vector type {float, double, int}
 */
void CukrVecspMallocGPU(CUKR_VECTOR_SP vec, int type);

// ------------------------------------------------------------------- //

/**
 * \brief Free the element corresonding to the
 *        specified precision and location
 * \param vec Pointer to the vector structure
 * \param loc Vector location {CPU,GPU}
 * \param type Vector type {double,qdouble,single,int}
 */
void CukrVecspFree(CUKR_VECTOR_SP vec, int loc, int type);

// ------------------------------------------------------------------- //

/**
 * \brief Free memory allocated to the data
 		  member of a vector corresponding to the
 *        desired precision, on the CPU. The
 *        function checks if the pointer
 *        corresponding to that particular location
 *        and precision is NULL of not. If npt NULL,
 *        on malloc will be performed. Therefore, this
 *        function can be called "just in case" to
 *        make sure memory allocation where it
 *        should be.
 * \param vec Pointer to the vector
 * \param type Vector type {double,qdouble,single,int}
 */
void CukrVecspFreeCPU(CUKR_VECTOR_SP vec, int type);

// ------------------------------------------------------------------- //

/**
 * \brief Free memory allocated to the data
 		  member of a vector corresponding to the
 *        desired precision, on the GPU
 * \param vec Pointer to the vector
 * \param type Vector type {double,qdouble,single,int}
 */
void CukrVecspFreeGPU(CUKR_VECTOR_SP vec, int type);




				/*****************************************/
				//              TRANSFER                 //
				/*****************************************/




/**
 * \brief Changes the location of a vector
 *        by copying the data from one location
 *        to another. The status of the vector will
 *        be updated to reflect the new location.
 * \param vec Pointer to the vector
 * \param dest_loc Destination vector location (CPU,GPU)
 * \param dest_type Destination vector type
 *        {double,qdouble,single,int}
 * \param update_attrs If nonzero, the location and
 *        the precision of the vector will be updated
 *        after the transfer. A false value should only
 *        be used for operations like print, where one
 *        would want to print the elements from a
 *        different locatio (which requires transfering
 *        the data to the CPU) but wishes to see the
 *        original status of the vector, not the updated
 *        one.
 */
void CukrVecspTransfer(CUKR_VECTOR_SP vec, int dest_loc, int dest_type);




				/*****************************************/
				//               GET/SET                 //
				/*****************************************/




/**
 * \brief Returns the element found at the
 *        position <index> in the vector vec.
 * \param vec The vector
 * \param index The index where to put the value
 * \param type An existing data type in which to
 * 		  store the value or default to read from
 *        the highest available one
 * \return val The value of the element
 */
CUKR_DOUBLE CukrVecspGetElement(CUKR_VECTOR_SP vec, int index, int type);

// ------------------------------------------------------------------- //

/**
 * \brief Sets the element with index <index>
 *        to the value specified by <val>
 *        Currently, setting can only
 *        be done on the CPU, so data will be
 *        set on the CPU and the status set to
 *        there.
 * \param vec The vector
 * \param index The index where to put the value
 * \param val The value the vector
 *        elements are set to
 * \param type An existing data type in which to
 * 		  store the value or default for all
 *        available precisions
 */
void CukrVecspSetElement(CUKR_VECTOR_SP vec, int index, CUKR_DOUBLE val, int type);



/**
 * \todo  Should make a simple kernel that
 *        sets all the values of a vector to
 *        a specicific value, so that we can
 *        set directly on the GPU.
 * \brief Sets all the elements
 *        of the vector to <val>.
 *        Currently, setting can only
 *        be done on the CPU, so data will be
 *        set on the CPU and the status set to
 *        there.
 * \param vec The vector
 * \param val The value the vector
 *        elements are set to
 * \param type An existing data type in which to
 * 		  store the values or default for the
 *        highest one available
 */
void CukrVecspSetFromVal(CUKR_VECTOR_SP vec, CUKR_DOUBLE val, int type);

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets all elements of a
 *        vector from a double precision vector
 *        <val>. The vector must reside on the host.
 *        Since data
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecspSetFromVectorD(CUKR_VECTOR_SP vec, int index, int count, CUKR_DOUBLE* val, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets all elements of a
 *        vector from a quasi-double precision
 *        vector <val>.
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecspSetFromVectorQ(CUKR_VECTOR_SP vec, int index, int count, CUKR_SINGLE* hval, CUKR_SINGLE* tval, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets all elements of a
 *        vector from a single precision vector
 *        <val>.
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecspSetFromVectorS(CUKR_VECTOR_SP vec, int index, int count, CUKR_SINGLE* val, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets all elements of a
 *        vector from a integer vector <val>.
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecspSetFromVectorI(CUKR_VECTOR_SP vec, int index, int count, CUKR_INT* val, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to a double precision vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecspGetToVectorD(CUKR_VECTOR_SP vec, int index, int count, CUKR_DOUBLE* val, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to a quasi-double precision vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecspGetToVectorQ(CUKR_VECTOR_SP vec, int index, int count, CUKR_SINGLE* hval, CUKR_SINGLE* tval, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to a single precision vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecspGetToVectorS(CUKR_VECTOR_SP vec, int index, int count, CUKR_SINGLE* val, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to an integer vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecspGetToVectorI(CUKR_VECTOR_SP vec, int index, int count, CUKR_INT* val, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Sets all elements of the
 *        vector to random values, on the CPUs.
 *        The rest of the precisions and
 *        locations are invalidated.
 * \param vec The vector
 * \param min,max The bounds for the random number
 * \param seed The seed of the random number gen.
 * \param type Data type where to write the data
 *             or default to match the solver type
 */
void CukrVecspSetRandom(CUKR_VECTOR_SP vec, CUKR_DOUBLE min, CUKR_DOUBLE max, int seed, int type);






				/*****************************************/
				//                 I/O                   //
				/*****************************************/





/**
 * \brief Reads data into a previously
 *        created vector structure from
 *        an opened file stream. Can be used
 *        for example when reading data for
 *        a CSR matrix, which consists of
 *        three vectors one after another.
 *        Calling this function three times,
 *        once for each vector, would do the
 *        trick.
 * \param f The open file stream
 * \param type The type of data that is
 *        supposed to be read (as double,
 *        qdouble, single, int)
 */
void CukrVecspReadDataFromFile(CUKR_VECTOR_SP vec, FILE* f, int type);

// ------------------------------------------------------------------- //

/**
 * \brief Reads data into a previously
 *        created vector structure from
 *        a file.
 * \param filename The file name
 * \param type The type of data that is
 *        supposed to be read (as double,
 *        qdouble, single, int) or default
 *        to match the solver
 */
void CukrVecspReadDataFromFilename(CUKR_VECTOR_SP vec, char* filename, int type);

// ------------------------------------------------------------------- //

/**
 * \brief Write the values of a vector
 *        to an opened file stream.
 * \param vec The vector
 * \param f The open file stream
 */
void CukrVecspWriteDataToFile(CUKR_VECTOR_SP vec, FILE* f);

// ------------------------------------------------------------------- //

/**
 * \brief Write the values of a vector
 *        to an opened file stream.
 * \param vec The vector
 * \param f The file name
 */
void CukrVecspWriteDataToFilename(CUKR_VECTOR_SP vec, char* filename);

// ------------------------------------------------------------------- //

/**
 * \brief Prints the values of a vector
 * \param vec Pointer to the vector being
 *        printed
 */
void CukrVecspPrint(CUKR_VECTOR_SP vec);

// ------------------------------------------------------------------- //

/**
 * \brief Prints the values of a vector
 * \param vec The vector being printed
 * \param filename The full path of the file
 *        to write the vector to
 */
void CukrVecspPrintToFilename(CUKR_VECTOR_SP vec, char* filename);


// ------------------------------------------------------------------- //

/**
 * \brief Prints the values of a vector
 * \param vec Pointer to the vector being
 *        printed
 * \param f An open file stream to write to instead
 */
void CukrVecspPrintToFile(CUKR_VECTOR_SP vec, FILE* f);

// ------------------------------------------------------------------- //

/**
 * \brief Prints the values of a vector
 * \param vec Pointer to the vector being
 *        printed
 * \param f An open file stream to write to instead
 * \param indent An indentation string, to be inserted
 *        before every print.
 */
void CukrVecspPrintToFileWithIndent(CUKR_VECTOR_SP vec, FILE* f, char* indent);




#endif /*CUKR_VECTOR_SP_H_*/
