/**
 * @file mat_vec/vector.c
 * @author Serban Georgescu
 * @since 2008
 * 
 * Interface to the SP/MP/MPI vector
 * functions
 */

#include "vector.h"




				/*****************************************/
				//       CONSTRUCTOR / DESTRUCTOR        //
				/*****************************************/         





/**
 * \brief Creates a vector of the specified 
 *        size, on the CPU, with the specified 
 *        type. Since by default the vectors are
 *        created on the CPU, the location need
 *        not be specified.
 * \param vec Pointer to the vector structure 
 *        being created
 * \param n Vector size
 * \param type Vector type (double, qddouble, 
 *        single, int)
 * \param shtype Sharing type (distributed or shared)
 * \param name The vector name
 */
void CukrVecCreate(CUKR_VECTOR* vec, CUKR_INT n, int type, int shtype, char* name)
{
	#ifdef CUKR_MPI
		CukrVecmpiCreate(vec, n, type, shtype, name);
	#else
		#ifdef CUKR_MP
			CukrVecmpCreate(vec, n, type, shtype, name);
		#else
			CukrVecspCreate(vec, n, type, name);
		#endif
	#endif	
}

// ------------------------------------------------------------------- //

/**
 * \brief Creates a clone of a vector. 
 *        The function currently copies
 *        the data in the highest precision
 *        available, and leaves the rest
 *        of the precision invalidated, so
 *        they will be transformed/convered
 *        to only if needed.
 * \param src The source vector
 * \param dst Pointer to the destination vector
 * \param name The new name for the destination
 */
void CukrVecCreateClone(CUKR_VECTOR src, CUKR_VECTOR* dst, char* name)
{
	#ifdef CUKR_MPI
		CukrVecmpiCreateClone(src, dst, name);
	#else
		#ifdef CUKR_MP
			CukrVecmpCreateClone(src, dst, name);
		#else
			CukrVecspCreateClone(src, dst, name);
		#endif
	#endif	
}

// ------------------------------------------------------------------- //

/**
 * \brief Frees all the memory used 
 *        by the vector
 * \param vec Pointer to the vector to 
 *        be destroyed 
 */
void CukrVecDestroy(CUKR_VECTOR* vec)
{
	#ifdef CUKR_MPI
		CukrVecmpiDestroy(vec);
	#else
		#ifdef CUKR_MP
			CukrVecmpDestroy(vec);
		#else
			CukrVecspDestroy(vec);
		#endif
	#endif	
}





				/*****************************************/
				//               GET/SET                 // 
				/*****************************************/ 




/**
 * \brief Sets the element with index <index>
 *        to the value specified by <val> 
 *        Currently, setting can only
 *        be done on the CPU, so data will be 
 *        set on the CPU and the status set to
 *        there. 
 * \param vec The vector
 * \param index The index where to put the value
 * \param val The value the vector 
 *        elements are set to
 * \param type An existing data type in which to 
 * 		  store the value or default for all 
 *        available precisions
 */
void CukrVecSetElement(CUKR_VECTOR vec, int index, CUKR_DOUBLE val, int type)
{
	#ifdef CUKR_MPI
		CukrVecmpiSetElement(vec, index, val, type);
	#else
		#ifdef CUKR_MP
			CukrVecmpSetElement(vec, index, val, type);
		#else
			CukrVecspSetElement(vec, index, val, type);
		#endif
	#endif		
}


/**
 * \todo  Should make a simple kernel that
 *        sets all the values of a vector to
 *        a specicific value, so that we can
 *        set directly on the GPU. 
 * \brief Sets all the elements
 *        of the vector to <val>. 
 *        Currently, setting can only
 *        be done on the CPU, so data will be 
 *        set on the CPU and the status set to
 *        there. 
 * \param vec The vector
 * \param val The value the vector 
 *        elements are set to
 * \param type An existing data type in which to 
 * 		  store the values or default for the 
 *        highest one available
 */
void CukrVecSetFromVal(CUKR_VECTOR vec, CUKR_DOUBLE val, int type)
{
	#ifdef CUKR_MPI
		CukrVecmpiSetFromVal(vec, val, type);
	#else
		#ifdef CUKR_MP
			CukrVecmpSetFromVal(vec, val, type);
		#else
			CukrVecspSetFromVal(vec, val, type);
		#endif
	#endif	
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets count elements of a 
 *        vector from a double precision 
 *        vector.
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecSetFromVectorD(CUKR_VECTOR vec, int index, int count, CUKR_DOUBLE* val, int loc)
{
	#ifdef CUKR_MPI
		CukrVecmpiSetFromVectorD(vec, index, count, val, loc);
	#else
		#ifdef CUKR_MP
			CukrVecmpSetFromVectorD(vec, index, count, val, loc);
		#else
			CukrVecspSetFromVectorD(vec, index, count, val, loc);
		#endif
	#endif			
}
// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets count elements of a 
 *        vector from a quasi-double precision 
 *        vector.
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecSetFromVectorQ(CUKR_VECTOR vec, int index, int count, CUKR_SINGLE* hval, CUKR_SINGLE* tval, int loc)
{
	#ifdef CUKR_MPI
		CukrVecmpiSetFromVectorQ(vec, index, count, hval, tval, loc);
	#else
		#ifdef CUKR_MP
			CukrVecmpSetFromVectorQ(vec, index, count, hval, tval, loc);
		#else
			CukrVecspSetFromVectorQ(vec, index, count, hval, tval, loc);
		#endif
	#endif			
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets count elements of a 
 *        vector from a single precision vector.
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecSetFromVectorS(CUKR_VECTOR vec, int index, int count, CUKR_SINGLE* val, int loc)
{
	#ifdef CUKR_MPI
		CukrVecmpiSetFromVectorS(vec, index, count, val, loc);
	#else
		#ifdef CUKR_MP
			CukrVecmpSetFromVectorS(vec, index, count, val, loc);
		#else
			CukrVecspSetFromVectorS(vec, index, count, val, loc);
		#endif
	#endif			
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets count elements of a 
 *        vector from an integer vector.
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecSetFromVectorI(CUKR_VECTOR vec, int index, int count, CUKR_INT* val, int loc)
{
	#ifdef CUKR_MPI
		CukrVecmpiSetFromVectorI(vec, index, count, val, loc);
	#else
		#ifdef CUKR_MP
			CukrVecmpSetFromVectorI(vec, index, count, val, loc);
		#else
			CukrVecspSetFromVectorI(vec, index, count, val, loc);
		#endif
	#endif			
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to a double precision vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecGetToVectorD(CUKR_VECTOR vec, int index, int count, CUKR_DOUBLE* val, int loc)
{
	#ifdef CUKR_MPI
		CukrVecmpiGetToVectorD(vec, index, count, val, loc);
	#else
		#ifdef CUKR_MP
			CukrVecmpGetToVectorD(vec, index, count, val, loc);
		#else
			CukrVecspGetToVectorD(vec, index, count, val, loc);
		#endif
	#endif			
}
// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to a quasi-double precision vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecGetToVectorQ(CUKR_VECTOR vec, int index, int count, CUKR_SINGLE* hval, CUKR_SINGLE* tval, int loc)
{
	#ifdef CUKR_MPI
		CukrVecmpiGetToVectorQ(vec, index, count, hval, tval, loc);
	#else
		#ifdef CUKR_MP
			CukrVecmpGetToVectorQ(vec, index, count, hval, tval, loc);
		#else
			CukrVecspGetToVectorQ(vec, index, count, hval, tval, loc);
		#endif
	#endif			
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to a single precision vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecGetToVectorS(CUKR_VECTOR vec, int index, int count, CUKR_SINGLE* val, int loc)
{
{
	#ifdef CUKR_MPI
		CukrVecmpiGetToVectorS(vec, index, count, val, loc);
	#else
		#ifdef CUKR_MP
			CukrVecmpGetToVectorS(vec, index, count, val, loc);
		#else
			CukrVecspGetToVectorS(vec, index, count, val, loc);
		#endif
	#endif			
}	
}

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to an integer vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecGetToVectorI(CUKR_VECTOR vec, int index, int count, CUKR_INT* val, int loc)
{
	#ifdef CUKR_MPI
		CukrVecmpiGetToVectorI(vec, index, count, val, loc);
	#else
		#ifdef CUKR_MP
			CukrVecmpGetToVectorI(vec, index, count, val, loc);
		#else
			CukrVecspGetToVectorI(vec, index, count, val, loc);
		#endif
	#endif			
}

// ------------------------------------------------------------------- //

/**
 * \brief Sets all elements of the
 *        vector (in double) to random
 *        values, on the CPU. The rest of 
 * 		  the precisions and locations are 
 *        invalidated.
 * \param vec The vector
 * \param min,max The bounds for the random number
 * \param seed The seed of the random number generator
 * \param type An existing data type in which to 
 * 		  store the values or default for the highest
 *        one available.
 */
void CukrVecSetRandom(CUKR_VECTOR vec, CUKR_DOUBLE min, CUKR_DOUBLE max, int seed, int type)
{
	#ifdef CUKR_MPI
		CukrVecmpiSetRandom(vec, min, max, seed, type);
	#else
		#ifdef CUKR_MP
			CukrVecmpSetRandom(vec, min, max, seed, type);
		#else
			CukrVecspSetRandom(vec, min, max, seed, type);
		#endif
	#endif		
}





				/*****************************************/
				//                 I/O                   // 
				/*****************************************/ 




  
/**
 * \brief Prints the values of a vector
 * \param vec Pointer to the vector being
 *        printed
 */
void CukrVecPrint(CUKR_VECTOR vec)
{
	#ifdef CUKR_MPI
		CukrVecmpiPrint(vec);
	#else
		#ifdef CUKR_MP
			CukrVecmpPrint(vec);
		#else
			CukrVecspPrint(vec);
		#endif
	#endif		
}

// ------------------------------------------------------------------- //

/**
 * \brief Prints the values of a vector
 * \param vec The vector being printed
 * \param filename The full path of the file
 *        to write the vector to
 */
void CukrVecPrintToFilename(CUKR_VECTOR vec, char* filename)
{
	#ifdef CUKR_MPI
		CukrVecmpiPrintToFilename(vec, filename);
	#else
		#ifdef CUKR_MP
			CukrVecmpPrintToFilename(vec, filename);
		#else
			CukrVecspPrintToFilename(vec, filename);
		#endif
	#endif	
}


// ------------------------------------------------------------------- //

/**
 * \brief Reads data into a previously
 *        created vector structure from 
 *        an opened file stream. Can be used
 *        for example when reading data for
 *        a CSR matrix, which consists of
 *        three vectors one after another.
 *        Calling this function three times,
 *        once for each vector, would do the 
 *        trick.
 * \param filename File containing the data
 * \param type The type of data that is
 *        supposed to be read (as double,
 *        qdouble, single, int) or default
 *        for the highest precision
 */
void CukrVecReadDataFromFilename(CUKR_VECTOR vec, char* filename, int type)
{
	#ifdef CUKR_MPI
		CukrVecmpiReadDataFromFilename(vec, filename, type);
	#else
		#ifdef CUKR_MP
			CukrVecmpReadDataFromFilename(vec, filename, type);
		#else
			CukrVecspReadDataFromFilename(vec, filename, type);
		#endif
	#endif		
}

// ------------------------------------------------------------------- //

/**
 * \brief Write the values of a CUKR vector
 *        to file 
 * \param vec The vector
 * \param filename The file
 */
void CukrVecWriteDataToFilename(CUKR_VECTOR vec, char* filename)
{
	#ifdef CUKR_MPI
		CukrVecmpiWriteDataToFilename(vec, filename);
	#else
		#ifdef CUKR_MP
			CukrVecmpWriteDataToFilename(vec, filename);
		#else
			CukrVecspWriteDataToFilename(vec, filename);
		#endif
	#endif		
}
