/**
 * @file mat_vec/vector.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Interface to the SP/MP/MPI vector
 * functions
 */

#ifndef CUKR_VECTOR_H_
#define CUKR_VECTOR_H_

#include "../cukr.h"
#include "vector_defs.h"

/* Decide which include file to use,
 * depending on defines set at built time.*/

#ifdef CUKR_MPI
	#include "mpi/vector_mpi.h"
#else
	#ifdef CUKR_MP
		#include "mp/vector_mp.h"
	#else
		#include "sp/vector_sp.h"
	#endif
#endif





				/*****************************************/
				//       CONSTRUCTOR / DESTRUCTOR        //
				/*****************************************/         





/**
 * \brief Creates a vector of the specified 
 *        size, on the CPU, with the specified 
 *        type. Since by default the vectors are
 *        created on the CPU, the location need
 *        not be specified.
 * \param vec Pointer to the vector structure 
 *        being created
 * \param n Vector size
 * \param type Vector type (double, qddouble, 
 *        single, int) or default to match the
 *        solver precision
 * \param shtype Sharing type (distributed or shared)
 * \param name The vector name
 */
void CukrVecCreate(CUKR_VECTOR* vec, CUKR_INT n, int type, int shtype, char* name);

// ------------------------------------------------------------------- //

/**
 * \brief Creates a clone of a vector. 
 *        The function currently copies
 *        the data in the highest precision
 *        available, and leaves the rest
 *        of the precision invalidated, so
 *        they will be transformed/convered
 *        to only if needed.
 * \param src The source vector
 * \param dst Pointer to the destination vector
 * \param name The new name for the destination
 */
void CukrVecCreateClone(CUKR_VECTOR src, CUKR_VECTOR* dst, char* name);

// ------------------------------------------------------------------- //

/**
 * \brief Frees all the memory used 
 *        by the vector
 * \param vec Pointer to the vector to 
 *        be destroyed 
 */
void CukrVecDestroy(CUKR_VECTOR* vec);





				/*****************************************/
				//               GET/SET                 // 
				/*****************************************/ 




/**
 * \brief Returns the element found at the
 *        position <index> in the vector vec. 
 * \param vec The vector
 * \param index The index where to put the value
 * \param type An existing data type in which to 
 * 		  store the value or default to read from
 *        the highest available one
 * \return val The value of the element
 */
CUKR_DOUBLE CukrVecGetElement(CUKR_VECTOR vec, int index, int type);

// ------------------------------------------------------------------- //

/**
 * \brief Sets the element with index <index>
 *        to the value specified by <val> 
 *        Currently, setting can only
 *        be done on the CPU, so data will be 
 *        set on the CPU and the status set to
 *        there. 
 * \param vec The vector
 * \param index The index where to put the value
 * \param val The value the vector 
 *        elements are set to
 * \param type An existing data type in which to 
 * 		  store the value or default for all 
 *        available precisions
 */
void CukrVecSetElement(CUKR_VECTOR vec, int index, CUKR_DOUBLE val, int type);

// ------------------------------------------------------------------- //

/**
 * \todo  Should make a simple kernel that
 *        sets all the values of a vector to
 *        a specicific value, so that we can
 *        set directly on the GPU. 
 * \brief Sets all the elements
 *        of the vector to <val>. 
 *        Currently, setting can only
 *        be done on the CPU, so data will be 
 *        set on the CPU and the status set to
 *        there. 
 * \param vec The vector
 * \param val The value the vector 
 *        elements are set to
 * \param type An existing data type in which to 
 * 		  store the values or default for the 
 *        highest one available
 */
void CukrVecSetFromVal(CUKR_VECTOR vec, CUKR_DOUBLE val, int type);

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets count elements of a 
 *        vector from a double precision 
 *        vector.
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecSetFromVectorD(CUKR_VECTOR vec, int index, int count, CUKR_DOUBLE* val, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets count elements of a 
 *        vector from a quasi-double precision 
 *        vector.
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecSetFromVectorQ(CUKR_VECTOR vec, int index, int count, CUKR_SINGLE* hval, CUKR_SINGLE* tval, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets count elements of a 
 *        vector from a single precision vector.
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecSetFromVectorS(CUKR_VECTOR vec, int index, int count, CUKR_SINGLE* val, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads and sets count elements of a 
 *        vector from an integer vector.
 * \param vec The vector
 * \param index The index in the vector from where
 *        to start adding values
 * \param count The number of the values to be 
 *        added
 * \param val The vector values are read from
 * \param loc The location of the source vector
 */
void CukrVecSetFromVectorI(CUKR_VECTOR vec, int index, int count, CUKR_INT* val, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to a double precision vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecGetToVectorD(CUKR_VECTOR vec, int index, int count, CUKR_DOUBLE* val, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to a quasi-double precision vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecGetToVectorQ(CUKR_VECTOR vec, int index, int count, CUKR_SINGLE* hval, CUKR_SINGLE* tval, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to a single precision vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecGetToVectorS(CUKR_VECTOR vec, int index, int count, CUKR_SINGLE* val, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Reads count elements from a cukr vector
 *        to an integer vector.
 * \param vec The vector
 * \param index The index in vec from where
 *        to start reading values
 * \param count The number of values to be 
 *        read
 * \param val The destination vector
 * \param loc The location of the dest. vector
 */
void CukrVecGetToVectorI(CUKR_VECTOR vec, int index, int count, CUKR_INT* val, int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Sets all elements of the
 *        vector to random values, on the CPUs. 
 *        The rest of the precisions and 
 *        locations are invalidated.
 * \param vec The vector
 * \param min,max The bounds for the random number
 * \param seed The seed of the random number gen.
 * \param type Data type where to write the data
 *             or default to match the solver type
 */
void CukrVecSetRandom(CUKR_VECTOR vec, CUKR_DOUBLE min, CUKR_DOUBLE max, int seed, int type);




				/*****************************************/
				//                 I/O                   // 
				/*****************************************/ 


// ------------------------------------------------------------------- //

/**
 * \brief Reads data into a previously
 *        created vector structure from 
 *        an opened file stream
 * \param filename File containing the data
 * \param type The type of data that is
 *        supposed to be read (as double,
 *        qdouble, single, int) or default
 *        to match the solver
 */
void CukrVecReadDataFromFile(CUKR_VECTOR vec, char* filename, int type);

// ------------------------------------------------------------------- //

/**
 * \brief Reads data into a previously
 *        created vector structure from 
 *        the specified file name
 * \param filename File containing the data
 * \param type The type of data that is
 *        supposed to be read (as double,
 *        qdouble, single, int) or default
 *        to match the solver
 */
void CukrVecReadDataFromFilename(CUKR_VECTOR vec, char* filename, int type);

// ------------------------------------------------------------------- //
  
/**
 * \brief Prints the values of a vector
 * \param vec Pointer to the vector being
 *        printed
 */
void CukrVecPrint(CUKR_VECTOR vec);

// ------------------------------------------------------------------- //

/**
 * \brief Prints the values of a vector
 * \param vec The vector being printed
 * \param filename The full path of the file
 *        to write the vector to
 */
void CukrVecPrintToFilename(CUKR_VECTOR vec, char* filename);


// ------------------------------------------------------------------- //

/**
 * \brief Write the values of a CUKR vector
 *        to file 
 * \param vec The vector
 * \param filename The file
 */
void CukrVecWriteDataToFilename(CUKR_VECTOR vec, char* filename);



#endif /*CUKR_VECTOR_H_*/
