/**
 * @file monitors/reports/blasx_report.c
 * @author Serban Georgescu
 * @since 2008 
 * 
 * Conains routine for reporting the 
 * perfomance of the BLAS subsystem.
 * 
 */
 
#include <stdio.h>
#include <stdlib.h>
#include "../../blas/blas.h" 
#include "../../util/log.h"
#include "../../db/db.h"
#include "reports.h"

/* Get access to the option database */
extern CukrDB cukr_db;

// ------------------------------------------------------------------- //

/*
 * \brief Write the performance report for
 *        all blas functions, for the specified
 *        thread (uses thread-level counters)
 * \param fout Pointer to the opened output
 *             file
 * \param blas The BLAS holding the data
 * \param tid The thread to which the perf. 
 *            counters belong
 */
void write_report_blas_performance(FILE* fout, CUKR_BLAS blas, int tid)
{ 
	CUKR_DOUBLE flops, elapsed;
	int op, pr;
	char ops[10][10]  = {"DOT", "AXPY", "AYPX", "SCAL", "COPY", "SPMV_CSR"};
	char prec[10][10] = {"D", "Q", "S"};
	
	/* For each operation */
	for (op = 0; op <= 5; op++)
		/* For each precision */
		for (pr = 0; pr <= 2; pr++) {
			
			/* Get performance */
			flops   = blas->counters[tid][CUKR_BLAS_IDX_FLOPS][op][pr];
			elapsed = blas->counters[tid][CUKR_BLAS_IDX_TIME] [op][pr];
			
			/* Write to file */	
			if (flops != 0 && elapsed != 0)
				fprintf (fout, "%s%s=%g\n", prec[pr], ops[op], 1e-3 * flops / elapsed);
		}
} 

/*
 * \brief Write the performance report for
 *        all blas functions, for the current
 *        node (uses node-level counters)
 * \param fout Pointer to the opened output
 *             file
 * \param blas The BLAS holding the data
 */
void write_report_blas_performance_mp(FILE* fout, CUKR_BLAS blas)
{ 
	CUKR_DOUBLE flops, elapsed;
	int op;
	char ops[10][10] = {"DOT", "AXPY", "AYPX", "SCAL", "COPY", "SPMV_CSR"};
	
	/* For each operation */
	for (op = 0; op <= 5; op++) {		
		/* Get performance */
		flops   = blas->counters_mp[CUKR_BLAS_IDX_FLOPS][op];
		elapsed = blas->counters_mp[CUKR_BLAS_IDX_TIME] [op];
		
		/* Write to file */	
		if (flops != 0 && elapsed != 0)
			fprintf (fout, "%s=%g\n", ops[op], 1e-3 * flops / elapsed);
	}
} 

/*
 * \brief Write the performance report for
 *        all blas functions, for the entire
 *        cluster (uses cluster-level counters)
 * \param fout Pointer to the opened output
 *             file
 * \param blas The BLAS holding the data
 */
void write_report_blas_performance_mpi(FILE* fout, CUKR_BLAS blas)
{ 
	CUKR_DOUBLE flops, elapsed;
	int op;
	char ops[10][10] = {"DOT", "AXPY", "AYPX", "SCAL", "COPY", "SPMV_CSR"};
	
	/* For each operation */
	for (op = 0; op <= 5; op++) {		
		/* Get performance */
		flops   = blas->counters_mpi[CUKR_BLAS_IDX_FLOPS][op];
		elapsed = blas->counters_mpi[CUKR_BLAS_IDX_TIME] [op];
		
		/* Write to file */	
		if (flops != 0 && elapsed != 0)
			fprintf (fout, "%s=%g\n", ops[op], 1e-3 * flops / elapsed);
	}
} 

/*
 * \brief Write a report of the specified
 *        type for all BLAS functions, for
 *        the specified thread 
 *        (uses thread-level counters)
 * \param fout Pointer to the opened output
 *             file
 * \param blas The BLAS holding the data
 * \param tid The thread to which the perf. 
 *            counters belong
 * \param report_idx The index of the report
 *                   to be written 
 */
void write_report_blas(FILE* fout, CUKR_BLAS blas, int tid, int report_idx)
{ 
	int op, pr;
	char ops[10][10] = {"DOT", "AXPY", "AYPX", "SCAL", "COPY", "SPMV_CSR"};
	char prec[10][10] = {"D", "Q", "S"};
	
	/* For each operation */
	for (op = 0; op <= 5; op++)
		/* For each precision */
		for (pr = 0; pr <= 2; pr++) {
			
			/* Get value */
			float val = blas->counters[tid][report_idx][op][pr];
			
			/* Write to file, if nonzero */	
			if (val)
				fprintf (fout, "%s%s=%g\n", prec[pr], ops[op], val);
		}
} 

/*
 * \brief Write a report of the specified
 *        type for all BLAS functions, for
 *        the current node 
 *        (uses node-level counters)
 * \param fout Pointer to the opened output
 *             file
 * \param blas The BLAS holding the data
 * \param report_idx The index of the report
 *                   to be written 
 */
void write_report_blas_mp(FILE* fout, CUKR_BLAS blas, int report_idx)
{ 
	int op;
	char ops[10][10] = {"DOT", "AXPY", "AYPX", "SCAL", "COPY", "SPMV_CSR"};
	
	/* For each operation */
	for (op = 0; op <= 5; op++) {
		/* Get value */
		float val = blas->counters_mp[report_idx][op];
		
		/* Write to file, if nonzero */	
		if (val)
			fprintf (fout, "%s=%g\n", ops[op], val);
	}
	
	
	/* Add communication report to timing report */	
	if (report_idx == CUKR_BLAS_IDX_TIME) 
	{    	                             
		if (blas->comm_inner != 0)
			fprintf (fout, "COMM_INNER=%g\n", blas->comm_inner);	    	                             	    	                             
	}
} 

/*
 * \brief Write a report of the specified
 *        type for all BLAS functions, for
 *        the entire cluster 
 *        (uses node-level counters)
 * \param fout Pointer to the opened output
 *             file
 * \param blas The BLAS holding the data
 * \param report_idx The index of the report
 *                   to be written 
 */
void write_report_blas_mpi(FILE* fout, CUKR_BLAS blas, int report_idx)
{ 
	int op;
	char ops[10][10] = {"DOT", "AXPY", "AYPX", "SCAL", "COPY", "SPMV_CSR"};
	
	/* For each operation */
	for (op = 0; op <= 5; op++) {
		/* Get value */
		float val = blas->counters_mpi[report_idx][op];
		
		/* Write to file, if nonzero */	
		if (val)
			fprintf (fout, "%s=%g\n", ops[op], val);
	}
	
	/* Add communication report to timing report */	
	if (report_idx == CUKR_BLAS_IDX_TIME) 
	{    	                             
		if (blas->comm_inner != 0)
			fprintf (fout, "COMM_INNER=%g\n", blas->comm_inner);
			
		if (blas->comm_outer != 0)
			fprintf (fout, "COMM_OUTER=%g\n", blas->comm_outer);	    	                             	    	                             
	}
} 

// ------------------------------------------------------------------- //

/*
 * \brief Reads the thread-leve blas counters 
 *        and generates a performance report.
 * \params blas The BLAS whose performance is being
 *              reported
 * \params tid The id of the thread for which 
 *             the performance is being reported
 * \params filename The file where the performance 
 *                  report is to be saved. If null, 
 *                  the report will go to the display.
 */    
void CukrReportBlasPerformance(CUKR_BLAS blas, int tid, char* filename)
{
	FILE *fout;
	
	/* If the filename is not null, try to open the file 
	 * from node0*/ 
	
	if (filename != NULL)
	{
		if (!cukr_db.mpi_rank)
			if ( (fout = fopen(filename, "w")) == NULL) {
				CUKR_FATALX("Could not open file for reporting: %s\n", filename);
			}
	}
	/* If null, output to stdout */	
	else
	  fout = stdout;
	
	
	/* Write to file only on node 0 */
	if (!cukr_db.mpi_rank)
	{  
		/* Write the name of the BLAS */
		fprintf(fout, "BLAS=%s\n", blas->name);
		
		/* Write performance report */
		fprintf (fout, "\n\nPERFORMANCE REPORT (in Mflop/s) \n\n");
		write_report_blas_performance(fout, blas, tid);
		
		/* Write timing report */
		fprintf (fout, "\n\nTIMING REPORT (in msec) \n\n");
		write_report_blas(fout, blas, tid, CUKR_BLAS_IDX_TIME);
		
		/* Write floating-point operations report */
		fprintf (fout, "\n\nFLOATING-POINT OPERATIONS REPORT (in flops ) \n\n");
		write_report_blas(fout, blas, tid, CUKR_BLAS_IDX_FLOPS);
		
		/* Write integer operations report */
		fprintf (fout, "\n\nINTEGER OPERATIONS REPORT (in iops ) \n\n");
		write_report_blas(fout, blas, tid, CUKR_BLAS_IDX_INTOPS);
		
		/* Write float loads report */
		fprintf (fout, "\n\nFLOAT LOADS REPORT (in loads ) \n\n");
		write_report_blas(fout, blas, tid, CUKR_BLAS_IDX_FLLOADS);
		
		/* Write integer loads report */
		fprintf (fout, "\n\nINTEGER LOADS REPORT (in loads ) \n\n");
		write_report_blas(fout, blas, tid, CUKR_BLAS_IDX_INTLOADS);
		
		/* Write float stores report */
		fprintf (fout, "\n\nFLOAT STORES REPORT (in stores ) \n\n");
		write_report_blas(fout, blas, tid, CUKR_BLAS_IDX_FLSTORES);
		
		/* Write integer loads report */
		fprintf (fout, "\n\nINTEGER LOADS REPORT (in loads ) \n\n");
		write_report_blas(fout, blas, tid, CUKR_BLAS_IDX_INTSTORES);
		 
		fclose(fout);
	}
} 

// ------------------------------------------------------------------- //

/*
 * \brief Reads the node-leve blas counters 
 *        and generates a performance report.
 * \params blas The BLAS whose performance is being
 *              reported
 * \params filename The file where the performance 
 *                  report is to be saved. If null, 
 *                  the report will go to the display.
 */    
void CukrReportBlasPerformanceMP(CUKR_BLAS blas, char* filename)
{
	FILE *fout;
	
	/* If the filename is not null, try to open the file 
	 * from node0*/ 
	
	if (filename != NULL)
	{
		if (!cukr_db.mpi_rank)
			if ( (fout = fopen(filename, "w")) == NULL) {
				CUKR_FATALX("Could not open file for reporting: %s\n", filename);
			}
	}
	/* If null, output to stdout */	
	else
	  fout = stdout;
	
	
	/* Write to file only on node 0 */
	if (!cukr_db.mpi_rank)
	{  
		/* Write the name of the BLAS */
		fprintf(fout, "BLAS=%s\n", blas->name);
		
		/* Write performance report */
		fprintf (fout, "\n\nPERFORMANCE REPORT (in Mflop/s) \n\n");
		write_report_blas_performance_mp(fout, blas);
		
		/* Write timing report */
		fprintf (fout, "\n\nTIMING REPORT (in msec) \n\n");
		write_report_blas_mp(fout, blas, CUKR_BLAS_IDX_TIME);
		 
		fclose(fout);
	}
} 

// ------------------------------------------------------------------- //

/*
 * \brief Reads the cluster-leve blas counters 
 *        and generates a performance report.
 * \params blas The BLAS whose performance is being
 *              reported
 * \params filename The file where the performance 
 *                  report is to be saved. If null, 
 *                  the report will go to the display.
 */    
void CukrReportBlasPerformanceMPI(CUKR_BLAS blas, char* filename)
{
	FILE *fout;
	
	/* If the filename is not null, try to open the file 
	 * from node0*/ 
	
	if (filename != NULL)
	{
		if (!cukr_db.mpi_rank)
			if ( (fout = fopen(filename, "w")) == NULL) {
				CUKR_FATALX("Could not open file for reporting: %s\n", filename);
			}
	}
	/* If null, output to stdout */	
	else
	  fout = stdout;
	
	
	/* Write to file only on node 0 */
	if (!cukr_db.mpi_rank)
	{  
		/* Write the name of the BLAS */
		fprintf(fout, "BLAS=%s\n", blas->name);
		
		/* Write performance report */
		fprintf (fout, "\n\nPERFORMANCE REPORT (in Mflop/s) \n\n");
		write_report_blas_performance_mpi(fout, blas);
		
		/* Write timing report */
		fprintf (fout, "\n\nTIMING REPORT (in msec) \n\n");
		write_report_blas_mpi(fout, blas, CUKR_BLAS_IDX_TIME);
		 
		fclose(fout);
	}
} 
