/**
 * @file monitors/reports/reports.h
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains the reporting routines. 
 * 
 * Reporting routines are called at the end
 * of the run, to extract performance data.
 */

#ifndef REPORTS_H_
#define REPORTS_H_

#include "../../blas/blas.h"
#include "../../solvers/solver.h"


/*
 * \brief Reads the thread-level blas counters 
 *        and generates a performance report.
 * \params blas The BLAS whose performance is being
 *              reported
 * \params tid The id of the thread for which 
 *             the performance is being reported
 * \params filename The file where the performance 
 *                  report is to be saved. If null, 
 *                  the report will go to the display.
 */    
void CukrReportBlasPerformance(CUKR_BLAS blas, int tid, char* filename);

// ------------------------------------------------------------------- //

/*
 * \brief Reads the thread-level blasx counters  
 *        and generates a performance report.
 * \params blas The BLASX whose performance is being
 *              reported
 * \params tid The id of the thread for which 
 *             the performance is being reported
 * \params filename The file where the performance 
 *                  report is to be saved. If null, 
 *                  the report will go to the display.
 */    
void CukrReportBlasxPerformance(CUKR_BLASX blasx, int tid, char* filename);

// ------------------------------------------------------------------- //

/*
 * \brief Reads the node-level blas counters 
 *        and generates a performance report.
 * \params blas The BLAS whose performance is being
 *              reported
 * \params filename The file where the performance 
 *                  report is to be saved. If null, 
 *                  the report will go to the display.
 */    
void CukrReportBlasPerformanceMP(CUKR_BLAS blas, char* filename);

// ------------------------------------------------------------------- //

/*
 * \brief Reads the node-level blasx counters 
 *        and generates a performance report.
 * \params blas The BLASX whose performance is being
 *              reported
 * \params filename The file where the performance 
 *                  report is to be saved. If null, 
 *                  the report will go to the display.
 */    
void CukrReportBlasxPerformanceMP(CUKR_BLASX blasx, char* filename);

// ------------------------------------------------------------------- //

/*
 * \brief Reads the cluster-level blas counters 
 *        and generates a performance report.
 * \params blas The BLAS whose performance is being
 *              reported
 * \params filename The file where the performance 
 *                  report is to be saved. If null, 
 *                  the report will go to the display.
 */    
void CukrReportBlasPerformanceMPI(CUKR_BLAS blas, char* filename);

// ------------------------------------------------------------------- //

/*
 * \brief Reads the cluster-level blasx counters 
 *        and generates a performance report.
 * \params blas The BLASX whose performance is being
 *              reported
 * \params filename The file where the performance 
 *                  report is to be saved. If null, 
 *                  the report will go to the display.
 */    
void CukrReportBlasxPerformanceMPI(CUKR_BLASX blas, char* filename);

// ------------------------------------------------------------------- //

/**
 * \brief Generate solver report
 * \param solver The solver
 * \param filename The name of the file
 *        in which to write the report
 */
void CukrReportSolver(CUKR_SOLVER solver, char* filename);

// ------------------------------------------------------------------- //

/**
 * \brief Generate runtime report
 * \param filename The name of the file
 *        in which to write the report
 */
void CukrReportRuntime(char* filename);

#endif /*REPORTS_H_*/
