/**
 * @file monitors/reports/blas_report.c
 * @author Serban Georgescu
 * @since 2008 
 * 
 * Conains routine for reporting the 
 * perfomance and behaviour of the solver
 * 
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "../../blas/blas.h" 
#include "../../util/log.h"
#include "../../db/db.h"
#include "reports.h"

/** Get access to database */
extern CukrDB cukr_db;


// ------------------------------------------------------------------- //

/**
 * \brief Generate solver report
 * \param solver The solver
 */
void CukrReportSolver(CUKR_SOLVER solver, char* filename)
{
	FILE *fout;
	
	/* 
	 * Open the report file 
	 */
	 
	/* If the filename is not null, try to open the file */ 
	if (filename != NULL)
	{	
		if (!cukr_db.mpi_rank)
			if ( (fout = fopen(filename, "w")) == NULL) {
				CUKR_FATALX("Could not open file for reporting: %s\n", filename);
			}
	}
	/* If null, output to stdout */	
	else
	  fout = stdout;
	  
	/*
	 * Write about the solver behaviour
	 */
	
	/* Output only at rank 0 */
	if (!cukr_db.mpi_rank) {
		 
		/* Solver mode */
		if (solver->itref) 
			fprintf(fout, "SOLVER_MODE:ITREF\n");
		else
			fprintf(fout, "SOLVER_MODE:NORMAL\n");
		
		/* Converted or diverged */
		if (solver->itref) {
			if (solver->reason == CUKR_DIVERGED_DTOL)
				fprintf(fout, "CONVERGENCE_INNER:DIVERGED_DTOL\n");
			else if (solver->reason == CUKR_DIVERGED_ITS)
				fprintf(fout, "CONVERGENCE_INNER:DIVERGED_ITS\n");
			else if (solver->reason == CUKR_CONVERGED_RTOL)
				fprintf(fout, "CONVERGENCE_INNER:CONVERGED_RTOL\n");
			else if (solver->reason == CUKR_CONVERGED_ABSTOL)
				fprintf(fout, "CONVERGENCE_INNER:CONVERGED_ABSTOL\n");
				
			if (solver->reason_out == CUKR_DIVERGED_DTOL)
				fprintf(fout, "CONVERGENCE_OUTER:DIVERGED_DTOL\n");
			else if (solver->reason_out == CUKR_DIVERGED_ITS)
				fprintf(fout, "CONVERGENCE_OUTER:DIVERGED_ITS\n");
			else if (solver->reason_out == CUKR_CONVERGED_RTOL)
				fprintf(fout, "CONVERGENCE_OUTER:CONVERGED_RTOL\n");
			else if (solver->reason_out == CUKR_CONVERGED_ABSTOL)
				fprintf(fout, "CONVERGENCE_OUTER:CONVERGED_ABSTOL\n");		
		}
		else {
			if (solver->reason == CUKR_DIVERGED_DTOL)
				fprintf(fout, "CONVERGENCE:DIVERGED_DTOL\n");
			else if (solver->reason == CUKR_DIVERGED_ITS)
				fprintf(fout, "CONVERGENCE:DIVERGED_ITS\n");
			else if (solver->reason == CUKR_CONVERGED_RTOL)
				fprintf(fout, "CONVERGENCE:CONVERGED_RTOL\n");
			else if (solver->reason == CUKR_CONVERGED_ABSTOL)
				fprintf(fout, "CONVERGENCE:CONVERGED_ABSTOL\n");
		}
		
		/* Write iteration count */
		if (solver->itref) {
			fprintf(fout, "INNER_ITS:%g\n", solver->counters[SOLVER_IDX_INNER + SOLVER_OFFSET_ITS]);
			fprintf(fout, "OUTER_ITS:%g\n", solver->counters[SOLVER_IDX_OUTER + SOLVER_OFFSET_ITS]);
		}	  
		else
			fprintf(fout, "ITS:%g\n", solver->counters[SOLVER_IDX_NORMAL + SOLVER_OFFSET_ITS]);
			
		/* Write elapsed time */
		CUKR_DOUBLE elapsed = 0;
		if (solver->itref) {
			elapsed = solver->counters[SOLVER_IDX_INNER + SOLVER_OFFSET_TIME] +
			          solver->counters[SOLVER_IDX_OUTER + SOLVER_OFFSET_TIME];
			fprintf(fout, "INNER_TIME:%g\n", solver->counters[SOLVER_IDX_INNER + SOLVER_OFFSET_TIME]);
			fprintf(fout, "OUTER_TIME:%g\n", solver->counters[SOLVER_IDX_OUTER + SOLVER_OFFSET_TIME]);
		}	  
		else {
			elapsed = solver->counters[SOLVER_IDX_NORMAL + SOLVER_OFFSET_TIME];
			fprintf(fout, "TIME:%g\n", solver->counters[SOLVER_IDX_NORMAL + SOLVER_OFFSET_TIME]);
		}
		
		/* Write final residual */
		if (solver->itref)
			fprintf(fout, "FINAL_RES_NRM:%g\n", sqrt(solver->rnorm2_out / solver->rnorm02_out)); 
		else
			fprintf(fout, "FINAL_RES_NRM:%g\n", sqrt(solver->rnorm2 / solver->rnorm02 )); 
		
		/*
		 * Write about the solver performance
		 */
		 
		fprintf(fout, "MFLOPS:%g\n", 1e-3 * solver->flops / elapsed); 
		fclose(fout);	
	
	}
}
