/**
 * @file pc/jacobi_pc.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines the actual implementation
 * of the Jacobi type preconditioner 
 * 
 * The solver is written using the
 * abstract BLAS interface, therefore
 * it runs both on CPU, GPU or using
 * both in combination.
 */

#define LOGGER_NAME "CUKr.pc.jacobi_pc"

#include <stdlib.h>
#include <stdio.h>
#ifndef __MACH__
	#include <malloc.h>
#endif
#include "jacobi_pc.h"
#include "../../blasx/blasx1.h"
#include "../../util/log.h"

/*
 * \brief Setup the preconditioner
 * \param pc Preconditioner context
 */
void jacobi_setup(CUKR_PC pc)
{
	char buffer[255];
	
	/* Alloc memory for the structure */
	if ( (pc->data = (JACOBI_PC_DATA)malloc(sizeof(struct _JACOBI_PC_DATA))) == NULL) {
		CUKR_FATALX("Error allocating memory for the Jacobi "\
					"data structure for prec. %s", pc->name);
	}
	JACOBI_PC_DATA data = (JACOBI_PC_DATA)(pc->data);
	
	/*
	 * Set the name 
	 */
	
	sprintf(pc->name, "%s", "Jacobi");  
	 
	
	/* 
	 * Get the diagonal of the system matrix
	 * and put it into the data struct 
	 */
	 
	/* Create the diagonal vector */
	sprintf(buffer, "%s.diag", pc->solver->mat->name);
	data->invdiag = NULL;
	CukrVecCreate(&(data->invdiag), pc->solver->rhs->n, CUKR_TYPE_DOUBLE, CUKR_SHTYPE_DISTRIBUTED, buffer);  
	
	/* Get the diagonal from the matrix
	 * in inverted form */
	CukrMatGetDiag(pc->solver->mat, data->invdiag, 0, CUKR_TRUE); 
}

/*
 * \brief Setup the preconditioner
 * \param pc Preconditioner context
 * \param vec The vector subjected to the
 *        prec.
 * \param res The result vector
 */
void jacobi_apply(CUKR_PC pc, CUKR_VECTOR vec, CUKR_VECTOR res)
{
	/* Access data store */
	JACOBI_PC_DATA data = (JACOBI_PC_DATA)(pc->data);
	
	/* Pointwise product between the
	 * vector and the inverted diagonal */ 
	CukrPWPR(pc->solver->blas, vec, data->invdiag, res);
	pc->solver->flops += vec->n;

}

/*
 * \brief Setup the preconditioner from options
 * \param pc Preconditioner context
 */
void jacobi_setfromopts(CUKR_PC pc)
{
	; 
}


/*
 * \brief Clean-up and restroy the prec.
 * \param pc Preconditioner context
 */
void jacobi_destroy(CUKR_PC pc)
{
	/* Log */
	LOG(DEBUG, "Destroying Jacobi preconditioner implementation");
	
	/* Access data store */
	JACOBI_PC_DATA data = (JACOBI_PC_DATA)(pc->data);
	
	/* Destroy the diagonal vector */
	CukrVecDestroy(&data->invdiag);
	
	/* Free the data store */
	free(data);	
	
	/* Log */
	LOG(DEBUG, "Jacobi preconditioner implementation destroyed");
}

/**
 * \brief Bind the Jacobi prec to the
 *        preconditioner context
 * \param pc The prec. context
 */
void pc_create_jacobi(CUKR_PC pc)
{
	pc->ops.setfromopts = jacobi_setfromopts;
	pc->ops.setup = jacobi_setup;
	pc->ops.apply = jacobi_apply;
	pc->ops.destroy = jacobi_destroy;
}
 
