/**
 * @file pc/pc.c
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines the CUKR_PC abstract
 * interace.
 * 
 * The CUKR_PC structure stores all
 * member variables (characteristic) and
 * function pointers necessary to define
 * a preconditioner. Since all prec. must 
 * implement the functions defined here,
 * the class functions as a prec. 
 * interface. 
 */

#define LOGGER_NAME "CUKr.pc.pc"

#include <stdio.h>
#include <stdlib.h> 
#ifndef __MACH__
	#include <malloc.h>
#endif
#include <string.h>
#include "../util/log.h"
#include "../util/utils.h"
#include "../db/db.h"
#include "pc.h"
#include "jacobi/jacobi_pc.h"

//========================== FORWARD DECLARATIONS ========================= //

void CukrPcSetOptions(CUKR_PC pc);



//============================= PUBLIC FUNCTIONS ========================== //


/**
 * \brief Create a prec structure. The 
 *        actual prec type is decided 
 *        through command-line options.
 * \param pc Pointer to the prec.
 *        object being created 
 * \param solver The parent solver 
 * \param name Prec. name
 */
void CukrPcCreate(CUKR_PC *pc, CUKR_SOLVER solver)
{
	int i;
	
	/* 
	 * Validate parameters 
	 */
	
	/* Log */
	LOG(DEBUG, "Creating preconditioner structure");          
	
	/*
	 * Alloc memory for the prec
	 * Set the name (first thing after malloc)
	 */
	 
	/* Alloc memory for the solver structure*/
	if ( ((*pc) = (CUKR_PC)malloc(sizeof(struct _CUKR_PC))) == NULL) {
		CUKR_FATAL("Error allocating memory for preconditioner object "\
					"while creating preconditioner");
	} 
		
	/* Set the parent solver */
	(*pc)->solver = solver;
	
	/* Set the setup flag to false */
	(*pc)->setted_up = CUKR_FALSE;
	
	/* Null the data member */
	(*pc)->data = NULL;
	
	/*
	 * Read options from the options database.
	 * It shoud be done now, since depending on
	 * some options (e.g. itref) additional
	 * memory needs to be allocated
	 */ 
	 
	/* Read general solver options from */
	CukrPcSetOptions(*pc);
	 
	/* Zero the counters */
	for (i = 0; i < CUKR_PC_CNTR_VECTOR_LENGTH; i++)
		(*pc)->counters[i] = 0;
	
}

// ------------------------------------------------------------------- //
                   
                        
/**
 * \brief Apply the preconditioner
 * \param pc The prec. object 
 * \param vec The source vector
 * \param res The result of the prec. operation
 */
void CukrPcApply(CUKR_PC pc, CUKR_VECTOR vec, CUKR_VECTOR res)
{
	/* If no preconditioner, the result will
	 * be the input */
	if (pc->pc_type == CUKR_PC_TYPE_NONE) {
		return;
	} 
	
	/* If the preconditioner is not yet
	 * set up, set it up */
	if (!pc->setted_up) {
		(pc->ops.setfromopts)(pc);
		(pc->ops.setup)(pc);
		 pc->setted_up = CUKR_TRUE;
	}
	/* Now apply it */
	(pc->ops.apply)(pc, vec, res); 
}

// ------------------------------------------------------------------- //

 
/**
 * \brief Destroy a prec. object
 * \param pc The pc object 
 */
void CukrPcDestroy(CUKR_PC* pc)
{
		
	/* Copy the name and tyoe, so they don't 
	 * get destroyed */
	char name[255];
	memcpy(name, (*pc)->name, 255);
	int pc_type = (*pc)->pc_type;
	
	
	if (pc_type != CUKR_PC_TYPE_NONE) {
		LOGX(DEBUG, "Destroying preconditioner %s", name);
	}
	else {
		LOG(DEBUG, "Destroying preconditioner structure");
	}
	
	/* Destroy the prec-specific stuff
	 * if there is something */
	if (pc_type != CUKR_PC_TYPE_NONE)
		((*pc)->ops.destroy)(*pc); 
	
	/* Destroy the solver structure */
	free(*pc);
	
	/* Log */
	if (pc_type != CUKR_PC_TYPE_NONE) {
		LOGX(DEBUG, "Preconditioner %s destroyed", name);
	}
	else {
		LOG(DEBUG, "Preconditioner structure destroyed");
	}
}


   //============================ PRIVATE FUNCTIONS =========================== //
   
/**
 * \brief Read prec. options from optiondb
 *        and set them in the prec. structure
 * \param solver The prec. object
 */
void CukrPcSetOptions(CUKR_PC pc)   
{
	/*
	 * Find the desired prec. in the
	 * option database and bind it
	 */
	 
	 const char* pc_str = CukrDBGetS(CUKR_DB_PC);
	 
	 /* If none, set it as such */
	 if (!CukrStricmp("NONE", pc_str))
	 {
		LOG(DEBUG, "Not using preconditioner");
		pc->pc_type = CUKR_PC_TYPE_NONE;
	 }
	 
	 /* If diagonal, set it as such */
	 else if (!CukrStricmp("JACOBI", pc_str))
	 {
		LOG(DEBUG, "Using JACOBI preconditiioner");
		pc->pc_type = CUKR_PC_TYPE_JACOBI;
		pc_create_jacobi(pc);
	 }
	 
	 /* If not, error message */
	 else
	 {
	 	CUKR_FATALX("Unknown prec. type: %s", pc_str);
	 }	
}
   
