/**
 * @file pc/pc.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines the CUKR_PC abstract
 * interace.
 * 
 * The CUKR_PC structure stores all
 * member variables (characteristic) and
 * function pointers necessary to define
 * a preconditioner. Since all prec. must 
 * implement the functions defined here,
 * the class functions as a prec. 
 * interface. 
 */

#ifndef CUKR_PC_H_
#define CUKR_PC_H_

 
/* ========= Forward declarations for structures ======== */
typedef struct _CUKR_PC* CUKR_PC;
typedef struct _CUKR_PC_OPS CUKR_PC_OPS;
/* ======================================================= */


#include "../mat_vec/matrix.h"
#include "../mat_vec/vector.h"
#include "../solvers/solver.h" 


		// ======================== DEFINES ======================== //

/** No preconditioner       */
#define CUKR_PC_TYPE_NONE   0

/** Jacobi (diagonal) prec. */
#define CUKR_PC_TYPE_JACOBI 1


/** Total no. of performance
 * counters */
#define CUKR_PC_CNTR_VECTOR_LENGTH 256

/** #1. PC Setup counter */
#define CUKR_PC_OP_SETUP  0
/** #2. PC Apply counter */
#define CUKR_PC_OP_APPLY  100

/** #1. Elapsed time counter */
#define CUKR_PC_IDX_TIME   0
/** #2. GFlops counter       */
#define CUKR_PC_IDX_MFLOPS 1   



        // ================== INTERFACE DECLARATIONS =============== //


/* Define all the function pointers
 * to the function needed to implement
 * the solver */
  

/* Function pointers for prec. 
 * operations */ 
typedef void (*CUKR_PC_OPS_SETUP)(CUKR_PC);   
typedef void (*CUKR_PC_OPS_APPLY)(CUKR_PC, CUKR_VECTOR, CUKR_VECTOR); 
typedef void (*CUKR_PC_OPS_SETFROMOPTS)(CUKR_PC);
typedef void (*CUKR_PC_OPS_DESTROY)(CUKR_PC);  
 
/** Sturcture holding the prec. 
 * operations (function pointers to
 * to what each prec. has to implement)
 */
struct _CUKR_PC_OPS {
	CUKR_PC_OPS_SETUP       setup;
	CUKR_PC_OPS_APPLY       apply;
	CUKR_PC_OPS_SETFROMOPTS setfromopts;
	CUKR_PC_OPS_DESTROY     destroy;
}; 
 
/** The actual prec. interface, 
 * containing members and function
 * pointers 
 */ 
struct _CUKR_PC{
	
	// ------- MEMBERS --------- //
	
	/** Name */
	char name[255];
	
	/** Preconditioner type */
	int pc_type;
	
	/** Parent solver*/
	CUKR_SOLVER solver;
	
	/** Prec. operations */
	CUKR_PC_OPS ops; 
	
	/** Flag indicating whether the
	 * prec. is set up or not */
	int setted_up; 
	
	/** Performance counters */
	double counters[CUKR_PC_CNTR_VECTOR_LENGTH];
	
	/** Data holder */
	void *data;
	
};


        // ============== PRECONDITIONER RELATED FUNCTIONS ============== //
        
        
        

/* Functions common to all prec.
 * Basically, get or set some member
 * of the solver structure */
 
/**
 * \brief Create a prec structure. The 
 *        actual prec type is decided 
 *        through command-line options.
 * \param pc Pointer to the prec.
 *        object being created 
 * \param solver The parent solver 
 */
void CukrPcCreate(CUKR_PC *pc, CUKR_SOLVER solver);

// ------------------------------------------------------------------- //
                   
                        
/**
 * \brief Apply the preconditioner
 * \param pc The prec. object 
 * \param vec The source vector
 * \param res The result of the prec. operation
 */
void CukrPcApply(CUKR_PC pc, CUKR_VECTOR vec, CUKR_VECTOR res);

// ------------------------------------------------------------------- //


/**
 * \brief Destroy a prec. object
 * \param pc The pc object 
 */
void CukrPcDestroy(CUKR_PC* pc); 

#endif /*CUKR_PC_H_*/
