/**
 * @file solvers/cg_solver.c
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines the actual implementation
 * of the CG (Conjugate Gradients) 
 * solver. 
 * 
 * The solver is written using the
 * abstract BLAS interface, therefore
 * it runs both on CPU, GPU or using
 * both in combination.
 */

#define LOGGER_NAME "CUKr.solvers.cg_solver"

#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include <math.h>
#include "cg_solver.h"
#include "../../blas/blas1.h" 
#include "../../blas/blas2.h"
#include "../../util/log.h"
#include "../../util/utils.h"
#include "../../db/db.h"
#include "../../monitors/monitors.h"
 
/* Get access to the option database */
extern CukrDB cukr_db;  
 
/**
 * \brief Solver setup
 * \param solver The solver context
 */
void cg_ops_setup(CUKR_SOLVER solver)
{ 
	LOG(DEBUG, "Setting up CG solver implementation");
	
	/* Alloc memory for the structure */
	if ( (solver->data = (CG_SOLVER_DATA)malloc(sizeof(struct _CG_SOLVER_DATA))) == NULL) {
		CUKR_FATALX("Error allocating memory for the CG "\
					"data structure solver %s", solver->name);
	}
	
	/* Init temp vectors */
	CG_SOLVER_DATA data = (CG_SOLVER_DATA)solver->data;
	data->r = data->d = data->q = data->s = NULL;
	 
	/* Get vector size */
	int n = solver->mat->rows;
	
	/* Create the temp vectors
	 * It is enough for only one to be shared,
	 * the one which gets multiplied by the matrix */
	CukrVecCreate(&data->d, n, solver->precision, CUKR_SHTYPE_SHARED,      "d");
	CukrVecCreate(&data->r, n, solver->precision, CUKR_SHTYPE_DISTRIBUTED, "r");
	CukrVecCreate(&data->q, n, solver->precision, CUKR_SHTYPE_DISTRIBUTED, "q");
	CukrVecCreate(&data->s, n, solver->precision, CUKR_SHTYPE_DISTRIBUTED, "s");
	
	LOG(DEBUG, "Done setting up CG solver implementation");
} 

// ------------------------------------------------------------------- //

/**
 * \brief The actual equation solving
 *        routine
 * \param solver The solver context
 */
void cg_ops_solve(CUKR_SOLVER solver)
{	
	int i;
	
	LOG(DEBUG, "Running solve procedure for CG "\
			   "solver implementation");
		
	/* Access the data structure */
	CG_SOLVER_DATA data = (CG_SOLVER_DATA)(solver->data);
	
	/* Make shortucts for easy reading */
	CUKR_BLAS blas = solver->blas;
	CUKR_VECTOR r = data->r;
	CUKR_VECTOR q = data->q;
	CUKR_VECTOR s = data->s;
	CUKR_VECTOR d = data->d;
	CUKR_VECTOR x = solver->sol;
	
	CUKR_DOUBLE rnorm2_old, alpha = 1, alpha_old = 1, beta = 1;
	CUKR_DOUBLE rtol2   = solver->rtol * solver->rtol; 
	CUKR_DOUBLE abstol2 = solver->abstol * solver->abstol; 
	CUKR_DOUBLE divtol2 = solver->divtol * solver->divtol; 	
	
	/* Aux for eigenvalue computation */
	CUKR_DOUBLE a, b;
	
	/* Log */
	#ifdef CUKR_LOGGING
		char str_type[255], str_loc[255];
		CukrTypeIdToStr(solver->precision, str_type);
		CukrLocIdToStr(solver->blas->location, str_loc);
		LOGX(INFO, "Solver %s will be run with rtol=%g, "\
		           "abstol=%g, divtol=%g, blas=%s (on %s "\
		           "in %s) and solve eq. %s%s=%s", 
		           solver->name, solver->rtol, solver->abstol,
		           solver->divtol, solver->blas->name, str_loc,
		           str_type, solver->mat->name, solver->sol_init->name,
		           solver->rhs->name);
	#endif
		
	/* Set the no. of iterations to zero */
	solver->its = 0;
	
	/* 
	 * r = b - Ax 
	 */
	
	CukrCOPY(blas, solver->rhs, r);
	
	#ifdef CUKR_MP
		/* If x is a shared vector, do the
		 * multiplication the regular way */
		if (x->shtype == CUKR_SHTYPE_SHARED)
			CukrSPMV(blas, -1.0, solver->mat, x, 1.0, r);
			
		/* If distributed (to save memory) copy it
		 * into d first and then do it */
		else {		
			CukrCOPY(blas, x, d);
			CukrSPMV(blas, -1.0, solver->mat, d, 1.0, r);
		} 
	#else
		CukrSPMV(blas, -1.0, solver->mat, x, 1.0, r);	
	#endif
	
	solver->flops += 2 * solver->mat->nz;
	
	/*
	 * If preconditioned
	 */
		
	if (solver->pc->pc_type != CUKR_PC_TYPE_NONE) 
	{ 
	    /* Apply precondtioner, if needed (d = M^-1 * r) */
		CukrPcApply(solver->pc, r, d);		
	}
	
	/*
	 * If not preconditioned
	 */
	 
	else 
	{
		/* r = d */		
		CukrCOPY(blas, r, d);				
	} 
	
	/* solver->rnorm = r'd */
	solver->rnorm02 = CukrDOT(blas, r, d);
	solver->rnorm2  = solver->rnorm02; 	
	solver->flops += 2 * r->n;
	 
	/* Add residual norm to history */
	solver->res_hist[solver->its] = sqrt(solver->rnorm02);
	
	/* For eig */
	alpha_old = 1;
	beta = solver->rnorm02;
	
	/* 
	 * Run the iterations 
	 */
	while( 
			(
	           solver->rnorm2 >= abstol2 
			   && 
			   (
			     solver->rnorm2 < divtol2 
			     || 
			     solver->its == 0 
			    )
			 )  
	        && 
		    ( 
		      ( 
		         solver->its < solver->max_it 
		         && 
		         solver->rnorm2 >= rtol2 * solver->rnorm02
		      ) 
		      ||     
		      ( 
		        solver->itref 
		        && 
		        !solver->its_out 
		        && 
		        solver->its < solver->min_it
		      ) 
		    )
		 )
		
	{							 	
		/* q = Ad */								 
		CukrSPMV(blas, 1.0, solver->mat, d, 0.0, q);
		solver->flops += 2 * solver->mat->nz;		
			
		/* alpha = solver->rnorm / d'q */ 
		alpha = solver->rnorm2 / CukrDOT(blas, d, q);
		solver->flops += 2 * r->n;
		
		/* Store eigenvalue info */		 
		 if (solver->eig) {
		 	if (solver->its)
		 		solver->eig_e[solver->its] = sqrt(beta) / alpha_old;
		 	solver->eig_d[solver->its] = sqrt(beta) * solver->eig_e[solver->its] + 1.0 / alpha;
		 }
		
		/* x = x + alpha*d */
		CukrAXPY(blas, alpha, d, x);
		solver->flops += 2 * r->n;
		
		/* r = r - alpha*q */
		CukrAXPY(blas, -alpha, q, r);	
		solver->flops += 2 * r->n;
		
		/*
		 * If preconditioned
		 */
		
		if (solver->pc->pc_type != CUKR_PC_TYPE_NONE)
		{  	 
			/* Apply precondtioner (s = M^-1 * r) */ 
			CukrPcApply(solver->pc, r, s);   
		 
			/* solver->rnorm = r's */
			rnorm2_old = solver->rnorm2;
			solver->rnorm2 = CukrDOT(blas, r, s);
			solver->flops += 2 * r->n;
		
			/* d = s + beta*d */
			beta = solver->rnorm2 / rnorm2_old;			
			CukrAYPX(blas, beta, s, d);
			solver->flops += 2 * r->n;
		}
				
		
		/*
		 * If not preconditioned
		 */
		 
		else
		{   
			/* solver->rnorm = r'r */
			rnorm2_old = solver->rnorm2;
			solver->rnorm2 = CukrDOT(blas, r, r);
			solver->flops += 2 * r->n;
		
			/* d = r + beta*d */
			beta = solver->rnorm2 / rnorm2_old;
			CukrAYPX(blas, beta, r, d);
			solver->flops += 2 * r->n;	
		}			
				 
		/* 
		 * Do stuff 
		 */		 	
		
		/* Increase iteration count */
		solver->its++;
		solver->its_total++;
		 
		 /* Update alpha_old */
		 alpha_old = alpha;
			 
		 /* Add residual norm to history */
		 solver->res_hist[solver->its] = sqrt(solver->rnorm2);	 
		 if (!cukr_db.mpi_rank && CukrDBGetCountS(CUKR_DB_MONITOR)) 
		 {
		 	/* Write iteration number */		 			
		 	printf("It = %5d ", solver->its);
		 
		 	/* Call registered monitors to write 
		  	* additional data */		  
		 	for (i = 0 ; i < solver->numbermonitors; i++)
		 		(solver->monitors[i])(solver, CUKR_INNER);   
		 
		 	/* End of line, after all monitors have 
		  	* written */	
		  	printf("\n");				   
		  } 		  
	}

	
	/* 
	 * Check termination reason and return
	 * the appropriate code 
	 */
	
	/* Check for rtol */
	if (solver->rnorm2 < rtol2 * solver->rnorm02)
		solver->reason = CUKR_CONVERGED_RTOL;
	/* If not, check for abstol */
	else if (solver->rnorm2 < abstol2)
		solver->reason = CUKR_CONVERGED_ABSTOL;
	/* If the max. no. of iterations was 
	 * exceeded, means divergence by its */ 
	else if (solver->its >= solver->max_it) 
		solver->reason = CUKR_DIVERGED_ITS;
	/* If not, it must be divtol */
	else
		solver->reason = CUKR_DIVERGED_DTOL;
	
	LOG(DEBUG, "Done running solve procedure for "\
			   "CG solver implementation");
}

// ------------------------------------------------------------------- //

/**
 * \brief Get various solver-specific
 *        settings from the option database
 * \param solver The solver context
 */
void cg_ops_setfromopts(CUKR_SOLVER solver)
{
	int i;
	
	/* Look for valid monitors to register */ 
	for (i = 0 ; i < CukrDBGetCountS(CUKR_DB_MONITOR); i++)
	{
		/* Register rres monitor */
		if (!CukrStricmp("rres", CukrDBGetExS(CUKR_DB_MONITOR, i)))
			CukrSolverAddMonitor(solver, CukrMonitorRRes);
		/* Register absres monitor */
		else if (!CukrStricmp("absres", CukrDBGetExS(CUKR_DB_MONITOR, i)))
			CukrSolverAddMonitor(solver, CukrMonitorAbsRes);
		/* If none, error */	
		else {
			CUKR_FATALX("Unknown monitor: %s", 
			    CukrDBGetExS(CUKR_DB_MONITOR, i));
		}
	}
	
}

// ------------------------------------------------------------------- //

/**
 * \brief Destroy solver-specific stuff
 * \param solver The solver context
 */
void cg_ops_destroy(CUKR_SOLVER solver)
{
	LOG(DEBUG, "Destroying CG solver implementation");
	
	/* Access the data structure */
	CG_SOLVER_DATA data = (CG_SOLVER_DATA)(solver->data);
	
	/* Destroy the temp vectors */
	CukrVecDestroy(&data->r);
	CukrVecDestroy(&data->d);
	CukrVecDestroy(&data->q);
	CukrVecDestroy(&data->s);
	
	
	/* Free the data store */
	free(solver->data);	
	
	/* Log */
	LOG(DEBUG, "CG solver implementation destroyed");
}
 
// ------------------------------------------------------------------- //
/**
 * \brief Bind the CG solver to the
 *        solver context
 * \param solver The solver context
 */
void solver_create_cg(CUKR_SOLVER solver)
{
	solver->ops.setup       = cg_ops_setup;
	solver->ops.solve       = cg_ops_solve;
	solver->ops.setfromopts = cg_ops_setfromopts;
	solver->ops.destroy     = cg_ops_destroy;
}
