/**
 * @file solvers/solver.c
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines the CUKR_SOLVER abstract
 * interace and associated functions.
 * 
 * The CUKR_SOLVER structure stores all
 * member variables (characteristic) and
 * function pointers necessary to define
 * a solver. Since all solvers must 
 * implement the functions defined here,
 * the class functions as a solver 
 * interface. 
 * 
 */

#define LOGGER_NAME "CUKr.solvers.solver"
#define RESULT_FILENAME "result.txt"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef __MACH__
	#include <malloc.h>
#endif
#include <math.h> 
#include "solver.h"
#include "cg/cg_solver.h" 
#include "../blas/blas1.h"
#include "../blas/blas2.h" 
#include "../util/log.h"
#include "../util/utils.h"
#include "../util/macros.h"
#include "../db/db.h"

/*
 * Forward declarations
 */

int cukr_solver_compare(const void * a, const void * b) {return ( abs(*(CUKR_DOUBLE*)a) - abs(*(CUKR_DOUBLE*)b) );}
static int petsc_LINPACKcgtql1(int *n, CUKR_DOUBLE *d, CUKR_DOUBLE *e, int *ierr);
static CUKR_DOUBLE petsc_LINPACKcgpthy(CUKR_DOUBLE*, CUKR_DOUBLE*); 

/*
 * PETSc macros, for LINPACKcgtql1
 */
 
#define PetscMin(a,b)   (((a)<(b)) ? (a) : (b))
#define PetscMax(a,b)   (((a)<(b)) ?  (b) : (a))  
#define PetscAbsReal(a) (((a)<0)   ? -(a) : (a))
#define PetscReal       CUKR_DOUBLE
#define PetscInt        CUKR_INT
#define PetscFunctionBegin
#define PetscFunctionReturn(a) return(a)
#define PetscErrorCode int

/*
 * Globals 
 */

/* Get access to the option database */
extern CukrDB cukr_db; 

//========================== FORWARD DECLARATIONS ========================= //

void CukrSolverSetOptions(CUKR_SOLVER solver);
void CukrSolverSetInnerRtolAuto(CUKR_SOLVER solver, CUKR_DOUBLE condest);

//============================= PUBLIC FUNCTIONS ========================== //
 

/**
 * \brief Create a solver structure. The 
 *        actual solver type is decided 
 *        through command-line options.
 * \param solver Pointer to the solver
 *        object 
 * \param name Solver name
 */
void CukrSolverCreate(CUKR_SOLVER* solver, char* name)
{
	int i;
	
	/* 
	 * Validate parameters 
	 */
	 
	/* Check for valid name */ 
	if (name == NULL) {
		CUKR_FATAL("Solver must have a non-null name");
	}
	
	/* Log */
	LOGX(DEBUG, "Creating solver structure for solver %s",
	            name);          
	
	/*
	 * Alloc memory for the solver
	 * Set the name and stuff
	 */
	 
	/* Alloc memory for the solver structure*/
	if ( ((*solver) = (CUKR_SOLVER)malloc(sizeof(struct _CUKR_SOLVER))) == NULL) {
		CUKR_FATALX("Error allocating memory for solver object "\
					"while creating solver %s", name);
	} 
	
	/* Set the name and stuff*/
	sprintf((*solver)->name, "%s", name);  
	(*solver)->setted_up = CUKR_FALSE;
	(*solver)->numbermonitors = 0;
	(*solver)->flops = 0;
	
	/* 
	 * Default options 
	 */
	 
	 (*solver)->itref  = CUKR_FALSE;
	 (*solver)->dscale = CUKR_FALSE;
	 (*solver)->eig    = CUKR_FALSE;
	 (*solver)->inner_rtol_auto = CUKR_FALSE;
	 (*solver)->outer_rtol_auto = CUKR_FALSE;
	
	
	/*
	 * Read options from the options database.
	 * It shoud be done now, since depending on
	 * some options (e.g. itref) additional
	 * memory needs to be allocated
	 */ 
	 
	/* Read general solver options from */
	CukrSolverSetOptions(*solver);
	
	/*
	 * Alloc additional memory, where needed 
	 */
	
	/* Alloc memory for residual history 
	 * of the actual (inner if itref) solver */
	if ( ((*solver)->res_hist = (CUKR_DOUBLE*)malloc(CUKR_SOLVER_MAX_HISTORY * sizeof(CUKR_DOUBLE))) == NULL) {
			CUKR_FATALX("Error allocating memory for solver.res_history "\
				        "while creating solver %s", name);
	}  
	
	/* Alloc memory for residual history 
	 * of the actual (inner if itref) solver */
	if ( ((*solver)->res_hist_out = (CUKR_DOUBLE*)malloc(CUKR_SOLVER_MAX_HISTORY * sizeof(CUKR_DOUBLE))) == NULL) {
			CUKR_FATALX("Error allocating memory for solver.res_history_out "\
				        "while creating solver %s", name);
	}  
	
	/* If itref,
	 * alloc memory for the residual history
	 * for the correction iterations */
	if ((*solver)->itref) {
		if ( ((*solver)->res_hist_out = (CUKR_DOUBLE*)malloc((*solver)->max_it_out * sizeof(CUKR_DOUBLE))) == NULL) {
				CUKR_FATALX("Error allocating memory for solver.res_history_out "\
					        "while creating solver %s", name);
		}  
	}
	
	/*
	 * Zero the counters and so on
	 */
	 
	 /* Null the data containers */
	 (*solver)->mat = NULL;
	 (*solver)->rhs = NULL;
	 (*solver)->rhs_out = NULL;
	 (*solver)->sol = NULL;
	 (*solver)->sol_out = NULL;
	 (*solver)->sol_init = NULL;
	 
	 /* Zero the counters */
	 (*solver)->its = 0;
	 (*solver)->its_out = 0;
	 (*solver)->its_total = 0;
	 (*solver)->reason = 0;
	 
	 for (i = 0; i < CUKR_SOLVER_CNTR_VECTOR_LENGTH; i++)
	 	(*solver)->counters[i] = 0;
	 
	/*
	 * Set solver implementation from options
	 */
	  
	 /* Process solver-specific options */
	 (*solver)->ops.setfromopts(*solver);
	 
	 /*
	  * Create preconditioner context
	  * and set if from options. If the prec.
	  * is not needed, this procedure will
	  * do nothing.
	  */
	 CukrPcCreate(&(*solver)->pc, *solver);
	   
}

// ------------------------------------------------------------------- //

/**
 * \brief Set the system matrix of the
 *        system to be solved.
 * \param solver The solver object 
 * \param mat The system matrix
 */   
void CukrSolverSetOperators(CUKR_SOLVER solver, CUKR_MATRIX mat)
{
	solver->mat = mat; 
}

// ------------------------------------------------------------------- //

/**
 * \brief Set solver tolerances
 * \param solver The solver object 
 * \param rtol Relative converge tolerance
 * \param abstol Absolute convergence tolerance
 * \param divtol Divergence tolerance
 * \param max_it Maximum no. of iterations 
 *        allowed
 */
void CukrSolverSetTolerances(CUKR_SOLVER solver, CUKR_DOUBLE rtol,
                             CUKR_DOUBLE abstol, CUKR_DOUBLE divtol,
                             int max_it)
{
	solver->rtol   = rtol;
	solver->abstol = abstol;
	solver->divtol = divtol;
	solver->max_it = max_it;
}
                             	
                        
// ------------------------------------------------------------------- //                        
                        
/**
 * \brief Solve the system
 * \param solver The solver
 *        object 
 * \param rhs The rhs vector
 * \param sol_init The initial solution 
 *        vector. If null, the init. solution 
 *        will be considered to be zero.
 */
void CukrSolverSolve(CUKR_SOLVER solver, CUKR_VECTOR rhs, 
                     CUKR_VECTOR sol_init)
{
	char reason_str[1024];
	int i;
	
	/* Time keeping variables */
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	
	LOGX(DEBUG, "Starting solving procedure for solver %s", 
	            solver->name);
	
	/* 
	 * Check for valid operator
	 */
	 
	if (solver->mat == NULL) {
		CUKR_FATALX("No matrix specified for solver %s. "\
		            "Use <CukrSetOperator> function before "
		            "calling the CukrSolve routine", solver->name);
	}
	
	/*
	 * Implementation specific setup
	 */
		 
	if (!solver->setted_up) { 
		(solver->ops.setup)(solver);
		solver->setted_up = CUKR_TRUE;
	} 	
	 	
	/*
	 * Copy rhs and initial solution
	 * so the original ones do not 
	 * get modified
	 */
	
	/* If no itref, copy to inner */
	if (!solver->itref) {
		CukrVecCreateClone(rhs,      &solver->rhs, "b");
		CukrVecCreateClone(sol_init, &solver->sol, "x");	
	}
	/* If itref, copy to outer */
	else
	{
		CukrVecCreateClone(rhs, &solver->rhs_out, "b_hi");
		CukrVecCreateClone(sol_init, &solver->sol_out, "x_hi");
		CukrVecCreateClone(sol_init, &solver->sol, "x_lo");
	}
	
	solver->sol_init = sol_init;
	
	/*
	 * Solve
	 */
	
	/* If no itref, just hand it to the
	 * solver routine */
	if (!solver->itref) 
	{	
		CLOCK_START;
		(solver->ops.solve)(solver);
		CLOCK_END;
		solver->counters[SOLVER_IDX_NORMAL + SOLVER_OFFSET_TIME] = CLOCK_SPAN;
		solver->counters[SOLVER_IDX_NORMAL + SOLVER_OFFSET_ITS] = solver->its;
		cukr_db.solverTime = CLOCK_SPAN;
		
		/* Log the convergence reason */
		LOGX(INFO, "%s", CukrSolverGetConvergedReasonAsStr(solver, CUKR_OUTER));
	}
	/* 
	 * Else, do itref 
	 */
	 
	else 
	{	
		/* Log */
		char str_type[255], str_loc[255];
		CukrTypeIdToStr(solver->blas_out->precision, str_type);
		CukrLocIdToStr(solver->blas_out->location, str_loc);
		LOGX(INFO, "Solver %s will be run using iterative " \
		 	   	   "refinement, with rtol=%g, " \
	               "abstol=%g, divtol=%g, blas=%s (on %s "\
	               "in %s) and solve eq. %s%s=%s", 
	               solver->name, solver->rtol_out, solver->abstol_out,
	               solver->divtol_out, solver->blas_out->name, str_loc,
	               str_type, solver->mat->name, solver->sol_out->name,
	               solver->rhs_out->name);
		
		CLOCK_START;
		
		/*
		 * For auto rtol, 
		 */
				
		/* Get params and declare vars */ 
		double solver_rtol_auto_th  = CukrDBGetD(CUKR_DB_SOLVER_OUTER_RTOL_AUTO_THRESHOLD); 
		double solver_rtol_auto_rel = CukrDBGetD(CUKR_DB_SOLVER_OUTER_RTOL_AUTO_RELAXATION);
		double outer_res_improvement = 1e16;	
		
		/* Set rtol_out to maximum (1e-16) */
		if (solver->outer_rtol_auto) 
			solver->rtol_out = 1e-16;
		
			
		/*
		 * If solver->sol_out is of distributed type, 
		 * we need the shared vector used by the inner
		 * solver to use instead
		 */
		 
		 #ifdef CUKR_MP
			 CUKR_VECTOR shared_vec;		 
			 if (solver->sol_out->shtype == CUKR_SHTYPE_DISTRIBUTED) 
			 {
				 const char *solver_str = CukrDBGetS(CUKR_DB_SOLVER);
				 
				 /* If CG solver, use d */
		 		 if (!CukrStricmp("CG", solver_str)) {
		 		 	 CG_SOLVER_DATA data = (CG_SOLVER_DATA)(solver->data);
		 			 shared_vec = data->d; 
		 		 }	
		 		 /* If none, error */
		 		 else {
		 		 	CUKR_FATALX("Unable to use the shared vector "\
		 		 	            "from the currently selected solver. "\
		 		 	            "Make the unknown vector %s SHARED.",
		 		 	            solver->rhs_out->name);
		 		 }
			 }
		  #endif 
		 	
		/* 
		 * The inner rhs will be the outer residual 
		 */		
		 
		CukrVecCreateClone(solver->rhs_out, &solver->rhs, "b_lo");
		
		#ifdef CUKR_MP
			/* If the solver->sol_out is shared,
			 * do the usual way */				
			if (solver->sol_out->shtype == CUKR_SHTYPE_SHARED) 
				CukrSPMV(solver->blas_out, -1.0, solver->mat, solver->sol_out, 1.0, solver->rhs);
				
			/* If the solver->sol_out is distributed,
			 * in order to save memory, used the shared
			 * vector used by the inner solver */
			else {
				CukrCOPY(solver->blas_out, solver->sol_out, shared_vec);
				CukrSPMV(solver->blas_out, -1.0, solver->mat, shared_vec, 1.0, solver->rhs);	
			} 
		#else
			CukrSPMV(solver->blas_out, -1.0, solver->mat, solver->sol_out, 1.0, solver->rhs);			
		#endif
		
		solver->flops += 2 * solver->mat->nz; 
		
		/* Compute initial defect */
		solver->rnorm02_out = CukrDOT(solver->blas_out, solver->rhs, solver->rhs);
		solver->rnorm2_out = solver->rnorm02_out;
		solver->flops += 2 * solver->rhs->n; 
		
		/* 
		 * Run the loop
		 */
		
		CUKR_DOUBLE rtol2   = solver->rtol_out   * solver->rtol_out; 
		CUKR_DOUBLE abstol2 = solver->abstol_out * solver->abstol_out; 
		CUKR_DOUBLE divtol2 = solver->divtol_out * solver->divtol_out;
		solver->its_out = 0;
		
		/* Update counters */
		CLOCK_END;
		solver->counters[SOLVER_IDX_OUTER + SOLVER_OFFSET_TIME] += CLOCK_SPAN;		
				   
		while ( (solver->its_out < solver->max_it_out)              &&
				(solver->rnorm2_out >= rtol2 * solver->rnorm02_out) &&
				(solver->rnorm2_out >= abstol2)                     &&
				(solver->rnorm2_out <  divtol2)                     && 
				( 
				  !solver->outer_rtol_auto                     || 
				  !solver->its_out                             || 
				  outer_res_improvement >= solver_rtol_auto_th || 
				  solver->rnorm2_out > solver->rnorm02_out * solver_rtol_auto_rel * (solver->condest / 1e+16)
				)
			  )
		{ 
			CLOCK_START;
			
			/* Scale defect, if not first inner iteration 
			 * and re-copy the initial solution */
			if (solver->rnorm02_out != solver->rnorm2_out) {
				CukrSCAL(solver->blas_out, 1.0/sqrt(solver->rnorm2_out), solver->rhs);
				CukrCOPY(solver->blas_out, solver->sol_init, solver->sol);
				solver->flops += solver->rhs->n; 
			}
			 
			/* 
			 * Run inner solver and check for divergence 
			 */
			
			/* Run inner solver */
			CLOCK_END;
			solver->counters[SOLVER_IDX_OUTER + SOLVER_OFFSET_TIME] += CLOCK_SPAN;
			
			CLOCK_START;
			(solver->ops.solve)(solver);
			CLOCK_END;
			solver->counters[SOLVER_IDX_INNER + SOLVER_OFFSET_TIME] += CLOCK_SPAN;
			
			CLOCK_START;
			
			/* Log convergence/divergence */
			LOGX(INFO, "%s", CukrSolverGetConvergedReasonAsStr(solver, CUKR_INNER));
			
			/* Check for divergence */
			if (solver->reason == CUKR_DIVERGED_ITS &&
				solver->reason == CUKR_DIVERGED_DTOL )
				{		
					/* Print the convergence reason */
					printf("%s\n", reason_str); 
					
					/* Exit */
					return;
				}
			
			/*
			 * Update & resume
			 */
			
			/* Update outer solution, using the scaling 
			 * factor for all bust the first iteration */
			if (solver->rnorm02_out != solver->rnorm2_out)
				CukrAXPY(solver->blas_out, sqrt(solver->rnorm2_out), solver->sol, solver->sol_out);
			else
				CukrAXPY(solver->blas_out, 1.0, solver->sol, solver->sol_out);
			solver->flops += 2 * solver->sol->n;
					
			/* 
			 * Compute defect in high precision 
			 */
			 
			CukrCOPY(solver->blas_out, solver->rhs_out, solver->rhs);
			
			#ifdef CUKR_MP			
				/* If the solver->sol_out is shared,
				 * do the usual way */				
				if (solver->sol_out->shtype == CUKR_SHTYPE_SHARED) 
					CukrSPMV(solver->blas_out, -1.0, solver->mat, solver->sol_out, 1.0, solver->rhs);
					
				/* If the solver->sol_out is distributed,
				 * in order to save memory, used the shared
				 * vector used by the inner solver */
				else {
					CukrCOPY(solver->blas_out, solver->sol_out, shared_vec);
					CukrSPMV(solver->blas_out, -1.0, solver->mat, shared_vec, 1.0, solver->rhs);	
				} 	
			#else
				CukrSPMV(solver->blas_out, -1.0, solver->mat, solver->sol_out, 1.0, solver->rhs);		
			#endif
			  
			solver->flops += 2 * solver->mat->nz;             
			
			/* Compute norm */
			CUKR_DOUBLE rnorm2_out = CukrDOT(solver->blas_out, solver->rhs, solver->rhs);
			outer_res_improvement = solver->rnorm2_out / rnorm2_out;
			solver->rnorm2_out = rnorm2_out;
			solver->flops += 2 * solver->rhs->n; 
			
			/* If either of the automatic option enabled,
			 * compute condest (1st iteration only) */
			 if (!solver->its_out && 
			     (solver->inner_rtol_auto || solver->outer_rtol_auto))
			 	 solver->condest = CukrSolverCondest(solver);   
				
			/* If first iteration and automatic inner_rtol
			 * enabled, set inner_rtol */
			 if (!solver->its_out && solver->inner_rtol_auto)
			 	CukrSolverSetInnerRtolAuto(solver, solver->condest); 
		
			 			
			/* 
			 * Do stuff 
			 */
			 
			/* Update iteration count */
			solver->its_out++; 	
			
			if (!cukr_db.mpi_rank && CukrDBGetCountS(CUKR_DB_MONITOR)) {
				
				/* Write iteration number */			 
				printf("\n>>> Outer iteration = %5d ( %5d inner its ) <<< ", 
						solver->its_out, solver->its_total);
			 
				/* Call registered monitors to write 
				 * additional data */
				for (i = 0 ; i < solver->numbermonitors; i++)
					(solver->monitors[i])(solver, CUKR_OUTER);  
				
				/* End of line, after all monitors have 
				 * written */
				printf("\n\n");
			}
			
			/* Update counters */
			CLOCK_END;
			solver->counters[SOLVER_IDX_OUTER + SOLVER_OFFSET_TIME] += CLOCK_SPAN;
		}
		
		/* Update counters */
		solver->counters[SOLVER_IDX_INNER + SOLVER_OFFSET_ITS] = solver->its_total;
		solver->counters[SOLVER_IDX_OUTER + SOLVER_OFFSET_ITS] = solver->its_out;
		
		/* 
		 * Establish convergence/divergence reason 
		 */
		
		/* Check for rtol */
		if (solver->rnorm2_out < rtol2 * solver->rnorm02_out)
			solver->reason_out = CUKR_CONVERGED_RTOL;
		/* If not, check for abstol */
		else if (solver->rnorm2_out < abstol2)
			solver->reason_out = CUKR_CONVERGED_ABSTOL;
		/* If the max. no. of iterations was 
		 * exceeded, means divergence by its */ 
		else if (solver->its_out >= solver->max_it_out)
			solver->reason_out = CUKR_DIVERGED_ITS;	
		/* If not, it must be divtol */
		else if (solver->rnorm2_out >= divtol2)
			solver->reason_out = CUKR_DIVERGED_DTOL;
		else
			solver->reason_out = CUKR_CONVERGED_BESTTOL;
		
		/* Log convergence reason */
		#ifdef CUKR_LOGGING
			
			LOGX(INFO, "%s", CukrSolverGetConvergedReasonAsStr(solver, CUKR_OUTER));
		#endif	
		/* Update solver time */
		cukr_db.solverTime = solver->counters[SOLVER_IDX_OUTER + SOLVER_OFFSET_TIME] +
							 solver->counters[SOLVER_IDX_INNER + SOLVER_OFFSET_TIME];
	} 
		 	
	
	/* Save result to file */
//	if (!solver->itref)
//		CukrVecmpiWriteDataToFile(solver->sol, RESULT_FILENAME);
//	else
//		CukrVecmpiWriteDataToFile(solver->sol_out, RESULT_FILENAME);
}                     

// ------------------------------------------------------------------- //

/*
 * \brief Returns the result vector,
 *        after the solve
 * \param solver The solver object
 * \return The result vector
 */
CUKR_VECTOR CukrSolverGetResult(CUKR_SOLVER solver)
{
	/* If not itref */
	if (!solver->itref)
		return solver->sol;
	/* If itref, return outer solution */
	else
		return solver->sol_out;	
}

// ------------------------------------------------------------------- //

/*
 * \brief Computes a condition number estimate
 * \param solver The solver object
 * \return The cond. est.
 */
 
CUKR_DOUBLE CukrSolverCondest(CUKR_SOLVER solver)
{
	/* Check that data was saved */
	if (!solver->eig)
		CUKR_FATALX("Cannot compute condition estimate for "\
		            "solver %s since eigenvalue information "\
		            "saving option was not set", solver->name);
	
	/* Time keeping variables */
	struct timeval t_start, t_end;
	struct timezone t_zone; 
	CLOCK_START;
				           
	/* Call the EIPACK routine */	   
	int err;	
	petsc_LINPACKcgtql1(&solver->its, solver->eig_d, solver->eig_e, &err);
	if (err) {
		CUKR_FATALX("Error from petsc_LINPACKcgtql1(); petsc eispack eigenvalue routine: %d", err)
	}          
	
	/* Count elapsed time */
	CLOCK_END;
	solver->counters[SOLVER_IDX_CONDEST + SOLVER_OFFSET_TIME] = CLOCK_SPAN;
	
	/* Return estimate */
	CUKR_DOUBLE condest = solver->eig_d[solver->its - 1] / solver->eig_d[0];
	return condest;
} 

// ------------------------------------------------------------------- //

/**
 * \brief After a solve, return the
 *        convergence/divergence reason
 * \param solver The solver object 
 * \param for_inner If true, in case of iterative
 *        refinement, reports are done for the 
 *        inner solver and if false, for the outer
 *        solver
 */               
const char* CukrSolverGetConvergedReasonAsStr(CUKR_SOLVER solver, int inner_outer)
{	
	/* If not itref, check for regular convergence */	
	if (!solver->itref) {
		switch (solver->reason)
		{
			case CUKR_DIVERGED_ITS: 
			{
				sprintf(solver->reason_str,
				       "Solver %s diverged. The maximum " \
				       "no. of iterations (%d) has been " \
				       "exceeded.", solver->name, solver->max_it);
				return solver->reason_str;       
			}
			case CUKR_DIVERGED_DTOL:
			{
				sprintf(solver->reason_str,
					   "Solver %s diverged. The residual " \
				       "exceeded the maximum allowed size " \
				       "(%g)", solver->name, solver->divtol);
				return solver->reason_str;       
			}
			case CUKR_CONVERGED_RTOL:
			{ 
				sprintf(solver->reason_str, 
					   "Solver %s converged in %d iterations. "\
				       "The relative residual was smaller than " \
				       "the required tolerance (%g)", solver->name, 
				       solver->its, solver->rtol);
				return solver->reason_str;       
			}
			case CUKR_CONVERGED_ABSTOL:
			{
				sprintf(solver->reason_str,
					   "Solver %s converged in %d iterations. "\
				       "The absolute residual was smaller than " \
				       "the required tolerance (%g)", solver->name, 
				       solver->its, solver->abstol);
				return solver->reason_str;       
			}			
			default: 
			{ 
				CUKR_FATALX("Solver %s returned an unknown reason (%d)",
				            solver->name, solver->reason);
			}		
		}
	}
	
	/* If itref */
	else {
		/* If called after calling the inner
		 * solver, treat only divergence */
		if (inner_outer == CUKR_INNER)
		{
			switch (solver->reason)
			{
				case CUKR_DIVERGED_ITS: 
				{
					sprintf(solver->reason_str,
					       "Inner solver %s diverged. The maximum " \
					       "no. of iterations (%d) in the inner solver " \
					       "has been exceeded.", solver->name, solver->max_it);
					return solver->reason_str;       
				}
				case CUKR_DIVERGED_DTOL:
				{
					sprintf(solver->reason_str,
					       "Inner solver %s diverged. The residual " \
					       "in the inner solver exceeded the maximum " \
					       "allowed size (%g)", solver->name, solver->divtol);
					return solver->reason_str;       
				}
				case CUKR_CONVERGED_RTOL:
				{ 
					sprintf(solver->reason_str, 
						   "Inner solver %s converged in %d iterations. "\
					       "The relative residual was smaller than " \
					       "the required tolerance (%g)", solver->name, 
					       solver->its, solver->rtol);
					return solver->reason_str;       
				}
				case CUKR_CONVERGED_ABSTOL:
				{
					sprintf(solver->reason_str,
						   "Inner solver %s converged in %d iterations. "\
					       "The absolute residual was smaller than " \
					       "the required tolerance (%g)", solver->name, 
					       solver->its, solver->abstol);
					return solver->reason_str;       
				} 
				default: 
				{ 
					CUKR_FATALX("Solver %s returned an unknown reason (%d)",
					            solver->name, solver->reason_out);
				}
			}	
		}
		/* For outer, treat both */
		else {
			switch (solver->reason_out)
			{
				case CUKR_DIVERGED_ITS: 
				{
					sprintf(solver->reason_str, 
					       "Solver %s diverged. The maximum " \
					       "no. of outer correction iterations (%d) has " \
					       "been exceeded.", solver->name, solver->max_it_out);
					return solver->reason_str;       
				}
				case CUKR_DIVERGED_DTOL:
				{
					sprintf(solver->reason_str,
					       "Solver %s diverged. The high-precision" \
					       "residual exceeded the maximum allowed size " \
					       "(%g)", solver->name, solver->divtol_out);
					return solver->reason_str;       
				}
				case CUKR_CONVERGED_RTOL:
				{ 
					sprintf(solver->reason_str,
					       "Solver %s converged in %d outer iterations "\
						   "and %d inner iterations. The relative residual "\
						   "was smaller than the required tolerance (%g)", 
						   solver->name, solver->its_out, solver->its_total, 
						   solver->rtol_out);
					return solver->reason_str;       
				}
				case CUKR_CONVERGED_ABSTOL:
				{
					sprintf(solver->reason_str,
					       "Solver %s converged in %d outer iterations "\
						   "and %d inner iterations. The absolute residual "\
						   "was smaller than the required tolerance (%g)", 
						   solver->name, solver->its_out, solver->its_total, 
						   solver->abstol_out);
					return solver->reason_str;       
				}
				case CUKR_CONVERGED_BESTTOL:
				{
					sprintf(solver->reason_str,
						   "Solver %s converged in %d outer iterations. "\
						   "and %d inner iteration. "\
					       "It was decided that the best possible "\
					       "outer residual reduction has been reached", 
					       solver->name, solver->its_out, solver->its_total);
					return solver->reason_str;       
				}
				default: 
				{ 
					CUKR_FATALX("Solver %s returned an unknown reason (%d)",
					            solver->name, solver->reason_out);
				}		
			}
		}
	}
}

// ------------------------------------------------------------------- //

/**
 * \brief After a solve, return the number
 *        of iterations spent
 * \param solver The solver object 
 */
int CukrSolverGetIterationNumber(CUKR_SOLVER solver)
{
	return solver->its;
}  

// ------------------------------------------------------------------- //
 
/**
 * \brief Register a monitor to show/record
 *        various info related to the solver
 * \param solver The solver object 
 * \param monitor The monitor to be added
 */
void CukrSolverAddMonitor(CUKR_SOLVER solver, CUKR_SOLVER_MONITOR mon)
{
	solver->monitors[solver->numbermonitors] = mon;
	solver->numbermonitors++;
}                  

// ------------------------------------------------------------------- //

/**
 * \brief Destroy a solver object
 * \param solver The solver object 
 */
void CukrSolverDestroy(CUKR_SOLVER* solver)
{
	/* Copy the name, so it doesn't get
	 * destroyed */
	char name[255];
	memcpy(name, (*solver)->name, 255);
	LOGX(DEBUG, "Destroying solver %s", name);
	
	/* Destroy preconditioner */
	CukrPcDestroy(&(*solver)->pc);
	
	/* Destroy the solver-specific stuff */
	(*solver)->ops.destroy(*solver);
	
	/* Destroy the inner history vector */
	if ((*solver)->res_hist != NULL)
		free((*solver)->res_hist);
		
	/* Destroy the outer history vector */
	if ((*solver)->res_hist_out != NULL)
		free((*solver)->res_hist_out);	
	
	/* Destroy solution vectors and RHSs */
	CukrVecDestroy(&(*solver)->sol);
	CukrVecDestroy(&(*solver)->sol_out);
	CukrVecDestroy(&(*solver)->rhs);
	CukrVecDestroy(&(*solver)->rhs_out);
	
	/* Destroy the solver structure */
	free(*solver);
	
	/* Log */
	LOGX(DEBUG, "Solver %s destroyed", name);
} 


//============================ PRIVATE FUNCTIONS =========================== //


/**
 * \brief Read solver options from optiondb
 *        and set them in the solver structure
 * \param solver The solver object
 */
void CukrSolverSetOptions(CUKR_SOLVER solver)
{  
	
	 LOGX(DEBUG, "Setting solver %s from options",
	  			 solver->name);
		
	
	 /* Iterative refinement */
	 if (CukrDBGetI(CUKR_DB_SOLVER_ITREF))
		solver->itref = CUKR_TRUE;
	
	 /*
	  * Find the desired solver in the
	  * option database and bind it
	  */
	 
	 const char *solver_str = CukrDBGetS(CUKR_DB_SOLVER);
	 
	 /* If CG solver, bind it */
	 if (!CukrStricmp("CG", solver_str))
	 {
	 	solver_create_cg(solver);
		LOG(DEBUG, "Using CG solver");
	 }
	 /* If not, error message */
	 else
	 {
	 	CUKR_FATALX("Unknown solver type: %s", solver_str);
	 }
		
	/* 
	 * Get tolerances and max/min iters 
	 */
	
	/* If no itref, take only the standard opts. 
	 * and store them in the regular members */
	if (!solver->itref) { 
		solver->rtol   = CukrDBGetD(CUKR_DB_SOLVER_RTOL);
		solver->abstol = CukrDBGetD(CUKR_DB_SOLVER_ABSTOL);
		solver->divtol = CukrDBGetD(CUKR_DB_SOLVER_DIVTOL);
		solver->max_it = CukrDBGetI(CUKR_DB_SOLVER_MAXIT);
		solver->min_it = 0;
	}
	/* If itref, make the standard opts to be 
	 * outer, and the inner ones to be the 
	 * regular members */
	else {
		solver->rtol   = CukrDBGetD(CUKR_DB_SOLVER_INNER_RTOL);
		solver->abstol = CukrDBGetD(CUKR_DB_SOLVER_INNER_ABSTOL);
		solver->divtol = CukrDBGetD(CUKR_DB_SOLVER_INNER_DIVTOL);
		solver->max_it = CukrDBGetI(CUKR_DB_SOLVER_INNER_MAXIT);
		solver->min_it = CukrDBGetI(CUKR_DB_SOLVER_INNER_MINIT);
		
		solver->rtol_out   = CukrDBGetD(CUKR_DB_SOLVER_RTOL);
		solver->abstol_out = CukrDBGetD(CUKR_DB_SOLVER_ABSTOL);
		solver->divtol_out = CukrDBGetD(CUKR_DB_SOLVER_DIVTOL);
		solver->max_it_out = CukrDBGetI(CUKR_DB_SOLVER_MAXIT);
				
	}	
	
	/* If itref and inner_rtol set to auto, enable
	 * eigenvalue computing support */
	 if (solver->itref && CukrDBGetI(CUKR_DB_SOLVER_INNER_RTOL_AUTO)) {
	 	 CukrDBSetI(CUKR_DB_SOLVER_EIGENINFO, 1);	 	 
	 	 solver->inner_rtol_auto = CUKR_TRUE;
	 }
	 
	 /* If itref and outer_rtol set to auto, enable
	 * eigenvalue computing support and a minimum no.
	 * of iterations of 100 */
	 if (solver->itref && CukrDBGetI(CUKR_DB_SOLVER_OUTER_RTOL_AUTO)) {
	 	 CukrDBSetI(CUKR_DB_SOLVER_EIGENINFO, 1);
	 	 solver->outer_rtol_auto = CUKR_TRUE;
	 	 if (CukrDBGetI(CUKR_DB_SOLVER_INNER_MINIT) < 100)
	 	 	 CukrDBSetI(CUKR_DB_SOLVER_INNER_MINIT, 100);
	 }
	
	/* 
	 * Eigenvalue computing support 
	 */
	 if (CukrDBGetI(CUKR_DB_SOLVER_EIGENINFO)) {	 	
	 		 	
	 	/* Set flag */
	 	solver->eig = CUKR_TRUE;
	 	
	 	/* Malloc strorage vectors */
	 	if ( (solver->eig_d = (CUKR_DOUBLE*)malloc((solver->max_it + 1) * sizeof(CUKR_DOUBLE))) == NULL)
	 		CUKR_FATALX("Error allocating memory for solver %s->eig_d",
	 		             solver->name); 	             
	 	if ( (solver->eig_e = (CUKR_DOUBLE*)malloc((solver->max_it + 1) * sizeof(CUKR_DOUBLE))) == NULL)
	 		CUKR_FATALX("Error allocating memory for solver %s->eig_e",
	 		             solver->name);	             
	 	
	 	/* Set first element of eig_e to 0 */
	 	solver->eig_e[0] = 0; 	 
	 }
	
	/*  
	 * Blas 
	 */
	 
	/* Set the BLAS acording to the desired 
	 * solver precision */
	
	const char* prec_str = CukrDBGetS(CUKR_DB_SOLVER_PRECISION); 
	 
	if (!CukrStricmp("DOUBLE", prec_str))
		solver->blas = cukr_db.blasD; 
	else if (!CukrStricmp("QDOUBLE", prec_str))
		solver->blas = cukr_db.blasQ; 
	else if (!CukrStricmp("SINGLE", prec_str))
		solver->blas = cukr_db.blasS;
	else {
		CUKR_FATALX("Unknown blas precision: %s", prec_str);
	}
	
	/* If itref, then set the outer blas to
	 * the previosly find blas (acording to
	 * the desired precision) and set the 
	 * blas to blasI */
	if (solver->itref){
		solver->blas_out = solver->blas;
		solver->blas = cukr_db.blasI;
	} 
		
	/* Set precision */
	solver->precision = solver->blas->precision;	  
}

// ------------------------------------------------------------------- //

/* \brief Based on condition number estimate,
 *        adjust inner_rtol automatically */
void CukrSolverSetInnerRtolAuto(CUKR_SOLVER solver, CUKR_DOUBLE condest)
{	
	/* Decide how to set based on rule */
	double new_inner_rtol;
	if  (condest > 1e6)
		new_inner_rtol = 1e-1;
	else if (condest > 1e3)
		new_inner_rtol = 1e-3;
	else
		new_inner_rtol = 1e-5;
		
	/* Output (for now) */
	printf("Condest: %g -> setting inner_rtol to: %g\n",
		   condest, new_inner_rtol);
	solver->rtol = new_inner_rtol;	   			
}



//============================ FROM PETSc =========================== //

static PetscErrorCode petsc_LINPACKcgtql1(PetscInt *n,PetscReal *d,PetscReal *e,PetscInt *ierr)
{
    /* System generated locals */
    PetscInt  i__1,i__2;
    PetscReal d__1,d__2,c_b10 = 1.0;

    /* Local variables */
     PetscReal c,f,g,h;
     PetscInt  i,j,l,m;
     PetscReal p,r,s,c2,c3 = 0.0;
     PetscInt  l1,l2;
     PetscReal s2 = 0.0;
     PetscInt  ii;
     PetscReal dl1,el1;
     PetscInt  mml;
     PetscReal tst1,tst2;

/*     THIS SUBROUTINE IS A TRANSLATION OF THE ALGOL PROCEDURE TQL1, */
/*     NUM. MATH. 11, 293-306(1968) BY BOWDLER, MARTIN, REINSCH, AND */
/*     WILKINSON. */
/*     HANDBOOK FOR AUTO. COMP., VOL.II-LINEAR ALGEBRA, 227-240(1971). */

/*     THIS SUBROUTINE FINDS THE EIGENVALUES OF A SYMMETRIC */
/*     TRIDIAGONAL MATRIX BY THE QL METHOD. */

/*     ON INPUT */

/*        N IS THE ORDER OF THE MATRIX. */

/*        D CONTAINS THE DIAGONAL ELEMENTS OF THE INPUT MATRIX. */

/*        E CONTAINS THE SUBDIAGONAL ELEMENTS OF THE INPUT MATRIX */
/*          IN ITS LAST N-1 POSITIONS.  E(1) IS ARBITRARY. */

/*      ON OUTPUT */

/*        D CONTAINS THE EIGENVALUES IN ASCENDING ORDER.  IF AN */
/*          ERROR EXIT IS MADE, THE EIGENVALUES ARE CORRECT AND */
/*          ORDERED FOR INDICES 1,2,...IERR-1, BUT MAY NOT BE */
/*          THE SMALLEST EIGENVALUES. */

/*        E HAS BEEN DESTROYED. */

/*        IERR IS SET TO */
/*          ZERO       FOR NORMAL RETURN, */
/*          J          IF THE J-TH EIGENVALUE HAS NOT BEEN */
/*                     DETERMINED AFTER 30 ITERATIONS. */

/*     CALLS CGPTHY FOR  DSQRT(A*A + B*B) . */

/*     QUESTIONS AND COMMENTS SHOULD BE DIRECTED TO BURTON S. GARBOW, */
/*     MATHEMATICS AND COMPUTER SCIENCE DIV, ARGONNE NATIONAL LABORATORY 
*/

/*     THIS VERSION DATED AUGUST 1983. */

/*     ------------------------------------------------------------------ 
*/
    PetscReal ds;

    PetscFunctionBegin;
    --e;
    --d;

    *ierr = 0;
    if (*n == 1) {
        goto L1001;
    }

    i__1 = *n;
    for (i = 2; i <= i__1; ++i) {
        e[i - 1] = e[i];
    }

    f = 0.;
    tst1 = 0.;
    e[*n] = 0.;

    i__1 = *n;
    for (l = 1; l <= i__1; ++l) {
        j = 0;
        h = (d__1 = d[l],PetscAbsReal(d__1)) + (d__2 = e[l],PetscAbsReal(d__2));
        if (tst1 < h) {
            tst1 = h;
        }
/*     .......... LOOK FOR SMALL SUB-DIAGONAL ELEMENT .......... */
        i__2 = *n;
        for (m = l; m <= i__2; ++m) {
            tst2 = tst1 + (d__1 = e[m],PetscAbsReal(d__1));
            if (tst2 == tst1) {
                goto L120;
            }
/*     .......... E(N) IS ALWAYS ZERO,SO THERE IS NO EXIT */
/*                THROUGH THE BOTTOM OF THE LOOP .......... */
        }
L120:
        if (m == l) {
            goto L210;
        }
L130:
        if (j == 30) {
            goto L1000;
        }
        ++j;
/*     .......... FORM SHIFT .......... */
        l1 = l + 1;
        l2 = l1 + 1;
        g = d[l];
        p = (d[l1] - g) / (e[l] * 2.);
        r = petsc_LINPACKcgpthy(&p,&c_b10);
        ds = 1.0; if (p < 0.0) ds = -1.0;
        d[l] = e[l] / (p + ds*r);
        d[l1] = e[l] * (p + ds*r);
        dl1 = d[l1];
        h = g - d[l];
        if (l2 > *n) {
            goto L145;
        }

        i__2 = *n;
        for (i = l2; i <= i__2; ++i) {
            d[i] -= h;
        }

L145:
        f += h;
/*     .......... QL TRANSFORMATION .......... */
        p = d[m];
        c = 1.;
        c2 = c;
        el1 = e[l1];
        s = 0.;
        mml = m - l;
/*     .......... FOR I=M-1 STEP -1 UNTIL L DO -- .......... */
        i__2 = mml;
        for (ii = 1; ii <= i__2; ++ii) {
            c3 = c2;
            c2 = c;
            s2 = s;
            i = m - ii;
            g = c * e[i];
            h = c * p;
            r = petsc_LINPACKcgpthy(&p,&e[i]);
            e[i + 1] = s * r;
            s = e[i] / r;
            c = p / r;
            p = c * d[i] - s * g;
            d[i + 1] = h + s * (c * g + s * d[i]);
        }
 
        p = -s * s2 * c3 * el1 * e[l] / dl1;
        e[l] = s * p;
        d[l] = c * p;
        tst2 = tst1 + (d__1 = e[l],PetscAbsReal(d__1));
        if (tst2 > tst1) {
            goto L130;
        }
L210:
        p = d[l] + f;
/*     .......... ORDER EIGENVALUES .......... */
        if (l == 1) {
            goto L250;
        }
/*     .......... FOR I=L STEP -1 UNTIL 2 DO -- .......... */
        i__2 = l;
        for (ii = 2; ii <= i__2; ++ii) {
            i = l + 2 - ii;
            if (p >= d[i - 1]) {
                goto L270;
            }
            d[i] = d[i - 1];
        }

L250:
        i = 1;
L270:
        d[i] = p;
    }

    goto L1001;
/*     .......... SET ERROR -- NO CONVERGENCE TO AN */
/*                EIGENVALUE AFTER 30 ITERATIONS .......... */
L1000:
    *ierr = l;
L1001:
    PetscFunctionReturn(0);
} /* cgtql1_ */

#undef __FUNCT__  
#define __FUNCT__ "LINPACKcgpthy"
static PetscReal petsc_LINPACKcgpthy(PetscReal *a,PetscReal *b)
{
    /* System generated locals */
    PetscReal ret_val,d__1,d__2,d__3;
 
    /* Local variables */
    PetscReal p,r,s,t,u;

    PetscFunctionBegin;
/*     FINDS DSQRT(A**2+B**2) WITHOUT OVERFLOW OR DESTRUCTIVE UNDERFLOW */


/* Computing MAX */
    d__1 = PetscAbsReal(*a),d__2 = PetscAbsReal(*b);
    p = PetscMax(d__1,d__2);
    if (!p) {
        goto L20;
    }
/* Computing MIN */
    d__2 = PetscAbsReal(*a),d__3 = PetscAbsReal(*b);
/* Computing 2nd power */
    d__1 = PetscMin(d__2,d__3) / p;
    r = d__1 * d__1;
L10:
    t = r + 4.;
    if (t == 4.) {
        goto L20;
    }
    s = r / t;
    u = s * 2. + 1.;
    p = u * p;
/* Computing 2nd power */
    d__1 = s / u;
    r = d__1 * d__1 * r;
    goto L10;
L20:
    ret_val = p;
    PetscFunctionReturn(ret_val);
} /* cgpthy_ */

