/**
 * @file pc/pc.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines the CUKR_SOLVER abstract
 * interace.
 * 
 * The CUKR_SOLVER structure stores all
 * member variables (characteristic) and
 * function pointers necessary to define
 * a solver. Since all solvers must 
 * implement the functions defined here,
 * the class functions as a solver 
 * interface. 
 */

#ifndef CUKR_SOLVER_H_
#define CUKR_SOLVER_H_

/* ========= Forward declarations for structures ======== */
typedef struct _CUKR_SOLVER*     CUKR_SOLVER;
typedef struct _CUKR_SOLVER_OPS CUKR_SOLVER_OPS;
/* ======================================================= */

#include "../mat_vec/matrix.h"
#include "../mat_vec/vector.h"
#include "../blas/blas.h"
#include "../pc/pc.h"

/* Define solver types. Currently
 * only the CG solver is supported. */


		// ======================== DEFINES ======================== //


/**
 * Denotes a CG solver
 */
#define CUKR_SOLVER_TYPE_CG 1


/*
 * Convergence/divergence reasons
 */

/** Converged due to small enough relative
 * residual */ 
#define CUKR_CONVERGED_RTOL    1
/** Converged due to small enough absolute
 * residual */ 
#define CUKR_CONVERGED_ABSTOL  2 
/** Converged due to enough iterations
 * being spent */ 
#define CUKR_CONVERGED_ITS     3
/** Converged due to the best possible
 * residual value being reached */ 
#define CUKR_CONVERGED_BESTTOL 4
/** Diverged due to residual exceeding the
 * threshold value */ 
#define CUKR_DIVERGED_DTOL   -1
/** Diverged due to exceeding the maximum
 * no. of allowed iterations */ 
#define CUKR_DIVERGED_ITS    -2

/*
 * Differentiate between inner/outer 
 * solver 
 */
/** Actual (inner, for itref) solver */
#define CUKR_INNER 0
/** Outer correction iterations */
#define CUKR_OUTER 1 
 
/*
 * Counters for the solver
 */ 
 
/** Max. no. of counters for the solver */
#define CUKR_SOLVER_CNTR_VECTOR_LENGTH 1000

/** Holds things related to the regular
 * solver mode (no itref) */
#define SOLVER_IDX_NORMAL  0

/** Holds things related to the inner
 * solver in case of itref */
#define SOLVER_IDX_INNER   100

/** Holds things related to the outer
 * solver in case of itref */
#define SOLVER_IDX_OUTER   200

/** Holds things related to the outer
 * solver in case of itref */
#define SOLVER_IDX_CONDEST 300

/** Offset for iteration count */
#define SOLVER_OFFSET_ITS 0

/** Solver for elapsed time */
#define SOLVER_OFFSET_TIME 10


 

 

		// ===================== CONFIGURATION ===================== //

/* Various contants used to configure
 * the solver behaviour */


/** 
 * Maximum no. of entries allowed in
 * history */
#define CUKR_SOLVER_MAX_HISTORY 1000000   
/** Maximum no. of allowed monitors */
#define CUKR_SOLVER_MAX_MONITORS 100


        // ================== INTERFACE DECLARATIONS =============== //


/* Define all the function pointers
 * to the function needed to implement
 * the solver */
  

/* Function pointers for solver 
 * operations */
typedef void (*CUKR_SOLVER_OPS_SETUP)(CUKR_SOLVER);  
typedef void (*CUKR_SOLVER_OPS_SOLVE)(CUKR_SOLVER); 
typedef void (*CUKR_SOLVER_OPS_SETFROMOPTS)(CUKR_SOLVER);
typedef void (*CUKR_SOLVER_OPS_DESTROY)(CUKR_SOLVER);  
 
/** Sturcture holding the solver 
 * operations (function pointers to
 * to what each solver has to implement)
 */
struct _CUKR_SOLVER_OPS {
	CUKR_SOLVER_OPS_SETUP       setup;
	CUKR_SOLVER_OPS_SOLVE       solve;
	CUKR_SOLVER_OPS_SETFROMOPTS setfromopts;
	CUKR_SOLVER_OPS_DESTROY     destroy;
}; 
 
/** Structure holding a solver monitor */
typedef void (*CUKR_SOLVER_MONITOR)(CUKR_SOLVER, int);

/** The actual solver interface, 
 * containing members and function
 * pointers 
 */ 
struct _CUKR_SOLVER{
	
	// ------- MEMBERS --------- //
	
	/** Name */
	char name[255];
	/** If the solver was setted up or not */
	int setted_up;
	
	/** Precision */
	int precision;
	/** The maximum no. of iterations for 
	 * actual (inner, in case of itref) solver*/
	int max_it;
	/** The minimum no. of iterations for 
	 * the inner solver (applies only in
	 * case of itref)*/
	int min_it;
	/** The maximum no. of iteration for the
	 * outer correction loop, if itref */
	int max_it_out;
	/** Current iteration for the actual
	 * (inner, in case of itref) solver.
	 * This value will reset at each call */
	int its;
	/** Current iteration for the actual 
	 * (inner, in case of itref) solver. 
	 * This value does not reset after each
	 * call, and showes the total number of
	 * iterations spent so far in the inner
	 * solver */  
	int its_total;
	/** Current iteration for the outer correction
	 * loop, if itref */ 
	int its_out; 
	/** Precondioner object */
	CUKR_PC pc;
	/** Preconditioner side (left,
	 * right, symmetric */
	int pc_side;
	   
	/** Relative tolerance for the
	 * actual (inner, in case of itref) solver*/
	CUKR_DOUBLE rtol;
	/** Relative tolerance for the
	 * outer correction loop, if itref */
	CUKR_DOUBLE rtol_out;
	/** Set inner rtol to auto or not */
	int inner_rtol_auto;
	/** Set outer rtol to auto or not */
	int outer_rtol_auto;
	/** Absolute tolerance for the
	 * actual (inner, in case of itref) solver*/
	CUKR_DOUBLE abstol;
	/** Absolute tolerance for the
	 * outer correction loop, if itref */
	CUKR_DOUBLE abstol_out;
	/** Divergence tolerance for the
	 * actual (inner, in case of itref) solver*/
	CUKR_DOUBLE divtol;
	/** Divergence tolerance for the
	 * outer correction loop, if itref */ 
	CUKR_DOUBLE divtol_out;
	/** Initial residual norm (squared)for the
	 * actual (inner, in case of itref) solver*/
	CUKR_DOUBLE rnorm02;
	/** Initial residual norm (squared) for the
	 * outer correction loop, if itref */
	CUKR_DOUBLE rnorm02_out;
	/** Current residual norm (squared) for the
	 * actual (inner, in case of itref) solver*/
	CUKR_DOUBLE rnorm2;
	/** Current residual norm (squared) for the
	 * outer correction loop, if itref */
	CUKR_DOUBLE rnorm2_out;
	/** Convergence reason for the actual
	 * (inner, in case of itref) solver*/
	int reason;
	/** Convergence reason for the
	 * outer correction loop, if itref */
	int reason_out;
	/** Residual history vector for the
	 * actual (inner, in case of itref) solver */
	CUKR_DOUBLE *res_hist;
	/** Residual history vector for the
	 * outer correction loop, if itref */
	CUKR_DOUBLE *res_hist_out;
	/** Current index in residual history
	 * vector for the actual (inner, in 
	 * case of itref) solver */ 
	int res_hist_index;
	/** Current index in residual history 
	 * vector for the outer correction loop, 
	 * if itref */
	int res_hist_index_out;
	/** Whether to reset the history
	 * after each solve or not */
	int res_hist_reset;
	/** Vector of monitors */
	CUKR_SOLVER_MONITOR monitors[CUKR_SOLVER_MAX_MONITORS];
	/** Number of currently registered
	 * monitors */
	int numbermonitors; 
	/** Type of norm to use for convergence
	 * testing */
	int normtype;
	/** Wheather to diagonally scale the 
	 * system or not */
	int dscale;
	/** Whether to diagonally unscale the
	 * system after solving */
	int dscaledix;
	/** Whether to use iterative refinement
	 * for the solution */
	int itref;
	/** Whether to save eigenvalue info or not */
	int eig;
	/** Condition number estimate for matrix */
	CUKR_DOUBLE condest;
	
	/** Vector for holding the tridiagonal matrix
	 * needed to estimate condition number */
	CUKR_DOUBLE *eig_e, *eig_ee, *eig_d, *eig_dd; 
	
	/** Pointer to the system matrix */
	CUKR_MATRIX mat;
	/** Pointer to the RHS vector for the
	 * actual (inner, in case of itref) solver */
	CUKR_VECTOR rhs;
	/** Pointer to the RHS vector for the
	 * outer correction loop, if itref */
	CUKR_VECTOR rhs_out;
	/** Initial solution vector */
	CUKR_VECTOR sol_init;
	/** Current solution vector for the
	 * actual (inner, in case of itref) solver */ 
	CUKR_VECTOR sol;
	/** Current solution vector for the
	 * outer correction loop, if itref */
	CUKR_VECTOR sol_out;
	
	/** BLAS to use for the actual
	 * (inner, in case of itref) solver */
	CUKR_BLAS blas;
	/** BLAS to use for the the
	 * outer correction loop, if itref */
	CUKR_BLAS blas_out;
	
	/** Solver operations */
	CUKR_SOLVER_OPS ops; 
	
	/** Data holder for implementations */
	void *data;
	
	/** Counters */
	CUKR_DOUBLE counters[CUKR_SOLVER_CNTR_VECTOR_LENGTH];
	
	/* Counter for the total flops executed */
	CUKR_DOUBLE flops;
	
	/* Reason for convergence/divergenc */
	char reason_str[255];
	
};


        // ================== SOLVER RELATED FUNCTIONS ============== //

/* Functions common to all solvers.
 * Basically, get or set some member
 * of the solver structure */
 
/**
 * \brief Create a solver structure. The 
 *        actual solver type is decided 
 *        through command-line options.
 * \param solver Pointer to the solver
 *        object 
 * \param name Solver name
 */
void CukrSolverCreate(CUKR_SOLVER* solver, char* name);

// ------------------------------------------------------------------- //

/**
 * \brief Set the system matrix of the
 *        system to be solved
 * \param solver The solver object 
 * \param mat The system matrix
 */   
void CukrSolverSetOperators(CUKR_SOLVER solver, CUKR_MATRIX mat); 

                        
// ------------------------------------------------------------------- //                        
                        
/**
 * \brief Solve the system
 * \param solver The solver
 *        object 
 * \param rhs The rhs vector
 * \param sol_init The initial 
 *        solution vector. If null, the
 *        init. solution will be considered
 *        to be zero.
 */
void CukrSolverSolve(CUKR_SOLVER solver, CUKR_VECTOR rhs, 
                     CUKR_VECTOR sol_init);

// ------------------------------------------------------------------- //

/*
 * \brief Computes a condition number estimate
 * \param solver The solver object
 * \return The cond. est.
 */
 
CUKR_DOUBLE CukrSolverCondest(CUKR_SOLVER solver);

// ------------------------------------------------------------------- //

/*
 * \brief Returns the result vector,
 *        after the solve
 * \param solver The solver object
 * \return The result vector
 */
CUKR_VECTOR CukrSolverGetResult(CUKR_SOLVER solver); 

// ------------------------------------------------------------------- //

/**
 * \brief After a solve, return the
 *        convergence/divergence reason
 * \param solver The solver object 
 */               
int CukrSolverGetConvergedReason(CUKR_SOLVER solver);

// ------------------------------------------------------------------- //

/**
 * \brief After a solve, return the number
 *        of iterations spent
 * \param solver The solver object 
 */
int CukrSolverGetIterationNumber(CUKR_SOLVER solver);  

// ------------------------------------------------------------------- //

/**
 * \brief Register a monitor to show/record
 *        various info related to the solver
 * \param solver The solver object 
 * \param monitor The monitor to be added
 */
void CukrSolverAddMonitor(CUKR_SOLVER solver, CUKR_SOLVER_MONITOR mon);                  

// ------------------------------------------------------------------- //

/**
 * \brief Destroy a solver object
 * \param solver The solver object 
 */
void CukrSolverDestroy(CUKR_SOLVER* solver); 

// ------------------------------------------------------------------- //

/**
 * \brief After a solve, return the
 *        convergence/divergence reason
 * \param solver The solver object 
 * \param for_inner If true, in case of iterative
 *        refinement, reports are done for the 
 *        inner solver and if false, for the outer
 *        solver
 */               
const char* CukrSolverGetConvergedReasonAsStr(CUKR_SOLVER solver, int inner_outer);

#endif /*CUKR_SOLVER_H_*/
