/**
 * @file util/threading.c
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains functions used to support
 * the multithreaded need for using
 * multiple GPUs
 * 	
 */

#define LOGGER_NAME "CUKr.util.threading"

#include <unistd.h>
#include <string.h>
#include <pthread.h>
#ifdef __APPLE__
	#include <mach/semaphore.h>
	#include <mach/task.h>
	#include <mach/mach.h>
	#define SEM_T semaphore_t
#else
	#include <semaphore.h>
	#define SEM_T sem_t
#endif
#include "threading.h" 
#include "../util/log.h"
#include "../util/utils.h" 
#include "../util/macros.h"
#include "../init/init.h"
#include "../db/db.h"
#include "../blas/impl/generic/generic_blas.h"
#include "../blas/sp/blas1_sp.h"
#include "../blasx/sp/blasx1_sp.h"
#include "../blas/sp/blas2_sp.h"

/* Get access to the database */ 
extern CukrDB cukr_db;  

#define _REENTRANT

/* Declare the task store */
CUKR_THREADS_TASKSTORE cukr_taskstore[CUKR_MAX_PE];

/* Declare the returned values store */
CUKR_THREADS_RETSTORE cukr_retstore[CUKR_MAX_PE];

/* Declare the vector of thread ids as global.
 * Must do that in order for "real" addresses
 * to be given to the thread start function.
 * If the vector is declared locally, the 
 * thread functions get junk.
 */
int cukr_threadIds[CUKR_MAX_PE];
pthread_t cukr_threads[CUKR_MAX_PE];

/* Semaphores to signal that here is a task
 * ready to be executed.*/
SEM_T cukr_sem_jobstart[CUKR_MAX_PE];

/* Semaphore to signal that a task is finished.
 * It starts at a value equal with the no. of
 * threads, and each one decreses it after 
 * executing the task.*/
SEM_T cukr_sem_jobdone[CUKR_MAX_PE];

/* Mutex and condition, used for broadcasing
 * the start of a job */
pthread_cond_t  cukr_cond  = PTHREAD_COND_INITIALIZER;
pthread_mutex_t cukr_mutex = PTHREAD_MUTEX_INITIALIZER;

/* Portable semaphores */

void platform_sem_create(void * semStructure, int initialValue)
{
    #ifdef __APPLE__
    semaphore_create(mach_task_self(), (semaphore_t *)semStructure, SYNC_POLICY_FIFO, initialValue);
    #else
    int pshared = 0;
    sem_init((sem_t *)semStructure, pshared, initialValue);
    #endif
}

void platform_sem_post(void * semStructure)
{
    #ifdef __APPLE__
    semaphore_signal(*((semaphore_t *)semStructure));
    #else
    sem_post((sem_t *)semStructure);
    #endif
}

void platform_sem_wait(void * semStructure)
{
    #ifdef __APPLE__
    semaphore_wait(*((semaphore_t *)semStructure));
    #else
    sem_wait((sem_t *)semStructure);
    #endif
}



/**
 * \brief Empties the tasklist for the
 *        specified thread
 * \param tid The thread id 
 */
void EmptyTasks(int tid)
{
	cukr_taskstore[tid].funcid  = CUKR_FUNCID_NONE;
	cukr_taskstore[tid].funcptr = NULL;
}

/**
 * \brief The main loop of the worker threads
 * \param tid The thread id of the thread 
 *        being created
 */
void* CukrThreadFunc(void *ptid) 
{
	
	/* Store thread id locally */
	int tid = *((int*)ptid);
	int val[] = {tid};

	/* Set the key to the active thread */
	pthread_setspecific(cukr_db.thread_key, val);

	LOGX(DEBUG, "Thread %d started", tid);
	printf("Thread %d started\n", tid);

	/* Set the GPU id */
	#ifdef CUKR_USE_GPU
		if (cukr_db.compute_on_gpu)
			CukrSetDevice(tid);
	#endif	
	
	/* Run forever, until the kill function
	 * id signal comes */ 
	while(1)
	{
		/* Sleep until semaphore is green */
		//printf("Thread %d wait in while(1)\n", tid);
		platform_sem_wait(&(cukr_sem_jobstart[tid]));
		//platform_sem_wait(&sem);
		//printf("Thread %d passed wait after while(1)\n", tid);
		//pthread_mutex_lock(&cukr_mutex);
		//printf("wait\n");
		//pthread_cond_wait(&cukr_cond, &cukr_mutex);
		//pthread_mutex_unlock(&cukr_mutex);
		//printf("done\n");
					
		/* If nothing in tasklist, back to 
		 * sleep */
		if (cukr_taskstore[tid].funcid == CUKR_FUNCID_NONE)
			continue;
			
		/* If kill signal, exit loop */
		if (cukr_taskstore[tid].funcid == CUKR_FUNCID_KILL)
		{
			LOGX(DEBUG, "[T_%d] Stopping ... ", tid);
			break;
		}
				
		/* If something else, extract args and execute */
		switch (cukr_taskstore[tid].funcid)
		{
			/* Process dummy task */
			case CUKR_FUNCID_DUMMY:
			{											
				/* Execute function */ 
				//printf("[T_%d] DUMMY \n", tid);				
				break;
			}
						
			/* Process vector destroy */
			case CUKR_FUNCID_VEC_DESTROY:
			{											
				LOGX(DEBUG, "[T_%d] Destroying vector %s ... ",
						    tid, (*cukr_taskstore[tid].pvargs[0])->name);
				/* Execute function */ 
				((CUKR_VECSP_DESTROY)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].pvargs[0]);
				
				break;
			}
			/* Process vector malloc */
			case CUKR_FUNCID_VEC_MALLOC:
			{							
				LOGX(DEBUG, "[T_%d] Malloc vector %s ... ",
						    tid, (cukr_taskstore[tid].vargs[0])->name);
				
				/* Execute function */ 
				((CUKR_VECSP_MALLOC)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].vargs[0],
				                                                 cukr_taskstore[tid].iargs[0],
				                                                 cukr_taskstore[tid].iargs[1]);
				
				break;
			}
			/* Process vector transfer */
			case CUKR_FUNCID_VEC_TRANSFER:
			{							
				LOGX(DEBUG, "[T_%d] Transferring vector %s ... ",
						    tid, (cukr_taskstore[tid].vargs[0])->name);
				
				/* Execute function */ 
				((CUKR_VECSP_TRANSFER)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].vargs[0],
				                                                   cukr_taskstore[tid].iargs[0],
				                                                   cukr_taskstore[tid].iargs[1]);			
				break;
			}
			/* Process vector clone */
			case CUKR_FUNCID_VEC_CLONE:
			{					
				LOGX(DEBUG, "[T_%d] Cloning vector %s ... ",
						    tid, (cukr_taskstore[tid].vargs[0])->name);
				
				/* Execute function */ 
				((CUKR_VECSP_CLONE)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].vargs[0],
				                                                cukr_taskstore[tid].pvargs[0],
				                                                cukr_taskstore[tid].pcargs[0]);			
				break;
			}
			/* Process matrix destroy */
			case CUKR_FUNCID_MAT_DESTROY:
			{							
				LOGX(DEBUG, "[T_%d] Destroying matrix %s ... ",
						    tid, (*cukr_taskstore[tid].pmargs[0])->name);
				
				/* Execute function */ 
				((CUKR_MAT_DESTROY)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].pmargs[0]);
				
				break;
			}
			/* Process matrix transfer */
			case CUKR_FUNCID_MAT_TRANSFER:
			{							
				LOGX(DEBUG, "[T_%d] Transferring matrix %s ... ",
						    tid, (cukr_taskstore[tid].margs[0])->name);
				
				/* Execute function */ 
				((CUKR_MAT_TRANSFER)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].margs[0],
				                                                   cukr_taskstore[tid].iargs[0],
				                                                   cukr_taskstore[tid].iargs[1]);			
				break;
			}
			/* Process matrix transfer */
			case CUKR_FUNCID_MAT_CONVERT:
			{							
				LOGX(DEBUG, "[T_%d] Converting matrix %s ... ",
						    tid, (cukr_taskstore[tid].margs[0])->name);
				
				/* Execute function */ 
				((CUKR_MAT_CONVERT)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].margs[0],
				                                                  cukr_taskstore[tid].iargs[0],
				                                                  cukr_taskstore[tid].iargs[1]);			
				break;
			}
			/* Process DOT */
			case CUKR_FUNCID_DOT:
			{							
				LOGX(DEBUG, "[T_%d] Computing dot product "\
				            "between vectors %s and %s... ", tid, 
				            (cukr_taskstore[tid].vargs[0])->name,  (cukr_taskstore[tid].vargs[1])->name);
				
				/* Execute function */ 
				cukr_retstore[tid].dval = 
				((CUKR_BLAS1_DOT_SP)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].bargs[0],
				                                                cukr_taskstore[tid].vargs[0],
				                                                cukr_taskstore[tid].vargs[1]);
				/* Set the returned value flag */
				cukr_retstore[tid].flag_dval = CUKR_TRUE;                                                			
				break;
			}
			/* Process AXPY */
			case CUKR_FUNCID_AXPY:
			{							
				LOGX(DEBUG, "[T_%d] Computing AXPY "\
				            "between vectors %s and %s... ", tid, 
				            (cukr_taskstore[tid].vargs[0])->name,  (cukr_taskstore[tid].vargs[1])->name);
				
				/* Execute function */ 
				((CUKR_BLAS1_AXPY_SP)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].bargs[0],
																 cukr_taskstore[tid].dargs[0],
				                                                 cukr_taskstore[tid].vargs[0],
				                                                 cukr_taskstore[tid].vargs[1]);                                              			
				break;
			}
			/* Process AYPX */
			case CUKR_FUNCID_AYPX:
			{							
				LOGX(DEBUG, "[T_%d] Computing AYPX "\
				            "between vectors %s and %s... ", tid, 
				            (cukr_taskstore[tid].vargs[0])->name,  (cukr_taskstore[tid].vargs[1])->name);
				
				/* Execute function */ 
				((CUKR_BLAS1_AYPX_SP)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].bargs[0],
																 cukr_taskstore[tid].dargs[0],
				                                                 cukr_taskstore[tid].vargs[0],
				                                                 cukr_taskstore[tid].vargs[1]);                                              			
				break;
			}
			/* Process AYPX */
			case CUKR_FUNCID_SCAL:
			{							
				LOGX(DEBUG, "[T_%d] Scaling vector%s... ", tid,
				            (cukr_taskstore[tid].vargs[0])->name);
				
				/* Execute function */ 
				((CUKR_BLAS1_SCAL_SP)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].bargs[0],
																 cukr_taskstore[tid].dargs[0],
				                                                 cukr_taskstore[tid].vargs[0]);                                              			
				break;
			}
			/* Process AYPX */
			case CUKR_FUNCID_COPY:
			{							
				LOGX(DEBUG, "[T_%d] Copying vector %s to %s... ", tid, 
				            (cukr_taskstore[tid].vargs[0])->name,  (cukr_taskstore[tid].vargs[1])->name);
				
				/* Execute function */ 
				((CUKR_BLAS1_COPY_SP)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].bargs[0],
				                                                 cukr_taskstore[tid].vargs[0],
				                                                 cukr_taskstore[tid].vargs[1]);                                              			
				break;
			}
			/* Process PWPR */
			case CUKR_FUNCID_PWPR:
			{							
				LOGX(DEBUG, "[T_%d] Computing PWPR "\
				            "between vectors %s and %s -> %s... ", tid, 
				            (cukr_taskstore[tid].vargs[0])->name,  (cukr_taskstore[tid].vargs[1])->name, (cukr_taskstore[tid].vargs[2])->name);
				
				/* Execute function */ 
				((CUKR_BLASX1_PWPR_SP)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].bargs[0],
																  cukr_taskstore[tid].vargs[0],
				                                                  cukr_taskstore[tid].vargs[1],
				                                                  cukr_taskstore[tid].vargs[2]);                                              			
				break;
			}
			
			/* Process multi-gpu SPMV */
			case CUKR_FUNCID_SPMV:
			{							
				LOGX(DEBUG, "[T_%d] Computing SPMV between matrix %s, "\
				            "vector %s -> result in vector %s", tid, 
				            (cukr_taskstore[tid].margs[0])->name, (cukr_taskstore[tid].vargs[0])->name, (cukr_taskstore[tid].vargs[1])->name);
				
				/* Execute function */ 
				((CUKR_BLAS2_SPMV_SP)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].bargs[0],
								   						         cukr_taskstore[tid].dargs[0],
																 cukr_taskstore[tid].margs[0],
																 cukr_taskstore[tid].vargs[0],
				                                                 cukr_taskstore[tid].dargs[1],
				                                                 cukr_taskstore[tid].vargs[1]);                                              			
				break;
			}
			
			/* Process multi-gpu vector sync (send) */
			case CUKR_FUNCID_VECMP_SYNC_SEND:
			{							
				LOGX(DEBUG, "[T_%d] Synchronizing (send) vector %s", tid, 
				            (cukr_taskstore[tid].vargs[0])->name);
				
				/* Execute function */ 
				((CUKR_VECMP_SYNC_SEND)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].vargs[0],
																      cukr_taskstore[tid].vargs[1],
																      cukr_taskstore[tid].ccargs[0],
				                                                      cukr_taskstore[tid].iargs[0],
				                                                      cukr_taskstore[tid].iargs[1]);                                              			
				break;
			}
			
			/* Process multi-gpu vector sync (receive) */
			case CUKR_FUNCID_VECMP_SYNC_RECV:
			{							
				LOGX(DEBUG, "[T_%d] Synchronizing (receive) vector %s", tid, 
				            (cukr_taskstore[tid].vargs[0])->name);
				
				/* Execute function */ 
				((CUKR_VECMP_SYNC_RECV)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].vargs[0],
																      cukr_taskstore[tid].vargs[1],
																      cukr_taskstore[tid].ccargs[0],
				                                                      cukr_taskstore[tid].iargs[0],
				                                                      cukr_taskstore[tid].iargs[1]);                                              			
				break;
			}
			
			/* Process get element */
			case CUKR_FUNCID_VEC_GET_ELEMENT:
			{							
				LOGX(DEBUG, "[T_%d] Reading element %d from vector %s",
				            tid, cukr_taskstore[tid].iargs[0],
				            (cukr_taskstore[tid].vargs[0])->name);
				
				/* Execute function */ 
				cukr_retstore[tid].dval = 
				((CUKR_VECSP_GET_ELEMENT)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].vargs[0],
				                                                        cukr_taskstore[tid].iargs[0],
				                                                        cukr_taskstore[tid].iargs[1]);
				/* Set the returned value flag */
				cukr_retstore[tid].flag_ival = CUKR_TRUE;                                                			
				break;
			}
			
			/* Process get element */
			case CUKR_FUNCID_VEC_SET_ELEMENT:
			{							
				LOGX(DEBUG, "[T_%d] Setting element %d from vector %s to value %g", tid,
				            cukr_taskstore[tid].iargs[0], (cukr_taskstore[tid].vargs[0])->name,
				            cukr_taskstore[tid].dargs[0]);
				
				/* Execute function */ 
				((CUKR_VECSP_SET_ELEMENT)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].vargs[0],
				                                                        cukr_taskstore[tid].iargs[0],
				                                                        cukr_taskstore[tid].dargs[0],
				                                                        cukr_taskstore[tid].iargs[1]);                                               			
				break;
			}			
			/* Process get set from vector */
			case CUKR_FUNCID_VEC_SET_FROMVEC_D:
			{							
				LOGX(DEBUG, "[T_%d] Setting vector %s from vector of doubles",
				            tid, (cukr_taskstore[tid].vargs[0])->name);
				
				/* Execute function */ 
				((CUKR_VECSP_SET_FROMVEC_D)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].vargs[0],
				                                                          cukr_taskstore[tid].iargs[0],
				                                                          cukr_taskstore[tid].iargs[1],
				                                                          cukr_taskstore[tid].pdargs[0],
				                                                          cukr_taskstore[tid].iargs[2]);                                              			
				break;
			}
			/* Process get set from vector */
			case CUKR_FUNCID_VEC_SET_FROMVEC_Q:
			{							
				LOGX(DEBUG, "[T_%d] Setting vector %s from vector of quasi-doubles",
				            tid, (cukr_taskstore[tid].vargs[0])->name);
				
				/* Execute function */ 
				((CUKR_VECSP_SET_FROMVEC_Q)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].vargs[0],
				                                                          cukr_taskstore[tid].iargs[0],
				                                                          cukr_taskstore[tid].iargs[1],
				                                                          cukr_taskstore[tid].psargs[0],
				                                                          cukr_taskstore[tid].psargs[0],
				                                                          cukr_taskstore[tid].iargs[2]);                                              			
				break;
			}
			/* Process get set from vector */
			case CUKR_FUNCID_VEC_SET_FROMVEC_S:
			{							
				LOGX(DEBUG, "[T_%d] Setting vector %s from vector of singles",
				            tid, (cukr_taskstore[tid].vargs[0])->name);
				
				/* Execute function */ 
				((CUKR_VECSP_SET_FROMVEC_S)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].vargs[0],
				                                                          cukr_taskstore[tid].iargs[0],
				                                                          cukr_taskstore[tid].iargs[1],
				                                                          cukr_taskstore[tid].psargs[0],
				                                                          cukr_taskstore[tid].iargs[2]);                                              			
				break;
			}
			/* Process get to vector */
			case CUKR_FUNCID_VEC_GET_TOVEC_D:
			{							
				LOGX(DEBUG, "[T_%d] Getting vector %s in vector of doubles",
				            tid, (cukr_taskstore[tid].vargs[0])->name);
				
				/* Execute function */ 
				((CUKR_VECSP_GET_TOVEC_D)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].vargs[0],
				                                                        cukr_taskstore[tid].iargs[0],
				                                                        cukr_taskstore[tid].iargs[1],
				                                                        cukr_taskstore[tid].pdargs[0],
				                                                        cukr_taskstore[tid].iargs[2]);                                              			
				break;
			}			
			/* Process get to vector */
			case CUKR_FUNCID_VEC_GET_TOVEC_Q:
			{							
				LOGX(DEBUG, "[T_%d] Getting vector %s in vector of quasi-doubles",
				            tid, (cukr_taskstore[tid].vargs[0])->name);
				
				/* Execute function */ 
				((CUKR_VECSP_GET_TOVEC_Q)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].vargs[0],
				                                                        cukr_taskstore[tid].iargs[0],
				                                                        cukr_taskstore[tid].iargs[1],
				                                                        cukr_taskstore[tid].psargs[0],
				                                                        cukr_taskstore[tid].psargs[1],
				                                                        cukr_taskstore[tid].iargs[2]);                                              			
				break;
			}
			/* Process get to vector */
			case CUKR_FUNCID_VEC_GET_TOVEC_S:
			{							
				LOGX(DEBUG, "[T_%d] Getting vector %s in vector of single",
				            tid, (cukr_taskstore[tid].vargs[0])->name);
				
				/* Execute function */ 
				((CUKR_VECSP_GET_TOVEC_S)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].vargs[0],
				                                                        cukr_taskstore[tid].iargs[0],
				                                                        cukr_taskstore[tid].iargs[1],
				                                                        cukr_taskstore[tid].psargs[0],
				                                                        cukr_taskstore[tid].iargs[2]);                                              			
				break;
			}
			/* Process get to vector */
			case CUKR_FUNCID_VEC_GET_TOVEC_I:
			{							
				LOGX(DEBUG, "[T_%d] Getting vector %s in vector of doubles",
				            tid, (cukr_taskstore[tid].vargs[0])->name);
				
				/* Execute function */ 
				((CUKR_VECSP_GET_TOVEC_I)(cukr_taskstore[tid].funcptr))(cukr_taskstore[tid].vargs[0],
				                                                        cukr_taskstore[tid].iargs[0],
				                                                        cukr_taskstore[tid].iargs[1],
				                                                        cukr_taskstore[tid].piargs[0],
				                                                        cukr_taskstore[tid].iargs[2]);                                              			
				break;
			}			
			
			/* If none, fatal error */
			default: 
			{
				CUKR_FATALX("Unknown function id: %d ", 
				            cukr_taskstore[tid].funcid);
			}
			
		}
				
		/* Clean tasklist */
		EmptyTasks(tid); 		
				
		/* Gree semaphore for job-done for
		 * this thread */ 
		platform_sem_post(&cukr_sem_jobdone[tid]);
	}
	
	/* Kill the thread */
	LOGX(DEBUG, "[T_%d] Stopped.", tid);
	platform_sem_post(&cukr_sem_jobdone[tid]);
    pthread_exit(NULL);
}

// ------------------------------------------------------------------- //

/**
 * \brief Creates and starts the worker threads
 * \param count The number of threads to start
 */ 
void CukrThreadsCreate(int count)
{
	int i;
	
	LOGX(INFO, "Creating %d worker threads", count);
	
	/* Empty the tasklists */
	for(i = 0; i < count; i++)
		EmptyTasks(i);
	
	/* Init semaphores */
	//platform_sem_create(&sem, 0);
	for(i = 0; i < count; i++) {
		platform_sem_create(&(cukr_sem_jobstart[i]), 0);
		platform_sem_create(&(cukr_sem_jobdone[i]) , 0);
	}
	
	/* Init retstore with zeros */
	for (i = 0; i < count; i++) {
		cukr_retstore[i].dval = 0;
		cukr_retstore[i].flag_dval = 0;
	}
	
	/* Starts the threads */
	for(i = 0; i < count; i++) {
		cukr_threadIds[i] = i; 
        pthread_create(&cukr_threads[i], NULL, CukrThreadFunc, (void*)(&cukr_threadIds[i]));
	}	
        
    LOGX(INFO, "%d worker threads successfully created", count);
}

// ------------------------------------------------------------------- //

/**
 * \brief Stops all worker threads
 * \param count The total no. of threads
 */ 
void CukrThreadsStop(int count)
{
	volatile int i;
	
	LOGX(INFO, "Sending stop signal to all (%d) " \
	           "worker threads", count);
	
	/* Send a dummy function ID as a signal to 
	 * stop the threads */
	
	for (i = 0; i < count; i++)
		CukrThreadsRun(i, NULL, CUKR_FUNCID_KILL);
		
	/* Wait for all threads to finish */
	CukrThreadsWaitCompletion(count);	
}

// ------------------------------------------------------------------- //

/**
 * \brief Add a function pointer to the tasks
 *        to be processed by a thread
 * \param tid The id of the thread that will 
 *        process the task
 * \param funcptr The pointer to the function to
 *        be executed
 * \param FUNCT_ID the id of the function to be 
 *        executed, so we know to which type to cast
 *        it and how to give the parameters
 * \param ... All the parameter needed by the function
 *        to be executed
 */  

void CukrThreadsRun(int tid, void* funcptr, int funcid, ...)
{
	LOGX(DEBUG, "Adding task (id=%d) to thread %d", 
	            funcid, tid);

	/* Update counter */
	cukr_db.threadsRunCalls++;
		
	/* Get the argument list */
	va_list arglist;
	va_start(arglist, funcid);

	/* Depending on the function type,
	 * extract and store the parameters */
	if (funcid != CUKR_FUNCID_KILL && funcid != CUKR_FUNCID_NONE && funcid != CUKR_FUNCID_DUMMY)
	{
		switch (funcid)
		{
			/* For vector destroy */
			case CUKR_FUNCID_VEC_DESTROY:
			{				
				cukr_taskstore[tid].pvargs[0] = va_arg(arglist, CUKR_VECTOR_SP*);
				break;
			}
			/* For vector malloc */
			case CUKR_FUNCID_VEC_MALLOC:
			case CUKR_FUNCID_VEC_TRANSFER:
			{
				cukr_taskstore[tid].vargs[0] = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].iargs[0] = va_arg(arglist, int);
				cukr_taskstore[tid].iargs[1] = va_arg(arglist, int);
				break;
			}
			/* For vector clone */
			case CUKR_FUNCID_VEC_CLONE:
			{
				cukr_taskstore[tid].vargs[0]  = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].pvargs[0] = va_arg(arglist, CUKR_VECTOR_SP*);
				cukr_taskstore[tid].pcargs[0] = va_arg(arglist, char*);
				break;
			}
			/* For matrix destroy */
			case CUKR_FUNCID_MAT_DESTROY:
			{
				cukr_taskstore[tid].pmargs[0] = va_arg(arglist, CUKR_MATRIX_SP*);
				break;
			}
			/* For matrix transfer */
			case CUKR_FUNCID_MAT_TRANSFER:
			case CUKR_FUNCID_MAT_CONVERT:
			{
				cukr_taskstore[tid].margs[0] = va_arg(arglist, CUKR_MATRIX_SP);
				cukr_taskstore[tid].iargs[0] = va_arg(arglist, int);
				cukr_taskstore[tid].iargs[1] = va_arg(arglist, int);
				break;
			}
			case CUKR_FUNCID_DOT:
			{
				cukr_taskstore[tid].bargs[0] = va_arg(arglist, CUKR_BLAS);
				cukr_taskstore[tid].vargs[0] = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].vargs[1] = va_arg(arglist, CUKR_VECTOR_SP);
				break;
			}
			case CUKR_FUNCID_AXPY:
			case CUKR_FUNCID_AYPX:
			{
				cukr_taskstore[tid].bargs[0] = va_arg(arglist, CUKR_BLAS);
				cukr_taskstore[tid].dargs[0] = va_arg(arglist, CUKR_DOUBLE);
				cukr_taskstore[tid].vargs[0] = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].vargs[1] = va_arg(arglist, CUKR_VECTOR_SP);
				break;
			}
			case CUKR_FUNCID_SCAL:
			{
				cukr_taskstore[tid].bargs[0] = va_arg(arglist, CUKR_BLAS);
				cukr_taskstore[tid].dargs[0] = va_arg(arglist, CUKR_DOUBLE);
				cukr_taskstore[tid].vargs[0] = va_arg(arglist, CUKR_VECTOR_SP);
				break;
			}
			case CUKR_FUNCID_COPY:
			{
				cukr_taskstore[tid].bargs[0] = va_arg(arglist, CUKR_BLAS);
				cukr_taskstore[tid].vargs[0] = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].vargs[1] = va_arg(arglist, CUKR_VECTOR_SP);
				break;
			}
			case CUKR_FUNCID_PWPR:
			{
				cukr_taskstore[tid].bargs[0] = va_arg(arglist, CUKR_BLAS);
				cukr_taskstore[tid].vargs[0] = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].vargs[1] = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].vargs[2] = va_arg(arglist, CUKR_VECTOR_SP);
				break;
			}
			case CUKR_FUNCID_SPMV:
			{
				cukr_taskstore[tid].bargs[0] = va_arg(arglist, CUKR_BLAS);
				cukr_taskstore[tid].dargs[0] = va_arg(arglist, CUKR_DOUBLE);
				cukr_taskstore[tid].margs[0] = va_arg(arglist, CUKR_MATRIX_SP);
				cukr_taskstore[tid].vargs[0] = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].dargs[1] = va_arg(arglist, CUKR_DOUBLE);
				cukr_taskstore[tid].vargs[1] = va_arg(arglist, CUKR_VECTOR_SP);
				break;
			}
			case CUKR_FUNCID_VECMP_SYNC_RECV:
			case CUKR_FUNCID_VECMP_SYNC_SEND:
			{
				cukr_taskstore[tid].vargs[0]  = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].vargs[1]  = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].ccargs[0] = va_arg(arglist, CUKR_CONTAINER_VECTOR_COMMENTRY);
				cukr_taskstore[tid].iargs[0]  = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].iargs[1]  = va_arg(arglist, CUKR_INT);
				break;
			}
			case CUKR_FUNCID_VEC_GET_ELEMENT:
			{
				cukr_taskstore[tid].vargs[0] = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].iargs[0] = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].iargs[1] = va_arg(arglist, CUKR_INT);
				break;
			}
			case CUKR_FUNCID_VEC_SET_ELEMENT:
			{
				cukr_taskstore[tid].vargs[0] = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].iargs[0] = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].dargs[0] = va_arg(arglist, CUKR_DOUBLE);
				cukr_taskstore[tid].iargs[1] = va_arg(arglist, CUKR_INT);
				break;
			}
			case CUKR_FUNCID_VEC_SET_FROMVEC_D:
			{
				cukr_taskstore[tid].vargs[0]  = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].iargs[0]  = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].iargs[1]  = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].pdargs[0] = va_arg(arglist, CUKR_DOUBLE*);
				cukr_taskstore[tid].iargs[2]  = va_arg(arglist, CUKR_INT);
				break;
			}
			case CUKR_FUNCID_VEC_SET_FROMVEC_Q:
			{
				cukr_taskstore[tid].vargs[0]  = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].iargs[0]  = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].iargs[1]  = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].psargs[0] = va_arg(arglist, CUKR_SINGLE*);
				cukr_taskstore[tid].psargs[0] = va_arg(arglist, CUKR_SINGLE*);
				cukr_taskstore[tid].iargs[2]  = va_arg(arglist, CUKR_INT);
				break;
			}
			case CUKR_FUNCID_VEC_SET_FROMVEC_S:
			{
				cukr_taskstore[tid].vargs[0]  = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].iargs[0]  = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].iargs[1]  = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].psargs[0] = va_arg(arglist, CUKR_SINGLE*);
				cukr_taskstore[tid].iargs[2]  = va_arg(arglist, CUKR_INT);
				break;
			}
			case CUKR_FUNCID_VEC_SET_FROMVEC_I:
			{
				cukr_taskstore[tid].vargs[0]  = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].iargs[0]  = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].iargs[1]  = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].piargs[0] = va_arg(arglist, CUKR_INT*);
				cukr_taskstore[tid].iargs[2]  = va_arg(arglist, CUKR_INT);
				break;
			}
			case CUKR_FUNCID_VEC_GET_TOVEC_D:
			{
				cukr_taskstore[tid].vargs[0]  = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].iargs[0]  = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].iargs[1]  = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].pdargs[0] = va_arg(arglist, CUKR_DOUBLE*);
				cukr_taskstore[tid].iargs[2]  = va_arg(arglist, CUKR_INT);
				break;
			}
			case CUKR_FUNCID_VEC_GET_TOVEC_Q:
			{
				cukr_taskstore[tid].vargs[0]  = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].iargs[0]  = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].iargs[1]  = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].psargs[0] = va_arg(arglist, CUKR_SINGLE*);
				cukr_taskstore[tid].psargs[1] = va_arg(arglist, CUKR_SINGLE*);
				cukr_taskstore[tid].iargs[2]  = va_arg(arglist, CUKR_INT);
				break;
			}
			case CUKR_FUNCID_VEC_GET_TOVEC_S:
			{
				cukr_taskstore[tid].vargs[0]  = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].iargs[0]  = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].iargs[1]  = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].psargs[0] = va_arg(arglist, CUKR_SINGLE*);
				cukr_taskstore[tid].iargs[2]  = va_arg(arglist, CUKR_INT);
				break;
			}
			case CUKR_FUNCID_VEC_GET_TOVEC_I:
			{
				cukr_taskstore[tid].vargs[0]  = va_arg(arglist, CUKR_VECTOR_SP);
				cukr_taskstore[tid].iargs[0]  = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].iargs[1]  = va_arg(arglist, CUKR_INT);
				cukr_taskstore[tid].piargs[0] = va_arg(arglist, CUKR_INT*);
				cukr_taskstore[tid].iargs[2]  = va_arg(arglist, CUKR_INT);
				break;
			} 
			
			/* If none, fatal error */
			default: 
			{
				CUKR_FATALX("Unknown function id: %d ", funcid);
			}
		}
		va_end(arglist); 
	}
	
	/* Set all return flags to false */
	cukr_retstore[tid].flag_dval = CUKR_FALSE; 
	cukr_retstore[tid].flag_ival = CUKR_FALSE; 
	
	/* Add the function id and pointer */
	cukr_taskstore[tid].funcptr = funcptr;
	cukr_taskstore[tid].funcid = funcid;
	
	/* Red semaphore for job done */
	platform_sem_create(&cukr_sem_jobdone[tid], 0);
	
	/* Green semaphore for job executing 
	 * and red semaphore for stopping*/
	platform_sem_post(&cukr_sem_jobstart[tid]);
}  

// ------------------------------------------------------------------- //

/**
 * \brief Wait for all count threads to 
 *        complete their tasks
 * \param count The total thread count 
 */
void CukrThreadsWaitCompletion(int count)
{	
	int i;

	for (i = 0; i < count; i++)
		platform_sem_wait(&cukr_sem_jobdone[i]);
			
}

// ------------------------------------------------------------------- //

/**
 * \brief Wait only the specified thread
 *        to complete its task
 * \param tid The thread id 
 */
void CukrThreadsWaitCompletionForThread(int tid)
{	
	platform_sem_wait(&cukr_sem_jobdone[tid]);
}

// ------------------------------------------------------------------- //

/**
 * \brief Get the value (type double) returned 
 *        by the task executed by the specified
 *        thread
 * \param tid The thread id
 */
CUKR_DOUBLE CukrThreadsGetRetval_double(int tid)
{
	/* Error if the return flag is not set */
	if (cukr_retstore[tid].flag_dval == CUKR_FALSE) {
		CUKR_FATAL("Return value of double type expected "\
		           "but nothing was returned. ");
	}
	
	/* Return the value */
	return cukr_retstore[tid].dval;
}
