/**
 * @file util/threading.h
 * @author Serban Georgescu
 * @since 2008
 *
 * Contains functions used to support
 * the multithreaded need for using
 * multiple GPUs
 * 	
 */
 
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include "../blas/blas.h"
#include "../mat_vec/mp/vector_mp.h"
#include "../mat_vec/mp/matrix_mp.h"

#ifndef CUKR_THREADING_H_
#define CUKR_THREADING_H_


/** 
 * The maximum no. of parameter a function
 * is allowed to have 
 * */
#define CUKR_TASK_MAXARGS 256

/** 
 *  The taskstore structure is the place
 *  where a task to be ex void* resultecuted by the worker
 *  threads (both function pointers and 
 *  parameters) are stored.
 */
typedef struct CUKR_THREADS_TASKSTORE_ {
	/** The function ID */
	int   funcid;
	/** Pointer to the function implementation */
	void* funcptr;
	/** List to hold pointers to vectors */
	CUKR_VECTOR_SP* 				   pvargs[CUKR_TASK_MAXARGS];
	CUKR_VECTOR_SP   					vargs[CUKR_TASK_MAXARGS];
	CUKR_MATRIX_SP* 				   pmargs[CUKR_TASK_MAXARGS];
	CUKR_MATRIX_SP   					margs[CUKR_TASK_MAXARGS];
	CUKR_BLAS     					bargs[CUKR_TASK_MAXARGS];
	CUKR_INT      					iargs[CUKR_TASK_MAXARGS];
	CUKR_DOUBLE   					dargs[CUKR_TASK_MAXARGS];
	CUKR_DOUBLE*  				   pdargs[CUKR_TASK_MAXARGS];
	CUKR_SINGLE*  				   psargs[CUKR_TASK_MAXARGS];
	CUKR_INT*  				       piargs[CUKR_TASK_MAXARGS];
	CUKR_CONTAINER_VECTOR_COMMENTRY ccargs[CUKR_TASK_MAXARGS];
	char*        pcargs[CUKR_TASK_MAXARGS];
} CUKR_THREADS_TASKSTORE;

/**
 * The retval structure holds the values
 * returned by the functions exectued by
 * the threads. It has separate places
 * for all data types that are returned,
 * one for each execution thread.
 */
typedef struct CUKR_THREADS_RETSTORE_ {
	
	/** Flags to show the presence or absence
	 * of a valid return value */
	CUKR_INT flag_dval, flag_ival; 
	
	/** List to hold values returned for each thread */
	CUKR_DOUBLE dval;
} CUKR_THREADS_RETSTORE;

/*
 * Function IDs
 */

/** Dummy function id, for test */
#define CUKR_FUNCID_DUMMY -2
/** Dummy function id, used to show an
 * empty task list */
#define CUKR_FUNCID_NONE  -1
/** Dummy function id, used a kill signal */ 
#define CUKR_FUNCID_KILL   0 

#define CUKR_FUNCID_VEC_MALLOC          1
#define CUKR_FUNCID_VEC_DESTROY     	2
#define CUKR_FUNCID_VEC_TRANSFER    	3
#define CUKR_FUNCID_VEC_CLONE       	4

#define CUKR_FUNCID_VEC_GET_ELEMENT     11
#define CUKR_FUNCID_VEC_SET_ELEMENT     12
#define CUKR_FUNCID_VEC_GET_TOVEC_D     13
#define CUKR_FUNCID_VEC_GET_TOVEC_Q     14
#define CUKR_FUNCID_VEC_GET_TOVEC_S     15
#define CUKR_FUNCID_VEC_GET_TOVEC_I     16
#define CUKR_FUNCID_VEC_SET_FROMVEC_D   17
#define CUKR_FUNCID_VEC_SET_FROMVEC_Q   18
#define CUKR_FUNCID_VEC_SET_FROMVEC_S   19
#define CUKR_FUNCID_VEC_SET_FROMVEC_I   20

#define CUKR_FUNCID_MAT_DESTROY   		102
#define CUKR_FUNCID_MAT_TRANSFER  		103
#define CUKR_FUNCID_MAT_CONVERT   		105

#define CUKR_FUNCID_DOT           		200
#define CUKR_FUNCID_AXPY          		210
#define CUKR_FUNCID_AYPX          		220
#define CUKR_FUNCID_SCAL          		230
#define CUKR_FUNCID_COPY          		240
#define CUKR_FUNCID_SPMV				250

#define CUKR_FUNCID_PWPR          		300

#define CUKR_FUNCID_VECMP_SYNC_SEND     400
#define CUKR_FUNCID_VECMP_SYNC_RECV     410


// ------------------------------------------------------------------- //

/**
 * \brief Creates and starts the worker threads
 * \param count The number of threads to start
 */ 
void CukrThreadsCreate(int count);

// ------------------------------------------------------------------- //

/**
 * \brief Stops all worker threads
 * \param count The total no. of threads
 */ 
void CukrThreadsStop(int count);

// ------------------------------------------------------------------- //

/**
 * \brief Add a function pointer to the tasks
 *        to be processed by a thread
 * \param tid The id of the thread that will 
 *        process the task
 * \param funcptr The pointer to the function to
 *        be executed
 * \param FUNCT_ID the id of the function to be 
 *        executed, so we know to which type to cast
 *        it and how to give the parameters
 * \param ... All the parameter needed by the function
 *        to be executed
 */
void CukrThreadsRun(int tid, void* funcptr, int FUNCT_ID, ...);  

// ------------------------------------------------------------------- //

/**
 * \brief Wait for all count threads to 
 *        complete their tasks
 * \param count The total thread count 
 */
void CukrThreadsWaitCompletion(int count);

// ------------------------------------------------------------------- //

/**
 * \brief Wait only the specified thread
 *        to complete its task
 * \param tid The thread id 
 */
void CukrThreadsWaitCompletionForThread(int tid);

// ------------------------------------------------------------------- //

/**
 * \brief Get the value (type double) returned 
 *        by the task executed by the specified
 *        thread
 * \param tid The thread id
 */
CUKR_DOUBLE CukrThreadsGetRetval_double(int tid);


#endif /*CUKR_THREADING_H_*/
