/**
 * @file utils/containers.c
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines containers (e.g. vector 
 * container, that can hold arbitrary
 * data and automatically do memory
 * managemenet and stuff)
*/

#define LOGGER_NAME "CUKr.util.containers"

#include <stdio.h>
#include <stdlib.h>
#ifndef __MACH__
	#include <malloc.h>
#endif
#include "log.h"
#include "containers.h"

int compare (const void * a, const void * b)
{
  return ( *(int*)a - *(int*)b );
}


// -------------------------------------------------------------------------------------//
// =============================  CONTAINER: VECTOR of INT ============================ //
// -------------------------------------------------------------------------------------//



/**
 * \brief Does initialization stuff to make
 *        a vector container ready for use.
 * 		  Start with START_SIZE elements
 * \param v The vector container
 */
void CukrContainersVectorOfIntCreate(CUKR_CONTAINER_VECTOR_INT *v)
{	
	/* 
	 * Allocate container structure 
	 */
	
	if ( (*v = (CUKR_CONTAINER_VECTOR_INT)malloc(sizeof(struct _CUKR_CONTAINER_VECTOR_INT))) == NULL) {
		CUKR_FATAL("Error allocating memory for container structure" );
	} 
	
	/*
	 * Init 
	 */ 
	
	(*v)->size = 0;
	(*v)->allocated = CUKR_CONTAINERS_VECTORINT_START_SIZE;
	if ( ((*v)->elements = (int*)malloc(CUKR_CONTAINERS_VECTORINT_START_SIZE * sizeof(CUKR_INT))) == NULL) {
		CUKR_FATAL("Error allocating initial memory for container");
	} 
	LOG(DEBUG, "Vector of int container created");
}

// ------------------------------------------------------------------- //

/**
 * \brief Free all the data allocated inside
 *        the vector container
 * \param v The vector container
 */
void CukrContainersVectorOfIntDestroy(CUKR_CONTAINER_VECTOR_INT *v)
{
	/* Free elements */
	free((*v)->elements);
	
	/* Delete container structure */
	free(*v);
	
	LOG(DEBUG, "Vector of int container destroyed");
}

// ------------------------------------------------------------------- //

/**
 * \brief Append an element at the end  
 *        of the container
 * \param v The vector container
 * \param elem The element to add
 */
void CukrContainersVectorOfIntAppend(CUKR_CONTAINER_VECTOR_INT v, int elem)
{
	/* If more memory needed, double it */
	if (v->size == v->allocated) {	
		if ( (v->elements = (int*)realloc(v->elements, 2*v->allocated*sizeof(CUKR_INT))) == NULL) {
			CUKR_FATALX("Error allocating %d int elements for container",
						2 * v->allocated);
		}
		v->allocated *= 2; 
	}
	
	/* Add the new element */
	v->elements[v->size++] = elem;
}

// ------------------------------------------------------------------- //

/**
 * \brief Append an element at the end  
 *        of the container. Make sure it
 *        is unique. 
 * \param v The vector container
 * \param elem The element to add
 */
void CukrContainersVectorOfIntAppendU(CUKR_CONTAINER_VECTOR_INT v, int elem)
{
	int i;
		
	/* Make sure the element is not there
	 * already */	 
	for (i = 0; i < v->size; i++)
		if (v->elements[i] == elem)
			return;
			
	/* If unique, append */
	CukrContainersVectorOfIntAppend(v, elem);		 
}

// ------------------------------------------------------------------- //

/**
 * \brief Sort the element of a vector
 *        in increasing order
 * \param v The vector container
 */
void CukrContainersVectorOfIntSortI(CUKR_CONTAINER_VECTOR_INT v)
{
	int i, aux, swapped;
	
	/* Sort using quick-sort */
	qsort (v->elements, v->size, sizeof(int), compare);
}

// ------------------------------------------------------------------- //

/**
 * \brief Sort the element of a vector 
 *        in increasing order and remove
 *        all duplicates
 * \param v The vector container
 */
void CukrContainersVectorOfIntSortUnique(CUKR_CONTAINER_VECTOR_INT v)
{
	int i;
	
	/* Return if no elements */
	if (!v->size)
		return;
	
	/* Sort using quick-sort */
	qsort (v->elements, v->size, sizeof(int), compare);
	
	/* Shrink vector to contain only unique elements */
	int unique_elements = 1;
	for (i = 1; i < v->size; i++)
		if (v->elements[i] !=  v->elements[unique_elements - 1])
			 v->elements[unique_elements++] = v->elements[i];
	v->size = unique_elements;			
}

// ------------------------------------------------------------------- //

/**
 * \brief Print all elements inside a 
 *        vector container
 * \param v The vector container
 */
void CukrContainersVectorOfIntPrint(CUKR_CONTAINER_VECTOR_INT v)
{
	CukrContainersVectorOfIntPrintToFile(v, stdout);
}

// ------------------------------------------------------------------- //

/**
 * \brief Print all elements inside a 
 *        vector container
 * \param f An open file stream to write to
 */
void CukrContainersVectorOfIntPrintToFile(CUKR_CONTAINER_VECTOR_INT v, FILE *f)
{
	CukrContainersVectorOfIntPrintToFileWithIndent(v, stdout, "");
}

// ------------------------------------------------------------------- //

/**
 * \brief Print all elements inside a 
 *        vector container
 * \param v The vector container
 * \param f An open file stream to write to
 * \param indent An indentation string, to be inserted
 *        before every print
 */
void CukrContainersVectorOfIntPrintToFileWithIndent(CUKR_CONTAINER_VECTOR_INT v, FILE *f, char *indent)
{
	int i;
	
	fprintf(f, "%sVector container with %d elem (%d allocated)\n",
	            indent, v->size, v->allocated);
	fprintf(f, "%s", indent);
	if (!v->size)
		fprintf(f, "(empty)");
	else	                       
		for (i = 0; i < v->size; i++)
			fprintf(f, "%d ", v->elements[i]);
	fprintf(f, "\n");       
}



// -------------------------------------------------------------------------------------//
// ==========================  CONTAINER: VECTOR of COMMENTRY  ======================== //
// -------------------------------------------------------------------------------------//

/**
 * \brief Does initialization stuff to make
 *        a vector container ready for use
 * \param v The vector container
 */
void CukrContainersVectorOfCommentryCreate(CUKR_CONTAINER_VECTOR_COMMENTRY *v)
{
	/* 
	 * Allocate container structure 
	 */
	
	if ( (*v = (CUKR_CONTAINER_VECTOR_COMMENTRY)malloc(sizeof(struct _CUKR_CONTAINER_VECTOR_COMMENTRY))) == NULL) {
		CUKR_FATAL("Error allocating memory for container structure" );
	} 
	
	/* 
	 * Set stuff
	 */
	 
	(*v)->size = 0;
	(*v)->comm_volume = 0;
	
	/*
	 * Allocate initial amount of elemens
	 */
	
	(*v)->allocated = CUKR_CONTAINERS_VECTORCOMMENTRY_START_SIZE;
	if ( ((*v)->elements = (CUKR_COMMENTRY*)malloc(CUKR_CONTAINERS_VECTORCOMMENTRY_START_SIZE * sizeof(struct _CUKR_COMMENTRY))) == NULL) {
		CUKR_FATAL("Error allocating initial memory for container");
	}
	LOG(DEBUG, "Vector of comm-entry container created");	
} 

// ------------------------------------------------------------------- //

/**
 * \brief Free all the data allocated inside
 *        the vector container
 * \param v The vector container
 */
void CukrContainersVectorOfCommentryDestroy(CUKR_CONTAINER_VECTOR_COMMENTRY *v)
{
	int i;
	
	/* Free elements */
	free((*v)->elements);
	
	/* Delete container structure */
	free(*v);
	
	LOG(DEBUG, "Vector of comm-entry container destroyed");
}

// ------------------------------------------------------------------- //

/**
 * \brief Append an element at the end  
 *        of the container
 * \param v The vector container
 * \param elem The element to add
 */
void CukrContainersVectorOfCommentryAppend(CUKR_CONTAINER_VECTOR_COMMENTRY v, 
                                           CUKR_COMMENTRY elem)
{
	/* If more memory needed, double it */
	if (v->size == v->allocated) {	
		if ( (v->elements = (CUKR_COMMENTRY*)realloc(v->elements, 2*v->allocated*sizeof(struct _CUKR_COMMENTRY))) == NULL) {
			CUKR_FATALX("Error allocating %d int elements for container",
						2 * v->allocated);
		}
		v->allocated *= 2; 
	}
	
	/* Add the new element */
	v->elements[v->size++] = elem;
	
	/* Update the communiaction volume */
	v->comm_volume += elem.size;
} 

// ------------------------------------------------------------------- //

/**
 * \brief Print all elements inside a 
 *        vector container
 * \param v The vector container
 */
void CukrContainersVectorOfCommentryPrint(CUKR_CONTAINER_VECTOR_COMMENTRY v)
{
	CukrContainersVectorOfCommentryPrintToFile(v, stdout);
}

// ------------------------------------------------------------------- //

/**
 * \brief Print all elements inside a 
 *        vector container
 * \param v The vector container
 */
void CukrContainersVectorOfCommentryPrintToFile(CUKR_CONTAINER_VECTOR_COMMENTRY v, FILE *f)
{
	CukrContainersVectorOfCommentryPrintToFileWithIndent(v, f, "");
}

// ------------------------------------------------------------------- //

/**
 * \brief Print all elements inside a 
 *        vector container
 * \param v The vector container
 */
void CukrContainersVectorOfCommentryPrintToFileWithIndent(CUKR_CONTAINER_VECTOR_COMMENTRY v, FILE *f, char *indent)
{
	int i;
	
	fprintf(f, "%sVector container with %d elem (%d allocated)\n",
	           indent, v->size, v->allocated);
	fprintf(f, "%s", indent);
	if (!v->size)
		fprintf(f, "(empty)");
	else	           
		for (i = 0; i < v->size; i++) 
			fprintf(f, "(pos:%d, size:%d) ", v->elements[i].pos, v->elements[i].size);
	fprintf(f, "\n");
		 
}

// -------------------------------------------------------------------------------------//
// ==========================  CONTAINER: VECTOR of MATENTRY  ======================== //
// -------------------------------------------------------------------------------------//

/**
 * \brief Does initialization stuff to make
 *        a vector container ready for use
 * \param v The vector container
 */
void CukrContainersVectorOfMatentryCreate(CUKR_CONTAINER_VECTOR_MATENTRY *v)
{
	/* 
	 * Allocate container structure 
	 */
	
	if ( (*v = (CUKR_CONTAINER_VECTOR_MATENTRY)malloc(sizeof(struct _CUKR_CONTAINER_VECTOR_MATENTRY))) == NULL) {
		CUKR_FATAL("Error allocating memory for container structure" );
	} 
	
	/* 
	 * Set stuff
	 */
	 
	(*v)->size = 0;
	
	/*
	 * Allocate initial amount of elemens
	 */
	
	(*v)->allocated = CUKR_CONTAINERS_VECTORMATENTRY_START_SIZE;
	if ( ((*v)->elements = (CUKR_MATENTRY*)malloc(CUKR_CONTAINERS_VECTORMATENTRY_START_SIZE * sizeof(struct _CUKR_MATENTRY))) == NULL) {
		CUKR_FATAL("Error allocating initial memory for container");
	}
	LOG(DEBUG, "Vector of mat-entry container created");	
} 

// ------------------------------------------------------------------- //

/**
 * \brief Free all the data allocated inside
 *        the vector container
 * \param v The vector container
 */
void CukrContainersVectorOfMatentryDestroy(CUKR_CONTAINER_VECTOR_MATENTRY *v)
{
	int i;
	
	/* Free elements */
	free((*v)->elements);
	
	/* Delete container structure */
	free(*v);
	
	LOG(DEBUG, "Vector of mat-entry container destroyed");
}

// ------------------------------------------------------------------- //

/**
 * \brief Append an element at the end  
 *        of the container
 * \param v The vector container
 * \param elem The element to add
 */
void CukrContainersVectorOfMatentryAppend(CUKR_CONTAINER_VECTOR_MATENTRY v, 
                                          CUKR_MATENTRY elem)
{
	/* If more memory needed, double it */
	if (v->size == v->allocated) {	
		if ( (v->elements = (CUKR_MATENTRY*)realloc(v->elements, 2*v->allocated*sizeof(struct _CUKR_MATENTRY))) == NULL) {
			CUKR_FATALX("Error allocating %d int elements for container",
						2 * v->allocated);
		}
		v->allocated *= 2; 
	}
	
	/* Add the new element */
	v->elements[v->size++] = elem;
} 

// ------------------------------------------------------------------- //

int compare_matentry (const void * a, const void * b)
{
   CUKR_MATENTRY ma = *(CUKR_MATENTRY*)a;
   CUKR_MATENTRY mb = *(CUKR_MATENTRY*)b;
   
   /* Row comes first */
   if (ma.row != mb.row) 
  	 return ma.row - mb.row;
  
   /* If same, do for columns */
   else
   	return ma.col - mb.col;	  
}

/**
 * \brief Sort the element of a vector 
 *        in increasing order and remove
 *        all duplicates
 * \param v The vector container
 */
void CukrContainersVectorOfMatentrySortUnique(CUKR_CONTAINER_VECTOR_MATENTRY v)
{
	int i;
	
	/* Return if no elements */
	if (!v->size)
		return;
	
	/* Sort using quick-sort */
	qsort (v->elements, v->size, sizeof(struct _CUKR_MATENTRY), compare_matentry);
	
	/* Shrink vector to contain only unique elements */
	int unique_elements = 1;
	for (i = 1; i < v->size; i++) {		
		if (v->elements[i].col != v->elements[unique_elements - 1].col || 
			v->elements[i].row != v->elements[unique_elements - 1].row   ) 
			v->elements[unique_elements++].val = v->elements[i].val;
		}
	v->size = unique_elements; 			
}

// ------------------------------------------------------------------- //

/**
 * \brief Print all elements inside a 
 *        vector container
 * \param v The vector container
 */
void CukrContainersVectorOfMatentryPrint(CUKR_CONTAINER_VECTOR_MATENTRY v)
{
	CukrContainersVectorOfMatentryPrintToFile(v, stdout);
}

// ------------------------------------------------------------------- //

/**
 * \brief Print all elements inside a 
 *        vector container
 * \param v The vector container
 */
void CukrContainersVectorOfMatentryPrintToFile(CUKR_CONTAINER_VECTOR_MATENTRY v, FILE *f)
{
	CukrContainersVectorOfMatentryPrintToFileWithIndent(v, f, "");
}

// ------------------------------------------------------------------- //

/**
 * \brief Print all elements inside a 
 *        vector container
 * \param v The vector container
 */
void CukrContainersVectorOfMatentryPrintToFileWithIndent(CUKR_CONTAINER_VECTOR_MATENTRY v, FILE *f, char *indent)
{
	int i;
	
	fprintf(f, "%sVector container with %d elem (%d allocated)\n",
	           indent, v->size, v->allocated);
	fprintf(f, "%s", indent);
	if (!v->size)
		fprintf(f, "(empty)");
	else	           
		for (i = 0; i < v->size; i++) 
			fprintf(f, "(row:%d,col:%d->val:%g ) ", v->elements[i].row, v->elements[i].col, v->elements[i].val);
	fprintf(f, "\n");
		 
}
