/**
 * @file utils/containers.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines containers (e.g. vector 
 * container, that can hold arbitrary
 * data and automatically do memory
 * managemenet and stuff)
*/

#ifndef CUKR_CONTAINERS_H_
#define CUKR_CONTAINERS_H_

/* =============== FORWARD DECLARATIONS ============== */

typedef struct _CUKR_COMMENTRY CUKR_COMMENTRY;
typedef struct _CUKR_MATENTRY  CUKR_MATENTRY;
typedef struct _CUKR_CONTAINER_VECTOR_INT       *CUKR_CONTAINER_VECTOR_INT;
typedef struct _CUKR_CONTAINER_VECTOR_COMMENTRY *CUKR_CONTAINER_VECTOR_COMMENTRY;
typedef struct _CUKR_CONTAINER_VECTOR_MATENTRY  *CUKR_CONTAINER_VECTOR_MATENTRY;

/* =================================================== */

#include "../mat_vec/mp/matrix_mp.h"
#include "../cukr.h"


// -------------------------------------------------------------------------------------//
// =============================  CONTAINER: VECTOR of INT ============================ //
// -------------------------------------------------------------------------------------//

/*
 * CONFIGURATION
 */

/** No of elements to start with 
 * for a INT container */
#define CUKR_CONTAINERS_VECTORINT_START_SIZE 10 

/** 
 * Defines a vector container for INT.
 */
struct _CUKR_CONTAINER_VECTOR_INT {
	
	/** The number of elements present in the
	 * container */
	int size;
	
	/** The number of elements for which memory
	 * is allocated ( >= size) */
	int allocated;
	
	/** The vector where pointers to the elements
	 * are stored */
	CUKR_INT *elements;
};


// ------------------------------------------------------------------- //

/**
 * \brief Does initialization stuff to make
 *        a vector container ready for use
 * \param v The vector container
 */
void CukrContainersVectorOfIntCreate(CUKR_CONTAINER_VECTOR_INT *v); 

// ------------------------------------------------------------------- //

/**
 * \brief Free all the data allocated inside
 *        the vector container
 * \param v The vector container
 */
void CukrContainersVectorOfIntDestroy(CUKR_CONTAINER_VECTOR_INT *v); 

// ------------------------------------------------------------------- //

/**
 * \brief Append an element at the end  
 *        of the container
 * \param v The vector container
 * \param elem The element to add
 */
void CukrContainersVectorOfIntAppend(CUKR_CONTAINER_VECTOR_INT v, int elem); 

// ------------------------------------------------------------------- //

/**
 * \brief Append an element at the end  
 *        of the container. Make sure it
 *        is unique. 
 * \param v The vector container
 * \param elem The element to add
 */
void CukrContainersVectorOfIntAppendU(CUKR_CONTAINER_VECTOR_INT v, int elem);

// ------------------------------------------------------------------- //

/**
 * \brief Sort the element of a vector
 *        in increasing order
 * \param v The vector container
 */
void CukrContainersVectorOfIntSortI(CUKR_CONTAINER_VECTOR_INT v);

// ------------------------------------------------------------------- //

/**
 * \brief Sort the element of a vector 
 *        in increasing order and remove
 *        all duplicates
 * \param v The vector container
 */
void CukrContainersVectorOfIntSortUnique(CUKR_CONTAINER_VECTOR_INT v);

// ------------------------------------------------------------------- //

/**
 * \brief Print all elements inside a 
 *        vector container
 * \param v The vector container
 */
void CukrContainersVectorOfIntPrint(CUKR_CONTAINER_VECTOR_INT v);

// ------------------------------------------------------------------- //

/**
 * \brief Print all elements inside a 
 *        vector container
 * \param f An open file stream to write to
 */
void CukrContainersVectorOfIntPrintToFile(CUKR_CONTAINER_VECTOR_INT v, FILE*f);

// ------------------------------------------------------------------- //

/**
 * \brief Print all elements inside a 
 *        vector container
 * \param v The vector container
 * \param f An open file stream to write to
 * \param indent An indentation string, to be inserted
 *        before every print
 */
void CukrContainersVectorOfIntPrintToFileWithIndent(CUKR_CONTAINER_VECTOR_INT v, FILE *f, char *indent);


// -------------------------------------------------------------------------------------//
// ==========================  CONTAINER: VECTOR of COMMENTRY  ======================== //
// -------------------------------------------------------------------------------------//

/*
 * CONFIGURATION
 */

/** No of elements to start with 
 * for a COMM-ENTRY container */
#define CUKR_CONTAINERS_VECTORCOMMENTRY_START_SIZE 10 


/** 
 * In order to make the transfer more
 * efficient, we would rather combine more
 * elements into one transfer than send all
 * one by one. 
 * A comm-entry represents such a group of 
 * elements to be transfered.
 */
struct _CUKR_COMMENTRY {
	
	/* The no. of elements to transfer */
	int size;
	
	/* The starting position int the 
	 * source data vector */
	int pos;
};

/** 
 * Defines a vector container for comm-entry
 * structures. 
 */
struct _CUKR_CONTAINER_VECTOR_COMMENTRY {
	
	/** The number of elements present in the
	 * container */
	int size;
	
	/** The number of elements for which memory
	 * is allocated ( >= size) */
	int allocated;
	
	/** The vector where pointers to the elements
	 * are stored */
	CUKR_COMMENTRY *elements;
	
	/** The total number of elements to be
	 * communicated (sum of all sizes from 
	 * all entries) */
	int comm_volume; 
};

// ------------------------------------------------------------------- //

/**
 * \brief Does initialization stuff to make
 *        a vector container ready for use
 * \param v The vector container
 */
void CukrContainersVectorOfCommentryCreate(CUKR_CONTAINER_VECTOR_COMMENTRY *v); 

// ------------------------------------------------------------------- //

/**
 * \brief Free all the data allocated inside
 *        the vector container
 * \param v The vector container
 */
void CukrContainersVectorOfCommentryDestroy(CUKR_CONTAINER_VECTOR_COMMENTRY *v); 

// ------------------------------------------------------------------- //

/**
 * \brief Append an element at the end  
 *        of the container
 * \param v The vector container
 * \param elem The element to add
 */
void CukrContainersVectorOfCommentryAppend(CUKR_CONTAINER_VECTOR_COMMENTRY v, 
                                           CUKR_COMMENTRY elem); 

// ------------------------------------------------------------------- //

/**
 * \brief Print all elements inside a 
 *        vector container
 * \param v The vector container
 */
void CukrContainersVectorOfCommentryPrint(CUKR_CONTAINER_VECTOR_COMMENTRY v);

// ------------------------------------------------------------------- //

/**
 * \brief Print all elements inside a 
 *        vector container
 * \param v The vector container
 */
void CukrContainersVectorOfCommentryPrintToFile(CUKR_CONTAINER_VECTOR_COMMENTRY v, FILE *f);

// ------------------------------------------------------------------- //

/**
 * \brief Print all elements inside a 
 *        vector container
 * \param v The vector container
 */
void CukrContainersVectorOfCommentryPrintToFileWithIndent(CUKR_CONTAINER_VECTOR_COMMENTRY v, FILE *f, char *indent);


// -------------------------------------------------------------------------------------//
// ==========================  CONTAINER: VECTOR of MATENTRY  ======================== //
// -------------------------------------------------------------------------------------//

/*
 * CONFIGURATION
 */

/** No of elements to start with 
 * for a MAT-ENTRY container */
#define CUKR_CONTAINERS_VECTORMATENTRY_START_SIZE 10 


/** 
 * In order to make the transfer more
 * efficient, we would rather combine more
 * elements into one transfer than send all
 * one by one. 
 * A comm-entry represents such a group of 
 * elements to be transfered.
 */
struct _CUKR_MATENTRY {
	
	/* Matrix value */
	CUKR_DOUBLE val;
	
	/* Position of the element */
	int row, col;
};

/** 
 * Defines a vector container for comm-entry
 * structures. 
 */
struct _CUKR_CONTAINER_VECTOR_MATENTRY {
	
	/** The number of elements present in the
	 * container */
	int size;
	
	/** The number of elements for which memory
	 * is allocated ( >= size) */
	int allocated;
	
	/** The vector where pointers to the elements
	 * are stored */
	CUKR_MATENTRY *elements;
};

// ------------------------------------------------------------------- //

/**
 * \brief Does initialization stuff to make
 *        a vector container ready for use
 * \param v The vector container
 */
void CukrContainersVectorOfMatentryCreate(CUKR_CONTAINER_VECTOR_MATENTRY *v); 

// ------------------------------------------------------------------- //

/**
 * \brief Free all the data allocated inside
 *        the vector container
 * \param v The vector container
 */
void CukrContainersVectorOfMatentryDestroy(CUKR_CONTAINER_VECTOR_MATENTRY *v); 

// ------------------------------------------------------------------- //

/**
 * \brief Append an element at the end  
 *        of the container
 * \param v The vector container
 * \param elem The element to add
 */
void CukrContainersVectorOfMatentryAppend(CUKR_CONTAINER_VECTOR_MATENTRY v, 
                                          CUKR_MATENTRY elem); 

// ------------------------------------------------------------------- //

/**
 * \brief Print all elements inside a 
 *        vector container
 * \param v The vector container
 */
void CukrContainersVectorOfMatentryPrint(CUKR_CONTAINER_VECTOR_MATENTRY v);

// ------------------------------------------------------------------- //

/**
 * \brief Print all elements inside a 
 *        vector container
 * \param v The vector container
 */
void CukrContainersVectorOfMatentryPrintToFile(CUKR_CONTAINER_VECTOR_MATENTRY v, FILE *f);

// ------------------------------------------------------------------- //

/**
 * \brief Print all elements inside a 
 *        vector container
 * \param v The vector container
 */
void CukrContainersVectorOfMatentryPrintToFileWithIndent(CUKR_CONTAINER_VECTOR_MATENTRY v, FILE *f, char *indent);


#endif /*CUKR_CONTAINERS_H_*/
