/**
 * @file utils/utils.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines various functions, useful
 * throughout the implementation
*/

#define LOGGER_NAME "CUKr.util.utils"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "utils.h"
#include "utils_cuda.h"
#include "../mat_vec/matrix_defs.h"
#include "../mat_vec/sp/matrix_sp.h"
#include "../mat_vec/mp/vector_mp.h"
#include "../solvers/solver.h"
#include "log.h"

/**
 * \brief Does a case insensitive comparison between
 *        str1 and str2. Returns -1 if str1 < str2, 
 *        0 if str1 = str2 and 1 if str1 > str2
 * \param str1, str2 The strings being compared
 */
int CukrStricmp(const char* str1, const char* str2)
{
	int i;
	
	if (strlen(str1) < strlen(str2))
		return -1;
	else if (strlen(str1) > strlen(str2))
		return  1;
	else {
		for (i = 0; i < strlen(str1); i++)
			if (toupper(str1[i]) < toupper(str2[i]))
				return -1;
			else if (toupper(str1[i]) > toupper(str2[i]))
				return  1; 
	}
	
	return 0;
 }

// ------------------------------------------------------------------- //

/**
 * \brief Extracts the location from a status
 *        string, as a human readable string
 * \param loc The location id
 * \param loc_str Pointer to a string where
 *        the location is to be stored
 */
void CukrLocIdToStr(int loc, char* loc_str)
{
    char s[255];
	
	sprintf(s, "");
	
	if (loc == CUKR_LOCATION_CPU)
		sprintf(s, "%s CPU", s);
	if (loc == CUKR_LOCATION_GPU)
		sprintf(s, "%s GPU", s);
		
	/* Error if empty */
	if (!CukrStricmp(s, "")){
		CUKR_FATALX("Invalid status id: %d", loc);
	}	
	
	/* Else, some formatting */	
	sprintf(loc_str, "(%s)", s);
}

// ------------------------------------------------------------------- //

/**
 * \brief Extracts the type from a status
 *        string, as a human readable string
 * \param type The type id
 * \param type_str Pointer to a string where
 *        the type is to be stored
 */
void CukrTypeIdToStr(int type, char* type_str)
{
	char s[255];
	
	sprintf(s, "");
	
	if (type == CUKR_TYPE_DOUBLE)
		sprintf(s, "DOUBLE");
	else if (type == CUKR_TYPE_QDOUBLE)
		sprintf(s, "QDOUBLE");
	else if (type == CUKR_TYPE_SINGLE)
		sprintf(s, "SINGLE");
	else if (type == CUKR_TYPE_INT)
		sprintf(s, "INT");
	else if (type == CUKR_TYPE_NONE)
		sprintf(s, "NONE");
		
	/* Error if empty */
	if (!CukrStricmp(s, "")){
		CUKR_FATALX("Invalid status id: %d", type);
	}	
	
	/* Else, some formatting */	
	sprintf(type_str, "(%s)", s);
}

// ------------------------------------------------------------------- //

/**
 * \brief Converts the precision value to the
 *        blas precision used for the counters
 * \param type The type id
 */
int CukrTypeToBlasPrec(int type)
{	
	if (type == CUKR_TYPE_DOUBLE)
		return 0;
	else if (type == CUKR_TYPE_QDOUBLE)
		return 1;
	else if (type == CUKR_TYPE_SINGLE)
		return 2;
	else {
		CUKR_FATALX("The provided type cannot be" \
		            "converted to blas precision", type);
	}	

	return -1;
}

// ------------------------------------------------------------------- //


/**
 * \brief Extracts the sharing type from 
 *        the id, as a human readable string
 * \param shtype The type id
 * \param shtype_str Pointer to a string where
 *        the type is to be stored
 */
void CukrShtypeIdToStr(int shtype, char* shtype_str)
{
	if (shtype == CUKR_SHTYPE_DISTRIBUTED)
		sprintf(shtype_str, "DISTRIBUTED");
	else if (shtype == CUKR_SHTYPE_SHARED)
		sprintf(shtype_str, "SHARED");
	else {
		CUKR_FATALX("Unknown sharing type %d", shtype);
	}
}

// ------------------------------------------------------------------- //

/**
 * \brief Given a string like "double", 
 *        return the type id (e.g. 
 *        CUKR_DOUBLE) 
 * \param type_str The type string
 */
int CukrTypeStrToId(const char* type_str)
{
	if (!CukrStricmp(type_str, "double") ||
		!CukrStricmp(type_str, "DOUBLE")   )      return CUKR_TYPE_DOUBLE;
    else if (!CukrStricmp(type_str, "qdouble") ||
			 !CukrStricmp(type_str, "QDOUBLE")  ) return CUKR_TYPE_QDOUBLE;		
	else if (!CukrStricmp(type_str, "single") ||
		     !CukrStricmp(type_str, "SINGLE")   ) return CUKR_TYPE_SINGLE;
	else if (!CukrStricmp(type_str, "int") ||
		     !CukrStricmp(type_str, "INT")   )    return CUKR_TYPE_INT;
	else {
		CUKR_FATALX("Invalid type: %s", type_str);
	}

	return -1;
}

// ------------------------------------------------------------------- //

/**
 * \brief Converts a status id to a
 *        human readable string
 * \param status The status id
 * \param status_str Pointer to a string where
 *        the status is to be stored
 */
void CukrStatusIdToStr(int status, char* status_str)
{	
	char s[255];
	
	sprintf(s, "");
	
	/* Get status */
	if (status & CUKR_STATE_CPU_DOUBLE)
		sprintf(s, "%s HOST_DOUBLE", s);
	if (status & CUKR_STATE_CPU_QDOUBLE)
		sprintf(s, "%s HOST_QDOUBLE", s);
	if (status & CUKR_STATE_CPU_SINGLE)
		sprintf(s, "%s HOST_SINGLE", s);
	if (status & CUKR_STATE_CPU_INT)
		sprintf(s, "%s HOST_INT", s);
	if (status & CUKR_STATE_GPU_DOUBLE)
		sprintf(s, "%s GPU_DOUBLE", s);
	if (status & CUKR_STATE_GPU_QDOUBLE)
		sprintf(s, "%s GPU_QDOUBLE", s);
	if (status & CUKR_STATE_GPU_SINGLE)
		sprintf(s, "%s GPU_SINGLE", s);
	if (status & CUKR_STATE_GPU_INT)
		sprintf(s, "%s GPU_INT", s);
	if (status == CUKR_STATE_NONE)
		sprintf(s, "%s NONE", s);
	
	/* Error if empty */
	if (!CukrStricmp(s, "")){
		CUKR_FATALX("Invalid status id: %d", status);
	}	
	
	/* Else, some formatting */	
	sprintf(status_str, "(%s )", s);
}

// ------------------------------------------------------------------- //

/**
 * \brief Checks for a valid status and 
 *        throws and error if the status
 *        is invalid
 * \param status The status id
 */
int CukrIsValidStatusId(int status)
{	
	if (status != CUKR_STATE_CPU_DOUBLE  && 
		status != CUKR_STATE_CPU_QDOUBLE &&
	    status != CUKR_STATE_CPU_SINGLE  &&
	    status != CUKR_STATE_CPU_INT     &&
	    status != CUKR_STATE_GPU_DOUBLE  && 
	    status != CUKR_STATE_GPU_QDOUBLE && 
	    status != CUKR_STATE_GPU_SINGLE  &&
	    status != CUKR_STATE_GPU_INT      )
	{
		return CUKR_FALSE;
	}
	else 
		return CUKR_TRUE;
}

/**
 * \brief Checks for a valid type and 
 *        throws and error if the type
 *        is invalid
 * \param status The type id
 */
int CukrIsValidTypeId(int type)
{	
	if (type != CUKR_TYPE_DOUBLE  &&
		type != CUKR_TYPE_QDOUBLE &&  
	    type != CUKR_TYPE_SINGLE  &&
	    type != CUKR_TYPE_INT      )
	{
		return CUKR_FALSE;
	}
	else 
		return CUKR_TRUE;
}

// ------------------------------------------------------------------- //

/**
 * \brief Checks for a valid location and 
 *        returns zero if the location is
 *        not valid
 * \param loc The location id
 */
int CukrIsValidLocationId(int loc)
{	
	if (loc != CUKR_LOCATION_CPU && 
	    loc != CUKR_LOCATION_GPU)
	{
		return CUKR_FALSE;
	}
	else 
		return CUKR_TRUE;
}

// ------------------------------------------------------------------- //

/**
 * \brief Checks for a valid solver and 
 *        returns zero if the solver is
 *        not valid
 * \param sol The solver id
 */
int CukrIsValidSolverId(int sol)
{
	if (sol == CUKR_SOLVER_TYPE_CG)
		return CUKR_TRUE;
	else
		return CUKR_FALSE;
}

// ------------------------------------------------------------------- //

/**
 * \brief Builds a status id from a 
 *        type id and a location id
 * \param loc The location id
 * \param type The type id
 */
int CukrLocIdAndTypeIdToStatusId(int loc, int type)
{
	if (loc == CUKR_LOCATION_CPU)
	{
		if (type == CUKR_TYPE_DOUBLE)
			return CUKR_STATE_CPU_DOUBLE;
		else if (type == CUKR_TYPE_QDOUBLE)
			return CUKR_STATE_CPU_QDOUBLE;
		else if (type == CUKR_TYPE_SINGLE)
			return CUKR_STATE_CPU_SINGLE;
		else if (type == CUKR_TYPE_INT)
			return CUKR_STATE_CPU_INT;
		else if (type == CUKR_TYPE_NONE)
			return CUKR_STATE_NONE;	
	}
	else if (loc == CUKR_LOCATION_GPU)
	{
		if (type == CUKR_TYPE_DOUBLE)
			return CUKR_STATE_GPU_DOUBLE;
	    else if (type == CUKR_TYPE_QDOUBLE)
			return CUKR_STATE_GPU_QDOUBLE;			
		else if (type == CUKR_TYPE_SINGLE)
			return CUKR_STATE_GPU_SINGLE;
		else if (type == CUKR_TYPE_INT)
			return CUKR_STATE_GPU_INT;
		else if (type == CUKR_TYPE_NONE)
			return CUKR_STATE_NONE;		
	}
	
	/* If we go to here means nothing worked
	 * so something is wrong */
	
	CUKR_FATALX("Invalid location - type combination: " \
	            "loc=%d, type=%d", loc, type);

	return -1;
}

// ------------------------------------------------------------------- //

/**
 * \brief Look into the status id and 
 *        get the highest floating point
 *        precision available. For int,
 * 		   it returns int. 
 * \param status The status id
 */
int GetHighestPrecFromStatusId(int status)
{
	if (status & CUKR_TYPE_DOUBLE)
		return CUKR_TYPE_DOUBLE;
	if (status & CUKR_TYPE_QDOUBLE)
		return CUKR_TYPE_QDOUBLE;
	if (status & CUKR_TYPE_SINGLE)
		return CUKR_TYPE_SINGLE;
	if (status & CUKR_TYPE_INT)
		return CUKR_TYPE_INT;	
	
	/* If none, then error */
	CUKR_FATALX("The specified status (%d) does not " \
	            "have data in any data types",
	            status);

	return -1;
}

// ------------------------------------------------------------------- //

/**
 * \brief Look into the status id and 
 *        get the highest floating point
 *        precision available on the CPU
 * 		  For int, it returns int. 
 * \param status The status id
 */
int GetHighestPrecOnCpuFromStatusId(int status)
{
	if (status & CUKR_STATE_CPU_DOUBLE)
		return CUKR_TYPE_DOUBLE;
	if (status & CUKR_STATE_CPU_QDOUBLE)
		return CUKR_TYPE_QDOUBLE;
	if (status & CUKR_STATE_CPU_SINGLE)
		return CUKR_TYPE_SINGLE;
	if (status & CUKR_STATE_CPU_INT)
		return CUKR_TYPE_INT;	
	
	return CUKR_TYPE_NONE;		
}

// ------------------------------------------------------------------- //

/**
 * \brief Look into the status id and 
 *        get the highest floating point
 *        precision available on the GPU
 * 		  For int, it returns int. 
 * \param status The status id
 */
int GetHighestPrecOnGpuFromStatusId(int status)
{
	if (status & CUKR_STATE_GPU_DOUBLE)
		return CUKR_TYPE_DOUBLE;
	if (status & CUKR_STATE_GPU_QDOUBLE)
		return CUKR_TYPE_QDOUBLE;
	if (status & CUKR_STATE_GPU_SINGLE)
		return CUKR_TYPE_SINGLE;
	if (status & CUKR_STATE_GPU_INT)
		return CUKR_TYPE_INT;	
	
	return CUKR_TYPE_NONE;		
}

// ------------------------------------------------------------------- //

/**
 * \brief Transforms the matrix format id to a 
 *        human readable string
 * \param type The format id
 * \param loc_str Pointer to a string where
 *        the format is to be stored
 */
void CukrMatFormatIdToStr(int format, char* fmt_str)
{
	if (format == CUKR_MATRIX_FORMAT_CSR)
		sprintf(fmt_str, "CSR");
	else if (format == CUKR_MATRIX_FORMAT_BCSR)
		sprintf(fmt_str, "BCSR");
	else if (format == CUKR_MATRIX_FORMAT_HYB)
		sprintf(fmt_str, "HYB");
	else if (format == CUKR_MATRIX_FORMAT_BHYB)
		sprintf(fmt_str, "BHYB");
	else {
		CUKR_FATALX("Invalid matrix format id: %d", format);
	}
}

// ------------------------------------------------------------------- //

/**
 * \brief Transforms the matrix format string  
 *        into a matrix format id
 * \param fmt_str The format string
 */
int CukrMatFormatStrToId(const char* fmt_str)
{
	if (!CukrStricmp(fmt_str, "CSR") || !CukrStricmp(fmt_str, "csr"))
		return CUKR_MATRIX_FORMAT_CSR;
	else if (!CukrStricmp(fmt_str, "BCSR") || !CukrStricmp(fmt_str, "bcsr"))
		return CUKR_MATRIX_FORMAT_BCSR;
	else if (!CukrStricmp(fmt_str, "HYB") || !CukrStricmp(fmt_str, "hyb"))
		return CUKR_MATRIX_FORMAT_HYB;
	else if (!CukrStricmp(fmt_str, "BHYB") || !CukrStricmp(fmt_str, "bhyb"))
		return CUKR_MATRIX_FORMAT_BHYB;
	/* If nothing found */
	else {
		CUKR_FATALX("Invalid matrix format string: %s", fmt_str);
	}	

	return -1;
}


// ------------------------------------------------------------------- //

/**
 * \brief Transforms the matrix format 
 *        variation id to a human readable string
 * \param type The format variation id
 * \param loc_str Pointer to a string where
 *        the format variation is to be stored
 */
void CukrMatFormatVariationIdToStr(int variation, char* var_str)
{
	if (variation == CUKR_MATRIX_FORMAT_VARIATION_NONE)
		sprintf(var_str, "NONE");
	else if (variation == CUKR_MATRIX_FORMAT_VARIATION_CSR4)
		sprintf(var_str, "CSR4");
	else if (variation == CUKR_MATRIX_FORMAT_VARIATION_BCSR4)
		sprintf(var_str, "BCSR4");
	else if (variation == CUKR_MATRIX_FORMAT_VARIATION_BHYB4)
		sprintf(var_str, "BHYB4");
	else {
		CUKR_FATALX("Invalid matrix format variation id: %d", variation);
	}
}

// ------------------------------------------------------------------- //

/**
 * \brief Transforms the matrix format 
 *        variation string into the id 
 * \param type The format variation id
 * \param var_str The format variation string
 */
int CukrMatFormatVariationStrToId(const char* var_str)
{
	/* For none */
	if (!CukrStricmp(var_str, "NONE"))
		return CUKR_MATRIX_FORMAT_VARIATION_NONE;
	
	/* For CSR4 */
	else if (!CukrStricmp(var_str, "CSR4"))
		return CUKR_MATRIX_FORMAT_VARIATION_CSR4;

	/* For BCSR4 */
	else if (!CukrStricmp(var_str, "BCSR4"))
		return CUKR_MATRIX_FORMAT_VARIATION_BCSR4;

	/* For BHYB4 */
	else if (!CukrStricmp(var_str, "BHYB4"))
		return CUKR_MATRIX_FORMAT_VARIATION_BHYB4;

	/* If nothing found */
	else {
		CUKR_FATALX("Invalid matrix format variation string: %s", var_str);
	}

	return -1;
}

// ------------------------------------------------------------------- //

/**
 * \brief Set the current gpu to gpuid
 * \param gpuid The #gpu to use
 */
void CukrSetDevice(int gpuid)
#ifndef CUKR_USE_GPU
{
	CUKR_FATAL("GPU support was not compiled in the library");
}
#else
	#ifdef CUKR_USE_CUDA
{
	/* Call the CUDA function */
	CukrSetDevice_cu(gpuid);
	LOGX(DEBUG, "Current GPU set to %d", gpuid);
}
	#else // CUKR_USE_OPENCL
{
	printf("gpuid: %d\n", gpuid);
		// fill in, if necessary
}
	#endif /* CUKR_USE_CUDA */
#endif /* CUKR_USE_GPU */

// ------------------------------------------------------------------- //

/**
 * \brief Returns the word size used
 *        by a BLAS
 * \param blas The blas
 */
float CukrBlasWordSize(CUKR_BLAS blas)
{
	switch (blas->precision) {
		case CUKR_TYPE_DOUBLE:
			return 8;
		case CUKR_TYPE_QDOUBLE:
			return 8;
		case CUKR_TYPE_SINGLE:
			return 4;
		default:
			return -1;	
	}
}

// ------------------------------------------------------------------- //

/* 
 * \brief Split-path routine 
 * \param path The path to be split
 * \param drv The drive letter part
 * \param dir The dir part
 * \param name The filename part
 * \param ext The extension part
 */
void CukrSplitPath(const char* path, char* drv, char* dir, char* name, char* ext)
{
    const char* end; /* end of processed string */
    const char* p;   /* search pointer */
    const char* s;   /* copy pointer */

    /* extract drive name */
    if (path[0] && path[1]==':') {
    if (drv) {
                 *drv++ = *path++;
                 *drv++ = *path++;
                 *drv = '\0';
         }
    } else if (drv)
      *drv = '\0';
 
         /* search for end of string or stream separator */
         for(end=path; *end && *end!=':'; )
                 end++;
 
     /* search for begin of file extension */
     for(p=end; p>path && *--p!='\\' && *p!='/'; )
             if (*p == '.') {
                     end = p;
                     break;
             }
 
     if (ext)
             for(s=end; (*ext=*s++); )
                     ext++;
 
     /* search for end of directory name */
     for(p=end; p>path; )
             if (*--p=='\\' || *p=='/') {
                         p++;
                         break;
                 }
 
         if (name) {
                 for(s=p; s<end; )
                         *name++ = *s++;
 
                 *name = '\0';
     }
 
     if (dir) {
         for(s=path; s<p; )
             *dir++ = *s++;
         *dir = '\0';
     }
}
