/**
 * @file utils/utils.h
 * @author Serban Georgescu
 * @since 2008
 * 
 * Defines various functions, useful
 * throughout the implementation
*/

#ifndef CUKR_UTILS_H_
#define CUKR_UTILS_H_

#include "../solvers/solver.h"

/**
 * \brief Does a case insensitive comparison between
 *        str1 and str2. Returns -1 if str1 < str2, 
 *        0 if str1 = str2 and 1 if str1 > str2
 * \param str1, str2 The strings being compared
 */
int CukrStricmp(const char* str1, const char* str2);

/**
 * \brief Extracts the location from a status
 *        string, as a human readable string
 * \param loc The location id
 * \param loc_str Pointer to a string where
 *        the location is to be stored
 */
void CukrLocIdToStr(int loc, char* loc_str);

// ------------------------------------------------------------------- //

/**
 * \brief Extracts the type from a status
 *        id, as a human readable string
 * \param type The type id
 * \param type_str Pointer to a string where
 *        the type is to be stored
 */
void CukrTypeIdToStr(int type, char* type_str);

// ------------------------------------------------------------------- //

/**
 * \brief Converts the precision value to the
 *        blas precision used for the counters
 * \param type The type id
 */
int CukrTypeToBlasPrec(int type);

// ------------------------------------------------------------------- //

/**
 * \brief Extracts the sharing type from 
 *        the id, as a human readable string
 * \param shtype The type id
 * \param shtype_str Pointer to a string where
 *        the type is to be stored
 */
void CukrShtypeIdToStr(int shtype, char* shtype_str);

// ------------------------------------------------------------------- //

/**
 * \brief Given a string like "double", 
 *        return the type id (e.g. 
 *        CUKR_DOUBLE) 
 * \param type_str The type string
 */
int CukrTypeStrToId(const char* type_str); 

// ------------------------------------------------------------------- //

/**
 * \brief Converts a status id to a
 *        human readable string
 * \param status The status id
 * \param status_str Pointer to a string where
 *        the status is to be stored
 */
void CukrStatusIdToStr(int status, char* status_str);

// ------------------------------------------------------------------- //

/**
 * \brief Checks for a valid status and 
 *        returns zero if the status is
 *        not valid
 * \param status The status id
 */
int CukrIsValidStatusId(int status);

// ------------------------------------------------------------------- //

/**
 * \brief Checks for a valid type and 
 *        returns zero if the type is
 *        not valid
 * \param type The type id
 */
int CukrIsValidTypeId(int type);

// ------------------------------------------------------------------- //

/**
 * \brief Checks for a valid location and 
 *        returns zero if the location is
 *        not valid
 * \param loc The location id
 */
int CukrIsValidLocationId(int loc);

// ------------------------------------------------------------------- //

/**
 * \brief Checks for a valid solver and 
 *        returns zero if the solver is
 *        not valid
 * \param sol The solver id
 */
int CukrIsValidSolverId(int sol);

// ------------------------------------------------------------------- //
 
/**
 * \brief Builds a status id from a 
 *        type id and a location id
 * \param loc The location id
 * \param type The type id
 */
int CukrLocIdAndTypeIdToStatusId(int loc, int type);

// ------------------------------------------------------------------- //

/**
 * \brief Look into the status id and 
 *        get the highest floating point
 *        precision available. For int,
 *        it returns int. 
 * \param status The status id
 */
int GetHighestPrecFromStatusId(int status);

// ------------------------------------------------------------------- //

/**
 * \brief Look into the status id and 
 *        get the highest floating point
 *        precision available on the CPU.
 * 		  For int, it returns int. 
 * \param status The status id
 */
int GetHighestPrecOnCpuFromStatusId(int status);

// ------------------------------------------------------------------- //

/**
 * \brief Look into the status id and 
 *        get the highest floating point
 *        precision available on the GPU
 * 		  For int, it returns int. 
 * \param status The status id
 */
int GetHighestPrecOnGpuFromStatusId(int status);

// ------------------------------------------------------------------- //

/**
 * \brief Transforms the matrix format id to a 
 *        human readable string
 * \param type The format id
 * \param fmt_str Pointer to a string where
 *        the format is to be stored
 */
void CukrMatFormatIdToStr(int format, char* fmt_str);

// ------------------------------------------------------------------- //

/**
 * \brief Transforms the matrix format string  
 *        into a matrix format id
 * \param fmt_str The format string
 */
int CukrMatFormatStrToId(const char* fmt_str);

// ------------------------------------------------------------------- //

/**
 * \brief Transforms the matrix format 
 *        variation id to a human readable string
 * \param type The format variation id
 * \param loc_str Pointer to a string where
 *        the format variation is to be stored
 */
void CukrMatFormatVariationIdToStr(int variation, char* var_str);

// ------------------------------------------------------------------- //

/**
 * \brief Transforms the matrix format 
 *        variation string into the id 
 * \param type The format variation id
 * \param var_str The format variation string
 */
int CukrMatFormatVariationStrToId(const char* var_str);

// ------------------------------------------------------------------- //

/**
 * \brief Set the current gpu to gpuid
 * \param gpuid The #gpu to use
 */
void CukrSetDevice(int gpuid);

// ------------------------------------------------------------------- //

/**
 * \brief Returns the word size used
 *        by a BLAS
 * \param blas The blas
 */
float CukrBlasWordSize(CUKR_BLAS blas);

// ------------------------------------------------------------------- //

/* 
 * \brief Split-path routine 
 * \param path The path to be split
 * \param drv The drive letter part
 * \param dir The dir part
 * \param name The filename part
 * \param ext The extension part
 */
void CukrSplitPath(const char* path, char* drv, char* dir, char* name, char* ext);

#endif /*CUKR_UTILS_H_*/
